/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
* Description: Crypto Engine Management
*
* Filename: trcryeng.c
* Author: Jin Zhang
* Date Created: 1/9/2003
 * $Source: source/cryptlib/trcrpeng.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:25CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


#include <trcrylib.h>




static const char  TM_CONST_QLF tlEncAlgDes[] =        "DES";
static const char  TM_CONST_QLF tlEncAlgIdea[] =       "IDEA";
static const char  TM_CONST_QLF tlEncAlgBlowfish[] =   "BLOWFISH";
static const char  TM_CONST_QLF tlEncAlgRC5R16B64[] =  "RC5-R16-B64";
static const char  TM_CONST_QLF tlEncAlg3Des[] =       "3DES";
static const char  TM_CONST_QLF tlEncAlgCast[] =       "CAST";
static const char  TM_CONST_QLF tlEncAlgAes[] =        "AES";
static const char  TM_CONST_QLF tlEncAlgNull[] =       "NULL";
static const char  TM_CONST_QLF tlEncAlgTwoFish[] =    "TWOFISH";
static const char  TM_CONST_QLF tlEncAlgNoSupport[] =  "NOT SUPPORTED";

static const char  TM_CONST_QLF tlHashAlgMd5[] =       "MD5";
static const char  TM_CONST_QLF tlHashAlgSha1[] =      "SHA1";
static const char  TM_CONST_QLF tlHashAlgTiger[] =     "TIGER";
static const char  TM_CONST_QLF tlHashAlgSha256[] =    "SHA256";
static const char  TM_CONST_QLF tlHashAlgSha384[] =    "SHA384";
static const char  TM_CONST_QLF tlHashAlgSha512[] =    "SHA512";
static const char  TM_CONST_QLF tlHashAlgRIPEMD[] =    "RIPEMD";
static const char  TM_CONST_QLF tlHashAlgNoSupport[] = "NOT SUPPORTED";

static ttCryptoEnginePtr tfCryptoEngineLookup(unsigned int engineId);
TM_NEARCALL static int tfCryptoDefaultGetRandomWord(ttUser32Bit   TM_FAR * dataPtr,
                                        ttUser16Bit            wordSize);
static int tfCryptoEngineRegisterLocked(
                           unsigned int                 engineId,
                           ttUserVoidPtr                initParamPtr,
                           ttCryptoEnginePtrPtr         newEnginePtrPtr,
                           ttCryptoEngineInitFuncPtr    engineInitFuncPtr,
                           ttCryptoGetRandomWordFuncPtr randomFuncPtr,
                           ttCryptoSessionOpenFuncPtr   sessionOpenFuncPtr,
                           ttCryptoSessionFuncPtr       sessionProcessFuncPtr,
                           ttCryptoSessionFuncPtr       sessionCloseFuncPtr);


/*
 * tfCryptoEngineLookup() Function Description
 * Called by tfCryptoEngineRegister or tfCryptoEngineAddAlgorithm
 * to find the crypto engine pointer given the engine id.
 *
 * Called with the Crypto Engine lock on.
 *
 * Parameter         description
 * engineId          Crypto Engine number, currently, we support 
 *                   TM_CRYPTO_ENGINE_SOFTWARE  and  TM_CRYPTO_ENGINE_HIFN7951
 *
 * Return value:     NULL pointer or pointer to the crypto engine
 */

static ttCryptoEnginePtr tfCryptoEngineLookup(unsigned int engineId)
{
    ttCryptoEngineListPtr    celPtr;
    ttCryptoEnginePtr        cenginePtr;
    
    cenginePtr = (ttCryptoEnginePtr)0;
    
    celPtr = ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
        ->clgCryptoEngineHeadPtr;
    while(celPtr)
    {
        if(celPtr->celCryptoEnginePtr->ceEngineId == engineId)
        {
            cenginePtr = celPtr->celCryptoEnginePtr;
            break;
        }
        celPtr = celPtr->celNextPtr;
    }
    if(!celPtr)
    {
        cenginePtr = (ttCryptoEnginePtr)0;
    }
    return cenginePtr;
}

static int tfCryptoDefaultGetRandomWord(ttUser32Bit   TM_FAR * dataPtr,
                                        ttUser16Bit            wordSize)
{
    tfGetRandomBytes((tt8BitPtr)dataPtr, wordSize * sizeof(tt32Bit));
    return TM_ENOERROR;
}


/*
 * tfCryptoEngineRegister() Function Description
 * Lock the crypto engine
 * Call common function tfCryptoEngineRegisterLocked()
 * unlock the crypto engine
 */

int tfCryptoEngineRegister(unsigned int                 engineId,
                           ttUserVoidPtr                initParamPtr,
                           ttCryptoEnginePtrPtr         newEnginePtrPtr,
                           ttCryptoEngineInitFuncPtr    engineInitFuncPtr,
                           ttCryptoGetRandomWordFuncPtr randomFuncPtr,
                           ttCryptoSessionOpenFuncPtr   sessionOpenFuncPtr,
                           ttCryptoSessionFuncPtr       sessionProcessFuncPtr,
                           ttCryptoSessionFuncPtr       sessionCloseFuncPtr)
{
    int errorCode;

    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    errorCode = tfCryptoEngineRegisterLocked(engineId,
                                  initParamPtr,
                                  newEnginePtrPtr,
                                  engineInitFuncPtr,
                                  randomFuncPtr,
                                  sessionOpenFuncPtr,
                                  sessionProcessFuncPtr,
                                  sessionCloseFuncPtr);
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return errorCode;
}

/*
 * tfCryptoEngineRegisterLocked() Function Description
 * Called by user space to register a crypto engine, or internally to register
 * the software crypto engine. User doesn't need to
 * call this function to register software crypto engine, unless user want
 * to replace Treck's software implementation. A crypto engine is able to
 * process multiple sessions, for each session, we define the following three
 * function pointers: openFuncPtr, processFuncPtr, closeFuncPtr. Session 
   example,
 * all packets to be processed using the same SA should use the same session.
 * When an SA is deleted, the session should be closed.
 *
 * Parameter         description
 * engineId          Crypto Engine number, currently, we support 
 *                   TM_CRYPTO_ENGINE_SOFTWARE  and  TM_CRYPTO_ENGINE_HIFN7951
 * initParamPtr      The initialization parameter
 * newEnginePtr      returns with pointer which points to the new engine
 * initFuncPtr       Crypto Engine initialization function
 * randomFuncPtr     If the hardware offers random number generation functionality,
 *                   input the function pointer. If it is NULL, we will use the
 *                   software random generator.
 * openFuncPtr       Crypto engine session open function pointer, which takes a
                     parameter of type ttSadbPtr, and constructs the session.
 * processFuncPtr    Crypto engine session process function pointer, which takes
                     a parameter of type ttCryptoRequestPtr
 * closeFuncPtr      Crypto engine session close function pointer, which takes
                     a parameter of type session. (for Hifn7951, it is 
                     (Hifn79xx_PktEngSessContext*)
 * needLock          need to lock crypto engine structure.

 *
 * Return value:     TM_ENOERROR, otherwise failed
 */
static int tfCryptoEngineRegisterLocked(
                           unsigned int                 engineId,
                           ttUserVoidPtr                initParamPtr,
                           ttCryptoEnginePtrPtr         newEnginePtrPtr,
                           ttCryptoEngineInitFuncPtr    engineInitFuncPtr,
                           ttCryptoGetRandomWordFuncPtr randomFuncPtr,
                           ttCryptoSessionOpenFuncPtr   sessionOpenFuncPtr,
                           ttCryptoSessionFuncPtr       sessionProcessFuncPtr,
                           ttCryptoSessionFuncPtr       sessionCloseFuncPtr)
{
    ttCryptoEngineListPtr    newListPtr;
    ttCryptoEnginePtr        newEnginePtr;
    int                      errorCode;


    errorCode = TM_ENOERROR;
    newEnginePtr = (ttCryptoEnginePtr)0;
    if(!tm_context(tvCryptoLibPtr))
    {
        errorCode = TM_EINVAL;
        goto REG_RETURN;
    }

/* firstly check if this engineId exists or not */

    if(engineId > TM_CRYPTO_ENGINE_MAX)
    {
        errorCode = -1;
        goto REG_RETURN;
    }

    if(tfCryptoEngineLookup(engineId) != (ttCryptoEnginePtr)0)
    {
/* this crypto engine has been registered before */
        errorCode = -1;
        goto REG_RETURN;
    }

    {
        newEnginePtr = tm_get_raw_buffer(sizeof(ttCryptoEngine));
        if(newEnginePtr == (ttCryptoEnginePtr)0)
        {
            errorCode = TM_ENOBUFS;
        }
        else
        {
            newEnginePtr->ceEngineId = engineId;
            newEnginePtr->ceSupportFlags = (tt32Bit)1;
            newEnginePtr->ceInitFuncPtr = engineInitFuncPtr;
            if(randomFuncPtr)
            {
                newEnginePtr->ceRandomWordFuncPtr = randomFuncPtr;
            }
            else
            {
                newEnginePtr->ceRandomWordFuncPtr = 
                             tfCryptoDefaultGetRandomWord;
            }
            newEnginePtr->ceSessionOpenFuncPtr = sessionOpenFuncPtr;
            newEnginePtr->ceSessionProcessFuncPtr = sessionProcessFuncPtr;
            newEnginePtr->ceSessionCloseFuncPtr = sessionCloseFuncPtr;


            newListPtr = tm_get_raw_buffer(sizeof(ttCryptoEngineList));
            if(newListPtr == (ttCryptoEngineListPtr)0)
            {
                errorCode = TM_ENOBUFS;
                tm_free_raw_buffer((ttRawBufferPtr)newEnginePtr);
                newEnginePtr = (ttCryptoEnginePtr)0;
            }
            else
            {
/* Initialize the engine */
                if(engineInitFuncPtr)
                {
                    errorCode = engineInitFuncPtr(initParamPtr);
                }

                if(errorCode == TM_ENOERROR)
                {
/* add the new crypto engine to the head of crypto engine list */
                    newListPtr->celCryptoEnginePtr = newEnginePtr;
                    newListPtr->celNextPtr = ((ttCryptoLibGlobalPtr)
                              tm_context(tvCryptoLibPtr))
                              ->clgCryptoEngineHeadPtr;
                    ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
                              ->clgCryptoEngineHeadPtr = newListPtr;
                }
                else
                {
                    tm_free_raw_buffer((ttRawBufferPtr)newEnginePtr);
                    tm_free_raw_buffer((ttRawBufferPtr)newListPtr);
                    newEnginePtr = (ttCryptoEnginePtr)0;
                }
            }
        }
    }

REG_RETURN:
    *newEnginePtrPtr = newEnginePtr;
    return errorCode;
}



/*
 * tfCryptoEngineDeRegister() Function Description
 * Called to de-register the crypto engine
 *
 * Parameter         description
 * engineId          Crypto Engine id used to find the previously registered 
 *                   crypto engine.
 *                   TM_CRYPTO_ENGINE_SOFTWARE  and  TM_CRYPTO_ENGINE_HIFN7951
 *
 * Return value:     TM_ENOERROR, otherwise failed to find
 */
int tfCryptoEngineDeRegister(unsigned int               engineId)
{
/*
 * TBD: To do this safely we need to add an owner count incremented for every
 * tfCryptoEngineGet(), and need to add a tfCryptoEngineUnGet() etc..
 */
    ttCryptoEngineListPtr    celPtr;
    ttCryptoEngineListPtr    prevPtr;
    int                      errorCode;

    errorCode = TM_ENOERROR;
    prevPtr = (ttCryptoEngineListPtr)0;
    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    if(!tm_context(tvCryptoLibPtr))
    {
        errorCode = TM_EINVAL;
        goto DEREG_RETURN;
    }

    celPtr = ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
             ->clgCryptoEngineHeadPtr;

    while(celPtr)
    {
        if(celPtr->celCryptoEnginePtr->ceEngineId == engineId)
        {
            break;
        }
        prevPtr = celPtr;
        celPtr = celPtr->celNextPtr;
    }

    if(celPtr)
    {
        if(!prevPtr)
        {
/* it is the head */
            ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
             ->clgCryptoEngineHeadPtr = celPtr->celNextPtr;
        }
        else
        {
            prevPtr->celNextPtr = celPtr->celNextPtr;
        }
/* now we are safe to free them */
        tm_free_raw_buffer((ttRawBufferPtr)celPtr->celCryptoEnginePtr);
        tm_free_raw_buffer((ttRawBufferPtr)celPtr);

    }
    else
    {
/* we didn't find anything */
        errorCode = TM_ENOENT;
    }
DEREG_RETURN:
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return errorCode;
}


/*
 * tfCryptoEngineAddAlgorithm() Function Description
 * Add supported algorithm to the crypto engine..
 *
 * Parameter         description
 * engineId          Crypto Engine number, currently, we support 
 *                   TM_CRYPTO_ENGINE_SOFTWARE  and  TM_CRYPTO_ENGINE_HIFN7951
 * algorithmId       Algorithm ID information, for example, SADB_AALG_MD5HMAC,
 *                   SADB_AALG_SHA1HMAC... SADB_EALG_3DESCBC...
 *
 * Return value:     TM_ENOERROR, otherwise failed
 */


int tfCryptoEngineAddAlgorithm(ttCryptoEnginePtr      cenginePtr,
                               unsigned int           algorithmId)
{
    int                 errorCode;

    errorCode = TM_ENOERROR;

    if(algorithmId > SADB_ALGORITHM_MAX 
               || cenginePtr == (ttCryptoEnginePtr)0)
    {
        errorCode = -1;
        goto ADDALG_RETURN;
    }

    cenginePtr->ceSupportFlags |= (TM_UL(1) << algorithmId);


ADDALG_RETURN:
    return errorCode;
}

/*DESCRIPTION: This function initialize the crypto library. It is called
 * when tfUseIpsec, or tfUsePki is called. If software engine is defined,
 * this call will register software engine and add all necessory algorithms
 */
int tfCryptoEngineInit(void)
{
    int               errorCode;
#ifdef TM_USE_MCF5235_CRYPTOENGINE
    ttCryptoEnginePtr mcf5235EnginePtr;
#endif /* TM_USE_MCF5235_CRYPTOENGINE */

    errorCode = TM_ENOERROR;

    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    if(!tm_context(tvCryptoLibPtr))
    {
        tm_context(tvCryptoLibPtr) = tm_get_raw_buffer(
                sizeof(ttCryptoLibGlobal));
        if(!tm_context(tvCryptoLibPtr))
        {
            errorCode = TM_ENOBUFS;
            goto CEINIT_RETURN;
        }
        tm_bzero(tm_context(tvCryptoLibPtr), sizeof(ttCryptoLibGlobal));


#ifdef TM_USE_MCF5235_CRYPTOENGINE
        if(!tfCryptoEngineLookup(TM_CRYPTO_ENGINE_MCF5235))
        {
            errorCode = tfCryptoEngineRegisterLocked(
                                               TM_CRYPTO_ENGINE_MCF5235,
                                               (ttUserVoidPtr)0,
                                               &mcf5235EnginePtr,
                                               tfMcf5235EngineInit,
                                               tfMcf5235GetRandomWord,
                                               (ttCryptoSessionOpenFuncPtr)0,
                                               tfMcf5235SessionProcess,
                                               (ttCryptoSessionFuncPtr)0);
            if(errorCode == TM_ENOERROR && mcf5235EnginePtr)
            {
/* add the following supported algorithms, you could just use public API
 * tfCryptoEngineAddAlgorithm one algorithm by one algoirthm, for example,
 * tfCryptoEngineAddAlgorithm(softEnginePtr, SADB_AALG_MD5HMAC );
 */
#ifdef TM_USE_MD5
                mcf5235EnginePtr->ceSupportFlags |=
                    (TM_UL(1) << SADB_AALG_MD5HMAC);
#endif /* TM_USE_MD5 */
/* DES and 3DES uses the same module */

            }
        }
#endif /* TM_USE_MCF5235_CRYPTOENGINE */
    }

CEINIT_RETURN:
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return errorCode;
}


void tfCryptoEngineUninitialize(void)
{
/*
 * TBD: To do this safely we need to add an owner count incremented for every
 * tfCryptoEngineGet(), and need to add a tfCryptoEngineUnGet() etc..
 */
    ttCryptoEngineListPtr     celPtr;
    ttCryptoEngineListPtr     prevCelPtr;

    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    if(tm_context(tvCryptoLibPtr))
    {
        celPtr = ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
                              ->clgCryptoEngineHeadPtr;
        while(celPtr)
        {
            prevCelPtr = celPtr;
            celPtr = celPtr->celNextPtr;
            tm_free_raw_buffer((ttRawBufferPtr)prevCelPtr->celCryptoEnginePtr);
            tm_free_raw_buffer((ttRawBufferPtr)prevCelPtr);
        }
        tm_free_raw_buffer(tm_context(tvCryptoLibPtr));
        tm_context(tvCryptoLibPtr) = 0;
    }
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return;
}

ttCryptoEnginePtr tfCryptoEngineGetFirst (void)
{
    ttCryptoEngineListPtr    celPtr;
    ttCryptoEnginePtr        cePtr;

    cePtr = (ttCryptoEnginePtr)0;
    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    if (tm_context(tvCryptoLibPtr) != (ttVoidPtr)0)
    {
        celPtr = ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
                                  ->clgCryptoEngineHeadPtr;
        if (celPtr != (ttCryptoEngineListPtr)0)
        {
            cePtr = celPtr->celCryptoEnginePtr;
        }
    }
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return cePtr;
}

/* find a crypto engine which supports the given algorithm*/
ttCryptoEnginePtr tfCryptoEngineGet (int       algorithm)
{
    return tfCryptoEngineGet2(algorithm, -1);
}

/* find a crypto engine which supports the given 2 algorithms */
ttCryptoEnginePtr tfCryptoEngineGet2 (int algorithm1, int algorithm2)
{
    ttCryptoEngineListPtr    celPtr;
    ttCryptoEnginePtr        cePtr;
    
    cePtr = (ttCryptoEnginePtr)0;
    tm_call_lock_wait(&(tm_context(tvCryptoLockEntry)));
    if (tm_context(tvCryptoLibPtr) != (ttVoidPtr)0)
    {
        celPtr = ((ttCryptoLibGlobalPtr)tm_context(tvCryptoLibPtr))
                                  ->clgCryptoEngineHeadPtr;
        while (celPtr)
        {
            cePtr = celPtr->celCryptoEnginePtr;
            if (cePtr->ceSupportFlags & (TM_UL(1) << algorithm1))
            {
                if (algorithm2 == -1)
                {
                    break;
                }
                else
                {
                    if (cePtr->ceSupportFlags & (TM_UL(1) << algorithm2))
                    {
                        break;
                    }
                }
            }
            celPtr = celPtr->celNextPtr;
        }
        if (celPtr == (ttCryptoEngineListPtr)0)
        {
            cePtr = (ttCryptoEnginePtr)0;
        }
    }
    tm_call_unlock(&(tm_context(tvCryptoLockEntry)));
    return cePtr;
}


#if (defined(TM_IPSEC_DEBUG) || defined(TM_IKE_DEBUG) || defined(TM_SSL_DEBUG))

#define TM_CRYPTO_NUM_32BITS 6 /* display at most 6 32-bits per line */
#define TM_CRYPTO_SIZE_32BIT 4 /* size of a 32-bit */
#define TM_CRYPTO_LEN_OFFSET 4 /* room to store offset in print buffer */
void tfCryptoDebugPrintf(ttCharPtrPtr  printBufPtrPtr,
                         tt8BitPtr     dataPtr,
                         tt32Bit       dataLength,
                         ttIntPtr      offPtr)
{
    ttCharPtr printBufPtr;
    int       printOffset;
    int       highDataByte;
    int       dataByte;
    int       i;
    int       j;
    int       k;
    static const char  TM_CONST_QLF tlAsciiHex[16] = \
            {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

    if (printBufPtrPtr != (ttCharPtrPtr)0)
    {
        printBufPtr = *printBufPtrPtr;
    }
    else
    {
        printBufPtr = (ttCharPtr)0;
    }
    if (printBufPtr == (ttCharPtr)0)
    {
        printOffset = TM_CRYPTO_LEN_OFFSET;
/*
 * Enough room for
 *   TM_CRYPTO_LEN_OFFSET bytes at the beginning (to store offset)
 * + TM_CRYPTO_NUM_32BITS in hex with space in between * and line return
 * + final null
 */
        printBufPtr = tm_get_raw_buffer(
                     TM_CRYPTO_LEN_OFFSET 
                   + (TM_CRYPTO_NUM_32BITS * ((TM_CRYPTO_SIZE_32BIT * 2) + 1))
                   + 1 );
    }
    else
    {
        printOffset = (int)(*((tt32BitPtr)(ttVoidPtr)printBufPtr));
    }
    if (printBufPtr != (ttCharPtr)0)
    {
        j = 0;
        if (offPtr != (ttIntPtr)0)
        {
            k = *offPtr;
        }
        else
        {
            k = 0;
        }
        for (i = 0; i < (int)dataLength; i++)
        {
            dataByte = ((int)(dataPtr[i])) & 0xFF;
            highDataByte = dataByte / 16; /* hex base */
            dataByte = dataByte & 0xF;
            printBufPtr[printOffset++] = tlAsciiHex[highDataByte]; 
            printBufPtr[printOffset++] = tlAsciiHex[dataByte]; /* low Byte */
            j++;
            k++;
            if (j >= TM_CRYPTO_SIZE_32BIT)
            {
                j = 0;
            }
            if (k >= (TM_CRYPTO_NUM_32BITS * TM_CRYPTO_SIZE_32BIT))
            {
                k = 0;
            }
            if (k == 0)
            {
/* Flush + allocate */
                *((tt32BitPtr)(ttVoidPtr)printBufPtr) = (tt32Bit)printOffset;
                tfCryptoDebugFlush(&printBufPtr);
                if (i == ((int)dataLength - 1))
                {
/* done with user data */
                    break;
                }
/*
 * new allocation 
 * Enough room for
 *   TM_CRYPTO_LEN_OFFSET bytes at the beginning (to store offset)
 * + TM_CRYPTO_NUM_32BITS in hex with space in between * and line return
 * + final null
 */
                printBufPtr = tm_get_raw_buffer(
                     TM_CRYPTO_LEN_OFFSET 
                   + (TM_CRYPTO_NUM_32BITS * ((TM_CRYPTO_SIZE_32BIT * 2) + 1))
                   + 1 );
                if (printBufPtr == (ttCharPtr)0)
                {
                    break;
                }
                printOffset = 4;
            }
            else if (j == 0)
            {
                printBufPtr[printOffset++] = ' ';
            }
        }
        if (printBufPtr != (ttCharPtr)0)
        {
/* Keep print offset into buffer */
            *((tt32BitPtr)(ttVoidPtr)printBufPtr) = (tt32Bit)printOffset;
        }
        if ( (offPtr != (ttIntPtr)0) && (printBufPtrPtr != (ttCharPtrPtr)0) )
        {
            *offPtr = k;
        }
        else
        {
/*
 * Flush.
 * Done with that data, or user is not saving the print buffer.
 * No choice but to flush it. printBufPtr is null upon return. 
 */ 
            tfCryptoDebugFlush(&printBufPtr);
            if (offPtr != (ttIntPtr)0)
/* Reset */
            {
                *offPtr = 0;
            }
        }
        if (printBufPtrPtr != (ttCharPtrPtr)0)
        {
            *printBufPtrPtr = printBufPtr;
        }
    }
    return;
}

void tfCryptoDebugFlush(ttCharPtrPtr printBufPtrPtr)
{
    ttCharPtr printBufPtr;
    int       printOffset;

    if (printBufPtrPtr != (ttCharPtrPtr)0)
    {
        printBufPtr = *printBufPtrPtr;
        if (printBufPtr != (ttCharPtr)0)
        {
            printOffset = (int)(*((tt32BitPtr)(ttVoidPtr)printBufPtr));
            if (printOffset > TM_CRYPTO_LEN_OFFSET)
            {
                if (printBufPtr[printOffset - 1] == ' ')
/* Replace space separator with '\n' */
                {
                    printOffset--;
                }
                printBufPtr[printOffset++] = '\n';
                printBufPtr[printOffset++] = '\0';
/* Flush */
                tm_debug_log0(&printBufPtr[4]);
            }
            tm_free_raw_buffer(printBufPtr);
            *printBufPtrPtr = (ttCharPtr)0;
        }
    }
    return;
}

#endif /*TM_IPSEC_DEBUG || TM_IKE_DEBUG || TM_SSL_DEBUG */



void tfCryptoEncAlgToStr(ttUser8Bit module, 
                         ttUser16Bit encAlg, 
                         ttUserConstCharPtrPtr bufferPtrPtr)
{
    switch(module)
    {
    case TM_CRYPTO_MODULE_IKE:
        {
            switch(encAlg)
            {
            case TM_IKE_DES_CBC:
                {
                    *bufferPtrPtr = tlEncAlgDes;
                    break;
                }
            case TM_IKE_IDEA_CBC:
                {
                    *bufferPtrPtr = tlEncAlgIdea;
                    break;
                }
            case TM_IKE_BLOWFISH_CBC:
                {
                    *bufferPtrPtr = tlEncAlgBlowfish;
                    break;
                }
            case TM_IKE_RC5_R16_B64_CBC:
                {
                    *bufferPtrPtr = tlEncAlgRC5R16B64;
                    break;
                }
            case TM_IKE_3DES_CBC:
                {
                    *bufferPtrPtr = tlEncAlg3Des;
                    break;
                }
            case TM_IKE_CAST_CBC:
                {
                    *bufferPtrPtr = tlEncAlgCast;
                    break;
                }
            case TM_IKE_AES_CBC:
                {
                    *bufferPtrPtr = tlEncAlgAes;
                    break;
                }
            case TM_IKE_NULL_ENCRYPT:
                {
                    *bufferPtrPtr = tlEncAlgNull;
                    break;
                }
            case TM_IKE_TWOFISH_CBC:
                {
                    *bufferPtrPtr = tlEncAlgTwoFish;
                    break;
                }
            default:
                {
                    *bufferPtrPtr = tlEncAlgNoSupport;
                    break;
                }            }

        }
    default:
        break;
    }
    *bufferPtrPtr = tlEncAlgDes;
}

void tfCryptoHashAlgToStr(ttUser8Bit module, 
                          ttUser16Bit hashAlg, 
                          ttUserConstCharPtrPtr bufferPtrPtr)
{
    switch(module)
    {
    case TM_CRYPTO_MODULE_IKE:
        {
            switch(hashAlg)
            {
            case TM_IKE_MD5:
                {
                    *bufferPtrPtr = tlHashAlgMd5;
                    break;
                }
            case TM_IKE_SHA1:
                {
                    *bufferPtrPtr = tlHashAlgSha1;
                    break;
                }
            case TM_IKE_SHA256:
                {
                    *bufferPtrPtr = tlHashAlgSha256;
                    break;
                }
            case TM_IKE_SHA384:
                {
                    *bufferPtrPtr = tlHashAlgSha384;
                    break;
                }
            case TM_IKE_SHA512:
                {
                    *bufferPtrPtr = tlHashAlgSha512;
                    break;
                }
            case TM_IKE_TIGER:
                {
                    *bufferPtrPtr = tlHashAlgTiger;
                    break;
                }
            case TM_IKE_RIPEMD:
                {
                    *bufferPtrPtr = tlHashAlgRIPEMD;
                    break;
                }
            default:
                {
                    *bufferPtrPtr = tlHashAlgNoSupport;
                    break;
                }
            }
        }
    default:
        break;
    }
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: trcrpeng.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 60778    Output checksum:  1275
 */
