/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: 16550 serial driver
 *
 * Filename: tr16550.c
 * Author: Jason
 * Date Created: 10/29/1999
 * $Source: source/drivers/16550/trc3uart.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:44CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 *   This driver was originally designed on a TMS320/C31 DSP target
 *   (Tiger 31/PC development board from DSP Research).  It uses
 *   memory-mapped I/O to access the 16550 registers.  Additionally, it
 *   deals with all of the byte/word issues with interacting with the
 *   32-bit C3x DSP.  For a 16550 device driver for general purpose
 *   processors, please see tr16550.c.
 *
 *   Due to differences in targets, this driver may require
 *   modifications for your specific hardware platform.  Sections of this
 *   driver that will require modifications include:
 *     - enabling and setting up the ISR (tf16550Open);
 *       restoring ISR on close (tf16550Close)
 *     - memory/port address for serial port (TM_16550_PORT_ADDR)
 *     - termination of the serial communication driver (tf16550Close).
 *
 *   This driver is designed for use as a single interface, i.e. with one
 *   serial port.  If multiple serial ports were to be used, each routine
 *   would have to be multiplexed on the interface handle that is passed in.
 *
 *   Currently, the 16550 is set to 8-N-1 operation at 1200 bps.  These values
 *   can be change by modifying tf16550Open and the constant TM_16550_BPS.  No
 *   type of hardware flow control is supported.
 *
 *   The only difference between this driver and the 8250 driver is that this
 *   takes advantage of the buffering (16-byte FIFO) capabilities of the
 *   16550. Both drivers should work with a standard 16550 serial controller.
 */

#include <trsocket.h>
#include "tr16550.h"
#include "tiger31.h"

#define TM_16550_BASE_ADDR 0x00fffff0

#define TM_16550_BPS           1200UL

#define TM_16550_BUFFER_SIZE    1024

/* Write out a word to specified port address*/
#define tm_out_word(tp_value,tp_port) \
*((unsigned int TM_FAR *) (TM_16550_BASE_ADDR + (tp_port))) = (tp_value)

/* Read in a word from specified port address */
#define tm_in_word(tp_port) \
*((unsigned int TM_FAR *) (TM_16550_BASE_ADDR + (tp_port)))

/*
 * Register offsets
 */

/*
 * RHR: Receive Holding Register  also:
 * THR: Transmit Holding Register
 */
#define TM_16550_RHR  0
#define TM_16550_THR  0

/* RHRI: Receive Holding Register Interrupt */
#define TM_16550_IER  1

#define TM_16550_DIV_LATCH_LSB  0
#define TM_16550_DIV_LATCH_MSB  1

/*
 * ISR: Interrupt Status Register  also:
 * FCR: FIFO Control Register
 */
#define TM_16550_ISR  2
#define TM_16550_FCR  2

/* LCR: Line Control Register */
#define TM_16550_LCR  3

/* MCR: Modem Control Register */
#define TM_16550_MCR  4

/* LSR: Line Status Register */
#define TM_16550_LSR  5

/* MSR: Modem Status Register */
#define TM_16550_MSR  6

/* SPR: Scratch Pad Register */
#define TM_16550_SPR  7




/*
 * Flags for the Line Control Register (LCR)
 */

/* Data bits per serial word */
#define TM_16550_8DATA_BITS     0x03
#define TM_16550_7DATA_BITS     0x02
#define TM_16550_6DATA_BITS     0x01
#define TM_16550_5DATA_BITS     0x00

/* Number of stop bits per word */
#define TM_16550_1STOP_BIT      0x00
#define TM_16550_2STOP_BIT      0x04

/* Parity */
#define TM_16550_NO_PARITY      0x00
#define TM_16550_ODD_PARITY     0x08
#define TM_16550_EVEN_PARITY    0x18
#define TM_16550_MARK_PARITY    0x28
#define TM_16550_SPC_PARITY     0x38

#define TM_16550_DLR_ENABLE     0x80

/*
 * Flags for the Line Status Register (LSR)
 */
#define TM_16550_XMIT_EMPTY     0x20
#define TM_16550_DATA_AVAIL     0x01

/*
 * Flags for the Modem Control Register (MCR)
 */
#define TM_16550_GPO2           0x00  /* No flow control. */
/* #define TM_16550_GPO2           0x22 /* Hardware flow control. */

/*
 * Flags for the Interrupt Enable Register (IER)
 */

/* Recieve Holding Register Interrupt, RxRDY */
#define TM_16550_RHRI_ON        0x01
#define TM_16550_RHRI_OFF       0x00

/* Transmit Holding Register Interrupt, TxRDY */
#define TM_16550_THRI           0x02

/* Receive Line Status Interrupt; ERROR */
#define TM_16550_RLSI           0x04

/* Modem Status Interrupt; DELTA */
#define TM_16550_MSI            0x08

/*
 * Flags for the FIFO Control Register (FCR)
 */

/* Enable the recieve and transmit FIFOs */
#define TM_16550_FIFO_ENB        0x01

/* Reset the recieve FIFO */
#define TM_16550_RECV_FIFO_RESET 0x02

/* Reset the transmit FIFO */
#define TM_16550_XMIT_FIFO_RESET 0x04

/* Sets the trigger level for the recieve FIFO interrupt (1,4,8 or 14) */
#define TM_16550_FIFO_LVL_01     0x00
#define TM_16550_FIFO_LVL_04     0x40
#define TM_16550_FIFO_LVL_08     0x80
#define TM_16550_FIFO_LVL_14     0xc0


/* Index indicating the last buffer that was freed by the stack */
int tl16550FreeBufferIndex;

/* Index indicating the next buffer to place incoming data into */
int tl16550IsrBufferIndex;

/* Buffer to recieve data */
char tl16550CharBuffer[2][TM_16550_BUFFER_SIZE];
int tl16550CharBufferLen[2];

/* Interface handle for the serial rate */
ttUserInterface tl16550InterfaceHandle;

/* old serial interrupt vector */
void TM_FAR * tl16550OldIntVector;

/*
 * tf16550Open
 *
 *   Intializes the 16550 serial controller and intializes the chip with
 *   baud rate and line control (stop bits, parity, etc).
 *
 */
int tf16550Open ( ttUserInterface interfaceHandle )
{
    unsigned int  lcrFlags;
    unsigned int  fcrFlags;
    unsigned int  baudRate;


    tl16550InterfaceHandle = interfaceHandle;
    tl16550IsrBufferIndex  = 1;
    tl16550FreeBufferIndex = 0;
    tl16550CharBufferLen[0] = 0;
    tl16550CharBufferLen[1] = 0;

/* Set the serial params: 8-N-1 in the line control register (LCR) */
    lcrFlags = TM_16550_8DATA_BITS | TM_16550_1STOP_BIT |
        TM_16550_NO_PARITY | TM_16550_DLR_ENABLE;
    tm_out_word(lcrFlags, TM_16550_LCR);

/* Setup the baud rate.  First, we divide the constant 0x1C200 (which
 * corresponds to 115200 Kbps) by the desired baud rate.  Above, by setting the
 * DLR enable bit, we allowed access to the DLR (Divisor Latch Register) which
 * treats register 0 (RHR) and register 1 (IER) as a single 16-bit word.  To
 * set the baud rate,  we write the value computed above (0x1C200/baud)
 * into the DLR
 */
    baudRate = (unsigned short) (0x1C200UL / (unsigned long) TM_16550_BPS);
    tm_out_word(baudRate, TM_16550_DIV_LATCH_LSB);
    tm_out_word((baudRate >> 8), TM_16550_DIV_LATCH_MSB);

/*
 * Now that we're done setting the baud rate, set the first 2 registers back
 * to their original function (disable the divisor latch register).
 */
    lcrFlags &= ~TM_16550_DLR_ENABLE;
    tm_out_word(lcrFlags, TM_16550_LCR);

/*
 * Set GPO2 int he MCR to allow interrupt servicing and set the IER to enable
 * recieve interrupt so we're notified of incoming characters.
 */
    tm_out_word(TM_16550_GPO2, TM_16550_MCR);
    tm_out_word(TM_16550_RHRI_ON, TM_16550_IER);

/*
 * Setup the FIFO operation: enable and reset the recv/xmit FIFOs and set the
 * trigger level to 14 bytes.
 */
    fcrFlags = TM_16550_FIFO_ENB | TM_16550_XMIT_FIFO_RESET |
               TM_16550_RECV_FIFO_RESET | TM_16550_FIFO_LVL_14;
    tm_out_word(fcrFlags, TM_16550_FCR);

/*
 * Setup the ISR. This is very hardware specific and will probably have to
 * be modified for your hardware.
 */

/* Set the interrupt vector */
    T31_setISR((long) &INT3, tf16550IsrHandler);

    return(TM_DEV_OKAY);
}


/*
 * tf16550Close
 *
 *   Closes the serial interface:  
 */
int tf16550Close ( ttUserInterface interfaceHandle )
{
    TM_UNREF_IN_ARG(interfaceHandle);

    return(TM_DEV_OKAY);
}

/*
 * tf16550Ioctl
 *
 *   No need for an IOCTL routine in our serial driver.
 */
int tf16550Ioctl ( ttUserInterface interfaceHandle,
                   int             flag,
                   void TM_FAR *   optionPtr,
                   int             optionLen)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(flag);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);

    return(TM_DEV_OKAY);
}


/*
 * tf16550Send
 *
 *   Transmits data to the 16550 serial controller.  Since the 16550 is
 *   completely byte oriented, we loop through the entire packet sending one
 *   byte at a time, waiting after each sent character to ensure that the
 *   previous character has completed transmitting.
 *
 */
int tf16550Send( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag)
{
    unsigned int offset;
    unsigned int dataWord;
    volatile unsigned int inWord;

    offset = 0;
    while (dataLength > 0)
    {
/*
 * Ensure that the previous character has finished transmitting - this is
 * indicated by the 'transmit buffer empty' bit in the LSR being set.
 */
        inWord = tm_in_word(TM_16550_LSR);
        while (!(inWord & TM_16550_XMIT_EMPTY))        
        {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
            tfInterfaceSpinLock(interfaceHandle);
            inWord = tm_in_word(TM_16550_LSR);            
        }

        dataWord = ((*dataPtr) >> ((TM_DSP_ROUND_PTR - offset) << 3)) & 0xff;
/* Send the current character to the 16550 */
        tm_out_word(dataWord, TM_16550_THR);

/* Move on to next character */
        if ( (++(offset) % TM_DSP_BYTES_PER_WORD) == 0)
        {
            offset = 0;
            dataPtr++;
        }
        
        dataLength--;
    }


/* Full packet has completed sending, inform the stack */
    if (flag & TM_USER_BUFFER_LAST)
    {
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }

    return(TM_DEV_OKAY);
}


/*
 * tf16550Receive
 *
 *   Reads data from the 16550 FIFO into our recieve buffer and pass this
 *   buffer to the stack.
 */
int tf16550Receive ( ttUserInterface        interfaceHandle,
                     char TM_FAR * TM_FAR * dataPtr,
                     int  TM_FAR *          dataLength,
                     ttUserBufferPtr        bufHandlePtr)
{
    int length;

    TM_UNREF_IN_ARG(interfaceHandle);

    T31_disable_interrupts();
/* reserve the current receive buffer for our use, and give the other receive
   buffer to the ISR so that it can continue receiving characters */
    tl16550FreeBufferIndex = tl16550IsrBufferIndex;
    if (tl16550IsrBufferIndex == 0)
    {
        tl16550IsrBufferIndex = 1;
    }
    else
    {
        tl16550IsrBufferIndex = 0;
    }
    T31_enable_interrupts();

    length = tl16550CharBufferLen[tl16550FreeBufferIndex];
    tl16550CharBufferLen[tl16550FreeBufferIndex] = 0;

    if (length > TM_16550_BUFFER_SIZE)
    {
        tfKernelError("tf16550Receive", "Buffer length exceeded!");
        tm_thread_stop;
    }

/* Return a pointer to the buffer and the length to the stack */
    *dataPtr = tl16550CharBuffer[tl16550FreeBufferIndex];
    *dataLength = length;
    *bufHandlePtr = (ttUserBufferPtr) 0;

    return (TM_DEV_OKAY);
}

unsigned int debugLog[128];
unsigned int debugIndex = 0;

/*
 * tf16550IsrHandler
 *
 *   Notifies the stack of incoming serial data.
 *
 */
void TM_INTERRUPT tf16550IsrHandler(void)
{
    unsigned int dataWord;
    unsigned int * bufDataPtr;
    int oldRcvBufLen;
    int newRcvBufLen;
    int i;


    oldRcvBufLen = tl16550CharBufferLen[tl16550IsrBufferIndex];
    newRcvBufLen = oldRcvBufLen;

/*
 * While there is still data available in the FIFO and we have room for it in
 * our buffer, copy data from the 16550 FIFO into our buffer.
 */
    while ( ( newRcvBufLen < TM_16550_BUFFER_SIZE )
            && ( tm_in_word(TM_16550_LSR) & TM_16550_DATA_AVAIL ) )
    {
        dataWord = tm_in_word(TM_16550_RHR) & 0xff;

/* Store the received byte into the correct byte (starting with the most
   significant) in the current word in the buffer. */
        bufDataPtr = (unsigned int *)
            (tl16550CharBuffer[tl16550IsrBufferIndex] + (newRcvBufLen >> 2));
        *bufDataPtr &= ~(0xff << ((TM_DSP_ROUND_PTR
            - ((newRcvBufLen) % TM_DSP_BYTES_PER_WORD)) << 3));
        *bufDataPtr |= (dataWord << ((TM_DSP_ROUND_PTR
            - ((newRcvBufLen) % TM_DSP_BYTES_PER_WORD)) << 3));

        newRcvBufLen++;
    }

    if (newRcvBufLen >= TM_16550_BUFFER_SIZE)     /* software buffer full... */
    {
        for (i=0 ; i < 17 ; i++)
        {                               /* ...so exhaust input hardware FIFO */
            tm_in_word(TM_16550_RHR);
        }
    }

    tl16550CharBufferLen[tl16550IsrBufferIndex] = newRcvBufLen;

    if (oldRcvBufLen != newRcvBufLen)
    {
/* notify the stack that we received some data */
        tfNotifyInterfaceIsr(tl16550InterfaceHandle, 1, 0, 0UL, 0UL);
    }
    
    return;
}


/*
 * tf16550FreeReceiveBuffer
 *
 *   Called when the stack is done with a recieved packet (ie, character).
 *   This should be immediately after it is passed up the stack, since both of
 *   the serial link layers (PPP,SLIP) copy immediately.  Therefore, we only
 *   use one buffer to recieve data into, so this routine does nothing,
 *   and isn't needed.
 *
 */
int tf16550FreeReceiveBuffer(ttUserInterface interfaceHandle,
                            char TM_FAR *   dataPtr)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtr);
    return TM_DEV_OKAY;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: trc3uart.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 12952    Output checksum: 42360
 */
