/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: ZiLOG eZ80 UART driver
 *
 * Filename: trzuart.c
 * Author: Jason & Qin Zhang
 * Date Created: 12/19/2002
 * $Source: source/drivers/16550/trzuart.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:45CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 *   This driver was designed on a generic Intel PC, and will require
 *   modifications for your specific hardware platform.  Sections of this
 *   driver that will require modifications include:
 *     - enabling and setting up the ISR (tfUartOpen);
 *       restoring ISR on close (tfUartClose)
 *     - memory/port address for serial port (TM_UART_BASE)
 *     - compiler specific implementations of INP and OUTP x86 assembly
 *       commands (tm_in_byte, tm_out_byte, etc)
 *     - termination of the serial communication driver (tfUartClose).
 *
 *   This driver is designed for use as a single interface, i.e. with one
 *   serial port.  If multiple serial ports were to be used, each routine
 *   would have to be multiplexed on the interface handle that is passed in.
 *
 *   Currently, the Uart is set to 8-N-1 operation at 1200 bps.  These values
 *   can be change by modifying tfUartOpen and the constant TM_UART_BPS.  No
 *   type of hardware flow control is supported.
 *
 *   The only difference between this driver and the 8250 driver is that this
 *   takes advantage of the buffering (16-byte FIFO) capabilities of the
 *   Uart. Both drivers should work with a standard Uart serial controller.
 */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include <trzuart.h>
#ifdef TM_KERNEL_RZK_EZ80
#include "zsysgen.h"
#endif /* TM_KERNEL_RZK_EZ80 */

/*
 * Use UART0 or UART1: 
 * if you want to use UART0, uncomment the following macro
 */
#define TM_UART0

/* Define UART register base address */
#define TM_UART0_BASE           0xC0
#define TM_UART1_BASE           0xD0

#ifdef TM_KERNEL_NONE_EZ80
#ifdef _EZ80L92
/* Define interrupt vector base */
#define TM_IV_BASE          0x180
/* Define UART interrupt vector number */
#define TM_IV_UART0         0x3C /* 0x18 * 5 / 2 */
#define TM_IV_UART1         0x41 /* 0x1A * 5 / 2 */
#endif /* _EZ80L92 */

#ifdef _EZ80F91
/* Define interrupt vector base */
#define TM_IV_BASE          0x0
/* Define UART interrupt vector offset */
#define TM_IV_UART0         0x70
#define TM_IV_UART1         0x74
#endif /* _EZ80F91 */

#else /* TM_KERNEL_NONE_EZ80 */

/* Define UART interrupt vector number */
#define TM_IV_UART0            12
#define TM_IV_UART1            13

#endif /* TM_KERNEL_NONE_EZ80 */

#ifdef TM_UART0
/* Select UART register base address */
#define TM_UART_BASE       TM_UART0_BASE
/* Select UART interrupt vector */
#define TM_IV_UART         TM_IV_UART0
#else /* TM_UART0 */
/* Select UART register base address */
#define TM_UART_BASE       TM_UART1_BASE
/* Select UART interrupt vector */
#define TM_IV_UART         TM_IV_UART1
#endif /* TM_UART0 */

#ifndef SYSTEM_CLOCKH
#define SYSTEM_CLOCKH      48000000
#endif /* SYSTEM_CLOCKH */

/* eZ90 L92 board system clock rate / 16, used to generate Baud Rate */
#define TM_UART_BASE_CLOCK		(SYSTEM_CLOCKH/16)
             
#define TM_UART_BPS             9600

#define TM_UART_BUFFER_SIZE     1024

/* Write out a word for specified port address */
#define tm_out_word(tp_value,tp_port) \
*((volatile unsigned short __INTIO *)(tp_port)) = (unsigned short)(tp_value)
    
/* Write out a byte to specified port address */
#define tm_out_byte(tp_value,tp_port) \
*((volatile unsigned char __INTIO *) (tp_port)) = (unsigned char) (tp_value)

/* Read in a word from specified port address */
#define tm_in_word(tp_port) \
(*((volatile unsigned short __INTIO *)(tp_port)))

/* Read in a byte from specified port address */
#define tm_in_byte(tp_port) \
(*((volatile unsigned char __INTIO *)(tp_port)))


/*
 * Register offsets
 */

/*
 * RBR: Receive Buffer Register  also:
 * THR: Transmit Holding Register
 */
#define TM_UART_RBR  0
#define TM_UART_THR  0

/* IER: Interrupt Enable Register */
#define TM_UART_IER  1

/*
 * IIR: Interrupt Identification Register  also:
 * FCR: FIFO Control Register
 */
#define TM_UART_IIR  2
#define TM_UART_FCR  2

/* LCR: Line Control Register */
#define TM_UART_LCR  3

/* MCR: Modem Control Register */
#define TM_UART_MCR  4

/* LSR: Line Status Register */
#define TM_UART_LSR  5

/* MSR: Modem Status Register */
#define TM_UART_MSR  6

/* SPR: Scratch Pad Register */
#define TM_UART_SPR  7

/*
 * Flags for the Line Control Register (LCR)
 */
/* Data bits per serial word */
#define TM_UART_8DATA_BITS     0x03
#define TM_UART_7DATA_BITS     0x02
#define TM_UART_6DATA_BITS     0x01
#define TM_UART_5DATA_BITS     0x00

/* Number of stop bits per word */
#define TM_UART_1STOP_BIT      0x00
#define TM_UART_2STOP_BIT      0x04

/* Parity */
#define TM_UART_NO_PARITY      0x00
#define TM_UART_ODD_PARITY     0x08
#define TM_UART_EVEN_PARITY    0x18
#define TM_UART_MARK_PARITY    0x28
#define TM_UART_SPC_PARITY     0x38

#define TM_UART_DLR_ENABLE     0x80

/*
 * Flags for the Line Status Register (LSR)
 */
#define TM_UART_XMIT_EMPTY     0x20
#define TM_UART_DATA_AVAIL     0x01

/*
 * Flags for the Modem Control Register (MCR)
 */
#define TM_UART_GPO2           0x0B    /* turn on OUT2, RTS, DTR */

/*
 * Flags for the Interrupt Enable Register (IER)
 */

/* Recieve Buffer Register Interrupt, RxRDY */
#define TM_UART_RBRI_ON        0x01
#define TM_UART_RBRI_OFF       0x00

/* Transmit Holding Register Interrupt, TxRDY */
#define TM_UART_THRI           0x02

/* Receive Line Status Interrupt; ERROR */
#define TM_UART_RLSI           0x04

/* Modem Status Interrupt; DELTA */
#define TM_UART_MSI            0x08

/*
 * Flags for the FIFO Control Register (FCR)
 */

/* Enable the recieve and transmit FIFOs */
#define TM_UART_FIFO_ENB        0x01

/* Reset the recieve FIFO */
#define TM_UART_RECV_FIFO_RESET 0x02

/* Reset the transmit FIFO */
#define TM_UART_XMIT_FIFO_RESET 0x03

/* Sets the trigger level for the recieve FIFO interrupt (1,4,8 or 14) */
#define TM_UART_FIFO_LVL_01     0x00
#define TM_UART_FIFO_LVL_04     0x40
#define TM_UART_FIFO_LVL_08     0x80
#define TM_UART_FIFO_LVL_14     0xc0

/*
 * Flags for the FIFO Control Register (FCR)
 */
/* Interrupt status code mask */
#define TM_UART_INSTS_MASK          0x0E

/* Receiver line status */
#define TM_UART_INSTS_LINESTATUS    0x06

/* Receive data ready or trigger level */
#define TM_UART_INSTS_RECEIVE       0x04

/* Character time-out */
#define TM_UART_INSTS_TIMEOUT       0x0C

/* Transmit buffer empty */
#define TM_UART_INSTS_TRANSMIT      0x02

/* Modem status */
#define TM_UART_INSTS_MODEMSTATUS   0x00

/* UART interrupt bit */
#define TM_UART_INTBIT              0x01

/* Define max receive buffers */
#define TM_UART_MAX_BUFFERS         2

/*
 * Global variable definitions
 */

/* Index indicating the last buffer that was freed by the stack */
int tlUartFreeBufferIndex;

/* Index indicating the next buffer to place incoming data into */
int tlUartIsrBufferIndex;

/* Buffer to recieve data */
char tlUartCharBuffer[TM_UART_MAX_BUFFERS][TM_UART_BUFFER_SIZE];
int tlUartCharBufferLen[TM_UART_MAX_BUFFERS];

/* Interface handle for the serial rate */
ttUserInterface tlUartInterfaceHandle;

/* External function */
extern void tfUartIsrProlog(void);
extern void tfZ80InstallIsr( ttUserIsrHandlerPtr   funcPtr,
		                     int                   jumpBase,
		                     tt8Bit                offset);

/* Function prototype */
void tfUartIsrReceive(void);

/*
 * tfUartOpen
 *
 *   Intializes the Uart serial controller and intializes the chip with
 *   baud rate and line control (stop bits, parity, etc).
 *
 */
int tfUartOpen ( ttUserInterface interfaceHandle )
{
    unsigned char          flags;
    unsigned short         baudRate;
    void TM_FAR * TM_FAR * intVectorPtr;

    tlUartInterfaceHandle = interfaceHandle;
    tlUartIsrBufferIndex  = 1;
    tlUartFreeBufferIndex = 0;
    tlUartCharBufferLen[0] = 0;
    tlUartCharBufferLen[1] = 0;

/* Disable all interrupt */
    tm_out_byte(0, TM_UART_BASE + TM_UART_IER);

/* Set DLAB bit to set Baud Rate */
    flags = TM_UART_DLR_ENABLE;
    tm_out_byte(flags, TM_UART_BASE + TM_UART_LCR);

/* Calculate Baud Rate */
    baudRate = (unsigned short) ( TM_UART_BASE_CLOCK / TM_UART_BPS);
    tm_out_word(baudRate, TM_UART_BASE + TM_UART_RBR);

/*
 * Now that we're done setting the baud rate, set the first 2 registers back
 * to their original function (disable the divisor latch register).
 */
/* Set the serial params: 8-N-1 in the line control register (LCR) */
    flags = TM_UART_8DATA_BITS | TM_UART_1STOP_BIT | TM_UART_NO_PARITY;
    tm_out_byte(flags, TM_UART_BASE + TM_UART_LCR);

/*
 * Set GPO2 int MCR: OUT2 + RTS + DTR
 */
    tm_out_byte(TM_UART_GPO2, TM_UART_BASE + TM_UART_MCR);

/*
 * Setup the FIFO operation: enable and reset the recv/xmit FIFOs and set the
 * trigger level to 14 bytes.
 */
    flags = TM_UART_FIFO_ENB | TM_UART_XMIT_FIFO_RESET |
               TM_UART_RECV_FIFO_RESET | TM_UART_FIFO_LVL_14;
    tm_out_byte(flags, TM_UART_BASE + TM_UART_FCR);

/*
 * Setup the ISR. This is very hardware specific and will probably have to
 * be modified for your hardware.
 */
#ifdef TM_KERNEL_RZK_EZ80
    tfKernelInstallIsrHandler(tfUartIsrProlog, TM_IV_UART);
#else /* TM_KERNEL_NONE_EZ80 */
    tfZ80InstallIsr( tfUartIsrHandler, TM_IV_BASE, TM_IV_UART);
#endif /* TM_KERNEL_RZK_EZ80 */

/*
 * Enable interrupts : Receive interrupt
 */
    tm_out_byte(TM_UART_RBRI_ON, TM_UART_BASE + TM_UART_IER);

    return(TM_DEV_OKAY);
}


/*
 * tfUartClose
 *
 *   Closes the serial interface:  restore old interrupt handler and disable
 *   IRQ4/IRQ3 interupts in the PIC.  As with the interrupt initialization
 *   code, this is very hardware specific.
 */
int tfUartClose ( ttUserInterface interfaceHandle )
{
    void TM_FAR * TM_FAR * intVectorPtr;

    TM_UNREF_IN_ARG(interfaceHandle);

/* Disable interrupts */
    tm_out_byte(0, TM_UART_BASE + TM_UART_IER);

    return(TM_DEV_OKAY);
}

/*
 * tfUartIoctl
 *
 *   No need for an IOCTL routine in our serial driver.
 */
int tfUartIoctl ( ttUserInterface interfaceHandle,
                   int             flag,
                   void TM_FAR *   optionPtr,
                   int             optionLen)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(flag);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);

    return(TM_DEV_OKAY);
}


/*
 * tfUartSend
 *
 *   Transmits data to the Uart serial controller.  Since the Uart is
 *   completely byte oriented, we loop through the entire packet sending one
 *   byte at a time, waiting after each sent character to ensure that the
 *   previous character has completed transmitting.
 *
 */
int tfUartSend( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag)
{
    while (dataLength > 0)
    {
/*
 * Ensure that the previous character has finished transmitting - this is
 * indicated by the 'transmit buffer empty' bit in the LSR being set.
 */
        while (!(tm_in_byte(TM_UART_BASE+TM_UART_LSR) &
                 TM_UART_XMIT_EMPTY))
        {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
            tfInterfaceSpinLock(interfaceHandle);
        }

/* Send the current character to the Uart */
        tm_out_byte(*dataPtr, TM_UART_BASE + TM_UART_THR);

/* Move on to next character */
        dataLength--;
        dataPtr++;
    }


/* Full packet has completed sending, inform the stack */
    if (flag & TM_USER_BUFFER_LAST)
    {
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }

    return(TM_DEV_OKAY);
}


/*
 * tfUartReceive
 *
 *   Reads data from the Uart FIFO into our recieve buffer and pass this
 *   buffer to the stack.
 */
int tfUartReceive ( ttUserInterface        interfaceHandle,
                     char TM_FAR * TM_FAR * dataPtr,
                     int  TM_FAR *          dataLength,
                     ttUserBufferPtr        bufHandlePtr)
{
    int length;

    TM_UNREF_IN_ARG(interfaceHandle);

/* reserve the current receive buffer for our use, and give the other receive
   buffer to the ISR so that it can continue receiving characters */
    tm_kernel_set_critical;
    tlUartFreeBufferIndex = tlUartIsrBufferIndex;
	tlUartIsrBufferIndex++;
    tlUartIsrBufferIndex %= TM_UART_MAX_BUFFERS;
	tm_kernel_release_critical;

    length = tlUartCharBufferLen[tlUartFreeBufferIndex];
    tlUartCharBufferLen[tlUartFreeBufferIndex] = 0;

    if (length > TM_UART_BUFFER_SIZE)
    {
        tfKernelError("tfUartReceive", "Buffer length exceeded!");
        tm_thread_stop;
    }

/* Return a pointer to the buffer and the length to the stack */
    *dataPtr = tlUartCharBuffer[tlUartFreeBufferIndex];
    *dataLength = length;
    *bufHandlePtr = (ttUserBufferPtr) 0;

    return (TM_DEV_OKAY);
}


/*
 * tfUartIsrHandler
 *
 *   UART receive interrupt handle function
 *
 */
void TM_INTERRUPT tfUartIsrHandler(void)
{
    int oldRcvBufLen;
    int newRcvBufLen;
    int i;

    oldRcvBufLen = tlUartCharBufferLen[tlUartIsrBufferIndex];
    newRcvBufLen = oldRcvBufLen;

/*
 * While there is still data available in the FIFO and we have room for it in
 * our buffer, copy data from the Uart FIFO into our buffer.
 */
    while ( ( newRcvBufLen < TM_UART_BUFFER_SIZE )
            && ( tm_in_byte( TM_UART_BASE + TM_UART_LSR )
                 & TM_UART_DATA_AVAIL ) )
    {
        tlUartCharBuffer[tlUartIsrBufferIndex][newRcvBufLen++]
            = tm_in_byte(TM_UART_BASE + TM_UART_RBR);
    }

    if (newRcvBufLen >= TM_UART_BUFFER_SIZE)     /* software buffer full... */
    {
        for (i=0 ; i < 17 ; i++)
        {                               /* ...so exhaust input hardware FIFO */
            tm_in_byte(TM_UART_BASE + TM_UART_RBR);
        }
    }

    tlUartCharBufferLen[tlUartIsrBufferIndex] = newRcvBufLen;

    if (oldRcvBufLen != newRcvBufLen)
    {
/* notify the stack that we received some data */
        tfNotifyInterfaceIsr(tlUartInterfaceHandle, 1, 0, 0UL, 0UL);
    }

    return;
}


/*
 * tfUartFreeReceiveBuffer
 *
 *   Called when the stack is done with a recieved packet (ie, character).
 *   This should be immediately after it is passed up the stack, since both of
 *   the serial link layers (PPP,SLIP) copy immediately.  Therefore, we only
 *   use one buffer to recieve data into, so this routine does nothing.
 *
 */
int tfUartFreeReceiveBuffer(ttUserInterface interfaceHandle,
                            char TM_FAR *   dataPtr)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtr);
    return TM_DEV_OKAY;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: trzuart.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 36471    Output checksum: 42160
 */
