/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: 8250 serial driver
 *
 * Filename: tr8250.c
 * Author: Jason
 * Date Created: 10/29/1999
 * $Source: source/drivers/8250/tr8250.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:48CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 *   This driver was designed on a generic Intel PC, and will require 
 *   modificaitons for your specific hardware platform.  Sections of this 
 *   driver that will require modifications include:
 *     - enabling and setting up the ISR (tf8250Open); 
 *       restoring ISR on close (tf8250Close)
 *     - memory address for serial port (TM_8250_PORT_ADDR)
 *     - compiler specific implementations of INP and OUTP x86 assembly
 *       commands (tm_in_byte, tm_out_byte, etc)
 *     - termination of the serial communication driver (tf8250Close).
 *
 *   This driver is designed for use as a single interface, i.e. with one 
 *   serial port.  If multiple serial ports were to be used, each routine
 *   would have to be multiplexed on the interface handle that is passed in.
 *
 *   Currently, the 8250 is set to 8-N-1 operation at 19200 bps.  These values
 *   can be change by modifying tf8250Open and the constant TM_8250_BPS.  No
 *   type of hardware flow control is supported.
 *
 *   Also, this driver should work without modification on a newer 16550 serial
 *   controller.  However, a 16550 driver is preferred to take advantage of
 *   it's buffering capabilities.
 *
 *   The data recieve logic uses two buffer indexes, recieve and ISR.  The ISR
 *   index tracks where the driver will place the next piece of incoming data
 *   in the ISR.  The recieve index points to the next character that will be
 *   passed up the Treck stack.  This index is not incremented in the recieve
 *   routine itself, but when the tf8250FreeReceiveBuffer() routine is called
 *   from the stack.  This routine will be called when the PPP/SLIP layer is
 *   done processing this character, and data can be recieved into this 
 *   location again.
 *
 *   For an excellent document on programming the 8250 UART, please refer to:
 *      http://www.whisqu.se/per/docs/general25.htm
 */

/* Treck Specific Includes */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/* Driver specific includes */
#include <tr8250.h>

/* This is for writing to and reading from ports in the Borland compiler
 * Change as necessary.
 */
#include <dos.h>
#include <conio.h>

/* Select which COM port to use (COM1, COM2) */
#define TM_8250_USE_COM1

/* Speed of this serial port */
#define TM_8250_BPS           19200UL

/* Size of receive buffer */
#define TM_8250_BUFFER_SIZE       5


/* Default 8250 port (COM2) [COM1 = 0x2e8] */
#define TM_8250_COM1_ADDR     0x3f8
#define TM_8250_COM2_ADDR     0x2f8
#define TM_8250_COM3_ADDR     0x3e8
#define TM_8250_COM4_ADDR     0x2e8

#define TM_8250_PIC_ADDR      0x21


/* Address for interrupt vector (IRQ4 for COM1, IRQ3 for COM2) */
#define TM_8250_IRQ4_ADDR      0x30
#define TM_8250_IRQ3_ADDR      0x2C

#define TM_8250_IRQ4_OFFSET    0x0c
#define TM_8250_IRQ3_OFFSET    0x0b

#define TM_8250_PIC_IRQ3       0x08
#define TM_8250_PIC_IRQ4       0x10 

/* Use COM1 (0x3f8, IRQ4) */
#ifdef TM_8250_USE_COM1
#define TM_8250_PORT_ADDR  TM_8250_COM1_ADDR
#define TM_8250_IRQ_ADDR   TM_8250_IRQ4_ADDR
#define TM_8250_IRQ_OFFSET TM_8250_IRQ4_OFFSET
#define TM_8250_PIC_IRQ    TM_8250_PIC_IRQ4
#endif /* TM_8250_USE_COM1 */

/* Use COM2 (0x2f8, IRQ3) */
#ifdef TM_8250_USE_COM2
#define TM_8250_PORT_ADDR  TM_8250_COM2_ADDR
#define TM_8250_IRQ_ADDR   TM_8250_IRQ3_ADDR
#define TM_8250_IRQ_OFFSET TM_8250_IRQ3_OFFSET
#define TM_8250_PIC_IRQ    TM_8250_PIC_IRQ3
#endif /* TM_8250_USE_COM2 */

/* Write out a word to specified port address*/
#define tm_out_word(tp_value,tp_port)    outpw(tp_port,tp_value)

/* Write out a byte to specified port address */
#define tm_out_byte(tp_value,tp_port)    outp(tp_port,tp_value)

/* Read in a word from specified port address */
#define tm_in_word  inpw

/* Read in a byte from specified port address */
#define tm_in_byte  inp



/* 
 * Register offsets 
 */

/* 
 * RHR: Receive Holding Register  also:
 * THR: Transmit Holding Register
 */
#define TM_8250_RHR  0
#define TM_8250_THR  0

/* RHRI: Receive Holding Register Interrupt */
#define TM_8250_IER  1

/* ISR: Interrupt Status Register */
#define TM_8250_ISR  2

/* LCR: Line Control Register */
#define TM_8250_LCR  3

/* MCR: Modem Control Register */
#define TM_8250_MCR  4

/* LSR: Line Status Register */
#define TM_8250_LSR  5

/* MSR: Modem Status Register */
#define TM_8250_MSR  6

/* SPR: Scratch Pad Register */
#define TM_8250_SPR  7
 



/* 
 * Flags for the Line Control Register (LCR)
 */

/* Data bits per serial word */
#define TM_8250_8DATA_BITS     0x03
#define TM_8250_7DATA_BITS     0x02
#define TM_8250_6DATA_BITS     0x01
#define TM_8250_5DATA_BITS     0x00

/* Number of stop bits per word */
#define TM_8250_1STOP_BIT      0x00
#define TM_8250_2STOP_BIT      0x04

/* Parity */
#define TM_8250_NO_PARITY      0x00   
#define TM_8250_ODD_PARITY     0x08
#define TM_8250_EVEN_PARITY    0x18
#define TM_8250_MARK_PARITY    0x28   
#define TM_8250_SPC_PARITY     0x38   

#define TM_8250_DLR_ENABLE     0x80

/*
 * Flags for the Line Status Register (LSR)
 */
#define TM_8250_XMIT_EMPTY     0x20

/*
 * Flags for the Modem Control Register (MCR)
 */
#define TM_8250_GPO2           0x08

/*
 * Flags for the Interrupt Enable Register (IER)
 */

/* Recieve Holding Register Interrupt, RxRDY */
#define TM_8250_RHRI           0x01

/* Transmit Holding Register Interrupt, TxRDY */
#define TM_8250_THRI           0x02

/* Receive Line Status Interrupt; ERROR */
#define TM_8250_RLSI           0x04

/* Modem Status Interrupt; DELTA */
#define TM_8250_MSI            0x08

/* Index indicating the next buffer to place incoming data into */
int tl8250IsrBufferIndex;

/* Index indicating the next buffer to pass to the Treck stack for processing */
int tl8250RecvBufferIndex;

unsigned char bufferFullFlag;

/* Buffer to recieve data */
char tl8250CharBuffer[TM_8250_BUFFER_SIZE];

/* Interface handle for the serial rate */
ttUserInterface tl8250InterfaceHandle;

/* old serial interrupt vector */
void TM_FAR * tl8250OldIntVector;


/*
 * tf8250Open
 *
 *   Intializes the 8250 serial controller and intializes the chip with
 *   baud rate and line control (stop bits, parity, etc).
 *   
 */
int tf8250Open(ttUserInterface interfaceHandle)
{
    unsigned char          lcrFlags;
    unsigned char          picFlags;
    unsigned short         baudRate;
    void TM_FAR * TM_FAR * intVectorPtr;
        

    tl8250InterfaceHandle = interfaceHandle;
    tl8250RecvBufferIndex = 0;
    tl8250IsrBufferIndex  = 0;
    bufferFullFlag        = 0;

/* Set the serial params: 8-N-1 in the line control register (LCR) */
    lcrFlags = TM_8250_8DATA_BITS | TM_8250_1STOP_BIT | 
        TM_8250_NO_PARITY | TM_8250_DLR_ENABLE;
    tm_out_byte(lcrFlags, TM_8250_PORT_ADDR + TM_8250_LCR);

/* Setup the baud rate.  First, we divide the constant 0x1C200 (which
 * corresponds to 115200 Kbps) by the desired baud rate.  Above, by setting the 
 * DLR enable bit, we allowed access to the DLR (Divisor Latch Register) which 
 * treats register 0 (RHR) and register 1 (IER) as a single 16-bit word.  To
 * set the baud rate,  we write the value computed above (0x1C200/baud) 
 * into the DLR
 */
    baudRate = (unsigned short) (0x1C200UL / (unsigned long) TM_8250_BPS);
    tm_out_word(baudRate, TM_8250_PORT_ADDR + TM_8250_RHR);

/*
 * Now that we're done setting the baud rate, set the first 2 registers back
 * to their original function (disable the divisor latch register).
 */
    lcrFlags &= ~TM_8250_DLR_ENABLE;
    tm_out_byte(lcrFlags, TM_8250_PORT_ADDR + TM_8250_LCR);

/*
 * Set GPO2 int he MCR to allow interrupt servicing and set the IER to enable
 * recieve interrupt so we're notified of incoming characters.
 */
    tm_out_byte(TM_8250_GPO2, TM_8250_PORT_ADDR + TM_8250_MCR);
    tm_out_byte(TM_8250_RHRI, TM_8250_PORT_ADDR + TM_8250_IER);

/*
 * Setup the ISR. This is very hardware specific and will probably have to
 * be modified for your hardware.  This code is for a generic Intel PC,
 * with COM1/COM3 on IRQ4 and COM2/COM4 on IRQ4.
 */


/* Save the old interrupt vector */
    intVectorPtr = (void TM_FAR *) TM_8250_IRQ_ADDR;
    tl8250OldIntVector = *intVectorPtr;

/* Set the interrupt vector */
    tfKernelInstallIsrHandler((ttUserIsrHandlerPtr) tf8250IsrHandler, 
                              TM_8250_IRQ_OFFSET);

/* Enable IRQ4/IRQ3 interrupts in the PIC */
    picFlags  = tm_in_byte(TM_8250_PIC_ADDR);
    picFlags &= ~TM_8250_PIC_IRQ;
    tm_out_byte(picFlags,TM_8250_PIC_ADDR);

    return(TM_DEV_OKAY);            
}


/*
 * tf8250Close
 *
 *   Closes the serial interface:  restore old interrupt handler and disable
 *   IRQ4/IRQ3 interupts in the PIC.  As with the interrupt initialization code,
 *   this is very hardware specific.
 */
int tf8250Close(ttUserInterface interfaceHandle)
{
    unsigned char          picFlags;
    void TM_FAR * TM_FAR * intVectorPtr;

    TM_UNREF_IN_ARG(interfaceHandle);

/* Restore old interrupt handler */
    intVectorPtr  = (void TM_FAR *) TM_8250_IRQ3_ADDR;
    *intVectorPtr = tl8250OldIntVector;

/* Disable IRQ4/IRQ3 interrupt in the PIC */
    picFlags  = tm_in_byte(TM_8250_PIC_ADDR);
    picFlags |= TM_8250_PIC_IRQ3;
    tm_out_byte(picFlags,TM_8250_PIC_ADDR);

    return(TM_DEV_OKAY);
}

/*
 * tf8250Ioctl
 *
 *   No need for an IOCTL routine in our serial driver.
 */
int tf8250Ioctl(ttUserInterface interfaceHandle,
                int             flag,
                void TM_FAR *   optionPtr,
                int             optionLen)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    flag            = flag;
    optionPtr       = optionPtr;
    optionLen       = optionLen;

    return(TM_DEV_OKAY);
}


/*
 * tf8250Send
 *
 *   Transmits data to the 8250 serial controller.  Since the 8250 is completely
 *   byte oriented, we loop through the entire packet sending one byte at a 
 *   time, waiting after each sent character to ensure that the previous 
 *   character has completed transmitting.
 *
 */
int tf8250Send( ttUserInterface interfaceHandle,
                char TM_FAR *   dataPtr,
                int             dataLength,
                int             flag )
{

    while (dataLength > 0)
    {
/* 
 * Ensure that the previous character has finished transmitting - this is 
 * indicated by the 'transmit buffer empty' bit in the LSR being set.
 */
        while ( ! (   tm_in_byte(TM_8250_PORT_ADDR+TM_8250_LSR)
                    & TM_8250_XMIT_EMPTY ) )
        {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
            tfInterfaceSpinLock(interfaceHandle);
        }
        
/* Send the current character to the 8250 */
        tm_out_byte(*dataPtr, TM_8250_PORT_ADDR + TM_8250_THR);

/* Move on to next character */
        dataLength--;
        dataPtr++;    
    }
    

/* Full packet has completed sending, inform the stack */
    if (flag & TM_USER_BUFFER_LAST)
    {
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }
    
    return(TM_DEV_OKAY);
}


/*
 * tf8250Receive
 *
 *   The 8250 serial chip can only return one character at a time, so for our
 *   recieve routine, we simply return a pointer to the next available 
 *   character in our recieve buffer.
 */                   
int tf8250Receive(ttUserInterface        interfaceHandle,
                  char TM_FAR * TM_FAR * dataPtr,
                  int  TM_FAR *          dataLength,
                  ttUserBufferPtr        bufHandlePtr)
{
    int errorCode;

    interfaceHandle=interfaceHandle;

    if ( (tl8250RecvBufferIndex != tl8250IsrBufferIndex) || 
                                                         (bufferFullFlag == 1) )
    {
        *dataPtr      = &(tl8250CharBuffer[tl8250RecvBufferIndex]);
        *dataLength   = 1;
        *bufHandlePtr = (ttUserBufferPtr) 0;

        errorCode = TM_DEV_OKAY;
    }
    else
    {
        errorCode = TM_DEV_ERROR;
    }

    return(errorCode);
}

/*
 * tf8250IsrHandler
 *
 *   Recieve interrupt handler.  Copies data from the 8250 to our recieve buffer
 *   and then notifies the stack that we have data available.
 *
 */
void TM_INTERRUPT tf8250IsrHandler(void)
{
    unsigned char tempData;

    if (bufferFullFlag == 0)
    {
        tl8250CharBuffer[tl8250IsrBufferIndex] = 
                                    tm_in_byte(TM_8250_PORT_ADDR + TM_8250_RHR);

        tl8250IsrBufferIndex = (tl8250IsrBufferIndex + 1) % TM_8250_BUFFER_SIZE;

/* The incoming data has filled up the recieve buffer. */
        if (tl8250IsrBufferIndex == tl8250RecvBufferIndex)
        {
            bufferFullFlag = 1;
        }

        tfNotifyReceiveInterfaceIsr(tl8250InterfaceHandle);
    }
    else
    {

/*
 * Even if we have no where to put the data, we still have to read it,
 * as this clears the interrupt and allows us to process future data.
 */
        tempData = tm_in_byte(TM_8250_PORT_ADDR + TM_8250_RHR);

/* Avoid compiler warning */
        TM_UNREF_IN_ARG(tempData);
    }

    tm_out_byte(0x20,0x20);

    return;
}


/*
 * tf8250FreeRecieveBuffer
 *
 *   Called when the stack is done with a recieved packet (character).  
 *   This should be shortly after it is passed up the stack, since both of the 
 *   serial link layers (PPP,SLIP) copy immediately.  This routine simply 
 *   increments the recieve index, which is used to ensure that the ISR doesn't 
 *   overwrite data that hasn't been freed.
 *
 */
int tf8250FreeReceiveBuffer(ttUserInterface interfaceHandle,
                            char TM_FAR *   dataPtr)
{
    interfaceHandle=interfaceHandle;
    dataPtr=dataPtr;

    tl8250RecvBufferIndex = (tl8250RecvBufferIndex + 1) % TM_8250_BUFFER_SIZE;

    tm_kernel_set_critical;
    bufferFullFlag = 0;
    tm_kernel_release_critical;

    return TM_DEV_OKAY;
}                            

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tr8250.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 31368    Output checksum: 43759
 */
