/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Intel 8254x Device Driver
 * Filename: tr8254x.c
 * Author: jshang
 * Date Created: 1/14/2003
 * $Source: source/drivers/8254x/tr8254x.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:50CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * There are a number of items in this driver that will need to be changed to
 * match your particular hardware platform:
 *
 *  1. Port I/O macros (tr8254x.h)
 *       In tr8254x.h there are a number of macros used for port I/O, such as
 *       tm_inl and tm_outl.  These should be modified to match your
 *       platform.
 *  2. configuration macros (ring size, etc)
 *       At the top of tr8254x.h, there are macros for various options
 *       related to this driver (transmit ring size, etc).
 *  3. VR_TO_BUS
 *       This macro translates a virtual (CPU) address to a physical (bus)
 *       address.  On many systems, this will do nothing.  On other systems 
 *       that utilize a MMU, this macro should be modified.
 *  4. tf8254xSetupPci()
 *       This sets up this PCI, retrieving its base addresses and interrupt 
 *       line information.  An example using a simple PCI implementation is
 *       included below.
 *  5. tf8254xSetupInterrupts()
 *       Setups the CPU or interupt controller to generate the appropriate
 *       interrupt.  
 */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include <tr8254x.h>
/* always enable dcache
 * notice: iq80321 dcache line size is 32 byte
 */

/* enable data cache */
#define TM_DCACHE_ENABLE

#ifdef TM_DCACHE_ENABLE
/* macro for dcache operations */
#include "includes.h"
#else /* !TM_DCACHE_ENABLE */
#define HAL_DCACHE_SYNC()
#define HAL_DCACHE_INVALIDATE(a,b)
#define HAL_DCACHE_STORE(a,b)

#endif /* TM_DCACHE_ENABLE */

/* this external variable is for Intel Test Demo to set/unset Offload, 
 * free to remove it and remove 485 lines
 */
extern int tv82544OffloadEnable;


/* test legacy TCP/UDP checksum offload */
/*#define TM_LEGACY_OFFLOAD*/



/* performance with Send Task is almost same as without Send Task */
#ifdef TM_TASK_SEND
int        tlNotifySend ;
#endif /* TM_TASK_SEND */

/* for IQ80321 board */
#define ioAddr      0xa0000000UL
/* general value is following */
/*#define ioAddr      (ttUser32Bit)infoPtr->i8254xMemAddrPtr*/


/* following header file and functions need to be
 * implemented based on target board 
 */

/* get mac address from eeprom */
void tf8254xGetMacAddr (ttUser32Bit, char *macAddr);

/* 82544 pci init in iq80321 board*/
int tf8254xSetupPci(unsigned char TM_FAR * intLine,
                    unsigned char TM_FAR * TM_FAR * memAddrPtr,
                    ttUser32Bit TM_FAR * tioAddr);

/* set up 8254x speed */
void tf8254xSetupSpeed(ttUser32Bit);

/* set ethernet ISR */
int tf8254xSetupInterrupts(unsigned char intLine, void (*fun)(void *),void *arg);


/* 8254x interrupt handler */
void tf8254xHandlerIsr(ttUserInterface);

/* packet receive checksum start byte */
#ifdef TM_DEV_RECV_OFFLOAD
#define TM_I82544_RECVXSUM_START   TM_ETHER_HEADER_SIZE
#endif /* TM_DEV_RECV_OFFLOAD */

#ifdef TM_DEV_SEND_OFFLOAD
/* setting hardware offload configuration */
static void tf8254xSetOffload (   ttUserInterface  interfaceHandle,
                                  ttUserPacketPtr  uPacketPtr );
#endif /* TM_DEV_SEND_OFFLOAD */

/* release buffer for completed sending packet */
static void tf8254xSendComplete(ttUserInterface interfaceHandle);

/* target board base address */
ttUser32Bit hal_pci_inbound_window_base;


/* 
 * tf8254xOpen()
 *  
 *      This routine initializes the 8254x chip.  This involves getting
 *   various PCI information, resetting and configuring the chip, allocating
 *   and initializing the data structures (info, transmit and receive rings, 
 *   etc) and installing the ISR handler.
 *
 *   1. Find where our device lives on the PCI bus.  For many embedded systems
 *      this would be unecessary, since this information is known ahead of
 *      time; the bus, device and func variables can be hard coded.
 *   2. Allocate memory for the information structure.  Initialize with
 *      necessary data - zero buffer pointers, fill in PCI related info 
 *      (interrupt line, BAR's, etc).
 *   3. Reset the 8254x .
 *   4. Set the base address for the Command and Receive Units (set to zero for
 *      pure linear addresses) and set the base address for the statistics 
 *      block.
 *   5. Initialize the transmit and receive buffer rings.
 *   6. Install the ISR handler.
 *   7. Start command processing.
 *
 */
int tf8254xOpen(ttUserInterface interfaceHandle)
{
    int                    errorCode;
    unsigned char          ringCount;
    unsigned char          intLine;
    tt8254xDescriptorPtr   frameDescPtr;
    ttRecvBufDescPtr       bufDescPtr;
    unsigned char TM_FAR * memAddrPtr;
    char                   phyAddr[TM_MAX_PHYS_ADDR];
    ttUser32Bit            tioAddr;
    ttUser32Bit            txctl;
    ttUser32Bit            rxctl;
    int                    i;    
    tt8254xInfoPtr         infoPtr;    
    
/* Get information about this PCI device (interrupt line, base addresses) */
    errorCode = tf8254xSetupPci(&intLine, &memAddrPtr, &tioAddr);
/* result is 0xa0000000 for IQ80321 */    
    hal_pci_inbound_window_base = 
        ((*((ttUser32Bit *)0xffffe120)) & CYG_PRI_CFG_BAR_MEM_MASK);
/* 
 * If a 8254x was found, create a structure for it and fill in some of the data
 */
    if (errorCode == TM_ENOERROR)
    {

        infoPtr = tfKernelMalloc(sizeof(tt8254xInfo)+31);
        if (infoPtr == (tt8254xInfoPtr) 0)
        {
/* Memory allocation failed: return with error */
            errorCode = TM_ENOBUFS;
            goto i8254xOpenEnd;
        }

        infoPtr = (tt8254xInfoPtr)(((tt32Bit)infoPtr + 31) & (~((tt32Bit)31)));
/* store the infoPtr */
        tfDeviceStorePointer(interfaceHandle, (ttVoidPtr)infoPtr);
        
        tfBzero(infoPtr, sizeof(tt8254xInfo));

        infoPtr->i8254xInterfaceHandle = interfaceHandle;

/* 8254x memory map base addr */
/* Set base addresses and interrupt line */
        infoPtr->i8254xMemAddrPtr = memAddrPtr;
        infoPtr->i8254xIoAddr     = tioAddr;
        infoPtr->i8254xIntLine    = intLine;

/* Initialize ring state variables */
        infoPtr->i8254xCmdRingTailIndex   = 0;
        infoPtr->i8254xCmdRingDirtyIndex  = 0;

        infoPtr->i8254xScatterIndex       = 0xff;
        infoPtr->i8254xRecvRingIndex      = 0;
        infoPtr->i8254xRecvRingDirtyIndex = TM_8254X_RECV_RING_SIZE-1;
        infoPtr->i8254xRecvRingIsrIndex   = 0;

#ifdef TM_DEV_RECV_OFFLOAD
        infoPtr->i8254xRecvCheckSumEnable = 0;
        infoPtr->i8254xRecvCheckSumCompute = 0;
#endif /* TM_DEV_RECV_OFFLOAD */

/* Issue software reset command to 8254x */
        rxctl  = tm_inl( ioAddr + TM_I82544_CTRL );
        tm_outl( rxctl | TM_I82544_CTRL_RST, ioAddr + TM_I82544_CTRL );
        rxctl = tm_inl( ioAddr + TM_I82544_CTRL );
/* setup speed */
        tf8254xSetupSpeed((ttUser32Bit)infoPtr->i8254xMemAddrPtr);

/* 
 * Initialize command/transmit ring 
 */
        for (ringCount=0; ringCount < TM_8254X_CMD_RING_SIZE; ringCount++) 
        {
            tm_writemem64( (tt32Bit)&infoPtr->i8254xCmdRing[ringCount] 
                            + TM_I82544_TD_BUFFER, 0 );
        }

/* Set the transmitter queue registers*/
        tm_outl( TM_VIRT_TO_BUS(infoPtr->i8254xCmdRing), ioAddr + TM_I82544_TDBAL );
        tm_outl( 0, ioAddr + TM_I82544_TDBAH );
        tm_outl( TM_8254X_CMD_RING_SIZE * TM_I82544_DESC_SIZE, ioAddr + TM_I82544_TDLEN );
        tm_outl( 0, ioAddr + TM_I82544_TDH );
        tm_outl( 0, ioAddr + TM_I82544_TDT );

/* Set IPG values*/
        tm_outl( 8 | (8<<10) | (6<<20), ioAddr + TM_I82544_TIPG );

/* Program tx ctrl register*/
        txctl = tm_inl( ioAddr + TM_I82544_TCTL );
/* collision threshold*/
        txctl |= (15<<4);   
/* collision distance*/
        txctl |= (64<<12);  
        txctl |= TM_I82544_TCTL_PSP;
        tm_outl( txctl, ioAddr + TM_I82544_TCTL );

/*
 * Initialize receive ring 
 */
        for (ringCount=0; ringCount < TM_8254X_RECV_RING_SIZE-1; ringCount++)
        {
            frameDescPtr = &(infoPtr->i8254xRecvDesc[ringCount]);
            bufDescPtr   = &(infoPtr->i8254xRecvBufDesc[ringCount]);
/* Allocate memory for this receive buffer, and set size in descriptor */    
            bufDescPtr->bufDataPtr = 
                                tfGetDriverBuffer(&(bufDescPtr->bufHandle),1600,16);
            if ( bufDescPtr->bufDataPtr == (char TM_FAR *)0 )
            {
                tfKernelError("tr8254xOpen", "can't allocate buffer");
                errorCode = TM_ENOBUFS;
                goto i8254xOpenEnd;
            }
/* dcache line size is 32 */
            bufDescPtr->bufDataPtr += TM_ETHER_IP_ALIGN+32; 

            tm_writemem64( (tt32Bit)frameDescPtr + TM_I82544_RD_BUFFER, 
                            TM_VIRT_TO_BUS(bufDescPtr->bufDataPtr) );
        }

/* 
 * For the first frame descriptor in the list (and only this one), link it to
 * the first buffer descriptor.
 */
/* Set the receiver queue registers*/
        tm_outl( TM_VIRT_TO_BUS(infoPtr->i8254xRecvDesc), ioAddr + TM_I82544_RDBAL );
        tm_outl( 0, ioAddr + TM_I82544_RDBAH );
        tm_outl( TM_8254X_RECV_RING_SIZE * TM_I82544_RD_SIZE, ioAddr + TM_I82544_RDLEN );
        tm_outl( 0, ioAddr + TM_I82544_RDH );
        tm_outl( TM_8254X_RECV_RING_SIZE-1, ioAddr + TM_I82544_RDT );
/* set receive delay timer as 0 */
        tm_outl( 0x80000000l, ioAddr + TM_I82544_RDTR);

/* zero out RAT*/
/* set chip Mac address in RAT */
        tf8254xGetPhysAddr(interfaceHandle,phyAddr);
        tm_outl((*(unsigned long*)phyAddr), ioAddr + TM_I82544_RAT);
        tm_outl( (0x80000000|(*(unsigned short *)&phyAddr[4])), 
                    ioAddr + TM_I82544_RAT + 4);
        for( i = 2; i < 32; i++ )
        {
            tm_outl( 0, ioAddr + TM_I82544_RAT +(i*4) );
        }

/* Zero out MTA*/
        for( i = 0; i < 128; i++ )
        {
            tm_outl( 0, ioAddr + TM_I82544_MTA +(i*4) );
        }
    
/* Set up receiver to accept broadcasts */
        rxctl = tm_inl( ioAddr + TM_I82544_RCTL );
        rxctl |=  ( TM_I82544_RCTL_BAM 
                    |TM_I82544_RCTL_MPE );
        tm_outl( rxctl, ioAddr + TM_I82544_RCTL );

        HAL_DCACHE_SYNC(); 
/* set ethernet ISR */
        tf8254xSetupInterrupts(infoPtr->i8254xIntLine,tf8254xHandlerIsr,
                                (void*)interfaceHandle);

/* enable receiver */
        rxctl = tm_inl( ioAddr + TM_I82544_RCTL );
        rxctl |= TM_I82544_RCTL_EN;
        tm_outl( rxctl, ioAddr + TM_I82544_RCTL );

/* enable transmitter */
        txctl = tm_inl( ioAddr + TM_I82544_TCTL );
        txctl |= TM_I82544_TCTL_EN;
        tm_outl( txctl, ioAddr + TM_I82544_TCTL );
    
        errorCode = TM_DEV_OKAY;
    }
    else
    {
/* Couldn't find the PCI device */
        errorCode = TM_DEV_ERROR;
    }
i8254xOpenEnd:
    return (errorCode);            
}

/*
 * tf8254xClose()
 *
 *   Closes the 8254x.  This routine should stop all 8254x operations and
 *   free all associated memory.
 *
 *   1. Disable interrupts from the 8254x.
 *   2. Stop the 8254x receive process.
 *   3. Free all transmit buffers.
 *   4. Free all receive buffers.
 */
int tf8254xClose(ttUserInterface interfaceHandle)
{
    unsigned char index;
    unsigned char dirtyIndex;
    ttUser32Bit rxctl, txctl;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);
    
/* Free all buffers in the transmit ring */
#ifndef TM_TASK_SEND
    tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */

/* Free all buffers in the receive ring */
    index      = infoPtr->i8254xRecvRingIndex;
    dirtyIndex = infoPtr->i8254xRecvRingDirtyIndex;

    while (index != dirtyIndex)
    {
        tfFreeDriverBuffer(infoPtr->i8254xRecvBufDesc[index].bufHandle);

        index = (unsigned char) ((index + 1) % TM_8254X_RECV_RING_SIZE);
    }

  
/* disable receiver*/
    rxctl = tm_inl( ioAddr + TM_I82544_RCTL );
    rxctl &= ~TM_I82544_RCTL_EN;
    tm_outl( rxctl, ioAddr + TM_I82544_RCTL );

/* Enable transmitter*/
    txctl = tm_inl( ioAddr + TM_I82544_TCTL );
    txctl &= ~TM_I82544_TCTL_EN;
    tm_outl( txctl, ioAddr + TM_I82544_TCTL );

    return(TM_DEV_OKAY);
}


/* 
 * tf8254xIoctl()
 *
 *   This routine performs four primary functions:  refilling the receive ring,
 *   and maintaining multicast addresses.  The refill portion simply walks
 *   through the recieve ring and allocates a new buffer for each buffer that
 *   is empty.  While doing this, it also moves the suspend bit, so that the
 *   chip will always stop receiving upon filling the next to last packet,
 *   thus preventing overruns.
 *   for send/recv offload functions, this function will return the offload
 *   functions of the chip to stack (GET), and set the offload functions by
 *   the command from stack (SET)
 */
int tf8254xIoctl(ttUserInterface interfaceHandle,
                 int             flag,
                 void TM_FAR *   optionPtr,
                 int             optionLen)
{
    int                     index;
    int                     dirtyIndex;
    ttUser32Bit             rxctl;
    ttRecvBufDescPtr        bufDescPtr;

#if defined(TM_DEV_SEND_OFFLOAD) || defined(TM_DEV_RECV_OFFLOAD)
    ttDevIoctlOffloadPtr    devoPtr;
#endif /* TM_DEV_SEND_OFFLOAD || TM_DEV_RECV_OFFLOAD */

    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);
    optionPtr=optionPtr;
    optionLen=optionLen;
/*
 * Refill our receive buffers.
 */
    if (flag & TM_8254X_IOCTL_REFILL)
    {
        index         = infoPtr->i8254xRecvRingIndex;
        dirtyIndex    = infoPtr->i8254xRecvRingDirtyIndex;

        while ( ((dirtyIndex +1 )% TM_8254X_RECV_RING_SIZE) != index )
        {
/* 
 * If this buffer has already been received by the Treck stack, 
 * allocate a new Ethernet buffer and place it in this slot in the 
 * receive ring.
 */
            if (infoPtr->i8254xRecvBufDesc[dirtyIndex].bufDataPtr == 0)
            {
                bufDescPtr = &(infoPtr->i8254xRecvBufDesc[dirtyIndex]);
            
                bufDescPtr->bufDataPtr = 
                                tfGetDriverBuffer(&(bufDescPtr->bufHandle),1600,16);

/* No memory available: break and hope we get some next time */ 
                if (bufDescPtr->bufDataPtr == (char TM_FAR *)0)
                {
                    tfKernelWarning("tf8254xIoctl","out of memory\n");
                    break;
                }
                bufDescPtr->bufDataPtr += TM_ETHER_IP_ALIGN+32; 
/* this invalidate must be called, otherwise must use recv_copy */
                HAL_DCACHE_INVALIDATE((tt32Bit)bufDescPtr->bufDataPtr, 1536);

                HAL_DCACHE_INVALIDATE(((tt32Bit)&infoPtr->i8254xRecvDesc[dirtyIndex]), 
                            TM_I82544_DESC_SIZE);
                tm_writemem64( (tt32Bit)(&infoPtr->i8254xRecvDesc[dirtyIndex]) 
                                + TM_I82544_RD_BUFFER, 
                            TM_VIRT_TO_BUS(bufDescPtr->bufDataPtr) );
                HAL_DCACHE_STORE(((tt32Bit)&infoPtr->i8254xRecvDesc[dirtyIndex]), 
                            TM_I82544_DESC_SIZE);
            }
            else
            {
/* 
 * This was a received error packet (ie, a packet that was not passed up the
 * stack) so rather than refilling it here (which would cause us to lose the
 * packet buffer) we simply skip over it.
 */
                tfKernelWarning("tf8254xIoctl","error in queue/bad packet\n");
            }
            dirtyIndex = 
                     (unsigned char) ((dirtyIndex+1) % TM_8254X_RECV_RING_SIZE);
/* update tail value */
        }

        tm_outl( dirtyIndex, 
                    (tt32Bit)infoPtr->i8254xMemAddrPtr + TM_I82544_RDT );

        infoPtr->i8254xRecvRingDirtyIndex = dirtyIndex;
    }


/*
 *  Perform send completes on any necessary buffers in the transmit ring.
 */
    if (flag & TM_8254X_IOCTL_SEND_COMPLETE)
    {
#ifndef TM_TASK_SEND
        tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */
    }

#if defined(TM_DEV_SEND_OFFLOAD) || defined(TM_DEV_RECV_OFFLOAD)
    if (flag & TM_DEV_IOCTL_OFFLOAD_GET)
    {
        if ( optionPtr != (void TM_FAR *)0 )
        {
            devoPtr = (ttDevIoctlOffloadPtr)optionPtr;
/* disable offload in order to compare to non-offload*/

            if (!tv82544OffloadEnable)
            {
                devoPtr->offloadFlags = 0;
                goto ioctlExit;
            }


            devoPtr->offloadFlags   = TM_DEVO_PACKET_RECV_CHKSUM_COMPUTE
                                     |TM_DEVO_IP_RECV_CHKSUM_VALIDATE
                                     |TM_DEVO_TCP_RECV_CHKSUM_VALIDATE
                                     |TM_DEVO_UDP_RECV_CHKSUM_VALIDATE
                                     |TM_DEVO_DITCH_BAD_CHECKSUM_PACKETS
                                     |TM_DEVO_TCP_CHKSUM_OFFLOAD
                                     |TM_DEVO_UDP_CHKSUM_OFFLOAD
#ifdef TM_LEGACY_OFFLOAD
                                     ;
#else /* !TM_LEGACY_OFFLOD */
                                     |TM_DEVO_IP_CHKSUM_OFFLOAD
                                     |TM_DEVO_TCP_SEGMENT_OFFLOAD_V4; 
#endif /* TM_LEGACY_OFFLOAD */
     
                                 
            devoPtr->offloadMaxSize = TM_I82544_MAX_PACKET_SIZE;
        }
    }

    if (flag & TM_DEV_IOCTL_OFFLOAD_SET)
    {
#ifdef TM_DEV_RECV_OFFLOAD
/* IP, TCP/UDP recv checksum validate */
        rxctl = tm_inl ( ioAddr + TM_I82544_RXCSUM );
        if ( (*(ttUser16Bit *)optionPtr & TM_DEVO_IP_RECV_CHKSUM_VALIDATE ))
        {
            rxctl |= TM_I82544_RXCSUM_IPOFL;
            infoPtr->i8254xRecvCheckSumEnable = 1;
        }
        if ( (*(ttUser16Bit *)optionPtr & (TM_DEVO_TCP_RECV_CHKSUM_VALIDATE
                          |TM_DEVO_UDP_RECV_CHKSUM_VALIDATE )))
        {
            rxctl |= TM_I82544_RXCSUM_TUOFL;
            infoPtr->i8254xRecvCheckSumEnable = 1;
        }
/* checksum start bit */
        rxctl |= TM_I82544_RECVXSUM_START;
        tm_outl ( rxctl, ioAddr + TM_I82544_RXCSUM);
/* store bad packet */        
        rxctl = tm_inl( ioAddr + TM_I82544_RCTL);
        if ( *(ttUser16Bit *)optionPtr & TM_DEVO_DITCH_BAD_CHECKSUM_PACKETS)
        {
            rxctl |= TM_I82544_RCTL_SBP;
        }
/* calculate packet checksum */        
        if ( *(ttUser16Bit *)optionPtr & TM_DEVO_PACKET_RECV_CHKSUM_COMPUTE)
        {
/* disable CRC checksum calculation  */
            rxctl |= TM_I82544_RCTL_SECRC;
            infoPtr->i8254xRecvCheckSumCompute = 1;
        }
        tm_outl ( rxctl, ioAddr + TM_I82544_RCTL);

#endif /* TM_DEV_RECV_OFFLOAD */
    }
#endif /* TM_DEV_SEND_OFFLOAD || TM_DEV_RECV_OFFLOAD */

ioctlExit:
    return (TM_DEV_OKAY);
}

/*
 * tf8254xSend()
 *
 *   This is a general sending packet function. The 82544 driver supports
 *   scatter sending, and suggest maximum scatter buffer number as 6 (may
 *   be more, if increases the transmit descripor number 
 *   1. Check to see if a transmit command descriptor is free, spin-locking
 *      if necessary.
 *   2. Attempt to send-complete any packets that have not yet been freed:
 *     2.1 Start off in the ring where we left off last time (dirty index).
 *     2.2 For each descriptor that is a Transmit command, call 
 *         tfSendCompleteInterface.
 *   3. Add the data pointer for this buffer in the appropriate spot in the 
 *      current descripor.
 *   4. If this is the last buffer in this packet (TM_USER_BUFFER_LAST flag 
 *      set), trigger the command, otherwise (TM_USER_BUFFER_MORE) store
 *      transmit buffer to the descripor
 *   
 */
int tf8254xSend(ttUserInterface interfaceHandle,
                char TM_FAR *   dataPtr,
                int             dataLength,
                int             flag)
{
    ttUser32Bit    txt,txh,dp,cmd;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);

    cmd = 0;

/* clear send complete */
#ifndef TM_TASK_SEND
    tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */

    if (flag & TM_USER_BUFFER_LAST)
    {
/* Set EOP bit on last frame */
        cmd |= TM_I82544_TD_CMD_EOP;
    }
    
    if ( infoPtr->i8254xScatterIndex == 0xff )
    {
/* there is not scatter buffer at before*/
        txt = infoPtr->i8254xCmdRingTailIndex;
    }
    else
    {
        txt = infoPtr->i8254xScatterIndex;
    }
/* if send queue is almost full */

    txh             = infoPtr->i8254xCmdRingDirtyIndex;

    while ( ((txh - txt +TM_8254X_CMD_RING_SIZE)%TM_8254X_CMD_RING_SIZE) 
                        < 5 )
    {
        if ( txh == txt ) 
        {
            break;
        }
        tfKernelWarning("tf82544xSend", " send buffer full\n");
/* spin lock */
        tfInterfaceSpinLock(interfaceHandle);
/*        txh = tm_inl ( ioAddr + TM_I82544_TDH); */ /* not safe */
#ifndef TM_TASK_SEND
        tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */
        txh = infoPtr->i8254xCmdRingDirtyIndex;
    }

    dp = ((tt32Bit)infoPtr->i8254xCmdRing + txt * TM_I82544_DESC_SIZE);

    HAL_DCACHE_STORE(((tt32Bit)dataPtr), dataLength);
    
    tm_writemem64( dp + TM_I82544_TD_BUFFER, 
        TM_VIRT_TO_BUS(((ttUser32Bit)dataPtr)));
    tm_writemem16( dp + TM_I82544_TD_LENGTH, dataLength );
    
       

/*  Get status back  */
    cmd |= TM_I82544_TD_CMD_IFCS| TM_I82544_TD_CMD_RS|TM_I82544_TD_CMD_RPS;
        
/*  Write command byte  */
    tm_writemem8( dp + TM_I82544_TD_CMD, cmd );

/* Zero out rest of fields */
    tm_writemem8( dp + TM_I82544_TD_STATUS, 0 );
    tm_writemem8( dp + TM_I82544_TD_CSO, 0 );
    tm_writemem8( dp + TM_I82544_TD_CSS, 0 );
    tm_writemem16( dp + TM_I82544_TD_SPECIAL, 0 );
 
    if (flag & TM_USER_BUFFER_LAST)
    {
        infoPtr->i8254xCmdRingStatus[txt] = 1;
    }

/* Increment tx tail pointer */
    txt = (txt + 1) % TM_8254X_CMD_RING_SIZE;

    if (flag & TM_USER_BUFFER_LAST)
    {
        HAL_DCACHE_STORE(((tt32Bit)&infoPtr->i8254xCmdRing[0]), 
                (TM_8254X_CMD_RING_SIZE*TM_I82544_DESC_SIZE));

/* And finally, cause the transmission to happen by setting the*/
/* tail pointer in the hardware.*/
        tm_outl( txt, ioAddr + TM_I82544_TDT );
        
        infoPtr->i8254xCmdRingTailIndex = txt;

        infoPtr->i8254xScatterIndex = 0xff;
    }
    else
    {
        infoPtr->i8254xScatterIndex = txt;
    }
    return(TM_DEV_OKAY);
}

/*
 * tf8254xScatDrvSend()
 *
 *   It is the packet transmit function with hardware offload (checksum and TCP
 *   segmentation, etc). 
 *   TM_USE_DRV_ONE_SCAT_SEND macro must be enable, and call 
 *   tfUseInterfaceOneScatSend() to load this driver, so that this
 *   transmit function have packet pointer parameter(uPacketPtr) 
 *   and all scatter frames in this packet will be sent once 
 *   ( if the chip does not support scatter send,
 *   need to set scatteredBufferCount as 1 in tf(Ng)OpenInterface).
 *
 *   1. Check to see if a transmit command descriptor is free, spin-locking
 *      if necessary.
 *   2. call tf8254xSetOffload to set/load offload parameters/registers
 *   3. Attempt to send-complete any packets that have not yet been freed:
 *     3.1 Start off in the ring where we left off last time (dirty index).
 *     3.2 For each descriptor that is a Transmit command, call 
 *         tfSendCompleteInterface.
 *   4. Add the data pointer for this buffer in the appropriate spot in the 
 *      current descripor.
 *   5. send
 */

#ifdef TM_USE_DRV_ONE_SCAT_SEND 
int tf8254xScatDrvSend (    ttUserInterface  interfaceHandle,
                            ttUserPacketPtr  uPacketPtr )
{
    ttUser32Bit    txt,txh,dp,cmd;
    ttUser32Bit    dataPtr;
    ttUser32Bit    value32;
    int              dataLen;
    int              scatterIndex;

    int              firstFrame;

#ifdef TM_LEGACY_OFFLOAD
    unsigned char    cmdCache;    
    ttDevSendOffloadPtr devoPtr;
#endif /* TM_LEGACY_OFFLOAD */

    unsigned char    cssCache;    
    unsigned char    csoCache;    

    int             startIndex;
    int             lengthIndex;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);

/* clear send complete */
    
#ifndef TM_TASK_SEND
    tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */

#if defined(TM_DEV_SEND_OFFLOAD) || defined(TM_DEV_RECV_OFFLOAD)
/* set offload behaviour using uPacketPtr */

    if((ttDevSendOffloadPtr)uPacketPtr->pktuDevOffloadPtr ==0)
    {
        infoPtr->i8254xTcpContextEnable  = 0;
        infoPtr->i8254xCacheDevoFlags    = 0;

    }
    else
    {
        if (!((infoPtr->i8254xCacheDevoFlags == 
            ((ttDevSendOffloadPtr)uPacketPtr->pktuDevOffloadPtr)->devoFlags) 
        && ((!tm_tcp_send_seg_offload(((ttDevSendOffloadPtr)
                uPacketPtr->pktuDevOffloadPtr)->devoFlags)))))
        {
            tf8254xSetOffload (interfaceHandle, uPacketPtr);
        }  
    }
        
#endif /* TM_DEV_SEND_OFFLOAD || TM_DEV_RECV_OFFLOAD */

    txt             = infoPtr->i8254xCmdRingTailIndex;

    txh             = infoPtr->i8254xCmdRingDirtyIndex;
    scatterIndex    = uPacketPtr->pktuLinkExtraCount; 

    firstFrame      = 1;
#ifdef TM_LEGACY_OFFLOAD
    cmdCache        = 0;
#endif /* TM_LEGACY_OFFLOAD */
    csoCache        = 0;
    cssCache        = 0;


/* if send queue is not enough */
    while ( ((TM_8254X_CMD_RING_SIZE + txh - txt -scatterIndex)
                %TM_8254X_CMD_RING_SIZE) < 5 )
    {
        if ( txh == txt ) 
        {
            break;
        }
        tfKernelWarning("tf8254xScatDrvSend", " send buffer full\n");
/* spin lock */
        tfInterfaceSpinLock(interfaceHandle);
/*        txh = tm_inl ( ioAddr + TM_I82544_TDH); */ /* not safe */
#ifndef TM_TASK_SEND
        tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */
        txh = infoPtr->i8254xCmdRingDirtyIndex;
    }

    startIndex = txt;
    lengthIndex = scatterIndex+1;

    while ( scatterIndex-- >= 0 )
    {
        cmd = 0;

        if (scatterIndex < 0 )
        {
/* Set EOP bit on last packet */
            cmd |= TM_I82544_TD_CMD_EOP|TM_I82544_TD_CMD_RPS;
        }
/*  Insert IFCS   */
        cmd |= TM_I82544_TD_CMD_IFCS;
        dp = ((tt32Bit)&infoPtr->i8254xCmdRing[txt]);
        dataPtr = (ttUser32Bit)uPacketPtr->pktuLinkDataPtr;
        dataLen = uPacketPtr->pktuLinkDataLength;


        if (( firstFrame ) && (txt%2))
        {
            HAL_DCACHE_INVALIDATE(dp, TM_I82544_DESC_SIZE);
        }

        HAL_DCACHE_STORE(((tt32Bit)dataPtr), dataLen);

        tm_writemem64( dp + TM_I82544_TD_BUFFER, TM_VIRT_TO_BUS(dataPtr) );
        
#ifdef TM_DEV_SEND_OFFLOAD
        if ( infoPtr->i8254xTcpContextEnable)
        {
            tm_writemem8( dp + TM_I82544_TD_DAT_POPTS, 
                            infoPtr->i8254xCachePopts);
            value32 = ((dataLen&0xfffff) | ((tt32Bit)0x0001<<20) 
                    | ((tt32Bit)(cmd|TM_I82544_TD_CMD_DEXT
                            |infoPtr->i8254xCacheDcmd))<<24);
            tm_writemem32( dp + TM_I82544_TD_DAT_DTALEN, value32);
            tm_writemem8(dp + TM_I82544_TD_DAT_STA, 0);
            tm_writemem16(dp + TM_I82544_TD_DAT_SPECIAL, 0);
        }
        else
#endif /* TM_DEV_SEND_OFFLOAD */
        {
            tm_writemem16( dp + TM_I82544_TD_LENGTH, dataLen );
/* Zero out rest of fields */
            tm_writemem8( dp + TM_I82544_TD_STATUS, 0 );

#if defined(TM_LEGACY_OFFLOAD) && defined (TM_DEV_SEND_OFFLOAD)
            devoPtr = (ttDevSendOffloadPtr)uPacketPtr->pktuDevOffloadPtr;
            if ((devoPtr != (ttDevSendOffloadPtr)0) && firstFrame)
            {
                if ( devoPtr->devoFlags)
                {
/* legacy checksum offload mode */
                    csoCache = devoPtr->devoTucso;
                    cssCache = devoPtr->devoTucss;
                    cmdCache = TM_I82544_TD_CMD_IC;
                }
            }
#endif /* defined(TM_LEGACY_OFFLOAD) && defined (TM_DEV_SEND_OFFLOAD) */
            tm_writemem8( dp + TM_I82544_TD_CSO, csoCache );
            tm_writemem8( dp + TM_I82544_TD_CSS, cssCache );

#ifdef TM_LEGACY_OFFLOAD
            cmd |= cmdCache;
#endif /* TM_LEGACY_OFFLOAD */
/*  Write command byte  */
            tm_writemem8( dp + TM_I82544_TD_CMD, cmd );
/*  Write special byte  */
            tm_writemem16( dp + TM_I82544_TD_SPECIAL, 0 );
        }
        
        if ( scatterIndex < 0)
        {
            infoPtr->i8254xCmdRingStatus[txt] = 1;
        }

/* Increment tx tail pointer */
        txt         = (txt + 1) % TM_8254X_CMD_RING_SIZE;
        uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 

        firstFrame  = 0;

    }

/* Finally, cause the transmission to happen by setting the
 * tail pointer in the hardware dcahe sync 
 */

    if ( txt % 2 )
    {
        tm_writemem8( (tt32Bit)&infoPtr->i8254xCmdRing[txt] 
                            + TM_I82544_TD_CON_TUCMD, 0);
    }

    if ( lengthIndex + startIndex <= TM_8254X_CMD_RING_SIZE )
    {
        HAL_DCACHE_STORE(((tt32Bit)&infoPtr->i8254xCmdRing[startIndex] ), 
            (lengthIndex * TM_I82544_DESC_SIZE));
    }
    else
    {
        HAL_DCACHE_STORE(((tt32Bit)&infoPtr->i8254xCmdRing[startIndex] ), 
            ((TM_8254X_CMD_RING_SIZE-startIndex) * TM_I82544_DESC_SIZE));
        HAL_DCACHE_STORE(((tt32Bit)&infoPtr->i8254xCmdRing[0] ), 
            ((lengthIndex + startIndex-TM_8254X_CMD_RING_SIZE) 
                    * TM_I82544_DESC_SIZE));
    }

    tm_outl( txt, ioAddr + TM_I82544_TDT );

    infoPtr->i8254xCmdRingTailIndex = txt;

    return(TM_DEV_OKAY);
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

/*
 * tf8254xSetOffload()
 *
 *   The function is setting and checking offload parameter
 *
 *   TM_USE_DRV_ONE_SCAT_SEND macro must be enable, and call 
 *   tfUseInterfaceOneScatSend() to load this driver, so that this
 *   sending function will have packet pointer parameter and all frames in this 
 *   packet will be sent once ( if the chip does not support scatter send,
 *   need to set scatteredBufferCount as 1 in tf(Ng)OpenInterface).
 *
 *   1. Check to see if a transmit command descriptor is free, spin-locking
 *      if necessary.
 *   2. call tf8254xSetOffload to set/load offload parameters/registers
 *   3. Attempt to send-complete any packets that have not yet been freed:
 *     3.1 Start off in the ring where we left off last time (dirty index).
 *     3.2 For each descriptor that is a Transmit command, call 
 *         tfSendCompleteInterface.
 *   4. Add the data pointer for this buffer in the appropriate spot in the 
 *      current descripor.
 *   5. send
 */

#ifdef TM_DEV_SEND_OFFLOAD
static void tf8254xSetOffload (  ttUserInterface  interfaceHandle,
                                 ttUserPacketPtr  uPacketPtr )
{
    ttDevSendOffloadPtr devoPtr;   
    ttUser32Bit       dp,txt,txh;
    unsigned char       value8;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);
 
    devoPtr = (ttDevSendOffloadPtr)uPacketPtr->pktuDevOffloadPtr;

    txt = infoPtr->i8254xCmdRingTailIndex ;
    txh = infoPtr->i8254xCmdRingDirtyIndex;

/* if send queue is not enough */
    while ( ((TM_8254X_CMD_RING_SIZE + txh - txt)
                %TM_8254X_CMD_RING_SIZE) < 5 )
    {
        if ( txh == txt ) 
        {
            break;
        }
        tfKernelWarning("tf8254xSetOffload", " send buffer full\n");
/* spin lock */
        tfInterfaceSpinLock(interfaceHandle);
/*        txh = tm_inl ( ioAddr + TM_I82544_TDH); */ /* not safe */
#ifndef TM_TASK_SEND
        tf8254xSendComplete(interfaceHandle);
#endif /* TM_TASK_SEND */
        txh = infoPtr->i8254xCmdRingDirtyIndex;
    }

    dp = ((tt32Bit)&infoPtr->i8254xCmdRing[txt] );

    if (txt%2)
    {
        HAL_DCACHE_INVALIDATE(dp, TM_I82544_DESC_SIZE);
    }

/* offload cache clear */
    infoPtr->i8254xCacheDcmd = 0;
    infoPtr->i8254xCachePopts = 0;
        
    if ( devoPtr->devoFlags != 0 )
    {
        infoPtr->i8254xTcpContextEnable = 1;
    }
 
    if ( devoPtr->devoFlags & TM_DEVO_IP_CHKSUM_OFFLOAD )
    {   
/* set IPCSS, IPCSO, IPCSE */ 
        tm_writemem8( dp + TM_I82544_TD_CON_IPCSS, devoPtr->devoIpcss );
        tm_writemem8( dp + TM_I82544_TD_CON_IPCSO, devoPtr->devoIpcso );
        tm_writemem16( dp + TM_I82544_TD_CON_IPCSE,devoPtr->devoIpcse );
        infoPtr->i8254xCachePopts |= TM_I82544_TD_DAT_POPTS_IXSM;
    }
    if ( devoPtr->devoFlags & 
            (TM_DEVO_TCP_CHKSUM_OFFLOAD|TM_DEVO_UDP_CHKSUM_OFFLOAD
                ))
    {
/* set TUCSS, TUCSO, TUCSE */
        tm_writemem8( dp + TM_I82544_TD_CON_TUCSS, devoPtr->devoTucss );
        tm_writemem8( dp + TM_I82544_TD_CON_TUCSO, devoPtr->devoTucso );
        tm_writemem16( dp + TM_I82544_TD_CON_TUCSE,devoPtr->devoTucse );
        infoPtr->i8254xCachePopts |= TM_I82544_TD_DAT_POPTS_TXSM;
    }
/* set DTYP */
    tm_writemem32(  dp + TM_I82544_TD_CON_PAYLEN, 0);
/* set cached DCMD, CMD for data context descriptor */    
      
    value8 = TM_I82544_TD_CON_TUCMD_DEXT
                |TM_I82544_TD_CON_TUCMD_RS;
    
    if ( devoPtr->devoFlags & TM_DEVO_TCP_SEGMENT_OFFLOAD_V4 )
    {
/* set MSS, HDRLEN, PAYLEN, STA */
        tm_writemem16(  dp + TM_I82544_TD_CON_MSS,  
                            devoPtr->devoFramePayload);
        tm_writemem8(   dp + TM_I82544_TD_CON_HDRLEN,
                            devoPtr->devoFrameHdrLen ); 
        tm_writemem32(  dp + TM_I82544_TD_CON_PAYLEN, 
                            (((ttDevSendOffloadPtr)uPacketPtr
                                    ->pktuDevOffloadPtr)->devoPayload 
                                    & 0xfffff));
        
        value8 |= TM_I82544_TD_CON_TUCMD_IP
                    |TM_I82544_TD_CON_TUCMD_TCP
                    |TM_I82544_TD_CON_TUCMD_TSE;

        infoPtr->i8254xCacheDcmd = TM_I82544_TD_DAT_DCMD_TSE;
        infoPtr->i8254xCachePopts = TM_I82544_TD_DAT_POPTS_IXSM
                                         |TM_I82544_TD_DAT_POPTS_TXSM;
    }
    tm_writemem8( dp + TM_I82544_TD_CON_STA, 0);
    tm_writemem8( dp + TM_I82544_TD_CON_TUCMD, value8);
    

/* load new context */    
    txt = (txt + 1) % TM_8254X_CMD_RING_SIZE;
/* dcahe sync */
    if ( txt % 2 )
    {
        tm_writemem8( (tt32Bit)&infoPtr->i8254xCmdRing[txt] 
                            + TM_I82544_TD_CON_TUCMD, 0);
    }
    HAL_DCACHE_STORE(dp, TM_I82544_DESC_SIZE);

    tm_outl( txt, ioAddr + TM_I82544_TDT );

    infoPtr->i8254xCmdRingTailIndex = txt;

/* cached new offload config */
    infoPtr->i8254xCacheDevoFlags     = devoPtr->devoFlags;
    infoPtr->i8254xCacheChainLength   = uPacketPtr->pktuChainDataLength; 

    return;
}
#endif /* TM_DEV_SEND_OFFLOAD */

/*
 * tf8254xReceive()
 *
 *      The 8254x uses a fairly straightforward receive mechanism, with the
 *   addition that it can handle gather reads.  However, this feature is not
 *   utilized at this time.  Like the transmit/action command, receive 
 *   descriptors are organized into a linked list, which in our case is 
 *   circular.  This list is composed of RFD's, Received Frame Descriptors. 
 *   These descriptors contain information (such as status) about the frame, and
 *   can contain data poiter.   
 *   Notice: devoRPtr defined for receive checksum offload, the devoRPtr pointer
 *   passed from stack will record the checksum checking flag( pass/fail of TCP/IP
 *   packet) and packet checksum value.
 *   1. Check current packet descripor (status) is ready 
 *   2. Set buffer/data pointer, and checksum offload pointer.
 *   3. Mark the frame receive datapointer as 0 (unavailable)
 */
int tf8254xReceive(ttUserInterface        interfaceHandle,
                   char TM_FAR * TM_FAR * dataPtr,
                   int  TM_FAR *          dataLength,
                   ttUserBufferPtr        bufHandlePtr
#ifdef TM_DEV_RECV_OFFLOAD
                 , ttDevRecvOffloadPtr    devoRPtr
#endif /* TM_DEV_RECV_OFFLOAD */
                   )
{
    ttRecvBufDescPtr   bufDescPtr;
    tt8254xDescriptorPtr frameDescPtr;
    unsigned short     status;
    int                errorCode;

    unsigned short     length;
    int                index;
    unsigned char      error;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);

/* receive queue check */
    if ( ((infoPtr->i8254xRecvRingDirtyIndex -
              infoPtr->i8254xRecvRingIsrIndex +
               TM_8254X_RECV_RING_SIZE)%TM_8254X_RECV_RING_SIZE)
                  < 20 )
    {
        if ( infoPtr->i8254xRecvRingIndex 
            != infoPtr->i8254xRecvRingIsrIndex )
        {
            if ( ((infoPtr->i8254xRecvRingIndex  -
              infoPtr->i8254xRecvRingIsrIndex +
               TM_8254X_RECV_RING_SIZE)%TM_8254X_RECV_RING_SIZE)
                  < 5 )
            {
                tfKernelWarning("tf8254xScatDrvSend", " recv buffer full\n");
            }

            tf8254xIoctl(interfaceHandle, TM_8254X_IOCTL_REFILL, 0, 0); 
        }
    }

    interfaceHandle=interfaceHandle;

/* 
 * By default (in case of an error) return an error, and a data pointer of
 * zero.
 */
    errorCode = TM_DEV_ERROR;

    *dataPtr      = 0;
    *dataLength   = 0;
    *bufHandlePtr = 0;

    index = infoPtr->i8254xRecvRingIndex;
    bufDescPtr = &(infoPtr->i8254xRecvBufDesc[index]);
    frameDescPtr = &(infoPtr->i8254xRecvDesc[index]);

    HAL_DCACHE_INVALIDATE(((tt32Bit)frameDescPtr), TM_I82544_DESC_SIZE);

    tm_readmem8( (tt32Bit)frameDescPtr + TM_I82544_RD_STATUS, status );
/* now assume there is no scatter buffers received */
    if(( status & TM_I82544_RD_STATUS_EOP )
        && (status & TM_I82544_RD_STATUS_DD))
    {
        tm_readmem16( (tt32Bit)frameDescPtr  
                            + TM_I82544_RD_LENGTH, length );
        if ((length <= TM_MAX_RX_PACKET_SIZE) 
                && ( bufDescPtr->bufDataPtr != (char TM_FAR *)0))
        {

/* store the cache line memory that is not in recv buffer */
            HAL_DCACHE_INVALIDATE(((tt32Bit)bufDescPtr->bufDataPtr) , length);

            *dataPtr      = bufDescPtr->bufDataPtr;
            *dataLength   = length;
            *bufHandlePtr = bufDescPtr->bufHandle;
            
            
#ifdef TM_DEV_RECV_OFFLOAD
            devoRPtr->devoRFlags = 0;
/* if not setting IXSM, check IPCS/IPE, TCPCS/TCPE */
            if (!(status & TM_I82544_RD_STATUS_IXSM))
            {
                if ( infoPtr->i8254xRecvCheckSumEnable )
                {
                    tm_readmem8( (tt32Bit)frameDescPtr 
                                + TM_I82544_RD_ERRORS, error);
                    if (status & TM_I82544_RD_STATUS_IPCS) 
                    {
                        if (error & TM_I82544_RD_ERRORS_IPE)
                        {
                            devoRPtr->devoRFlags |= 
                                TM_DEVOR_IP_RECV_CHKSUM_FAILED;
                        }
                        else
                        {
                            devoRPtr->devoRFlags |= 
                                TM_DEVOR_IP_RECV_CHKSUM_VALIDATED;
                        }
                    }
                    if (status & TM_I82544_RD_STATUS_TCPCS) 
                    {
                        if (error & TM_I82544_RD_ERRORS_TCPE)
                        {
                            devoRPtr->devoRFlags |=TM_DEVOR_TCP_RECV_CHKSUM_FAILED
                                            |TM_DEVOR_UDP_RECV_CHKSUM_FAILED;
                        }
                        else
                        {
                            devoRPtr->devoRFlags |=
                                TM_DEVOR_TCP_RECV_CHKSUM_VALIDATED 
                               |TM_DEVOR_UDP_RECV_CHKSUM_VALIDATED ;
                        }
                    }
                }
            }
            if ( infoPtr->i8254xRecvCheckSumCompute )
            {
                devoRPtr->devoRFlags |= TM_DEVOR_PACKET_RECV_CHKSUM_COMPUTED;
                tm_readmem16((tt32Bit)frameDescPtr + TM_I82544_RD_PKTXSUM, 
                         devoRPtr->devoRecvXsum); 
/* transform to network order */
                devoRPtr->devoRecvXsum = htons(devoRPtr->devoRecvXsum); 
                devoRPtr->devoRecvXsumLength = length 
                                            - TM_I82544_RECVXSUM_START;
                devoRPtr->devoRecvXsumOffset = TM_I82544_RECVXSUM_START;
            }
#endif /* TM_DEV_RECV_OFFLOAD */
/* clear zero in status */
            
/* Mark this buffer as unavailable to the chip and in need of a refill */
            bufDescPtr->bufDataPtr = 0;

            if ( index%2 )
            {
                tm_writemem64( (tt32Bit)frameDescPtr + TM_I82544_RD_BUFFER, 
                            TM_VIRT_TO_BUS(0) );
                tm_writemem64( (tt32Bit)frameDescPtr - TM_I82544_DESC_SIZE + TM_I82544_RD_BUFFER, 
                            TM_VIRT_TO_BUS(0) );

                tm_writemem8( (tt32Bit)frameDescPtr + TM_I82544_RD_STATUS, 0 );
                tm_writemem8( (tt32Bit)frameDescPtr - TM_I82544_DESC_SIZE + TM_I82544_RD_STATUS, 0 );
                HAL_DCACHE_STORE((tt32Bit)frameDescPtr, TM_I82544_DESC_SIZE);
            }


/* Move to the next buffer in the ring */
            infoPtr->i8254xRecvRingIndex = 
                        (int) ((index+1) % TM_8254X_RECV_RING_SIZE);
            errorCode = TM_DEV_OKAY;
        }
        else
        {
            tfKernelWarning("tf82544Receive","too big length\n");
        }
    }
    else
    {
/* 
 * ERROR!  We were interrupted because of a received packet, but there doesn't
 * seem to be anything here for us...
 */
        tfKernelWarning("tf82544Receive","no eop packet received\n");
    }

    return(errorCode);
}

/*
 * tf8254xGetPhysAddr()
 *
 *   Returns the physical address of this Ethernet interface.  This is just
 *   sample data and this routine should be modified to return real data.
 *
 */


int tf8254xGetPhysAddr(ttUserInterface interfaceHandle,
                       char TM_FAR *   physicalAddress)
{    
    static unsigned char macAddr[] = { 0x00,0x07,0xe9,0x03,0x39,0x1b};
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);
/* automatically get mac address from eeprom, right now is same as above
 * initial values
 */
    tf8254xGetMacAddr((ttUser32Bit)infoPtr->i8254xMemAddrPtr,macAddr);

    physicalAddress[0] = macAddr[0];
    physicalAddress[1] = macAddr[1];
    physicalAddress[2] = macAddr[2];
    physicalAddress[3] = macAddr[3];
    physicalAddress[4] = macAddr[4];
    physicalAddress[5] = macAddr[5];

    return(TM_DEV_OKAY);


}


/*
 * tf8254xHandlerIsr()
 *
 *   The only interrupt that we should handle here is a received frame 
 *   interrupt.  When this interrupt is received, we loop through the recieve
 *   ring and count how many packets have been received since the last 
 *   interrupt.  At the very end of the routine, we notify on all of these
 *   packets.  It is important for this to be done at the end of ISR 
 *   (please see Device Driver FAQ in Treck manual).
 */
void tf8254xHandlerIsr(ttUserInterface interfaceHandle)
{
    unsigned char  status;
    int            notifyCount;
    int            rxp,rxd; 
    ttUser32Bit  dp;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);

    notifyCount = 0;

/* clear all pending interrupt */
    dp = tm_inl( ioAddr + TM_I82544_ICR );
#ifdef TM_TASK_SEND
    if ( dp&0x1 )
    {
/* send complete */
        tf8254xSendComplete(infoPtr->i8254xInterfaceHandle);
    }
#endif /* TM_TASK_SEND */
    if ( dp & 0x80)
    {
        rxp = infoPtr->i8254xRecvRingIsrIndex; 
    
/* invalidate from RecvRingIsrIndex to RecvRingDirtyIndex */
        rxd = infoPtr->i8254xRecvRingDirtyIndex;

/* invalidate the recv descriptor without recvIndex or dirtyIndex */
        if ( rxp < rxd)
        {
            HAL_DCACHE_INVALIDATE(((tt32Bit)&infoPtr->i8254xRecvDesc[rxp]),
                (TM_I82544_DESC_SIZE*(((int)(rxd/2))*2-rxp)));
        }
        else
        {
            HAL_DCACHE_INVALIDATE(((tt32Bit)&infoPtr->i8254xRecvDesc[rxp]),
                (TM_I82544_DESC_SIZE*(TM_8254X_RECV_RING_SIZE-rxp)));
            if ( rxd > 1 )
            {
                HAL_DCACHE_INVALIDATE(((tt32Bit)&infoPtr->i8254xRecvDesc[0]),
                    (TM_I82544_DESC_SIZE*((int)(rxd/2))*2));
            }
        }

        while ( 1 )
        {
/* now assume there is not scatter buffer received */
/*  Get status*/

            tm_readmem8( (tt32Bit)&infoPtr->i8254xRecvDesc[rxp] 
                            + TM_I82544_RD_STATUS, status );
            if(( status & TM_I82544_RD_STATUS_EOP )
                && (status & TM_I82544_RD_STATUS_DD)
                && (infoPtr->i8254xRecvBufDesc[rxp].bufDataPtr !=0))
            {
                notifyCount++;
            }
            else
            {   
                break;
            }

/* Advance rxp pointer*/
            rxp = ( rxp + 1 ) % TM_8254X_RECV_RING_SIZE;
        }

        infoPtr->i8254xRecvRingIsrIndex = rxp;
    }
    tfNotifyInterfaceIsr( infoPtr->i8254xInterfaceHandle, 
                          notifyCount,
#ifdef TM_TASK_SEND
                          tlNotifySend, 2500,
#else  /* !TM_TASK_SEND */
                          0, 0UL, 
#endif /* TM_TASK_SEND */
                          0UL );
#ifdef TM_TASK_SEND
    tlNotifySend = 0;
#endif /* TM_TASK_SEND */

    return;
}

/*
 * tf8254xSendComplete()
 *
 * Free all buffer already sent.
 *
 */

static void tf8254xSendComplete(ttUserInterface interfaceHandle)
{
    int             dirtyIndex;
    unsigned char   value8;
    tt8254xInfoPtr         infoPtr;    
    infoPtr = tfDeviceGetPointer(interfaceHandle);

#ifdef TM_TASK_SEND
    tlNotifySend    = 0;
#endif /* TM_TASK_SEND */
    dirtyIndex = infoPtr->i8254xCmdRingDirtyIndex;


    while (dirtyIndex !=infoPtr->i8254xCmdRingTailIndex)
    {
#ifdef TM_TASK_SEND
        if ( ((dirtyIndex%2) == 0)
            &&(((infoPtr->i8254xCmdRingTailIndex - dirtyIndex 
                +TM_8254X_CMD_RING_SIZE)%TM_8254X_CMD_RING_SIZE) < 5))
        {
            break;
        }
#endif /* TM_TASK_SEND */
        if (infoPtr->i8254xCmdRingStatus[dirtyIndex] == 1)
        {
            HAL_DCACHE_INVALIDATE((
                (tt32Bit)&infoPtr->i8254xCmdRing[dirtyIndex]), 
                                    TM_I82544_DESC_SIZE);
            tm_readmem8((tt32Bit)&infoPtr->i8254xCmdRing[dirtyIndex] 
                        + TM_I82544_TD_CON_STA,value8);
            if ((value8&0x01) != 0 )
            {
                infoPtr->i8254xCmdRingStatus[dirtyIndex] = 0;

#ifdef TM_TASK_SEND
                tlNotifySend++;
#else   /*! TM_TASK_SEND */
                tfSendCompleteInterface(interfaceHandle,
                                    TM_DEV_SEND_COMPLETE_DRIVER);
#endif /*! TM_TASK_SEND */
                dirtyIndex = (int) ((dirtyIndex+1) 
                                        % TM_8254X_CMD_RING_SIZE);

            }
            else
            {
                break;
            }
        }
        else
        {
            dirtyIndex = (int) ((dirtyIndex+1) 
                                        % TM_8254X_CMD_RING_SIZE);
        }
    }
    infoPtr->i8254xCmdRingDirtyIndex = dirtyIndex;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tr8254x.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 46447    Output checksum: 21903
 */
