/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Intel 8255x Device Driver
 * Filename: tr8255x.c
 * Author: Jason
 * Date Created: 3/14/2000
 * $Source: source/drivers/8255x/tr8255x.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:51CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* There are a number of items in this driver that will need to be changed to
 * match your particular hardware platform:
 *
 *  1. Port I/O macros (tr8255x.h)
 *       In tr8255x.h there are a number of macros used for port I/O, such as
 *       tm_out_byte and tm_in_word.  These should be modified to match your
 *       platform (the included ones are for the Microsoft x86 C compiler).
 *  2. configuration macros (ring size, etc)
 *       At the top of tr8255x.h, there are macros for various options
 *       related to this driver (transmit ring size, etc).
 *  3. tm_virt_addr_to_phys
 *       This macro translates a virtual (CPU) address to a physical (bus)
 *       address.  On many systems, this will do nothing.  On other systems 
 *       that utilize a MMU, this macro should be modified.
 *  4. tm_clear_cpu_interrupts
 *       Clears the interrupt from the CPU (or interrupt controller) after
 *       an ISR has completed.  Should be implemented for your system - an
 *       example one for a 8259 PIC is included below.
 *  5. tf8255xSetupPci()
 *       This sets up this PCI, retrieving its base addresses and interrupt 
 *       line information.  An example using a simple PCI implementation is
 *       included below.
 *  6. tf8255xSetupInterrupts()
 *       Setups the CPU or interupt controller to generate the appropriate
 *       interrupt.  The routine below implements this for the 8259 PIC.
 *  7. tm_asm_invd 
 *       This macros invalidates the CPU's data cache.  This is necessary for
 *       cache coherency, since the 8255x uses DMA to transfer data.  
 *       Depending on your system, there are much more elegant ways to handle
 *       this than to invalidate the entire cache (Memory Write w/ Invalidate,
 *       invalidating of specific cache lines, etc).
 */

/*
 * Define when using the IQ80310 XScale evaluation board.
 * NOTE: When using GCC, this file should be compiled with the -O0 option
 *       to disable all optimizations, as there seems to be at least one
 *       optimization bug in GCC that causes this driver to fail.
 */
/* #define TM_XSCALE_IQ80310 */

/* For _outp, _inp, etc (Microsoft C compiler) */
#ifdef _MSC_VER
#include <conio.h>
#endif /* MSC_VER */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include <tr8255x.h>

#ifdef TM_XSCALE_IQ80310
#include <iq80310.h> 
#include <pci_bios.h>
#else /* !TM_XSCALE_IQ80310 */
#include <trpci.h>
#endif /* TM_XSCALE_IQ80310 */

/* 8255x interrupt handler */
void tf8255xHandlerIsr(void);

/* Actual (external) ISR routine that will call tf8255xHandlerIsr() */
#ifndef TM_XSCALE_IQ80310
extern void tf8255xIsr(void);
#endif /* TM_XSCALE_IQ80310 */

static void tf8255xSendComplete(ttUserInterface interfaceHandle);

/*
 * Data structure containing information about this interface 
 */
tt8255xInfoPtr tlInfoPtr;

/*
 * tf8255xSetupPci()
 *
 *   This routine finds the 8255x Ethernet device on the PCI bus, and
 *   returns the necessary information (interrupt line, base memory address
 *   and/or base I/O address).  Currently, this uses the simple Treck PCI
 *   module (source/drivers/pci/trpci.c) to obtain this information.  This
 *   routine should be changed per your environment, and can be as simple as
 *   returning hard coded values.
 *
 */
int tf8255xSetupPci(unsigned char TM_FAR * intLine,
                    unsigned char TM_FAR * TM_FAR * memAddrPtr,
                    unsigned long TM_FAR * ioAddr)
{
    unsigned char  pciIntLine;
    unsigned char  pciBus;
    unsigned char  pciFunc;
    unsigned char  pciDevice;
    unsigned long  tempLong;
    unsigned long  pciIoAddr;
    unsigned short pciCommand;
    int            errorCode;
     
#ifdef TM_XSCALE_IQ80310
/* Initialize the PCI buses */
    PCI_DATA            pnew_data;
    PCI_DEVICE_LOCATION devid;

    sys_pci_device_initialization(&pnew_data);
  

/* Get PCI bus, device and function number for the 8255x */

    errorCode = sys_find_pci_device(TM_PCI_ID_INTEL,
                                    TM_PCI_ID_INTEL_82559,
                                    0,
                                    &devid);
    
    if (errorCode == OK)
    {
        pciBus    = devid.bus_number;
        pciDevice = devid.device_number;
        pciFunc   = devid.device_number;
   
/* Get the I/O and memory base addresses for the 8255x */
        sys_read_config_dword(
            pciBus, pciDevice, pciFunc, REGION0_BASE_OFFSET, &tempLong);
        
        *memAddrPtr = (unsigned char TM_FAR *) tempLong;
                        
        sys_read_config_dword(
            pciBus, pciDevice, pciFunc, REGION1_BASE_OFFSET, &pciIoAddr);

/* Mask out the I/O space indication bit */
        pciIoAddr &= ~1;

        *ioAddr = pciIoAddr;

/* 
 * Enables the device to act as a bus master, and enable both the I/O space
 * and memory decoders. (These options are set above by the 
 * #define TM_PCI_MODE_CMD and will probably have to be changed per platform)
 */
        sys_read_config_word(pciBus,pciDevice,pciFunc,STATUS_OFFSET,
                              &pciCommand);
        pciCommand |= TM_PCI_MODE_CMD;      

        sys_write_config_word(pciBus,pciDevice,pciFunc,COMMAND_OFFSET,
                               &pciCommand);
        errorCode = TM_DEV_OKAY;
    }
    else
    {
        errorCode = TM_DEV_ERROR;
    }
    
#else /* TM_XSCALE_IQ80310 */
/* Get PCI bus, device and function number for the 8255x */
 
     errorCode = tfPciFindDevice(TM_PCI_ID_INTEL,TM_PCI_ID_INTEL_82559,
                                &pciBus, &pciDevice, &pciFunc);

    if (errorCode == TM_DEV_OKAY)
    {
/* Get the interrupt line to use */
        tm_pci_read_conf_byte(pciBus,pciDevice,pciFunc,TM_PCI_INT_LINE, 
                              pciIntLine);
        *intLine = pciIntLine;

/* Get the I/O and memory base addresses for the 8255x */
        tm_pci_read_conf_dword(pciBus,pciDevice,pciFunc,TM_PCI_BASE_ADDR_0,
                               tempLong);
        *memAddrPtr = (unsigned char TM_FAR *) tempLong;
                        
        tm_pci_read_conf_dword(pciBus,pciDevice,pciFunc,TM_PCI_BASE_ADDR_1,
                               pciIoAddr);

/* Mask out the I/O space indication bit */
        pciIoAddr &= ~1;

        *ioAddr = pciIoAddr;

/* 
 * Enables the device to act as a bus master, and enable both the I/O space
 * and memory decoders. (These options are set above by the 
 * #define TM_PCI_MODE_CMD and will probably have to be changed per platform)
 */
        tm_pci_read_conf_word(pciBus,pciDevice,pciFunc,TM_PCI_STAT_CMD_REG,
                              pciCommand);
        pciCommand |= TM_PCI_MODE_CMD;      

        tm_pci_write_conf_word(pciBus,pciDevice,pciFunc,TM_PCI_STAT_CMD_REG,
                               pciCommand);

    }
#endif /* TM_XSCALE_IQ80310 */
    
    return errorCode;
}

/*
 * tf8255xSetupInterrupts()
 *
 *   This routine sets up the ISR for the 8255x.  This is obviously very
 *   platform specific, and will need to be changed for your environment.
 *   This driver was developed on a PC platform and therefore utilizes the
 *   8259 PIC, with all IRQ vectors linearly set up starting at address 0x20.
 *   This routine needs to install the ISR, and enable any external hardware
 *   necessary to catch this interrupt;  tf8255xOpen handles setting up and
 *   enabling interrupts on the 8255x side of things.  
 *
 */
int tf8255xSetupInterrupts(unsigned char intLine)
{
    
#ifdef TM_XSCALE_IQ80310

/* Install and enable the Ethernet interrupt handler. */
    tfKernelInstallIsrHandler(tf8255xHandlerIsr, ENET_INT_ID);

#else /* !TM_XSCALE_IQ80310 */
   
    unsigned char inByte;
                              
/* 
 * Setup and install the ISR 
 */
    tfKernelInstallIsrHandler(tf8255xIsr,0x20 + intLine);

/* Enable interrupt in PIC */
    if (intLine < 8)
    {
        inByte = (unsigned char) _inp(TM_PIC1_IMR_ADDR);
        inByte &= ~( 1 << intLine );
        _outp( TM_PIC1_IMR_ADDR, inByte );
    }
    else
    {
        inByte = (unsigned char) _inp(TM_PIC2_IMR_ADDR);
        inByte &= ~( 1 << (intLine-8) );
        _outp( TM_PIC2_IMR_ADDR, inByte );
    }
#endif /* TM_XSCALE_IQ80310 */
    
    return TM_DEV_OKAY;
}     


/* 
 * tf8255xOpen()
 *  
 *      This routine initializes the 8255x chip.  This involves getting
 *   various PCI information, resetting and configuring the chip, allocating
 *   and initializing the data structures (info, transmit and receive rings, 
 *   etc) and installing the ISR handler.
 *
 *   1. Find where our device lives on the PCI bus.  For many embedded systems
 *      this would be unecessary, since this information is known ahead of
 *      time; the bus, device and func variables can be hard coded.
 *   2. Allocate memory for the information structure.  Initialize with
 *      necessary data - zero buffer pointers, fill in PCI related info 
 *      (interrupt line, BAR's, etc).
 *   3. Reset the 8255x, and wait 10 us (per databook).
 *   4. Set the base address for the Command and Receive Units (set to zero for
 *      pure linear addresses) and set the base address for the statistics 
 *      block.
 *   5. Initialize the transmit and receive buffer rings.
 *   6. Install the ISR handler.
 *   7. Setup our individual Ethernet address.
 *   8. Start command processing.
 *
 */
int tf8255xOpen(ttUserInterface interfaceHandle)
{
    int                    errorCode;
    int                    delayCount;
    unsigned char          ringCount;
    unsigned char          inByte;
    unsigned char          intLine;
    char                   physAddr[6];
    ttIndAddrSetupPtr      indAddrCmdPtr;
    ttRecvFrameDescPtr     frameDescPtr;
    ttRecvBufDescPtr       bufDescPtr;
    unsigned char TM_FAR * memAddrPtr;
    unsigned long          ioAddr;

    if (tfInterfaceScatterCount(interfaceHandle) > TM_8255X_MAX_SCATTER)
    {
        errorCode = TM_DEV_ERROR;
        goto i8255xOpenEnd;        
    }
    
/* Get information about this PCI device (interrupt line, base addresses) */
    errorCode = tf8255xSetupPci(&intLine, &memAddrPtr, &ioAddr);

/* 
 * If an 8255x was found, create a structure for it and fill in some of the data
 */
    if (errorCode == TM_ENOERROR)
    {

        tlInfoPtr = tfKernelMalloc(sizeof(tt8255xInfo));

        if (tlInfoPtr == (tt8255xInfoPtr) 0)
        {
/* Memory allocation failed: return with error */
            errorCode = TM_ENOBUFS;
            goto i8255xOpenEnd;
        }

        tfBzero(tlInfoPtr, sizeof(tt8255xInfo));

        tlInfoPtr->i8255xInterfaceHandle = interfaceHandle;

        tlInfoPtr->i8255xXmitQueueEnb = tfInterfaceXmitQueueOn(interfaceHandle);
        
/* Set base addresses and interrupt line */
        tlInfoPtr->i8255xMemAddrPtr = memAddrPtr;
        tlInfoPtr->i8255xIoAddr     = ioAddr;
        tlInfoPtr->i8255xIntLine    = intLine;

/* Initialize ring state variables */
        tlInfoPtr->i8255xCmdRingIndex       = 0;
        tlInfoPtr->i8255xCmdRingDirtyIndex  = 0;
        tlInfoPtr->i8255xRecvRingIndex      = 0;
        tlInfoPtr->i8255xRecvRingDirtyIndex = 0;
        tlInfoPtr->i8255xRecvRingIsrIndex   = 0;

/* Initialize misc other state information */
        tlInfoPtr->i8255xSpuriousInterrupts = 0;

/* Issue software reset command to 8255x */
        tm_out_dword(TM_SCB_PORT_CMD_RESET, TM_SCB_PORT);

/*
 * We should delay for 10us (the actual delay time is ten system clocks
 * and five transmit clocks, but the databook says that 10 us is a
 * conservative estamite, and the extra time doesn't hurt us since this
 * is just on initialization.
 */
        for (delayCount=0; delayCount<2500; delayCount++)
        {
        ;
        }

/* Unmask all interrupt types, but don't enable 8255x->CPU interrupts yet 
 * (0x0100 -> 2)
 */
        tm_out_byte(TM_SCB_CMD_DISABLE_INT, TM_SCB_INT_MASK);

/* Set the CU and RU base segment to 0 (pure linear addressing mode) */
        inByte = (unsigned char) tm_in_byte(TM_SCB_COMMAND);
        tm_set_scb_gen_addr(0UL);
        tm_issue_scb_cmd_wait(TM_SCB_CMD_LOAD_CU_BASE);
        tm_issue_scb_cmd_wait(TM_SCB_CMD_LOAD_RU_BASE);

/*
 * Intialize the lowel upper driver
 */
#ifdef TM_XSCALE_IQ80310
        tf80310FindLinkSpeed();
#endif /* TM_XSCALE_IQ80310 */

/* 
 * Initialize command/transmit ring 
 */
        for (ringCount=0; ringCount < TM_8255X_CMD_RING_SIZE; ringCount++) 
        {

/* Set all commands to complete, since none are being held by the chip */
            tlInfoPtr->i8255xCmdRing[ringCount].genStatus = TM_CMD_COMPLETE;

/* Set up circular link list of action commands */
            tlInfoPtr->i8255xCmdRing[ringCount].genLinkPtr = 
                tm_virt_addr_to_phys(&(tlInfoPtr->i8255xCmdRing[(ringCount+1) % 
                TM_8255X_CMD_RING_SIZE]));

        }

        tm_asm_invd;

/*
 * Initialize receive ring 
 */
        for (ringCount=0; ringCount < TM_8255X_RECV_RING_SIZE; ringCount++)
        {
            frameDescPtr = &(tlInfoPtr->i8255xRecvDesc[ringCount]);
            bufDescPtr   = &(tlInfoPtr->i8255xRecvBufDesc[ringCount]);

/* Amount of data currently in buffer = 0 */
            bufDescPtr->bufCount = 0;       

/* Link current buffer to next buffer in list */
            bufDescPtr->bufNextPtr = tm_virt_addr_to_phys(&(tlInfoPtr->
                                               i8255xRecvBufDesc[ringCount+1]));

/* Allocate memory for this receive buffer, and set size in descriptor */    
            bufDescPtr->bufDataVirtPtr = 
                                  tfGetEthernetBuffer(&(bufDescPtr->bufHandle));
            bufDescPtr->bufDataPtr = 
                               tm_virt_addr_to_phys(bufDescPtr->bufDataVirtPtr);
            bufDescPtr->bufSize = tm_flip_dword(1518);

/* Set frame to flexible mode */
            frameDescPtr->recvStatus = TM_RECV_FLEX;

/* Link current frame to next in list */
            frameDescPtr->recvLinkPtr = 
                tm_virt_addr_to_phys(&(tlInfoPtr->i8255xRecvDesc[
                    (ringCount+1) % TM_8255X_RECV_RING_SIZE]));

/* 
 * No data immediately following this frame descriptor (all data is stored in
 * external buffer).
 */
            frameDescPtr->recvBufDescPtr = (ttRecvBufDescPtr) 0xffffffff;
            frameDescPtr->recvCount      = 0;
            frameDescPtr->recvSize       = 0;
        }

/* 
 * For the first frame descriptor in the list (and only this one), link it to
 * the first buffer descriptor.
 */
        tlInfoPtr->i8255xRecvDesc[0].recvBufDescPtr = 
            tm_virt_addr_to_phys(&(tlInfoPtr->i8255xRecvBufDesc[0]));

/* 
 * Turn both the frame list and buffer list into circular linked lists by 
 * connecting the tail to the head.
 */
        tlInfoPtr->i8255xRecvBufDesc[TM_8255X_RECV_RING_SIZE-1].bufNextPtr = 
            tm_virt_addr_to_phys(&(tlInfoPtr->i8255xRecvBufDesc[0]));
        tlInfoPtr->i8255xRecvDesc[TM_8255X_RECV_RING_SIZE-1].recvLinkPtr = 
            tm_virt_addr_to_phys(&(tlInfoPtr->i8255xRecvDesc[0]));

/* 
 * Set the 8255x to suspend receive operation after receiving the next to last
 * packet.  This forces the 8255x to stop before it overruns buffers that
 * have not yet been refilled.
 */
        tm_8255x_suspend(tlInfoPtr->i8255xRecvDesc[TM_8255X_RECV_RING_SIZE-2]);

        tm_asm_invd;

/* Setup ethernet address */
        tf8255xGetPhysAddr(interfaceHandle, physAddr);
        indAddrCmdPtr = (ttIndAddrSetupPtr) 
                     &(tlInfoPtr->i8255xCmdRing[tlInfoPtr->i8255xCmdRingIndex]);
        tm_inc_cmd_index(tlInfoPtr);
        indAddrCmdPtr->indStatus = TM_CMD_SUSPEND | TM_CMD_IND_ADDR;
        indAddrCmdPtr->indAddr[0] = *((unsigned short *)&physAddr[0]);
        indAddrCmdPtr->indAddr[1] = *((unsigned short *)&physAddr[2]);
        indAddrCmdPtr->indAddr[2] = *((unsigned short *)&physAddr[4]);

        tm_asm_invd;

        errorCode = tf8255xSetupInterrupts(tlInfoPtr->i8255xIntLine);
/* 
 * Unmask appropriate interrupts, set address of command (transmit) ring and
 * start command processing.
 */
        tm_set_scb_gen_addr(
            tm_virt_addr_to_phys((unsigned long)&tlInfoPtr->i8255xCmdRing[0]));
        tm_issue_scb_cmd_wait(TM_SCB_CMD_CU_START);

        tm_out_byte(TM_SCB_CMD_ENABLE_INT, TM_SCB_INT_MASK);
    
/* Load address of receive ring pointer into register */
        tm_set_scb_gen_addr( tm_virt_addr_to_phys( (unsigned long)
                             tlInfoPtr->i8255xRecvDesc ) );
        tm_issue_scb_cmd_wait(TM_SCB_CMD_RU_START);

        errorCode = TM_DEV_OKAY;
    }
    else
    {
/* Couldn't find the PCI device */
        errorCode = TM_DEV_ERROR;
    }

i8255xOpenEnd:
    return (errorCode);            
}

/*
 * tf8255xClose()
 *
 *   Closes the 8255x.  This routine should stop all 8255x operations and
 *   free all associated memory.
 *
 *   1. Disable interrupts from the 8255x.
 *   2. Stop the 8255x receive process.
 *   3. Free all transmit buffers.
 *   4. Free all receive buffers.
 */
int tf8255xClose(ttUserInterface interfaceHandle)
{
    unsigned char index;
    unsigned char dirtyIndex;

 
/* Disable interrupts from the 8255x */
    tm_out_byte(TM_SCB_CMD_DISABLE_INT, TM_SCB_INT_MASK);

/* 
 * Stop the chip's receive process (we don't have to worry about the send
 * process, since by this point the stack will have stopped sending data).
 */
    tm_issue_scb_cmd_wait(TM_SCB_CMD_RU_ABORT);

/* Free all buffers in the transmit ring */
    tf8255xSendComplete(interfaceHandle);

/* Free all buffers in the receive ring */
    index      = tlInfoPtr->i8255xRecvRingIndex;
    dirtyIndex = tlInfoPtr->i8255xRecvRingDirtyIndex;

    while (index != dirtyIndex)
    {
        tfFreeDriverBuffer(tlInfoPtr->i8255xRecvBufDesc[index].bufHandle);

        index = (unsigned char) ((index + 1) % TM_8255X_RECV_RING_SIZE);
    }


    return(TM_DEV_OKAY);
}


/* 
 * tf8255xIoctl()
 *
 *   This routine performs two primary functions:  refilling the receive ring,
 *   and maintaining multicast addresses.  The refill portion simply walks
 *   through the recieve ring and allocates a new buffer for each buffer that
 *   is empty.  While doing this, it also moves the suspend bit, so that the
 *   chip will always stop receiving upon filling the next to last packet,
 *   thus preventing overruns.
 */
int tf8255xIoctl(ttUserInterface interfaceHandle,
                 int             flag,
                 void TM_FAR *   optionPtr,
                 int             optionLen)
{
    unsigned char           index;
    unsigned char           dirtyIndex;
    unsigned char           oldDirtyIndex;
    ttRecvBufDescPtr        bufDescPtr;

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    int                     errorCode;
    ttMultiAddrSetupPtr     multiCmdPtr;
    unsigned short TM_FAR * mcastAddrPtr;
#endif /* TM_IGMP or TM_6_USE_MLD */

    optionPtr=optionPtr;
    optionLen=optionLen;

/*
 * Refill our receive buffers.
 */
    if (flag & TM_8255X_IOCTL_REFILL)
    {
        index         = tlInfoPtr->i8255xRecvRingIndex;
        dirtyIndex    = tlInfoPtr->i8255xRecvRingDirtyIndex;
        oldDirtyIndex = dirtyIndex;

        while ( dirtyIndex != index )
        {

/* 
 * If this buffer has already been received by the Treck stack, 
 * allocate a new Ethernet buffer and place it in this slot in the 
 * receive ring.
 */
            if (tlInfoPtr->i8255xRecvBufDesc[dirtyIndex].bufDataPtr == 0)
            {
            
                bufDescPtr = &(tlInfoPtr->i8255xRecvBufDesc[dirtyIndex]);
            
                bufDescPtr->bufDataVirtPtr = 
                                  tfGetEthernetBuffer(&(bufDescPtr->bufHandle));
                bufDescPtr->bufDataPtr = 
                               tm_virt_addr_to_phys(bufDescPtr->bufDataVirtPtr);

/* No memory available: break and hope we get some next time */ 
                if (bufDescPtr->bufDataPtr == (char TM_FAR *)0)
                {
                    break;
                }


            }
            else
            {
/* 
 * This was a received error packet (ie, a packet that was not passed up the
 * stack) so rather than refilling it here (which would cause us to lose the
 * packet buffer) we simply skip over it.
 */
            }

            dirtyIndex = 
                     (unsigned char) ((dirtyIndex+1) % TM_8255X_RECV_RING_SIZE);
            

        }

/* 
 * Set the new suspend bit.  Its important to do this before we unset the old
 * one, otherwise there's a chance that the chip will have the opportunity to
 * overrun buffers.  This way, the worse that will happen is that the chip
 * will not be able to receive, even though free buffers are available.
 */
        if (dirtyIndex > 1)
        {
            tm_8255x_suspend(tlInfoPtr->i8255xRecvDesc[dirtyIndex-2]);
        }
        else
        {
            tm_8255x_suspend(
                tlInfoPtr->i8255xRecvDesc[TM_8255X_RECV_RING_SIZE-2+dirtyIndex]);
        }

/* Unset the old suspend bit.  See above. */
        if (dirtyIndex != oldDirtyIndex)
        {
            if (oldDirtyIndex > 1)
            {
                tm_8255x_unsuspend(tlInfoPtr->i8255xRecvDesc[oldDirtyIndex-2]);
            }
            else
            {
                tm_8255x_unsuspend(
                    tlInfoPtr->i8255xRecvDesc[
                        TM_8255X_RECV_RING_SIZE-2+oldDirtyIndex]);
            }
        }

        tlInfoPtr->i8255xRecvRingDirtyIndex = dirtyIndex;

        if (dirtyIndex != oldDirtyIndex)
        {
/* 
 * If we moved the suspend bit (meaning that the 8255x was in a suspended state
 * coming in here) we resume the chip's receive process.
 */
            unsigned char status;

            /* get the status of the receiver */
            status = tm_in_byte(TM_SCB_STATUS) & 0x3c;

            /* send the resume command only if the receiver is suspended */
            if (status == 0x4)
            {
                tm_issue_scb_cmd_wait(TM_SCB_CMD_RU_RESUME);

            }
        }
    }

/*
 *  Perform send completes on any necessary buffers in the transmit ring.
 */
    if (flag & TM_8255X_IOCTL_SEND_COMPLETE)
    {
        tf8255xSendComplete(interfaceHandle);
    }

    
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))

/*
 * Setup our list of multicast addresses to receive.
 */
    if ( (flag & TM_DEV_SET_MCAST_LIST) || (flag & TM_DEV_SET_ALL_MCAST) )
    {

/* 
 * We are in the middle of a scattered send, so we can't add a command buffer
 * to our command/transmit ring.  There's nothing we can do here, so we just 
 * have to return and let the IGMP error deal with it.
 */
        if (tlInfoPtr->i8255xScatterIndex != 0)
        {
            errorCode = TM_DEV_ERROR;
        }
        else
        {

/* Wait for a transmit/command buffer to become available */
            index = tlInfoPtr->i8255xCmdRingIndex;
            while (!(tlInfoPtr->i8255xCmdRing[index].genStatus & 
                                                              TM_CMD_COMPLETE))
            {
                tm_asm_invd;
            }

/* 
 * Go ahead and send complete on the transmit ring, to ensure that the next
 * buffer is available (ie, emulate the behavior of the send routine)
 */
            tf8255xSendComplete(interfaceHandle);

            tm_inc_cmd_index(tlInfoPtr);

            multiCmdPtr = 
                      (ttMultiAddrSetupPtr) &(tlInfoPtr->i8255xCmdRing[index]);

/* Multicast address setup command, suspend on completion */
            multiCmdPtr->multiStatus = TM_CMD_MULTI_ADDR | TM_CMD_SUSPEND;


            if ( (flag & TM_DEV_SET_MCAST_LIST) && (optionLen == 0) )
            {
/* Clear multicast address list (receive NO multicast packets) */
                multiCmdPtr->multiByteCount = 0;
            }
            else
            {

/* Set multicast list (recieve ALL multicast packets) */
                multiCmdPtr->multiByteCount = tm_flip_word(6);

                mcastAddrPtr = multiCmdPtr->multiAddr;

                *mcastAddrPtr++ = 0xFFFF;
                *mcastAddrPtr++ = 0xFFFF;
                *mcastAddrPtr++ = 0xFFFF;

            } 
            
/*
 * We need to make sure that only the last packet has its suspend bit set, so
 * unset the suspend bit from the previous command.
 */
            if (index > 0)
            {
                index--;
            }
            else
            {
                index = TM_8255X_CMD_RING_SIZE - 1;
            }

            tlInfoPtr->i8255xCmdRing[index].genStatus &= ~TM_CMD_SUSPEND;

            tm_asm_invd;
            tm_issue_scb_cmd_wait(TM_SCB_CMD_CU_RESUME); 
   
        }
    }

#endif /* TM_IGMP or TM_6_USE_MLD */

    return (TM_DEV_OKAY);
}

/*
 * tf8255xSend()
 *
 *      The 8255x handles transmission differently than other Ethernet chips.
 *   A transmit command is handled through the a normal action command (see
 *   top of driver for description of action commands).  This action command
 *   contains a pointer to a TBD (Transmit Buffer Descriptor) that is an 
 *   additional structure, containing any number of pointers to actual data.
 *   Currently, these TBD's are statically allocated, one TBD per command
 *   descriptor.  Each TBD contains room for 5 buffer pointers.  Therefore, the
 *   maximum capacity of this method is 25 packets, with each packet consisting
 *   of 5 scattered buffers.  In theory, it would be possible to dynamically
 *   allocate TBD's to make more efficient use of memory.  However, this is
 *   troublesome since we don't know the total number of buffers that will be
 *   sent to us from the Treck stack, only the maximum.  
 *      The 82558/82559 have an additional mode that allows two TBD entries to
 *   immediately follow the Transmit command descriptor, alleviating the need
 *   for seperate TBD's.  This feature is not currently utilized.
 *
 *   1. Check to see if a transmit command descriptor is free, spin-locking
 *      if necessary.
 *   2. Attempt to send-complete any packets that have not yet been freed:
 *     2.1 Start off in the ring where we left off last time (dirty index).
 *     2.2 Loop until we reach an active descriptor (i.e., on that the chip is
 *         not yet done with, indicated by the TM_CMD_COMPLETE_BIT) or we run
 *         out of 'dirty' entries. 
 *     2.3 For each descriptor that is a Transmit command, call 
 *         tfSendCompleteInterface.
 *   3. Add the data pointer for this buffer in the appropriate spot in the 
 *      current TBD.
 *   4. If this is the last buffer in this packet (TM_USER_BUFFER_LAST flag 
 *      set), fill out Transmit command structure and trigger the command
 *      unit to resume (ie, send the packet).
 *   
 */
int tf8255xSend(ttUserInterface interfaceHandle,
                char TM_FAR *   dataPtr,
                int             dataLength,
                int             flag)
{
    unsigned char    index;
    unsigned char    dirtyIndex;
    unsigned char    tbdIndex;
    ttTransmitCmdPtr xmitCmdPtr;
    unsigned long    spinLockCounter;
    int              errorCode;

    errorCode = TM_DEV_OKAY;
    

    index = tlInfoPtr->i8255xCmdRingIndex;
    tm_asm_invd;

/* Wait for free transmit descriptor */    
    if (!(tlInfoPtr->i8255xCmdRing[index].genStatus & TM_CMD_COMPLETE))
    {
        
/* No transmit descriptor is immediately available: */
        if (tlInfoPtr->i8255xXmitQueueEnb)
        {
/* Transmit queue is enabled - return an error to the stack, the current
   frame will be queued. */
            errorCode = TM_DEV_ERROR;
            goto i8255xEnd;
        }
        else
        {
/* No transmit queue - just spin until a transmit descriptor is available. */
            spinLockCounter=0;
            while (!(tlInfoPtr->i8255xCmdRing[index].genStatus
                     & TM_CMD_COMPLETE))
            {
                tm_asm_invd;
                spinLockCounter++;
                tfInterfaceSpinLock(interfaceHandle);
            }
        }
    }

/* Send complete on any necessary packets */
    dirtyIndex = tlInfoPtr->i8255xCmdRingDirtyIndex;
    while ( (dirtyIndex != index) && 
            (tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus & TM_CMD_COMPLETE))
    {

/* 
 * If this buffer is a transmit buffer and has completed being processed,
 * call send complete to notify the Treck stack that it's OK to free this 
 * packet.
 */
        if (tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus & TM_CMD_TRANSMIT)
        {

            tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus = TM_CMD_COMPLETE;

            tfSendCompleteInterface(interfaceHandle,
                                    TM_DEV_SEND_COMPLETE_DRIVER);
        }

        dirtyIndex = (unsigned char) ((dirtyIndex+1) % TM_8255X_CMD_RING_SIZE);
    }
    tlInfoPtr->i8255xCmdRingDirtyIndex = dirtyIndex;


/* Add current buffer to the current TBD */
    tbdIndex = tlInfoPtr->i8255xScatterIndex;
    tlInfoPtr->i8255xXmitBufDesc[index][tbdIndex].xmitDataPtr = 
                                                  tm_virt_addr_to_phys(dataPtr);
    tlInfoPtr->i8255xXmitBufDesc[index][tbdIndex].xmitSize = 
                                                      tm_flip_dword(dataLength);

/* Move to next TBD */
    tlInfoPtr->i8255xScatterIndex++;

    if (flag & TM_USER_BUFFER_LAST)
    {
        tm_inc_cmd_index(tlInfoPtr);

        xmitCmdPtr = (ttTransmitCmdPtr) &(tlInfoPtr->i8255xCmdRing[index]);

/* Transmit command, flexible mode, suspend on completion */
        xmitCmdPtr->xmitStatus = TM_CMD_TRANSMIT | TM_CMD_FLEX_XMIT | 
                                 TM_CMD_SUSPEND;                    

        xmitCmdPtr->xmitTbdPtr = 
                      tm_virt_addr_to_phys(tlInfoPtr->i8255xXmitBufDesc[index]);

/* All data in TBD's, no data at end of command structure */
        xmitCmdPtr->xmitByteCount = 0;

/* Set number of TBD's and Transmit FIFO threshold */
        xmitCmdPtr->xmitTbdNum = tm_flip_word( (unsigned short)
            ((tlInfoPtr->i8255xScatterIndex << 8) |
            (TM_8255X_XMIT_THRESHOLD >> 3)));

/* 
 * We need to make sure that only the last packet has its suspend bit set, so
 * unset the suspend bit from the previous command.
 */

        if (index > 0)
        {
            index--;
        }
        else
        {
            index = TM_8255X_CMD_RING_SIZE-1;
        }

        tlInfoPtr->i8255xCmdRing[index].genStatus &= ~TM_CMD_SUSPEND;


/* Reset the scatter index since we're done with this packet */        
        tlInfoPtr->i8255xScatterIndex = 0;

        tm_asm_invd;
        tm_issue_scb_cmd_wait(TM_SCB_CMD_CU_RESUME);        

    }

i8255xEnd:    
    return(errorCode);
}

/*
 * tf8255xReceive()
 *
 *      The 8255x uses a fairly straightforward receive mechanism, with the
 *   addition that it can handle gather reads.  However, this feature is not
 *   utilized at this time.  Like the transmit/action command, receive 
 *   descriptors are organized into a linked list, which in our case is 
 *   circular.  This list is composed of RFD's, Received Frame Descriptors. 
 *   These descriptors contain information (such as status) about the frame, and
 *   can optionally contain data (though in our case it doesn't).  Most 
 *   importantly they contain links to RBD's (Received Buffer Descriptors)
 *   which is a list of buffers that form this packet.  In our case, this list 
 *   just includes a single buffer. This buffering mechanism is extremely 
 *   similar to the transmit mechanism described above.
 *
 *   1. Check current packet for the 'complete' flag being set; if not set
 *      we got called from an interrupt without there being data available.
 *      Just break from this routine.
 *   2. Mark this frame as processed by unsetting the 'complete' flag.
 *   3. If this frame is OK, mark it unavailable to the chip (set the RBD 
 *      pointer in the current RFD to NULL) and pass the packet up the stack.
 *      Move receive index to next descriptor.
 *   4. If there was an out of memory error, refill some buffers here.
 *   5. If there was any sort of other error, just ignore it and move on to the
 *      next buffer.  Our refill routine will clean this up later.
 *      Move receive index to next descriptor.
 */
int tf8255xReceive(ttUserInterface        interfaceHandle,
                   char TM_FAR * TM_FAR * dataPtr,
                   int  TM_FAR *          dataLength,
                   ttUserBufferPtr        bufHandlePtr)
{
    unsigned char      index;
    ttRecvBufDescPtr   bufDescPtr;
    ttRecvFrameDescPtr frameDescPtr;
    unsigned short     status;
    int                errorCode;
    unsigned long *    tempDataPtr;
    ttPacketPtr        packetPtr;

    interfaceHandle=interfaceHandle;

    tm_asm_invd;

    index = tlInfoPtr->i8255xRecvRingIndex;

/* 
 * By default (in case of an error) return and error, and a data pointer of
 * zero.
 */
    errorCode = TM_DEV_ERROR;

    *dataPtr      = 0;
    *dataLength   = 0;
    *bufHandlePtr = 0;

    tm_asm_invd;
    if (tlInfoPtr->i8255xRecvDesc[index].recvStatus & TM_RECV_COMPLETE)
    {

        bufDescPtr = &(tlInfoPtr->i8255xRecvBufDesc[index]);
        frameDescPtr = &(tlInfoPtr->i8255xRecvDesc[index]);
        status = (unsigned short) (frameDescPtr->recvStatus & 0x0000ffffUL);

/* Mark this frame as processed */
        tm_8255x_processed(*frameDescPtr);

        if (status & TM_RECV_OK)
        {
/* Good packet received: pass it up the stack */
            tempDataPtr   = (unsigned long *) bufDescPtr->bufHandle;
            *dataPtr      = bufDescPtr->bufDataVirtPtr;
            *dataLength   = tm_flip_dword(bufDescPtr->bufCount);
            *dataLength  &= 0x00003fffUL;
            *bufHandlePtr = bufDescPtr->bufHandle;

            packetPtr = (ttPacketPtr) *bufHandlePtr;

/* Mark this buffer as unavailable to the chip and in need of a refill */
            bufDescPtr->bufDataPtr = 0;


            errorCode = TM_DEV_OKAY;
        }
        else 
        {
            if (status & TM_RECV_STATUS_NOMEM)
            {
/* No available resources for this packet.  Refill here? */                    
            }
            else
            {
/* 
 * Another sort of error occured.  Simply skip past this packet and let refill
 * clean it up.  In theory, this shouldn't occur because the chip is not set
 * to save bad frames.
 */            
            }
        }


/* Move to the next buffer in the ring */
        tlInfoPtr->i8255xRecvRingIndex = 
            (unsigned char) ((index+1) % TM_8255X_RECV_RING_SIZE);

    }
    else
    {
/* 
 * ERROR!  We were interrupted because of a received packet, but there doesn't
 * seem to be anything here for us...
 */
        tlInfoPtr->i8255xSpuriousInterrupts++;
    }

    return(errorCode);
}

/*
 * tf8255xGetPhysAddr()
 *
 *   Returns the physical address of this Ethernet interface.  This is just
 *   sample data and this routine should be modified to return real data.
 *
 */
int tf8255xGetPhysAddr(ttUserInterface interfaceHandle,
                       char TM_FAR *   physicalAddress)
{
    int errorCode;
    
    interfaceHandle=interfaceHandle;
    
#ifdef TM_XSCALE_IQ80310
    
    errorCode = tf80310GetPhysAddrFromEeprom (physicalAddress);
    
#else /* !TM_XSCALE_IQ80310 */

    errorCode = TM_DEV_OKAY;
    physicalAddress[0] = 0x11;
    physicalAddress[1] = 0x22;
    physicalAddress[2] = 0x33;
    physicalAddress[3] = 0x44;
    physicalAddress[4] = 0x55;
    physicalAddress[5] = 0x66;
    physicalAddress[6] = 0x77;
    
#endif /* TM_XSCALE_IQ80310 */

    return errorCode;


}

/*
 * tf8255xHandlerIsr()
 *
 *   The only interrupt that we should handle here is a received frame 
 *   interrupt.  When this interrupt is received, we loop through the recieve
 *   ring and count how many packets have been received since the last 
 *   interrupt.  At the very end of the routine, we notify on all of these
 *   packets.  It is important for this to be done at the end of ISR 
 *   (please see Device Driver FAQ in Treck manual).
 */
void tf8255xHandlerIsr(void)
{
    unsigned char  index;
    unsigned char  dirtyIndex;
    unsigned short status;
    int            notifyCount;

    notifyCount = 0;
    
    status = tm_in_word(TM_SCB_STATUS);

/* 
 * Frame received 
 */
    if (status & TM_SCB_FRAME_RECV_INT)
    {
        tm_asm_invd;
        index = tlInfoPtr->i8255xRecvRingIsrIndex;
        dirtyIndex = tlInfoPtr->i8255xRecvRingDirtyIndex;
        while (tlInfoPtr->i8255xRecvDesc[index].recvStatus & TM_RECV_COMPLETE)
        {

/* A frame has been receive; count so we can later notify the stack */
            notifyCount++;

            index = (unsigned char) ((index+1) % TM_8255X_RECV_RING_SIZE);

/* Have we reached the end of where the chip could have received data? */
            if (index == dirtyIndex)
            {
                break;
            }

        }

        tlInfoPtr->i8255xRecvRingIsrIndex = index;
    }
    else
    {
/* We received an interrupt that shouldn't (in theory) have been generated */
    }


/* For every frame received, notify the stack of an incoming frame */    
    tfNotifyInterfaceIsr( tlInfoPtr->i8255xInterfaceHandle, 
                          notifyCount,
                          0, 0UL, 0UL );

/* Acknowledge all 8255x interrupts */
    tm_out_word((unsigned short) (status & 0xfc00), TM_SCB_STATUS);

    tm_clear_cpu_interrupts;

    return;
}

void tf8255xSendComplete(ttUserInterface interfaceHandle)
{
    unsigned char index;
    unsigned char dirtyIndex;
    
    interfaceHandle=interfaceHandle;

    index = tlInfoPtr->i8255xCmdRingIndex;
    dirtyIndex = tlInfoPtr->i8255xCmdRingDirtyIndex;
    while ( (dirtyIndex != index) && 
            (tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus & TM_CMD_COMPLETE))
    {
        if (tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus & TM_CMD_TRANSMIT)
        {

            tlInfoPtr->i8255xCmdRing[dirtyIndex].genStatus = TM_CMD_COMPLETE;

            tfSendCompleteInterface(interfaceHandle,
                                    TM_DEV_SEND_COMPLETE_DRIVER);
        }

        dirtyIndex = (unsigned char) ((dirtyIndex+1) % TM_8255X_CMD_RING_SIZE);
    }
    tlInfoPtr->i8255xCmdRingDirtyIndex = dirtyIndex;

}




#ifdef TM_SWAP_NEEDED

unsigned short tf8255xInWord(int offset)
{
    unsigned short temp16;


#ifdef TM_USE_IO
    temp16 = _inpw( (unsigned short) (tlInfoPtr->i8255xIoAddr + offset) );
#else /* !TM_USE_IO */
    temp16 = *((unsigned short *) (tlInfoPtr->i8255xMemAddrPtr+offset));
#endif /* TM_USE_IO */

    temp16 = tm_flip_word(temp16);

    return temp16;

}

unsigned long tf8255xInDword(int offset)
{

    unsigned long temp32;

#ifdef TM_USE_IO
    temp32 = _inpd( (unsigned short) (tlInfoPtr->i8255xIoAddr + offset) );
#else /* !TM_USE_IO */
    temp32 = *((unsigned long *) (tlInfoPtr->i8255xMemAddrPtr+offset));
#endif /* TM_USE_IO */

    temp32 = tm_flip_dword(temp32);

    return temp32;
}

#endif /* TM_SWAP_NEEDED */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tr8255x.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 28709    Output checksum:  5052
 */
