/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2008
 * Copyright Elmic Systems Japan 1997 - 2008
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Driver for Asix AX88796B Ethernet chip
 *
 * Filename: tb88796.c
 * Author: Lorne
 * Date Created: 2008/10/27
 * $Source: source/drivers/ax88796/tb88796.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:52CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * This file is a heavily modified, combined version of ne.c and 8390.c.
 * Written 1992-94 by Donald Becker.
 *
 * Copyright 1993 United States Government as represented by the
 * Director, National Security Agency.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 */

/*
 * All throughout this code, a naming convention has been employed in
 * order to make it as painless as possible to use this driver for the
 * 8390 in other implementations besides the NE2000. Any hardware settings
 * that are specific to the NE2K have the prefix ne2k, any that are
 * universal to all 8390's have the prefix 8390, any that are specific to
 * the AX88796B use prefix 88796.
 */

/*
 * Includes
 */

/* Treck Specific Includes */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

#ifdef __SCRabbit
/* Softools Rabbit library */
#include <rabbit.h>
#include <SoftoolsIDblock.h>
#endif /* __SCRabbit */

/* NE2000 specific includes and defines */
#include <tr88796.h>



/*
 * Pre-Processor Definitions (may have to be modified)
 */

/* Define the following symbol to read the MAC address from PROM */
/* #define TM_USE_NE2K_PROM */

/*
 * Number of transmit and receive pages to use. We have a total of 0x40 pages
 * to play with. Transmit and receive each require a minimum of 6 pages.
 * RX is included here just to make understanding easier - it is not
 * actually used by the code. The code just looks at how many buffers there
 * are, and how many of them are used for transmit. All others are assumed
 * to be receive.
 */
#define TM_NE2K_TX_PAGES 0x0C /* 3 kbytes (12 * 256 bytes) */
#define TM_NE2K_RX_PAGES (0x40 - TM_NE2K_TX_PAGES)

#ifdef TM_RCM3200
/* Base I/O Port */
#define TM_NE2K_BASE_ADDRESS   0x4000
#endif /* TM_RCM3200 */

/*
 * This is defined to do nothing. If you want the driver to do something
 * else while it's waiting for various things to occur, define it here
 */
#define tm_ne2k_busy_wait

/*
 * These defines are all for Softools C for Rabbit.
 * Change them in accordance with your compiler.
 */
#ifdef __SCRabbit
/* Read in a byte from an external port address */
#define tm_in_ext_byte(p) _ine(p)
/* Write out a byte to an external port address */
#define tm_out_ext_byte(p, b) _oute((p), (b))
/* No operation */
#define tm_ne2k_noop() nop()
/* Millisecond delay */
#define tm_ne2k_delay(ms) delay(ms)
#else /* !__SCRabbit */
#error "Compiler dependent macros must be defined"
#endif /* __SCRabbit */


/*
 * Pre-Processor Definitions (should not have to be modified)
 */

/* The length (in bytes) of an ethernet address */
#define TM_8390_ETHER_ADDR_LEN TM_ETHERNET_PHY_ADDR_LEN
/* The length (in bytes) of the multicast register(s) */
#define TM_8390_MCAST_ADDR_LEN 8

/* The CRC32 reflected polynomial (used for multicast hashing) */
#define TM_NE2K_CRC32_POLY 0xEDB88320UL

/* The length of a buffer page */
#define TM_8390_BUFFER_PAGE_LEN 256

/* Minimum allowable size for a received packet */
#define TM_8390_MIN_ETHER_PACKET 60
/* Maximum allowable size for a received packet */
#define TM_8390_MAX_ETHER_PACKET 1518

/* Maximum number of times around the polling loop */
#define TM_8390_MAX_POLL_LOOP 1

/* Maximum receive packet queue */
#define TM_8390_RCVQ 16

#define TM_NE2K_DATAPORT        0x10  /* Data Port                     */
#define TM_NE2K_RESET           0x1f  /* Reset Port                    */
#define TM_NE2K_BUFFER_START    0x40  /* First page of buffer space    */
#define TM_NE2K_BUFFER_STOP     0x80  /* Last page of buffer space + 1 */
#define TM_NE2K_REGISTER_OFFSET 0x00  /* Where the registers start     */
#define TM_88796_DSR            0x17  /* Device Status Register        */
#define TM_88796_MCR            0x1B  /* MAC Configure Register        */
#define TM_88796_CTEPR          0x1C  /* Current Tx End Page           */
#define TM_88796_BER            0x1E  /* Big-Endian Register           */

#define TM_NE2K_TXBUF_START  TM_NE2K_BUFFER_START
#define TM_NE2K_TXBUF_STOP   (TM_NE2K_TXBUF_START + TM_NE2K_TX_PAGES)
#define TM_NE2K_RXBUF_START  TM_NE2K_TXBUF_STOP
#define TM_NE2K_RXBUF_STOP   TM_NE2K_BUFFER_STOP

/*
 * The command register. It's available on every page
 */
#define TM_8390_CR        (0x00 + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 0 Read-Only Registers
 */
/* Transmit Status Register                 */
#define TM_8390_PG0_TSR   (0x04 + TM_NE2K_REGISTER_OFFSET)
/* Number of Collision Regisger             */
#define TM_8390_PG0_NCR   (0x05 + TM_NE2K_REGISTER_OFFSET)
/* Current Page Register        */
#define TM_8390_PG0_CURR  (0x06 + TM_NE2K_REGISTER_OFFSET)
/* Current Remote DMA Address 0             */
#define TM_8390_PG0_CRDA0 (0x08 + TM_NE2K_REGISTER_OFFSET)
/* Current Remote DMA Address 1             */
#define TM_8390_PG0_CRDA1 (0x09 + TM_NE2K_REGISTER_OFFSET)
/* Receive Status Register                  */
#define TM_8390_PG0_RSR   (0x0c + TM_NE2K_REGISTER_OFFSET)
/* Tally Counter 0 (Frame Alignment Errors) */
#define TM_8390_PG0_CNTR0 (0x0d + TM_NE2K_REGISTER_OFFSET)
/* Tally Counter 1 (CRC Errors)             */
#define TM_8390_PG0_CNTR1 (0x0e + TM_NE2K_REGISTER_OFFSET)
/* Tally Counter 2 (Missed Packet Errors)   */
#define TM_8390_PG0_CNTR2 (0x0f + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 0 Read/Write Registers
 */
/* Page Start Register             */
#define TM_8390_PG0_PSTART (0x01 + TM_NE2K_REGISTER_OFFSET)
/* Page Stop Register              */
#define TM_8390_PG0_PSTOP  (0x02 + TM_NE2K_REGISTER_OFFSET)
/* Boundary Pointer                         */
#define TM_8390_PG0_BNRY   (0x03 + TM_NE2K_REGISTER_OFFSET)
/* Interrupt Status Register                */
#define TM_8390_PG0_ISR    (0x07 + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 0 Write-Only Registers
 */
/* Transmit Page Start Address     */
#define TM_8390_PG0_TPSR   (0x04 + TM_NE2K_REGISTER_OFFSET)
/* Transmit Byte Count Register 0  */
#define TM_8390_PG0_TBCR0  (0x05 + TM_NE2K_REGISTER_OFFSET)
/* Transmit Byte Count Register 1  */
#define TM_8390_PG0_TBCR1  (0x06 + TM_NE2K_REGISTER_OFFSET)
/* Remote Start Address Register 0 */
#define TM_8390_PG0_RSAR0  (0x08 + TM_NE2K_REGISTER_OFFSET)
/* Remote Start Address Register 1 */
#define TM_8390_PG0_RSAR1  (0x09 + TM_NE2K_REGISTER_OFFSET)
/* Remote Byte Count Register 0    */
#define TM_8390_PG0_RBCR0  (0x0a + TM_NE2K_REGISTER_OFFSET)
/* Remote Byte Count Register 1    */
#define TM_8390_PG0_RBCR1  (0x0b + TM_NE2K_REGISTER_OFFSET)
/* Receive Configuration Register  */
#define TM_8390_PG0_RCR    (0x0c + TM_NE2K_REGISTER_OFFSET)
/* Transmit Configuration Register */
#define TM_8390_PG0_TCR    (0x0d + TM_NE2K_REGISTER_OFFSET)
/* Data Configuration Register     */
#define TM_8390_PG0_DCR    (0x0e + TM_NE2K_REGISTER_OFFSET)
/* Interrupt Mask Register         */
#define TM_8390_PG0_IMR    (0x0f + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 1 Registers
 */
/* Physical Address Register 0  */
#define TM_8390_PG1_PAR0 (0x01 + TM_NE2K_REGISTER_OFFSET)
/* Physical Address Register 1  */
#define TM_8390_PG1_PAR1 (0x02 + TM_NE2K_REGISTER_OFFSET)
/* Physical Address Register 2  */
#define TM_8390_PG1_PAR2 (0x03 + TM_NE2K_REGISTER_OFFSET)
/* Physical Address Register 3  */
#define TM_8390_PG1_PAR3 (0x04 + TM_NE2K_REGISTER_OFFSET)
/* Physical Address Register 4  */
#define TM_8390_PG1_PAR4 (0x05 + TM_NE2K_REGISTER_OFFSET)
/* Physical Address Register 5  */
#define TM_8390_PG1_PAR5 (0x06 + TM_NE2K_REGISTER_OFFSET)
/* Current Page Register        */
#define TM_8390_PG1_CURR (0x07 + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 0 */
#define TM_8390_PG1_MAR0 (0x08 + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 1 */
#define TM_8390_PG1_MAR1 (0x09 + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 2 */
#define TM_8390_PG1_MAR2 (0x0a + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 3 */
#define TM_8390_PG1_MAR3 (0x0b + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 4 */
#define TM_8390_PG1_MAR4 (0x0c + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 5 */
#define TM_8390_PG1_MAR5 (0x0d + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 6 */
#define TM_8390_PG1_MAR6 (0x0e + TM_NE2K_REGISTER_OFFSET)
/* Multicast Address Register 7 */
#define TM_8390_PG1_MAR7 (0x0f + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 2 Registers
 */
/* Total Receive Buffer Free Page */
#define TM_88796_PG2_TFP (0x0A + TM_NE2K_REGISTER_OFFSET)
/* Chip version */
#define TM_88796_PG2_VER (0x0B + TM_NE2K_REGISTER_OFFSET)

/*
 * Page 3 Registers
 */
/* Wake-up Control and Status Register */
#define TM_88796_PG3_WUCSR (0x0A + TM_NE2K_REGISTER_OFFSET)
/* Power Management Register */
#define TM_88796_PG3_PMR   (0x0B + TM_NE2K_REGISTER_OFFSET)
/* Miscellaneous Control Register */
#define TM_88796_PG3_MISC  (0x0D + TM_NE2K_REGISTER_OFFSET)

/* Definitions for the Command Register (CR) */
#define TM_8390_CR_STP 0x01 /* Stop                 */
#define TM_8390_CR_STA 0x02 /* Start                */
#define TM_8390_CR_TXP 0x04 /* Transmit Packet      */
#define TM_8390_CR_RD0 0x08 /* Remote DMA Command 0 */
#define TM_8390_CR_RD1 0x10 /* Remote DMA Command 1 */
#define TM_8390_CR_RD2 0x20 /* Remote DMA Command 2 */
#define TM_8390_CR_PS0 0x40 /* Page Select 0        */
#define TM_8390_CR_PS1 0x80 /* Page Select 1        */

/* Definitions for the Interrupt Status Register (ISR) */
#define TM_8390_ISR_PRX 0x01 /* Packet Received     */
#define TM_8390_ISR_PTX 0x02 /* Packet Transmitted  */
#define TM_8390_ISR_RXE 0x04 /* Receive Error       */
#define TM_8390_ISR_TXE 0x08 /* Transmit Error      */
#define TM_8390_ISR_OVW 0x10 /* Overwrite Warning   */
#define TM_8390_ISR_CNT 0x20 /* Counter Overflow    */
#define TM_8390_ISR_RDC 0x40 /* Remote DMA Complete */
#define TM_8390_ISR_RST 0x80 /* Reset Status        */

#define TM_8390_ISR_SELECT  ( TM_8390_ISR_PRX \
							| TM_8390_ISR_PTX \
							| TM_8390_ISR_RXE \
							| TM_8390_ISR_TXE \
							| TM_8390_ISR_OVW )

/* Definitions for the Data Configuration Register (DCR) */
#define TM_8390_DCR_WTS  0x01 /* Word Transfer Select     */
#define TM_8390_DCR_BOS  0x02 /* Byte Order Select        */
#define TM_8390_DCR_LAS  0x04 /* Long Address Select      */
#define TM_8390_DCR_LS   0x08 /* Loopback Select          */
#define TM_8390_DCR_AR   0x10 /* Auto-Initialize Remote   */
#define TM_8390_DCR_FT0  0x20 /* FIFO Threshhold Select 0 */
#define TM_8390_DCR_FT1  0x40 /* FIFO Threshhold Select 1 */

/* Definitions for the Receive Configuration Register (RCR) */
#define TM_8390_RCR_SEP 0x01 /* Save Errored Packets */
#define TM_8390_RCR_AR  0x02 /* Accept Runt Packets  */
#define TM_8390_RCR_AB  0x04 /* Accept Broadcast     */
#define TM_8390_RCR_AM  0x08 /* Accept Multicast     */
#define TM_8390_RCR_PRO 0x10 /* Promiscuous Physical */
#define TM_8390_RCR_MON 0x20 /* Monitor Mode         */

/* Definitions for the Transmit Configuration Register (TCR) */
#define TM_8390_TCR_CRC  0x01 /* Inhibit CRC                */
#define TM_8390_TCR_LB0  0x02 /* Encoded Loopback Control 0 */
#define TM_8390_TCR_LB1  0x04 /* Encoded Loopback Control 1 */
#define TM_8390_TCR_ATD  0x08 /* Auto Transmit Disable      */
#define TM_8390_TCR_OFST 0x10 /* Collision Offset Enable    */

/* Definitions for the Receive Status Register (RSR) */
#define TM_8390_RSR_PRX 0x01 /* Packet Received Intact     */
#define TM_8390_RSR_CRC 0x02 /* CRC Error                  */
#define TM_8390_RSR_FAE 0x04 /* Frame Alignment Error      */
#define TM_8390_RSR_FO  0x08 /* FIFO Overrun               */
#define TM_8390_RSR_MPA 0x10 /* Missed Packet              */
#define TM_8390_RSR_PHY 0x20 /* Physical/Multicast Address */
#define TM_8390_RSR_DIS 0x40 /* Receiver Disabled          */
#define TM_8390_RSR_DFR 0x80 /* Deferring                  */

/* Definitions for the Transmit Status Register (TSR) */
#define TM_8390_TSR_PTX 0x01  /* Packet Transmitted      */
#define TM_8390_TSR_COL 0x04  /* Transmit Collided       */
#define TM_8390_TSR_ABT 0x08  /* Transmit Aborted        */
#define TM_8390_TSR_CRS 0x10  /* Carrier Sense Lost      */
#define TM_8390_TSR_FU  0x20  /* FIFO Underrun           */
#define TM_8390_TSR_CDH 0x40  /* CD Heartbeat            */
#define TM_8390_TSR_OWC 0x80  /* Out of Window Collision */

/* Definitions for the Device Status Register (DSR) */
#define TM_88796_DSR_I_LINK		0x01	/* PHY link up */
#define TM_88796_DSR_I_DUPLEX	0x02	/* PHY full duplex */
#define TM_88796_DSR_I_SPEED	0x04	/* PHY speed = 100 Mbps */
#define TM_88796_DSR_D_RDY		0x10	/* Device ready */
#define TM_88796_DSR_RD_RDY		0x20	/* Read data port ready */
#define TM_88796_DSR_RDMA_RDY	0x40	/* Remote DMA completed */
#define TM_88796_DSR_B_ENDIAN	0x80	/* Big-endian mode */

/* Definitions for Miscellaneous Control Register (MISC) */
#define TM_88796_MISC_TBR       0x01    /* Transmit Buffer Ring Enable */
#define TM_88796_MISC_BCB1      0x02    /* Burst Cycle Base on SA1 or SA0 */
#define TM_88796_MISC_GPTE      0x04    /* General Purpose Timer Enable */

/* Definitions for the MAC Configure Register (MCR) */
#define TM_88796_MCR_VLANE      0x08    /* VLAN Enable */
#define TM_88796_MCR_MPSEL      0x10    /* Media Select by Program */
#define TM_88796_MCR_BBTC       0x20    /* Back-To-Back Transmission */
#define TM_88796_MCR_BPLE       0x40    /* Back Pressure Leakage Enable */
#define TM_88796_MCR_DUPX       0x80    /* MAC Duplex Setting */

/* Definitions for the Current Transmit End Page Register (CTEPR) */
#define TM_88796_CTEPR_TXCQF    0x80    /* Tx Command Queue Full      */
#define TM_88796_CTEPR_CTEPR    0x7F    /* Current Tx end page mask   */

/* Definitions for the received frame header status */
#define TM_88796_HDR_PRX        0x01    /* Packet Received Intact     */
#define TM_88796_HDR_CRC        0x02    /* CRC Error                  */
#define TM_88796_HDR_FAE        0x04    /* Frame Alignment Error      */
#define TM_88796_HDR_MII        0x08    /* MII Error                  */
#define TM_88796_HDR_RUNT       0x10    /* Runt Packet                */
#define TM_88796_HDR_MCBC       0x20    /* Multicast/Broadcast Address*/


/*
 * Structure Definitions
 */

/*
 * Treck packet information for queuing.
 */
typedef struct tsPktInfo
{
	char TM_FAR *	data;
	int				length;
	ttUserBuffer	buffer;
} ttPktInfo;
typedef ttPktInfo TM_FAR * ttPktInfoPtr;

/*
 * This structure contains all of the information we need to track for
 * the device.
 */
typedef struct tsNE2kDev
{
/* The interface handle */
    ttUserInterface ifHandle;
/* The Ethernet address we read out of the hardware */
    unsigned char   ethAddr[TM_ETHERNET_PHY_ADDR_LEN];
/* Boolean value indicating that the physical link is ready */
	unsigned char	linkIsUp;
/* Receive packet queue */
	int				rxadd, rxremove, rxcount;
	ttPktInfo		rxpackets[TM_8390_RCVQ];
/* Current transmit page (where to put the next transmit packet) */
	unsigned char	txpage;
/* Device counters */
	tt88796Stats			 stats;
} ttNE2kDev;
typedef ttNE2kDev TM_FAR * ttNE2kDevPtr;


/*
 * Global Variables
 */

static ttNE2kDevPtr tlNe2kPtr;


/*
 * Forward Definitions
 */


static void tf88796BlockInput(unsigned int     readLen,
                             unsigned int     currOffset,
                             unsigned char TM_FAR * rxBuffer);

static void tf88796BlockOutput(unsigned int     length,
                             unsigned int     address,
                             unsigned char TM_FAR * buffer);

static int tf88796Init(void);
static void tf88796Dequeue(void);
static int tf88796CheckReceive(void);


/*
 * Code
 */

/*
 * tf88796Open()
 *
 *   Configure and start up the Ethernet interface.
 *   We take the address from the Ethernet board and set up the board.
 *   Parameters:
         interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       error:            TM_DEV_ERROR
 *       success:          TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Open(ttUserInterface interfaceHandle)
{
	int		retCode;

	retCode = TM_DEV_ERROR;
	tlNe2kPtr = (ttNE2kDevPtr)0;

/* Did the user turn on the transmit queue? */
    if (tfInterfaceXmitQueueOn(interfaceHandle))
    {
        tfKernelError("tf88796Open",
                        "Transmit queue not supported on this device");
		goto ne2kOpenEnd;
    }

/* Make sure that scatter sends are not enabled */
    if (tfInterfaceScatterCount(interfaceHandle) != 1)
    {
        tfKernelError("tf88796Open",
                        "Scatter send not supported on this device");
		goto ne2kOpenEnd;
    }

/* Get memory for the ttNE2kDev structure */
    tlNe2kPtr = (ttNE2kDevPtr)tm_get_raw_buffer(sizeof(ttNE2kDev));
    if (tlNe2kPtr == (void TM_FAR *)0)
    {
        tfKernelError("tf88796Open", "Memory allocation failed");
		goto ne2kOpenEnd;
    }
/* Zero out this memory area */
    tm_bzero( tlNe2kPtr, sizeof(ttNE2kDev) );
/* Save the interface handle */
    tlNe2kPtr->ifHandle = interfaceHandle;
/* Initial statistical values */
	tlNe2kPtr->stats.txminfree = TM_NE2K_TX_PAGES * TM_8390_BUFFER_PAGE_LEN;

/* Initialize hardware */
	retCode = tf88796Init();
	if (retCode != TM_DEV_OKAY)
	{
/* Hardware problem. Release the allocated memory. */
    	tm_free_raw_buffer((ttRawBufferPtr)tlNe2kPtr);
		tlNe2kPtr = (ttNE2kDevPtr)0;
	}

ne2kOpenEnd:
	return retCode;
}


/*
 * tf88796Init()
 *
 *	Initialize the Ethernet chip.
 *	This chip is an integrated MAC and PHY transceiver.
 *	After issuing a reset to the device the PHY will automatically negotiate
 *	the best link speed and duplex for the physical connection. This may take
 *	a few seconds during which time no packets can be sent or received.
 *	While the PHY is in autonegotiation, the MAC registers can be set up
 *	to reflect the desired MAC address and transmit/receive buffer sizes.
 *
 *	Parameters:
 *		interfaceHandle		The user interface handle for this device.
 *	Returns:
 *		error:				TM_DEV_ERROR
 *		success:			TM_DEV_OKAY
 */
static int tf88796Init(void)
{
	unsigned long ulCount;
	time_t		texpire;
	int			retCode;
	int			counter;

/* Prepare for an error exit */
    retCode = TM_DEV_ERROR;

/* Initialize hardware specific data */
	tlNe2kPtr->txpage = TM_NE2K_TXBUF_START;

/* Reset the card. This is done by reading from the reset port. */
    tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_RESET);

/* Busy-wait for the reset to finish */
	ulCount = 1000000L;
    while (    (   tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_DSR)
                  & TM_88796_DSR_D_RDY )
            == 0 )
    {
		if (--ulCount == 0)
			goto ne2kInitEnd;		/* Fail after some time */
        tm_ne2k_busy_wait;
    }

#ifdef __SCRabbit
/* Read in the Ethernet address from the Rabbit RCM3200 ID block */
    memcpy(tlNe2kPtr->ethAddr, idBlock->macAddr[0], TM_ETHERNET_PHY_ADDR_LEN);
#else /* !__SCRabbit */
#error "Cannot determine MAC address"
#endif /* __SCRabbit */

/* Stop the chip and set it to Page 0 */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP );
/* Set the chip for byte-wide transfers */
    tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_DCR, 0);
/* Clear Count Registers */
    tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0, 0);
    tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0);
/* Accept broadcast and multicast packets */
    tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR,
    				TM_8390_RCR_AB | TM_8390_RCR_AM);
/* Set transmit to internal loopback mode */
    tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, TM_8390_TCR_LB0);
/* Set the receive page limits */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_PSTART,
                      TM_NE2K_RXBUF_START );
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_PSTOP,
                      TM_NE2K_RXBUF_STOP );
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY,
                      TM_NE2K_RXBUF_START );

/* Change to Page 1 */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_PS0 + TM_8390_CR_STP );
/* Set physical address & disable multicast */
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_PAR0 + counter,
                          tlNe2kPtr->ethAddr[counter] );
    }
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + counter,
                          0 );
    }

/* Set the current receive page */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_CURR,
                      TM_NE2K_RXBUF_START );

/* Change to Page 3 */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP
                      + TM_8390_CR_PS0 + TM_8390_CR_PS1 );
/* Enable transmit buffer ring feature */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_88796_PG3_MISC,
                      TM_88796_MISC_TBR );

/* Set back to Page 0 */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP );

/* Enable back-to-back trasmission, use internal PHY */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_88796_MCR,
                      TM_88796_MCR_BBTC );

/*
 * Defer starting the device until we have a valid link.
 * The PHY may take a few seconds to negotiate speed and duplex with
 * the peer, at which point, the I_LINK bit will be set in the Device
 * Status Register (DSR).
 * If there is no cable connection, the I_LINK status will remain false.
 * The "link up" event will be serviced in tf88796Poll().
 */

/* Wait up to 5 seconds for the physical link to become active */
	texpire = readRTC(0) + 5;
	do {
		if (tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_DSR)
						& TM_88796_DSR_I_LINK)
			break;
		tm_ne2k_busy_wait;
	} while (readRTC(0) < texpire);

/* Indicate success */
    retCode = TM_DEV_OKAY;

ne2kInitEnd:
    return retCode;
}


/*
 * tf88796Close()
 *
 *   Shut down the chip.
 *   Parameters:
 *       interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Close (ttUserInterface interfaceHandle)
{
/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Stop chip and set it to Page 0 */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP );

	if (tlNe2kPtr)
	{
		/* Free any locally queued packets */
		tf88796Dequeue();

		tm_free_raw_buffer((ttRawBufferPtr)tlNe2kPtr);
		tlNe2kPtr = (ttNE2kDevPtr)0;
	}

    return(TM_DEV_OKAY);
}


/*
 * tf88796Dequeue()
 *
 *   Free any locally queued packets.
 *   Parameters:
 *       none
 *   Returns:
 *       none
 */
static void tf88796Dequeue(void)
{
	ttPktInfoPtr	pktInfoPtr;
	int				i, n;

	/* Number of receive packets on the local queue */
	n = tlNe2kPtr->rxcount;
	if (n > 0)
	{
		/* Packet information */
		pktInfoPtr = tlNe2kPtr->rxpackets;

		/* Head of the queue */
		i = tlNe2kPtr->rxremove;

		/* Free all queued packets */
		do
		{
			tfFreeDriverBuffer(pktInfoPtr[i].buffer);
			if (++i >= TM_8390_RCVQ)
				i = 0;
		} while (--n);

		/* Update the device state information */
		tlNe2kPtr->rxremove = i;
		tlNe2kPtr->rxcount = 0;
	}
}


/*
 * tf88796Poll()
 *
 *   This function replaces the interrupt handler. It polls the device for
 *   receive, transmit-complete, and receive overflow warning events.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       none
 *   Returns:
 *       none
 */
static int tf88796Poll(ttUserInterface interfaceHandle)
{
/* Used to store the state of the ISR */
    unsigned char   interrupts;
	unsigned char	maxloop;
	unsigned char	u8;
	int				retCode;

	retCode = TM_DEV_OKAY;

/* Sanity check */
	if (!tlNe2kPtr)
	{
		retCode = TM_DEV_ERROR;
		goto ne2kPollEnd;
	}

/* Enable the device if the physical link transitions to "up" */
	if (!tlNe2kPtr->linkIsUp)
	{
		if (tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_DSR)
						& TM_88796_DSR_I_LINK)
		{
			tlNe2kPtr->linkIsUp = 1;			/* Link is now up */
			/*
			 * Clear all device interrupts,
			 * Enable transmit,
			 * Start the device
			 */
			tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, 0xff);
			tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, 0);
			tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR,
							TM_8390_CR_RD2 + TM_8390_CR_STA);
		}

		/* Link is either down or just starting */
		goto ne2kPollEnd;
	}

/* Return if the device is not ready or the physical link is broken */
	if ((tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_DSR)
						& TM_88796_DSR_D_RDY) == 0)
	{
		++tlNe2kPtr->stats.dev_not_ready;
		goto ne2kPollEnd;
	}
	if ((tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_DSR)
						& TM_88796_DSR_I_LINK) == 0)
	{
		++tlNe2kPtr->stats.phy_not_ready;
		goto ne2kPollEnd;
	}

/* Bounded loop while there are events to service */
    for (maxloop = 0; maxloop < TM_8390_MAX_POLL_LOOP; ++maxloop)
    {
/* Read and clear interrupts; exit loop if there is nothing to do. */
		interrupts = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR);
		tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, interrupts);
		if ((interrupts & TM_8390_ISR_SELECT) == 0)
			break;

		++tlNe2kPtr->stats.intevent;

/* Process received frames */
		if (interrupts & TM_8390_ISR_PRX)
		{
			++tlNe2kPtr->stats.rxint;
			retCode = tf88796CheckReceive();
			if (retCode != TM_DEV_OKAY)
			{
				goto ne2kPollEnd;
			}
		}

/* Process receive events */
		if (interrupts & (TM_8390_ISR_RXE | TM_8390_ISR_OVW))
		{
			/* Read receive status register */
	        u8 = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSR);

			/* Count receive errors */
			if (interrupts & TM_8390_ISR_RXE)
			{
				++tlNe2kPtr->stats.rxerror;
			}
			if (interrupts & TM_8390_ISR_OVW)
			{
				++tlNe2kPtr->stats.overrun;
			}
			if (u8 & TM_8390_RSR_CRC)
			{
				++tlNe2kPtr->stats.rxcrc;
			}
			if (u8 & TM_8390_RSR_FAE)
			{
				++tlNe2kPtr->stats.rxalign;
			}
			if (u8 & TM_8390_RSR_MPA)
			{
				++tlNe2kPtr->stats.rxmissed;
			}
			if (u8 & TM_8390_RSR_DIS)
			{
				++tlNe2kPtr->stats.rxdisabled;
			}
		}

/* Process transmit events */
		if (interrupts & (TM_8390_ISR_PTX | TM_8390_ISR_TXE))
		{
			/* Read transmit status register */
	        u8 = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TSR);

			/* Count transmitted frames */
			if (interrupts & TM_8390_ISR_PTX)
			{
				++tlNe2kPtr->stats.txint;
			}

			/* Count transmit errors */
			if (interrupts & TM_8390_ISR_TXE)
			{
				++tlNe2kPtr->stats.txerror;
			}
			if (u8 & TM_8390_TSR_COL)
			{
				++tlNe2kPtr->stats.txcollision;
			}
			if (u8 & TM_8390_TSR_ABT)
			{
				++tlNe2kPtr->stats.txabort;
			}
			if (u8 & TM_8390_TSR_OWC)
			{
				++tlNe2kPtr->stats.txwindow;
			}
		}

    }

/* Notify Treck of the number of packets we have received */
	if (tlNe2kPtr->rxcount)
	{
		tfNotifyInterfaceTask(interfaceHandle, tlNe2kPtr->rxcount,
							0, 0UL, 0UL);
	}

ne2kPollEnd:
    return(retCode);
}


/*
 * tf88796BlockInput()
 *
 *   This function transfers data from device memory to
 *   standard memory.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       readLen    Number of bytes to transfer
 *       currOffset The place to begin reading from in the internal buffers
 *       rxBuffer   The buffer to copy into
 *   Returns:
 *       none
 */
static void tf88796BlockInput(unsigned int     readLen,
                             unsigned int     currOffset,
                             unsigned char TM_FAR * rxBuffer)
{
/* Store the packet length */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0,
                      readLen & 0xFF );
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1,
                      readLen >> 8 );
/* Store the address to read from */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0,
                      currOffset & 0xFF );
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1,
                      currOffset >> 8 );
/* Issue a Remote Read DMA command */
    tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD0);

	while (readLen--)
	{
/*
 * Get 8 bits from the dataport & save it to the appropriate place in the
 * buffer
 */
		*rxBuffer++ = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT);
	}
}


/*
 * tf88796BlockOutput()
 *
 *   This function transfers data from standard memory to device
 *   memory.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       length     Number of bytes to transfer
 *       address    Address in the device memory
 *       buffer     The buffer address
 *   Returns:
 *       none
 */
static void tf88796BlockOutput(unsigned int     length,
                               unsigned int     address,
                               unsigned char TM_FAR * buffer)
{
	unsigned char    byte;

/* Store the length to write */
	tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0,
						length & 0xFF);
	tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1,
						length >> 8);
/* Store the address to write to */
	tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0,
						address & 0xFF);
	tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1,
						address >> 8);
/* Issue a Remote Write DMA command */
	tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR,
						TM_8390_CR_RD1);

	while (length--)
	{
/* Write the next byte to the port */
		byte = *buffer++;
		tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT, byte);
	}
}


/*
 * tf88796CheckReceive()
 *
 *	This function checks for received frames in the device.
 *	If found, received frames are extracted from the device
 *	and added to a locally maintained receive queue.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *	Parameters:
 *		none
 *	Returns:
 *		error:		TM_DEV_ERROR
 *		success:	TM_DEV_OKAY
 */
static int tf88796CheckReceive(void)
{
	unsigned char		page;
	unsigned char		header[4];
	unsigned int		addr;
	int					length, extra, limit;
	char TM_FAR *		bufPtr;
	ttUserBuffer		bufHandle;
	ttPktInfoPtr		pktInfoPtr;
	int					maxloop;
	int					retCode;
	int					rxadd;

	retCode = TM_DEV_OKAY;

	/* Local copy of receive queue information */
	pktInfoPtr = tlNe2kPtr->rxpackets;
	rxadd = tlNe2kPtr->rxadd;

	/* Get the location of the next frame to retrieve */
	page = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY);

	/*
	 * Extract all received frames from the device, even if there is
	 * not enough memory or queue space for the received data.
	 * Maintain a loop count to allow an exit if there is a serious
	 * problem.
	 */
	for (maxloop = 50; ; )
	{
		/* Exit if there are no new frames available */
		if (page == tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_CURR))
			break;

		/* Page address */
		addr = page * TM_8390_BUFFER_PAGE_LEN;

		/* Retrieve the frame header */
		tf88796BlockInput(sizeof(header), addr, header);

		/* Record statistics */
		if (header[0] & TM_88796_HDR_PRX)
			++tlNe2kPtr->stats.rxgood;
		if (header[0] & TM_88796_HDR_CRC)
			++tlNe2kPtr->stats.rxcrc;
		if (header[0] & TM_88796_HDR_FAE)
			++tlNe2kPtr->stats.rxalign;
		if (header[0] & TM_88796_HDR_MII)
			++tlNe2kPtr->stats.rxmii;
		if (header[0] & TM_88796_HDR_RUNT)
			++tlNe2kPtr->stats.rxrunt;
		if (header[0] & TM_88796_HDR_MCBC)
			++tlNe2kPtr->stats.rxmcbc;

		/* Start of frame in device memory and frame length */
		addr += sizeof(header);
		length = 256 * (int)header[3] + (int)header[2];

		/* Check if there is room in the receive packet queue */
		if (tlNe2kPtr->rxcount < TM_8390_RCVQ)
		{
			/* Get a buffer for the data */
			bufPtr = tfGetDriverBuffer(&bufHandle, length, 1);
			if (bufPtr)
			{
				/* Distance to end of receive memory */
				limit = TM_8390_BUFFER_PAGE_LEN * TM_NE2K_RXBUF_STOP - addr;

				/* Check for frame wraparound */
				extra = length - limit;
				if (extra > 0)
				{
					/* Frame is split in two */
					tf88796BlockInput(limit, addr, bufPtr);
					tf88796BlockInput(extra,
							TM_NE2K_RXBUF_START * TM_8390_BUFFER_PAGE_LEN,
							bufPtr + limit);
				}
				else
				{
					/* Frame is contiguous */
					tf88796BlockInput(length, addr, bufPtr);
				}

				/* Add packet to receive queue */
				pktInfoPtr[rxadd].buffer = bufHandle;
				pktInfoPtr[rxadd].data = bufPtr;
				pktInfoPtr[rxadd].length = length;
				++tlNe2kPtr->rxcount;
				if (++rxadd >= TM_8390_RCVQ)
				{
					rxadd = 0;
				}
			}
		}

		/* Link to next frame page. Free up receive memory. */
		page = header[1];
		tm_out_ext_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY, page);

		/* Avoid an infinite loop if there is a serious problem */
		if (--maxloop == 0)
		{
			retCode = TM_DEV_ERROR;
			break;
		}
	}

	/* Update device state information */
	tlNe2kPtr->rxadd = rxadd;

	return retCode;
}



/*
 * tf88796Receive()
 *
 *   This function passes on received packets to the Treck stack.
 *   It is called after it has been determined that received frames
 *   exist in the device and the Treck stack has been notified of
 *   how many frames have been received.
 *   The received frames have already been extracted from the device
 *   and added to a local queue maintained by this driver.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         We fill in this pointer with a pointer to the data
 *                       we have just received
 *       dataLength      We fill in this length with the length of the data
 *                       we have just received
 *       ttUserBufferPtr We fill out this buffer handle with the buffer handle
 *                       we get back when we allocate memory
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
TM_NEARCALL int tf88796Receive(ttUserInterface        interfaceHandle,
                  char TM_FAR * TM_FAR * dataPtr,
                  int TM_FAR *           dataLength,
                  ttUserBufferPtr        bufHandlePtr)
{
/* The return value */
    int             retCode;
	ttPktInfoPtr	pktInfo;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Assume that we return OKAY unless otherwise instructed */
    retCode = TM_DEV_OKAY;

/* Sanity check */
	if (!tlNe2kPtr || tlNe2kPtr->rxcount < 1)
	{
		retCode = TM_DEV_ERROR;
        goto ne2kRecvEnd;
	}

/* Remove the next packet from the receive queue */
	pktInfo = tlNe2kPtr->rxpackets + tlNe2kPtr->rxremove;
	if (++tlNe2kPtr->rxremove >= TM_8390_RCVQ)
		tlNe2kPtr->rxremove = 0;
	--tlNe2kPtr->rxcount;

/* Extract the packet information */
	*dataPtr = pktInfo->data;
	*dataLength = pktInfo->length;
	*bufHandlePtr = pktInfo->buffer;

ne2kRecvEnd:
    return(retCode);
}


/*
 * tf88796Send()
 *
 *   This function stores a packet in the device's buffers for sending and
 *   triggers a send.
 *   This function will wait a limited amount of time for space in the
 *   transmit command queue.
 *
 *   This chip has a limited ability to queue the transmit start page (TPSR)
 *   and transmit byte count (TBCR0, TBCR1) each time a transmit start
 *   command is issued.
 *   This function will wait a limited amount of time for space in the
 *   transmit command queue.
 *
 *   This chip also provides the page address of the last frame transmitted.
 *   The available space in the transmit ring buffer can be determined by
 *   subtracting the page address of the last frame stored, which is
 *   maintained in the device state block, from the page address of the last
 *   frame transmitted, accounting for ring buffer wrap.
 *   This function will wait a limited amount of time for the required
 *   amount of space in the transmit ring buffer.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         The pointer to the data we're sending
 *       dataLength      The length of the data we're sending
 *       flag            A flag indicating various things about the data to
                         send
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
TM_NEARCALL int tf88796Send ( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag )
{
/* The return value for this function */
    int retCode;
/* Timeout calculation */
	time_t texpire;
/* End of free transmit buffer space */
	unsigned char txend;
/* Free transmit buffer space */
	int				txfree;
	int				extra;
	unsigned int	addr;
	int				limit;

/* Deal with 'unused parameter' warnings */
    TM_UNREF_IN_ARG(flag);

/* Prepare for possible error exit */
	retCode = TM_DEV_ERROR;

/* Sanity check */
	if (!tlNe2kPtr)
	{
		retCode = TM_DEV_ERROR;
        goto ne2kSendEnd;
	}

/* Limit the time spent waiting to transmit */
	texpire = readRTC(0) + 5;
/*
 * Wait until the link is up before proceeding
 */
	if (!tlNe2kPtr->linkIsUp)
	{
/*
 * Wait up to 5 seconds for the physical link to become active. Silently
 * drop the packet if the timeout expires.
 */
		for (;;)
		{
			retCode = tf88796Poll(interfaceHandle);
			if (retCode != TM_DEV_OKAY)
				goto ne2kSendEnd;
			if (tlNe2kPtr->linkIsUp)
				break;
			if (readRTC(0) >= texpire)
			{
				++tlNe2kPtr->stats.phy_not_ready;
				goto ne2kSendOkay;
			}
			tm_ne2k_busy_wait;
		}
	}

/* Wait for free space in the transmit command queue */
	while (tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_CTEPR)
							& TM_88796_CTEPR_TXCQF)
	{
		retCode = tf88796Poll(interfaceHandle);
		if (retCode != TM_DEV_OKAY)
			goto ne2kSendEnd;
		if (readRTC(0) >= texpire)
		{
			++tlNe2kPtr->stats.txtimeout;
			goto ne2kSendOkay;
		}
		tm_ne2k_busy_wait;
	}

/* Wait for free space in the transmit buffer */
	for (;;)
	{
		/* Current transmit end page */
		txend = tm_in_ext_byte(TM_NE2K_BASE_ADDRESS + TM_88796_CTEPR)
							& TM_88796_CTEPR_CTEPR;

		/* Exit if the register is inactive */
		if (txend < TM_NE2K_TXBUF_START || txend >= TM_NE2K_TXBUF_STOP)
			break;

		/* Free space in transmit buffer (account for wraparound) */
		txfree = (int)txend - (int)tlNe2kPtr->txpage;
		if (txfree < 0)
			txfree += TM_NE2K_TX_PAGES;
		txfree *= TM_8390_BUFFER_PAGE_LEN;
		if (txfree < tlNe2kPtr->stats.txminfree)
			tlNe2kPtr->stats.txminfree = txfree;
		if (txfree >= dataLength)
			break;

		retCode = tf88796Poll(interfaceHandle);
		if (retCode != TM_DEV_OKAY)
			goto ne2kSendEnd;
		if (readRTC(0) >= texpire)
		{
			++tlNe2kPtr->stats.txtimeout;
			goto ne2kSendOkay;
		}
		tm_ne2k_busy_wait;
	}

/* Output the packet to the device's buffers */
	addr = tlNe2kPtr->txpage * TM_8390_BUFFER_PAGE_LEN;
	limit = TM_NE2K_TXBUF_STOP * TM_8390_BUFFER_PAGE_LEN - addr;
	extra = dataLength - limit;
	if (extra > 0)
	{
		tf88796BlockOutput(limit, addr, dataPtr);
		tf88796BlockOutput(extra,
					TM_NE2K_TXBUF_START * TM_8390_BUFFER_PAGE_LEN,
					dataPtr + limit);
	}
	else
	{
		tf88796BlockOutput(dataLength, addr, dataPtr);
	}

/* Queue the transmit command */
/* Set the packet length */
	tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR0,
					dataLength & 0xff );
	tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR1,
					dataLength >> 8 );
/* Set the page to start sending from */
	tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TPSR,
					tlNe2kPtr->txpage );
/* Trigger the send */
	tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
					TM_8390_CR_RD2 + TM_8390_CR_TXP );

	++tlNe2kPtr->stats.transmit;

/* Advance to the next transmit page */
	txend = (addr + dataLength + TM_8390_BUFFER_PAGE_LEN - 1)
					/ TM_8390_BUFFER_PAGE_LEN;
	if (txend >= TM_NE2K_TXBUF_STOP)
		txend -= TM_NE2K_TX_PAGES;
	tlNe2kPtr->txpage = txend;

ne2kSendOkay:
/* Release ownership of the packet */
	tfSendCompleteInterface(interfaceHandle,
					TM_DEV_SEND_COMPLETE_DRIVER);

/* Indicate success */
    retCode = TM_DEV_OKAY;

ne2kSendEnd:
    return(retCode);
}


/*
 * tf88796Ioctl()
 *
 *   This function, in general, allows the user to talk directly to the
 *   chip, for any reason. We have implemented it here to allow the
 *   following:
 *		1) set multicast addresses,
 *		2) periodically poll the device to service activity,
 *		3) retrieve device statistics.
 *
 *   flag bit mask 1:
 *		Poll device for service.
 *		'optionPtr' and 'optionLen' are unused.
 *
 *   flag bit mask 2:
 *		Retrieve device statistics:
 *		'optionPtr' points to storage to receive the pointer to the
 *		common statistics structure (tt88796StatsPtr) for the device.
 *		'optionLen' = sizeof(tt88796StatsPtr)
 *
 *   TM_DEV_SET_MCAST_LIST:
 *     Adds 'optionLen' number of multicast entries to the multicast hash
 *     table.  This is done by
 *         1. Compute a 32-bit CRC on the address.
 *         2. Extract the most significant 6 bits of the computed CRC.
 *         3. Set the bit coresponding to the above number in the multicast
 *            address filter (ie, decode the 6 bit value above into a 64 bit
 *            value: 50 -> 0x0002 0000 0000 0000)
 *
 *   TM_DEV_SET_ALL_MCAST:
 *     Sets the chip to receive all multicast packets (ie, set the address
 *     filter to all ones).
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       interfaceHandle   The handle to the user interface for this device
 *       flag       A flag what we want this function to do
 *       dataPtr    The pointer to the data being passed in
 *       dataLength The length of the data (in this case, how many multicast
 *                  addresses are being added).
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Ioctl(ttUserInterface interfaceHandle,
                int             flag,
                void TM_FAR *   optionPtr,
                int             optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/* The calculated CRC */
    unsigned long   calcCRC;
/* A counter */
    unsigned char   i;
/* Which byte of the ethernet address we're working on */
    unsigned char   byte;
/* Which bit of the current byte we're working on */
    unsigned char   bit;
/* A pointer to the list of multicast addresses passed to us */
    unsigned char * multiAddr;
/* The actual hash table to be written to the chip */
    unsigned char   logicalAddrFilter[TM_8390_MCAST_ADDR_LEN];
#endif /* TM_IGMP or TM_6_USE_MLD */
	int				retCode;

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

	retCode = TM_DEV_OKAY;

/* Poll for Ethernet events. */
    if (flag & 1)
	{
    	retCode = tf88796Poll(interfaceHandle);
		goto ne2kIoctlEnd;
	}

/* Dump statistics */
    if (flag & 2)
	{
		if (optionPtr && optionLen >= sizeof(tt88796StatsPtr))
		{
			*(tt88796StatsPtr *)optionPtr = &tlNe2kPtr->stats;
		}
	}

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    if (flag & TM_DEV_SET_MCAST_LIST)
    {
/*
 * Since most Ethernet transmission (this device is no exception) starts with
 * the least significant bit of each byte, it makes sense to do the CRC
 * calculation based on a shift right model rather than the traditional
 * shift left model.
 * The advantage is that the input bytes do not have to be bit reversed prior
 * to being merged with the running CRC.
 * The disadvantage is that the resulting CRC has the most significant bit of
 * the CRC in the least significant bit of the running CRC variable and the
 * bits must be reversed upon extraction.
 * It also requires the use of a "reflected" CRC-32 polynomial, i.e. one in
 * which all the bits are in the reverse order.
 * This is preferable since only the 6 most significant bits of the resulting
 * CRC are needed to address into the 64-bit multicast bit array: 3 bits
 * address the the byte offset (0 to 7) and 3 bits address the bit offset
 * (0 to 7) within that byte.
 */

/* Zero out the hash table */
        tm_bzero( logicalAddrFilter, TM_8390_MCAST_ADDR_LEN );

/* Get the first address */
        multiAddr = (unsigned char TM_FAR *) optionPtr;

/* While we have more addresses to work on */
        while (optionLen-- > 0)
        {

/* Initialize the calculated CRC (with all ones) */
            calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
            for (byte = 0; byte < TM_8390_ETHER_ADDR_LEN; byte++)
            {

/* Merge input bits with current CRC */
				calcCRC ^= (unsigned long)*multiAddr++ & 0xFFUL;

/* For each bit in the current byte */
                for (bit = 0; bit < 8; bit++)
                {

/*
 * Shift CRC and, if the bit shifted out is 1, merge the reflected
 * polynomial into the CRC.
 */
					if (calcCRC & 1)
						calcCRC = (calcCRC >> 1) ^ TM_NE2K_CRC32_POLY;
					else
						calcCRC = calcCRC >> 1;
                }
            }

/* Extract the most significant CRC bits to get the byte offset (0..7) */
			for (byte = 0, i = 0; i < 3; ++i)
			{
				byte <<= 1;
				byte |= (unsigned char)(calcCRC & 1);
				calcCRC >>= 1;
			}

/*
 * Extract the next most significant CRC bits to get the bit offset
 * (0..7) within the byte.
 */
			for (bit = 0, i = 0; i < 3; ++i)
			{
				bit <<= 1;
				bit |= (unsigned char)(calcCRC & 1);
				calcCRC >>= 1;
			}

/* Set the appropriate bit in the appropriate byte of the array */
            logicalAddrFilter[byte] |= 1 << bit;
        }

/* Set chip to Page 1 */
		tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_PS0 );
/* Write the new multicast hash to the multicast address register */
        for (i = 0; i < TM_8390_MCAST_ADDR_LEN; i++)
        {
            tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + i,
                              logicalAddrFilter[i] );
        }

/* Set chip to Page 0 */
		tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 );
    }

    if (flag & TM_DEV_SET_ALL_MCAST)
    {
/* Set chip to Page 1 */
		tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_PS0 );
/* Set the logical address filter to all ones (receive ALL multicast packets)
*/
        for (i = 0; i < TM_8390_MCAST_ADDR_LEN; i++)
        {
            tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + i,
                              0xff );
        }

/* Set chip to Page 0 */
		tm_out_ext_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 );
    }
#endif /* TM_IGMP or TM_6_USE_MLD */

ne2kIoctlEnd:
    return(retCode);
}


/*
 * tf88796GetPhyAddr()
 *
 *   This function retrieves the adapter's physical address.
 *
 *   Parameters:
 *       interfaceHandle  The handle to the user interface for this device
 *       physicalAddress A buffer into which to store the physical address
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796GetPhyAddr(ttUserInterface interfaceHandle,
                     char TM_FAR *   physicalAddress)
{
/* A counter */
    unsigned char counter;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Copy the address into the buffer */
    for (counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        physicalAddress[counter] = tlNe2kPtr->ethAddr[counter];
    }

    return(TM_DEV_OKAY);
}

/***/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tb88796.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 32795    Output checksum: 25528
 */
