/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2008
 * Copyright Elmic Systems Japan 1997 - 2008
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Driver for Asix AX88796 Ethernet chip
 *
 * Filename: tr88796.c
 * Author: Lorne
 * Date Created: 2008/10/27
 * $Source: source/drivers/ax88796/tr88796.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2011/05/27 13:05:53CDT $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * This file is a heavily modified, combined version of ne.c and 8390.c.
 * Written 1992-94 by Donald Becker.
 *
 * Copyright 1993 United States Government as represented by the
 * Director, National Security Agency.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 */

/*
 * Includes
 */

/* Treck Specific Includes */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

#ifdef __SCRabbit
/* Softools Rabbit library */
#include <rabbit.h>
#include <SoftoolsIDblock.h>
#endif /* __SCRabbit */

/* NE2000 specific includes and defines */
#include <tr88796.h>



/*
 * Pre-Processor Definitions (may have to be modified)
 */

/*
 * Device memory allotment.
 * Internally, the AX88796 has 16Kbytes of shared transmit and receive
 * buffer memory.
 * The following definition determines how much memory will be used for
 * transmit; the remainder will be used for receive.
 * Define the maximum number of packets that need to be held in the
 * device memory awaiting transmission.
 * 1536 bytes of memory are reserved for each transmit packet.
 * Minimum value is 1. Default value is 2, which leaves 13K for receive.
 * Note: If this value is set too high, you may see degradation in reception
 * as a result of missed packets caused by receive buffer overrun.
 */
#define TM_88796_TX_PACKETS  2

/*
 * Packet queue sizes.
 * Buffering of packets moving from the Treck stack to the device
 * is necessary so that the application is not delayed while waiting
 * for the device to become ready.
 * Similarly, packets received by the device need to be quickly moved to a
 * packet queue prior to delivery to the Treck stack to minimize the
 * possibility of overflowing the receive buffer memory.
 * Packets that cannot be added to an already full queue will be dropped
 * and will need to be retransmitted.
 * Define the maximum number of packets allowed to be queued for transmit
 * and reception.
 */
#define TM_88796_TXQ_SIZE  16
#define TM_88796_RXQ_SIZE  16

/*
 * Maximum number of times around the polling and receive loops.
 * These values are arbitrarily chosen to prevent and long or potentially
 * endless loop condition.
 */
#define TM_88796_MAX_POLL_LOOP  2
#define TM_88796_MAX_RECV_LOOP  64

/* Base I/O port address of the AX88796 on the RCM3200 */
#define TM_88796_BASE_ADDRESS   0x4000

/*
 * These defines are all for Softools C for Rabbit.
 * Change them in accordance with your compiler.
 */
#ifdef __SCRabbit
/* Read a byte from a device register */
#define tm_in_dev_byte(r) _ine(TM_88796_BASE_ADDRESS + (int)(r))
/* Write a byte to a device register */
#define tm_out_dev_byte(r, b) _oute(TM_88796_BASE_ADDRESS + (int)(r), (b))
/* No operation */
#define tm_88796_noop() nop()
/* Millisecond delay */
#define tm_88796_delay(ms) delay(ms)
#else /* !__SCRabbit */
#error "Compiler dependent macros must be defined"
#endif /* __SCRabbit */


/*
 * Pre-Processor Definitions (should not have to be modified)
 */

/* The length (in bytes) of an ethernet address */
#define TM_88796_ETHER_ADDR_LEN TM_ETHERNET_PHY_ADDR_LEN
/* The length (in bytes) of the multicast register(s) */
#define TM_88796_MCAST_ADDR_LEN 8

/* The CRC32 reflected polynomial (used for multicast hashing) */
#define TM_88796_CRC32_POLY 0xEDB88320UL

/* The length of a buffer page */
#define TM_88796_BUFFER_PAGE_LEN 256

/* Minimum allowable size for a received packet */
#define TM_88796_MIN_ETHER_PACKET 60
/* Maximum allowable size for a received packet */
#define TM_88796_MAX_ETHER_PACKET 1518

/*
 * Device transmit and receive buffer limits.
 * Total memory available for packets in the device is 16Kbytes.
 * The internal device memory space has a logical page size of 256 bytes.
 * Physical memory starts at an internal offset of 0x4000 (page 0x40).
 * A packet occupies at most 6 pages (1536 bytes).
 */
#define TM_88796_BUFFER_START   0x40
#define TM_88796_BUFFER_STOP    0x80
#define TM_88796_TX_PAGES       (TM_88796_TX_PACKETS * 6)
#define TM_88796_RX_PAGES       (0x40 - TM_88796_TX_PAGES)
#define TM_88796_TXBUF_START    TM_88796_BUFFER_START
#define TM_88796_TXBUF_STOP     (TM_88796_TXBUF_START + TM_88796_TX_PAGES)
#define TM_88796_RXBUF_START    TM_88796_TXBUF_STOP
#define TM_88796_RXBUF_STOP     TM_88796_BUFFER_STOP

/* I/O port offsets */
#define TM_88796_DATAPORT   0x10    /* Data Port                        */
#define TM_88796_RESET      0x1f    /* Reset Port                       */

/*
 * Page-independent registers
 */
#define TM_88796_CR         0x00    /* Command register                 */
#define TM_88796_GPI        0x17    /* General Purpose Input            */

/*
 * Page 0 Read-Only Registers
 */
#define TM_88796_PG0_TSR    0x04    /* Transmit Status Register         */
#define TM_88796_PG0_NCR    0x05    /* Number of Collision Regisger     */
#define TM_88796_PG0_CURR   0x06    /* Current Page Register            */
#define TM_88796_PG0_CRDA0  0x08    /* Current Remote DMA Address 0     */
#define TM_88796_PG0_CRDA1  0x09    /* Current Remote DMA Address 1     */
#define TM_88796_PG0_RSR    0x0c    /* Receive Status Register          */

/*
 * Page 0 Read/Write Registers
 */
#define TM_88796_PG0_PSTART 0x01    /* Page Start Register              */
#define TM_88796_PG0_PSTOP  0x02    /* Page Stop Register               */
#define TM_88796_PG0_BNRY   0x03    /* Boundary Pointer                 */
#define TM_88796_PG0_ISR    0x07    /* Interrupt Status Register        */

/*
 * Page 0 Write-Only Registers
 */
#define TM_88796_PG0_TPSR   0x04    /* Transmit Page Start Address      */
#define TM_88796_PG0_TBCR0  0x05    /* Transmit Byte Count Register 0   */
#define TM_88796_PG0_TBCR1  0x06    /* Transmit Byte Count Register 1   */
#define TM_88796_PG0_RSAR0  0x08    /* Remote Start Address Register 0  */
#define TM_88796_PG0_RSAR1  0x09    /* Remote Start Address Register 1  */
#define TM_88796_PG0_RBCR0  0x0a    /* Remote Byte Count Register 0     */
#define TM_88796_PG0_RBCR1  0x0b    /* Remote Byte Count Register 1     */
#define TM_88796_PG0_RCR    0x0c    /* Receive Configuration Register   */
#define TM_88796_PG0_TCR    0x0d    /* Transmit Configuration Register  */
#define TM_88796_PG0_DCR    0x0e    /* Data Configuration Register      */
#define TM_88796_PG0_IMR    0x0f    /* Interrupt Mask Register          */

/*
 * Page 1 Registers
 */
#define TM_88796_PG1_PAR0   0x01    /* Physical Address Register 0      */
#define TM_88796_PG1_PAR1   0x02    /* Physical Address Register 1      */
#define TM_88796_PG1_PAR2   0x03    /* Physical Address Register 2      */
#define TM_88796_PG1_PAR3   0x04    /* Physical Address Register 3      */
#define TM_88796_PG1_PAR4   0x05    /* Physical Address Register 4      */
#define TM_88796_PG1_PAR5   0x06    /* Physical Address Register 5      */
#define TM_88796_PG1_CURR   0x07    /* Current Page Register            */
#define TM_88796_PG1_MAR0   0x08    /* Multicast Address Register 0     */
#define TM_88796_PG1_MAR1   0x09    /* Multicast Address Register 1     */
#define TM_88796_PG1_MAR2   0x0a    /* Multicast Address Register 2     */
#define TM_88796_PG1_MAR3   0x0b    /* Multicast Address Register 3     */
#define TM_88796_PG1_MAR4   0x0c    /* Multicast Address Register 4     */
#define TM_88796_PG1_MAR5   0x0d    /* Multicast Address Register 5     */
#define TM_88796_PG1_MAR6   0x0e    /* Multicast Address Register 6     */
#define TM_88796_PG1_MAR7   0x0f    /* Multicast Address Register 7     */

/* Definitions for the Command Register (CR) */
#define TM_88796_CR_STP     0x01    /* Stop                             */
#define TM_88796_CR_STA     0x02    /* Start                            */
#define TM_88796_CR_TXP     0x04    /* Transmit Packet                  */
#define TM_88796_CR_RD0     0x08    /* Remote DMA Command 0             */
#define TM_88796_CR_RD1     0x10    /* Remote DMA Command 1             */
#define TM_88796_CR_RD2     0x20    /* Remote DMA Command 2             */
#define TM_88796_CR_PS0     0x40    /* Page Select 0                    */
#define TM_88796_CR_PS1     0x80    /* Page Select 1                    */

/* Definitions for the Interrupt Status Register (ISR) */
#define TM_88796_ISR_PRX    0x01    /* Packet Received                  */
#define TM_88796_ISR_PTX    0x02    /* Packet Transmitted               */
#define TM_88796_ISR_RXE    0x04    /* Receive Error                    */
#define TM_88796_ISR_TXE    0x08    /* Transmit Error                   */
#define TM_88796_ISR_OVW    0x10    /* Overwrite Warning                */
#define TM_88796_ISR_CNT    0x20    /* Counter Overflow                 */
#define TM_88796_ISR_RDC    0x40    /* Remote DMA Complete              */
#define TM_88796_ISR_RST    0x80    /* Reset Status                     */

#define TM_88796_ISR_SELECT ( TM_88796_ISR_PRX \
                            | TM_88796_ISR_PTX \
                            | TM_88796_ISR_RXE \
                            | TM_88796_ISR_TXE \
                            | TM_88796_ISR_OVW )

/* Definitions for the Data Configuration Register (DCR) */
#define TM_88796_DCR_WTS    0x01    /* Word Transfer Select             */
#define TM_88796_DCR_BOS    0x02    /* Byte Order Select                */
#define TM_88796_DCR_LAS    0x04    /* Long Address Select              */
#define TM_88796_DCR_LS     0x08    /* Loopback Select                  */
#define TM_88796_DCR_AR     0x10    /* Auto-Initialize Remote           */
#define TM_88796_DCR_FT0    0x20    /* FIFO Threshhold Select 0         */
#define TM_88796_DCR_FT1    0x40    /* FIFO Threshhold Select 1         */

/* Definitions for the Receive Configuration Register (RCR) */
#define TM_88796_RCR_SEP    0x01    /* Save Errored Packets             */
#define TM_88796_RCR_AR     0x02    /* Accept Runt Packets              */
#define TM_88796_RCR_AB     0x04    /* Accept Broadcast                 */
#define TM_88796_RCR_AM     0x08    /* Accept Multicast                 */
#define TM_88796_RCR_PRO    0x10    /* Promiscuous Physical             */
#define TM_88796_RCR_MON    0x20    /* Monitor Mode                     */

/* Definitions for the Transmit Configuration Register (TCR) */
#define TM_88796_TCR_CRC    0x01    /* Inhibit CRC                      */
#define TM_88796_TCR_LB0    0x02    /* Encoded Loopback Control 0       */
#define TM_88796_TCR_LB1    0x04    /* Encoded Loopback Control 1       */
#define TM_88796_TCR_ATD    0x08    /* Auto Transmit Disable            */
#define TM_88796_TCR_OFST   0x10    /* Collision Offset Enable          */

/* Definitions for the Receive Status Register (RSR) */
#define TM_88796_RSR_PRX    0x01    /* Packet Received Intact           */
#define TM_88796_RSR_CRC    0x02    /* CRC Error                        */
#define TM_88796_RSR_FAE    0x04    /* Frame Alignment Error            */
#define TM_88796_RSR_FO     0x08    /* FIFO Overrun                     */
#define TM_88796_RSR_MPA    0x10    /* Missed Packet                    */
#define TM_88796_RSR_PHY    0x20    /* Physical/Multicast Address       */
#define TM_88796_RSR_DIS    0x40    /* Receiver Disabled                */
#define TM_88796_RSR_DFR    0x80    /* Deferring                        */

/* Definitions for the Transmit Status Register (TSR) */
#define TM_88796_TSR_PTX    0x01    /* Packet Transmitted               */
#define TM_88796_TSR_COL    0x04    /* Transmit Collided                */
#define TM_88796_TSR_ABT    0x08    /* Transmit Aborted                 */
#define TM_88796_TSR_CRS    0x10    /* Carrier Sense Lost               */
#define TM_88796_TSR_FU     0x20    /* FIFO Underrun                    */
#define TM_88796_TSR_CDH    0x40    /* CD Heartbeat                     */
#define TM_88796_TSR_OWC    0x80    /* Out of Window Collision          */

/* Definitions for the General Purpose Input Register (GPI) */
#define TM_88796_GPI_I_LINK     0x01    /* PHY link up                  */
#define TM_88796_GPI_I_DUPLEX   0x02    /* PHY full duplex              */
#define TM_88796_GPI_I_SPEED    0x04    /* PHY speed = 100 Mbps         */

/* Definitions for the received frame header status */
#define TM_88796_HDR_PRX        0x01    /* Packet Received Intact       */
#define TM_88796_HDR_CRC        0x02    /* CRC Error                    */
#define TM_88796_HDR_FAE        0x04    /* Frame Alignment Error        */
#define TM_88796_HDR_MII        0x08    /* MII Error                    */
#define TM_88796_HDR_RUNT       0x10    /* Runt Packet                  */
#define TM_88796_HDR_MCBC       0x20    /* Multicast/Broadcast Address  */


/*
 * Structure Definitions
 */

/*
 * Treck packet information for queuing.
 */
typedef struct tsPktInfo
{
    char TM_FAR *   data;
    int             length;
    ttUserBuffer    buffer;
} ttPktInfo;
typedef ttPktInfo TM_FAR * ttPktInfoPtr;

/*
 * This structure contains all of the information we need to track for
 * the device.
 */
typedef struct ts88796Dev
{
/* The interface handle */
    ttUserInterface ifHandle;
/* The Ethernet address we read out of the hardware */
    unsigned char   ethAddr[TM_ETHERNET_PHY_ADDR_LEN];
/* Boolean value indicating that the physical link is ready */
    unsigned char   linkIsUp;
/* Receive packet queue */
    int             rqAdd, rqRemove, rqCount;
    ttPktInfo       rqPackets[TM_88796_RXQ_SIZE];
/* Transmit packet queue */
    int             tqAdd, tqRemove, tqCount;
    ttPktInfo       tqPackets[TM_88796_TXQ_SIZE];
/* Transmit ring buffer management */
    int             tbAdd, tbCount;
/* Device counters */
    tt88796Stats    stats;
} tt88796Dev;
typedef tt88796Dev TM_FAR * tt88796DevPtr;


/*
 * Global Variables
 */

static tt88796Dev tl88796Mem;
static tt88796DevPtr tl88796Ptr;


/*
 * Forward Definitions
 */


static void tf88796BlockInput(unsigned int     readLen,
                             unsigned int     currOffset,
                             unsigned char TM_FAR * rxBuffer);

static void tf88796BlockOutput(unsigned int     length,
                             unsigned int     address,
                             unsigned char TM_FAR * buffer);

static int tf88796Init(void);
static void tf88796Dequeue(void);
static int tf88796PollReceive(void);
static int tf88796PollTransmit(void);


/*
 * Code
 */

/*
 * tf88796Open()
 *
 *   Configure and start up the Ethernet interface.
 *   We take the address from the Ethernet board and set up the board.
 *   Parameters:
         interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       error:            TM_DEV_ERROR
 *       success:          TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Open(ttUserInterface interfaceHandle)
{
    int     retCode;

    retCode = TM_DEV_ERROR;
    tl88796Ptr = (tt88796DevPtr)0;

/* Did the user turn on the transmit queue? */
    if (tfInterfaceXmitQueueOn(interfaceHandle))
    {
        tfKernelError("tf88796Open",
                        "Transmit queue not supported on this device");
        goto OpenEnd;
    }

/* Make sure that scatter sends are disabled */
    if (tfInterfaceScatterCount(interfaceHandle) != 1)
    {
        tfKernelError("tf88796Open",
                        "Scatter send not supported on this device");
        goto OpenEnd;
    }

/* Assign memory for the tt88796Dev structure */
    tl88796Ptr = &tl88796Mem;

/* Initialize device state information */
    tm_bzero( tl88796Ptr, sizeof(tt88796Dev) );
    tl88796Ptr->ifHandle = interfaceHandle;

/* Initialize hardware */
    retCode = tf88796Init();
    if (retCode != TM_DEV_OKAY)
    {
        tl88796Ptr = (tt88796DevPtr)0;
    }

OpenEnd:
    return retCode;
}


/*
 * tf88796Init()
 *
 *  Initialize the Ethernet chip.
 *  This chip is an integrated MAC and PHY transceiver.
 *  After issuing a reset to the device the PHY will automatically negotiate
 *  the best link speed and duplex for the physical connection. This may take
 *  a few seconds during which time no packets can be sent or received.
 *  While the PHY is in autonegotiation, the MAC registers can be set up
 *  to reflect the desired MAC address and transmit/receive buffer sizes.
 *
 *  Parameters:
 *      interfaceHandle     The user interface handle for this device.
 *  Returns:
 *      error:              TM_DEV_ERROR
 *      success:            TM_DEV_OKAY
 */
static int tf88796Init(void)
{
    int         retCode;
    int         counter;

    retCode = TM_DEV_ERROR;

/* Reset the card and wait until reset is complete. */
    tm_in_dev_byte(TM_88796_RESET);
    tm_88796_delay(2);

#ifdef __SCRabbit
/* Read in the Ethernet address from the Rabbit RCM3200 ID block */
    memcpy(tl88796Ptr->ethAddr, idBlock->macAddr[0], TM_ETHERNET_PHY_ADDR_LEN);
#else /* !__SCRabbit */
#error "Cannot determine MAC address"
#endif /* __SCRabbit */

/* Access page 0 register bank */
    tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_STP);
/* Byte-wide transfers */
    tm_out_dev_byte(TM_88796_PG0_DCR, 0);
/* Clear Count Registers */
    tm_out_dev_byte(TM_88796_PG0_RBCR0, 0);
    tm_out_dev_byte(TM_88796_PG0_RBCR1, 0);
/* Accept broadcast and multicast packets */
    tm_out_dev_byte(TM_88796_PG0_RCR, TM_88796_RCR_AB | TM_88796_RCR_AM);
/* Set transmit to internal loopback mode */
    tm_out_dev_byte(TM_88796_PG0_TCR, TM_88796_TCR_LB0);
/* Set the receive page limits */
    tm_out_dev_byte(TM_88796_PG0_PSTART, TM_88796_RXBUF_START);
    tm_out_dev_byte(TM_88796_PG0_PSTOP, TM_88796_RXBUF_STOP);
    tm_out_dev_byte(TM_88796_PG0_BNRY, TM_88796_RXBUF_START);

/* Change to Page 1 */
    tm_out_dev_byte(TM_88796_CR,
                      TM_88796_CR_RD2 + TM_88796_CR_PS0 + TM_88796_CR_STP);
/* Set physical address & disable multicast */
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_out_dev_byte(TM_88796_PG1_PAR0 + counter,
                          tl88796Ptr->ethAddr[counter]);
    }
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_out_dev_byte(TM_88796_PG1_MAR0 + counter, 0);
    }

/* Set the current receive page */
    tm_out_dev_byte(TM_88796_PG1_CURR, TM_88796_RXBUF_START + 1);

/* Change to Page 0 and start the device */
    tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_STA);
/* Clear all device interrupts */
    tm_out_dev_byte(TM_88796_PG0_ISR, 0xff);
/* Enable transmit */
    tm_out_dev_byte(TM_88796_PG0_TCR, 0);

/* Indicate success */
    retCode = TM_DEV_OKAY;

InitEnd:
    return retCode;
}


/*
 * tf88796Close()
 *
 *   Shut down the chip.
 *   Parameters:
 *       interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Close (ttUserInterface interfaceHandle)
{
/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Stop chip and set it to Page 0 */
    tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_STP);

    if (tl88796Ptr)
    {
        /* Release locally queued packets */
        tf88796Dequeue();

        tl88796Ptr = (tt88796DevPtr)0;
    }

    return(TM_DEV_OKAY);
}


/*
 * tf88796Dequeue()
 *
 *   Release all packets held in the driver's queues.
 *   Parameters:
 *       none
 *   Returns:
 *       none
 */
static void tf88796Dequeue(void)
{
    ttPktInfoPtr    pktInfoPtr;
    int             i, n;

    /* Release all packets held on the receive queue */
    n = tl88796Ptr->rqCount;
    if (n > 0)
    {
        pktInfoPtr = tl88796Ptr->rqPackets;
        i = tl88796Ptr->rqRemove;
        do
        {
            tfFreeDriverBuffer(pktInfoPtr[i].buffer);
            if (++i >= TM_88796_RXQ_SIZE)
                i = 0;
        } while (--n);

        tl88796Ptr->rqRemove = i;
        tl88796Ptr->rqCount = 0;
    }

    /* Release all packets held on the transmit queue */
    n = tl88796Ptr->tqCount;
    if (n > 0)
    {
        i = tl88796Ptr->tqRemove;
        do
        {
            tfSendCompleteInterface(tl88796Ptr->ifHandle,
                        TM_DEV_SEND_COMPLETE_DRIVER);
            if (++i >= TM_88796_TXQ_SIZE)
                i = 0;
        } while (--n);

        tl88796Ptr->tqRemove = i;
        tl88796Ptr->tqCount = 0;
    }
}


/*
 * tf88796Poll()
 *
 *   This function replaces the interrupt handler. It polls the device for
 *   receive, transmit-complete, and receive overflow warning events.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       none
 *   Returns:
 *       none
 */
static int tf88796Poll(ttUserInterface interfaceHandle)
{
/* Used to store the state of the ISR */
    unsigned char   interrupts;
    unsigned char   loops;
    unsigned char   u8;
    int             retCode;
    int             resend;

/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);

    retCode = TM_DEV_OKAY;

/* Sanity check */
    if (!tl88796Ptr)
    {
        retCode = TM_DEV_ERROR;
        goto PollEnd;
    }

/* Bounded loop while there are events to service */
    for (loops = 0; loops < TM_88796_MAX_POLL_LOOP; ++loops)
    {
/* Read and clear interrupts; exit loop if there is nothing to do. */
        interrupts = tm_in_dev_byte(TM_88796_PG0_ISR);
        tm_out_dev_byte(TM_88796_PG0_ISR, interrupts);
        if ((interrupts & TM_88796_ISR_SELECT) == 0)
            break;

        ++tl88796Ptr->stats.intevent;

/* Process receive error */
        if (interrupts & TM_88796_ISR_RXE)
        {
            ++tl88796Ptr->stats.rxerror;

            /* Decode receive status register */
            u8 = tm_in_dev_byte(TM_88796_PG0_RSR);

            if (u8 & TM_88796_RSR_CRC)
            {
                ++tl88796Ptr->stats.rxcrc;
            }
            if (u8 & TM_88796_RSR_FAE)
            {
                ++tl88796Ptr->stats.rxalign;
            }
            if (u8 & TM_88796_RSR_MPA)
            {
                ++tl88796Ptr->stats.rxmissed;
            }
            if (u8 & TM_88796_RSR_DIS)
            {
                ++tl88796Ptr->stats.rxdisabled;
            }
        }

/* Process received frames */
        if (interrupts & TM_88796_ISR_PRX)
        {
            ++tl88796Ptr->stats.rxint;

            /* Move received packets from the device and to the queue */
            retCode = tf88796PollReceive();
            if (retCode != TM_DEV_OKAY)
            {
                goto PollEnd;
            }
        }

/* Process buffer overrun */
        if (interrupts & TM_88796_ISR_OVW)
        {
            ++tl88796Ptr->stats.overrun;

            /* Record the state of the TXP bit in the command register */
            u8 = tm_in_dev_byte(TM_88796_CR);

            /* Stop the device and wait for at least 1.5ms */
            tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_STP);
            tm_88796_delay(3);

            /* Clear remote byte count */
            tm_out_dev_byte(TM_88796_PG0_RBCR0, 0);
            tm_out_dev_byte(TM_88796_PG0_RBCR1, 0);

            /* Check for a "resend" condition */
            resend = 0;
            if (u8 & TM_88796_CR_TXP)
            {
                u8 = tm_in_dev_byte(TM_88796_PG0_ISR);
                if ((u8 & (TM_88796_ISR_PTX | TM_88796_ISR_TXE)) == 0)
                    resend = 1;
            }

            /* Enter internal loopback mode */
            tm_out_dev_byte(TM_88796_PG0_TCR, TM_88796_TCR_LB0);

            /* Start the device */
            tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_STA);

            /* Clear overrun condition */
            tm_out_dev_byte(TM_88796_PG0_ISR, TM_88796_ISR_OVW);

            /* Exit loopback mode */
            tm_out_dev_byte(TM_88796_PG0_TCR, 0);

            /* Resend, if necessary */
            if (resend)
            {
                tm_out_dev_byte(TM_88796_CR,
                        TM_88796_CR_RD2 + TM_88796_CR_STA + TM_88796_CR_TXP);
            }
        }

/* Process transmit error */
        if (interrupts & TM_88796_ISR_TXE)
        {
            ++tl88796Ptr->stats.txerror;

            /* Decode transmit status register */
            u8 = tm_in_dev_byte(TM_88796_PG0_TSR);

            if (u8 & TM_88796_TSR_COL)
            {
                ++tl88796Ptr->stats.txcollision;
            }
            if (u8 & TM_88796_TSR_ABT)
            {
                ++tl88796Ptr->stats.txabort;
            }
            if (u8 & TM_88796_TSR_OWC)
            {
                ++tl88796Ptr->stats.txwindow;
            }
        }

/* Process transmit packet event */
        if (interrupts & TM_88796_ISR_PTX)
        {
            ++tl88796Ptr->stats.txint;

            /* Make room in the device for another packet */
            if (tl88796Ptr->tbCount > 0)
                --tl88796Ptr->tbCount;

            /* Move packets from the send queue to the device */
            retCode = tf88796PollTransmit();
            if (retCode != TM_DEV_OKAY)
            {
                goto PollEnd;
            }
        }
    }

PollEnd:
    return(retCode);
}


/*
 * tf88796BlockInput()
 *
 *   This function transfers data from device memory to
 *   standard memory.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       readLen    Number of bytes to transfer
 *       currOffset The place to begin reading from in the internal buffers
 *       rxBuffer   The buffer to copy into
 *   Returns:
 *       none
 */
static void tf88796BlockInput(unsigned int     readLen,
                              unsigned int     currOffset,
                              unsigned char TM_FAR * rxBuffer)
{
/* Store the packet length */
    tm_out_dev_byte(TM_88796_PG0_RBCR0, readLen & 0xFF);
    tm_out_dev_byte(TM_88796_PG0_RBCR1, readLen >> 8);
/* Store the address to read from */
    tm_out_dev_byte(TM_88796_PG0_RSAR0, currOffset & 0xFF);
    tm_out_dev_byte(TM_88796_PG0_RSAR1, currOffset >> 8);
/* Issue a Remote Read DMA command */
    tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD0);

    while (readLen--)
    {
/*
 * Get 8 bits from the dataport & save it to the appropriate place in the
 * buffer
 */
        *rxBuffer++ = tm_in_dev_byte(TM_88796_DATAPORT);
    }
}


/*
 * tf88796BlockOutput()
 *
 *   This function transfers data from standard memory to device
 *   memory.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       length     Number of bytes to transfer
 *       address    Address in the device memory
 *       buffer     The buffer address
 *   Returns:
 *       none
 */
static void tf88796BlockOutput(unsigned int     length,
                               unsigned int     address,
                               unsigned char TM_FAR * buffer)
{
    unsigned char    byte;

/* Store the length to write */
    tm_out_dev_byte(TM_88796_PG0_RBCR0, length & 0xFF);
    tm_out_dev_byte(TM_88796_PG0_RBCR1, length >> 8);
/* Store the address to write to */
    tm_out_dev_byte(TM_88796_PG0_RSAR0, address & 0xFF);
    tm_out_dev_byte(TM_88796_PG0_RSAR1, address >> 8);
/* Issue a Remote Write DMA command */
    tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD1);

    while (length--)
    {
/* Write the next byte to the port */
        byte = *buffer++;
        tm_out_dev_byte(TM_88796_DATAPORT, byte);
    }
}


/*
 * tf88796PollReceive()
 *
 *  This function checks for received frames in the device.
 *  If found, received frames are extracted from the device
 *  and added to a locally maintained receive queue.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *  Parameters:
 *      none
 *  Returns:
 *      error:      TM_DEV_ERROR
 *      success:    TM_DEV_OKAY
 */
static int tf88796PollReceive(void)
{
    unsigned char       page;
    unsigned char       header[4];
    unsigned int        addr;
    int                 length, extra, limit;
    char TM_FAR *       bufPtr;
    ttUserBuffer        bufHandle;
    ttPktInfoPtr        pktInfoPtr;
    int                 loops;
    int                 retCode;
    int                 rqAdd, rqCount;

    retCode = TM_DEV_OKAY;

    /* Local copy of receive queue information */
    pktInfoPtr = tl88796Ptr->rqPackets;
    rqAdd = tl88796Ptr->rqAdd;
    rqCount = tl88796Ptr->rqCount;

    /* Get the location of the next frame to retrieve */
    page = tm_in_dev_byte(TM_88796_PG0_BNRY) + 1;
    if (page >= TM_88796_RXBUF_STOP)
        page = TM_88796_RXBUF_START;

    /*
     * Extract all received frames from the device, even if there is
     * not enough memory or queue space for the received data.
     * Maintain a loop count to allow an exit if there is a serious
     * problem.
     */
    for (loops = 0; ; )
    {
        /* Exit if there are no new frames available */
        if (page == tm_in_dev_byte(TM_88796_PG0_CURR))
            break;

        /* Avoid an infinite loop if there is a serious problem */
        if (++loops >= TM_88796_MAX_RECV_LOOP)
        {
            retCode = TM_DEV_ERROR;
            break;
        }

        /* Page address */
        addr = page * TM_88796_BUFFER_PAGE_LEN;

        /* Retrieve the frame header */
        tf88796BlockInput(sizeof(header), addr, header);

        /* Start of frame in device memory and frame length */
        addr += sizeof(header);
        length = 256 * (int)header[3] + (int)header[2];

        /* Sanity check */
        if ((header[0] & 0x1f) == 0
        ||  header[1] < TM_88796_RXBUF_START
        ||  header[1] >= TM_88796_RXBUF_STOP
        ||  length < TM_88796_MIN_ETHER_PACKET
        ||  length > TM_88796_MAX_ETHER_PACKET)
        {
            /* Invalid header information */
            retCode = TM_DEV_ERROR;
            break;
        }

        /* Record statistics */
        if (header[0] & TM_88796_HDR_PRX)
            ++tl88796Ptr->stats.rxgood;
        if (header[0] & TM_88796_HDR_MCBC)
            ++tl88796Ptr->stats.rxmcbc;

        /* Check if there is room in the receive packet queue */
        if (rqCount >= TM_88796_RXQ_SIZE)
        {
            ++tl88796Ptr->stats.rxqovf;
        }
        else
        {
            /* Get a buffer for the data */
            bufPtr = tfGetDriverBuffer(&bufHandle, length, 2);
            if (!bufPtr)
            {
                ++tl88796Ptr->stats.rxnobuf;
            }
            else
            {
                /* Distance to end of receive memory */
                limit = TM_88796_BUFFER_PAGE_LEN * TM_88796_RXBUF_STOP - addr;

                /* Check for frame wraparound */
                extra = length - limit;
                if (extra > 0)
                {
                    /* Frame is split in two */
                    tf88796BlockInput(limit, addr, bufPtr);
                    tf88796BlockInput(extra,
                            TM_88796_RXBUF_START * TM_88796_BUFFER_PAGE_LEN,
                            bufPtr + limit);
                }
                else
                {
                    /* Frame is contiguous */
                    tf88796BlockInput(length, addr, bufPtr);
                }

                /* Add packet to receive queue */
                pktInfoPtr[rqAdd].buffer = bufHandle;
                pktInfoPtr[rqAdd].data = bufPtr;
                pktInfoPtr[rqAdd].length = length;
                ++rqCount;
                if (++rqAdd >= TM_88796_RXQ_SIZE)
                {
                    rqAdd = 0;
                }
            }
        }

        /* Link to next frame page. Free up receive memory. */
        page = header[1];
        if (page == TM_88796_RXBUF_START)
        {
            tm_out_dev_byte(TM_88796_PG0_BNRY, TM_88796_RXBUF_STOP - 1);
        }
        else
        {
            tm_out_dev_byte(TM_88796_PG0_BNRY, page - 1);
        }
    }

    /* Number of packets added to receive queue */
    loops = rqCount - tl88796Ptr->rqCount;

    /* Update device state information */
    tl88796Ptr->rqAdd = rqAdd;
    tl88796Ptr->rqCount = rqCount;

    /* Notify Treck of the number of packets we have received */
    if (loops)
    {
        tfNotifyInterfaceTask(tl88796Ptr->ifHandle, loops, 0, 0UL, 0UL);
    }

    return retCode;
}



/*
 * tf88796Receive()
 *
 *   This function passes on received packets to the Treck stack.
 *   It is called after it has been determined that received frames
 *   exist in the device and the Treck stack has been notified of
 *   how many frames have been received.
 *   The received frames have already been extracted from the device
 *   and added to a local queue maintained by this driver.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         We fill in this pointer with a pointer to the data
 *                       we have just received
 *       dataLength      We fill in this length with the length of the data
 *                       we have just received
 *       ttUserBufferPtr We fill out this buffer handle with the buffer handle
 *                       we get back when we allocate memory
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
TM_NEARCALL int tf88796Receive(ttUserInterface        interfaceHandle,
                               char TM_FAR * TM_FAR * dataPtr,
                               int TM_FAR *           dataLength,
                               ttUserBufferPtr        bufHandlePtr)
{
/* The return value */
    int             retCode;
    ttPktInfoPtr    pktInfoPtr;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Assume that we return OKAY unless otherwise instructed */
    retCode = TM_DEV_OKAY;

/* Sanity check */
    if (!tl88796Ptr || tl88796Ptr->rqCount < 1)
    {
        retCode = TM_DEV_ERROR;
        goto RecvEnd;
    }

/* Remove the next packet from the receive queue */
    pktInfoPtr = tl88796Ptr->rqPackets + tl88796Ptr->rqRemove;
    if (++tl88796Ptr->rqRemove >= TM_88796_RXQ_SIZE)
        tl88796Ptr->rqRemove = 0;
    --tl88796Ptr->rqCount;

/* Extract the packet information */
    *dataPtr = pktInfoPtr->data;
    *dataLength = pktInfoPtr->length;
    *bufHandlePtr = pktInfoPtr->buffer;

RecvEnd:
    return(retCode);
}


/*
 * tf88796PollTransmit()
 *
 *  This function moves packets that are queued for send into the device.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *  Parameters:
 *      none
 *  Returns:
 *      error:      TM_DEV_ERROR
 *      success:    TM_DEV_OKAY
 */
static int tf88796PollTransmit(void)
{
    unsigned int        page;
    unsigned int        addr;
    ttPktInfoPtr        pktInfoPtr;
    int                 retCode;

    retCode = TM_DEV_OKAY;

    /* Local copy of receive queue information */
    pktInfoPtr = tl88796Ptr->tqPackets;

    /*
     * Continue to move packets into the device until the queue is empty
     * or the device is full.
     */
    while (tl88796Ptr->tqCount > 0
                        && tl88796Ptr->tbCount < TM_88796_TX_PACKETS)
    {
        /* Next packet to be sent */
        pktInfoPtr = tl88796Ptr->tqPackets + tl88796Ptr->tqRemove;

        /* Copy packet contents to the next available send slot */
        page = tl88796Ptr->tbAdd * 6 + TM_88796_TXBUF_START;
        addr = page * TM_88796_BUFFER_PAGE_LEN;
        tf88796BlockOutput(pktInfoPtr->length, addr, pktInfoPtr->data);

        /* Prepare the device and trigger the send */
        tm_out_dev_byte(TM_88796_PG0_TBCR0, pktInfoPtr->length & 0xff);
        tm_out_dev_byte(TM_88796_PG0_TBCR1, pktInfoPtr->length >> 8);
        tm_out_dev_byte(TM_88796_PG0_TPSR, page);
        tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_TXP);

        /* Count packets sent */
        ++tl88796Ptr->stats.transmit;

        /* One packet added to the device */
        ++tl88796Ptr->tbCount;
        if (++tl88796Ptr->tbAdd >= TM_88796_TX_PACKETS)
            tl88796Ptr->tbAdd = 0;

        /* One packet removed from the send queue */
        --tl88796Ptr->tqCount;
        if (++tl88796Ptr->tqRemove >= TM_88796_TXQ_SIZE)
            tl88796Ptr->tqRemove = 0;

        /* Release ownership of the packet */
        tfSendCompleteInterface(tl88796Ptr->ifHandle,
                        TM_DEV_SEND_COMPLETE_DRIVER);
    }

    return retCode;
}


/*
 * tf88796Send()
 *
 *   This function stores a packet in the local send queue and attempts
 *   to send it.
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         The pointer to the data we're sending
 *       dataLength      The length of the data we're sending
 *       flag            A flag indicating various things about the data to
                         send
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
TM_NEARCALL int tf88796Send ( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag )
{
    ttPktInfoPtr        pktInfoPtr;
    int                 retCode;

/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(flag);

    retCode = TM_DEV_OKAY;

    /* Sanity check */
    if (!tl88796Ptr)
    {
        retCode = TM_DEV_ERROR;
        goto SendEnd;
    }

    /* Accept packet only if there is room in the transmit queue */
    if (tl88796Ptr->tqCount >= TM_88796_TXQ_SIZE)
    {
        /* Cannot accept packet; Release ownership of the packet */
        ++tl88796Ptr->stats.txqovf;
        tfSendCompleteInterface(tl88796Ptr->ifHandle,
                        TM_DEV_SEND_COMPLETE_DRIVER);
    }
    else
    {
        pktInfoPtr = tl88796Ptr->tqPackets + tl88796Ptr->tqAdd;
        pktInfoPtr->data = dataPtr;
        pktInfoPtr->length = dataLength;
        pktInfoPtr->buffer = 0;

        /* One packet added to the send queue */
        ++tl88796Ptr->tqCount;
        if (++tl88796Ptr->tqAdd >= TM_88796_TXQ_SIZE)
        {
            tl88796Ptr->tqAdd = 0;
        }
    }

    /* Move packets from the queue to the device */
    retCode = tf88796PollTransmit();

SendEnd:
    return retCode;
}


/*
 * tf88796Ioctl()
 *
 *   This function, in general, allows the user to talk directly to the
 *   chip, for any reason. We have implemented it here to allow the
 *   following:
 *      1) set multicast addresses,
 *      2) periodically poll the device to service activity,
 *      3) retrieve device statistics.
 *
 *   flag bit mask 1:
 *      Poll device for service.
 *      'optionPtr' and 'optionLen' are unused.
 *
 *   flag bit mask 2:
 *      Retrieve device statistics:
 *      'optionPtr' points to storage to receive the pointer to the
 *      common statistics structure (tt88796StatsPtr) for the device.
 *      'optionLen' = sizeof(tt88796StatsPtr)
 *
 *   TM_DEV_SET_MCAST_LIST:
 *     Adds 'optionLen' number of multicast entries to the multicast hash
 *     table.  This is done by
 *         1. Compute a 32-bit CRC on the address.
 *         2. Extract the most significant 6 bits of the computed CRC.
 *         3. Set the bit coresponding to the above number in the multicast
 *            address filter (ie, decode the 6 bit value above into a 64 bit
 *            value: 50 -> 0x0002 0000 0000 0000)
 *
 *   TM_DEV_SET_ALL_MCAST:
 *     Sets the chip to receive all multicast packets (ie, set the address
 *     filter to all ones).
 *
 *   On entry, register page 0 must be selected.
 *   On exit, register page 0 will be selected.
 *
 *   Parameters:
 *       interfaceHandle   The handle to the user interface for this device
 *       flag       A flag what we want this function to do
 *       dataPtr    The pointer to the data being passed in
 *       dataLength The length of the data (in this case, how many multicast
 *                  addresses are being added).
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796Ioctl(ttUserInterface interfaceHandle,
                int             flag,
                void TM_FAR *   optionPtr,
                int             optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/* The calculated CRC */
    unsigned long   calcCRC;
/* A counter */
    unsigned char   i;
/* Which byte of the ethernet address we're working on */
    unsigned char   byte;
/* Which bit of the current byte we're working on */
    unsigned char   bit;
/* A pointer to the list of multicast addresses passed to us */
    unsigned char * multiAddr;
/* The actual hash table to be written to the chip */
    unsigned char   logicalAddrFilter[TM_88796_MCAST_ADDR_LEN];
#endif /* TM_IGMP or TM_6_USE_MLD */
    int             retCode;

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

    retCode = TM_DEV_OKAY;

/* Poll for Ethernet events. */
    if (flag & 1)
    {
        retCode = tf88796Poll(interfaceHandle);
        goto IoctlEnd;
    }

/* Dump statistics */
    if (flag & 2)
    {
        if (optionPtr && optionLen >= sizeof(tt88796StatsPtr))
        {
            *(tt88796StatsPtr *)optionPtr = &tl88796Ptr->stats;
        }
    }

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    if (flag & TM_DEV_SET_MCAST_LIST)
    {
/*
 * Since most Ethernet transmission (this device is no exception) starts with
 * the least significant bit of each byte, it makes sense to do the CRC
 * calculation based on a shift right model rather than the traditional
 * shift left model.
 * The advantage is that the input bytes do not have to be bit reversed prior
 * to being merged with the running CRC.
 * The disadvantage is that the resulting CRC has the most significant bit of
 * the CRC in the least significant bit of the running CRC variable and the
 * bits must be reversed upon extraction.
 * It also requires the use of a "reflected" CRC-32 polynomial, i.e. one in
 * which all the bits are in the reverse order.
 * This is preferable since only the 6 most significant bits of the resulting
 * CRC are needed to address into the 64-bit multicast bit array: 3 bits
 * address the the byte offset (0 to 7) and 3 bits address the bit offset
 * (0 to 7) within that byte.
 */

/* Zero out the hash table */
        tm_bzero( logicalAddrFilter, TM_88796_MCAST_ADDR_LEN);

/* Get the first address */
        multiAddr = (unsigned char TM_FAR *) optionPtr;

/* While we have more addresses to work on */
        while (optionLen-- > 0)
        {

/* Initialize the calculated CRC (with all ones) */
            calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
            for (byte = 0; byte < TM_88796_ETHER_ADDR_LEN; byte++)
            {

/* Merge input bits with current CRC */
                calcCRC ^= (unsigned long)*multiAddr++ & 0xFFUL;

/* For each bit in the current byte */
                for (bit = 0; bit < 8; bit++)
                {

/*
 * Shift CRC and, if the bit shifted out is 1, merge the reflected
 * polynomial into the CRC.
 */
                    if (calcCRC & 1)
                        calcCRC = (calcCRC >> 1) ^ TM_88796_CRC32_POLY;
                    else
                        calcCRC = calcCRC >> 1;
                }
            }

/* Extract the most significant CRC bits to get the byte offset (0..7) */
            for (byte = 0, i = 0; i < 3; ++i)
            {
                byte <<= 1;
                byte |= (unsigned char)(calcCRC & 1);
                calcCRC >>= 1;
            }

/*
 * Extract the next most significant CRC bits to get the bit offset
 * (0..7) within the byte.
 */
            for (bit = 0, i = 0; i < 3; ++i)
            {
                bit <<= 1;
                bit |= (unsigned char)(calcCRC & 1);
                calcCRC >>= 1;
            }

/* Set the appropriate bit in the appropriate byte of the array */
            logicalAddrFilter[byte] |= 1 << bit;
        }

/* Set chip to Page 1 */
        tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_PS0);
/* Write the new multicast hash to the multicast address register */
        for (i = 0; i < TM_88796_MCAST_ADDR_LEN; i++)
        {
            tm_out_dev_byte(TM_88796_PG1_MAR0 + i, logicalAddrFilter[i]);
        }

/* Set chip to Page 0 */
        tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2);
    }

    if (flag & TM_DEV_SET_ALL_MCAST)
    {
/* Set chip to Page 1 */
        tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2 + TM_88796_CR_PS0);
/* Set the logical address filter to all ones (receive ALL multicast packets)
*/
        for (i = 0; i < TM_88796_MCAST_ADDR_LEN; i++)
        {
            tm_out_dev_byte(TM_88796_PG1_MAR0 + i, 0xff);
        }

/* Set chip to Page 0 */
        tm_out_dev_byte(TM_88796_CR, TM_88796_CR_RD2);
    }
#endif /* TM_IGMP or TM_6_USE_MLD */

IoctlEnd:
    return(retCode);
}


/*
 * tf88796GetPhyAddr()
 *
 *   This function retrieves the adapter's physical address.
 *
 *   Parameters:
 *       interfaceHandle  The handle to the user interface for this device
 *       physicalAddress A buffer into which to store the physical address
 *   Returns:
 *       TM_DEV_OKAY
 */
TM_NEARCALL int tf88796GetPhyAddr(ttUserInterface interfaceHandle,
                     char TM_FAR *   physicalAddress)
{
/* A counter */
    unsigned char counter;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Copy the address into the buffer */
    for (counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        physicalAddress[counter] = tl88796Ptr->ethAddr[counter];
    }

    return(TM_DEV_OKAY);
}

/***/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tr88796.c     File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 36130    Output checksum:  6406
 */
