/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Crystal LAN (cs8900) Device Driver
 * Filename: trcrystl.c
 * Author: Jason
 * Date Created: 11/9/99
 * $Source: source/drivers/crystal/trcrystl.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:55CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Originally developed on real-mode x86 platform, uC/OS pre-emptive kernel, 
 *   Borland compiler.   
 *
 *
 * Various paramters to set for a particular platform:
 *   TM_CRYSTAL_IOBASE             Base address for I/O with the CS8900
 *   TM_CRYSTAL_IRQ                Interrupt request line to use
 *   TM_CRYSTAL_NUM_RECV_BUF       Number of big receive buffers to allocate
 *   TM_CRYSTAL_NUM_SMALL_RECV_BUF Number of small receive buffers to allocate
 *   TM_CRYSTAL_SMALL_BUFFER_SIZE  size of small receive buffers
 *   TM_CRYSTAL_PHYSINT            Either TM_CRYSTAL_LINECTL_10BASET,
 *                                 TM_CRYSTAL_LINECTL_AUI, or
 *                                 TM_CRYSTAL_LINECTL_AUTOSEL
 *
 * Modifications likely to necessary:
 * Interrupt mechanism - As usual, interrupts vary widely from platform to
 * platform.  This driver was developed on a standard x86 PC (ethernet via
 * ISA) so it uses the standard PIC mechanism.  If this doesn't match your
 * platform, it will be necessary to change this in 3 places: 
 * tfCrystalOpen where the interrupt is installed, tfCrystalIsrHandler,
 * where the interrupt is handled and tfCrystalClose where the interrupt is
 * removed.
 *
 * For more information on the Crystal LAN chip, please see document DS271PP3
 * from Cirrus.
 *
 */

#include <trsocket.h>
/* For tm_kernel_set_critical, tm_kernel_release_critical: */
#include <trmacro.h>
#include <trtype.h>     /* for trproto.h */
#include <trproto.h>    /* for tfFreeSharedBuffer() */
#include <trcrystl.h>

#include <dos.h>        /* Needed for port I/O in Borland */

/* Base I/O Address of cs8900 */
#define TM_CRYSTAL_IOBASE  0x300

/* Default Interrupt Request Line */
#define TM_CRYSTAL_IRQ    5

/*
 * The higher this number, the better.  This driver was initially developed
 * on a x86 real-mode, so this number is very low because I was limited
 * to 64K of data space.
 */
#define TM_CRYSTAL_NUM_RECV_BUF       2
#define TM_CRYSTAL_NUM_SMALL_RECV_BUF 4

#define TM_CRYSTAL_SMALL_BUFFER_SIZE  128

/*
 * This defines the physical interface that the cs8900 will use, either AUI,
 * 10BaseT or Auto-Select.
 */
#define TM_CRYSTAL_PHYSINT TM_CRYSTAL_LINECTL_10BASET
/* #define TM_CRYSTAL_PHYSINT TM_CRYSTAL_LINECTL_AUI */
/* #define TM_CRYSTAL_PHYSINT TM_CRYSTAL_LINECTL_AUTOSEL */

#define TM_FALSE 0
#define TM_TRUE  1

/* 
 * Forces the driver to refill the receive queue inline (ie, on every receive) 
 * rather than only in the IOCTL routine.
 */
#define TM_CRYSTAL_REFILL_INLINE

/* 
 * I/O port addresses 
 */

/* 
 * Programmable Interrupt Controller addresses (Interrupt Status Register,
 * Interrupt Mask Register).
 */
#define TM_PIC1_ISR_ADDR 0x20
#define TM_PIC1_IMR_ADDR 0x21
#define TM_PIC2_ISR_ADDR 0xA0
#define TM_PIC2_IMR_ADDR 0xA1


/* CS8900 Port address (offset from base i/o address) */
/* Receieve/Transmit data to/from the cs8900 */
#define TM_CRYSTAL_RX_DATA      0x00
#define TM_CRYSTAL_TX_DATA      0x00
/* Interrupt status queue */
#define TM_CRYSTAL_ISQ          0x08
/* PacketPage address port */
#define TM_CRYSTAL_PKTPAGE_ADDR 0x0A
/* PacketPage data port */
#define TM_CRYSTAL_PKTPAGE_DATA 0x0C

/* 
 * Packet page registers 
 */
#define TM_CRYSTAL_PP_INTNUM    0x0022
#define TM_CRYSTAL_PP_DMA_NUM   0x0024
#define TM_CRYSTAL_PP_MEMBASE   0x002c
#define TM_CRYSTAL_PP_EEPROMCMD 0x0040
#define TM_CRYSTAL_PP_RXCFG     0x0102
#define TM_CRYSTAL_PP_RXCTL     0x0104
#define TM_CRYSTAL_PP_TXCFG     0x0106
#define TM_CRYSTAL_PP_BUFCFG    0x010A
#define TM_CRYSTAL_PP_LINECTL   0x0112
#define TM_CRYSTAL_PP_SELFCTL   0x0114
#define TM_CRYSTAL_PP_BUSCTL    0x0116
#define TM_CRYSTAL_PP_ISQ       0x0120
#define TM_CRYSTAL_PP_SELFSTAT  0x0136
#define TM_CRYSTAL_PP_BUSSTAT   0x0138
#define TM_CRYSTAL_PP_TXCMD     0x0144
#define TM_CRYSTAL_PP_TXLEN     0x0146
#define TM_CRYSTAL_PP_LAF       0x0150
#define TM_CRYSTAL_PP_HWADDR    0x0158


/* CS8900 Signature (used for chip verification) */
#define TM_CRYSTAL_SIGNATURE 0x3000


/* 
 * Bitfields for various register in the PacketPage structure.  For in depth
 * description, please see databook.
 */
/* Reset command */
#define TM_CRYSTAL_SELFCTL_RESET   0x0040

/* Done initializing, busy, external decoding logic present */
#define TM_CRYSTAL_SELFSTAT_INITD  0x0080
#define TM_CRYSTAL_SELFSTAT_SIBUSY 0x0100
#define TM_CRYSTAL_SELFSTAT_ELPRES 0x0800

/* Media type (10baseT, AUI), transmission/reception enable */
#define TM_CRYSTAL_LINECTL_AUI     0x0080
#define TM_CRYSTAL_LINECTL_10BASET 0x0000
#define TM_CRYSTAL_LINECTL_AUTOSEL 0x0100
#define TM_CRYSTAL_LINECTL_TXON    0x0080
#define TM_CRYSTAL_LINECTL_RXON    0x0040

/* Generate interrupts on good packet, enable DMA */
#define TM_CRYSTAL_RXCFG_OKIE      0x0100
#define TM_CRYSTAL_RXCFG_DMA       0x0200

/* Types of frames to accept: OK, multicast, individual, broadcast */
#define TM_CRYSTAL_RXCTL_ACC_OK    0x0100
#define TM_CRYSTAL_RXCTL_ACC_MCAST 0x0200
#define TM_CRYSTAL_RXCTL_ACC_IND   0x0400
#define TM_CRYSTAL_RXCTL_ACC_BCAST 0x0800

/* Generate interrupts on transmit OK */
#define TM_CRYSTAL_TXCFG_OKIE      0x0100

/* Enable Interrupts, memory mapped mode */
#define TM_CRYSTAL_BUSCTL_INT_ENB  0x8000
#define TM_CRYSTAL_BUSCTL_MEMMODE  0x0400

/* Start transmission of packet */
#define TM_CRYSTAL_TXCMD_TXSTART   0x00C0

/* Chip ready for transmit */
#define TM_CRYSTAL_BUSSTAT_TXRDY   0x0100

/* Generate interrupt on DMA transfer complete */
#define TM_CRYSTAL_BUFCFG_DMAIE    0x0080

/* 'Packet receieved' event */
#define TM_CRYSTAL_ISR_RXEVENT     0x0004

/* Enable external decoding logic */
#define TM_CRYSTAL_EEPROMCMD_ELSEL 0x0400

/* Send allocation request states */
#define TM_CRYSTAL_ALLOC_IDLE    0
#define TM_CRYSTAL_ALLOC_WAITING 1


/* Write out a word for specified port address */
#define tm_out_word(tp_value,tp_port)    outpw(tp_port,tp_value)

/* Write out a byte to specified port address */
#define tm_out_byte(tp_value,tp_port)    outp(tp_port,tp_value)

/* Read in a word from specified port address */
#define tm_in_word(tp_port)  inpw(tp_port)

/* Read in a byte from specified port address */
#define tm_in_byte(tp_port)  inp(tp_port)

/* Translate a segment/offset into an absolute address */
#define tm_get_abs_address(origAddress) \
            (((unsigned long)(origAddress)&0xffff0000UL)>>12) + \
            (((unsigned long)(origAddress))&0x0000ffffUL)

/* Polynomial used for 32-bit CRC calculation (used for multicast addresses) */
#define TM_CRC32_POLY 0x04c11db7UL

/* The CS8900's Interrupt Request Line */
static unsigned char tlCrystalIrq;

/* Treck interface handle for this cs8900 */
static ttUserInterface tlCrystalInterfaceHandle;

/* Indicate whether the user is using a Treck transmit queue */
static int tlCrystalXmitQueueOn;

/*
 * Indicates the status of a transmit buffer allocation (either idle, or 
 * waiting).  This is only used if the transmit queue is enabled.
 */
static int tlCrystalAllocState;

/* Counts the number of missed frames due to lack of buffer space */
unsigned long tlCrystalMissedFrames;
/*
 * Counts the number of calls to driver recv when there is no buffer ready
 * to be received
 */
unsigned long tlCrystalSpuriousInterrupts;

/*
 * Internal function prototypes
 */
/* Read 16-bit value from CS8900's PacketPage structure */
static unsigned short tfCrystalReadPktPage(unsigned short addr);

/* Writes 16-bit value to CS8900's PacketPage structure */
static void tfCrystalWritePktPage(unsigned short addr, unsigned short value);

/* CS8900 ISR handler */
static void TM_INTERRUPT tfCrystalIsrHandler(void);

/*
 * tfCrystalOpen()
 *
 * Verifies that a cs8900 is present, resets the chip, initializes various
 * registers and sets up the interrupt mechanisms.
 *
 */
int tfCrystalOpen(ttUserInterface interfaceHandle)
{
         int              errorCode;
         char             hardwareAddr[6];
volatile unsigned short   inputWord;
         unsigned char    intNumber;
         unsigned char    inputByte;
         unsigned long    delayCount;

    tlCrystalInterfaceHandle = interfaceHandle;

/*
 * Initialize empty ring of receive buffers,
 * and pre-allocate receive buffers
 */
    errorCode = tfPoolCreate( interfaceHandle,
                              TM_CRYSTAL_NUM_RECV_BUF,
                              TM_CRYSTAL_NUM_SMALL_RECV_BUF,
                              TM_ETHER_MAX_PACKET_CRC,
                              TM_CRYSTAL_SMALL_BUFFER_SIZE,
                              0,
                              TM_POOL_REFILL_IN_LINE );
    if (errorCode == TM_ENOERROR)
    {

        tlCrystalIrq = TM_CRYSTAL_IRQ;

        tlCrystalMissedFrames = 0;

/* Force the chip to reset */
        tfCrystalWritePktPage( TM_CRYSTAL_PP_SELFCTL,
                               TM_CRYSTAL_SELFCTL_RESET );

/* Wait for 125 usecs (per databook) */
        for (delayCount = 0; delayCount < 10000000UL; delayCount++)
        {
            ;
        }

/* Wait for chip to complete reset (!SIBUSY and INITD) */
        while ((tfCrystalReadPktPage(TM_CRYSTAL_PP_SELFSTAT) & 
               (TM_CRYSTAL_SELFSTAT_SIBUSY | TM_CRYSTAL_SELFSTAT_INITD)) !=
                TM_CRYSTAL_SELFSTAT_INITD)
        {
            ;
        }

/* 
 * Setup the init parameters on the chip 
 */
 

/* 
 * Set physical interface to use.  This is controlled by the TM_CRYSTAL_PHYSINT
 * #define set at the top of this file.
 */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LINECTL, TM_CRYSTAL_PHYSINT);


/* Might be necessary to set IOCHRDY */

/* Set cs8900 to generate interrupts upon receiving a good frame */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_RXCFG, TM_CRYSTAL_RXCFG_OKIE);

/*
 * Set cs8900 to receieve unicast, broadcast and multicast frames, and to 
 * discard error frames.
 */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_RXCTL, 
                              TM_CRYSTAL_RXCTL_ACC_OK |
                              TM_CRYSTAL_RXCTL_ACC_BCAST |
                              TM_CRYSTAL_RXCTL_ACC_IND);



/* Set cs8900 to generate interrupts when a frame has completed transmitting */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_TXCFG,
                              TM_CRYSTAL_TXCFG_OKIE);


/* Set the Ethernet hardware address */
        tfCrystalGetPhysicalAddress((ttUserInterface) 0, hardwareAddr);
        tfCrystalWritePktPage(TM_CRYSTAL_PP_HWADDR,
                              *((unsigned short *)&hardwareAddr[0]));
        tfCrystalWritePktPage(TM_CRYSTAL_PP_HWADDR+2,
                              *((unsigned short *)&hardwareAddr[2]));
        tfCrystalWritePktPage(TM_CRYSTAL_PP_HWADDR+4,
                              *((unsigned short *)&hardwareAddr[4]));


/* 
 * Initialize the interrupt mechanisms.  This is very platform specific to and
 * will need to be tailored to your environment.
 */

/* Set the ISA interrupt line to use, dependent on the IRQ we wish to use */
        switch (tlCrystalIrq)
        {
            case 5: 
                intNumber =  3;
                break;

            case 10:
                intNumber = 0;
                break;

            case 11:
                intNumber = 1;
                break;

            case 12:
                intNumber = 2;
                break;
        }

        tfCrystalWritePktPage(TM_CRYSTAL_PP_INTNUM, intNumber);

/* Install ISR Handler */
        tfKernelInstallIsrHandler((ttUserIsrHandlerPtr) tfCrystalIsrHandler, 
                                  0x08 + tlCrystalIrq);


/* Enable interrupt in PIC */
        if (tlCrystalIrq == 5)
        {
            inputByte = tm_in_byte(TM_PIC1_IMR_ADDR);
            inputByte &= ~(1<<tlCrystalIrq);
            tm_out_byte(inputByte, TM_PIC1_IMR_ADDR);
        }
        else
        {
            inputByte = tm_in_byte(TM_PIC2_IMR_ADDR);
            inputByte &= ~(1<<(tlCrystalIrq-7));
            tm_out_byte(inputByte, TM_PIC2_IMR_ADDR);
        }

/* Enable interrupts on the cs8900 */
        inputWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_BUSCTL);
        inputWord |= TM_CRYSTAL_BUSCTL_INT_ENB;
        tfCrystalWritePktPage(TM_CRYSTAL_PP_BUSCTL, inputWord);        

/* Enable transmit and receive on the cs8900 */
        inputWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_LINECTL);
        inputWord |= TM_CRYSTAL_LINECTL_TXON | TM_CRYSTAL_LINECTL_RXON;
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LINECTL, inputWord);
        inputWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_LINECTL);

        tlCrystalAllocState  = TM_CRYSTAL_ALLOC_IDLE;
        tlCrystalXmitQueueOn = tfInterfaceXmitQueueOn(interfaceHandle);
    }
    return(errorCode);            
}

/*
 * tfCrystalClose()
 *
 * Disable transmit and receieve on the cs8900 and prevent the chip from 
 * generating interrupts in the future.
 *
 */
int tfCrystalClose(ttUserInterface interfaceHandle)
{
    unsigned short   inputWord;

/* Disable interrupts on the cs8900 */
    inputWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_BUSCTL);
    inputWord &= ~TM_CRYSTAL_BUSCTL_INT_ENB;
    tfCrystalWritePktPage(TM_CRYSTAL_PP_BUSCTL, inputWord);        

/* Disable transmit and receive on the cs8900 */
    inputWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_LINECTL);
    inputWord &= ~(TM_CRYSTAL_LINECTL_TXON | TM_CRYSTAL_LINECTL_RXON);
    tfCrystalWritePktPage(TM_CRYSTAL_PP_LINECTL, inputWord);
/* Delete the interface recv pool, and free all pre-allocated recv buffers */
    (void)tfPoolDelete(interfaceHandle);

    return(TM_DEV_OKAY);
}


/* 
 * tfCrystalIoctl()
 *
 *   TM_CRYSTAL_IOCTL_REFILL:
 *     Cycles through the receive ring, allocating buffers for any empty 
 *     entries in the ring.
 *
 *   TM_DEV_SET_MCAST_LIST:
 *     Adds a number of multicast entries (==optionLen) to the CS8900 hash 
 *     table.  This is done by 
 *         1. Compute a 32-bit CRC on the address.
 *         2. Extract the most significant 6 bits of the computed CRC.
 *         3. Set the bit coresponding to the above number in the CS8900's
 *            address filter (ie, decode the 6 bit value above into a 64 bit
 *            value: 50 -> 0x0002 0000 0000 0000) 
 *
 *   TM_DEV_SET_ALL_MCAST:
 *     Sets the CS8900 to receive all multicast packets (ie, set the address
 *     filter to all ones).
 *
 */
int tfCrystalIoctl ( ttUserInterface interfaceHandle,
                     int             flag,
                     void TM_FAR *   optionPtr,
                     int             optionLen )
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    unsigned long               calcCRC;
    unsigned char               currByte;
    unsigned char               byte, bit;
    unsigned long               hashEntry;
    unsigned char      TM_FAR * multiAddr;
    unsigned short              logicalAddrFilter[4];
    unsigned short              inWord;
    int                         savedOptionLen;
#endif /* TM_IGMP or TM_6_USE_MLD */

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    savedOptionLen = optionLen;
    if (flag & TM_DEV_SET_MCAST_LIST)
    {

/* 
 * Clear the CS8900's address filter (since this IOCTL call will add all the
 * entries we need.
 */
        logicalAddrFilter[0] = 0;
        logicalAddrFilter[1] = 0;
        logicalAddrFilter[2] = 0;
        logicalAddrFilter[3] = 0;

/* Get the first address */
        multiAddr = (unsigned char TM_FAR *) optionPtr;

        while (optionLen-- > 0)
        {
            
  
/* Initialize the calculated CRC (with all ones) */
            calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
            for (byte=0;byte<6;byte++)
            {
                currByte = multiAddr[byte];

/* For each bit in the current byte */
                for (bit=0;bit<8;bit++)
                {

/* Shift the calculated CRC left by one */
                    calcCRC <<= 1;

/* If the current bit is one... */
                    if ((currByte >> bit) & 0x01)
                    {

/* ... XOR the calculated CRC with the CRC-32 polynomial */
                        calcCRC ^= TM_CRC32_POLY;

                    }                   
                }
            }

/* Get the most significant 6 bits from the calculated CRC */        
            hashEntry = (calcCRC & 0xfc000000UL) >> 26;

/* 
 * Decode the above number and set the appropriate bit in the CS8900 logical
 * address filter 
 */
            logicalAddrFilter[(unsigned char) (hashEntry >> 4)] |=
                1 << ((unsigned char) (hashEntry % 16) + 1);

            multiAddr += 6;

        }

        if (savedOptionLen != 0)
        {
/* Enable receiving multicast packets */
            inWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_RXCTL);
            inWord |= TM_CRYSTAL_RXCTL_ACC_MCAST;
            tfCrystalWritePktPage(TM_CRYSTAL_PP_RXCTL, inWord);            

/* Set the new logical address filter in the CS89000 */
            tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF,   logicalAddrFilter[0]);
            tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+2, logicalAddrFilter[1]);
            tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+4, logicalAddrFilter[2]);
            tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+6, logicalAddrFilter[3]);
        }
        else
        {
/* Disable receiving multicast packets */
            inWord = tfCrystalReadPktPage(TM_CRYSTAL_PP_RXCTL);
            inWord &= ~TM_CRYSTAL_RXCTL_ACC_MCAST;
            tfCrystalWritePktPage(TM_CRYSTAL_PP_RXCTL, inWord);
        }

    }
    if (flag & TM_DEV_SET_ALL_MCAST)
    {

/* Set the logical address filter to all ones (receive ALL multicast packets) */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF,   0xffff);
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+2, 0xffff);
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+4, 0xffff);
        tfCrystalWritePktPage(TM_CRYSTAL_PP_LAF+6, 0xffff);
    }
#endif /* TM_IGMP or TM_6_USE_MLD */

    return(TM_DEV_OKAY);
}

/*
 * tfCrystalSend()
 *
 * Requests transmit approval and poll the chip until it is approved.  Copy the
 * data (via ports) to the CS8900.
 *
 */
int tfCrystalSend ( ttUserInterface interfaceHandle,
                    char TM_FAR *   dataPtr,
                    int             dataLength,
                    int             flag )
{
    int errorCode;

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(flag);

    errorCode = TM_DEV_OKAY;

/* We are currently waiting on a send allocation request */
    if ( tlCrystalAllocState == TM_CRYSTAL_ALLOC_WAITING )
    {
        if ( tfCrystalReadPktPage(TM_CRYSTAL_PP_BUSSTAT) &
                TM_CRYSTAL_BUSSTAT_TXRDY )
        {
/* Transmit approved: set state and continue to send packet */
            tlCrystalAllocState = TM_CRYSTAL_ALLOC_IDLE;
        }
        else
        {
/*
 * Transmit still not approved:  return and let the transmit queue deal
 * with it.
 */
            errorCode = TM_DEV_ERROR;

        }
    }

/* No pending transmit request; issue a new one for this packet */
    else
    {

/* Request approval to send a frame: write TX-Command and then the length */
        tfCrystalWritePktPage(TM_CRYSTAL_PP_TXCMD, TM_CRYSTAL_TXCMD_TXSTART);
        tfCrystalWritePktPage(TM_CRYSTAL_PP_TXLEN, dataLength);

/* 
 * Poll the cs8900 until it is possible to transmit the frame (the Rdy4TxNOW
 * bit in the bus status register is set).
 */
        if ( !(   tfCrystalReadPktPage(TM_CRYSTAL_PP_BUSSTAT)
                & TM_CRYSTAL_BUSSTAT_TXRDY ) )
        {

/* 
 * Allocation failed: either simply return an error (if the transmit queue
 * is enabled) or spin lock otherwise.
 */
            if (tlCrystalXmitQueueOn)
            {
                tlCrystalAllocState = TM_CRYSTAL_ALLOC_WAITING;

                errorCode = TM_DEV_ERROR;
            }
            else
            {
                do
                {
/* Yield processor to other tasks while we're waiting */
                    tfInterfaceSpinLock(interfaceHandle);

                } while ( !(   tfCrystalReadPktPage(TM_CRYSTAL_PP_BUSSTAT)
                         & TM_CRYSTAL_BUSSTAT_TXRDY ) );
            }
        }
    }

    if (errorCode == TM_DEV_OKAY)
    {
/* Transmit approved: Copy data to cs8900 */
        while (dataLength > 0)
        {
            tm_out_word(*((unsigned short *) dataPtr),
                        TM_CRYSTAL_IOBASE+TM_CRYSTAL_TX_DATA);
            dataLength -= 2;
            dataPtr += 2;
        }

/* We're done with the packet, so signal the stack to free the packet */
        tfSendCompleteInterface(tlCrystalInterfaceHandle, 
                                TM_DEV_SEND_COMPLETE_DRIVER);

    }

    return errorCode;
}

/*
 * tfCrystalReceive()
 *
 * Since we're forced to copy in the ISR, not much to do in the receive
 * routine. Just detach the buffer from our ring and pass it up the stack.
 * If desired, we also refill this buffer in this routine.
 */
int tfCrystalReceive(ttUserInterface        interfaceHandle,
                     char TM_FAR * TM_FAR * dataPtrPtr,
                     int  TM_FAR *          dataLengthPtr,
                     ttUserBufferPtr        bufHandlePtr)
{
    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          bufHandlePtr );
}


/*
 * Get the physical address
 * return it to the Treck stack
 * The address of the card needs to be changed - this is just test data
 */
int tfCrystalGetPhysicalAddress(ttUserInterface interfaceHandle,
                                char TM_FAR *   physicalAddress)
{

/* This line is to get rid of compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

    physicalAddress[0] = 0x00;
    physicalAddress[1] = 0x11;
    physicalAddress[2] = 0x22;
    physicalAddress[3] = 0x33;
    physicalAddress[4] = 0x44;
    physicalAddress[5] = 0x55;

    return TM_ENOERROR;
}

/*
 * tfCrystalIsrHandler
 *
 *     On every interrupt from the CS8900, there could be a number of actual 
 *   interrupts to process.  These events are retrieved through ISQ (Interrupt
 *   Status Queue).  Once a zero is read from this register, the CS8900 
 *   considers this interrupt complete and is free to generate future 
 *   interrupts.
 *     Unfortunately, one of the things that the CS8900 does upon completion
 *   of an interrupt is clear the current packet from its internal buffer.
 *   Therefore, it is necessary to copy the data inside of the ISR, to our ring
 *   of pre-allocated receive buffers.  This ring is used to avoid allocating 
 *   memory in the ISR, and should be refilled periodically using 
 *   tfInterfaceIoctl.
 *     Also, this ISR is not very deterministic, due to the structure of the 
 *   CS8900 interrupt service queue.  It is entirely possible that while we 
 *   are processing (copying) a received packet that the chip recieves another
 *   packet, and when we go through the loop again there is another packet.  So
 *   this could go on for some time, and its almost impossible to predict how
 *   long this will take...
 */
void TM_INTERRUPT tfCrystalIsrHandler(void)
{
    unsigned short TM_FAR * dataPtr;
    unsigned short          dataWord; /* room to discard packet if needed */
    unsigned short          isrEvent;
    int                     notifyCount;
    int                     copyLength;
    int                     isOdd;


    tm_kernel_set_critical;

    isrEvent = tm_in_word(TM_CRYSTAL_IOBASE + 8);

    notifyCount = 0;
    while (isrEvent != 0)
    {

/* Receive event */
        if ((isrEvent & 0x003f) == TM_CRYSTAL_ISR_RXEVENT)
        {

/* 
 * Good frame received (Every frame that we are interrupted on should be good,
 * so anything other than this is an error).
 */
            if (isrEvent & 0x0100)
            {

/* If DMA is not enabled, we need to copy the data into one of the buffers in
 * our receive queue.
 */
/* Discard the status */
                tm_in_word(TM_CRYSTAL_IOBASE + TM_CRYSTAL_RX_DATA);
/* length of this packet */
                copyLength =
                        tm_in_word(TM_CRYSTAL_IOBASE + TM_CRYSTAL_RX_DATA);
                isOdd = copyLength & 1;
/* 
 * If there is room in our pool (ie, there is a buffer that has been
 * allocated, and is owned by the chip), copy the data 
 */
/* Get pointer to pre-allocated buffer entry */
                dataPtr = (unsigned short TM_FAR *) tfPoolIsrGetBuffer(
                                                     tlCrystalInterfaceHandle,
                                                     copyLength );
                if ( dataPtr != (unsigned short TM_FAR *)0 )
                {
                    copyLength = copyLength >> 1;


/* Copy the current packet to our receive buffer */
                    while (copyLength--)
                    {
                        *dataPtr++ = tm_in_word(   TM_CRYSTAL_IOBASE
                                                 + TM_CRYSTAL_RX_DATA );
                    }
                    if (isOdd)
                    {
                        *((char TM_FAR*)dataPtr) = tm_in_byte(
                                                        TM_CRYSTAL_IOBASE
                                                      + TM_CRYSTAL_RX_DATA );
                    }
/* Keep track of number of packets that we need to notify the stack with */
                    notifyCount++;
                }
                else
                {
                    copyLength = copyLength >> 1;
/* No room to receive this packet: drop it */
/* discard the current packet to dismiss the interrupt */
                    while (copyLength--)
                    {
                        dataWord = tm_in_word(   TM_CRYSTAL_IOBASE
                                               + TM_CRYSTAL_RX_DATA );
                    }                 
                    if (isOdd)
                    {
                        dataWord = tm_in_byte(   TM_CRYSTAL_IOBASE
                                               + TM_CRYSTAL_RX_DATA );
                    }
                    TM_UNREF_IN_ARG(dataWord); /* avoid compiler warning */
                    tlCrystalMissedFrames++;
                }


            }
            else
            {
/* ERROR! We should never get here: we received a packet  */
                ;
            }

        }

        isrEvent = tm_in_word(TM_CRYSTAL_IOBASE + 8);

    }

/* Acknowledge interrupt to the PIC */
    if (tlCrystalIrq == 5)
    {
        tm_out_byte(TM_PIC1_ISR_ADDR, TM_PIC1_ISR_ADDR);
    }
    else
    {
        tm_out_byte(TM_PIC2_ISR_ADDR, TM_PIC2_ISR_ADDR);
    }
    tm_kernel_release_critical;
/* 
 * Now that the ISR is mostly done, we notify the stack for each packet that
 * we received.  This is done at the end of the ISR because the call to
 * tfNotifyInterfaceIsr tends to release our critical section 
 * (especially in uC/OS on the x86).
 */
    tfNotifyInterfaceIsr(tlCrystalInterfaceHandle, notifyCount, 0, 0UL, 0UL);
    return;
}

/*
 * tfCrystalWritePktPage()
 *
 * Writes a value to the cs8900 PacketPage structure.  This routine is needed
 * because we can either access this structure via I/O mode.
 * The cs8900 uses the standard 'first write to a address port to specify the 
 * location and then write to a data port to write to that location' method.
 */ 
void tfCrystalWritePktPage(unsigned short addr, unsigned short value)
{
    tm_out_word(addr, TM_CRYSTAL_IOBASE + TM_CRYSTAL_PKTPAGE_ADDR);
    tm_out_word(value, TM_CRYSTAL_IOBASE + TM_CRYSTAL_PKTPAGE_DATA);

    return;
}


/*
 * tfCrystalReadPktPage()
 *
 * Reads a value from the cs8900 PacketPage structure.  This routine is needed
 * because we can either access this structure via I/O mode.
 * The cs8900 uses the standard 'first write to a address port to specify the 
 * location and then write to a data port to write to that location' method.
 */ 
unsigned short tfCrystalReadPktPage(unsigned short addr)
{
    unsigned short inWord;

    tm_out_word(addr, TM_CRYSTAL_IOBASE + TM_CRYSTAL_PKTPAGE_ADDR);
    inWord = tm_in_word(TM_CRYSTAL_IOBASE + TM_CRYSTAL_PKTPAGE_DATA);

    return (inWord);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: trcrystl.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 14617    Output checksum: 42416
 */
