/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated.
 * Copyright laws and International Treaties protect the contents of this
 * file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Texas Instruments TMS320DM641 EMAC device driver
 *
 * Device driver for the Ethernet MAC (EMAC) controller integrated with
 * Texas Instrument's TMS320DM642 Digital Media Processor.  This driver was
 * developed on the DM642 EVM using TI's Code Composer tools.  The DM642 EMAC
 * is described in detail in SPRU628.  The DM642 EMAC supports up to eight
 * prioritized TX and RX channels, which could be used for Quality of
 * Serivce (QoS).  This driver however only uses one channel (#0).
 *
 * This driver depends on interrupts being setup through the CCS configuration
 * tool.  For an example configuration, please see the sample code in:
 *   source/kernel/dsp_bios/dm642evm/
 *
 * Filename: trdm6mac.c
 * Author: Jason
 * Date Created: 01/20/2004
 * $Source: source/drivers/dm642mac/trdm6mac.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:35:59CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Treck includes. */
#include <trsocket.h>
#include "trdm6mac.h"

/* Treck internal API includes, for tm_assert macro. */
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* Needed for TI interrupt control function (IRQ_enable, IRQ_disable) */
#define CHIP_DM642
#include <csl_irq.h>
#include <csl_cache.h>
#include <csl_mdio.h>

#define TM_EMAC_MAX_PACKET_LEN  1518
/* receive packet buffers must be a multiple of the cache line size */
#define TM_EMAC_RX_BUF_LEN      1536

/* The number of hardware TX queues to use. Must be from 3 to 8. */
#define TM_EMAC_TX_QUEUES       8

/* DM642 L2 cache line size in bytes, from csl_cache.h */
#define TM_CACHE_L2_LINESIZE    128

/* Base address of EMAC registers (default for DM642 EVM) */
#define TM_EMAC_REG_BASE        0x01c80000

/* Base address of EMAC frame descriptors, in non-cacheable memory. Refer to
   TI header file csl_emachal.h */
#define TM_EMAC_DSC_BASE_ADDR   0x01c81000

/* EMAC Descriptor Size and Element Count */
#define TM_EMAC_DSC_SIZE        4096
#define TM_EMAC_DSC_ENTRY_SIZE  20 /* extra for emacDescUserBufPtr */
#define TM_EMAC_DSC_ENTRY_COUNT 95

#if ((TM_EMAC_DSC_SIZE/TM_EMAC_DSC_ENTRY_SIZE) < TM_EMAC_DSC_ENTRY_COUNT)
#error "too many descriptors, reduce TM_EMAC_DSC_ENTRY_COUNT!"
#endif

/*
 * Maximum number of receive and transmit descriptors.
 */
#define TM_EMAC_MAX_RX_DESC 40
#define TM_EMAC_MIN_RX_DESC 15
#define TM_EMAC_MAX_TX_DESC (TM_EMAC_DSC_ENTRY_COUNT - TM_EMAC_MIN_RX_DESC)

/* Macros to access EMAC registers. */
#define tm_emac_set_dword( regaddr, data ) \
    (*(volatile unsigned int TM_FAR *)(regaddr)) = (unsigned int) (data)

#define tm_emac_get_dword( regaddr) \
    (*(volatile unsigned int TM_FAR *)(regaddr))


/* Nothing below should normally need to be modified... */

/* EMAC transmit descriptor flags. */
#define TM_EMAC_XMIT_SOP    0x8000    /* Start of packet        */
#define TM_EMAC_XMIT_EOP    0x4000    /* End of packet          */
#define TM_EMAC_XMIT_OWNER  0x2000    /* Buffer owned by EMAC   */
#define TM_EMAC_XMIT_EOQ    0x1000    /* End of queue           */

/* EMAC receive descriptor flags. */
#define TM_EMAC_RECV_OWNER  0x2000    /* Buffer owned by EMAC   */
#define TM_EMAC_RECV_EOQ    0x1000    /* End of queue           */

/* EMAC register definitions. */
#define TM_EMAC_REG_TXCONTROL          (TM_EMAC_REG_BASE + 0x0004)
#define TM_EMAC_REG_RXCONTROL          (TM_EMAC_REG_BASE + 0x0014)
#define TM_EMAC_REG_RXMBP_ENABLE       (TM_EMAC_REG_BASE + 0x0100)
#define TM_EMAC_REG_UNICAST_SET        (TM_EMAC_REG_BASE + 0x0104)
#define TM_EMAC_REG_UNICAST_CLEAR      (TM_EMAC_REG_BASE + 0x0108)
#define TM_EMAC_REG_RXBUFOFFSET        (TM_EMAC_REG_BASE + 0x0110)
#define TM_EMAC_REG_MACCONTROL         (TM_EMAC_REG_BASE + 0x0160)
#define TM_EMAC_REG_MACSTATUS          (TM_EMAC_REG_BASE + 0x0164)
#define TM_EMAC_REG_TX_INTMASK_SET     (TM_EMAC_REG_BASE + 0x0178)
#define TM_EMAC_REG_TX_INTMASK_CLEAR   (TM_EMAC_REG_BASE + 0x017c)
#define TM_EMAC_REG_MACINVECTOR        (TM_EMAC_REG_BASE + 0x0180)
#define TM_EMAC_REG_RX_INTMASK_SET     (TM_EMAC_REG_BASE + 0x0198)
#define TM_EMAC_REG_RX_INTMASK_CLEAR   (TM_EMAC_REG_BASE + 0x019c)
#define TM_EMAC_REG_MAC_INTMASK_CLEAR  (TM_EMAC_REG_BASE + 0x1ac)
#define TM_EMAC_REG_MACADDR_LOW        (TM_EMAC_REG_BASE + 0x01b0)
#define TM_EMAC_REG_MACADDR_MID        (TM_EMAC_REG_BASE + 0x01d0)
#define TM_EMAC_REG_MACADDR_HI         (TM_EMAC_REG_BASE + 0x01d4)
#define TM_EMAC_REG_MC_HASH0           (TM_EMAC_REG_BASE + 0x01d8)
#define TM_EMAC_REG_MC_HASH1           (TM_EMAC_REG_BASE + 0x01dc)
#define TM_EMAC_REG_TXHDP              (TM_EMAC_REG_BASE + 0x0600)
#define TM_EMAC_REG_RXHDP              (TM_EMAC_REG_BASE + 0x0620)
#define TM_EMAC_RX0INTACK              (TM_EMAC_REG_BASE + 0x0660) 
#define TM_EMAC_REG_EWCTL              (TM_EMAC_REG_BASE + 0x3004)

/*
 * EMAC register flags, defining certain bits in EMAC registers.
 */

/* EMAC wrapper control bits. */
#define TM_EMAC_EWCTL_INT_DISABLE 0x0
#define TM_EMAC_EWCTL_INT_ENABLE  0x1
#define TM_EMAC_EWCTL_MDIO_RST    0x2
#define TM_EMAC_EWCTL_EMAC_RST    0x4

/* Multicast/broadcast enable bits. */ 
#define TM_EMAC_MBP_BROADCAST_EN 0x2000
#define TM_EMAC_MBP_MULTICAST_EN 0x0020

/* TX, RX control bits. */
#define TM_EMAC_TXCONTROL_ENABLE 0x1
#define TM_EMAC_RXCONTROL_ENABLE 0x1

/* MAC/PHY control bits. */
#define TM_EMAC_MACCONTROL_MII_EN       0x20
#define TM_EMAC_MACCONTROL_FULLDUPLEX   0x01

/* MAC interrupt vector bits. */
#define TM_EMAC_MACINVECTOR_RX 0x00000100

/* Swap bytes for big-endian (32 bit only) */
#ifdef TM_LITTLE_ENDIAN
#define TM_SWIZZLE(x) x
#else /* ! TM_LITTLE_ENDIAN */
#define TM_SWIZZLE(x) ((x>>24)|(x<<24)|((x>>8)&0xFF00)|((x<<8)&0xFF0000))
#endif /* ! TM_LITTLE_ENDIAN */

/*
 * General buffer descriptor format (applies to both TX and RX)
 */
typedef struct tsEmacPktDesc
{
/* Next descriptor pointer. */
    struct tsEmacPktDesc TM_FAR * emacDescNextPtr;

/* Buffer pointer. */
    char                 TM_FAR * emacDescBufPtr;

/* Buffer offset and buffer length. */
    unsigned int                  emacDescOffsetBufLength;

/* Flags and packet length. */
    unsigned int                  emacDescFlagsPktLength;

/* This field is specific to this driver, which is OK since it is after the
 *  structure used by the EMAC.  This field is only used in RX to store the
 *  Treck zero copy buffer ptr.
 */
    ttUserBuffer                  emacDescUserBufPtr; 
} ttEmacPktDesc;

typedef ttEmacPktDesc TM_FAR * ttEmacPktDescPtr;

/*
 * EMAC control structure, containing the various queues and counters used
 * to manage packet flow in the system.
 */
typedef struct tsEmacInfo
{
    Handle           emacMdioPtr; /* Handle to TI's MDIO Module */

/* Static pool of buffer descriptors.  These are allocated on demand from
 * TX and RX paths as needed. */   
    int              emacAvailBufDescsCount;

/* Transmit queue head and tail pointers. */
    ttEmacPktDescPtr emacFirstXmitPtr[TM_EMAC_TX_QUEUES];
    ttEmacPktDescPtr emacLastXmitPtr[TM_EMAC_TX_QUEUES];

/* Counters to show which transmit queue to use */
    unsigned int     emacEmptyTxQueue;
    unsigned int     emacFullTxQueue;

/* "pre" transmit queue head and tail pointers.  This queue is used to store
 *  scattered buffers prior to transmit. */
    ttEmacPktDescPtr emacFirstXmitInputPtr;
    ttEmacPktDescPtr emacLastXmitInputPtr;

/* Receive queue head and tail pointers. */
    ttEmacPktDescPtr emacFirstRecvPtr;
    ttEmacPktDescPtr emacLastRecvPtr;
    ttEmacPktDescPtr emacIsrFirstRecvPtr;
    unsigned int     emacRecvDescCount;

/* Queue to manage buffer descriptor pool. */
    ttEmacPktDescPtr emacBufDescQ;
} ttEmacInfo;

typedef ttEmacInfo TM_FAR * ttEmacInfoPtr;

/* Local utility functions to manage buffer queues. */
static void tfPushBufDesc(ttEmacInfoPtr infoPtr, ttEmacPktDescPtr descPtr);
static ttEmacPktDescPtr tfPopBufDesc(ttEmacInfoPtr infoPtr);
static void tfRefillRecvList(ttEmacInfoPtr infoPtr);
static void tfCleanupXmitQueue(ttEmacInfoPtr   infoPtr,
                               ttUserInterface interfaceHandle);


/* Stored statically for use in ISR handler. */
ttUserInterface  emacInterfaceHandle;
#ifdef TM_ERROR_CHECKING
ttDeviceEntryPtr emacDevEntryPtr;
#endif /* TM_ERROR_CHECKING */

extern unsigned char macAddrFromApp[6];
ttEmacInfo emacInfo;

/*
 * tfDm642EmacOpen()
 *
 * Initializes local control structures, resets and initializes EMAC, sets up
 * interrupts and enables TX and RX.
 */
int tfDm642EmacOpen(ttUserInterface interfaceHandle)
{
    int           i;
    ttEmacInfoPtr infoPtr;
    unsigned int  temp;
    char          physAddr[6];
    unsigned char *emacDescPtr;

    infoPtr = &emacInfo;
    tm_bzero(infoPtr, sizeof(ttEmacInfo));
    emacInterfaceHandle = interfaceHandle;
#ifdef TM_ERROR_CHECKING
    emacDevEntryPtr = (ttDeviceEntryPtr) interfaceHandle;
#endif /* TM_ERROR_CHECKING */
    
/* Populate packet descriptor queue. */
    emacDescPtr = (unsigned char *) TM_EMAC_DSC_BASE_ADDR;
    for (i=0;i<TM_EMAC_DSC_ENTRY_COUNT;i++)
    {
        tfPushBufDesc(infoPtr, (ttEmacPktDescPtr) emacDescPtr);
        emacDescPtr += TM_EMAC_DSC_ENTRY_SIZE;
    }

/* Disable interrupts and put the EMAC and MDIO into reset. */
    tm_emac_set_dword( TM_EMAC_REG_EWCTL,
                       TM_EMAC_EWCTL_INT_DISABLE |
                       TM_EMAC_EWCTL_MDIO_RST    |
                       TM_EMAC_EWCTL_EMAC_RST );

/* Wait approx. 100 cycles */
    for( i=0; i<5; i++ )
    {
        temp = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    }

/* Take EMAC and MDIO out of reset but keep interrupts disabled. */
    tm_emac_set_dword( TM_EMAC_REG_EWCTL,
                       TM_EMAC_EWCTL_INT_DISABLE );
    
/* Wait approx. 100 cycles */
    for( i=0; i<5; i++ )
    {
        temp = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    }

    /* Start the MII Configuration */
    infoPtr->emacMdioPtr = MDIO_open(
        MDIO_MODEFLG_AUTONEG | MDIO_MODEFLG_HD10 | MDIO_MODEFLG_FD10
        | MDIO_MODEFLG_HD100 | MDIO_MODEFLG_FD100);

/* Reset MAC control, TX control and RX control. */
    tm_emac_set_dword( TM_EMAC_REG_MACCONTROL, 0);
    tm_emac_set_dword( TM_EMAC_REG_TXCONTROL, 0);    
    tm_emac_set_dword( TM_EMAC_REG_RXCONTROL, 0);
    tm_emac_set_dword( TM_EMAC_REG_MACSTATUS, 0);    

/* Initialize head descriptor pointers to NULL. (have to initialize all
 * channels, even though we're only using channel #1.
 */
    for (i=0;i<8;i++)
    {
        tm_emac_set_dword( TM_EMAC_REG_TXHDP + i*4, 0);
        tm_emac_set_dword( TM_EMAC_REG_RXHDP + i*4, 0);
    }

/* If using EMAC statistics, the statistic registers should be cleared here. */
    
/* Set MAC address. */
    tfDm642EmacGetPhysicalAddress( interfaceHandle, physAddr );
    tm_emac_set_dword( TM_EMAC_REG_MACADDR_LOW, physAddr[5]);

    tm_emac_set_dword( TM_EMAC_REG_MACADDR_MID, physAddr[4]);

    temp = *((unsigned int TM_FAR *) &physAddr[0]);
    tm_emac_set_dword(TM_EMAC_REG_MACADDR_HI, TM_SWIZZLE(temp));

/* Initialize receive buffer offset to zero. */
    tm_emac_set_dword( TM_EMAC_REG_RXBUFOFFSET, 0);

/* Clear multicast hash registers. */
    tm_emac_set_dword( TM_EMAC_REG_MC_HASH0, 0);
    tm_emac_set_dword( TM_EMAC_REG_MC_HASH1, 0);

/* Reset Multicast/Broadcast/Promiscuous enable register. */
    tm_emac_set_dword( TM_EMAC_REG_RXMBP_ENABLE, TM_EMAC_MBP_BROADCAST_EN);

/* Disable unicast on all channels, and then enable it only on channel #0 */
    tm_emac_set_dword( TM_EMAC_REG_UNICAST_CLEAR, 0xFF);
    tm_emac_set_dword( TM_EMAC_REG_UNICAST_SET, 1);
    
/* Could setup loopback mode in MACCONTROL here */
/* NOTE: full-duplex mode is set in tfDm642EmacIoctl */

/* Disable TX and RX interrupts on all channels, and then enable
 * RX interrupts on channel #0 only. */
    tm_emac_set_dword(TM_EMAC_REG_TX_INTMASK_CLEAR, 0xFF);
    tm_emac_set_dword(TM_EMAC_REG_RX_INTMASK_CLEAR, 0xFF);
    tm_emac_set_dword(TM_EMAC_REG_MAC_INTMASK_CLEAR, 0x3);
    tm_emac_set_dword(TM_EMAC_REG_RX_INTMASK_SET, 0x1);

/* Initially fill up the receive queue, and prepare the MAC for packet
 * reception. */
    tfRefillRecvList(infoPtr);

/* Enable transmit and receive, and then enable the MII. */
    temp = tm_emac_get_dword(TM_EMAC_REG_TXCONTROL);
    temp |= TM_EMAC_TXCONTROL_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_TXCONTROL, temp);
    
    temp = tm_emac_get_dword(TM_EMAC_REG_RXCONTROL);
    temp |= TM_EMAC_RXCONTROL_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_RXCONTROL, temp);

    temp = tm_emac_get_dword(TM_EMAC_REG_MACCONTROL);
    temp |= TM_EMAC_MACCONTROL_MII_EN;
    tm_emac_set_dword(TM_EMAC_REG_MACCONTROL, temp);

/* Enables EMAC interrupts on the DM642. */
    IRQ_enable(IRQ_EVT_MACINT);
    
/* Enable global interrupt in EMAC wrapper */
    temp = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    temp |= TM_EMAC_EWCTL_INT_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_EWCTL, temp);

/* The EMAC is now fully configured and ready for transmit and receive,
   with RX interrupts enabled. */

    return TM_DEV_OKAY;
}

/*
 * tfDm642EmacClose()
 *
 * Disables TX and RX on the EMAC, and disable RX interrupts.  Free all
 * empty buffers in the receive queue back to the stack.  No need to mess
 * with any of the internal queues of buffer descriptors, as they will be
 * fully re-initialized if the driver is opened again.
 *
 */
int tfDm642EmacClose(ttUserInterface interfaceHandle)
{
    ttEmacPktDescPtr currDescPtr;
    ttEmacInfoPtr    infoPtr;
    unsigned int     temp;

    infoPtr = &emacInfo;
    
/* Enable global interrupt in EMAC wrapper */
    temp = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    temp &= ~TM_EMAC_EWCTL_INT_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_EWCTL, temp);
    
/* Disables EMAC interrupts on the DM642. */
    IRQ_disable(IRQ_EVT_MACINT);

/* Disable the MII, and then disable  transmit and receive. */
    temp = tm_emac_get_dword(TM_EMAC_REG_MACCONTROL);
    temp &= ~TM_EMAC_MACCONTROL_MII_EN;
    tm_emac_set_dword(TM_EMAC_REG_MACCONTROL, temp);

    temp = tm_emac_get_dword(TM_EMAC_REG_TXCONTROL);
    temp &= ~TM_EMAC_TXCONTROL_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_TXCONTROL, temp);
    
    temp = tm_emac_get_dword(TM_EMAC_REG_RXCONTROL);
    temp &= ~TM_EMAC_RXCONTROL_ENABLE;
    tm_emac_set_dword(TM_EMAC_REG_RXCONTROL, temp);

/* Free all packets in the receive queue back to the stack. */
    currDescPtr = infoPtr->emacFirstRecvPtr;
    while (currDescPtr != (ttEmacPktDescPtr)0)
    {
        if (currDescPtr->emacDescUserBufPtr != (ttUserBuffer)0)
        {
            tfFreeDriverBuffer(currDescPtr->emacDescUserBufPtr);
        }
        currDescPtr = currDescPtr->emacDescNextPtr;
    }
    
    /* Close the MDIO Module */
    MDIO_close(infoPtr->emacMdioPtr);

    return TM_DEV_OKAY;
}

/*
 * tfDm642EmacIoctl()
 *
 * Handles general I/O control functions for the EMAC driver.  Currently,
 * the only function needed is to refill the receive queue.  Asynchronously
 * refilling the recieve queue with empty buffers is important to avoid a
 * race condition if the EMAC completely fills the receive queue.
 *
 */
int tfDm642EmacIoctl ( ttUserInterface interfaceHandle,
                  int             flag,
                  void TM_FAR *   optionPtr,
                  int             optionLen )
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    unsigned int    mcastHash0;
    unsigned int    mcastHash1;
    unsigned char  *multiAddr;
    int             loopIndex;
    unsigned char   currByte;
    unsigned char   tempHash;
#endif /* (TM_IGMP or TM_6_USE_MLD) */
    unsigned int    mdioStatus, linkStatus;
    ttEmacInfoPtr   infoPtr;
    unsigned int    temp;
    
    infoPtr = &emacInfo;

/* cleanup EMAC transmit frame descriptors and call tfSendCompleteInterface
   for packets that have completed transmit */
    if (flag & TM_EMAC_IOCTL_REFILL_TX)
    {
        tfCleanupXmitQueue(infoPtr, interfaceHandle);
    }

/* refill empty EMAC receive frame descriptors with Treck packet buffers */
    if (flag & TM_EMAC_IOCTL_REFILL_RX)
    {
        tfRefillRecvList(infoPtr);
    }

/* check for and set full duplex or half duplex EMAC operation */
    if (flag & TM_EMAC_IOCTL_SET_DUPLEX)
    {
        /* Signal the MDIO */
        mdioStatus = MDIO_timerTick(infoPtr->emacMdioPtr);

        /* On a new link, set the EMAC duplex */
        if( mdioStatus == MDIO_EVENT_LINKUP )
        {
            MDIO_getStatus(infoPtr->emacMdioPtr, 0, &linkStatus );

            if( linkStatus == MDIO_LINKSTATUS_FD10 ||
                linkStatus == MDIO_LINKSTATUS_FD100 )
            {
/* enable full-duplex operation */
                temp = tm_emac_get_dword(TM_EMAC_REG_MACCONTROL);
                temp |= TM_EMAC_MACCONTROL_FULLDUPLEX;
                tm_emac_set_dword(TM_EMAC_REG_MACCONTROL, temp);
            }
            else
            {
/* disable full-duplex operation */
                temp = tm_emac_get_dword(TM_EMAC_REG_MACCONTROL);
                temp &= ~TM_EMAC_MACCONTROL_FULLDUPLEX;
                tm_emac_set_dword(TM_EMAC_REG_MACCONTROL, temp);
            }
        }
    }

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/*
 * Setup our list of multicast addresses to receive.
 */
    if (flag & TM_DEV_SET_MCAST_LIST)
    {
/* disable multicast reception when optionLen is 0 */
        mcastHash0 = 0;
        mcastHash1 = 0;

/* Get the first address */
        multiAddr = (unsigned char *) optionPtr;

/* for each multicast address in the multicast address list */
        for (;optionLen > 0; optionLen--)
        {
            tempHash = 0;

/* For all bytes in the (6-byte) address */
            for (loopIndex = 0; loopIndex < 2; loopIndex++)
            {
                currByte = *multiAddr++;
                tempHash ^= (currByte >> 2) ^ (currByte << 4);
                currByte = *multiAddr++;
                tempHash ^= (currByte >> 4) ^ (currByte << 2);
                currByte = *multiAddr++;
                tempHash ^= (currByte >> 6) ^ currByte;
            }

            if (tempHash & 0x20)
            {
                mcastHash1 |= (1 << (tempHash & 0x1f));
            }
            else
            {
                mcastHash0 |= (1 << (tempHash & 0x1f));
            }
        }

/* Setup the Rx Filter to receive broadcast and multicast packets */
        tm_emac_set_dword(
            TM_EMAC_REG_RXMBP_ENABLE,
            (TM_EMAC_MBP_BROADCAST_EN + TM_EMAC_MBP_MULTICAST_EN) );

        tm_emac_set_dword( TM_EMAC_REG_MC_HASH0, mcastHash0);
        tm_emac_set_dword( TM_EMAC_REG_MC_HASH1, mcastHash1);
    }

/*
 * Set to receive all multicast packets.
 */
    if (flag & TM_DEV_SET_ALL_MCAST)
    {
/* Setup the Rx Filter to receive all broadcast and multicast packets */
        tm_emac_set_dword(
            TM_EMAC_REG_RXMBP_ENABLE,
            (TM_EMAC_MBP_BROADCAST_EN + TM_EMAC_MBP_MULTICAST_EN) );

        tm_emac_set_dword( TM_EMAC_REG_MC_HASH0, 0xffffffff);
        tm_emac_set_dword( TM_EMAC_REG_MC_HASH1, 0xffffffff);
    }
#endif /* (TM_IGMP or TM_6_USE_MLD) */

    return TM_DEV_OKAY;
}

/*
 * tfDm642EmacSend()
 *
 * Queues a packet (possibly consisting of a number of scattered buffers) onto
 * the EMAC for transmission.  If other packets are still awaiting
 * transmission, the new packet is added to the tail of the queue.
 * Otherwise, the new packet is sent directly to the EMAC.
 *
 */
int tfDm642EmacSend ( ttUserInterface interfaceHandle,
                      char TM_FAR *   dataPtr,
                      int             dataLength,
                      int             flag )
{
    ttEmacInfoPtr    infoPtr;
    ttEmacPktDescPtr newDescPtr;
    ttEmacPktDescPtr newLastDescPtr;
    unsigned int     descFlags;
    unsigned short   newDataLength;
    int              cacheCleanWordLen;
    void            *cacheCleanAddr;
    unsigned int     currTxQueue;
    unsigned int     nextTxQueue;
    
    infoPtr = &emacInfo;
    newDescPtr = (ttEmacPktDescPtr)0;
    
/* Clean up the transmit queue */
    tfCleanupXmitQueue(infoPtr, interfaceHandle);

/* Get packet descriptor(s) from queue. */
/* avoid transmits starving receives */
    if (infoPtr->emacAvailBufDescsCount > TM_EMAC_MIN_RX_DESC)
    {
        newDescPtr = tfPopBufDesc(infoPtr);
    }

    if (newDescPtr == (ttEmacPktDescPtr)0)
    {
/* No buffer descriptors are available.  Cleanup the transmit queue to free
 * any already transmitted packets, or to wait for currently queue packets to
 * be transmitted, making their buffer descriptors available for use. */
        while (newDescPtr == (ttEmacPktDescPtr) 0)
        {
            tfCleanupXmitQueue(infoPtr, interfaceHandle);
/* Re-attempt to allocate buffer descriptor. */
/* avoid transmits starving receives */
            if (infoPtr->emacAvailBufDescsCount > TM_EMAC_MIN_RX_DESC)
            {
                newDescPtr = tfPopBufDesc(infoPtr);
            }
        }
    }

/* Fill out new packet descriptor. */
    newDescPtr->emacDescNextPtr = (ttEmacPktDescPtr) 0;
    newDescPtr->emacDescBufPtr  = dataPtr;
    newDescPtr->emacDescOffsetBufLength = dataLength;

/* flush the tx packet data in the L2 cache to external memory */
    cacheCleanWordLen = (dataLength + TM_CACHE_L2_LINESIZE + 3) >> 2;
    cacheCleanAddr = (void*) (((unsigned int) dataPtr)
                              & ~(TM_CACHE_L2_LINESIZE - 1));
    CACHE_clean(CACHE_L2, cacheCleanAddr, cacheCleanWordLen);

/*
 * Packet transmission is handled by two queues.  The first queue (the
 * "transmit" queue) is synchronized with the EMAC.  If no packets are in the
 * queue, a new queue is started with the new packet as the head and the
 * transmitter is restarted.  However, if packets are queued awaiting
 * transmission, the new packet is simply added to the queue at the tail and
 * there is no need to restart the transmitter (one exception to this involves
 * the EOQ flag - see below.
 *
 * The second queue (the "transmit input" queue) is managed completely by the
 * driver.  This queue serves to queue buffers in a packet that consists of
 * multiple buffers (a "scattered send").  This queue is formed prior to
 * transmission, and when the final buffer in a packet is received, the entire
 * queue is added to the main transmit queue, just as a single packet would be
 * as described above.  While the transmit input queue queues multiple buffers
 * in a packet, it is also used when a packet consists of a single buffer, in
 * which case the queue also only contains a single buffer.  In either case,
 * the methodology is the same.
 */
    
/* Add buffer to transmit input queue. */
    if (infoPtr->emacFirstXmitInputPtr == (ttEmacPktDescPtr)0)
    {
/* First buffer in this packet, so start the queue and set the start of packet
   flag (SOP). */
        infoPtr->emacFirstXmitInputPtr = newDescPtr;
        infoPtr->emacLastXmitInputPtr  = newDescPtr;
        descFlags = (TM_EMAC_XMIT_OWNER | TM_EMAC_XMIT_SOP);
        newDescPtr->emacDescFlagsPktLength = dataLength | (descFlags << 16);
    }
    else
    {
/* Additional buffers in a packet; add them to the queue and update the
   total length. */
        infoPtr->emacLastXmitInputPtr->emacDescNextPtr = newDescPtr;
        infoPtr->emacLastXmitInputPtr = newDescPtr;
        newDescPtr->emacDescFlagsPktLength = dataLength;
        newDataLength =
            infoPtr->emacFirstXmitInputPtr->emacDescFlagsPktLength & 0xffff;
        newDataLength += dataLength;
        infoPtr->emacFirstXmitInputPtr->emacDescFlagsPktLength =
            newDataLength | ((TM_EMAC_XMIT_OWNER | TM_EMAC_XMIT_SOP) << 16);
    }

    if (flag == TM_USER_BUFFER_LAST)
    {
/* Last buffer in the packet.  Set the EOP flag and prepare to submit the
 * entire transmit input queue (which may consist of as little as one buffer)
 * to the EMAC.  Reset the transmit input queue. */
        newDescPtr->emacDescFlagsPktLength |= (TM_EMAC_XMIT_EOP << 16);
        newDescPtr = infoPtr->emacFirstXmitInputPtr;
        newLastDescPtr = infoPtr->emacLastXmitInputPtr;
        infoPtr->emacFirstXmitInputPtr = (ttEmacPktDescPtr) 0;
        infoPtr->emacLastXmitInputPtr = (ttEmacPktDescPtr) 0;        
    }
    else
    {
/* More buffers to follow... */
        return TM_DEV_OKAY;
    }

/* Get the current and next hardware TX queues */
    currTxQueue = infoPtr->emacEmptyTxQueue;
    nextTxQueue = (currTxQueue + 1) % TM_EMAC_TX_QUEUES;

/* If another hardware queue is available, put a packet in this one
 * and move to the next */
    if (nextTxQueue != infoPtr->emacFullTxQueue)
    {
        infoPtr->emacFirstXmitPtr[nextTxQueue] = newDescPtr;
        infoPtr->emacLastXmitPtr[nextTxQueue]  = newLastDescPtr;
        infoPtr->emacEmptyTxQueue = nextTxQueue;
/* start the transmitter */
        tm_emac_set_dword( TM_EMAC_REG_TXHDP + nextTxQueue*4, newDescPtr );
    }
/* Else all hardware queues are in use. It's safe to append to the end of
 * this one - the hardware's working on a much earlier queue */
    else
    { 
        infoPtr->emacLastXmitPtr[currTxQueue]->emacDescNextPtr = newDescPtr;
        infoPtr->emacLastXmitPtr[currTxQueue] = newLastDescPtr;
    }

    return TM_DEV_OKAY;
}

#ifdef TM_USE_DRV_ONE_SCAT_SEND
/*
 * tfDm642EmacScatSend
 * Scatter send function.
 */
int tfDm642EmacScatSend(
    ttUserInterface interfaceHandle,
    ttUserPacketPtr  uPacketPtr )
{
    char TM_FAR *    dataPtr;
    int              dataLength;
    int              scatDataLen;
    int              scatterIndex;
    int              errorCode;
    ttEmacInfoPtr    infoPtr;
    ttEmacPktDescPtr newDescPtr;
    ttEmacPktDescPtr newLastDescPtr;
    unsigned int     descFlags;
    int              cacheCleanWordLen;
    void            *cacheCleanAddr;
    unsigned int     currTxQueue;
    unsigned int     nextTxQueue;

    scatterIndex = uPacketPtr->pktuLinkExtraCount;
    dataLength = uPacketPtr->pktuChainDataLength;

    if (scatterIndex == 0)
    {
/* if this is not a scattered send, then call the regular driver send */
        errorCode = tfDm642EmacSend(
            interfaceHandle,
            (char TM_FAR *) uPacketPtr->pktuLinkDataPtr,
            dataLength, TM_USER_BUFFER_LAST); 
    }
    else
    {
        errorCode = TM_DEV_OKAY;
        infoPtr = &emacInfo;

        for (; scatterIndex >= 0; scatterIndex--)
        {
            dataPtr = (char TM_FAR *) uPacketPtr->pktuLinkDataPtr;
            scatDataLen = uPacketPtr->pktuLinkDataLength;
            uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 

/* Get packet descriptor(s) from queue. */
            newDescPtr = (ttEmacPktDescPtr)0;
/* avoid transmits starving receives */
            if (infoPtr->emacAvailBufDescsCount > TM_EMAC_MIN_RX_DESC)
            {
                newDescPtr = tfPopBufDesc(infoPtr);
            }

            if (newDescPtr == (ttEmacPktDescPtr)0)
            {
/* No buffer descriptors are available.  Cleanup the transmit queue to free
 * any already transmitted packets, or to wait for currently queue packets to
 * be transmitted, making their buffer descriptors available for use. */
                while (newDescPtr == (ttEmacPktDescPtr) 0)
                {
                    tfCleanupXmitQueue(infoPtr, interfaceHandle);
/* Re-attempt to allocate buffer descriptor. */
/* avoid transmits starving receives */
                    if (infoPtr->emacAvailBufDescsCount > TM_EMAC_MIN_RX_DESC)
                    {
                        newDescPtr = tfPopBufDesc(infoPtr);
                    }
                }
            }

/* Fill out new packet descriptor. */
            newDescPtr->emacDescNextPtr = (ttEmacPktDescPtr) 0;
            newDescPtr->emacDescBufPtr  = dataPtr;
            newDescPtr->emacDescOffsetBufLength = scatDataLen;

/* flush the tx packet data in the L2 cache to external memory */
            cacheCleanWordLen = (scatDataLen + TM_CACHE_L2_LINESIZE + 3) >> 2;
            cacheCleanAddr = (void*) (((unsigned int) dataPtr)
                                      & ~(TM_CACHE_L2_LINESIZE - 1));
            CACHE_clean(CACHE_L2, cacheCleanAddr, cacheCleanWordLen);

/*
 * Packet transmission is handled by two queues.  The first queue (the
 * "transmit" queue) is synchronized with the EMAC.  If no packets are in the
 * queue, a new queue is started with the new packet as the head and the
 * transmitter is restarted.  However, if packets are queued awaiting
 * transmission, the new packet is simply added to the queue at the tail and
 * there is no need to restart the transmitter (one exception to this involves
 * the EOQ flag - see below.
 *
 * The second queue (the "transmit input" queue) is managed completely by the
 * driver.  This queue serves to queue buffers in a packet that consists of
 * multiple buffers (a "scattered send").  This queue is formed prior to
 * transmission, and when the final buffer in a packet is received, the entire
 * queue is added to the main transmit queue, just as a single packet would be
 * as described above.  While the transmit input queue queues multiple buffers
 * in a packet, it is also used when a packet consists of a single buffer, in
 * which case the queue also only contains a single buffer.  In either case,
 * the methodology is the same.
 */
    
/* Add buffer to transmit input queue. */
            if (infoPtr->emacFirstXmitInputPtr == (ttEmacPktDescPtr)0)
            {
/* First buffer in this packet, so start the queue and set the start of packet
   flag (SOP). */
                infoPtr->emacFirstXmitInputPtr = newDescPtr;
                infoPtr->emacLastXmitInputPtr  = newDescPtr;
                descFlags = (TM_EMAC_XMIT_OWNER | TM_EMAC_XMIT_SOP);
                newDescPtr->emacDescFlagsPktLength =
                    dataLength | (descFlags << 16);
            }
            else
            {
/* Additional buffers in a packet; add them to the queue and update the
   total length. */
                infoPtr->emacLastXmitInputPtr->emacDescNextPtr = newDescPtr;
                infoPtr->emacLastXmitInputPtr = newDescPtr;
                newDescPtr->emacDescFlagsPktLength = scatDataLen;
            }

            if (scatterIndex == 0)
            {
/* Last buffer in the packet.  Set the EOP flag and prepare to submit the
 * entire transmit input queue (which may consist of as little as one buffer)
 * to the EMAC.  Reset the transmit input queue. */
                newDescPtr->emacDescFlagsPktLength |= (TM_EMAC_XMIT_EOP << 16);
                newDescPtr = infoPtr->emacFirstXmitInputPtr;
                newLastDescPtr = infoPtr->emacLastXmitInputPtr;
                infoPtr->emacFirstXmitInputPtr = (ttEmacPktDescPtr) 0;
                infoPtr->emacLastXmitInputPtr = (ttEmacPktDescPtr) 0;        
            }
        }

/* Clean the Xmit queue and update which hardware TX queues are free */
        tfCleanupXmitQueue(infoPtr, interfaceHandle);

        currTxQueue = infoPtr->emacEmptyTxQueue;
        nextTxQueue = (currTxQueue + 1) % TM_EMAC_TX_QUEUES;

/* If another hardware queue is available, put a packet in this one
 * and move to the next */
        if (nextTxQueue != infoPtr->emacFullTxQueue)
        {
            infoPtr->emacFirstXmitPtr[nextTxQueue] = newDescPtr;
            infoPtr->emacLastXmitPtr[nextTxQueue]  = newLastDescPtr;
            infoPtr->emacEmptyTxQueue = nextTxQueue;
/* start the transmitter */
            tm_emac_set_dword( TM_EMAC_REG_TXHDP + nextTxQueue*4, newDescPtr );
        }
/* Else all hardware queues are in use. It's safe to append to the end of
 * this one - the hardware's working on a much earlier queue */
        else
        {
            infoPtr->emacLastXmitPtr[currTxQueue]->emacDescNextPtr = newDescPtr;
            infoPtr->emacLastXmitPtr[currTxQueue] = newLastDescPtr;
        }

    }

    return errorCode;
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

/*
 * tfDm642EmacRecv()
 *
 * Removes a packet from the receive queue and pass it up to the Treck stack,
 * and then refill the receive queue.
 *
 */
int tfDm642EmacRecv(ttUserInterface        interfaceHandle,
                    char TM_FAR * TM_FAR * dataPtrPtr,
                    int  TM_FAR *          dataLengthPtr,
                    ttUserBufferPtr        bufHandlePtr)
{
    ttEmacPktDescPtr currDescPtr;
    ttEmacInfoPtr    infoPtr;
    ttEmacPktDescPtr nextDescPtr;
    
    infoPtr = &emacInfo;
    currDescPtr = infoPtr->emacFirstRecvPtr;

    tm_assert(
        tfDm642EmacRecv,
        (!(currDescPtr->emacDescFlagsPktLength & (TM_EMAC_RECV_OWNER << 16))));

/*
 * Check for End Of Queue indicator to determine if the receiver needs to be
 * restarted (see note in tfEmacSend).
 */
    if (currDescPtr->emacDescFlagsPktLength & (TM_EMAC_RECV_EOQ << 16))
    {
/* restart receive at the next receive descriptor */
        nextDescPtr = currDescPtr->emacDescNextPtr;
        if (nextDescPtr != (ttEmacPktDescPtr) 0)
        {
            infoPtr->emacIsrFirstRecvPtr = nextDescPtr;
            tm_emac_set_dword(TM_EMAC_REG_RXHDP, nextDescPtr);
        }
    }

/* Get data from RX descriptor to pass buffer up to the stack. */
    *dataPtrPtr = currDescPtr->emacDescBufPtr;
    *dataLengthPtr = currDescPtr->emacDescFlagsPktLength & 0xffff;
    *bufHandlePtr = currDescPtr->emacDescUserBufPtr;

/*
 * Remove the RX descriptor from the receive queue and return it to the
 * general buffer queue.
 */
    infoPtr->emacFirstRecvPtr = currDescPtr->emacDescNextPtr;
    infoPtr->emacRecvDescCount--;
    tfPushBufDesc(infoPtr, currDescPtr);

    tfRefillRecvList(infoPtr);
    
    return TM_DEV_OKAY;
}


/*
 * tfDm642EmacIsrHandler()
 *
 * Handles and acknowledges EMAC interrupts to the DM642.  For RX interrupts,
 * notify the stack that there is a new packet available for reception.
 * Ignore all other types of interrupts.
 * 
 */
void tfDm642EmacIsrHandler(void)
{
    unsigned int regValue;
    unsigned int intSource;
    int countRxComplete;
    ttEmacPktDescPtr rxDescPtr;
    ttEmacPktDescPtr rxCompleteDescPtr;

/* Disable EMAC interrupts. */
    regValue = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    regValue &= ~TM_EMAC_EWCTL_INT_ENABLE;
    tm_emac_set_dword( TM_EMAC_REG_EWCTL, regValue);

    intSource = tm_emac_get_dword( TM_EMAC_REG_MACINVECTOR );

/* RX interrupt, notify stack of incoming packet, ACK interrupt. */
    if (intSource & TM_EMAC_MACINVECTOR_RX)
    {
        regValue = tm_emac_get_dword( TM_EMAC_RX0INTACK );
        tm_emac_set_dword( TM_EMAC_RX0INTACK, regValue );

        rxDescPtr = emacInfo.emacIsrFirstRecvPtr;
/* rxCompleteDescPtr is the last RX frame descriptor completed by the MAC */
        rxCompleteDescPtr = (ttEmacPktDescPtr) regValue;

/* count how many packets we have received, always at least 1 */
        for (countRxComplete = 1; 
             rxDescPtr != rxCompleteDescPtr;
             rxDescPtr = rxDescPtr->emacDescNextPtr)
             
        {
            countRxComplete++;
        }

/* notify Treck TCP how many packets we have received */
        tfNotifyInterfaceIsr(
            emacInterfaceHandle, countRxComplete, 0, 0, 0);

/* advance to the next RX frame descriptor we expect to receive into */
        emacInfo.emacIsrFirstRecvPtr = rxCompleteDescPtr->emacDescNextPtr;
    }

/* Ignore any other sorts of interrupts (we shouldn't get any anyways) */

/* Re-enable EMAC interrupts. */
    regValue = tm_emac_get_dword( TM_EMAC_REG_EWCTL );
    regValue |= TM_EMAC_EWCTL_INT_ENABLE;
    tm_emac_set_dword( TM_EMAC_REG_EWCTL, regValue);
}

/*
 * tfDm642EmacGetPhysicalAddress()
 *
 * Returns the physical (MAC) address of the controller.
 *
 */
int tfDm642EmacGetPhysicalAddress(ttUserInterface interfaceHandle,
                                  char TM_FAR *   physicalAddress)
{
/* This line is to get rid of compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Return a sample address; in a real application the MAC address would
 * probably be read from an EEPROM, etc. */
    physicalAddress[0] = macAddrFromApp[0];
    physicalAddress[1] = macAddrFromApp[1];
    physicalAddress[2] = macAddrFromApp[2];
    physicalAddress[3] = macAddrFromApp[3];
    physicalAddress[4] = macAddrFromApp[4];
    physicalAddress[5] = macAddrFromApp[5];

    return TM_DEV_OKAY;
    
}

/*
 * Internal helper functions.
 */

static void tfCleanupXmitQueue(ttEmacInfoPtr   infoPtr,
                               ttUserInterface interfaceHandle)
{
    ttEmacPktDescPtr currDescPtr;
    ttEmacPktDescPtr nextDescPtr;
    unsigned int     descFlags;
    unsigned int     emacOwnsFlags;
    unsigned int     currTxQueue;

/* Find which queue is the first "used" queue */
    currTxQueue = infoPtr->emacFullTxQueue;

/* Loop through all of the hardware queues currently in use */
    while (currTxQueue != infoPtr->emacEmptyTxQueue)
    {
    
/* Clean up this transmitter queue, calling tfSendComplete */
        currDescPtr = infoPtr->emacFirstXmitPtr[currTxQueue];
        while (currDescPtr != (ttEmacPktDescPtr) 0)
        {
            nextDescPtr = currDescPtr->emacDescNextPtr;
            emacOwnsFlags = TM_EMAC_XMIT_OWNER | TM_EMAC_XMIT_SOP;
            descFlags = currDescPtr->emacDescFlagsPktLength >> 16;
            if ( (descFlags & emacOwnsFlags) == emacOwnsFlags)
            {
/* Reached the first descriptor still owned by the controller; stop. */
                break;
            }
            else
            {
                if (descFlags & TM_EMAC_XMIT_EOP)
                {
/* A full packet has been transmitted, so notify the stack. */
                    tfSendCompleteInterface(interfaceHandle,
                                            TM_DEV_SEND_COMPLETE_DRIVER);
                }

/* Controller has transmitted the buffer attached to this descriptor, so
 * free the descriptor back to the queue to be reallocated.
 */
                tfPushBufDesc(infoPtr, currDescPtr);

/* 
 * Even if we're adding a packet to the transmitter's queue, there's a
 * possibility that the transmitter is stopped, due to a race condition
 * with setting the 'next descriptor pointer' (see SPRU628, section 2.3.2).
 * This is resolved by checking the queue for a descriptor with the
 * EOQ (End Of Queue) flag set which indicates that the transmitter believes
 * it reached the end of the queue and has stopped.  In this case, we need
 * to restart the transmitter with the descriptor following the one with
 * the EOQ flag set.
 */
                if ((descFlags & TM_EMAC_XMIT_EOQ)
                    && (nextDescPtr != (ttEmacPktDescPtr) 0))
                {
                    tm_emac_set_dword( TM_EMAC_REG_TXHDP + currTxQueue*4, nextDescPtr );
                }
            }
        
            currDescPtr = nextDescPtr;
        }

/* Update the transmit queue. */
    	infoPtr->emacFirstXmitPtr[currTxQueue] = currDescPtr;

/* If we cleaned up the entire list, clear the last descriptor pointer
 * and update the dirty hardware queue indicator */
        if (currDescPtr == (ttEmacPktDescPtr) 0)
        {
            infoPtr->emacLastXmitPtr[currTxQueue] = (ttEmacPktDescPtr) 0;
            currTxQueue = (currTxQueue + 1) % TM_EMAC_TX_QUEUES;
        }
        else
        {
/* We didn't clean this entire queue, so we can't try to clean
 * the next queue */
            break;
        }
    }

/* Note which queue is the first "used" queue */
    infoPtr->emacFullTxQueue = currTxQueue;
}    

/* Refill the receive queue with empty packets. */
static void tfRefillRecvList(ttEmacInfoPtr infoPtr)
{
    ttEmacPktDescPtr newDescPtr;
    ttEmacPktDescPtr lastDescPtr;

    lastDescPtr = infoPtr->emacLastRecvPtr;

    while (infoPtr->emacRecvDescCount < TM_EMAC_MAX_RX_DESC)
    {
/* Create new RX descriptor and allocate buffer to recieve data into. */
        newDescPtr = tfPopBufDesc(infoPtr);
        if (newDescPtr == (ttEmacPktDescPtr)0)
        {
/* No new buffer descriptor is available, so stop refilling. */
            break;
        }
        
        tm_bzero(newDescPtr, sizeof(ttEmacPktDesc));

/* Allocate data buffer from the stack and attach it to the buffer desc.
   NOTE: buffer length must be a multiple of the cache line size,
   and buffer must be aligned on a cache line boundary (128-byte aligned) */
        newDescPtr->emacDescBufPtr = (char TM_FAR *)
            tfGetDriverBuffer(
                &newDescPtr->emacDescUserBufPtr,
                TM_EMAC_RX_BUF_LEN,
                TM_CACHE_L2_LINESIZE);
        if (newDescPtr->emacDescBufPtr == (char TM_FAR *) 0)
        {
/* No data buffer is available from stack.  Return the buffer descriptor to
 * the pool and stop refilling. */
            tfPushBufDesc(infoPtr, newDescPtr);
            break;
        }

	tm_assert(
            tfRefillRecvList,
            ((((int) newDescPtr->emacDescBufPtr) & 127) == 0));

/* invalidate the rx packet buffer in the L2 cache */
        CACHE_clean(
            CACHE_L2,
            (void TM_FAR *) newDescPtr->emacDescBufPtr,
            TM_EMAC_RX_BUF_LEN/4);

/* TM_ETHER_IP_ALIGN is added to the Ethernet buffer so that the TCP/IP
   header is aligned on a 4-byte boundary. */
        newDescPtr->emacDescBufPtr += TM_ETHER_IP_ALIGN;

        newDescPtr->emacDescOffsetBufLength = TM_EMAC_MAX_PACKET_LEN;
        newDescPtr->emacDescFlagsPktLength = TM_EMAC_RECV_OWNER << 16;

        if (lastDescPtr == (ttEmacPktDescPtr)0)
        {
/* start a new queue of rx frame descriptors. */
            infoPtr->emacLastRecvPtr = newDescPtr;
            infoPtr->emacFirstRecvPtr = newDescPtr;
            infoPtr->emacIsrFirstRecvPtr = newDescPtr;
/* Give the MAC the address of the first receive frame descriptor */
            tm_emac_set_dword(TM_EMAC_REG_RXHDP, newDescPtr);
        }
        else
        {
/* add this rx frame descriptor to the end of an existing queue */
            lastDescPtr->emacDescNextPtr = newDescPtr;
            infoPtr->emacLastRecvPtr = newDescPtr;

/*
 * Check for End Of Queue indicator to determine if the receiver needs to be
 * restarted (see note above in tfEmacSend).
 */
            if (lastDescPtr->emacDescFlagsPktLength & (TM_EMAC_RECV_EOQ << 16))
            {
/* restart receive at the newly added receive descriptor */
                infoPtr->emacIsrFirstRecvPtr = newDescPtr;
                tm_emac_set_dword(TM_EMAC_REG_RXHDP, newDescPtr);

/* reset the EOQ bit here, so that we don't try to restart receive twice */
                lastDescPtr->emacDescFlagsPktLength &=
                    ~(TM_EMAC_RECV_EOQ << 16);
            }
        }
        lastDescPtr = newDescPtr;

        infoPtr->emacRecvDescCount++;            
    }
}

/* Return a buffer descriptor to the static descriptor pool. */
static void tfPushBufDesc(ttEmacInfoPtr infoPtr, ttEmacPktDescPtr descPtr)
{
    descPtr->emacDescNextPtr = infoPtr->emacBufDescQ;
    infoPtr->emacBufDescQ = descPtr;
    infoPtr->emacAvailBufDescsCount++;
    tm_assert(
        tfPushBufDesc,
        (infoPtr->emacAvailBufDescsCount <= TM_EMAC_DSC_ENTRY_COUNT));
}


/* Get a buffer descriptor from the static descriptor pool. */
static ttEmacPktDescPtr tfPopBufDesc(ttEmacInfoPtr infoPtr)
{
    ttEmacPktDescPtr descPtr;

    descPtr = infoPtr->emacBufDescQ;

    if (descPtr != (ttEmacPktDescPtr)0)
    {
        infoPtr->emacBufDescQ = descPtr->emacDescNextPtr;
        tm_assert(tfPopBufDesc, (infoPtr->emacAvailBufDescsCount != 0));
        infoPtr->emacAvailBufDescsCount--;
    }

    return descPtr;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: trdm6mac.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 56056    Output checksum: 40100
 */
