/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: ZiLOG eZ80 F91 EMAC Ethernet driver
 *
 * Filename: tremac.c
 * Author: Qin Zhang
 * Date Created: 01/29/2003
 * $Source: source/drivers/emac/tremac.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:00CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Developed on an ZiLOG eZ80 F91 board ( EMAC Ethernet controller integrated
 * in ZiLOG eZ80 F91 processor) using ZDS II compiler
 */

/*
 * Include the Treck Interface types
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <tremac.h>

#ifdef TM_KERNEL_RZK_EZ80
#include "zsysgen.h"
#endif /* TM_KERNEL_RZK_EZ80 */

/* Interrupt Vectors for EMAC */
#define TM_IV_EMAC_RX    0x40
#define TM_IV_EMAC_TX    0x44
#define TM_IV_EMAC_SYS   0x48


#define TM_EMAC_PACKET_TYPE_OFFSET_1 12    /* Offset of start of packet-type
                                             * field from beginning of buffer */
#define TM_EMAC_PACKET_TYPE_OFFSET_2 13    /* Offset of end of packet-type
                                             * field from beginning of buffer */
#define TM_EMAC_INVALID_PACKET_TYPE  ((unsigned char)0xFF)
                                            /* An invalid packet type to get the
                                             * stack to discard packets with
                                             * errors in them */

/*
 * EMAC register definitions
 */
/* EMAC Test Register. Rest: 00 */
#define TM_EMAC_TEST        0x20
/* EMAC Configuration Register. Rest: 00 */
#define TM_EMAC_CFG1        0x21
/* EMAC Configuration Register. Rest: 37 */
#define TM_EMAC_CFG2        0x22
/* EMAC Configuration Register. Rest: 0F */
#define TM_EMAC_CFG3        0x23
/* EMAC Configuration Register. Rest: 00 */
#define TM_EMAC_CFG4        0x24
/* EMAC Station Address--Byte 0. Rest: 00 */
#define TM_EMAC_STAD_0      0x25
/* EMAC Station Address--Byte 1. Rest: 00 */
#define TM_EMAC_STAD_1      0x26
/* EMAC Station Address--Byte 2. Rest: 00 */
#define TM_EMAC_STAD_2      0x27
/* EMAC Station Address--Byte 3. Rest: 00 */
#define TM_EMAC_STAD_3      0x28
/* EMAC Station Address --Byte 4. Rest: 00 */
#define TM_EMAC_STAD_4      0x29
/* EMAC Station Address--Byte 5. Rest: 00 */
#define TM_EMAC_STAD_5      0x2A
/* EMAC Transmit Pause Timer Value--Low Byte. Rest: 00 */
#define TM_EMAC_TPTV_L      0x2B
/* EMAC Transmit Pause Timer Value--High Byte. Rest: 00 */
#define TM_EMAC_TPTV_H      0x2C
/* EMAC Inter-Packet Gap. Rest: 15 */
#define TM_EMAC_IPGT        0x2D
/* EMAC Non-Back-Back IPG. Rest: 0C */
#define TM_EMAC_IPGR1       0x2E
/* EMAC Non-Back-Back IPG. Rest: 12 */
#define TM_EMAC_IPGR2       0x2F
/* EMAC Maximum Frame Length--Low Byte. Rest: 00 */
#define TM_EMAC_MAXF_L      0x30
/* EMAC Maximum Frame Length--High Byte. Rest: 06 */
#define TM_EMAC_MAXF_H      0x31
/* EMAC Address Filter Register. Rest: 00 */
#define TM_EMAC_AFR         0x32
/* EMAC Hash Table--Byte 0. Rest: 00 */
#define TM_EMAC_HTBL_0      0x33
/* EMAC Hash Table--Byte 1. Rest: 00 */
#define TM_EMAC_HTBL_1      0x34
/* EMAC Hash Table--Byte 2. Rest: 00 */
#define TM_EMAC_HTBL_2      0x35
/* EMAC Hash Table--Byte 3. Rest: 00 */
#define TM_EMAC_HTBL_3      0x36
/* EMAC Hash Table--Byte 4. Rest: 00 */
#define TM_EMAC_HTBL_4      0x37
/* EMAC Hash Table--Byte 5. Rest: 00 */
#define TM_EMAC_HTBL_5      0x38
/* EMAC Hash Table --Byte 6. Rest: 00 */
#define TM_EMAC_HTBL_6      0x39
/* EMAC Hash Table--Byte 7. Rest: 00 */
#define TM_EMAC_HTBL_7      0x3A
/* EMAC MII Management Register. Rest: 00 */
#define TM_EMAC_MIIMGT      0x3B
/* EMAC PHY Configuration Data--Low Byte. Rest: 00 */
#define TM_EMAC_CTLD_L      0x3C
/* EMAC PHY Configuration Data--High Byte. Rest: 00 */
#define TM_EMAC_CTLD_H      0x3D
/* EMAC PHY Register Address Register. Rest: 00 */
#define TM_EMAC_RGAD        0x3E
/* EMAC PHY Unit Select Address Register. Rest: 00 */
#define TM_EMAC_FIAD        0x3F
/* EMAC Transmit Polling Timer Register. Rest: 00 */
#define TM_EMAC_PTMR        0x40
/* EMAC Reset Control Register. Rest: 00 */
#define TM_EMAC_RST         0x41
/* EMAC Transmit Lower Boundary Point--Low Byte. Rest: 00 */
#define TM_EMAC_TLBP_L      0x42
/* EMAC Transmit Lower Boundary Point--High Byte. Rest:00 */
#define TM_EMAC_TLBP_H      0x43
/* EMAC Boundary Pointer Low Byte. Rest: 00 */
#define TM_EMAC_BP_L        0x44
/* EMAC Boundary Pointer High Byte. Rest: C0 */
#define TM_EMAC_BP_H        0x45
/* EMAC Boundary Pointer Upper Byte. Rest: FF */
#define TM_EMAC_BP_U        0x46
/* EMAC Receive High Boundary Pointer--Low Byte. Rest: 00 */
#define TM_EMAC_RHBP_L      0x47
/* EMAC Receive High Boundary Pointer--High Byte. Rest: 00 */
#define TM_EMAC_RHBP_H      0x48
/* Receive Read Pointer Low Byte. Rest: 00 */
#define TM_EMAC_RRP_L       0x49
/* EMAC Receive Read Pointer High Byte. Rest: 00 */
#define TM_EMAC_RRP_H       0x4A
/* EMAC Buffer Size Register. Rest: 00 */
#define TM_EMAC_BUFSZ       0x4B
/* EMAC Interrupt Enable Register. Rest: 00 */
#define TM_EMAC_IEN         0x4C
/* EMAC Interrupt Status Register. Rest: 00 */
#define TM_EMAC_ISTAT       0x4D
/* EMAC PHY Read Status Data--Low Byte. Rest: 00 */
#define TM_EMAC_PRSD_L      0x4E
/* EMAC PHY Read Status Data--High Byte. Rest: 00 */
#define TM_EMAC_PRSD_H      0x4F
/* EMAC MII Status Register. Rest: 00 */
#define TM_EMAC_MIISTAT     0x50
/* EMAC Receive Write Pointer--Low Byte. Rest: 00 */
#define TM_EMAC_RWP_L       0x51
/* EMAC Receive Write Pointer--High Byte. Rest: 00 */
#define TM_EMAC_RWP_H       0x52
/* EMAC Transmit Read Pointer--Low Byte. Rest: 00 */
#define TM_EMAC_TRP_L       0x53
/* EMAC Transmit Read Pointer--High Byte. Rest: 00 */
#define TM_EMAC_TRP_H       0x54
/* EMAC Receive Blocks Left Register -- Low byte. Rest: 00 */
#define TM_EMAC_BLKSLFT_L   0x55
/* EMAC Receive Blocks Left Register -- High byte. Rest: 00 */
#define TM_EMAC_BLKSLFT_H   0x56
/* EMAC FIFO Data--Low Byte. Rest: 0X */
#define TM_EMAC_FDATA_L     0x57
/* EMAC FIFO Data--High Byte. Rest: XX */
#define TM_EMAC_FDATA_H     0x58
/* EMAC FIFO Flags Register. Rest: 33 */
#define TM_EMAC_FFLAGS      0x59
/* RAM control register */
#define TM_RAM_CTL          0xB4
/* RAM address upper byte register */
#define TM_RAM_ADDR_U       0xB5
/* MBIST Control */
#define TM_MBIST_GPR        0xB6
#define TM_MBIST_EMR        0xB7

/*
 * Register data definitions
 */
#define TM_RAM_ADDR_U_DATA  0x7A

/*
 * Build in memory boundary
 */
/* General purpose memory boundary */
#define TM_GRAM_LOW     (((unsigned int)TM_RAM_ADDR_U_DATA << 16) + 0xE000)
#define TM_GRAM_HIGH    (((unsigned int)TM_RAM_ADDR_U_DATA << 16) + 0xFFFF)

/* EMAC memory boundary */
#define TM_ERAM_LOW     (((unsigned int)TM_RAM_ADDR_U_DATA << 16) + 0xC000)
#define TM_ERAM_HIGH    (((unsigned int)TM_RAM_ADDR_U_DATA << 16) + 0xDFFF)

/* Boolean macros */
#define TM_EMAC_TRUE                 1     /* TRUE */
#define TM_EMAC_FALSE                0     /* FALSE */

/* EMAC buffer size: 32, 64, 128 or 256 bytes */
#define TM_EMAC_BUF_SIZE    32

/* The nubmer of Tx EMAC packet buffer */
#define TM_BUF_NUM_TX    2

/* The nubmer of Rx EMAC packet buffer */
#define TM_BUF_NUM_RX    3

/* Total number of EMAC packet buffer */
#define TM_BUF_NUM       (TM_BUF_NUM_TX + TM_BUF_NUM_RX)

/* Write out a word for specified port address */
#define tm_out_word(tp_value,tp_port) \
(*((volatile unsigned short __INTIO *)(tp_port)) = (unsigned short)(tp_value))
    
/* Write out a byte to specified port address */
#define tm_out_byte(tp_value,tp_port) \
(*((volatile unsigned char __INTIO *)(tp_port)) = (unsigned char) (tp_value))

/* Read in a word from specified port address */
#define tm_in_word(tp_port) \
(*((volatile unsigned short __INTIO *)(tp_port)))

/* Read in a byte from specified port address */
#define tm_in_byte(tp_port) \
(*((volatile unsigned char __INTIO *)(tp_port)))


/* Shared memory buffer is owned by EMAC */
#define tm_is_emac_own(stat)        (stat & 0x8000)

/* Give the memory buffer to EMAC to start a transmission */
#define tm_set_emac_own(stat)       ((stat) |= 0x8000)

/*
 * typedefs
 */
/* EMAC packet descriptor */
typedef struct tsEmacPkt
{
/* Pointer to the next descriptor bufIdx */
    struct tsEmacPkt *  epktNextPtr;
/* the pakcet size following this descriptor */
    ttUser16Bit         epktPktSize;
/* Descriptor status */
    ttUser16Bit         epktStat;
/* The ethernet packet */
/* 1529 = 1536 - (sizeof(epktNextPtr) + 4); */
    char                epktPkt[1529];
} ttEmacPkt;

typedef ttEmacPkt * ttEmacPktPtr;


/* EMAC info struct define */
typedef struct tsEmacInfo
{
/* Low boundary pointer of TX buffer */
    ttEmacPktPtr    infTxBufPtr[TM_BUF_NUM_TX];
/* Low boundary pointer of RX buffer */
    ttEmacPktPtr    infRxBufPtr[TM_BUF_NUM_RX];
/* Tx packet buffer flag */
    int             infTxFlag[TM_BUF_NUM_TX];
/* Current TX buffer pointer index */
    int             infCurTxIndex;
/* Current RX buffer pointer, for ISR */
    int             infCurRxIndex;
/* Receive index, for tfEmacReceive */
    int             infRecvIndex;
/* Interface handle */
    ttUserInterface infEmacIfHandle;
/* Counts the number of missed frames due to lack of buffer space */
    unsigned long   infMissedFrames;
/*
 * Counts the number of calls to driver recv when there is no buffer ready
 * to be received
 */
    unsigned long   infSpuriousInterrupts;
/* Indicate whether the user is using a Treck transmit queue */
    int             infXmitQueueOn;
} ttEmacInfo;

typedef ttEmacInfo * ttEmacInfoPtr;

/*
 * extern function definitions
 */
/* Emac ISR prolog for RZK */
#ifdef TM_KERNEL_RZK_EZ80
extern void tfEmacIsrProlog(void);
#else /* TM_KERNEL_RZK_EZ80 */
extern void tfZ80InstallIsr( ttUserIsrHandlerPtr   funcPtr,
                             int                   ivBase,
                             ttUser8Bit            offset);
#endif /* TM_KERNEL_RZK_EZ80 */


/*
 * Global variables
 */
ttEmacInfoPtr tlEmacPtr;


/*
 * Get the physical address
 * return it to the Treck stack
 * The address of the card needs to be changed - this is just test data
 */
int tfEmacGetPhysicalAddress( ttUserInterface emacHandle,
                              char TM_FAR *   phyAddrPtr)
{
/* This line is to get rid of compiler warnings */
    TM_UNREF_IN_ARG(emacHandle);

    phyAddrPtr[0] = 0x00;
    phyAddrPtr[1] = 0x11;
    phyAddrPtr[2] = 0x22;
    phyAddrPtr[3] = 0x33;
    phyAddrPtr[4] = 0x44;
    phyAddrPtr[5] = 0x55;

    return TM_ENOERROR;
}


/*
 * tfEmacOpen()
 *
 * Config eZ80 F91 EMAC, initializes various
 * registers and sets up the interrupt mechanisms.
 *
 */
int tfEmacOpen(ttUserInterface interfaceHandle)
{
    char                    hardwareAddr[6];
    unsigned char           regValue;
    unsigned int            i;
    int	                    tmp;
    int                     errorCode;

    tlEmacPtr = (ttEmacInfoPtr)tm_get_raw_buffer( sizeof(ttEmacInfo) );
    if (tlEmacPtr == (ttEmacInfoPtr)0) 
    {
        errorCode = TM_DEV_ERROR;
    }
    else
    {
        tlEmacPtr->infEmacIfHandle = interfaceHandle;
        tlEmacPtr->infMissedFrames = 0;
        tlEmacPtr->infSpuriousInterrupts = 0;

/* Set RAM address upper byte */
        regValue = (ttUser8Bit)TM_RAM_ADDR_U_DATA;
        tm_out_byte(regValue, TM_RAM_ADDR_U);

/* Enable internal RAM */
        regValue = 0xC0; /* Enable GP RAM and EMAC RAM */
        tm_out_byte(regValue, TM_RAM_CTL);

/* Reset EMAC */
        regValue = 0x3F;
        tm_out_byte(regValue, TM_EMAC_RST);
        regValue = 0x00;
        tm_out_byte(regValue, TM_EMAC_RST);

/* wait 10ms */
        i = 0;
        while(i<1000)
        {
            i++;
        }

/* Set buffer pointer registers */
/* Set TLBP */
        regValue = (char)(TM_ERAM_LOW & 0xFF);
        tm_out_byte(regValue, TM_EMAC_TLBP_L);

        regValue = (char)((TM_ERAM_LOW & 0xFF00) >> 8);
        tm_out_byte(regValue, TM_EMAC_TLBP_H);

/* Set BP */
        regValue =
            (char)((TM_ERAM_LOW + TM_BUF_NUM_TX * sizeof(ttEmacPkt)) & 0xFF);
        tm_out_byte(regValue, TM_EMAC_BP_L);

        regValue = (char)(((TM_ERAM_LOW +
            TM_BUF_NUM_TX * sizeof(ttEmacPkt)) & 0xFF00) >> 8);
        tm_out_byte(regValue, TM_EMAC_BP_H);

/* Set RHBP */
        regValue =
            (char)((TM_ERAM_LOW + TM_BUF_NUM * sizeof(ttEmacPkt)) & 0xFF);
        tm_out_byte(regValue, TM_EMAC_RHBP_L);

        regValue = (char)
            (((TM_ERAM_LOW + TM_BUF_NUM * sizeof(ttEmacPkt)) & 0xFF00) >> 8);
        tm_out_byte(regValue, TM_EMAC_RHBP_H);

#ifdef TM_TEST_ERAM
/* Start EMAC RAM test */
        regValue = 0x80;
        tm_out_byte(regValue, TM_MBIST_EMR);

/* Test end */
        regValue = tm_in_byte(TM_MBIST_EMR);
        while(regValue & 0x80)
        {
            for(i = 0; i< 1000; i++);
            regValue = tm_in_byte(TM_MBIST_EMR);
        }

/* Check result */
        if(regValue != 0x60)
        {
            tfKernelError("tfEmacOpen", "EMAC RAM test failed!");
        }
#endif /* TM_TEST_ERAM */

/* Initialize TX buffer ring */
        for( i = 0; i < TM_BUF_NUM_TX; i++ )
        {
	        tmp = (int)(TM_ERAM_LOW + i * sizeof(ttEmacPkt));
            tlEmacPtr->infTxBufPtr[i] = (ttEmacPktPtr)tmp;
/*            (ttEmacPktPtr)(TM_ERAM_LOW + i * sizeof(ttEmacPkt)); */
            
/* Link all TX EMAC buffer to form a TX ring */
            tmp =TM_ERAM_LOW + ((i + 1)%TM_BUF_NUM_TX) * sizeof(ttEmacPkt);
            (tlEmacPtr->infTxBufPtr[i])->epktNextPtr = (ttEmacPktPtr)tmp;
/*            (ttEmacPktPtr)(TM_ERAM_LOW + ((i + 1)%TM_BUF_NUM_TX) *
                                                    sizeof(ttEmacPkt)); */
            (tlEmacPtr->infTxBufPtr[i])->epktStat = 0;
        }
        tlEmacPtr->infCurTxIndex = 0;

/* Initialize RX buffer ring */
        for( i = 0; i < TM_BUF_NUM_RX; i++ )
        {
            tlEmacPtr->infRxBufPtr[i] = (ttEmacPktPtr)
                (TM_ERAM_LOW + (i + TM_BUF_NUM_TX) * sizeof(ttEmacPkt));
/* Link all TX EMAC buffer to form a TX ring */
            (tlEmacPtr->infRxBufPtr[i])->epktNextPtr = (ttEmacPktPtr)
                (TM_ERAM_LOW + ((i + 1)%TM_BUF_NUM_RX + TM_BUF_NUM_TX) *
                                                        sizeof(ttEmacPkt));
            (tlEmacPtr->infRxBufPtr[i])->epktStat = 0;
        }
        tlEmacPtr->infCurRxIndex = 0;
        tlEmacPtr->infRecvIndex = 0;

/* Config EMAC config register 1 */
        regValue = 0x10; /* CRC appended */
        tm_out_byte(regValue, TM_EMAC_CFG1);

/* Config EMAC config register 2 */
        regValue = 0x37; /* reset value */
        tm_out_byte(regValue, TM_EMAC_CFG2);

/* Config EMAC config register 3 */
        regValue = 0x0F; /* reset value */
        tm_out_byte(regValue, TM_EMAC_CFG3);

/* Config EMAC config register 4 */
        regValue = 0x01; /* Rx Enable */
        tm_out_byte(regValue, TM_EMAC_CFG4);

/* Set the Ethernet hardware address */
        (void)tfEmacGetPhysicalAddress((ttUserInterface) 0, hardwareAddr);
        tm_out_byte(hardwareAddr[0], TM_EMAC_STAD_0);
        tm_out_byte(hardwareAddr[1], TM_EMAC_STAD_1);
        tm_out_byte(hardwareAddr[2], TM_EMAC_STAD_2);
        tm_out_byte(hardwareAddr[3], TM_EMAC_STAD_3);
        tm_out_byte(hardwareAddr[4], TM_EMAC_STAD_4);
        tm_out_byte(hardwareAddr[5], TM_EMAC_STAD_5);
    
/* Set Transmit Polling Timer Register:3 * (SYSCLK/256) */
        regValue = 0x03;
        tm_out_byte(regValue, TM_EMAC_PTMR);

/* Set Rx Filter: Accept Multi-cast & Broad-cast */
        regValue = 0x03;
        tm_out_byte(regValue, TM_EMAC_AFR);

/* Install ISR handler */
#ifdef TM_KERNEL_RZK_EZ80
/* Old version of RZK, ISR handler will be call by ISR prolog */
        tfKernelInstallIsrHandler(tfEmacIsrProlog, TM_IV_EMAC_RX);
/* New version of RZK, install ISR handler directly */
/*        tfKernelInstallIsrHandler(tfEmacIsrHandler,TM_IV_EMAC_RX); */
#else /*  TM_KERNEL_RZK_EZ80 */
        tfZ80InstallIsr(tfEmacIsrHandler, 0, TM_IV_EMAC_RX);
#endif /* TM_KERNEL_RZK_EZ80 */

/* Enable RX_DONE & RX_OVR interrupts of EMAC */
        regValue = 0x0C;
        tm_out_byte(regValue, TM_EMAC_IEN);

        errorCode = TM_DEV_OKAY;
    }

    return(errorCode);            
}

/*
 * tfEmacClose()
 *
 * Disable transmit and receieve on EMAC and prevent the chip from 
 * generating interrupts in the future.
 *
 */
int tfEmacClose(ttUserInterface interfaceHandle)
{
    unsigned char regValue;

    TM_UNREF_IN_ARG(interfaceHandle);
    tlEmacPtr->infEmacIfHandle = (ttUserInterface)0;

/* Disable interrupts on EMAC */
    regValue = 0x00;
    tm_out_byte(regValue, TM_EMAC_IEN);

/* Disable transmit on EMAC */
    regValue = 0x00;
    tm_out_byte(regValue, TM_EMAC_PTMR);

    return(TM_DEV_OKAY);
}


/* 
 * tfEmacIoctl()
 *
 */
int tfEmacIoctl ( ttUserInterface interfaceHandle,
                  int             flag,
                  void TM_FAR *   optionPtr,
                  int             optionLen )
{
/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

    return(TM_DEV_OKAY);
}


/*
 * tfEmacSend()
 *
 * Requests transmit approval and poll the chip until it is approved.  Copy the
 * data to the EMAC RAM. Don't support scatter send, so TM_USER_BUFFER_LAST is 
 * not checked.
 *
 */
int tfEmacSend ( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag )
{
    int             errorCode;
    int             bufIdx;
    char          * bufPtr;

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(flag);

    errorCode = TM_DEV_OKAY;

    bufIdx = tlEmacPtr->infCurTxIndex;
    bufPtr = (tlEmacPtr->infTxBufPtr[bufIdx])->epktPkt;
/* Wait until the hardware has freed up current TX ring entries */
    if(tm_is_emac_own((tlEmacPtr->infTxBufPtr[bufIdx])->epktStat))
    {
        if (tlEmacPtr->infXmitQueueOn)
        {
/* If user uses a transmit queue, queue the buffer in the transmit queue */
            errorCode = TM_DEV_ERROR;
#ifdef TM_ERROR_CHECKING
            tfKernelWarning("tfEmacSend", "Xmit queue.");
#endif /* TM_ERROR_CHECKING */
        }
        else
        {
            do
            {
/*
 * If do not use transmit task, yield the CPU.
 */
#ifdef TM_ERROR_CHECKING
                tfKernelWarning("tfEmacSend", "spin lock.!!!!!!!!!!!!!!");
#endif /* TM_ERROR_CHECKING */
                (void)tfInterfaceSpinLock(interfaceHandle);
            }
            while(tm_is_emac_own((tlEmacPtr->infTxBufPtr[bufIdx])->epktStat));
        }
    }

    if(errorCode == TM_DEV_OKAY)
    {
/* Call send complete on dirty packets. Only call send complete once per
 * scattered packet. This loop runs while the address isn't set to 0 (to keep
 * us from looping) and the hardware doesn't own the buffer in question
 */
        tfSendCompleteInterface(interfaceHandle,TM_DEV_SEND_COMPLETE_DRIVER);
    
    
/* Update current buffer index */
        tlEmacPtr->infCurTxIndex = (bufIdx + 1) % TM_BUF_NUM_TX;
 
/* Set the necessary fields in the transmit descriptor
 * Caution: the write order is important here.
 */
        (tlEmacPtr->infTxBufPtr[bufIdx])->epktPktSize = (ttUser16Bit)dataLength;

/* Copy data into the buffer */    
        tm_bcopy( dataPtr, bufPtr, dataLength);

/* Give the buffer's ownership to EMAC to start transmission */
        tm_set_emac_own((tlEmacPtr->infTxBufPtr[bufIdx])->epktStat);
    }

    return errorCode;
}


/*
 * tfEmacReceive()
 *
 * Since we're forced to copy in the ISR, not much to do in the receive
 * routine. Just detach the buffer from our ring and pass it up the stack.
 * If desired, we also refill this buffer in this routine.
 */
int tfEmacReceive(ttUserInterface        interfaceHandle,
                  char TM_FAR * TM_FAR * dataPtrPtr,
                  int  TM_FAR *          dataLengthPtr,
                  ttUserBufferPtr        bufHandlePtr)
{
    int             bufIdx;
    int             pktStat;
    int             errorCode;

/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Get the current receive buffer index */
    bufIdx = tlEmacPtr->infRecvIndex;

/* Get the packet status */
    pktStat = (tlEmacPtr->infRxBufPtr[bufIdx])->epktStat;

#ifdef TM_ERROR_CHECKING
/* Make sure that the packet belongs to us, otherwise return an error */
    if (!(pktStat & 0x80))
    {
        tfKernelWarning("tfEmacReceive", "We do not own this buffer");
        errorCode = TM_DEV_ERROR;
    }
    else
#endif /* TM_ERROR_CHECKING */
    {
/* Set the passed-in variables to appropriate values */
        *dataLengthPtr = (tlEmacPtr->infRxBufPtr[bufIdx])->epktPktSize;
        *dataPtrPtr = (tlEmacPtr->infRxBufPtr[bufIdx])->epktPkt;
        *bufHandlePtr = (ttUserBufferPtr)0;
/* Check if there was an error in receive */
        if (pktStat & 0x203F)
        {    
/* The 12th and 13th bytes of an ethernet packet indicate type. If we set
 * them to something invalid (0xffff in this case), the stack will simply
 * discard the packet
 */
            *((*dataPtrPtr)+TM_EMAC_PACKET_TYPE_OFFSET_1) =
                                      (char)TM_EMAC_INVALID_PACKET_TYPE;
            *((*dataPtrPtr)+TM_EMAC_PACKET_TYPE_OFFSET_2) = 
                                       (char)TM_EMAC_INVALID_PACKET_TYPE;
        }
/* Set the address in the descriptor ring to 0, so Ioctl knows to clean it up */
        bufIdx = (bufIdx+1) & TM_BUF_NUM_RX;
        errorCode = TM_DEV_OKAY;
    }

/* Update the current bufIdx */
    tlEmacPtr->infRecvIndex++;

    return(errorCode);
}


/*
 * tfEmacIsrHandler
 */
void TM_INTERRUPT tfEmacIsrHandler(void)
{
    int                 bufIdx;
    int                 recvNotifyCount;
    ttUser16Bit         pktStat;
    unsigned char       regValue;
    ttUser8Bit          rrpLow;
    ttUser8Bit          rrpHigh;

    recvNotifyCount = 0;
/* Handle the RX_DONE and RX_OVR interrupt */
    regValue = tm_in_byte(TM_EMAC_ISTAT);
    if(regValue & 0x0C)
    {
        bufIdx = tlEmacPtr->infCurRxIndex;
#ifdef TM_ERROR_CHECKING
        rrpLow = tm_in_byte(TM_EMAC_RRP_L);
        rrpHigh = tm_in_byte(TM_EMAC_RRP_H);
        if(  ((ttUser8Bit)((unsigned int)tlEmacPtr->infRxBufPtr[bufIdx])
                                                                != rrpLow)
          || ((ttUser8Bit)((unsigned int)tlEmacPtr->infRxBufPtr[bufIdx] >> 8)
                                                                != rrpHigh) )
        {
            tfKernelError("tfEmacIsrHandler", "Corrupted memory");
        }
#endif /* TM_ERROR_CHECKING */

        pktStat = (tlEmacPtr->infRxBufPtr[bufIdx])->epktStat;
        if( (regValue & 0x08) && (pktStat & 0x80) )
        {
/* RX_DONE interrupt received & received packet is OK */
            recvNotifyCount ++;
            tlEmacPtr->infCurRxIndex ++;
            tlEmacPtr->infCurRxIndex %= TM_BUF_NUM_RX;
/* Clear RX_DONE bit */
            regValue &= 0xF7;
            tm_out_byte(regValue, TM_EMAC_ISTAT);
/* update EMAC RRP */
            bufIdx ++;
            rrpLow = (ttUser8Bit)
                ((unsigned int)tlEmacPtr->infRxBufPtr[bufIdx] & 0xFF);
            tm_out_byte(rrpLow, TM_EMAC_RRP_L);
            rrpHigh = (ttUser8Bit)
                (((unsigned int)tlEmacPtr->infRxBufPtr[bufIdx]>>8)&0xFF);
            tm_out_byte(rrpHigh, TM_EMAC_RRP_H);
        }
    }

    if(recvNotifyCount)
    {
        tfNotifyInterfaceIsr( tlEmacPtr->infEmacIfHandle,
                              recvNotifyCount,
                              0, 0UL, 0UL );
    }

    return;
}


/*
 * tfEmacFreeRecvBuf
 * Free receive packet buffer
 */
int tfEmacFreeRecvBuf( ttUserInterface  interfaceHandle,
                       char           * bufPtr )
{
    ttEmacPktPtr    pktPtr;
    int             errorCode;

    TM_UNREF_IN_ARG(interfaceHandle);
    pktPtr = (ttEmacPktPtr)(bufPtr - 7);
#ifdef TM_ERROR_CHECKING
    if(pktPtr->epktStat & 0x80)
    {
        tfKernelWarning("tfEmacFreeRecvBuf", "We do not own this buffer");
        errorCode = TM_DEV_ERROR;
    }
    else
#endif /* TM_ERROR_CHECKING */
    {
/* Mark this buffer as unused buffer */
        pktPtr->epktStat = 0;
        errorCode = TM_DEV_OKAY;
    }

    return errorCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:39      For: Xilinx-Samsung
 * File Name: tremac.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 48409    Output checksum: 46576
 */
