/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * $Source: source/drivers/mpc860/trmpc860.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:11CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/* 
 * SOURCE FILE FOR Motorola MPC860 QUICC Sample Driver
 *
 * This device driver is derived from the Ethernet Bridge example from
 * Motorola.
 * It has been modified to interface to the Treck TCP/IP stack and to
 * provide readability.
 *
 * Since it is a derived work, it remains the property of Motorola Corp.
 *
 * We have left the definintions of the mpc860 registers as they
 * are in the QUICC user manual
 *
 * This is why they do not match TRECK coding standards.
 *
 * Modification History
 * Date: $Date: 2010/01/18 06:36:11CST $
 * Author:
 * Version:  $Revision: 6.0.2.2 $
 * Release:  $ProjectRevision: 6.0.1.23 $
 * Description:
 */


#include <trsocket.h>
#include <trmpc860.h>


/* 
 * IMPORTANT
 * If the mpc860 device only receives a few frames and stops forever,
 * make sure that you are calling the tfMpc860RefillScc1RecvPool()
 * function from a low priority task
 */


/*    
 * Since we do not accept SHORT frames, we use the SHORT bit to 
 * mean that this frame has been notified
 * IF THIS DRIVER IS CHANGED TO ALLOW SHORT FRAMES, ANOTHER BIT
 * WILL NEED TO BE USED
 */
#define TM_MPC860_NOTIFIED TM_MPC860_RECV_E_SHORT
#define TM_MPC860_ERROR_RECV (TM_MPC860_RECV_ERROR & (~TM_MPC860_NOTIFIED))

/*
 * Local Function Prototypes
 */
void tfMpc860CreateXmitPool(int channel, int maxBd);
void tfMpc860CreateRecvPool(int channel, int maxBd);
void tfMpc860IssueCommand(unsigned short command);
void tfMpc860Busy(ttUserInterface interfaceHandle, int channel);
void tfMpc860RecvFrame(ttUserInterface interfaceHandle,
                      int             channel);
void tfMpc860XmitError(ttUserInterface interfaceHandle,int channel);
void tfMpc860Scc1SendComplete(ttUserInterface interfaceHandle);
void tfMpc860EtherUp(int channel);
void tfMpc860EtherScc1BoardInit(void);
void tfMpc860EtherScc1Init(void);
void tfMpc860RefillScc1RecvPool(void);
void tfMpc860Down(int channel);

int tfMpc860SerialClose(ttUserInterface interfaceHandle);
void tfMpc860SerialUp(int channel);
int tfMpc860SerialOpen(ttUserInterface interfaceHandle);
void tfMpc860SerialScc4BoardInit(void);
void tfMpc860SerialScc4Init(void);


static ttMpc860Ptr tlMpc860Ptr;

static ttMpc860BufDescPtr tlMpc860XmitBufDescPtr;
static ttMpc860BufDescPtr tlMpc860XmitIsrBufDescPtr;
static ttMpc860BufDescPtr tlMpc860XmitBufDescStartPtr;

static ttMpc860BufDescPtr tlMpc860RecvBufDescPtr;
static ttMpc860BufDescPtr tlMpc860RecvIsrBufDescPtr;
static ttMpc860BufDescPtr tlMpc860RecvBufDescStartPtr; 
static ttMpc860BufDescPtr tlScatterBufferStart;

static ttMpc860BufDescPtr tlMpc860RecvRefillBufDescPtr; 
unsigned long     tvMpc860RecvErrorCount;
unsigned long     tvMpc860RecvPacketCount;
unsigned long     tvMpc860SendPacketCount;
unsigned long     tvMpc860BusyPacketCount;
unsigned long     tvMpc860XmitWait;
unsigned long     tvMpc860IsrRecvErrorCount;

/*
 * Four Interface handles for four SCC's 
 */
static ttUserInterface tlInterfaceArray[4];
/*
 * Four transmit queue flags for four SCC's
 */
static int             tlMpc860XmitQueueArray[4];

/*
 * Table of ttUserBuffer Handles
 */
static ttUserBuffer    tlUserBufferList[TM_MPC860_ETHER_MAX_RECV_BUFD];

/*
 * Our index to the ttUserBuffer's
 */
static ttUserBufferPtr tlUserBufferPtr; 
static ttUserBufferPtr tlUserRefillBufferPtr; 



/*
 * Return the Physical Address for the device
 * For the MPC860, it should be read from EEPROM
 * because every ethernet address is Unique
 *
 * THIS ROUTINE MUST BE MODIFIED!
 *
 * To get an ethernet address address block for 
 * your company, you need to contact IEEE Standards
 * or visit http://standards.ieee.org/faqs/OUI.html
 * for information on getting an Company ID/OUI
 */
int tfMpc860GetPhyAddr ( ttUserInterface   userInterface,
                         char TM_FAR     * address )
{
/* 
 * We don't use the interface handle here, but we keep the compiler
 * from complaining about unused formal parameters
 */
    TM_UNREF_IN_ARG(userInterface);
    address[0] = 0x00;
    address[1] = (char)(unsigned char)0xea;
    address[2] = (char)(unsigned char)0xc0;
    address[3] = (char)(unsigned char)0xf0;
    address[4] = 0x08;
    address[5] = 0x60;
    return 0;
}

/* 
 * Initialize SCC1 to be an ethernet channel
 */
void tfMpc860EtherScc1Init()
{
    char                          etherAddr[6];
    struct ethernet_pram TM_FAR * enet_pram;
    struct scc_regs      TM_FAR * regs;

/* Retrieve the Ethernet Address */
    tfMpc860GetPhyAddr((ttUserInterface)0, etherAddr);

    enet_pram = &tlMpc860Ptr->pram[0].enet_scc;          
    regs = &tlMpc860Ptr->scc_regs[0];

/* Disable transmit/receive */
    regs->scc_gsmrl = 0;

/* 
 * Initialize paramater ram 
 */

/* points to start of receiving ring */
    enet_pram->rbase =   TM_MPC860_SCC1_RECV_BASE;
/* points to start of transmiting ring */
    enet_pram->tbase =   TM_MPC860_SCC1_XMIT_BASE;   
/* Function Codes for Transmit and Receive */
    enet_pram->rfcr =    TM_MPC860_SCC1_RECV_FUNC_CODE;
    enet_pram->tfcr =    TM_MPC860_SCC1_XMIT_FUNC_CODE;
/* We always receive a full ethernet frame so we set it to 1518 */
    enet_pram->mrblr =   TM_MPC860_SCC1_MAX_RECV_BUF_LEN;
/* A good ethernet CRC */
    enet_pram->c_mask =  TM_MPC860_ETHER_CRC_MASK;
/* The CRC we start with */
    enet_pram->c_pres =  TM_MPC860_ETHER_CRC_PRES;
/* These three registers are cleared for clarity sake */
    enet_pram->crcec =   TM_MPC860_ETHER_CRCEC;
    enet_pram->alec =    TM_MPC860_ETHER_ALEC;
    enet_pram->disfc =   TM_MPC860_ETHER_DISFC;
/* Pad Character */
    enet_pram->pads =    TM_MPC860_ETHER_PADS;
/* Max Retries when we have a collision */
    enet_pram->ret_lim = TM_MPC860_ETHER_RETRY_LIMIT;
/* Maximum Ethernet Frame Length (Always 1518) */
    enet_pram->mflr =    TM_MPC860_ETHER_MAX_FRAME_LEN;
/* Minimum Ethernet Frame Length (Always 64) */
    enet_pram->minflr =  TM_MPC860_ETHER_MIN_FRAME_LEN;
/* Maximum DMA Counts */
    enet_pram->maxd1 =   TM_MPC860_ETHER_MAX_DMA1_COUNT;
    enet_pram->maxd2 =   TM_MPC860_ETHER_MAX_DMA2_COUNT; 
/* Group Addresses */
    enet_pram->gaddr1 =  TM_MPC860_ETHER_GROUP_ADDR1;    
    enet_pram->gaddr2 =  TM_MPC860_ETHER_GROUP_ADDR2;    
    enet_pram->gaddr3 =  TM_MPC860_ETHER_GROUP_ADDR3;    
    enet_pram->gaddr4 =  TM_MPC860_ETHER_GROUP_ADDR4;    
/* 
 * Our physical (Ethernet) address (IN LITTLE ENDIAN)
 * If the physical address is:
 * 0x112233445566
 *
 * It is stored like this:
 * paddr_l=0x2211
 * paddr_m=0x4433
 * paddr_h=0x6655
 * 
 * If not done correctly, the MPC860 will NOT receive any packets
 * with its destination address 
 */
    enet_pram->paddr_l = (unsigned short)
                         (   ((unsigned short)etherAddr[0]        & 0x00ff)
                           | ((((unsigned short)etherAddr[1])<<8) & 0xff00) );
    enet_pram->paddr_m = (unsigned short)
                         (   ((unsigned short)etherAddr[2]        & 0x00ff)
                           | ((((unsigned short)etherAddr[3])<<8) & 0xff00) );
    enet_pram->paddr_h = (unsigned short)
                         (   ((unsigned short)etherAddr[4]        & 0x00ff)
                           | ((((unsigned short)etherAddr[5])<<8) & 0xff00) );

    enet_pram->p_per =   TM_MPC860_ETHER_P_PER;
/* Individual Hash Addresses */
    enet_pram->iaddr1 =  TM_MPC860_ETHER_INDV_ADDR1; 
    enet_pram->iaddr2 =  TM_MPC860_ETHER_INDV_ADDR2; 
    enet_pram->iaddr3 =  TM_MPC860_ETHER_INDV_ADDR3; 
    enet_pram->iaddr4 =  TM_MPC860_ETHER_INDV_ADDR4; 
/* Temp Registers */
    enet_pram->taddr_h = TM_MPC860_ETHER_T_ADDR_H;   
    enet_pram->taddr_m = TM_MPC860_ETHER_T_ADDR_M;
    enet_pram->taddr_l = TM_MPC860_ETHER_T_ADDR_L;
    tvMpc860RecvPacketCount = 0L;
    tvMpc860SendPacketCount = 0L;
    tvMpc860XmitWait = 0L;
    tvMpc860RecvErrorCount = 0L;
    tvMpc860IsrRecvErrorCount = 0L;
    tlScatterBufferStart = (ttMpc860BufDescPtr)0;

/* Tell the CP to initialize our Recv/Xmit parameters */
    tfMpc860IssueCommand(TM_MPC860_INIT_RXTX_SCC1);
}

/* 
 * Initialize SCC1 to be an ethernet channel    
 */
void tfMpc860EtherScc1BoardInit()
{
/*  tlMpc860Ptr->sim_mcr =   0x608f; */

/*
 * IMPLEMENTATION NOTE:
 *
 * This routine is very board specific
 * In this example we use the Atlas Communications Engines
 * DARWIN devlopment board for an example
 * We assume the the ethernet is in SLOT 0 and at SCC1
 * on the DARWIN
 *
 * Please refer to your boards design to modify this routine
 */

/* clear previous scc1 settings */
    tlMpc860Ptr->si_sicr &= ~(0x3f);
/* Atlas DARWIN Setup */
    tlMpc860Ptr->si_sicr |=  TM_MPC860_SI_SCC1_RCS_CLK1 |
                             TM_MPC860_SI_SCC1_TCS_CLK2;
                             
/* Disconnect the clock routing source */
    tlMpc860Ptr->si_sicr &=  ~TM_MPC860_SI_SCC1_CONNECT;
                            
/* Make PortA Pins 6,7,14,15 for onchip use */
    tlMpc860Ptr->pio_padir &= ~(TM_MPC860_PORTA_PA6 |
                                TM_MPC860_PORTA_PA7 |
                                TM_MPC860_PORTA_PA14 |
                                TM_MPC860_PORTA_PA15);
    
/* 
 * Use PortA pins 15,14 for SCC1 Ethernet RX/TX and 
 * pins 6,7 for Ethernet TCLK and RCLK 
 */
    tlMpc860Ptr->pio_papar |= TM_MPC860_PORTA_PA6 |
                              TM_MPC860_PORTA_PA7 |
                              TM_MPC860_PORTA_PA14 |
                              TM_MPC860_PORTA_PA15;
/* Turn off the Open Drain */
    tlMpc860Ptr->pio_paodr &= ~TM_MPC860_PORTA_PA14;
                            
/* Clear the pending and in-service registers */
    tlMpc860Ptr->intr_cipr = TM_MPC860_SCC1_INT_MASK;
    tlMpc860Ptr->intr_cisr = TM_MPC860_SCC1_INT_MASK;
/* Set the CP mask register to accept interrupts from SCC1 */
    tlMpc860Ptr->intr_cimr |= TM_MPC860_SCC1_INT_MASK;

    tlMpc860Ptr->sdma_sdcr = TM_MPC860_SDMA_ARB_ID;

    
 /* PortC Pins 10 & 11 are onchip */
    tlMpc860Ptr->pio_pcdir &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);
 /* PortC Pins 10 & 11 are onchip */
    tlMpc860Ptr->pio_pcpar &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);
                                 
/*
 * Use Pins 10 and 11 (SCC1 CD and CTS) for Ethernet RENA and CLSN  
 */
    tlMpc860Ptr->pio_pcso |=   TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11;
                              
/* Clear the data register for pins 10 and 11 */
    tlMpc860Ptr->pio_pcdat &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);

/*
 * Port B External Signals 
 * PB14,PB15 are clock select
 * PB26 is Link Enable
 * PB27 Is Full Duplex Select
 */
    tlMpc860Ptr->pip_pbdir |= TM_MPC860_PORTB_PB14 |
                              TM_MPC860_PORTB_PB15 |
                              TM_MPC860_PORTB_PB27 |
                              TM_MPC860_PORTB_PB26;
                                
    tlMpc860Ptr->pip_pbpar &= ~(TM_MPC860_PORTB_PB14 |
                                TM_MPC860_PORTB_PB15 |
                                TM_MPC860_PORTB_PB27 |
                                TM_MPC860_PORTB_PB26);
/* Set clock select to 00 */
    tlMpc860Ptr->pip_pbdat &= ~(TM_MPC860_PORTB_PB14 |
                                TM_MPC860_PORTB_PB15);
/* Set Link Enable */
    tlMpc860Ptr->pip_pbdat |=  TM_MPC860_PORTB_PB26;

/* Turn OFF Full Duplex Mode */
    tlMpc860Ptr->pip_pbdat &=  ~TM_MPC860_PORTB_PB27;
    


/*
 * IMPLEMENTATION NOTE:
 *
 * The Actual SCC1 ISR Wrapper Function in installed here
 * It must call the tfMpc860Scc1HandlerIsr function
 * This routine may be in "C" if it is supported
 * or assembly if not.  It is VERY RTOS dependent
 * so it cannot be provided with the driver
 *
 * With some RTOS's a wrapper function may not be
 * needed.  All the driver cares about is that the
 * function tfMpc860Scc1HandlerIsr function is called
 * when the interrupt occurs.
 *
 */

/* Port B pin 19 is Ethernet TENA for DARWIN */
    tlMpc860Ptr->pip_pbpar |= TM_MPC860_PORTB_PB19;
    tlMpc860Ptr->pip_pbdir |= TM_MPC860_PORTB_PB19;

/* Install the wrapper function */
    tfKernelInstallIsrHandler(tfMpc860Scc1Isr, TM_MPC860_SCCA_INT);
}

/* 
 * Enable the SCC's  to recieve and transmit 
 */
void tfMpc860EtherUp(int channel)
{
    struct scc_regs TM_FAR * regs;

    regs = &tlMpc860Ptr->scc_regs[channel];

/* clear all events on the SCC event register and CP event reg*/
    regs->scc_scce = 0xffff;

/* Clear the appropriate bit in the CISR */
    switch (channel)
    {
        case TM_MPC860_SCC1_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC1_INT_MASK;
            break;
        case TM_MPC860_SCC2_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC2_INT_MASK;
            break;
        case TM_MPC860_SCC3_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC3_INT_MASK;
            break;
        case TM_MPC860_SCC4_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC4_INT_MASK;
            break;
        default:
            break;
    }


/* 
 * mask the events we sich to interrupt on
 */
    regs->scc_sccm = 
/* An error in transmitting has occured */
                     TM_MPC860_INTR_ETHER_E_XMIT |
/* A buffer has completed transmitting */
                     TM_MPC860_INTR_ETHER_XMIT_B |
/* A ethernet frame has been received */
                     TM_MPC860_INTR_ETHER_RECV_F |
/* The busy condition has occured */
                     TM_MPC860_INTR_ETHER_BUSY;
    
/* SCC1 data sync reg  (The pattern for Ethernet)*/
    regs->scc_dsr =  TM_MPC860_SCC_FRAME_SYNC_ETHER;

/* protocol specific mode register */
    regs->scc_psmr = 
/* To disable the reception of Broadcast packets uncomment this line */
/*                     TM_MPC860_ETHER_BROADCAST | */
/* Turn on the CRC for ethernet */
                     TM_MPC860_ETHER_ENABLE_CRC | 
/* Ignore 22 bits for ethernet */
                     TM_MPC860_ETHER_NBITS_IGNORED;


/* SCC general mode reg */
    regs->scc_gsmrh = 0;

/* SCC general mode reg */
    regs->scc_gsmrl = 
/* Transmit Clock Inversion */
                      TM_MPC860_XMIT_CLOCK_INV | 
/* Preamble Bit Length (48 always for ethernet) */
                      TM_MPC860_PREAMBLE_48 | 
/* Preamble Bit Pattern (Alternating '1' and '0' for ethernet */
                      TM_MPC860_PREAMBLE_PAT_10 | 
/* Ethernet Mode */
                      TM_MPC860_ETHERNET_PORT;

/* enable SCC1 receive/transmit operation */
    regs->scc_gsmrl |= TM_MPC860_ENABLE_RECV | TM_MPC860_ENABLE_XMIT;
        
}

void tfMpc860Down(int channel)
{
    struct scc_regs *regs;


    regs = &tlMpc860Ptr->scc_regs[channel];

/* clear all events on the SCC event register and CP event reg*/
    regs->scc_scce = 0xffff;


/* Disable SCCx receive/transmit operation */
    regs->scc_gsmrl &= ~(TM_MPC860_ENABLE_RECV | TM_MPC860_ENABLE_XMIT);
        
}

/*
 * Disable the Ethernet on SCC1
 */
int tfMpc860EtherClose (ttUserInterface interfaceHandle)
{

    ttMpc860BufDescPtr      bufDescPtr;
    ttUserBufferPtr         userBufPtr;
    int                     i;


/* Shut down the SCC1 Ethernet hardware */
    tfMpc860Down(TM_MPC860_SCC1_CHANNEL);

/* 
 * Free leftover memory:
 *   First, send complete on any remaining packets in the transmit ring.
 */
    bufDescPtr = tlMpc860XmitBufDescStartPtr;
    for (i = 0; i < TM_MPC860_ETHER_MAX_XMIT_BUFD; i++)
    {
/* Call send complete to free this entry */
        if ( ((bufDescPtr->bdStatus) & TM_MPC860_XMIT_LAST ) &&
             (bufDescPtr->bdDataPtr != (char TM_FAR *) 0) )
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER );
        }
        bufDescPtr++;
    }


/* Next, free any buffers in the receive ring */
    bufDescPtr = tlMpc860RecvBufDescStartPtr;
    userBufPtr = &(tlUserBufferList[0]);
    for (i = 0; i < TM_MPC860_ETHER_MAX_RECV_BUFD; i++)
    {
        if ((bufDescPtr->bdDataPtr) != (char TM_FAR *) 0)
        {
            tfFreeDriverBuffer(*userBufPtr);
        }
        bufDescPtr++;
        userBufPtr++;
    }
        

    return TM_DEV_OKAY;
}

/* 
 * The main initialization routine for the mpc860 and the scc's 
 */
int tfMpc860EtherOpen(ttUserInterface interfaceHandle)
{
    unsigned long  mpc860Registers;
    int            errorCode;

    mpc860Registers = TM_MPC860_IMMR_LOC;
    tlMpc860Ptr = (ttMpc860 *)(mpc860Registers&0xfffffffcUL);

    errorCode = TM_DEV_OKAY;
/* 
 * Check that the number of transmit descriptors exceed to the number
 * of buffers per frame that the user specified in tfOpenInterface.
 */
    if (tfInterfaceScatterCount(interfaceHandle) >
            TM_MPC860_ETHER_MAX_XMIT_BUFD)
    {
        tfKernelWarning("tfMpc860EtherOpen",
                        "Scatter buffers exceed descriptors");
        errorCode = TM_DEV_ERROR;
    }

    if (errorCode == TM_DEV_OKAY)
    {
/* Initialize default Ethernet parameters */
        tfMpc860EtherScc1Init();
/* Create our transmit pool */
        tfMpc860CreateXmitPool( TM_MPC860_SCC1_CHANNEL,
                                TM_MPC860_ETHER_MAX_XMIT_BUFD );
/* Create our receive pool (with receive buffers) */
        tfMpc860CreateRecvPool( TM_MPC860_SCC1_CHANNEL,
                                TM_MPC860_ETHER_MAX_RECV_BUFD );
/* Initial board specific parameters */
        tfMpc860EtherScc1BoardInit();
/* Bring the ethernet scc port online */
        tfMpc860EtherUp(TM_MPC860_SCC1_CHANNEL);

/* Save our interfaceHandle for ISR use */
        tlInterfaceArray[TM_MPC860_SCC1_CHANNEL] = interfaceHandle;
/* Save whether the user enabled the Treck transmit queue */
        tlMpc860XmitQueueArray[TM_MPC860_SCC1_CHANNEL] =
                                    tfInterfaceXmitQueueOn(interfaceHandle);
    }

    return errorCode;
}


/* 
 * Interrupt service routine called from actual
 * ISR handlers
 */ 
void tfMpc860Scc1HandlerIsr()
{
    ttUserInterface          interfaceHandle;
    struct scc_regs TM_FAR * regs;
    int                      channel;
    unsigned short TM_FAR  * bdStatusPtr;
#ifdef TM_TASK_SEND
    void   TM_FAR * TM_FAR * bdDataPtrPtr;
#endif /* TM_TASK_SEND */
    register unsigned short  event;
    int                      sentNotifyCount;
    unsigned long            sentNotifyBytes;
    int                      recvNotifyCount;
    
    channel = TM_MPC860_SCC1_CHANNEL;
    interfaceHandle = tlInterfaceArray[channel];
/* Points to SCC regs */
    regs = &tlMpc860Ptr->scc_regs[channel]; 

/* Check to see what events have occured */
    event = regs->scc_scce;
/* Cleared event bits we hill handle */
    regs->scc_scce = event;
    recvNotifyCount = 0;
    sentNotifyCount = 0;
    sentNotifyBytes = 0UL;


#ifdef TM_TASK_SEND
    if( event & TM_MPC860_INTR_ETHER_XMIT_B ) 
    {
        bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
        while (    ( (*bdStatusPtr & TM_MPC860_XMIT_READY) == 0 )
                && ( *bdDataPtrPtr != (void TM_FAR *)0 ) )
        {
/* Zero it out so we know that we have freed this one before */
            *bdDataPtrPtr = (void TM_FAR *)0;
            if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
            {
                tlMpc860XmitIsrBufDescPtr = tlMpc860XmitBufDescStartPtr;
            }
            else
            {
                tlMpc860XmitIsrBufDescPtr++;
            }
            bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
            bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
            sentNotifyCount++;
            sentNotifyBytes = (unsigned long)
                  (   sentNotifyBytes
                    + (unsigned long)tlMpc860XmitIsrBufDescPtr->bdLength );
        }
    }
#endif /* TM_TASK_SEND */
    if ( event & TM_MPC860_INTR_ETHER_RECV_F )
    {
/*
 * loop on frames with the Empty bit and SHORT (notify) bit is NOT set
 * "==0" .
 * Only notify while the frames are not empty
 */
        bdStatusPtr = &tlMpc860RecvIsrBufDescPtr->bdStatus;
        while (!(*bdStatusPtr & (TM_MPC860_RECV_EMPTY|TM_MPC860_NOTIFIED)))
        {
            recvNotifyCount++;
            *bdStatusPtr |= TM_MPC860_NOTIFIED;
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860RecvIsrBufDescPtr = tlMpc860RecvBufDescStartPtr;
            }
            else
            {
                tlMpc860RecvIsrBufDescPtr++;
            }
            bdStatusPtr = &tlMpc860RecvIsrBufDescPtr->bdStatus;
        }
    }
    if ( event &  TM_MPC860_INTR_ETHER_E_XMIT )
    {
/* An error on Transmit has occured */
        tfMpc860XmitError(interfaceHandle,channel);
    }
    if ( event & TM_MPC860_INTR_ETHER_BUSY )
    {
/* Busy Condition has been set */
        tfMpc860Busy(interfaceHandle,channel);
    }
    if ( (recvNotifyCount != 0) || (sentNotifyCount != 0) )
    {
        tfNotifyInterfaceIsr( interfaceHandle, recvNotifyCount,
                              sentNotifyCount, sentNotifyBytes, 0UL );
    }
/* Clear the appropriate bit in the CISR */
    tlMpc860Ptr->intr_cisr |= TM_MPC860_SCC1_INT_MASK;
}

/*
 * Send a packet to the network
 */
int tfMpc860EtherSend ( ttUserInterface interfaceHandle,
                        char   TM_FAR * bufferPtr,
                        int             length,
                        int             flag )
{
    ttMpc860BufDescPtr        bufDescPtr;
    ttMpc860BufDescPtr        startBufDescPtr;
    unsigned short  TM_FAR  * bdStatusPtr;
    void    TM_FAR * TM_FAR * bdDataPtrPtr;
    int                       errorCode;

    errorCode = TM_DEV_OKAY;
/* Wait for a free buffer descriptor */    
    while (tlMpc860XmitBufDescPtr->bdStatus & TM_MPC860_XMIT_READY)
    {
        if (tlMpc860XmitQueueArray[TM_MPC860_SCC1_CHANNEL])
        {
/* If user uses a transmit queue, queue the buffer in the transmit queue */
            errorCode = TM_DEV_ERROR;
            goto mpc860SendExit;
        }
        else
        {
/* Otherwise */
            do
            {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
                tfInterfaceSpinLock(interfaceHandle);
                tvMpc860XmitWait++;
            }
            while (tlMpc860XmitBufDescPtr->bdStatus & TM_MPC860_XMIT_READY);
        }
    }
#ifndef TM_TASK_SEND
/* Inlined send complete */
    bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
    bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
    while (    (    ((*bdStatusPtr & TM_MPC860_XMIT_READY) == 0)
                 && (tlMpc860XmitIsrBufDescPtr != tlScatterBufferStart) )
            && (*bdDataPtrPtr != (void TM_FAR *)0) )
/* 
 * Watch out for a scattered frame we have not yet finished
 * loading up
 */
    {
/* Zero it out so we know that we have freed this one before */
        *bdDataPtrPtr = (void TM_FAR *)0;
        if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
        {
            tlMpc860XmitIsrBufDescPtr = tlMpc860XmitBufDescStartPtr;
        }
        else
        {
            tlMpc860XmitIsrBufDescPtr++;
        }
/* Only call send complete when the last piece has been transmitted */
        if (*bdStatusPtr & TM_MPC860_XMIT_LAST)
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER );
        }
        bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
    }
#endif /* NOT TM_TASK_SEND */
    bufDescPtr = tlMpc860XmitBufDescPtr++;
    bdStatusPtr = &bufDescPtr->bdStatus;
    if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
    {
        tlMpc860XmitBufDescPtr = tlMpc860XmitBufDescStartPtr;
    }
    bufDescPtr->bdDataPtr = (void TM_FAR *)bufferPtr;
    bufDescPtr->bdLength = (unsigned short)length;
/* Clear the previous status (but preserve the WRAP bit) */
    bufDescPtr->bdStatus &= TM_MPC860_XMIT_WRAP;
/* If this is the last buffer in the frame then xmit it */
    if (flag == TM_USER_BUFFER_LAST)
    {
/* Now mark the last bd ready */
        bdStatusPtr = &bufDescPtr->bdStatus;
        *bdStatusPtr |= ( TM_MPC860_XMIT_READY |
                          TM_MPC860_XMIT_PAD |

#ifdef TM_TASK_SEND
                          TM_MPC860_XMIT_INTR |
#endif /* TM_TASK_SEND */

                          TM_MPC860_XMIT_LAST |
                          TM_MPC860_XMIT_CRC);
/* Check to see if we are xmitting scattered data */
        if (tlScatterBufferStart != (ttMpc860BufDescPtr)0)
        {
/* We have at least one more piece to xmit if we are here */
            startBufDescPtr = tlScatterBufferStart;
            tlScatterBufferStart = (ttMpc860BufDescPtr)0;
/* 
 * Go ahead and xmit the first packets (with more set) 
 * by walking the list backwards
 * We do this so that we don't need a critical section
 * in order to keep from getting gaps inbetween the 
 * pieces
 * All we need to do here is update the
 * ready flag
 */
            do
            {
                if (bufDescPtr == tlMpc860XmitBufDescStartPtr)
                {
/* We are at the front of the list so go to the end */
                    bufDescPtr += (TM_MPC860_ETHER_MAX_XMIT_BUFD - 1);
                }
                else
                {
/* Go back one buffer descriptor */
                    bufDescPtr--;
                }
                bdStatusPtr = &bufDescPtr->bdStatus;
                *bdStatusPtr |= TM_MPC860_XMIT_READY;
            }
            while (bufDescPtr != startBufDescPtr);
        }        
    }
    else
    {
/* More buffers to come */
        if (tlScatterBufferStart == (ttMpc860BufDescPtr)0)
        {
            tlScatterBufferStart = bufDescPtr;
        }
    }
    tvMpc860SendPacketCount++;

mpc860SendExit:
    return errorCode;
}


/*
 * A Transmit Error has occured.  Currently, we don't examine what type
 * of error it was - if you cared, you could spin through the ring and
 * look at the error conditions.
 *
 * The most important thing to be done here is to restart the transmitter.
 * However, when the CPM does this, it doesn't move its 'current' pointer
 * forward to the next buffer, as we have (in the previous tfMpc860EtherSend).
 * Since the driver and the hardware are out of sync, we'll eventually
 * run into a deadlock situation.  The best way to handle this is simply
 * nudge the CPM's current pointer forward one entry.  This is definitely
 * must simpler than removing the error packet from the ring and resetting
 * our counter.  
 * 
 */ 
void tfMpc860XmitError(ttUserInterface interfaceHandle,int channel)
{
    struct ethernet_pram * enet_pram;
    ttMpc860BufDescPtr     bufDescPtr;
    int                    command;


    TM_UNREF_IN_ARG(interfaceHandle);

    enet_pram = &tlMpc860Ptr->pram[channel].enet_scc;

/* issue the RESTART TRANSMIT command to the CP */
    command  = TM_MPC860_RESTART_TX | ( channel << 6);
    tfMpc860IssueCommand((unsigned short)command);

    bufDescPtr = TTBD_ADDR(tlMpc860Ptr,enet_pram);
        
    if (bufDescPtr->bdStatus & TM_MPC860_XMIT_WRAP)
    {
        enet_pram->tbptr -= sizeof (ttMpc860BufDesc) * 
            (TM_MPC860_ETHER_MAX_XMIT_BUFD - 1);
    }
    else
    {
        enet_pram->tbptr += sizeof (ttMpc860BufDesc);
    }

    bufDescPtr = TTBD_ADDR(tlMpc860Ptr,enet_pram);
        
    if (bufDescPtr->bdStatus & TM_MPC860_XMIT_WRAP)
    {
        enet_pram->tbptr -= sizeof (ttMpc860BufDesc) * 
            (TM_MPC860_ETHER_MAX_XMIT_BUFD - 1);
    }
    else
    {
        enet_pram->tbptr += sizeof (ttMpc860BufDesc);
    }

}

/*
 * tfMpc860EtherReceive
 * Receive a buffer from the chip and pass back to the stack
 */
int tfMpc860EtherReceive(ttUserInterface  interfaceHandle,
                         char TM_FAR    **dataPtrPtr,
                         int  TM_FAR     *lengthPtr,
                         ttUserBufferPtr  userBufferHandlePtr)
{
    ttMpc860BufDescPtr      bufDescPtr;
    unsigned short TM_FAR * bdStatusPtr;
    int                     errorCode;
    
    TM_UNREF_IN_ARG(interfaceHandle);
    errorCode = TM_DEV_OKAY;
    
    bdStatusPtr = &tlMpc860RecvBufDescPtr->bdStatus;

/* Skip pass error packets who are not empty */
    if (    ( !(*bdStatusPtr &  TM_MPC860_RECV_EMPTY) )
         && ( *bdStatusPtr & TM_MPC860_ERROR_RECV ) )
    {
        tvMpc860RecvErrorCount++;
/* Return the userBuffer Handle for this data pointer */
        if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
        {
            tlMpc860RecvBufDescPtr = tlMpc860RecvBufDescStartPtr;
            tlUserBufferPtr = tlUserBufferList;
        }
        else
        {
            tlMpc860RecvBufDescPtr++;
            tlUserBufferPtr++;
        }
        errorCode = TM_EWOULDBLOCK;
    }
    else
    {
/* Check to see if we own the Buffer Descriptor */
        if (*bdStatusPtr & TM_MPC860_RECV_EMPTY)
        {
        
/* Will only happen when refill cleans up error before we are called here */
            tvMpc860RecvErrorCount++;
/* Return the userBuffer Handle for this data pointer */
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860RecvBufDescPtr = tlMpc860RecvBufDescStartPtr;
                tlUserBufferPtr = tlUserBufferList;
            }
            else
            {
                tlMpc860RecvBufDescPtr++;
                tlUserBufferPtr++;
            }
            errorCode = TM_EWOULDBLOCK;
        }
        else
        {
            bufDescPtr = tlMpc860RecvBufDescPtr++;
/* Return the userBuffer Handle for this data pointer */
            *userBufferHandlePtr = *tlUserBufferPtr++;
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860RecvBufDescPtr = tlMpc860RecvBufDescStartPtr;
                tlUserBufferPtr = tlUserBufferList;
            }
/* We are returning the pointer to the data */
            *dataPtrPtr = (char TM_FAR *)(bufDescPtr->bdDataPtr);

/* Zero out the bufferPtr so we can refill it later */
            bufDescPtr->bdDataPtr = (void TM_FAR *)0;

/* We are returning the length */
            *lengthPtr  = bufDescPtr->bdLength;
/* Up our packet count */
            tvMpc860RecvPacketCount++;
        }
    }
    return(errorCode);
}

/* 
 * Interrupt called on a busy condition on the line. 
 * Lost some frames here.
 */
void tfMpc860Busy(ttUserInterface interfaceHandle, int channel)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(channel);
    tvMpc860BusyPacketCount++;
}

/* 
 * tfMpc860IssueCmd() is used to send commands to the communication processor
 * and waits till the command is completed
 */
void tfMpc860IssueCommand(unsigned short command)
{
    volatile unsigned short TM_FAR * commandReg;

/* Save the command register for the CP */  
    commandReg = &tlMpc860Ptr->cp_cr;
/* Issue the command */
    *commandReg = (unsigned short)(command | TM_MPC860_COMMAND_FLAG);
/* wait for the CP zero the CMD_FLG */
    while( (*(volatile short *)commandReg) & TM_MPC860_COMMAND_FLAG )
    {
/* 
 * If this loop never exits:
 *
 * 1) check to make sure that the TM_MPC860_IMMR_LOC is correct 
 *
 * 2) that the tlMpc860 value is correct and has been initialized
 *
 * 3) Make sure the compiler has generated good code
 *    Some compilers have been known to just AND a CPU register with
 *    TM_MPC860_COMMAND_FLAG instead of ANDING the contents of the 
 *    memory location for commandReg with TM_MPC860_COMMAND_FLAG.
 *    WORK AROUND: move the check into a seperate function
 *    and call that function from inside this loop.
 */
       ;
    } 
}


/* 
 * This routine is preparing for each SCC a pool of READY buffer 
 * descriptors that are ready to recieve. The last buffer descriptor
 * has the wrap bit set.
 */
void tfMpc860CreateRecvPool(int channel, int maxBd)
{
    ttMpc860BufDescPtr            tempBufDescPtr;
    struct ethernet_pram TM_FAR * enetPramAddr;
    ttUserBuffer                  userBuffer;
    void                 TM_FAR * dataPtr;
    int i;

    enetPramAddr = &tlMpc860Ptr->pram[channel].enet_scc;

    tempBufDescPtr = TM_MPC860_RECV_BD_ADDR(tlMpc860Ptr,enetPramAddr);
/* Save it off for our use in the recv routine */    
    tlMpc860RecvRefillBufDescPtr = tempBufDescPtr;
    tlMpc860RecvBufDescStartPtr = tempBufDescPtr;
    tlMpc860RecvIsrBufDescPtr = tempBufDescPtr;
    tlMpc860RecvBufDescPtr = tempBufDescPtr;
    tlUserRefillBufferPtr = tlUserBufferList;
    tlUserBufferPtr = tlUserBufferList;

    for(i = 0; i < maxBd; i++, tempBufDescPtr++ )
    {
/* bdDataPtr need to be aligned on a 4-byte boundary */
        dataPtr = (void TM_FAR *)
                    (tfGetEthernetBuffer(&userBuffer) - TM_ETHER_IP_ALIGN);
        if (userBuffer == (ttUserBuffer)0)
        {
/* No memory so stop the create here */
            break;
        }
        tempBufDescPtr->bdDataPtr = dataPtr;
        tlUserBufferList[i] = userBuffer;
        tempBufDescPtr->bdLength = 0;
        tempBufDescPtr->bdStatus = (TM_MPC860_RECV_EMPTY|TM_MPC860_RECV_INTR);
    }
    --tempBufDescPtr;
 /* signify last bit on */
    tempBufDescPtr->bdStatus |= TM_MPC860_RECV_WRAP;
}

/*
 * SET GROUP ADDRESS COMMAND for SCC1:
 * (pages 7-7, 7-8, 7-253, 7-254, and 7-258 of the MC68360 manual)
 * Reset bit (bit 15) is 0. Bits 12-14 are reserved, and zero.
 * Opcode (bits 8 - 11) is 8.
 * Channel number bits 4-7 is 0.
 * Flag (bit 0) is set in tfQuiccIssueCommand
 */
#define TM_MPC860_SET_GROUP_ADDRESS          0x0800

/*
 * General function
 * Can be used to configure the device
 * or refill the receive pool
 */
int tfMpc860Ioctl (ttUserInterface interfaceHandle, int flag,
                   void TM_FAR * optionPtr, int optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))

    struct ethernet_pram TM_FAR * enetPramAddrPtr;
    struct scc_regs      TM_FAR * regsPtr;
    unsigned char        TM_FAR * etherPtr;
    unsigned char        TM_FAR * msbHashPtr;
    int                           index;

    TM_UNREF_IN_ARG(interfaceHandle);
    if ( (flag == TM_DEV_SET_MCAST_LIST) || (flag == TM_DEV_SET_ALL_MCAST) )
    {
/* SCC1 Registers */
        regsPtr = &tlMpc860Ptr->scc_regs[TM_MPC860_SCC1_CHANNEL];

/*
 * Disable SCC1 receive/transmit operation, before clearing the Group
 * address hash table
 */
        regsPtr->scc_gsmrl &= ~(TM_MPC860_ENABLE_RECV | TM_MPC860_ENABLE_XMIT);

        enetPramAddrPtr = &tlMpc860Ptr->pram[TM_MPC860_SCC1_CHANNEL].enet_scc;
/*
 * Clear hash table. If optionLen iz zero, this will clear reception
 * of all multicast addresses (as intended).
 */
        enetPramAddrPtr->gaddr1 = 0;
        enetPramAddrPtr->gaddr2 = 0;
        enetPramAddrPtr->gaddr3 = 0;
        enetPramAddrPtr->gaddr4 = 0;

        if (flag == TM_DEV_SET_MCAST_LIST)
        {
/* Most significant byte of entry in the hash table is in low memory. */
            msbHashPtr = (unsigned char TM_FAR *)
                                           (&(enetPramAddrPtr->taddr_h));
            etherPtr = (unsigned char TM_FAR *)optionPtr;
/* Add multicast address list */
            while (optionLen > 0)
            {
                index = 0;
                while (index < 6)
                {
/* Copy Ethernet address MSB (high memory) to Hash table MSB (low memory) */
                    msbHashPtr[index] = etherPtr[6 - 1 - index];
                    index++;
                }
                tfMpc860IssueCommand(TM_MPC860_SET_GROUP_ADDRESS);
/* Next Ethernet address in the list */
                optionLen--;
                etherPtr += 6;
            }
        }
        else
        {
            if (flag == TM_DEV_SET_ALL_MCAST)
            {
/* Enable reception of all multicast addresses */
                enetPramAddrPtr->gaddr1 = 0xFFFF;
                enetPramAddrPtr->gaddr2 = 0xFFFF;
                enetPramAddrPtr->gaddr3 = 0xFFFF;
                enetPramAddrPtr->gaddr4 = 0xFFFF;
            }
        }
/* Enable SCC1 receive/transmit operation */
        regsPtr->scc_gsmrl |= (TM_MPC860_ENABLE_RECV | TM_MPC860_ENABLE_XMIT);
    }

#else /* !(TM_IGMP or TM_6_USE_MLD) */

    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(interfaceHandle);

#endif /* !(TM_IGMP or TM_6_USE_MLD) */
    if (flag & TM_MPC860_REFILL_SCC1)
    {
        tfMpc860RefillScc1RecvPool();
    }
#ifndef TM_TASK_SEND
    if (flag & TM_MPC860_SEND_COMPLETE)
    {
        tfMpc860Scc1SendComplete(interfaceHandle);
    }
#endif /* NOT TM_TASK_SEND */
    return TM_DEV_OKAY;
}

/*
 * Walk through receive list and allocate new buffers
 * after they have been processed
 */
void tfMpc860RefillScc1RecvPool()
{
    ttUserBuffer            userBuffer;
    unsigned short TM_FAR * bdStatusPtr;
    void           TM_FAR * dataPtr;
    void           TM_FAR * TM_FAR * bdDataPtrPtr;

    bdDataPtrPtr = &tlMpc860RecvRefillBufDescPtr->bdDataPtr;
    bdStatusPtr  = &tlMpc860RecvRefillBufDescPtr->bdStatus;
/* Loop while it is empty or an error */
    while (    (*bdDataPtrPtr == (void TM_FAR *)0)
            || (*bdStatusPtr & TM_MPC860_ERROR_RECV) )
    {
/* Check to see if it is an error */
        if (!(*bdStatusPtr & TM_MPC860_ERROR_RECV))
        {
/* If it is not get a new buffer, otherwise just reuse it */
/* bdDataPtr need to be aligned on a 4-byte boundary */
            dataPtr = (void TM_FAR *)
                    (tfGetEthernetBuffer(&userBuffer) - TM_ETHER_IP_ALIGN);
            if (userBuffer == (ttUserBuffer)0)
            {
/* No memory so stop the refill */
                break;
            }
            tlMpc860RecvRefillBufDescPtr->bdDataPtr = dataPtr;
            *tlUserRefillBufferPtr = userBuffer;
        }
        tlMpc860RecvRefillBufDescPtr->bdLength = 0;
/* Clear the status Except for the wrap bit */
        *bdStatusPtr &= TM_MPC860_RECV_WRAP;
/* Mark this BD as ready to use */
        *bdStatusPtr |= (TM_MPC860_RECV_EMPTY|TM_MPC860_RECV_INTR);
        if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
        {
            tlMpc860RecvRefillBufDescPtr = tlMpc860RecvBufDescStartPtr;
            tlUserRefillBufferPtr = tlUserBufferList;
        }
        else
        {
            tlMpc860RecvRefillBufDescPtr++;
            tlUserRefillBufferPtr++;
        }
        bdDataPtrPtr = &tlMpc860RecvRefillBufDescPtr->bdDataPtr;
        bdStatusPtr  = &tlMpc860RecvRefillBufDescPtr->bdStatus;
    }
}

#ifndef TM_TASK_SEND
/*
 * Walk through send list and post send completes
 */
void tfMpc860Scc1SendComplete(ttUserInterface interfaceHandle)
{
    unsigned short TM_FAR *  bdStatusPtr;
    void            TM_FAR * TM_FAR * bdDataPtrPtr;

    bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
    bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
    while (    ((*bdStatusPtr & TM_MPC860_XMIT_READY) == 0)
            && (tlMpc860XmitIsrBufDescPtr != tlScatterBufferStart)
            && (*bdDataPtrPtr != (void TM_FAR *)0) )
    {
/* Zero it out so we know that we have freed this one before */
        *bdDataPtrPtr = (void TM_FAR *)0;
        if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
        {
            tlMpc860XmitIsrBufDescPtr = tlMpc860XmitBufDescStartPtr;
        }
        else
        {
            tlMpc860XmitIsrBufDescPtr++;
        }
/* Only call send complete when the last piece has been transmitted */
        if (*bdStatusPtr & TM_MPC860_XMIT_LAST)
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER );
        }
        bdStatusPtr = &tlMpc860XmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860XmitIsrBufDescPtr->bdDataPtr;
    }
}
#endif /* NOT TM_TASK_SEND */


/* 
 * Prepare a transmit pool, set the status bits in the buffer descriptors
 * to be not READY to transmit and set the last buffer descriptor to
 * be wrapped
 */
void tfMpc860CreateXmitPool(int channel, int maxBd)
{
    int                           i;
    ttMpc860BufDescPtr            tempBufDescPtr;
    struct ethernet_pram TM_FAR * enetPramAddr;

/* Save the ethernet parameter address */
    enetPramAddr = &tlMpc860Ptr->pram[channel].enet_scc;

    tempBufDescPtr = TM_MPC860_XMIT_BD_ADDR(tlMpc860Ptr, enetPramAddr);

/* Save it off for our use in the xmit routine */    
    tlMpc860XmitIsrBufDescPtr = tempBufDescPtr;
    tlMpc860XmitBufDescStartPtr = tempBufDescPtr;
    tlMpc860XmitBufDescPtr = tempBufDescPtr;
    for( i = 0; i < maxBd; i++, tempBufDescPtr++ )
    {
        tempBufDescPtr->bdStatus = 0;
        tempBufDescPtr->bdLength = 0;
        tempBufDescPtr->bdDataPtr = (void TM_FAR *)0;
    }
    --tempBufDescPtr;
/* signify last bit on */
    tempBufDescPtr->bdStatus |= TM_MPC860_XMIT_WRAP; 
}

/*
 *
 *
 * SERIAL INTERFACE
 *
 *
 */

/* 
 * Initialize SCC4 to be an UART channel
 */
void tfMpc860SerialScc4Init()
{
    struct uart_pram TM_FAR * rs232_pram;
    struct scc_regs TM_FAR  * regs;


    rs232_pram = &tlMpc860Ptr->pram[3].scc.pscc.u;
    regs = &tlMpc860Ptr->scc_regs[3];

/* Disable transmit/receive */
    regs->scc_gsmrl = 0;



/* points to start of receiving ring */
    rs232_pram->rbase =   TM_MPC860_SCC4_RECV_BASE;
/* points to start of transmiting ring */
    rs232_pram->tbase =   TM_MPC860_SCC4_XMIT_BASE;   

/* Function Codes for Transmit and Receive */
    rs232_pram->rfcr =    TM_MPC860_SCC4_RECV_FUNC_CODE;
    rs232_pram->tfcr =    TM_MPC860_SCC4_XMIT_FUNC_CODE;

/* We always receive a full ethernet frame so we set it to 1518 */
    rs232_pram->mrblr =   TM_MPC860_SCC4_MAX_RECV_BUF_LEN;
/* maximum idle characters */
    rs232_pram->max_idl = TM_MPC860_SCC4_MAX_IDL16;        

/* Tell the CP to initialize our Recv/Xmit parameters */
    tfMpc860IssueCommand(TM_MPC860_INIT_RXTX_SCC4);
}




/* 
 * Initialize SCC1 to be an ethernet channel    
 */
void tfMpc860SerialScc4BoardInit()
{
/*  tlMpc860Ptr->sim_mcr =   0x608f; */

/*
 * IMPLEMENTATION NOTE:
 *
 * This routine is very board specific
 * In this example we use the Atlas Communications Engines
 * HSB devlopment board for an example
 * Please refer to your boards design to modify this routine
 */

/* Atlas HSB Setup */
    tlMpc860Ptr->si_sicr |=   TM_MPC860_SI_SCC4_RCS_BRG1 |
                             TM_MPC860_SI_SCC4_TCS_BRG1;

/* Set the divider to 0x130 and RESET and no DIV16 
 * NOTE: The divider starts in bit location 1 not 0
 * The formula is
 * Divider=ClockRate/DesiredBaudRate/16-1*2
 * Example for 24.576 clock
 * 318=((24576000/9600)/16)-1*2
 * 0x130=318
 */
    tlMpc860Ptr->brgc1 = 0x130|TM_MPC860_BRGC_RST;                            

/* Make PortA Pins 6,7 for onchip use */
    tlMpc860Ptr->pio_padir &= ~(TM_MPC860_PORTA_PA6 |
                               TM_MPC860_PORTA_PA7);
    
/* 
 * pins 6,7 for SCC4 Xmit and Recv
 */
    tlMpc860Ptr->pio_papar |= TM_MPC860_PORTA_PA6 |
                             TM_MPC860_PORTA_PA7;
                            
/* Clear the pending and in-service registers */
    tlMpc860Ptr->intr_cipr = TM_MPC860_SCC4_INT_MASK;
    tlMpc860Ptr->intr_cisr = TM_MPC860_SCC4_INT_MASK;
/* Set the CP mask register to accept interrupts from SCC4 */
    tlMpc860Ptr->intr_cimr |= TM_MPC860_SCC4_INT_MASK;

    tlMpc860Ptr->sdma_sdcr = TM_MPC860_SDMA_ARB_ID;

    
/* PortC Pins 10 & 11 ar CD and CTS */
    tlMpc860Ptr->pio_pcpar &=   ~(TM_MPC860_PORTC_PC10 |
                                 TM_MPC860_PORTC_PC11);

    tlMpc860Ptr->pio_pcdir &=   ~(TM_MPC860_PORTC_PC10 |
                                 TM_MPC860_PORTC_PC11);
                                 
/*
 * Use Pins 10 and 11 (SCC4 CTS and CD)
 */
    tlMpc860Ptr->pio_pcso |=  TM_MPC860_PORTC_PC10|
                              TM_MPC860_PORTC_PC11;


/* Port B pin 12 is Ethernet TENA for HSB */
    tlMpc860Ptr->pip_pbpar |= TM_MPC860_PORTB_PB15;
    tlMpc860Ptr->pip_pbdir |= TM_MPC860_PORTB_PB15;
    

/*
 * IMPLEMENTATION NOTE:
 *
 * The Actual SCC4 ISR Wrapper Function in installed here
 * It must call the tfMpc860Scc1HandlerIsr function
 * This routine may be in "C" if it is supported
 * or assembly if not.  It is VERY RTOS dependent
 * so it cannot be provided with the driver
 *
 * With some RTOS's a wrapper function may not be
 * needed.  All the driver cares about is that the
 * function tfMpc860Scc1HandlerIsr function is called
 * when the interrupt occurs.
 *
 */

/* Install the wrapper function */
/* NOTE: TO ADD ISR HANDLER */
/*    tfKernelInstallIsrHandler(tfMpc860Scc4Isr,TM_MPC860_SCCD_INT); */
}

/* 
 * The main initialization routine for the mpc860 and the scc's 
 */
int tfMpc860SerialOpen(ttUserInterface interfaceHandle)
{
    unsigned long  mpc860Registers;

    mpc860Registers = TM_MPC860_IMMR_LOC;
    tlMpc860Ptr = (ttMpc860 *)(mpc860Registers&0xfffffffcUL);

/* Interface handle is not used, but could be for multiple ethernets */
    TM_UNREF_IN_ARG(interfaceHandle);
/* Initialize default Ethernet parameters */
    tfMpc860SerialScc4Init();
/* Initial board specific parameters */
    tfMpc860SerialScc4BoardInit();
/* Bring the ethernet scc port online */
    tfMpc860SerialUp(TM_MPC860_SCC4_CHANNEL);

/* Save our interfaceHandle for ISR use */
    tlInterfaceArray[TM_MPC860_SCC4_CHANNEL] = interfaceHandle;
    return TM_DEV_OKAY;
}

/*
 * Disable the Serial (RS-232) on SCC4
 */
int tfMpc860SerialClose(ttUserInterface interfaceHandle)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    tfMpc860Down(TM_MPC860_SCC4_CHANNEL);
    return 0;
}

/* 
 * Enable the SCC's  to recieve and transmit 
 */
void tfMpc860SerialUp(int channel)
{
    struct scc_regs * TM_FAR regs;


    regs = &tlMpc860Ptr->scc_regs[channel];

/* clear all events on the SCC event register and CP event reg*/
    regs->scc_scce = 0xffff;

/* Clear the appropriate bit in the CISR */
    switch (channel)
    {
        case TM_MPC860_SCC1_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC1_INT_MASK;
            break;
        case TM_MPC860_SCC2_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC2_INT_MASK;
            break;
        case TM_MPC860_SCC3_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC3_INT_MASK;
            break;
        case TM_MPC860_SCC4_CHANNEL:
            tlMpc860Ptr->intr_cisr = TM_MPC860_SCC4_INT_MASK;
            break;
        default:
            break;
    }


/* 
 * mask the events we sich to interrupt on
 */
    regs->scc_sccm = 
/* A buffer has completed transmitting */
                     TM_MPC860_INTR_SERIAL_XMIT_B |
/* A ethernet frame has been received */
                     TM_MPC860_INTR_SERIAL_RECV_B |
/* The busy condition has occured */
                     TM_MPC860_INTR_SERIAL_BUSY;
    
/* protocol specific mode register */
    regs->scc_psmr = 
/* CTS Flow Control On */
                     TM_MPC860_UART_FLOW |
/* CHaracters are 8 Bits */
                     TM_MPC860_CHAR_8;
                     


/* SCC general mode reg */
/* Set receive FIFO to 1 byte for UART mode */
    regs->scc_gsmrh = TM_MPC860_UART_RECV_FIFO_1;

/* SCC general mode reg */
    regs->scc_gsmrl = 
/* 16x Sample Rate */
                      TM_MPC860_UART_RECV_SAMPLE | 
                      TM_MPC860_UART_XMIT_SAMPLE | 
/* Serial UART Mode */
                      TM_MPC860_UART;

/* enable SCC1 receive/transmit operation */
    regs->scc_gsmrl |= TM_MPC860_ENABLE_RECV | TM_MPC860_ENABLE_XMIT;
     
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trmpc860.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 57845    Output checksum: 41477
 */
