/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * $Source: source/drivers/mpc860/trmpcfec.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:12CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/* 
 * SOURCE FILE FOR Motorola MPC860 QUICC Sample Driver
 *
 * This device driver is derived from the Ethernet Bridge example from
 * Motorola. Also uses code derived from Motorola's FEC example (tdemo.c)
 * It has been modified to interface to the Treck TCP/IP stack and to
 * provide readability.                                            
 *
 * Since it is a derived work, it remains the property of Motorola Corp.
 *
 * We have left the definintions of the mpc860 registers as they
 * are in the QUICC user manual
 *
 * This is why they do not match TRECK coding standards.
 *
 * This driver differs from the standard MPC860 driver (trmpc860.c) in that
 * it utilizes the Fast Ethernet Controller (FEC) which is a completely 
 * seperate component from the standard SCCs (and is in fact external to the
 * CPM entirely).  Currently, this driver *only* utilizes the FEC, though
 * it is certainly possible to also include code for the SCCs as well.  This
 * would allow you to use the DiamondT to run a 100BaseT & 10BaseT Ethernet and
 * a serial connection simultaneously (though there are some caveats: see
 * Motorola documentation).
 *
 * This driver also accounts for the fact that caching may be enabled on the 
 * MPC860 core.  Currently, we do a number of things to make this work.  First,
 * all memory is marked as write-through.  This solves the problem with 
 * transmit buffers.  Next, on reciept of a packet, we invalidate the cache,
 * thus synchronizing data cache and memory and allowing us to access the 
 * received packet.  Finally, all transmit/receive descriptors are placed in
 * uncached memory.  A good rule of thumb for the MPC860 is that anytime the
 * chip deals with memory, the cache and memory are unsynced (since the chip
 * DMAs and the MPC860 doesn't support cache bus snooping) and something must
 * be done to work around it.  This driver works; however, there are probably 
 * more elegant ways to handle caching.
 *
 */


#include <trsocket.h>
#include <trmpc860.h>

/* 
 * IMPORTANT
 * If the mpc860 device only receives a few frames and stops forever,
 * make sure that you are calling the tfMpc860RefillFECRecvPool()
 * function from a low priority task
 */


/*    
 * Since we do not accept SHORT frames, we use the SHORT bit to 
 * mean that this frame has been notified
 * IF THIS DRIVER IS CHANGED TO ALLOW SHORT FRAMES, ANOTHER BIT
 * WILL NEED TO BE USED
 */
#define TM_MPC860_NOTIFIED TM_MPC860_RECV_E_SHORT
#define TM_MPC860_ERROR_RECV 0x0037

/*
 * This mask is used to specify that a specific address is in uncached
 * memory space.  In our example, memory from 0x00000000 - 0x7fffffff is
 * cached memory, and memory from 0x80000000 - 0xffffffff is uncached.
 * So by applying this address mask, we can specify that certain memory
 * (eg, transmit/receive descriptors) are uncached.
 */
#define TM_MPC860_UNCACHED_ADDR_MASK 0x80000000UL

/*
 * Local Function Prototypes
 */

/* Fast Ethernet Controller interface */
void tfMpc860FECSendComplete(ttUserInterface interfaceHandle);
void tfMpc860FECUp(void);
void tfMpc860FECBoardInit(void);
void tfMpc860FECInit(void);
void tfMpc860RefillFECRecvPool(void);

/* General functions */     
void tfMpc860CreateXmitPool(int maxBd);
void tfMpc860CreateRecvPool(int maxBd);
void tfMpc860FECDown(void);

static ttMpc860Ptr tlMpc860Ptr;

/* Ethernet driver local variables */
static ttMpc860BufDescPtr tlMpc860FECXmitBufDescPtr;
static ttMpc860BufDescPtr tlMpc860FECXmitIsrBufDescPtr;
static ttMpc860BufDescPtr tlMpc860FECXmitBufDescStartPtr;

static ttMpc860BufDescPtr tlMpc860FECRecvBufDescPtr;
static ttMpc860BufDescPtr tlMpc860FECRecvIsrBufDescPtr;
static ttMpc860BufDescPtr tlMpc860FECRecvBufDescStartPtr; 
static ttMpc860BufDescPtr tlScatterBufferStart;

static ttMpc860BufDescPtr tlMpc860FECRecvRefillBufDescPtr; 


/*
 * Unlike the CPM, the FEC needs the transmit and receive buffer descriptors
 * to reside in external memory, so we need to declare them here.  We 
 * allocate them later.
 */

ttMpc860BufDesc *tlMpc860RecvBufDescPtr;
ttMpc860BufDesc *tlMpc860XmitBufDescPtr;

unsigned long     tvMpc860RecvErrorCount;
unsigned long     tvMpc860RecvPacketCount;
unsigned long     tvMpc860SendPacketCount;
unsigned long     tvMpc860BusyPacketCount;
unsigned long     tvMpc860XmitWait;
unsigned long     tvMpc860IsrRecvErrorCount;

/*
 * Four Interface handles for four SCC's 
 */
static ttUserInterface tlFECInterfaceHandle;
static int             tlFECXmitQueueOn;

/*
 * Table of ttUserBuffer Handles
 */
static ttUserBuffer    tlUserBufferList[TM_MPC860_FEC_MAX_RECV_BUFD];

/*
 * Our index to the ttUserBuffer's
 */
static ttUserBufferPtr tlUserBufferPtr; 
static ttUserBufferPtr tlUserRefillBufferPtr; 


/*
 *
 * Ethernet interface
 *
 */

/*
 * Return the Physical Address for the device
 * For the MPC860, it should be read from EEPROM
 * because every ethernet address is Unique
 *
 * THIS ROUTINE MUST BE MODIFIED!
 *
 * To get an ethernet address address block for 
 * your company, you need to contact IEEE Standards
 * or visit http://standards.ieee.org/faqs/OUI.html
 * for information on getting an Company ID/OUI
 */
int tfMpc860GetPhyAddr ( ttUserInterface userInterface,
                         char   TM_FAR * address )
{
/* 
 * We don't use the interface handle here, but we keep the compiler
 * from complaining about unused formal parameters
 */
    TM_UNREF_IN_ARG(userInterface);
    address[0] = 0x00;
    address[1] = 0xea;
    address[2] = 0xc0;
    address[3] = 0xf0;
    address[4] = 0x08;
    address[5] = 0x60;
    return 0;
}


/* 
 * Initialize FEC
 */
void tfMpc860FECInit (void)
{
    char   etherAddr[6];

/* Reset the FEC, which also disables send & recieve */
    tlMpc860Ptr->t_ecntrl = TM_MPC860T_ECNTRL_ETHERNETRESET;

/* Retrieve the Ethernet Address */
    tfMpc860GetPhyAddr((ttUserInterface)0, etherAddr);

/* Set the FEC's physical address */
/* First four bytes of address */
    tlMpc860Ptr->t_addr_low =
                 (   ( (unsigned short)etherAddr[1] & 0x00ff )
                   | ( ( ( (unsigned short)etherAddr[0] ) << 8 ) & 0xff00 ) );
    tlMpc860Ptr->t_addr_low <<= 16;
    tlMpc860Ptr->t_addr_low |=
                 (   ( (unsigned short)etherAddr[3] & 0x00ff )
                   | ( ( ( (unsigned short)etherAddr[2] ) << 8 ) & 0xff00 ) );
/* Last two bytes of address */
    tlMpc860Ptr->t_addr_high =
               (   ( (unsigned short)etherAddr[5]        & 0x00ff )
                 | ( (((unsigned short)etherAddr[4]) << 8) & 0xff00 ) ) << 16;

/* 
 * Set 860T To generate the following interrupts:
 *  - Transmit frame
 *  - Recieve frame
 *  - Bus error
 */
    tlMpc860Ptr->t_imask = TM_MPC860T_IEVENT_RECEIVEFRAMEINT;



/* Clear interrupt event register */
    tlMpc860Ptr->t_ievent = 0xffffffffUL;
        
/* Set interrupt level? (t_ivec) */
    tlMpc860Ptr->t_ivec = (TM_MPC860_FEC_LEVEL << TM_MPC860T_IVEC_LEVEL);

/* Enable the FEC interrupt in the SIU */       
    tlMpc860Ptr->siu_simask |= TM_MPC860_FEC_MASK;

    asm(" mtspr 80,0");
/*
 * The FEC hash table is used for multicasting Ethernet frames.  This is
 * handled much differently than in the CPM.  See Motorola's MPC860T supplement
 * to the PowerQUICC User's Manual for more information.
 */
    tlMpc860Ptr->t_hash_table_high = TM_MPC860T_HASH_TABLE_HIGH;
    tlMpc860Ptr->t_hash_table_low = TM_MPC860T_HASH_TABLE_LOW;

/* We always receive a full ethernet frame so we set buffer size to 1518 */

    tlMpc860Ptr->t_r_buff_size =  TM_MPC860_FEC_MAX_RECV_BUF_LEN;

    tlMpc860RecvBufDescPtr = 
          tfKernelMalloc(TM_MPC860_FEC_MAX_RECV_BUFD*sizeof(ttMpc860BufDesc));

    tlMpc860XmitBufDescPtr = 
          tfKernelMalloc(TM_MPC860_FEC_MAX_XMIT_BUFD*sizeof(ttMpc860BufDesc));


    if (tlMpc860RecvBufDescPtr == 0 || tlMpc860XmitBufDescPtr == 0)
    {
        tfKernelError("tfMpc860FECInit",
                      "Not enough memory for descriptor rings");
    }

/* points to start of receiving ring */
    tlMpc860Ptr->t_r_des_start = (unsigned long) tlMpc860RecvBufDescPtr;
/* points to start of transmiting ring */
    tlMpc860Ptr->t_x_des_start = (unsigned long) tlMpc860XmitBufDescPtr;   

/*
 * We want the receieve and transmit buffer descriptors to be uncached, since
 * they are DMA'ed to by the FEC.  We accomplish this by setting the MSB of
 * the address which, in our memory map at least, places the memory in the
 * uncached region.  This will obviously need to modified for your specific
 * memory layout.  If caching is disabled, the address mask can be zero.
 */
    tlMpc860RecvBufDescPtr = (ttMpc860BufDesc TM_FAR *)
     ( (unsigned long)tlMpc860RecvBufDescPtr | TM_MPC860_UNCACHED_ADDR_MASK );

    tlMpc860XmitBufDescPtr = (ttMpc860BufDesc TM_FAR *)
     ( (unsigned long)tlMpc860XmitBufDescPtr | TM_MPC860_UNCACHED_ADDR_MASK );

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/* FEC reciever mode: MII interface, promiscuous receive, full duplex */
/* TBD: we should use the CRC32 polynomial 0x04c11db7 to compute the hash,
   and program the hash table registers similar to what is done by the SMC
   device driver. For now, to get MLD to work for IPv6, we just set the
   receiver to be in promiscuous receive mode. */
    tlMpc860Ptr->t_r_cntrl = TM_MPC860T_MIIMODE | TM_MPC860T_PROMISCUOUS;
#else /* !(TM_IGMP or TM_6_USE_MLD) */
/* FEC reciever mode: MII interface, full duplex */
    tlMpc860Ptr->t_r_cntrl = TM_MPC860T_MIIMODE;
#endif /* !(TM_IGMP or TM_6_USE_MLD) */


/* FEC transmitter mode: not promiscuous, MII interface, full duplex */
    tlMpc860Ptr->t_x_cntrl = 0; /* TM_MPC860T_FULLDUPLEXENABLE; */

/* Set byte ordering in DMA to Motorola, and DMA function code to 0 */
    tlMpc860Ptr->t_fun_code = TM_MPC860T_DATABO0 | TM_MPC860T_DATABO1 |
                              TM_MPC860T_DESCBO0 | TM_MPC860T_DESCBO1;

/* Maximum Ethernet Frame Length (Always 1518) */
/*    tlMpc860Ptr->t_r_hash = TM_MPC860_FEC_MAX_FRAME_LEN; */


    tvMpc860RecvPacketCount = 0L;
    tvMpc860SendPacketCount = 0L;
    tvMpc860XmitWait = 0L;
    tvMpc860RecvErrorCount = 0L;
    tvMpc860IsrRecvErrorCount = 0L;
    tlScatterBufferStart = (ttMpc860BufDescPtr)0;
        
}


/* 
 * Initialize FEC to be an ethernet channel    
 */
void tfMpc860FECBoardInit (void)
{
/*
 * IMPLEMENTATION NOTE:
 *
 * This routine is very board specific
 * In this example we use the Atlas Communications Engines
 * DARWIN devlopment board for an example
 * We assume the the ethernet is in SLOT 0 and at SCC1
 * on the DARWIN
 *
 * Please refer to your boards design to modify this routine
 */


/* clear previous scc1 settings */
    tlMpc860Ptr->si_sicr &= ~(0x3f);
/* Atlas DARWIN Setup */
    tlMpc860Ptr->si_sicr |=  TM_MPC860_SI_SCC1_RCS_CLK1 |
                             TM_MPC860_SI_SCC1_TCS_CLK2;
                             
/* Disconnect the clock routing source */
    tlMpc860Ptr->si_sicr &=  ~TM_MPC860_SI_SCC1_CONNECT;
                            
/* Make PortA Pins 6,7,14,15 for onchip use */
    tlMpc860Ptr->pio_padir &= ~(TM_MPC860_PORTA_PA6 |
                                TM_MPC860_PORTA_PA7 |
                                TM_MPC860_PORTA_PA14 |
                                TM_MPC860_PORTA_PA15);

    
/* 
 * Use PortA pins 15,14 for SCC1 Ethernet RX/TX and 
 * pins 6,7 for Ethernet TCLK and RCLK 
 */
    tlMpc860Ptr->pio_papar |= TM_MPC860_PORTA_PA6 |
                              TM_MPC860_PORTA_PA7 |
                              TM_MPC860_PORTA_PA14 |
                              TM_MPC860_PORTA_PA15;
/* Turn off the Open Drain */
    tlMpc860Ptr->pio_paodr &= ~TM_MPC860_PORTA_PA14;
                            
/* Clear the pending and in-service registers */
    tlMpc860Ptr->intr_cipr = TM_MPC860_SCC1_INT_MASK;
    tlMpc860Ptr->intr_cisr = TM_MPC860_SCC1_INT_MASK;
/* Set the CP mask register to accept interrupts from SCC1 */
    tlMpc860Ptr->intr_cimr |= TM_MPC860_SCC1_INT_MASK;

    tlMpc860Ptr->sdma_sdcr = TM_MPC860_SDMA_ARB_ID;

    
 /* PortC Pins 10 & 11 are onchip */
    tlMpc860Ptr->pio_pcdir &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);
 /* PortC Pins 10 & 11 are onchip */
    tlMpc860Ptr->pio_pcpar &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);
                                 
/*
 * Use Pins 10 and 11 (SCC1 CD and CTS) for Ethernet RENA and CLSN  
 */
    tlMpc860Ptr->pio_pcso |=   TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11;
                              
/* Clear the data register for pins 10 and 11 */
    tlMpc860Ptr->pio_pcdat &=  ~(TM_MPC860_PORTC_PC10 | TM_MPC860_PORTC_PC11);

/*
 * Port B External Signals 
 * PB14,PB15 are clock select
 * PB26 is Link Enable
 * PB27 Is Full Duplex Select
 */
    tlMpc860Ptr->pip_pbdir |= TM_MPC860_PORTB_PB14 |
                              TM_MPC860_PORTB_PB15 |
                              TM_MPC860_PORTB_PB27 |
                              TM_MPC860_PORTB_PB26;
                                
    tlMpc860Ptr->pip_pbpar &= ~(TM_MPC860_PORTB_PB14 |
                                TM_MPC860_PORTB_PB15 |
                                TM_MPC860_PORTB_PB27 |
                                TM_MPC860_PORTB_PB26);
/* Set clock select to 00 */
    tlMpc860Ptr->pip_pbdat &= ~(TM_MPC860_PORTB_PB14 |
                                TM_MPC860_PORTB_PB15);
/* Set Link Enable */
    tlMpc860Ptr->pip_pbdat |=  TM_MPC860_PORTB_PB26;

/* Turn OFF Full Duplex Mode */
    tlMpc860Ptr->pip_pbdat &=  ~TM_MPC860_PORTB_PB27;
    


/*
 * IMPLEMENTATION NOTE:
 *
 * The Actual SCC1 ISR Wrapper Function in installed here
 * It must call the tfMpc860Scc1HandlerIsr function
 * This routine may be in "C" if it is supported
 * or assembly if not.  It is VERY RTOS dependent
 * so it cannot be provided with the driver
 *
 * With some RTOS's a wrapper function may not be
 * needed.  All the driver cares about is that the
 * function tfMpc860Scc1HandlerIsr function is called
 * when the interrupt occurs.
 *
 */

/* Port B pin 19 is Ethernet TENA for DARWIN */
    tlMpc860Ptr->pip_pbpar |= TM_MPC860_PORTB_PB19;
    tlMpc860Ptr->pip_pbdir |= TM_MPC860_PORTB_PB19;

/* Configure Port D pins to enable MII interface */
    tlMpc860Ptr->pio_pdpar = TM_MPC860_PORTD_PD3 | TM_MPC860_PORTD_PD4 |
                             TM_MPC860_PORTD_PD5 | TM_MPC860_PORTD_PD6 |
                             TM_MPC860_PORTD_PD7 | TM_MPC860_PORTD_PD8 |
                             TM_MPC860_PORTD_PD9 | TM_MPC860_PORTD_PD10 |
                             TM_MPC860_PORTD_PD11 | TM_MPC860_PORTD_PD12 |
                             TM_MPC860_PORTD_PD13 | TM_MPC860_PORTD_PD14 |
                             TM_MPC860_PORTD_PD15;
    tlMpc860Ptr->pio_pddir = TM_MPC860_PORTD_PD3 | TM_MPC860_PORTD_PD4 |
                             TM_MPC860_PORTD_PD5 | TM_MPC860_PORTD_PD9 |
                             TM_MPC860_PORTD_PD11 | TM_MPC860_PORTD_PD12;

/* Install the wrapper function */
    tfKernelInstallIsrHandler(tfMpc860FECIsr,TM_MPC860_FEC_INT);

}

/* 
 * Enable the SCC's  to recieve and transmit 
 */
void tfMpc860FECUp (void)
{

/* Reset the FEC, which also disables send & recieve */
    tlMpc860Ptr->t_ecntrl = TM_MPC860T_ECNTRL_ETHERNETENABLE;

    tlMpc860Ptr->t_r_des_active = TM_MPC860T_BD_ACTIVE;
    tlMpc860Ptr->t_x_des_active = TM_MPC860T_BD_ACTIVE;

}

/*
 * Disable the Ethernet on FEC
 */
int tfMpc860FECClose(ttUserInterface interfaceHandle)
{

    ttMpc860BufDescPtr      bufDescPtr;
    ttUserBufferPtr         userBufPtr;
    int                     i;


/* Shut down the SCC1 Ethernet hardware */
    tfMpc860FECDown();

/* 
 * Free leftover memory:
 *   First, send complete on any remaining packets in the transmit ring.
 */
    bufDescPtr = tlMpc860FECXmitBufDescStartPtr;
    for (i = 0; i < TM_MPC860_FEC_MAX_XMIT_BUFD; i++)
    {
/* Call send complete to free this entry */
        if ( ((bufDescPtr->bdStatus) & TM_MPC860_XMIT_LAST ) &&
             (bufDescPtr->bdDataPtr != (char TM_FAR *) 0) )
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER );
        }
        bufDescPtr++;
    }


/* Next, free any buffers in the receive ring */
    bufDescPtr = tlMpc860FECRecvBufDescStartPtr;
    userBufPtr = &(tlUserBufferList[0]);
    for (i = 0; i < TM_MPC860_FEC_MAX_RECV_BUFD; i++)
    {
        if ((bufDescPtr->bdDataPtr) != (char TM_FAR *) 0)
        {
            tfFreeDriverBuffer(*userBufPtr);
        }
        bufDescPtr++;
        userBufPtr++;
    }
        

    return TM_DEV_OKAY;

}

/* 
 * The main initialization routine for the mpc860 and the scc's 
 */
int tfMpc860FECOpen(ttUserInterface interfaceHandle)
{

    unsigned long  mpc860Registers;
    int            errorCode;

    mpc860Registers = TM_MPC860_IMMR_LOC;
    tlMpc860Ptr = (ttMpc860 *)(mpc860Registers & 0xfffffffcUL);

    errorCode = TM_DEV_OKAY;

/* 
 * Check that the number of transmit descriptors exceed to the number
 * of buffers per frame that the user specified in tfOpenInterface.
 */
    if (tfInterfaceScatterCount(interfaceHandle) >
            TM_MPC860_ETHER_MAX_XMIT_BUFD)
    {
        tfKernelWarning("tfMpc860EtherOpen",
                        "Scatter buffers exceed descriptors");
        errorCode = TM_DEV_ERROR;
    }

    if (errorCode == TM_DEV_OKAY)
    {
/* Interface handle is not used, but could be for multiple ethernets */
        TM_UNREF_IN_ARG(interfaceHandle);
/* Initialize default Ethernet parameters */
        tfMpc860FECInit();
/* Create our transmit pool */
        tfMpc860CreateXmitPool(TM_MPC860_FEC_MAX_XMIT_BUFD);
/* Create our receive pool (with receive buffers) */
        tfMpc860CreateRecvPool(TM_MPC860_FEC_MAX_RECV_BUFD);
/* Initial board specific parameters */
        tfMpc860FECBoardInit();
/* Bring the ethernet scc port online */
        tfMpc860FECUp();

/* Save our interfaceHandle for ISR use */
        tlFECInterfaceHandle = interfaceHandle;
/* Save whether the user enabled the Treck transmit queue */
        tlFECXmitQueueOn = tfInterfaceXmitQueueOn(interfaceHandle);
    }

    return errorCode;
}

/* 
 * Interrupt service routine called from actual
 * ISR handlers
 */ 
void tfMpc860FECHandlerIsr (void)
{
    register unsigned long   event;
    ttUserInterface          interfaceHandle;
    unsigned short  TM_FAR * bdStatusPtr;
    int                      recvNotifyCount;
    int                      sentNotifyCount;
    unsigned long            sentNotifyBytes;
#ifdef TM_TASK_SEND
    void  TM_FAR * TM_FAR  * bdDataPtrPtr;
#endif /* TM_TASK_SEND */
    
    interfaceHandle = tlFECInterfaceHandle;

/* Check to see what events have occured */
    event = tlMpc860Ptr->t_ievent;

/* Clear event bits that we have handled */
    tlMpc860Ptr->t_ievent = event;

    recvNotifyCount = 0;
    sentNotifyCount = 0;
    sentNotifyBytes = 0UL;


#ifdef TM_TASK_SEND
    if( event & TM_MPC860T_INTR_ETHER_XMIT_B ) 
    {
        bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
        while (    ( ( *bdStatusPtr & TM_MPC860_XMIT_READY ) == 0 )
                && ( *bdDataPtrPtr != (void TM_FAR *)0 ) )
        {
/* Zero it out so we know that we have freed this one before */
            *bdDataPtrPtr = void TM_FAR *)0;
            if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
            {
                tlMpc860FECXmitIsrBufDescPtr = tlMpc860FECXmitBufDescStartPtr;
            }
            else
            {
                tlMpc860FECXmitIsrBufDescPtr++;
            }
            bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
            bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
            sentNotifyCount++;
            sentNotifyBytes = (unsigned long)
                  (   sentNotifyBytes
                    + (unsigned long)tlMpc860FECXmitIsrBufDescPtr->bdLength );
        }
    }
#endif /* TM_TASK_SEND */

    if ( event & TM_MPC860T_INTR_ETHER_RECV_F )
    {
/* loop on frames with the Empty bit and SHORT (notify) bit is NOT set "==0" */
/* Only notify while the frames are not empty */
        bdStatusPtr = &tlMpc860FECRecvIsrBufDescPtr->bdStatus;
        while (!(*bdStatusPtr & (TM_MPC860_RECV_EMPTY|TM_MPC860_NOTIFIED)))
        {
            recvNotifyCount++;
            *bdStatusPtr |= TM_MPC860_NOTIFIED;
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860FECRecvIsrBufDescPtr = tlMpc860FECRecvBufDescStartPtr;
            }
            else
            {
                tlMpc860FECRecvIsrBufDescPtr++;
            }
            bdStatusPtr = &tlMpc860FECRecvIsrBufDescPtr->bdStatus;
        }

    }
    if ( (recvNotifyCount != 0) || (sentNotifyCount != 0) )
    {
        tfNotifyInterfaceIsr( interfaceHandle, recvNotifyCount,
                              sentNotifyCount, sentNotifyBytes, 0UL );
    }
}

/*
 * Send a packet to the network
 */
int tfMpc860FECSend ( ttUserInterface interfaceHandle,
                      char   TM_FAR * bufferPtr,
                      int             length,
                      int             flag )
{
    ttMpc860BufDescPtr      bufDescPtr;
    ttMpc860BufDescPtr      startBufDescPtr;
    unsigned short TM_FAR * bdStatusPtr;
    void  TM_FAR * TM_FAR * bdDataPtrPtr;
    int                     errorCode;

    errorCode = TM_ENOERROR;
/* Wait for a free buffer descriptor */    
    if (tlMpc860FECXmitBufDescPtr->bdStatus & TM_MPC860_XMIT_READY)
    {
        if (tlFECXmitQueueOn)
        {
/* If user uses a transmit queue, queue the buffer in the transmit queue */
            errorCode = TM_DEV_ERROR;
            goto mpcFECSendExit;
        }
        else
        {
/* Otherwise */
            do
            {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
                tfInterfaceSpinLock(interfaceHandle);
                tvMpc860XmitWait++;
            }
            while (   tlMpc860FECXmitBufDescPtr->bdStatus
                    & TM_MPC860_XMIT_READY );
        }
    }
#ifndef TM_TASK_SEND

/* Inlined send complete */
    bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
    bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
    while (    (    (*bdStatusPtr & TM_MPC860_XMIT_READY) == 0
                 && tlMpc860FECXmitIsrBufDescPtr != tlScatterBufferStart )
            && ( *bdDataPtrPtr != (void TM_FAR *)0 ) )
/* 
 * Watch out for a scattered frame we have not yet finished
 * loading up
 */
    {
/* Zero it out so we know that we have freed this one before */
        *bdDataPtrPtr = (void TM_FAR *)0;
        if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
        {
            tlMpc860FECXmitIsrBufDescPtr = tlMpc860FECXmitBufDescStartPtr;
        }
        else
        {
            tlMpc860FECXmitIsrBufDescPtr++;
        }
/* Only call send complete when the last piece has been transmitted */
        if (*bdStatusPtr & TM_MPC860_XMIT_LAST)
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER);
        }
        bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
    }
#endif /* NOT TM_TASK_SEND */
    bufDescPtr = tlMpc860FECXmitBufDescPtr++;
    bdStatusPtr = &bufDescPtr->bdStatus;
    if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
    {
        tlMpc860FECXmitBufDescPtr = tlMpc860FECXmitBufDescStartPtr;
    }
    bufDescPtr->bdDataPtr = (void TM_FAR *)bufferPtr;
    bufDescPtr->bdLength = length;
/* Clear the previous status (but preserve the WRAP bit) */
    bufDescPtr->bdStatus &= TM_MPC860_XMIT_WRAP;
/* If this is the last buffer in the frame then xmit it */
    if (flag == TM_USER_BUFFER_LAST)
    {
/* Now mark the last bd ready */
        bdStatusPtr = &bufDescPtr->bdStatus;
        *bdStatusPtr |= (TM_MPC860_XMIT_READY |
                        TM_MPC860_XMIT_PAD |

#ifdef TM_TASK_SEND
                        TM_MPC860_XMIT_INTR | 
#endif /* TM_TASK_SEND */

                        TM_MPC860_XMIT_LAST |
                        TM_MPC860_XMIT_CRC);

/* Check to see if we are xmitting scattered data */
        if (tlScatterBufferStart != (ttMpc860BufDescPtr)0)
        {
/* We have at least one more piece to xmit if we are here */
            startBufDescPtr = tlScatterBufferStart;
            tlScatterBufferStart = (ttMpc860BufDescPtr)0;
/* 
 * Go ahead and xmit the first packets (with more set) 
 * by walking the list backwards
 * We do this so that we don't need a critical section
 * in order to keep from getting gaps inbetween the 
 * pieces
 * All we need to do here is update the
 * ready flag
 */
            do
            {
                if (bufDescPtr == tlMpc860FECXmitBufDescStartPtr)
                {
/* We are at the front of the list so go to the end */
                    bufDescPtr += TM_MPC860_FEC_MAX_XMIT_BUFD;
                }
                else
                {
/* Go back one buffer descriptor */
                    bufDescPtr--;
                }
                bdStatusPtr = &bufDescPtr->bdStatus;
                *bdStatusPtr |= TM_MPC860_XMIT_READY;
            } while (bufDescPtr != startBufDescPtr);
        }

/* Inform the MPC860T that there are new transmit frames to process */
        tlMpc860Ptr->t_x_des_active = TM_MPC860T_BD_ACTIVE;
    }
    else
    {
/* More buffers to come */
        if (tlScatterBufferStart == (ttMpc860BufDescPtr)0)
        {
            tlScatterBufferStart=bufDescPtr;
        }
    }

    tvMpc860SendPacketCount++;

mpcFECSendExit:
    return errorCode;
}


/*
 * tfMpc860Receive
 * Receive a buffer from the chip and pass back to the stack
 */
int tfMpc860FECReceive ( ttUserInterface        interfaceHandle,
                         char TM_FAR * TM_FAR * dataPtrPtr,
                         int  TM_FAR          * lengthPtr,
                         ttUserBufferPtr        userBufferHandlePtr )
{
    int                     errorCode;
    ttMpc860BufDescPtr      bufDescPtr;
    unsigned short TM_FAR * bdStatusPtr;
    
    TM_UNREF_IN_ARG(interfaceHandle);
    errorCode = TM_ENOERROR;
    
    bdStatusPtr = &tlMpc860FECRecvBufDescPtr->bdStatus;

/* Skip pass error packets who are not empty */
    if (    ( !( *bdStatusPtr &  TM_MPC860_RECV_EMPTY ) )
         && ( *bdStatusPtr & TM_MPC860_ERROR_RECV ) )
    {
        tvMpc860RecvErrorCount++;
                
/* Return the userBuffer Handle for this data pointer */
        if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
        {
            tlMpc860FECRecvBufDescPtr = tlMpc860FECRecvBufDescStartPtr;
            tlUserBufferPtr = tlUserBufferList;
        }
        else
        {
            tlMpc860FECRecvBufDescPtr++;
                        tlUserBufferPtr++;
        }
        errorCode = TM_EWOULDBLOCK;
    }
    else
    {
/* Check to see if we own the Buffer Descriptor */
        if (*bdStatusPtr & TM_MPC860_RECV_EMPTY)
        {
        
/* Will only happen when refill cleans up error before we are called here */
            tvMpc860RecvErrorCount++;
/* Return the userBuffer Handle for this data pointer */
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860FECRecvBufDescPtr = tlMpc860FECRecvBufDescStartPtr;
                tlUserBufferPtr = tlUserBufferList;
            }
            else
            {
                tlMpc860FECRecvBufDescPtr++;
                tlUserBufferPtr++;
            }
            errorCode = TM_EWOULDBLOCK;
        }
        else
        {
            bufDescPtr = tlMpc860FECRecvBufDescPtr++;
            *userBufferHandlePtr = *tlUserBufferPtr++;
            if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
            {
                tlMpc860FECRecvBufDescPtr = tlMpc860FECRecvBufDescStartPtr;
                tlUserBufferPtr = tlUserBufferList;
            }

/* 
 * Invalidate the cache 
 *
 *   This is very specific to the DDI compiler, and should be changed for
 *   your specific platform.
 */
            asm(" xor r3,r3,r3");
            asm(" lis r3, 0x0C00");
            asm(" mtspr 80,r3");
                        
/* We are returning the pointer to the data */
            *dataPtrPtr = (char TM_FAR *)(bufDescPtr->bdDataPtr);

/* Zero out the bufferPtr so we can refill it later */
            bufDescPtr->bdDataPtr=(void TM_FAR *)0;

/* We are returning the length */
            *lengthPtr  = bufDescPtr->bdLength;

/* Up our packet count */
            tvMpc860RecvPacketCount++;
        }
    }

    return(errorCode);
}

/*
 * General function
 * Can be used to configure the device
 * or refill the receive pool
 */
int tfMpc860Ioctl ( ttUserInterface interfaceHandle, int flag,
                    void TM_FAR * optionPtr, int optionLen )
{

    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(interfaceHandle);

    if (flag & TM_MPC860_REFILL_FEC)
    {
        tfMpc860RefillFECRecvPool();
    }
#ifndef TM_TASK_SEND
    if (flag & TM_MPC860_SEND_COMPLETE)
    {
        tfMpc860FECSendComplete(interfaceHandle);
    }
#endif /* NOT TM_TASK_SEND */
    return TM_ENOERROR;
}

/*
 * Walk through receive list and allocate new buffers
 * after they have been processed
 */
void tfMpc860RefillFECRecvPool (void)
{
    ttUserBuffer            userBuffer;
    unsigned short TM_FAR * bdStatusPtr;
    void           TM_FAR * dataPtr;
    void           TM_FAR * TM_FAR * bdDataPtrPtr;
        
/* Instruct the MPC860T to process receive frames */
    tlMpc860Ptr->t_r_des_active = TM_MPC860T_BD_ACTIVE;

    bdDataPtrPtr = &tlMpc860FECRecvRefillBufDescPtr->bdDataPtr;
    bdStatusPtr  = &tlMpc860FECRecvRefillBufDescPtr->bdStatus;
/* Loop while it is empty or an error */
    while (    ( ( *bdDataPtrPtr == (void TM_FAR *)0 )
            || ( *bdStatusPtr & TM_MPC860_ERROR_RECV ) ) )
    {
/* Check to see if it is an error */
        if ( !(*bdStatusPtr & TM_MPC860_ERROR_RECV) )
        {

/* If it is not get a new buffer, otherwise just reuse it */
/* bdDataPtr need to be aligned on a 4-byte boundary */
            dataPtr = (void TM_FAR *)
                    (tfGetEthernetBuffer(&userBuffer) - TM_ETHER_IP_ALIGN);
            if (userBuffer == (ttUserBuffer)0)
            {
/* No memory so stop the refill */
                break;
            }

            tlMpc860FECRecvRefillBufDescPtr->bdDataPtr = dataPtr;
            *tlUserRefillBufferPtr = userBuffer;

        }
        tlMpc860FECRecvRefillBufDescPtr->bdLength = 0;
/* Clear the status Except for the wrap bit */
        *bdStatusPtr &= TM_MPC860_RECV_WRAP;
/* Mark this BD as ready to use */
        *bdStatusPtr |= (TM_MPC860_RECV_EMPTY|TM_MPC860_RECV_INTR);
        if (*bdStatusPtr & TM_MPC860_RECV_WRAP)
        {
            tlMpc860FECRecvRefillBufDescPtr = tlMpc860FECRecvBufDescStartPtr;
            tlUserRefillBufferPtr = tlUserBufferList;
        }
        else
        {
            tlMpc860FECRecvRefillBufDescPtr++;
            tlUserRefillBufferPtr++;
        }
        bdDataPtrPtr = &tlMpc860FECRecvRefillBufDescPtr->bdDataPtr;
        bdStatusPtr  = &tlMpc860FECRecvRefillBufDescPtr->bdStatus;
    }
}

#ifndef TM_TASK_SEND
/*
 * Walk through send list and post send completes
 */
void tfMpc860FECSendComplete(ttUserInterface interfaceHandle)
{
    unsigned short TM_FAR * bdStatusPtr;
    void  TM_FAR * TM_FAR * bdDataPtrPtr;

    bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
    bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
/* An ethernet frame has completed transmitting so notify the user*/
    while (    (*bdStatusPtr & TM_MPC860_XMIT_READY) == 0
            && tlMpc860FECXmitIsrBufDescPtr != tlScatterBufferStart
            && *bdDataPtrPtr != (void TM_FAR *)0 )
    {
/* Zero it out so we know that we have freed this one before */
        *bdDataPtrPtr = (void TM_FAR *)0;
        if (*bdStatusPtr & TM_MPC860_XMIT_WRAP)
        {
            tlMpc860FECXmitIsrBufDescPtr = tlMpc860FECXmitBufDescStartPtr;
        }
        else
        {
            tlMpc860FECXmitIsrBufDescPtr++;
        }
/* Only call send complete when the last piece has been transmitted */
        if (*bdStatusPtr & TM_MPC860_XMIT_LAST)
        {
            tfSendCompleteInterface( interfaceHandle,
                                     TM_DEV_SEND_COMPLETE_DRIVER );
        }
        bdStatusPtr = &tlMpc860FECXmitIsrBufDescPtr->bdStatus;
        bdDataPtrPtr = &tlMpc860FECXmitIsrBufDescPtr->bdDataPtr;
    }
}
#endif /* NOT TM_TASK_SEND */



/*
 *
 * GENERAL FUNCTIONS
 *
 */

void tfMpc860FECDown (void)
{
/* Reset the FEC, which also disables send & recieve */
    tlMpc860Ptr->t_ecntrl = TM_MPC860T_ECNTRL_ETHERNETRESET;

/* Clear interrupt event register */
    tlMpc860Ptr->t_ievent = 0xffffffffUL;
}

/* 
 * This routine is preparing for each SCC a pool of READY buffer 
 * descriptors that are ready to recieve. The last buffer descriptor
 * has the wrap bit set.
 */
void tfMpc860CreateRecvPool (int maxBd)
{
    int                    i;
    ttMpc860BufDescPtr     tempBufDescPtr;
    ttUserBuffer           userBuffer;
    void          TM_FAR * dataPtr;

    tempBufDescPtr = tlMpc860RecvBufDescPtr;
/* Save it off for our use in the recv routine */    
    tlMpc860FECRecvRefillBufDescPtr = tempBufDescPtr;
    tlMpc860FECRecvBufDescStartPtr = tempBufDescPtr;
    tlMpc860FECRecvIsrBufDescPtr = tempBufDescPtr;
    tlMpc860FECRecvBufDescPtr = tempBufDescPtr;
    tlUserRefillBufferPtr = tlUserBufferList;
    tlUserBufferPtr = tlUserBufferList;
    for ( i = 0; i < maxBd; i++, tempBufDescPtr++ )
    {

/* bdDataPtr need to be aligned on a 4-byte boundary */
        dataPtr = (void TM_FAR *)
                    (tfGetEthernetBuffer(&userBuffer) - TM_ETHER_IP_ALIGN);
        if (userBuffer == (ttUserBuffer)0)
        {
/* No memory so stop the create here */
            break;
        }
        tempBufDescPtr->bdDataPtr = dataPtr;
        tlUserBufferList[i] = userBuffer;
        tempBufDescPtr->bdLength = 0;
        tempBufDescPtr->bdStatus = (   TM_MPC860_RECV_EMPTY
                                     | TM_MPC860_RECV_INTR );
    }
    --tempBufDescPtr;
 /* signify last bit on */
    tempBufDescPtr->bdStatus |= TM_MPC860_RECV_WRAP;
}
        
/* 
 * Prepare a transmit pool, set the status bits in the buffer descriptors
 * to be not READY to transmit and set the last buffer descriptor to
 * be wrapped
 */
void tfMpc860CreateXmitPool(int maxBd)
{
    int                           i;
    ttMpc860BufDescPtr            tempBufDescPtr;

    tempBufDescPtr = tlMpc860XmitBufDescPtr;

/* Save it off for our use in the xmit routine */    
    tlMpc860FECXmitIsrBufDescPtr = tempBufDescPtr;
    tlMpc860FECXmitBufDescStartPtr = tempBufDescPtr;
    tlMpc860FECXmitBufDescPtr = tempBufDescPtr;

    for( i = 0; i < maxBd; i++, tempBufDescPtr++ )
    {
        tempBufDescPtr-> bdStatus = 0;
        tempBufDescPtr-> bdLength = 0;
        tempBufDescPtr-> bdDataPtr = (void TM_FAR *)0;
    }
    --tempBufDescPtr;

/* signify last bit on */
    tempBufDescPtr->bdStatus |= TM_MPC860_XMIT_WRAP; 
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trmpcfec.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 18566    Output checksum: 55604
 */
