/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2010
 * Copyright Elmic Systems Japan 1997 - 2010
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Device driver for TI's TMX320VC5510 DSP starter kit, used with
 *  LogicIO Ethernet daughtercard (Macronix MX98728EC GMAC).
 *  Adapted from example code provided by Texas Instruments.
 *
 * Filename: trtic5mx.c
 * Author: Ed Remmell, adapted from code written by Michael A. Denio of TI
 * Date Created:
 * $Source: source/drivers/mx98728/trtic5mx.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2010/07/16 15:37:33CDT $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trtic5mx.h>

/* DSP/BIOS includes */
#include <std.h>
#include <hst.h>
#include <swi.h>
#include <tsk.h>
#include <log.h>
#include <sts.h>

/* Elmic stack includes */
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* constant declarations and enumerations */
/* GMAC interrupts we care about */
#define TM_INTMASK (TM_GM_RIM | TM_GM_TIM | TM_GM_TEIM)
#define TM_IFLAG   (1<<0)

/*
 * Our receive pool consists of two sized buffers: small and large.
 * These macros determine the size and number of these buffers in our
 * pre-allocated receive pool.
 * We need to allocate extra bytes in the packet for:
 *  - unpacking the Ethernet addresses (12 bytes),
 *  - a dummy 16-bit value in ttEtherHeader (DSP only),
 *  - up to 14 bytes of over-copy as a result copying data out of the
 *    device in 16 byte blocks.
 */
#define TM_GMAC_EXTRA_BYTES         28
#define TM_GMAC_LARGE_RECV_POOL     5
#define TM_GMAC_SMALL_RECV_POOL     5
#define TM_GMAC_LARGE_RECV_BUF_SIZE (1518 + TM_GMAC_EXTRA_BYTES)
#define TM_GMAC_SMALL_RECV_BUF_SIZE (128 + TM_GMAC_EXTRA_BYTES)

/* 
 * This macro determines on what sort of word boundary our buffers are aligned 
 * (word, long word, etc).
 */
#define TM_GMAC_BUF_ALIGNMENT       2 /* 4-byte alignment on 16-bit DSP */

/* uncomment the following macro to restrict the Macronix GMAC to only
   negotiate half-duplex mode */
#define TM_GMAC_HALF_DUPLEX_TX

/* type definitions */

/* local function prototypes */
static void gmInitialize();
static void gmStart();
static void gmStop();
static int gmResetRxFifo();

/* local variable definitions */
/* current TX state */
#define STATE_IDLE     0
#define STATE_SEND     2
static volatile tt32Bit TxState;
static tt32Bit *tlTxFrameBuf[2];
static int tlTxQueuedFrameLen[2]; /* 0 if no packet queued in tlTxFrameBuf */
static int tlNextSendFrameBufIndex;
static int tlIsrNextSendFrameBufIndex;

/* Simple Tx and Rx statistics */
static tt32Bit tlRxGood;
static tt32Bit tlRxBad;
static tt32Bit tlTxGood;
static tt32Bit tlTxBad;
static tt32Bit tlTxBusy;
static tt32Bit tlFatalError;
static tt32Bit tlResetRxFifo;
static tt32Bit tlIsrTx;

volatile int tlTxFree;

static ttUserInterface tlTiMxInterfaceHandle = 0;

/* static const tables */

/* macro definitions */

/* macros used to read from and write to GMAC registers */
#define READ32(x)    (*(volatile tt32Bit *)x)
#define WRITE32(x,y) (*(volatile tt32Bit *)(x) = (y))

/* Data Swizzle Macro (32 bit only) */
#define SWIZZLE(x) ((x>>24)|(x<<24)|((x>>8)&0xFF00)|((x<<8)&0xFF0000))

#define TM_EMIF_CE2_1_REG   *(volatile ioport tt16Bit*)(TM_EMIF_CE2_1)
#define TM_EMIF_CE2_2_REG   *(volatile ioport tt16Bit*)(TM_EMIF_CE2_2)
#define TM_EMIF_CE2_3_REG   *(volatile ioport tt16Bit*)(TM_EMIF_CE2_3)


/* Simple delay function */
static void Delay(tt32Bit t)
{
    volatile tt32Bit i;
    while( t-- )
        for( i=0; i<250000; i++);
}

/****************************************************************************
* FUNCTION: tfTiMxOpen
*
* PURPOSE: Initialize and open the GMAC.
*
* PARAMETERS:
*   interfaceHandle
*
* RETURNS:
*   TM_DEV_OKAY: successful
*   TM_DEV_ERROR: failed to read GMAC ID string
*
* NOTES:
*
****************************************************************************/
int tfTiMxOpen(ttUserInterface interfaceHandle)
{
    tt32Bit idString;
    
/* initialize statistics */
    tlRxGood = 0;
    tlRxBad = 0;
    tlFatalError = 0;
    tlResetRxFifo = 0;
    tlIsrTx = 0;
    tlTxGood = 0;
    tlTxBad = 0;
    tlTxBusy = 0;

    tlTxQueuedFrameLen[0] = 0;
    tlTxQueuedFrameLen[1] = 0;
    tlNextSendFrameBufIndex = 0;
    tlIsrNextSendFrameBufIndex = 0;

    tlTiMxInterfaceHandle = interfaceHandle;

/* malloc is passed the number of MADUs (16-bit char's) to alloc */
#ifdef TM_USE_SHEAP
    tlTxFrameBuf[0] = (tt32Bit *) tfSheapMalloc(TM_ETHER_MAX_PACKET_CRC/2);
#else
    tlTxFrameBuf[0] = (tt32Bit *) malloc(TM_ETHER_MAX_PACKET_CRC/2);
#endif
    tm_assert(tfTiMxOpen, !(((tt32Bit) tlTxFrameBuf[0]) & 1));
    tm_assert(tfTiMxOpen, (((tt32Bit) tlTxFrameBuf[0]) != 0));

#ifdef TM_USE_SHEAP
    tlTxFrameBuf[1] = (tt32Bit *)tfSheapMalloc(TM_ETHER_MAX_PACKET_CRC/2);
#else
    tlTxFrameBuf[1] = (tt32Bit *) malloc(TM_ETHER_MAX_PACKET_CRC/2);
#endif

    tm_assert(tfTiMxOpen, !(((tt32Bit) tlTxFrameBuf[1]) & 1));
    tm_assert(tfTiMxOpen, (((tt32Bit) tlTxFrameBuf[1]) != 0));
    
/* Set EMIF CE2 values */
/* The following is appropriate if your DSP clock frequency is 150 MHz */
/*  TM_EMIF_CE2_1_REG = 0x2673;
    TM_EMIF_CE2_2_REG = 0x5673;
    TM_EMIF_CE2_3_REG = 0x00FF; */

/* The following is appropriate if your DSP clock frequency is 200 MHz:

   CE21:
   MTYPE:  010b:   asynchronous memory, 32-bit data bus width
   RDSETUP:    8:  read setup period 8 CPU clock cycles
   RDSTROBE:   39: read strobe period 39 CPU clock cycles
   RDHOLD: 3:  read hold period 3 CPU clock cycles

   CE22:
   RDEXHLD:    1:  read extended hold period 2 CPU clock cycles
   WREXHLD:    1:  write extended hold period 2 CPU clock cycles
   WRSETUP:    8:  write setup period 8 CPU clock cycles
   WRSTROBE:   35: write strobe period 35 CPU clock cycles
   WRHOLD: 3:  write hold period 3 CPU clock cycles

   CE23:
   TIMOUT: 255

   Read cycle length = RDSETUP + RDSTROBE + RDHOLD = 50 CPU clocks.
   Write cycle length = WRSETUP + WRSTROBE + WRHOLD = 46 CPU clocks. */
    TM_EMIF_CE2_1_REG = 0x289f;
    TM_EMIF_CE2_2_REG = 0x588f;
    TM_EMIF_CE2_3_REG = 0x00FF;

/* try to read back the ID string, should be 0x0001 followed by "MX" */
    idString = READ32( TM_GMAC_ID);
    if (idString != 0x00014d58)
    {
        return TM_DEV_ERROR;
    }

/* Initialize receive buffer pool */
    (void)tfPoolCreate(interfaceHandle,
                       TM_GMAC_LARGE_RECV_POOL,
                       TM_GMAC_SMALL_RECV_POOL,
                       TM_GMAC_LARGE_RECV_BUF_SIZE,
                       TM_GMAC_SMALL_RECV_BUF_SIZE,
                       TM_GMAC_BUF_ALIGNMENT,
                       TM_POOL_REFILL_IN_LINE );

/* Attempt to reset the entire part */
    gmInitialize();

/* Enable board interrupt */
    C55_enableIMR1( TM_IFLAG );

/* Start */
    gmStart();

    return(TM_DEV_OKAY);
}

/*
 * tfTiMxClose()
 *
 *   This function is called when the user wishes to close this device.
 *   We need to clean up any allocated memory, disable any chip interrupts,
 *   and disable transmit/receive on the chip.  While we do need to free memory
 *   that was allocated for the receive pool (by calling tfPoolDelete) it is
 *   not necessary to clean up any transmit buffer, since they are freed 
 *   immediately upon return from tfTiMxSend().
 *
 */
int tfTiMxClose(ttUserInterface interfaceHandle)
{
/* Normally we'd shut off the part, but here we'll only stop it. This is
   because it can't handle a soft reset. */
    gmStop();

/* 
 * Free memory from pre-allocated buffers in the receive ring 
 */
    (void)tfPoolDelete(interfaceHandle);

    HWI_disable();
    tlTiMxInterfaceHandle = 0;
    HWI_enable();
#ifdef TM_USE_SHEAP
    tfSheapFree(tlTxFrameBuf[0]);
    tfSheapFree(tlTxFrameBuf[1]);
#else
    free(tlTxFrameBuf[0]);
    free(tlTxFrameBuf[1]);
#endif
    return TM_DEV_OKAY;
}

/* 
 * tfTiMxIoctl()
 *
 *   This I/O control routine performs two functions:
 *
 *   1.  Setup the list of multicast addresses.
 *   2.  Set the SMC to recieve all multicast packets.  This is done by writing
 *       all 1's to the SMC's four multicast hash registers.
 *
 */
int tfTiMxIoctl(ttUserInterface interfaceHandle,
               int             flag,
               void TM_FAR *   optionPtr,
               int             optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    tt32Bit control;

/*
 * Setup our list of multicast addresses to receive.
 */
    if (flag & TM_DEV_SET_MCAST_LIST)
    {
/* TBD: implement multicast list */
/* Setup the Rx Filter to receive all broadcast and multicast packets */
        control = READ32( TM_GMAC_CONTROL );
        control |= TM_GM_AB | TM_GM_PM;
        WRITE32( TM_GMAC_CONTROL, control );
    }

/*
 * Set the SMC to receive all multicast packets.
 */
    if (flag & TM_DEV_SET_ALL_MCAST)
    {
/* Setup the Rx Filter to receive all broadcast and multicast packets */
        control = READ32( TM_GMAC_CONTROL );
        control |= TM_GM_AB | TM_GM_PM;
        WRITE32( TM_GMAC_CONTROL, control );
    }

#endif /* (TM_IGMP or TM_6_USE_MLD) */

    return TM_DEV_OKAY;
}

/****************************************************************************
* FUNCTION: tfTiMxGetPhysAddr
*
* PURPOSE: Read the MAC address from the GMAC.
*
* PARAMETERS:
*   interfaceHandle
*   physicalAddress: output parameter used to store the MAC address
*
* RETURNS:
*   TM_DEV_OKAY: successful
*
* NOTES:
*
****************************************************************************/
int tfTiMxGetPhysAddr(
    ttUserInterface interfaceHandle,
    char TM_FAR *   physicalAddress)
{
    tt32Bit tmp;

/* Read our MAC address */
    tmp = READ32( TM_GMAC_ADDR1 );
    physicalAddress[0] = (tt8Bit) (tmp & 0xFF);
    physicalAddress[1] = (tt8Bit) ((tmp >> 8) & 0xFF);
    physicalAddress[2] = (tt8Bit) ((tmp >> 16) & 0xFF);
    physicalAddress[3] = (tt8Bit) ((tmp >> 24) & 0xFF);
    tmp = READ32( TM_GMAC_ADDR2 );
    physicalAddress[4] = (tt8Bit) (tmp & 0xFF);
    physicalAddress[5] = (tt8Bit) ((tmp >> 8) & 0xFF);

    return TM_DEV_OKAY;
}

/*
 * tfTiMxReceive()
 *
 *   This routine is called by the Treck stack to receive a packet.  Since
 *   the data has already been copied inside of the ISR, all we need to do
 *   in this routine is to pass the buffer up to the Treck stack.  
 */
int tfTiMxReceive(ttUserInterface        interfaceHandle,
                  char TM_FAR * TM_FAR * dataPtrPtr,
                  int  TM_FAR *          dataLengthPtr,
                  ttUserBufferPtr        bufHandlePtr)
{
    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          bufHandlePtr );
}

/*
 * tfTiMxSend()
 *
 * This function is called by the Treck stack to send a packet.
 *
 * NOTE: We only use the TX direct FIFO mode in this driver. This has the
 * benefit of elminating the TX local DMA completely, which leaves the packet
 * memory's total bandwidth to the RX local DMA and the host. Therefore,
 * receiving at the full line speed is more achievable this way. The
 * disadvantage is that we will get lower throughput with transmits. But,
 * this GMAC appears to be buggy, and we want to avoid corruption of the RX
 * FIFO which is even worse (it causes us to have to reset the RX FIFO); in
 * TX direct FIFO mode, corruption of the RX FIFO is not observed to occur
 * provided we wait for the TX complete (or TX error) interrupt before sending
 * the next packet.
 */
int tfTiMxSend(ttUserInterface interfaceHandle,
               char TM_FAR *   dataCharPtr,
               int             dataLength,
               int             flag)
{
    int         errorCode;
    tt32Bit     tmp;
    tt16BitPtr  data16Ptr;
    int         frameBufIndex;
    tt32Bit     *frameBufPtr;

/* Flags not used since we don't support scattered sends */
    flag=flag;
    tm_assert(tfTiMxSend, (flag & TM_USER_BUFFER_LAST));

    errorCode = TM_DEV_OKAY;
    data16Ptr = (tt16BitPtr) dataCharPtr;

    if( tlFatalError )
    {
/* wait until the transmitter is available */
        tlTxBusy++;
        errorCode = TM_DEV_ERROR;
    }
    else
    {
        HWI_disable();
/* disable interrupts, because we use tlTxQueuedFrameLen to signal gmIsr that
   there is a queued packet available to transmit */
        if( !tlTxFree )
        {
            if (!tlTxQueuedFrameLen[tlNextSendFrameBufIndex])
            {
                tlTxQueuedFrameLen[tlNextSendFrameBufIndex] = dataLength;
                frameBufPtr = tlTxFrameBuf[tlNextSendFrameBufIndex];

                if (tlNextSendFrameBufIndex == 0)
                {
                    tlNextSendFrameBufIndex = 1;
                }
                else
                {
                    tlNextSendFrameBufIndex = 0;
                }
            }
            else
            {
/* frames are already queued for the ISR to transmit */
                tlTxBusy++;
                errorCode = TM_DEV_ERROR;
            }
            
            if (errorCode == TM_DEV_OKAY)
            {
/* prepare the packet for the ISR to transmit */
                frameBufIndex = 0;

/*
 * On the TI C5x platform, the Ethernet addresses in the Ethernet header in
 * the packet from the chip is *not* packed; that is, the first BYTE of the
 * Ethernet header is in the first 16-bit WORD of the data buffer, the second
 * BYTE is in the second WORD, and so on, for the first 12 bytes (words) of
 * the Ethernet header.  So we must deal with this part of the data buffer
 * specially and write it out differently than the rest of the packet, which
 * is packed into words. Note that we add a dummy WORD at the start of the
 * structure ttEtherHeader so that the structure length (in WORDs) is 14.
 */

/* The GMAC can only be accessed as a 32-bit device by the TI C5x DSP */
/* Skip dummy word in ttEtherHeader */
                data16Ptr++;

/* the source and destination Ethernet addresses are unpacked */
/* Write out Destination Address */
/* Note that the byte-order is reversed ("SWIZZLE'd"). */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                frameBufPtr[frameBufIndex++] = tmp; 

/* Write out last 2 bytes of Destination Address, plus the first 2 bytes of
   the Source Address SWIZZLE'd */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                frameBufPtr[frameBufIndex++] = tmp; 

/* Write out the rest of the Source Address SWIZZLE'd */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                frameBufPtr[frameBufIndex++] = tmp; 

/* Adjust for the number of bytes we've already written */
                dataLength -= 12;

/* Write out the rest of the packet data (packed) */
                dataLength = (dataLength+3)>>2;
                while( dataLength-- )
                {
                    tmp = ((tt32Bit) *data16Ptr++) << 16;
                    tmp = ((tt32Bit) *data16Ptr++) | tmp;
                    frameBufPtr[frameBufIndex++] = SWIZZLE(tmp); 
                }
            }
            HWI_enable();

            if (errorCode == TM_DEV_OKAY)
            {
/* We're done with this packet, so inform the Treck stack */
                tfSendCompleteInterface(
                    interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
            }
        }
        else
        {
            HWI_enable();

/* Set the frame size */
            tmp = READ32(TM_GMAC_AUX2)&0xffff | ((tt32Bit)dataLength << 16);
            WRITE32( TM_GMAC_AUX2, tmp );

/*
 * On the TI C5x platform, the Ethernet addresses in the Ethernet header in
 * the packet from the chip is *not* packed; that is, the first BYTE of the
 * Ethernet header is in the first 16-bit WORD of the data buffer, the second
 * BYTE is in the second WORD, and so on, for the first 12 bytes (words) of
 * the Ethernet header.  So we must deal with this part of the data buffer
 * specially and write it out differently than the rest of the packet, which
 * is packed into words. Note that we add a dummy WORD at the start of the
 * structure ttEtherHeader so that the structure length (in WORDs) is 14.
 */

/* The GMAC can only be accessed as a 32-bit device by the TI C5x DSP */
/* Skip dummy word in ttEtherHeader */
            data16Ptr++;

/* the source and destination Ethernet addresses are unpacked */
/* Write out Destination Address */
/* Note that the byte-order is reversed ("SWIZZLE'd"). */
            tmp = ((tt32Bit) *data16Ptr++) & 0xff;
            tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
            WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Write out last 2 bytes of Destination Address, plus the first 2 bytes of
   the Source Address SWIZZLE'd */
            tmp = ((tt32Bit) *data16Ptr++) & 0xff;
            tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
            WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Write out the rest of the Source Address SWIZZLE'd */
            tmp = ((tt32Bit) *data16Ptr++) & 0xff;
            tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
            tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
            WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Adjust for the number of bytes we've already written */
            dataLength -= 12;

/* Write out the rest of the packet data (packed) */
            dataLength = (dataLength+3)>>2;
            while( dataLength-- )
            {
                tmp = ((tt32Bit) *data16Ptr++) << 16;
                tmp = ((tt32Bit) *data16Ptr++) | tmp;
                WRITE32( TM_GMAC_WRTXFIFOD, SWIZZLE(tmp) );
            }

            HWI_disable();
/* Send the frame */
            TxState = STATE_SEND;
            tlTxFree = 0;

            tmp = READ32(TM_GMAC_CONTROL)|TM_GM_ST1;
            WRITE32( TM_GMAC_CONTROL, tmp );
            HWI_enable();

/* We're done with this packet, so inform the Treck stack */
            tfSendCompleteInterface(
                interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
        }
    }

    return errorCode;
}

#ifdef TM_USE_DRV_ONE_SCAT_SEND
/*
 * tfTiMxScatSend
 * Scatter send function.
 *
 * NOTE: We only use the TX direct FIFO mode in this driver. This has the
 * benefit of elminating the TX local DMA completely, which leaves the packet
 * memory's total bandwidth to the RX local DMA and the host. Therefore,
 * receiving at the full line speed is more achievable this way. The
 * disadvantage is that we will get lower throughput with transmits. But,
 * this GMAC appears to be buggy, and we want to avoid corruption of the RX
 * FIFO which is even worse (it causes us to have to reset the RX FIFO); in
 * TX direct FIFO mode, corruption of the RX FIFO is not observed to occur
 * provided we wait for the TX complete (or TX error) interrupt before sending
 * the next packet.
 */
int tfTiMxScatSend(ttUserInterface interfaceHandle,
                   ttUserPacketPtr  uPacketPtr )
{
    int         errorCode;
    tt32Bit     tmp;
    tt16BitPtr  data16Ptr;
    int         dataLength;
    int         scatterIndex;
    int         firstScatDataLen;
    int         scatDataLen;
    int         frameBufIndex;
    tt32Bit     *frameBufPtr;

    scatterIndex = uPacketPtr->pktuLinkExtraCount;

    if (scatterIndex == 0)
    {
/* if this is not a scattered send, then call the regular driver send */
        errorCode = tfTiMxSend(
            interfaceHandle,
            (char TM_FAR *) uPacketPtr->pktuLinkDataPtr,
            uPacketPtr->pktuChainDataLength, TM_USER_BUFFER_LAST); 
    }
    else
    {
/* NOTE: we only support a buffersPerFrameCount of 2 */
        tm_assert(tfTiMxScatSend, (scatterIndex == 1));
        errorCode = TM_DEV_OKAY;
        data16Ptr = (tt16BitPtr) uPacketPtr->pktuLinkDataPtr;
        scatDataLen = uPacketPtr->pktuLinkDataLength;
        firstScatDataLen = scatDataLen;
        dataLength = uPacketPtr->pktuChainDataLength;

        if( tlFatalError )
        {
/* wait until the transmitter is available */
            tlTxBusy++;
            errorCode = TM_DEV_ERROR;
        }
        else
        {
            HWI_disable();

            if( !tlTxFree )
            {
/* disable interrupts, because we use tlTxQueuedFrameLen to signal gmIsr that
   there is a queued packet available to transmit */
                if (!tlTxQueuedFrameLen[tlNextSendFrameBufIndex])
                {
                    tlTxQueuedFrameLen[tlNextSendFrameBufIndex] = dataLength;
                    frameBufPtr = tlTxFrameBuf[tlNextSendFrameBufIndex];

                    if (tlNextSendFrameBufIndex == 0)
                    {
                        tlNextSendFrameBufIndex = 1;
                    }
                    else
                    {
                        tlNextSendFrameBufIndex = 0;
                    }
                }
                else
                {
/* frames are already queued for the ISR to transmit */
                    tlTxBusy++;
                    errorCode = TM_DEV_ERROR;
                }
            
                if (errorCode == TM_DEV_OKAY)
                {
/* prepare the packet for the ISR to transmit */
                    frameBufIndex = 0;

/*
 * On the TI C5x platform, the Ethernet addresses in the Ethernet header in
 * the packet from the chip is *not* packed; that is, the first BYTE of the
 * Ethernet header is in the first 16-bit WORD of the data buffer, the second
 * BYTE is in the second WORD, and so on, for the first 12 bytes (words) of
 * the Ethernet header.  So we must deal with this part of the data buffer
 * specially and write it out differently than the rest of the packet, which
 * is packed into words. Note that we add a dummy WORD at the start of the
 * structure ttEtherHeader so that the structure length (in WORDs) is 14.
 */

/* The GMAC can only be accessed as a 32-bit device by the TI C5x DSP */
/* Skip dummy word in ttEtherHeader */
                    data16Ptr++;

/* the source and destination Ethernet addresses are unpacked */
/* Write out Destination Address */
/* Note that the byte-order is reversed ("SWIZZLE'd"). */
                    tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                    tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 16))
                           & 0x00ff0000) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 24))
                           & 0xff000000) | tmp;
                    frameBufPtr[frameBufIndex++] = tmp; 

/* Write out last 2 bytes of Destination Address, plus the first 2 bytes of
   the Source Address SWIZZLE'd */
                    tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                    tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 16))
                           & 0x00ff0000) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 24))
                           & 0xff000000) | tmp;
                    frameBufPtr[frameBufIndex++] = tmp; 

/* Write out the rest of the Source Address SWIZZLE'd */
                    tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                    tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 16))
                           & 0x00ff0000) | tmp;
                    tmp = (((((tt32Bit) *data16Ptr++) << 24))
                           & 0xff000000) | tmp;
                    frameBufPtr[frameBufIndex++] = tmp; 

/* Adjust for the number of bytes we've already written */
                    scatDataLen -= 12;

/* Write out the rest of the packet data (packed).
   NOTE: for scattered-send, the first piece will always have an even
   number of bytes and consists of the Ethernet, IP and TCP headers. */
                    tm_assert(tfTiMxScatSend, (!(scatDataLen & 1)));
                    scatDataLen = scatDataLen>>2;
                    while( scatDataLen-- )
                    {
                        tmp = ((tt32Bit) *data16Ptr++) << 16;
                        tmp = ((tt32Bit) *data16Ptr++) | tmp;
                        frameBufPtr[frameBufIndex++] = SWIZZLE(tmp); 
                    }
                    if (firstScatDataLen & 2)
                    {
/* write the final odd 16-bit word of the headers, if there is one */
                        tmp = ((tt32Bit) *data16Ptr++) << 16;
                        uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 
                        data16Ptr = (tt16BitPtr) uPacketPtr->pktuLinkDataPtr;
                        scatDataLen = uPacketPtr->pktuLinkDataLength + 2;

                        scatDataLen = (scatDataLen+3)>>2;
                        while( scatDataLen-- )
                        {
/* write the rest of the packet data */
                            tmp = ((tt32Bit) *data16Ptr++) | tmp;
                            frameBufPtr[frameBufIndex++] = SWIZZLE(tmp); 
                            tmp = ((tt32Bit) *data16Ptr++) << 16;
                        }
                    }
                    else
                    {
                        uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 
                        data16Ptr = (tt16BitPtr) uPacketPtr->pktuLinkDataPtr;
                        scatDataLen = uPacketPtr->pktuLinkDataLength;

                        scatDataLen = (scatDataLen+3)>>2;
                        while( scatDataLen-- )
                        {
/* write the rest of the packet data */
                            tmp = ((tt32Bit) *data16Ptr++) << 16;
                            tmp = ((tt32Bit) *data16Ptr++) | tmp;
                            frameBufPtr[frameBufIndex++] = SWIZZLE(tmp); 
                        }
                    }
                }
                HWI_enable();

                if (errorCode == TM_DEV_OKAY)
                {
/* We're done with this packet, so inform the Treck stack */
                    tfSendCompleteInterface(
                        interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
                }
            }
            else
            {
                HWI_enable();

/* GMAC transmitter is available */
/* Set the frame size */
                tmp = READ32(TM_GMAC_AUX2)
                    &0xffff | ((tt32Bit)dataLength << 16);
                WRITE32( TM_GMAC_AUX2, tmp );

/*
 * On the TI C5x platform, the Ethernet addresses in the Ethernet header in
 * the packet from the chip is *not* packed; that is, the first BYTE of the
 * Ethernet header is in the first 16-bit WORD of the data buffer, the second
 * BYTE is in the second WORD, and so on, for the first 12 bytes (words) of
 * the Ethernet header.  So we must deal with this part of the data buffer
 * specially and write it out differently than the rest of the packet, which
 * is packed into words. Note that we add a dummy WORD at the start of the
 * structure ttEtherHeader so that the structure length (in WORDs) is 14.
 */

/* The GMAC can only be accessed as a 32-bit device by the TI C5x DSP */
/* Skip dummy word in ttEtherHeader */
                data16Ptr++;

/* the source and destination Ethernet addresses are unpacked */
/* Write out Destination Address */
/* Note that the byte-order is reversed ("SWIZZLE'd"). */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Write out last 2 bytes of Destination Address, plus the first 2 bytes of
   the Source Address SWIZZLE'd */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Write out the rest of the Source Address SWIZZLE'd */
                tmp = ((tt32Bit) *data16Ptr++) & 0xff;
                tmp = ((((tt32Bit) *data16Ptr++) << 8) & 0x0000ff00) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 16)) & 0x00ff0000) | tmp;
                tmp = (((((tt32Bit) *data16Ptr++) << 24)) & 0xff000000) | tmp;
                WRITE32( TM_GMAC_WRTXFIFOD, tmp);

/* Adjust for the number of bytes we've already written */
                scatDataLen -= 12;

/* Write out the rest of the packet data (packed).
   NOTE: for scattered-send, the first piece will always have an even
   number of bytes and consists of the Ethernet, IP and TCP headers. */
                tm_assert(tfTiMxScatSend, (!(scatDataLen & 1)));
                scatDataLen = scatDataLen>>2;
                while( scatDataLen-- )
                {
                    tmp = ((tt32Bit) *data16Ptr++) << 16;
                    tmp = ((tt32Bit) *data16Ptr++) | tmp;
                    WRITE32( TM_GMAC_WRTXFIFOD, SWIZZLE(tmp) );
                }
                if (firstScatDataLen & 2)
                {
/* write the final odd 16-bit word of the headers, if there is one */
                    tmp = ((tt32Bit) *data16Ptr++) << 16;
                    uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 
                    data16Ptr = (tt16BitPtr) uPacketPtr->pktuLinkDataPtr;
                    scatDataLen = uPacketPtr->pktuLinkDataLength + 2;

                    scatDataLen = (scatDataLen+3)>>2;
                    while( scatDataLen-- )
                    {
/* write the rest of the packet data */
                        tmp = ((tt32Bit) *data16Ptr++) | tmp;
                        WRITE32( TM_GMAC_WRTXFIFOD, SWIZZLE(tmp) );
                        tmp = ((tt32Bit) *data16Ptr++) << 16;
                    }
                }
                else
                {
                    uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 
                    data16Ptr = (tt16BitPtr) uPacketPtr->pktuLinkDataPtr;
                    scatDataLen = uPacketPtr->pktuLinkDataLength;

                    scatDataLen = (scatDataLen+3)>>2;
                    while( scatDataLen-- )
                    {
/* write the rest of the packet data */
                        tmp = ((tt32Bit) *data16Ptr++) << 16;
                        tmp = ((tt32Bit) *data16Ptr++) | tmp;
                        WRITE32( TM_GMAC_WRTXFIFOD, SWIZZLE(tmp) );
                    }
                }

                HWI_disable();
/* Send the frame */
                TxState = STATE_SEND;
                tlTxFree = 0;

                tmp = READ32(TM_GMAC_CONTROL)|TM_GM_ST1;
                WRITE32( TM_GMAC_CONTROL, tmp );
                HWI_enable();

/* We're done with this packet, so inform the Treck stack */
                tfSendCompleteInterface(
                    interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
            }
        }
    }

    return errorCode;
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

/* Reset GMAC and clear its external memory */
static void gmInitialize()
{
    tt32Bit  i;
    tt32Bit  bytesWritten;
    tt32Bit  page;

/* Hard Reset */
    WRITE32( TM_GMAC_RESET_ON, 0 );
    Delay(4);
    WRITE32( TM_GMAC_RESET_OFF, 0 );
    Delay(4);

/* Soft Reset */
    WRITE32( TM_GMAC_CONTROL, TM_GM_RESET );
    Delay(3);
    WRITE32( TM_GMAC_CONTROL, 0 );

/* Set TLBP and TWP */
    WRITE32( TM_GMAC_TXRING1, 0x0 );

/* Set TRP */
    WRITE32( TM_GMAC_TXRING2, 0x0 );

/* initialize all 128K bytes of packet memory:
   it seems like we can only write to the first 32-bit word on each page so
   this code isn't really doing what it should */
    bytesWritten = 0;
    page = 0;
    for( i=0; i<32768; i++ )
    {
        WRITE32( TM_GMAC_TWD, 0 );
        bytesWritten += 4;
        if( !(bytesWritten & 0xFF) )
        {
            page += 0x10000;
            page &= 0xfff0000;
            WRITE32( TM_GMAC_TXRING1, page );
        }
    }

/* Set auto-recover on Rx DMA */
    WRITE32( TM_GMAC_AUX3, TM_GM_AUTORCVR );

/* Set RRP and RHBP
   NOTE: we use all of the 128K bytes of onboard packet memory for receive */
    WRITE32( TM_GMAC_RXRING2, 0x01FF0000 );

/* Set RWP */
    WRITE32( TM_GMAC_RXRING1, 0x00000000 );

/* Enable Autoneg, linktest, 16 bit mode, and TX FIFO count registers */
/* reject received packets with errors */
/* reduce RX FIFO threshold to try to avoid overrun */
    WRITE32(
        TM_GMAC_CONFIG,
        TM_GM_LTE | TM_GM_ANE | TM_GM_PBW | TM_GM_TXFIFOCNTEN | TM_GM_ARXERRB
        | TM_GM_RTHD1);

/* Clear interrupt mask and set BP */
    WRITE32( TM_GMAC_ISR, 0x00000000 );

/* Reset the FIFOs */
#ifdef TM_GMAC_HALF_DUPLEX_TX 
/* NOTE: we auto-negotiate half-duplex mode, because (according to TI) this
   GMAC has interoperability problems with certain hubs in full-duplex mode */
    WRITE32( TM_GMAC_AUX2, 0xe800 );
    Delay(2);
    WRITE32( TM_GMAC_AUX2, 0x2800 );
#else /* !TM_GMAC_HALF_DUPLEX_TX */
    WRITE32( TM_GMAC_AUX2, 0xFC00 );
    Delay(2);
    WRITE32( TM_GMAC_AUX2, 0x3C00 );
#endif /* !TM_GMAC_HALF_DUPLEX_TX */

/* Enable interrupts (and set BP) */
    WRITE32( TM_GMAC_ISR, 0x00000000 | TM_INTMASK );

/* Tansmitter is not free until we start */
    tlTxFree = 0;

/* Reset any pending error */
    tlFatalError = 0;
}

/* Start GMAC Rx and Tx state machines */
static void gmStart()
{
    tt32Bit control;

/* Initialize states */
    TxState = STATE_IDLE;
    tlTxFree = 1;

/* Start receive */
    control = READ32( TM_GMAC_CONTROL );
    control |= TM_GM_SR;
#if LCONF_TEST
    control |= TM_GM_LB0;
#endif
/* Setup the Rx Filter to receive all broadcast and multicast packets */
    control |= TM_GM_AB | TM_GM_PM;
    WRITE32( TM_GMAC_CONTROL, control );
}

/* Stop GMAC, abort current Rx and Tx */
static void gmStop()
{
/* Disable board interrupt */
    (void)C55_disableIMR1( TM_IFLAG );

/* Stop anyone from entering device send */
    tlTxFree = 0;

/* Let tasks fall out of TX or RX */
    tlFatalError = 1;
    TSK_sleep( 100 );

/* Stop the device */
    WRITE32( TM_GMAC_CONTROL, 0 );
    TSK_sleep( 100 );

/* Reset the part */
/* Hard Reset */
    WRITE32( TM_GMAC_RESET_ON, 0 );
    Delay(4);
    WRITE32( TM_GMAC_RESET_OFF, 0 );
    Delay(4);

/* Soft Reset */
    WRITE32( TM_GMAC_CONTROL, TM_GM_RESET );
    Delay(3);
    WRITE32( TM_GMAC_CONTROL, 0 );
}

/* kludge for buggy GMAC */
static int gmResetRxFifo()
{
    tt32Bit prevControl; /* previous value of GMAC_CONTROL */

/* stop receive */
    prevControl = READ32( TM_GMAC_CONTROL );
    WRITE32( TM_GMAC_CONTROL, (prevControl & ~TM_GM_SR));

/* reset the RX FIFO in this buggy GMAC */
#ifdef TM_GMAC_HALF_DUPLEX_TX 
/* NOTE: we auto-negotiate half-duplex mode, because (according to TI) this
   GMAC has interoperability problems with certain hubs in full-duplex mode */
    WRITE32( TM_GMAC_AUX2, 0xa800 );
#else /* !TM_GMAC_HALF_DUPLEX_TX */
    WRITE32( TM_GMAC_AUX2, 0xBC00 );
#endif /* !TM_GMAC_HALF_DUPLEX_TX */
    tlResetRxFifo++;

    Delay(2);

/* reenable the RX FIFO */
#ifdef TM_GMAC_HALF_DUPLEX_TX 
/* NOTE: we auto-negotiate half-duplex mode, because (according to TI) this
   GMAC has interoperability problems with certain hubs in full-duplex mode */
    WRITE32( TM_GMAC_AUX2, 0x2800 );
#else /* !TM_GMAC_HALF_DUPLEX_TX */
    WRITE32( TM_GMAC_AUX2, 0x3C00 );
#endif /* !TM_GMAC_HALF_DUPLEX_TX */

/* receive any pending packets (done in gmIsr) */
/* later when we exit gmIsr we will restart receive */

    return prevControl;
}

/* General purpose ISR function. Here we will signal the scheduler that a
   packet is available on GMAC. */
void gmIsr()
{
    tt32Bit flags;
    tt32Bit data1, data2, data3, data4;
    tt32Bit status;
    tt16Bit *data16Ptr;
    int     length;
    int     tlen;
    int     rxPacketCount;
    tt32Bit prevControl;
    int     frameBufDataIndex;
    tt32Bit *frameBufPtr;
    int     queuedFrameLen;

    HWI_disable(); /* very important for GMAC stability */

    rxPacketCount = 0;
    prevControl = 0;
    
/* read the interrupt flags */
    flags = READ32( TM_GMAC_ISR ) & 0x0000FF00;

    if (flags & TM_GM_RI)
    {
/* we have a receive packet interrupt, wait for when the packet data is
   ready to be read */
        data1 = 0;
        while(READ32( TM_GMAC_AUX1 ) & TM_GM_DREQB)
        {
/* NOTE: this GMAC has a problem where the GM_DREQB signal will stay high.
   Sometimes we can work-around this issue by resetting the RX FIFO, but this
   does not always work. */
            if( ++data1 == 1000 )
            {
/* reset the RX FIFO in this buggy GMAC */
                prevControl = gmResetRxFifo();
                break;
            }
        }
    }

    while (flags & (TM_GM_RI | TM_GM_TI | TM_GM_TEI))
    {
/* ACK the interrupt */
        WRITE32( TM_GMAC_ISR, ( 0x00000000 | flags | TM_INTMASK ) );

/* the GMAC signals interrupts to the DSP in a level sensitive fashion, while
   the DSP only supports edge-triggered interrupts. If the GMAC interrupt
   signal were to stay low in a back-to-back interrupt, the DSP may become out
   of synch and unable to detect additional interrupts. To avoid losing an
   interrupt, we loop here while the GMAC tells us there is an interrupt
   pending */

/* check for Tx Complete or Tx Error */
        if (flags & (TM_GM_TI | TM_GM_TEI))
        {
            if (flags & TM_GM_TEI)
            {
/* record that a transmit error occurred */
                tlTxBad++;
            }
            else
            {
                tlTxGood++;
            }
            queuedFrameLen = tlTxQueuedFrameLen[tlIsrNextSendFrameBufIndex];
            if (queuedFrameLen)
            {
                frameBufPtr = tlTxFrameBuf[tlIsrNextSendFrameBufIndex];
/* give this frame buffer back to tfTiMxSend, once interrupts are enabled */
                tlTxQueuedFrameLen[tlIsrNextSendFrameBufIndex] = 0;

                if (tlIsrNextSendFrameBufIndex == 0)
                {
                    tlIsrNextSendFrameBufIndex = 1;
                }
                else
                {
                    tlIsrNextSendFrameBufIndex = 0;
                }

/* a frame is queued for us to transmit, it has already been byte-swapped */
/* write the frame size to the GMAC */
                data1 = READ32(TM_GMAC_AUX2)&0xffff
                    | ((tt32Bit)queuedFrameLen << 16);
                WRITE32( TM_GMAC_AUX2, data1);

                frameBufDataIndex = 0;
/* convert length in bytes to length in 32-bit longs */
                queuedFrameLen = (queuedFrameLen+3)>>2;
                while(queuedFrameLen--)
                {
/* write out the frame data to the GMAC */
                    data1 = frameBufPtr[frameBufDataIndex++];
                    WRITE32( TM_GMAC_WRTXFIFOD, data1);
                }

/* initiate the GMAC send */
                data1 = READ32(TM_GMAC_CONTROL)|TM_GM_ST1;
                WRITE32( TM_GMAC_CONTROL, data1);
                tlIsrTx++;
            }
            else
            {
/* NOTE: updating tlTxFree here *AND ONLY HERE* based on the transmit complete
   interrupt helps overall stability of the GMAC */
                TxState = STATE_IDLE;
                tlTxFree = 1;
            }
        }

/* check for received packets */
        data1 = READ32(TM_GMAC_RXRING1) >> 16;
        data2 = READ32(TM_GMAC_RXRING2) & 0xFFFF;
        while ((data1 != data2) && !(READ32( TM_GMAC_AUX1 ) & TM_GM_DREQB ))
        {
/* Rx a packet */

/* Read the first 16 bytes */
            data1 = READ32( TM_GMAC_RRD );
            data2 = READ32( TM_GMAC_RRD );
            data3 = READ32( TM_GMAC_RRD );
            data4 = READ32( TM_GMAC_RRD );

/* Decode length, nextpage, and status */
            length   = (int)((data1 >> 12) & 0x0FFF);
            status   = (int)((data1 >> 24) & 0x0FF);

            if(   (length < TM_ETHER_HEADER_SIZE)
               || (length > 1518) )
            {
/* reset the RX FIFO in this buggy GMAC */
                prevControl = gmResetRxFifo();
                goto checkForInt;
            }

            data16Ptr = (unsigned short TM_FAR *) 0;

            if( !(status & 0x80) )
            {
                tlRxGood++;
                
                if ((tlFatalError == 0) && tlTiMxInterfaceHandle)
                {
/* Try and get a buffer. If we can't, toss the packet */
                    data16Ptr = (tt16BitPtr)
                        tfPoolIsrGetBuffer(tlTiMxInterfaceHandle,
                                           length + TM_GMAC_EXTRA_BYTES);
                }
            }
            else
            {
                tlRxBad++;
            }

/* Pull the packet out - good or bad */
            tlen  = length - 12;
            if( !data16Ptr )
            {
/* Get the rest of the packet and throw it away */
                while( tlen > 0 )
                {
                    data1 = 0;
                    while(READ32( TM_GMAC_AUX1 ) & TM_GM_DREQB)
                    {
                        if( ++data1 == 1000 )
                        {
/* odd - we can't read the rest of this packet from the GMAC */
/* reset the RX FIFO in this buggy GMAC */
                            prevControl = gmResetRxFifo();
                            goto checkForInt;
                        }
                    }
/* When DREQB is deasserted, there is packet data available for us to read */

                    data1 = READ32( TM_GMAC_RRD );
                    data2 = READ32( TM_GMAC_RRD );
                    data3 = READ32( TM_GMAC_RRD );
                    data4 = READ32( TM_GMAC_RRD );
                    tlen -= 16;
                }
            }
            else
            {
/*
 * On the TI C5x platform, the Ethernet addresses in the Ethernet header in
 * the packet from the chip is *not* packed; that is, the first BYTE of the
 * Ethernet header is in the first 16-bit WORD of the data buffer, the second
 * BYTE is in the second WORD, and so on, for the first 12 bytes (words) of
 * the Ethernet header.  So we must deal with this part of the data buffer
 * specially and write it differently than the rest of the packet, which
 * is packed into words. Note that we add a dummy WORD at the start of the
 * structure ttEtherHeader so that the structure length (in WORDs) is 14.
 */

/* Skip dummy word in ttEtherHeader */
                data16Ptr++;

/* Copy the Ethernet Destination Address and Source Address as unpacked and
   SWIZZLE'd */
                *data16Ptr++ = (tt16Bit) (data2 & 0xff);
                *data16Ptr++ = (tt16Bit) ((data2 >> 8) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data2 >> 16) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data2 >> 24) & 0xff);

                *data16Ptr++ = (tt16Bit) (data3 & 0xff);
                *data16Ptr++ = (tt16Bit) ((data3 >> 8) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data3 >> 16) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data3 >> 24) & 0xff);

                *data16Ptr++ = (tt16Bit) (data4 & 0xff);
                *data16Ptr++ = (tt16Bit) ((data4 >> 8) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data4 >> 16) & 0xff);
                *data16Ptr++ = (tt16Bit) ((data4 >> 24) & 0xff);

/* Get the rest of the packet */
                while( tlen > 0 )
                {
                    data1 = 0;
                    while(READ32( TM_GMAC_AUX1 ) & TM_GM_DREQB)
                    {
                        if( ++data1 == 1000 )
                        {
/* odd - we can't read the rest of this packet from the GMAC */
/* reset the RX FIFO in this buggy GMAC */
                            prevControl = gmResetRxFifo();
/* give back the pool receive buffer */
                            tfPoolIsrFreeBuffer(tlTiMxInterfaceHandle,
                                            length + TM_GMAC_EXTRA_BYTES);
                            goto checkForInt;
                        }
                    }
/* When DREQB is deasserted, there is packet data available for us to read */

                    data1 = READ32( TM_GMAC_RRD );
                    data2 = READ32( TM_GMAC_RRD );
                    data3 = READ32( TM_GMAC_RRD );
                    data4 = READ32( TM_GMAC_RRD );
                        
/* we need to swizzle the data read in */
                    *data16Ptr++ = (tt16Bit) ((data1 << 8) & 0x0000ff00)
                        | (tt16Bit) ((data1 >> 8) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data1 >> 8) & 0x0000ff00)
                        | (tt16Bit) ((data1 >> 24) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data2 << 8) & 0x0000ff00)
                        | (tt16Bit) ((data2 >> 8) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data2 >> 8) & 0x0000ff00)
                        | (tt16Bit) ((data2 >> 24) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data3 << 8) & 0x0000ff00)
                        | (tt16Bit) ((data3 >> 8) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data3 >> 8) & 0x0000ff00)
                        | (tt16Bit) ((data3 >> 24) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data4 << 8) & 0x0000ff00)
                        | (tt16Bit) ((data4 >> 8) & 0x000000ff);
                    *data16Ptr++ = (tt16Bit) ((data4 >> 8) & 0x0000ff00)
                        | (tt16Bit) ((data4 >> 24) & 0x000000ff);

                    tlen -= 16;
                }

                rxPacketCount++;
            }

/* check if we can receive another packet */
            data1 = READ32(TM_GMAC_RXRING1) >> 16;
            data2 = READ32(TM_GMAC_RXRING2) & 0xFFFF;
        }

checkForInt:
/* check for another interrupt */
        flags = READ32( TM_GMAC_ISR ) & 0x0000FF00;
    }

    if (prevControl)
    {
/* restart receive */
        prevControl |= TM_GM_SR;
        WRITE32( TM_GMAC_CONTROL, prevControl);
    }

    if (rxPacketCount)
    {
/* Notify the stack for each packet that we received. */
        tfNotifyInterfaceIsr(
            tlTiMxInterfaceHandle, rxPacketCount, 0, 0UL, 0UL);
    }
    return;
}

/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trtic5mx.c    File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 22796    Output checksum: 56028
 */
