/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Driver for NE2000 and DP8390 Ethernet chips
 *
 * Filename: trne2000.c
 * Author: Jared
 * Date Created: 12/09/1999
 * $Source: source/drivers/ne2000/trne2000.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:15CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * This file is a heavily modified, combined version of ne.c and 8390.c.
 * Written 1992-94 by Donald Becker.
 *
 * Copyright 1993 United States Government as represented by the
 * Director, National Security Agency.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 */

/*
 * All throughout this code, a naming convention has been employed in
 * order to make it as painless as possible to use this driver for the
 * 8390 in other implementations besides the NE2000. Any hardware settings
 * that are specific to the NE2K have the prefix ne2k, and any that are
 * universal to all 8390's have the prefix 8390. Fairly self explanatory.
 */


/*
 * Includes
 */

/* Treck Specific Includes */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/* NE2000 specific includes and defines */
#include <trne2000.h>

/* Needed for Borland BCC 5.0. */
#include <conio.h>
#include <dos.h>


/*
 * Pre-Processor Definitions (may have to be modified)
 */


/*
 * Number of transmit and receive pages to use. We have a total of 0x40 pages
 * to play with. Transmit and receive each require a minimum of 6 pages.
 * RX is included here just to make understanding easier - it is not
 * actually used by the code. The code just looks at how many buffers there
 * are, and how many of them are used for transmit. All others are assumed
 * to be receive.
 */
#define TM_NE2K_TX_PAGES 0x20 /* 8 kbytes (32 * 256 bytes) */
#define TM_NE2K_RX_PAGES 0x20 /* 8 kbytes (32 * 256 bytes) */

/* Base I/O Port */
#define TM_NE2K_BASE_ADDRESS   0x240
/* Our interrupt */
#define TM_NE2K_INT            0x0d
/* Address of the PIC */
#define TM_NE2K_PIC_ADDR       0x21
/* Address to clear the PIC */
#define TM_NE2K_PIC_CLEAR_ADDR 0x20
/* Value to write to the PIC to clear it */
#define TM_NE2K_PIC_CLEAR_VAL  0x20
/* Our IRQ */
#define TM_NE2K_IRQ            5
/*
 * The start of the PROM data. Incedentally, this is where the Ethernet
 * address lives
 */
#define TM_NE2K_PROM_START   0x00

/*
 * If there's been an overflow in the receive ring, we have to do a silly
 * dance as prescribed by NatSemi. One of the steps is to delay for
 * a while. We use this define as the number to count to in a while
 * loop. It's much, much too large, but better safe than sorry.
 */
#define TM_NE2K_DELAY_ISR    1000000L

/*
 * Various clones of the 8390 have a bug in them, which can be handled
 * by doing a dummy read before you write to the data port.
 */
/* #define TM_8390_HANDLE_RB4W_BUG */

/*
 * This is defined to do nothing. If you want the driver to do something
 * else while it's waiting for various things to occur, define it here
 */
#define tm_ne2k_busy_wait

/*
 * These defines are all for Borland BCC 5.0. Change them in accordance with
 * your compiler
 */
#ifdef __BORLANDC__
/* Write out a word to specified port address*/
#define tm_ne2k_out_word outport
/* Write out a byte to specified port address */
#define tm_ne2k_out_byte outportb
/* Read in a word from specified port address */
#define tm_ne2k_in_word  inport
/* Read in a byte from specified port address */
#define tm_ne2k_in_byte  inportb
#else /* _MSC_VER */
/* Write out a word to specified port address*/
#define tm_ne2k_out_word outpw
/* Write out a byte to specified port address */
#define tm_ne2k_out_byte outp
/* Read in a word from specified port address */
#define tm_ne2k_in_word  inpw
/* Read in a byte from specified port address */
#define tm_ne2k_in_byte  inp
#endif


/*
 * Pre-Processor Definitions (should not have to be modified)
 */


#define TM_8390_TRUE  1
#define TM_8390_FALSE 0

/* The length (in bytes) of an ethernet address */
#define TM_8390_ETHER_ADDR_LEN 6
/* The length (in bytes) of the multicast register(s) */
#define TM_8390_MCAST_ADDR_LEN 8

/* The CRC32 Polynomial (used for multicast hashing) */
#define TM_NE2K_CRC32_POLY 0x04c11db7UL

/*
 * These are the values stored in the isFull member of the ttNE2kSendRingEntry
 * structure (defined below).
 * 0 - this page is empty
 * 1 - this page is full
 * 2 - this page is empty, but for the next page, go back to the beginning of
 *     the ring.
 */
#define TM_8390_SEND_RING_EMPTY 0x00
#define TM_8390_SEND_RING_FULL  0x01
#define TM_8390_SEND_RING_WRAP  0x02

/* Values to check if a lock is 'locked' or 'unlocked' */
#define TM_8390_LOCKED   1
#define TM_8390_UNLOCKED 0

/* The length of a buffer page */
#define TM_8390_BUFFER_PAGE_LEN 256

/* Minimum allowable size for a received packet */
#define TM_8390_MIN_ETHER_PACKET 60
/* Maximum allowable size for a received packet */
#define TM_8390_MAX_ETHER_PACKET 1518

#define TM_NE2K_DATAPORT        0x10  /* Data Port                     */
#define TM_NE2K_RESET           0x1f  /* Reset Port                    */
#define TM_NE2K_BUFFER_START    0x40  /* First page of buffer space    */
#define TM_NE2K_BUFFER_STOP     0x80  /* Last page of buffer space + 1 */
#define TM_NE2K_REGISTER_OFFSET 0x00  /* Where the registers start     */

/*
 * The command register. It's available on every page
 */
#define TM_8390_CR        0x00 + TM_NE2K_REGISTER_OFFSET

/*
 * Page 0 Read-Only Registers
 */
/* Current Local DMA Address 0              */
#define TM_8390_PG0_CLDA0 0x01 + TM_NE2K_REGISTER_OFFSET
/* Current Local DMA Address 1              */
#define TM_8390_PG0_CLDA1 0x02 + TM_NE2K_REGISTER_OFFSET
/* Transmit Status Register                 */
#define TM_8390_PG0_TSR   0x04 + TM_NE2K_REGISTER_OFFSET
/* Number of Collision Regisger             */
#define TM_8390_PG0_NCR   0x05 + TM_NE2K_REGISTER_OFFSET
/* FIFO                                     */
#define TM_8390_PG0_FIFO  0x06 + TM_NE2K_REGISTER_OFFSET
/* Current Remote DMA Address 0             */
#define TM_8390_PG0_CRDA0 0x08 + TM_NE2K_REGISTER_OFFSET
/* Current Remote DMA Address 1             */
#define TM_8390_PG0_CRDA1 0x09 + TM_NE2K_REGISTER_OFFSET
/* Receive Status Register                  */
#define TM_8390_PG0_RSR   0x0c + TM_NE2K_REGISTER_OFFSET
/* Tally Counter 0 (Frame Alignment Errors) */
#define TM_8390_PG0_CNTR0 0x0d + TM_NE2K_REGISTER_OFFSET
/* Tally Counter 1 (CRC Errors)             */
#define TM_8390_PG0_CNTR1 0x0e + TM_NE2K_REGISTER_OFFSET
/* Tally Counter 2 (Missed Packet Errors)   */
#define TM_8390_PG0_CNTR2 0x0f + TM_NE2K_REGISTER_OFFSET

/*
 * Page 0 Read/Write Registers
 */
/* Boundary Pointer                         */
#define TM_8390_PG0_BNRY  0x03 + TM_NE2K_REGISTER_OFFSET
/* Interrupt Status Register                */
#define TM_8390_PG0_ISR   0x07 + TM_NE2K_REGISTER_OFFSET

/*
 * Page 0 Write-Only Registers
 */
/* Page Start Register             */
#define TM_8390_PG0_PSTART 0x01 + TM_NE2K_REGISTER_OFFSET
/* Page Stop Register              */
#define TM_8390_PG0_PSTOP  0x02 + TM_NE2K_REGISTER_OFFSET
/* Transmit Page Start Address     */
#define TM_8390_PG0_TPSR   0x04 + TM_NE2K_REGISTER_OFFSET
/* Transmit Byte Count Register 0  */
#define TM_8390_PG0_TBCR0  0x05 + TM_NE2K_REGISTER_OFFSET
/* Transmit Byte Count Register 1  */
#define TM_8390_PG0_TBCR1  0x06 + TM_NE2K_REGISTER_OFFSET
/* Remote Start Address Register 0 */
#define TM_8390_PG0_RSAR0  0x08 + TM_NE2K_REGISTER_OFFSET
/* Remote Start Address Register 1 */
#define TM_8390_PG0_RSAR1  0x09 + TM_NE2K_REGISTER_OFFSET
/* Remote Byte Count Register 0    */
#define TM_8390_PG0_RBCR0  0x0a + TM_NE2K_REGISTER_OFFSET
/* Remote Byte Count Register 1    */
#define TM_8390_PG0_RBCR1  0x0b + TM_NE2K_REGISTER_OFFSET
/* Receive Configuration Register  */
#define TM_8390_PG0_RCR    0x0c + TM_NE2K_REGISTER_OFFSET
/* Transmit Configuration Register */
#define TM_8390_PG0_TCR    0x0d + TM_NE2K_REGISTER_OFFSET
/* Data Configuration Register     */
#define TM_8390_PG0_DCR    0x0e + TM_NE2K_REGISTER_OFFSET
/* Interrupt Mask Register         */
#define TM_8390_PG0_IMR    0x0f + TM_NE2K_REGISTER_OFFSET

/*
 * Page 1 Registers
 */
/* Physical Address Register 0  */
#define TM_8390_PG1_PAR0 0x01 + TM_NE2K_REGISTER_OFFSET
/* Current Page Register        */
#define TM_8390_PG1_CURR 0x07 + TM_NE2K_REGISTER_OFFSET
/* Multicast Address Register 0 */
#define TM_8390_PG1_MAR0 0x08 + TM_NE2K_REGISTER_OFFSET
/* Multicast Address Register 5 */
#define TM_8390_PG1_MAR5 0x0d + TM_NE2K_REGISTER_OFFSET

/* Definitions for the Command Register (CR) */
#define TM_8390_CR_STP 0x01 /* Stop                 */
#define TM_8390_CR_STA 0x02 /* Start                */
#define TM_8390_CR_TXP 0x04 /* Transmit Packet      */
#define TM_8390_CR_RD0 0x08 /* Remote DMA Command 0 */
#define TM_8390_CR_RD1 0x10 /* Remote DMA Command 1 */
#define TM_8390_CR_RD2 0x20 /* Remote DMA Command 2 */
#define TM_8390_CR_PS0 0x40 /* Page Select 0        */
#define TM_8390_CR_PS1 0x80 /* Page Select 1        */

/* Definitions for the Interrupt Status Register (ISR) */
#define TM_8390_ISR_PRX 0x01 /* Packet Received     */
#define TM_8390_ISR_PTX 0x02 /* Packet Transmitted  */
#define TM_8390_ISR_RXE 0x04 /* Receive Error       */
#define TM_8390_ISR_TXE 0x08 /* Transmit Error      */
#define TM_8390_ISR_OVW 0x10 /* Overwrite Warning   */
#define TM_8390_ISR_CNT 0x20 /* Counter Overflow    */
#define TM_8390_ISR_RDC 0x40 /* Remote DMA Complete */
#define TM_8390_ISR_RST 0x80 /* Reset Status        */

/* Definitions for the Data Configuration Register (DCR) */
#define TM_8390_DCR_WTS  0x01 /* Word Transfer Select     */
#define TM_8390_DCR_BOS  0x02 /* Byte Order Select        */
#define TM_8390_DCR_LAS  0x04 /* Long Address Select      */
#define TM_8390_DCR_LS   0x08 /* Loopback Select          */
#define TM_8390_DCR_AR   0x10 /* Auto-Initialize Remote   */
#define TM_8390_DCR_FT0  0x20 /* FIFO Threshhold Select 0 */
#define TM_8390_DCR_FT1  0x40 /* FIFO Threshhold Select 1 */

/* Definitions for the Receive Configuration Register (RCR) */
#define TM_8390_RCR_SEP 0x01 /* Save Errored Packets */
#define TM_8390_RCR_AR  0x02 /* Accept Runt Packets  */
#define TM_8390_RCR_AB  0x04 /* Accept Broadcast     */
#define TM_8390_RCR_AM  0x08 /* Accept Multicast     */
#define TM_8390_RCR_PRO 0x10 /* Promiscuous Physical */
#define TM_8390_RCR_MON 0x20 /* Monitor Mode         */

/* Definitions for the Transmit Configuration Register (TCR) */
#define TM_8390_TCR_CRC  0x01 /* Inhibit CRC                */
#define TM_8390_TCR_LB0  0x02 /* Encoded Loopback Control 0 */
#define TM_8390_TCR_LB1  0x04 /* Encoded Loopback Control 1 */
#define TM_8390_TCR_ATD  0x08 /* Auto Transmit Disable      */
#define TM_8390_TCR_OFST 0x10 /* Collision Offset Enable    */

/* Definitions for the Receive Status Register (RSR) */
#define TM_8390_RSR_PRX 0x01 /* Packet Received Intact     */
#define TM_8390_RSR_CRC 0x02 /* CRC Error                  */
#define TM_8390_RSR_FAE 0x04 /* Frame Alignment Error      */
#define TM_8390_RSR_FO  0x08 /* FIFO Overrun               */
#define TM_8390_RSR_MPA 0x10 /* Missed Packet              */
#define TM_8390_RSR_PHY 0x20 /* Physical/Multicast Address */
#define TM_8390_RSR_DIS 0x40 /* Receiver Disabled          */
#define TM_8390_RSR_DFR 0x80 /* Deferring                  */

/* Definitions for the Transmit Status Register (TSR) */
#define TM_8390_TSR_PTX 0x01  /* Packet Transmitted      */
#define TM_8390_TSR_COL 0x04  /* Transmit Collided       */
#define TM_8390_TSR_ABT 0x08  /* Transmit Aborted        */
#define TM_8390_TSR_CRS 0x10  /* Carrier Sense Lost      */
#define TM_8390_TSR_FU  0x20  /* FIFO Underrun           */
#define TM_8390_TSR_CDH 0x40  /* CD Heartbeat            */
#define TM_8390_TSR_OWC 0x80  /* Out of Window Collision */


/*
 * Structure Definitions
 */


/*
 * Structure representing the header info stored before each received
 * packet by the 8390
 */
typedef struct ts8390HeaderInfo
{
/* Status of the received packet */
    unsigned char      status;
/* Offset of the start page of the next packet */
    unsigned char      nextPacket;
/* Length (in bytes) of packet + header */
    unsigned short int length;
} tt8390HeaderInfo;
typedef tt8390HeaderInfo TM_FAR * tt8390HeaderInfoPtr;

/*
 * The 8390, being a very low-price, low-capability chip, does not do any of
 * the accounting for its send ring. Which means that we get to. We create
 * an array of these structures, one per transmit page, in order to keep
 * track of vital ring information.
 */
typedef struct tsNE2kSendRingEntry
{
/* The length of the packet stored in this page */
    unsigned short int packetLen;
/* Is this page full? */
    unsigned char      isFull;
} ttNE2kSendRingEntry;
typedef ttNE2kSendRingEntry TM_FAR * ttNE2kSendRingEntryPtr;

/*
 * In order to allow scatter-send, we have an array, the number of elements of
 * which is the maximum number of buffers that can comprise a packet. Thus, if
 * you allow a maximum of 5 buffers per scatter packets, this array will have
 * 5 elements. Each element in the array is one of these structures, which
 * holds a pointer to the data, and the length of the data.
 */
typedef struct tsNE2kScatterHdrEntry
{
/* A pointer to a portion of the packet */
    void TM_FAR *      dataPtr;
/* The length of this portion of the packet */
    unsigned short int dataLen;
} ttNE2kScatterHdrEntry;
typedef ttNE2kScatterHdrEntry TM_FAR * ttNE2kScatterHdrEntryPtr;

/*
 * This structure contains all of the information we need to track for
 * the device.
 */
typedef struct tsNE2kDev
{
/*
 * The old ISR. We save it when we 'open' the chip, and restore it when we
 * close it
 */
#ifdef __BORLANDC__
    void interrupt (TM_CODE_FAR * oldIsr)();
#else /* _MSC_VER */
    ttUserIsrHandlerPtr oldIsr;
#endif
/* The pointer to the array of send-ring entries */
    ttNE2kSendRingEntryPtr   sendRingPtr;
/* The pointer to the array that holds buffer information for scatter sends */
    ttNE2kScatterHdrEntryPtr scatterBuffPtr;
/*
 * The interface handle. Generally, it is passed into any driver function,
 * but for this chip we need it in the ISR, to which no parameters are
 * passed
 */
    ttUserInterface          ifHandle;
/* For scatter send, this is the number of buffers that comprise a packet */
    int                      sendPieces;
/*
 * For scatter send, this is the maximum number of buffers that comprise a
 * packet
 */
    int                      maxSendPieces;
/* The offset of the page we're sending from */
    unsigned short int       curSend;
/* The offset of the page we previously sent from */
    unsigned short int       prevSend;
/*
 * The offset of the page we're currently filling. It will be sent as soon as 
 * we get a chance.
 */
    unsigned short int       fillSend;
/* For scatter-send, we need to know how long a packet is. */
    short int                packetLength;
/* The current receive page */
    unsigned char            currentRcvPage;
/* The Ethernet address we read out of the hardware */
    unsigned char            ethAddr[TM_ETHERNET_PHY_ADDR_LEN];
/*
 * This is another part of the silly dance we have to do if there's a receive
 * overflow. We have to keep track of if an overflow occurred, and if so,
 * whether we were transmitting at the time.
 */
    unsigned char            overflowFlags;
/* We need to guarantee we only send one packet at a time */
    volatile unsigned char   sendingLock;
    char                     ne2kXmitOn;
    char                     ne2kFiller[2];
} ttNE2kDev;
typedef ttNE2kDev TM_FAR * ttNE2kDevPtr;


/*
 * Global Variables
 */

ttNE2kDevPtr tlNe2kPtr;

/*
 * Forward Definitions
 */


static void tfNE2kBlockInput(unsigned short int     readLen,
                             unsigned short int     currOffset,
                             unsigned char TM_FAR * rxBuffer);

static void tfNE2kBlockOutput(unsigned char startPage);

static void interrupt tfNE2kIsr(void);


/*
 * Code
 */


/*
 * tfNE2kOpen()
 *
 *   Configure and start up the Ethernet interface.
 *   We take the address from the Ethernet board and set up the board.
 *   Then we store the interrupt address and enable the interrupt.
 *   Parameters:
         interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       error:            TM_DEV_ERROR
 *       success:          TM_DEV_OKAY
 */
int tfNE2kOpen(ttUserInterface interfaceHandle)
{
#ifndef __BORLANDC__
    void TM_FAR * TM_FAR * intVectorPtr;
#endif

/* Variable to save the old command register in case of a problem */
    unsigned char oldCmdReg;
/* Variable to save the old Tally Counter 0 register in case of a problem */
    unsigned char oldCntrReg;
/* The current interrupt mask for the PIC (x86 only) */
    unsigned char currIntMask;
/* A counter */
    unsigned char counter;
/* The return code */
    int           retCode;
    unsigned int  totalAllocSize;
    unsigned int  scatterAllocSize;
    int           scatterCount;

/* Did the user turn on the transmit queue? */
    if (tfInterfaceXmitQueueOn(interfaceHandle))
    {
/* This driver does not support a Treck transmit queue */
        tfKernelWarning("tfNe2kOpen",
                        "Cannot use Treck transmit queue with NE2K driver\n");
/* Tell stack there is a problem */
        retCode = TM_DEV_ERROR;
        goto ne2kOpenEnd;
    }
/*
 * The maximum allowed buffers that represent a single packet for scatter
 * send.
 */
    scatterCount = tfInterfaceScatterCount(interfaceHandle);
    if (scatterCount < 1)
    {
/* Tell stack there is a problem */
        retCode = TM_DEV_ERROR;
        goto ne2kOpenEnd;
    }
/* We assume the return code is OK unless it's specifically changed */
    retCode = TM_DEV_OKAY;

/* Get memory for the array of scatterHdrEntry structures */
    scatterAllocSize = scatterCount * sizeof(ttNE2kScatterHdrEntry);
/*
 * And, get memory for the ring of sendRingEntry structs, and for
 * ttNE2kDev structure.
 */
    totalAllocSize =   sizeof(ttNE2kDev)
                     + scatterAllocSize
                     + TM_NE2K_TX_PAGES * sizeof(ttNE2kSendRingEntry);
/* Pointer to the allocated memory area */
    tlNe2kPtr = (ttNE2kDevPtr)tm_get_raw_buffer(totalAllocSize);
    if (tlNe2kPtr == (void TM_FAR *)0)
    {
/* Tell stack there is a problem */
        retCode = TM_DEV_ERROR;
        goto ne2kOpenEnd;
    }
/* Zero out this memory area */
    tm_bzero( tlNe2kPtr, totalAllocSize );
    tlNe2kPtr->maxSendPieces = scatterCount;
/* Array of scatterHdrEntry structures */
    tlNe2kPtr->scatterBuffPtr = (ttNE2kScatterHdrEntryPtr)
                           ( ((char TM_FAR *)tlNe2kPtr) + sizeof(ttNE2kDev) );
   
/* ring of sendRingEntry structs */
    tlNe2kPtr->sendRingPtr = (ttNE2kSendRingEntryPtr)
                                (   ((char TM_FAR *)tlNe2kPtr->scatterBuffPtr)
                                  + scatterAllocSize );

/* Save the interface handle so we can get to it in the ISR */
    tlNe2kPtr->ifHandle = interfaceHandle;

/*
 * This code checks if there's an NE2000 hanging out at the base address.
 * This is magic from the Linux driver.
 */
/*
 * Save whatever was in the command register in case this isn't actually
 * an NE2000.
 */
    oldCmdReg = (unsigned char) tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS);
/* Stop chip and set it to Page 1 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_PS0 + TM_8390_CR_STP );
/* Read from Multicast Register 5 */
    oldCntrReg = (unsigned char)
        tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR5);
/* Write all 1's to Multicast Register 5 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR5, 0xff);
/* Set the chip back to Page 0 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR, TM_8390_CR_RD2);
/* Read from a counter register, which should clear it */
    tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_CNTR0);
/* Check if the counter register was actually cleared */
    if( tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_CNTR0) != 0)
    {
/* It wasn't. Restore the registers, and set an error code */
        tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS, oldCmdReg);
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_CNTR0,
                          oldCntrReg );
/* Tell stack there is a problem */
        retCode = TM_DEV_ERROR;
        goto ne2kOpenEnd;
    }

/*
 * Reset the card. This is done by reading from the reset port, and
 * writing that value back to the reset port
 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_NE2K_RESET,
                      tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_RESET) );

/* Busy-wait for the reset to finish */
    while (    (   tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR)
                  & TM_8390_ISR_RST )
            == 0 )
    {
        tm_ne2k_busy_wait;
    }

/*
 * Retrieve the physical address from the card. To do this correctly we must
 * first set up card registers
 */
/* Set the card to Page 0 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR, TM_8390_CR_RD2 +
                                                        TM_8390_CR_STP);
/*
 * Set the chip for normal operation, 4-word FIFO threshhold, and word-wide
 * transfers
 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_DCR,
                      TM_8390_DCR_FT1 + TM_8390_DCR_LS + TM_8390_DCR_WTS );
/* Clear the count registers */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0, 0);
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0);
/* Mask all interrupts in the IMR */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_IMR, 0x00);
/* Clear all interrupts in the ISR */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, 0xff);
/* Set receive to monitor mode */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR, TM_8390_RCR_MON);
/* Set transmit to loopback mode */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, TM_8390_TCR_LB0);
/* Set the number of bytes we want to transfer */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0,
                      TM_ETHERNET_PHY_ADDR_LEN );
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0);
/* Set the address to read from */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0,
                      TM_NE2K_PROM_START );
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1,
                      TM_NE2K_PROM_START );
/* Start reading the PROM */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD0 + TM_8390_CR_STA );

/* Read in the Ethernet address */
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tlNe2kPtr->ethAddr[counter] = (unsigned char)
                     tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT);
    }

/* Store the first receive page we'll use */
    tlNe2kPtr->currentRcvPage = TM_NE2K_BUFFER_START + TM_NE2K_TX_PAGES;

/* Stop the chip and set it back to Page 0 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP );
/*
 * Set the chip for normal operation, 4-word FIFO threshhold, and word-wide
 * transfers
 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_DCR,
                      TM_8390_DCR_FT1 + TM_8390_DCR_LS + TM_8390_DCR_WTS );
/* Clear Count Registers */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0, 0);
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0);
/* Set receive to monitor mode */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR, TM_8390_RCR_MON);
/* Set transmit to loopback mode */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, TM_8390_TCR_LB0);
/* Set the first transmit page */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TPSR,
                      TM_NE2K_BUFFER_START );
/* Set the first receive page */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_PSTART,
                      TM_NE2K_BUFFER_START + TM_NE2K_TX_PAGES );
/* Set the receive boundary page */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY,
                      TM_NE2K_BUFFER_STOP - 1);
/* Set the receive stop page */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_PSTOP,
                      TM_NE2K_BUFFER_STOP );

/* Mask all interrupts in the IMR */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_IMR, 0);
/* Clear all interrupts in the ISR */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, 0xff);

/* Change to Page 1 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_PS0 + TM_8390_CR_STP );
/* Set physical address & disable multicast */
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_PAR0 + counter,
                          tlNe2kPtr->ethAddr[counter] );
    }
    for(counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + counter,
                          0 );
    }

/* Set the current page */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_CURR,
                      tlNe2kPtr->currentRcvPage );
/* Reset to Page 0 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STP );

/* Set up interrupts */
/* Save the old ISR */
#ifdef __BORLANDC__
    tlNe2kPtr->oldIsr = getvect(TM_NE2K_INT);
#else /* _MSC_VER */
    intVectorPtr = (void TM_FAR *) ((unsigned long) TM_NE2K_INT * 4UL);
    tlNe2kPtr->oldIsr = (ttUserIsrHandlerPtr) *intVectorPtr;
#endif
/* Unmask our interrupt */
    currIntMask = (unsigned char) tm_ne2k_in_byte(TM_NE2K_PIC_ADDR);
    currIntMask &= (~(1 << TM_NE2K_IRQ)) & 0xff;
    tm_ne2k_out_byte(TM_NE2K_PIC_ADDR,currIntMask);
/* Install the ISR handler */
    tfKernelInstallIsrHandler((ttUserIsrHandlerPtr)tfNE2kIsr,TM_NE2K_INT);

/* Clear all interrupts */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, 0xff);
/* Enable the interrupts that concern us */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_IMR,
                      TM_8390_ISR_PRX + TM_8390_ISR_PTX + TM_8390_ISR_OVW );
/* Start the chip */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STA );
/* Enable transmit */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, 0);
/*
 * Set the Receive Configuration Register to receive packets and accept
 * broadcasts
 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR, TM_8390_RCR_AB);

ne2kOpenEnd:
    return(retCode);
}


/*
 * tfNE2kClose()
 *
 *   This code removes the interrupt from the PIC and restores the old
 *   interrupt table. We could also shut down the chip, but it really doesn't
 *   seem necessary.
 *   Parameters:
 *       interfaceHandle   The user interface handle for this device.
 *   Returns:
 *       TM_DEV_OKAY
 */
int tfNE2kClose (ttUserInterface interfaceHandle)
{
/* We need to save the current interrupt mask for a little while */
    unsigned char currIntMask;

/* Deal with 'unused variable' compiler warning */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Get the current interrupt mask */
    currIntMask = (unsigned char) tm_ne2k_in_byte(TM_NE2K_PIC_ADDR);
/* Mask out our interrupt */
    currIntMask |= (1 << TM_NE2K_IRQ) & 0xff;
/* Save the modified interrupt mask */
    tm_ne2k_out_byte(TM_NE2K_PIC_ADDR, currIntMask);
/* Save the old vector back into the vector table */
    tfKernelInstallIsrHandler( (ttUserIsrHandlerPtr)tlNe2kPtr->oldIsr,
                               TM_NE2K_INT );

    tm_free_raw_buffer((ttRawBufferPtr)tlNe2kPtr);
    tlNe2kPtr = (ttNE2kDevPtr)0;
    return(TM_DEV_OKAY);
}


/*
 * tfNE2kIsr()
 *
 *   This funciton is the interrupt handler. It deals with receive, transmit-
 *   complete, and receive overflow warning interrupts.
 *   Parameters:
 *       none
 *   Returns:
 *       none
 */
static void interrupt tfNE2kIsr(void)
{
/* Used to store the state of the ISR */
    unsigned char     interrupts;
/* The command register, so we can restore it when we're done */
    unsigned char     cmdReg;
/* Whether we were sending when the interrupt came in */
    unsigned char     sending;
/* A counter */
    long int          counter;
/*
 * The number of packets we've received, and thus, second parameter
 * to tfNotifyInterfaceIsr()
 */
    int              recvCount;

    recvCount = 0;

/* Save the state of the command register */
    cmdReg = (unsigned char)
        tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR);
/* Set to page 0 to read interrupt status */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      cmdReg & ~(TM_8390_CR_PS0 | TM_8390_CR_PS1) );

/*
 * Record what interrupts need to be serviced. We can ignore reset, but the
 * rest might be of interest to us.
 * We have to loop here, as new interrupts can come in while we're working
 * on the old ones. The 8390 is a pretty hokey chip.
 */
    while((interrupts = (unsigned char)
           (tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR) &
            ~TM_8390_ISR_RST)) != 0)
    {
/* Is there an overrun warning interrupt? */
        if (interrupts & TM_8390_ISR_OVW)
        {
/*
 * This is the silly dance that NatSemi says we have to do if there's an 
 * overrun in the receive ring
 */
/* Check if we are sending */
            sending = (unsigned char)
                ((tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR)
                  & TM_8390_CR_TXP) ? TM_8390_TRUE : TM_8390_FALSE);

/* Stop the chip */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_STP );

/* Delay for a while */
            for (counter = 0; counter < (TM_NE2K_DELAY_ISR); counter++)
            {
#ifdef __BORLANDC__
                asm nop;
#else /* _MSC_VER */
                __asm nop;
#endif
            }
/* Reset counters */
            tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0, 0);
            tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0);

/* If we were sending... */
            if (sending)
            {
/* Check if we need to resend the packet (i.e. Have we completed send?) */
                if (tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR) & 
                    (TM_8390_ISR_PTX | TM_8390_ISR_TXE))
                {
/*
 * Once we have cleared some received packets and restarted the chip, we
 * check this flag to see if we need to resend the packet
 */
                    tlNe2kPtr->overflowFlags |= TM_8390_CR_TXP;
                }
            }

/* Go into loopback mode */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR,
                              TM_8390_TCR_LB0 );
/* Start the chip */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_STA );

/* Clear the interrupt */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR,
                              TM_8390_ISR_OVW + TM_8390_ISR_PRX );

/*
 * The next time receive is called, this flag will cause it to reenable
 * the chip
 */
            tlNe2kPtr->overflowFlags |= TM_8390_ISR_OVW;
        }
        else
        {
/* Is there a receive interrupt? */
            if (interrupts & TM_8390_ISR_PRX)
            {
/* Clear the interrupt */
                tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR,
                                  TM_8390_ISR_PRX );
/*
 * Record the fact that we've got a new packet, so we can notify the stack
 * of it later. We should be able to notify here, but some RTOS's
 * inadvertantly re-enable interrupts in the notification function. And that's
 * bad.
 */
                recvCount++;
            }
        }

/* Is there a transmit complete interrupt? */
        if (interrupts & TM_8390_ISR_PTX)
        {
/* Check if we're at the end of the send ring */
            if (    tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].isFull
                 == TM_8390_SEND_RING_WRAP)
            {
/* Mark this packet as empty */
                tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].isFull = 
                                                     TM_8390_SEND_RING_EMPTY;
/* Go to the beginning of the ring */
                tlNe2kPtr->curSend = 0;
            }
/* Store the location of the previously sent packet */
            tlNe2kPtr->prevSend = tlNe2kPtr->curSend;
/*
 * Get the start of the next packet to send. To do this, we subtract one from
 * the actual length of the packet and shift right by 8, which gives us 0 for
 * lengths of 1 through 256, 1 for lengths of 257 through 512, etc. This number
 * being 1 less than the pages the buffer actually takes up, we add one at the
 * end.
 */
            tlNe2kPtr->curSend += 
               ((tlNe2kPtr->sendRingPtr[tlNe2kPtr->prevSend].packetLen -
                 1) >> 8) + 1;

/*
 * If the last packet went all the way to the end of the buffer-space, we
 * have to go back to the beginning
 */
            if (tlNe2kPtr->curSend >= TM_NE2K_TX_PAGES)
            {
                tlNe2kPtr->curSend = 0;
            }
/* 
 * If we've made a note to ourselves that the next packet couldn't fit at
 * the end of the buffer ring, we have to go back to the beginning
 */
            if (    tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].isFull
                 == TM_8390_SEND_RING_WRAP)
            {
/* Mark this buffer as empty */
                tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].isFull = 
                                                     TM_8390_SEND_RING_EMPTY;
/* Go back to the beginning of the ring */
                tlNe2kPtr->curSend = 0;
            }

/*
 * We need to find out how many pages the last packet took up so that we
 * can free them. Becaue of the way the loop (see below) works, we actually
 * want the number to be 1 less than the number of buffers used
 */
            counter = 
                (tlNe2kPtr->sendRingPtr[tlNe2kPtr->prevSend].packetLen -
                 1) >> 8;
/* Mark all of the now-empty pages as empty */
            for (; counter >= 0; counter--)
            {
                tlNe2kPtr->sendRingPtr[tlNe2kPtr->prevSend +
                                          (unsigned short int)counter].isFull
                                                   = TM_8390_SEND_RING_EMPTY;
            }
/* Clear the interrupt */
            tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR,
                             TM_8390_ISR_PTX);
/* Check if there's another packet to send... */
            if (    tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].isFull
                 == TM_8390_SEND_RING_FULL )
            {
/* Store the packet length */
                tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR0,
                    tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].packetLen
                  & 0xff);
                tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR1,
                    tlNe2kPtr->sendRingPtr[tlNe2kPtr->curSend].packetLen
                 >> 8 );
/* Store the buffer address */
                tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TPSR,
                                 tlNe2kPtr->curSend + TM_NE2K_BUFFER_START);
/*
 * Restore the old command register, with the addition of making sure it's in
 * Page 0 and starting a transmit
 */
                tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                                    (   cmdReg
                                      & ~(TM_8390_CR_PS0 | TM_8390_CR_PS1) )
                                  | TM_8390_CR_TXP );
            }
            else
            {
/* We're not sending anything, so we can unlock sendingLock */
                tlNe2kPtr->sendingLock = TM_8390_UNLOCKED;
            }
 
        }

/*
 * The RTL8019AS NIC seems to have an awkward glitch.  It seems that
 * periodically the RDC bit gets set in the ISR, this does NOT seem
 * to cause the interrupt to stay active, however, it does block this
 * loop from exiting.  By writing all unused bits here, we assure that
 * the bits are cleared and if a spurious bit gets through, it will not
 * stop the operation.
 */
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR,
                            TM_8390_ISR_RXE
                          + TM_8390_ISR_TXE
                          + TM_8390_ISR_CNT
                          + TM_8390_ISR_RDC );
    }

/* Clear the interrupt from the PIC*/
    tm_ne2k_out_byte(TM_NE2K_PIC_CLEAR_ADDR, TM_NE2K_PIC_CLEAR_VAL);

/* Restore the Command Register */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR, cmdReg);

/* Call notify with the number of packets we have received */
    tfNotifyInterfaceIsr( tlNe2kPtr->ifHandle, recvCount, 0, 0UL, 0UL );
}


/*
 * tfNE2kBlockInput()
 *
 *   This function transfers data from the NE2000's internal buffers to
 *   standard memory.
 *
 *   If the user is compiling with the Borland compiler, we have provided
 *   a method to do the transfer in hand-optimized assembly. This speeds
 *   things up IMMENSELY. I would very highly suggest that you port this
 *   assembly code to whatever platform you are using. If the the user does
 *   not wish to make use of the assembly, another method is provided.
 *
 *   Parameters:
 *       readLen    Number of bytes to transfer
 *       currOffset The place to begin reading from in the internal buffers
 *       rxBuffer   The buffer to copy into
 *   Returns:
 *       none
 */
static void tfNE2kBlockInput(unsigned short int     readLen,
                             unsigned short int     currOffset,
                             unsigned char TM_FAR * rxBuffer)
{
/* If we don't use the assembly code, we need a counter */
#ifndef __BORLANDC__
    unsigned short int counter;
#endif

/* Set chip to Page 0 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_STA );
/* Store the lower 8 bits of the packet length */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0,
                      readLen & 0xFF );
/* Store the upper 8 bits of the packet length */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1,
                      readLen >> 8 );
/* Store the lower 8 bits of the address to read from */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0,
                      currOffset & 0xFF );
/* Store the upper 8 bits of the address to read from */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1,
                      currOffset >> 8 );
/* Start the transfer */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD0 + TM_8390_CR_STA );

/* This is the assembly version of the transfer */
#ifdef __BORLANDC__
/* Save off the registers we will be using */
/* dx contains the i/o port */
    asm push    dx
/* di has the lower 8 bits of the address */
    asm push    di
/* es has the upper 8 bits of the address */
    asm push    es
/* cx will be the counter */
    asm push    cx

/* Load the address in rxBuffer into es:di */
    asm les     di, rxBuffer
/* Save the i/o port into dx */
    asm mov     dx, TM_NE2K_BASE_ADDRESS+TM_NE2K_DATAPORT
/* Save the count into cx */
    asm mov     cx, readLen
/* We do this a 16 bits at a time, so divide the counter by 2 */
    asm shr     cx,1
/* Do the input */
    asm rep     insw

/* Restore all of the registers we used */
    asm pop     cx
    asm pop     es
    asm pop     di
    asm pop     dx
/* Here's the transfer in non-assembly form */
#else /* _MSC_VER */
    for (counter = 0; counter < (readLen>>1); counter ++)
    {
/*
 * Get 16 bits from the dataport & save it to the appropriate place in the
 * buffer
 */
        ((unsigned short int TM_FAR *)rxBuffer)[counter] = (unsigned short int)
            tm_ne2k_in_word(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT);
    }
#endif
/*
 * In both assembly and non-assembly versions, if we had an odd number of
 * bytes, we have to pick up the last one
 */
    if (readLen & 1)
    {
        rxBuffer[readLen-1] = (unsigned char)
            tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT);
    }

/* Acknowledge the DMA transfer */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, TM_8390_ISR_RDC);

    return;
}


/*
 * tfNE2kBlockOutput()
 *
 *   This function transfers data from standard memory to the NE2000's
 *   internal buffers.
 *
 *   If the user is compiling with the Borland compiler, we have provided
 *   a method to do the transfer in hand-optimized assembly. This speeds
 *   things up IMMENSELY. I would very highly suggest that you port this
 *   assembly code to whatever platform you are using. If the the user does
 *   not wish to make use of the assembly, another method is provided.
 *
 *   Parameters:
 *       startPage   The page to start writing the data to
 *   Returns:
 *       none
 */
static void tfNE2kBlockOutput(unsigned char startPage)
{
#ifdef __BORLANDC__
    unsigned char TM_FAR * txBuffer;
#else /* _MSC_VER */
/* If we're not using the assembly version, we need a counter */
    unsigned short int     counter;
#endif
    unsigned short int     bytes;
/*
 * We support scattered data, so we need to know how many different pieces
 * a packet is in
 */
    unsigned short int     buffCount;

/* Set chip to Page 0 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_STA + TM_8390_CR_RD2 );

/*
 * This bug doesn't always occur - but sometimes a dummy read must be done
 * before we can write data to the chip. The actual code is magic from
 * the Linux driver.
 */
#ifdef TM_8390_HANDLE_RB4W_BUG
/*
 * Tell the chip you want to read 66 bytes of data, starting at offset 66.
 * As mentioned above, this is magic from the Linux driver. I'm not sure
 * if the numbers are significant, but my intuition is that they're not
 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0, 0x42);
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1, 0x00);
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0, 0x42);
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1, 0x00);
/* Start the 'transfer' */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD0 + TM_8390_CR_STA );
/* This is a short delay so the dummy read has a chance to occur */
    asm jmp j1
j1:
    asm jmp j2
j2:
    asm jmp j3
j3:
/* Acknowledge the DMA command */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, TM_8390_ISR_RDC);
#endif /* #ifdef TM_8390_HANDLE_RB4W_BUG */

/* Store the lower 8 bits of the length to write */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR0,
                      tlNe2kPtr->packetLength & 0xFF );
/* Store the upper 8 bits of the length to write */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RBCR1,
                      tlNe2kPtr->packetLength >> 8 );
/* 
 * Store the lower 8 bits of the address to write to. This is always 0, as
 * we always write on a page boundary
 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR0, 0);
/* Store the upper 8 bits of the address to write to */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RSAR1, startPage);
/* Start the transfer */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD1 + TM_8390_CR_STA );

/* For each piece of the buffer... */
    for (buffCount = 0; ((int) buffCount) < tlNe2kPtr->sendPieces; buffCount++)
    {
/* The length we need to write is the length of this buffer */
        bytes = tlNe2kPtr->scatterBuffPtr[buffCount].dataLen;

/* This is the assembly version of the transfer */
#ifdef __BORLANDC__
        txBuffer = tlNe2kPtr->scatterBuffPtr[buffCount].dataPtr;
        

/* Save the registers we're going to use */
/* dx will be the i/o port */
        asm push    dx
/* si will be the lower 8 bits of the address of the buffer */
        asm push    si
/* ds will be the upper 8 bits of the address of the buffer */
        asm push    ds
/* cx will be the counter */
        asm push    cx

/* Load the address of the buffer into ds:si */
        asm lds     si, txBuffer
/* Load the i/o port into dx */
        asm mov     dx, TM_NE2K_BASE_ADDRESS+TM_NE2K_DATAPORT
/* Load the count into cx */
        asm mov     cx, bytes
/* Divide the count in half, as we're doing 16bits at a time */
        asm shr     cx, 1
/* Do the transfer */
        asm rep     outsw

/* Restore all of the saved registers */
        asm pop     cx
        asm pop     ds
        asm pop     si
        asm pop     dx
#else /* _MSC_VER */
/* The non-optimized version of the transfer... */
        for (counter=0; counter< (bytes >> 1); counter++)
        {
/* Write the next word to the port */
            tm_ne2k_out_word(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT, 
                ( (unsigned short int TM_FAR *)
                  tlNe2kPtr->scatterBuffPtr[buffCount].dataPtr )[counter]);
        }
#endif
/*
 * For both the optimized and the unoptimized version, if we had an odd number
 * of bytes, we have to pick up the last byte and write it to the port
 */
        if (bytes & 1)
        {
            tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_NE2K_DATAPORT, 
                ( (unsigned char TM_FAR *)
                  tlNe2kPtr->scatterBuffPtr[buffCount].dataPtr )[bytes-1]);
        }

    }

/* Acknowledge the DMA transfer */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_ISR, TM_8390_ISR_RDC);

    return;
}



/*
 * tfNE2kReceive()
 *
 *   This function receives a packet. It is called after we've received an
 *   interrupt indicating there is data to be handled.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         We fill in this pointer with a pointer to the data
 *                       we have just received
 *       dataLength      We fill in this length with the length of the data
 *                       we have just received
 *       ttUserBufferPtr We fill out this buffer handle with the buffer handle
 *                       we get back when we allocate memory
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
int tfNE2kReceive(ttUserInterface        interfaceHandle,
                  char TM_FAR * TM_FAR * dataPtr,
                  int TM_FAR *           dataLength,
                  ttUserBufferPtr        bufHandlePtr)
{
/* The return value */
    int                retCode;
/* The page that the chip is receiving into currently */
    unsigned char      rxPage;
/* The page we're reading from */
    unsigned char      thisFrame;
/* The next page to read from */
    unsigned char      nextFrameTest;
/* Various packet information stored by the chip */
    tt8390HeaderInfo   frameInfo;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Assume that we return OKAY unless otherwise instructed */
    retCode = TM_DEV_OKAY;

/* Set chip to Page 1 */
    tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                      TM_8390_CR_RD2 + TM_8390_CR_PS0);
/* Get the first 'empty' receive page */
    rxPage = (unsigned char)
        tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG1_CURR);
/* Set chip to Page 0 */
    tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_CR, TM_8390_CR_RD2);

/* 
 * Get the address of the next page that needs to be read. It is always
 * one ahead of the boundary page.
 */
    thisFrame = (unsigned char)
        (tm_ne2k_in_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY) + 1);
/*
 * If we're past the end of the buffer ring, go back to the beginning of the
 * receive section
 */
    if(thisFrame >= TM_NE2K_BUFFER_STOP)
    {
        thisFrame = TM_NE2K_BUFFER_START + TM_NE2K_TX_PAGES;
    }

/*
 * If the page that we're supposed to read from is the same page the chip
 * is currently receiving into, we don't actually have anything to read
 */
    if(thisFrame == rxPage)
    {
        retCode = TM_DEV_ERROR;
        goto ne2kRecvEnd;
    }

/* Get the frame header */
    tfNE2kBlockInput((unsigned short int) (sizeof(frameInfo)),
                     (unsigned short int) (thisFrame << 8),
                     (unsigned char TM_FAR *)&frameInfo);

/*
 * Calculate the address of the next frame from the packet length given in the
 * info prepended to the packet by the 8390.
 */
    nextFrameTest = (unsigned char)
        (((((frameInfo.length + sizeof(frameInfo)) -
            1) >> 8) + 1) + thisFrame);
/*
 * If we go past the end of the buffers, go back to the appropriate buffer
 * at the beginning of the ring. The receive logic is capable of wrapping
 * a packet, so we don't necessarily go back to the first receive page
 */
    if( nextFrameTest >= TM_NE2K_BUFFER_STOP )
    {
        nextFrameTest -= TM_NE2K_BUFFER_STOP;
        nextFrameTest += TM_NE2K_BUFFER_START + TM_NE2K_TX_PAGES;
    }

/*
 * We check to make sure the packet header info makes sense. If it doesn't,
 * we throw the packet away. The 4 tests performed are:
 *   1)  This one I'm not sure on. It came from the Linux driver, and exactly
 *       what the status field does is never actually discussed in the
 *       DP8390D datasheet. I decided to leave it in, as Donald Becker knows
 *       what he's doing.
 *   2)  We check to see if the next packet pointer we calculated based on
 *       the length is the same length the chip reported
 *   3)  Make sure the length isn't too short
 *   4)  Make sure the length isn't too long
 */
    if(((frameInfo.status & 0x1f) == 0x01) &&
       (frameInfo.nextPacket == nextFrameTest) &&
       (frameInfo.length >= TM_8390_MIN_ETHER_PACKET) &&
       (frameInfo.length <= TM_8390_MAX_ETHER_PACKET))
    {
/* The header info is ok, so get a buffer to place the data into */
        *dataPtr = tfGetEthernetBuffer(bufHandlePtr);
/* We couldn't get a buffer; return an error */
        if(*dataPtr == (char TM_FAR *)0)
        {
            retCode = TM_DEV_ERROR;
            goto ne2kRecvEnd;
        }

/* Get the packet */
        tfNE2kBlockInput((unsigned short int) frameInfo.length,
                         (unsigned short int)
                         ((thisFrame << 8) + sizeof(frameInfo)),
                         (unsigned char TM_FAR *)*dataPtr);

/* Store the data length */
        *dataLength = frameInfo.length;
/* Set the next-packet pointer */
        tlNe2kPtr->currentRcvPage = frameInfo.nextPacket;
    }
    else
    {
/*
 * The frame is bad. We're not sure how many frames got screwed up, so just
 * clear all received frames out
 */
        tlNe2kPtr->currentRcvPage = rxPage;
        retCode = TM_DEV_ERROR;
    }

/*
 * Store the boundary page (which, in turn, tells us where the next page to
 * read lives)
 */
    if ( tlNe2kPtr->currentRcvPage ==    TM_NE2K_BUFFER_START
                                          + TM_NE2K_TX_PAGES )
    {
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY,
                          TM_NE2K_BUFFER_STOP - 1 );
    }
    else
    {
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_BNRY,
                          tlNe2kPtr->currentRcvPage - 1 );
    }


/*
 * If the NIC was stopped due to a buffer overflow, check for a resend, exit
 * loopback, and then resend if required.
 */
    if (tlNe2kPtr->overflowFlags & TM_8390_ISR_OVW)
    {
/* Take chip out of loopback */
        tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TCR, 0);
        if (tlNe2kPtr->overflowFlags & TM_8390_CR_TXP)
        {
/* If we need to retransmit, do so */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                                TM_8390_CR_RD2
                              + TM_8390_CR_STA
                              + TM_8390_CR_TXP );
        }
        else
        {
/* Otherwise, just start the chip */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_STA );
        }
/* Clear the 'we had an overflow' flag */
        tlNe2kPtr->overflowFlags = 0;
    }


/*
 * Sometimes, it appears that we get only one interrupt, but multiple packets
 * come in. So here we check if we have any delinquent packets loitering
 * about. If we do, the call to tfNotifyInterfaceTask will generate
 * another call to this function. If we somehow manage to clear out a
 * packet that hasn't been interrupted for yet, but then it DOES get an
 * interrupt, it's no huge problem - this function will return TM_DEV_ERROR
 * to tfRecvInterface, which should just ignore it.
 * NOTE that we CANNOT call tfNotifyInterfaceIsr since we are not inside an
 * ISR.
 */
    if (tlNe2kPtr->currentRcvPage != rxPage)
    {
        tfNotifyInterfaceTask(tlNe2kPtr->ifHandle, 1, 0, 0UL, 0UL);
    }
ne2kRecvEnd:
    return(retCode);
}


/*
 * tfNE2kSend()
 *
 *   This function stores a packet in the NE2000's buffers for sending. If the
 *   chip is not currently busy sending anything, it also triggers a send. If
 *   the chip is busy, the packet send will be triggered in the ISR after all
 *   of the preceding packets in the NE2000's buffers have been sent.
 *
 *   Parameters:
 *       interfaceHandle The handle to the user interface for this device
 *       dataPtr         The pointer to the data we're sending
 *       dataLength      The length of the data we're sending
 *       flag            A flag indicating various things about the data to
                         send
 *   Returns:
 *       success:        TM_DEV_OKAY
 *       error:          TM_DEV_ERROR
 */
int tfNE2kSend ( ttUserInterface interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag )
{
/* The length of the packet, in pages (e.g., 4 pages long) */
    short int lengthInPages;
/* A pointer to the page we're concerned with */
    short int currPage;
/* The return value for this function */
    short int retCode;

/* Deal with 'unused parameter' warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* We assume the return value is TM_DEV_OKAY unless told otherwise */
    retCode = TM_DEV_OKAY;

/*
 * Make sure we're not receiving too many scatter pieces. Should never
 * happen, as the Treck stack makes sure of it
 */
    if (tlNe2kPtr->sendPieces < tlNe2kPtr->maxSendPieces)
    {
/* Save this buffer's length in the array */
        tlNe2kPtr->scatterBuffPtr[tlNe2kPtr->sendPieces].dataLen =
            (unsigned short int) dataLength;
/* Save this buffer's address in the array */
        tlNe2kPtr->scatterBuffPtr[tlNe2kPtr->sendPieces].dataPtr = 
                                                                     dataPtr;
/* Increment the total length of the packet by this buffer's length */
        tlNe2kPtr->packetLength += (short int) dataLength;
/* Increment the number of pieces in this packet by 1 */
        tlNe2kPtr->sendPieces++;
    }
    else
    {
/* Uh-oh, we've been given too many pieces. Should never happen. */
        retCode = TM_DEV_ERROR;
        tfKernelWarning("tfNe2kSend",
                        "Scatter count in open does not match current one\n");
        goto ne2kSendEnd;
    }
/* If this is the last buffer in a packet... */
    if (flag & TM_USER_BUFFER_LAST)
    {
/*
 * Check if there's not enough room for the packet before we hit the end
 * of the send buffers. If not, go back to the beginning of the send buffers.
 */
        if (   tlNe2kPtr->packetLength
               > (int) ((   TM_NE2K_TX_PAGES - tlNe2kPtr->fillSend)
                        * TM_8390_BUFFER_PAGE_LEN ))
        {
/* Wait until the page is unused before we modify it */
            while (tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].isFull)
            {
                tfInterfaceSpinLock(interfaceHandle);
            }
/*
 * We intentionally set the data length here to be too long, even though we're
 * not actually filling out the packet. This will cause the send routine in
 * the ISR to go back to the start of the ring
 */
            tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].packetLen = 
                                                  tlNe2kPtr->packetLength;
/*
 * Set a flag instructing the send routine in the ISR to go back to the
 * beginning of the ring
 */
            tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].isFull = 
                                                      TM_8390_SEND_RING_WRAP;
/* The next page to be filled is page 0 */
            tlNe2kPtr->fillSend = 0;
        }
/* Initialize lengthInPages to 0 */
        lengthInPages = 0;
/*
 * This loop verifies that we have enough free buffers to copy the packet.
 * We first figure out what the last page in the packet is, and then walk
 * through the buffer ring until we have enough free pages to store the
 * packet
 */
        for (currPage = ((tlNe2kPtr->packetLength + 255) /
                         TM_8390_BUFFER_PAGE_LEN) + tlNe2kPtr->fillSend;
             ((short int) tlNe2kPtr->fillSend) < currPage;
             tlNe2kPtr->fillSend++)
        {
/* Wait for the buffer to become free */
            while (tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].isFull)
            {
                tfInterfaceSpinLock(interfaceHandle);
            }
/* Mark it as used (it's not being used quite yet, but it will be very soon) */
            tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].isFull = 1;
/*
 * Fill out the packet length.
 */
            tlNe2kPtr->sendRingPtr[tlNe2kPtr->fillSend].packetLen = 
                                                   tlNe2kPtr->packetLength;
/*
 * Increment lengthInPages by one, so we know how many pages long the packet
 * is when we're done with the loop
 */
            lengthInPages++;
        }

/*
 * Mask out all interrupts in the IMR. Bad things happen if the chip
 * interrupts itself while it's sending
 */
        tm_ne2k_out_byte(TM_NE2K_BASE_ADDRESS + TM_8390_PG0_IMR, 0);

/* Output the packet to the device's buffers */
        tfNE2kBlockOutput((unsigned char)
                          (TM_NE2K_BUFFER_START + (tlNe2kPtr->fillSend - 
                                                   lengthInPages)));

/*
 * If the device isn't currently sending anything, we can go ahead and send
 * this packet. If not, the ISR will pick it up eventually. We need a critical
 * section here to make sure that once we've tested the lock, nobody else
 * can look at it before we've set it to 'locked'
 */
        tm_kernel_set_critical;
        if (tlNe2kPtr->sendingLock == TM_8390_UNLOCKED)
        {
            tlNe2kPtr->sendingLock = TM_8390_LOCKED;
/* Now that it's locked, we can release the critical section */
            tm_kernel_release_critical;
/* Set the chip to page 0 */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 );
/* Set the lower 8 bits of the packet length */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR0,
                              tlNe2kPtr->packetLength & 0xff );
/* Set the upper 8 bits of the packet length */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TBCR1,
                              tlNe2kPtr->packetLength >> 8 );
/* Set the page to start sending from */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_TPSR,
                                TM_NE2K_BUFFER_START
                              + (tlNe2kPtr->fillSend - lengthInPages) );
/* Trigger the send */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                                TM_8390_CR_RD2
                              + TM_8390_CR_TXP
                              + TM_8390_CR_STA );
        }
        else
        {
/* We're done with the comparison, free the critical section */
            tm_kernel_release_critical;
        }

/* Reenable interrupts on the chip */
        tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_IMR,
                            TM_8390_ISR_PRX
                          + TM_8390_ISR_PTX
                          + TM_8390_ISR_OVW );

/* Make sure we haven't run past the end of the ring */
        tlNe2kPtr->fillSend %= TM_NE2K_TX_PAGES;


/*
 * We transferred the current packet, so we reset packetLength and sendPieces
 * back to 0
 */
        tlNe2kPtr->packetLength = 0;
        tlNe2kPtr->sendPieces = 0;
/*
 * We just transferred the packet to buffer memory, so we can send-complete
 * for it.
 */
        tfSendCompleteInterface(tlNe2kPtr->ifHandle,
                                TM_DEV_SEND_COMPLETE_DRIVER);
    }
ne2kSendEnd:
    return(retCode);
}


/* 
 * tfNE2kIoctl()
 *
 *   This function, in general, allows the user to talk directly to the
 *   chip, for any reason. We have implemented it here to allow the
 *   user to set multicast addresses. It should be noted that at the time
 *   this function was implemented, IGMP was not fully tested, therefore
 *   making it impossible to test this function. Use with caution!
 *
 *   TM_DEV_SET_MCAST_LIST:
 *     Adds 'optionLen' number of multicast entries to the multicast hash 
 *     table.  This is done by 
 *         1. Compute a 32-bit CRC on the address.
 *         2. Extract the most significant 6 bits of the computed CRC.
 *         3. Set the bit coresponding to the above number in the multicast
 *            address filter (ie, decode the 6 bit value above into a 64 bit
 *            value: 50 -> 0x0002 0000 0000 0000)
 *
 *   TM_DEV_SET_ALL_MCAST:
 *     Sets the chip to receive all multicast packets (ie, set the address
 *     filter to all ones).
 *
 *   Parameters:
 *       interfaceHandle   The handle to the user interface for this device
 *       flag       A flag what we want this function to do
 *       dataPtr    The pointer to the data being passed in
 *       dataLength The length of the data (in this case, how many multicast
 *                  addresses are being added).
 *   Returns:
 *       TM_DEV_OKAY
 */
int tfNE2kIoctl(ttUserInterface interfaceHandle,
                int             flag,
                void TM_FAR *   optionPtr,
                int             optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/* The calculated CRC */
    unsigned long   calcCRC;
/* The current byte we're looking at */
    unsigned char   currByte;
/* A counter */
    unsigned char   i;
/* Which byte of the ethernet address we're working on */
    unsigned char   byte;
/* Which bit of the current byte we're working on */
    unsigned char   bit;
/*
 * The top 6 bits of the calculated CRC, from which we generate the hash table
 */
    unsigned long   hashEntry;
/* A pointer to the list of multicast addresses passed to us */
    unsigned char * multiAddr;
/* The actual hash table to be written to the chip */
    unsigned char   logicalAddrFilter[TM_8390_MCAST_ADDR_LEN];
    int             savedOptionLen;
#endif /* TM_IGMP or TM_6_USE_MLD */

/* To avoid compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    savedOptionLen = optionLen;
    if (flag & TM_DEV_SET_MCAST_LIST)
    {

/* Zero out the hash table */
        tm_bzero( logicalAddrFilter, TM_8390_MCAST_ADDR_LEN );

/* Get the first address */
        multiAddr = (unsigned char TM_FAR *) optionPtr;

/* While we have more addresses to work on */
        while (optionLen-- > 0)
        {
            
/* Initialize the calculated CRC (with all ones) */
            calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
            for (byte=0; byte<TM_8390_ETHER_ADDR_LEN; byte++)
            {
                currByte = multiAddr[byte];

/* For each bit in the current byte */
                for (bit=0; bit < TM_8390_MCAST_ADDR_LEN; bit++)
                {

/* Shift the calculated CRC left by one */
                    calcCRC <<= 1;

/* If the current bit is one... */
                    if ((currByte >> bit) & 0x01)
                    {

/* ... XOR the calculated CRC with the CRC-32 polynomial */
                        calcCRC ^= TM_NE2K_CRC32_POLY;

                    }                   
                }
            }

/* Get the most significant 6 bits from the calculated CRC */        
            hashEntry = (calcCRC & 0xfc000000UL) >> 26;


/* 
 * Decode the above number and set the appropriate bit in the CS8900 logical
 * address filter 
 */          
            logicalAddrFilter[(unsigned char) (hashEntry >> 3)] |=
                                        1 << ((unsigned char) hashEntry %
                                              TM_8390_MCAST_ADDR_LEN);

/* Increment the pointer to the next address in the list */
            multiAddr += TM_8390_MCAST_ADDR_LEN;

        }

/* If they actually gave us addresses... */
        if (savedOptionLen != 0)
        {
/* Set chip to Page 0 */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2);
/* Set the chip to accept both multicast and broadcast packets */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR,
                              TM_8390_RCR_AM | TM_8390_RCR_AB );
/* Set chip to Page 1 */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 + TM_8390_CR_PS0 );
/* Write the new multicast hash to the multicast address register */
            for (i=0; i < TM_8390_MCAST_ADDR_LEN; i++)
            {
                tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + i,
                                  logicalAddrFilter[i] );
            }
        }
/* Else they want multicast receiving turned off */
        else
        {
/* Disable receiving multicast packets */
/* Set chip to Page 0 */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_CR,
                              TM_8390_CR_RD2 );
/* Set the chip to not accept multicast packets */
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG0_RCR,
                              TM_8390_RCR_AB );
        }

    }

    if (flag & TM_DEV_SET_ALL_MCAST)
    {
/* Set the logical address filter to all ones (receive ALL multicast packets)
*/
        for (i=0; i<TM_8390_MCAST_ADDR_LEN; i++)
        {
            tm_ne2k_out_byte( TM_NE2K_BASE_ADDRESS + TM_8390_PG1_MAR0 + i,
                              0xff );
        }
    }
#endif /* TM_IGMP or TM_6_USE_MLD */

    return(TM_DEV_OKAY);
}


/*
 * tfNE2kGetPhyAddr()
 *
 *   This function retrieves the adapter's physical address.
 *
 *   Parameters:
 *       interfaceHandle  The handle to the user interface for this device
 *       physicalAddress A buffer into which to store the physical address
 *   Returns:
 *       TM_DEV_OKAY
 */
int tfNE2kGetPhyAddr(ttUserInterface interfaceHandle,
                     char TM_FAR *   physicalAddress)
{
/* A counter */
    unsigned char counter;

/* Deal with 'unused parameter' compiler warnings */
    TM_UNREF_IN_ARG(interfaceHandle);

/* Copy the address into the buffer */
    for (counter = 0; counter < TM_ETHERNET_PHY_ADDR_LEN; counter++)
    {
        physicalAddress[counter] = tlNe2kPtr->ethAddr[counter];
    }

    return(TM_DEV_OKAY);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trne2000.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 19535    Output checksum: 28485
 */
