/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * $Source: source/drivers/packet/trpacket.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:16CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * Packet Driver Interface for Treck Real-Time TCP/IP
 *
 * This packet driver interface was taken from the NCSA Telnet
 * Source Code.
 *
 * Since it is a derived work, it still remains the property of
 * NCSA.
 *
 * This version only supports ethernet and only works under Intel
 * real mode, because packet drivers are Intel real mode.
 *
 * Note:  If the driver only receives a few packets and stops
 * make sure that tfPacketIoctl is being called to refill the
 * receive pool.  It should be called either from the main() loop
 * or a task that calls it periodically
 *
 * Last modified by: Mike
 */

/*
 * Includes
 */

#define TM_PD_DEBUG

#include <trsocket.h>

#include <trmacro.h>
#include <trtype.h>
#include <trproto.h> /* for tfFreeSharedBuffer() */

#include <trpacket.h>
#ifdef TM_PD_DEBUG
#include <stdio.h>
#endif /* TM_PD_DEBUG */
#include <dos.h>

/* Call back function macros */
/*
 * Crynwr packet driver can detect if the packet recv ISR function
 * calls IRET, or RET. Other packet drivers just assume that it returns
 * RET.
 */
#ifndef TM_CRYNWR_DRIVER
/* Cannot detect that we call IRET. So we have to call RET */
#define ttPacketCBFuncPtr   ttPacketWrapFuncPtr
#define tfPacketCBFunc      tfPacketRecvCB
#else /* TM_CRYNWR_DRIVER */
/* Can detect that we call IRET */
#define ttPacketCBFuncPtr   ttPacketIsrFuncPtr
#define tfPacketCBFunc      tfPacketRecvIsr
#endif /* TM_CRYNWR_DRIVER */

/*
 * Constants
 */
static const char TM_FAR tlPacketIpType[]={0x08,0x00};
static const char TM_FAR tlPacketArpType[]={0x08,0x06};

/*
 * Other Locals
 */
static ttUserInterface tlInterfaceHandle;
static unsigned char   tlPdDriverClass;
static int             tlIpHandle;
static int             tlArpHandle;

/*
 * Internal Functions
 */

int tfPacketReleaseType(int handle);

int tfPacketAccessType( unsigned char      interfaceClass,
                        unsigned short     interfaceType,
                        unsigned char      interfaceNumber,
                        const char TM_FAR *typePtr,
                        unsigned short     typeLength,
                        int TM_FAR        *errorCodePtr,
                        ttPacketCBFuncPtr  recvFuncPtr);

#ifndef TM_CRYNWR_DRIVER
/* call back function for non CRYNWR packet driver */
void tfPacketRecvCB(void);
#endif /* TM_CRYNWR_DRIVER */

/*
 * This routine is called by the packet driver at
 * ISR time
 */
#ifdef __BORLANDC__
void interrupt tfPacketRecvIsr(unsigned short bp,
                               unsigned short di,
                               unsigned short si,
                               unsigned short ds,
                               unsigned short es,
                               unsigned short dx,
                               unsigned short cx,
                               unsigned short bx,
                               unsigned short ax);
#else /* _MSC_VER */
void _interrupt TM_FAR tfPacketRecvIsr(unsigned short es,
                                       unsigned short ds,
                                       unsigned short di,
                                       unsigned short si,
                                       unsigned short bp,
                                       unsigned short sp,
                                       unsigned short bx,
                                       unsigned short dx,
                                       unsigned short cx,
                                       unsigned short ax,
                                       unsigned short ip,
                                       unsigned short cs,
                                       unsigned short flags);
#endif /* _MSC_VER */

/*
 * Error codes returned by packet drivers
 */
/*
 * ErrorCode        Meaning
 *  0         "No error at all."
 *  1         "Invalid handle number"
 *  2         "No interfaces of specified class found"
 *  3         "No interfaces of specified type found"
 *  4         "No interfaces of specified number found"
 *  5         "Bad packet type specified"
 *  6         "This interface does not support multicast"
 *  7         "This packet driver cannot terminate"
 *  8         "An invalid receiver mode was specified"
 *  9         "Operation failed because of insufficient space"
 * 10         "The type had previously been accessed, and not released."
 * 11         "The command was out of range, or not implemented"
 * 12         "The packet couldn't be sent (usually hardware error)"
 * 13         "Hardware address couldn't be changed (more than 1 handle open)"
 * 14         "Hardware address has bad length or format"
 * 15         "Couldn't reset/initialize interface (more than 1 handle open)"
 * 16         "An invalid iocb was specified"
 */

#define TM_PD_ENOTRELEASED  10

/*
 * Open the Packet driver
 * Registers for receiving IP and ARP packets
 */
int tfPacketOpen(ttUserInterface interfaceHandle)
{
    ttPacketDriverInfo     info;
    int                    errorCode;
/* x86 Registers */
    TM_REGS_TYPE           inRegs;
    TM_REGS_TYPE           outRegs;
    TM_SEG_REGS_TYPE       segRegs;
#ifdef TM_PD_DEBUG
    unsigned long TM_FAR * vectorPtr;
#endif /* TM_PD_DEBUG */

/* Initialize the receive pool */
    errorCode = tfPoolCreate( interfaceHandle,
                              TM_PD_RECV_BIG_BUFFERS,
                              TM_PD_RECV_SMALL_BUFFERS,
                              TM_ETHER_MAX_PACKET_CRC,
                              TM_PD_SMALL_BUFFER_SIZE,
                              0,
                              TM_POOL_REFILL_IN_LINE );
    if (errorCode != TM_ENOERROR)
    {
        goto packetOpenEnd;
    }
    tlInterfaceHandle = interfaceHandle;
    tlArpHandle = TM_PD_ERROR;
    tlIpHandle = TM_PD_ERROR;

    inRegs.x.ax = TM_PD_FUNCTION_INFO;
    inRegs.x.bx = 0;

/*
 * Check to make sure the packet driver is registered at
 * TM_PD_INTERRUPT_NUMBER
 */
#ifdef TM_PD_DEBUG
    vectorPtr = (unsigned long TM_FAR *)
                                 ((unsigned long) TM_PD_INTERRUPT_NUMBER * 4UL);
    if (*vectorPtr == 0UL)
    {
        errorCode = TM_DEV_ERROR;
        printf("tfPacketOpen: packet driver interrupt vector 0x%02X is NULL\n",
                TM_PD_INTERRUPT_NUMBER);
        goto packetOpenFinish;
    }
#endif /* TM_PD_DEBUG */
    tm_software_irq_x(TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs, &segRegs);
    if (outRegs.x.cflag != 0)
    {
        errorCode = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: no packet driver at 0x%x, error %d\n",
                TM_PD_INTERRUPT_NUMBER, errorCode);
#endif /* TM_PD_DEBUG */
        goto packetOpenFinish;
    }
    else
    {
/*
 * If we did not have an error, save the information for the
 * tfPacketAccessType() call
 */
        errorCode = TM_PD_ENOERROR;
        info.pdVersion = outRegs.x.bx;
        info.pdClass   = outRegs.h.ch;
        info.pdType    = outRegs.x.dx;
        info.pdNumber  = outRegs.h.cl;
        info.pdNamePtr  = (unsigned char TM_FAR *)
                                     (tm_make_far_ptr(segRegs.ds,outRegs.x.si));
        info.pdFunctionality = outRegs.h.al;
#ifdef TM_PD_DEBUG
 printf("Version=%X\nClass=%X\nType=%X\nNumber=%XName=%s\nFunctionalty=%X\n",
            info.pdVersion,
            info.pdClass,
            info.pdType,
            info.pdNumber,
            info.pdNamePtr,
            info.pdFunctionality);
#endif /* TM_PD_DEBUG */
    }
/* Check to make sure that it has functionality */
    if (info.pdFunctionality == TM_PD_FUNC_NOT_INSTALLED)
    {
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: Error Not Installed!\n");
#endif /* TM_PD_DEBUG */
        errorCode = TM_EINVAL;
        goto packetOpenFinish;
    }

    tlPdDriverClass = info.pdClass;
    if (tlPdDriverClass != TM_PD_CLASS_ETHERNET)
    {
        errorCode = TM_EINVAL;
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: Packet Driver Not Ethernet!\n");
#endif /* TM_PD_DEBUG */
        goto packetOpenFinish;
    }
/* Register with the packet driver that we will handle IP datagrams */
    tlIpHandle = tfPacketAccessType(info.pdClass,
                                    info.pdType,
                                    info.pdNumber,
                                    tlPacketIpType,
                                    TM_PD_IP_LENGTH,
                                    &errorCode,
                                    tfPacketCBFunc);
/* Make sure that the register worked */
    if (tlIpHandle == TM_PD_ERROR)
    {
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: ipHandle error %d\n", errorCode);
#endif /* TM_PD_DEBUG */
        goto packetOpenFinish;
    }
/* Register with the packet driver that we will handle ARP datagrams */
    tlArpHandle = tfPacketAccessType(info.pdClass,
                                     info.pdType,
                                     info.pdNumber,
                                     tlPacketArpType,
                                     TM_PD_ARP_LENGTH,
                                     &errorCode,
                                     tfPacketCBFunc);
/* Make sure that the register worked */
    if (tlArpHandle == TM_PD_ERROR)
    {
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: arpHandle error %d\n", errorCode);
#endif /* TM_PD_DEBUG */
        goto packetOpenFinish;
    }
packetOpenFinish:
    if (errorCode != TM_ENOERROR)
    {
        (void)tfPoolDelete(interfaceHandle);
    }
packetOpenEnd:
    return (errorCode);
}

/*
 * Close the packet driver
 */
int tfPacketClose(ttUserInterface interfaceHandle)
{
    int             errorCode;

    errorCode = TM_PD_ENOERROR;
/* release IP Packet Handling */
    if (tlIpHandle >= 0)
    {
        errorCode = tfPacketReleaseType(tlIpHandle);
#ifdef TM_PD_DEBUG
         if (errorCode != TM_PD_ENOERROR)
        {
            printf("tfPacketClose IP handle: %d ReleaseType error=%x\n",
                    tlIpHandle, errorCode);
        }
#endif /* TM_PD_DEBUG */
    }
/* Release ARP Packet Handling */
    if (tlArpHandle >= 0)
    {
        errorCode = tfPacketReleaseType(tlArpHandle);
#ifdef TM_PD_DEBUG
        if (errorCode != TM_PD_ENOERROR)
        {
            printf("tfPacketClose ARP hanle: %d ReleaseType error=%x\n",
                    tlArpHandle, errorCode);
        }
#endif /* TM_PD_DEBUG */
    }
    (void)tfPoolDelete(interfaceHandle);
    return (errorCode);
}

/*
 * Send a packet to the packet driver
 * Packet drivers only support contiguous data
 * so the flag is always the last packet
 */
int tfPacketSend(ttUserInterface interfaceHandle,
                 char    TM_FAR *dataPtr,
                 int             dataLength,
                 int             flag)
{
    int              errorCode;
    TM_REGS_TYPE     inRegs;
    TM_REGS_TYPE     outRegs;
    TM_SEG_REGS_TYPE segRegs;

    TM_UNREF_IN_ARG(flag);

/* Our function code is send a packet */
    inRegs.x.ax = TM_PD_FUNCTION_SEND;
/* Pass the length */
    inRegs.x.cx = dataLength;
/* Pass the segment/offset of dataPtr */
    segRegs.ds  = tm_far_ptr_segment(dataPtr);
    inRegs.x.si = tm_far_ptr_offset(dataPtr);

/* Call the packet driver to Send the packet */
    tm_software_irq_x(TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs, &segRegs);

/* Check for error by checking the carry bit */
    if (outRegs.x.cflag != 0)
    {
        errorCode = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: Error Sending %d",errorCode);
#endif /* TM_PD_DEBUG */
    }
    else
    {
        errorCode = TM_PD_ENOERROR;
/* Can only call tfSendCompleteInterface, if we return no error */
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }
    return (errorCode);
}

/*
 * Get the physical address for the packet driver
 */
int tfPacketGetPhyAddr( ttUserInterface interfaceHandle,
                        char TM_FAR *physicalAddress )
{
    int              errorCode;
    TM_REGS_TYPE     inRegs;
    TM_REGS_TYPE     outRegs;
    TM_SEG_REGS_TYPE segRegs;

    TM_UNREF_IN_ARG(interfaceHandle);

    inRegs.x.ax = TM_PD_FUNCTION_GET_ADDR;
    inRegs.x.bx = tlIpHandle;
    inRegs.x.cx = TM_PD_ETHER_ADDR_LEN;
    segRegs.es  = tm_far_ptr_segment(physicalAddress);
    inRegs.x.di = tm_far_ptr_offset(physicalAddress);

/* Call the packet driver to get the physical address */
    tm_software_irq_x(TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs, &segRegs);

    if (outRegs.x.cflag != 0)
    {
        errorCode = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
#ifdef TM_PD_DEBUG
        printf("tfPacketOpen: Error Getting Ether Address %d",errorCode);
#endif /* TM_PD_DEBUG */
    }
    else
    {
        errorCode=TM_PD_ENOERROR;
    }

    return (errorCode);
}

/*
 * Receive a packet back into Treck
 */
int tfPacketRecv(ttUserInterface  interfaceHandle,
                 char TM_FAR    **dataPtrPtr,
                 int TM_FAR      *dataLengthPtr,
                 ttUserBufferPtr  userBufferHandlePtr)
{
    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          userBufferHandlePtr );
}

/*
 * General function
 * Can be used to configure the device
 * or refill the receive pool
 */
int tfPacketIoctl ( ttUserInterface interfaceHandle, int flag,
                    void TM_FAR * optionPtr, int optionLen )
{

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    TM_REGS_TYPE      inRegs;
    TM_REGS_TYPE      outRegs;
    TM_SEG_REGS_TYPE  segRegs;
#endif /* TM_IGMP or TM_6_USE_MLD */
    int               errorCode;

    TM_UNREF_IN_ARG(interfaceHandle);
    errorCode = TM_ENOERROR;
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    if (flag == TM_DEV_SET_MCAST_LIST)
    {
/* set multicast list */
#define TM_PD_FUNCTION_SET_MCAST 0x1600
/* Our function code is set multicast list */
        inRegs.x.ax = TM_PD_FUNCTION_SET_MCAST;
/* Pass the length */
        inRegs.x.cx = optionLen * 6;
/* Pass the segment/offset of optionPtr */
        segRegs.es  = tm_far_ptr_segment(optionPtr);
        inRegs.x.di = tm_far_ptr_offset(optionPtr);

/* Call the packet driver to Send the packet */
        tm_software_irq_x( TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs,
                           &segRegs );

/* Check for error by checking the carry bit */
        if (outRegs.x.cflag != 0)
        {
            errorCode = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
#ifdef TM_PD_DEBUG
            printf("tfPacketOpen: Error Setting multicast list %d",errorCode);
#endif /* TM_PD_DEBUG */
        }
    }
    if (flag == TM_DEV_SET_ALL_MCAST)
    {
        errorCode = TM_EPROTONOSUPPORT;
    }
#else /* !(TM_IGMP or TM_6_USE_MLD) */

    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);
    TM_UNREF_IN_ARG(flag);

#endif /* !(TM_IGMP or TM_6_USE_MLD) */
    return errorCode;
}

/*
 * Called by the packet driver during an ISR
 * First Call: get a memory buffer to write into
 * Second Call: Mark the receive as complete
 */
#ifdef __BORLANDC__
void interrupt tfPacketRecvIsr(
                                unsigned short bp,
                                unsigned short di,
                                unsigned short si,
                                unsigned short ds,
                                unsigned short es,
                                unsigned short dx,
                                unsigned short cx,
                                unsigned short bx,
                                unsigned short ax)
#else /* _MSC_VER */
void _interrupt TM_FAR tfPacketRecvIsr(unsigned int es,
                                       unsigned int ds,
                                       unsigned int di,
                                       unsigned int si,
                                       unsigned int bp,
                                       unsigned int sp,
                                       unsigned int bx,
                                       unsigned int dx,
                                       unsigned int cx,
                                       unsigned int ax,
                                       unsigned int ip,
                                       unsigned int cs,
                                       unsigned int flags)
#endif /* _MSC_VER */
{
    char TM_FAR *dataPtr;

/* get rid of compiler warnings */
    TM_UNREF_IN_ARG(ds);
    TM_UNREF_IN_ARG(si);
    TM_UNREF_IN_ARG(bp);
    TM_UNREF_IN_ARG(bx);
    TM_UNREF_IN_ARG(dx);
    TM_UNREF_IN_ARG(es);
    TM_UNREF_IN_ARG(di);
#ifdef _MSC_VER
    TM_UNREF_IN_ARG(sp);
    TM_UNREF_IN_ARG(ip);
    TM_UNREF_IN_ARG(cs);
    TM_UNREF_IN_ARG(flags);
#endif /* _MSC_VER */


    if (!ax)
    {
/* First Call: Get a pre-allocated buffer from the Treck interface recv pool */
        dataPtr = (char TM_FAR *) tfPoolIsrGetBuffer( tlInterfaceHandle,
                                                      cx );
        if ( dataPtr != (char TM_FAR *)0 )
        {
/* Space available so return it */
            es = tm_far_ptr_segment(dataPtr);
            di = tm_far_ptr_offset(dataPtr);
        }
        else
        {
/* No Space */
            es = 0;
            di = 0;
        }
    }
    else
    {
/* Notify the treck stack about the recv buffer */
        tfNotifyInterfaceIsr(tlInterfaceHandle, 1, 0, 0UL, 0UL);
    }
    return;
}


/*
 * Register an Access Type
 * Tell the Packet Driver what kind of packets that we will process
 */
int tfPacketAccessType( unsigned char      interfaceClass,
                        unsigned short     interfaceType,
                        unsigned char      interfaceNumber,
                        const char TM_FAR *typePtr,
                        unsigned short     typeLength,
                        int TM_FAR        *errorCodePtr,
                        ttPacketCBFuncPtr  recvFuncPtr)
{
    TM_REGS_TYPE      inRegs;
    TM_REGS_TYPE      outRegs;
    TM_SEG_REGS_TYPE  segRegs;
    int               handle;

/* Save the function code for accessType */
    inRegs.x.ax = TM_PD_FUNCTION_ACCESS_TYPE|interfaceClass;
/* Pass the parameters in the registers */
    inRegs.x.bx = interfaceType;
    inRegs.h.dl = interfaceNumber;
    segRegs.ds  = tm_far_ptr_segment(typePtr);
    inRegs.x.si = tm_far_ptr_offset(typePtr);
    inRegs.x.cx = typeLength;
    segRegs.es  = tm_far_ptr_segment(recvFuncPtr);
    inRegs.x.di = tm_far_ptr_offset(recvFuncPtr);

/* Call the packet driver to register this access type */
    tm_software_irq_x(TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs, &segRegs);

/* Check for error by checking the carry bit */
    if (outRegs.x.cflag != 0)
    {
/* Return the errorCode */
        *errorCodePtr = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
        handle = TM_PD_ERROR;
    }
    else
    {
        handle = (int)(outRegs.x.ax);
    }

    return (handle);
}

/*
 * Release this network layer
 */
int tfPacketReleaseType(int handle)
{
    int              errorCode;
    TM_REGS_TYPE     inRegs;
    TM_REGS_TYPE     outRegs;

/* Function is Release type */
    inRegs.x.ax = TM_PD_FUNCTION_RELEASE_TYPE;
/* Handle to Release type */
    inRegs.x.bx = (unsigned short)handle;

/* Call the packet driver to Release this handle */
    tm_software_irq(TM_PD_INTERRUPT_NUMBER, &inRegs, &outRegs);

/* Check the carry bit for an error */
    if (outRegs.x.cflag != 0)
    {
        errorCode = (int)(outRegs.h.dh) & 0xFF; /* dh is error code */
    }
    else
    {
        errorCode = TM_PD_ENOERROR;
    }
    return (errorCode);
}


#ifndef TM_CRYNWR_DRIVER
/*
 * Since the tfPacketRecvIsr routine is defined with the 'interrupt'
 * specifier, it performs an IRET rather than a normal RETF.  This causes 
 * problems with some packet drivers, most notably the 3c509 packet driver from
 * 3COM.  This wrapper function solves the problem by pushing the flags before
 * calling tfPacketRecvIsr, and then doing a normal RETF back to the device's
 * packet driver.  The MSC compiler attempts to save the SI/DI registers, which
 * is a problem since the DI register is used to return data to the device's 
 * PD.  To solve this we remove the values for SI/DI that MSC has placed on the
 * stack and replace them with our own.
 */
void tfPacketRecvCB( void )
{

    _asm
    {
        pushf
        call far ptr tfPacketRecvIsr
#ifdef _MSC_VER
        pop ax
        pop ax
#ifdef _DEBUG
#error This driver does not support receiving packets in _DEBUG mode!
#else /* !_DEBUG */
        push di
        push si
#endif /* !_DEBUG */
#endif /* _MSC_VER */
    }

    return;
}
#endif /* TM_CRYNWR_DRIVER */


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trpacket.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 23648    Output checksum:  1006
 */
