/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: General PCI routines
 * Filename: trpci.c
 * Author: Jason
 * Date Created: 3/14/2000
 * $Source: source/drivers/pci/trpci.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:17CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * This module includes routines and macros that provide very basic PCI 
 * configuration functionality; primarily find the device, and retrieving 
 * base addresses and interrupt line information.  This module is used for
 * the Intel 8255x Ethernet device driver.
 *
 * For a good overview of the PCI bus archictecture, please refer to:
 *      "PCI System Archicecture, Fourth Edition", by Shanley and Anderson.
 */

/* For _outp, _inp, etc (Microsoft C compiler) */
#include <conio.h>

#include <trpci.h>
#include <trsocket.h>

/*
 * Specifies whether a base address lies in memory-mapped register space or
 * I/O register space.
 */
#define TM_PCI_IO_SPACE  0x00000001UL
#define TM_PCI_MEM_SPACE 0x00000000UL


/*
 * Maximum number of buses, functions, etc 
 */
#define TM_PCI_MAX_BUSES   1
#define TM_PCI_MAX_FUNCS   8
#define TM_PCI_MAX_DEVS   32


/*
 * tfPciFindDevice()
 *
 *   Returns the bus, device and function of the specified device.  This 
 *   function simply iterates through every possible device on the PCI bus to 
 *   find the one we're looking for.  The way is guaranteed to work; however, 
 *   there are probably more efficient ways to do this, such as the PCI BIOS, 
 *   which can perform this functionality in a single call.  However, not every 
 *   system (especially embedded ones) has this capability.  Also, in a 
 *   hard-wired system it is likely that you know the something about the 
 *   device, bus number, etc at compile time, so this function could simply
 *   return these numbers for the specified device.  At the very least it should
 *   be possible to limit the number of buses that this function searches
 *   (currently 256).  However, since this function is only called once upon
 *   startup, its probably not very important that it be incredibly efficient.
 *
 */
int tfPciFindDevice(unsigned short  vendorId,
                    unsigned short  deviceId,
                    unsigned char * bus,
                    unsigned char * device,
                    unsigned char * func)
{
    unsigned char busNum;
    unsigned char devNum;
    unsigned char funcNum;
    unsigned long devVendId;
    unsigned long devClassRev;
    int           errorCode;
    int           count;


    count = 0;
    errorCode = TM_ENODEV;
    for (busNum = 0; busNum < TM_PCI_MAX_BUSES; busNum++)
    {
        for (devNum = 0; devNum < TM_PCI_MAX_DEVS; devNum++)
        {
            for (funcNum = 0; funcNum < TM_PCI_MAX_FUNCS; funcNum++)
            {

/* Read device and vendor ID from current PCI configuration block */
                tm_pci_read_conf_dword(busNum,devNum,funcNum,
                                       TM_PCI_VENDDEV_ID,devVendId);
                tm_pci_read_conf_dword(busNum,devNum,funcNum,
                                       TM_PCI_CLASS_REV,devClassRev);

/* 
 * Check to see if device and vendor IDs match, if so break from loop and
 * return bus, device and function numbers.
 */
                if (((unsigned short)((devVendId >> 16) & 0x0000FFFF) ==
                                                                   deviceId) &&
                   ((unsigned short) (devVendId & 0x0000FFFF) == vendorId))
                {
                    *bus    = busNum;
                    *device = devNum;
                    *func   = funcNum;

                    errorCode = TM_ENOERROR;

                    goto PCI_FIND_RETURN;
                    
                }
            }
        }
    }   

PCI_FIND_RETURN:
    return errorCode;
} 
 

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trpci.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 64110    Output checksum: 29703
 */
