/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: PRISM 802.11b device driver
 *
 * Filename: trprism.c
 * Author: Jason Schmidlapp
 * Date Created: 4/9/2003
 * $Source: source/drivers/prism/trprism.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:19CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * This was originally tested using a D-LINK DWL-520 PCI board (which uses the
 * PRISM 2.5 chipset) on an Intel XScale IQ80310 development board.
 *
 * This implementation is for stations only - that is, it does not include any
 * Access Point (AP) specific functionality.
 *
 * Specific 802.11b options, such as the desired SSID, can be set through
 * tfIoctlInterface.  This API can also be used to retrieve information about
 * the current state of the wireless connections.  The options available to
 * tfIoctlInterface are outlined below:
 *
 * TM_PRISM_IOCTL_SET_PORT_TYPE
 *   option type: int
 *   possible values: TM_PRISM_AD_HOC_MODE, TM_PRISM_INFRASTRUCTURE_MODE,
 *                    TM_PRISM_AUTO_PORT_MODE
 * Specifies the mode of station operation: Ad-hoc for use on networks without
 * an Access Point (AP) or Infrastructure mode when an AP is to be used.  Auto
 * port mode will first attempt to contact an AP and if this fails, will use
 * an ad-hoc network.
 *
 * TM_PRISM_IOCTL_SET_IBSS_CHAN
 *   option type: unsigned char
 *   possible values: 0 - 14
 * Sets the communications channel to use when this station creates and IBSS
 * in ad-hoc or auto-port mode.
 *
 * TM_PRISM_IOCTL_SET_SSID
 *   option type: null terminated string
 * The Service Set Identifier ("network name") that the station should attempt
 * to connect to.  This parameter may also be null or a zero length string; in
 * this case the station will attempt to connect to any SSID that is found.
 *
 * TM_PRISM_IOCTL_SET_OWN_SSID
 *   option type: null terminated string
 * The SSID to be used when the station creates an IBSS in ad-hoc or
 * auto-port mode.
 *
 * TM_PRISM_IOCTL_SET_WEP_KEY
 *   option type: unsigned char buffer. 5 bytes for 64-bit encryption; 13 byte
 *                                      for 128-bit encryption.
 * Specifies the encryption key to use when WEP is enabled.  WEP may operate in
 * either 64-bit or 128-bit mode, depending on the configuration of the AP.
 * For 64-bit mode, you must specify a 40-bit key.  For 128-bit mode, you must
 * specify a 104-bit key.  This key is shared by the station and the AP, and
 * so is probably assigned by the network administrator.  If you wish to use
 * WEP, after setting the encryption key, you should enable WEP with the
 * TM_PRISM_IOCTL_SET_WEP_ENABLE option (see below).
 *
 * TM_PRISM_IOCTL_SET_WEP_ENABLE
 *   option type: unsigned char
 *   possible values: boolean
 * A boolean value that either enables or disables the use of WEP.
 *
 * TM_PRISM_IOCTL_GET_PORT_STATUS
 *   option type: int
 *   possible values:
 *     TM_PRISM_PORT_DISABLED:     Port is disabled.
 *     TM_PRISM_PORT_SEARCHING:    Station is searching for initial connection.
 *     TM_PRISM_PORT_CONN_IBSS:    Ad-hoc mode: connected to IBSS
 *     TM_PRISM_PORT_CONN_ESS:     Infrastructure mode: connected to BSS
 *     TM_PRISM_PORT_OUT_OF_RANGE: Infrastructure mode: AP is out of range
 *     TM_PRISM_PORT_CONN_WDS:     Shouldn't occur (not implemented)
 *     TM_PRISM_PORT_HOST_AP:      Should never occur, since this is a station
 *                                 only implementation.
 * This option retrieves the current status of the wireless connection.
 *
 * TM_PRISM_IOCTL_GET_SSID
 *   option type: string buffer, at least 33 bytes long
 * Gets the Service Set Identifier to which the station is connected.
 *
 * TM_PRISM_IOCTL_GET_BSSID
 *   option type: string buffer, at least 6 bytes long
 * Gets the ID of the Basic Service Set (BSS) that the station is currently
 * connected.  For instance, the MAC address of the AP which the station is
 * using.
 *
 * TM_PRISM_IOCTL_GET_COMM_QUALITY
 *   option type: int
 *   possible values: 0-92
 * This option indicates the quality of communicates between the station and
 * the current BSS.  This value is based on average signal level and average
 * noise level recorded by the firmware.
 *
 * TM_PRISM_IOCTL_GET_TX_RATE
 *   option type: int
 *   possible values: 0-100
 * Gets the current transmission data rate, in Mbits/sec.
 * 
 *
 * There are a few platform dependencies in this device driver that would
 * have to be addressed when porting this driver to a new target.  Of course,
 * there dependencies have been kept to a minimum...
 *    1. tfPrismSetupPci()
 *       This routine scans the PCI bus for the PRISM device, and does
 *       any setup necessary for the driver to access the device over the
 *       PCI bus.
 *    2. tm_prism_out_word, tm_prism_in_word.
 *       These are the fundamental macros used by the driver to read and write
 *       registers on the PRISM controller.  Currently, they are implemented
 *       to access memory-mapped registers, based on the address returned by
 *       the PCI scan.  A new target may do this differently - for instance,
 *       may use programmed I/O instead.
 *    3. Byte ordering.
 *       This driver was developed on a little endian platform (Intel XScale).
 *       To port to a big endian target, the tm_prism_out_word and
 *       tm_prism_in_word macros would need to be modified to swap the words
 *       before writing/reading to the PRISM, since the PRISM chipset is little
 *       endian.  (Note that some big endian processors (e.g. some PowerPC's)
 *       do all programmed I/O in a little endian fashion)
 *    4. Interrupts.
 *       tfPrismOpen should setup an interrupt handler for the PRISM
 *       controller.  For an example of this, see where
 *       tfKernelInstallIsrHandler is called for the XScale IQ80310 platform.
 *
 * 
 * Design notes:
 * 1. The PRISM controller has two BAP's (buffer access paths) available for
 *    use, numbered 0 and 1.  All data accesses from the ISR (e.g. receiving
 *    data) is done through BAP #1, while all other data accesses (e.g. sending
 *    data) is done thorugh BAP #0).  This is to prevent concurrent accesses
 *    of the same BAP.
 */ 

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

#include <trprism.h>

#ifdef TM_XSCALE_IQ80310
#include <iq80310.h> 
#include <pci_bios.h>
#endif /* TM_XSCALE_IQ80310 */

/*
 * Default mode, either Ad-Hoc (no Access Point) or Infrastructure (use AP).
 * This can be changed at runtime using an ioctl option.
 */
/* #define TM_PRISM_DEFAULT_MODE  TM_PRISM_AD_HOC_MODE */
#define TM_PRISM_DEFAULT_MODE  TM_PRISM_INFRASTRUCTURE_MODE
/* #define TM_PRISM_DEFAULT_MODE  TM_PRISM_AUTO_PORT_MODE */

/* Default desired SSID. */
#define TM_PRISM_DEFAULT_SSID "elmic_wlan" /* max length = 32 bytes */

/* Default SSID to use in ad-hoc mode when creating a BSS */
#define TM_PRISM_DEFAULT_OWN_SSID "elmic_wlan" /* max length = 32 bytes */

/* Default channel to use in ad-hoc mode. */
#define TM_PRISM_DEFAULT_BSS_CHAN 7

/* Bit map specifying the supported transmission rates: */
#define TM_PRISM_DEFAULT_TX_RATES   TM_PRISM_TX_RATE_1_MBPS  \
                                  | TM_PRISM_TX_RATE_2_MBPS  \
                                  | TM_PRISM_TX_RATE_5_MBPS  \
                                  | TM_PRISM_TX_RATE_11_MBPS

/*
 * Our receive pool consists of two sized buffers: small and large.
 * These macros determine the size and number of these buffers in our
 * pre-allocated receive pool.
 */
#define TM_PRISM_LARGE_RECV_POOL     4
#define TM_PRISM_SMALL_RECV_POOL     10
#define TM_PRISM_LARGE_RECV_BUF_SIZE 1518
#define TM_PRISM_SMALL_RECV_BUF_SIZE 128

/* 
 * This macro determines on what sort of byte boundary our buffers are aligned 
 * (word, long word, etc).
 */
#define TM_PRISM_BUF_ALIGNMENT       4

/* Maximum number of pre-allocated transmit FID's */
#define TM_PRISM_MAX_XMIT_FIDS 10

/* Maximum number of scattered buffers per transmitted frame. */
#define TM_PRISM_BUFS_PER_FRAME 5

/* Writes a word to the PRISM memory-mapped registers */
#define tm_prism_out_word(value, offset) \
    *((unsigned short *) (tlPrismInfoPtr->prismMemAddrPtr+offset)) = value

/* Reads a word from the PRISM memory-mapped registers */
#define tm_prism_in_word(offset) \
    *((unsigned short *) (tlPrismInfoPtr->prismMemAddrPtr+offset))


/***** END USER CONFIG ******/


/* Waits for the busy bit in the specified register to be cleared. */
#define tm_prism_wait_not_busy(regNum, tempCmdReg)      \
do                                                      \
{                                                       \
    tempCmdReg = tm_prism_in_word(regNum);              \
} while (tempCmdReg & TM_PRISM_REG_CMD_BUSY)

/* Waits for the command to complete, and then acknowledge the command. */
#define tm_prism_cmd_wait_and_ack(tempCmdReg, result)               \
do                                                                  \
{                                                                   \
    tempCmdReg = tm_prism_in_word(TM_PRISM_REG_EVSTAT);             \
} while (!(tempCmdReg & TM_PRISM_REG_EVSTAT_CMD));                  \
result = tm_prism_in_word(TM_PRISM_REG_STATUS);                     \
tm_prism_out_word(TM_PRISM_REG_EVACK_CMD, TM_PRISM_REG_EVACK)

/* Moves the index to the next item in the queue. */
#define tm_prism_inc_queue_index(index) \
index = ((index) + 1) % TM_PRISM_MAX_XMIT_FIDS

/* Tests whether an error occured while writing to the offset register. */
#define tm_prism_offset_reg_error(reg) ((reg) & 0x4000)

/*
 * Debug output macros.
 * To view debugging output via printf, define TM_PRISM_DEBUG_PRINTF.
 */ 

#ifdef TM_PRISM_DEBUG_PRINTF
#define tm_prism_debug_output(debugString) \
printf(debugString)
#else /* !TM_PRISM_DEBUG_PRINTF */
#define tm_prism_debug_output(debugString)
#endif /* TM_PRISM_DEBUG_PRINTF */

#ifdef TM_PRISM_DEBUG_PRINTF
#define tm_prism_debug_output_short(debugString, debugShort) \
printf(debugString, debugShort)
#else /* !TM_PRISM_DEBUG_PRINTF */
#define tm_prism_debug_output_short(debugString, debugShort)
#endif /* TM_PRISM_DEBUG_PRINTF */

/* constant declarations and enumerations */

/*
 * Type definitions
 */

/* PRISM driver state information. */
typedef struct tsPrismInfo
{

/* Turbo Treck interface handle. */
    ttUserInterface        prismInterfaceHandle;

/* Base memory address for PRISM registers (from PCI) */
    unsigned char TM_FAR * prismMemAddrPtr;

/* List of scattered buffers in the current frame being transmitted. */
    unsigned char TM_FAR * prismScatDataPtr[TM_PRISM_BUFS_PER_FRAME];

/* Lengths of the scattered buffers in the current frame. */
    unsigned int           prismScatDataLen[TM_PRISM_BUFS_PER_FRAME];
    
/* Base I/O port address for PRISM registers (from PCI) */
    unsigned long          prismIoAddr;

/* Number of buffers in the current frame. */
    unsigned int           prismScatIndex;

/* Total size of all buffers in the current frame. */
    unsigned int           prismScatFrameSize;

/* Queue of PRISM controller FID's used to transmit frames. */
    int                    prismXmitFidQueue[TM_PRISM_MAX_XMIT_FIDS];

/* Head of PRISM transmit FID queue. */
    int                    prismXmitFidHead;

/* Tail of PRISM transmit FID queue. */
    int                    prismXmitFidTail;
    
} ttPrismInfo;

typedef ttPrismInfo TM_FAR * ttPrismInfoPtr;

/* 802.3 LLC + SNAP header. */
typedef struct tsLlcSnapHeader
{
    unsigned char  llcDsap;
    unsigned char  llcSsap;
    unsigned char  llcCtl;
    unsigned char  snapOui[3];
    unsigned short snapType;
} ttLlcSnapHeader;

#define TM_LLC_SNAP_HDR_LEN 8

/* PRISM transmit control structure. */
typedef struct tsPrismTransmitFrame
{
    
/* Frame transmit status. */
    unsigned short xmitStatus;

/* Reserved, must be set to zero. */
    unsigned short xmitReserved1;
    unsigned short xmitReserved2;

/* Reserver for software, unused by this driver. */
    unsigned short xmitSwSupport1;
    unsigned short xmitSwSupport2;

/* Frame rate to transmit frame - AP only (so unused by this driver). */
    unsigned short xmitTxRate;

/* Frame processing controls: port number, frame format, interrupt flags. */
    unsigned short xmitTxControl;

/* 802.11b header.  Unused by this driver. */
    unsigned short xmitFrameControl;
    unsigned short xmitDurationId;
    unsigned short xmitAddress1[3];
    unsigned short xmitAddress2[3];
    unsigned short xmitAddress3[3];
    unsigned short xmitSeqControl;
    unsigned short xmitAddress4[3];
    unsigned short xmitDataLen;

/* 802.3 header. */
/* Destination Ethernet address */
    unsigned short xmitDestAddr[3];
/* Source Ethernet address */
    unsigned short xmitSrcAddr[3];
/* Length of Ethernet frame, not including 802.3 header */
    unsigned short xmitLength;
    
} ttPrismTransmitFrame;

#define TM_PRISM_XMIT_CTL_LEN         60
#define TM_PRISM_XMIT_CTL_NO_8023_LEN 46

typedef struct tsPrismReceiveFrame
{

/* Received frame status. */
    unsigned short recvStatus;

/* Timestamp (32-bits, 1 microsecond resolution) */
    unsigned short recvTimestamp1;
    unsigned short recvTimestamp2;

/* The signal and silence levels reported by the RF modem just prior to
   packet reception. */
    unsigned short recvSigSilence;

/* The real bitrate that frame was received at by RF modem. */
    unsigned short recvRate;

/* Reserver - currently unused. */
    unsigned short recvReserved1;
    unsigned short recvReserved2;    

/* 802.11b header.  Unused by this driver. */
    unsigned short recvFrameControl;
    unsigned short recvDurationId;
    unsigned short recvAddress1[3];
    unsigned short recvAddress2[3];
    unsigned short recvAddress3[3];
    unsigned short recvSeqControl;
    unsigned short recvAddress4[3];
    unsigned short recvDataLen;

/* 802.3 header. */
/* Destination Ethernet address */
    unsigned short recvDestAddr[3];
/* Source Ethernet address */
    unsigned short recvSrcAddr[3];
/* Length of Ethernet frame, not including 802.3 header */
    unsigned short recvLength;
    
} ttPrismReceiveFrame;

#define TM_PRISM_RECV_CTL_LEN         60
#define TM_PRISM_RECV_CTL_NO_8023_LEN 46

/* local function prototypes */
static int tfPrismSetupPci(unsigned char TM_FAR * intLine,
                           unsigned char TM_FAR * TM_FAR * memAddrPtr,
                           unsigned long TM_FAR * ioAddr);
static int tfPrismAllocateFID(int size);
static void tfPrismDoAccessCommand(unsigned short          rid,
                                   unsigned short TM_FAR * recDataPtr,
                                   int                     recDataLen,
                                   int                     readWriteFlag);
static void tfPrismProcessInfoFrame(void);

/* Local helper routines to set PRISM options. */
static void tfPrismSetBssChan(unsigned char channel);
static void tfPrismSetSSID(char TM_FAR * ssid);
static void tfPrismSetOwnSSID(char TM_FAR * ssid);
static void tfPrismSetPortType(int portType);
static void tfPrismSetWEPKey(unsigned char TM_FAR * keyPtr, int keyLength);
static void tfPrismEnableWEP(void);
static void tfPrismDisableWEP(void);

/* Local helper routines to retrieve PRISM info. */
static int tfPrismGetPortStatus(void);
static void tfPrismGetCurrentSSID(char TM_FAR * ssid);
static void tfPrismGetCurrentBSSID(char TM_FAR * bssid);
static int tfPrismGetCommsQuality(void);
static int tfPrismGetTxRate(void);


/* local variable definitions */
static ttPrismInfoPtr tlPrismInfoPtr;

/* static const tables */

/* macro definitions */

/* #define TM_PRISM_NO_INTERRUPTS */

#define TM_PCI_ID_INTERSIL      0x1260 /* PCI Vendor ID. */
#define TM_PCI_ID_INTERSIL_3842 0x3873 /* PCI Device ID. */

/* 
 * PCI bus commands 
 */
/* PCI command: enable bus mastering */
#define TM_PCI_BUS_MASTER_CMD 0x0004
/* PCI command: enable I/O space registers */
#define TM_PCI_IO_SPACE_CMD   0x0001
/* PCI command: enable memory mapped registers */
#define TM_PCI_MEM_SPACE_CMD  0x0002

/* Complete command to send to the PCI device */
#define TM_PCI_MODE_CMD       TM_PCI_IO_SPACE_CMD | TM_PCI_MEM_SPACE_CMD | \
                              TM_PCI_BUS_MASTER_CMD

/*
 * PRISM register definitions.
 */ 
#define TM_PRISM_REG_CMD      0x00  /* Command register               */
#define TM_PRISM_REG_PARAM0   0x04  /* Command parameter 0            */
#define TM_PRISM_REG_STATUS   0x10  /* Command result status          */ 
#define TM_PRISM_REG_INFOFID  0x20  /* Info FID                       */
#define TM_PRISM_REG_SELECT0  0x30  /* BAP select register 0          */
#define TM_PRISM_REG_SELECT1  0x34  /* BAP select register 1          */
#define TM_PRISM_REG_OFFSET0  0X38  /* BAP offset register 0          */
#define TM_PRISM_REG_OFFSET1  0X3c  /* BAP offset register 1          */
#define TM_PRISM_REG_RX_FID   0x40  /* Received FID                   */
#define TM_PRISM_REG_ALLOCFID 0x44  /* FID allocation result register */
#define TM_PRISM_REG_EVSTAT   0x60  /* Event status register          */
#define TM_PRISM_REG_INTEN    0x64  /* Interrupt enable register      */
#define TM_PRISM_REG_EVACK    0x68  /* Event acknowledgement register */
#define TM_PRISM_REG_DATA0    0x6c  /* BAP data register 0            */
#define TM_PRISM_REG_DATA1    0x70  /* BAP data register 1            */

/* Command register is busy */
#define TM_PRISM_REG_CMD_BUSY 0x8000

/*
 * Event status bits. 
 */
#define TM_PRISM_REG_EVSTAT_RX    0x0001
#define TM_PRISM_REG_EVSTAT_TX    0x0002
#define TM_PRISM_REG_EVSTAT_TXEXC 0x0004 
#define TM_PRISM_REG_EVSTAT_ALLOC 0x0008
#define TM_PRISM_REG_EVSTAT_CMD   0x0010
#define TM_PRISM_REG_EVSTAT_INFO  0x0080
#define TM_PRISM_REG_EVSTAT_TICK  0x8000

/*
 * Event acknowledgment bits.
 */ 
#define TM_PRISM_REG_EVACK_RX     0x0001
#define TM_PRISM_REG_EVACK_TX     0x0002
#define TM_PRISM_REG_EVACK_TXEXC  0x0004 
#define TM_PRISM_REG_EVACK_ALLOC  0x0008
#define TM_PRISM_REG_EVACK_CMD    0x0010
#define TM_PRISM_REG_EVACK_INFO   0x0080
#define TM_PRISM_REG_EVACK_TICK   0x8000

/*
 * Interrupt mask bits
 */ 
#define TM_PRISM_REG_INTEN_RX     0x0001
#define TM_PRISM_REG_INTEN_ALLOC  0x0008

/*
 * PRISM command codes.
 */ 
#define TM_PRISM_CMD_CODE_INIT     0x0000  /* Initialize controller.      */
#define TM_PRISM_CMD_CODE_ENABLE   000001  /* Enable MAC port.            */
#define TM_PRISM_CMD_CODE_DISABLE  0x0002  /* Disable MAC port.           */
#define TM_PRISM_CMD_CODE_ALLOC    0x000A  /* Allocate FID.               */
#define TM_PRISM_CMD_CODE_TRANSMIT 0x010B  /* Transmit frame (w/ reclaim) */
#define TM_PRISM_CMD_CODE_WACCESS  0x0121  /* Access, with Write bit set. */
#define TM_PRISM_CMD_CODE_ACCESS   0x0021  /* Access, w/o Write bit set.  */

/*
 * Resource IDentifiers (RID's):
 */
#define TM_PRISM_RID_COMM_TALLIES 0xf100
#define TM_PRISM_RID_LINK_STATUS  0xf200
#define TM_PRISM_RID_PORT_TYPE    0xfc00
#define TM_PRISM_RID_OWN_MAC_ADDR 0xfc01
#define TM_PRISM_RID_DESIRED_SSID 0xfc02
#define TM_PRISM_RID_OWN_CHAN     0xfc03
#define TM_PRISM_RID_OWN_SSID     0xfc04
#define TM_PRISM_RID_WEP_KEY_ID   0xfc23
#define TM_PRISM_RID_WEP_KEY0     0xfc24
#define TM_PRISM_RID_WEP_FLAGS    0xfc28
#define TM_PRISM_RID_CREATE_IBSS  0xfc81
#define TM_PRISM_RID_TX_RATE_CTL  0xfc84
#define TM_PRISM_RID_PORT_STATUS  0xfd40
#define TM_PRISM_RID_CURR_SSID    0xfd41
#define TM_PRISM_RID_CURR_BSSID   0xfd42
#define TM_PRISM_RID_COMM_QUALITY 0xfd43
#define TM_PRISM_RID_CURR_TX_RATE 0xfd44

/* Station port types: */
#define TM_PRISM_PORT_TYPE_IBSS  0   /* Ad-hoc & auto-port mode */
#define TM_PRISM_PORT_TYPE_BSS   1   /* Infrastructure mode     */ 

/* Access command flags */
#define TM_PRISM_ACCESS_WRITE  0  /* Set configuration data */
#define TM_PRISM_ACCESS_READ   1  /* Get configuration data */

/* Length of Ethernet header, in bytes. */
#define TM_PRISM_ETHER_HDR_LEN 14

#define TM_PRISM_TX_CONTROL_8023_HDR 0x0000

/*
 * WEP configuration flags:
 */ 
#define TM_PRISM_WEP_FLAGS_ENABLE   0x0001
#define TM_PRISM_WEP_FLAGS_DISABLE  0x0000

/*
 * Bit-map values for supported 802.11b rates.
 */ 
#define TM_PRISM_TX_RATE_1_MBPS  0x1
#define TM_PRISM_TX_RATE_2_MBPS  0x2
#define TM_PRISM_TX_RATE_5_MBPS  0x4
#define TM_PRISM_TX_RATE_11_MBPS 0x8


/*
 * tfPrismOpen 
 *
 * Initializes the PRISM 802.11b controller.  Get the relevant PCI
 * information, intialize the chip and enable interrupts.
 *
 * 1. Scan PCI bus for PRISM hardware and get the necessary PCI information.
 * 2. Allocate PRISM driver state structure and fill in information from PCI.
 * 3. Wait for basic hardware initialization to complete (busy bit in command
 *    register is clear).
 * 4. Issue command to initialize the PRISM MAC.
 * 5. Disable interrupts, and clear any outstanding interrupts.
 * 6. Set the port type to infrastructure (using Access Point) or
 *    ad-hoc (no AP).  This can be changed later through an IOCTL command.
 * 7. Set the channel and SSID to be used when creating a BSS in ad-hoc mode.
 * 6. If the controller is in infrastructure mode, set our desired SSID.
 * 7. Initialize receive buffer pool.
 * 8. Install an interrupt handler for this PCI line. 
 * 9. Clear outstanding interrupts again and enable interrupts on the
 *    controller.
 * 10. Enable port zero (the only one available for a station).
 */     
int tfPrismOpen(ttUserInterface interfaceHandle)
{
    int                     defaultMode;
    int                     errorCode;
    unsigned char           intLine;
    unsigned char TM_FAR *  memAddrPtr;
    unsigned long           ioAddr;
    unsigned int            qIndex;
    unsigned int            allocSize;
    unsigned short          result;
    unsigned short          txRateControl;
    volatile unsigned short cmdReg;


    if ( tfInterfaceScatterCount(interfaceHandle) > TM_PRISM_BUFS_PER_FRAME )
    {
        tfKernelWarning("tfPrismOpen",
                        "Too many scattered buffers for this driver.");
        errorCode = TM_DEV_ERROR;
        goto prismOpenEnd;
    }
    
/* 1. Scan PCI bus for PRISM hardware and get the necessary PCI information. */
    errorCode = tfPrismSetupPci(&intLine, &memAddrPtr, &ioAddr);

    if (errorCode == TM_DEV_OKAY)
    {
/* 2. Allocate PRISM driver state structure and fill in information from PCI. */
        tlPrismInfoPtr =
            (ttPrismInfoPtr) tm_kernel_malloc(sizeof(ttPrismInfo));

        if (tlPrismInfoPtr == (ttPrismInfoPtr)0)
        {
            errorCode = TM_ENOBUFS;
            goto prismOpenEnd;
        }

/* Ensure that base memory address is long word aligned. */
        memAddrPtr = (unsigned char *) ((unsigned long)memAddrPtr & 0xfffffff0);
        
        tm_bzero(tlPrismInfoPtr, sizeof(ttPrismInfo));

        tlPrismInfoPtr->prismInterfaceHandle = interfaceHandle;
        tlPrismInfoPtr->prismMemAddrPtr      = memAddrPtr;
        tlPrismInfoPtr->prismIoAddr          = ioAddr;
            
/*
 * 3. Wait for basic hardware initialization to complete (busy bit in command
 *    register is clear).
 */
        tm_prism_wait_not_busy(TM_PRISM_REG_CMD, cmdReg);
        
/* 4. Issue command to initialize the PRISM MAC. */
        tm_prism_out_word(TM_PRISM_CMD_CODE_INIT, TM_PRISM_REG_CMD);
        tm_prism_cmd_wait_and_ack(cmdReg, result);
        
/* 5. Disable interrupts, and clear any outstanding interrupts. */
        tm_prism_out_word(0, TM_PRISM_REG_INTEN);
        tm_prism_out_word(0xffff, TM_PRISM_REG_EVACK);
        
/*
 * 6. Set the port type to infrastructure (using Access Point),
 *    ad-hoc (no AP) or auto-port (use AP if present, other use ad-hoc).  This
 *    can be changed later through an IOCTL command.
 */
        defaultMode = TM_PRISM_DEFAULT_MODE;        
        tfPrismSetPortType(defaultMode);

/* 7. Set the channel and SSID to be used when creating a BSS in ad-hoc mode. */
        tfPrismSetBssChan(TM_PRISM_DEFAULT_BSS_CHAN);
        tfPrismSetOwnSSID(TM_PRISM_DEFAULT_OWN_SSID);
        
        tfPrismSetSSID(TM_PRISM_DEFAULT_SSID);

/* Set the possible data rates to be used for data transmission. */
        txRateControl = TM_PRISM_DEFAULT_TX_RATES;
        tfPrismDoAccessCommand(TM_PRISM_RID_TX_RATE_CTL,
                               &txRateControl,
                               sizeof(txRateControl),
                               TM_PRISM_ACCESS_WRITE);
        
/*
 * Allocate a pool of FID's for transmission.
 */
        tm_bzero(tlPrismInfoPtr->prismXmitFidQueue,
                 sizeof(tlPrismInfoPtr->prismXmitFidQueue));
/* Maximum packet size that could be transmitted */
        allocSize = TM_ETHER_MAX_PACKET_CRC + TM_PRISM_XMIT_CTL_LEN
            + TM_LLC_SNAP_HDR_LEN - TM_PRISM_ETHER_HDR_LEN;
        for (qIndex=0; qIndex < TM_PRISM_MAX_XMIT_FIDS - 1; qIndex++)
        {
            tlPrismInfoPtr->prismXmitFidQueue[qIndex] =
                tfPrismAllocateFID(allocSize);
        }

        tlPrismInfoPtr->prismXmitFidHead = 0;
        tlPrismInfoPtr->prismXmitFidTail = TM_PRISM_MAX_XMIT_FIDS - 1;
        
/* 7. Initialize receive buffer pool. */
        (void)tfPoolCreate(interfaceHandle,
                           TM_PRISM_LARGE_RECV_POOL,
                           TM_PRISM_SMALL_RECV_POOL,
                           TM_PRISM_LARGE_RECV_BUF_SIZE,
                           TM_PRISM_SMALL_RECV_BUF_SIZE,
                           TM_PRISM_BUF_ALIGNMENT,
                           TM_POOL_REFILL_IN_LINE );

/* 8. Install an interrupt handler for this PCI line. */
#ifndef TM_PRISM_NO_INTERRUPTS
#ifdef TM_XSCALE_IQ80310
        tfKernelInstallIsrHandler(tfPrismHandlerIsr, SINTB_INT_ID);
#else /* !TM_XSCALE_IQ80310 */
#error "Unsupported hardware - must add interrupt handler for PRISM controller"
#endif /* TM_XSCALE_IQ80310 */
#endif /* TM_PRISM_NO_INTERRUPTS */
        
/*
 * 9. Clear outstanding interrupts again and enable interrupts on the
 *    controller.
 */
        tm_prism_out_word(0xffff, TM_PRISM_REG_EVACK);
#ifndef TM_PRISM_NO_INTERRUPTS
        tm_prism_out_word(TM_PRISM_REG_INTEN_RX | TM_PRISM_REG_INTEN_ALLOC,
                          TM_PRISM_REG_INTEN);
#endif /* TM_PRISM_NO_INTERRUPTS */
        
/* 10. Enable port zero (the only one available for a station). */
        tm_prism_wait_not_busy(TM_PRISM_REG_CMD, cmdReg);
        
        tm_prism_out_word(TM_PRISM_CMD_CODE_ENABLE, TM_PRISM_REG_CMD);
        tm_prism_cmd_wait_and_ack(cmdReg, result);
    
    }

prismOpenEnd:
    return errorCode;
}

/*
 * tfPrismGetPhysAddr
 *
 *   Returns the physical address of this Ethernet interface.  
 *
 */
int tfPrismGetPhysAddr(ttUserInterface interfaceHandle,
                       char TM_FAR *   physicalAddress)
{
    int errorCode;
    
    interfaceHandle=interfaceHandle;

    tfPrismDoAccessCommand(TM_PRISM_RID_OWN_MAC_ADDR,
                           (unsigned short TM_FAR *) physicalAddress,
                           TM_ETHERNET_PHY_ADDR_LEN,
                           TM_PRISM_ACCESS_READ);

    return errorCode;
}

/*
 * tfPrismSend
 *
 * Submits an Ethernet frame to the PRISM controller for transmission.  This
 * includes converting the header from DIX Ethernet to 802.3 Ethernet format,
 * filling out a transmit structure and copying this to the controller.
 *
 * 1. Allocate FID.
 * 2. Write the allocated FID into the BAP select register and wait it
 *    to be accepted (busy bit cleared).
 * 3. Fill out transmit control structure, indicating that a 802.3 header will
 *    be used.
 * 4. The frame sent by the will be in DIX Ethernet format, rather than 802.3
 *    format.  Since the packet isn't being DMA'd directly to the controller,
 *    but is instead being copied word by word, this isn't as big of a deal.
 *    We just fill out the LLC and SNAP headers and send these in between the
 *    DIX Ethernet header and the start of the payload (the IP header).
 * 5. Everything is prepared to send.  Write a zero offset into the BAP offset
 *    register and wait for it be accepted (busy bit cleared).
 * 6. Copy the transmit control structure up to, but not including, the 802.3
 *    Ethernet header into the BAP.
 * 7. Copy the source and destination address from the original DIX Ethernet
 *    header into the BAP.
 * 8. Copy the packet length (big endian) into the BAP.  This forms the end of
 *    the base 802.3 Ethernet header.
 * 9. Copy the LLC and SNAP headers into the BAP.
 * 10. Copy the packet payload into the BAP (this starts at the IP or ARP
 *     header).
 * 11. Write the FID into the Parameter #0 register and issue a transmit
 *     command.
 */
int tfPrismSend(ttUserInterface interfaceHandle,
                char TM_FAR *   dataPtr,
                int             dataLength,
                int             flag)
{
    ttPrismTransmitFrame    xmitFrame;
    ttLlcSnapHeader         llcSnapHdr;
    int                     transmitFID;
    int                     i;
    int                     copyLength;
    int                     errorCode;
    unsigned short          regValue;
    unsigned short TM_FAR * shortDataPtr;    
    unsigned short TM_FAR * origEtherTypePtr;
    unsigned short          tempShort;
    unsigned char TM_FAR *  tempCharPtr; 
    unsigned short          result;
    int                     fidQTail;
    unsigned int            scatIndex;
    int TM_FAR *            fidQPtr;

    errorCode = TM_DEV_OKAY;

    /* Add this buffer to the list of scattered buffers for the current frame. */
    scatIndex = tlPrismInfoPtr->prismScatIndex;
    tlPrismInfoPtr->prismScatDataPtr[scatIndex] = (unsigned char *)dataPtr;
    tlPrismInfoPtr->prismScatDataLen[scatIndex] = dataLength;
    tlPrismInfoPtr->prismScatFrameSize += dataLength;
    tlPrismInfoPtr->prismScatIndex++;

    /* Frame is not yet complete, so return and wait for more buffers. */
    if (flag == TM_USER_BUFFER_MORE) {
        goto prismSendReturn;
    }

/* 1. Allocate FID. */

/* Get FID from the head of the queue. */
    tm_kernel_set_critical;
    fidQTail = tlPrismInfoPtr->prismXmitFidTail;
    tm_kernel_release_critical;

    while (tlPrismInfoPtr->prismXmitFidHead == fidQTail)
    {
/* FID queue is empty.  Wait for a FID to be reclaimed. */
        tfInterfaceSpinLock(interfaceHandle);
        tm_kernel_set_critical;
        fidQTail = tlPrismInfoPtr->prismXmitFidTail;
        tm_kernel_release_critical;

#ifdef TM_PRISM_NO_INTERRUPTS
/* Since interrupts aren't being used, we need poll to see if an alloc
   event has been issued; otherwise, we can get stuck here with an
   empty FID queue. */
        tfPrismHandlerIsr();
#endif /* TM_PRISM_NO_INTERRUPTS */

    }

    fidQPtr = tlPrismInfoPtr->prismXmitFidQueue;
    transmitFID = fidQPtr[tlPrismInfoPtr->prismXmitFidHead];

/* Indicate that this FID belongs to the chip, until it is reclaimed. */
    fidQPtr[tlPrismInfoPtr->prismXmitFidHead] = 0;

/* Move the head of the queue to the next entry. */
    tm_prism_inc_queue_index(tlPrismInfoPtr->prismXmitFidHead);

/* Add this buffer to the list of scattered buffers for the current frame. */
    scatIndex = tlPrismInfoPtr->prismScatIndex;
    tlPrismInfoPtr->prismScatDataPtr[scatIndex] = dataPtr;
    tlPrismInfoPtr->prismScatDataLen[scatIndex] = dataLength;

    tlPrismInfoPtr->prismScatFrameSize += dataLength;
    tlPrismInfoPtr->prismScatIndex++;

/* Frame is not yet complete, so return and wait for more buffers. */
    if (flag == TM_USER_BUFFER_MORE)
    {
        goto prismSendReturn;
    }
    
/*
 * 3. Fill out transmit control structure, indicating that a 802.3 header will
 *    be used.
 */
    tm_bzero(&xmitFrame, sizeof(xmitFrame));
    xmitFrame.xmitTxControl = TM_PRISM_TX_CONTROL_8023_HDR;
        
/*
 * 4. The frame sent by the will be in DIX Ethernet format, rather than 802.3
 *    format.  Since the packet isn't being DMA'd directly to the controller,
 *    but is instead being copied word by word, this isn't as big of a deal.
 *    We just fill out the LLC and SNAP headers and send these in between the
 *    DIX Ethernet header and the start of the payload (the IP header).
 *    NOTE: If the stack supports 802.3 formatted frames in the future, much of
 *    this can be removed.
 *    
 */
    llcSnapHdr.llcDsap = 0xaa;
    llcSnapHdr.llcSsap = 0xaa;
    llcSnapHdr.llcCtl = 3;

    llcSnapHdr.snapOui[0] = 0;
    llcSnapHdr.snapOui[1] = 0;
    llcSnapHdr.snapOui[2] = 0;

    origEtherTypePtr = (unsigned short TM_FAR *)
        (tlPrismInfoPtr->prismScatDataPtr[0] + TM_PRISM_ETHER_HDR_LEN - sizeof(unsigned short));
    llcSnapHdr.snapType = *origEtherTypePtr;

/* 2. Write the allocated FID into the BAP select register  */
    tm_prism_out_word(transmitFID, TM_PRISM_REG_SELECT0);
    
/*
 * 5. Everything is prepared to send.  Write a zero offset into the BAP offset
 *    register and wait for it be accepted (busy bit cleared).
 */
    tm_prism_out_word(0, TM_PRISM_REG_OFFSET0);
    tm_prism_wait_not_busy(TM_PRISM_REG_OFFSET0, regValue);
    if (tm_prism_offset_reg_error( regValue ))
    {
        errorCode = TM_DEV_ERROR;
        goto prismSendReturn;        
    }

/*
 * 6. Copy the transmit control structure up to, but not including, the 802.3
 *    Ethernet header into the BAP.
 */
    shortDataPtr = (unsigned short TM_FAR *) &xmitFrame;
    copyLength = TM_PRISM_XMIT_CTL_NO_8023_LEN >> 1; 
    for (i=0; i<copyLength; i++)
    {
        tm_prism_out_word(shortDataPtr[i], TM_PRISM_REG_DATA0);
    }
    
/*
 * 7. Copy the source and destination address from the original DIX Ethernet
 *    header into the BAP.
 *
 */
    shortDataPtr = (unsigned short TM_FAR *) tlPrismInfoPtr->prismScatDataPtr[0];
    copyLength = (TM_PRISM_ETHER_HDR_LEN - sizeof(unsigned short)) >> 1;
    for (i=0; i<copyLength; i++)
    {
        tm_prism_out_word(shortDataPtr[i], TM_PRISM_REG_DATA0);        
    }
    
/*
 * 8. Copy the packet length (big endian) into the BAP.  This forms the end of
 *    the base 802.3 Ethernet header.
 */
    copyLength =
        htons(  tlPrismInfoPtr->prismScatFrameSize - TM_PRISM_ETHER_HDR_LEN
              + TM_LLC_SNAP_HDR_LEN);
    tm_prism_out_word(copyLength, TM_PRISM_REG_DATA0);

/* 9. Copy the LLC and SNAP headers into the BAP. */
    shortDataPtr = (unsigned short TM_FAR *) &llcSnapHdr;
    copyLength = TM_LLC_SNAP_HDR_LEN >> 1;
    for (i=0; i<copyLength; i++)
    {
        tm_prism_out_word(shortDataPtr[i], TM_PRISM_REG_DATA0);
    }
    
/*
 * 10. Copy the packet payload into the BAP (this starts at the IP or ARP
 *     header).  Since there may be more than one buffer per frame, we need
 *     to loop through all of the scattered buffers.  This can get a bit
 *     messy with odd sized frames: see below.
 *
 *     NOTE: This is the most time consuming part of the send routine, so
 *           anything that can be done to optimize this (such as using DMA?)
 *           will definitely benefit performance.
 */
    for (scatIndex=0; scatIndex < tlPrismInfoPtr->prismScatIndex; scatIndex++)
    {
        shortDataPtr = (unsigned short TM_FAR *)
            tlPrismInfoPtr->prismScatDataPtr[scatIndex];
        copyLength = (tlPrismInfoPtr->prismScatDataLen[scatIndex] >> 1);

/* If this is the first buffer in this frame, move past the Ethernet header. */
        if (scatIndex == 0)
        {
            shortDataPtr += (TM_PRISM_ETHER_HDR_LEN>>1);
            copyLength -= (TM_PRISM_ETHER_HDR_LEN>>1);
        }

        if ((unsigned) shortDataPtr & 1)
        {
/* This buffer is misaligned, due to a previous buffer that was odd sized.
   So, to avoid a misaligned memory access, we have to do this the ugly way
   and copy data a byte at a time... */
            copyLength <<= 1;
            dataPtr = (unsigned char TM_FAR *) shortDataPtr;
            for (i=0; i < copyLength; i += 2)
            {
                tempCharPtr = (unsigned char TM_FAR *) &tempShort;
                tempCharPtr[0] = dataPtr[i];
                tempCharPtr[1] = dataPtr[i+1];
                tm_prism_out_word(tempShort, TM_PRISM_REG_DATA0);
            }
        }
        else
        {
/* A nice aligned buffer, so copy it a word at a time. */
            for (i=0; i < copyLength; i++)
            {
                tm_prism_out_word(shortDataPtr[i], TM_PRISM_REG_DATA0);
            }
        }

/* If the packet size is odd, the last byte should be sent in the least
   significant byte of the last data word. */
        if (tlPrismInfoPtr->prismScatDataLen[scatIndex] & 1)
        {
            dataPtr = tlPrismInfoPtr->prismScatDataPtr[scatIndex];
            dataLength = tlPrismInfoPtr->prismScatDataLen[scatIndex];
            tempShort = 0;
            tempCharPtr = (unsigned char TM_FAR *) &tempShort;
            tempCharPtr[0] = dataPtr[dataLength-1];

/* If there is another scattered buffer following this one, we need to
   include the first byte from that buffer when we copy this word to the
   controller.  This causes the next buffer to be misaligned (see above). */
            if (scatIndex != tlPrismInfoPtr->prismScatIndex)
            {
                dataPtr = tlPrismInfoPtr->prismScatDataPtr[scatIndex+1];
                tempCharPtr[1] = dataPtr[0];
                tlPrismInfoPtr->prismScatDataPtr[scatIndex+1]++;
                tlPrismInfoPtr->prismScatDataLen[scatIndex+1]--;
            }

            tm_prism_out_word(tempShort, TM_PRISM_REG_DATA0);
        }
    }
    
/*
 * 11. Write the FID into the Parameter #0 register and issue a transmit
 *     command.
 */
/* Make sure that previous command has completed. */
    tm_prism_wait_not_busy(TM_PRISM_REG_CMD, regValue);

/* Issue allocation command, and wait for it to be accepted. */
    tm_prism_out_word(transmitFID, TM_PRISM_REG_PARAM0);
    tm_prism_out_word(TM_PRISM_CMD_CODE_TRANSMIT, TM_PRISM_REG_CMD);
    tm_prism_cmd_wait_and_ack(regValue, result);

    if (result != 0x000b)
    {
/* Transmit command failed, put FID back in queue. */
        fidQPtr = tlPrismInfoPtr->prismXmitFidQueue;
        tm_kernel_set_critical;
        fidQPtr[tlPrismInfoPtr->prismXmitFidTail] = transmitFID;
        tm_prism_inc_queue_index(tlPrismInfoPtr->prismXmitFidTail);
        tm_kernel_release_critical;
    }
    
    if (flag & TM_USER_BUFFER_LAST)
    {
/* Reset scattered buffer info */
        tlPrismInfoPtr->prismScatIndex = 0;
        tlPrismInfoPtr->prismScatFrameSize = 0;

/* Free the packet sent by stack, since we've copied it to the controller. */
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }

prismSendReturn:
    return errorCode;
}

/*
 * tfPrismReceive
 * 
 * Retrieves a buffer from the receive pool, which was filled in during the
 * ISR, and passes it up to the stack.
 * 
 */ 
int tfPrismReceive(ttUserInterface        interfaceHandle,
                   char TM_FAR * TM_FAR * dataPtrPtr,
                   int  TM_FAR *          dataLengthPtr,
                   ttUserBufferPtr        bufHandlePtr)
{
    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          bufHandlePtr );
}

/*
 * tfPrismHandlerIsr
 *
 * Processes PRISM interrupts, including the a received packet event.
 *
 * 1. If a packet has been received...
 *     1.1. Get FID of received packet and place it into the BAP select reg.
 *     1.2. Read the receive control frame, including the 802.3 header.
 *     1.3. Allocate a buffer to place the incoming data.  If this fails,
 *          just ACK the receive event, which should free the PRISM memory
 *          containing the incoming packet.
 *     1.4. Read the LLC and SNAP headers.  Get the frame type from the SNAP
 *          header and discard the LLC and SNAP headers.
 *     1.5. Copy the Ethernet destination address, source address and frame
 *          type into the stack bufer.
 *     1.6. Read the rest of the packet from the BAP (PRISM controller).
 *     1.7. Acknowledge this receive event.
 * 2. Allocation event - issued by the PRISM controller when a transmit FID
 *    can be reclaimed.
 * 3. For every frame received, notify the stack of an incoming frame.    
 */
void tfPrismHandlerIsr(void)
{
    ttPrismReceiveFrame     recvFrame;
    ttLlcSnapHeader         llcSnapHdr;
    unsigned short TM_FAR * shortDataPtr;
    unsigned char TM_FAR *  bufDataPtr;    
    unsigned short          event;
    unsigned short          recvFID;
    int                     xmitFID;
    volatile unsigned short regValue;
    unsigned short          recvLen;
    int                     recvNotifyCount;
    int                     i;
    int                     copyLength;

    
    recvNotifyCount = 0;

    event = tm_prism_in_word(TM_PRISM_REG_EVSTAT);

    while ( event & (TM_PRISM_REG_EVSTAT_RX |
                     TM_PRISM_REG_EVSTAT_INFO |
                     TM_PRISM_REG_EVSTAT_ALLOC))
    {
        
/* Received packet. */
        if (event & TM_PRISM_REG_EVSTAT_RX)
        {

/*     1.1. Get FID of received packet and place it into the BAP select reg. */
            recvFID = tm_prism_in_word(TM_PRISM_REG_RX_FID);
            tm_prism_out_word(recvFID, TM_PRISM_REG_SELECT1);
            tm_prism_out_word(0, TM_PRISM_REG_OFFSET1);
            tm_prism_wait_not_busy(TM_PRISM_REG_OFFSET1, regValue);

            if ( !tm_prism_offset_reg_error(regValue))
            {
            
/*     1.2. Read the receive control frame, including the 802.3 header. */
                copyLength = TM_PRISM_RECV_CTL_LEN >> 1;
                shortDataPtr = (unsigned short TM_FAR *) &recvFrame;
                for (i=0; i<copyLength; i++)
                {
                    shortDataPtr[i] = tm_prism_in_word(TM_PRISM_REG_DATA1);
                }

                recvLen = htons(recvFrame.recvLength);
/*
 *     1.3. Allocate a buffer to place the incoming data.  If this fails,
 *          just ACK the receive event, which should free the PRISM memory
 *          containing the incoming packet.
 */
                bufDataPtr = (unsigned char TM_FAR *) 
                    tfPoolIsrGetBuffer(
                        tlPrismInfoPtr->prismInterfaceHandle,
                        recvLen + TM_PRISM_ETHER_HDR_LEN - TM_LLC_SNAP_HDR_LEN);
            }
            else
            {
/* Couldn't read data from the controller due to an OFFSET register error.
   Just ACK this event which should free up the packet on the controller. */
                bufDataPtr = (unsigned char TM_FAR *) 0;
            }
            
            if (bufDataPtr != (unsigned char TM_FAR *) 0)
            {
/*
 *     1.4. Read the LLC and SNAP headers.  Get the frame type from the SNAP
 *          header and discard the LLC and SNAP headers.
 */
                shortDataPtr = (unsigned short TM_FAR *) &llcSnapHdr;
                copyLength   =  TM_LLC_SNAP_HDR_LEN >> 1;
                for (i=0; i<copyLength; i++)
                {
                    shortDataPtr[i] = tm_prism_in_word(TM_PRISM_REG_DATA1);
                }
                
/*
 *     1.5. Copy the Ethernet destination address, source address and frame
 *          type into the stack bufer.
 */
                tm_bcopy(&recvFrame.recvDestAddr[0],
                         bufDataPtr,
                         TM_PRISM_ETHER_HDR_LEN - sizeof(unsigned short));
                shortDataPtr = (unsigned short TM_FAR *)
                    (  bufDataPtr + TM_PRISM_ETHER_HDR_LEN
                     - sizeof(unsigned short));
                *shortDataPtr = llcSnapHdr.snapType;
                
/*     1.6. Read the rest of the packet from the BAP (PRISM controller). */
                shortDataPtr = (unsigned short TM_FAR *)
                    (bufDataPtr + TM_PRISM_ETHER_HDR_LEN);
                copyLength = (recvLen - TM_LLC_SNAP_HDR_LEN)  >> 1;
                for (i=0; i<copyLength; i++)
                {
                    shortDataPtr[i] = tm_prism_in_word(TM_PRISM_REG_DATA1);
                }

/* Odd sized incoming frames. */
                if (recvLen & 1)
                {
                    recvLen = recvLen + TM_PRISM_ETHER_HDR_LEN
                        - TM_LLC_SNAP_HDR_LEN;
                    regValue = tm_prism_in_word(TM_PRISM_REG_DATA1);
                    bufDataPtr[recvLen-1] =
                        ((unsigned char TM_FAR *)(&regValue))[0];
                }
                
                recvNotifyCount++;
            }

/*     1.7. Acknowledge this receive event. */
            tm_prism_out_word(TM_PRISM_REG_EVACK_RX, TM_PRISM_REG_EVACK);
        }

/* Process and acknowledge information frames from PRISM controller. */        
        if (event & TM_PRISM_REG_EVSTAT_INFO)
        {
            tfPrismProcessInfoFrame();
        }
        
/*
 * 2. Allocation event - issued by the PRISM controller when a transmit FID
 *    can be reclaimed.
 */ 
        if (event & TM_PRISM_REG_EVSTAT_ALLOC)
        {
            xmitFID = tm_prism_in_word(TM_PRISM_REG_ALLOCFID);

/* Add a reclaimed FID to the tail of the queue.  Note that this doesn't do
   any error checking currently, and assumes that the queue is in sync.
   (i.e. that the controller hasn't done something strange like issue an
   ALLOC event before a FID has been removed from the queue. */
            tlPrismInfoPtr->prismXmitFidQueue[
                tlPrismInfoPtr->prismXmitFidTail] = xmitFID;
            
            tm_prism_inc_queue_index(tlPrismInfoPtr->prismXmitFidTail);

            tm_prism_out_word(TM_PRISM_REG_EVACK_ALLOC, TM_PRISM_REG_EVACK);
        }

        
/* In case another event has occured while handling the ones above... */
        event = tm_prism_in_word(TM_PRISM_REG_EVSTAT);        
    }

/* 3. For every frame received, notify the stack of an incoming frame */
    tfNotifyInterfaceIsr( tlPrismInfoPtr->prismInterfaceHandle, 
                          recvNotifyCount,
                          0, 0UL, 0UL );

}

/*
 * tfPrismClose
 *
 * Closes the PRISM 802.11b interface.
 *
 */
int tfPrismClose(ttUserInterface interfaceHandle)
{

    unsigned short cmdReg;
    unsigned short result; 
/* 
 * Free memory from pre-allocated buffers in the receive ring 
 */
    (void)tfPoolDelete(interfaceHandle);

/* Disable MAC port on controller. */
    tm_prism_wait_not_busy(TM_PRISM_REG_CMD, cmdReg);
        
    tm_prism_out_word(TM_PRISM_CMD_CODE_DISABLE, TM_PRISM_REG_CMD);
    tm_prism_cmd_wait_and_ack(cmdReg, result);

/* Disable interrupts, and clear any outstanding interrupts. */
    tm_prism_out_word(0, TM_PRISM_REG_INTEN);
    tm_prism_out_word(0xffff, TM_PRISM_REG_EVACK);
    
    return TM_DEV_OKAY;    
}

/*
 * tfPrismIoctl
 *
 * Controls various options on the PRISM controller as well as issuing certain
 * 802.11b related commands.
 *
 */ 
int tfPrismIoctl(ttUserInterface interfaceHandle,
                 int             flag,
                 void TM_FAR *   optionPtr,
                 int             optionLen)
{
    int           errorCode;
    int           paramInt;
    unsigned char param8;

    errorCode = TM_ENOERROR;

/* Gets the status of the wireless connection. */
    if (flag == TM_PRISM_IOCTL_GET_PORT_STATUS)
    {
        if (optionLen == sizeof(int))
        {
            *((int TM_FAR *) optionPtr) = tfPrismGetPortStatus();
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Get the current Service Set Identifier (SSID) which the station is
   connected to. */
    if (flag == TM_PRISM_IOCTL_GET_SSID)
    {
        if (optionLen > 32)
        {
            tfPrismGetCurrentSSID((char TM_FAR *) optionPtr);
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Get the MAC address used as the identifier of the Basic Service Set which
   the station is connected to. */
    if (flag == TM_PRISM_IOCTL_GET_BSSID)
    {
        if (optionLen >= 6)
        {
            tfPrismGetCurrentBSSID((char TM_FAR *) optionPtr);
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Gets the current quality of communications with the BSS.  This is based
   on the average signal and noise level.  The value ranges between 0-92. */
    if (flag == TM_PRISM_IOCTL_GET_COMM_QUALITY)
    {
        if (optionLen == sizeof(int))
        {
            paramInt = tfPrismGetCommsQuality();
            *((int TM_FAR *)optionPtr) = paramInt;
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Gets the current transmission rate, in Mbits/sec. */
    if (flag == TM_PRISM_IOCTL_GET_TX_RATE)
    {
        if (optionLen == sizeof(int))
        {
            paramInt = tfPrismGetTxRate();
            *((int TM_FAR *)optionPtr) = paramInt;
        }
    }

    
/* Sets the connection type (ad-hoc, infrastructure) for this station. */
    if (flag == TM_PRISM_IOCTL_SET_PORT_TYPE)
    {
        if (optionLen == sizeof(int))
        {
            paramInt = *((int TM_FAR *) optionPtr);
            if (    (paramInt == TM_PRISM_INFRASTRUCTURE_MODE)
                 || (paramInt == TM_PRISM_AD_HOC_MODE)
                 || (paramInt == (TM_PRISM_AUTO_PORT_MODE)) )
            {
                tfPrismSetPortType(paramInt);
            }
            else
            {
                errorCode = TM_EINVAL;
            }
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Sets the channel to be used when creating a BSS in ad-hoc mode. */
    if (flag == TM_PRISM_IOCTL_SET_IBSS_CHAN)
    {
        if (optionLen == sizeof(unsigned char))
        {
            param8 = *((unsigned char TM_FAR *) optionPtr);
            tfPrismSetBssChan(param8);
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* A null terminated string that indicates the Service Set that this station
   wishes to connect to.  This parameter may be null, or zero length, which
   indicates that the station should attempt to connect to any BSS. */
    if (flag == TM_PRISM_IOCTL_SET_SSID)
    {
        tfPrismSetSSID((char TM_FAR *) optionPtr);        
    }

/* A null terminated string that indicates the Service Set ID to use when
   creating an IBSS in ad-hoc mode. */
    if (flag == TM_PRISM_IOCTL_SET_OWN_SSID)
    {
        tfPrismSetOwnSSID((char TM_FAR *) optionPtr);
    }

/* Sets the WEP key for 64-bit encryption (40-bit key) or 128-bit
   encryption (104-bit key). */
    if (flag == TM_PRISM_IOCTL_SET_WEP_KEY)
    {
        if ( (optionLen == 5) || (optionLen == 13) )
        {
            tfPrismSetWEPKey((unsigned char TM_FAR *) optionPtr, optionLen);
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

/* Enables or disables  WEP encryption by the controller. */
    if (flag == TM_PRISM_IOCTL_SET_WEP_ENABLE)
    {
        if (optionLen == sizeof(unsigned char))
        {
            param8 = *((unsigned char TM_FAR *)optionPtr);
            if (param8)
            {
                tfPrismEnableWEP();
            }
            else
            {
                tfPrismDisableWEP();
            }
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }

    
/* The PRISM controller does not currnetly support per address multicast
   filtering.  That is, it always receives ALL multicast packets.  So both
   of these commands (issued by IGMP) to indicate which multicast addresses
   to receive are useless.  Multicast packets not destined for us will be
   filtered higher up in the stack. */
    if ( (flag & TM_DEV_SET_MCAST_LIST) || (flag & TM_DEV_SET_ALL_MCAST) )
    {
        errorCode = TM_ENOERROR;
    }

    return errorCode;
}

/*
 * tfPrismProcessInfoFrame
 * 
 * Processes an information frame, which can be asycnhronously generated by
 * the PRISM controller to notify us of certain events.  This routine doesn't
 * currently do anything too useful, but will display information from a
 * link status frame, and display the RID value for all other frame types.
 *
 * If more detailed logging was desired, it would be possible to read the
 * CommTallies frame here to get detailed counts of various PRISM operations
 * (packets sent, dropped, etc).
 * 
 */ 
static void tfPrismProcessInfoFrame(void)
{
    unsigned short infoFID;
    unsigned short regValue;
    
    infoFID = tm_prism_in_word(TM_PRISM_REG_INFOFID);
    tm_prism_out_word(infoFID, TM_PRISM_REG_SELECT1);
    tm_prism_out_word(0, TM_PRISM_REG_OFFSET1);
    tm_prism_wait_not_busy(TM_PRISM_REG_OFFSET1, regValue);
    if (tm_prism_offset_reg_error(regValue))
    {
/* Can't read info frame from the controller; just ACK the event and return. */
        goto prismInfoFrameReturn;        
    }
    

/* Discard length. */
    regValue = tm_prism_in_word(TM_PRISM_REG_DATA1);

/* Get info RID */
    regValue = tm_prism_in_word(TM_PRISM_REG_DATA1);

    if (regValue == TM_PRISM_RID_LINK_STATUS)
    {
        regValue = tm_prism_in_word(TM_PRISM_REG_DATA1);

        switch (regValue)
        {
            case TM_PRISM_CONN:
                tm_prism_debug_output("Link status: connected\n");
                break;
                
            case TM_PRISM_DISCONN:
                tm_prism_debug_output("Link status: disconnected\n");
                break;
                            
            case TM_PRISM_AP_CHANGE:
                tm_prism_debug_output("Link status: AP change\n");
                break;
                                       
            case TM_PRISM_AP_OUT_OF_RANGE:
                tm_prism_debug_output("Link status: AP out of range\n");
                break;
                            
            case TM_PRISM_AP_IN_RANGE:
                tm_prism_debug_output("Link status: AP in range\n");
                break;
                                       
            case TM_PRISM_ASSC_FAIL:
                tm_prism_debug_output("Link status: association failed\n");
                break;

            default:
                tm_prism_debug_output("Unrecognized link status.\n");
                            
        }
    }
    else
    {
/* The CommTallies frame is generated too often, so don't display. */
        if (regValue != TM_PRISM_RID_COMM_TALLIES)
        {
            tm_prism_debug_output_short("Info RID = %x\n", regValue);
        }
    }

prismInfoFrameReturn:
    
/* Acknowledge info frame. */
    tm_prism_out_word(TM_PRISM_REG_EVACK_INFO, TM_PRISM_REG_EVACK);

}

/*
 * tfPrismSetupPci()
 *
 *   This routine finds the PRISM 802.11b device on the PCI bus, and
 *   returns the necessary information (interrupt line, base memory address
 *   and/or base I/O address).  This is currently implemented for the
 *   Intel IQ80310 platform (pci_serv.c); if you are using a different target,
 *   you need to implement the PCI scan functions.
 */
static int tfPrismSetupPci(unsigned char TM_FAR * intLine,
                           unsigned char TM_FAR * TM_FAR * memAddrPtr,
                           unsigned long TM_FAR * ioAddr)
{
    unsigned char  pciBus;
    unsigned char  pciFunc;
    unsigned char  pciDevice;
    unsigned long  tempLong;
    unsigned long  pciIoAddr;
    unsigned short pciCommand;
    int            errorCode;
     
#ifdef TM_XSCALE_IQ80310
/* Initialize the PCI buses */
    PCI_DATA            pnew_data;
    PCI_DEVICE_LOCATION devid;

    sys_pci_device_initialization(&pnew_data);
  

/* Get PCI bus, device and function number for the PRISM controller */
    errorCode = sys_find_pci_device(TM_PCI_ID_INTERSIL,
                                    TM_PCI_ID_INTERSIL_3842,
                                    0,
                                    &devid);
    
    if (errorCode == OK)
    {
        pciBus    = devid.bus_number;
        pciDevice = devid.device_number;
        pciFunc   = devid.device_number;
   
/* Get the I/O and memory base addresses for the 8255x */
        sys_read_config_dword(
            pciBus, pciDevice, pciFunc, REGION0_BASE_OFFSET, &tempLong);
        
        *memAddrPtr = (unsigned char TM_FAR *) tempLong;
                        
        sys_read_config_dword(
            pciBus, pciDevice, pciFunc, REGION1_BASE_OFFSET, &pciIoAddr);

/* Mask out the I/O space indication bit */
        pciIoAddr &= ~1;

        *ioAddr = pciIoAddr;

/* 
 * Enables the device to act as a bus master, and enable both the I/O space
 * and memory decoders. (These options are set above by the 
 * #define TM_PCI_MODE_CMD and will probably have to be changed per platform)
 */
        sys_read_config_word(pciBus,pciDevice,pciFunc,STATUS_OFFSET,
                              &pciCommand);
        pciCommand |= TM_PCI_MODE_CMD;      

        sys_write_config_word(pciBus,pciDevice,pciFunc,COMMAND_OFFSET,
                               &pciCommand);
        errorCode = TM_DEV_OKAY;
    }
    else
    {
        errorCode = TM_DEV_ERROR;
    }
    
#else /* TM_XSCALE_IQ80310 */
    
#error "Unsupported hardware - need to implement PCI scan function."
    
#endif /* TM_XSCALE_IQ80310 */
    
    return errorCode;
}

/*
 * tfPrismSetSSID()
 * 
 * Sets the desired service set that the station is to connect to.
 * A null or zero length string indicates any SSID.
 */ 
static void tfPrismSetSSID(char TM_FAR * ssid)
{
    int  ssidLength;
    char ssidParamBuf[34];
    int  i;

/* Null or zero length SSID indicates that the station should connect to any
   available SSID.  Execute the access command with a zero length parameter. */
    if (ssid != (char TM_FAR *)0)
    {
        ssidLength = tm_strlen(ssid);
    }
    else
    {
        ssidLength = 0;
    }

/* Copy the user's SSID into the parameter buffer, placing the length of the
   SSID in the first word of the buffer. */
    *((unsigned short TM_FAR *) &ssidParamBuf[0]) = ssidLength;
    for (i=0; i<ssidLength; i++)
    {
        ssidParamBuf[i+2] = ssid[i];
    }
    
    tfPrismDoAccessCommand(TM_PRISM_RID_DESIRED_SSID,
                           (unsigned short TM_FAR *) ssidParamBuf,
                           ssidLength + 2,
                           TM_PRISM_ACCESS_WRITE);

}

/*
 * tfPrismSetOwnSSID()
 * 
 * Sets the service set ID that the station will use when creating an IBSS.
 */ 
static void tfPrismSetOwnSSID(char TM_FAR * ssid)
{
    int  ssidLength;
    char ssidParamBuf[34];
    int  i;
    
    if (ssid != (char TM_FAR *)0)
    {
        ssidLength = tm_strlen(ssid);
    }
    else
    {
        ssidLength = 0;
    }

/* Copy the user's SSID into the parameter buffer, placing the length of the
   SSID in the first word of the buffer. */
    *((unsigned short TM_FAR *) &ssidParamBuf[0]) = ssidLength;
    for (i=0; i<ssidLength; i++)
    {
        ssidParamBuf[i+2] = ssid[i];
    }

    tfPrismDoAccessCommand(TM_PRISM_RID_OWN_SSID,
                           (unsigned short TM_FAR *) ssidParamBuf,
                           ssidLength+2,
                           TM_PRISM_ACCESS_WRITE);

}

/*
 * tfPrismSetPortType()
 * 
 * Sets the port type (infrastructure or ad-hoc) of the 802.11b controller.
 */ 
static void tfPrismSetPortType(int portType)
{
    unsigned short regValue;

    if (portType == TM_PRISM_INFRASTRUCTURE_MODE)
    {
        regValue = TM_PRISM_PORT_TYPE_BSS;
    }
    else
    {
/* Ad-hoc and auto-port modes. */
        regValue = TM_PRISM_PORT_TYPE_IBSS;
    }

    tfPrismDoAccessCommand(TM_PRISM_RID_PORT_TYPE,
                           &regValue,
                           sizeof(regValue),
                           TM_PRISM_ACCESS_WRITE);

/* For auto-port mode, configure the controller so that an AP (ESS) is used
   if present; otherwise an ad-hoc network (IBSS) is joined or created. */
    if (portType == TM_PRISM_AUTO_PORT_MODE)
    {
        regValue = 1;
        tfPrismDoAccessCommand(TM_PRISM_RID_CREATE_IBSS,
                               &regValue,
                               sizeof(regValue),
                               TM_PRISM_ACCESS_WRITE);
    }
}


/*
 * tfPrismSetBssChan()
 * 
 * Sets the channel used for creating a BSS in ad-hoc mode.
 */ 
static void tfPrismSetBssChan(unsigned char channel)
{
    unsigned short channelId;

    channelId = (unsigned short) channel;
    tfPrismDoAccessCommand(TM_PRISM_RID_OWN_CHAN,
                           &channelId,
                           sizeof(channelId),
                           TM_PRISM_ACCESS_WRITE);
    
}

/*
 * tfPrismSetWEPKey()
 *
 * Sets the WEP encryption key used to encrypt outgoing packets and
 * decrypt incoming packets.  All encryption/decryption is done by the
 * firmware.
 * While the PRISM provides four slots for WEP keys, we always use the
 * first slot. (#0).
 */ 
static void tfPrismSetWEPKey(unsigned char TM_FAR * keyPtr, int keyLength)
{
    unsigned short         defaultKeyId;
    unsigned short         wepKey[7];
    unsigned char TM_FAR * keyDataPtr;
    int                    i;

/* Copy the user's key to our parameter buffer, padding it out to an even
   number of bytes so that tfPrismDoAccessCommand can process it correctly. */
    keyDataPtr = (unsigned char TM_FAR *) wepKey;
    for (i=0;i<keyLength;i++)
    {
        keyDataPtr[i] = keyPtr[i];
    }

    keyDataPtr[i] = 0;
    
    tfPrismDoAccessCommand(TM_PRISM_RID_WEP_KEY0,
                           wepKey,
                           keyLength+1,
                           TM_PRISM_ACCESS_WRITE);

    defaultKeyId = 0;
    tfPrismDoAccessCommand(TM_PRISM_RID_WEP_KEY_ID,
                           &defaultKeyId,
                           sizeof(defaultKeyId),
                           TM_PRISM_ACCESS_WRITE);

}

/*
 * tfPrismEnableWEP()
 */ 
static void tfPrismEnableWEP(void)
{
    unsigned short wepFlags;
    
    wepFlags = TM_PRISM_WEP_FLAGS_ENABLE;
    tfPrismDoAccessCommand(TM_PRISM_RID_WEP_FLAGS,
                           &wepFlags,
                           sizeof(wepFlags),
                           TM_PRISM_ACCESS_WRITE);
}

/*
 * tfPrismDisableWEP()
 */ 
static void tfPrismDisableWEP(void)
{
    unsigned short wepFlags;
    
    wepFlags = TM_PRISM_WEP_FLAGS_DISABLE;
    tfPrismDoAccessCommand(TM_PRISM_RID_WEP_FLAGS,
                           &wepFlags,
                           sizeof(wepFlags),
                           TM_PRISM_ACCESS_WRITE);
}

/*
 * tfPrismGetCurrentSSID()
 *
 * Retrieves the current SSID that the station is connected to. 'ssid' should
 * point to a string buffer at least 33 bytes long.
 */ 
static void tfPrismGetCurrentSSID(char TM_FAR * ssid)
{
    char tempResult[34];
    int  i;
    int  ssidLength;
    
    tfPrismDoAccessCommand(TM_PRISM_RID_CURR_SSID,
                           (unsigned short TM_FAR *) tempResult,
                           32,
                           TM_PRISM_ACCESS_READ);
    ssidLength = *((unsigned short TM_FAR *) &tempResult);

    if (ssidLength <= 32)
    {
/* Copy the string into the user's buffer, moving past the length. */
        for (i=0; i<ssidLength; i++)
        {
            ssid[i] = tempResult[2+i];
        }
        ssid[i] = '\0';
    }
    else
    {
/* SSID length from the controller is bad; just return a zero length string. */
        ssid[0] = '\0';
    }

}

/*
 * tfPrismGetCurrentBSSID()
 *
 * Gets the current BSSID (MAC adress) that the station is connected to.
 */ 
static void tfPrismGetCurrentBSSID(char TM_FAR * bssid)
{
    tfPrismDoAccessCommand(TM_PRISM_RID_CURR_BSSID,
                           (unsigned short TM_FAR *) bssid,
                           6,
                           TM_PRISM_ACCESS_READ);
}

/*
 * tfPrismGetPortStatus()
 *
 * Returns the current status of the wireless connection.
 */ 
static int tfPrismGetPortStatus(void)
{
    unsigned short portStatus;

    tfPrismDoAccessCommand(TM_PRISM_RID_PORT_STATUS,
                           &portStatus,
                           sizeof(portStatus),
                           TM_PRISM_ACCESS_READ);

    return portStatus;
}

/*
 * tfPrismGetCommsQuality()
 *
 * Returns the current quality of communications with the BSS.  The returned
 * value is based on the average signal and noise levels and ranges between
 * 0-92.
 */
static int tfPrismGetCommsQuality(void)
{
    unsigned short commQualityArr[3];

    tfPrismDoAccessCommand(TM_PRISM_RID_COMM_QUALITY,
                           commQualityArr,
                           sizeof(commQualityArr),
                           TM_PRISM_ACCESS_READ);

/* The controller returns three pieces of information: the communications
   quality and the average signal and average noise level that the quality
   number was based upon.  We're really only in the first number, so just
   return this to the caller. */

    return (int)commQualityArr[0];
}

/*
 * tfPrismGetTxRate()
 *
 * Returns the current data rate that is being used to transmit packet.
 * This number is in Mbits/sec and may range between 1-100.
 */
static int tfPrismGetTxRate(void)
{
    unsigned short txRate;

    tfPrismDoAccessCommand(TM_PRISM_RID_CURR_TX_RATE,
                           &txRate,
                           sizeof(txRate),
                           TM_PRISM_ACCESS_READ);

    return txRate;
}


/*
 * tfPrismDoAccessCommand()
 * 
 * Formulates a structure for an access command and issues the command.
 * recDataPtr points to data to include as the parameter in the
 * resource structure.
 * This routine assumes that the length of the record is even, so any
 * routine that calls this routine must ensure that the buffer is properly
 * padded, if necessary. (e.g. tfPrismSetWEPKey)
 */
static void tfPrismDoAccessCommand(unsigned short          rid,
                                   unsigned short TM_FAR * recDataPtr,
                                   int                     recDataLen,
                                   int                     readWriteFlag)
{
         int            i;
         unsigned short result;
         unsigned short resultLength;
         unsigned short recLenWords;
         unsigned short tempShort;
volatile unsigned short regValue;

/* Convert bytes into 16-bit words */
    recLenWords = recDataLen >> 1;
    
/* Copy configuration structure to BAP */
    if (readWriteFlag == TM_PRISM_ACCESS_WRITE)
    {
        tm_prism_out_word(rid, TM_PRISM_REG_SELECT0);
    
        tm_prism_out_word(0, TM_PRISM_REG_OFFSET0);
        tm_prism_wait_not_busy(TM_PRISM_REG_OFFSET0, regValue);
        if (tm_prism_offset_reg_error(regValue))
        {
/* Error writing data for Access command; return. */
            goto prismAccessCmdReturn;
        }
        
        tm_prism_out_word(recLenWords+1, TM_PRISM_REG_DATA0);
        tm_prism_out_word(rid, TM_PRISM_REG_DATA0);

/* Copy configuration parameter(s) to BAP */
        for (i=0; i<recLenWords; i++)
        {
            tm_prism_out_word(recDataPtr[i], TM_PRISM_REG_DATA0);       
        }
    }

/* Make sure that previous command has completed. */
    tm_prism_wait_not_busy(TM_PRISM_REG_CMD, regValue);
    
/* Issue 'Access' command. */
    tm_prism_out_word(rid, TM_PRISM_REG_PARAM0);

    if (readWriteFlag == TM_PRISM_ACCESS_WRITE)
    {
        tm_prism_out_word(TM_PRISM_CMD_CODE_WACCESS, TM_PRISM_REG_CMD);
    }
    else
    {
        tm_prism_out_word(TM_PRISM_CMD_CODE_ACCESS, TM_PRISM_REG_CMD);
    }
    tm_prism_cmd_wait_and_ack(regValue, result);

/* Copy configuration structure from BAP */
    if (readWriteFlag == TM_PRISM_ACCESS_READ)
    {
        tm_prism_out_word(rid, TM_PRISM_REG_SELECT0);
    
        tm_prism_out_word(0, TM_PRISM_REG_OFFSET0);
        tm_prism_wait_not_busy(TM_PRISM_REG_OFFSET0, regValue);
        if (tm_prism_offset_reg_error(regValue))
        {
/* Error read dataa for Access command; return. */
            goto prismAccessCmdReturn;
        }

        resultLength = (tm_prism_in_word(TM_PRISM_REG_DATA0) - 1);
/* Discard the RID value. */
        regValue = tm_prism_in_word(TM_PRISM_REG_DATA0);

/* Copy configuration parameter(s) to BAP */
        for (i=0; i<resultLength; i++)
        {
            recDataPtr[i] = tm_prism_in_word(TM_PRISM_REG_DATA0);
        }
    }

prismAccessCmdReturn:
    return;
}

/*
 * tfPrismAllocateFID()
 * 
 * Allocate a FID of the given size from the controller.
 */ 
static int tfPrismAllocateFID(int size)
{
    volatile unsigned short regValue;
             unsigned short result;

/* Round allocation size to an even boundary. */
    if (size & 1)
    {
        size++;
    }
    
/* Make sure that previous command has completed. */
    tm_prism_wait_not_busy(TM_PRISM_REG_CMD, regValue);

/* Issue allocation command, and wait for it to be accepted. */
    tm_prism_out_word(size, TM_PRISM_REG_PARAM0);
    tm_prism_out_word(TM_PRISM_CMD_CODE_ALLOC, TM_PRISM_REG_CMD);
    tm_prism_cmd_wait_and_ack(regValue, result);

/*
 * Allocation command has been accepted, but we still have to wait for
 * the actual allocation to complete.
 */
    do
    {
        regValue = tm_prism_in_word(TM_PRISM_REG_EVSTAT);
    } while (!(regValue & TM_PRISM_REG_EVSTAT_ALLOC));

    regValue = tm_prism_in_word(TM_PRISM_REG_ALLOCFID);
    
    tm_prism_out_word(TM_PRISM_REG_EVACK_ALLOC, TM_PRISM_REG_EVACK);

    return regValue;
}

/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trprism.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 38133    Output checksum: 44096
 */
