/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2005
 * Copyright Elmic Systems Japan 1997 - 2005
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 * Stretch S56XX Ethernet Driver
 *
 * Filename: tr56xx.c
 * Author: Jared August
 * Date Created: 8/26/2005
 * $Source: source/drivers/s56xx/trs56xx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:20CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <sx-errors.h>
#include <sx-ppi.h>
#include <sx-misc.h>
#include <sx-timer.h>
#include "trsocket.h"
#include "trs56xx.h"

#ifdef TM_56XX_USE_JUMBO_FRAMES
#define TM_56XX_MTU_SIZE 9000
#else /* TM_56XX_USE_JUMBO_FRAMES */
#define TM_56XX_MTU_SIZE 1518
#endif /* TM_56XX_USE_JUMBO_FRAMES */

#define TM_56XX_BUFFER_SIZE (TM_56XX_MTU_SIZE+18)

void tf56xxRxIsr(void * interfaceHandle);
void tf56xxTxIsr(void * interfaceHandle);

/* Information we need to keep for each Ethernet port */
tt56xxPortInfo tl56xxPort[TM_56XX_NUM_PORTS]
               __attribute__ ((section(".dram.data")));
/* Number of unique Ethernet ports configured */
unsigned char __attribute__ ((section(".dram.data"))) tl56xxPortCount = 0;

/* Buffer descriptors */
#define TM_56XX_DATARAM __attribute__ ((section(".dram.data"))) \
                        __attribute__ ((aligned(16)))
sx_dmac_descriptor rx_descriptors[TM_56XX_RX_RING_SIZE*TM_56XX_NUM_PORTS]
                                 TM_56XX_DATARAM;
sx_dmac_descriptor tx_descriptors[TM_56XX_TX_RING_SIZE*TM_56XX_NUM_PORTS]
                                 TM_56XX_DATARAM;

#if (TM_56XX_NUM_PORTS > 4) || (TM_56XX_NUM_PORTS < 1)
#error TM_56XX_NUM_PORTS must be defined to be in the range 1-4
#endif /* (TM_56XX_NUM_PORTS > 4) || (TM_56XX_NUM_PORTS < 1) */

/* Open the interface */
int tf56xxOpen(ttUserInterface interfaceHandle)
{
	sx_ppi_config     ppi_cfg;
    sx_uint32         sx_err;
    sx_ppi_link_state linkState;
    int               counter;
    tt56xxPortInfoPtr thisPort;
    unsigned char     incPortCount;

/* See if this interface has been opened before */
    thisPort = tfDeviceGetPointer(interfaceHandle);
    if (thisPort == (tt56xxPortInfoPtr)0)
    {
/* If it's new, make sure we haven't opened all of the allotted interfaces
 * already */
        if (tl56xxPortCount < TM_56XX_NUM_PORTS)
        {
            thisPort = &(tl56xxPort[tl56xxPortCount]);
            incPortCount = 1;
        }
        else
        {
            return TM_DEV_ERROR;
        }
    }
    else
    {
        incPortCount = 0;
    }

 /* Set the Ethernet address */
    thisPort->phyAddr[0] = 0x0;
    thisPort->phyAddr[1] = 0x1;
    thisPort->phyAddr[2] = 0x2;
    thisPort->phyAddr[3] = 0x3;
    thisPort->phyAddr[4] = 0x4;
    thisPort->phyAddr[5] = tl56xxPortCount;

 /* Save the interface handle */
    thisPort->interfaceHandle = interfaceHandle;

/* Set up the PPI configuration structure */
    ppi_cfg.which_ppi           = tl56xxPortCount;
	ppi_cfg.ppi_mode            = SX_PPI_MODE_GMII;
	ppi_cfg.streaming_mode      = 0;

	ppi_cfg.num_receive_desc    = TM_56XX_RX_RING_SIZE;
	ppi_cfg.receive_descriptors = &(rx_descriptors[TM_56XX_RX_RING_SIZE*
                                                   tl56xxPortCount]);
#ifdef TM_56XX_POLLED_MODE
	ppi_cfg.receive_cb          = 0;
	ppi_cfg.receive_cb_arg      = 0;
	ppi_cfg.receive_cb_freq     = 0;
	ppi_cfg.receive_cb_timeout  = 0;
#else /* TM_56XX_POLLED_MODE */
	ppi_cfg.receive_cb          = tf56xxRxIsr;
	ppi_cfg.receive_cb_arg      = thisPort;
	ppi_cfg.receive_cb_freq     = TM_56XX_RX_ISR_AGG_COUNT;
	ppi_cfg.receive_cb_timeout  = TM_56XX_RX_ISR_AGG_TIMEOUT;
#endif /* TM_56XX_POLLED_MODE */

	ppi_cfg.num_send_desc       = TM_56XX_TX_RING_SIZE;
	ppi_cfg.send_descriptors    = &(tx_descriptors[TM_56XX_TX_RING_SIZE*
                                                   tl56xxPortCount]);
#ifdef TM_56XX_POLLED_MODE
	ppi_cfg.send_cb             = 0;
	ppi_cfg.send_cb_arg         = 0;
	ppi_cfg.send_cb_freq        = 0;
	ppi_cfg.send_cb_timeout     = 0;
#else /* TM_56XX_POLLED_MODE */
	ppi_cfg.send_cb             = tf56xxTxIsr;
	ppi_cfg.send_cb_arg         = thisPort;
	ppi_cfg.send_cb_freq        = TM_56XX_TX_ISR_AGG_COUNT;
	ppi_cfg.send_cb_timeout     = TM_56XX_TX_ISR_AGG_TIMEOUT;
#endif /* TM_56XX_POLLED_MODE */

	ppi_cfg.mtu                 = TM_56XX_MTU_SIZE;
	ppi_cfg.promiscuous         = 0;
	ppi_cfg.half_duplex         = 0;
	ppi_cfg.loopback            = 0;
	ppi_cfg.rx_skip             = 0;
	ppi_cfg.rx_stride           = 0;
	ppi_cfg.tx_skip             = 0;
	ppi_cfg.tx_stride           = 0;

    ppi_cfg.station_address[0]  = thisPort->phyAddr[0];
    ppi_cfg.station_address[1]  = thisPort->phyAddr[1];
    ppi_cfg.station_address[2]  = thisPort->phyAddr[2];
    ppi_cfg.station_address[3]  = thisPort->phyAddr[3];
    ppi_cfg.station_address[4]  = thisPort->phyAddr[4];
    ppi_cfg.station_address[5]  = thisPort->phyAddr[5];

/* Initialize the PPI */
	sx_err = sx_ppi_init(&ppi_cfg, &(thisPort->ppi_handle));
    if (sx_err != SXERR_NONE)
    {
        printf("sx_ppi_init failed with error code %d\r\n", sx_err);
        while (1);
	}
    
/* Allocate receive buffers and put them in the receive ring */
    for (counter = 0; counter < TM_56XX_RX_RING_SIZE; counter++)
    {
        thisPort->rxRingInfo[counter].dataPtr =
            tfGetDriverBuffer(&(thisPort->rxRingInfo[counter].bufHandle),
                              TM_56XX_BUFFER_SIZE,
                              2);
        if (thisPort->rxRingInfo[counter].dataPtr == 0)
        {
            printf("tfGetEthernetBuffer failed!\r\n");
            while(1);
        }
        sx_err = sx_ppi_sched_receive(
               thisPort->ppi_handle,
               thisPort->rxRingInfo[counter].dataPtr,
               &(thisPort->rxRingInfo[counter].bufLength));
        if (sx_err != 0)
        {
            printf("sx_ppi_sched_receive failed with error code %d!\r\n",
                   sx_err);
            while (1);
    	}
    }

/* Wait until the link is up */
    do
    {
        sx_err = sx_ppi_get_link_state(thisPort->ppi_handle, &linkState);
        if (sx_err != SXERR_NONE)
        {
            printf("sx_ppi_get_link_state failed with error code %d\r\n",
                   sx_err);
            while (1);
    	}
    } while (linkState == SX_PPI_LINK_DOWN);

/* Wait a while longer for good measure */
    sx_delay_ms(250);    

/* Save off the pointer to this interface's info structure */
    tfDeviceStorePointer(interfaceHandle, &(tl56xxPort[tl56xxPortCount]));

/* Increment the number of configured ports */
    if (incPortCount)
    {
        tl56xxPortCount++;
    }

	return TM_DEV_OKAY;
}

/* Close an Ethernet port down */
int tf56xxClose(ttUserInterface interfaceHandle)
{
    sx_uint32         sx_err;
    tt56xxPortInfoPtr thisPort;
    int               counter;
    int               errCode;

/* Get the important info for this port */
    thisPort = tfDeviceGetPointer(interfaceHandle);

/* Close the PPI */
    sx_err = sx_ppi_close(thisPort->ppi_handle);
    if (sx_err != SXERR_NONE)
    {
        printf("sx_ppi_close failed with error code %d!\r\n", sx_err);
        while(1);
    }

/* Free all the driver buffers */
    for (counter = 0; counter < TM_56XX_RX_RING_SIZE; counter++)
    {
        errCode = tfFreeDriverBuffer(thisPort->rxRingInfo[counter].bufHandle);
        if (errCode != 0)
        {
            printf("tfFreeDriverBuffer failed!\r\n");
            while(1);
        }
    }

	return TM_DEV_OKAY;
}

/* Send a packet */
int tf56xxSend(ttUserInterface interfaceHandle,
			   char TM_FAR *   dataPtr,
			   int             dataLength,
			   int             flag)
{
    sx_uint32         sx_err;
    tt56xxPortInfoPtr thisPort;

/* Get the important info for this port */
    thisPort = tfDeviceGetPointer(interfaceHandle);

/* Flush the cache */
    sx_dcache_region_writeback(dataPtr, dataLength);
/* Send the packet */
	sx_err = sx_ppi_sched_send(thisPort->ppi_handle, dataPtr, dataLength, 1);
	if (sx_err != 0)
	{
		printf("sx_ppi_sched_send failed with error code %d\r\n", sx_err);
        while(1);
	}
	return TM_DEV_OKAY;
}

/* Receive a packet */
int tf56xxRecv(ttUserInterface        interfaceHandle,
			   char TM_FAR * TM_FAR * dataPtr,
			   int TM_FAR *           dataLength,
			   ttUserBufferPtr        bufHandlePtr)
{
    sx_uint32         sx_err;
    tt56xxPortInfoPtr thisPort;

/* Get the important info for this port */
    thisPort = tfDeviceGetPointer(interfaceHandle);

/* Set the data pointer, data length, and buffer handle */
    *dataPtr      = thisPort->rxRingInfo[thisPort->dirtyRxIndex].dataPtr;
    *dataLength   = thisPort->rxRingInfo[thisPort->dirtyRxIndex].bufLength;
    *bufHandlePtr = thisPort->rxRingInfo[thisPort->dirtyRxIndex].bufHandle;

/* Get a new buffer for the rx ring */
    thisPort->rxRingInfo[thisPort->dirtyRxIndex].dataPtr = tfGetDriverBuffer(
                   &(thisPort->rxRingInfo[thisPort->dirtyRxIndex].bufHandle),
                   TM_56XX_BUFFER_SIZE,
                   2);
    if (thisPort->rxRingInfo[thisPort->dirtyRxIndex].dataPtr == 0)
    {
        printf("tfGetEthernetBuffer failed!\r\n");
        while(1);
    }
/* Put the new buffer on the rx ring */
    sx_err = sx_ppi_sched_receive(
                 thisPort->ppi_handle,
                 thisPort->rxRingInfo[thisPort->dirtyRxIndex].dataPtr,
                 &(thisPort->rxRingInfo[thisPort->dirtyRxIndex].bufLength));
    if (sx_err != 0)
    {
        printf("sx_ppi_sched_receive failed with error code %d!\r\n",
               sx_err);
        while (1);
	}

/* Increment the dirty index */
    thisPort->dirtyRxIndex++;
    if (thisPort->dirtyRxIndex == TM_56XX_RX_RING_SIZE)
    {
        thisPort->dirtyRxIndex = 0;
    }

/* Invalidate the cache for the new packet */
    sx_dcache_region_invalidate(*dataPtr, *dataLength);

    return TM_DEV_OKAY;
}

/* The ioctl function - not necessary for this driver */
int tf56xxIoctl(ttUserInterface interfaceHandle,
				int             flag,
				void TM_FAR *   optionPtr,
				int             optionLen)
{
	return TM_DEV_OKAY;
}

/* Return the physical address for this Ethernet port */
int tf56xxGetPhyAddr(ttUserInterface interfaceHandle,
					 char TM_FAR *   physicalAddress)
{
    tt56xxPortInfoPtr thisPort;

    thisPort = tfDeviceGetPointer(interfaceHandle);

    physicalAddress[0] = thisPort->phyAddr[0];
    physicalAddress[1] = thisPort->phyAddr[1];
    physicalAddress[2] = thisPort->phyAddr[2];
    physicalAddress[3] = thisPort->phyAddr[3];
    physicalAddress[4] = thisPort->phyAddr[4];
    physicalAddress[5] = thisPort->phyAddr[5];

	return TM_DEV_OKAY;
}

#ifndef TM_56XX_POLLED_MODE
/* The receive interrupt handler */
void tf56xxRxIsr(void * port)
{
    sx_ppi_stat       stat;
    tt56xxPortInfoPtr thisPort;

    thisPort = port;

/* Get how many packets were received */
    (void)sx_ppi_receive_stat(thisPort->ppi_handle, &stat);
/* Tell the stack how many were received */
    tfNotifyInterfaceIsr(thisPort->interfaceHandle,
                         stat.incr_count,
                         0,
                         0,
                         0);
}

/* The send complete interrupt handler */
void tf56xxTxIsr(void * port)
{
    sx_ppi_stat       stat;
    tt56xxPortInfoPtr thisPort;

    thisPort = port;

/* Get how many packets were transmitted */
    (void)sx_ppi_send_stat(thisPort->ppi_handle, &stat);
/* Tell the stack how many were transmitted */
    tfNotifyInterfaceIsr(thisPort->interfaceHandle,
                         0,
                         stat.incr_count,
/* Note: 2048 is the low-water mark to get tfCheckSentInterface() or
 *       tfWaitSentInterface() to wake up. This number isn't used for
 *       anything else, so we set it big enough here to wake up the
 *       app-level functions every time */
                         2049,
                         0);
}
#endif /* TM_56XX_POLLED_MODE */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trs56xx.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 37353    Output checksum: 57689
 */
