/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: SMC9000 Device Driver
 * Filename: trsmc9.c
 * Author: Jason
 * Date Created: 2/1/2000
 * $Source: source/drivers/smc/trsmc.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:22CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* This driver should function on any of the SMSC 'LAN9000' family of
 * Ethernet controllers.  There are a number of things that it may be 
 * necessary to modify to match your platform:
 *
 *  1. Specify the base I/O address and the IRQ with TM_SMC_IO_ADDR and 
 *      TM_SMC_IRQ.
 *  2. tfSmcSetupHardware sets up any external hardware necessary for the
 *      SMC (e.g., setting various pin direction, etc)
 *  3. tfSmcSetupInterrupts installs an ISR to handle interrupts from the
 *      SMC as well as enabling interrupts on both the interrupt controller
 *      and the CPU (but not the Ethernet chip itself).
 *  4. tm_clear_cpu_interrupts should acknowledge interrupts to the interrupt
 *      controller (not the SMC).
 *  5. tm_smc_in... / tm_smc_out... macros should be modified to match your
 *      platform and compiler.
 *  6. tfSmcGetPhysAddr should return the hardware address of this Ethernet
 *      controller (either hardcoded, read from EEPROM, etc).
 *
 *
 * NOTES:
 * ------ 
 *  1. Doesn't probe for a SMC9000 chip; I/O address, IRQ, etc must be
 *      specified using the macros below.
 *  2. For maximum performance, please see the 'OPTIMIZATION NOTE' in
 *      tfSmcSend and tfSmcReceive.
 *  3. This driver was developed on a standard Pentium class PC using the
 *      SMSC 91c100 Ethernet controller and the Borland compiler.
 *
 */

#include <trsocket.h>
#include <trsmc.h>

/* For tm_get_raw_buffer */
#include <trmacro.h>
#include <trtype.h>
#include <trglobal.h>
#include <trproto.h>
#include <trsmc.h>

/* For outp,outpw,inp,inpw for Borland compiler */
#include <conio.h>

/* Base I/O address and IRQ */
#define TM_SMC_IO_ADDR  0x300
#define TM_SMC_IRQ      3


/*
 * Percentage of total on-chip memory to use for transmit buffers.
 * Setting to zero would not limit buffer usage (ie, either transmit
 * or receive could fill all of on-chip RAM with buffers).
 *
 * It is important that the number defined below doesn't have parantheses
 * around it...
 */
#define TM_XMIT_MEM_SIZE 1/2



/*
 * Our receive pool consists of two sized buffers: small and large.
 * These macros determine the size and number of these buffers in our
 * pre-allocated receive pool.
 */
#define TM_SMC_LARGE_RECV_POOL     4
#define TM_SMC_SMALL_RECV_POOL     10
#define TM_SMC_LARGE_RECV_BUF_SIZE 1518
#define TM_SMC_SMALL_RECV_BUF_SIZE 128

/* 
 * This macro determines on what sort of byte boundary our buffers are aligned 
 * (word, long word, etc).
 */
#define TM_SMC_BUF_ALIGNMENT       4

/* Various 8259 PIC register definitions */
#define TM_PIC1_ISR_ADDR 0x20
#define TM_PIC2_ISR_ADDR 0xA0
#define TM_PIC_CLEAR_INT 0x20
#define TM_PIC1_IMR_ADDR 0x21
#define TM_PIC2_IMR_ADDR 0xA1

/* Acknowledges interrupts to the PIC */
#define tm_clear_cpu_interrupts(intLine) \
    if ( (intLine) < 8 ) \
    { \
        outp( TM_PIC1_ISR_ADDR, TM_PIC_CLEAR_INT ); \
    } \
    else \
    { \
        outp( TM_PIC2_ISR_ADDR, TM_PIC_CLEAR_INT ); \
        outp( TM_PIC1_ISR_ADDR, TM_PIC_CLEAR_INT ); \
    }

/* Port I/O commands */
#define tm_smc_in_byte(offset) inp(TM_SMC_IO_ADDR+offset)
#define tm_smc_in_word(offset) inpw(TM_SMC_IO_ADDR+offset)

#define tm_smc_out_byte(value,offset) outp(TM_SMC_IO_ADDR+offset,value)
#define tm_smc_out_word(value,offset) outpw(TM_SMC_IO_ADDR+offset,value)

#define tm_smc_select_bank(x) tm_smc_out_word(x,TM_SMC_BANK_SELECT)

#define tm_smc_flip_word(x) ( (((x) & 0xff) << 8) | (((x) & 0xff00) >> 8) )


/* 
 * SMC9000 Register Offsets 
 */

/* (All Banks) */
#define TM_SMC_BANK_SELECT 0xE

/* (Bank 0) */
#define TM_SMC_TCR 0x0
#define TM_SMC_RCR 0x4
#define TM_SMC_MIR 0x8
#define TM_SMC_MCR 0xA

/* (Bank 1) */
#define TM_SMC_CONFIG     0x0
#define TM_SMC_IND_ADDR_0 0x4
#define TM_SMC_IND_ADDR_2 0x6
#define TM_SMC_IND_ADDR_4 0x8
#define TM_SMC_CONTROL    0xC

/* (Bank 2) */
#define TM_SMC_MMU_CMD   0x0
#define TM_SMC_PKT_NUM   0x2
#define TM_SMC_ALLOC_RES 0x3
#define TM_SMC_FIFO      0x4
#define TM_SMC_POINTER   0x6
#define TM_SMC_DATA1     0x8
#define TM_SMC_DATA2     0xA
#define TM_SMC_INT_STAT  0xC
#define TM_SMC_INT_ACK   0xC
#define TM_SMC_INT_MASK  0xD

/* (Bank 3) */
#define TM_SMC_MULTICAST0 0x0
#define TM_SMC_MULTICAST2 0x2
#define TM_SMC_MULTICAST4 0x4
#define TM_SMC_MULTICAST6 0x6

/* 
 * Interrupt status/mask bits 
 */
/* Frame received */
#define TM_INT_RX           0x01
/* Transmit error */
#define TM_INT_TX           0x02
/* Tx FIFO empty */
#define TM_INT_TX_EMPTY     0x04
/* Transmit buffer allocation complete */
#define TM_INT_ALLOC        0x08
/* Receive overrun */
#define TM_INT_RX_OVRN      0x10
/* Ethernet protocol handler interrupt */
#define TM_INT_EPH          0x20
/* Early receive interrupt */
#define TM_INT_ERCV         0x40
/* Transmit Idle interrupt */
#define TM_INT_TX_IDLE      0x80



/* Flag indicating that an odd byte is present in this frame */
#define TM_CTL_ODD_BYTE     0x2000

/* 'Magic number' used to identify the SMC9000 chip */
#define TM_SMC_MAGIC_NUM    0x3300

/* SMC command: perform a soft reset of the chip */
#define TM_SMC_SOFT_RST     0x8000

/* 
 * SMC control register flags 
 */
/* Automatically release sent packets */
#define TM_AUTO_RELEASE     0x0800
/* Receive bad packets, rather than immediately reject them */
#define TM_RECV_BAD_PACKETS 0x4000
/* Enable transmission of packets */
#define TM_XMIT_ENABLE      0x0001
/* Enable reception of packets */
#define TM_RECV_ENABLE      0x0100

/* Automatically increment the internal SMC pointer as we send data */
#define TM_PTR_AUTO_INC     0x4000

/* 
 * MMU commands 
 */
/* Allocate chip memory for transmit */
#define TM_MMU_ALLOC        0x0020
/* Reset the MMU */
#define TM_MMU_RESET        0x0040
/* Release received packet */
#define TM_MMU_REL_RX_PKT   0x0080
/* Free a packet queued for transmission */
#define TM_MMU_FREE_PKT     0x00a0
/* Enqueue a packet for transmission */
#define TM_MMU_ENQUEUE      0x00c0


/*
 * Pointer register flags
 */
/* Specifies that a read operation (vs a write operation) is to follow */
#define TM_PTR_READ_OPERATION 0x2000
/* Auto increment the pointer for each read operation */
#define TM_PTR_AUTO_INCREMENT 0x4000
/* Use the top of the RX FIFO as our pointer, not the one set in this
   register */
#define TM_PTR_RCV_RX_FIFO    0x8000


/* Polynomial used for 32-bit CRC calculation (used for multicast addresses) */
#define TM_CRC32_POLY 0x04c11db7UL


/* Send allocation request states */
#define TM_SMC_ALLOC_IDLE    0
#define TM_SMC_ALLOC_WAITING 1


#define tm_smc_alloc_approved(statReg) ((statReg) & TM_INT_ALLOC)


/*
 * SMC information structure 
 */
typedef struct tsSmcInfo
{

/* Treck interface handle for this device */
    ttUserInterface smcInterfaceHandle;

/* 
 * Current state of our send allocation request, either waiting or idle 
 * If send queue is not enabled, the state will always be idle.
 */
    unsigned char   smcAllocState;

/* Flag indicating if a transmit queue is available for this interface */
    unsigned char   smcXmitQueueEnb;

/* Interupt line (eg, IRQ) used for the SMC */
    unsigned char   smcIntLine;

} ttSmcInfo;

typedef ttSmcInfo TM_FAR * ttSmcInfoPtr;


ttSmcInfoPtr tlSmcPtr;


void interrupt tfSmcHandlerIsr(void);

/*
 * tfSmcSetupHardware()
 *
 *   Setup any hardware related to the SMC9000 (assign pins, etc).  This will
 *   be called as soon as the SMC open routine begins (ie, before any other
 *   activity: initialization of the chip, buffer management) takes place.
 *
 */
int tfSmcSetupHardware(void)
{
    return TM_DEV_OKAY;
}


/*
 * tfSmcSetupInterrupts()
 *
 *   Setup and enable any hardware needed to enable 'intLine'.  For instance,
 *   the routine below sets up the interrupts on a typical x86 PC (masks the
 *   interrupt in the interrupt controller (PIC), etc).  
 *   Install the ISR handler to the appropriate vector location
 *   (ie, IRQ3 @ 0xB, IRQ10 @ 0x72).
 *   This routine is only to setup the interrupts for hardware external to the
 *   SMC (the PIC, processor, etc).  The normal initialization routine 
 *   (tfSmcOpen) will set up interrupts inside the SMC controller.
 */
int tfSmcSetupInterrupts(unsigned char intLine)
{
    unsigned char   inByte;

/* If IRQ is 0-7, use PIC1 & vector base 0x08 */
    if ( intLine < 8 )
    {
        inByte = inp(TM_PIC1_IMR_ADDR);
        inByte &= ~( 1 << intLine );
        outp(TM_PIC1_IMR_ADDR, inByte);
    
        tfKernelInstallIsrHandler((ttUserIsrHandlerPtr) tfSmcHandlerIsr,
                                  0x08 + intLine);

    }
/* If IRQ is 8-15, use PIC2 & vector base 0x70 */
    else
    {
        inByte = inp(TM_PIC2_IMR_ADDR);
        inByte &= ~( 1 << (intLine-8) );
        outp(TM_PIC2_IMR_ADDR, inByte);

        tfKernelInstallIsrHandler((ttUserIsrHandlerPtr) tfSmcHandlerIsr,
                                  0x70 + (intLine - 8) );

    }

    return TM_DEV_OKAY;
}     


/* 
 * tfSmcOpen()
 *
 * 1. Verify that this is a SMC9000 chip that we're dealing with (ie,eheck 
 *    the I/O address).
 * 2. Get the amount of memory available on the chip.
 * 3. Reset the chip.
 * 4. Set various registers to default values.
 * 5. Reset the chip's MMU.
 * 6. Setup the Ethernet HW address 
 * 7. Setup interrupt handler, etc.
 * 8. Enable transmit and receive
 * 9. Enable interrupts.
 */
int tfSmcOpen(ttUserInterface interfaceHandle)
{
    unsigned short  inWord;
    int             errorCode;
    unsigned short  delayCount;
    char            physAddr[6];
    unsigned short  memSize;

/* PARAMETER CHECK: make sure that scatter sends are not enabled */
    if ( tfInterfaceScatterCount(interfaceHandle) > 1 )
    {
        tfKernelWarning("tfSmcOpen",
                        "Scatter send not supported on this device");

        errorCode = TM_DEV_ERROR;
        goto smcOpenEnd;
    }

/* Perform any external hardware initilization necessary */
    (void)tfSmcSetupHardware();

/* Verify that this is a SMC9000 chip (ie, check the I/O address). 
 * To do this, we read a 16-bit value from the bank select register.
 * The upper byte of this word should always be 0x33.
 */
    inWord = tm_smc_in_word(TM_SMC_BANK_SELECT);
    if ( (inWord & 0xff00) == TM_SMC_MAGIC_NUM )
    {

/* Allocate memory for the info structure */
        tlSmcPtr = tm_get_raw_buffer( sizeof(ttSmcInfo) );

        if (tlSmcPtr == 0)
        {
/* No memory for info structure; quit now and return error */
            errorCode = TM_DEV_ERROR;
            goto smcOpenEnd;
        }        
        (void)tfDeviceStorePointer(interfaceHandle, (ttVoidPtr)tlSmcPtr); 

/* 
 * Initialize the information structure 
 */
        tlSmcPtr->smcAllocState      = TM_SMC_ALLOC_IDLE;
        tlSmcPtr->smcInterfaceHandle = interfaceHandle;
        tlSmcPtr->smcXmitQueueEnb    = tfInterfaceXmitQueueOn(interfaceHandle);
        tlSmcPtr->smcIntLine         = TM_SMC_IRQ;
      
/* Reset the chip */
        tm_smc_select_bank(0);
        tm_smc_out_word( TM_SMC_SOFT_RST, TM_SMC_RCR );

/* Delay for some amount of time to allow the chip to reset... */
        for (delayCount=0;delayCount<10000;delayCount++)
        {
            ;
        }

/* Clear the RCR register - this terminates the reset process */
        tm_smc_out_word(0, TM_SMC_RCR);        

/* Partition the on-chip RAM in sections for transmit and receive buffers */
        memSize = (tm_smc_in_word(TM_SMC_MIR) & 0x00ff) * TM_XMIT_MEM_SIZE;
        inWord = tm_smc_in_word(TM_SMC_MCR);
        inWord = (inWord & 0xff00) | memSize;
        tm_smc_out_word(inWord, TM_SMC_MCR);
       
/* Setup the control register:
 *  + Don't receive bad packets
 *  + Automatically free transmitted packets from on-board memory.
 */ 
        tm_smc_select_bank(1);
        inWord = tm_smc_in_word(TM_SMC_CONTROL);
        inWord |= TM_AUTO_RELEASE;
        inWord &= ~TM_RECV_BAD_PACKETS;
        tm_smc_out_word(inWord, TM_SMC_CONTROL);

/* Reset the chip's MMU */
        tm_smc_select_bank(2);
        tm_smc_out_word(TM_MMU_RESET, TM_SMC_MMU_CMD);

/* Initialize receive buffer pool */
        (void)tfPoolCreate(interfaceHandle,
                     TM_SMC_LARGE_RECV_POOL,
                     TM_SMC_SMALL_RECV_POOL,
                     TM_SMC_LARGE_RECV_BUF_SIZE,
                     TM_SMC_SMALL_RECV_BUF_SIZE,
                     TM_SMC_BUF_ALIGNMENT,
                     TM_POOL_REFILL_IN_LINE );

/* Setup Ethernet HW address */
        tm_smc_select_bank(1);
        (void)tfSmcGetPhysAddr(interfaceHandle, physAddr);
        tm_smc_out_word(*((unsigned short *) &physAddr[0]), TM_SMC_IND_ADDR_0);
        tm_smc_out_word(*((unsigned short *) &physAddr[2]), TM_SMC_IND_ADDR_2);
        tm_smc_out_word(*((unsigned short *) &physAddr[4]), TM_SMC_IND_ADDR_4);

/* Setup interrupt handler, etc */
        errorCode = tfSmcSetupInterrupts(tlSmcPtr->smcIntLine);


        if (errorCode == TM_DEV_OKAY)
        {
/* Enable transmit and receive */
            tm_smc_select_bank(0);
            tm_smc_out_word(TM_XMIT_ENABLE, TM_SMC_TCR);
            tm_smc_out_word(TM_RECV_ENABLE, TM_SMC_RCR);        

            tm_smc_select_bank(1);
            inWord = tm_smc_in_word(TM_SMC_CONFIG);

/* Enable interrupts (on receive and transmit error only) */
            tm_smc_select_bank(2);
            tm_smc_out_word(TM_INT_EPH | TM_INT_RX, TM_SMC_INT_MASK);

/* Device initialized successfully */
            errorCode = TM_DEV_OKAY;
        }        

    }
    else
    {
/* Couldn't verify existence of SMC9000 chip */
        errorCode = TM_DEV_ERROR;
    }


smcOpenEnd:
    return errorCode;          
}



/*
 * tfSmcClose()
 *
 *   This function is called when the user wishes to close this device.
 *   We need to clean up any allocated memory, disable any chip interrupts,
 *   and disable transmit/receive on the chip.  While we do need to free memory
 *   that was allocated for the receive pool (by calling tfPoolDelete) it is
 *   not necessary to clean up any transmit buffer, since they are freed 
 *   immediately upon return from tfSmcSend().
 *
 */
int tfSmcClose(ttUserInterface interfaceHandle)
{
    unsigned short  inWord;

/* Disable all interrupts */
    tm_smc_select_bank(2);
    tm_smc_out_word(0, TM_SMC_INT_MASK);

/* Disable transmit */
    tm_smc_select_bank(0);
    inWord = tm_smc_in_word(TM_SMC_TCR);
    inWord &= ~TM_XMIT_ENABLE;
    tm_smc_out_word(inWord, TM_SMC_TCR);
    
/* Disable receive */    
    inWord = tm_smc_in_word(TM_SMC_RCR);
    inWord &= ~TM_RECV_ENABLE;
    tm_smc_out_word(inWord, TM_SMC_RCR);
/* 
 * Free memory from pre-allocated buffers in the receive ring 
 */
    (void)tfPoolDelete(interfaceHandle);

/* Free device driver variable */
    tlSmcPtr = (ttSmcInfoPtr)tfDeviceClearPointer(interfaceHandle);
    if ( tlSmcPtr != (ttSmcInfoPtr)0 )
    {
/* Free malloc'd memory */
        tm_free_raw_buffer((void TM_FAR *)tlSmcPtr);
        tlSmcPtr = (ttSmcInfoPtr)0;
    }

    return TM_DEV_OKAY;
}



/* 
 * tfSmcIoctl()
 *
 *   This I/O control routine performs two functions:
 *
 *   1.  Setup the list of multicast addresses.
 *   2.  Set the SMC to recieve all multicast packets.  This is done by writing
 *       all 1's to the SMC's four multicast hash registers.
 *
 */
int tfSmcIoctl(ttUserInterface interfaceHandle,
               int             flag,
               void TM_FAR *   optionPtr,
               int             optionLen)
{
#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
    unsigned long               calcCRC;
    unsigned char               currByte;
    unsigned char               byte, bit;
    unsigned long               hashEntry;
    unsigned char      TM_FAR * multiAddr;
    unsigned short              logicalAddrFilter[4];
    int                         index;
#endif /* TM_IGMP or TM_6_USE_MLD */



    interfaceHandle=interfaceHandle;

#if (defined(TM_IGMP) || defined(TM_6_USE_MLD))
/*
 * Setup our list of multicast addresses to receive.
 */
    if (flag & TM_DEV_SET_MCAST_LIST)
    {

/* 
* Clear the SMC's address filter (since this IOCTL call will add all the
* entries we need.
*/
        logicalAddrFilter[0] = 0;
        logicalAddrFilter[1] = 0;
        logicalAddrFilter[2] = 0;
        logicalAddrFilter[3] = 0;

/* Get the first address */
        multiAddr = (unsigned char TM_FAR *) optionPtr;

        while (optionLen-- > 0)
        {
        

/* Initialize the calculated CRC (with all ones) */
            calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
            for (byte=0;byte<6;byte++)
            {
                currByte = multiAddr[byte];

/* For each bit in the current byte */
                for (bit=0;bit<8;bit++)
                {

/* Shift the calculated CRC left by one */
                    calcCRC <<= 1;

/* If the current bit is one... */
                    if ((currByte >> bit) & 0x01)
                    {

/* ... XOR the calculated CRC with the CRC-32 polynomial */
                        calcCRC ^= TM_CRC32_POLY;

                    }                   
                }
            }

/* Get the most significant 6 bits from the calculated CRC */        
            hashEntry = (calcCRC & 0xfc000000UL) >> 26;
            index = (hashEntry >> 4) & 0xFF;
/* 
* Decode the above number and set the appropriate bit in the SMC logical
* address filter 
*/
            logicalAddrFilter[index] |=
                1 << ((unsigned char) (hashEntry % 16));

            multiAddr += 6;

        }


        tm_smc_select_bank(3);
        tm_smc_out_word(logicalAddrFilter[0], TM_SMC_MULTICAST0);
        tm_smc_out_word(logicalAddrFilter[1], TM_SMC_MULTICAST2);
        tm_smc_out_word(logicalAddrFilter[2], TM_SMC_MULTICAST4);
        tm_smc_out_word(logicalAddrFilter[3], TM_SMC_MULTICAST6);
        tm_smc_select_bank(2);
    }

/*
 * Set the SMC to receive all multicast packets.
 */
    if (flag & TM_DEV_SET_ALL_MCAST)
    {
        tm_smc_select_bank(3);            
        tm_smc_out_word(0xFFFF, TM_SMC_MULTICAST0);
        tm_smc_out_word(0xFFFF, TM_SMC_MULTICAST2);
        tm_smc_out_word(0xFFFF, TM_SMC_MULTICAST4);
        tm_smc_out_word(0xFFFF, TM_SMC_MULTICAST6);
        tm_smc_select_bank(2);
    }

#else /* !(TM_IGMP or TM_6_USE_MLD) */
/* Avoid compiler warnings when IGMP is disabled */
    interfaceHandle=interfaceHandle;
    optionPtr=optionPtr;
    optionLen=optionLen;
    flag=flag;
#endif /* !(TM_IGMP or TM_6_USE_MLD) */

    return TM_DEV_OKAY;
}






/*
 * tfSmcReceive()
 *
 *   This routine is called by the Treck stack to receive a packet.  Since
 *   the data has already been copied inside of the ISR, all we need to do
 *   in this routine is to pass the buffer up to the Treck stack.  
 */
int tfSmcReceive(ttUserInterface        interfaceHandle,
                 char TM_FAR * TM_FAR * dataPtrPtr,
                 int  TM_FAR *          dataLengthPtr,
                 ttUserBufferPtr        bufHandlePtr)
{
    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          bufHandlePtr );
}


/*
 * tfSmcGetPhysAddr()
 *
 *   Returns the physical address of this Ethernet interface.  This is just
 *   sample data and this routine should be modified to return real data.
 *
 */
int tfSmcGetPhysAddr(ttUserInterface interfaceHandle,
                     char TM_FAR *   physicalAddress)
{
    interfaceHandle=interfaceHandle;

    physicalAddress[0] = 0x00;
    physicalAddress[1] = 0x80;
    physicalAddress[2] = 0x0F;
    physicalAddress[3] = 0xFD;
    physicalAddress[4] = 0xB0;
    physicalAddress[5] = 0x22;

    return TM_DEV_OKAY;
}

/*
 * tfSmcHandlerIsr()
 *
 * A number of SMC registers need to be saved and restored in this routine: 
 * the bank select register, the pointer register and the packet number
 * register.
 *
 * Two types of SMC interrupts must be dealt with in this ISR.  The first is a 
 * transmit error.  This interrupt is fired when there is a error transmitting
 * a packet.  The only thing that has to be done to deal with this interrupt
 * is to free the packet from the chip.
 *
 * The second, and most important type, is the receive interrupt.  This 
 * interrupt is fired when a packet is received and ready to be read by the
 * CPU.  We first check to make sure that we have a buffer available in our
 * pool; if none is available, we drop the packet.  Otherwise we copy the
 * entire packet into our pool, which will be passed up to the Treck stack
 * in the receive routine. In either case, at the end of this section of the
 * ISR we instruct the chip to release the most recently received packet.
 *xxx
 */
void interrupt tfSmcHandlerIsr(void)
{
    unsigned char           intStatus;
    unsigned char           tcr;
    unsigned short          regBank;
    unsigned short          regPointer;
    unsigned short          pktNumber;
    unsigned short          pktLength;
    unsigned short          copyLength;
    unsigned short TM_FAR * dataPtr;
    unsigned char           recvNotifyCount;
    unsigned short          inWord;

/* Save on chip registers (current bank, pointer, and packet number) */
    regBank = tm_smc_in_word(TM_SMC_BANK_SELECT);

/* Register bank 2 used through entire routine */
    tm_smc_select_bank(2);

    regPointer = tm_smc_in_word(TM_SMC_POINTER);

    pktNumber = tm_smc_in_word(TM_SMC_PKT_NUM);

    recvNotifyCount = 0;

    do
    {

/* Read the interrupt status register */
        intStatus = (unsigned char)(  tm_smc_in_byte(TM_SMC_INT_STAT) 
									& (TM_INT_EPH | TM_INT_RX) );

/* 
 * Ethernet protocol handler interrupt
 */
        if (intStatus & TM_INT_EPH)
        {
/* Check the the TXENA bit of the TCR(Transmission control register) */

/* Switch to Register bank 0 - control registers */
            tm_smc_select_bank(0);
            tcr = tm_smc_in_byte(TM_SMC_TCR);
            if ((tcr & TM_XMIT_ENABLE)== TM_8BIT_ZERO)
            {
/* The TXENA bit is cleared, this means a transmission error has occured
 * We need to free this packet from the chip's transmit FIFO.  
 * Set the current packet number to the packet at the 
 * top of the transmit FIFO, then instruct the MMU to free this errored
 * packet.  Currently, we don't record the error condition, but this is
 * where that could be done.
 */
                pktNumber = tm_smc_in_byte(TM_SMC_FIFO) & 0x7f;
                tm_smc_out_byte(pktNumber, TM_SMC_PKT_NUM);
                tm_smc_out_byte(TM_MMU_FREE_PKT, TM_SMC_MMU_CMD);

/* Re enable the transmission */
                tm_smc_select_bank(0);
                tm_smc_out_word(TM_XMIT_ENABLE | tm_smc_in_word(TM_SMC_TCR), 
                                TM_SMC_TCR);  
                tm_smc_select_bank(2);
            }

/* Switch back to Register bank 2 - interrupt etc */
            tm_smc_select_bank(2);
        }
       
/*
 * Received packet interrupt 
 */
        else if (intStatus & TM_INT_RX)
        {

/* 
 * Setup the read operation: Read from the top of the RX FIFO and auto
 * increment the pointer for each read.
 */
            tm_smc_out_word( TM_PTR_RCV_RX_FIFO |
                                TM_PTR_AUTO_INCREMENT |
                                TM_PTR_READ_OPERATION,
                             TM_SMC_POINTER );


/* Read the status of this packet */
            inWord = tm_smc_in_word(TM_SMC_DATA1);

/* Get the length of this packet */
            pktLength = tm_smc_in_word(TM_SMC_DATA1) & 0x07ff;

/* Remove the length of the status word, etc from the packet length */
            pktLength -= 6;

/* Add length of odd byte */
            if (inWord & 0x2000)
            {
                pktLength++;
            }


            dataPtr = (unsigned short TM_FAR *) 
                        tfPoolIsrGetBuffer(tlSmcPtr->smcInterfaceHandle,
                                           pktLength);
                
/* Copy the data from the chip into our buffer */
            if (dataPtr != (unsigned short TM_FAR *) 0)
            {

                copyLength = pktLength >> 1;
                while (copyLength--)
                {
                    *dataPtr++ = tm_smc_in_word(TM_SMC_DATA1);
                }

/* 
 * OPTIMIZATION NOTE: To acheive the best performance with this driver, the
 * above loop should be optimized in assembly.  For instance, on the x86 
 * platform, a significant performance boost could be had by using the string
 * input instruction.  Also, the SMC also makes a second data port 
 * (TM_SMC_DATA2) available at the location immediately following the first
 * data port.  This would allow a single 32-bit value to be read at once,
 * if possible on your platform which would also significantly improve
 * performance.
 */
            
/* Catch the odd byte (if present) at the end of the packet */
                if (pktLength & 1)
                {
                    *((unsigned char TM_FAR *) dataPtr) = 
                    tm_smc_in_word(TM_SMC_DATA1) & 0x00FF;
                }

/* Count the number of frames to notify for */
                recvNotifyCount++;
            }
            else
            {
/* No room to receive this packet: drop it */                
                inWord++;    
            }

/* 
 * Whether or not we could receive this packet, instruct the chip to release
 * this packet from its internal RAM - this acknowledges the interrupt.
 */
            tm_smc_out_word(TM_MMU_REL_RX_PKT, TM_SMC_MMU_CMD);

        }

    }
    while (intStatus != 0);


/* Restore on chip registers (register bank, pointer, packet number) */
    tm_smc_out_word(regPointer,TM_SMC_POINTER);
    tm_smc_out_word(pktNumber,TM_SMC_PKT_NUM);
    tm_smc_select_bank(regBank);

/* Clear the CPU interrupts */
    tm_clear_cpu_interrupts(tlSmcPtr->smcIntLine);

/* 
 * Now that the ISR is mostly done, we notify the stack for each packet that
 * we received.  This is done at the end of the ISR because the call to
 * tfNotifyInterfaceIsr tends to release our critical section 
 * (especially in uC/OS on the x86).
 */
    tfNotifyInterfaceIsr(tlSmcPtr->smcInterfaceHandle, 
                         recvNotifyCount, 0, 0UL, 0UL);
    return;
}



/*
 * tfSmcSend()
 *
 *   The logic determining how and when to allocate memory, and if this memory
 *   allocate request is the most complicated part of this routine.  This 
 *   driver will work with either the transmit task enabled or disabled.  For
 *   efficiency reasons, it is recommended that the transmit task be enabled.
 *   Here's a breakdown of the logic:
 *
 *      1. If we are currently waiting for allocation notification...
 *           (this will only happen if the transmit queue is enabled)
 *        A. If our previous allocation was approved, acknowledge the 
 *           allocation, continue and send the packet.
 *        B. If we are still waiting on an allocation, return an error 
 *           (the upper layer transmit queue will handle queuing the packet)
 *      2. Otherwise, if we are not waiting on a previous allocate:
 *        A. Request an allocation from the chip.
 *        B. If the this allocation was approved, acknowledge the allocation,
 *           continue and send the packet.
 *        C. If this allocation fails...
 *          a. If the transmit queue is enabled, mark the fact that we're 
 *             waiting for an allocation notification, and return an error.
 *             The upper layer transmit queue will handle the current packet.
 *          b. If the transmit queue is not enabled, spin lock until the 
 *             allocation is approved.
 *
 *   So, if the transmit queue is not enabled, all we do is request some
 *   memory, wait for it to be approved and send the packet.  Things only get
 *   complicated when the transmit queue is enabled.
 *
 *   Once this is done, and we have some on-chip memory, we copy the packet to
 *   the chip by writing to the DATA1 port.  We first write out the status and
 *   length words, and then actual data.  When we reach the last word, we
 *   check if there is an extra byte (ie, length is odd).  If there is we copy
 *   this as well as the control byte indicating that this byte is present.  
 *   Otherwise we simply send a zero.  Finally we instruct the chip to enqueue
 *   this packet in the outgoing FIFO and call the tfSendCompleteInterface to
 *   inform the Treck stack that we're finished with this packet.
 *
 */
int tfSmcSend(ttUserInterface interfaceHandle,
              char TM_FAR *   dataPtr,
              int             dataLength,
              int             flag)
{
    int                     mmuPages;
    volatile unsigned char  intStatus;
    unsigned char           packetNum;
    int                     pktCounter;
    unsigned short TM_FAR * pktDataPtr;
    unsigned short          outWord;
    int                     errorCode;


/* Flags not used since we don't support scattered sends */
    flag=flag;

    errorCode = TM_DEV_OKAY;

/* We are currently waiting on a send allocation request */
    if ( tlSmcPtr->smcAllocState == TM_SMC_ALLOC_WAITING )
    {

        tm_smc_select_bank(2);                    
        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Previous allocate approved: acknowledge it, set state and continue */    
            tm_smc_out_byte(TM_INT_ALLOC, TM_SMC_INT_STAT);

            tlSmcPtr->smcAllocState = TM_SMC_ALLOC_IDLE;
        }
        else
        {
/* Previous allocate still in progress, return error to transmit queue */
            errorCode = TM_DEV_ERROR;
        }
    }

/* No pending allocation request; issue new one for this packet */
    else   
    {
    
/* 
 * The number of 256 byte buffers to allocate for this packet, including
 * the control, status and length words
 */    
        if (dataLength & 1)
        {
/* odd sized packet, extra byte will be picked up by control word */
            mmuPages = ((dataLength+5) >> 8) & 0x0007;
        }
        else
        {
            mmuPages = ((dataLength+6) >> 8) & 0x0007;
        }

/* Allocate on chip memory for this buffer */    
        tm_smc_select_bank(2);    
        tm_smc_out_word(TM_MMU_ALLOC | mmuPages, TM_SMC_MMU_CMD);

        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Allocation approved: acknowledge and go on to send packet */
            tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );

        }
        else
        {

/* 
 * Allocation failed: either simply return an error (if the transmit queue
 * is enabled) or spin lock otherwise.
 */

            if ( tlSmcPtr->smcXmitQueueEnb )
            {
                tlSmcPtr->smcAllocState = TM_SMC_ALLOC_WAITING;

                errorCode = TM_DEV_ERROR;
            }
            else
            {
/* 
 * Wait (spin lock) until on-chip memory is available.  To determine if the
 * on-chip memory has been allocated, we poll the interrupt status register 
 * (note however, that an interrupt is *not* generated, because we have it
 * masked off). 
 */
                do 
                {

/* Yield processor to other tasks while we're waiting */
                    (void)tfInterfaceSpinLock(interfaceHandle);

/* Poll to check in allocation has completed (read the interrupt status
   byte) */
                    tm_smc_select_bank(2);                    
                    intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);


                } while ( !tm_smc_alloc_approved(intStatus) );

/* Allocation approved: acknowledge and go on to send packet */
                tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );
            }

        }

    }


/* No memory available; do not send packet and return with an error */
    if (errorCode != TM_DEV_OKAY)
    {
        goto smcSendReturn;
    }


/* Memory allocation is now complete: we can write our packet to the chip */

/* 
 * Get the packet number of the last allocated packet; this is where we'll copy
 * the data.
 */    
    packetNum = tm_smc_in_byte(TM_SMC_ALLOC_RES);

#ifdef TM_ERROR_CHECKING
    if (packetNum & 0x80)
    {
/* ALLOCATION FAILED!  This is a critical error, since the chip told
 * us that it was done...
 */                
        tfKernelError( "tfSmcSend", "Inconsistent allocation result" );
    }
#endif /* TM_ERROR_CHECKING */
    
/* Tell the chip that we're writing the packet that was just allocated above */
    tm_smc_out_byte(packetNum, TM_SMC_PKT_NUM);
    
/* Instruct the chip to auto-increment the data pointer as we copy data to
   it */ 
    tm_smc_out_word(TM_PTR_AUTO_INC, TM_SMC_POINTER);

/* 
 * The transmit packet format on the SMC chip consists of a status word, a 
 * byte count word, less than 1200 words of data followed by a control byte and
 * the last data byte, if the length is odd.
 */

/* Write the status word to the chip */
    tm_smc_out_word(0, TM_SMC_DATA1);

/* Write the length (length of packet, plus 6 bytes of overhead) to the chip */
    tm_smc_out_word(dataLength + 6, TM_SMC_DATA1);

/* 
 * Write out the packet data to the chip 
 */
    pktCounter = dataLength >> 1;
    pktDataPtr = (unsigned short *) dataPtr;

    while (pktCounter--)
    {
        tm_smc_out_word(*(pktDataPtr++), TM_SMC_DATA1);        
    }
    
/* 
 * OPTIMIZATION NOTE: To acheive the best performance with this driver, the
 * above loop should be optimized in assembly.  For instance, on the x86 
 * platform, a significant performance boost could be had by using the string
 * output instruction.  Also, the SMC also makes a second data port 
 * (TM_SMC_DATA2) available at the location immediately following the first
 * data port.  This would allow a single 32-bit value to be written at once,
 * if possible on your platform which would also significantly improve
 * performance.
 */

/* 
 * If we have an odd byte to deal with, write it out to the chip, along with
 * the control byte that indicates that there's an odd byte of data.  Otherwise
 * just write a zero to the data port.
 */    
    if ((dataLength & 1) == 1)
    {
        outWord = (unsigned short)(*( (unsigned char *) pktDataPtr));
        outWord = TM_CTL_ODD_BYTE | (outWord & 0xFF);
        tm_smc_out_word(outWord, TM_SMC_DATA1);
    }
    else
    {
        tm_smc_out_word(0, TM_SMC_DATA1);
    }


/* Enqueue this packet in the outgoing FIFO */
    tm_smc_select_bank(2);
    tm_smc_out_word(TM_MMU_ENQUEUE, TM_SMC_MMU_CMD);

/* We're done with this packet, so inform the Treck stack */
    tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);

smcSendReturn:

    return errorCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trsmc.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  3920    Output checksum: 21425
 */
