/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: SMC9000 Device Driver modified for chip SMSC's LAN91C113I
 *              Infineon C161P processor, Keil uVision2 IDE. CMX RTOS
 *
 * Filename: trsmcc16.c
 * Author: Jason (original), Bryan Wang (modified)
 * Date Created: 10/9/2000
 * $Source: source/drivers/smc/trsmc113.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:24CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/* This driver should function on any of the SMSC 'LAN9000' family of
 * Ethernet controllers.  There are a number of things that it may be 
 * necessary to modify to match your platform:
 *
 *  1. Specify the base I/O address 
 *
 *  2. tfSmcSetupHardware sets up any external hardware necessary for the
 *     SMC (e.g., setting various pin direction, etc)
 *
 *  3. tfSmcSetupInterrupts installs an ISR to handle interrupts from the
 *     SMC as well as enabling interrupts on both the interrupt controller
 *     and the CPU (but not the Ethernet chip itself).
 * 
 *  4. tm_clear_cpu_interrupts should acknowledge interrupts to the interrupt
 *     controller (not the SMC).
 *
 *  5. tm_smc_in... / tm_smc_out... macros should be modified to match your
 *     platform and compiler.
 *
 *  6. tfSmcGetPhysAddr should return the hardware address of this Ethernet
 *     controller (either hardcoded, read from EEPROM, etc).
 *
 *
 * NOTES:
 * ------ 
 *  1. Doesn't probe for a SMC9000 chip; I/O address, IRQ, etc must be
 *     specified using the macros below.
 *
 *  2. For maximum performance, please see the 'OPTIMIZATION NOTE' in
 *     tfSmcSend and tfSmcReceive.
 *
 */

/* Use the fast routine for sending and receiving packets */

#include <trsocket.h>
#include <trsmc.h>
#include <trsmc113.h>
#include <cxfuncs.h>        /* CMX RTOS */
#include <timer.h>          /* For wait */

/* For tm_get_raw_buffer */
#include <trmacro.h>
#include <trtype.h>
#include <trglobal.h>
#include <trproto.h>

/* Base address for memory mapped registers. */
#define TM_SMC_BASE_ADDR            0x208000

/*
 * Our receive pool consists of two sized buffers: small and large.
 * These macros determine the size and number of these buffers in our
 * pre-allocated receive pool.
 */
#define TM_SMC_LARGE_RECV_POOL     4
#define TM_SMC_SMALL_RECV_POOL     10
#define TM_SMC_LARGE_RECV_BUF_SIZE 1518
#define TM_SMC_SMALL_RECV_BUF_SIZE 128

/* 
 * This macro determines on what sort of byte boundary our buffers are aligned 
 * (word, long word, etc).
 */
#define TM_SMC_BUF_ALIGNMENT       4

/* Acknowledges interrupts */
#define tm_clear_cpu_interrupts(intLine)

/* Memory mapped register commands */
#define tm_smc_in_byte(offset) tfSmc113InByte(offset)
#define tm_smc_in_word(offset) tfSmc113InWord(offset)
    
#define tm_smc_out_byte(value,offset) tfSmc113OutByte(offset, value)
#define tm_smc_out_word(value,offset) tfSmc113OutWord(offset, value)

#define tm_smc_select_bank(x) tm_smc_out_word(x,TM_SMC_BANK_SELECT)
#define tm_smc_flip_word(x) ( (((x) & 0xff) << 8) | (((x) & 0xff00) >> 8) )

/* 
 * SMC9000 Register Offsets 
 */

/* (All Banks) */
#define TM_SMC_BANK_SELECT 0xE

/* (Bank 0) */
#define TM_SMC_TCR 0x0
#define TM_SMC_EPH 0x2
#define TM_SMC_RCR 0x4
#define TM_SMC_MIR 0x8
#define TM_SMC_PHY 0xA

/* (Bank 1) */
#define TM_SMC_CONFIG       0x0
#define TM_SMC_IND_ADDR_0   0x4
#define TM_SMC_IND_ADDR_1   0x5
#define TM_SMC_IND_ADDR_2   0x6
#define TM_SMC_IND_ADDR_3   0x7
#define TM_SMC_IND_ADDR_4   0x8
#define TM_SMC_IND_ADDR_5   0x9
#define TM_SMC_CONTROL      0xC

/* (Bank 2) */
#define TM_SMC_MMU_CMD      0x0
#define TM_SMC_PKT_NUM      0x2
#define TM_SMC_ALLOC_RES    0x3
#define TM_SMC_FIFO         0x4
#define TM_SMC_POINTER      0x6
#define TM_SMC_DATA1        0x8
#define TM_SMC_DATA2        0xA
#define TM_SMC_INT_STAT     0xC
#define TM_SMC_INT_ACK      0xC
#define TM_SMC_INT_MASK     0xD

/* (Bank 3) */
#define TM_SMC_MULTICAST0   0x0
#define TM_SMC_MULTICAST2   0x2
#define TM_SMC_MULTICAST4   0x4
#define TM_SMC_MULTICAST6   0x6
#define TM_SMC_MGMT         0x8

/* 
 * Interrupt status/mask bits 
 */
/* Frame received */
#define TM_INT_RX           0x01
/* Transmit error */
#define TM_INT_TX           0x02
/* Tx FIFO empty */
#define TM_INT_TX_EMPTY     0x04
/* Transmit buffer allocation complete */
#define TM_INT_ALLOC        0x08
/* Receive overrun */
#define TM_INT_RX_OVRN      0x10
/* Ethernet protocol handler interrupt */
#define TM_INT_EPH          0x20
/* Early receive interrupt */
#define TM_INT_ERCV         0x40
/* Transmit Idle interrupt */
#define TM_INT_TX_IDLE      0x80

/* TCR register bits */
#define TM_TCR_TXENA        0x01


/* Flag indicating that an odd byte is present in this frame */
#define TM_CTL_ODD_BYTE     0x2000

/* 'Magic number' used to identify the SMC9000 chip */
#define TM_SMC_MAGIC_NUM    0x3300

/* SMC command: perform a soft reset of the chip */
#define TM_SMC_SOFT_RST     0x8000

/* 
 * SMC control register flags 
 */
/* Automatically release sent packets */
#define TM_AUTO_RELEASE     0x0800
/* Receive bad packets, rather than immediately reject them */
#define TM_RECV_BAD_PACKETS 0x4000
/* Transmite error enable */
#define TM_TE_ENABLE        0x0020
/* Link error enable */
#define TM_LE_ENABLE        0x0080

/* Enable transmission of packets */
#define TM_XMIT_ENABLE      0x0001
/* Enable reception of packets */
#define TM_RECV_ENABLE      0x0100
/* Enables all multiast  */
#define TM_RECV_ALL_MULTI   0x0040
/* Enables promiscous mode */
#define TM_RECV_PRMS        0x0020

/* 
 * SMC configuration register flags 
 */
/* Automatically release sent packets */
#define TM_EXT_PHY          0x0200

/* Automatically increment the internal SMC pointer as we send data */
#define TM_PTR_AUTO_INC     0x4000

/* 
 * MMU commands 
 */
/* BUSY bit, a release command is stil being processed */
#define TM_MMU_BUSY        0x0001
/* Allocate chip memory for transmit */
#define TM_MMU_ALLOC        0x0020
/* Reset the MMU */
#define TM_MMU_RESET        0x0040
/* Release received packet */
#define TM_MMU_REL_RX_PKT   0x0080
/* Free a packet queued for transmission */
#define TM_MMU_FREE_PKT     0x00a0
/* Enqueue a packet for transmission */
#define TM_MMU_ENQUEUE      0x00c0


/* Management interface register bits and constants */
#define TM_MI_MDO           0x01
#define TM_MI_MDI           0x02
#define TM_MI_MCLK          0x04
#define TM_MI_MDOE          0x08
#define TM_MI_MALL          0x0F
#define TM_MI_OPWRITE       0x01
#define TM_MI_OPREAD        0x02

/* internal PHY addresses */
#define TM_SMC_INTERNAL_PHY 0x00

/* PHY registers */
#define TM_PHY_CONTROL_REG  0x00
#define TM_PHY_STATUS_REG   0x01
/* auto negotiation  capcity advertisement register */
#define TM_PHY_NEG_ADV_REG  0x04
/* auto negotiation remote end capcity register */
#define TM_PHY_NEG_RES_REG  0x05

/* Receive/PHY control register */
#define TM_PHY_ANEG        0x8000


/* PHY control register bits */
#define TM_PHY_ANEG_EN      0x1000
#define TM_PHY_ANEG_RST     0x0200
#define TM_PHY_RESET        0x8000
#define TM_PHY_DIS          0x0400

/* PHY status register bits*/
#define TM_PHY_AUTO_NEG_COMPLETE    0x0008

/*
 * PHY capcity register bits
 */

/* 100Based TX, full duplex */
#define TM_PHY_TX_FDX    0x0100
/* 802.3 CSMA operation capble */
#define TM_PHY_CSMA         0x0001



/*
 * Pointer register flags
 */
/* Specifies that a read operation (vs a write operation) is to follow */
#define TM_PTR_READ_OP      0x2000
/* Auto increment the pointer for each read operation */
#define TM_PTR_AUTO_INC     0x4000
/* Use the top of the RX FIFO as our pointer, not the one set in this
   register */
#define TM_PTR_RCV_RX_FIFO  0x8000

/* Polynomial used for 32-bit CRC calculation (used for multicast addresses) */
#define TM_CRC32_POLY 0x04c11db7UL

/* Send allocation request states */
#define TM_SMC_ALLOC_IDLE    0
#define TM_SMC_ALLOC_WAITING 1

#define tm_smc_alloc_approved(statReg) ((statReg) & TM_INT_ALLOC)

/*
 * SMC information structure 
 */
typedef struct tsSmcInfo
{

/* Treck interface handle for this device */
    ttUserInterface smcInterfaceHandle;

/* 
 * Current state of our send allocation request, either waiting or idle 
 * If send queue is not enabled, the state will always be idle.
 */
    unsigned char   smcAllocState;

/* Flag indicating if a transmit queue is available for this interface */
    unsigned char   smcXmitQueueEnb;
} ttSmcInfo;
typedef ttSmcInfo TM_FAR * ttSmcInfoPtr;

/* Global SMC info pointer pointing to the current device */
ttSmcInfoPtr tlSmcPtr;

char tlMacAddress[] = "TTTTTT";

/* Send a single bit to PHY via MII */
void tfPhyBit(tt16Bit bits)
{
    tm_smc_out_word(bits, TM_SMC_MGMT);
    tm_smc_out_word(bits | TM_MI_MCLK, TM_SMC_MGMT);
}

/* 
 * Read or write a PHY register through MII interface bit streaming
 *      opCode = TM_MI_OPWRITE, return 1;
 *      opCode = TM_MI_OPREAD,  return PHY register value;
 */  
tt16Bit tfPhyAccess(
    tt8Bit  phyAddr,    /* PHY to access */
    tt8Bit  regAddr,    /* PHY register to access */
    tt8Bit  opCode,     /* read or write */
    tt16Bit data)       /* data to write */
{
    int     i;
    tt16Bit mgmt;
    tt16Bit regBank;

/* Save the old bank selection */
    regBank = tm_smc_in_word(TM_SMC_BANK_SELECT);

/* Change to bank 3, to use the MI register  */
    tm_smc_select_bank(3);

/* Filter unused bits from input variables */

    phyAddr &= 0x1F;
    regAddr &= 0x1F;
    opCode  &= 0x03;

/* read in the management register and clear the last 4 bits
 * for MI accessing, (0xXXX0)
 */
    mgmt = tm_smc_in_word(TM_SMC_MGMT) & (TM_MI_MALL ^ 0xFFFF);

/* Output preamble for for sync (32 bits 1 )*/
    for (i = 0; i< 32; i++)
    {
        tfPhyBit(mgmt | TM_MI_MDOE | TM_MI_MDO);
    }


/* Output start of Frame (01) */
    for (i = 0; i<=1; i++)
    {
        tfPhyBit(mgmt | TM_MI_MDOE | i);
    }

/* Output code (01 for write, 10 for read) */
    for (i = 1; i>=0; i--)
    {
        tfPhyBit(mgmt | TM_MI_MDOE | ((opCode >> i) & 0x01));
    }

/* Output address of PHY (5 bits) */
    for (i = 4; i>=0; i--)
    {
        tfPhyBit(mgmt | TM_MI_MDOE | ((phyAddr >> i) & 0x01));
    }

/* Output address of register (5 bits) */
    for (i = 4; i>=0; i--)
    {
        tfPhyBit(mgmt | TM_MI_MDOE | ((regAddr >> i) & 0x01));
    }

    if (opCode == TM_MI_OPREAD)
    {

/* Read operation */

/* Implement turnaround (Z0) */
        tfPhyBit(mgmt);    /*(Z) */

/* Commend out for National Semiconductor PHY BUG, leave for other PHYs */
        tfPhyBit(mgmt | TM_MI_MDOE); /*(0) */

/* Read data */

        data = 0;

        for (i = 15; i>=0; i--)
        {
            tfPhyBit(mgmt);
            data |= ((tm_smc_in_word(TM_SMC_MGMT) & TM_MI_MDI) >> 1) << i;
        }

    }
    else
    {
    
/* Write operation */

/* Implement turnaround (10) */
        for (i = 1; i>=0; i--)
        {
            tfPhyBit(mgmt | TM_MI_MDOE | (i & 0x01));
        }

/* Write data */
        for (i = 15; i>=0; i--)
        {
            tfPhyBit(mgmt | TM_MI_MDOE | ((data >> i) & 0x01));
        }

        data = 0;
    }

/* restore the register bank selection */
    tm_smc_select_bank(regBank);

    return data;
}


/* Reset PHY */
int tfPhyReset(tt8Bit phyAddr)
{
    int i;
    tt16Bit regValue;

    tfPhyAccess(phyAddr, TM_PHY_CONTROL_REG, TM_MI_OPWRITE, TM_PHY_RESET);
    for (i = 5; i>=0; i--)
    {
        regValue = tfPhyAccess(phyAddr, TM_PHY_CONTROL_REG, TM_MI_OPREAD, TM_PHY_RESET);
        if (!(regValue & TM_PHY_RESET))
        {
            return TM_DEV_OKAY;
        }
        DelayMs(10);
    }
    return TM_DEV_ERROR;
}


/* Reset PHY */
#define TM_DEF_PHY_COLTROL 0x1000       
int tfPhyTest(tt8Bit phyAddr)
{
    int i;
    tt16Bit regValue;

    for (i = 5; i>=0; i--)
    {
        regValue = tfPhyAccess(phyAddr, TM_PHY_CONTROL_REG, TM_MI_OPREAD, TM_PHY_RESET);
        if (regValue == TM_DEF_PHY_COLTROL)
        {
            return TM_DEV_OKAY;
        }
        DelayMs(10);
    }
    return TM_DEV_ERROR;
}


/* 
 * tfSmcOpen()
 *
 * 1. Verify that this is a SMC9000 chip that we're dealing with (ie,eheck 
 *    the I/O address).
 * 2. Get the amount of memory available on the chip.
 * 3. Reset the chip.
 * 4. Set various registers to default values.
 * 5. Reset the chip's MMU.
 * 6. Setup the Ethernet HW address 
 * 7. Setup interrupt handler, etc.
 * 8. Enable transmit and receive
 * 9. Enable interrupts.
 * 10. configure PHY
 */
int tfSmcOpen(ttUserInterface interfaceHandle)
{
    unsigned short  inWord;
    int             errorCode;
    char            physAddr[6];
    int             i;

/* Verify that this is a SMC9000 chip (ie, check the I/O address). 
 * To do this, we read a 16-bit value from the bank select register.
 * The upper byte of this word should always be 0x33.
 */
    errorCode = 0;
    tm_smc_select_bank(3);

    inWord = tm_smc_in_word(TM_SMC_BANK_SELECT);
    if ( (inWord & 0xff00) == TM_SMC_MAGIC_NUM )
    {

/* Allocate memory for the info structure */
        tlSmcPtr = tm_get_raw_buffer( sizeof(ttSmcInfo) );

        if (tlSmcPtr == 0)
        {
/* No memory for info structure; quit now and return error */
            errorCode = TM_DEV_ERROR;
            goto smcOpenEnd;
        }        
        (void)tfDeviceStorePointer(interfaceHandle, (ttVoidPtr)tlSmcPtr); 

/* 
 * Initialize the information structure 
 */
        tlSmcPtr->smcAllocState      = TM_SMC_ALLOC_IDLE;
        tlSmcPtr->smcInterfaceHandle = interfaceHandle;
        tlSmcPtr->smcXmitQueueEnb    = (unsigned char)
                                        tfInterfaceXmitQueueOn(interfaceHandle);
      
/* Reset the chip */
        tm_smc_select_bank(0);
        tm_smc_out_word( TM_SMC_SOFT_RST, TM_SMC_RCR );

/* Delay for 0.5 seconds tp allow the chip to reset... */
        DelayMs (50);

/* Clear the RCR register - this terminates the reset process */
        tm_smc_out_word(0, TM_SMC_RCR);        

/* Set the PHY control register to auto select 10/100 MBPS and Full/Half duplex */
        inWord = tm_smc_in_word(TM_SMC_PHY);
        inWord = inWord  | TM_PHY_ANEG;
        tm_smc_out_word(inWord, TM_SMC_PHY);


/* Setup the control register:
 *  + Don't receive bad packets
 *  + Automatically free transmitted packets from on-board memory.
 */ 
        tm_smc_select_bank(1);
        inWord = tm_smc_in_word(TM_SMC_CONTROL);
        inWord |= TM_AUTO_RELEASE;
        inWord |= TM_TE_ENABLE;
        inWord |= TM_LE_ENABLE;
        inWord &= ~TM_RECV_BAD_PACKETS;
        tm_smc_out_word(inWord, TM_SMC_CONTROL);

/* Reset the chip's MMU */
        tm_smc_select_bank(2);
        tm_smc_out_word(TM_MMU_RESET, TM_SMC_MMU_CMD);

/* Initialize receive buffer pool */
        (void)tfPoolCreate(interfaceHandle,
                     TM_SMC_LARGE_RECV_POOL,
                     TM_SMC_SMALL_RECV_POOL,
                     TM_SMC_LARGE_RECV_BUF_SIZE,
                     TM_SMC_SMALL_RECV_BUF_SIZE,
                     TM_SMC_BUF_ALIGNMENT,
                     TM_POOL_REFILL_IN_LINE );

/* Setup Ethernet HW address */
        tm_smc_select_bank(1);
        (void)tfSmcGetPhysAddr(interfaceHandle, physAddr);
        tm_smc_out_word(*((unsigned short *) &physAddr[0]), TM_SMC_IND_ADDR_0);
        tm_smc_out_word(*((unsigned short *) &physAddr[2]), TM_SMC_IND_ADDR_2);
        tm_smc_out_word(*((unsigned short *) &physAddr[4]), TM_SMC_IND_ADDR_4);


/* To initialize the processor such that it will vector when the 91C113
 * asserts an interrupt requires the following code:
 */

        EXICON |= 0x4000; /* setup fast interrupt 7*/
        DP2    &= 0x7FFF; /* Ensure it is set as an input*/
        CC15IC =  0x50  ; /* IE = 1, ILVL = 4 , GLVL = 0*/

/* Enable interrupts (on receive and transmit error only) */
        tm_smc_select_bank(2);
        tm_smc_out_byte(TM_INT_EPH | TM_INT_RX, TM_SMC_INT_MASK);

/* Enable transmit and receive */
        tm_smc_select_bank(0);
        tm_smc_out_word(TM_XMIT_ENABLE, TM_SMC_TCR);
        tm_smc_out_word(TM_RECV_ENABLE, TM_SMC_RCR);        

/*
	promiscous receive mode
        tm_smc_out_word(TM_RECV_ENABLE | TM_RECV_PRMS, TM_SMC_RCR);        
*/

/* Disable external PHY and Enable internal PHY */
        tm_smc_select_bank(1);
        inWord = tm_smc_in_word(TM_SMC_CONFIG);
        inWord &= ~TM_EXT_PHY;
        tm_smc_out_word(inWord, TM_SMC_CONFIG);

/* Reset the internal PHY */
        errorCode = tfPhyReset(TM_SMC_INTERNAL_PHY);

        if (errorCode != TM_DEV_OKAY)
        {
            errorCode = TM_DEV_ERROR;
            goto smcOpenEnd;
        }

/* Remove the internal PHY out of the isolation mode so it starts to work
 * set the auto negotiatino bit on 
 */
        tfPhyAccess(TM_SMC_INTERNAL_PHY, 
                    TM_PHY_CONTROL_REG,  
                    TM_MI_OPWRITE, 
                    TM_PHY_ANEG_EN);

/* Delay for 0.5 seconds to allow the set to take effect on the PHY */
        DelayMs (50);

        tfPhyTest(TM_SMC_INTERNAL_PHY);

        if (errorCode != TM_DEV_OKAY)
        {
            errorCode = TM_DEV_ERROR;
            goto smcOpenEnd;
        }

/* Negociate the link parameters */


/* Set the desired auto parameters to the PHY auto_negociation 
 * advertisement register 
 */
        tfPhyAccess(TM_SMC_INTERNAL_PHY,
                    TM_PHY_NEG_ADV_REG,  
                    TM_MI_OPWRITE, 
                    TM_PHY_TX_FDX | TM_PHY_CSMA);

/* Start the auto negociation process */
        tfPhyAccess(TM_SMC_INTERNAL_PHY,
                    TM_PHY_CONTROL_REG,  
                    TM_MI_OPWRITE, 
                    TM_PHY_ANEG_RST);

/* Wait a max of 10 seconds for the auto negociation to complete */
        errorCode = TM_DEV_ERROR;

        for (i=0; i< 100; i++)
        {
            inWord = tfPhyAccess(
                        TM_SMC_INTERNAL_PHY,
                        TM_PHY_STATUS_REG,  
                        TM_MI_OPREAD, 
                        TM_PHY_ANEG_EN);

            if (inWord & TM_PHY_AUTO_NEG_COMPLETE)
            {
/* Auto negociation has completed */
                errorCode = TM_DEV_OKAY;
                break;
            }
            DelayMs(10);
        }

/* Read the remote end capacity information */
        inWord = tfPhyAccess(
                    TM_SMC_INTERNAL_PHY,
                    TM_PHY_NEG_RES_REG,  
                    TM_MI_OPREAD, 
                    TM_PHY_ANEG_EN);

        if (errorCode != TM_DEV_OKAY)
        {
            errorCode = TM_DEV_ERROR;
            goto smcOpenEnd;
        }
    }
    else
    {
/* Couldn't verify existence of SMC9000 chip */
        errorCode = TM_DEV_ERROR;
    }

smcOpenEnd:
    return errorCode;          
}


/*
 * tfSmcClose()
 *
 *   This function is called when the user wishes to close this device.
 *   We need to clean up any allocated memory, disable any chip interrupts,
 *   and disable transmit/receive on the chip.  While we do need to free memory
 *   that was allocated for the receive pool (by calling tfPoolDelete) it is
 *   not necessary to clean up any transmit buffer, since they are freed 
 *   immediately upon return from tfSmcSend().
 *
 */
int tfSmcClose(ttUserInterface interfaceHandle)
{
    unsigned short  inWord;

/* Disable all interrupts */
    tm_smc_select_bank(2);
    tm_smc_out_word(0, TM_SMC_INT_MASK);

/* Disable transmit */
    tm_smc_select_bank(0);
    inWord = tm_smc_in_word(TM_SMC_TCR);
    inWord &= ~TM_XMIT_ENABLE;
    tm_smc_out_word(inWord, TM_SMC_TCR);
    
/* Disable receive */    
    inWord = tm_smc_in_word(TM_SMC_RCR);
    inWord &= ~TM_RECV_ENABLE;
    tm_smc_out_word(inWord, TM_SMC_RCR);
/* 
 * Free memory from pre-allocated buffers in the receive ring 
 */
    (void)tfPoolDelete(interfaceHandle);

/* Free device driver variable */
    tlSmcPtr = (ttSmcInfoPtr)tfDeviceClearPointer(interfaceHandle);
    if ( tlSmcPtr != (ttSmcInfoPtr)0 )
    {
/* Free malloc'd memory */
        tm_free_raw_buffer((void TM_FAR *)tlSmcPtr);
        tlSmcPtr = (ttSmcInfoPtr)0;
    }

    return TM_DEV_OKAY;
}



/* 
 * tfSmcIoctl()
 *
 *   This I/O control routine performs two functions:
 *
 *   1.  Setup the list of multicast addresses.
 *   2.  Set the SMC to recieve all multicast packets.  This is done by writing
 *       all 1's to the SMC's four multicast hash registers.
 *
 */
int tfSmcIoctl(ttUserInterface interfaceHandle,
               int             flag,
               void TM_FAR *   optionPtr,
               int             optionLen)
{
#ifdef TM_IGMP
    unsigned long               calcCRC;
    unsigned char               currByte;
    unsigned char               byte, bit;
    unsigned long               hashEntry;
    unsigned char      TM_FAR * multiAddr;
    unsigned short              logicalAddrFilter[4];
    int                         index;
#endif /* TM_IGMP */

interfaceHandle=interfaceHandle;

#ifdef TM_IGMP
    {
/*
 * Setup our list of multicast addresses to receive.
 */
        if (flag & TM_DEV_SET_MCAST_LIST)
        {

/* 
 * Clear the SMC's address filter (since this IOCTL call will add all the
 * entries we need.
 */
            logicalAddrFilter[0] = 0;
            logicalAddrFilter[1] = 0;
            logicalAddrFilter[2] = 0;
            logicalAddrFilter[3] = 0;

/* Get the first address */
            multiAddr = (unsigned char TM_FAR *) optionPtr;

            while (optionLen-- > 0)
            {
            
  
/* Initialize the calculated CRC (with all ones) */
                calcCRC = 0xffffffffUL;

/* For all the bytes in the (6 byte) address */
                for (byte=0;byte<6;byte++)
                {
                    currByte = multiAddr[byte];

/* For each bit in the current byte */
                    for (bit=0;bit<8;bit++)
                    {

/* Shift the calculated CRC left by one */
                        calcCRC <<= 1;

/* If the current bit is one... */
                        if ((currByte >> bit) & 0x01)
                        {

/* ... XOR the calculated CRC with the CRC-32 polynomial */
                            calcCRC ^= TM_CRC32_POLY;

                        }                   
                    }
                }

/* Get the most significant 6 bits from the calculated CRC */        
                hashEntry = (calcCRC & 0xfc000000UL) >> 26;

/* 
 * Decode the above number and set the appropriate bit in the SMC logical
 * address filter 
 */
                index = (hashEntry >> 4) & 0xFF;
#ifdef TM_ERROR_CHECKING
                if (index > 3)
                {
/* Because we only keep 2 bits, this should never happen */
                    tfKernelError("tfSmcIoctl", "Stack corruption\n");
                }
#endif /* TM_ERROR_CHECKING */
                logicalAddrFilter[index] |=
                    1 << ((unsigned char) (hashEntry % 16));

                multiAddr += 6;

            }
            tm_smc_select_bank(3);
            tm_smc_out_word(logicalAddrFilter[0], TM_SMC_MULTICAST0);            
            tm_smc_out_word(logicalAddrFilter[1], TM_SMC_MULTICAST2);            
            tm_smc_out_word(logicalAddrFilter[2], TM_SMC_MULTICAST4);            
            tm_smc_out_word(logicalAddrFilter[3], TM_SMC_MULTICAST6);            
            tm_smc_select_bank(2);                    
        }

/*
 * Set the SMC to receive all multicast packets.
 */
        if (flag & TM_DEV_SET_ALL_MCAST)
        {
            tm_smc_select_bank(0);
            tm_smc_out_word(TM_RECV_ALL_MULTI, TM_SMC_RCR);        
        }
    }

#else /* TM_IGMP */
/* Avoid compiler warnings when IGMP is disabled */
    interfaceHandle=interfaceHandle;
    optionPtr=optionPtr;
    optionLen=optionLen;
    flag=flag;
#endif /* !TM_IGMP */

    return TM_DEV_OKAY;
}


/*
 * tfSmcReceive()
 *
 *   This routine is called by the Treck stack to receive a packet.  Since
 *   the data has already been copied inside of the ISR, all we need to do
 *   in this routine is to pass the buffer up to the Treck stack.  
 */
int tfSmcReceive(ttUserInterface        interfaceHandle,
                 char TM_FAR * TM_FAR * dataPtrPtr,
                 int  TM_FAR *          dataLengthPtr,
                 ttUserBufferPtr        bufHandlePtr)
{

    return tfPoolReceive( interfaceHandle,
                          dataPtrPtr,
                          dataLengthPtr,
                          bufHandlePtr );

}


/*
 * tfSmcGetPhysAddr()
 *
 *   Returns the physical address of this Ethernet interface.  This is just
 *   sample data and this routine should be modified to return real data.
 *   Should be read from EEPROM or other source
 */
int tfSmcGetPhysAddr(ttUserInterface interfaceHandle,
                     char TM_FAR *   physicalAddress)
{
    interfaceHandle=interfaceHandle;

/* Set the ethernet address according the the global configuration structure */
	tm_memcpy(physicalAddress, tlMacAddress, 6);

    return TM_DEV_OKAY;
}
/*
 * tfSmcSend()
 *
 *   The logic determining how and when to allocate memory, and if this memory
 *   allocate request is the most complicated part of this routine.  This 
 *   driver will work with either the transmit task enabled or disabled.  For
 *   efficiency reasons, it is recommended that the transmit task be enabled.
 *   Here's a breakdown of the logic:
 *
 *      1. If we are currently waiting for allocation notification...
 *           (this will only happen if the transmit queue is enabled)
 *        A. If our previous allocation was approved, acknowledge the 
 *           allocation, continue and send the packet.
 *        B. If we are still waiting on an allocation, return an error 
 *           (the upper layer transmit queue will handle queuing the packet)
 *      2. Otherwise, if we are not waiting on a previous allocate:
 *        A. Request an allocation from the chip.
 *        B. If the this allocation was approved, acknowledge the allocation,
 *           continue and send the packet.
 *        C. If this allocation fails...
 *          a. If the transmit queue is enabled, mark the fact that we're 
 *             waiting for an allocation notification, and return an error.
 *             The upper layer transmit queue will handle the current packet.
 *          b. If the transmit queue is not enabled, spin lock until the 
 *             allocation is approved.
 *
 *   So, if the transmit queue is not enabled, all we do is request some
 *   memory, wait for it to be approved and send the packet.  Things only get
 *   complicated when the transmit queue is enabled.
 *
 *   Once this is done, and we have some on-chip memory, we copy the packet to
 *   the chip by writing to the DATA1 port.  We first write out the status and
 *   length words, and then actual data.  When we reach the last word, we
 *   check if there is an extra byte (ie, length is odd).  If there is we copy
 *   this as well as the control byte indicating that this byte is present.  
 *   Otherwise we simply send a zero.  Finally we instruct the chip to enqueue
 *   this packet in the outgoing FIFO and call the tfSendCompleteInterface to
 *   inform the Treck stack that we're finished with this packet.
 *
 */

int tfSmcSend(ttUserInterface interfaceHandle,
              char TM_FAR *   dataPtr,
              int             dataLength,
              int             flag)
{
    int                     mmuPages;
    volatile unsigned char  intStatus;
    unsigned char           packetNum;
    int                     errorCode;

/* Flags not used since we don't support scattered sends */
    flag=flag;

    errorCode = TM_DEV_OKAY;

/* We are currently waiting on a send allocation request */
    if ( tlSmcPtr->smcAllocState == TM_SMC_ALLOC_WAITING )
    {
        tm_smc_select_bank(2);                    
        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Previous allocate approved: acknowledge it, set state and continue */    
            tm_smc_out_byte(TM_INT_ALLOC, TM_SMC_INT_STAT);

            tlSmcPtr->smcAllocState = TM_SMC_ALLOC_IDLE;
        }
        else
        {
/* Previous allocate still in progress, return error to transmit queue */
            errorCode = TM_DEV_ERROR;
        }
    }

/* No pending allocation request; issue new one for this packet */
    else   
    {
/* 
 * The number of 256 byte buffers to allocate for this packet, including
 * the control, status and length words
 */    
        if (dataLength & 1)
        {
/* odd sized packet, extra byte will be picked up by control word */
            mmuPages = ((dataLength+5) >> 8) & 0x0007;
        }
        else
        {
            mmuPages = ((dataLength+6) >> 8) & 0x0007;
        }

/* Allocate on chip memory for this buffer */    
        tm_smc_select_bank(2);    
        tm_smc_out_word(TM_MMU_ALLOC | mmuPages, TM_SMC_MMU_CMD);

        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Allocation approved: acknowledge and go on to send packet */
            tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );

        }
        else
        {

/* 
 * Allocation failed: either simply return an error (if the transmit queue
 * is enabled) or spin lock otherwise.
 */
            if ( tlSmcPtr->smcXmitQueueEnb )
            {
                tlSmcPtr->smcAllocState = TM_SMC_ALLOC_WAITING;

                errorCode = TM_DEV_ERROR;
            }
            else
            {
/* 
 * Wait (spin lock) until on-chip memory is available.  To determine if the
 * on-chip memory has been allocated, we poll the interrupt status register 
 * (note however, that an interrupt is *not* generated, because we have it
 * masked off). 
 */
                tm_smc_select_bank(2);                    
                do 
                {
/* Yield processor to other tasks while we're waiting, SMC LAN 113 manages 
 * the internal FIFO buffer automatically, if we coudn't get buffer for send,
 * The buffer might be used out by received packets, so here we need to 
 * give the receive task a chance to clear out the buffer
 */
                    (void)tfInterfaceSpinLock(interfaceHandle);

/* Poll the interrupt status bit to check if allocation has completed */
                    intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

                } while ( !tm_smc_alloc_approved(intStatus) );

/* Allocation approved: acknowledge and go on to send packet */
                tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );
            }

        }

    }

/* No memory available; do not send packet and return with an error */
    if (errorCode != TM_DEV_OKAY)
    {
        goto smcSendReturn;
    }

/* Memory allocation is now complete: we can write our packet to the chip */

/* 
 * Get the packet number of the last allocated packet; this is where we'll copy
 * the data.
 */    
    packetNum = tm_smc_in_byte(TM_SMC_ALLOC_RES);

#ifdef TM_ERROR_CHECKING
    if (packetNum & 0x80)
    {
/* ALLOCATION FAILED!  This is a critical error, since the chip told
 * us that it was done...
 */                
        tfKernelError( "tfSmcSend", "Inconsistent allocation result" );
    }
#endif /* TM_ERROR_CHECKING */
    
/* Tell the chip that we're writing the packet that was just allocated above */
    tm_smc_out_byte(packetNum, TM_SMC_PKT_NUM);
    
/* Instruct the chip to auto-increment the data pointer as we copy data to it*/
    tm_smc_out_word(TM_PTR_AUTO_INC, TM_SMC_POINTER);

/* 
 * The transmit packet format on the SMC chip consists of a status word, a 
 * byte count word, less than 1200 words of data followed by a control byte and
 * the last data byte, if the length is odd.
 */

/* High speed data transfer routine  */
    tfSmc113SndData (dataPtr, dataLength);

/* Enqueue this packet in the outgoing FIFO */
    tm_smc_select_bank(2);
    tm_smc_out_word(TM_MMU_ENQUEUE, TM_SMC_MMU_CMD);

/* We're done with this packet, so inform the Treck stack */
    tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);

smcSendReturn:

    return errorCode;
}


#ifdef TM_USE_DRV_ONE_SCAT_SEND
/*
 * tfScmScatSend
 * Scatter send function.
 */
int tfSmcScatSend(ttUserInterface interfaceHandle,
                  ttUserPacketPtr  uPacketPtr )
{
    char TM_FAR *           dataPtr;
    int                     dataLength;
    int                     scatDataLen;
    int                     scatterIndex;
    int                     mmuPages;
    volatile unsigned char  intStatus;
    unsigned char           packetNum;
    unsigned short          outWord;
    int                     scatOdd;
    int                     errorCode;

    errorCode = TM_DEV_OKAY;
    dataLength = uPacketPtr->pktuChainDataLength;

/* We are currently waiting on a send allocation request */
    if ( tlSmcPtr->smcAllocState == TM_SMC_ALLOC_WAITING )
    {

        tm_smc_select_bank(2);                    
        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Previous allocate approved: acknowledge it, set state and continue */    
            tm_smc_out_byte(TM_INT_ALLOC, TM_SMC_INT_STAT);

            tlSmcPtr->smcAllocState = TM_SMC_ALLOC_IDLE;
        }
        else
        {
/* Previous allocate still in progress, return error to transmit queue */
            errorCode = TM_DEV_ERROR;
        }
    }

/* No pending allocation request; issue new one for this packet */
    else   
    {
    
/* 
 * The number of 256 byte buffers to allocate for this packet, including
 * the control, status and length words
 */    
        if (dataLength & 1)
        {
/* odd sized packet, extra byte will be picked up by control word */
            mmuPages = ((dataLength+5) >> 8) & 0x0007;
        }
        else
        {
            mmuPages = ((dataLength+6) >> 8) & 0x0007;
        }

/* Allocate on chip memory for this buffer */    
        tm_smc_select_bank(2);    
        tm_smc_out_word(TM_MMU_ALLOC | mmuPages, TM_SMC_MMU_CMD);

        intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

        if ( tm_smc_alloc_approved(intStatus) )
        {

/* Allocation approved: acknowledge and go on to send packet */
            tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );

        }
        else
        {
/* 
 * Allocation failed: either simply return an error (if the transmit queue
 * is enabled) or spin lock otherwise.
 */
            if ( tlSmcPtr->smcXmitQueueEnb )
            {
                tlSmcPtr->smcAllocState = TM_SMC_ALLOC_WAITING;
                errorCode = TM_DEV_ERROR;
            }
            else
            {
/* 
 * Wait (spin lock) until on-chip memory is available.  To determine if the
 * on-chip memory has been allocated, we poll the interrupt status register 
 * (note however, that an interrupt is *not* generated, because we have it
 * masked off). 
 */
                tm_smc_select_bank(2);                    
                do 
                {
/* Yield processor to other tasks while we're waiting, SMC LAN 113 manages 
 * the internal FIFO buffer automatically, if we coudn't get buffer for send,
 * The buffer might be used out by received packets, so here we need to 
 * give the receive task a chance to clear out the buffer
 */
                    (void)tfInterfaceSpinLock(interfaceHandle);

/* Poll the interrupt status bit to check if allocation has completed */
                    intStatus = tm_smc_in_byte(TM_SMC_INT_STAT);

                } while ( !tm_smc_alloc_approved(intStatus) );

/* Allocation approved: acknowledge and go on to send packet */
                tm_smc_out_byte( TM_INT_ALLOC, TM_SMC_INT_STAT );
            }

        }

    }


/* No memory available; do not send packet and return with an error */
    if (errorCode != TM_DEV_OKAY)
    {
        goto smcSendReturn;
    }


/* Memory allocation is now complete: we can write our packet to the chip */

/* 
 * Get the packet number of the last allocated packet; this is where we'll copy
 * the data.
 */    
    packetNum = tm_smc_in_byte(TM_SMC_ALLOC_RES);

#ifdef TM_ERROR_CHECKING
    if (packetNum & 0x80)
    {
/* ALLOCATION FAILED!  This is a critical error, since the chip told
 * us that it was done...
 */                
        tfKernelError( "tfSmcScatSend", "Inconsistent allocation result" );
    }
#endif /* TM_ERROR_CHECKING */
    
/* Tell the chip that we're writing the packet that was just allocated above */
    tm_smc_out_byte(packetNum, TM_SMC_PKT_NUM);
    
/* Instruct the chip to auto-increment the data pointer as we copy data to
   it */ 
    tm_smc_out_word(TM_PTR_AUTO_INC, TM_SMC_POINTER);

/* 
 * The transmit packet format on the SMC chip consists of a status word, a 
 * byte count word, less than 1200 words of data followed by a control byte and
 * the last data byte, if the length is odd.
 */

/* 
 * Write out the packet data to the chip 
 */
    scatterIndex = uPacketPtr->pktuLinkExtraCount;

	if (scatterIndex == 0)
	{
/* The packet is not scattered, 
 * lets's use the optimized non scattered send routine
 */
        tfSmc113SndData(
            (char *)uPacketPtr->pktuLinkDataPtr,
            uPacketPtr->pktuLinkDataLength);
	}
	else 
	{
/* This packet is really scattered, */

/* Write the status word to the chip */
        tm_smc_out_word(0, TM_SMC_DATA1);

/* Write the length (length of packet, plus 6 bytes of overhead) to the chip */
        if(dataLength & 0x1)
        {
            tm_smc_out_word(dataLength + 5, TM_SMC_DATA1);
        }
        else
        {
            tm_smc_out_word(dataLength + 6, TM_SMC_DATA1);
        }

        scatOdd = 0;
	    outWord = 0;
        while ( scatterIndex-- >= 0 )
        {
            dataPtr = (char *)uPacketPtr->pktuLinkDataPtr;
            scatDataLen = uPacketPtr->pktuLinkDataLength;
            uPacketPtr  = uPacketPtr->pktuLinkNextPtr; 
            if(scatOdd)
            {
                outWord |= ((unsigned short)(*dataPtr) << 8);
                tm_smc_out_word(outWord, TM_SMC_DATA1);
                dataPtr++;
                scatDataLen--;
                scatOdd = 0;
            }

    /* Fast routine to send the even bytes only, 
     * not the length, not the final odd byte
     */
            tfSmc113SndScatData(dataPtr, scatDataLen & 0xFFFE);

            if(scatDataLen & 0x01)
            {
                scatOdd = 1;
                outWord = (unsigned short)(*(dataPtr + scatDataLen -1));
            }
        }
    

    /* 
     * If we have an odd byte to deal with, write it out to the chip, along with
     * the control byte that indicates that there's an odd byte of data.  Otherwise
     * just write a zero to the data port.
     */    
        if ((dataLength & 1) == 1)
        {
            outWord |= TM_CTL_ODD_BYTE;
            tm_smc_out_word(outWord, TM_SMC_DATA1);
        }
        else
        {
            tm_smc_out_word(0, TM_SMC_DATA1);
        }
    }

/* Enqueue this packet in the outgoing FIFO */
    tm_smc_select_bank(2);
    tm_smc_out_word(TM_MMU_ENQUEUE, TM_SMC_MMU_CMD);

/* We're done with this packet, so inform the Treck stack */
    tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);

smcSendReturn:

    return errorCode;
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */


/*
 * tfSmcIsrHandler()
 *
 * A number of SMC registers need to be saved and restored in this routine: 
 * the bank select register, the pointer register and the packet number
 * register.
 *
 * Two types of SMC interrupts must be dealt with in this ISR.  The first is a 
 * transmit error.  This interrupt is fired when there is a error transmitting
 * a packet.  The only thing that has to be done to deal with this interrupt
 * is to free the packet from the chip.
 *
 * The second, and most important type, is the receive interrupt.  This 
 * interrupt is fired when a packet is received and ready to be read by the
 * CPU.  We first check to make sure that we have a buffer available in our
 * pool; if none is available, we drop the packet.  Otherwise we copy the
 * entire packet into our pool, which will be passed up to the Treck stack
 * in the receive routine. In either case, at the end of this section of the
 * ISR we instruct the chip to release the most recently received packet.
 *xxx
 */

/* To correctly provide an ISR for the MAC the following construct MUST be used */
void SMSC113_ISR (void) interrupt 0x1f
{
/* Disable interrupt */
    IEN = 0;
    _push_(DPP0);

/* Force DPP0 to correct value */
    DPP0 = C_PAGEDPP0; 

/* Call the actuall Treck ISR handler */
    tfSmcIsrHandler();

    ADEIR = 1;
    DPP0 = _pop_();
/* enable ineterrupt */
	IEN = 1;
}
  
void tfSmcIsrHandler(void)
{
    unsigned char           intStatus;
    unsigned char           tcr;
    unsigned short          regBank;
    unsigned short          regPointer;
    unsigned short          pktNumber;
    unsigned short          pktLength;
    unsigned short TM_FAR * dataPtr;
    unsigned char           recvNotifyCount;
    unsigned short          inWord;

/* Save on chip registers (current bank, pointer, and packet number) */
    regBank = tm_smc_in_word(TM_SMC_BANK_SELECT);

/* Register bank 2 -- interrupt etc */
    tm_smc_select_bank(2);

    regPointer = tm_smc_in_word(TM_SMC_POINTER);
    pktNumber = tm_smc_in_word(TM_SMC_PKT_NUM);
    recvNotifyCount = 0;

    do
    {

/* Read the interrupt status register */
        intStatus = (unsigned char)(  tm_smc_in_byte(TM_SMC_INT_STAT) 
                                    & (TM_INT_EPH | TM_INT_RX) );
/* 
 * Transmit error interrupt 
 */
        if (intStatus & TM_INT_EPH)
        {
/* Check the the TXENA bit of the TCR(Transmission control register) */

/* Switch to Register bank 0 - control registers */
            tm_smc_select_bank(0);
            tcr = tm_smc_in_byte(TM_SMC_TCR);
            if ((tcr & TM_XMIT_ENABLE)== TM_8BIT_ZERO)
            {
/* The TXENA bit is cleared, this means a transmission error has occured
 * We need to free this packet from the chip's transmit FIFO.  
 * Set the current packet number to the packet at the 
 * top of the transmit FIFO, then instruct the MMU to free this errored
 * packet.  Currently, we don't record the error condition, but this is
 * where that could be done.
 */
                pktNumber = tm_smc_in_byte(TM_SMC_FIFO) & 0x7f;
                tm_smc_out_byte(pktNumber, TM_SMC_PKT_NUM);
                tm_smc_out_byte(TM_MMU_FREE_PKT, TM_SMC_MMU_CMD);
                tm_smc_select_bank(0);
                tm_smc_out_word(TM_XMIT_ENABLE | tm_smc_in_word(TM_SMC_TCR), TM_SMC_TCR);  
                tm_smc_select_bank(2);
            }

/* Switch back to Register bank 2 - interrupt etc */
            tm_smc_select_bank(2);
        }        
/*
 * Received packet interrupt 
 */
        else if (intStatus & TM_INT_RX)
        {

/* 
 * Setup the read operation: Read from the top of the RX FIFO and auto
 * increment the pointer for each read.
 */
            tm_smc_out_word( TM_PTR_RCV_RX_FIFO |
                             TM_PTR_AUTO_INC |
                             TM_PTR_READ_OP,
                             TM_SMC_POINTER );


/* Read the status of this packet */
            inWord = tm_smc_in_word(TM_SMC_DATA1);

/* Get the length of this packet */
            pktLength = tm_smc_in_word(TM_SMC_DATA1);
            pktLength &= 0x07ff;

/* Remove the length of the status word, etc from the packet length */
            if (pktLength >= 6)
                pktLength -= 6;

/* Add length of odd byte */
            if (inWord & 0x1000)
            {
                pktLength++;
            }

            dataPtr = (unsigned short TM_FAR *) 
                        tfPoolIsrGetBuffer(tlSmcPtr->smcInterfaceHandle,
                                           pktLength);
                
/* Copy the data from the chip into our buffer */
            if (dataPtr != (unsigned short TM_FAR *) 0)
            {

/* High speed data transfer routine  */
                tfSmc113RcvData ((ttUserCharPtr)dataPtr, pktLength);

/* Count the number of frames to notify for */
                recvNotifyCount++;
            }
/* 
 * Whether or not we could receive this packet, instruct the chip to release
 * this packet from its internal RAM - this acknowledges the interrupt.
 */
            tm_smc_out_word(TM_MMU_REL_RX_PKT, TM_SMC_MMU_CMD);
        }
    }
    while (intStatus != 0);

/* Restore on chip registers (register bank, pointer, packet number) */
    tm_smc_out_word(regPointer,TM_SMC_POINTER);
    tm_smc_out_word(pktNumber,TM_SMC_PKT_NUM);
    tm_smc_select_bank(regBank);

/* Clear the CPU interrupts */
    tm_clear_cpu_interrupts(tlSmcPtr->smcIntLine);

/* 
 * Now that the ISR is mostly done, we notify the stack for each packet that
 * we received.  This is done at the end of the ISR because the call to
 * tfNotifyInterfaceIsr tends to release our critical section 
 * (especially in uC/OS on the x86).
 */
    if(recvNotifyCount)
    {
        tfNotifyInterfaceIsr(tlSmcPtr->smcInterfaceHandle, 
                             recvNotifyCount, 0, 0UL, 0UL);
    }
    return;
}



/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trsmc113.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 15979    Output checksum: 21802
 */
