/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: AMR Evaluator-7T board UART driver
 *
 * Filename: trarm7t.c
 * Author: Bryan Wang
 * Date Created: 6/4/2003
 * $Source: source/drivers/uart/trarm7t.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:32CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 *   This driver was ported for AMR Evaluator-7T board UART, and will require
 *   modifications for your specific hardware platform.  Sections of this
 *   driver that will require modifications include:
 *     - enabling and setting up the ISR (tfUartOpen);
 *       restoring ISR on close (tfUartClose)
 *     - memory/port address for serial port (TM_UART_BASE)
 *     - compiler specific implementations of INP and OUTP x86 assembly
 *       commands (tm_in_byte, tm_out_byte, etc)
 *     - termination of the serial communication driver (tfUartClose).
 *
 *   This driver is designed for use as a single interface, i.e. with one
 *   serial port.  If multiple serial ports were to be used, each routine
 *   would have to be multiplexed on the interface handle that is passed in.
 *
 *   Currently, the Uart is set to 8-N-1 operation at 9600  bps.  These values
 *   can be change by modifying tfUartOpen and the constant TM_UART_BPS.  No
 *   type of hardware flow control is supported.
 *
 *   The only difference between this driver and the 8250 driver is that this
 *   takes advantage of the buffering (16-byte FIFO) capabilities of the
 *   Uart. Both drivers should work with a standard Uart serial controller.
 */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* Use UART0 or UART1 */
#define TM_UART0

/* Define UART register base address */
#define TM_SYSCFG  (0x03ff0000)
#define TM_UART0_BASE (TM_SYSCFG + 0xD000)
#define TM_UART1_BASE  (TM_SYSCFG + 0xE000)

#ifdef TM_UART0
#define TM_UART_BASE       TM_UART0_BASE
#else /* TM_UART0 */
#define TM_UART_BASE       TM_UART1_BASE
#endif /* TM_UART0 */

#ifndef TM_SYSTEM_CLOCKH
#define TM_SYSTEM_CLOCKH    50000000
#endif /* TM_SYSTEM_CLOCKH */


/* --------------  ARM 7t ------------------ */

/* Serial settings....................... */

#define ULCON   0x00
#define UCON    0x04
#define USTAT   0x08
#define UTXBUF  0x0C
#define URXBUF  0x10
#define UBRDIV  0x14

/* Line control register bits............ */
#define ULCR8bits       (3)
#define ULCRS1StopBit   (0)
#define ULCRNoParity    (0)

/* UART Control Register bits............ */
#define UCRRxM      (1)
#define UCRRxSI     (1 << 2)
#define UCRTxM      (1 << 3)
#define UCRDSR      (1 << 5)
#define UCRLPB      (1 << 7)

/* UART Status Register bits */
#define USROverrun      (1 << 0)
#define USRParity       (1 << 1)
#define USRFraming      (1 << 2)
#define USRBreak        (1 << 3)
#define USRDTR          (1 << 4)
#define USRRxData       (1 << 5) 
#define USRTxHoldEmpty  (1 << 6)
#define USRTxEmpty      (1 << 7)

/* UART primitives */

#define GET_STATUS(p)   (*(volatile unsigned  *)((p) + USTAT))
#define RX_DATA(s)      ((s) & USRRxData)
#define RX_OVER_RUN(s)  ((s) & USROverrun)
#define TX_READY(s)     ((s) & USRTxHoldEmpty)
#define PUT_CHAR(p,c)   (*(unsigned  *)((p) + UTXBUF) = (unsigned )(c))
#define GET_CHAR(p)     (*(volatile unsigned  *)((p) + URXBUF))

/* baut rate macros  for 50mHx MCLOCK frequency */
#define BAUD_1200       (1301 << 4)
#define BAUD_2400       (650  << 4)
#define BAUD_4800       (325  << 4)
#define BAUD_9600       (162  << 4)
#define BAUD_19200      (80  << 4)
#define BAUD_38400      (40  << 4)
#define BAUD_57600      (26  << 4)
#define BAUD_115200     (13  << 4)
#define BAUD_230400     (6  << 4)
#define BAUD_460800     (2  << 4)

/* board system clock rate / 16, used to generate Baud Rate */
#define TM_UART_BPS             BAUD_9600
#define TM_UART_BUFFER_SIZE     1024
#define TM_UART_MAX_BUFFERS     2

/* For debugging purpose */
extern void ConsolePrint(char * string);

/* Local variable definitions */

/* Buffer to hold the received data */
char tlUartBuffer[TM_UART_MAX_BUFFERS][TM_UART_BUFFER_SIZE];

/* Length of the receiving buffers */
int  tlUartBufferLen[TM_UART_MAX_BUFFERS];

/* Index indicating the buffer that the stack is current reading from*/
int  tlUartReadBufferIndex;

/* Index indicating the buffer that the driver is writing into  */
int  tlUartWriteBufferIndex;

/* Interface handle for the serial port */
ttUserInterface tlUartInterfaceHandle;

/* Isr handler set this to 1 if the overrun bit of the UART
 * status register is on, tfUartReceive clear this flag and prints
 */
int tlUartRxOverRun;
/*
 * tfUartOpen
 *   Intializes the Uart serial controller and intializes the chip with
 *   baud rate and line control (stop bits, parity, etc).
 */
int tfUartOpen ( ttUserInterface interfaceHandle )
{
    tlUartInterfaceHandle = interfaceHandle;

/* Disable interrupts  */
   *(volatile unsigned *) (TM_UART_BASE + UCON) = 0;

    tlUartWriteBufferIndex = 1;
    tlUartReadBufferIndex  = 0;
    tlUartBufferLen[0] = 0;
    tlUartBufferLen[1] = 0;
    tlUartRxOverRun = 0;

/* Set port for 8 bit, one stop, no parity  */
   *(volatile unsigned *) (TM_UART_BASE + ULCON) = (ULCR8bits);

/* Set baud rate  */
   *(volatile unsigned *) (TM_UART_BASE + UBRDIV) = TM_UART_BPS;
   
/* Enable interrupt operation on UART */
   *(volatile unsigned *) (TM_UART_BASE + UCON) = 
       UCRRxM | UCRTxM | UCRRxSI | UCRDSR;

    return(TM_DEV_OKAY);
}


/*
 * tfUartClose
 *
 *   Closes the serial interface:  disable interrupts
 */
int tfUartClose ( ttUserInterface interfaceHandle )
{
    TM_UNREF_IN_ARG(interfaceHandle);

/* Disable interrupts  */
   *(volatile unsigned *) (TM_UART_BASE + UCON) = 0;

    return(TM_DEV_OKAY);
}

/*
 * tfUartIoctl
 *
 *   No need for an IOCTL routine in our serial driver.
 */
int tfUartIoctl ( ttUserInterface interfaceHandle,
                   int             flag,
                   void TM_FAR *   optionPtr,
                   int             optionLen)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(flag);
    TM_UNREF_IN_ARG(optionPtr);
    TM_UNREF_IN_ARG(optionLen);

    return(TM_DEV_OKAY);
}


/*
 * tfUartSend
 *
 *   Transmits data to the Uart serial controller.  Since the Uart is
 *   completely byte oriented, we loop through the entire packet sending one
 *   byte at a time, waiting after each sent character to ensure that the
 *   previous character has completed transmitting.
 *
 */
int tfUartSend( ttUserInterface  interfaceHandle,
                 char TM_FAR *   dataPtr,
                 int             dataLength,
                 int             flag)
{

/*    
    int i;
    char msg[256];

    if (dataLength != 0)
    {
        tm_sprintf(msg, "sent: length = %4d, data = 0x", dataLength);
        for( i = 0; i < dataLength; i++)
        {
            tm_sprintf(msg + tm_strlen(msg), " %x", (dataPtr)[i] );
        }
        tm_sprintf(msg + tm_strlen(msg), "\n");
            
        ConsolePrint(msg);
    }
*/
    while (dataLength > 0)
    {
/*
 * Ensure that the previous character has finished transmitting - this is
 * indicated by the 'transmit buffer empty' bit in the LSR being set.
 */
        while ( TX_READY(GET_STATUS(TM_UART_BASE)) == 0)
        {
/*
 * Do nothing function if user does not use transmit task. Otherwise will
 * yield the CPU.
 */
            tfInterfaceSpinLock(interfaceHandle);
        }

        tm_kernel_set_critical;
/* Send the current character to the Uart */
        PUT_CHAR(TM_UART_BASE, *dataPtr++);
/* Move on to next character */
        dataLength--;
        tm_kernel_release_critical;
    }

/* Full packet has completed sending, inform the stack */
    if (flag & TM_USER_BUFFER_LAST)
    {
        tfSendCompleteInterface(interfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }

    return(TM_DEV_OKAY);
}


/*
 * tfUartReceive
 *
 *   Reads data from the Uart FIFO into our recieve buffer and pass this
 *   buffer to the stack.
 */
int tfUartReceive ( ttUserInterface        interfaceHandle,
                    char TM_FAR * TM_FAR * dataPtrPtr,
                    int  TM_FAR *          dataLength,
                    ttUserBufferPtr        bufHandlePtr)
{
    int length;
    int i;
    char msg[256];

    tm_assert(tfUartReceive, interfaceHandle == tlUartInterfaceHandle);

/* Switch the read-write buffers */
    tm_kernel_set_critical;
    tlUartReadBufferIndex  = tlUartWriteBufferIndex;
    tlUartWriteBufferIndex = (tlUartWriteBufferIndex + 1) % TM_UART_MAX_BUFFERS;
    tm_kernel_release_critical;

    length = tlUartBufferLen[tlUartReadBufferIndex];
    tlUartBufferLen[tlUartReadBufferIndex] = 0;

    if (length > TM_UART_BUFFER_SIZE)
    {
        tfKernelError("tfUartReceive", "Buffer length exceeded!");
        tm_thread_stop;
    }

/* Return a pointer to the buffer and the length to the stack */
    *dataPtrPtr = tlUartBuffer[tlUartReadBufferIndex];
    *dataLength = length;
    *bufHandlePtr = (ttUserBufferPtr) 0;

    if (tlUartRxOverRun == 1)
    {
        tlUartRxOverRun = 0;
        ConsolePrint("UART overrun !\n");   
    }

/* Print to the console for debugging */
/*
    if (length != 0)
    {
        tm_sprintf(msg, "recv: read buffer = %d, write buffer = %d, length = %4d, data = \n", 
                   tlUartReadBufferIndex, 
                   tlUartWriteBufferIndex,
                   length);
                  
        for( i = 0; i < length; i++)
        {
            tm_sprintf(msg + tm_strlen(msg), " %x", (*dataPtrPtr)[i] );
            if (i % 16 == 0)
            {
                tm_sprintf(msg + tm_strlen(msg), "\n");
            }
            else if (i % 8 == 0)
            {
                tm_sprintf(msg + tm_strlen(msg), " -");
            }
        }
        tm_sprintf(msg + tm_strlen(msg), "\n\n");
            
        ConsolePrint(msg);
    }
*/
    
    return (TM_DEV_OKAY);
}


/*
 * tfUartIsrHandler
 *
 *   UART receive interrupt handle function
 */
void tfUartIsrHandler(void)
{
    int oldBufferLen;
    int newBufferLen;

    oldBufferLen = tlUartBufferLen[tlUartWriteBufferIndex];
    newBufferLen = oldBufferLen;

/*
 * While the receive buffer register is ready and we have room in our 
 * buffer, keep reading the data, receive register status is automatically
 * reset when data is read.
 */
 
    if (RX_OVER_RUN(GET_STATUS(TM_UART_BASE)) !=0)
    {
        tlUartRxOverRun = 1;
    }
 
    while (    newBufferLen < TM_UART_BUFFER_SIZE
            && RX_DATA(GET_STATUS(TM_UART_BASE)) !=0 )

    {
        tlUartBuffer[tlUartWriteBufferIndex][newBufferLen++]
            = GET_CHAR(TM_UART_BASE);
    }

    if (newBufferLen != oldBufferLen)
    {
        tlUartBufferLen[tlUartWriteBufferIndex] = newBufferLen;
/* notify the stack that we received some data */
        tfNotifyInterfaceIsr(tlUartInterfaceHandle, 1, 0, 0UL, 0UL);
    }

    return;
}


/*
 * tfUartFreeReceiveBuffer
 *
 *   Called when the stack is done with a recieved packet (ie, character).
 *   This should be immediately after it is passed up the stack, since both of
 *   the serial link layers (PPP,SLIP) copy immediately.  Therefore, we only
 *   use one buffer to recieve data into, so this routine does nothing.
 *
 */
int tfUartFreeReceiveBuffer(
    ttUserInterface interfaceHandle,
    char TM_FAR *   dataPtr)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtr);
    return TM_DEV_OKAY;
}


int tfUartGetPhysicalAddress(
    ttUserInterface interfaceHandle,
    char TM_FAR *   physicalAddress)
{

    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(physicalAddress);
   
    return TM_DEV_OKAY;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trarm7t.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 38126    Output checksum: 59453
 */
