/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for AMX KADAK RTOS
 * Filename: tramx.c
 * Author: Odile and Jared
 * Date Created: 10/01/98
 * $Source: source/kernel/amx/aj/tramx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:33CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
                                                                                          
#include "amx831cf.h"           /* AMX C Header                 */
#include "amx831sd.h"           /* AMX Structure Definitions    */
#include "amx831ec.h"           /* AMX Error Codes              */
#include <trsocket.h>           /* Treck user include file      */

#ifndef TM_KERN_OKAY
#define TM_KERN_OKAY   0
#endif /* TM_KERN_OKAY */

#ifndef TM_KERN_ERROR
#define TM_KERN_ERROR -1
#endif /* TM_KERN_ERROR */

static char           tlErrorFunction[80];
static char           tlErrorType[80];
static char           tlWarnFunction[80];
static char           tlWarnType[80];
static struct amxisps tlIsproot;        /* ISP Root */

/*
 * Initialize the OS
 */
void  tfKernelInitialize(void)
{
}

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc (unsigned size)
{
    char TM_FAR *blockp;
    long memsize;

    if ((ajmget((long)size, &blockp, &memsize)) == AERMNA)
    {
        tfKernelError("tfKernelMalloc","OS unable to allocate memory");
    }
    return (blockp);
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree (void TM_FAR *memoryBlockPtr)
{
    if ((ajmfre((char TM_FAR *) memoryBlockPtr)) != AEROK)
    {
        tfKernelError("tfKernelFree",
                      "Invalid Memory Block Pointer or Memory Not In Use");
    }
}


/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */

/*
 * Create a counting semaphore and increment semaphore count
 */
int tfKernelCreateCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int              errorCode;

    errorCode = ajsmcre( (AMXID *)(&countingSemaphorePtr->genIntParm),
                         0,
                         (char TM_FAR *)0 );
    if(errorCode == AEROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int errorCode;

    errorCode = ajsmwat( (AMXID)countingSemaphorePtr->genIntParm,
                         0,
                         0 );
    if (errorCode == AEROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore, up to tickTimeout in ticks.
 */
int tfKernelPendCountSemTimeout ( ttUserGenericUnionPtr countingSemaphorePtr, 
                                  ttUser32Bit tickTimeout )
{
    int errorCode;

    errorCode = ajsmwat( (AMXID)countingSemaphorePtr->genIntParm,
                         (long)tickTimeout,
                         0 );
    if (errorCode == AEROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        if (errorCode == AERTMO)
        {
            errorCode = TM_ETIMEDOUT;
        }
        else
        {
            errorCode = TM_KERN_ERROR;
        }
    }
    return errorCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int errorCode;

    errorCode = ajsmsig((AMXID)countingSemaphorePtr->genIntParm);
    if (errorCode == AEROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

void tfKernelError ( const char TM_FAR * functionNamePtr,
                     const char TM_FAR * errorMessagePtr )
{
    char TM_FAR * ptr;

    ptr = tlErrorFunction;
    while (*functionNamePtr)
    {
        *ptr++ = *functionNamePtr++;
    }
    ptr = tlErrorType;
    while (*errorMessagePtr)
    {
        *ptr++ = *errorMessagePtr++;
    }
   
    tm_thread_stop;
}

void tfKernelWarning ( const char TM_FAR * functionNamePtr,
                       const char TM_FAR * warnMessagePtr )
{
    char TM_FAR * ptr;

    ptr = tlWarnFunction;
    while (*functionNamePtr)
    {
        *ptr++ = *functionNamePtr++;
    }
    ptr = tlWarnType;
    while (*warnMessagePtr)
    {
        *ptr++ = *warnMessagePtr++;
    }
}



/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent (ttUserGenericUnionPtr eventPtr)
{
    int              errorCode;

    errorCode = ajsmcre( (AMXID TM_FAR *)(&(eventPtr->genIntParm)),
                         0,
                         (char TM_FAR *)0 );
    if(errorCode != AEROK)
    {
        tfKernelError("tfKernelCreateEvent", "Failed to create event");
        tm_thread_stop;
    }
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent (ttUserGenericUnionPtr eventPtr)
{
    int              errorCode;

    errorCode = ajsmwat((AMXID)eventPtr->genIntParm, 0UL, 0);
    if(errorCode != AEROK)
    {
        tfKernelError("tfKernelPendEvent", "Failed to wait on event");
        tm_thread_stop;
    }
}

/*
 * Resume waiting task that is waiting on this event
 */
void tfKernelIsrPostEvent (ttUserGenericUnionPtr eventPtr)
{
    int         errorCode;

    errorCode = ajsmsig((AMXID)eventPtr->genIntParm);
    if (errorCode != AEROK)
    {
        tfKernelError("tfKernelIsrPostEvent",
                "Invalid semaphore id or no AMX message envelopes available");
    }
}

void tfKernelInstallIsrHandler ( ttUserIsrHandlerPtr funcPtr,
                                 unsigned long       offset )
{
    ajispm(funcPtr,&tlIsproot);                  /* Make ISP root */
/* Install root into vector table */
    if ((ajivtw((int)offset, (void (*)())&tlIsproot)) == AERIPR)
    {
        tfKernelError("tfKernelInstallIsrHandler",
                      "Access to Interrupt Vector Table denied");
    }
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
    return;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: tramx.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 31191    Output checksum: 59759
 */
