/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for AMX KADAK RTOS
 * Filename: tramx.c
 * Author: Odile
 * Date Created: 10/01/98
 * $Source: source/kernel/amx/cj/tramx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:34CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <cjzzz.h>          /* AMX include files       */

#include <trsocket.h>       /* Treck user include file */

#ifndef TM_USE_SHEAP

#ifndef TM_CJ_POOL_SIZE
#define TM_CJ_POOL_SIZE     128000UL
#endif /* CJ_POOL_SIZE */

CJ_ID          tlCjMemId;
unsigned long  tlCjMemPool[TM_CJ_POOL_SIZE];

#endif /* TM_USE_SHEAP */

#ifndef TM_KERN_OKAY
#define TM_KERN_OKAY   0
#endif /* TM_KERN_OKAY */

#ifndef TM_KERN_ERROR
#define TM_KERN_ERROR -1
#endif /* TM_KERN_ERROR */

static int  tlSemCount;
static char tlErrorFunction[80];
static char tlErrorType[80];
static char tlWarnFunction[80];
static char tlWarnType[80];

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc (unsigned size)
{
#ifdef TM_USE_SHEAP
    return tfSheapMalloc(size);
#else /* TM_USE_SHEAP */
    void        * memoryBlockPtr;
    unsigned long lsize;
    int           errorCode;

    errorCode = cjmmget( tlCjMemId, (unsigned long)size, &memoryBlockPtr,
                         &lsize );
    if (errorCode != CJ_EROK)
    {
        memoryBlockPtr = (void TM_FAR *)0;
    }
    return (void TM_FAR *)memoryBlockPtr;
#endif /* TM_USE_SHEAP */
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree (void TM_FAR *memoryBlockPtr)
{
#ifdef TM_USE_SHEAP
    tfSheapFree(memoryBlockPtr);
#else /* !TM_USE_SHEAP */
    (void)cjmmfree(memoryBlockPtr);
#endif /* TM_USE_SHEAP */
}

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical (void)
{
    cjcfdi();
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical (void)
{
    cjcfei();
}


/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int              errorCode;
    int              semTemp;
    char    TM_FAR * tagPtr;

    tagPtr = "T000";
    tagPtr[3] = (tlSemCount % 10) + '0'; /* unit digit */
    semTemp = tlSemCount/10;
    tagPtr[2] = (semTemp % 10) + '0'; /* teen digit */
    semTemp = semTemp/10;
    tagPtr[1] = (semTemp % 10) + '0'; /* hundred digit */
    errorCode = cjsmcreate( (CJ_ID *)countingSemaphorePtr,
                            tagPtr,
                            0 );
    tlSemCount++;
    if(errorCode == CJ_EROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int errorCode;

    errorCode = cjsmwait( (CJ_ID)countingSemaphorePtr->genIntParm,
                          0,
                          (CJ_TIME)0);
    if (errorCode == CJ_EROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore, up to tickTimeout in ticks.
 */
int tfKernelPendCountSemTimeout ( ttUserGenericUnionPtr countingSemaphorePtr, 
                                  ttUser32Bit tickTimeout )
{
    int errorCode;

    errorCode = cjsmwait( (CJ_ID)countingSemaphorePtr->genIntParm,
                          0,
                          (CJ_TIME)tickTimeout);
    if (errorCode == CJ_EROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        if (errorCode == CJ_WRTMOUT)
        {
            errorCode = TM_ETIMEDOUT;
        }
        else
        {
            errorCode = TM_KERN_ERROR;
        }
    }
    return errorCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int errorCode;

    errorCode = cjsmsignal((CJ_ID)countingSemaphorePtr->genIntParm);
    if (errorCode == CJ_EROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int errorCode;

    errorCode = cjsmdelete((CJ_ID)countingSemaphorePtr->genIntParm);
    if (errorCode == CJ_EROK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}


void  tfKernelError ( const char TM_FAR * functionNamePtr,
                      const char TM_FAR * errorMessagePtr )
{
    char TM_FAR * ptr;

    ptr = tlErrorFunction;
    while (*functionNamePtr)
    {
        *ptr++ = *functionNamePtr++;
    }
    ptr = tlErrorType;
    while (*errorMessagePtr)
    {
        *ptr++ = *errorMessagePtr++;
    }
   
    tm_thread_stop;
}

void  tfKernelWarning ( const char TM_FAR *functionNamePtr,
                        const char TM_FAR *warnMessagePtr )
{
    char TM_FAR * ptr;

    ptr = tlWarnFunction;
    while (*functionNamePtr)
    {
        *ptr++ = *functionNamePtr++;
    }
    ptr = tlWarnType;
    while (*warnMessagePtr)
    {
        *ptr++ = *warnMessagePtr++;
    }
}



void  tfKernelInitialize (void)
{
#ifndef TM_USE_SHEAP
    int errorCode;

    errorCode = cjmmcreate( &tlCjMemId, "TRMM", (void *)&tlCjMemPool[0],
                            TM_CJ_POOL_SIZE);
    if (errorCode != CJ_EROK)
    {
        tfKernelError("tfKernelInitialize", "cjmmcreate failed");
    }
#endif /* TM_USE_SHEAP */

    tlSemCount = 0;
}

/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent (ttUserGenericUnionPtr eventPtr)
{
    int errorCode;

    errorCode = cjsmcreate( (CJ_ID *)eventPtr,
                            "TREV",
                            0 );
    if (errorCode != CJ_EROK)
    {
        tfKernelError("tfKernelCreateEvent", "Failed to create event");
        tm_thread_stop;
    }
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent (ttUserGenericUnionPtr eventPtr)
{
    int errorCode;

    errorCode = cjsmwait((CJ_ID)eventPtr->genIntParm, 0, (CJ_TIME)0);
    if (errorCode != CJ_EROK)
    {
        tfKernelError("tfKernelPendEvent", "Failed to wait on event");
        tm_thread_stop;
    }
}

/*
 * Resume waiting task that is waiting on this event
 */
void tfKernelIsrPostEvent (ttUserGenericUnionPtr eventPtr)
{
    (void)cjsmsignal((CJ_ID)eventPtr->genIntParm);
}

/*
 * Install an ISR Handler 
 */
void tfKernelInstallIsrHandler ( ttUserIsrHandlerPtr funcPtr,
                                 unsigned long       offset )
{
    offset = offset/4;
    cjksivtwr((int)offset, (CJ_ISPPROC)funcPtr);
    return;
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
    return;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: tramx.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 39779    Output checksum: 13267
 */
