/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trkernel.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/kernel/cmx/C166/trkernel.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:36:37CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Modify the include file for your Kernel/OS */
/*#include <kernel.h> */

/*#define TM_USE_SEM_FOR_EVENT*/

/* For Treck Kernel interface */
#include <trsocket.h>

/* For CMX RTOS APIs */
#include <cxfuncs.h>

/* For malloc, free */
#include <stdlib.h>

/* For printf to the serial consol */
#include <stdio.h> 

/* For critical section */
#include <C16xLib_Embedded.h>

/* For Treck simple heap */
extern ttUser32Bit TM_FAR buf0[];
extern ttUser32Bit TM_FAR buf1[];

/* numbers of dynamic sheap buffers we're allocated */
static int TM_GLOBAL_QLF tlSheapBufCount = 0;
/* Numbers of semaphore we've used */
static int TM_GLOBAL_QLF tlSemCount = 0;
/* Numbers of events we've used */
static int TM_GLOBAL_QLF tlEventCount = 0;

/* 
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
    ttUser32Bit TM_FAR * buf;

    switch(tlSheapBufCount)
    {
    case 0:
        buf = buf0;
        break;
    case 1:
        buf = buf1;
        break;
/*
    case 2:
        buf = buf2;
        break;
    case 3:
        buf = buf3;
        break;
*/
    default:
        tfKernelError("tfKernelSheapCreate",
                      "Two many sheap blocks requested");
        buf = NULL;
    }
    tlSheapBufCount++;
    *sheapSizePtr = (int)(TM_SHEAP_SIZE/TM_SHEAP_NUM_PAGE);
    return buf;
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */


/*
 * Stop all other calls into Treck by a kernel context switch
 */
void tfKernelSetCritical()
{
    CriticalSection_Start();
 }

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical()
{
    CriticalSection_End();
}


/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
/* CMS requires the applcation to keep track of the reference to the semaphor. */
    int errorCode;

    if ( K_Semaphore_Create(tlSemCount, 0)  == K_OK)
    {
        countingSemaphore->gen16BitParm = tlSemCount;
        tlSemCount ++;
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }

    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;

/* Wait on the semaphore indefinately */
    if (K_Semaphore_Wait(countingSemaphore->gen16BitParm, 0) == K_OK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }

    return errorCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to a timeout value (in ticks)
 */
int tfKernelPendCountSemTimeout( ttUserGenericUnionPtr countingSemaphore, 
                                 ttUser32Bit tickTimeout )
{
    int errorCode;

/* Wait on the semaphore for  tickTimeout in system ticks*/
    if (K_Semaphore_Wait(countingSemaphore->gen16BitParm, tickTimeout) == K_OK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }

    return errorCode;

}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;

    if (K_Semaphore_Post(countingSemaphore->gen16BitParm) == K_OK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }

    return errorCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;
/* CMX doesn't have API for deleting a semaphor, however, reset might be the
 * next match, it flushes the semaphore queue, even if it is owned by other
 * tasks
 */
    if (K_Semaphore_Reset(countingSemaphore->gen16BitParm, 1) == K_OK)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }

    return errorCode;
}


void  tfKernelError(const char TM_FAR * functionName,
                    const char TM_FAR *errorMessage)
{
/* print to the serial console and stop */
    printf("tfKernelError! %s %s\n", functionName, errorMessage);
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
/* print to the serial console and stop */
    printf("tfKernelWarning! %s %s\n", functionName, errorMessage);
}

void  tfKernelInitialize(void)
{
/* Initialize the CMX kernel, we did this in the main() */
/*    void K_OS_Init(void); */

}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
/* release control to the next same priority or lower priority task able to run. */
    K_Task_Coop_Sched();
}

/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    eventPtr->gen16BitParm = tlEventCount++;
}

/*
 * Task wait on an Event
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    TM_UNREF_IN_ARG(eventPtr);
    K_Event_Wait(eventPtr->gen8BitParm,
                 0,       /* wait indefinately*/
                 3 );     /* clear the event before waiting and after awaken*/
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Task wait on an Event up to a timeout value (in ticks).
 */
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                               ttUser32Bit tickTimeout )
{
    K_Event_Wait(eventPtr->gen8BitParm ,
                 tickTimeout,/* wait */, 
                 3 );        /* clear the event before waiting and after awaken*/
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    K_Event_Signal(3,/* All tasks wating for this event */
                   0, 
                   eventPtr->gen8BitParm);
}

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    K_Event_Signal(3,/* All tasks wating for this event */
                   0,
                   eventPtr->gen8BitParm);
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:40      For: Xilinx-Samsung
 * File Name: trkernel.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  2981    Output checksum: 46120
 */
