/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 *
 * Filename: trdspbi.c
 * Author: Ed Remmell (adapted from previous work done by Brent Yates, 
 *  Copyright Embedded Systems Design 2000)
 * Date Created:
 * $Source: source/kernel/dsp_bios/c5510dsk/trdspbi.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:37:13CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

////////////////////////////////////////////////////////////////////////////////
//  INCLUDED FILES                                                            //
////////////////////////////////////////////////////////////////////////////////
/*
 * Include DSP BIOS Stuff 
 */
#include <c55x.h>
#include <std.h>
#include <stdlib.h>
#include <log.h>
#include <tsk.h>
#include <sem.h>
#include <lck.h>

#include "tcpcfg.h"

/*
 * Include Treck Stuff
 */
#include <trsocket.h>

////////////////////////////////////////////////////////////////////////////////
//  LOCAL CONSTANT DECLARATIONS                                               //
////////////////////////////////////////////////////////////////////////////////
#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1



////////////////////////////////////////////////////////////////////////////////
// The following functions are called by the Treck Protocol Stack             //
////////////////////////////////////////////////////////////////////////////////

static int pageNumber = 0;

#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
    void *memBlockPtr;
/* *sheapSizePtr can be changed by the user */

    if (pageNumber == 0)
    {
/* allocate first page out of on-chip DARAM */
        memBlockPtr = malloc(*sheapSizePtr);
    }
    else
    {
/* allocate additional pages out of slower external SDRAM */
        memBlockPtr = (void *) (0x1e0000 + ((pageNumber-1) * (*sheapSizePtr)));
    }
    pageNumber++;
/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
#ifdef _DEBUG
    if (!memBlockPtr)
    {
        tfKernelError(
            "tfKernelMalloc", "Failed to malloc memory!");
    }
    else if (((unsigned long) memBlockPtr) & 1)
    {
        tfKernelError(
            "tfKernelMalloc", "Memory block is not 32-bit aligned!");
    }
#endif

    return((ttUser32Bit TM_FAR *) memBlockPtr);
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */

/*
 * DSP BIOS Initialization needed for Treck Integration
 */
void tfKernelInitialize()
{
    // No initialization needed for DSP BIOS
}

/*
 * Yields the CPU to other tasks at the same or higher priority
 * Only used if the user uses a transmit task and wishes to call the
 * treck function "trInterfaceSpinLock" inside the device driver
 * funciton to allow other tasks to run while waiting for the device
 * driver to be ready to transmit.
 */
void tfKernelTaskYield(void)
{
/* NOTE: TSK_yield() will not work, because it only yields to another
   task of the same priority. We run each Elmic stack task at a different
   priority, so that we get preemptive RTOS behavior from DSP/BIOS. */
    TSK_sleep(1);
}

/* 
 * Report Errors and Warnings
 */
// debugging aids
void    tfKernelError(
    const char TM_FAR * function, const char TM_FAR * error)
{
    LOG_printf(&LOG_system, "Fatal error in %s: %s \n", function, error);
    while (1) {
        TSK_sleep(100); // watchdog will reset from this loop
    }
}

void    tfKernelWarning (
    const char TM_FAR * function, const char TM_FAR * error)
{
#ifdef _DEBUG
    LOG_printf(&LOG_system, "Warning in %s: %s \n", function, error);
#endif
}

#ifndef TM_TRECK_PREEMPTIVE_KERNEL
/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void *memBlockPtr;

/* size is specified in MADUs, i.e. units of 16-bits which is the Minimally
   Addressable Data Unit on the TI C5x DSP */
    memBlockPtr = malloc(size);
#ifdef _DEBUG
    if (!memBlockPtr)
    {
        tfKernelError(
            "tfKernelMalloc", "Failed to malloc memory!");
    }
    else if (((unsigned long) memBlockPtr) & 1)
    {
        tfKernelError(
            "tfKernelMalloc", "Memory block is not 32-bit aligned!");
    }
#endif

    return(memBlockPtr);
}

/*
 * Free the memory back to the kernel
 */
void tfKernelFree(void TM_FAR* memoryBlockPtr)
{
    free(memoryBlockPtr);
}
#endif /* ! TM_TRECK_PREEMPTIVE_KERNEL */

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    HWI_disable();
}

/*
 * Resume normal operation (from tfKernelSetCritical)
 */
void tfKernelReleaseCritical()
{
    HWI_enable();
}

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_Handle sem;

    // Initialize the semaphore to zero per Treck spec
    sem = SEM_create(0, NULL);
    if (sem != NULL)
    {
        countingSemaphore->genVoidParmPtr = (void TM_FAR *)sem;
    }
    else
    {
        tfKernelError("tfKernelCreateCountSem",
                      "Unable to Create DSP Bios Semaphore");
        tm_thread_stop;
    }
    return(TM_KERN_OKAY);
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_pend((SEM_Handle)countingSemaphore->genVoidParmPtr, SYS_FOREVER);
    return(TM_KERN_OKAY);
}

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_post((SEM_Handle)countingSemaphore->genVoidParmPtr);
    return(TM_KERN_OKAY);
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_delete((SEM_Handle)countingSemaphore->genVoidParmPtr);
    return(TM_KERN_OKAY);
}

/*
 * Install an ISR Handler 
 * This function is not called directly by the Treck protocol stack, but
 * is called (optionally) by a device driver to install the ISR handler.
 * funcPtr is the Function To Be Called and the Handler
 * offset is the offset into the Interrupt Vetor Table
 */
void tfKernelInstallIsrHandler(ttUserIsrHandlerPtr funcPtr, u_long offSet)
{
    return;
}

/*******************************************************************
 * Since we have a preemptive kernel and plan to use a task to handle
 * receiving packets and pumping the stack, we must define the following
 * three functions: tfKernelCreateEvent, tfKernelPendEvent, tfKernelIsrPostEvent
 *******************************************************************/

/*
 * Create an event structure for pend/post from an ISR
 * Called from task level
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_Handle sem;

    // Initialize the semaphore to zero per Treck spec
    sem = SEM_create(0, NULL);
    if (sem != NULL)
    {
        eventPtr->genVoidParmPtr = (void TM_FAR *)sem;
    }
    else
    {
        tfKernelError("tfKernelCreateCountEvent",
                      "Unable to Create DSP Bios Semaphore");
        tm_thread_stop;
    }
}

/*
 * Wait on an Event from an ISR
 * Called from task level
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_pend((SEM_Handle)eventPtr->genVoidParmPtr, SYS_FOREVER);
}

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_post((SEM_Handle)eventPtr->genVoidParmPtr);
}

/*
 * Resume waiting tasks that ware waiting on this event
 * signalled from ISR
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_ipost((SEM_Handle)eventPtr->genVoidParmPtr);
}







/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:41      For: Xilinx-Samsung
 * File Name: trdspbi.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 54923    Output checksum: 64990
 */
