/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Sample main application of the Treck TCP/IP stack on the
 *  Texas Instruments DM642 EVM using the onchip Ethernet MAC and the
 *  DSP/BIOS opreating system.
 *
 * Filename: main.c
 * Author: Ed Remmell
 * Date Created:
 * $Source: source/kernel/dsp_bios/dm642evm/main.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:37:36CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* TI header files */
#include <std.h>
#include <log.h>
#include <tsk.h>
#include <sem.h>
#include <stdio.h>

#include <csl_i2c.h>
#include <csl_emifa.h>
#include <csl_cache.h>
#include <csl_gpio.h>
#include "evmdm642_eeprom.h"
#include "evmdm642.h"
#include "evmdm642_apll.h"

/* Treck header files */
#include <trsocket.h>
#include <trsystem.h>
#include "trdm6mac.h"


/* #define only one of the following: */
#define TM_DEMO_USE_DHCP
/* #define TM_DEMO_USE_STATIC_IP */

/* Define this to use a transmit task (see the User's Manual for more info) */
/* #define TM_USE_XMIT_TASK */

/* #define TM_PING_TEST */

#if (defined(TM_DEMO_USE_DHCP) && defined(TM_DEMO_USE_STATIC_IP))
#error "define only one of TM_DEMO_USE_DHCP or TM_DEMO_USE_STATIC_IP!"
#endif

#ifdef TM_DEMO_USE_STATIC_IP
#define TM_OUR_IP_ADDRESS     "192.168.100.10"
#define TM_OUR_IP_NETMASK     "255.255.255.0"
#define TM_GATEWAY_IP_ADDRESS "192.168.100.1"
#endif /* TM_DEMO_USE_STATIC_IP */

#define TM_DEMO_MCAST_GROUP   "224.168.100.10"

#define TM_ECHO_PORT          7
#define TM_DISCARD_PORT       9
#define TM_TEST_BUFFER_SIZE   4096


#if (defined TM_USE_XMIT_TASK) && (!defined TM_TASK_XMIT)
#error You must define TM_TASK_XMIT in trsystem.h and recompile the stack if TM_USE_XMIT_TASK is enabled.
#endif

void EEPROM_init();

/* Boolean indicating whether the treck stack has been initialized. */
unsigned char treckStackStarted;

/* Interface handle for the DM642 EMAC. */
ttUserInterface interfaceHandle;

/* our local IP address and netmask */
ttUserIpAddress localIpAddr;
ttUserIpAddress localNetMask;
ttUserIpAddress gatewayIpAddr;
int addrConfigStatus;

unsigned char macAddrFromApp[6];

/* Tasks are started by DSP/BIOS after main() exits. */
Void main()
{
    treckStackStarted = 0;
    interfaceHandle = 0;
    addrConfigStatus = 0;

/* Retrieve the Ethernet address from the I2C EEPROM */
	EEPROM_init();
    EVMDM642_EEPROM_read(0x00007F00, (Uint32)macAddrFromApp, 6);
}

#ifdef TM_DEMO_USE_DHCP
/* notification function for DHCP address configuration */
void myDhcpNotifyFunc(
    ttUserInterface interfaceHandle, int errorCode)
{
    switch(errorCode)
    {
    case TM_ENOERROR:
/* we have successfully auto-configured an address using DHCP */
        addrConfigStatus = 1;
        errorCode = tfGetIpAddress(
            interfaceHandle,
            &localIpAddr,
            0);
        errorCode = tfGetNetMask(
            interfaceHandle,
            &localNetMask,
            0);
        break;

    case TM_ETIMEDOUT:
        printf("DHCP client timed-out, no DHCP server was found.\n");
        break;

    default:
        printf("DHCP client error %u.\n", errorCode);
        break;
    }

    return;
}
#endif /* TM_DEMO_USE_DHCP */

#ifdef TM_PING_TEST
static int lastPingTransmittedCount;
static int lastPingReceivedCount;
static void tfxPingCB(int pingSd)
{
    ttPingInfo pingInfo;

    tfPingGetStatistics(pingSd, &pingInfo);

    if (pingInfo.pgiSendErrorCode == TM_ENOERROR)
    {
        if (lastPingTransmittedCount != pingInfo.pgiTransmitted)
        {
/* sent Echo Request to gateway */
            lastPingTransmittedCount = pingInfo.pgiTransmitted;
        }
    }

    if (pingInfo.pgiRecvErrorCode == TM_ENOERROR)
    {
        if (lastPingReceivedCount != pingInfo.pgiReceived)
        {
            lastPingReceivedCount = pingInfo.pgiReceived;

            printf("Gateway replied to ping! sent: %u, received %u\n",
                lastPingTransmittedCount, lastPingReceivedCount);
        }
    }
}
#endif /* TM_PING_TEST */

/* Main treck task that initializes the stack, and then starts sample
 * application.
 */
void taskTreckMain(void)
{
    ttUserLinkLayer             linkLayerHandle;
    struct sockaddr_storage     localSockAddr;
    int                         errorCode;
    struct in_addr              tempInAddr;
#ifdef TM_PING_TEST
    int                         pingSd;
#endif /* TM_PING_TEST */
#ifdef TM_USE_XMIT_TASK
	short optionValue;
#endif /* TM_USE_XMIT_TASK */
    
    printf("Entering Treck main web server task.\n");
    
    tfInitTreck();
    tfStartTreck();
    
/* Use the Ethernet link layer */    
    linkLayerHandle = tfUseEthernet();
    
/* Add the network interface */
    interfaceHandle=tfAddInterface(
        "EMAC.1",
        linkLayerHandle,
        tfDm642EmacOpen,
        tfDm642EmacClose,
        tfDm642EmacSend,
        tfDm642EmacRecv,
        (ttDevFreeRecvFuncPtr)0,
        tfDm642EmacIoctl,
        tfDm642EmacGetPhysicalAddress,
        &errorCode);

#ifdef TM_USE_XMIT_TASK
	optionValue = 1;
	errorCode = tfInterfaceSetOptions(interfaceHandle, TM_DEV_OPTIONS_XMIT_TASK, &optionValue, sizeof(short));
#endif

#ifdef TM_USE_DRV_ONE_SCAT_SEND
    errorCode = tfUseInterfaceOneScatSend
        (interfaceHandle, tfDm642EmacScatSend);
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

    treckStackStarted = 1;

/* configure our local IP address */
    
#ifdef TM_DEMO_USE_STATIC_IP
    localIpAddr = inet_addr(TM_OUR_IP_ADDRESS);
    localNetMask = inet_addr(TM_OUR_IP_NETMASK);
    gatewayIpAddr = inet_addr(TM_GATEWAY_IP_ADDRESS);

/* Open the ethernet interface */
    errorCode=tfOpenInterface(
        interfaceHandle,
        localIpAddr,
        localNetMask,
        TM_DEV_MCAST_ENB | TM_DEV_SCATTER_SEND_ENB,  
        5);

    if (errorCode == TM_ENOERROR)
    {
        addrConfigStatus = 1;
    }

    tfAddDefaultGateway(interfaceHandle, gatewayIpAddr);
#endif /* TM_DEMO_USE_STATIC_IP */
#ifdef TM_DEMO_USE_DHCP
    printf("Starting DHCP auto-configuration...\n");

    tfUseDhcp(interfaceHandle, myDhcpNotifyFunc);

/* Open the ethernet interface */
    errorCode=tfOpenInterface(
        interfaceHandle,
        0,
        0,
        TM_DEV_MCAST_ENB | TM_DEV_SCATTER_SEND_ENB | TM_DEV_IP_DHCP,  
        5);

    while (!addrConfigStatus)
    {
        TSK_sleep(1);
    }
#endif /* TM_DEMO_USE_DHCP */

    memset(&localSockAddr, 0, sizeof(localSockAddr));
    localSockAddr.ss_family = PF_INET;
    localSockAddr.ss_len = sizeof(localSockAddr);
    localSockAddr.addr.ipv4.sin_addr.s_addr = localIpAddr;
    localSockAddr.addr.ipv4.sin_port = htons(80);
    
    printf(
        "  local IP address:   %s\n",
        inet_ntoa(localSockAddr.addr.ipv4.sin_addr));
    tempInAddr.s_addr = localNetMask;
    printf("  local network mask: %s\n", inet_ntoa(tempInAddr));

    errorCode = tfGetDefaultGateway(&gatewayIpAddr);
    if (errorCode == TM_ENOERROR)
    {
        tempInAddr.s_addr = gatewayIpAddr;
        printf("  gateway IP address: %s\n", inet_ntoa(tempInAddr));
    }

#ifdef TM_IGMP
/* set a default interface for sending multicast packets */
    errorCode = tfSetMcastInterface(interfaceHandle, 0);
    if(errorCode != TM_ENOERROR)
    {
        printf(
            "Couldn't set multicast interface '%s'\n",
            tfStrError(errorCode));
        return;
    }
#endif /* TM_IGMP */

#ifdef TM_PING_TEST
    lastPingTransmittedCount = 0;
    lastPingReceivedCount = 0;

/* ping just so we know everything is working */
    tempInAddr.s_addr = gatewayIpAddr;
    pingSd = tfPingOpenStart(
        inet_ntoa(tempInAddr), 0, 0, (ttPingCBFuncPtr) tfxPingCB);
#endif /* TM_PING_TEST */
    
}

/* TCP echo server task */
void taskTreckTcpEcho(void)
{
    int                         errorCode;
    int                         serverSd;
    int                         acceptSd;
    int                         optValue;
    struct sockaddr_storage     remoteAddr;
    int                         remoteAddrLen;
    struct sockaddr_storage     localSockAddr;
    ttUserMessage               bufferHandle;
    char *                      dataPtr;
    int                         bytesRecv;
    int                         bytesSent;

    printf("Entering Treck TCP echo server task.\n");

/* wait for interface to be opened */
    while (!addrConfigStatus)
    {
        TSK_sleep(1);
    }

    memset(&localSockAddr, 0, sizeof(localSockAddr));
    localSockAddr.ss_family = PF_INET;
    localSockAddr.ss_len = sizeof(localSockAddr);
    localSockAddr.addr.ipv4.sin_addr.s_addr = 0;
    localSockAddr.addr.ipv4.sin_port = htons(TM_ECHO_PORT);
    
    serverSd = socket(PF_INET, SOCK_STREAM, IP_PROTOTCP);
    if(serverSd == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf("taskTreckTcpEcho failed to create socket, %d.\n", errorCode);
        return;
    }

/* increase the size of the socket send and receive buffers */
    optValue = 32768;
    errorCode = setsockopt(
        serverSd, SOL_SOCKET, SO_RCVBUF, (char*)&optValue,
        sizeof(optValue));
    errorCode = setsockopt(
        serverSd, SOL_SOCKET, SO_SNDBUF, (char*)&optValue,
        sizeof(optValue));

/* Turn off the Nagle algorithm */
	optValue = 1;
	errorCode = setsockopt(
		serverSd, IP_PROTOTCP, TCP_NODELAY, (char*)&optValue,
		sizeof(optValue));

    errorCode = bind(
        serverSd, (struct sockaddr*)&localSockAddr, sizeof(localSockAddr));
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf("taskTreckTcpEcho failed to bind, %d.\n", errorCode);
        tfClose(serverSd);
        return;
    }

    errorCode = listen(serverSd, 2);
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf(
            "taskTreckTcpEcho failed to listen, %d.\n", errorCode);
        tfClose(serverSd);
        return;
    }

    for(;;)
    {
        acceptSd = accept(
            serverSd,
            (struct sockaddr TM_FAR *)0,
            (int TM_FAR             *)0 );

        if(acceptSd == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(serverSd);
            printf(
                "taskTreckTcpEcho failed to accept, %d.\n", errorCode);
            break;
        }
        else
        {
            remoteAddrLen = sizeof(remoteAddr);
            errorCode = getpeername(
                acceptSd, (struct sockaddr *)&remoteAddr, &remoteAddrLen);
            if (errorCode == TM_SOCKET_ERROR)
            {
                errorCode = tfGetSocketError(acceptSd);
                printf(
                    "TCP echo client Get Peer name, %s connected\n",
                    tfStrError(errorCode));
            }
            else
            {
                printf("TCP echo client %s connected\n",
                       inet_ntoa(remoteAddr.addr.ipv4.sin_addr));
            }

            for(;;)
            {
                bytesRecv = tfZeroCopyRecv(
                    acceptSd,
                    &bufferHandle,
                    &dataPtr,
                    1500,
                    0);
        
                if (bytesRecv > 0)
                {
                    bytesSent = tfZeroCopySend(
                        acceptSd,
                        bufferHandle,
                        bytesRecv,
                        0);

                    if (bytesSent == TM_SOCKET_ERROR)
                    {
                        errorCode = tfGetSocketError(acceptSd);
                        switch(errorCode)
                        {
                        case TM_ECONNRESET:
                            printf("TCP echo client connection reset.\n");
                            break;
                            
                        default:
                            printf(
                                "taskTreckTcpEcho failed to send, %d.\n",
                                errorCode);
                            break;
                        }
                        tfClose(acceptSd);
                        break;
                    }
                }
                else
                {
                    if (bytesRecv == TM_SOCKET_ERROR)
                    {
                        errorCode = tfGetSocketError(acceptSd);
                        switch(errorCode)
                        {
                        case TM_ECONNRESET:
                            printf("TCP echo client connection reset.\n");
                            break;
                            
                        default:
                            printf(
                                "taskTreckTcpEcho failed to recv, %d.\n",
                                errorCode);
                            break;
                        }
                        tfClose(acceptSd);
                        break;
                    }
                    else
                    {
                        printf("TCP echo client disconnected.\n");
                        tfClose(acceptSd);
                        break;
                    }
                }
            }
        }
    }

    tfClose(serverSd);
    return;
}

/* UDP echo server task */
void taskTreckUdpEcho(void)
{
    int                         errorCode;
    int                         serverSd;
    struct sockaddr_storage     remoteAddr;
    int                         remoteAddrLen;
    struct sockaddr_storage     localSockAddr;
    ttUserMessage               bufferHandle;
    char *                      dataPtr;
    int                         bytesRecv;
    int                         bytesSent;
#ifdef TM_IGMP
    struct ip_mreq		mcastGroup;
#endif /* TM_IGMP */
    
    printf("Entering Treck UDP echo server task.\n");

/* wait for interface to be opened */
    while (!addrConfigStatus)
    {
        TSK_sleep(1);
    }
    
    memset(&localSockAddr, 0, sizeof(localSockAddr));
    localSockAddr.ss_family = PF_INET;
    localSockAddr.ss_len = sizeof(localSockAddr);
    localSockAddr.addr.ipv4.sin_addr.s_addr = 0;
    localSockAddr.addr.ipv4.sin_port = htons(TM_ECHO_PORT);
    
startUdpTest:
    serverSd = socket(PF_INET, SOCK_DGRAM, IP_PROTOUDP);
    if(serverSd == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf("taskTreckUdpEcho failed to create socket, %d.\n", errorCode);
        return;
    }

    errorCode = bind(
        serverSd, (struct sockaddr*)&localSockAddr, sizeof(localSockAddr));
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf("taskTreckUdpEcho failed to bind, %d.\n", errorCode);
        tfClose(serverSd);
        return;
    }

#ifdef TM_IGMP
/* join a multicast group, so we can receive UDP multicast packets */
    mcastGroup.imr_multiaddr.s_addr = inet_addr(TM_DEMO_MCAST_GROUP);
    mcastGroup.imr_interface.s_addr = localIpAddr;
    errorCode = setsockopt(
	serverSd, IP_PROTOIP, IPO_ADD_MEMBERSHIP, (const char*) &mcastGroup,
        sizeof(mcastGroup));
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf(
            "taskTreckUdpEcho failed to join multicast group, %d.\n",
            errorCode);
        tfClose(serverSd);
        return;
    }
#endif /* TM_IGMP */
    
    remoteAddrLen = sizeof(remoteAddr);
    for(;;)
    {
        bytesRecv = tfZeroCopyRecvFrom(
            serverSd,
            &bufferHandle,
            &dataPtr,
            1500,
            0,
            (struct sockaddr *) &remoteAddr,
            &remoteAddrLen);
        
        if (bytesRecv > 0)
        {
            bytesSent = tfZeroCopySendTo(
                serverSd,
                bufferHandle,
                bytesRecv,
                0,
                (struct sockaddr *) &remoteAddr,
                remoteAddrLen);
            
            if (bytesSent == TM_SOCKET_ERROR)
            {
                errorCode = tfGetSocketError(serverSd);
                printf(
                    "taskTreckUdpEcho failed to send, %d.\n",
                    errorCode);
            }
        }
        else
        {
            if (bytesRecv == TM_SOCKET_ERROR)
            {
                errorCode = tfGetSocketError(serverSd);
                tfClose(serverSd);
                goto startUdpTest;
            }
            else
            {
/* we were just sent a 0-byte UDP message, we can't send a 0-byte
   UDP message so instead we just free the received message here */
                if (bufferHandle != (ttUserMessage) 0)
                {
                    tfFreeZeroCopyBuffer(bufferHandle);
                }
            }
        }
    }
}

void taskTreckTcpDiscard(void)
{
    int                         errorCode;
    int                         serverSd;
    int                         acceptSd;
    int                         optValue;
    struct sockaddr_storage     remoteAddr;
    int                         remoteAddrLen;
    struct sockaddr_storage     localSockAddr;
    ttUserMessage               bufferHandle;
    char *                      dataPtr;
    int                         bytesRecv;

    printf("Entering Treck TCP discard server task.\n");

/* wait for interface to be opened */
    while (!addrConfigStatus)
    {
        TSK_sleep(1);
    }
    
    memset(&localSockAddr, 0, sizeof(localSockAddr));
    localSockAddr.ss_family = PF_INET;
    localSockAddr.ss_len = sizeof(localSockAddr);
    localSockAddr.addr.ipv4.sin_addr.s_addr = 0;
    localSockAddr.addr.ipv4.sin_port = htons(TM_DISCARD_PORT);
    
    serverSd = socket(PF_INET, SOCK_STREAM, IP_PROTOTCP);
    if(serverSd == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf(
            "taskTreckTcpDiscard failed to create socket, %d.\n", errorCode);
        return;
    }

/* increase the size of the socket receive buffers */
    optValue = 32768;
    errorCode = setsockopt(
        serverSd, SOL_SOCKET, SO_RCVBUF, (char*)&optValue,
        sizeof(optValue));

    errorCode = bind(
        serverSd, (struct sockaddr*)&localSockAddr, sizeof(localSockAddr));
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf("taskTreckTcpDiscard failed to bind, %d.\n", errorCode);
        tfClose(serverSd);
        return;
    }

    errorCode = listen(serverSd, 2);
    if(errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(serverSd);
        printf(
            "taskTreckTcpDiscard failed to listen, %d.\n", errorCode);
        tfClose(serverSd);
        return;
    }

    for(;;)
    {
        acceptSd = accept(
            serverSd,
            (struct sockaddr TM_FAR *)0,
            (int TM_FAR             *)0 );

        if(acceptSd == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(serverSd);
            printf(
                "taskTreckTcpDiscard failed to accept, %d.\n", errorCode);
            break;
        }
        else
        {
            remoteAddrLen = sizeof(remoteAddr);
            errorCode = getpeername(
                acceptSd, (struct sockaddr *)&remoteAddr, &remoteAddrLen);
            if (errorCode == TM_SOCKET_ERROR)
            {
                errorCode = tfGetSocketError(acceptSd);
                printf(
                    "TCP discard client Get Peer name, %s connected\n",
                    tfStrError(errorCode));
            }
            else
            {
                printf("TCP discard client %s connected\n",
                       inet_ntoa(remoteAddr.addr.ipv4.sin_addr));
            }

            for(;;)
            {
                bytesRecv = tfZeroCopyRecv(
                    acceptSd,
                    &bufferHandle,
                    &dataPtr,
                    1500,
                    0);
        
                if (bytesRecv > 0)
                {
                    if (bufferHandle != (ttUserMessage) 0)
                    {
                        tfFreeZeroCopyBuffer(bufferHandle);
                    }
                }
                else
                {
                    if (bytesRecv == TM_SOCKET_ERROR)
                    {
                        errorCode = tfGetSocketError(acceptSd);
                        switch(errorCode)
                        {
                        case TM_ECONNRESET:
                            printf("TCP discard client connection reset.\n");
                            break;
                            
                        default:
                            printf(
                                "taskTreckTcpDiscard failed to recv, %d.\n",
                                errorCode);
                            break;
                        }
                        tfClose(acceptSd);
                        break;
                    }
                    else
                    {
                        printf("TCP discard client disconnected.\n");
                        tfClose(acceptSd);
                        break;
                    }
                }
            }
        }
    }

    tfClose(serverSd);

    return;
}

void taskXmit(void)
{
#ifdef TM_USE_XMIT_TASK	
	while (!treckStackStarted)
	{
	    TSK_sleep(1);
	}

	while(1)
	{
/* Wait for the post from the Treck stack */
		tfWaitXmitInterface (interfaceHandle);
/* Move the data from the stack to the device driver send
 * function
 */
		tfXmitInterface (interfaceHandle);
	}
#endif /* TM_USE_XMIT_TASK */
	return;
}

void taskTreckTelnet(void)
{
	while (!addrConfigStatus)
	{
	    TSK_sleep(1);
	}

    printf("Entering Treck telnet server task.\n");

	tfTeldUserStart(0, 10, 5, TM_BLOCKING_ON);
}

/*
 * 10 ms timer periodic task that updates and executes Treck timers, and
 * refills EMAC receive queue.
 */
void taskTreckTimer(void)
{
    unsigned int loResTicksPerTreckTick;
	int loopCntForIoctl;
    
    printf("Entering Treck timer task.\n");

    loResTicksPerTreckTick =
        (CLK_countspms() * TM_TICK_LENGTH) / (CLK_getprd() + 1);
    if (loResTicksPerTreckTick == 0)
    {
        loResTicksPerTreckTick = 1;
    }
    
	while (!treckStackStarted)
	{
		TSK_sleep(1);
	}

	loopCntForIoctl = 0;
    while(1)
    {
        TSK_sleep(loResTicksPerTreckTick);
        
        tfTimerUpdate();
        tfTimerExecute();

        loopCntForIoctl++;
        if (loopCntForIoctl >= 10)
        {
            loopCntForIoctl = 0;
			if (interfaceHandle)
			{
				tfIoctlInterface(
					interfaceHandle, TM_EMAC_IOCTL_REFILL, (void *) 0, 0);
			}
        }
    }
}

/*
 * Treck receive task.
 */
void taskTreckRecv(void)
{
    printf("Entering Treck receive task.\n");

/* wait for interface to be opened */
    while (interfaceHandle == (ttUserInterface)0)
    {
        TSK_sleep(1);
    }
    
    while (1)
    {
        tfWaitReceiveInterface(interfaceHandle);
        tfRecvInterface(interfaceHandle);
    }
}    


/* I2C handle */
extern I2C_Handle EVMDM642_I2C_hI2C;

/* GPIO handle */
extern GPIO_Handle EVMDM642_GPIO_hGPIO;

/*
 * Function to allow access to the EEPROM, where the MAC address is stored
 */
void EEPROM_init()
{
    volatile Uint32 test;
    
    I2C_Config i2cCfg = {
        0x0000007f, /* I2COAR -    Not used if master */
        0x00000000, /* I2CIER -    Disable interrupts, use polling */
        0x0000001b, /* I2CCLKL -   Low period for 100KHz operation */
        0x0000001b, /* I2CCLKH -   High period for 100KHz operation */
        0x00000002, /* I2CCNT -    Data words per transmission */
        0x0000001a, /* I2CSAR -    Slave address */
        0x00004680, /* I2CMDR -    Mode */
        0x00000019  /* I2CPSC -    Prescale 300MHz to 12MHz */
    };
        
    /* Unlock PERCFG through PCFGLOCK */
    *((unsigned long *)0x1b3f018) = 0x10c0010c;
    
    /* Enable VP0-VP2, I2C and McASP0 in PERCFG */
    *((unsigned long *)0x1b3f000) = 0x79;
    
    /* Read back PERCFG */
    test = *((unsigned long *)0x1b3f000);
    
    /* Wait at least 128 CPU cycles */
    EVMDM642_wait(128);

    /* Open I2C handle */
    EVMDM642_I2C_hI2C = I2C_open(I2C_PORT0, I2C_OPEN_RESET);

    /* Configure I2C controller */
    I2C_config(EVMDM642_I2C_hI2C, &i2cCfg);

    /* Take I2C out of reset */        
    I2C_outOfReset(EVMDM642_I2C_hI2C);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: main.c        File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 62733    Output checksum: 58115
 */
