/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated.
 * Copyright laws and International Treaties protect the contents of this
 * file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for DSP/BIOS by Texas Instruments
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 *
 * Filename: trdspbios.c
 * Author: Jason
 * Date Created:
 * $Source: source/kernel/dsp_bios/dm642evm/trdspbios.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:37:38CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include DSP/BIOS stuff. 
 */
#include <std.h>
#include <sem.h>
#include <stdio.h>

/*
 * Include Treck Stuff
 */
#include <trsocket.h>



#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1


void tfKernelInitialize()
{
/* Nothing needs to be done for DSP/BIOS. */
}

/*
 * Yields the CPU to other tasks at the same or higher priority
 */
void tfKernelTaskYield(void)
{
    TSK_yield();
}


/* 
 * Report an Error
 */
void tfKernelError(const char TM_FAR *functionName,
                   const char TM_FAR *errorMessage)
{
    printf("fatal error: %s --- %s\n",functionName, errorMessage); 
    tm_thread_stop;
}

void tfKernelWarning(const char TM_FAR *functionName,
                     const char TM_FAR *warnMessage)
{
    printf("warning: %s --- %s\n",functionName, warnMessage); 
}

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void *memBlockPtr;

    memBlockPtr = malloc(size);
#ifdef _DEBUG
    if (!memBlockPtr)
    {
        tfKernelError(
            "tfKernelMalloc", "Failed to malloc memory!");
    }
    else if (((unsigned int) memBlockPtr) & 3)
    {
        tfKernelError(
            "tfKernelMalloc", "Memory block is not 32-bit aligned!");
    }
#endif

    return(memBlockPtr);
}

/*
 * Free the memory back to the kernel
 */

void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    free(memoryBlockPtr);
}


/* OK to use global value to save CSR for critical sections, since the
 * Treck code *never* nests critical sections.
 */
Uns savedCSR;

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    savedCSR = HWI_disable();
}

/*
 * Resume normal operation (from tfKernelSetCritical)
 */
void tfKernelReleaseCritical()
{
    HWI_restore(savedCSR);
}


/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_Handle semHandle;
    int        retCode;


/* Create semaphore with an initial count of zero, and no special attributes. */
    semHandle = SEM_create(0, (SEM_Attrs *) 0);

    if (semHandle == (SEM_Handle) 0)
    {
        retCode = TM_KERN_ERROR;
        printf("Error creating semaphore\n");
    }
    else
    {
        countingSemaphore->genVoidParmPtr = (void TM_FAR *) semHandle;
        retCode = TM_KERN_OKAY;
    }

    return retCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    Bool result;
    int retCode;

    result = SEM_pend((SEM_Handle) countingSemaphore->genVoidParmPtr,
                      SYS_FOREVER);

    if (result)
    {
        retCode = TM_KERN_OKAY;
    }
    else
    {
        printf("Failed to pend on semaphore.\n");
        retCode = TM_KERN_ERROR;
    }

    return retCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to tickTimeout (in ticks)
 */
int tfKernelPendCountSemTimeout(ttUserGenericUnionPtr countingSemaphore,
                                ttUser32Bit tickTimeout)
{
    Bool result;
    int retCode;

    result = SEM_pend((SEM_Handle) countingSemaphore->genVoidParmPtr,
                      tickTimeout);

    if (result)
    {
        retCode = TM_KERN_OKAY;
    }
    else
    {
        retCode = TM_ETIMEDOUT;
    }

    return retCode;    
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_post( (SEM_Handle) countingSemaphore->genVoidParmPtr);

    return TM_KERN_OKAY;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    SEM_delete( (SEM_Handle)countingSemaphore->genVoidParmPtr);

    return TM_KERN_OKAY;
}

/*
 * Install an ISR Handler 
 */
void tfKernelInstallIsrHandler(ttUserIsrHandlerPtr funcPtr, ttUser32Bit offSet)
{
    return;
}


/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_Handle semHandle;

/* Create semaphore with an initial count of zero, and no special attributes. */
    semHandle = SEM_create(0, (SEM_Attrs *) 0);

    eventPtr->genVoidParmPtr = (SEM_Handle) semHandle;
    
    return;
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_pend((SEM_Handle) eventPtr->genVoidParmPtr, SYS_FOREVER);

    return;
}

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_post((SEM_Handle)eventPtr->genVoidParmPtr);
}

/*
 * Resume waiting tasks that ware waiting on this event
 * signalled from ISR
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    SEM_ipost((SEM_Handle)eventPtr->genVoidParmPtr);
}

/* 
 * Treck Stack calls this fucntion to ask for the system time. It is 
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network, 
 * or from application user interface.
 * 
 * Parameters:
 *  secondsPtr: 
 *      Pointer to store number of seconds elapsed since midnight (00:00:00), 
 *      January 1, 1970, Coordinated Universal Time (UTC). 
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT). 
 *      This matches the return value of the ANSI C library 
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year. 
 *      In case this value ever wrap arounds, we need another space to hold the 
 *      a bigger units of time. This is the purpose of daysPtr. 
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr  
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *  
 */
int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    int errorCode; 
  
    errorCode = TM_KERN_OKAY;
    *daysPtr = 0;
    *secondsPtr = time(NULL);

    return errorCode;
} 





/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: trdspbios.c   File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 25500    Output checksum: 15259
 */
