/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2005
 * Copyright Elmic Systems Japan 1997 - 2005
 * Portions Copyright Stretch Incorporated 2005
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 * Checksum function for the Stretch S5000. This includes both code optimized
 * for Stretch and the original 'C' version of the checksum, if the user
 * doesn't want to use ISEF for this.
 *
 * Filename: checksum.c
 * Author: Jared August
 * Date Created: 8/26/2005
 * $Source: source/kernel/none/s56xx/checksum.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:08CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

#include <sx-misc.h>
#include "libei.h"
#include "main.h"

/*
 * tfPacketChecksum() Function Description
 * This function performs checksum on (IP, ICMP, IGMP, UDP and TCP)
 * packets. Note that we cannot assume that data is contiguous. We follow
 * the chain of ttPacket (using the pktNextPtr), and compute the checksum
 * on each contiguous data area. We try and unroll the loop as much as we can.
 * This version handles odd byte lengths in inner links, and odd linkDataPtr.
 *
 * Parameter       Description
 * packetPtr       Pointer to a ttPacket. packetPtr->pktLinkDataPtr
 *                 points to the data to be checksummed.
 * length          Length of data to be checksumed.
 * pshHdrPtr       Pointer to pseudo header
 * pshLen          length of pseudo-header
 * Returns
 * Value   Meaning
 * 16bit one's complement checksum
 */

tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen    totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
#ifdef TM_56XX_USE_ISEF_CSUM
	WR                 wrData1, wrData2, wrData3, wrDataOut; 
	unsigned short int acc; 
	unsigned int       i, pre_aligned_length, post_aligned_length, rounds;
#endif /* TM_56XX_USE_ISEF_CSUM */
    tt16BitPtr         link16BitDataPtr;
    tt32Bit            sum;
    tt32Bit            offset;
    ttPktLen           linkLength;
    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    }                  oddByte16Bit;
    tt8Bit             sumByteSwapped;

    sum = TM_UL(0);
/* 
 * The pseudo header length must be a either 12, or 40, 
 * if pshHdrPtr is non null 
 */
    tm_assert( tfPacketChecksum,
               (    (pshLen == TM_16BIT_ZERO) 
                 || (    (pshHdrPtr != TM_16BIT_NULL_PTR)
                      && ((pshLen == 40) || (pshLen == 12)) ) ) );
/* sum in the pseudo-header, if necessary */
    if (pshLen != 0)
    {
/* unroll the Pseudo header checksum */
/* 6 words (12 bytes) */
/* Should work OK for 16-bit DSP's */
        sum = pshHdrPtr[0]; sum += pshHdrPtr[1]; sum += pshHdrPtr[2];
        sum += pshHdrPtr[3]; sum += pshHdrPtr[4]; sum += pshHdrPtr[5];
    }
    sumByteSwapped = TM_8BIT_ZERO;
    linkLength = (ttPktLen)0;
    do
    {
/* Point to data in this link */
        link16BitDataPtr = (tt16BitPtr)packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
        if (linkLength == (ttPktLen)0)
        {
            linkLength = packetPtr->pktLinkDataLength;
        }
        else
        {
/* previous link had odd byte length, store 2nd byte to make a word */
            oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
            sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
            link16BitDataPtr = (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
            linkLength = packetPtr->pktLinkDataLength - 1;
            totalDataLength--;
        }
        if (linkLength > totalDataLength)
        {
/* Do not span more than asked */
            linkLength = totalDataLength;
        }
/* left over data to be checksumed after this link */
        totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
        if ( (tm_is_odd((ttUserPtrCastToInt)
                        link16BitDataPtr)) && (linkLength != 0) )
        {
/* save odd byte for later addition */
            oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
/* Update pointer, and length. Pointer now is on even byte boundary */
            link16BitDataPtr =
                               (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
            linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
            tm_byte_swap_sum(sum);
            sumByteSwapped = TM_8BIT_YES;
        }
#ifdef TM_56XX_USE_ISEF_CSUM
/* Optimized checksum routine to make use of ISEF capabilities */

/* Data must be aligned on a 128-bit boundary. Figure out how much to feed
 * into the ISEF instruction on the first pass, so all other passes will be
 * correctly aligned */
    	pre_aligned_length = ((unsigned int)link16BitDataPtr) & 0x000000f; 
	    pre_aligned_length = 48 - pre_aligned_length;
        if (pre_aligned_length > linkLength)
        {
            pre_aligned_length = linkLength;
        }

/*	Take care of unaligned data at the beginning of the buffer */
    	WRGET0INIT(0, link16BitDataPtr); 
	    WRGET0X(&wrData1, pre_aligned_length); 
    	WRGET0X(&wrData2, pre_aligned_length - 16);
        WRGET0X(&wrData3, pre_aligned_length - 32);
    	CHKSUM16_INIT(wrData1, wrData2, wrData3, &wrDataOut); 

/* This data is now 128-bit aligned */
	    rounds = (linkLength - pre_aligned_length) / 48; 
	    for (i = 0; i < rounds; i++)
        {
            WRGET0I(&wrData1, 16); 
     	    WRGET0I(&wrData2, 16); 
    	    WRGET0I(&wrData3, 16); 
	        CHKSUM16(wrData1, wrData2, wrData3, &wrDataOut); 
	    }	  

/* Take care of data after the aligned area */
	    post_aligned_length = linkLength - pre_aligned_length - (rounds*48); 
	    WRGET0X(&wrData1, post_aligned_length); 
	    WRGET0X(&wrData2, post_aligned_length-16); 
	    WRGET0X(&wrData3, post_aligned_length-32); 
	    CHKSUM16(wrData1, wrData2, wrData3, &wrDataOut); 
	    WRAS16I(wrDataOut, &acc, 0); 

/* Add the results into the running sum */
        sum += acc;

/* Update the data pointer and link length */
        link16BitDataPtr += (linkLength >> 1);
        linkLength = 0;
#else /* TM_56XX_USE_ISEF_CSUM */
        while (linkLength >= 32)
        {
/*  Unroll 16 words in one loop pass */
            sum += link16BitDataPtr[0]; sum += link16BitDataPtr[1];
            sum += link16BitDataPtr[2]; sum += link16BitDataPtr[3];
            sum += link16BitDataPtr[4]; sum += link16BitDataPtr[5];
            sum += link16BitDataPtr[6]; sum += link16BitDataPtr[7];
            sum += link16BitDataPtr[8]; sum += link16BitDataPtr[9];
            sum += link16BitDataPtr[10]; sum += link16BitDataPtr[11];
            sum += link16BitDataPtr[12]; sum += link16BitDataPtr[13];
            sum += link16BitDataPtr[14]; sum += link16BitDataPtr[15];
            link16BitDataPtr += 16;
            linkLength -= 32;
            if (sum & TM_UL(0x80000000))
/* If high order bit set. Fold 32 bit sum into 16 bit */
            {
                tm_fold_sum(sum);
            }
        }
        while (linkLength >= 8)
        {
/*  Unroll 4 words in one loop pass */
            sum += link16BitDataPtr[0]; sum += link16BitDataPtr[1];
            sum += link16BitDataPtr[2]; sum += link16BitDataPtr[3];
            link16BitDataPtr += 4;
            linkLength -= 8;
        }
        while (linkLength > 1)
        {
/* 1 word at a time */
            sum += *link16BitDataPtr++;
            linkLength -= 2;
        }
#endif /* TM_56XX_USE_ISEF_CSUM */
        if (sumByteSwapped != TM_8BIT_ZERO)
        {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
            tm_byte_swap_sum(sum);
/* toggle byte swapping */
            sumByteSwapped = TM_8BIT_ZERO;
            if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
            {
                oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
                linkLength = 0;
            }
            else
            {
/* remember that we have our even byte stored in oddChar[0] */
                linkLength = 1;
            }
        }
        else
        {
            if (linkLength != 0)
            {
/* Odd byte length. Save the character to add it later on */
                oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
            }
        }

/* check this before computing next packetPtr */
        if (totalDataLength == (ttPktLen)0)
        {
/* break out of the loop. Do not check 0 length after the loop */
            goto outFold;
        }
        packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
    }
    while (packetPtr != (ttPacketPtr)0);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
#ifdef TM_LINT
LINT_BOOLEAN_BEGIN
#endif /* TM_LINT */

    if (totalDataLength != (ttPktLen)0)
#ifdef TM_LINT
LINT_BOOLEAN_END
#endif /* TM_LINT */

    {
        tfKernelError("tfPacketChecksum",
            "Data in Packet chain smaller than length to be checksumed");
    }
#endif
outFold:
    if (linkLength != TM_UL(0))
    {
/*  Add left-over byte */
        oddByte16Bit.oddChar[1] = 0;
        sum = sum + oddByte16Bit.oddWord;
    }
/*  Fold 32-bit sum to 16 bits, adding possible carries from fold */
    tm_add_carries_sum(sum);
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: checksum.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 32974    Output checksum: 50649
 */
