/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2005
 * Copyright Elmic Systems Japan 1997 - 2005
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 * Stretch S56XX TCP discard server demo application
 *
 * Filename: main_rx.c
 * Author: Jared August
 * Date Created: 8/26/2005
 * $Source: source/kernel/none/s56xx/main_rx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:13CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <sx-ppi.h>

#include "trsystem.h"
#include "trsocket.h"
#include "trs56xx.h"

/* The IP address for the first Ethernet port. Subsequent ports will have an
 * address incremented by 0.0.1.0 (e.g., 192.168.1.77, 192.168.2.77)*/
#define OUR_IP_ADDR "192.168.0.77"
#define OUR_NETMASK "255.255.255.0"
/* Our listening port */
#define OUR_PORT    2000

/* How much data to receive during the test */
#define GIGS_TO_RECEIVE 5

#define ONE_GIG      0x40000000UL

#ifdef TM_56XX_USE_JUMBO_FRAMES
#define BUF_SIZE 8942
#else
#define BUF_SIZE 1460
#endif /* TM_56XX_USE_JUMBO_FRAMES */

/* Number of incoming connection requests we've received */
int __attribute__ ((section(".dram.data"))) incoming = 0;

/* Need the driver control structure to poll for events */
extern tt56xxPortInfo tl56xxPort[TM_56XX_NUM_PORTS];

/* Timer "ISR" - update the Treck timers, set a new timer event */
void timerHandler(void * junk)
{
    tfTimerUpdateIsr();
    sx_event_set_timer(TM_TICK_LENGTH, timerHandler, 0);
}

/* This function will be called when an incoming connection arrives,
 * or there is an error */
void socketConnCB(int sd, int flags)
{
    int errCode;

    if (flags & TM_CB_ACCEPT)
    {
/* Increment the number of incoming connection requests */
        incoming++;
    }
    if (flags & TM_CB_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(sd);
        printf("socket %d returned error %d\r\n", sd, errCode);
        while(1);
    }
}

int main(int argc, char ** argv)
{
    struct sockaddr_in   srvrAddr;
    struct sockaddr_in   srcAddr[TM_56XX_NUM_PORTS];
    ttUserLinkLayer      ethernetHandle;
    ttUserInterface      interfaceHandle[TM_56XX_NUM_PORTS];
    ttUserMessage        bufHandle;
    char                 *bufPtr;
    int                  errCode;
    int                  testSocket[TM_56XX_NUM_PORTS];
    int                  srvrSocket;
    int                  optValInt;
    unsigned long        optValUL;
    int                  i;
    int                  addrLen;
    char                 intfName[5];
    unsigned long        dataReceived;
    unsigned long        gigsReceived;
#ifdef TM_56XX_POLLED_MODE
    sx_ppi_stat          stat;
#endif /* TM_56XX_POLLED_MODE */

/* Initialize the Treck stack */
    tfStartTreck();

/* Start the timer */
    sx_event_set_timer(TM_TICK_LENGTH, timerHandler, 0);

/* Get an Ethernet link layer handle */
    ethernetHandle = tfUseEthernet();
    if (ethernetHandle == (ttUserLinkLayer)0)
    {
        printf("tfUseEthernet failed!\r\n");
        while(1);
    }

/* Set up most of the interface name */
    intfName[0] = 'e';
    intfName[1] = 't';
    intfName[2] = 'h';
    intfName[4] = (char)0;

/* Open each port we're using... */
    for (i = 0; i < TM_56XX_NUM_PORTS; i++)
    {
/* Store the port number in the interface name */
        intfName[3] = '0' + i;
/* Add the interface */
        interfaceHandle[i] = tfAddInterface(intfName,
                                            ethernetHandle,
                                            tf56xxOpen,
                                            tf56xxClose,
                                            tf56xxSend,
                                            tf56xxRecv,
                                            (ttDevFreeRecvFuncPtr)0,
                                            tf56xxIoctl,
                                            tf56xxGetPhyAddr,
                                            &errCode);
        if (interfaceHandle[i] == (ttUserInterface)0)
        {
            printf("tfAddInterface for interface %d failed with error code" \
                   " %d!\r\n",
                   i,
                   errCode);
            while(1);
        }

#ifdef TM_56XX_USE_JUMBO_FRAMES
/* If we're using jumbo frames, set the correct MTU */
        errCode = tfSetIfMtu(interfaceHandle[i], BUF_SIZE+40);
        if (errCode != TM_ENOERROR)
        {
            printf("tfSetIfMtu failed with error code %d!\r\n", errCode);
            while(1);
        }
#endif /* TM_56XX_USE_JUMBO_FRAMES */

/* Open the interface. The first interface will be on the X.X.0.0
 * subnet, the second will be on X.X.1.0, etc. */
        errCode = tfOpenInterface(interfaceHandle[i],
                                  inet_addr("192.168.0.77") + htonl(i << 8),
                                  inet_addr("255.255.255.0"),
                                  0,
                                  1);
        if (errCode != 0)
        {
            printf("tfOpenInterface failed with error code %d!\r\n", errCode);
            while(1);
        }
    }

/* Specify the address family */
    srvrAddr.sin_family = AF_INET;
/* Specify the destination port */
    srvrAddr.sin_port = htons(2000);
/* Specify the destination IP address */
    srvrAddr.sin_addr.s_addr = INADDR_ANY;
/* Create a socket */
    srvrSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
/* Verify the socket was created correctly. If not, return immediately */
    if (srvrSocket == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(testSocket[i]);
        printf("socket failed with error code %d!\r\n", errCode);
        while(1);
    }

/* Disable the TCP timestamp */
    optValInt = 0;
    errCode = setsockopt(srvrSocket,
                         IP_PROTOTCP,
                         TM_TCP_TS,
                         (const char *)&optValInt,
                         sizeof(optValInt));
    if (errCode == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(srvrSocket);
        printf("setsockopt failed with error code %d!\r\n", errCode);
        while(1);
    }
/* Set the socket send queue */
    optValUL = BUF_SIZE*2;
    errCode = setsockopt(srvrSocket,
                         SOL_SOCKET,
                         SO_SNDBUF,
                         (const char *)&optValUL,
                         sizeof(optValUL));
    if (errCode == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(srvrSocket);
        printf("setsockopt failed with error code %d!\r\n", errCode);
        while(1);
    }
/* Set the socket receive queue */
    optValUL = BUF_SIZE*12;
    errCode = setsockopt(srvrSocket,
                         SOL_SOCKET,
                         SO_RCVBUF,
                         (const char *)&optValUL,
                         sizeof(optValUL));
    if (errCode == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(srvrSocket);
        printf("setsockopt failed with error code %d!\r\n", errCode);
        while(1);
    }

/* Bind the socket to the port */
    errCode = bind(srvrSocket,
                   (const struct sockaddr *)&srvrAddr,
                   sizeof(srvrAddr));
    if (errCode == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(srvrSocket);
        printf("bind failed with error code %d!\r\n", errCode);
        while(1);
    }

/* Register a callback so the application is notified of an incoming
 * connection request or if an error occurs */
    tfRegisterSocketCB(srvrSocket,
                       socketConnCB,
                       TM_CB_SOCKET_ERROR|TM_CB_ACCEPT);

/* Listen for incoming connections */
    errCode = listen(srvrSocket,
                     TM_56XX_NUM_PORTS);
    if (errCode == TM_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(srvrSocket);
        printf("listen failed with error code %d!\r\n", errCode);
        while(1);
    }

/* Wait until each port has an incoming connection */
    while (incoming < TM_56XX_NUM_PORTS)
    {
/* Execute any expired timers */
        tfTimerExecute();

/* For each Ethernet port... */
        for (i = 0; i < TM_56XX_NUM_PORTS; i++)
        {
#ifdef TM_56XX_POLLED_MODE
/* Get the number of incoming packets waiting to be processed */
            (void)sx_ppi_receive_stat(tl56xxPort[i].ppi_handle, &stat);
/* Process them */
            while (stat.incr_count--)
            {
                tfRecvInterface(interfaceHandle[i]);
            }
/* Get the number of packets that the driver has finished sending */
            (void)sx_ppi_send_stat(tl56xxPort[i].ppi_handle, &stat);
/* Tell the stack they've been sent */
            while (stat.incr_count--)
            {
                tfSendCompleteInterface(interfaceHandle[i], TM_DEV_SEND_COMPLETE_APP);
            }
#else /* TM_56XX_POLLED_MODE */
/* Check if we've had any receive interrupts */
            while (!tfCheckReceiveInterface(interfaceHandle[i]))
            {
/* Process the receive interrupts */
                tfRecvInterface(interfaceHandle[i]);
            }
/* Check if we've had any transmit complete interrupts */
            while (!tfCheckSentInterface(interfaceHandle[i]))
            {
/* Process the transmit complete interrupts */
                tfSendCompleteInterface(interfaceHandle[i], TM_DEV_SEND_COMPLETE_APP);
            }
#endif /* TM_56XX_POLLED_MODE */
        }
    }

/* Accept all of the pending connections */
    for (i = 0; i < TM_56XX_NUM_PORTS; i++)
    {
        addrLen = sizeof(srcAddr[i]);
        testSocket[i] = accept(srvrSocket,
                               (struct sockaddr *)&(srcAddr[i]),
                               &addrLen);
        if (testSocket[i] == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(srvrSocket);
            printf("accept failed with error code %d!\r\n", errCode);
            while(1);
        }
    }

    dataReceived = 0;
    gigsReceived = 0;

/* Run until we've received enough data */
    while (gigsReceived < GIGS_TO_RECEIVE)
    {
        tfTimerExecute();

        for (i = 0; i < TM_56XX_NUM_PORTS; i++)
        {
#ifdef TM_56XX_POLLED_MODE
            (void)sx_ppi_receive_stat(tl56xxPort[i].ppi_handle, &stat);
            while (stat.incr_count--)
            {
                tfRecvInterface(interfaceHandle[i]);
            }
            (void)sx_ppi_send_stat(tl56xxPort[i].ppi_handle, &stat);
            while (stat.incr_count--)
            {
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#else /* TM_56XX_POLLED_MODE */
            while (!tfCheckReceiveInterface(interfaceHandle[i]))
            {
                tfRecvInterface(interfaceHandle[i]);
            }
            while (!tfCheckSentInterface(interfaceHandle[i]))
            {
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#endif /* TM_56XX_POLLED_MODE */

/* Receive any data in the socket queue */
            errCode = tfZeroCopyRecv(testSocket[i],
                                     &bufHandle,
                                     &bufPtr,
                                     BUF_SIZE,
                                     0);

/* Check if there was an error while receiving. If so, break from the loop */
            if (errCode <= 0)
            {
                if (errCode == TM_SOCKET_ERROR)
                {
                    errCode = tfGetSocketError(testSocket[i]);
                    if (errCode != TM_EWOULDBLOCK)
                    {
                        printf("recv failed with error code %d!\r\n",
                               errCode);
                        while(1);
                    }
                }
                else
                {
                    tfClose(testSocket[i]);
                    printf("socket closed!\r\n");
                }
            }
/* If no error, free the buffer we acquired */
            else
            {
                tfFreeZeroCopyBuffer(bufHandle);
                dataReceived += errCode;
            }
        }
        if (dataReceived >= ONE_GIG)
        {
            gigsReceived++;
            dataReceived -= ONE_GIG;
        }
    }
    return(0);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: main_rx.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 19635    Output checksum: 31241
 */
