/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2005
 * Copyright Elmic Systems Japan 1997 - 2005
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 * Stretch S56XX TCP discard client demo application
 *
 * Filename: main_tx.c
 * Author: Jared August
 * Date Created: 8/26/2005
 * $Source: source/kernel/none/s56xx/main_tx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:14CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <sx-ppi.h>

#include "trsystem.h"
#include "trsocket.h"
#include "trs56xx.h"

/* The IP address for the first Ethernet port. Subsequent ports will have an
 * address incremented by 0.0.1.0 (e.g., 192.168.1.77, 192.168.2.77) */
#define OUR_IP_ADDR  "192.168.0.77"
#define OUR_NETMASK  "255.255.255.0"
/* The destination IP address for the first Ethernet port. Subsequent ports
 * will send to an address incremented by 0.0.1.0 (e.g, 192.168.1.50,
 * 192.168.2.50) */
#define DEST_IP_ADDR "192.168.0.50"
/* Destination TCP port */
#define DEST_PORT    2000

/* How much data to send during the test */
#define GIGS_TO_SEND 5

#define ONE_GIG      0x40000000UL

#ifdef TM_56XX_USE_JUMBO_FRAMES
#define BUF_SIZE 8942
#else
#define BUF_SIZE 1460
#endif /* TM_56XX_USE_JUMBO_FRAMES */

/* How many connections have been established */
int __attribute__ ((section(".dram.data"))) connected = 0;

/* Need the driver control structure to poll for events */
extern tt56xxPortInfo tl56xxPort[TM_56XX_NUM_PORTS];

/* Timer "ISR" - update the Treck timers, set a new timer event */
void timerHandler(void * junk)
{
    tfTimerUpdateIsr();
    sx_event_set_timer(TM_TICK_LENGTH, timerHandler, 0);
}

/* This function will be called when connect() completes on a socket,
 * or there is an error */
void socketConnCB(int sd, int flags)
{
    int errCode;

    if (flags & TM_CB_CONNECT_COMPLT)
    {
/* Increment the number of successful connections */
        connected++;
    }
    if (flags & TM_CB_SOCKET_ERROR)
    {
        errCode = tfGetSocketError(sd);
        printf("socket %d returned error %d\r\n", sd, errCode);
        while(1);
    }
}

int main(int argc, char ** argv)
{
    struct sockaddr_in   destAddr[TM_56XX_NUM_PORTS];
    ttUserLinkLayer      ethernetHandle;
    ttUserInterface      interfaceHandle[TM_56XX_NUM_PORTS];
    ttUserMessage        bufHandle;
    char                 *bufPtr;
    int                  errCode;
    int                  testSocket[TM_56XX_NUM_PORTS];
    int                  optValInt;
    unsigned long        optValUL;
    int                  i;
    char                 intfName[5];
    unsigned long        dataSent;
    unsigned long        gigsSent;
#ifdef TM_56XX_POLLED_MODE

    sx_ppi_stat          stat;
#endif /* TM_56XX_POLLED_MODE */

/* Initialize the Treck stack */
    tfStartTreck();

/* Start the timer */
    sx_event_set_timer(TM_TICK_LENGTH, timerHandler, 0);

/* Get an Ethernet link layer handle */
    ethernetHandle = tfUseEthernet();
    if (ethernetHandle == (ttUserLinkLayer)0)
    {
        printf("tfUseEthernet failed!\r\n");
        while(1);
    }

/* Set up most of the interface name */
    intfName[0] = 'e';
    intfName[1] = 't';
    intfName[2] = 'h';
    intfName[4] = (char)0;

/* Open each port we're using... */
    for (i = 0; i < TM_56XX_NUM_PORTS; i++)
    {
/* Store the port number in the interface name */
        intfName[3] = '0' + i;
/* Add the interface */
        interfaceHandle[i] = tfAddInterface(intfName,
                                            ethernetHandle,
                                            tf56xxOpen,
                                            tf56xxClose,
                                            tf56xxSend,
                                            tf56xxRecv,
                                            (ttDevFreeRecvFuncPtr)0,
                                            tf56xxIoctl,
                                            tf56xxGetPhyAddr,
                                            &errCode);
        if (interfaceHandle[i] == (ttUserInterface)0)
        {
            printf("tfAddInterface for interface %d failed with error code" \
                   " %d!\r\n",
                   i,
                   errCode);
            while(1);
        }

#ifdef TM_56XX_USE_JUMBO_FRAMES
/* If we're using jumbo frames, set the correct MTU */
        errCode = tfSetIfMtu(interfaceHandle[i], BUF_SIZE+40);
        if (errCode != TM_ENOERROR)
        {
            printf("tfSetIfMtu failed with error code %d!\r\n", errCode);
            while(1);
        }
#endif /* TM_56XX_USE_JUMBO_FRAMES */

/* Open the interface. The first interface will be on the X.X.0.0
 * subnet, the second will be on X.X.1.0, etc. */
        errCode = tfOpenInterface(interfaceHandle[i],
                                  inet_addr(OUR_IP_ADDR) + htonl(i << 8),
                                  inet_addr(OUR_NETMASK),
                                  0,
                                  1);
        if (errCode != 0)
        {
            printf("tfOpenInterface failed with error code %d!\r\n", errCode);
            while(1);
        }
    }

/* Set up a TCP client connection on each interface */
    for (i = 0; i < TM_56XX_NUM_PORTS; i++)
    {

/* Specify the address family */
        destAddr[i].sin_family = AF_INET;
/* Specify the destination port */
        destAddr[i].sin_port = htons(DEST_PORT);
/* Specify the destination IP address */
        destAddr[i].sin_addr.s_addr = inet_addr(DEST_IP_ADDR) + htonl(i << 8);
/* Create a socket */
        testSocket[i] = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if (testSocket[i] == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(testSocket[i]);
            printf("socket failed with error code %d!\r\n", errCode);
            while(1);
        }

/* Disable the TCP timestamp */
        optValInt = 0;
        errCode = setsockopt(testSocket[i],
                             IP_PROTOTCP,
                             TM_TCP_TS,
                             (const char *)&optValInt,
                             sizeof(optValInt));
        if (errCode == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(testSocket[i]);
            printf("setsockopt failed with error code %d!\r\n", errCode);
            while(1);
        }
/* Set the socket send queue */
        optValUL = BUF_SIZE*12;
        errCode = setsockopt(testSocket[i],
                             SOL_SOCKET,
                             SO_SNDBUF,
                             (const char *)&optValUL,
                             sizeof(optValUL));
        if (errCode == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(testSocket[i]);
            printf("setsockopt failed with error code %d!\r\n", errCode);
            while(1);
        }
/* Set the socket receive queue */
        optValUL = BUF_SIZE*2;
        errCode = setsockopt(testSocket[i],
                             SOL_SOCKET,
                             SO_RCVBUF,
                             (const char *)&optValUL,
                             sizeof(optValUL));
        if (errCode == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(testSocket[i]);
            printf("setsockopt failed with error code %d!\r\n", errCode);
            while(1);
        }

/* Register a callback so the application is notified when connect()
 * completes or an error occurs */
        tfRegisterSocketCB(testSocket[i],
                           socketConnCB,
                           TM_CB_SOCKET_ERROR|TM_CB_CONNECT_COMPLT);

/* Connect to the server */
        errCode = connect(testSocket[i],
                          (struct sockaddr *)&(destAddr[i]),
                          sizeof(destAddr[i]));
        if (errCode == TM_SOCKET_ERROR)
        {
            errCode = tfGetSocketError(testSocket[i]);
            if (errCode != TM_EINPROGRESS)
            {
                printf("connect failed with error code %d!\r\n", errCode);
                while(1);
            }
        }
    }

/* Wait until all connections have completed */
    while(connected < TM_56XX_NUM_PORTS)
    {
/* Execute any expired timers */
        tfTimerExecute();

/* For each Ethernet port... */
        for (i = 0; i < TM_56XX_NUM_PORTS; i++)
        {
#ifdef TM_56XX_POLLED_MODE
/* Get the number of incoming packets waiting to be processed */
            (void)sx_ppi_receive_stat(tl56xxPort[i].ppi_handle, &stat);
/* Process them */
            while (stat.incr_count--)
            {
                tfRecvInterface(interfaceHandle[i]);
            }
/* Get the number of packets that the driver has finished sending */
            (void)sx_ppi_send_stat(tl56xxPort[i].ppi_handle, &stat);
/* Tell the stack they've been sent */
            while (stat.incr_count--)
            {
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#else /* TM_56XX_POLLED_MODE */
/* Check if we've had any receive interrupts */
            while (!tfCheckReceiveInterface(interfaceHandle[i]))
            {
/* Process the receive interrupts */
                tfRecvInterface(interfaceHandle[i]);
            }
/* Check if we've had any transmit complete interrupts */
            while (!tfCheckSentInterface(interfaceHandle[i]))
            {
/* Process the transmit complete interrupts */
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#endif /* TM_56XX_POLLED_MODE */

        }
    }

    dataSent = 0;
    gigsSent = 0;

/* Run until we've sent enough data */
    while (gigsSent < GIGS_TO_SEND)
    {
        tfTimerExecute();

        for (i = 0; i < TM_56XX_NUM_PORTS; i++)
        {
#ifdef TM_56XX_POLLED_MODE
            (void)sx_ppi_receive_stat(tl56xxPort[i].ppi_handle, &stat);
            while (stat.incr_count--)
            {
                tfRecvInterface(interfaceHandle[i]);
            }
            (void)sx_ppi_send_stat(tl56xxPort[i].ppi_handle, &stat);
            while (stat.incr_count--)
            {
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#else /* TM_56XX_POLLED_MODE */
            while (!tfCheckReceiveInterface(interfaceHandle[i]))
            {
                tfRecvInterface(interfaceHandle[i]);
            }
            while (!tfCheckSentInterface(interfaceHandle[i]))
            {
                tfSendCompleteInterface(interfaceHandle[i],
                                        TM_DEV_SEND_COMPLETE_APP);
            }
#endif /* TM_56XX_POLLED_MODE */

/* Get a zero copy buffer */
            bufHandle = tfGetZeroCopyBuffer(BUF_SIZE, &bufPtr);
            if (bufHandle == (ttUserMessage)0)
            {
                printf("tfGetZeroCopyBuffer failed!\r\n");
                while(1);
            }

/* Send another packet to the destination specified above */
            errCode = tfZeroCopySend(testSocket[i],
                                     bufHandle,
                                     BUF_SIZE,
                                     0);

/* Check if there was an error while sending. If so, break from the loop */
            if (errCode <= 0)
            {
                if (errCode == TM_SOCKET_ERROR)
                {
/* If there's an error, free the buffer */
                    tfFreeZeroCopyBuffer(bufHandle);
                    errCode = tfGetSocketError(testSocket[i]);
/* If the error isn't TM_EWOULDBLOCK, die */
                    if (errCode != TM_EWOULDBLOCK)
                    {
                        printf("send failed with error code %d!\r\n",
                               errCode);
                        while(1);
                    }
                }
                else
                {
/* The socket has closed on us */
                    tfFreeZeroCopyBuffer(bufHandle);
                    tfClose(testSocket[i]);
                    printf("socket closed!\r\n");
                }
            }
            else
            {
                dataSent += errCode;
            }
        }
        if (dataSent >= ONE_GIG)
        {
            gigsSent++;
            dataSent -= ONE_GIG;
        }
    }
    return(0);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: main_tx.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 30557    Output checksum: 41651
 */
