/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for Xilinx MicroBlaze
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trxilmb.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/kernel/none/xillinx/trmb.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:28CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

extern int rxnotify, rxpoll, txnotify, txpoll;

/*
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    return (ttUser32Bit TM_FAR *)(malloc((*sheapSizePtr)));
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void TM_FAR * memoryPtr = (void TM_FAR *)0;
    xil_printf("tfKernelMalloc called\r\n");
    memoryPtr = (void *)(tfSheapMalloc((unsigned)size));
    return memoryPtr;
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    xil_printf("tfKernelFree called\r\n");
    tfSheapFree(memoryBlockPtr);
    return;
}

/*
 * Stop all other calls into Treck by
 * a kernel context switch
 */
void tfKernelSetCritical(void)
{
    microblaze_disable_interrupts();
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical(void)
{
    microblaze_enable_interrupts();
}

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
    (void)tfKernelPrint(TM_PRINT_TYPE_FAIL,
                        "tfKernelError! %s: %s\n",
                        functionName,
                        errorMessage);
    while(1);
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
    (void)tfKernelPrint(TM_PRINT_TYPE_WARN,
                        "tfKernelWarning! %s: %s\n",
                        functionName,
                        errorMessage);
}

void  tfKernelInitialize(void)
{
    return;
}

/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    return;
}

/*
 * Task wait on an Event
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    return;
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
    return;
}

#define TM_COLOR_PRINT_MAX_LINE 256
/* print a string in the right location with good color */
/* void tfKernelPrint(int msgType, char *msgStr) */
int tfKernelPrint(int msgType, const char TM_FAR * format, ...)
{
    int                         nResult;
    char                        buffer[TM_COLOR_PRINT_MAX_LINE];
    va_list                     args;

    va_start(args, format);
    nResult = tfVSPrintF(buffer, format, args);
    va_end(args);
    msgType = msgType;
    xil_printf("%s\r\n", buffer);
    return 0;
}

/*
 * Treck Stack calls this fucntion to ask for the system time. It is
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network,
 * or from application user interface.
 *
 * Parameters:
 *  secondsPtr:
 *      Pointer to store number of seconds elapsed since midnight (00:00:00),
 *      January 1, 1970, Coordinated Universal Time (UTC).
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT).
 *      This matches the return value of the ANSI C library
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year.
 *      In case this value ever wrap arounds, we need another space to hold the
 *      a bigger units of time. This is the purpose of daysPtr.
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *
 */
int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    return TM_KERN_ERROR;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: trmb.c        File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 31209    Output checksum: 37167
 */
