/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Sample main application of the Treck TCP/IP stack on the
 *  Intel "Forest Lake" 80314 target, no RTOS.
 *
 * Filename: main.c
 * Author: Ed Remmell
 * Date Created:
 * $Source: source/kernel/none/xscale/iq80314/gnu/main.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:31CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Treck header files */
#include <trsocket.h>
#include "tr80314.h"


/* #define only one of the following: */
#define TM_DEMO_USE_DHCP
/* #define TM_DEMO_USE_STATIC_IP */


/* #define TM_PING_TEST */

#if (defined(TM_DEMO_USE_DHCP) && defined(TM_DEMO_USE_STATIC_IP))
#error "define only one of TM_DEMO_USE_DHCP or TM_DEMO_USE_STATIC_IP!"
#endif

#ifdef TM_DEMO_USE_STATIC_IP
#define TM_OUR_IP_ADDRESS     "192.168.100.10"
#define TM_OUR_IP_NETMASK     "255.255.255.0"
#define TM_GATEWAY_IP_ADDRESS "192.168.100.1"
#endif /* TM_DEMO_USE_STATIC_IP */

#define TM_DEMO_MCAST_GROUP   "224.168.100.10"

#define TM_ECHO_PORT          7
#define TM_DISCARD_PORT       9
#define TM_TEST_BUFFER_SIZE   4096


/* Boolean indicating whether the treck stack has been initialized. */
unsigned char treckStackStarted;

/* Interface handle for the Intel 80314 EMAC. */
ttUserInterface interfaceHandle;

/* our local IP address and netmask */
ttUserIpAddress localIpAddr;
ttUserIpAddress localNetMask;
ttUserIpAddress gatewayIpAddr;
int addrConfigStatus;

#ifdef TM_DEMO_USE_DHCP
/* notification function for DHCP address configuration */
void myDhcpNotifyFunc(
    ttUserInterface interfaceHandle, int errorCode)
{
    switch(errorCode)
    {
    case TM_ENOERROR:
/* we have successfully auto-configured an address using DHCP */
        addrConfigStatus = 1;
        errorCode = tfGetIpAddress(
            interfaceHandle,
            &localIpAddr,
            0);
        errorCode = tfGetNetMask(
            interfaceHandle,
            &localNetMask,
            0);
        break;

    case TM_ETIMEDOUT:
        printf("DHCP client timed-out, no DHCP server was found.\n");
        break;

    default:
        printf("DHCP client error %u.\n", errorCode);
        break;
    }

    return;
}
#endif /* TM_DEMO_USE_DHCP */

#ifdef TM_PING_TEST
static int lastPingTransmittedCount;
static int lastPingReceivedCount;
static void tfxPingCB(int pingSd)
{
    ttPingInfo pingInfo;

    tfPingGetStatistics(pingSd, &pingInfo);

    if (pingInfo.pgiSendErrorCode == TM_ENOERROR)
    {
        if (lastPingTransmittedCount != pingInfo.pgiTransmitted)
        {
/* sent Echo Request to gateway */
            lastPingTransmittedCount = pingInfo.pgiTransmitted;
        }
    }

    if (pingInfo.pgiRecvErrorCode == TM_ENOERROR)
    {
        if (lastPingReceivedCount != pingInfo.pgiReceived)
        {
            lastPingReceivedCount = pingInfo.pgiReceived;

            printf("Gateway replied to ping! sent: %u, received %u\n",
                lastPingTransmittedCount, lastPingReceivedCount);
        }
    }
}
#endif /* TM_PING_TEST */

/* Main treck task that initializes the stack, and then starts sample
 * application.
 */
int main()
{
    ttUserLinkLayer             linkLayerHandle;
    ttHttpdUserSrvHandle        serverHandle;
    struct sockaddr_storage     localSockAddr;
    int                         errorCode;
    struct in_addr              tempInAddr;
    int                         configFlags;
    int                         scatteredBufferCount;
#ifdef TM_PING_TEST
    int                         pingSd;
#endif /* TM_PING_TEST */
    
    treckStackStarted = 0;
    interfaceHandle = 0;
    addrConfigStatus = 0;
    
    tfInitTreck();
    tfStartTreck();
    
/* Use the Ethernet link layer */    
    linkLayerHandle = tfUseEthernet();
    
    interfaceHandle=tfAddInterface(
            "forestlake",
            linkLayerHandle,
            tf80314Open,
            tf80314Close,
            tf80314Send,
            tf80314Receive,
            (ttDevFreeRecvFuncPtr)0,
            tf80314Ioctl,
            tf80314GetPhysAddr,
            &errorCode);

#ifdef TM_USE_DRV_ONE_SCAT_SEND
    errorCode = tfUseInterfaceOneScatSend(interfaceHandle, tf80314ScatSend);
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

#ifdef TM_USE_DRV_ONE_SCAT_SEND
    configFlags = TM_DEV_SCATTER_SEND_ENB;
#else /* ! TM_USE_DRV_ONE_SCAT_SEND */
    configFlags = 0;
#endif /* ! TM_USE_DRV_ONE_SCAT_SEND */
    if (configFlags & TM_DEV_SCATTER_SEND_ENB)
    {
        scatteredBufferCount = 6;
    }
    else
    {
        scatteredBufferCount = 1;
    }

    treckStackStarted = 1;

/* configure our local IP address */
    
#ifdef TM_DEMO_USE_STATIC_IP
    localIpAddr = inet_addr(TM_OUR_IP_ADDRESS);
    localNetMask = inet_addr(TM_OUR_IP_NETMASK);
    gatewayIpAddr = inet_addr(TM_GATEWAY_IP_ADDRESS);

/* Open the ethernet interface */
    errorCode=tfOpenInterface(
        interfaceHandle,
        localIpAddr,
        localNetMask,
        TM_DEV_MCAST_ENB | TM_DEV_SCATTER_SEND_ENB,  
        5);

    if (errorCode == TM_ENOERROR)
    {
        addrConfigStatus = 1;
    }

    tfAddDefaultGateway(interfaceHandle, gatewayIpAddr);
#endif /* TM_DEMO_USE_STATIC_IP */
#ifdef TM_DEMO_USE_DHCP
    printf("Starting DHCP auto-configuration...\n");

    tfUseDhcp(interfaceHandle, myDhcpNotifyFunc);

/* Open the ethernet interface */
    errorCode=tfOpenInterface(
        interfaceHandle,
        0,
        0,
        TM_DEV_MCAST_ENB | TM_DEV_SCATTER_SEND_ENB | TM_DEV_IP_DHCP,  
        5);

    while (!addrConfigStatus)
    {
        tfTimerExecute();
        (void)tfIoctlInterface(
            interfaceHandle,
            TM_DEV_IOCTL_REFILL_POOL_FLAG|TM_80314_IOCTL_SEND_COMPLETE,
            (void TM_FAR *)0, 0);
        if (tfCheckReceiveInterface(interfaceHandle) == TM_ENOERROR)
        {
            tfRecvInterface(interfaceHandle); 
        }
    }
#endif /* TM_DEMO_USE_DHCP */

    memset(&localSockAddr, 0, sizeof(localSockAddr));
    localSockAddr.ss_family = PF_INET;
    localSockAddr.ss_len = sizeof(localSockAddr);
    localSockAddr.addr.ipv4.sin_addr.s_addr = localIpAddr;
    localSockAddr.addr.ipv4.sin_port = htons(80);
    
    printf(
        "  local IP address:   %s\n",
        inet_ntoa(localSockAddr.addr.ipv4.sin_addr));
    tempInAddr.s_addr = localNetMask;
    printf("  local network mask: %s\n", inet_ntoa(tempInAddr));

    errorCode = tfGetDefaultGateway(&gatewayIpAddr);
    if (errorCode == TM_ENOERROR)
    {
        tempInAddr.s_addr = gatewayIpAddr;
        printf("  gateway IP address: %s\n", inet_ntoa(tempInAddr));
    }

#ifdef TM_IGMP
/* set a default interface for sending multicast packets */
    errorCode = tfSetMcastInterface(interfaceHandle, 0);
    if(errorCode != TM_ENOERROR)
    {
        printf(
            "Couldn't set multicast interface '%s'\n",
            tfStrError(errorCode));
        return -1;
    }
#endif /* TM_IGMP */

    errorCode = tfHttpdUserInit(1);
    if(errorCode != TM_ENOERROR)
    {
        printf(
            "Couldn't initilize the web service '%s'\n",
            tfStrError(errorCode));
        return -1;
    }
    
#ifdef TM_PING_TEST
    lastPingTransmittedCount = 0;
    lastPingReceivedCount = 0;

/* ping just so we know everything is working */
    tempInAddr.s_addr = gatewayIpAddr;
    pingSd = tfPingOpenStart(
        inet_ntoa(tempInAddr), 0, 0, (ttPingCBFuncPtr) tfxPingCB);
#endif /* TM_PING_TEST */
    
    serverHandle = tfHttpdUserStart( 
        10,              /* max total Connections, */
        5,               /* max connections per IP */
        360,             /* idleTimeOut, 10 minutes */
        TM_BLOCKING_OFF,  /* blockingState, */
        "TreckDemo",     /* rootPtr, */
        "/index.htm",    /* default page */
        "/cgi-bin",      /* CGI dir */
        &localSockAddr,  /* localSockAddrPtr to listen, */
        &errorCode);     /* error code */
    
    if (errorCode == TM_ENOERROR)
    {
        tfHttpdUserAuthConfig( 
            serverHandle,
            "/auth.htm",
            "Basic Authentication Test",
            "treck",
            "treck",
            TM_HTTPD_AUTH_BAA);
    }
    
    while(1)
    {
/* execute the web server */
        tfHttpdUserExecute(serverHandle);

        tfTimerExecute();
        (void)tfIoctlInterface(
            interfaceHandle,
            TM_DEV_IOCTL_REFILL_POOL_FLAG|TM_80314_IOCTL_SEND_COMPLETE,
            (void TM_FAR *)0, 0);
        if (tfCheckReceiveInterface(interfaceHandle) == TM_ENOERROR)
        {
            tfRecvInterface(interfaceHandle); 
        }
    }
}

/* Implement the interface function that handles the HTTP events.
 * If CGI is to be used, this function is implemented in txcgi.c
 */
void tfHttpdSignalEvent(
    ttHttpdUserConHandle    conHandle,
    int                     event )
{
/*  We could get the connection information and log it */
    ttHttpdUserConEntryPtr  conEntryPtr;
    
    conEntryPtr = tfHttpdUserGetConInfo(conHandle);
    if (conEntryPtr != (ttHttpdUserConEntryPtr)0)
    {
/* log the information in the con entry*/        
    }
    
    switch(event)
    {
    case TM_HTTPD_CON_OPENED:
/* log the information in the con entry*/        
        break;
    case TM_HTTPD_CON_CLOSE:
/* log the information in the con entry*/     
        break;
    default:
        break;
    }
}



/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:42      For: Xilinx-Samsung
 * File Name: main.c        File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 51850    Output checksum: 54280
 */
