/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for Intel "Forest Lake" 80314 port
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 *
 * Filename: trkernel.c
 * Author: Paul
 * Date Created:
 * $Source: source/kernel/none/xscale/iq80314/gnu/trkernel.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:34CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Modify the include file for your Kernel/OS */
/*#include <kernel.h> */


#include <stdio.h>
#include <stdlib.h>

#include <trsocket.h>
#include <tr80314.h>

#define _Write_CPSR(VAL) \
asm("msr\tcpsr, %0" : : "rI" (VAL) : "cc", "memory")

#if defined(__cplusplus)

#define REGISTER

#else

#define REGISTER register

#endif

/* 
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    return (ttUser32Bit TM_FAR *)(malloc((*sheapSizePtr)));
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void TM_FAR * memoryPtr = (void TM_FAR *)0;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

/* To avoid compiler warning */
    TM_UNREF_IN_ARG(size);

    tm_get_current_context(contextHandle);
/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
/*    memoryPtr = (void TM_FAR *)(malloc((unsigned)size)); */
    tm_set_current_context(contextHandle);
    return memoryPtr;
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

/* To avoid compiler warning */
    TM_UNREF_IN_ARG(memoryBlockPtr);

/*
 * NOTE 'C' library free() is not re-entrant. If free() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
/*   free(memoryBlockPtr); */
     tm_get_current_context(contextHandle);
/*     free(memoryBlockPtr); */
     tm_set_current_context(contextHandle);
}

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    ttUser32Bit cpsr;

    tm_get_current_context(contextHandle);
  
  
/* Make sure that interrupts are disabled */
    cpsr = _Read_CPSR;
    _Write_CPSR(cpsr | _IRQ_DISABLE | _FIQ_DISABLE);

    tm_set_current_context(contextHandle);
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical()
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    ttUser32Bit cpsr;
    tm_get_current_context(contextHandle);
    /*
    Example:
    kernelEnableInterrupt();
    */
/* Make sure that interrupts are enabled */
    cpsr = _Read_CPSR;
    cpsr &= ~(_IRQ_DISABLE | _FIQ_DISABLE);
    _Write_CPSR(cpsr);

    tm_set_current_context(contextHandle);
}


/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    int kernelSemaphore;
    int errorCode;

    tm_get_current_context(contextHandle);
    kernelSemaphore=kernelCreateSemaphore();
    tm_set_current_context(contextHandle);
    if(kernelSemaphore == -1)
    {
        errorCode = TM_KERN_ERROR;
    }
    else
    {
        countingSemaphore->genIntParm=kernelSemaphore;
        errorCode = TM_KERN_OKAY;
    }
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(countingSemaphore);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    int     errorCode;
    tm_get_current_context(contextHandle);
    if (kernelPend(countingSemaphore->genIntParm)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(countingSemaphore);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to a timeout value (in ticks)
 */
int tfKernelPendCountSemTimeout( ttUserGenericUnionPtr countingSemaphore, 
                                 ttUser32Bit tickTimeout )
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    int     errorCode;
    tm_get_current_context(contextHandle);
    if (kernelPendTimeout(countingSemaphore->genIntParm, tickTimeout)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_ETIMEDOUT;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(countingSemaphore);
    TM_UNREF_IN_ARG(tickTimeout);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    int errorCode;
    tm_get_current_context(contextHandle);
    if (kernelPost(countingSemaphore->genIntParm)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(countingSemaphore);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    tm_get_current_context(contextHandle);
    if (kernelDeleteSemaphore(countingSemaphore->genIntParm)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(countingSemaphore);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}


void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)printf("tfKernelError! %s: %s\n", functionName, errorMessage);
    tm_set_current_context(contextHandle);
    tm_thread_stop;
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)printf("tfKernelWarning! %s: %s\n", functionName, errorMessage);
    tm_set_current_context(contextHandle);
}

void  tfKernelInitialize(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    tm_set_current_context(contextHandle);
}

/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(eventPtr);
    tm_set_current_context(contextHandle);
}

/*
 * Task wait on an Event
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(eventPtr);
    tm_set_current_context(contextHandle);
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Task wait on an Event up to a timeout value (in ticks).
 */
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                               ttUser32Bit tickTimeout )
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    /*
    Example:
    int     errorCode;
    tm_get_current_context(contextHandle);
    if (kernelEventPendTimeout(eventPtr->genIntParm, timeout)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(eventPtr);
    TM_UNREF_IN_ARG(tickTimeout);
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(eventPtr);
    tm_set_current_context(contextHandle);
}

/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    TM_UNREF_IN_ARG(eventPtr);
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    tm_set_current_context(contextHandle);
    return;
}

/* 
 * Treck Stack calls this fucntion to ask for the system time. It is 
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network, 
 * or from application user interface.
 * 
 * Parameters:
 *  secondsPtr: 
 *      Pointer to store number of seconds elapsed since midnight (00:00:00), 
 *      January 1, 1970, Coordinated Universal Time (UTC). 
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT). 
 *      This matches the return value of the ANSI C library 
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year. 
 *      In case this value ever wrap arounds, we need another space to hold the 
 *      a bigger units of time. This is the purpose of daysPtr. 
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr  
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *  
 */

int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    int errorCode; 
  
    errorCode = TM_KERN_OKAY;
    *daysPtr = 0;
    *secondsPtr = time(NULL);

    return errorCode;
} 

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trkernel.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 21620    Output checksum: 64756
 */
