/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated. 
 * Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for Quadros
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trquadro.c
 * Author: Odile 
 * Date Created: 05/11/04
 * $Source: source/kernel/quadros/motorola/coldfire/metrowerks/trquadro.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:37CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include Quadros Stuff 
 */
#include "rtxcapi.h"
#include "rtxcstdio.h"
#include "enable.h"
#include "kpart.h" /* QNSHEAP */
#include <string.h>
#include <extras.h>

/*
 * Include Treck Stuff
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trglobal.h>


#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1

int debugAlloc = 0;

//static unsigned long count = 0; /* holds the semaphore creation count */

void tfKernelInitialize()
{
}

/*
 * Yields the CPU to other tasks at the same or higher priority
 */
void tfKernelTaskYield(void)
{
    KS_YieldTask();
}


/* 
 * Report an Error
 */
void tfKernelError(
    const char TM_FAR *functionName, const char TM_FAR *errorMessage)
{
    rtxc_printf("fatal error: %s --- %s\n",functionName, errorMessage); 
    tm_thread_stop;
}

void tfKernelWarning(
    const char TM_FAR *functionName, const char TM_FAR *warnMessage)
{
    rtxc_printf("warning: %s --- %s\n",functionName, warnMessage); 
}

/*
 * Allocate a memory block of size from the kernel
 */

void TM_FAR *tfKernelMalloc(unsigned size)
{
    return tfSheapMalloc(size);
}

ttUser32BitPtr tfKernelSheapCreate(ttUser32BitPtr pageSizePtr)
{
    PARTPROP partProp;

    partProp.size = 0;
    KS_GetPartProp(QNSHEAP, &partProp);
    *pageSizePtr = (ttUser32Bit)partProp.size;
    return((ttUser32BitPtr)KS_AllocBlk(QNSHEAP));
}

/*
 * Free the memory back to the kernel
 */

void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    (void)tfSheapFree(memoryBlockPtr);
}

#ifndef TM_GHSARM7_ASM_CRITICAL
/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    DISABLE;
}

/*
 * Resume normal operation (from tfKernelSetCritical)
 */
void tfKernelReleaseCritical()
{
    ENABLE;
}
#endif /* TM_GHSARM7_ASM_CRITICAL */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    KSRC        status;
    SEMAPROP    sprop;
    int         retCode = TM_KERN_OKAY;
    //char semaName [20];
    //char semaCount[10];

    //strcpy(&semaName[0], "TRECK");        /* setup the semaphore name */
    //itoa(count, &semaCount[0], 10);       /* convert the count to a string */
    //strcat(&semaName[0], &semaCount[0]);  /* concatenate the sema name and count */

    status = KS_OpenSema((char *)0, (SEMA *)&countingSemaphore->genIntParm);
    if(status != RC_GOOD)
    {
        tfKernelError("tfKernelCreateCountSem",
                      "Unable to Create Quadros Semaphore");
        retCode = TM_KERN_ERROR;
        tm_thread_stop;
    }

    sprop.attributes = (0
      & ~ATTR_FIFO_ORDER         /* Waiting Order by task priority */
      & ~ATTR_MULTIPLE_WAITERS); /* Signal type "single" */
    KS_DefSemaProp((SEMA)countingSemaphore->genIntParm, &sprop); /* IT-ID: 12 */

    //count++;                              /* incremen the semaphore creation count */

    return(retCode);
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int     retCode;

    retCode = TM_KERN_OKAY;

    KS_TestSemaW((SEMA)countingSemaphore->genIntParm);
    return(retCode);
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to tickTimeout (in ticks)
 */
int tfKernelPendCountSemTimeout(ttUserGenericUnionPtr countingSemaphore, 
                                ttUser32Bit           tickTimeout)
{
    KSRC          status;
    int           retCode;

    retCode = TM_KERN_OKAY;

    tickTimeout = (tickTimeout * tvTimerTickLength) / CLKTICK;
    retCode = TM_KERN_OKAY;
    status = KS_TestSemaT((SEMA)countingSemaphore->genIntParm, (TICKS)tickTimeout);
    if (status != RC_GOOD)
    {
        if (status == RC_TICKOUT)
        {
            retCode = TM_ETIMEDOUT;
        }
        else
        {
            retCode = TM_KERN_ERROR;
        }
    }
    return retCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int     retCode;

    retCode = TM_KERN_OKAY;

    KS_SignalSema((SEMA)countingSemaphore->genIntParm);
    return retCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    KSRC    status;
    int     retCode;

    retCode = TM_KERN_OKAY;
    status = KS_CloseSema((SEMA)countingSemaphore->genIntParm);
    if (status != RC_GOOD)
    {
        tfKernelError("tfKernelDeleteCountSem",
                      "Unable to Delete a Quadros Semaphore");
        tm_thread_stop;
    }
    return retCode;
}

/*
 * Install an ISR Handler 
 */
void tfKernelInstallIsrHandler(ttUserIsrHandlerPtr funcPtr, 
                               unsigned long offset)
{
    setvect((int)offset, funcPtr);
    return;
}


/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    KSRC        status;
    SEMAPROP    sprop;
    //int  retCode = TM_KERN_OKAY;
    //char semaName [20];
    //char semaCount[10];

    //strcpy(&semaName[0], "TRECK");        /* setup the semaphore name */
    //itoa(count, &semaCount[0], 10);       /* convert the count to a string */
    //strcat(&semaName[0], &semaCount[0]);  /* concatenate the sema name and count */

    status = KS_OpenSema((char *)0, (SEMA *)&eventPtr->genIntParm);
    if(status != RC_GOOD)
    {
        tfKernelError("tfKernelCreateEvent",
                      "Unable to Create Quadros Semaphore");
        tm_thread_stop;
    }

    sprop.attributes = (0
      & ~ATTR_FIFO_ORDER         /* Waiting Order by task priority */
      & ~ATTR_MULTIPLE_WAITERS); /* Signal type "single" */
    KS_DefSemaProp((SEMA)eventPtr->genIntParm, &sprop); /* IT-ID: 12 */

    //count++;                              /* increment the name counter */
}

/*
 * Wait on an Event from a Task
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    KS_TestSemaW((SEMA)eventPtr->genIntParm);
}

/*
 * Resume waiting tasks that are waiting on this event
 * Called from an ISR
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    IS_SignalSema((SEMA)eventPtr->genIntParm);
}

/*
 * Resume waiting tasks that are waiting on this event
 * Called from a Task
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    KS_SignalSema((SEMA)eventPtr->genIntParm);
}

/* 
 * Treck Stack calls this fucntion to ask for the system time. It is 
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network, 
 * or from application user interface.
 * 
 * Parameters:
 *  secondsPtr: 
 *      Pointer to store number of seconds elapsed since midnight (00:00:00), 
 *      January 1, 1970, Coordinated Universal Time (UTC). 
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT). 
 *      This matches the return value of the ANSI C library 
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year. 
 *      In case this value ever wrap arounds, we need another space to hold the 
 *      a bigger units of time. This is the purpose of daysPtr. 
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr  
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *  
 */
int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    int errorCode; 
  
    errorCode = TM_KERN_OKAY;
    *daysPtr = 0;
/*    *secondsPtr = time(NULL); */
    *secondsPtr = tvTime/1000; /* NOTE: this is not the correct time */

    return errorCode;
} 

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trquadro.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 31401    Output checksum: 63816
 */
