/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: RTXC32 kernel interface
 * Filename: trrtxc.c
 * Author: Odile
 * Date Created: 07/26/04
 * $Source: source/kernel/rtxc32/trrtxc.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:39CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Kernel interface for RTXC32 and
 * Counting semaphore implementation for RTXC32 which does not support this
 * concept. We support up to TM_NUMCOUNTSEM counting semaphores. For
 * OS primitive, we use binary semaphores, since RTXC32 supports them.
 * Each of the TRECK task will have an event associated with it.
 * When a task creates a counting semaphore, we reserve a space in our
 * array of counting semaphores, and initialize the counting semaphore to
 * zero. When a task tries and pends on the counting semaphore, if the
 * counting semaphore is non zero, we decrease it and grant the access
 * to the caller, otherwise (counting semaphore is zero), we will queue
 * that task in the task queue associated with the counting semaphore,
 * ordered by priority order, and then we will call tfKernelPendEvent()
 * passing a pointer to the binary semaphore associated with the task, and
 * the task will pend.
 * When a post on a counting semaphore occurs, if there is any task
 * in the task queue associated with the counting semaphore (i.e a pending
 * task), we go ahead and remove the first task in the task queue (and only
 * the first task), retrieve the associated binary semaphore, and post on the
 * associated binary semaphore by calling tfKernelTaskPostEvent() passing a
 * pointer to the binary semaphore associated with the task, and the task
 * will resume. Otherwise, no task is pending, we just go ahead and increase
 * the counting semaphore.
 *
 * This version assumes that counting sempahores are created, but
 * never deleted, which is the case with the counting sempahores used
 * by the Treck stack, since the Treck stack recycles the created counting
 * semaphores internally.
 *
 *========================================================================
 * I. Configuration/macro definitions needed at compile time:
 * I.1 Number of tasks making calls to the Treck task. This includes
 *       the Treck timer task, Treck receive task, (possibly the Treck
 *       send complete task if any, the Treck transmit task if any), and
 *       any user application making socket API calls.
 *       Default is 12. 
 *
 *   #define TM_NUMTSK 12
 *
 *   Modify this value in the trrtxc.c module if needed.
 *
 * I.2 binary semaphores to be generated by the RTXC32 tools.
 *
 *  The RTXC32 tools should generate contiguous binary semaphores to be used
 *  by the Treck stack, and to be found in RTXC32 csema.h.
 *  The tool should generate contiguous semaphores for the Treck stack.
 *  The number of semaphores needed == number of Treck tasks (TM_NUMTSK) + up to
 *  3 additional ones per interface, depending on what task macro is defined;
 *  i.e. one for the recv task (if TM_TASK_RECV is defined), one for the 
 *  send complete task (if TM_TASK_SEND is defined), and one
 *  for the transmit task (it TM_TASK_XMIT is defined.)
 *  The RTXC32 kernel interface code assumes that there are 15 binary semaphores
 *  reserved for the Treck stack, and that their values are given by the
 *  macro definitions TRXSM00, ..., TRXSM14
 *
 *  Modify the name of these macros in the trrtxc.c module if needed, so that
 *  they match the ones in csema.h. We need to use the first one, and the 
 *  last one.
 *
 * II. Calls needed to be made by the user at run time
 *
 * II.1. The user need to call tfTaskInit() prior to running any task, and
 *       prior to calling tfStartTreck() (and tfInitTreckMultipleContext()
 *       if using multiple contexts).
 *
 * Example code:
 *
 * tfTaskInit();
 *
 * II.2. At the beginning of each Treck task, and before the Treck task makes
 *      any call to the stack, the user needs to call
 *      tfTaskRegister() passing the Task ID pointer as initialized
 *      by the call to tfKernelGetCurrentTaskId implemented in trrtxc.c, and 
 *      passing the task priority. There we store the TASK ID, priority 
 *      mapping and try and create a binary semaphore that will be associated
 *      with the TASK ID. The task index of the task is returned. A task index 
 *      of TM_NUMTSK or bigger indicates failure (overflow).
 *
 * Example code to be inserted at the beginning of the task:
 *    
 * #include <trsocket.h>
 * ttUserGenericUnionPtr taskIdPtr;
 * int                   taskPri;
 *
 * Initialize taskPri, with the priority of the task:
 *    taskPri = ??;
 *    tfKernelGetCurrentTaskId(taskIdPtr);
 *    tfTaskRegister(taskIdPtr, taskPri);
 *
 *========================================================================
 * Priority macro:
 *      tm_task_priority_lower_same(taskAPri, taskBPri)
 *      RTXC32 task priorities are descending(i.e a lower priority task has a
 *      higher priority value), so the macro is defined as follows:
 *
 *  #define tm_task_priority_lower_same(taskAPri, taskBPri)  \
 *                                       ((taskAPri) >= (taskBPri))
 *
 * TaskID Data type. RTXC32's TASK ID is stored as an integer. 
 *      So taskIdPtr->genIntParam, and tskIdUnion.genIntParm are used.
 *      If this is no longer the case replace taskIdPtr->genInParm, and
 *      tskIdUnion.genIntParm, with the appropriate data type (as defined
 *      in ttUserGenericUnion), in tfKernelGetCurrentTaskId, and
 *      tfTaskIdToIndex().
 *
 * Functions/procedures needed for the counting sempahore implementation
 *    and provided here:
 *
 *   1 int tfKernelGetCurrentTaskId(ttUserGenericUnionPtr taskIdPtr)
 *       Get the TASK ID of the running task. Store it in taskIdPtr with
 *       the proper data type, and return TM_KERN_OKAY.
 *
 *   2 void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
 *        Wait on a binary semaphore (pointed to by eventPtr) posted from
 *        another task. This function is also called directly by the stack for
 *        event synchronizations between the device driver and the
 *        recv task, and send task (if used), and possibly transmit task
 *        (if used).
 *
 *   3 void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
 *         Resume waiting tasks that ware waiting on the binary sempahore
 *         pointed to by eventPtr.
 *         This function is also called directly by the stack for
 *         event synchronizations between the device driver and the
 *         recv task, and send task (if used), and possibly transmit task
 *         (if used)
 *         This is similar to tfKernelPostIsrEvent(),
 *         except that this will be posted from a task, and not from
 *         an ISR, since RTXC32 has different calls when a Post occurs
 *         from an ISR or occurs from a task.
 */


#include <trsocket.h>
#include <trmacro.h>

#ifndef TM_USE_KERNEL_CRITICAL_STATUS
//#error You need to define TM_USE_KERNEL_CRITICAL_STATUS in trsystem.h
#endif /* TM_USE_KERNEL_CRITICAL_STATUS */

/*
 * Number of tasks making calls to the Treck task. This includes
 * the Treck timer task, Treck receive task, (possibly the Treck
 * send complete task if any, Treck transmit task if any), and any
 * user application making socket API calls.
 * Default is 12.
 */
#define TM_NUMTSK            12

/* Treck does not use more counting semaphore than Treck tasks */
#define TM_NUMCOUNTSEM       TM_NUMTSK

#include "rtxcapi.h"
#include "csema.h"
#include "enable.h"

typedef int SEMA;
typedef int SSTATE;

/* Static Name Array */

/* Generated by RTXC32 tools, and to be found in csema.h */
/* The tool should generate contiguous semaphores for the Treck stack.
 * The number of semaphores needed == number of treck tasks (TM_NUMTSK) + up to
 * 3 additional ones per interface, depending on what task macro is defined;
 * i.e. one for the recv task (if TM_TASK_RECV is defined), one for the 
 * send complete task (if TM_TASK_SEND is defined), and one
 * for the transmit task (it TM_TASK_XMIT is defined.)
 */
/*
 * These macros are commented out, because they should be defined in 
 * csema.h
 */
//#define TRXSM00    1 
//#define TRXSM14    15
/* END of Generated by RTXC32 tools */
#ifndef TRXSM00
#error You need to generate binary semaphores whose definitions are in csema.h
#endif /* TRXSM00 */

#define TM_TRECK_SEMA_START TRXSM00
#define TM_TRECK_SEMA_LAST  TRXSM14

/*
 * Local macros
 */

#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1

#define TM_INVALID  (unsigned char)0
#define TM_VALID    (unsigned char)1

/*
 * Task A has lower (or same) priority than Task B.
 * The following definition for RTXC32 (using bigger or equal sign ) assumes
 * that priorities are descending (i.e a a lower priority task has a higher
 * priority value).
 */
#define tm_task_priority_lower_same(taskAPri, taskBPri)  \
                                        ((taskAPri) >= (taskBPri))

/*
 * Local types
 */

/*
 * Task entry structure. We keep track of TASK ID, TASK Priority
 * and event associated with the task. The next pointer field is used to
 * link tasks waiting on a counting semaphores.
 */
typedef struct tsTaskEntry
{
/*
 * Pointer to next task waiting on counting semaphore.
 * Has to be first entry
 */
    struct tsTaskEntry      TM_FAR * tskNextPtr;
/* event to pend on */
    ttUserGenericUnion               tskEventUnion;
/* Task ID */
    ttUserGenericUnion               tskIdUnion;
/* Task priority */
    int                              tskPri;
    unsigned char                    tskEventValid;
    unsigned char                    tskRegistered;
    unsigned char                    tskFiller[2];
} ttTaskEntry;

typedef ttTaskEntry TM_FAR * ttTaskEntryPtr;

/*
 * Counting semaphore data structure. We keep track of the counting
 * semaphore count, and the queue of tasks waiting on the counting
 * semaphore.
 */

typedef struct tsCountSemEntry
{
/*
 * Pointer to queue of tasks waiting on counting semaphore.
 * Has to be first entry
 */
    ttTaskEntryPtr       csTaskQueuePtr;
/* count value of counting semaphore */
    int                  csCount;
} ttCountSemEntry;

typedef ttCountSemEntry TM_FAR * ttCountSemEntryPtr;

typedef void           TM_FAR  * ttVoidPtr;

/*
 * Local variables
 */

/* Table of tasks. */
ttTaskEntry        tlTaskArray[TM_NUMTSK];

/* Table of counting semaphores. Counting semaphore ID is index */
ttCountSemEntry    tlCountSemArray[TM_NUMCOUNTSEM];

/*
 * Counting Sempahore index (starting search of next available counting
 * sempahore ID)
 */
unsigned int       tlCountSemIndex;

/*
 * Task index (starting search of next available task slot)
 */
unsigned int       tlTaskIndex;

void tfTaskInit (void);
int tfTaskRegister (ttUserGenericUnionPtr taskIdPtr, int taskPri);

static int tfKernelGetCurrentTaskId(ttUserGenericUnionPtr taskIdPtr);

static int  tfTaskIdToIndex (ttUserGenericUnionPtr taskIdPtr);

int tlRtxcSemaIndex;

void  tfKernelInitialize(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
/* Empty */
    tm_set_current_context(contextHandle);
}

/*
 * Get the TASK ID of the currently running task.
 * Store task ID of current task in one of the field of the union pointed
 * to by taskIdPtr
 */
static int tfKernelGetCurrentTaskId(ttUserGenericUnionPtr taskIdPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    taskIdPtr->genIntParm = KS_inqtask();
    tm_set_current_context(contextHandle);
    return TM_KERN_OKAY;
}

/*
 * Wait on an Event posted from another task
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    TM_UNREF_IN_ARG(eventPtr);
    KS_wait(eventPtr->genIntParm);
    tm_set_current_context(contextHandle);
    return;
}

#ifdef TM_USE_PEND_TIMEOUT
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                              ttUser32Bit tickTimeout )
{
    int retCode;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);

    retCode = KS_waitt(eventPtr->genIntParm, tickTimeout);
    if (retCode == RC_GOOD)
    {
        retCode = TM_KERN_OKAY;
    }
    else
    {
        retCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return retCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks that ware waiting on this event
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    KS_signal((SEMA)eventPtr->genIntParm);
    tm_set_current_context(contextHandle);
    return;
}

/*
 * tfTaskInit need to be called prior to starting any Treck task and
 * prior to calling tfStartTreck()
 */
void tfTaskInit (void)
{
    tm_bzero(&tlTaskArray[0], sizeof(tlTaskArray));
    tm_bzero(&tlCountSemArray[0], sizeof(tlCountSemArray));
    tlCountSemIndex = 0;
    tlTaskIndex = 0;
    tlRtxcSemaIndex = TM_TRECK_SEMA_START;
}

/*
 * tfTaskRegister need to be called after starting each task.
 */
int tfTaskRegister (ttUserGenericUnionPtr taskIdPtr, int taskPri)
{
    int taskIndex;

    taskIndex = tfTaskIdToIndex(taskIdPtr);
    if (taskIndex < (unsigned int)TM_NUMTSK)
    {
        tlTaskArray[taskIndex].tskPri = taskPri;
    }
    return taskIndex;
}

/*
 * Find the slot corresponding to task ID. If no slot is found for taskId,
 * allocate one (if there is room), store the Task ID and create an
 * event that we store.
 */
static int  tfTaskIdToIndex (ttUserGenericUnionPtr taskIdPtr)
{
    ttTaskEntryPtr   taskPtr;
    unsigned int     taskIndex;

    tm_kernel_set_critical;
/* Search for taskId slot */
    for ( taskIndex = 0; taskIndex < tlTaskIndex; taskIndex++ )
    {
        if (    ( tlTaskArray[taskIndex].tskRegistered != TM_INVALID )
             && (    (    taskIdPtr->genIntParm
                       == tlTaskArray[taskIndex].tskIdUnion.genIntParm ) ) )
        {
            break;
        }
    }
    if (    (taskIndex < (unsigned int)TM_NUMTSK) /* room */
        &&  (taskIndex == tlTaskIndex) ) /* no slot allocated yet */
/*
 * If there is room for a new slot, and the task is not already in the task
 * array, store the task ID, and reserve the new slot for this task
 */
    {
        tlTaskArray[taskIndex].tskIdUnion.genIntParm =
                                                      taskIdPtr->genIntParm;
        tlTaskArray[taskIndex].tskRegistered = TM_VALID;
        tlTaskIndex++;
    }
    tm_kernel_release_critical;
/* Create associated event if not already done so */
    if (taskIndex < (unsigned int)TM_NUMTSK)
    {
        taskPtr = &tlTaskArray[taskIndex];
        if ((taskPtr->tskEventValid == TM_INVALID))
        {
/* create an event if one has not been created yet */
            tfKernelCreateEvent(&taskPtr->tskEventUnion);
            taskPtr->tskEventValid = TM_VALID;
        }
    }
#ifdef TM_ERROR_CHECKING
    else
    {
        tfKernelError( "tfTaskIdToIndex",
                       "More tasks started than available slots");
    }
#endif /* TM_ERROR_CHECKING */
    return taskIndex;
}
/*
 * Counting Semaphore Operations
 */

/*
 * Create a counting semaphore. Search for a free (not in use) slot in
 * our table of counting semaphores. This version assumes that the user
 * will never delete a counting semaphore, which is the case with
 * Treck tasks.
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphorePtr)
{
    int             retCode;

    retCode = TM_KERN_ERROR;
    tm_kernel_set_critical;
    if (tlCountSemIndex < (unsigned int)TM_NUMCOUNTSEM)
    {
        countingSemaphorePtr->genIntParm = (int)tlCountSemIndex++;
        retCode = TM_KERN_OKAY;
    }
    tm_kernel_release_critical;
    return retCode;
}

/*
 * Wait on a counting semaphore. If counting semaphore is zero, we
 * queue our task to the counting semaphore, and wait on the task
 * event. If counting semaphore is non zero, we get the semaphore
 * (decrease it by one), and return.
 */
int tfKernelPendCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
#ifdef TM_USE_PEND_TIMEOUT
{
    return tfKernelPendCountSemTimeout(countingSemaphorePtr, TM_UL(0));
}

/*
 * Wait on a counting semaphore. If counting semaphore is zero, we
 * queue our task to the counting semaphore, and wait on the task
 * event. If counting semaphore is non zero, we get the semaphore
 * (decrease it by one), and return.
 */
int tfKernelPendCountSemTimeout ( ttUserGenericUnionPtr countingSemaphorePtr, 
                                  ttUser32Bit tickTimeout )
#endif /* TM_USE_PEND_TIMEOUT */
{
    ttCountSemEntryPtr countSemPtr;
    ttTaskEntryPtr     taskPtr;
    ttTaskEntryPtr     taskQueuePtr;
    ttTaskEntryPtr     prevTaskQueuePtr;
    ttUserGenericUnion curTaskId;
    unsigned int       taskIndex;
    unsigned int       countSemId;
    int                priority;
    int                retCode;

    retCode = TM_KERN_ERROR;
/* Get current task ID from the OS */
    retCode = tfKernelGetCurrentTaskId(&curTaskId);
    if (retCode == TM_KERN_OKAY)
    {
        retCode = TM_KERN_ERROR;
/* Map from current task ID to our Task Array Pointer */
        taskIndex = tfTaskIdToIndex(&curTaskId);
        if (taskIndex < (unsigned int)TM_NUMTSK)
        {
/* Counting semaphore ID */
            countSemId = (unsigned int)countingSemaphorePtr->genIntParm;
            if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
            {
                countSemPtr = &tlCountSemArray[countSemId];
                taskPtr = &tlTaskArray[taskIndex];
                tm_kernel_set_critical;
                if (countSemPtr->csCount > 0)
                {
/* Decrease the counting semaphore. We are granted access */
                    countSemPtr->csCount--;
                    tm_kernel_release_critical;
                    retCode = TM_KERN_OKAY;
                }
                else
                {
/*
 * We need to pend. We will insert our task into the counting semaphore
 * queue in priority order.
 */
                    taskQueuePtr = countSemPtr->csTaskQueuePtr;
                    if ( taskQueuePtr == (ttTaskEntryPtr)0 )
                    {
/* Empty counting semaphore task queue */
                        countSemPtr->csTaskQueuePtr = taskPtr;
                        taskPtr->tskNextPtr = (ttTaskEntryPtr)0;
                    }
                    else
                    {
                        priority = taskPtr->tskPri;
                        prevTaskQueuePtr = (ttTaskEntryPtr)countSemPtr;
                        while (    (taskQueuePtr != (ttTaskEntryPtr)0)
                                && (tm_task_priority_lower_same(
                                                    priority,
                                                    taskQueuePtr->tskPri)) )
                        {
                            prevTaskQueuePtr = taskQueuePtr;
                            taskQueuePtr = taskQueuePtr->tskNextPtr;
                        }
/* Insert in counting semaphore task queue */
                        taskPtr->tskNextPtr = prevTaskQueuePtr->tskNextPtr;
                        prevTaskQueuePtr->tskNextPtr = taskPtr;
                    }
                    tm_kernel_release_critical;
#ifdef TM_USE_PEND_TIMEOUT
                    if (tickTimeout != TM_UL(0))
                    {
                        retCode = tfKernelPendEventTimeout(
                                    &taskPtr->tskEventUnion, tickTimeout);
                        if (retCode != TM_KERN_OKAY)
/* Timeout occured. Remove counting semaphore from the task queue */
                        {
                            retCode = TM_ETIMEDOUT;
                            tm_kernel_set_critical;
                            taskQueuePtr = countSemPtr->csTaskQueuePtr;
                            if (taskQueuePtr == taskPtr) 
                            {
/* At head of queue, special case: head of queue point to next entry */
                                countSemPtr->csTaskQueuePtr = 
                                                    taskPtr->tskNextPtr; 
                            }
                            else
                            {
                                while (taskQueuePtr != (ttTaskEntryPtr)0)
                                {
                                    prevTaskQueuePtr = taskQueuePtr;
                                    taskQueuePtr = taskQueuePtr->tskNextPtr;
                                    if (taskQueuePtr == taskPtr) 
                                    {
/* Remove from queue, using previous pointer */
                                        prevTaskQueuePtr->tskNextPtr = 
                                                          taskPtr->tskNextPtr;
                                        break;
                                    }
                                }
                            }
                            tm_kernel_release_critical;
                        }
                    }
                    else
#endif /* TM_USE_PEND_TIMEOUT */
                    {
                        tfKernelPendEvent(&taskPtr->tskEventUnion);
                        retCode = TM_KERN_OKAY;
                    }
                }
            }
#ifdef TM_ERROR_CHECKING
            else
            {
/* Should not happen */
                tfKernelError("tfKernelPendCountSem",
                              "Invalid Counting Semaphore");
            }
#endif /* TM_ERROR_CHECKING */
        }
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPendCountSem",
                      "Invalid current task ID");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Resume first task waiting on a counting semaphore
 */
int tfKernelPostCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    ttCountSemEntryPtr  countSemPtr;
    ttTaskEntryPtr      taskQueuePtr;
    unsigned int        countSemId;
    int                 retCode;

    retCode = TM_KERN_ERROR;
    countSemId = (unsigned int)countingSemaphorePtr->genIntParm;
    if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
    {
        countSemPtr = &tlCountSemArray[countSemId];
        tm_kernel_set_critical;
        taskQueuePtr = countSemPtr->csTaskQueuePtr;
/* If no task is waiting on the semaphore yet. Just increase the count */
        if (taskQueuePtr == (ttTaskEntryPtr)0)
        {
            countSemPtr->csCount++;
            tm_kernel_release_critical;
        }
        else
/* Remove and Reschedule task at head of the queue (highest priority) */
        {
/* Remove task at head of the queue */
            countSemPtr->csTaskQueuePtr = taskQueuePtr->tskNextPtr;
            tm_kernel_release_critical;
/* Move task to run queue */
            tfKernelTaskPostEvent(&taskQueuePtr->tskEventUnion);
        }
        retCode = TM_KERN_OKAY;
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPostCountSem", "Invalid Counting Semaphore");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Delete a counting semaphore
 * Unused function by TRECK since semaphores are recycled internally
 */
int tfKernelDeleteCountSem (ttUserGenericUnionPtr countingSemaphorePtr)
{
    int             retCode;

    TM_UNREF_IN_ARG(countingSemaphorePtr);
    retCode = TM_KERN_ERROR;
    return retCode;
}

void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    eventPtr->genIntParm = tlRtxcSemaIndex;
    if (tlRtxcSemaIndex > TM_TRECK_SEMA_LAST)
    {
        tfKernelError("tfKernelCreateEvent", "Ran out of RTXC semaphores\n");
    }
    tlRtxcSemaIndex++;
}
/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    KS_ISRsignal((SEMA)eventPtr->genIntParm);
    tm_set_current_context(contextHandle);
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    KS_yield();
    tm_set_current_context(contextHandle);
    return;
}

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    tm_set_current_context(contextHandle);
    tm_thread_stop;
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    tm_set_current_context(contextHandle);
}

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    KDISABLE;
//    PUSHSR_KDISABLE(tvKernelCriticalStatus); /* save current SR */
    tm_set_current_context(contextHandle);
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    KENABLE;
//    POSPSR(tvKernelCriticalStatus);
    tm_set_current_context(contextHandle);
}

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    return tfSheapMalloc(size);
}

/*
 * Free the memory back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    (void)tfSheapFree(memoryBlockPtr);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trrtxc.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 18285    Output checksum: 14374
 */
