/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks
 *              NOTE: This File defines the hooks to RZK RTOS.
 *
 * Filename: trrzk.c
 * Author: Qin Zhang
 * Date Created: 8/5/02
 * $Source: source/kernel/rzk/ez80/trrzk.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:41CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Include files */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trglobal.h>
#include "ZTypes.h"
#include "ZSemaphore.h"
#include "ZEventGroup.h"
#include "ZThread.h"
#include "ZInterrupt.h"
//#include "ZRegion.h"

#define TM_INIT_SEMAPHORE 		0
//#define TM_USE_TRECK_SEMAPHORE

extern RZK_STATUS_t       RZKGetErrorNum();

/*
 * Local macros 
 */
/* RZK Event Mask */
#define TM_RZK_EVENT_MASK   0x01

/* Mininum semaphore created in initialization */
#define TM_RZK_MIN_SEMAPHORE    5

/* Semaphore nitialization parameters */
#define TM_SEMAPHORE_MAX          10
#define TM_SEMAPHORE_INI          1
#define TM_SEMAPHORE_NUMTHREAD    0

#ifdef TM_USE_TRECK_SEMAPHORE
extern void tfTaskInit(void);
#endif /* TM_USE_TRECK_SEMAPHORE */

/* 
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
#ifdef TM_DYNAMIC_CREATE_SHEAP 
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    return (ttUser32Bit TM_FAR *)(malloc((*sheapSizePtr)));
}
#endif TM_DYNAMIC_CREATE_SHEAP 

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void TM_FAR * memoryPtr;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
/*    return (void TM_FAR *)(malloc((unsigned)size)); */
#ifdef TM_USE_SHEAP
    memoryPtr = (void TM_FAR *)tfSheapMalloc(size);
#else /* !TM_USE_SHEAP */
    memoryPtr = (void TM_FAR *)(void TM_FAR *)(malloc((unsigned)size));
#endif /* !TM_USE_SHEAP */
    tm_set_current_context(contextHandle);
    return memoryPtr;
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
/*
 * NOTE 'C' library free() is not re-entrant. If free() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
/*   free(memoryBlockPtr); */
    tm_get_current_context(contextHandle);
#ifdef TM_USE_SHEAP
	tfSheapFree(memoryBlockPtr);
#else /* !TM_USE_SHEAP */
    free(memoryBlockPtr);
#endif /* !TM_USE_SHEAP */
    tm_set_current_context(contextHandle);
	return;
}

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    /*
    Example:
    kernelDisableInterrupt();
    */
    tm_set_current_context(contextHandle);
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical()
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    /*
    Example:
    kernelEnableInterrupt();
    */
    tm_set_current_context(contextHandle);
}

#ifndef TM_USE_TRECK_SEMAPHORE
/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */
/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int             errorCode; 
    RZK_STATUS_t    rzkStatus;
    RZK_SEMAPHOREPARAMS_t semParam;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    countingSemaphore->genVoidParmPtr = RZKCreateSemaphore (
                                            (unsigned char *)"TRECK",
                                            TM_SEMAPHORE_MAX,
                                            RECV_ORDER_FIFO );
    if(countingSemaphore->genVoidParmPtr == NULL)
    {
        errorCode = TM_KERN_ERROR;
	    rzkStatus = RZKGetErrorNum();
    }
    else
    {
        semParam.uInitialCount = TM_SEMAPHORE_MAX;
        semParam.uDynamicCount = TM_SEMAPHORE_INI;
        semParam.nNumThreads = TM_SEMAPHORE_NUMTHREAD;

        rzkStatus = RZKSetSemaphoreParameters(countingSemaphore->genVoidParmPtr,
                                     &semParam);
        if (rzkStatus == RZKERR_SUCCESS)
        {
            errorCode = TM_KERN_OKAY;
            rzkStatus = RZKAcquireSemaphore( countingSemaphore->genVoidParmPtr,
                                             INFINITE_SUSPEND);
            if (rzkStatus == RZKERR_SUCCESS)
            {
                errorCode = TM_KERN_OKAY;
            }
            else
            {
                errorCode = TM_KERN_ERROR;
            }
        }
        else
        {
            errorCode = TM_KERN_ERROR;
            rzkStatus = RZKGetErrorNum();
        }
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int             errorCode;
    RZK_STATUS_t    rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    rzkStatus = RZKAcquireSemaphore( countingSemaphore->genVoidParmPtr,
                                     INFINITE_SUSPEND);
    if (rzkStatus == RZKERR_SUCCESS)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSemTimeout( ttUserGenericUnionPtr countingSemaphore,
                                 ttUser32Bit           tickTimeout )
{
    int             errorCode;
    RZK_STATUS_t    rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    rzkStatus = RZKAcquireSemaphore( countingSemaphore->genVoidParmPtr,
                                     tickTimeout );
    if (rzkStatus == RZKERR_SUCCESS)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        if (rzkStatus == RZKERR_TIMEOUT)
        {
            errorCode = TM_ETIMEDOUT;
        }
        else
        {
            errorCode = TM_KERN_ERROR;
        }
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;
    RZK_STATUS_t    rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
	rzkStatus = RZKReleaseSemaphore(countingSemaphore->genVoidParmPtr);
    if (rzkStatus==RZKERR_SUCCESS)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    tm_get_current_context(contextHandle);
    if (RZKDeleteSemaphore(countingSemaphore->genVoidParmPtr)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}
#endif /* TM_USE_TRECK_SEMAPHORE */

void  tfKernelError(char TM_FAR *functionName,
                    char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
//    printf("tfKernelError! %s: %s\n", functionName, errorMessage);
    tm_set_current_context(contextHandle);
    tm_thread_stop;
}

void  tfKernelWarning(char TM_FAR *functionName,
                      char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
//    printf("tfKernelWarning! %s: %s\n", functionName, errorMessage);
    tm_set_current_context(contextHandle);
    return;
}

void  tfKernelInitialize(void)
{
    int		        counter;
    ttCountSemPtr   semaphorePtr;
    RZK_NAME_t      semName[MAX_OBJECT_NAME_LEN];
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);

#ifdef TM_USE_TRECK_SEMAPHORE
    tfTaskInit();
#endif /* TM_USE_TRECK_SEMAPHORE */

#if TM_INIT_SEMAPHORE
    semName[0] = 'S';
    semName[1] = 'E';
    semName[2] = 'M';
    semName[4] = '\0';
    for ( counter = 0; counter < TM_RZK_MIN_SEMAPHORE; counter++ )
    {
        semaphorePtr = (ttCountSemPtr)
                       tm_kernel_malloc((unsigned)sizeof(ttCountSem));
        semName[3] = '0' + counter;
        semaphorePtr->couCountSem.genVoidParmPtr = 
                    RZKCreateSemaphore( semName,
                                        1,
                                        RECV_ORDER_PRIORITY );
        semaphorePtr->couNext = tvCountSemListPtr;
        tvCountSemListPtr = semaphorePtr;
    }
#endif
    tm_set_current_context(contextHandle);
    return;
}

/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    RZK_STATUS_t    rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext   contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    eventPtr->genVoidParmPtr = RZKCreateEventGroup( (unsigned char *)"TRECK",
                                                    TM_RZK_EVENT_MASK );
    if(eventPtr->genVoidParmPtr == NULL)
    {
	    rzkStatus = RZKGetErrorNum();
    }
    tm_set_current_context(contextHandle);
    return;
}

/*
 * Task wait on an Event
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)RZKPendOnEventGroup( (RZK_EVENTHANDLE_t)eventPtr->genVoidParmPtr,
                               TM_RZK_EVENT_MASK,
                               INFINITE_SUSPEND,
                               EVENT_CONSUME );
    tm_set_current_context(contextHandle);
    return;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Task wait on an Event
 */
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                              ttUser32Bit tickTimeout )
{
    int            errorCode;
    RZK_STATUS_t   rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */

    errorCode = TM_KERN_OKAY;
    tm_get_current_context(contextHandle);
    rzkStatus = RZKPendOnEventGroup( (RZK_EVENTHANDLE_t)eventPtr->genVoidParmPtr,
                                     TM_RZK_EVENT_MASK,
                                     tickTimeout,
                                     EVENT_CONSUME );
    if (rzkStatus != RZKERR_SUCCESS)
    {
        if (rzkStatus == RZKERR_TIMEOUT)
        {
            errorCode = TM_ETIMEDOUT;
        }
        else
        {
            errorCode = TM_KERN_ERROR;
        }
    }
    tm_set_current_context(contextHandle);
    return errorCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    RZK_STATUS_t    rzkStatus;
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    rzkStatus = RZKPostToEventGroup( (RZK_EVENTHANDLE_t)eventPtr->genVoidParmPtr,
                               TM_RZK_EVENT_MASK,
                               EVENT_OR );
    tm_set_current_context(contextHandle);
    return;
}

/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    (void)RZKPostToEventGroup( (RZK_EVENTHANDLE_t)eventPtr->genVoidParmPtr,
                               TM_RZK_EVENT_MASK,
                               EVENT_OR );
    return;
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)RZKYieldThread();
    tm_set_current_context(contextHandle);
    return;
}


/*
 * Install ISR handler
 */
void tfKernelInstallIsrHandler ( ttUserIsrHandlerPtr funcPtr,
                                 unsigned long       offset )
{
    RZK_PTR_t  isrHandle;

    isrHandle = RZKInstallInterruptHandler(funcPtr, offset);
    if( isrHandle == NULL )
    {
        tfKernelError("tfKernelInstallIsrHandler",
                      "RZK install interrupt handler failed");
    }
    else
    {
        RZKEnableInterrupts( offset );
    }
    return;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trrzk.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 24302    Output checksum: 61520
 */
