/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: eZ80 L92 + RZK demo application
 *
 * Filename: txrzkapp.c
 * Author: Qin Zhang
 * Date Created:
 * $Source: source/kernel/rzk/ez80/txrzkapp.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:43CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h>

#ifdef TM_DEBUG_LOGGING
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#endif /* TM_DEBUG_LOGGING */

#include <trcrystl.h>
#include <trzuart.h>

#ifdef TM_KERNEL_RZK_EZ80
#include "ZSysgen.h"
#include "ZTypes.h"
#include "ZThread.h"
#include "ZMessageQ.h"
#include "ZSemaphore.h"     
#include "ZMemory.h" 
#include "ZMailbox.h"  
#include "ZTimer.h"
#include "ZInterrupt.h"
#include "ZEventgroup.h"
#include "ZSignals.h"
#include "ZDevice.h"
#endif /* TM_KERNEL_RZK_EZ80 */

/*
 * Local Macro Definitions
 */
/*
 * System options
 */
/* 1: Enable FTP server; 0: Disable FTP server */
#define TM_THREAD_FTPD      1
/* 1: Enable UDP echo server; 0: Disable UDP echo server */
#define TM_THREAD_UDPECHO   1
/* 1: Enable TCP echo server; 0: Disable TCP echo server */
#define TM_THREAD_TCPECHO   1
/* 1: Enable PPP demo application; 0: Disable PPP demo application */
#define TM_THREAD_PPP       0

/*
 * Use async interface: if you want to use serial interface, uncomment
 * the following macro
 */
/* #define TM_USE_SERIAL_INTERFACE */

/*
 * Use Ethernet interface: if you do not want to use Ethernet interface,
 * comment out the following macro
 */
#define TM_USE_ETHERNET_INTERFACE

/* Performance test thread */
#define TM_THREAD_PERFTEST      0

/* Ram file system size: one block = 128 Bytes */
#define TM_RAMFS_BLOCKS     256

#if TM_THREAD_PPP
#ifndef TM_USE_SERIAL_INTERFACE
#define TM_USE_SERIAL_INTERFACE
#endif /* TM_USE_SERIAL_INTERFACE */
#endif /* TM_THREAD_PPP */

#ifdef TM_PPP_LQM
/* Parameters controlling determination of link quality */
#define TM_LQM_HYSTERESIS_MAX_FAILURES 3
#define TM_LQM_HYSTERESIS_SAMPLES      4
#endif /* TM_PPP_LQM */

/* IP address & mask */
#ifdef TM_USE_SERIAL_INTERFACE
#define TM_SERIAL_OUR_IP_ADDR       "192.168.200.240"
#define TM_SERIAL_PEER_IP_ADDR      "192.168.200.241"
#endif /* TM_USE_SERIAL_INTERFACE */

#ifdef TM_USE_ETHERNET_INTERFACE
#define TM_ETHER_OUR_IP_ADDR        "192.168.100.10"
#define TM_ETHER_PEER_IP_ADDR       "192.168.100.20"
#endif /* TM_USE_ETHERNET_INTERFACE */

#define TM_NET_MASK                 "255.255.255.0"

/* Echo port define */
#define TM_ECHO_PORT        7
#define TM_DISCARD_PORT     9

/* System internal timer parameters */
#define TM_TIMER_INIDELAY           1
#define TM_TIMER_PERIOD             5

/* Timer thread period */
#define TM_TIMERTHREAD_WAITTICKS	5

/* Stack size definitions for threads */
#define TM_RECVTHREAD_STACKSIZE     1024
#define TM_TIMERTHREAD_STACKSIZE    1536
#define TM_FTPDTHREAD_STACKSIZE     1024
#define TM_UDPECHOTHREAD_STACKSIZE  1024
#define TM_TCPECHOTHREAD_STACKSIZE  1024
#define TM_PPPTHREAD_STACKSIZE      1024
#define TM_PERFTESTTHREAD_STACKSIZE 1024
#define TM_LOGTHREAD_STACKSIZE      1024

/* Threads priority 4(high) <-> 127(low) */
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
#define TM_FTPDTHREAD_PRIORITY          70
#define TM_UDPECHOTHREAD_PRIORITY       70
#define TM_TCPECHOTHREAD_PRIORITY       70
#define TM_PPPTHREAD_PRIORITY           70
#define TM_ETHER_RECVTHREAD_PRIORITY    50
#define TM_SERIAL_RECVTHREAD_PRIORITY   60
#define TM_TIMERTHREAD_PRIORITY         30
#define TM_PERFTESTTHREAD_PRIORITY      65
#define TM_LOGTHREAD_PRIORITY           75
#else /* !TM_TRECK_PREEMPTIVE_KERNEL */
#define TM_FTPDTHREAD_PRIORITY          65
#define TM_UDPECHOTHREAD_PRIORITY       65
#define TM_TCPECHOTHREAD_PRIORITY       65
#define TM_PPPTHREAD_PRIORITY           65
#define TM_ETHER_RECVTHREAD_PRIORITY    65
#define TM_SERIAL_RECVTHREAD_PRIORITY   65
#define TM_TIMERTHREAD_PRIORITY         65
#define TM_PERFTESTTHREAD_PRIORITY      65
#define TM_LOGTHREAD_PRIORITY           65
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */

/* Round robin time slices for threads: 0 -> default */
#define TM_FTPDTHREAD_SLICES        1
#define TM_UDPECHOTHREAD_SLICES     1
#define TM_TCPECHOTHREAD_SLICES     1
#define TM_PPPTHREAD_SLICES         1
#define TM_RECVTHREAD_SLICES        1
#define TM_TIMERTHREAD_SLICES       1
#define TM_PERFTESTTHREAD_SLICES    0
#define TM_LOGTHREAD_SLICES         1

/*
 * Config PRT0(TIMER0) on eZ80 board 
 */
#define TM_Z80_TMR0_CTL     ((TM_INTIO)0x80)  /* control reg          */
#define TM_Z80_TMR0_RR_L    ((TM_INTIO)0x81)  /* reload reg low       */
#define TM_Z80_TMR0_RR_H    ((TM_INTIO)0x82)  /* reload reg high      */
#define TM_Z80_TMR_ISS      ((TM_INTIO)0x92)  /* input source select  */

/*
 * Config PRT1(TIMER1) on eZ80 board & install timer ISR
 */
#define TM_Z80_TMR1_CTL     ((TM_INTIO)0x83)  /* control reg          */
#define TM_Z80_TMR1_RR_L    ((TM_INTIO)0x84)  /* reload reg low       */
#define TM_Z80_TMR1_RR_H    ((TM_INTIO)0x85)  /* reload reg high      */

/*
 * Config Crystal cs8900a ISR
 */
#define TM_Z80_PD_DR        ((TM_INTIO)0xA2)
#define TM_Z80_PD_DDR       ((TM_INTIO)0xA3)
#define TM_Z80_PD_ALT1      ((TM_INTIO)0xA4)
#define TM_Z80_PD_ALT2      ((TM_INTIO)0xA5)

/* IR_CTL: Infrared Encoder/Decoder Control Register */
#define TM_Z80_IR_CTL       ((TM_INTIO)0xBF)

/*
 * Timer1 regs value
 */
/*
 * Reload Value = SYSTEM CLOCK / DIV / TIMER RATE
 *              = 48M/256(DIV)/500
 *              = 375 = 177H
 */
#define TM_TMR0_RR_L_DATA   0x77
#define TM_TMR0_RR_H_DATA   0x01

/*
 * Reload Value = SYSTEM CLOCK / DIV / TIMER RATE
 *              = 48M/256(DIV)/100
 *              = 1875 = 753H
 */
#define TM_TMR1_RR_L_DATA   0x53
#define TM_TMR1_RR_H_DATA   0x07

/* Timer0 Contrl reg value */
#define TM_TMR0_CTL_DATA    0x5F

/* Timer1 Contrl reg value */
#define TM_TMR1_CTL_DATA    0x5F

/* Timer0 ISS reg value */
#define TM_TMR0_ISS_MASK    0x03
#define TM_TMR0_ISS_DATA    0x00 /* 00 : system clock */

/* Timer1 ISS reg value */
#define TM_TMR1_ISS_MASK    0x0C
#define TM_TMR1_ISS_DATA    0x00 /* 00 : system clock */

/* CS3 registers value */
#define TM_Z80_PD_DR_DATA        0xFC
#define TM_Z80_PD_DDR_DATA       0xEF
#define TM_Z80_PD_ALT1_DATA      0x10
#define TM_Z80_PD_ALT2_DATA      0x13

/* eZ80 reg macro */
#define tm_z80_reg(x)  ((x)[0])

/* RZK interrupt id */
#define TM_RZK_TIMER1	6

/* Tyep defines */
typedef volatile unsigned char __INTIO * TM_INTIO;

/* User socket address pointer union */
typedef union tuUserSockPtr
{
    struct sockaddr_in TM_FAR   *sockInPtr;
    struct sockaddr TM_FAR      *sockPtr;
} ttUserSockPtrUnion;

/* Extern function declaration */
extern void tfTimerIsrProlog(void);
extern int tfRamfsInit(ttUser16Bit blCount);

#ifdef TM_KERNEL_RZK_EZ80
extern RZK_STATUS_t       RZKGetErrorNum(void);
extern RZK_THREADHANDLE_t RZKGetCurrentThread(void);
#endif /* TM_KERNEL_RZK_EZ80 */

/*
 * Global Variable Definitions
 */
#ifdef TM_USE_SERIAL_INTERFACE
/* Serial interface handle */
ttUserInterface     tlSerialIfHandle = 0;

/* Serial receive thread handle */
ttUserGenericUnion  tlSerialRecvHandle;
#endif /* TM_USE_SERIAL_INTERFACE */

#if (TM_THREAD_PPP || defined(TM_USE_SERIAL_INTERFACE))
/* PPP connection status */
int tlPppConnectedStatus = 0;
#endif /* (TM_THREAD_PPP || (defined(TM_USE_SERIAL_INTERFACE)) */

#ifdef TM_USE_ETHERNET_INTERFACE
/* Ethernet interface handle */
ttUserInterface     tlEthernetIfHandle = 0;

/* Ethernet receive thread handle */
ttUserGenericUnion  tlEthernetRecvHandle;
#endif /* TM_USE_ETHERNET_INTERFACE */

/* Timer thread handle */
ttUserGenericUnion  tlTimerHandle;

/* Stack for threads */
#ifdef TM_USE_ETHERNET_INTERFACE
/* Ethernet receive thread stack definition */
char tlEthernetRecvThreadStack[TM_RECVTHREAD_STACKSIZE];
#endif /* TM_USE_ETHERNET_INTERFACE */

#ifdef TM_USE_SERIAL_INTERFACE
/* Serial receive thread stack definition */
char tlSerialRecvThreadStack[TM_RECVTHREAD_STACKSIZE];
#endif /* TM_USE_SERIAL_INTERFACE */

char tlTimerThreadStack[TM_TIMERTHREAD_STACKSIZE];

#if TM_THREAD_FTPD
ttUserGenericUnion  tlFtpdHandle;
char tlFtpdThreadStack[TM_FTPDTHREAD_STACKSIZE];
#endif /* TM_THREAD_FTPD */

#if TM_THREAD_UDPECHO
ttUserGenericUnion  tlUdpechoHandle;
char tlUdpechoThreadStack[TM_UDPECHOTHREAD_STACKSIZE];
#endif /* TM_THREAD_UDPECHO */

#if TM_THREAD_TCPECHO
ttUserGenericUnion  tlTcpechoHandle;
char tlTcpechoThreadStack[TM_TCPECHOTHREAD_STACKSIZE];
#endif /* TM_THREAD_TCPECHO */

#if TM_THREAD_PPP
ttUserGenericUnion  tlPppHandle;
char tlPppThreadStack[TM_PPPTHREAD_STACKSIZE];
#endif /* TM_THREAD_PPP */

#if TM_THREAD_PERFTEST
ttUserGenericUnion  tlPerfTestHandle;
char tlPerfTestThreadStack[TM_PERFTESTTHREAD_STACKSIZE];
#endif /* TM_THREAD_PERFTEST */

#ifdef TM_DEBUG_LOGGING
ttUserGenericUnion  tlLogHandle;
struct sockaddr_in  tlToAddress;
int tlLastSeqNo;
int tlLoggingSd;
volatile int treckError;
char tlLogMsg[(TM_DEF_MAX_LOG_MSG_LEN * 2) + 22];
char tlLogThreadStack[TM_LOGTHREAD_STACKSIZE];
#endif /* TM_DEBUG_LOGGING */

#ifdef TM_KERNEL_RZK_EZ80
RZK_DCB_t pDCB[];
#endif /* TM_KERNEL_RZK_EZ80 */

/*
 * External function definitions
 */
#ifdef TM_USE_TRECK_SEMAPHORE
extern int tfTaskRegister (ttUserGenericUnionPtr taskIdPtr, int taskPri);
#endif /* TM_USE_TRECK_SEMAPHORE */

/*
 * Function prototype definitions
 */
void RZKApplicationEntry(void);
#ifdef TM_USE_SERIAL_INTERFACE
void tfInitUart0 ( void );
ttUserInterface tfStartupSerial( ttUserIpAddress ipAddress,
                                 ttUserIpAddress netMask);
void tfSerialRecvThread(void);
void tfPppNotifyFunc(ttUserInterface pInterface, unsigned int flags);
#endif /* TM_USE_SERIAL_INTERFACE */

#ifdef TM_USE_ETHERNET_INTERFACE
void tfInitEthernet(void);
ttUserInterface tfStartupEthernet( ttUserIpAddress ipAddress,
                                   ttUserIpAddress netMask);
void tfEthernetRecvThread(void);
#endif /* TM_USE_ETHERNET_INTERFACE */

#if TM_THREAD_PPP
void tfPppTest(void);
#endif /* TM_THREAD_PPP */

void tfTimerThread(void);
void Timer1ISR(void);
void tfConfigTmr0(void);
void tfInstallTimerIsr(void);

void tfTimerIsrHandler(void);

#if TM_THREAD_FTPD
void tfFtpdThread(void);
#endif /* TM_THREAD_FTPD */

#if TM_THREAD_UDPECHO
void tfUdpechoThread(void);
#endif /* TM_THREAD_UDPECHO */

#if TM_THREAD_TCPECHO
void tfTcpechoThread(void);
#endif /* TM_THREAD_TCPECHO */

#if TM_THREAD_PPP
void tfPppThread(void);
#endif /* TM_THREAD_PPP */

#if TM_THREAD_PERFTEST
void tfPerfTestThread(void);
#endif /* TM_THREAD_PERFTEST */

#ifdef TM_DEBUG_LOGGING
void tfLogThread(void);
#endif /* TM_DEBUG_LOGGING */

#ifdef TM_PPP_LQM
ttUser8Bit myLinkQualityMonitor(
    ttUserInterface interfaceHandle,
    int reasonCode,
    unsigned long timeElapsedMsec,
    ttLqrCountDeltasPtr countDeltasPtr,
    ttConstLqrCountsPtr countsPtr,
    ttUser32Bit outLqrs,
    ttUser32Bit outPackets,
    ttUser32Bit outOctets );
#endif /* TM_PPP_LQM */

/* Functions */
void RZKApplicationEntry(void)
{
	int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_STATUS_t        rzkStatus;
#endif /* TM_KERNEL_RZK_EZ80 */

    tfInitTreck();

/* Set tick length */
    errorCode = tfInitTreckOptions(TM_OPTION_TICK_LENGTH, 10);

/* Start the Treck TCP/IP Stack */
	(void)tfStartTreck();

/* Config timer1 ISR & install TRECK timer ISR handler */
    tfInstallTimerIsr();

#ifdef TM_USE_ETHERNET_INTERFACE
/* Initialize Ethernet */
    tfInitEthernet();

/* Start the Treck stack: add and open the ethernet interface */
    tlEthernetIfHandle = tfStartupEthernet( inet_addr(TM_ETHER_OUR_IP_ADDR),
                                            inet_addr(TM_NET_MASK));

    if( tlEthernetIfHandle == 0)
    {
        tfKernelError("RZKApplicationEntry","Start Ethernet failed");
    }
#endif /* TM_USE_ETHERNET_INTERFACE */

#ifdef TM_USE_SERIAL_INTERFACE
/* Initialize UART */
    tfInitUart0();

/* Start the Treck stack: add and open the PPP interface */
    tlSerialIfHandle = tfStartupSerial( inet_addr(TM_SERIAL_OUR_IP_ADDR),
                                        inet_addr(TM_NET_MASK));

    if( tlSerialIfHandle == 0)
    {
        tfKernelError("RZKApplicationEntry","Start serial interface failed");
    }
#endif /* TM_USE_SERIAL_INTERFACE */

/* Generate threads */
#if TM_THREAD_FTPD
/* Create & start FTPD thread */
    tlFtpdHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"FTPD",    /* name     */
                                tfFtpdThread,               /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlFtpdThreadStack+TM_FTPDTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_FTPDTHREAD_PRIORITY,     /* priority */
                                TM_FTPDTHREAD_SLICES,       /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlFtpdHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlFtpdHandle, TM_FTPDTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_THREAD_FTPD */

#if TM_THREAD_UDPECHO
/* Create & start UDP echo server */
    tlUdpechoHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"UDPECHO", /* name     */
                                tfUdpechoThread,            /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlUdpechoThreadStack+TM_UDPECHOTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_UDPECHOTHREAD_PRIORITY,  /* priority */
                                TM_UDPECHOTHREAD_SLICES,    /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlUdpechoHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlUdpechoHandle, TM_UDPECHOTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_THREAD_UDPECHO */

#if TM_THREAD_TCPECHO
/* Create & start TCP echo server */
    tlTcpechoHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"TCPECHO", /* name     */
                                tfTcpechoThread,               /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlTcpechoThreadStack+TM_TCPECHOTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_TCPECHOTHREAD_PRIORITY,  /* priority */
                                TM_TCPECHOTHREAD_SLICES,    /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlTcpechoHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlTcpechoHandle, TM_TCPECHOTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_THREAD_TCPECHO */

#if TM_THREAD_PPP
/* Create & start PPP application thread */
    tlPppHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"PPP",     /* name     */
                                tfPppThread,                /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlPppThreadStack+TM_PPPTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_PPPTHREAD_PRIORITY,  /* priority */
                                TM_PPPTHREAD_SLICES,    /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlPppHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlPppHandle, TM_PPPTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_THREAD_PPP */

#ifdef TM_DEBUG_LOGGING
/* Create & start LOG thread */
    tlLogHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"LOG",     /* name     */
                                tfLogThread,                /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlLogThreadStack+TM_LOGTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_LOGTHREAD_PRIORITY,      /* priority */
                                TM_LOGTHREAD_SLICES,        /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION      /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_LOG */
                                );
    if(tlLogHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlLogHandle, TM_LOGTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_DEBUG_LOGGING */


#if TM_THREAD_PERFTEST
/* Create & start perfermance test thread */
    tlPerfTestHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"PERFTEST",/* name     */
                                tfPerfTestThread,           /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlPerfTestThreadStack+TM_PERFTESTTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_PERFTESTTHREAD_PRIORITY, /* priority */
                                TM_PERFTESTTHREAD_SLICES,   /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlPerfTestHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlPerfTestHandle, TM_PERFTESTTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_THREAD_PERFTEST */

#ifdef TM_USE_ETHERNET_INTERFACE
/* Create & start ethernet receive thread */
    tlEthernetRecvHandle.genVoidParmPtr = RZKCreateThread( 
                        (unsigned char *)"ETHRECV",/* name     */
                        tfEthernetRecvThread,   /* entry    */
                        NULL,                   /* parm     */
                        NULL,                   /* clean    */
                        (tlEthernetRecvThreadStack+TM_RECVTHREAD_STACKSIZE),
                                                /* stack    */
                        TM_ETHER_RECVTHREAD_PRIORITY, /* priority */
                        TM_RECVTHREAD_SLICES,   /* slices   */
                        RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                        | RZK_THREAD_ROUNDROBIN 
                        | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                        );
    if(tlEthernetRecvHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlEthernetRecvHandle, TM_ETHER_RECVTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_USE_ETHERNET_INTERFACE */

#ifdef TM_USE_SERIAL_INTERFACE
/* Create & start ethernet receive thread */
    tlSerialRecvHandle.genVoidParmPtr = RZKCreateThread( 
                        (unsigned char *)"SERIRECV",/* name     */
                        tfSerialRecvThread,     /* entry    */
                        NULL,                   /* parm     */
                        NULL,                   /* clean    */
                        (tlSerialRecvThreadStack+TM_RECVTHREAD_STACKSIZE),
                                                /* stack    */
                        TM_SERIAL_RECVTHREAD_PRIORITY, /* priority */
                        TM_RECVTHREAD_SLICES,   /* slices   */
                        RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                        | RZK_THREAD_ROUNDROBIN 
                        | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                        );
    if(tlSerialRecvHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlSerialRecvHandle, TM_SERIAL_RECVTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */
#endif /* TM_USE_SERIAL_INTERFACE */

/* Create & start timer thread */
    tlTimerHandle.genVoidParmPtr = RZKCreateThread( 
                                (unsigned char *)"TIMER",/* name     */
                                tfTimerThread,              /* entry    */
                                NULL,                       /* parm     */
                                NULL,                       /* clean    */
                                (tlTimerThreadStack+TM_TIMERTHREAD_STACKSIZE),
                                                            /* stack    */
                                TM_TIMERTHREAD_PRIORITY,    /* priority */
                                TM_TIMERTHREAD_SLICES,      /* slices   */
                                RZK_THREAD_AUTOSTART
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
                                | RZK_THREAD_ROUNDROBIN 
                                | RZK_THREAD_PREEMPTION       /* mode     */
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
                                );
    if(tlTimerHandle.genVoidParmPtr == NULL)
    {
        rzkStatus = RZKGetErrorNum();
    }
#ifdef TM_USE_TRECK_SEMAPHORE
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
    else
    {
        (void)tfTaskRegister(&tlTimerHandle, TM_TIMERTHREAD_PRIORITY);
    }
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */
#endif /* TM_USE_TRECK_SEMAPHORE */

    return;
}


/*
 * System internal timer call back function(ISR)
 */
void tfTimerIsrHandler(void)
{
/* Update timer value */
	tfTimerUpdateIsr();
    return;
}


/*
 * Timer thread entry function
 */
void tfTimerThread(void)
{
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  timerThreadHandle;

    timerThreadHandle = RZKGetCurrentThread();
#endif /* TM_KERNEL_RZK_EZ80 */

    while(1)
    {
        tfTimerExecute();
#ifdef TM_KERNEL_RZK_EZ80
        (void)RZKSuspendThread( timerThreadHandle,
                                TM_TIMERTHREAD_WAITTICKS );
#endif /* TM_KERNEL_RZK_EZ80 */
    }
}


/*
 * Receive task entry function
 */
#ifdef TM_USE_ETHERNET_INTERFACE
void tfEthernetRecvThread(void)
{
    while(1)
    {
        tfWaitReceiveInterface(tlEthernetIfHandle);
		tfRecvInterface(tlEthernetIfHandle);
    }
    return;
}
#endif /* TM_USE_ETHERNET_INTERFACE */

#ifdef TM_USE_SERIAL_INTERFACE
void tfSerialRecvThread(void)
{
    while(1)
    {
        tfWaitReceiveInterface(tlSerialIfHandle);
		tfRecvInterface(tlSerialIfHandle);
    }
    return;
}
#endif /* TM_USE_SERIAL_INTERFACE */

/*
 * Thread Entry Functions
 */
#if TM_THREAD_FTPD
/* FTPD thread entry function */
void tfFtpdThread(void)
{
    int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;
#endif /* TM_KERNEL_RZK_EZ80 */

/* Initialize RAM file system */
    (void)tfRamfsInit(TM_RAMFS_BLOCKS);

    errorCode = tfFtpdUserStart(TM_FS_ALLCMND_MASK,/* File command allowed */
                                2,                 /* Max connections      */
                                2,                 /* Max concurrent conns */
                                1200,              /* Idle timeout(seconds)*/
                                TM_BLOCKING_ON );  /* Blocking state       */

#ifdef TM_KERNEL_RZK_EZ80
    thisThreadHandle = RZKGetCurrentThread();
    (void)RZKTerminateThread( thisThreadHandle );
#endif /* TM_KERNEL_RZK_EZ80 */
    return;
}
#endif /* TM_THREAD_FTPD */


#if TM_THREAD_UDPECHO
/* UDP echo server thread entry function */
void tfUdpechoThread(void)
{
    int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;
#endif /* TM_KERNEL_RZK_EZ80 */
	
    errorCode = tfTestTreck(  TM_TEST_UDP_ECHO_SERVER,
                              TM_ETHER_OUR_IP_ADDR,
                              htons(7),
                              64,
                              0,
                              (unsigned long int)(TM_TEST_FLAG_ZEROCOPY
							  | TM_TEST_FLAG_UDP_CS_OFF),
                              (ttUserTestHandlePtr)0,
                              (void TM_FAR *)0UL);

#ifdef TM_KERNEL_RZK_EZ80
    thisThreadHandle = RZKGetCurrentThread();
    (void)RZKTerminateThread( thisThreadHandle );
#endif /* TM_KERNEL_RZK_EZ80 */
    return;
}
#endif /* TM_THREAD_UDPECHO */


#if TM_THREAD_TCPECHO
/* TCP echo server thread entry function */
void tfTcpechoThread(void)
{
    int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;
#endif /* TM_KERNEL_RZK_EZ80 */

    errorCode = tfTestTreck(  TM_TEST_TCP_ECHO_SERVER,
                              TM_ETHER_OUR_IP_ADDR,
                              htons(TM_ECHO_PORT),
                              1,
                              0,
                              (unsigned long int)TM_TEST_FLAG_ZEROCOPY,
                              (ttUserTestHandlePtr)0,
                              (void TM_FAR *)0UL);

#ifdef TM_KERNEL_RZK_EZ80
    thisThreadHandle = RZKGetCurrentThread();
    (void)RZKTerminateThread( thisThreadHandle );
#endif /* TM_KERNEL_RZK_EZ80 */
    return;
}
#endif /* TM_THREAD_TCPECHO */

#if TM_THREAD_PPP
/* PPP thread entry function */
void tfPppThread(void)
{
    int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;
#endif /* TM_KERNEL_RZK_EZ80 */

    tfPppTest();

#ifdef TM_KERNEL_RZK_EZ80
    thisThreadHandle = RZKGetCurrentThread();
    (void)RZKTerminateThread( thisThreadHandle );
#endif /* TM_KERNEL_RZK_EZ80 */
    return;
}
#endif /* TM_THREAD_PPP */


#if TM_THREAD_PERFTEST

#define TM_PERF_TEST	4
/* Performance test thread entry function */
void tfPerfTestThread(void)
{
    int                 errorCode;
#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;
#endif /* TM_KERNEL_RZK_EZ80 */

#if (TM_PERF_TEST == 1)
/* 1.1 UDP short packet send test: disable UDP checksum */
    errorCode = tfTestTreck(  TM_TEST_UDP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              36,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)(TM_TEST_FLAG_ZEROCOPY
                              | TM_TEST_FLAG_UDP_CS_OFF), /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 1) */

#if (TM_PERF_TEST == 2)
/* 1.2 UDP short packet send test: enable UDP checksum */
    errorCode = tfTestTreck(  TM_TEST_UDP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              36,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)TM_TEST_FLAG_ZEROCOPY, /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 2) */

#if (TM_PERF_TEST == 3)
/* 2.1 UDP long packet send test: disable UCP checksum */
    errorCode = tfTestTreck(  TM_TEST_UDP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              1472,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)(TM_TEST_FLAG_ZEROCOPY
                              | TM_TEST_FLAG_UDP_CS_OFF), /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 3) */

#if (TM_PERF_TEST == 4)
/* 2.2 UDP long packet send test: enable UCP checksum */
    errorCode = tfTestTreck(  TM_TEST_UDP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              1472,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)TM_TEST_FLAG_ZEROCOPY, /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 4) */

#if (TM_PERF_TEST == 5)
/* 3.1 TCP short packet send test: TCP_NODELAY */
    errorCode = tfTestTreck(  TM_TEST_TCP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              12,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)(TM_TEST_FLAG_ZEROCOPY
                              | TM_TEST_FLAG_TCP_NODELAY), /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 5) */

#if (TM_PERF_TEST == 6)
/* 4 TCP long packet send test */
    errorCode = tfTestTreck(  TM_TEST_TCP_SEND,         /* test type */
                              TM_ETHER_PEER_IP_ADDR,  /* ip address */
                              htons(TM_DISCARD_PORT),   /* port number */
                              1448,                       /* data size */
                              10000,                    /* test count */
                              (unsigned long int)TM_TEST_FLAG_ZEROCOPY, /* flags */
                              (ttUserTestHandlePtr)0,   /* call back */
                              (void TM_FAR *)0UL);      /* parameter */
#endif /* (TM_PERF_TEST == 7) */

#ifdef TM_KERNEL_RZK_EZ80
    thisThreadHandle = RZKGetCurrentThread();
    (void)RZKTerminateThread( thisThreadHandle );
#endif /* TM_KERNEL_RZK_EZ80 */
    return;
}
#endif /* TM_THREAD_PERFTEST */

#ifdef TM_DEBUG_LOGGING
int walkLog(int msgSeqNo, ttConstCharPtr pszLogMsg, ttUserGenericUnion genParm)
{
    int len;
    int errorCode = 0;

    if (tlLastSeqNo < msgSeqNo)
    {
        len = tfSPrintF(tlLogMsg, "LOG seqNo=%d, msg=%s", msgSeqNo, pszLogMsg);
        errorCode = sendto(
            tlLoggingSd, 
            tlLogMsg, 
            len, 
            0,
            (const struct sockaddr TM_FAR *) &tlToAddress,
            (int)sizeof(struct sockaddr));

	    if (errorCode == -1)
        {
            treckError=tfGetSocketError(tlLoggingSd);
        }
        else
        {
            errorCode = TM_ENOERROR;
            tlLastSeqNo = msgSeqNo;
        }
    }

    return errorCode; /* continue the walk */
}

/* Log thread */
void tfLogThread(void)
{
    int                 ret;
    RZK_THREADHANDLE_t  thisThreadHandle;

    thisThreadHandle = RZKGetCurrentThread();

    tlLastSeqNo = -1;
    tlLoggingSd=socket(PF_INET,SOCK_DGRAM,0);
    tlToAddress.sin_family=PF_INET;
    tlToAddress.sin_addr.s_addr=inet_addr(TM_ETHER_PEER_IP_ADDR);
    tlToAddress.sin_port=htons(TM_DISCARD_PORT);

    while (1)
    {
        if (tlLoggingSd != TM_SOCKET_ERROR)
        {
            (void)RZKSuspendThread ( thisThreadHandle, 1 );

            ret = tfLogWalk(
                (ttLogWalkFuncPtr) walkLog,
                tm_context(tvLogCtrlBlkPtr), tlLastSeqNo, tlLogHandle);
            if( ret != TM_ENOERROR )
            {
                tfKernelError("tfLogThread","Log walk failed");
            }
        }
    }

	return;
}
#endif /* TM_DEBUG_LOGGING */


void tfConfigTmr0(void)
{
    unsigned char tmp;

/* Disable timer 1 */
    tm_z80_reg(TM_Z80_TMR0_CTL) = 0x00;

/* Set timer 1 input source select */
    tmp = tm_z80_reg(TM_Z80_TMR_ISS);
    tmp &= ~TM_TMR0_ISS_MASK;
    tmp |= (TM_TMR0_ISS_MASK & TM_TMR0_ISS_DATA);
    tm_z80_reg(TM_Z80_TMR_ISS) = tmp;

/* Set timer 1 reload register */
    tm_z80_reg(TM_Z80_TMR0_RR_L) = TM_TMR0_RR_L_DATA;
    tm_z80_reg(TM_Z80_TMR0_RR_H) = TM_TMR0_RR_H_DATA;

/* Enable timer 1 */
    tm_z80_reg(TM_Z80_TMR0_CTL) = TM_TMR0_CTL_DATA;

    return;
}


void tfInstallTimerIsr(void)
{
    unsigned char tmp;

/* Disable timer 1 */
    tm_z80_reg(TM_Z80_TMR1_CTL) = 0x00;

/* Set timer 1 input source select */
    tmp = tm_z80_reg(TM_Z80_TMR_ISS);
    tmp &= ~TM_TMR1_ISS_MASK;
    tmp |= (TM_TMR1_ISS_MASK & TM_TMR1_ISS_DATA);
    tm_z80_reg(TM_Z80_TMR_ISS) = tmp;

/* Set timer 1 reload register */
    tm_z80_reg(TM_Z80_TMR1_RR_L) = TM_TMR1_RR_L_DATA;
    tm_z80_reg(TM_Z80_TMR1_RR_H) = TM_TMR1_RR_H_DATA;

/* Install timer 1 ISR */
#ifdef TM_KERNEL_RZK_EZ80
    RZKInstallInterruptHandler(tfTimerIsrProlog, TM_RZK_TIMER1);
/*    RZKInstallInterruptHandler(tfTimerIsrHandler, TM_RZK_TIMER1); */
#endif /* TM_KERNEL_RZK_EZ80 */

/* Enable timer 1 */
    tm_z80_reg(TM_Z80_TMR1_CTL) = TM_TMR1_CTL_DATA;

    return;
}


#ifdef TM_USE_ETHERNET_INTERFACE
/*
 * tfInitEthernet description:
 * Initialize board settings for Crystal 8900a chipset: Config PD4 for
 * cs8900a interrupt input.
 */
void tfInitEthernet(void)
{
    unsigned char value;

/* Config PD4 for cs8900a interrupt */
    value = tm_z80_reg(TM_Z80_PD_DR);
    tm_z80_reg(TM_Z80_PD_DR)    = value | 0x10;
    value = tm_z80_reg(TM_Z80_PD_DDR);
    tm_z80_reg(TM_Z80_PD_DDR)   = value & 0xEF;
    value = tm_z80_reg(TM_Z80_PD_ALT1);
    tm_z80_reg(TM_Z80_PD_ALT1)  = value | 0x10;
    value = tm_z80_reg(TM_Z80_PD_ALT2);
    tm_z80_reg(TM_Z80_PD_ALT2)  = value | 0x10;

#if 0
#ifdef TM_KERNEL_RZK_EZ80
    isrHandle = RZKInstallInterruptHandler(tfCrystalIsrProlog, TM_RZK_CSIO);
/*    isrHandle = RZKInstallInterruptHandler(tfCrystalIsrHandler, TM_RZK_CSIO); */
    if( isrHandle == NULL )
    {
        tfKernelError("tfKernelInstallIsrHandler",
                      "RZK install interrupt handler failed");
    }
#endif /* TM_KERNEL_RZK_EZ80 */
#endif

    return;
}


/*
 * Perform system startup (init link layer, etc)
 */
ttUserInterface tfStartupEthernet( ttUserIpAddress ipAddress,
                                   ttUserIpAddress netMask)
{
	ttUserLinkLayer linkLayerHandle;
    ttUserInterface interfaceHandle;
	int             errorCode;

/* Use the Ethernet Link Layer (Including ARP) */
	linkLayerHandle=tfUseEthernet();
	
/* Add our interface */
    interfaceHandle = tfAddInterface( "CRYSTAL", 
                                      linkLayerHandle,
                                      tfCrystalOpen,
                                      tfCrystalClose,
									  tfCrystalSend,
                                      tfCrystalReceive,
                                      (ttDevFreeRecvFuncPtr)0,
                                      tfCrystalIoctl,
                                      tfCrystalGetPhysicalAddress,
                                      &errorCode);
    if( errorCode != TM_ENOERROR )
    {
        tfKernelError("tfStartupEthernet","Failed to add ethernet interface");
    }
	
/* Config the Interface */
	errorCode = tfConfigInterface(  interfaceHandle,
			                        ipAddress,
			                        netMask,
			                        0,
			                        1,
			                        0);
	
	if( errorCode != TM_ENOERROR )
	{
        errorCode = tfCloseInterface(interfaceHandle);
        interfaceHandle = (ttUserInterface)0;
	}
	
	return(interfaceHandle);
}
#endif /* TM_USE_ETHERNET_INTERFACE */


#ifdef TM_USE_SERIAL_INTERFACE
/* Fot test UART */
/* #define TM_UART_TEST */
void tfInitUart0 ( void ) 
{
    unsigned char       value;
#ifdef TM_UART_TEST
	char				testSend[] = "Qin";
#endif /* TM_UART_TEST */

/*  uzictl0 = UARTMODE; */
/*    tm_z80_reg((TM_INTIO)0xCF) = 0x01; */

/* PD0 = uart0_tx, PD1 = uart0_rx */
    value = tm_z80_reg(TM_Z80_PD_ALT1);
	tm_z80_reg(TM_Z80_PD_ALT1) = value & 0xFC;
    value = tm_z80_reg(TM_Z80_PD_ALT2);
	tm_z80_reg(TM_Z80_PD_ALT2) = value | 0x03;

/* PD_DDR is not read/write, but WRITE-ONLY */
    value = tm_z80_reg(TM_Z80_PD_DDR);
	tm_z80_reg(TM_Z80_PD_DDR) = value | 0x03;

/*
 * Config Infrared Encoder/Decoder Control Register:
 * Disable Encoder/Decoder, Pass IR_RXD data to UART0 RxD
 */
/*    tm_z80_reg(TM_Z80_IR_CTL) = 0x02; */

#ifdef TM_UART_TEST
    tfUartOpen(tlSerialIfHandle);
    (void)tfUartSend(tlSerialIfHandle, testSend, 3, 0);
    tm_thread_stop;
#endif /* TM_UART_TEST */
    return;
}  


/*
 * Perform system startup (init link layer, etc)
 */
ttUserInterface tfStartupSerial( ttUserIpAddress ipAddress,
                                 ttUserIpAddress netMask)
{
	ttUserLinkLayer linkLayerHandle;
    ttUserInterface interfaceHandle;
	int             errorCode;
	ttUser8Bit      timeoutValue;
#ifdef TM_PPP_LQM
    ttUser32Bit     reportingPeriod;
#endif /* TM_PPP_LQM */
    ttUser8Bit      magicNumberStatus;
	ttUserIpAddress peerIpAddress;

/* Use the Serial Link Layer */
#ifdef TM_ASYNC_SERVER
    linkLayerHandle = tfUseAsyncServerPpp( tfPppNotifyFunc );
#else /* TM_ASYNC_SERVER */
    linkLayerHandle = tfUseAsyncPpp( tfPppNotifyFunc );
#endif /* TM_ASYNC_SERVER */
	
/* Add our interface */
    interfaceHandle = tfAddInterface( "SERIAL", 
                                      linkLayerHandle,
                                      tfUartOpen,
                                      tfUartClose,
									  tfUartSend,
                                      tfUartReceive,
                                      tfUartFreeReceiveBuffer,
                                      tfUartIoctl,
                                      0,
                                      &errorCode);
    if( errorCode != TM_ENOERROR )
    {
        tfKernelError("tfStartupSerial","Failed to add serial interface");
    }

/* Set options */
#ifdef TM_PPP_LQM
    if (errorCode == TM_ENOERROR)
    {
        errorCode = tfUsePppLqm(
            interfaceHandle,
            (ttUser32Bit) 3000);

        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","Use PPP LQM failed");
        }
    }
#endif /* TM_PPP_LQM */

    if (errorCode == TM_ENOERROR)
    {
        timeoutValue = 20;
        errorCode = tfPppSetOption( interfaceHandle, TM_PPP_LCP_PROTOCOL,
					 TM_PPP_OPT_WANT, TM_LCP_TIMEOUT, 
					 (const char TM_FAR *)&timeoutValue, 
					 1 );
        
        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        magicNumberStatus = 1;
        errorCode = tfPppSetOption(
            interfaceHandle,
            TM_PPP_LCP_PROTOCOL,
            TM_PPP_OPT_WANT,
            TM_LCP_MAGIC_NUMBER,
            (const char*) &magicNumberStatus,
            1);
        
        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }

#ifdef TM_PPP_LQM
    if (errorCode == TM_ENOERROR)
    {
        reportingPeriod = 5000;
        errorCode = tfPppSetOption(
            interfaceHandle,
            TM_PPP_LCP_PROTOCOL,
            TM_PPP_OPT_WANT,
            TM_LCP_QUALITY_PROTOCOL,
            (const char*) &reportingPeriod,
            4);
        
        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        reportingPeriod = 5000;
        errorCode = tfPppSetOption(
            interfaceHandle,
            TM_PPP_LCP_PROTOCOL,
            TM_PPP_OPT_ALLOW,
            TM_LCP_QUALITY_PROTOCOL,
            (const char*) &reportingPeriod,
            4);
        
        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }
#endif /* TM_PPP_LQM */

    if (errorCode == TM_ENOERROR)
    {
        peerIpAddress=inet_addr(TM_SERIAL_PEER_IP_ADDR);
        errorCode = tfPppSetOption(
            interfaceHandle,
            TM_PPP_IPCP_PROTOCOL,
            TM_PPP_OPT_ALLOW,
            TM_IPCP_IP_ADDRESS,
            (char *) &peerIpAddress,
            4);

        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        errorCode = tfPppSetOption(
            interfaceHandle,
            TM_PPP_IPCP_PROTOCOL,
            TM_PPP_OPT_WANT,
            TM_IPCP_IP_ADDRESS,
            (char *) &ipAddress,
            4);

        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial","PPP set option failed");
        }
    }

    if (errorCode == TM_ENOERROR)
    {
/* Configure the interface */
        errorCode=tfConfigInterface(
            interfaceHandle,
            ipAddress,
            netMask,
            0,
            1,
			0);
        
        if (errorCode != TM_ENOERROR)
        {
            tfKernelWarning("tfStartupSerial",
                            "Config serial interface failed");
        }
    }

	if( errorCode != TM_ENOERROR )
	{
	    (void)tfCloseInterface(interfaceHandle);
#ifdef TM_PPP_LQM
        (void)tfFreePppLqm(interfaceHandle);
#endif /* TM_PPP_LQM */
        interfaceHandle = (ttUserInterface)0;
	}
	
	return(interfaceHandle);
}


void tfPppNotifyFunc(ttUserInterface interfaceHandle, unsigned int flag)
{
    TM_UNREF_IN_ARG(interfaceHandle);

    switch (flag)
    {
#ifdef TM_PPP_LQM
    case TM_LL_LQM_LINK_BAD:
        tfKernelWarning("tfPppNotifyFunc","Link quality is bad.");
/* TBD: attempt recovery of the PPP link */
        break;

    case TM_LL_LQM_UP:
        tfKernelWarning("tfPppNotifyFunc","LQM enabled on the link.");
        break;

    case TM_LL_LQM_DISABLED:
        tfKernelWarning("tfPppNotifyFunc","LQM disabled on the link.");
        break;
#endif /* TM_PPP_LQM */

    case TM_LL_OPEN_COMPLETE:
        tfKernelWarning("tfPppNotifyFunc","PPP link tlOpened.");
        tlPppConnectedStatus = 1;
        break;

    case TM_LL_CLOSE_COMPLETE:
        tfKernelWarning("tfPppNotifyFunc","PPP link closed.");
        tlPppConnectedStatus = 0;
        break;

    default:
        tfKernelWarning("tfPppNotifyFunc","error flag.");
        break;
    }
    return;
}
#endif /* TM_USE_SERIAL_INTERFACE */


#if TM_THREAD_PPP
#define TM_TEST_MAX_PPP_PACKET_LEN  500
void tfPppTest(void)
{
    int                 loopCounter;
    int                 errorCode;
	int                 sock;
	int                 dataSock;
    struct sockaddr_in  addr;
	int                 addrLen;
	char				dataPtr[TM_TEST_MAX_PPP_PACKET_LEN];
    ttUserSockPtrUnion  tempSockAddr;

#ifdef TM_KERNEL_RZK_EZ80
    RZK_THREADHANDLE_t  thisThreadHandle;

    thisThreadHandle = RZKGetCurrentThread();
#endif /* TM_KERNEL_RZK_EZ80 */

/* wait for the PPP link to come up */
    while (!tlPppConnectedStatus)
    {
#ifdef TM_KERNEL_RZK_EZ80
        (void)RZKSuspendThread(thisThreadHandle, 5);
#endif /* TM_KERNEL_RZK_EZ80 */
    }

#ifdef TM_PPP_LQM
    errorCode = tfLqmRegisterMonitor(
        tlSerialIfHandle,
        myLinkQualityMonitor,
        TM_LQM_HYSTERESIS_MAX_FAILURES,
        TM_LQM_HYSTERESIS_SAMPLES);
#endif /* TM_PPP_LQM */

/* Discard server */
    sock = socket( PF_INET, SOCK_STREAM, IP_PROTOTCP);
	if(sock == TM_SOCKET_ERROR)
	{
	    errorCode = TM_SOCKET_ERROR;
    }
	else
	{
	    addr.sin_family      = PF_INET;
	    addr.sin_addr.s_addr = inet_addr("0.0.0.0");
	    addr.sin_port        = htons(TM_DISCARD_PORT);
		tempSockAddr.sockInPtr = &addr;
        errorCode = bind( sock, tempSockAddr.sockPtr, sizeof(addr) );
		if(errorCode == TM_ENOERROR)
		{
		    errorCode = listen(sock, 10);
            if (errorCode != TM_ENOERROR)
            {
                errorCode = tfGetSocketError(sock);
            }
			else
			{
			    for(;;)
				{
				    addrLen = sizeof(addr);
		            dataSock = accept( sock, tempSockAddr.sockPtr, &addrLen );
		            if (dataSock < 0)
		            {
		                errorCode = tfGetSocketError(sock);
						break;
		            }
		            else 
		            {
                        errorCode = recv(
                            dataSock, dataPtr, TM_TEST_MAX_PPP_PACKET_LEN, 0);
		                if (errorCode < 0)
		                {
		                    errorCode = tfGetSocketError(dataSock);
		                    break;
		                }
					}
			    }
			}
        }
	}

    return;
}


#ifdef TM_PPP_LQM
ttUser8Bit myLinkQualityMonitor(
    ttUserInterface interfaceHandle,
    int reasonCode,
    unsigned long timeElapsedMsec,
    ttLqrCountDeltasPtr countDeltasPtr,
    ttConstLqrCountsPtr countsPtr,
    ttUser32Bit outLqrs,
    ttUser32Bit outPackets,
    ttUser32Bit outOctets )
{
    int linkQuality;
    int outboundLqrsInPipeline;
    
    linkQuality = 0;
    switch(reasonCode)
    {
    case TM_LQM_MONITOR_LQR:
/* calculate link quality of the outgoing link */
        if (countDeltasPtr->deltaLastOutPackets >
            (countDeltasPtr->deltaPeerInPackets 
             + countDeltasPtr->deltaPeerInDiscards))
        {
            linkQuality += countDeltasPtr->deltaLastOutPackets
                - (countDeltasPtr->deltaPeerInPackets 
                   + countDeltasPtr->deltaPeerInDiscards);
        }

/* calculate link quality of the incoming link */
        if (countDeltasPtr->deltaPeerOutPackets 
            > countDeltasPtr->deltaSaveInPackets)
        {
            linkQuality += countDeltasPtr->deltaPeerOutPackets
                - countDeltasPtr->deltaSaveInPackets;
        }

/* calculate the number of outbound LQRs still in the pipeline */
        outboundLqrsInPipeline = (int) ((ttUser32Bit) 0xffffffff 
                                        & (outLqrs - countsPtr->lastOutLqrs));
        if (outboundLqrsInPipeline > 1)
        {
            linkQuality += outboundLqrsInPipeline - 1;
        }
        break;

    case TM_LQM_MONITOR_TIMEOUT:
        tfKernelWarning("myLinkQualityMonitor","LQM timeout occurred.");
        linkQuality = 1;
        break;

    default:
        tfKernelWarning("myLinkQualityMonitor","LQM invalid reasonCode.");
        break;
    }
    
    /* make it take at least 2 sample periods to declare the link bad */
    if (linkQuality > TM_LQM_HYSTERESIS_MAX_FAILURES)
    {
        linkQuality = TM_LQM_HYSTERESIS_MAX_FAILURES;
    }
    return (ttUser8Bit) linkQuality;
}
#endif /* TM_PPP_LQM */
#endif /* TM_THREAD_PPP */


#ifndef TM_USE_ETHERNET_INTERFACE
void tfCrystalIsrHandler(void)
{
    return;
}
#endif /* !TM_USE_ETHERNET_INTERFACE */


#ifndef TM_USE_SERIAL_INTERFACE
void tfUartIsrHandler(void)
{
    return;
}
#endif /* !TM_USE_SERIAL_INTERFACE */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: txrzkapp.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  4931    Output checksum: 43448
 */
