/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * $Source: source/kernel/threadx/arm/arm7/ghs/demo.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:46CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * Demo of the Turbo Treck TCP/IP stack and the ThreadX RTOS with the
 * Green Hills toolset.
 */

#include <trsocket.h> 
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include "trcrystl.h"
#include "tx_api.h"

/* If this file is being compiled with EventAnalyzer support... */
#ifdef TX_ENABLE_EVENT_LOGGING

/* then declare the linker-defined begin and end symbols for the .eventlog
   section.  */
extern char __ghsbegin_eventlog[];
extern char __ghsend_eventlog[];
#endif

/* declare the linker_defined end symbol for the .free_mem section.  */
extern char __ghsend_free_mem[];

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define THREAD_STACK_SIZE 2048

#define TIMER_TASK_PRI 2
#define RECV_TASK_PRI  3
#define APP_TASK_PRI   4

/* Define the thread object control blocks.  */
TX_THREAD   treckAppTcb;
TX_THREAD   treckTimerTcb;
TX_THREAD   treckRecvTcb;

/* Support for multiple interfaces */
ttUserInterface interfaceHandleArray[5];
#define ETHERNET_INTERFACE_1  0

/* Define thread and other function prototypes.  */
void treckAppTask(ULONG input);
void treckTimerTask(ULONG input);
void treckRecvTask(ULONG input);

void testDns(void);
void pingCallback(int sockDesc);

int sendtoNoCopy(struct sockaddr_in * toAddress, int size);
int sendNoCopy(struct sockaddr_in * toAddress, 
               int                  size, 
               int                  noDelayFlag, 
               int                  echoFlag);
int recvfromNoCopy(int size);


/* Define main entry point.  */

void main(int argc, char **argv) {
    puts("Hello, single-threaded world");    /* ThreadX not yet running */

    /* Start ThreadX.  This call never returns */
    tx_kernel_enter();
}


/* Define what the initial system looks like.  */

void tx_application_define(void *threadx_avail_mem) {

    UINT status;

    /* Define pointers to the first available memory
       and the end of free memory.  */
    char *free_mem = (char *)threadx_avail_mem;
    char *end_of_free_mem = (char *)&__ghsend_free_mem;

    /* If the remaining portion of the free_mem section is
       very small or negative, assume that the user placed a small
       free_mem section at the end of the linked image intending
       that memory usage would extend past the end of the section.
       In that case, just use 0xffffffff as the end of memory.  */
    if((signed int)(end_of_free_mem - free_mem) < 256) {
        end_of_free_mem = (char *)0xffffffff;
    }

    /* Create all threads in the system.  */
    status = tx_thread_create(&treckAppTcb, 
                              "Turbo Treck App", 
                              treckAppTask, 
                              0,
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              APP_TASK_PRI, 
                              APP_TASK_PRI,
                              TX_NO_TIME_SLICE, 
                              TX_AUTO_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

    /* If no more memory is available, stop creating objects.  */
    if(((unsigned int)(free_mem + THREAD_STACK_SIZE) >
            (unsigned int)end_of_free_mem))
        goto tx_free_mem_exhausted;
 
    /* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;

    status = tx_thread_create(&treckTimerTcb, 
                              "Turbo Treck Timer", 
                              treckTimerTask, 
                              0,
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              TIMER_TASK_PRI, 
                              TIMER_TASK_PRI,
                              TX_NO_TIME_SLICE, 
                              TX_DONT_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

    /* If no more memory is available, stop creating objects.  */
    if(((unsigned int)(free_mem + THREAD_STACK_SIZE) >
            (unsigned int)end_of_free_mem))
        goto tx_free_mem_exhausted;

    /* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;    

    status = tx_thread_create(&treckRecvTcb, 
                              "Turbo Treck Recv", 
                              treckRecvTask,
                              ETHERNET_INTERFACE_1, 
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              RECV_TASK_PRI,
                              RECV_TASK_PRI, 
                              TX_NO_TIME_SLICE, 
                              TX_DONT_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

    /* If no more memory is available, stop creating objects.  */
    if(((unsigned int)(free_mem + THREAD_STACK_SIZE) >
            (unsigned int)end_of_free_mem))
        goto tx_free_mem_exhausted;

    /* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;        

    return;
tx_create_error:
    /* Report the error.  */
    puts("Error in tx_application_define: a tx_*_create function returned\n"
            " a value other than TX_SUCCESS.");

    return;
tx_free_mem_exhausted:

    /* Report the error.  */
    puts("Error in tx_application_define: no more memory available in "
            "free_mem section.  Check your linker file");

    return;
}


void treckAppTask(ULONG input)
{
    ttUserLinkLayer    linkLayerHandle;
    ttUserInterface    interfaceHandle;
    ttUserIpAddress    ipAddress;
    ttUserIpAddress    netMask;
    int                errorCode;
    int                testType;
    int                testSize;    
    struct sockaddr_in toAddress;

/* Avoid compiler warning */
    input=input;

    printf("Turbo Treck task starting\n");

    tfInitTreck();

    tfStartTreck();
    
/* Use the Ethernet link layer */    
    linkLayerHandle = tfUseEthernet();

/* Add the network interface */
    interfaceHandle=tfAddInterface(
            "CS8900.1",
            linkLayerHandle,
            tfCrystalOpen,
            tfCrystalClose,
            tfCrystalSend,
            tfCrystalReceive,
            (ttDevFreeRecvFuncPtr)0,
            tfCrystalIoctl,
            tfCrystalGetPhysicalAddress,
            &errorCode);

    interfaceHandleArray[ETHERNET_INTERFACE_1] = interfaceHandle;

    ipAddress = inet_addr("10.2.2.52");
    netMask   = inet_addr("255.255.255.0");    

    errorCode=tfOpenInterface(
            interfaceHandle,
            ipAddress,
            netMask,
            0,  
            1);   

    tfAddDefaultGateway(interfaceHandle, inet_addr("10.2.2.160"));
 
/* Start the timer task */    
    tx_thread_resume(&treckTimerTcb);
    tx_thread_resume(&treckRecvTcb);
    
    toAddress.sin_family      = PF_INET;
    toAddress.sin_addr.s_addr = inet_addr("10.2.2.50");
    toAddress.sin_port        = htons(9);

    testType = 3;
    switch (testType)
    {
/* Test #1 - small UDP zero copy send */        
        case 1:
        testSize = 16;
        sendtoNoCopy(&toAddress, testSize);
        break;

/* Test #2 - large UDP zero copy send */
    case 2:
        testSize = 1472;
        sendtoNoCopy(&toAddress, testSize);
        break;
        
/* Test #3 - TCP zero copy send */      
    case 3:
        testSize = 1448;
        sendNoCopy(&toAddress, testSize, 0, 0);
        break;

/* Test #4 - TCP zero copy echo */      
    case 4:
        testSize = 512;
        toAddress.sin_port = htons(7);
        sendNoCopy(&toAddress, testSize, 0, 1);
        break;

/* Test #5 - Ping */
    case 5:    
        errorCode = tfPingOpenStart (
            "63.144.81.35",
            0,
            100,
            pingCallback );
/* Fall through and just wait for replies */
   
        case 9:
        while (1)
        {
        tx_thread_sleep(100);
        }
    }       
    
} 

int udpRecvCount;

/*
 * recvfromNoCopy
 */
int recvfromNoCopy(int size)
{
    char *         testPtr;
    int                errorCode;
    ttUserMessage      bufferHandle;
    int            udpSd;
    struct sockaddr_in bindAddr;

    udpSd = socket(PF_INET, SOCK_DGRAM, 0);

    bindAddr.sin_family      = PF_INET;
    bindAddr.sin_addr.s_addr = 0UL;
    bindAddr.sin_port        = htons(9);
    
    errorCode = bind(udpSd,
                     (struct sockaddr *) &bindAddr,
                     sizeof(bindAddr));
    
    udpRecvCount = 0;
    while (1) 
    {
        errorCode = tfZeroCopyRecvFrom(udpSd,
                                       &bufferHandle,
                                       &testPtr,
                                       size,
                                       0,
                                       0,
                                       0);
    
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(udpSd);
            if (errorCode != TM_EWOULDBLOCK)
            {
                printf("tfZeroCopyRecvFrom error = %d\n", errorCode);
                break;
            }
        }
        else
        {
            tfFreeZeroCopyBuffer(bufferHandle);
        }

        udpRecvCount++;

    }
    
    return errorCode;
}    
    
int udpSendCount;

/*
 * sendtoNoCopy
 */
int sendtoNoCopy(struct sockaddr_in *toAddress, int size)
{
    char *         testPtr;
    int            errorCode;
    ttUserMessage  bufferHandle;
    int            udpSd;

    udpSd = socket(PF_INET, SOCK_DGRAM, 0);
    
    udpSendCount = 0;
    while (1) 
    {
        bufferHandle=tfGetZeroCopyBuffer(size, &testPtr);

        errorCode = tfZeroCopySendTo(udpSd,
                                     bufferHandle,
                                     size,
                                     0,
                                     (const struct sockaddr TM_FAR *) toAddress,
                                     (int) sizeof(struct sockaddr) ); 

        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(udpSd);
            if (errorCode != TM_EWOULDBLOCK)
            {
                printf("tfZeroCopySendTo error = %d\n", errorCode);
                break;
            }
        }

        udpSendCount++;

    }
    
    return errorCode;
}

int tcpSendCount;

char sendBuffer[1000];

/*
 * sendNoCopy
 */
int sendNoCopy(struct sockaddr_in * toAddress, 
               int                  size, 
               int                  noDelayFlag, 
               int                  echoFlag)
{
    char *         testPtr;
    int            errorCode;
    ttUserMessage  bufferHandle;
    int            tcpSd;
    int            optParam;
    
    tcpSd = socket(PF_INET, SOCK_STREAM, 0);

    if (noDelayFlag)
    {
    optParam = 1;
    errorCode = setsockopt(tcpSd,
                           IP_PROTOTCP,
                           TCP_NODELAY,
                           (const char *) &optParam,
                           sizeof(optParam));
    
    if (errorCode != TM_ENOERROR)
    {
        goto sendNoCopyFinish;
    }
    }


    errorCode = connect(tcpSd,
                        (struct sockaddr *) toAddress,
                        sizeof(struct sockaddr_in));

    if (errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(tcpSd);
        printf("connect error = %d\n", errorCode);  
        goto sendNoCopyFinish;
    }
    
    tcpSendCount = 0;
    testPtr = sendBuffer;
    while (1) 
    {

        bufferHandle=tfGetZeroCopyBuffer(size, &testPtr);

        errorCode = tfZeroCopySend(tcpSd,
                                   bufferHandle,
                                   size,
                                   0);

        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(tcpSd);
            printf("tfZeroCopySend error = %d\n", errorCode);
            goto sendNoCopyFinish;
        }

        if (echoFlag)
        {

            errorCode = tfZeroCopyRecv(tcpSd,
                                       &bufferHandle,
                                       &testPtr,
                                       size,
                                       0);
        
            if (errorCode == TM_SOCKET_ERROR)
            {       
                errorCode = tfGetSocketError(tcpSd);
                printf("tfZeroCopyRecv error = %d\n", errorCode);
                goto sendNoCopyFinish;
            }
            else
            {
                if (bufferHandle != (ttUserMessage) 0)
                {
                    tfFreeZeroCopyBuffer(bufferHandle);
                }

            }
        
        }

        tcpSendCount++;
    }

sendNoCopyFinish:    
    return errorCode;
}    



/*
 * Treck timer task
 */
void treckTimerTask(ULONG input)
{
    ttUserInterface interfaceHandle;

    interfaceHandle = interfaceHandleArray[ETHERNET_INTERFACE_1];
    
    while (1)
    {
        tfTimerUpdate();
        tfTimerExecute();
        tfIoctlInterface(interfaceHandle, TM_DEV_IOCTL_REFILL_POOL_FLAG,
                         (void *) 0, 0);
        tx_thread_sleep(1);
    }
}    

/*
 * treckRecvTask
 */
void treckRecvTask(ULONG interfaceIndex)
{
    ttUserInterface interfaceHandle;

    interfaceHandle = interfaceHandleArray[interfaceIndex];

    while (1)
    {
        tfWaitReceiveInterface(interfaceHandle);
        tfRecvInterface(interfaceHandle);
    }
}    

void pingCallback(int sockDesc)
{
    printf("Ping reply\n");
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: demo.c        File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 29028    Output checksum: 46986
 */
