/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for ThreadX by Express Logic
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trthread.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/kernel/threadx/arm/arm7/ghs/trthread.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:50CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include ThreadX Stuff 
 */
#include <tx_api.h>
/*
 * Include Treck Stuff
 */
#include <trsocket.h>



#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1

int tlThreadXCountSemNum;
int tlThreadXEventNum;

char tlCountSemNameArray[5][15] =
    {"TreckCountSem0",
     "TreckCountSem1",
     "TreckCountSem2",
     "TreckCountSem3",
     "TreckCountSem4"};

char tlEventNameArray[5][12] =
    {"TreckEvent0",
     "TreckEvent1",
     "TreckEvent2",
     "TreckEvent3",
     "TreckEvent4"};

/*
 * ThreadX Byte Pool
 * Similar to Simple Heap
 * Size set to 100K Bytes
 * Longs are used to maintain alignment
 */
#define TM_BYTE_POOL_SIZE 100L*1024L

long tvBytePool[(TM_BYTE_POOL_SIZE)/4];

static TX_BYTE_POOL tlThreadxBytePool;

int debugAlloc = 0;

void tfKernelInitialize()
{
    UINT errorCode;
    
    errorCode=tx_byte_pool_create(&tlThreadxBytePool,
                                  "Treck Heap",
                                  (VOID *)tvBytePool,
                                  (ULONG)TM_BYTE_POOL_SIZE);

    
    if (errorCode != TX_SUCCESS)
    {
        tfKernelError("tfKernelInitialize","Faild to create byte pool");
        tm_thread_stop;
    }

    tlThreadXCountSemNum = 0;
    tlThreadXEventNum    = 0;
}

/*
 * Yields the CPU to other tasks at the same or higher priority
 */
void tfKernelTaskYield(void)
{
    tx_thread_relinquish();
}


/* 
 * Report an Error
 */
void tfKernelError(
    const char TM_FAR *functionName, const char TM_FAR *errorMessage)
{
    printf("fatal error: %s --- %s\n",functionName, errorMessage); 
    tm_thread_stop;
}

void tfKernelWarning(
    const char TM_FAR *functionName, const char TM_FAR *warnMessage)
{
    printf("warning: %s --- %s\n",functionName, warnMessage); 
}

/*
 * Allocate a memory block of size from the kernel
 */

void TM_FAR *tfKernelMalloc(unsigned size)
{
    void TM_FAR *memoryBlockPtr;
    UINT errorCode;

/*
 * ThreadX does the long word alignment so we don't have to 
 */
    errorCode=tx_byte_allocate(&tlThreadxBytePool,
                               (VOID *)&memoryBlockPtr,
                               (ULONG)size,
                               TX_NO_WAIT);
    if (errorCode!=TX_SUCCESS)
    {
        memoryBlockPtr=(void TM_FAR *)0;
    }

    return(memoryBlockPtr);
}

/*
 * Free the memory back to the kernel
 */

void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    (void)tx_byte_release((VOID *)memoryBlockPtr);
}

#ifndef TM_GHSARM7_ASM_CRITICAL
/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    (void)tx_interrupt_control(TX_INT_DISABLE);
}

/*
 * Resume normal operation (from tfKernelSetCritical)
 */
void tfKernelReleaseCritical()
{
    (void)tx_interrupt_control(TX_INT_ENABLE);
}
#endif /* TM_GHSARM7_ASM_CRITICAL */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    TX_SEMAPHORE   *semaphorePtr;
    int             retCode;
    UINT            status;
    char *      countSemStr;
    
    retCode=TM_KERN_OKAY;


    countSemStr = tlCountSemNameArray[tlThreadXCountSemNum];
    tlThreadXCountSemNum++;
    
    /* Initialize the semaphore to zero */
    semaphorePtr=tfKernelMalloc(sizeof(TX_SEMAPHORE));
    status =  tx_semaphore_create(semaphorePtr,countSemStr,0);
    if (status == TX_SUCCESS)
    {
        countingSemaphore->genVoidParmPtr = (void TM_FAR *)semaphorePtr;
    }
    else
    {
        tfKernelError("tfKernelCreateCountSem",
                      "Unable to Create ThreadX Semaphore");
        tm_thread_stop;
    }
    return retCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    UINT    kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    kernelError=tx_semaphore_get(
            (TX_SEMAPHORE *)countingSemaphore->genVoidParmPtr, 
            TX_WAIT_FOREVER);
    if (kernelError != TX_SUCCESS)
    {
        tfKernelError("tfKernelPendCountSem",
                      "Unable to Pend on ThreadX Semaphore");
        tm_thread_stop;
    }
    return(retCode);
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to tickTimeout (in ticks)
 */
int tfKernelPendCountSemTimeout(ttUserGenericUnionPtr countingSemaphore, ttUser32Bit tickTimeout)
{
    UINT    kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    kernelError=tx_semaphore_get(
            (TX_SEMAPHORE *)countingSemaphore->genVoidParmPtr, 
            tickTimeout);
    if (kernelError != TX_SUCCESS)
    {
        if (kernelError == TX_WAIT_ABORTED)
        {
            retCode = TM_ETIMEDOUT;
        }
        else
        {
            retCode = TM_KERN_ERROR;
        }
    }
    return(retCode);
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    UINT    kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    kernelError = tx_semaphore_put(
            (TX_SEMAPHORE *)countingSemaphore->genVoidParmPtr);
    if (kernelError != TX_SUCCESS)
    {
        tfKernelError("tfKernelPostCountSem",
                      "Unable to Post on ThreadX Semaphore");
        tm_thread_stop;
    }
    return(retCode);
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    UINT    kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    kernelError = tx_semaphore_delete(
            (TX_SEMAPHORE *)countingSemaphore->genVoidParmPtr);
    if (kernelError != TX_SUCCESS)
    {
        tfKernelError("tfKernelDeleteCountSem",
                      "Unable to Delete a ThreadX Semaphore");
        tm_thread_stop;
    }
    return(retCode);
}

/*
 * Install an ISR Handler 
 */
void tfKernelInstallIsrHandler(ttUserIsrHandlerPtr funcPtr, unsigned long offSet)
{
    return;
}


/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    TX_SEMAPHORE   *semaphorePtr;
    UINT            status;
    char *      eventStr;
    

    eventStr = tlEventNameArray[tlThreadXEventNum];
    tlThreadXEventNum++;
    
    /* Initialize the semaphore to zero */
    semaphorePtr=tfKernelMalloc(sizeof(TX_SEMAPHORE));
    status =  tx_semaphore_create(semaphorePtr,eventStr,0);
    if (status == TX_SUCCESS)
    {
        eventPtr->genVoidParmPtr = (void TM_FAR *)semaphorePtr;
    }
    else
    {
        tfKernelError("tfKernelCreateCountEvent",
                      "Unable to Create ThreadX Semaphore");
        tm_thread_stop;
    }
    return;
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    UINT    kernelError;

    kernelError=tx_semaphore_get(
            (TX_SEMAPHORE *)eventPtr->genVoidParmPtr, 
            TX_WAIT_FOREVER);
    if (kernelError != TX_SUCCESS)
    {
        tfKernelError("tfKernelPendEvent",
                      "Unable to Pend on ThreadX Semaphore");
        tm_thread_stop;
    }
    return;
}

/*
 * Resume waiting tasks that ware waiting on this event
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    UINT    kernelError;

    kernelError = tx_semaphore_put(
            (TX_SEMAPHORE *)eventPtr->genVoidParmPtr);
    if (kernelError != TX_SUCCESS)
    {
        tfKernelError("tfKernelIsrPostEvent",
                      "Unable to Post on ThreadX Semaphore");
        tm_thread_stop;
    }
    return;
}







/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trthread.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 21898    Output checksum: 53568
 */
