/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * $Source: source/kernel/threadx/motorola/cpu32/ghs/demo6.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:38:57CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * Demo of the Turbo Treck TCP/IP stack and the ThreadX RTOS with the
 * Green Hills toolset.
 */

#include <trsocket.h> 
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include <trquicc.h>
#include "tx_api.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* declare the linker_defined end symbol for the .free_mem section.  */
extern char __ghsend_free_mem[];

/* Enables Ethernet and/or serial interfaces */
#define USE_ETHERNET
/* #define USE_PPP */

#define OPEN_V4_INTERFACE
#define OPEN_V6_INTERFACE

#define V6_MANUAL_CONF

#define LOCAL_IP4_ADDR "10.2.2.50"
#define LOCAL_IP4_PREFIX_LEN 24 /* equivalent to 255.255.255.0 */
#define DEFAULT_IP4_GW_ADDR "10.2.2.160"

#define DEST_IP4_ADDR "10.2.2.160"
#define DEST_PORT 9


#define TEST_TYPE 3
#define TEST_ADDR_FAMILY AF_INET6
/* #define TEST_ADDR_FAMILY AF_INET */

/* * * * * * */

#define THREAD_STACK_SIZE 2048

/* Task priorities (timer task should always be highest) */
#define TIMER_TASK_PRI     2
#define ETH_RECV_TASK_PRI  3
#define PPP_RECV_TASK_PRI  4
#define APP_TASK_PRI       5

/* Define the thread object control blocks.  */
TX_THREAD   treckAppTcb;
TX_THREAD   treckTimerTcb;
TX_THREAD   treckEthRecvTcb;
TX_THREAD   treckPppRecvTcb;

/* Support for multiple interfaces */
ttUserInterface interfaceHandleArray[2];
#define ETHERNET_INTERFACE_1  0
#define PPP_INTERFACE_1       1

/* Define thread and other function prototypes.  */
void treckAppTask(ULONG input);
void treckTimerTask(ULONG input);
void treckRecvTask(ULONG input);

void pingCallback(int sockDesc);
int sendtoNoCopy(struct sockaddr * toAddress, int size);
int sendNoCopy(struct sockaddr *   toAddress, 
               int                  size, 
               int                  noDelayFlag, 
               int                  echoFlag);
int recvfromNoCopy(int size);

int openEthernetInterface(void);
int openPppInterface(void);

void ipv6Notify (ttUserInterface           interfaceHandle,
                 unsigned int              multiHomeIndex,
                 struct sockaddr_storage * ipv6AddrPtr,
                 int                       event );

extern void tfQuiccSetupTimer(void);

int pppStarted;
int ipv6Started;

/*
 * Main entry point.
 */
void main(int argc, char **argv) {
    puts("Hello, single-threaded world");    /* ThreadX not yet running */

    tfQuiccSetupTimer();
    
    /* Start ThreadX.  This call never returns */
    tx_kernel_enter();
}


/*
 * Define what the initial system looks like (tasks, etc)
 */
void tx_application_define(void *threadx_avail_mem) {

    UINT status;

/* Define pointers to the first available memory and the end of free memory. */
    char *free_mem = (char *)threadx_avail_mem;

/* Application task */
    status = tx_thread_create(&treckAppTcb, 
                              "Turbo Treck App", 
                              treckAppTask, 
                              0,
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              APP_TASK_PRI, 
                              APP_TASK_PRI,
                              TX_NO_TIME_SLICE, 
                              TX_AUTO_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

/* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;

/* Timer task */
    status = tx_thread_create(&treckTimerTcb, 
                              "Turbo Treck Timer", 
                              treckTimerTask, 
                              0,
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              TIMER_TASK_PRI, 
                              TIMER_TASK_PRI,
                              TX_NO_TIME_SLICE, 
                              TX_DONT_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

/* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;    

/* Ethernet receive task */
#ifdef USE_ETHERNET
    status = tx_thread_create(&treckEthRecvTcb, 
                              "Turbo Treck Recv", 
                              treckRecvTask,
                              ETHERNET_INTERFACE_1, 
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              ETH_RECV_TASK_PRI,
                              ETH_RECV_TASK_PRI, 
                              TX_NO_TIME_SLICE, 
                              TX_DONT_START);
    if (status != TX_SUCCESS)  goto tx_create_error;
    
/* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;        
#endif /* USE_ETHERNET */

/* PPP receive task */
#ifdef USE_PPP
    status = tx_thread_create(&treckPppRecvTcb, 
                              "Turbo PPP Treck Recv", 
                              treckRecvTask,
                              PPP_INTERFACE_1, 
                              free_mem, 
                              THREAD_STACK_SIZE, 
                              PPP_RECV_TASK_PRI,
                              PPP_RECV_TASK_PRI, 
                              TX_NO_TIME_SLICE, 
                              TX_DONT_START);
    if (status != TX_SUCCESS)  goto tx_create_error;

/* Advance the memory pointer past the thread's stack.  */
    free_mem += THREAD_STACK_SIZE;        
#endif /* USE_PPP */

    return;
tx_create_error:
/* Report the error.  */
    puts("Error in tx_application_define: a tx_*_create function returned\n"
            " a value other than TX_SUCCESS.");

    return;
}


void treckAppTask(ULONG input)
{
    int                 testType;
    int                 testSize;
    int                 errorCode;
    struct sockaddr_in  to4Address;
    struct sockaddr_in6 to6Address;
    struct sockaddr *   toAddrPtr;

/* Avoid compiler warning */
    input=input;

    printf("Turbo Treck task starting\n");

    tfInitTreck();

    tfStartTreck();

    interfaceHandleArray[ETHERNET_INTERFACE_1] = (ttUserInterface) 0;
    interfaceHandleArray[PPP_INTERFACE_1]      = (ttUserInterface) 0;
 
/* Start the timer task */    
    tx_thread_resume(&treckTimerTcb);

/* Init the Ethernet interface */
#ifdef USE_ETHERNET
    tx_thread_resume(&treckEthRecvTcb);

    errorCode = openEthernetInterface();

    if (errorCode != 0)
    {
        printf("Failed to open Ethernet interface: %d\n", errorCode);
    }
#endif /* USE_ETHERNET */

/* Init the PPP interface */
#ifdef USE_PPP
    tx_thread_resume(&treckPppRecvTcb);

    errorCode = openPppInterface();

    if (errorCode != 0)
    {
        printf("Failed to open PPP interface: %d\n", errorCode);
    }
#endif /* USE_PPP */

/* Destination address for the tests below */
    if (TEST_ADDR_FAMILY == AF_INET6)
    {
        tfBzero(&to6Address, sizeof(to6Address));
        
        to6Address.sin6_len = sizeof(struct sockaddr_in6);
        to6Address.sin6_family = AF_INET6;
        to6Address.sin6_port = htons(9);

        to6Address.sin6_addr.ip6Addr.ip6U16[0] = 0xfe80;
        to6Address.sin6_addr.ip6Addr.ip6U16[1] = 0x0;
        to6Address.sin6_addr.ip6Addr.ip6U16[2] = 0x0;
        to6Address.sin6_addr.ip6Addr.ip6U16[3] = 0x0;
        to6Address.sin6_addr.ip6Addr.ip6U16[4] = 0x0210;
        to6Address.sin6_addr.ip6Addr.ip6U16[5] = 0x5aff;
        to6Address.sin6_addr.ip6Addr.ip6U16[6] = 0xfe04;
        to6Address.sin6_addr.ip6Addr.ip6U16[7] = 0xc266;        

        errorCode =
            tf6SockaddrSetScopeId(interfaceHandleArray[ETHERNET_INTERFACE_1],
                                  (struct sockaddr_storage *) &to6Address);
            
        toAddrPtr = (struct sockaddr *) &to6Address;
    }
    else
    {
        to4Address.sin_family      = PF_INET;
        to4Address.sin_addr.s_addr = inet_addr(DEST_IP4_ADDR);
        to4Address.sin_port        = htons(9);

        toAddrPtr = (struct sockaddr *) &to4Address;
    }
    
    testType = TEST_TYPE;
    switch (testType)
    {
/* Test #1 - small UDP zero copy send */        
        case 1:
            testSize = 16;
            sendtoNoCopy(toAddrPtr, testSize);
            break;

/* Test #2 - large UDP zero copy send */
        case 2:
            testSize = 1472;
            sendtoNoCopy(toAddrPtr, testSize);
            break;
        
/* Test #3 - TCP zero copy send */      
        case 3:
            testSize = 1448;
            sendNoCopy(toAddrPtr, testSize, 0, 0);
            break;

/* Test #4 - TCP zero copy echo */      
        case 4:
            testSize = 512;
            to4Address.sin_port = htons(7);
            sendNoCopy(toAddrPtr, testSize, 0, 1);
            break;

/* Test #5 - Ping */
        case 5:    
            tfPingOpenStart ("10.2.2.160",
                             0,
                             100,
                             pingCallback );
/* Fall through and just wait for replies */
   
        case 9:
            while (1)
            {
                tx_thread_sleep(100);
            }
    }       
    
} 


/*
 * ipv6Notify
 */ 
void ipv6Notify (ttUserInterface           interfaceHandle,
                 unsigned int              multiHomeIndex,
                 struct sockaddr_storage * ipv6AddrPtr,
                 int                       event )
{
    if (event == TM_6_DEV_ADDR_CONFIG_COMPLETE)
    {
        ipv6Started = 1;
    }
}

/*
 * Start up an Ethernet interface.
 */ 
int openEthernetInterface(void)
{
    ttUserLinkLayer         linkLayerHandle;
    ttUserInterface         interfaceHandle;
    int                     errorCode;
    struct sockaddr_storage configAddr;
    struct in6_addr *       ip6AddrPtr;
    
/* Use the Ethernet link layer */    
    linkLayerHandle = tfUseEthernet();

/* Add the network interface */
    interfaceHandle=tfAddInterface(
            "QUICC.SCC1",
            linkLayerHandle,
            tfQuiccEtherOpen,
            tfQuiccEtherClose,
            tfQuiccEtherSend,
            tfQuiccEtherReceive,
            (ttDevFreeRecvFuncPtr)0,
            tfQuiccIoctl,
            tfQuiccGetPhyAddr,
            &errorCode);

    if (errorCode == TM_ENOERROR)
    {
        interfaceHandleArray[ETHERNET_INTERFACE_1] = interfaceHandle;

#ifdef OPEN_V4_INTERFACE
        tfBzero( (void *) &configAddr, sizeof(configAddr) );
        configAddr.addr.ipv4.sin_addr.s_addr = inet_addr(LOCAL_IP4_ADDR);
        configAddr.addr.ipv4.sin_family = PF_INET;
        configAddr.addr.ipv4.sin_len    = sizeof(struct sockaddr_in);

        puts("Opening IPv4 Ethernet interface.");
        errorCode = tfNgConfigInterface( interfaceHandle,
                                         &configAddr,
                                         LOCAL_IP4_PREFIX_LEN, 
                                         TM_DEV_MCAST_ENB,
                                         0,
                                         1,
                                         (tt6DevAddrNotifyFuncPtr) 0,
                                         0);

        puts("IPv4 Ethernet interface open.");
        
        tfAddDefaultGateway(interfaceHandle, inet_addr(DEFAULT_IP4_GW_ADDR));
#endif /* OPEN_V4_INTERFACE */

#ifdef OPEN_V6_INTERFACE
        if ( errorCode == TM_ENOERROR )
        {

#ifdef V6_MANUAL_CONF
/* Manually configure an address */        
            ip6AddrPtr = &configAddr.addr.ipv6.sin6_addr;

            ip6AddrPtr->ip6Addr.ip6U16[0] = 0x3ffe;
            ip6AddrPtr->ip6Addr.ip6U16[1] = 0x0302;
            ip6AddrPtr->ip6Addr.ip6U16[2] = 0x011;
            ip6AddrPtr->ip6Addr.ip6U16[3] = 0x4;
            ip6AddrPtr->ip6Addr.ip6U16[4] = 0x2ea;
            ip6AddrPtr->ip6Addr.ip6U16[5] = 0xc0ff;
            ip6AddrPtr->ip6Addr.ip6U16[6] = 0xfef0;
            ip6AddrPtr->ip6Addr.ip6U16[7] = 0x0;
#else /* !V6_MANUAL_CONF */
            tfBzero( (void *) &configAddr, sizeof(configAddr) );
#endif /* V6_MANUAL_CONF */
            configAddr.addr.ipv6.sin6_family = PF_INET6;

            puts("Opening IPv6 Ethernet interface.");
            ipv6Started = 0;
            errorCode = tfNgConfigInterface( interfaceHandle,
                                             &configAddr,
                                             64, /* prefix length */
                                             TM_DEV_MCAST_ENB,
                                             0,
                                             1,
                                             ipv6Notify,
                                             0 );

            if (errorCode == TM_EINPROGRESS)
            {
               errorCode = TM_ENOERROR;
            }
            
/* Wait for IPv6 to open */
            puts("Waiting for IPv6 Ethernet interface to open...");
            while (ipv6Started == 0)
            {
                tx_thread_sleep(1);
            }
            puts("IPv6 Ethernet interface open");
            
        }
#endif /* OPEN_V6_INTERFACE */
        
    }
    
    return errorCode;
}


/*
 * PPP notification routine.
 */ 
void pppNotify(ttUserInterface interfaceHandle, unsigned int flags)
{
    if (flags == TM_LL_OPEN_COMPLETE)
    {
        pppStarted=1;
    }

    if (flags == TM_LL_CLOSE_COMPLETE)
    {
        pppStarted = 0;
    }

    if (flags == TM_LL_IP4_OPEN_COMPLETE)
    { 
        puts("IPCP up");
    }

    if (flags == TM_LL_IP6_OPEN_COMPLETE)
    {
        puts("IPV6CP up");
    }
}

/*
 * Start up a PPP interface.  Returns when the connection is established or
 * an error has occurred.
 */ 
int openPppInterface(void)
{
    struct sockaddr_storage configAddr;
    ttUserInterface         interfaceHandle;
    ttUserLinkLayer         linkLayerHandle;
    ttUserIpAddress         ipAddress;
    unsigned short          opt16;
    int                     errorCode;
    char                    username[] = "testuser";
    char                    password[] = "testpass";
    
    linkLayerHandle = tfUseAsyncPpp(pppNotify);
    
/* Add the Interface */
    interfaceHandle=tfAddInterface(
            "QUICC.SCC4",
            linkLayerHandle,
            tfQuiccSerialOpen,
            tfQuiccSerialClose,
            tfQuiccSerialSend,
            tfQuiccSerialReceive,
            (ttDevFreeRecvFuncPtr)0,
            tfQuiccSerialIoctl,
            (ttDevGetPhysAddrFuncPtr)0,
            &errorCode);

    interfaceHandleArray[PPP_INTERFACE_1] = interfaceHandle;

    pppStarted = 0;

#ifdef OPEN_V4_INTERFACE
/* Allow server to set it's IP address. */
    if (errorCode == TM_ENOERROR)
    {
        ipAddress = 0UL;    
        errorCode = tfPppSetOption(interfaceHandle,
                                   TM_PPP_IPCP_PROTOCOL,
                                   TM_PPP_OPT_ALLOW,
                                   TM_IPCP_IP_ADDRESS,
                                   (const char *) &ipAddress,
                                   sizeof(ipAddress));
    }

/* Request an IP address from the server. */
    if (errorCode == TM_ENOERROR)
    {    
        ipAddress = 0UL;    
        errorCode = tfPppSetOption(interfaceHandle,
                                   TM_PPP_IPCP_PROTOCOL,
                                   TM_PPP_OPT_WANT,
                                   TM_IPCP_IP_ADDRESS,
                                   (const char *) &ipAddress,
                                   sizeof(ipAddress));
    }

/* Use PAP for authentication to server */
    if (errorCode == TM_ENOERROR)
    {    
        opt16 = TM_PPP_PAP_PROTOCOL;
        errorCode = tfPppSetOption(interfaceHandle,
                                   TM_PPP_LCP_PROTOCOL,
                                   TM_PPP_OPT_ALLOW,
                                   TM_LCP_AUTH_PROTOCOL,
                                   (const char *) &opt16,
                                   2);
    }

/* Set PAP username */
    if (errorCode == TM_ENOERROR)
    {    
        errorCode = tfPppSetOption(interfaceHandle,
                                   TM_PPP_PAP_PROTOCOL,
                                   TM_PPP_OPT_WANT,
                                   TM_PAP_USERNAME,
                                   username,
                                   tfStrLen(username));
    }

/* Set PAP password */
    if (errorCode == TM_ENOERROR)
    {    
        errorCode = tfPppSetOption(interfaceHandle,
                                   TM_PPP_PAP_PROTOCOL,
                                   TM_PPP_OPT_WANT,
                                   TM_PAP_PASSWORD,
                                   password,
                                   tfStrLen(password));
    }

/* Config the Interface (open connection) */
    if (errorCode == TM_ENOERROR)
    {
        puts("Opening IPv4 PPP interface");
        tfBzero( (void *) &configAddr, sizeof(configAddr) );
        configAddr.addr.ipv4.sin_family = PF_INET;
        configAddr.addr.ipv4.sin_len    = sizeof(struct sockaddr_in);

        errorCode = tfNgConfigInterface( interfaceHandle,
                                         &configAddr,
                                         64, /* prefix length */
                                         TM_DEV_MCAST_ENB,
                                         0,
                                         1,
                                         (tt6DevAddrNotifyFuncPtr) 0,
                                         0);
    }
#endif /* OPEN_V4_INTERFACE */

#ifdef OPEN_V6_INTERFACE
    if ( (errorCode == TM_ENOERROR) || (errorCode == TM_EINPROGRESS))
    {
        puts("Opening IPv6 PPP interface");
        tfBzero( (void *) &configAddr, sizeof(configAddr) );
        configAddr.addr.ipv6.sin6_family = PF_INET6;

        errorCode = tfNgConfigInterface( interfaceHandle,
                                         &configAddr,
                                         64, /* prefix length */
                                         TM_DEV_MCAST_ENB,
                                         0,
                                         1,
                                         ipv6Notify,
                                         0 );
    }
#endif /* OPEN_V6_INTERFACE */

/* Wait for connection to complete */
    if (errorCode == TM_ENOERROR)
    {    
        while (pppStarted == 0)
        {
            tx_thread_sleep(1);
        }

        puts("PPP interface open");
    }
    
    
    return errorCode;
}

int udpRecvCount;

/*
 * Zero copy UDP receive
 */
int recvfromNoCopy(int size)
{
    char *         testPtr;
    int                errorCode;
    ttUserMessage      bufferHandle;
    int            udpSd;
    struct sockaddr_in bindAddr;

    udpSd = socket(PF_INET, SOCK_DGRAM, 0);

    bindAddr.sin_family      = PF_INET;
    bindAddr.sin_addr.s_addr = 0UL;
    bindAddr.sin_port        = htons(9);
    
    errorCode = bind(udpSd,
                     (struct sockaddr *) &bindAddr,
                     sizeof(bindAddr));
    
    udpRecvCount = 0;
    while (1) 
    {
        errorCode = tfZeroCopyRecvFrom(udpSd,
                                       &bufferHandle,
                                       &testPtr,
                                       size,
                                       0,
                                       0,
                                       0);
    
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(udpSd);
            if (errorCode != TM_EWOULDBLOCK)
            {
                printf("tfZeroCopyRecvFrom error = %d\n", errorCode);
                break;
            }
        }
        else
        {
            tfFreeZeroCopyBuffer(bufferHandle);
        }

        udpRecvCount++;

    }
    
    return errorCode;
}    
    
int udpSendCount;

/*
 * Zero copy UDP send
 */
int sendtoNoCopy(struct sockaddr *toAddress, int size)
{
    char *         testPtr;
    int            errorCode;
    ttUserMessage  bufferHandle;
    int            udpSd;

    udpSd = socket(toAddress->sa_family, SOCK_DGRAM, 0);
    
    udpSendCount = 0;
    while (1) 
    {
        bufferHandle=tfGetZeroCopyBuffer(size, &testPtr);

        errorCode = tfZeroCopySendTo(udpSd,
                                     bufferHandle,
                                     size,
                                     0,
                                     (const struct sockaddr TM_FAR *) toAddress,
                                     (int) sizeof(struct sockaddr) ); 

        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(udpSd);
            if (errorCode != TM_EWOULDBLOCK)
            {
                printf("tfZeroCopySendTo error = %d\n", errorCode);
                break;
            }
        }

        udpSendCount++;

        tx_thread_sleep(1000);
    }
    
    return errorCode;
}

int tcpSendCount;

char sendBuffer[1000];

/*
 * Zero copy TCP send
 */
int sendNoCopy(struct sockaddr *    toAddress, 
               int                  size, 
               int                  noDelayFlag, 
               int                  echoFlag)
{
    char *         testPtr;
    int            errorCode;
    ttUserMessage  bufferHandle;
    int            tcpSd;
    int            optParam;

    tcpSd = socket(toAddress->sa_family, SOCK_STREAM, 0);

    if (noDelayFlag)
    {
        optParam = 1;
        errorCode = setsockopt(tcpSd,
                               IP_PROTOTCP,
                               TCP_NODELAY,
                               (const char *) &optParam,
                               sizeof(optParam));
    
        if (errorCode != TM_ENOERROR)
        {
            goto sendNoCopyFinish;
        }
    }


    errorCode = connect(tcpSd, toAddress, toAddress->sa_len);

    if (errorCode != TM_ENOERROR)
    {
        errorCode = tfGetSocketError(tcpSd);
        printf("connect error = %d\n", errorCode);  
        goto sendNoCopyFinish;
    }
    
    tcpSendCount = 0;
    testPtr = sendBuffer;
    while (1) 
    {

        bufferHandle=tfGetZeroCopyBuffer(size, &testPtr);

        errorCode = tfZeroCopySend(tcpSd,
                                   bufferHandle,
                                   size,
                                   0);

        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(tcpSd);
            printf("tfZeroCopySend error = %d\n", errorCode);
            goto sendNoCopyFinish;
        }

        if (echoFlag)
        {

            errorCode = tfZeroCopyRecv(tcpSd,
                                       &bufferHandle,
                                       &testPtr,
                                       size,
                                       0);
        
            if (errorCode == TM_SOCKET_ERROR)
            {       
                errorCode = tfGetSocketError(tcpSd);
                printf("tfZeroCopyRecv error = %d\n", errorCode);
                goto sendNoCopyFinish;
            }
            else
            {
                if (bufferHandle != (ttUserMessage) 0)
                {
                    tfFreeZeroCopyBuffer(bufferHandle);
                }

            }
        
        }

        tcpSendCount++;
    }

sendNoCopyFinish:    
    return errorCode;
}    

/*
 * Treck timer task
 */
void treckTimerTask(ULONG input)
{
    ttUserInterface interfaceHandle;

    while (1)
    {
        tfTimerUpdate();
        tfTimerExecute();

#ifdef USE_ETHERNET
        interfaceHandle = interfaceHandleArray[ETHERNET_INTERFACE_1];

        if (interfaceHandle != (ttUserInterface)0)
        {
            tfIoctlInterface(interfaceHandle,
                             TM_QUICC_REFILL_SCC1|TM_QUICC_SEND_COMPLETE,
                             (void *) 0, 0);
        }
#endif /* USE_ETHERNET */

#ifdef USE_PPP
        interfaceHandle = interfaceHandleArray[PPP_INTERFACE_1];

        if (interfaceHandle != (ttUserInterface) 0)
        {
            tfIoctlInterface(interfaceHandle,
                             TM_QUICC_REFILL_SCC4|TM_QUICC_SEND_COMPLETE,
                             (void *) 0, 0);
        }
#endif /* USE_PPP */
        
        tx_thread_sleep(1);
    }
}    

/*
 * Receive task (used for both PPP & Ethernet)
 */
void treckRecvTask(ULONG interfaceIndex)
{
    ttUserInterface interfaceHandle;

    while (1)
    {
        if (interfaceHandle != (ttUserInterface) 0)
        {
            tfWaitReceiveInterface(interfaceHandle);
            tfRecvInterface(interfaceHandle);
        }
        else
        {
            tx_thread_sleep(1);
            interfaceHandle = interfaceHandleArray[interfaceIndex];
        }
    }
}    

/*
 * Called by the stack when a PING reply is received.
 */ 
void pingCallback(int sockDesc)
{
    printf("Ping reply\n");
}


/*************************/



/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: demo6.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  7958    Output checksum: 64144
 */
