/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Counting Semaphore implementation without task priorities
 * Filename: trctsem2.c
 * Author: Odile
 * Date Created: 09/01/98
 * $Source: source/kernel/trctsem2.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:39:16CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * Counting semaphore implementation for OS that do not support this
 * concept. We support up to TM_NUMCOUNTSEM counting semaphores. For
 * OS primitive, we use events, and expect the OS to support events.
 * When tfTaskInit() is called by the user (prior to calling tfStartTreck()),
 * up to TM_NUMEVT events are created by calling the OS create event
 * routine, the event values are stored in the local event array,
 * and each entry in the local event array is marked as unused
 * (TM_EVENT_INVALID).
 * When a task creates a counting semaphore, we reserve a space in our
 * array of counting semaphores, and initialize the counting semaphore to
 * zero. When a task tries and pends on the counting semaphore, if the
 * counting semaphore is non zero, we decrease it and grant the access
 * to the caller, otherwise (counting semaphore is zero), we will find
 * an unused event in our local event array, and queue that event entry,
 * in FIFO order, in the event queue associated with the counting
 * semaphore, and then we will call tfKernelPendEvent()
 * passing a pointer to the event associated with the task, and the task
 * will pend.
 * When a post on a counting semaphore occurs, if there is any event
 * in the event queue associated with the counting semaphore (i.e a pending
 * task), we go ahead and remove the first event from the event queue
 * (and only the first event), and post on that event by calling
 * tfKernelTaskPostEvent() passing a pointer to the event associated
 * with the task, and the task will resume. Otherwise,
 * no task is pending, we just go ahead and increase the counting semaphore.
 *
 * This version assumes that counting sempahores are created, but
 * never deleted, which is the case with the counting sempahores used
 * by the Treck stack, since the Treck stack recycles the created counting
 * semaphores internally.
 *
 * I. Functions/procedures needed to be written by the user
 *
 *   I.1 void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
 *        Wait on an Event (pointed to by eventPtr) posted from another task
 *        This would probably be similar if not the same as
 *        tfKernelPendEvent().
 *
 *   I.2 void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
 *         Resume waiting tasks that ware waiting on the event pointed
 *         to by eventPtr.
 *         This would probably be similar to tfKernelPostIsrEvent(),
 *         except that this will be posted from a task, and not from
 *         an ISR, and most OS have different calls when a Post occurs
 *         from an ISR or occurs from a task.
 *
 * II. Calls needed to be made by the user at run time
 *
 *      The user need to call tfTaskInit() prior to running any Treck task,
 *      and prior to calling tfStartTreck().
 *      There the local variables are initialized, and events are pre-created.
 *
 * III. Macros needed to be modified by the user:
 *
 *      Number of events needed for the Treck stack. This number should match
 *      the number of tasks making calls to the Treck task. This includes
 *      the Treck timer task, Treck receive task, (possibly the Treck
 *      send complete task if any), and any user application making
 *      socket API calls. Default is 12.
 *   #define TM_NUMEVT 12
 */

#include <trsocket.h>
#include <trmacro.h>

/*
 * Local macros
 */
#define TM_EVENT_INVALID  (unsigned char)0
#define TM_EVENT_VALID    (unsigned char)1

/*
 * Number of events should match number of tasks making calls to the Treck
 * task. This includes the Treck timer task, Treck receive task,
 * (possibly the Treck send complete task if any), and any user
 * application making socket API calls. Default is 12.
 */
#define TM_NUMEVT            12

/* Treck does not use more counting semaphores than Treck tasks */
#define TM_NUMCOUNTSEM       TM_NUMEVT

/*
 * Local types
 */

/*
 * Event data structure. Each valid entry corresponds to an OS event.
 * A valid entry can be queued to a counting semaphore.
 */
typedef struct tsEventEntry
{
    struct tsEventEntry      TM_FAR * evNextPtr;
    ttUserGenericUnion                evEventUnion;
    unsigned char                     evEventValid;
    unsigned char                     evFiller[3];
} ttEventEntry;

typedef ttEventEntry TM_FAR * ttEventEntryPtr;

/*
 * Counting semaphore data structure. We keep track of the counting
 * semaphore count, and the queue of events waiting on the counting
 * semaphore.
 */
typedef struct tsCountSemEntry
{
    ttEventEntryPtr      csEventQueuePtr;
    ttEventEntryPtr      csEventLastQueuePtr;
    int                  csCount;
} ttCountSemEntry;
typedef ttCountSemEntry TM_FAR * ttCountSemEntryPtr;

typedef void           TM_FAR  * ttVoidPtr;

/*
 * Local variables
 */

/* Table of events (primitives for counting semaphores). */
ttEventEntry       tlEventArray[TM_NUMEVT];

/* Table of counting semaphores. Counting semaphore ID is index */
ttCountSemEntry    tlCountSemArray[TM_NUMCOUNTSEM];
/*
 * Counting Sempahore index (starting search of next available counting
 * sempahore ID)
 */
unsigned int       tlCountSemIndex;

void tfTaskInit (void);

/*
 * tfTaskInit() function description:
 * Function to be called by the user prior to calling tfStartTreck(), and
 * prior to running any Treck task.
 * Local tables are initialized to zero.
 * Up to TM_NUMEVT are created by calling the OS create event routine, and
 * their values stored in our local event array, and each entry in the
 * local event array is marked as unused (TM_EVENT_INVALID) (by bzero).
 */
void tfTaskInit (void)
{
    int index;

    tm_bzero(&tlEventArray[0], sizeof(tlEventArray));
    tm_bzero(&tlCountSemArray[0], sizeof(tlCountSemArray));
    tlCountSemIndex = 0;
    for (index = 0; index < TM_NUMEVT; index++)
    {
        tfKernelCreateEvent(&(tlEventArray[index].evEventUnion));
    }
}

/*
 * Wait on an Event posted from another task
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    TM_UNREF_IN_ARG(eventPtr);
/*  OSWaitSetEvent(eventPtr->genIntParm); */
    return;
}

#ifdef TM_USE_PEND_TIMEOUT
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                              ttUser32Bit tickTimeout )
{
    int retCode;

    TM_UNREF_IN_ARG(eventPtr);
    TM_UNREF_IN_ARG(tickTimeout);
    retCode = TM_KERN_OKAY;
/*  retCode = OSWaitSetEvent(eventPtr->genIntParm, tickTimeout); */
    return retCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks that ware waiting on this event
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    TM_UNREF_IN_ARG(eventPtr);
/*  OSIncEvent(eventPtr->genIntParm); */
    return;
}

/*
 * tfKernelCreateCountSem function description:
 * Create a counting semaphore. Search for a free (not in use) slot in
 * our table of counting semaphores. This version assumes that the user
 * will never delete a counting semaphore, which is the case with
 * Treck tasks.
 * Parameters:
 * countingSemaphorePtr Pointer to a counting semaphore union
 * Return value:
 * TM_KERN_OKAY  success
 * TM_KERN_ERROR failure
 */
int tfKernelCreateCountSem( ttUserGenericUnionPtr countingSemaphorePtr )
{
    int             retCode;

    retCode = TM_KERN_ERROR;
    tm_kernel_set_critical;
    if (tlCountSemIndex < (unsigned int)TM_NUMCOUNTSEM)
    {
        countingSemaphorePtr->genIntParm = (int)tlCountSemIndex++;
        retCode = TM_KERN_OKAY;
    }
    tm_kernel_release_critical;
    return retCode;
}

/*
 * tfKernelPendCount
 * Wait on a counting semaphore. If counting semaphore is zero, we
 * find a unused event (TM_EVENT_INVALID) in our table of pre-created events,
 * and queue the event to the counting semaphore, and wait on the task
 * event. If counting semaphore is non zero, we get the semaphore
 * (decrease it by one), and return.
 * Parameters:
 * countingSemaphorePtr Pointer to a counting semaphore union
 * Return value:
 * TM_KERN_OKAY  success
 * TM_KERN_ERROR failure
 */
int tfKernelPendCountSem( ttUserGenericUnionPtr countingSemaphorePtr )
#ifdef TM_USE_PEND_TIMEOUT
{
    return tfKernelPendCountSemTimeout(countingSemaphorePtr, TM_UL(0));
}

/*
 * Wait on a counting semaphore. If counting semaphore is zero, we
 * queue our task to the counting semaphore, and wait on the task
 * event. If counting semaphore is non zero, we get the semaphore
 * (decrease it by one), and return.
 */
int tfKernelPendCountSemTimeout ( ttUserGenericUnionPtr countingSemaphorePtr,
                                  ttUser32Bit tickTimeout )
#endif /* TM_USE_PEND_TIMEOUT */
{
    ttCountSemEntryPtr countSemPtr;
    ttEventEntryPtr    eventPtr;
    unsigned int       eventIndex;
    unsigned int       countSemId;
    int                retCode;
#ifdef TM_USE_PEND_TIMEOUT
    ttEventEntryPtr    eventQueuePtr;
    ttEventEntryPtr    prevEventQueuePtr;
#endif /* TM_USE_PEND_TIMEOUT */

    retCode = TM_KERN_ERROR;
/* Counting semaphore ID */
    countSemId = (unsigned int)countingSemaphorePtr->genIntParm;
    if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
    {
        countSemPtr = &tlCountSemArray[countSemId];
        tm_kernel_set_critical;
        if (countSemPtr->csCount > 0)
        {
/* Decrease the counting semaphore. We are granted access */
            countSemPtr->csCount--;
            tm_kernel_release_critical;
            retCode = TM_KERN_OKAY;
        }
        else
        {
            for ( eventIndex = 0; eventIndex < TM_NUMEVT; eventIndex++ )
            {
                if (    tlEventArray[eventIndex].evEventValid
                     == TM_EVENT_INVALID )
                {
/* Grab an unused event */
                    tlEventArray[eventIndex].evEventValid = TM_EVENT_VALID;
                    break;
                }
            }
            if (eventIndex < (unsigned int)TM_NUMEVT)
            {
                eventPtr = &tlEventArray[eventIndex];
                if ( countSemPtr->csEventQueuePtr == (ttEventEntryPtr)0 )
                {
/* Empty counting semaphore event queue */
                    countSemPtr->csEventQueuePtr = eventPtr;
                    countSemPtr->csEventLastQueuePtr = eventPtr;
                    eventPtr->evNextPtr = (ttEventEntryPtr)0;
                }
                else
                {
/* Insert at the end of counting semaphore event queue */
                    countSemPtr->csEventLastQueuePtr->evNextPtr = eventPtr;
                    countSemPtr->csEventLastQueuePtr = eventPtr;
                    eventPtr->evNextPtr = (ttEventEntryPtr)0;
                }
                tm_kernel_release_critical;
#ifdef TM_USE_PEND_TIMEOUT
                if (tickTimeout != TM_UL(0))
                {
                    retCode = tfKernelPendEventTimeout(
                                &eventPtr->evEventUnion, tickTimeout);
                    if (retCode != TM_KERN_OKAY)
/* Timeout occured. Remove counting semaphore from the event queue */
                    {
                        retCode = TM_ETIMEDOUT;
                        tm_kernel_set_critical;
                        eventQueuePtr = countSemPtr->csEventQueuePtr;
                        prevEventQueuePtr = (ttEventEntryPtr)0;
                        if (eventQueuePtr == eventPtr) 
                        {
/* At head of queue, special case: head of queue point to next entry */
                            countSemPtr->csEventQueuePtr = eventPtr->evNextPtr;
                        }
                        else
                        {
                            while (eventQueuePtr != (ttEventEntryPtr)0)
                            {
                                prevEventQueuePtr = eventQueuePtr;
                                eventQueuePtr = eventQueuePtr->evNextPtr;
                                if (eventQueuePtr == eventPtr) 
                                {
/* Remove from queue, using previous pointer */
                                    prevEventQueuePtr->evNextPtr = 
                                                           eventPtr->evNextPtr;
                                    break;
                                }
                            }
                        }
                        if ( countSemPtr->csEventLastQueuePtr == eventPtr )
                        {
/* At end of queue: update pointer to last item in queue */
                            countSemPtr->csEventLastQueuePtr = 
                                                           prevEventQueuePtr;
                        }
                        tm_kernel_release_critical;
                    }
                }
                else
#endif /* TM_USE_PEND_TIMEOUT */
                {
                    tfKernelPendEvent(&eventPtr->evEventUnion);
                    retCode = TM_KERN_OKAY;
                }
/* Release the event */
                eventPtr->evEventValid = TM_EVENT_INVALID;
            }
            else
            {
                tm_kernel_release_critical;
#ifdef TM_ERROR_CHECKING
                tfKernelError("tfKernelPendCountSem","Invalid Event Flag");
#endif /* TM_ERROR_CHECKING */
            }
        }
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPendCountSem","Invalid Counting Semaphore");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Resume first task waiting on a counting semaphore
 * Parameters:
 * countingSemaphorePtr Pointer to a counting semaphore union
 * Return value:
 * TM_KERN_OKAY  success
 * TM_KERN_ERROR failure
 */
int tfKernelPostCountSem( ttUserGenericUnionPtr countingSemaphorePtr )
{
    ttCountSemEntryPtr  countSemPtr;
    ttEventEntryPtr     eventQueuePtr;
    unsigned int        countSemId;
    int                 retCode;

    retCode = TM_KERN_ERROR;
    countSemId = (unsigned int)countingSemaphorePtr->genIntParm;
    if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
    {
        countSemPtr = &tlCountSemArray[countSemId];
        tm_kernel_set_critical;
        eventQueuePtr = countSemPtr->csEventQueuePtr;
/* If no task is waiting on the semaphore yet. Just increase the count */
        if (eventQueuePtr == (ttEventEntryPtr)0)
        {
            countSemPtr->csCount++;
            tm_kernel_release_critical;
        }
        else
/* Remove and Reschedule task at head of the queue */
        {
/* Remove event entry at head of the queue */
            countSemPtr->csEventQueuePtr = eventQueuePtr->evNextPtr;
            tm_kernel_release_critical;
/* Move task to run queue */
            tfKernelTaskPostEvent(&eventQueuePtr->evEventUnion);
        }
        retCode = TM_KERN_OKAY;
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPostCountSem", "Invalid Counting Semaphore");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Delete a counting semaphore
 * Unused function by TRECK since semaphores are recycled internally
 * Parameters:
 * countingSemaphorePtr Pointer to a counting semaphore union
 * Return value:
 * TM_KERN_ERROR failure
 */
int tfKernelDeleteCountSem( ttUserGenericUnionPtr countingSemaphorePtr )
{
    int             retCode;

    TM_UNREF_IN_ARG(countingSemaphorePtr);
    retCode = TM_KERN_ERROR;
    return retCode;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:43      For: Xilinx-Samsung
 * File Name: trctsem2.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 38777    Output checksum: 29689
 */
