/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for uC/OS
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trucos.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/kernel/ucos/motorola/ppc/ghs/trucos.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:41:28CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <includes.h> /* uC/OS include */
#include <trsocket.h>



#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1


static char tlUcosErrorFunction[80];
static char tlUcosErrorType[80];
static char tlUcosWarnFunction[80];
static char tlUcosWarnType[80];


/*
 * Initialize the uC/OS kernel.  In this environment, nothing is necessary
 * as the kernel as already been started.
 */
void tfKernelInitialize()
{
}

/*
 * Yields the CPU to another task - does nothing under uC/OSS because of
 * its exclusive priority structure.
 */
void tfKernelTaskYield(void)
{
}

/* 
 * Report an Error
 */
void tfKernelError(
    const char TM_FAR *functionName, const char TM_FAR *errorMessage)
{
    char TM_FAR *ptr;
    ptr=tlUcosErrorFunction;
    while(*functionName)
        *ptr++ = *functionName++;
    ptr=tlUcosErrorType;
    while(*errorMessage)
        *ptr++ = *errorMessage++;
   
/*    printf("%s --- %s\n",functionName, errorMessage); */
    tm_thread_stop;
}

void tfKernelWarning(
    const char TM_FAR *functionName, const char TM_FAR *warnMessage)
{
    char TM_FAR *ptr;
    ptr=tlUcosWarnFunction;
    while(*functionName)
        *ptr++ = *functionName++;
    ptr=tlUcosWarnType;
    while(*warnMessage)
        *ptr++ = *warnMessage++;
}

/*
 * Allocate a memory block of size from the kernel
 */

/* 
 * Since uC/OS does not provide a heap, we call the simple heap
 * routine provided in the Treck libraries.
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    return(tfSheapMalloc(size));
}

/*
 * Since uC/OS does not provide a heap, we call the simple heap
 * routine provided in the Treck libraries.
 */

void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    tfSheapFree(memoryBlockPtr);
}

#ifndef TM_SDS68K_ASM_CRITICAL
/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    OS_ENTER_CRITICAL();
}

/*
 * Resume normal operation (from tfKernelSetCritical)
 */
void tfKernelReleaseCritical()
{
    OS_EXIT_CRITICAL();
}
#endif /* TM_SDS68K_ASM_CRITICAL */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    OS_EVENT TM_FAR *semaphorePtr;
    int             retCode;
    
    retCode=TM_KERN_OKAY;
    

    /* Initialize the semaphore to zero */
    semaphorePtr =  OSSemCreate(0);
    if (semaphorePtr != (OS_EVENT TM_FAR *)0)
    {
        countingSemaphore->genVoidParmPtr = (void TM_FAR *)semaphorePtr;
    }
    else
    {
        tfKernelError("tfKernelCreateCountSem","Unable to Create uC/OS Semaphore");
        tm_thread_stop;
    }
    return retCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    UBYTE   kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    OSSemPend((OS_EVENT TM_FAR *)countingSemaphore->genVoidParmPtr, 0, &kernelError);
    if (kernelError != OS_NO_ERR)
    {
        tfKernelError("tfKernelPendCountSem","Unable to Pend on uC/OS Semaphore");
        tm_thread_stop;
    }
    return(retCode);
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to a timeout value (in ticks)
 */
int tfKernelPendCountSemTimeout(ttUserGenericUnionPtr countingSemaphore, 
                                ttUser32Bit           tickTimeout)
{
    UBYTE   kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    OSSemPend( (OS_EVENT TM_FAR *)countingSemaphore->genVoidParmPtr, tickTimeout,
               &kernelError);
    if (kernelError != OS_NO_ERR)
    {
        if (kernelError == OS_TIMEOUT)
        {
            retCode = TM_ETIMEDOUT;
        }
        else
        {
            tfKernelError( "tfKernelPendCountSem",
                           "Unable to Pend on uC/OS Semaphore");
        }
    }
    return(retCode);
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    UBYTE   kernelError;
    int     retCode;

    retCode=TM_KERN_OKAY;
    kernelError = OSSemPost((OS_EVENT TM_FAR *)countingSemaphore->genVoidParmPtr);
    if (kernelError != OS_NO_ERR)
    {
        tfKernelError("tfKernelPendCountSem","Unable to Post on uC/OS Semaphore");
        tm_thread_stop;
    }
    return(retCode);
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    /* No uC/OS facility to delete a semaphore */
    return(TM_KERN_OKAY);
}

/*
 * Install an ISR Handler 
 */
void tfKernelInstallIsrHandler(ttUserIsrHandlerPtr funcPtr, unsigned long offSet)
{
    OSInstallISR((void TM_FAR *)funcPtr, offSet);
    return;
}

/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    OS_EVENT TM_FAR *semaphorePtr;
    
    

    /* Initialize the semaphore to zero */
    semaphorePtr =  OSSemCreate(0);
    if (semaphorePtr != (OS_EVENT TM_FAR *)0)
    {
        eventPtr->genVoidParmPtr = (void TM_FAR *)semaphorePtr;
    }
    else
    {
        tfKernelError("tfKernelCreateIsrEvent","Unable to Create uC/OS Semaphore");
        tm_thread_stop;
    }
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    UBYTE   kernelError;

    OSSemPend((OS_EVENT TM_FAR *)eventPtr->genVoidParmPtr, 0, &kernelError);
    if (kernelError != OS_NO_ERR)
    {
        tfKernelError("tfKernelPendIsrEvent","Unable to Pend on uC/OS Semaphore");
        tm_thread_stop;
    }
}

/*
 * Resume waiting tasks that ware waiting on this event.  One of these
 * functions is called from an ISR, the other from a task.  In uC/OS, the
 * call is the same (OSSemPost).
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    UBYTE   kernelError;

    kernelError = OSSemPost((OS_EVENT TM_FAR *)eventPtr->genVoidParmPtr);
    if (kernelError != OS_NO_ERR)
    {
        tfKernelError("tfKernelPostIsrEvent","Unable to Post on uC/OS Semaphore");
        tm_thread_stop;
    }
    return;
}

void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    UBYTE   kernelError;

    kernelError = OSSemPost((OS_EVENT TM_FAR *)eventPtr->genVoidParmPtr);
    if (kernelError != OS_NO_ERR)
    {
        tfKernelError("tfKernelTaskPostEvent","Unable to Post on uC/OS Semaphore");
        tm_thread_stop;
    }
    return;
}






/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trucos.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 58226    Output checksum: 40804
 */
