/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Original Author(s):    hmt, gthomas
 * Description:  hardware driver for 82559 Intel PRO/100+ ethernet on
 *               a Intel XScale IQ80310 development board
 *              
 * Modified by Andre Demarque and Jason Schmidlapp of Elmic Systems.
 * 
 * $Source: source/kernel/ucos/xscale/iq80310/gnu/iq80310.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:41:37CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
/*
 * -------------------------------------------
 * The contents of this file are subject to the Red Hat eCos Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.redhat.com/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * The Original Code is eCos - Embedded Configurable Operating System,
 * released September 30, 1998.
 *
 * The Initial Developer of the Original Code is Red Hat.
 *
 * Portions created by Red Hat are
 * Copyright (C) 1998, 1999, 2000 Red Hat, Inc.
 * All Rights Reserved.      
 * -------------------------------------------                              
 *
 * Portions of this software may have been derived from OpenBSD or other sources,
 * and are covered by the appropriate copyright disclaimers included herein.
 *
 * -------------------------------------------
 *
 * Original Author(s):    hmt, gthomas
 * Original Contributors: Ron Spence, Pacific Softworks; msalter
 * Date:         2000-02-01
 * Purpose:      
 * Description:  hardware driver for 82559 Intel PRO/100+ ethernet on
 *               a Intel XScale IQ80310 development board
 *              
 * Modified by Andre Demarque and Jason Schmidlapp of Elmic Systems.
 * 
 * Modification History
 * Date:     $Date: 2010/01/18 06:41:37CST $
 * Author:
 * Version:  $Revision: 6.0.2.2 $
 * Release:  $ProjectRevision: 6.0.1.23 $
 * Description:
 *
 */ 

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
#include <tr8255x.h>
#include "iq80310.h"

/* type definitions */

/* MDI Control Register */
typedef union
{
    struct
    {
        unsigned int data     : 16;  /* data to write or data read */
        unsigned int regAdd   : 5;   /* PHY register address */
        unsigned int phyAdd   : 5;   /* PHY address */
        unsigned int op       : 2;   /* opcode, 1 for write, 2 for read */
        unsigned int ready    : 1;   /* 1 = operation complete */
        unsigned int intEnab  : 1;   /* 1 = interrupt at end of cycle */
        unsigned int rsrv     : 2;   /* reserved */
    } bits;
    unsigned int word;
} MDICTL;


/* local function prototypes */
static int tf80310Delay(int delay);
static int tf80310ReadEeprom(long ioaddr, int location, int addr_len);
static int tf80310GetEepromSize(long ioaddr);
static void writeMDI (unsigned char  phyAdd,
                      unsigned char  regAdd,
                      unsigned short data);
static unsigned short readMDI ( unsigned char phyAdd, unsigned char regAdd);

/* local variable definitions */
extern tt8255xInfoPtr tlInfoPtr;

/* macro definitions */

/* Display debug information about 82559 EEPROM */
/* #define DEBUG_EE */

/* Dispaly debug information about 82559 PHY interface (MDI) */
#define DEBUG_MDI

/* PHY speeds */
#define SPEED_NOLINK  0 
#define SPEED_10M     10
#define SPEED_100M    100

/*
 * Delay for a while - used while reading from EEPROM.
 */ 
static int tf80310Delay(int delay)
{
  volatile int i;
/*
 * the loop is going to take 3 ticks.  At 600 MHz, to give uS, multiply
 * by 600/3 = 200. No volatile is needed on i; gcc recognizes delay
 * loops and does NOT elide them.  However, we include it, just in case...
 */ 
  for ( i = 200 * delay; i ; i--)
  {
    ;
  }
}

/*
 * Sample routines to access the 82559 EEPROM.
 */ 

/* Reads data from the 82559 EEPROM. */ 
static int tf80310ReadEeprom(long ioaddr, int location, int addr_len)
{
    unsigned short retval = 0;
    int read_cmd = location | EE_READ_CMD(addr_len);
    int i, tries = 10;

 try_again:
/* Should already be not-selected, but anyway: */
    tm_out_word(EE_ENB & ~EE_CS, TM_SCB_eeprom);
    tf80310Delay(EEPROM_ENABLE_DELAY);
    tm_out_word(EE_ENB, TM_SCB_eeprom);
    tf80310Delay(EEPROM_ENABLE_DELAY);
    
/* Shift the read command bits out, changing only one bit per time. */
    for (i = EE_TOP_CMD_BIT(addr_len); i >= 0; i--)
    {
        short dataval = (read_cmd & (1 << i)) ? EE_DATA_WRITE : 0;
        tm_out_word(EE_ENB | dataval               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB | dataval | EE_SHIFT_CLK, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB | dataval               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
    }

/* clear the dataval, leave the clock low */
    tm_out_word(EE_ENB, TM_SCB_eeprom);
    tf80310Delay(1);
    
    retval = tm_in_word(TM_SCB_eeprom);
/*
 * This should show a zero in the data read bit to confirm that the
 * address transfer is compelete.  If not, go to the start and try
 * again!
 */ 
    if ( (0 != (retval & EE_DATA_READ)) && (tries-- > 0) )
    {
/* Terminate the EEPROM access. */
        tm_out_word(EE_ENB & ~EE_CS, TM_SCB_eeprom);
        tf80310Delay(EEPROM_DONE_DELAY);
#ifdef DEBUG_EE
        printf( "Warning: Retrying EEPROM read word %d, address %x, try %d\n",
                   location,  TM_SCB_eeprom, tries+1 );
#endif
        goto try_again;
    }

/*
 * This fires with one device on one of the customer boards!
 * (but is OK on all other h/w.
 */ 
    if ( (EE_DATA_READ & retval) != 0 )
    {
#ifdef DEBUG_EE
        printf( "Read EEPROM: Dummy data bit not 0, reg %x\n" , retval );
#endif
    }
    tf80310Delay(1);
    retval = 0;

    for (i = EE_TOP_DATA_BIT; i >= 0; i--)
    {
        tm_out_word(EE_ENB | EE_SHIFT_CLK, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        retval =
            (retval << 1) | ((tm_in_word(TM_SCB_eeprom) & EE_DATA_READ) ? 1 : 0);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
    }
    
/* Terminate the EEPROM access. */
    tm_out_word(EE_ENB & ~EE_CS, TM_SCB_eeprom);
    tf80310Delay(EEPROM_DONE_DELAY);
    
    return retval;
}



/* Gets the EEPROM address size. */
static int tf80310GetEepromSize(long ioaddr)
{
    unsigned short retval = 0;
    int i, addrbits;

/* Should already be not-selected, but anyway: */
    tm_out_word(EE_ENB & ~EE_CS, TM_SCB_eeprom);
    tf80310Delay(EEPROM_ENABLE_DELAY);
    tm_out_word(EE_ENB, TM_SCB_eeprom);
    tf80310Delay(EEPROM_ENABLE_DELAY);
    
/* Shift the read command bits out. */
    for (i = 2; i >= 0; i--)
    {
        short dataval = (6 & (1 << i)) ? EE_DATA_WRITE : 0;
        tm_out_word(EE_ENB | dataval               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB | dataval | EE_SHIFT_CLK, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB | dataval               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
    }
/* Now clock out address zero, looking for the dummy 0 data bit */
    for ( i = 1; i <= 12; i++ )
    {
        tm_out_word(EE_ENB               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB | EE_SHIFT_CLK, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB               , TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY); 
        retval = tm_in_word(TM_SCB_eeprom) & EE_DATA_READ;
        if ( 0 == retval )
            break; /* The dummy zero est arrive' */
    }

#ifdef DEBUG_EE
    printf( "eeprom data bits %d (ioaddr %x)\n", i, TM_SCB_eeprom );
#endif
    if ( 6 != i && 8 != i )
    {
#ifdef DEBUG_EE
        printf( "*****EEPROM data bits not 6 or 8*****\n" );
#endif
        i = 6;
    }
    addrbits = i;

/* clear the dataval, leave the clock low to read in the data regardless */
    tm_out_word(EE_ENB, TM_SCB_eeprom);
    tf80310Delay(1);
    
    retval = tm_in_word(TM_SCB_eeprom);
    if ( (EE_DATA_READ & retval) != 0 )
    {
#ifdef DEBUG_EE
        printf( "Size EEPROM: Dummy data bit not 0, reg %x\n" , retval );
#endif
    }
    tf80310Delay(1);
    
    for (i = EE_TOP_DATA_BIT; i >= 0; i--)
    {
        tm_out_word(EE_ENB | EE_SHIFT_CLK, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        retval = tm_in_word(TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
        tm_out_word(EE_ENB, TM_SCB_eeprom);
        tf80310Delay(EEPROM_SK_DELAY);
    }
    
/* Terminate the EEPROM access. */
    tm_out_word(EE_ENB & ~EE_CS, TM_SCB_eeprom);
    tf80310Delay(EEPROM_DONE_DELAY);
    
    return addrbits;
}



/* Reads the Ethernet MAC address from the 82559 EEPROM. */ 
int tf80310GetPhysAddrFromEeprom(char TM_FAR * ethernetAddr)
{
    int  i;
    int  count;
/* (this is the length of the *EEPROM*s address, not MAC address) */
    int  addr_length;
    char mac_address[6];
    int  mac_addr_ok;
    int  errorCode;

    errorCode = TM_DEV_OKAY;
    mac_addr_ok = 0;


/* read eeprom and get 82559's mac address */
    addr_length = tf80310GetEepromSize((long) tlInfoPtr->i8255xMemAddrPtr);

/* (this is the length of the *EEPROM*s address, not MAC address)
 * We only need the first 6 bytes, but read the entire contents of the
 * EEPROM, as an example in case other info needs to be retrieved.
 */
    for (i = 0, count = 0; count < 64; count++)
    {
        unsigned short value;
        
/* read word from eeprom */
        value = tf80310ReadEeprom((unsigned long) tlInfoPtr->i8255xMemAddrPtr,
                                  count,
                                  addr_length);
        
#ifdef DEBUG_EE
        printf( "%2d: %04x\n", count, value );
#endif
        if (count < 3)
        {
            mac_address[i++] = value & 0xFF;
            mac_address[i++] = (value >> 8) & 0xFF;
        }
    }

#ifdef DEBUG_EE
    printf("MAC Address = %02X %02X %02X %02X %02X %02X\n",
           mac_address[0], mac_address[1],
           mac_address[2], mac_address[3],
           mac_address[4], mac_address[5]);
#endif

/* Return the MAC address. */
    ethernetAddr[0] = mac_address[0];
    ethernetAddr[1] = mac_address[1];
    ethernetAddr[2] = mac_address[2];
    ethernetAddr[3] = mac_address[3];
    ethernetAddr[4] = mac_address[4];
    ethernetAddr[5] = mac_address[5];

    return errorCode;
}


/*
 * Sample routines to access the PHY component, through the 82559's
 * Management Data Interface (MDI).
 */ 

/* Read from the PHY component */
static unsigned short readMDI ( unsigned char phyAdd, unsigned char regAdd)
{
    
    MDICTL mdiCtrl;
    int num_ms = 0; 

/* prepare for the MDI operation */
    mdiCtrl.bits.ready = MDI_NOT_READY;
    mdiCtrl.bits.intEnab = MDI_POLLED;  /* no interrupts */
    mdiCtrl.bits.op = MDI_READ_OP;
    mdiCtrl.bits.phyAdd = phyAdd & 0x1f;
    mdiCtrl.bits.regAdd = regAdd & 0x1f;

/* start the operation */
    tm_out_dword(mdiCtrl.word, TM_SCB_CtrlMDI);

/* delay a bit */
    tf80310Delay (1000);

/* poll for completion */
    mdiCtrl.word = tm_in_dword(TM_SCB_CtrlMDI);

    while ((mdiCtrl.bits.ready == MDI_NOT_READY) && (num_ms != 2000))
    {
/* wait max 2 secs */
        mdiCtrl.word = tm_in_dword(TM_SCB_CtrlMDI);
        tf80310Delay(1000);
        num_ms++;
    }   
        
    if (num_ms >= 2000)
    {
#ifdef DEBUG_MDI
        printf("readMDI Timeout!\n");
#endif /* DEBUG_MDI */
        return -1;
    }
    else
    {
        return (unsigned short)mdiCtrl.bits.data;
    }
}


/* Write to the PHY component */
static void writeMDI (unsigned char  phyAdd,
                      unsigned char  regAdd,
                      unsigned short data)
{
    register MDICTL mdiCtrl;
    int num_ms = 0;

/* prepare for the MDI operation */
    mdiCtrl.bits.ready = MDI_NOT_READY;
    mdiCtrl.bits.intEnab = MDI_POLLED;  /* no interrupts */
    mdiCtrl.bits.op = MDI_WRITE_OP;
    mdiCtrl.bits.phyAdd = phyAdd & 0x1f;
    mdiCtrl.bits.regAdd = regAdd & 0x1f;
    mdiCtrl.bits.data = data & 0xffff;

/* start the operation */
    tm_out_dword(mdiCtrl.word, TM_SCB_CtrlMDI);

/* delay a bit */
    tf80310Delay(1000);

/* poll for completion */
    mdiCtrl.word = tm_in_dword(TM_SCB_CtrlMDI);

    while ((mdiCtrl.bits.ready == MDI_NOT_READY) && (num_ms != 2000))
    {
        mdiCtrl.word = tm_in_dword(TM_SCB_CtrlMDI);
        tf80310Delay(1000);
        num_ms++;
    }   
    if (num_ms >= 2000)
    {
#ifdef DEBUG_MDI
        printf("writeMDI Timeout!\n");
#endif /* DEBUG_MDI */
    }
 
    return;
}

/*
 * Get the link speed (10,100,no link) from the PHY component.
 */ 
int tf80310FindLinkSpeed( void )
{

    unsigned long ioaddr;
    unsigned int  phy_id;
    unsigned short phy_addr_reg, temp1, temp2;
    int broadcom_flag = FALSE;
    int link_speed = SPEED_NOLINK;

/* get 82559's I/O address */
    ioaddr = (unsigned long) tlInfoPtr->i8255xMemAddrPtr; 

    phy_id = readMDI(MDI_DEFAULT_PHY_ADDR, MDI_PHY_ID_1) << 16;
    phy_id |= readMDI(MDI_DEFAULT_PHY_ADDR, MDI_PHY_ID_2);

    if ((phy_id & 0xfffffff0) == I82555_PHY_ID)
    {
#ifdef DEBUG_MDI
        printf("Intel 82555/558 PHY detected...\n");
#endif
/* dummy read for reliable status */
        (void)readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);
    
        temp1 = readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);

        phy_addr_reg = readMDI (MDI_DEFAULT_PHY_ADDR, I82555_STATCTRL_REG);

        if (temp1 & MDI_STAT_LINK) /* speed only valid with good LNK */
        {
            link_speed =
                (phy_addr_reg & I82555_100_MBPS) ? SPEED_100M : SPEED_10M;
        }   
#ifdef DEBUG_MDI
        else
        {
            printf("Connect Speed is NOT VALID phy_addr_reg =%x\n",phy_addr_reg);
        }
#endif
    }

    if ((phy_id & 0xfffffff0) == ICS1890_PHY_ID)
    {
#ifdef DEBUG_MDI
        printf ("Integrated Circuit Systems ICS1890 PHY detected...\n");
        printf ("Revision = %c\n", 'A' + (phy_id & REVISION_MASK));
#endif
/* dummy read for reliable status */
        (void)readMDI (MDI_DEFAULT_PHY_ADDR, ICS1890_QUICKPOLL_REG);
        temp1 = readMDI (MDI_DEFAULT_PHY_ADDR, ICS1890_QUICKPOLL_REG);

        if (temp1 & QUICK_LINK_VALID) /* speed only valid with good LNK */
        {
            link_speed = (temp1 & QUICK_100_MBPS) ? SPEED_100M : SPEED_10M;
        }
#ifdef DEBUG_MDI
        else
        {
            printf ("Connect Speed is NOT VALID\n");
        }
#endif
    }

    if ((phy_id & 0xfffffff0) == DP83840_PHY_ID)
    {
#ifdef DEBUG_MDI
        printf ("National DP83840 PHY detected...\n");
        printf ("Revision = %c\n", 'A' + (phy_id & REVISION_MASK));
#endif

/* dummy read for reliable status */
        (void)readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);
        temp1 = readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);

        phy_addr_reg = readMDI (MDI_DEFAULT_PHY_ADDR, DP83840_PHY_ADDR_REG);

        if (temp1 & MDI_STAT_LINK) /* speed only valid with good LNK */
        {
            link_speed =
                (phy_addr_reg & PHY_ADDR_SPEED_10_MBPS) ? SPEED_10M : SPEED_100M;
        }
#ifdef DEBUG_MDI
        else
        {
            printf("Connect Speed is NOT VALID\n");
        }
#endif
    }

    if ((phy_id & 0xfffffff0) == I82553_PHY_ID)
    {
#ifdef DEBUG_MDI
        printf("Intel 82553 PHY detected...\n");
        printf ("Revision = %c\n", 'A' + (phy_id & REVISION_MASK));
#endif
        broadcom_flag = TRUE;
    }

    if (phy_id == I82553_REVAB_PHY_ID)
    {
#ifdef DEBUG_MDI
        printf("Intel 82553 PHY detected...\n");
        printf("Revision = B\n");
#endif
        broadcom_flag = TRUE;
    }

    if (broadcom_flag == TRUE)
    {
        temp2 = readMDI (MDI_DEFAULT_PHY_ADDR, I82553_PHY_EXT_REG0);

/* dummy read for reliable status */
        (void)readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);
        temp1 = readMDI (MDI_DEFAULT_PHY_ADDR, MDI_PHY_STAT);

        if (temp1 & MDI_STAT_LINK) /* speed only valid with good LNK */
        {
            link_speed = (temp2 & EXT_REG0_100_MBPS) ? SPEED_100M : SPEED_10M;
#ifdef DEBUG_MDI
        }
        else
        { 
           printf("Connect Speed is NOT VALID\n");
#endif
        }
    }

    return link_speed;
}



/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: iq80310.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 15952    Output checksum: 56139
 */
