@;/**********************************************************************
@; $Source: source/kernel/ucos/xscale/iq80310/gnu/os_cpu_a.s $
@; Author: Jason
@; 
@; Project: LH79531 EVB
@;
@; Description:
@;      uC/OS-II V2.03 LH79531 processor specific assembly code
@;
@;      Functions defined in this module:
@; 
@;      void IRQContextSwap (void)   /* perform interrupt level context switch */
@;      void OS_TASK_SW(void)        /* task level context switch */
@;      void ARMDisableInt(void)     /* disable interrupts when in SVC */
@;      void ARMEnableInt(void)      /* enable interrupts when in SVC */
@;      void OSStartHighRdy(void)    /* start highest priority task */
@; 
@; Notes:
@;   The code herein assumes that all threads use r13 as
@;   the stack-pointer, and that it is a APCS conformant stack.
@;   i.e. there is never any data stored beneath the current
@;   stack-pointer.
@;   The above needs to be true to enable the context switches
@;   started as a return from an interrupt to use the current
@;   threads stack as the state save area.
@; 
@; Modification History
@;   $Revision: 6.0.2.2 $
@;   $Date: 2010/01/18 06:41:44CST $
@;   $Author: odile $
@;   $ProjectName: /home/mks/proj/tcpip.pj $
@;   $ProjectRevision: 6.0.1.23 $
@; 
@;    Rev 1.1   Sep 05 2001 11:23:46   BarnettH
@; Revised IRQContextSwap to more accurately preserve
@; condition code flags over a context switch, and eliminate
@; some instructions.
@; 
@;    Rev 1.0   Sep 08 2000 08:31:56   BarnettH
@; Initial revision.
@;
@;  COPYRIGHT (C) 2000 SHARP MICROELECTRONICS OF THE AMERICAS, INC.
@;                     CAMAS, WA 
@;********************************************************************/

@       ; Equates
#define IRQNoInt    0xC0
#define ModeMask    0x1F

#define SVC32Mode   0x13
#define IRQ32Mode   0x12
#define FIQ32Mode   0x11

@       AREA    |subr|, CODE, READONLY

@       ; External symbols needed by the module
        .extern OSTCBCur
        .extern OSTCBHighRdy
        .extern OSPrioCur
        .extern OSPrioHighRdy
        .extern OSRunning
        .extern OSTaskSwHook
        .extern iq80310_irq_handler
        .extern iq80310_fiq_handler
        .extern swap_context_flag

@       ; Export the functions needed by other modules
@       EXPORT  OS_TASK_SW
@       EXPORT  OSStartHighRdy
@       EXPORT  RB_IRQHandler
@       EXPORT  ARMDisableInt
@       EXPORT  ARMEnableInt

@;/***********************************************************************
@;
@; Function: IRQContextSwap
@;
@; Purpose:
@;      To perform a context switch from the interrupt level.
@;
@; Processing:
@;    See uC/OS-II Interrupt Level Context Switch flow chart
@;
@; Parameters: void
@;
@; Outputs:  None
@;
@; Returns:  void
@;
@; Notes:
@;   Sets up the stacks and registers to call the task level
@;   context switch
@;
@;*********************************************************************/
        .global IRQContextSwap
IRQContextSwap:
        @;
        LDMFD   sp!,{r12}               @; pop SPSR value from stack
        @; if NE then we need to check if we were a nested interrupt
        AND     r11,r12,#0x1f           @; mask out all but the mode bits
        TEQNE   r11,#0x12               @; check for interrupted IRQ thread
        @; if EQ then we can return immediately
        MSREQ   SPSR_cf,r12             @; restore the SPSR
        LDMEQFD sp!,{r0-r12,pc}^        @; and return to the interrupted thread

        @; We need to perform a context switch now.
        @; r12 = SPSR describing the interrupted thread.
        @; r11 = interrupted thread processor mode

        @; We rely on the interrupted thread having IRQs enabled (since we
        @; would never have reached this point otherwise).
        @;
        @; We have recovered the SPSR value,
        @; so only r0-r12,lr are on the stack.

        LDR     r4,[sp,#(13 * 4)]       @; get return address

        @; r12 contains the CPSR (spsr_irq) describing the interrupted thread
        MRS     r0,CPSR                 @; save current mode info.
        ORR     r1,r12,#0xc0            @; disable IRQ in interrupted SPSR
        MSR     CPSR_cf,r1              @; and change to interrupted thread mode

@       ; We are now in the interrupted thread mode with IRQs disabled.

        MRS     r1,SPSR                 @; copy current SPSR
        MOV     r2,r12                  @; copy interrupted thread SPSR as CPSR
        MOV     r3,lr                   @; copy the current lr
        STMFD   sp!,{r1,r2,r3,r4}       @; and construct return stack
        MSR     CPSR_cf,r0              @; return to IRQ mode

        @; We are now in IRQ mode, IRQs disabled
        @; r12 = SPSR describing interrupted thread (interrupts enabled)
        ORR	r12,r12,#0xc0	
        MSR     SPSR_cf,r12             @; restore SPSR_irq ready for return
        LDMFD   sp!,{r0-r12,lr}         @; restore all the registers    
        SUBS    pc,pc,#0                @; and return to the interrupted mode
        NOP                             @; flush the pipeline (3 NOPS for 7TDMI)
        NOP
        NOP

        @; we are now executing in the interrupted mode with IRQs enabled
        BL      OS_TASK_SW              @; perform the context switch

        LDMFD   sp!,{lr}                @; pop SPSR
        MSR     SPSR_cf,lr              @; recover the SPSR when the thread 
                                        @; was interrupted
        LDMFD   sp!,{lr}                @; pop CPSR
        MSR     CPSR_cf,lr              @; recover the CPSR when the thread 
                                        @; was interrupted
        LDMFD   sp!,{lr,pc}             @; return to the interrupted thread
        @;
        @; End of IRQContextSwap
        @;

@;/***********************************************************************
@;
@; Function: OS_TASK_SW 
@;
@; Purpose:
@;      To perform a context switch from the Task Level.
@;
@; Processing:
@;    See uC/OS-II Task Level Context Switch flow chart
@;
@; Parameters: void
@;
@; Outputs:  None
@;
@; Returns:  void
@;
@; Notes:
@;   On entry, OSTCBCur and OSPrioCur hold the current TCB and priority
@;   and OSTCBHighRdy and OSPrioHighRdy contain the same for the task
@;   to be switched to.
@; 
@;   The following code assumes that the virtual memory is directly
@;   mapped into  physical memory. If this is not true, the cache must 
@;   be flushed at context switch to avoid address aliasing.
@;
@;*********************************************************************/

        .global OS_TASK_SW
OS_TASK_SW:
        STMFD   sp!, {r0-r12, lr}       @; push register file and ret address
        MRS     r4, CPSR
        STMFD   sp!, {r4}               @; push CPSR
        MRS     r4, SPSR
        STMFD   sp!, {r4}               @; push SPSR

        @; OSPrioCur = OSPrioHighRdy
        LDR     r4, =OSPrioCur
        LDR     r5, =OSPrioHighRdy
        LDRB    r6, [r5]
        STRB    r6, [r4]
        
        @; Get current task TCB address
        LDR     r4, =OSTCBCur
        LDR     r5, [r4]
        STR     sp, [r5]                @; store sp in preempted tasks s TCB

        BL      OSTaskSwHook            @; call Task Switch Hook

        @; Get highest priority task TCB address
        LDR     r6, =OSTCBHighRdy
        LDR     r6, [r6]
        LDR     sp, [r6]                @; get new task s stack pointer

        @; OSTCBCur = OSTCBHighRdy
        STR     r6, [r4]                @; set new current task TCB address
        LDMFD   sp!, {r4}
        MSR     SPSR_cf, r4     
        LDMFD   sp!, {r4}
        MSR     CPSR_cf, r4     
        LDMFD   sp!, {r0-r12, pc}
        @;
        @; End of OS_TASK_SW
        @;

@;/***********************************************************************
@;
@; Function: OSStartHighRdy
@;
@; Purpose:
@;         To start the task with the highest priority during OS startup
@;
@; Processing:
@;    See uC/OS-II Task Level Context Switch flow chart
@;
@; Parameters: void
@;
@; Outputs:  None
@;
@; Returns:  void
@;
@; Notes:
@;   Called once during OSStart()
@;
@;*********************************************************************/

        .global OSStartHighRdy
OSStartHighRdy:
        @; Call OSTaskSwHook
        BL      OSTaskSwHook
        @; Set OSRunning to TRUE
        LDR     r4, =OSRunning
        MOV     r5, #1
        STRB    r5, [r4]
        @; Get current task TCB address
        LDR     r4, =OSTCBCur
        @; Get highest priority task TCB address
        LDR     r5, =OSTCBHighRdy
        LDR     r5, [r5]                @; get stack pointer
        LDR     sp, [r5]                @; switch to the new stack

        STR     r5, [r4]                @; set new current task TCB address

        LDMFD   sp!, {r4}               @; get new SPSR from top of stack
        MSR     SPSR_cf, r4             @; SPSR should be SVC32Mode
        LDMFD   sp!, {r4}               @; get new state from top of stack
        MSR     CPSR_cf, r4             @; CPSR should be SVC32Mode
        LDMFD   sp!, {r0-r12, pc}       @; start the new task
        @;
        @; End of OSStartHighRdy
        @;
@/***********************************************************************
@;
@; Function: IQ80310_IRQHandler
@;
@; Purpose:
@;         The LH79531 EVB (RainBow) IRQ interrupt handler
@;
@; Processing:
@;    Saves context
@;    Calls the IRQ dispatcher
@;    Checks if context switch necessary
@;    If not, restores context and returns from interrupt
@;    If switch required, branches without link to IRQContextSwap
@;       which performs context switch if interrupts not nested
@;       and returns from interrupt to new context
@;
@; Parameters: void
@;
@; Outputs:  None
@;
@; Returns:  void
@;
@; Notes:
@;   (1) IRQContextSwap does not return to this handler.  The RFI is
@;   performed directly by IRQContextSwap
@;
@*********************************************************************/

	
                .global IQ80310_IRQHandler
IQ80310_IRQHandler:
        /* Interrupt Handler Prologue */
        SUB     lr, lr, #4
        STMFD   sp!, {r0-r12, lr}       /* push registers and return address */
        MRS     r4, SPSR                /* get SPSR */
        STMFD   sp!, {r4}               /* push SPSR */
        /* End Interrupt Handler Prologue */

        /* Call main IRQ service dispatcher */
        BL      iq80310_irq_handler

        LDR     r12, =swap_context_flag 
        LDR     r11, [r12]
        TST     r11, #1
        BEQ     .L99 
        MOV     r11, #0
        STR     r11, [r12]
        B       IRQContextSwap

        /* Interrupt Handler Epilogue */
.L99:   LDMFD   sp!, {r4}               /* pop SPSR */  
        MSR     SPSR_cf, r4             /* restore SPSR*/
        LDMFD   sp!, {r0-r12, pc}^      /* restore registers and RFI */
        /* End Interrupt Handler Epilogue */
        /* */
        /* End of IRQHandler */
        

        .globl IQ80310_FIQHandler
IQ80310_FIQHandler:
        /* Cyclone FIQ handler */
        /* save registers onto stack */
        ldr     sp, =__fiq_stack        /* point stack pointer at FIQ stack */
        sub     lr, lr, #4              /* adjust link register (return to 
	                                   last address) */
        stmfd   sp!, {r12, lr}          /* push r12 and link reg onto stack */
        mrs     r12, spsr               /* store spsr in r12 */
        stmfd   sp!, {r12}              /* push spsr onto stack */
        stmfd   sp!, {r0-r7}            /* push r0 thru r7 (r8 - r14 are 
	                                   banked) */
        bl      iq80310_fiq_handler
        /* restore registers and return */
        ldmfd   sp!, {r0-r7}            /* restore r0 thru r7 (r8 - r14 are 
	                                   banked) */
        ldmfd   sp!, {r12}
        msr     spsr, r12               /* restore SPSR */
        ldmfd   sp!, {r12,pc}^          /* restore r12 and PC, return to inst 
	                                   before exception occurred */

	
/*;/************************************************************************/
@;
@; Functions: DisableInt
@;                EnableInt
@;
@; Purpose:
@;    Disable and enable IRQ and FIQ preserving current CPU mode.
@;
@; Processing:
@;    Push the CPSR onto the stack
@;    Disable IRQ and FIQ interrupts
@;    Return 
@;
@; Parameters: void
@;
@; Outputs:  None
@;
@; Returns:  void
@;
@; Notes:
@;   (1) Can be called from SVC mode to protect Critical Sections. 
@;   (2) Do not use these calls at interrupt level.
@;   (3) Used in pairs within the same function level;
@;   (4) Will restore interrupt state when called; i.e., if interrupts
@;       are disabled when DisableInt is called, interrupts will still
@;       still be disabled when the matching EnableInt is called.
@;   (5) Uses the method described by Labrosse as "Method 2".
@;
@;*********************************************************************/

	.global ARMDisableInt
ARMDisableInt:
	MRS	r1, CPSR
	ORR	r1, r1, #0xc0           @; mask IRQ Int s
	MSR	CPSR_cf, r1		@; disable IRQ Int s
	MOV	pc, lr
    
	@;
	@; End of ARMDisableInt
	@;

	.global ARMEnableInt
ARMEnableInt:
	MRS	r1, CPSR
	BIC	r1, r1, #0xc0
	MSR	CPSR_cf, r1		@; restore original CPSR
	MOV	pc, lr
	@;
	@; End of ARMEnableInt

/*
 * Low-lebvel interrupt support for IQ80310 diags
 */
        .extern iq80310_irq_handler
        .extern iq80310_fiq_handler


        .globl __ignore_abort
__ignore_abort:
        subs    pc,lr,#4

        .globl _cspr_enable_fiq_int
_cspr_enable_fiq_int:
        mrs     r0, cpsr
        bic     r0, r0, #0x40
        msr     cpsr, r0
        mov     pc, lr

        .globl _cspr_enable_irq_int
_cspr_enable_irq_int:
        mrs     r0, cpsr
        bic     r0, r0, #0x80
        msr     cpsr, r0
        mov     pc, lr

        .bss
        .rept 1024
        .word 0
        .endr
  __irq_stack:
        .rept 1024
        .word 0
        .endr
  __fiq_stack:
	.rept 1024
	.word 0
	.endr

        .global _flushICache
_flushICache:
        mcr     p15, 0, r0, c7, c5, 0;  /* invalidate the I cache and BTB */
        mcr     p15, 0, r0, c8, c5, 0;  /* invalidate the I cache TLB */
        mov     pc, lr;                 /* return */

@       END
@       ;
@       ; End of os_cpu_a.s
@       ;
