/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: USX Kernel Interface
 * Filename: trusx.c
 * Author: Odile
 * Date Created: 08/28/98
 * $Source: source/kernel/usx/trusx.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:12CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


#include <trsocket.h>
#include <trmacro.h>
#include <mtlib.h>

/*
 * Local macros
 */

#define TM_KERN_OKAY   0
#define TM_KERN_ERROR -1

#define TM_NUMCOUNTSEM       NUMTSK

#define TM_EVENT_INVALID    -1

#define TM_NUMEVENT         NUMTSK

static unsigned int         tlEvtIndex;

int tfGetNextEvt (void);

/*
 * Local types
 */


typedef struct tsTaskEntry
{
/*
 * Pointer to next task waiting on counting semaphore.
 * Has to be first entry
 */
    struct tsTaskEntry      TM_FAR * tskNextPtr;
/* 0 .. 255 */
    unsigned short                   tskPri;
/* 0 .. 255 */
    short                            tskEvent;
} ttTaskEntry;

typedef ttTaskEntry TM_FAR * ttTaskEntryPtr;

typedef struct tsCountSemEntry
{
/*
 * Pointer to queue of tasks waiting on counting semaphore.
 * Has to be first entry
 */
    ttTaskEntryPtr       csTaskQueuePtr;
/* count value of counting semaphore */
    short                csCount;
} ttCountSemEntry;

typedef ttCountSemEntry TM_FAR * ttCountSemEntryPtr;

typedef void           TM_FAR  * ttVoidPtr;

/*
 * Local variables
 */

/* Table of tasks. Task ID is index */
ttTaskEntry        tlTaskArray[NUMTSK];

/* Table of counting semaphores. Counting semaphore ID is index */
ttCountSemEntry    tlCountSemArray[TM_NUMCOUNTSEM];

/*
 * Counting Sempahore index (starting search of next available counting
 * sempahore ID)
 */
unsigned int       tlCountSemIndex;

static char tlErrorFunction[80];
static char tlErrorType[80];
static char tlWarnFunction[80];
static char tlWarnType[80];

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    return (void TM_FAR *)(reqmem((GLOBAL | COLOR0), size ));
}

/*
 * Free a memory block of size back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    relmem(memoryBlockPtr);
}


/*
 * Stop all other calls into Treck by
 * a kernel context switch
 */
void tfKernelSetCritical()
{
  MASK_INTS();
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical()
{
  UNMASK_INTS;
}

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
    char TM_FAR *ptr;

    tm_kernel_set_critical;
    ptr=tlErrorFunction;
    while(*functionName)
    {
        *ptr++ = *functionName++;
    }
    ptr=tlErrorType;
    while(*errorMessage)
    {
        *ptr++ = *errorMessage++;
    }
    tm_kernel_release_critical;

    tm_thread_stop;
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *warnMessage)
{
    char TM_FAR *ptr;

    tm_kernel_set_critical;
    ptr = tlWarnFunction;
    while (*functionName)
    {
        *ptr++ = *functionName++;
    }
    ptr = tlWarnType;
    while (*warnMessage)
    {
        *ptr++ = *warnMessage++;
    }
    tm_kernel_release_critical;
}


/*
 * Create an event structure for pend/post from an ISR
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    eventPtr->genIntParm = tfGetNextEvt();
}

/*
 * Wait on an Event from an ISR
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    wteset_dec((unsigned int)eventPtr->genIntParm, 0);
}

/*
 * Resume waiting tasks that ware waiting on this event
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    MTqcmd_c(INCEVT, (unsigned int)eventPtr->genIntParm);
}

/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */

/*
 * Create a counting semaphore. Search for a free (not in use) slot in
 * the table of counting semaphores.
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    unsigned int    countSemId;
    int             retCode;

    retCode = TM_KERN_ERROR;
    tm_kernel_set_critical;
    countSemId = tlCountSemIndex++;
    tm_kernel_release_critical;
    if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
    {
        countingSemaphore->genIntParm = (int)countSemId;
        retCode = TM_KERN_OKAY;
    }
#ifdef TM_ERROR_CHECKING
    else
    {
        tfKernelError("tfKernelCreateCountSem", "No more counting semaphore");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;

}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem (ttUserGenericUnionPtr countingSemaphore)
#ifdef TM_USE_PEND_TIMEOUT
{
    return tfKernelPendCountSemTimeout(countingSemaphorePtr, TM_UL(0));
}

/*
 * Wait on a counting semaphore. If counting semaphore is zero, we
 * queue our task to the counting semaphore, and wait on the task
 * event. If counting semaphore is non zero, we get the semaphore
 * (decrease it by one), and return.
 */
int tfKernelPendCountSemTimeout ( ttUserGenericUnionPtr countingSemaphorePtr, 
                                  ttUser32Bit tickTimeout )
#endif /* TM_USE_PEND_TIMEOUT */
{
    ttCountSemEntryPtr countSemPtr;
    ttTaskEntryPtr     taskPtr;
    ttTaskEntryPtr     taskQueuePtr;
    ttTaskEntryPtr     prevTaskQueuePtr;
    unsigned int       countSemId;
    int                retCode;
    unsigned short     priority;

    retCode = TM_KERN_ERROR;
/* Map current task to task pointer in our task array */
    if ( (unsigned int)cur_task < (unsigned int)NUMTSK )
    {
        taskPtr = &tlTaskArray[cur_task];
        if (taskPtr->tskEvent == TM_EVENT_INVALID)
        {
            taskPtr->tskEvent = (short)tfGetNextEvt();
        }
/* Counting semaphore ID */
        countSemId = (unsigned int)countingSemaphore->genIntParm;
        if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
        {
            countSemPtr = &tlCountSemArray[countSemId];
            tm_kernel_set_critical;
            if (countSemPtr->csCount > 0)
            {
/* Decrease the counting semaphore. We are granted access */
                countSemPtr->csCount--;
                tm_kernel_release_critical;
                retCode = TM_KERN_OKAY;
            }
            else
            {
/*
 * We need to pend. We will insert our task into the counting semaphore
 * queue in priority order.
 */
                taskQueuePtr = countSemPtr->csTaskQueuePtr;
                if ( taskQueuePtr == (ttTaskEntryPtr)0 )
                {
/* Empty counting semaphore task queue */
                    countSemPtr->csTaskQueuePtr = taskPtr;
                    taskPtr->tskNextPtr = (ttTaskEntryPtr)0;
                }
                else
                {
                    priority = taskPtr->tskPri;
                    prevTaskQueuePtr = (ttTaskEntryPtr)countSemPtr;
                    while (    (taskQueuePtr != (ttTaskEntryPtr)0)
                            && (priority <= taskQueuePtr->tskPri) )
                    {
                        prevTaskQueuePtr = taskQueuePtr;
                        taskQueuePtr = taskQueuePtr->tskNextPtr;
                    }
/* Insert in counting semaphore task queue */
                    taskPtr->tskNextPtr = prevTaskQueuePtr->tskNextPtr;
                    prevTaskQueuePtr->tskNextPtr = taskPtr;
                }
                tm_kernel_release_critical;
#ifdef TM_USE_PEND_TIMEOUT
                if (tickTimeout != TM_UL(0))
                {
                    retCode = wteset_dec(&taskPtr->tskEvent, tickTimeout);
                    if (retCode != TM_KERN_OKAY)
/* Timeout occured. Remove counting semaphore from the task queue */
                    {
                        retCode = TM_ETIMEDOUT;
                        tm_kernel_set_critical;
                        taskQueuePtr = countSemPtr->csTaskQueuePtr;
                        if (taskQueuePtr == taskPtr) 
                        {
/* At head of queue, special case: head of queue point to next entry */
                            countSemPtr->csTaskQueuePtr = 
                                                    taskPtr->tskNextPtr; 
                        }
                        else
                        {
                            while (taskQueuePtr != (ttTaskEntryPtr)0)
                            {
                                prevTaskQueuePtr = taskQueuePtr;
                                taskQueuePtr = taskQueuePtr->tskNextPtr;
                                if (taskQueuePtr == taskPtr) 
                                {
/* Remove from queue, using previous pointer */
                                    prevTaskQueuePtr->tskNextPtr = 
                                                          taskPtr->tskNextPtr;
                                    break;
                                }
                            }
                        }
                        tm_kernel_release_critical;
                    }
                }
                else
#endif /* TM_USE_PEND_TIMEOUT */
                {
                    (void)wteset_dec((unsigned int)taskPtr->tskEvent, 0);
                    retCode = TM_KERN_OKAY;
                }
            }
        }
#ifdef TM_ERROR_CHECKING
        else
        {
/* Should not happen */
            tfKernelError("tfKernelPendCountSem",
                          "Invalid Counting Semaphore");
        }
#endif /* TM_ERROR_CHECKING */
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPendCountSem",
                      "Invalid current task ID");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem (ttUserGenericUnionPtr countingSemaphore)
{
    ttCountSemEntryPtr  countSemPtr;
    ttTaskEntryPtr      taskQueuePtr;
    unsigned int        countSemId;
    int                 retCode;

    retCode = TM_KERN_ERROR;
    countSemId = (unsigned int)countingSemaphore->genIntParm;
    if ( countSemId < (unsigned int)TM_NUMCOUNTSEM )
    {
        countSemPtr = &tlCountSemArray[countSemId];
        tm_kernel_set_critical;
        taskQueuePtr = countSemPtr->csTaskQueuePtr;
/* If no task is waiting on the semaphore yet. Just increase the count */
        if (taskQueuePtr == (ttTaskEntryPtr)0)
        {
            countSemPtr->csCount++;
            tm_kernel_release_critical;
        }
        else
/* Remove and Reschedule task at head of the queue (highest priority) */
        {
/* Remove task at head of the queue */
            countSemPtr->csTaskQueuePtr = taskQueuePtr->tskNextPtr;
            tm_kernel_release_critical;
/* Move task to run queue */
            (void)incevt((unsigned int)taskQueuePtr->tskEvent);
        }
        retCode = TM_KERN_OKAY;
    }
#ifdef TM_ERROR_CHECKING
    else
    {
/* Should not happen */
        tfKernelError("tfKernelPostCountSem", "Invalid Counting Semaphore");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * Delete a counting semaphore
 * Unused function by TRECK since semaphores are recycled internally
 */
int tfKernelDeleteCountSem (ttUserGenericUnionPtr countingSemaphore)
{
    int             retCode;

    TM_UNREF_IN_ARG(countingSemaphore);
    retCode = TM_KERN_ERROR;
    return retCode;
}

void  tfKernelInitialize (void)
{
}

/*
 * tfTaskInit need to be called prior to starting any Treck task and
 * prior to calling tfStartTreck()
 */
void tfTaskInit (void)
{
    unsigned int i;

    for (i = 0; i < NUMTSK; i++)
    {
        tlTaskArray[i].tskNextPtr = (ttTaskEntryPtr)0;
        tlTaskArray[i].tskPri = (unsigned int)0;
        tlTaskArray[i].tskEvent = TM_EVENT_INVALID;
    }
    tm_bzero(&tlCountSemArray[0], sizeof(tlCountSemArray));
    tlCountSemIndex = 0;
    tlEvtIndex = 0;
}

/*
 * tfTaskInitPriority need to be called after starting each task.
 */
int tfTaskInitPriority (unsigned short taskId, unsigned short taskPri)
{
    int retCode;

    retCode = TM_KERN_ERROR;
    if ( taskId < (unsigned short)NUMTSK )
    {
        tlTaskArray[taskId].tskPri = taskPri;
        retCode = TM_KERN_OKAY;
    }
#ifdef TM_ERROR_CHECKING
    else
    {
        tfKernelError("tfTaskSetPriority", "Invalid Task ID");
    }
#endif /* TM_ERROR_CHECKING */
    return retCode;
}

/*
 * This should be made common code if any other non treck task use
 * events
 */
int tfGetNextEvt (void)
{
    int evt;

    tm_kernel_set_critical;
    if (tlEvtIndex < (unsigned int)TM_NUMEVENT)
    {
        evt = tlEvtIndex++;
    }
#ifdef TM_KERNEL_ERROR
    else
    {
        evt = TM_EVENT_INVALID;
        tfKernelError("tfGetNextEvt", "No more event");
    }
#endif /* TM_KERNEL_ERROR */
    tm_kernel_release_critical;
    return evt;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trusx.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 46644    Output checksum:  8115
 */
