/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks for Win32
 *
 * Filename: trwin32.c
 * Author: Bryan Wang
 * Date Created: 08/15/2002
 * $Source: source/kernel/win32/intelx86/trwin32.c $
 *
 * Modification History
 * $Revision: 6.0.2.5 $
 * $Date: 2011/02/14 16:41:03CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#ifdef _WIN32
/* compile and run on Win32 platform */
/* to avoid an warning in MS VC++ header file */
struct _RPC_ASYNC_STATE;
#include <windows.h>
#include <time.h>
#endif /* _WIN32 */


#ifdef TM_TRECK_DEMO
    void tfxSendMail(
        char * from,
        char * to,
        char * subject,
        char * body);
#endif /* TM_TRECK_DEMO */


#include <stdio.h>
#include <trsocket.h>

#ifdef TM_KERNEL_WIN32_X86
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


CRITICAL_SECTION tlCriticalSection;

void  tfKernelInitialize(void)
{
#ifdef TM_MULTIPLE_CONTEXT
/*
 * Allocate the Thread local storage index so that threads can store 
 * their context handle into it, and retrive it later when a context
 * global is needed.
 */
    tvTlsIndex = TlsAlloc();
    tm_assert(tfKernelInitialize, tvTlsIndex != -1)
#endif /* TM_MULTIPLE_CONTEXT */

    InitializeCriticalSection(&tlCriticalSection);
}

void  tfKernelUnInitialize(void)
{
    return;
}

#ifdef TM_DYNAMIC_CREATE_SHEAP
/* 
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    return (ttUser32Bit TM_FAR *)(malloc((*sheapSizePtr)));
}

#ifdef TM_USE_STOP_TRECK
void tfKernelSheapFree (ttUser32Bit TM_FAR * sheapPtr)
{
    free(sheapPtr);
    return;
}
#endif /* TM_USE_STOP_TRECK */

#endif /* TM_DYNAMIC_CREATE_SHEAP */

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical()
{
    EnterCriticalSection(&tlCriticalSection);
}


/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical()
{
    LeaveCriticalSection(&tlCriticalSection);
}

#ifndef TM_USE_SHEAP
void TM_FAR * tfKernelMalloc(unsigned int size)
{
    ttCharPtr allocPtr;

    tfKernelSetCritical();
    allocPtr = malloc(size);
    tfKernelReleaseCritical();
    return (void TM_FAR *)allocPtr;
}

void tfKernelFree(ttCharPtr allocPtr)
{
    tfKernelSetCritical();
    free(allocPtr);
    tfKernelReleaseCritical();
}
#endif /* TM_USE_SHEAP */

int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    void *              semaphoreHandle;
    int                 retCode;

    retCode = TM_KERN_OKAY;

    /* Initialize the semaphore to zero */
    semaphoreHandle = CreateSemaphore(
                            NULL,   /* Security policy  */ 
                            0,      /* initial count    */
                            1,      /* maximum count    */
                            NULL);  /* object name      */

    if (semaphoreHandle != 0)
    {
        countingSemaphore->genVoidParmPtr = semaphoreHandle;
    }
    else
    {
        tfKernelError( "tfKernelCreateCountSem",
                       "Unable to Create Windows Semaphore");
        retCode = TM_KERN_ERROR;
    }

    return retCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int retCode;

    retCode = TM_KERN_OKAY;

    retCode = CloseHandle(countingSemaphore->genVoidParmPtr );
    if (retCode != 0)
    {
        countingSemaphore->genVoidParmPtr = 0 ;
    }
    else
    {
        retCode = TM_KERN_ERROR;
    }
    return retCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int errorCode;
    int retCode;

    retCode = TM_KERN_OKAY;

    errorCode = WaitForSingleObject(
                    countingSemaphore->genVoidParmPtr,/* handle to object  */
                    INFINITE);                        /* time-out interval */ 

    if (errorCode == WAIT_FAILED)
    {
        retCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPendCountSem",
                       "Unable to Pend on Windows Semaphore");
    }

    return retCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to a timeout value in ticks.
 */
int tfKernelPendCountSemTimeout( ttUserGenericUnionPtr countingSemaphore, 
                          ttUser32Bit timeOut )
{
    int errorCode;
    int retCode;

    retCode = TM_KERN_OKAY;

/* Convert timeout in milliseconds */
    timeOut = timeOut * tvTimerTickLength;
    errorCode = WaitForSingleObject(
                    countingSemaphore->genVoidParmPtr,/* handle to object  */
                    timeOut);                        /* time-out interval */ 

    if (errorCode == WAIT_FAILED)
    {
        retCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPendCountSemTimeout",
                       "Unable to Pend on Windows Semaphore");
    }
    else
    {
        if (errorCode == WAIT_TIMEOUT)
        {
            retCode = TM_ETIMEDOUT;
        }
    }
    return retCode;
}
#endif /* TM_USE_PEND_TIMEOUT */


/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    int     retCode;
    int     errorCode;

    retCode = TM_KERN_OKAY;
    errorCode = ReleaseSemaphore(
                    countingSemaphore->genVoidParmPtr, /* semaphore handle */
                    1,                                 /* count increment  */
                    NULL);                             /* previous count   */
    
    if (errorCode == 0)
    {
        retCode = TM_KERN_ERROR;
    }

    return retCode;
}


/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
    eventPtr->genVoidParmPtr = 
            CreateEvent(NULL, FALSE, FALSE, (void*)0);
    tm_assert(tfKernelCreateEvent, eventPtr->genVoidParmPtr != 0);
}

#ifdef TM_USE_STOP_TRECK
/*
 * Delete an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
int tfKernelDeleteEvent(ttUserGenericUnionPtr eventPtr)
{
    int errorCode;

    tm_assert(tfKernelDeleteEvent, eventPtr != 0);
    tm_assert(tfKernelDeleteEvent, eventPtr->genVoidParmPtr != 0);
    errorCode = CloseHandle(eventPtr->genVoidParmPtr);
    if (errorCode != 0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    return errorCode;
}
#endif /* TM_USE_STOP_TRECK */

/*
 * Task wait on an Event
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
    int errorCode;
    errorCode = WaitForSingleObject(eventPtr->genVoidParmPtr, INFINITE);
    tm_assert(tfKernelPendEvent, errorCode != WAIT_FAILED);
}

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
    int errorCode;
    errorCode = SetEvent(eventPtr->genVoidParmPtr);
    tm_assert(tfKernelTaskPostEvent, errorCode != 0);
}

/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
    int errorCode;
    errorCode = SetEvent(eventPtr->genVoidParmPtr);
    tm_assert(tfKernelTaskPostEvent, errorCode != 0);
}

/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
    Sleep(1);
}

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
#ifdef TM_TRECK_DEMO
/* if this is a treck demo, show the kernel error in a message box */
    char    error[256];
    char    msg[256];

    strcpy(error, functionName);
    strcat(error, "() -- ");
    strcat(error, errorMessage);
    strcat(error, "\n");


    strcpy(msg, "This following Treck internal error has occured.\n");
    strcat(msg, "Please report this to support@treck.com\n");
    strcat(msg, error);

    MessageBox(NULL, 
                msg, 
                "Treck Stack Kernel Error",
                MB_OK | MB_ICONEXCLAMATION);
    PostQuitMessage(1);

#else /* TM_TRECK_DEMO */
    (void)tfKernelPrint(TM_PRINT_TYPE_FAIL,
                        "tfKernelError! %s: %s\n",
                        functionName,
                        errorMessage);
    tm_thread_stop;
#endif /* TM_TRECK_DEMO */
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
    char    msg[256];

    strcpy(msg, functionName);
    strncat(msg, errorMessage, 256 - strlen(functionName));

#ifndef TM_IGNORE_WARNING

#if (defined(_AFXDLL) && defined(_DEBUG))
/* if this is a MFC application, show the kernel error in a message box */
    MessageBox(NULL, msg, "treck kernel warning", MB_OK | MB_ICONEXCLAMATION);
#else /* _AFXDLL && _DEBUG*/
    (void)tfKernelPrint(TM_PRINT_TYPE_WARN,
                        "tfKernelWarning! %s: %s\n",
                        functionName,
                        errorMessage);
#endif /* _AFXDLL &&_DEBUG*/

#endif /* !TM_IGNORE_WARNING */
}

#ifdef TM_MULTIPLE_CONTEXT
ttUserContext tfKernelGetContext()
{
    void * voidPtr;

    voidPtr = TlsGetValue(tvTlsIndex); 
    if(voidPtr == (void *)0)
    {
        voidPtr = tvCurrentContext;
    }
    tvCurrentContext = (ttContextPtr)voidPtr;
    return (ttUserContext)voidPtr;
}

void tfKernelSetContext(ttUserContext context)
{
    int errorCode;                
  
    errorCode = TlsSetValue(tvTlsIndex, context); 
    tm_assert(tfKernelSetContext, errorCode != 0);    
} 
#endif /* TM_MULTIPLE_CONTEXT */

#define TM_COLOR_PRINT_MAX_LINE 256
int tfKernelPrint(int msgType, const char TM_FAR * format, ...)
{
    int                         nResult;

#ifdef _WIN32
    HANDLE                      outputHandle;
    CONSOLE_SCREEN_BUFFER_INFO  bufferInfo;
    COORD*                      pPos;
    WORD                        foregroundColor;
#endif /* _WIN32 */

    char                        buffer[TM_COLOR_PRINT_MAX_LINE];

    va_list                     args;
    
    va_start(args, format);
    nResult = tfVSPrintF(buffer, format, args);
    va_end(args);

#ifdef _WIN32
    outputHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    GetConsoleScreenBufferInfo(outputHandle, &bufferInfo);
    pPos = &bufferInfo.dwCursorPosition;


    switch(msgType)
    {
    case TM_PRINT_TYPE_TEST:
        foregroundColor = FOREGROUND_GREEN;
#ifdef TM_AUTO_TEST
        pPos->X         = 0;
#endif /* TM_AUTO_TEST */
        break;
    case TM_PRINT_TYPE_INFO:
        foregroundColor = FOREGROUND_GREEN;
        break;
    case TM_PRINT_TYPE_PASS:
        foregroundColor = FOREGROUND_GREEN;
#ifdef TM_AUTO_TEST
        pPos->X         = 74;
        pPos->Y         -= 1;
#endif /* TM_AUTO_TEST */
        break;
    case TM_PRINT_TYPE_FAIL:
        foregroundColor = FOREGROUND_RED;
#ifdef TM_AUTO_TEST
        pPos->X         = 74;
        pPos->Y         -= 1;
#endif /* TM_AUTO_TEST */
        break;
    case TM_PRINT_TYPE_WARN:
        foregroundColor = FOREGROUND_BLUE;
#ifdef TM_AUTO_TEST
        pPos->X         = 74;
#endif /* TM_AUTO_TEST */
        break;
    case TM_PRINT_TYPE_STAT:
        foregroundColor = FOREGROUND_GREEN | FOREGROUND_RED;
#ifdef TM_AUTO_TEST
        pPos->X         = 0;
#endif /* TM_AUTO_TEST */
        break;

    default:
        foregroundColor = FOREGROUND_GREEN | FOREGROUND_RED |FOREGROUND_BLUE;
#ifdef TM_AUTO_TEST
        pPos->X         = 0;
        pPos->Y         = 0;
#endif /* TM_AUTO_TEST */
        break;
    }

    SetConsoleCursorPosition(outputHandle, *pPos);
    
    foregroundColor |= FOREGROUND_INTENSITY;
    SetConsoleTextAttribute(
            outputHandle,
            foregroundColor);

    printf("%s", buffer);

    SetConsoleTextAttribute(
            GetStdHandle(STD_OUTPUT_HANDLE),
            FOREGROUND_GREEN | FOREGROUND_RED |FOREGROUND_BLUE);

#else /* _WIN32 */
    TM_UNREF_IN_ARG(msgType);
    TM_UNREF_IN_ARG(format);
    nResult = 0;

    printf("%s", buffer);
#endif /* _WIN32 */
    return nResult;
}

/* 
 * Treck Stack calls this fucntion to ask for the system time. It is 
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network, 
 * or from application user interface.
 * 
 * Parameters:
 *  secondsPtr: 
 *      Pointer to store number of seconds elapsed since midnight (00:00:00), 
 *      January 1, 1970, Coordinated Universal Time (UTC). 
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT). 
 *      This matches the return value of the ANSI C library 
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year. 
 *      In case this value ever wrap arounds, we need another space to hold the 
 *      a bigger units of time. This is the purpose of daysPtr. 
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr  
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *  
 */


int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    time_t  times; /* long integer */
#ifdef _DEBUG
    ttTime  utcTime;
#endif /* _DEBUG */
    int     errorCode; 
  
    errorCode = TM_KERN_OKAY;
    times= time(&times);
    *daysPtr = 0;
    *secondsPtr = (ttUser32Bit)times; /* cast down on 64-bit CPUs */
#ifdef _DEBUG
    tfGetUtcTime(daysPtr, secondsPtr, &utcTime);
#endif /* _DEBUG */

    return errorCode;
} 


#else /* TM_KERNEL_WIN32_X86 */
#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */
static int tlWin32Dummy = 0;
#endif /* TM_KERNEL_WIN32_X86 */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trwin32.c     File Revision: 6.0.2.5 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 13815    Output checksum: 43604
 */
