/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Win32 device driver for serial port, uses COM1 set to 9600
 *  baud, 8-N-1
 *
 * Filename: trwuart.c
 * Author: Ed Remmell
 * Date Created:
 * $Source: source/kernel/win32/intelx86/trwuart.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:21CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#ifdef _WIN32
/* compile and run on Win32 platform */
/* to avoid an warning in MS VC++ header file */
struct _RPC_ASYNC_STATE;
#include  <windows.h>
#endif /* _WIN32 */

#include <trsocket.h>

#include <stdio.h>
#include <conio.h>

#define TM_COMM_NUM_RX_BUFS                     5
#define TM_COMM_RX_BUF_LEN                      2000
#define TM_COMM_TX_BUF_LEN                      2000
#define TM_WIN32_TASK_STACK_SIZE                128*1024

/* Win32 task that reads characters from the UART */
void tfWin32CommRecvIsr();

int tfWin32CommDrvOpen(ttUserInterface interfaceHandle);
int tfWin32CommDrvClose(ttUserInterface interfaceHandle);

int tfWin32CommDrvSend(
    ttUserInterface     interfaceHandle,
    char   TM_FAR       *dataPtr,
    int                 dataLength,
    int                 flag );

int tfWin32CommDrvRecv(
    ttUserInterface     interfaceHandle,
    char     TM_FAR * TM_FAR * dataPtrPtr,
    int               TM_FAR * dataSizePtr,
    ttUserBufferPtr     packetPtrPtr );

int tfWin32CommIoctl (
    ttUserInterface interfaceHandle,
    int             flag,
    void TM_FAR *   optionPtr,
    int             optionLen );

int tfWin32CommFreeReceiveBuffer(
    ttUserInterface interfaceHandle, char TM_FAR *   dataPtr);

static HANDLE  tlCfd; /* comm port file descriptor */
typedef struct tagTtRxBuffer {
    int	    bufLen;
    int     fullStatus;
    char    buf[TM_COMM_RX_BUF_LEN];
} ttRxBuffer;
    
static ttRxBuffer       tlRxBuffers[TM_COMM_NUM_RX_BUFS];
static int              tlNextEmptyRxBufIndex;
static int              tlNextFullRxBufIndex;
static HANDLE           tlReceiveThreadHandle = (HANDLE)0;
static ttUserInterface  tlInterfaceHandle = (ttUserInterface) 0;

int tfWin32CommDrvOpen(ttUserInterface interfaceHandle)
{
    int                 errorCode;
    int                 errno;
    int                 i;
    COMMTIMEOUTS        ct; 
    DCB                 dcb;
    DWORD               idThread;

/* initialization */
    errorCode = TM_DEV_OKAY; /* assume successful */
    tlNextEmptyRxBufIndex = 0;
    tlNextFullRxBufIndex = 0;
    for (i = 0; i < TM_COMM_NUM_RX_BUFS; i++)
    {
        tlRxBuffers[i].fullStatus = 0;
    }
/* loop until the IRQ task finishes */
    while (tlReceiveThreadHandle);
    tlInterfaceHandle = interfaceHandle; 
    
/* setup COM port */
    
    if ( (tlCfd=CreateFile(
              "COM1:",
              GENERIC_READ|GENERIC_WRITE,
              0,
              NULL,
              OPEN_EXISTING,
              FILE_FLAG_OVERLAPPED,
              NULL))==INVALID_HANDLE_VALUE )
    { 
        errno = GetLastError();
        printf("CreatFile failed. Errno=%d\n",errno);

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    } 
    printf("CreatFile OK\n"); 
     
    if ( !SetupComm(tlCfd, TM_COMM_RX_BUF_LEN, TM_COMM_TX_BUF_LEN)!=0 )
    { 
        errno = GetLastError(); 
        printf("SetupComm failed. Errno=%d\n",errno);

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    } 
    printf("SetupComm OK\n");

    if ( !GetCommState(tlCfd,&dcb) ) { 
        errno = GetLastError();
        printf("GetCommState failed. Errno=%d\n",errno); 

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    } 
    printf("GetCommState OK\n"); 

    dcb.BaudRate = CBR_9600;
    dcb.fBinary = TRUE; /* enable binary mode transfers */
    dcb.ByteSize = 8; /* 8 data bits */
    dcb.fParity = TRUE; /* enable parity checking */
    dcb.Parity = NOPARITY;
    dcb.StopBits = ONESTOPBIT;
    dcb.fOutxDsrFlow = FALSE;         /* disable DSR output flow control */
    dcb.fRtsControl = RTS_CONTROL_HANDSHAKE; /* hardware flow control */
    dcb.fOutxCtsFlow = TRUE; /* hardware flow control */
    dcb.fDtrControl = DTR_CONTROL_ENABLE;
    dcb.fDsrSensitivity = FALSE; /* ignore DSR for input */
    dcb.fInX = FALSE; /* disable XON/XOFF flow control */
    dcb.fOutX = FALSE;/* disable XON/XOFF flow control */
    dcb.fTXContinueOnXoff = TRUE;     /* XOFF continues Tx */
    dcb.fAbortOnError = FALSE;
    dcb.fErrorChar = FALSE;           /* Disable error replacement */
    dcb.fNull = FALSE;                /* Disable null stripping */

    if ( !SetCommState(tlCfd,&dcb) ) {
        errno = GetLastError();
        printf("SetCommState failed. Errno=%d\n",errno);

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    } 
    printf("SetCommState OK\n"); 
    printf("DCBlength=0x%04x\n",dcb.DCBlength);
    printf("BaudRate=%d\n",dcb.BaudRate); 
    printf("fBinary=0x%04x\n",dcb.fBinary); 
    printf("fParity=0x%04x\n",dcb.fParity); 
    printf("fOutxCtsFlow=0x%04x\n",dcb.fOutxCtsFlow); 
    printf("fOutxDsrFlow=0x%04x\n",dcb.fOutxDsrFlow); 
    printf("fDtrControl=0x%04x\n",dcb.fDtrControl); 
    printf("fDsrSensitivity=0x%04x\n",dcb.fDsrSensitivity); 
    printf("fTXContinueOnXoff=0x%04x\n",dcb.fTXContinueOnXoff);
    printf("fOutX=0x%04x\n",dcb.fOutX);
    printf("fInX=0x%04x\n",dcb.fInX); 
    printf("fErrorChar=0x%04x\n",dcb.fErrorChar); 
    printf("fNull=0x%04x\n",dcb.fNull);
    printf("fRtsControl=0x%04x\n",dcb.fRtsControl); 
    printf("fAbortOnError=0x%04x\n",dcb.fAbortOnError); 
    printf("XonLim=0x%02x\n",dcb.XonLim); 
    printf("XoffLim=0x%02x\n",dcb.XoffLim); 
    printf("ByteSize=0x%02x\n",dcb.ByteSize); 
    printf("Parity=0x%02x\n",dcb.Parity); 
    printf("StopBits=0x%02x\n",dcb.StopBits);
    printf("XonChar=0x%02x\n",dcb.XonChar); 
    printf("XoffChar=0x%02x\n",dcb.XoffChar); 
    printf("ErrorChar=0x%02x\n",dcb.ErrorChar); 
    printf("EofChar=0x%02x\n",dcb.EofChar); 
    printf("EvtChar=0x%02x\n",dcb.EvtChar); 

    if ( !GetCommTimeouts(tlCfd,&ct) ) { 
        errno = GetLastError(); 
        printf("GetCommTimeouts failed. Errno=%d\n",errno); 

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    }
    printf("Got CT\n"); 

    ct.ReadIntervalTimeout = MAXDWORD; /* return whatever bytes are in buffer*/
    ct.ReadTotalTimeoutMultiplier = 0;
    ct.ReadTotalTimeoutConstant = 0; 
    ct.WriteTotalTimeoutMultiplier = 2; /* wait 2 msec per-byte for writing */
    ct.WriteTotalTimeoutConstant = 50; /* additional 50 msec delay per write */
    if ( !SetCommTimeouts(tlCfd,&ct) ) {
        errno = GetLastError(); 
        printf("SetCommTimeouts failed. Errno=%d\n",errno); 

        errorCode = TM_DEV_ERROR;
        goto exitDrvOpen;
    } 
    printf("Set CT\n"); 
    printf("ReadIntervalTimeout=%d\n",ct.ReadIntervalTimeout); 
    printf("ReadTotalTimeoutMultiplier=%d\n",ct.ReadTotalTimeoutMultiplier); 
    printf("ReadTotalTimeoutConstant=%d\n",ct.ReadTotalTimeoutConstant); 
    printf("WriteTotalTimeoutMultiplier=%d\n",ct.WriteTotalTimeoutMultiplier); 
    printf("WriteTotalTimeoutConstant=%d\n",ct.WriteTotalTimeoutConstant); 

/* create a task to receive characters from the UART */
    tlReceiveThreadHandle = CreateThread(   
        NULL, 
        TM_WIN32_TASK_STACK_SIZE,
        (LPTHREAD_START_ROUTINE )tfWin32CommRecvIsr, 
        NULL, 
        0,  
        &idThread);
    if ( tlReceiveThreadHandle == NULL)
    {
        tfKernelError("tfWin32CommDrvOpen", 
                        " Create tfWin32CommRecvIsr error");
    }
    else
    {
        (void) SetThreadPriority(
            tlReceiveThreadHandle, 
            THREAD_PRIORITY_NORMAL);
    }
    
exitDrvOpen:
    return errorCode;
}

int tfWin32CommDrvClose(ttUserInterface interfaceHandle)
{
    if ( CloseHandle(tlCfd) )
        printf("Close OK\n");
    else
        printf("Close failed\n");
    
	tlInterfaceHandle = (ttUserInterface) 0;

    return TM_DEV_OKAY;
}

VOID CALLBACK tfWin32CommDrvSendCompletionRoutine(
  DWORD dwErrorCode,
  DWORD dwNumberOfBytesTransfered,
  LPOVERLAPPED lpOverlapped
)
{
    int flag;

    flag = (int) lpOverlapped->hEvent;
    if (flag & TM_USER_BUFFER_LAST)
    {
/* NOTE: the only place where we can enter an alertable wait state is in
   the driver send function, so we know that the device driver is already
   locked when this callback executes */
        tfSendCompleteInterface(
            tlInterfaceHandle, TM_DEV_SEND_COMPLETE_DRIVER);
    }

    free(lpOverlapped);
    return;
}

int tfWin32CommDrvSend  (   
    ttUserInterface     interfaceHandle,
    char     TM_FAR *   dataPtr,
    int                 dataLength,
    int                 flag )
{
    int                 errorCode;
    OVERLAPPED         *lpOverlapped;

    errorCode = TM_DEV_OKAY; /* assume successful */
    lpOverlapped = malloc(sizeof(OVERLAPPED));
    if (!lpOverlapped)
    {
        errorCode = TM_DEV_ERROR;
        goto exitDrvSend;
    }
    tfBzero(lpOverlapped, sizeof(OVERLAPPED));
/* we use lpOverlapped->hEvent to pass data to the completion callback */
    lpOverlapped->hEvent = (HANDLE) flag;
    
/* use WriteFileEx because it is asynchronous */
    if ( !WriteFileEx(
             tlCfd, dataPtr, (DWORD) dataLength, lpOverlapped,
             tfWin32CommDrvSendCompletionRoutine) )
    {
        printf("COMM driver failed to send!\n");
        errorCode = TM_DEV_ERROR;
    }
/* enter an alertable wait state, so that our completion callback gets
   called */
    SleepEx(1, TRUE);
    
exitDrvSend:    
    return errorCode;
}

int tfWin32CommDrvRecv(
    ttUserInterface     interfaceHandle,
    char     TM_FAR * TM_FAR * dataPtrPtr,
    int               TM_FAR * dataSizePtr,
    ttUserBufferPtr     packetPtrPtr )
{
    int errorCode;

    interfaceHandle=interfaceHandle;

    if (tlRxBuffers[tlNextFullRxBufIndex].fullStatus)
    {
        *dataPtrPtr   = tlRxBuffers[tlNextFullRxBufIndex].buf;
        *dataSizePtr  = tlRxBuffers[tlNextFullRxBufIndex].bufLen;
        *packetPtrPtr = (ttUserBufferPtr) 0;

        errorCode = TM_DEV_OKAY;
    }
    else
    {
        errorCode = TM_DEV_ERROR;
    }

    return(errorCode);
}

/*
 * tfWin32CommIoctl
 *
 *   No need for an IOCTL routine in our serial driver.
 */
int tfWin32CommIoctl (
    ttUserInterface interfaceHandle,
    int             flag,
    void TM_FAR *   optionPtr,
    int             optionLen )
{
    TM_UNREF_IN_ARG(interfaceHandle);
    flag            = flag;
    optionPtr       = optionPtr;
    optionLen       = optionLen;

    return(TM_DEV_OKAY);
}

/*
 * tfWin32CommRecvThread() Function Description
 */
void tfWin32CommRecvIsr ()
{
    OVERLAPPED          overlapped;
    unsigned long       nBytesRead;
    
    while (tlInterfaceHandle)
    {
/* make sure we don't overrun the Treck receive task */
        if (!tlRxBuffers[tlNextEmptyRxBufIndex].fullStatus)
        {
            tfBzero(&overlapped, sizeof(OVERLAPPED));
/* NOTE: overlapped.hEvent can be used to pass data to the completion
   callback */
    
            if (ReadFile(
                    tlCfd,
                    tlRxBuffers[tlNextEmptyRxBufIndex].buf,
                    TM_COMM_RX_BUF_LEN, /* size of our RX buffer */
                    &nBytesRead,
                    &overlapped))
            { 
                if (nBytesRead)
                {
/* we read some data from the UART */
                    tlRxBuffers[tlNextEmptyRxBufIndex].bufLen =
                        (int) nBytesRead;
/* mark this buffer full */
                    tlRxBuffers[tlNextEmptyRxBufIndex].fullStatus = 1; 

/* notify Treck TCP that we received some data */
                    tfNotifyInterfaceIsr(tlInterfaceHandle, 1, 0, 0UL, 0UL);

/* advance our empty RX buffer index to the next */
                    tlNextEmptyRxBufIndex = 
                        ((tlNextEmptyRxBufIndex + 1) % TM_COMM_NUM_RX_BUFS);
                }
            }
        }
    }
    tlReceiveThreadHandle = (HANDLE)0;
}

/*
 * tf16550FreeReceiveBuffer
 *
 *   Called when the stack is done with a recieved packet (ie, character).
 *   This should be immediately after it is passed up the stack, since both of
 *   the serial link layers (PPP,SLIP) copy immediately.
 *
 */
int tfWin32CommFreeReceiveBuffer(
    ttUserInterface interfaceHandle, char TM_FAR *   dataPtr)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtr);

/* mark this buffer empty */
    tlRxBuffers[tlNextFullRxBufIndex].fullStatus = 0;

/* advance our full RX buffer index to the next */
    tlNextFullRxBufIndex = 
        ((tlNextFullRxBufIndex + 1) % TM_COMM_NUM_RX_BUFS);
    
    return TM_DEV_OKAY;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trwuart.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 48870    Output checksum: 54309
 */
