/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: trkernel.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/kernel/xilkernel/trxmk.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:27CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/* Modify the include file for your Kernel/OS */
#include "xparameters.h"
#include "xmk.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <os_config.h>
#include <sys/process.h>
#include <pthread.h>

#include <semaphore.h>
#include <sys/timer.h>

#include <sys/bufmalloc.h>

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

#ifdef TM_TRECK_PREEMPTIVE_KERNEL

extern int yield(void);

#ifdef __PPC__
#define TM_KERNEL_XILKERNEL_PPC
#else /* __PPC__ */
#define TM_KERNEL_XILKERNEL_MB
#endif /* __PPC__ */

extern void microblaze_enable_interrupts(void);     /* Enable Interrupts */
extern void microblaze_disable_interrupts(void);	/* Disable Interrupts */

/* 
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */

#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate(ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
	int errorCode;
	xil_printf("tfKernelSheapCreate: allocating %d bytes\r\n", (*sheapSizePtr));
//    errorCode = (ttUser32Bit TM_FAR *)(bufmalloc(MEMBUF_ANY, (*sheapSizePtr)));
    errorCode = (ttUser32Bit TM_FAR *)(malloc(*sheapSizePtr));	
    xil_printf("tfKernelSheapCreate: returned %d\r\n", errorCode);
    
	return errorCode;
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR *tfKernelMalloc(unsigned size)
{
    void TM_FAR * memoryPtr = (void TM_FAR *)0;
//    memoryPtr = (void *)(bufmalloc(MEMBUF_ANY, size)); 
    memoryPtr = (void *)(malloc(size)); 
	if (memoryPtr)
	{
	    xil_printf("tfKernelMalloc %d bytes (0x%08X)\r\n", size, memoryPtr);
	}
	else
	{
		xil_printf("tfKernelMalloc FAILED to allocate %d bytes\r\n", size);
	}
    return memoryPtr;
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
    xil_printf("tfKernelFree (0x%08X)\r\n", memoryBlockPtr);
//	buffree(MEMBUF_ANY, memoryBlockPtr);
	free(memoryBlockPtr);
    return;
}

/*
 * Stop all other calls into Treck by 
 * a kernel context switch
 */
void tfKernelSetCritical(void)
{
#ifdef TM_KERNEL_XILKERNEL_MB
    microblaze_disable_interrupts();
#endif /* TM_KERNEL_XILKERNEL_MB */
#ifdef TM_KERNEL_XILKERNEL_PPC
	XExc_mDisableExceptions(XEXC_ALL);
#endif /* TM_KERNEL_XILKERNEL_PPC */
}

/*
 * Resume normal operation (from tm_kernel_set_critical)
 */
void tfKernelReleaseCritical(void)
{
#ifdef TM_KERNEL_XILKERNEL_MB
    microblaze_enable_interrupts();
#endif /* TM_KERNEL_XILKERNEL_MB */
#ifdef TM_KERNEL_XILKERNEL_PPC
	XExc_mEnableExceptions(XEXC_ALL);
#endif /* TM_KERNEL_XILKERNEL_PPC */
}


/*
 * Counting Semaphore Operations
 *
 * The examples assume the following:
 * A semaphore is of type int
 * Success from a Kernel call is 0
 * Failure from a Kernel Call is -1
 *
 */

/*
 * Create a counting semaphore
 */
int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
	int errorCode;
	
    /* Initialize the semaphore to zero */
	errorCode = sem_init((sem_t*)&countingSemaphore->genUintParm,0,0);

	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else
	{
		errorCode = TM_KERN_ERROR;
		tfKernelError( "tfKernelCreateCountSem",
                       "Unable to Create Xilkernel Semaphore");
	}
    return errorCode;
}

/*
 * Wait on a counting semaphore
 */
int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
	int errorCode;
	errorCode = sem_wait((sem_t*)&countingSemaphore->genUintParm);
	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPendCountSem","Unable to Pend on Xilkernel Semaphore");
	}
    return errorCode;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Wait on a counting semaphore up to a timeout value (in ticks)
 */
int tfKernelPendCountSemTimeout( ttUserGenericUnionPtr countingSemaphore, 
                                 ttUser32Bit tickTimeout )
{
	int errorCode;

    errorCode = sem_timedwait((sem_t*)&countingSemaphore->genUintParm,
    						 tickTimeout * TM_TICK_LENGTH);
	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else if (errorCode == ETIMEDOUT)
	{
		errorCode = TM_ETIMEDOUT;
		tfKernelWarning( "tfKernelPendCountSemTimeout",
						 "Timeout expired before lock was obtained");
	}
	else
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPendCountSemTimeout",
                       "Unable to Pend on Xilkernel Semaphore");
	}
    return errorCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume waiting tasks on a counting semaphore
 */
int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
	int errorCode;

	errorCode = sem_post((sem_t*)&countingSemaphore->genUintParm);
	
	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPostCountSem",
                       "Unable to Post on Xilkernel Semaphore");
	}
    return errorCode;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
	int errorCode;
   
	errorCode = sem_destroy((sem_t*)&countingSemaphore->genUintParm);

	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelDeleteCountSem",
                       "Unable to Destroy Xilkernel Semaphore");
	}
    return errorCode;

}

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
    (void)tfKernelPrint(TM_PRINT_TYPE_FAIL, 
                 "tfKernelError! %s: %s\n", functionName, errorMessage);
     
    while(1)
	{
		yield();
	}
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
    (void)tfKernelPrint(
            TM_PRINT_TYPE_WARN,
            "tfKernelWarning! %s: %s\n", functionName, errorMessage);
}

void  tfKernelInitialize(void)
{
	/* Initialize the kernel here if necessary */
}

// *******************************************************

/*
 * Create an event structure for pend/post from an ISR, and
 * also for XMIT task pend/post.
 */
void tfKernelCreateEvent(ttUserGenericUnionPtr eventPtr)
{
	int errorCode;
   
    /* Initialize the semaphore to zero */
    errorCode = sem_init((sem_t*)&eventPtr->genUintParm,0,0);	 

	if (errorCode == -1)
	{
        tfKernelError( "tfKernelCreateEvent",
                       "Unable to Create Xilkernel Semaphore");
	}
    return;
}

/*
 * Task wait on an Event, probably the RX Task
 */
void tfKernelPendEvent(ttUserGenericUnionPtr eventPtr)
{
	int errorCode;
	errorCode = sem_wait((sem_t*)&eventPtr->genUintParm);
	
	if (errorCode == -1)
	{
        tfKernelError( "tfKernelPendEvent",
                       "Unable to Pend on Xilkernel Semaphore");
	}
    return;
}

#ifdef TM_USE_PEND_TIMEOUT
/*
 * Task wait on an Event up to a timeout value (in ticks).
 */
int tfKernelPendEventTimeout( ttUserGenericUnionPtr eventPtr, 
                               ttUser32Bit tickTimeout )
{
	int errorCode;

    errorCode = sem_timedwait((sem_t*)&eventPtr->genUintParm,
    					 tickTimeout * TM_TICK_LENGTH);
	if (errorCode != -1)
	{
		errorCode = TM_KERN_OKAY;
	}
	else if (errno == ETIMEDOUT)
	{
		errorCode = TM_ETIMEDOUT;
		tfKernelWarning( "tfKernelPendEventTimeout",
						 "Timeout expired before lock was obtained");
	}
	else
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPendEventTimeout",
                       "Unable to Pend on Xilkernel Semaphore");
	}
    return errorCode;
}
#endif /* TM_USE_PEND_TIMEOUT */

/*
 * Resume task waiting on this Event. Called from a task.
 */
void tfKernelTaskPostEvent(ttUserGenericUnionPtr eventPtr)
{
	int errorCode;
	errorCode = sem_post((sem_t*)&eventPtr->genUintParm);

	if (errorCode == -1)
	{
        tfKernelError( "tfKernelTaskPostEvent",
                       "Unable to Post on Xilkernel Semaphore");
	}
    return;
}

/*
 * Resume task waiting on this event. Called from an ISR.
 */
void tfKernelIsrPostEvent(ttUserGenericUnionPtr eventPtr)
{
	int errorCode;
	errorCode = sem_post((sem_t*)&eventPtr->genUintParm);
	if (errorCode == -1)
	{
		errorCode = TM_KERN_ERROR;
        tfKernelError( "tfKernelPostISRCountSem",
                       "Unable to Post on Xilkernel Semaphore");
	}
}


/*
 * Yield the CPU. Used by tfInterfaceSpinLock().
 */
void tfKernelTaskYield(void)
{
	yield();
    return;
}


int tfKernelPrint(int msgType, const char TM_FAR * format, ...)
{
    int                         nResult;
    char                        buffer[256];
    va_list                     args;
   
    va_start(args, format);
    nResult = tfVSPrintF(buffer, format, args);
    va_end(args);

    msgType = msgType;
    format = format;
    nResult = 0;

    xil_printf("%s\r\n", buffer);
    return nResult;
}


/* 
 * Treck Stack calls this fucntion to ask for the system time. It is 
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network, 
 * or from application user interface.
 * 
 * Parameters:
 *  secondsPtr: 
 *      Pointer to store number of seconds elapsed since midnight (00:00:00), 
 *      January 1, 1970, Coordinated Universal Time (UTC). 
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT). 
 *      This matches the return value of the ANSI C library 
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year. 
 *      In case this value ever wrap arounds, we need another space to hold the 
 *      a bigger units of time. This is the purpose of daysPtr. 
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr  
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *  
 */
int tfKernelGetSystemTime(ttUser32Bit * daysPtr, ttUser32Bit * secondsPtr)
{
    int errorCode; 
    errorCode = TM_KERN_OKAY;
    *daysPtr = 0;
    time((time_t *)secondsPtr);

    return errorCode;
} 

/* Processed by Treck Shipping Tool v2.8 
 * at: 2004/05/12 09:57:31  for: xilinx
 * File Name: trkernel.c    File Revision: 4.0.2.3 
 * Project Name: tcpip.pj   Project Release: 4.0.2.10
 * Input checksum: 44623,   Output checksum: 44867
 */

#endif /* TM_TRECK_PREEMPTIVE_KERNEL */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trxmk.c       File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 47669    Output checksum: 17061
 */
