/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Optimized checksum routine for the ARM7 platform and 
 *              Green Hills compiler.
 * Filename: trcsarmg.c
 * Author: Jason
 * Date Created: 11/13/2001
 * $Source: source/ocksum/trcsarmg.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:28CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * This is an assembly routine to calculate the checksum of a packet.
 * This routine will only total the long word aligned portion of the data;
 * The code in the tfPacketChecksum routine will handle the remaining.  This
 * routine should be inlined by the GHS compiler in the tfPacketChecksum 
 * function below, and should not generate a function call.  Obviously a 
 * function call here would significantly affect performance, so if this 
 * happens, check the GHS optimization settings.
 */

asm unsigned long tm_asm_checksum (int loopCount, char * dataPtr, int checksum)
{
%mem loopCount %reg dataPtr %reg checksum

/* Set counter to number of times to iterate through loop */
    LDR r0,loopCount

CHECKSUM_LOOP:

/* Load current long word from packet */
    LDR r1, [dataPtr]

/* Add current long word to checksum */
    ADDS checksum, r1, checksum

/* If the carry bit is set, add in the carry from the last addition */    
    ADCCS checksum, checksum, 0

/* Move data pointer to next long word */
    ADD dataPtr, dataPtr, 4

/* Decrement loop counter */
    SUBS r0, r0, 1

    BNE CHECKSUM_LOOP

/* compiler expects value to be returned in 'r0' register */    
    MOV r0, checksum

        
%reg loopCount %reg dataPtr %reg checksum

/* Set counter to number of times to iterate through loop */
    MOV r0,loopCount

CHECKSUM_LOOP:

/* Load current long word from packet */
    LDR r1, [dataPtr]

/* Add current long word to checksum */
    ADDS checksum, r1, checksum

/* If the carry bit is set, add in the carry from the last addition */    
    ADCCS checksum, checksum, 0

/* Move data pointer to next long word */
    ADD dataPtr, dataPtr, 4

/* Decrement loop counter */
    SUBS r0, r0, 1

    BNE CHECKSUM_LOOP

/* compiler expects value to be returned in 'r0' register */    
    MOV r0, checksum

}

        
tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
    tt16BitPtr          link16BitDataPtr;
    register tt32BitPtr link32BitDataPtr;
    register tt32Bit    sum;
    ttPktLen            linkLength;
    ttPktLen            numberWords;
    register int        bigLoop;
#ifdef TM_DEV_SEND_OFFLOAD
    tt8Bit              partialChecksum;
#endif /* TM_DEV_SEND_OFFLOAD */

    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    } oddByte16Bit;
    tt8Bit        sumByteSwapped;

    sum = 0UL;

/* sum in the pseudo-header, if necessary */
    while (pshLen > 1)
    {
        sum += *pshHdrPtr++;
        pshLen -= 2;
    }

/* packetPtr is assumed to be non null unless checksum is offloaded */
#ifdef TM_DEV_SEND_OFFLOAD
/* packetPtr could be null if checksum is offloaded. */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        partialChecksum = TM_8BIT_YES;
    }
    else
#endif /* TM_DEV_SEND_OFFLOAD */
    {
#ifdef TM_DEV_SEND_OFFLOAD
        partialChecksum = TM_8BIT_ZERO;
#endif /* TM_DEV_SEND_OFFLOAD */
        linkLength = (ttPktLen)0;
        sumByteSwapped = TM_8BIT_ZERO;
        do
        {
/* Point to data in this link */
            link16BitDataPtr = (tt16BitPtr)packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
            if (linkLength == (ttPktLen)0)
            {
                linkLength = packetPtr->pktLinkDataLength;
            }
            else
            {
/* previous link had odd byte length, store 2nd byte to make a word */
                oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
                link16BitDataPtr = (tt16BitPtr)
                                            ((tt8BitPtr)link16BitDataPtr + 1);
                linkLength = packetPtr->pktLinkDataLength - 1;
                totalDataLength--;
            }
            if (linkLength > totalDataLength)
            {
/* Do not span more than asked */
                linkLength = totalDataLength;
            }
/* left over data to be checksumed after this link */
            totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
            if ((tm_is_odd(link16BitDataPtr)) && (linkLength != 0))
            {
/* save odd byte for later addition */
                oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
/* Update pointer, and length. Pointer now is on even byte boundary */
                link16BitDataPtr =
                               (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
                linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
                tm_byte_swap_sum(sum);
                sumByteSwapped = TM_8BIT_YES;
            }

/* 
 * The assembly loop below needs to deal with everything long word aligned,
 * so if we're not on a long word boundary, add in the extra 16-bit value to
 * the checksum and move the pointer forward.
 */
            if ( (((tt32Bit) link16BitDataPtr) & 2) && (linkLength >= 2) )
            {
                sum += *link16BitDataPtr++;
                linkLength -= 2;
            }
        
/* 
 * Compute number of words and number of iterations through the assembly
 * loop below.
 */
            numberWords = linkLength >> 1;
            bigLoop = (unsigned short)(numberWords >> 1);


/* 
 * Call optimized assembly checksum routine.
 *
 * NOTE: This *shouldn't* result in a function call (ie, it should be inlined).
 * If it does generate a function call, check your compiler optimizations.
 */
            if (bigLoop > 0)
            {
                link32BitDataPtr = (tt32BitPtr) link16BitDataPtr;
                sum = tm_asm_checksum(bigLoop, link32BitDataPtr, sum);      
                tm_add_carries_sum(sum);
                linkLength -= bigLoop * 4;
                link16BitDataPtr = (tt16BitPtr) (link32BitDataPtr);
            }

/* 
 * Catch an extra 16-bit word here.  After this, the only data to be summed
 * are 32-bit long words and possibly and extra byte.  
 */
            if (numberWords & 1)
            {
                sum += *link16BitDataPtr++;
                linkLength -= 2;
            }
        
/* We need to know if linkLength was odd */
            linkLength &= 1;
            if (sumByteSwapped != TM_8BIT_ZERO)
            {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
                tm_byte_swap_sum(sum);
/* toggle byte swapping */
                sumByteSwapped = TM_8BIT_ZERO;
                if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
                {
                    oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                    sum = sum + oddByte16Bit.oddWord;
                    linkLength = 0;
                }
                else
                {
/* remember that we have our even byte stored in oddChar[0] */
                    linkLength = 1;
                }
            }
            else
            {
                if (linkLength != 0)
                {
/* Odd byte length. Save the character to add it later on */
                    oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
                }
            }
/* check this before computing next packetPtr */
            if (totalDataLength == (ttPktLen)0)
            {
/* break out of the loop. Do not check 0 length after the loop */
                goto outFold;
            }
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (packetPtr != TM_PACKET_NULL_PTR);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
        if (totalDataLength != (ttPktLen)0)
        {
            tfKernelError("tfPacketChecksum",
                "Data in Packet chain smaller than length to be checksumed");
        }
#endif
outFold:
        if (linkLength != 0UL)
        {
/*  Add left-over byte */
            oddByte16Bit.oddChar[1] = 0;
            sum = sum + oddByte16Bit.oddWord;
        }
    }
/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
    tm_add_carries_sum(sum);
#ifdef TM_DEV_SEND_OFFLOAD
    if (partialChecksum == TM_8BIT_ZERO)
#endif /* TM_DEV_SEND_OFFLOAD */
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcsarmg.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 20773    Output checksum: 60595
 */
