/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Optimized checksum routine for the CPU32 platform and the 
 *              SDS compiler.
 * Filename: trcsc32s.c
 * Author: Jason
 * Date Created: 4/5/2000
 * $Source: source/ocksum/trcsc32s.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:30CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>

#ifdef TM_LINT
LINT_UNACCESS_SYM_BEGIN(bigLoop)
#endif /* TM_LINT */

tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
/*
 * LINT NOTE: link32BitDataPtr must *NOT* be replaced by linkDataPtr.u32Ptr
 *            in order to avoid LINT warnings.  The inline assembly below does
 *            not accept linkDataPtr.u32Ptr as a valid pointer type and will
 *            generate an error.
 */
    tt32BitPtr         link32BitDataPtr;
    ttPtrUnion         linkDataPtr;
    tt32Bit            sum;
    ttPktLen           linkLength;
    ttPktLen           numberWords;
    unsigned short int bigLoop;
    unsigned short int switchOffset;
    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    }                   oddByte16Bit;
    tt8Bit              sumByteSwapped;
#ifdef TM_DEV_SEND_OFFLOAD
    tt8Bit          partialChecksum;
#endif /* TM_DEV_SEND_OFFLOAD */

    sum = 0UL;

/* sum in the pseudo-header, if necessary */
    while (pshLen > 1)
    {
        sum += *pshHdrPtr++;
        pshLen -= 2;
    }

/* packetPtr is assumed to be non null unless checksum is offloaded */
#ifdef TM_DEV_SEND_OFFLOAD
/* packetPtr could be null if checksum is offloaded. */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        partialChecksum = TM_8BIT_YES;
    }
    else
#endif /* TM_DEV_SEND_OFFLOAD */
    {
#ifdef TM_DEV_SEND_OFFLOAD
        partialChecksum = TM_8BIT_ZERO;
#endif /* TM_DEV_SEND_OFFLOAD */
        linkLength = (ttPktLen)0;
        sumByteSwapped = TM_8BIT_ZERO;
        do
        {
/* Point to data in this link */
            linkDataPtr.u8Ptr = packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
            if (linkLength == (ttPktLen)0)
            {
                linkLength = packetPtr->pktLinkDataLength;
            }
            else
            {
/* previous link had odd byte length, store 2nd byte to make a word */
                oddByte16Bit.oddChar[1] = *(linkDataPtr.u8Ptr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
                linkDataPtr.u8Ptr++;
                linkLength = packetPtr->pktLinkDataLength - 1;
                totalDataLength--;
            }
            if (linkLength > totalDataLength)
            {
/* Do not span more than asked */
                linkLength = totalDataLength;
            }
/* left over data to be checksumed after this link */
            totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
            if ((tm_is_odd(linkDataPtr.u8Ptr)) && (linkLength != 0))
            {
/* save odd byte for later addition */
                oddByte16Bit.oddChar[0] = *(linkDataPtr.u8Ptr);
/* Update pointer, and length. Pointer now is on even byte boundary */
                linkDataPtr.u8Ptr++;
                linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
                tm_byte_swap_sum(sum);
                sumByteSwapped = TM_8BIT_YES;
            }


/*
 * Chances are, the linkLength isn't going to be evenly divisible by 64. Thus,
 * we do calculations on the remainder first, and then use an optimized
 * assembly loop to handle the majority of the packet. While it would in fact
 * be possible, as well as marginally faster, to handle ALL of the
 * checksumming in assembly, this part is left in C to make porting to your
 * (the customer's) platform easier. If you are a confident assembly
 * programmer, this switch can be incorporated into the rest of the loop
 * (Duff's Device is one example of how it can be done).
 */
/* How far to jump into the switch statement */
            numberWords = linkLength >> 1;

/*
 * How many times we are going to have to go through the big loop. Each time
 * through the loop, we add in 64 more bytes.
 */
            bigLoop = (unsigned short)((numberWords + 31) >> 5);
            switchOffset = (unsigned short)(numberWords % 32);

            switch (switchOffset)
            {
                case 31:       sum += *(linkDataPtr.u16Ptr+30);
                    /* fall through */
                case 30:       sum += *(linkDataPtr.u16Ptr+29);
                    /* fall through */
                case 29:       sum += *(linkDataPtr.u16Ptr+28);
                    /* fall through */
                case 28:       sum += *(linkDataPtr.u16Ptr+27);
                    /* fall through */
                case 27:       sum += *(linkDataPtr.u16Ptr+26);
                    /* fall through */
                case 26:       sum += *(linkDataPtr.u16Ptr+25);
                    /* fall through */
                case 25:       sum += *(linkDataPtr.u16Ptr+24);
                    /* fall through */
                case 24:       sum += *(linkDataPtr.u16Ptr+23);
                    /* fall through */
                case 23:       sum += *(linkDataPtr.u16Ptr+22);
                    /* fall through */
                case 22:       sum += *(linkDataPtr.u16Ptr+21);
                    /* fall through */
                case 21:       sum += *(linkDataPtr.u16Ptr+20);
                    /* fall through */
                case 20:       sum += *(linkDataPtr.u16Ptr+19);
                    /* fall through */
                case 19:       sum += *(linkDataPtr.u16Ptr+18);
                    /* fall through */
                case 18:       sum += *(linkDataPtr.u16Ptr+17);
                    /* fall through */
                case 17:       sum += *(linkDataPtr.u16Ptr+16);
                    /* fall through */
                case 16:       sum += *(linkDataPtr.u16Ptr+15);
                    /* fall through */
                case 15:       sum += *(linkDataPtr.u16Ptr+14);
                    /* fall through */
                case 14:       sum += *(linkDataPtr.u16Ptr+13);
                    /* fall through */
                case 13:       sum += *(linkDataPtr.u16Ptr+12);
                    /* fall through */
                case 12:       sum += *(linkDataPtr.u16Ptr+11);
                    /* fall through */
                case 11:       sum += *(linkDataPtr.u16Ptr+10);
                    /* fall through */
                case 10:       sum += *(linkDataPtr.u16Ptr+9);
                    /* fall through */
                case  9:       sum += *(linkDataPtr.u16Ptr+8);
                    /* fall through */
                case  8:       sum += *(linkDataPtr.u16Ptr+7);
                    /* fall through */
                case  7:       sum += *(linkDataPtr.u16Ptr+6);
                    /* fall through */
                case  6:       sum += *(linkDataPtr.u16Ptr+5);
                    /* fall through */
                case  5:       sum += *(linkDataPtr.u16Ptr+4);
                    /* fall through */
                case  4:       sum += *(linkDataPtr.u16Ptr+3);
                    /* fall through */
                case  3:       sum += *(linkDataPtr.u16Ptr+2);
                    /* fall through */
                case  2:       sum += *(linkDataPtr.u16Ptr+1);
                    /* fall through */
                case  1:       sum += *(linkDataPtr.u16Ptr);
/* Increment the pointer */
                               linkDataPtr.u16Ptr += switchOffset;
/* Decrement the loop control variable */
                               bigLoop--;
                               break;
                default:;
            }

            link32BitDataPtr = linkDataPtr.u32Ptr;


/*
 * The SDS 68K compiler guarantees that we can use the D0,D1,A0 and A1 
 * registers in 'asm' statements, so there is no need to save them here.
 * The only register that needs to be saved is D2.
 */
            asm(" move.l D2,-(A7)");

/* D0 = current checksum total */
            asm(" move.l {sum},D0");

/* D1 = number of times to iterate through below loop */
            asm(" move.w {bigLoop},D1"); 

/* A0 = data pointer to start checksumming at */
            asm(" movea.l {link32BitDataPtr},A0");

/* Clear carry flag */
            asm(" move.w #00,CCR");

/* D2 = 0 */
            asm(" eor d2,d2");

/* 
 * Checksum 64 bytes (16 long words) at a time.  We know from the above
 * switch statement that the pointer we're given (link32BitDataPtr) will
 * be 64-byte 'aligned'.  We loop for 'bigLoop' number of times. 
 */
            asm(" jmp CHECKSUM_END");

            asm("CHECKSUM_INNER_LOOP:");

/* 
 * For each long word, we do a long add with the current checksum and the
 * current data.  Then we do an 'add with extend' with the current checksum
 * and zero, to catch the carry bit from the previous operation.
 */ 
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");
            asm(" add.l (A0)+,D0");
            asm(" addx.l D2,D0");

            asm("CHECKSUM_END:");
            asm(" dbra D1,CHECKSUM_INNER_LOOP");

/* Move the final checksum back into the 'sum' variable. */
            asm(" move.l D0,{sum}");

/* Restore the D2 register */
            asm(" movem.l (A7)+,D2");

/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
            tm_add_carries_sum(sum);

/* Increment the pointer to the correct point */
            linkDataPtr.u16Ptr += (numberWords - switchOffset);
/*
 * We're done with the 16-bit variable. Move the sum back to the standard
 * (32-bit) variable for the rest of the function
 */

/* We need to know if linkLength was odd */
            linkLength &= 1;
            if (sumByteSwapped != TM_8BIT_ZERO)
            {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
                tm_byte_swap_sum(sum);
/* toggle byte swapping */
                sumByteSwapped = TM_8BIT_ZERO;
                if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
                {
                    oddByte16Bit.oddChar[1] = *(linkDataPtr.u8Ptr);
/* add to sum */
                    sum = sum + oddByte16Bit.oddWord;
                    linkLength = 0;
                }
                else
                {
/* remember that we have our even byte stored in oddChar[0] */
                    linkLength = 1;
                }
            }
            else
            {
                if (linkLength != 0)
                {
/* Odd byte length. Save the character to add it later on */
                    oddByte16Bit.oddChar[0] = *(linkDataPtr.u8Ptr);
                }
            }
/* check this before computing next packetPtr */
            if (totalDataLength == (ttPktLen)0)
            {
/* break out of the loop. Do not check 0 length after the loop */
                goto outFold;
            }
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (packetPtr != TM_PACKET_NULL_PTR);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
#ifdef TM_LINT
LINT_BOOLEAN_BEGIN
#endif /* TM_LINT */

        if (totalDataLength != (ttPktLen)0)
#ifdef TM_LINT
LINT_BOOLEAN_END
#endif /* TM_LINT */

        {
            tfKernelError("tfPacketChecksum",
                "Data in Packet chain smaller than length to be checksumed");
        }
#endif
outFold:
        if (linkLength != 0UL)
        {
/*  Add left-over byte */
            oddByte16Bit.oddChar[1] = 0;
            sum = sum + oddByte16Bit.oddWord;
        }
    }
/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
    tm_add_carries_sum(sum);
#ifdef TM_DEV_SEND_OFFLOAD
    if (partialChecksum == TM_8BIT_ZERO)
#endif /* TM_DEV_SEND_OFFLOAD */
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}
#ifdef TM_LINT
LINT_UNACCESS_SYM_END(bigLoop)
#endif /* TM_LINT */


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcsc32s.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 48974    Output checksum: 34589
 */
