/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Optimized checksum routine for the PowerPC platform and 
 *              GCC compiler.
 * Filename: trcsppcg.c
 * Author: Ed Remmell
 * Date Created: 04/05/2000
 * $Source: source/ocksum/trcsppcg.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:32CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>

tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
    tt16BitPtr          link16BitDataPtr;
    register tt32Bit    link32BitDataPtr;
    register tt32Bit    sum;
    register int        bigLoop;
    ttPktLen            linkLength;
    ttPktLen            numberWords;
    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    }                   oddByte16Bit;
    tt8Bit              sumByteSwapped;
#ifdef TM_DEV_SEND_OFFLOAD
    tt8Bit          partialChecksum;
#endif /* TM_DEV_SEND_OFFLOAD */

    sum = 0UL;

/* sum in the pseudo-header, if necessary */
    while (pshLen > 1)
    {
        sum += *pshHdrPtr++;
        pshLen -= 2;
    }

/* packetPtr is assumed to be non null unless checksum is offloaded */
#ifdef TM_DEV_SEND_OFFLOAD
/* packetPtr could be null if checksum is offloaded. */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        partialChecksum = TM_8BIT_YES;
    }
    else
#endif /* TM_DEV_SEND_OFFLOAD */
    {
#ifdef TM_DEV_SEND_OFFLOAD
        partialChecksum = TM_8BIT_ZERO;
#endif /* TM_DEV_SEND_OFFLOAD */
        linkLength = (ttPktLen)0;
        sumByteSwapped = TM_8BIT_ZERO;
        do
        {
/* Point to data in this link */
            link16BitDataPtr = (tt16BitPtr)packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
            if (linkLength == (ttPktLen)0)
            {
                linkLength = packetPtr->pktLinkDataLength;
            }
            else
            {
/* previous link had odd byte length, store 2nd byte to make a word */
                oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
                link16BitDataPtr = (tt16BitPtr)
                                            ((tt8BitPtr)link16BitDataPtr + 1);
                linkLength = packetPtr->pktLinkDataLength - 1;
                totalDataLength--;
            }
            if (linkLength > totalDataLength)
            {
/* Do not span more than asked */
                linkLength = totalDataLength;
            }
/* left over data to be checksumed after this link */
            totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
            if ((tm_is_odd(link16BitDataPtr)) && (linkLength != 0))
            {
/* save odd byte for later addition */
                oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
/* Update pointer, and length. Pointer now is on even byte boundary */
                link16BitDataPtr =
                               (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
                linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
                tm_byte_swap_sum(sum);
                sumByteSwapped = TM_8BIT_YES;
            }

/* 
 * Compute number of words and number of iterations through the assembly
 * loop below.
 */
            numberWords = linkLength >> 1;
            bigLoop = (unsigned short)(numberWords >> 1);

/* 
 * Catch an extra 16-bit word here.  After this, the only data to be summed
 * are 32-bit long words and possibly and extra byte.
 */
            if (numberWords & 1)
            {
                sum += *link16BitDataPtr++;
                linkLength -= 2;
            }

            if (bigLoop > 0)
            {
                link32BitDataPtr = (tt32Bit)((void *)link16BitDataPtr);

/*
 * This is GNU compiler in-line assembly for the PPC to calculate the
 * checksum of a packet. This code only totals the long-word-aligned portion
 * of the data, and uses a tight loop which adds long words to the current
 * checksum total. This tight (ie, NOT unrolled) loop should be more efficient
 * on most PPC platforms due to the PPC's branch prediction and speculative
 * loads.  This code could be optimized even further by doing more detailed
 * software pipelining, but this varies significantly between different PPC
 * models. 
 */

/*
; Set branch counter to number of times to iterate through loop 
    mtspr   9,bigLoop

    addi    link32BitDataPtr,link32BitDataPtr,-4
    
; Clear the carry bit
    addic   sum,sum,0
                      
CHECKSUM_LOOP:
    lwzu    r6,4(link32BitDataPtr)
    adde    sum,r6,sum
    bdnz    CHECKSUM_LOOP

; In case there was a carry due to the last addition, add the carry bit back 
; in to the sum ( sum = sum + 0 + XER[carry] ).
    xor     r6,r6,r6
    adde    sum,r6,sum
*/
                __asm__("
                        mtspr   9,%2
                        addi    %1,%1,-4
                        addic   %0,%0,0
CHECKSUM_LOOP:
                        lwzu    r6,4(%1)
                        adde    %0,r6,%0
                        bdnz    CHECKSUM_LOOP
                        xor     r6,r6,r6
                        adde    %0,r6,%0"
                        : "+r" (sum), "+r" (link32BitDataPtr)
                        : "r" (bigLoop)
                        : "ctr", "r6"
                        );

                linkLength -= (unsigned)bigLoop * 4;
                link16BitDataPtr = (tt16BitPtr)((void *) (link32BitDataPtr
                                                          + sizeof(tt32Bit)));
            }

/* We need to know if linkLength was odd */
            linkLength &= 1;
            if (sumByteSwapped != TM_8BIT_ZERO)
            {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
                tm_byte_swap_sum(sum);
/* toggle byte swapping */
                sumByteSwapped = TM_8BIT_ZERO;
                if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
                {
                    oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                    sum = sum + oddByte16Bit.oddWord;
                    linkLength = 0;
                }
                else
                {
/* remember that we have our even byte stored in oddChar[0] */
                    linkLength = 1;
                }
            }
            else
            {
                tm_add_carries_sum(sum);
                if (linkLength != 0)
                {
/* Odd byte length. Save the character to add it later on */
                    oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
                }
            }
/* check this before computing next packetPtr */
            if (totalDataLength == (ttPktLen)0)
            {
/* break out of the loop. Do not check 0 length after the loop */
                goto outFold;
            }
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (packetPtr != TM_PACKET_NULL_PTR);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
#ifdef TM_LINT
LINT_BOOLEAN_BEGIN
#endif /* TM_LINT */

        if (totalDataLength != (ttPktLen)0)
#ifdef TM_LINT
LINT_BOOLEAN_END
#endif /* TM_LINT */

        {
            tfKernelError("tfPacketChecksum",
                "Data in Packet chain smaller than length to be checksumed");
        }
#endif
outFold:
        if (linkLength != 0UL)
        {
/*  Add left-over byte */
            oddByte16Bit.oddChar[1] = 0;
            sum = sum + oddByte16Bit.oddWord;
        }
    }
/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
    tm_add_carries_sum(sum);
#ifdef TM_DEV_SEND_OFFLOAD
    if (partialChecksum == TM_8BIT_ZERO)
#endif /* TM_DEV_SEND_OFFLOAD */
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcsppcg.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 25494    Output checksum: 60723
 */
