/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Optimized checksum function for TI C6000 DSP and TI Code
 *              Composer
 *
 * Filename: trcstic6.c
 * Author: Ed Remmell
 * Date Created:
 * $Source: source/ocksum/trcstic6.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:33CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>

/* assembly language implementation of checksum, see trcstic6.sa */
TM_PROTO_EXTERN tt32Bit tfTiC6xAsmChecksum(
    tt32Bit sum, tt32BitPtr link32BitDataPtr, tt32Bit wordCount);

tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
    register tt16BitPtr link16BitDataPtr;
    register tt32BitPtr link32BitDataPtr;
    register tt32Bit    sum;
    register tt32Bit    wordCount;
    ttPktLen            linkLength;
    ttPktLen            shortCount;
    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    }                   oddByte16Bit;
    tt8Bit              sumByteSwapped;
#ifdef TM_DEV_SEND_OFFLOAD
    tt8Bit          partialChecksum;
#endif /* TM_DEV_SEND_OFFLOAD */

    sum = 0UL;

/* sum in the pseudo-header, if necessary */
    if (pshLen != 0)
    {
/* unroll the Pseudo header checksum */
/* 6 words (12 bytes) */
/* Should work OK for 16-bit DSP's */
        sum = pshHdrPtr[0]; sum += pshHdrPtr[1]; sum += pshHdrPtr[2];
        sum += pshHdrPtr[3]; sum += pshHdrPtr[4]; sum += pshHdrPtr[5];
    }

/* packetPtr is assumed to be non null unless checksum is offloaded */
#ifdef TM_DEV_SEND_OFFLOAD
/* packetPtr could be null if checksum is offloaded. */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        partialChecksum = TM_8BIT_YES;
    }
    else
#endif /* TM_DEV_SEND_OFFLOAD */
    {
#ifdef TM_DEV_SEND_OFFLOAD
        partialChecksum = TM_8BIT_ZERO;
#endif /* TM_DEV_SEND_OFFLOAD */
        linkLength = (ttPktLen)0;
        sumByteSwapped = TM_8BIT_ZERO;
        do
        {
/* Point to data in this link */
            link16BitDataPtr = (tt16BitPtr)packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
            if (linkLength == (ttPktLen)0)
            {
                linkLength = packetPtr->pktLinkDataLength;
            }
            else
            {
/* previous link had odd byte length, store 2nd byte to make a word */
                oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
                link16BitDataPtr = (tt16BitPtr)
                                            ((tt8BitPtr)link16BitDataPtr + 1);
                linkLength = packetPtr->pktLinkDataLength - 1;
                totalDataLength--;
            }
            if (linkLength > totalDataLength)
            {
/* Do not span more than asked */
                linkLength = totalDataLength;
            }
/* left over data to be checksumed after this link */
            totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
            if ((tm_is_odd(link16BitDataPtr)) && (linkLength != 0))
            {
/* save odd byte for later addition */
                oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
/* Update pointer, and length. Pointer now is on even byte boundary */
                link16BitDataPtr =
                               (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
                linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
                tm_byte_swap_sum(sum);
                sumByteSwapped = TM_8BIT_YES;
            }

/* Compute number of 16-bit short words left to checksum */
            shortCount = linkLength >> 1;

/* 
 * Catch an extra 16-bit short word here.  After this, the only data to be
 * summed are 32-bit long words and possibly an extra byte.
 */
            if (shortCount && (((tt32Bit) link16BitDataPtr) & 2))
            {
/* link16BitDataPtr must be 32-bit aligned for the assembly checksum */
                sum += *link16BitDataPtr++;
                linkLength -= 2;

                shortCount--;
            }

/* compute number of 32-bit words to checksum in assembly */
            wordCount = (unsigned short)(shortCount >> 1);

            asm(";cksumBigLoop");
            if (wordCount > 0)
            {
/* adjust for the number of 32-bit words we are about to checksum */
                linkLength -= (unsigned)wordCount << 2;
                link32BitDataPtr = (tt32BitPtr) link16BitDataPtr;
                link16BitDataPtr = (tt16BitPtr) (link32BitDataPtr + wordCount);

/* 
 * Call optimized assembly checksum routine, see trcstic6.sa
 */
                sum = tfTiC6xAsmChecksum(sum, link32BitDataPtr, wordCount);
            }
            asm(";END cksumBigLoop");

/* 
 * Catch an extra 16-bit short word here. After this, the only data to be
 * summed is possibly an extra byte.
 */
            if (shortCount & 1)
            {
                sum += *link16BitDataPtr++;
                linkLength -= 2;
            }

/* We need to know if linkLength was odd */
            linkLength &= 1;
            if (sumByteSwapped != TM_8BIT_ZERO)
            {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
                tm_byte_swap_sum(sum);
/* toggle byte swapping */
                sumByteSwapped = TM_8BIT_ZERO;
                if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
                {
                    oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                    sum = sum + oddByte16Bit.oddWord;
                    linkLength = 0;
                }
                else
                {
/* remember that we have our even byte stored in oddChar[0] */
                    linkLength = 1;
                }
            }
            else
            {
                tm_add_carries_sum(sum);
                if (linkLength != 0)
                {
/* Odd byte length. Save the character to add it later on */
                    oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
                }
            }
/* check this before computing next packetPtr */
            if (totalDataLength == (ttPktLen)0)
            {
/* break out of the loop. Do not check 0 length after the loop */
                goto outFold;
            }
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (packetPtr != TM_PACKET_NULL_PTR);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
#ifdef TM_LINT
LINT_BOOLEAN_BEGIN
#endif /* TM_LINT */

        if (totalDataLength != (ttPktLen)0)
#ifdef TM_LINT
LINT_BOOLEAN_END
#endif /* TM_LINT */

        {
            tfKernelError("tfPacketChecksum",
                "Data in Packet chain smaller than length to be checksumed");
        }
#endif
outFold:
        if (linkLength != 0UL)
        {
/*  Add left-over byte */
            oddByte16Bit.oddChar[1] = 0;
            sum = sum + oddByte16Bit.oddWord;
        }
    }
/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
    tm_add_carries_sum(sum);
#ifdef TM_DEV_SEND_OFFLOAD
    if (partialChecksum == TM_8BIT_ZERO)
#endif /* TM_DEV_SEND_OFFLOAD */
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}
/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcstic6.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 21053    Output checksum:  5383
 */
