/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Packet Checksum for x86/Borland/Microsoft real mode only
 * Filename: trcsz80.c
 * Author: 
 * Date Created: 10/23/02
 * $Source: source/ocksum/trcsz80.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:37CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <eZ80L92.h>

/*
 * tfPacketChecksum() Function Description
 * This function performs checksum on (IP, ICMP, IGMP, UDP and TCP)
 * packets. Note that we cannot assume that data is contiguous. We follow
 * the chain of ttPacket (using the pktNextPtr), and compute the checksum
 * on each contiguous data area. We try and unroll the loop as much as we can.
 * This version handles odd byte lengths in inner links, and odd linkDataPtr.
 *
 * Parameter       Description
 * packetPtr       Pointer to a ttPacket. packetPtr->pktLinkDataPtr
 *                 points to the data to be checksummed.
 * length          Length of data to be checksumed.
 * Returns
 * Value   Meaning
 * 16bit one's complement checksum
 */

tt16Bit tfPacketChecksum (
    ttPacketPtr packetPtr,
    ttPktLen totalDataLength,
    tt16BitPtr  pshHdrPtr,
    tt16Bit     pshLen )
{
    tt16BitPtr      link16BitDataPtr;
    tt32Bit         sum;
    ttPktLen        linkLength;
    union
    {
        tt16Bit  oddWord;
        tt8Bit   oddChar[2];
    } oddByte16Bit;
    tt8Bit          sumByteSwapped;
#ifdef TM_DEV_SEND_OFFLOAD
    tt8Bit          partialChecksum;
#endif /* TM_DEV_SEND_OFFLOAD */

    sum = 0UL;

/* sum in the pseudo-header, if necessary */
    while (pshLen > 1)
    {
        sum += *pshHdrPtr++;
        pshLen -= 2;
    }

/* packetPtr is assumed to be non null unless checksum is offloaded */
#ifdef TM_DEV_SEND_OFFLOAD
/* packetPtr could be null if checksum is offloaded. */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        partialChecksum = TM_8BIT_YES;
    }
    else
#endif /* TM_DEV_SEND_OFFLOAD */
    {
#ifdef TM_DEV_SEND_OFFLOAD
        partialChecksum = TM_8BIT_ZERO;
#endif /* TM_DEV_SEND_OFFLOAD */
        sumByteSwapped = TM_8BIT_ZERO;
        linkLength = (ttPktLen)0;
        do
        {
/* Point to data in this link */
            link16BitDataPtr = (tt16BitPtr)packetPtr->pktLinkDataPtr;
/* Loop through scattered data */
            if (linkLength == (ttPktLen)0)
            {
                linkLength = packetPtr->pktLinkDataLength;
            }
            else
            {
/* previous link had odd byte length, store 2nd byte to make a word */
                oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                sum = sum + oddByte16Bit.oddWord;
/* Update pointer, and length. Pointer now is on odd byte boundary */
                link16BitDataPtr = (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
                linkLength = packetPtr->pktLinkDataLength - 1;
                totalDataLength--;
            }
            if (linkLength > totalDataLength)
            {
/* Do not span more than asked */
                linkLength = totalDataLength;
            }
/* left over data to be checksumed after this link */
            totalDataLength -= linkLength;
/* Move pointer to even byte boundary */
            if((tm_is_odd((tt32Bit)link16BitDataPtr)) && (linkLength != 0))
            {
/* save odd byte for later addition */
                oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
/* Update pointer, and length. Pointer now is on even byte boundary */
                link16BitDataPtr =
                                   (tt16BitPtr)((tt8BitPtr)link16BitDataPtr + 1);
                linkLength--;
/* Byte swap the sum, to collect odd bytes with odd bytes */
                tm_byte_swap_sum(sum);
                sumByteSwapped = TM_8BIT_YES;
            }
            while (linkLength >= 64)
            {

/*  Unroll 16 words in one loop pass */
/*          sum += link16BitDataPtr[0]; sum += link16BitDataPtr[1];
                sum += link16BitDataPtr[2]; sum += link16BitDataPtr[3];
                sum += link16BitDataPtr[4]; sum += link16BitDataPtr[5];
                sum += link16BitDataPtr[6]; sum += link16BitDataPtr[7];
                sum += link16BitDataPtr[8]; sum += link16BitDataPtr[9];
                sum += link16BitDataPtr[10]; sum += link16BitDataPtr[11];
                sum += link16BitDataPtr[12]; sum += link16BitDataPtr[13];
                sum += link16BitDataPtr[14]; sum += link16BitDataPtr[15];
*/
/*Following optimized assembly code calculates the sum of 32 words in one loop */

			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLD	BC,(IY)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tLD	DE,HL") ;		/* link16BitDataPtr[0] */
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+2") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[1] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+4") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[2] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+6") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[3] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+8") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[4] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+10") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[5] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+12") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[6] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+14") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[7] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+16") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[8] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+18") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[9] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+20") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[10] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+22") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[11] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+24") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[12] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+26") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[13] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+28") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[14] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+30") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[15] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+32") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[16] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tLEA	HL,IY+34") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[17] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+36") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[18] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+38") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[19] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+40") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[20] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+42") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[21] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+44") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[22] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+46") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[23] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+48") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[24] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+50") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[25] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+52") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[26] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+54") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[27] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+56") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[28] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+58") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[29] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+60") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[30] */
			_asm("\tLD	DE,HL") ;
			_asm("\tLEA	HL,IY+62") ;
			_asm("\tLD	BC,(HL)") ;
			_asm("\tLD	IY,(IX-10)") ;
			_asm("\tPUSH.S BC") ;
			_asm("\tPOP.S HL") ;
			
			_asm("\tADD	HL,DE") ;		/* link16BitDataPtr[31] */
			_asm("\tLD	BC,HL") ;
			_asm("\tCALL __itol") ;
			_asm("\tLD	HL,(IX-4)") ;
			_asm("\tLD	E,(IX-1)") ;
			_asm("\tCALL __ladd") ;
			_asm("\tLD	(IX-4),BC") ;
			_asm("\tLD	(IX-1),A") ;

                link16BitDataPtr += 32;
                linkLength -= 64;
                if (sum & TM_UL(0x80000000))
/* If high order bit set. Fold 32 bit sum into 16 bit */
                {
                    tm_fold_sum(sum);
                }
            }
            while (linkLength >= 8)
            {
/*  Unroll 4 words in one loop pass */
                sum += link16BitDataPtr[0] + link16BitDataPtr[1] \
                +link16BitDataPtr[2]+link16BitDataPtr[3];


                link16BitDataPtr += 4;
                linkLength -= 8;
            }
            while (linkLength > 1)
            {
/* 1 word at a time */
                sum += *link16BitDataPtr++;
                linkLength -= 2;
            }
            if (sumByteSwapped != TM_8BIT_ZERO)
            {
/*
 * We want to add the saved even byte, so byte swap the sum to start
 * collecting the even bytes.
 */
                tm_byte_swap_sum(sum);
/* toggle byte swapping */
                sumByteSwapped = TM_8BIT_ZERO;
                if (linkLength != 0)
/* extra byte at the end. We can pair it with our saved even byte */
                {
                    oddByte16Bit.oddChar[1] = *((tt8BitPtr)link16BitDataPtr);
/* add to sum */
                    sum = sum + oddByte16Bit.oddWord;
                    linkLength = 0;
                }
                else
                {
/* remember that we have our even byte stored in oddChar[0] */
                    linkLength = 1;
                }
            }
            else
            {
                if (linkLength != 0)
                {
/* Odd byte length. Save the character to add it later on */
                    oddByte16Bit.oddChar[0] = *((tt8BitPtr)link16BitDataPtr);
                }
            }

/* check this before computing next packetPtr */
            if (totalDataLength == (ttPktLen)0)
            {
/* break out of the loop. Do not check 0 length after the loop */
                goto outFold;
            }
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (packetPtr != TM_PACKET_NULL_PTR);
#ifdef TM_ERROR_CHECKING
/* double check that all data has been checksumed */
#ifdef TM_LINT
LINT_BOOLEAN_BEGIN
#endif /* TM_LINT */
        if (totalDataLength != (ttPktLen)0)
#ifdef TM_LINT
LINT_BOOLEAN_END
#endif /* TM_LINT */
        {
            tfKernelError("tfPacketChecksum",
                "Data in Packet chain smaller than length to be checksumed");
        }
#endif /* TM_ERROR_CHECKING */
outFold:
        if (linkLength != TM_UL(0))
        {
/*  Add left-over byte */
            oddByte16Bit.oddChar[1] = 0;
            sum = sum + oddByte16Bit.oddWord;
        }
    }
/*  Fold 32-bit sum to 16 bits, adding posssible carries from fold */
    tm_add_carries_sum(sum);
#ifdef TM_DEV_SEND_OFFLOAD
    if (partialChecksum == TM_8BIT_ZERO)
#endif /* TM_DEV_SEND_OFFLOAD */
    {
        sum = ~sum;
    }
    return (tt16Bit)sum;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcsz80.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  9128    Output checksum: 29990
 */
