/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Sockets Interface (tfAddInterfaceMhomeAddress)
 *
 * Filename: traddmhm.c
 * Author: Odile
 * Date Created: 10/01/99
 * $Source: source/sockapi/traddmhm.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2010/09/20 14:28:28CDT $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>


#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * tfAddInterfaceMhomeAddress() function description:
 * Add a multihome IP address to the interface without adding it to
 * the routing table
 * Parameters:
 * interfaceHandle  User device interface handle
 * ipAddress        User IP address
 * mhomeIndex       Mhome index of the configuration on the interface
 * Return value:
 * TM_ENOERROR      No error
 * TM_EADDRNOTAVAIL Invalid IP address
 * TM_EINVAL        Invalid interface
 * TM_EALREADY      Interface already configured for that mhome address
 */
int tfAddInterfaceMhomeAddress ( ttUserInterface interfaceId,
                                 ttUserIpAddress ipAddress,
                                 unsigned char   mHomeIndex )
{
    return tf4AddInterfaceMhomeAddress(
        interfaceId, ipAddress, (tt16Bit) mHomeIndex);
}

int tf4AddInterfaceMhomeAddress ( ttUserInterface interfaceId,
                                 ttUserIpAddress ipAddress,
                                 tt16Bit mHomeIndex )
{
    ttDeviceEntryPtr  devEntryPtr;
	int               errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceId;
    errorCode = TM_ENOERROR;
/*
 * Parameter check: check that Ip address is not limited broadcast,
 * or zero IP address, or class D or E address.
 */
    if (    tm_ip_match(ipAddress, TM_IP_LIMITED_BROADCAST)
         || tm_ip_zero(ipAddress)
         || tm_ip_is_class_d_e(ipAddress) )
    {
        errorCode = TM_EADDRNOTAVAIL;
        goto addMhomeFinish;
    }
/*
 * Parameters check: check that mhome is within range allowed, that
 * interface is valid
 */
    if (    ((int)mHomeIndex >= TM_MAX_IPS_PER_IF)
         || ((int)mHomeIndex == 0)
         || (tfValidInterface(devEntryPtr) != TM_ENOERROR) )
    {
        errorCode = TM_EINVAL;
        goto addMhomeFinish;
    }
/* lock the entry */
    tm_call_lock_wait(&(devEntryPtr->devLockEntry));
    if ( ! (tm_ip_dev_conf_flag(devEntryPtr, mHomeIndex) & TM_DEV_IP_CONFIG) )
    {
        tm_ip_copy( (tt4IpAddress)ipAddress,
                    tm_ip_dev_addr(devEntryPtr, mHomeIndex) );
        tm_ip_copy( (tt4IpAddress)TM_IP_HOST_MASK,
                    tm_ip_dev_mask(devEntryPtr, mHomeIndex) );
        tm_4_ip_dev_prefixLen(devEntryPtr,mHomeIndex) = (tt8Bit) 32;
        tm_ip_copy(TM_IP_ALL_ONES,
                           tm_ip_dev_dbroad_addr(devEntryPtr, mHomeIndex));
/* Ip address has been configured successfully */
        tm_ip_dev_conf_flag(devEntryPtr, mHomeIndex) = TM_DEV_IP_CONFIG;
#ifndef TM_SINGLE_INTERFACE_HOME
/* Keep track of maximum configured interface */
        if (mHomeIndex >= devEntryPtr->devMhomeUsedEntries)
        {
            devEntryPtr->devMhomeUsedEntries = (tt16Bit)((int)mHomeIndex + 1);
        }
#endif /* TM_SINGLE_INTERFACE_HOME */
    }
    else
    {
/*
 * If this interface/mHomeIndex index has already been configured, return with
 * TM_EALREADY error
 */
        errorCode = TM_EALREADY;
    }
/* unlock the entry */
    tm_call_unlock(&(devEntryPtr->devLockEntry));
addMhomeFinish:
    return errorCode;
#undef devEntryPtr
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: traddmhm.c    File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  9582    Output checksum: 38179
 */
