/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (tfAddStaticRoute)
 *
 * Filename: traddsrt.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/traddsrt.c $
 *
 * Modification History
 * $Revision: 6.0.2.4 $
 * $Date: 2011/01/17 13:29:49CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


#ifdef TM_SINGLE_INTERFACE_HOME
/* Route IP address does not match my network IP address */
#define tm_rt_network_not_match(devPtr, ipAddr) \
        (    ((~devPtr->devNetMask) & ipAddr ) \
          || tm_ip_match(devPtr->devNetMask, TM_IP_HOST_MASK) )
/* Route IP address does not match my directed broadcast IP address */
#define tm_rt_dbroad_not_match(devPtr, ipAddr) \
        (    !tm_4_ll_is_broadcast(devPtr) \
          || tm_ip_not_match(ipAddr, devPtr->devDBroadIpAddr))
#endif /* TM_SINGLE_INTERFACE_HOME */

#ifndef TM_SINGLE_INTERFACE_HOME
static int tfRtNetworkNotMatch(ttDeviceEntryPtr devPtr,
                               ttUserIpAddress  ipAddr,
                               tt16Bit          ifaceMhome);
static int tfRtDbroadNotMatch(ttDeviceEntryPtr devPtr,
                              ttUserIpAddress  ipAddr,
                              tt16Bit          ifaceMhome);

/* Route IP address does not match my network IP address */
static int tfRtNetworkNotMatch(ttDeviceEntryPtr devPtr,
                               ttUserIpAddress  ipAddr,
                               tt16Bit          ifaceMhome)
{
    int     retVal;

    retVal = tfMhomeNetMatch(devPtr, ipAddr, &ifaceMhome);
    if (   (retVal != TM_ENOERROR)
        || ((~(devPtr->devNetMaskArray[ifaceMhome])) & ipAddr)
        || tm_ip_match(devPtr->devNetMaskArray[ifaceMhome],
                       TM_IP_HOST_MASK) )
    {
        retVal = 1;
    }
    else
    {
        retVal = 0;
    }

    return retVal;
}


/* Route IP address does not match my directed broadcast IP address */
static int tfRtDbroadNotMatch(ttDeviceEntryPtr devPtr,
                              ttUserIpAddress  ipAddr,
                              tt16Bit          ifaceMhome)
{
    int     retVal;
    int     errorCode;

    retVal = 1; /* assume no directed broadcast match */
    if (tm_4_ll_is_broadcast(devPtr))
    {
        errorCode = tfMhomeDBroadMatch(devPtr, ipAddr, &ifaceMhome);
        if (errorCode == TM_ENOERROR)
        {
            retVal = 0; /* directed broadcast match */
        }
    }
    return retVal;
}
#endif /* !TM_SINGLE_INTERFACE_HOME */

/*
 * Called from the socket interface to add a static route.
 * Parameters:
 * InterfaceId: pointer to the device entry
 * destIpAddress: destination IP address
 * destNetMask: destination Network mask
 * gateway: IP address of the gateway for this route
 * Hops: Number of gateway hops to destination
 * Return values:
 * TM_EINVAL if any of the first 4 parameters is zero
 * for additional error values see tfRtAddEntry()
 */
int tfAddStaticRoute (ttUserInterface interfaceId,
                      ttUserIpAddress destIpAddress,
                      ttUserIpAddress destNetMask,
                      ttUserIpAddress gateway,
                      int             hops)
{
    ttDeviceEntryPtr    devPtr;
    int                 errorCode;
    int                 retCode;
    tt16Bit             flags;
#ifndef TM_SINGLE_INTERFACE_HOME
    tt16Bit             mHomeIndex;
#endif /* !TM_SINGLE_INTERFACE_HOME */

#ifndef TM_SINGLE_INTERFACE_HOME
    mHomeIndex = 0;
#endif /* !TM_SINGLE_INTERFACE_HOME */
    devPtr = (ttDeviceEntryPtr)interfaceId;
    errorCode = tfValidConfigInterface( devPtr, TM_16BIT_ZERO );
    if (errorCode == TM_ENOERROR)
    {
        errorCode = TM_EINVAL;
        if (  (tm_ip_not_zero(destIpAddress))
           && (tm_ip_not_zero(destNetMask))
           && (tm_ip_not_zero(tm_ip_net( (tt4IpAddress)destIpAddress,
                                         (tt4IpAddress)destNetMask )))
           && (   (tm_ip_not_zero(gateway))
               || (!(tm_4_ll_is_lan(devPtr))) )
           && (hops > 0) )
        {
            retCode = 1;
            if (   (!tm_ll_is_pt2pt(devPtr))
                && (!(tm_ip_match(TM_IP_LOOP_BACK_ADDR, gateway))))
            {
/*
 * BUG ID 201: Broadcast and network host address should not be used as
 * static routes. (Only relevant for non pt2pt link layers.)
 */
/* Route IP address does not match my network IP address */
#ifdef TM_SINGLE_INTERFACE_HOME
                retCode = tm_rt_network_not_match(devPtr,gateway);
#else /* !TM_SINGLE_INTERFACE_HOME */
                retCode = tfRtNetworkNotMatch(devPtr,gateway,mHomeIndex);
#endif /* !TM_SINGLE_INTERFACE_HOME */
                if (retCode)
                { 
/* Route IP address does not match my directed broadcast IP address */
#ifdef TM_SINGLE_INTERFACE_HOME
                    retCode = tm_rt_dbroad_not_match(devPtr, gateway);
#else /* !TM_SINGLE_INTERFACE_HOME */
                    retCode = tfRtDbroadNotMatch(devPtr, gateway, mHomeIndex);
#endif /* !TM_SINGLE_INTERFACE_HOME */
                }
            }
            if (retCode)
            {
                if (tm_ip_is_loop_back(destIpAddress))
                {
                    flags = TM_RTE_REJECT|TM_RTE_INDIRECT|TM_RTE_STATIC;
                }
                else
                {
                    flags = TM_RTE_INDIRECT|TM_RTE_STATIC;
                }
                errorCode = tf4RtAddRoute((ttDeviceEntryPtr)interfaceId, 
                                          destIpAddress, destNetMask,
                                          gateway, TM_16BIT_ZERO,
                                          TM_16BIT_ZERO, hops , TM_RTE_INF,
                                          flags, TM_8BIT_YES);
            }
        }
    }
    return errorCode;
}

/****************************************************************************
* FUNCTION: tfNgAddStaticRoute
*
* PURPOSE: Next Generation API replaces tfAddStaticRoute
*
* PARAMETERS:
*   None
*
* RETURNS:
*   Nothing
*
* NOTES:
*
****************************************************************************/
int tfNgAddStaticRoute(
    ttUserInterface interfaceHandle,
    const struct sockaddr_storage TM_FAR * destIpAddrPtr,
    int prefixLen,
    const struct sockaddr_storage TM_FAR * gatewayPtr,
    int hops )
{
    int                 errorCode;

    errorCode   = TM_ENOERROR; /* assume success */

/* parameter validation */
    if ((destIpAddrPtr == (ttConstSockAddrStoragePtr)0)
        || (gatewayPtr == (ttConstSockAddrStoragePtr)0)
        || (destIpAddrPtr->ss_family != gatewayPtr->ss_family))
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        switch (destIpAddrPtr->ss_family)
        {
        case AF_INET:
            if (prefixLen > 32)
            {
                errorCode = TM_EINVAL;
            }
            else
            {
                errorCode = tfAddStaticRoute(
                    interfaceHandle,
                    destIpAddrPtr->addr.ipv4.sin_addr.s_addr,
                    tvRt32ContiguousBitsPtr[prefixLen],
                    gatewayPtr->addr.ipv4.sin_addr.s_addr,
                    hops);
            }
            break;


        default:
            errorCode = TM_EAFNOSUPPORT;
            break;
        }
    }

    return errorCode;
}



/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: traddsrt.c    File Revision: 6.0.2.4 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  5629    Output checksum: 45508
 */
