/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2009
 * Copyright Elmic Systems Japan 1997 - 2009
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     Address Selection Policy Table Interface
 * Filename:        traplcy.c
 * Author:          Peter
 * Date Created:    11/29/09
 * $Source: source/sockapi/traplcy.c $
 *
 * Modification History
 * $Revision: 6.0.2.5 $
 * $Date: 2010/08/20 01:20:11CDT $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_RFC3484
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

typedef struct tsAddrPolicyInit
{
    char    apiPrefix[TM_6_IP_HOST_PREFIX_LENGTH];
    int     apiPrefixLen;
    tt8Bit  apiPrecedence;
    tt8Bit  apiLabel;
} ttAddrPolicyInit;


/*
 * Default initialization table per RFC 3484 section 2.1
 * NOTE: This table MUST be ordered by decending prefix length
 */
static const ttAddrPolicyInit tvAddrPolicyTableInit[] =
{
/* Loopback */
    { "::1", /* The IPv6 prefix */
      128,   /* The prefix length */
      50,    /* The Precedence */
      0      /* The Label */
    },
/* IPv4 compatible */
    { "::",         96, 20, 3 },
/* IPv4 mapped */
    { "::ffff:0:0", 96, 10, 4 },
/* 6to4 */
    { "2002::",     16, 30, 2 },
/* other */
    { "::",          0, 40, 1 }
};

static int tfAddrPolicyTableFindEntryByPrefixCB(
                                        ttNodePtr      nodePtr,
                                        ttGenericUnion genParam);

static int tfAddrPolicyTableFindFirstSmallerEntry(
                                        ttNodePtr      nodePtr,
                                        ttGenericUnion genParam);

static int tfAddrPolicyTableFindEntryByNodeCB(
                                        ttNodePtr      nodePtr,
                                        ttGenericUnion genParam);

/* Initialize default address selection policy table */
int tfAddrPolicyInit (void)
{
    ttNodePtr               nodePtr;
    ttAddrPolicyEntryPtr    addrPolicyEntryPtr;
    int                     i;
    int                     initTableSize;
    int                     prevPrefixLen;
    int                     retCode;
    int                     errorCode;

#ifdef TM_LOCK_NEEDED
    tm_bzero(&tm_context(tvAddrPolicyTableLockEntry), sizeof(ttLockEntry));
#endif /* TM_LOCK_NEEDED */
    tfListInit(&(tm_context(tvAddrPolicyTable)));

    initTableSize = sizeof(tvAddrPolicyTableInit) / sizeof(ttAddrPolicyInit);
    prevPrefixLen = TM_6_IP_HOST_PREFIX_LENGTH;
    errorCode = TM_ENOERROR;
    for (i=0; i < initTableSize; i++)
    {
        nodePtr = tm_get_raw_buffer(sizeof(ttAddrPolicyEntry));
        if (nodePtr != TM_NODE_NULL_PTR)
        {
            addrPolicyEntryPtr = (ttAddrPolicyEntryPtr)nodePtr;
            tm_bzero(nodePtr, sizeof(ttAddrPolicyEntry));
            retCode = inet_pton(AF_INET6,
                                tvAddrPolicyTableInit[i].apiPrefix,
                                &(addrPolicyEntryPtr->apePrefix));
            if (retCode != 1)
            {
/* IPv6 Prefix was not a properly formatted IPv6 address */
                errorCode = TM_EINVAL;
                break;
            }
            if (prevPrefixLen < tvAddrPolicyTableInit[i].apiPrefixLen)
            {
/* Initialization table was not in decending order of prefix length */
                errorCode = TM_EINVAL;
                break;
            }
            prevPrefixLen = tvAddrPolicyTableInit[i].apiPrefixLen;
            addrPolicyEntryPtr->apePrefixLen =
                        tvAddrPolicyTableInit[i].apiPrefixLen;
            addrPolicyEntryPtr->apePrecedence =
                        tvAddrPolicyTableInit[i].apiPrecedence;
            addrPolicyEntryPtr->apeLabel =
                        tvAddrPolicyTableInit[i].apiLabel;
            tfListAddToTail(&(tm_context(tvAddrPolicyTable)), nodePtr);
        }
        else
        {
            errorCode = TM_ENOBUFS;
            break;
        }
    }
    if (errorCode != TM_ENOERROR)
    {
        tfListFree(&(tm_context(tvAddrPolicyTable)));
    }
    return errorCode;
}

/****************************************************************************
* FUNCTION: tfAddrPolicyTableClear
*
* PURPOSE:
*   Clear all entries from the DNS Policy Table
*
* PARAMETERS:
*   flags: Currently unused
*
* RETURNS:
*   TM_ENOERROR: Operation completed successfully
*
****************************************************************************/
int tfAddrPolicyTableClear(int flags)
{
    TM_UNREF_IN_ARG(flags);

    tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));
    if (tm_context(tvAddrPolicyTable).listCount != 0)
    {
        tfListFree(&(tm_context(tvAddrPolicyTable)));
    }
    tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));
    return TM_ENOERROR;
}

/****************************************************************************
* FUNCTION: tfAddrPolicyTableAddEntry
*
* PURPOSE:
*   Add a (blank) entry to the DNS Policy Table
*
* PARAMETERS:
*   ttUserAddrPolicyEntryPtr: Pointer to the user policy structure to add
*   structType:              The type of structure to add
*                            Currently, TM_ADDR_USER_POLICY_STRUCT_1 is the
*                            only valid value for this variable.
*   flags:                   Currently unused
*
* RETURNS:
*   ttUserAddrPolicyEntryHandle: Handle to the new policy that must be
*                               passed back to the other DNS Policy Table
*                               API functions
*   NULL:                       The memory could not be allocated or
*                               tfStartTreck() has not been called
*
****************************************************************************/
ttUserAddrPolicyEntryHandle tfAddrPolicyTableAddEntry(
                                ttUserAddrPolicyEntryPtr userPolicyPtr,
                                int                     structType,
                                int                     flags)
{
    ttNodePtr       nodePtr;
    ttNodePtr       insertNodePtr;
    ttGenericUnion  genParam;

    TM_UNREF_IN_ARG(flags);

    if (   (userPolicyPtr == (ttUserAddrPolicyEntryPtr)0)
        || (structType != TM_ADDR_USER_POLICY_STRUCT_1) )
    {
        nodePtr = TM_NODE_NULL_PTR;
    }
    else if (userPolicyPtr->aupePrefixLen > TM_6_IP_HOST_PREFIX_LENGTH)
    {
        nodePtr = TM_NODE_NULL_PTR;
    }
    else
    {
        nodePtr = tm_get_raw_buffer(sizeof(ttAddrPolicyEntry));
        if (nodePtr != TM_NODE_NULL_PTR)
        {
            tm_bzero(nodePtr, sizeof(ttAddrPolicyEntry));
            tm_memcpy(&(((ttAddrPolicyEntryPtr)nodePtr)->apePrefix),
                      &(userPolicyPtr->aupePrefix),
                      sizeof(tt6IpAddress));
            ((ttAddrPolicyEntryPtr)nodePtr)->apePrefixLen =
                                         userPolicyPtr->aupePrefixLen;
            ((ttAddrPolicyEntryPtr)nodePtr)->apePrecedence =
                                         userPolicyPtr->aupePrecedence;
            ((ttAddrPolicyEntryPtr)nodePtr)->apeLabel=
                                         userPolicyPtr->aupeLabel;

            tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));
            genParam.genVoidParmPtr = (ttVoidPtr)nodePtr;
/* Make sure a node with this prefix/prefix length combination is unique */
            insertNodePtr = tfListWalk(&(tm_context(tvAddrPolicyTable)),
                                       tfAddrPolicyTableFindEntryByPrefixCB,
                                       genParam);
            if (insertNodePtr == TM_NODE_NULL_PTR)
            {
                genParam.genIntParm = userPolicyPtr->aupePrefixLen;
                insertNodePtr = tfListWalk(
                                    &(tm_context(tvAddrPolicyTable)),
                                    tfAddrPolicyTableFindFirstSmallerEntry,
                                    genParam);
                if (insertNodePtr == TM_NODE_NULL_PTR)
/* Add to Tail (before head) */
                {
                    insertNodePtr =
                            &(tm_context(tvAddrPolicyTable).listHeadNode);
                }
                tfListAddAtNode(&(tm_context(tvAddrPolicyTable)), nodePtr,
                                insertNodePtr);
            }
            else
            {
                tm_free_raw_buffer(nodePtr);
                nodePtr = TM_NODE_NULL_PTR;
            }
            tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));
            if (nodePtr != TM_NODE_NULL_PTR)
            {
                tf6UpdateCacheLabels();
            }
        }
    }
    return (ttUserAddrPolicyEntryHandle)nodePtr;
}


/****************************************************************************
* FUNCTION: tfAddrPolicyTableDisplayEntry
*
* PURPOSE:
*   Display an RFC3484 Policy Table Entry
*   This function will return a single entry of the Policy Table. Given
*   a "current entry", this function will return the next entry in the table.
*   If the "current entry" is null, it will return the first entry.
*   If the "current entry" is the last entry, it will return null.
*
* PARAMETERS:
*   ttUserAddrPolicyEntryHandle: Handle, as returned by tfAddrPolicyTableAddEntry
*   ttUserAddrPolicyEntryPtr:    Pointer to a user policy structure to fill out
*   structType:                 The type of structure to fill out
*                               Currently, TM_ADDR_USER_POLICY_STRUCT_1 is the
*                               only valid value for this variable.
*   flags:                      Currently unused
*
* RETURNS:
*   ttUserAddrPolicyEntryHandle: Handle to the policy whose parameters are
*                               contained in the ttUserAddrPolicyEntryPtr
*   NULL:                       The specified handle was invalid or was the
*                               last entry in the list.
*
****************************************************************************/
ttUserAddrPolicyEntryHandle tfAddrPolicyTableDisplayEntry(
                                ttUserAddrPolicyEntryHandle  apeHandle,
                                ttUserAddrPolicyEntryPtr     userPolicyPtr,
                                int                          structType,
                                int                          flags)
{
    ttAddrPolicyEntryPtr        addrPolicyEntryPtr;
    ttNodePtr                   nodePtr;
    ttGenericUnion              genParam;
    int                         errorCode;

    TM_UNREF_IN_ARG(flags);

    nodePtr = TM_NODE_NULL_PTR;
    errorCode = TM_ENOERROR;

    if (   (tm_context(tvAddrPolicyTable).listCount == 0)
        || (userPolicyPtr == (ttUserAddrPolicyEntryPtr)0)
        || (structType != TM_ADDR_USER_POLICY_STRUCT_1) )
    {
        errorCode = TM_EINVAL;
    }

    tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));

/* If the user specified a handle, make sure it is valid */
    if (errorCode == TM_ENOERROR)
    {
        if (apeHandle != (ttUserAddrPolicyEntryHandle)0)
        {
            genParam.genVoidParmPtr = (ttVoidPtr)apeHandle;
/* Make sure this node exists */
            nodePtr = tfListWalk(&(tm_context(tvAddrPolicyTable)),
                                 tfAddrPolicyTableFindEntryByNodeCB,
                                 genParam);
            if (nodePtr != TM_NODE_NULL_PTR)
            {
/* We found the node, make sure it is not the last node */
                nodePtr = nodePtr->nodeNextPtr;
                if (nodePtr ==
                      &((tm_context(tvAddrPolicyTable)).listHeadNode))
                {
                    nodePtr = TM_NODE_NULL_PTR;
                    errorCode = TM_EINVAL;
                }
            }
            else
            {
/* Node is not in the list */
                errorCode = TM_EINVAL;
            }
        }
/* If the user did not specify a handle, return the head (if it exists) */
        else if ((tm_context(tvAddrPolicyTable)).listCount > 0)
        {
            nodePtr = (tm_context(tvAddrPolicyTable)).listHeadNode.nodeNextPtr;
        }
        else
        {
            nodePtr = TM_NODE_NULL_PTR;
            errorCode = TM_EINVAL;
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        addrPolicyEntryPtr = (ttAddrPolicyEntryPtr)nodePtr;
        tm_memcpy(&(userPolicyPtr->aupePrefix),
                  &(addrPolicyEntryPtr->apePrefix),
                  sizeof(tt6IpAddress));
        userPolicyPtr->aupePrefixLen = addrPolicyEntryPtr->apePrefixLen;
        userPolicyPtr->aupePrecedence = addrPolicyEntryPtr->apePrecedence;
        userPolicyPtr->aupeLabel = addrPolicyEntryPtr->apeLabel;
    }

    tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));

    return (ttUserAddrPolicyEntryHandle)nodePtr;
}


/****************************************************************************
* FUNCTION: tfAddrPolicyTableModifyEntry
*
* PURPOSE:
*   Set options of an entry in the DNS Policy Table
*
* PARAMETERS:
*   ttUserAddrPolicyEntryHandle: Handle, as returned by tfAddrPolicyTableAddEntry
*   optionName:                 Name of the option to set
*   optionValuePtr:             Pointer to the value of the option to set
*   optionLength:               Length of the data pointed to by optionValuePtr
*   flags:                      Currently unused
*
* RETURNS:
*   TM_ENOERROR:    Operation completed successfully
*   TM_EINVAL:      One of the parameters was invalid
*   TM_EPERM:       Empty address policy table
*
****************************************************************************/
int tfAddrPolicyTableModifyEntry(ttUserAddrPolicyEntryHandle apeHandle,
                                int                        optionName,
                                ttUserVoidPtr              optionValuePtr,
                                int                        optionLength,
                                int                        flags)
{
    ttAddrPolicyEntryPtr addrPolicyEntryPtr;
    ttNodePtr           nodePtr;
    ttGenericUnion      genParam;
    int                 errorCode;

    TM_UNREF_IN_ARG(flags);

    errorCode = TM_ENOERROR;
    nodePtr = TM_NODE_NULL_PTR;

    if (tm_context(tvAddrPolicyTable).listCount == 0)
    {
        errorCode = TM_EPERM;
    }
    else if (   (apeHandle == (ttUserAddrPolicyEntryHandle)0)
             || (optionName == 0)
             || (optionValuePtr == (ttUserVoidPtr)0)
             || (optionLength == 0) )
    {
        errorCode = TM_EINVAL;
    }

    tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));
    if (errorCode == TM_ENOERROR)
    {
        genParam.genVoidParmPtr = (ttVoidPtr)apeHandle;
/* Make sure this node exists */
        nodePtr = tfListWalk(&(tm_context(tvAddrPolicyTable)),
                             tfAddrPolicyTableFindEntryByNodeCB,
                             genParam);
        if (nodePtr == TM_NODE_NULL_PTR)
        {
/* Node is not in the list */
            errorCode = TM_EINVAL;
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        addrPolicyEntryPtr = (ttAddrPolicyEntryPtr)nodePtr;
        switch (optionName)
        {
        case TM_ADDR_POLICY_TABLE_OPT_PREFIX:
            if (optionLength == sizeof(tt6IpAddress))
            {
                tm_memcpy(&(addrPolicyEntryPtr->apePrefix),
                          optionValuePtr,
                          sizeof(tt6IpAddress));
            }
            else
            {
                errorCode = TM_EINVAL;
            }
            break;
        case TM_ADDR_POLICY_TABLE_OPT_PREFIX_LEN:
            if (   (optionLength == sizeof(int))
                && ((int)(*((ttIntPtr)(optionValuePtr))) <=
                                    TM_6_IP_HOST_PREFIX_LENGTH) )
            {
                addrPolicyEntryPtr->apePrefixLen =
                                    (int)(*((ttIntPtr)(optionValuePtr)));
            }
            else
            {
                errorCode = TM_EINVAL;
            }
            break;
        case TM_ADDR_POLICY_TABLE_OPT_PRECEDENCE:
            if (optionLength == sizeof(ttUser8Bit))
            {
                addrPolicyEntryPtr->apePrecedence =
                                    (tt8Bit)(*((tt8BitPtr)(optionValuePtr)));
            }
            else
            {
                errorCode = TM_EINVAL;
            }
            break;
        case TM_ADDR_POLICY_TABLE_OPT_LABEL:
            if (optionLength == sizeof(ttUser8Bit))
            {
                addrPolicyEntryPtr->apeLabel =
                                    (tt8Bit)(*((tt8BitPtr)(optionValuePtr)));
            }
            else
            {
                errorCode = TM_EINVAL;
            }
            break;
        default:
            errorCode = TM_EINVAL;
            break;
        }
    }
    tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));
    if (    (optionName == TM_ADDR_POLICY_TABLE_OPT_LABEL)
         && (errorCode == TM_ENOERROR) )
    {
        tf6UpdateCacheLabels();
    }
    return errorCode;
}


/****************************************************************************
* FUNCTION: tfAddrPolicyTableRemoveEntry
*
* PURPOSE:
*   Remove an entry from the DNS Policy Table
*
* PARAMETERS:
*   ttUserAddrPolicyEntryHandle: Handle, as returned by tfAddrPolicyTableAddEntry
*   flags:                      Currently unused
*
* RETURNS:
*   TM_ENOERROR: Operation completed successfully
*   TM_EINVAL:   ttUserAddrPolicyEntryHandle was invalid
*   TM_EPERM:    Empty address policy table
*
****************************************************************************/
int tfAddrPolicyTableRemoveEntry(ttUserAddrPolicyEntryHandle apeHandle,
                                 int                        flags)
{
    ttNodePtr       nodePtr;
    ttGenericUnion  genParam;
    int             errorCode;

    TM_UNREF_IN_ARG(flags);

    nodePtr = TM_NODE_NULL_PTR;
    errorCode = TM_ENOERROR;

    if (tm_context(tvAddrPolicyTable).listCount == 0)
    {
        errorCode = TM_EPERM;
    } else if (apeHandle == (ttUserAddrPolicyEntryHandle)0)
    {
        errorCode = TM_EINVAL;
    }

    tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));
    if (errorCode == TM_ENOERROR)
    {
        genParam.genVoidParmPtr = (ttVoidPtr)apeHandle;
/* Make sure this node exists */
        nodePtr = tfListWalk(&(tm_context(tvAddrPolicyTable)),
                             tfAddrPolicyTableFindEntryByNodeCB,
                             genParam);
        if (nodePtr == TM_NODE_NULL_PTR)
        {
/* Node is not in the list */
            errorCode = TM_EINVAL;
        }
    }

    if (errorCode == TM_ENOERROR)
    {
        tfListRemove(&(tm_context(tvAddrPolicyTable)), nodePtr);
        tm_free_raw_buffer(nodePtr);
    }
    tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));

    return errorCode;
}

/* Get address label from address selection policy table */
tt8Bit tfAddrPolicyGetLabel(tt6ConstIpAddressPtr ipAddrPtr)
{
    ttNodePtr               nodePtr;
    struct sockaddr_storage sockAddr;
    ttGenericUnion          genParam;
    tt8Bit                  label;

    label = (tt8Bit)1; /* default */
    tm_bzero(&sockAddr, sizeof(struct sockaddr_storage));
    sockAddr.ss_family = AF_INET6;
    tm_6_ip_copy(ipAddrPtr, &(sockAddr.addr.ipv6.sin6_addr));
    genParam.genVoidParmPtr = (ttVoidPtr)&sockAddr;
    tm_call_lock_wait(&tm_context(tvAddrPolicyTableLockEntry));
    nodePtr = tfListWalk(&(tm_context(tvAddrPolicyTable)),
                         tfAddrPolicyTableFindMatchByPrefixCB,
                         genParam);
    if (nodePtr != (ttNodePtr)0)
    {
        label = ((ttAddrPolicyEntryPtr)nodePtr)->apeLabel;
    }
    tm_call_unlock(&tm_context(tvAddrPolicyTableLockEntry));
    return label;
}

/* Find the best match in the list for the given tt6IpAddressPtr */
int tfAddrPolicyTableFindMatchByPrefixCB(ttNodePtr      nodePtr,
                                        ttGenericUnion genParam)
{
    struct sockaddr_storage TM_FAR *    basePtr;
    tt6IpAddressPtr                     candidatePtr;
    tt6IpAddress                        base6Addr;
    tt6IpAddressPtr                     base6AddrPtr;
    unsigned int                        compareSizeBytes;
    unsigned int                        compareSizeBits;
    unsigned int                        i;
    int                                 memcmpResult;
    u_char                              bitcmpResult;
    tt8Bit                              found;

    memcmpResult = 0;
    found = TM_8BIT_NO;

    basePtr = (struct sockaddr_storage TM_FAR *)(genParam.genVoidParmPtr);
    tm_bzero(&base6Addr, sizeof(tt6IpAddress));
    if (basePtr->ss_family == AF_INET)
    {
        tm_6_addr_to_ipv4_mapped((basePtr->addr.ipv4.sin_addr.s_addr),
                                 &(base6Addr));
                                
        base6AddrPtr = &(base6Addr);
    }
    else
    {
        base6AddrPtr = &(basePtr->addr.ipv6.sin6_addr);
    }
    candidatePtr = &(((ttAddrPolicyEntryPtr)nodePtr)->apePrefix);

    compareSizeBytes = ((ttAddrPolicyEntryPtr)nodePtr)->apePrefixLen / 8;
    if (compareSizeBytes > 0)
    {
        memcmpResult = tm_memcmp((ttVoidPtr)(base6AddrPtr->s6LAddr),
                                 (ttVoidPtr)(candidatePtr->s6LAddr),
                                 compareSizeBytes);
    }
    if (memcmpResult == 0)
    {
        compareSizeBits = ((ttAddrPolicyEntryPtr)nodePtr)->apePrefixLen % 8;
        if (compareSizeBits == 0)
        {
            found = TM_8BIT_YES;
        }
        else
        {
            bitcmpResult = (u_char) (base6AddrPtr->s6_addr[compareSizeBytes] ^
                                     candidatePtr->s6_addr[compareSizeBytes]);
            if (bitcmpResult == 0)
            {
                found = TM_8BIT_YES;
            }
            else
            {
                for (i=0; i<compareSizeBits; i++)
                {
                    if ((bitcmpResult & 0x80) == 0)
                    {
                        bitcmpResult <<= 1;
                    }
                    else
                    {
/* Does not match */
                        break;
                    }
                }
                if (i == compareSizeBits)
                {
                    found = TM_8BIT_YES;
                }
            }
        }
    }

    return (int)found;
}


/* Find a node in the list by prefix and prefix length */
static int tfAddrPolicyTableFindEntryByPrefixCB(ttNodePtr      nodePtr,
                                                ttGenericUnion genParam)
{
    tt6IpAddressPtr     basePtr;
    tt6IpAddressPtr     candidatePtr;
    int                 memcmpResult;
    tt8Bit              found;

    basePtr = &(((ttAddrPolicyEntryPtr)(genParam.genVoidParmPtr))->apePrefix);
    candidatePtr = &(((ttAddrPolicyEntryPtr)nodePtr)->apePrefix);
    memcmpResult = tm_memcmp((ttVoidPtr)basePtr,
                             (ttVoidPtr)candidatePtr,
                             sizeof(tt6IpAddress));

    if (   (memcmpResult == 0)
        && ((((ttAddrPolicyEntryPtr)nodePtr)->apePrefixLen) ==
            (((ttAddrPolicyEntryPtr)(genParam.genVoidParmPtr))->apePrefixLen)) )
    {
        found = TM_8BIT_YES;
    }
    else
    {
        found = TM_8BIT_NO;
    }

    return (int)found;
}

/* Return the first node that has a smaller prefix length */
static int tfAddrPolicyTableFindFirstSmallerEntry(ttNodePtr      nodePtr,
                                                  ttGenericUnion genParam)
{
    tt8Bit  found;

    if ((((ttAddrPolicyEntryPtr)nodePtr)->apePrefixLen) < (genParam.genIntParm))
    {
        found = TM_8BIT_YES;
    }
    else
    {
        found = TM_8BIT_NO;
    }

    return (int)found;
}

/* Find the node in the list */
static int tfAddrPolicyTableFindEntryByNodeCB(ttNodePtr      nodePtr,
                                              ttGenericUnion genParam)
{
    tt8Bit  found;

    if (nodePtr == (ttNodePtr)(genParam.genVoidParmPtr))
    {
        found = TM_8BIT_YES;
    }
    else
    {
        found = TM_8BIT_NO;
    }

    return (int)found;
}

#else /* !TM_USE_RFC3484 */

LINT_UNUSED_HEADER
/* To allow link for builds when TM_USE_RFC3484 is not defined */
int tvAplcyDummy = 0;

#endif /* !TM_USE_RFC3484 */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: traplcy.c     File Revision: 6.0.2.5 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  8903    Output checksum: 19868
 */
