/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (bind)
 *
 * Filename: trbind.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trbind.c $
 *
 * Modification History
 * $Revision: 6.0.2.4 $
 * $Date: 2011/02/14 15:58:48CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

#define TM_CHECK_CONFIG_IP_ADDRESS 0

/*
 * Bind the socket to a port number (without it we are using the random
 * port number), and to an IP address.
 */
int bind(int                            socketDescriptor, 
         const struct sockaddr TM_FAR * localAddressPtr, 
         int                            localAddressLength)
{
/* Common routine with tfBindNoCheck() */
    return tfBindFlag( socketDescriptor,
                       localAddressPtr,
                       localAddressLength,
                       TM_CHECK_CONFIG_IP_ADDRESS );
}

/* Common routine called by both bind(), and tfBindNoCheck() */
int tfBindFlag(int                            socketDescriptor,
               const struct sockaddr TM_FAR * localAddressPtr,
               int                            localAddressLength,
               int                            flag )
{
    ttSocketEntryPtr    socketEntryPtr;
    ttDeviceEntryPtr    anyDevPtr;
    ttSockTuple         sockTuple;
    ttSockAddrPtrUnion  localAddressInPtr;
    int                 errorCode;
    int                 retCode;
    ttIpPort            portNumber;
    tt16Bit             treeFlag;
    tt8Bit              addrFamily;
    tt8Bit              noAddrMatch;
#ifndef TM_SINGLE_INTERFACE_HOME
auto    tt16Bit         anyMhomeIndex;
#endif /* TM_SINGLE_INTERFACE_HOME */
#ifdef TM_4_USE_SCOPE_ID
    unsigned int        srcDevIndex;
#endif /* TM_4_USE_SCOPE_ID */
    
    errorCode = TM_ENOERROR;
#ifdef TM_4_USE_SCOPE_ID
    srcDevIndex = 0;
#endif /* TM_4_USE_SCOPE_ID */

/* to suppress compiler warnings only, doesn't need to be initialized indeed */
    localAddressInPtr.sockPtr = (struct sockaddr TM_FAR *)localAddressPtr;
    
/* PARAMETER CHECK */

/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    socketEntryPtr = tfSocketCheckAddrLock( socketDescriptor,
                                            localAddressPtr,
                                            localAddressLength,
                                            &errorCode
                                            );
    {
        if (socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR)
        {
/* errorCode set by tfSocketCheckAndLock */
            goto bindFinish;
        }
        if ( localAddressPtr == (const struct sockaddr TM_FAR *)0 )
        {
            errorCode = TM_EINVAL;
            goto bindFinish;
        }

        addrFamily = localAddressPtr->sa_family;
        tm_bcopy(&(socketEntryPtr->socTuple), &sockTuple, sizeof(sockTuple));
    
        {
            portNumber = (ttIpPort)(localAddressInPtr.sockInPtr->sin_port);
            tm_ip_copy(
                (tt4IpAddress)(localAddressInPtr.sockInPtr->sin_addr.s_addr),
                tm_4_ip_addr(sockTuple.sotLocalIpAddress));
        }

        if (flag == TM_CHECK_CONFIG_IP_ADDRESS)
        {
            {
                if (    tm_ip_not_zero(
                                  tm_4_ip_addr(sockTuple.sotLocalIpAddress))
                     && !tm_ip_is_multicast(
                                  tm_4_ip_addr(sockTuple.sotLocalIpAddress)) )
                {
/* Check that IP address is a valid interface address */
#ifdef TM_SINGLE_INTERFACE_HOME
                    anyDevPtr = tfIfaceConfig(0);
                    if ( anyDevPtr != (ttDeviceEntryPtr)0 )
                    {
                        if ( !tm_ip_match(
                                    tm_4_ip_addr(sockTuple.sotLocalIpAddress),
                                    tm_ip_dev_addr(anyDevPtr, 0)) )
                        {
                            anyDevPtr = (ttDeviceEntryPtr)0;
                        }
                    }
#else /* TM_SINGLE_INTERFACE_HOME */
#ifdef TM_USE_STRONG_ESL
                    if (tm_8bit_all_bits_set(socketEntryPtr->socFlags2,
                                             TM_SOCF2_BINDTODEVICE))
                    {
/* Bound to device */
                        anyDevPtr = socketEntryPtr->socRteCacheStruct.rtcDevPtr;
                        if (tfMhomeAddrMatch(
                                anyDevPtr,
                                tm_4_ip_addr(sockTuple.sotLocalIpAddress),
                                &anyMhomeIndex) != TM_ENOERROR )
                        {
/* could not find the specified IPv4 address on the interface
   specified by SO_BINDTODEVICE */
                            errorCode = TM_EINVAL;
                            goto bindFinish;
                        }
                    }
                    else
                    {
                        anyDevPtr = (ttDeviceEntryPtr)0;
                    }
                    if (anyDevPtr == (ttDeviceEntryPtr)0)
/* Not bound to device */
#endif /* TM_USE_STRONG_ESL */
                    {
#ifdef TM_4_USE_SCOPE_ID
                        if (srcDevIndex != 0)
                        {
                            anyDevPtr = 
                                tm_context(tvDevIndexToDevPtr)[srcDevIndex];
                            if (tfMhomeAddrMatch(
                                    anyDevPtr,
                                    tm_4_ip_addr(sockTuple.sotLocalIpAddress),
                                    &anyMhomeIndex) != TM_ENOERROR )
                            {
/* could not find the specified link-local scope IPv4 address on the interface
   specified by sin6_scope_id */
                                anyDevPtr = (ttDeviceEntryPtr)0;
                            }
                        }
                        else
#endif /* TM_4_USE_SCOPE_ID */
                        {
                            anyDevPtr = tfIfaceMatch(
                                    tm_4_ip_addr(sockTuple.sotLocalIpAddress),
                                    tfMhomeAddrMatch,
                                    &anyMhomeIndex);
                        }
                    }
#endif /* !TM_SINGLE_INTERFACE_HOME */
                    if ( anyDevPtr == (ttDeviceEntryPtr)0 )
                    {
                        errorCode = TM_EINVAL;
                        goto bindFinish;
                    }
                }
            }
        }

#ifdef TM_4_USE_SCOPE_ID
        if (srcDevIndex != 0)
        {
/* Treat the link-local scope IPv4 address as an IPv4-mapped IPv6 address for
   storing on the socket and reinsertion into the socket tree. We must
   embedded the scope ID first, however. */
            addrFamily = AF_INET6;
            tm_4_addr_add_scope(ipAddrPtr, srcDevIndex);
        }
        else
#endif /* TM_4_USE_SCOPE_ID */

        if ( tm_16bit_one_bit_set(socketEntryPtr->socFlags, TM_SOCF_BIND) )
        {
            {
                noAddrMatch = (tt8Bit)tm_ip_not_match(
                              tm_4_ip_addr(sockTuple.sotLocalIpAddress),
                              tm_4_ip_addr(socketEntryPtr->socOurIfIpAddress));
            }
        
            if (    (!(socketEntryPtr->socOptions & SO_REUSEADDR))
                 || (socketEntryPtr->socOurLayer4Port != portNumber)
                 || noAddrMatch )
            {
                errorCode = TM_EINVAL;
            }
            goto bindFinish;
        }
/* If no address was specified, use the default interface address */    
        {
            if (    (addrFamily == AF_INET)
                 && (tm_ip_zero(tm_4_ip_addr(sockTuple.sotLocalIpAddress))) )
            {
                tm_ip_copy(tm_4_ip_addr(socketEntryPtr->socOurIfIpAddress),
                           tm_4_ip_addr(sockTuple.sotLocalIpAddress));
            }
        }

        {
        }
        if (portNumber != (ttIpPort)0)
        {
            sockTuple.sotLocalIpPort = portNumber;
        }
        if (    (portNumber != 0)
             || (socketEntryPtr->socProtocolNumber != IPPROTO_TCP)
             || (tm_16bit_one_bit_set(socketEntryPtr->socFlags, 
                                      TM_SOCF_IN_TREE))
           )
        {
            treeFlag = (tt16Bit)(   socketEntryPtr->socCoFlags
                                  | (   socketEntryPtr->socOptions
                                      & (SO_REUSEADDR | SO_REUSEPORT) ) );
            if (    (portNumber == 0)
                 && (tm_16bit_bits_not_set(socketEntryPtr->socFlags,
                                           TM_SOCF_IN_TREE)))
            {
                treeFlag = (tt16Bit)(treeFlag | TM_SOC_TREE_RANDOM_PORT);
            }
            retCode = tfSocketTreeReInsert(socketEntryPtr,
                                           &sockTuple,
                                           treeFlag);
            
            if ( retCode != TM_SOC_RB_OKAY )
            {
                errorCode = TM_EADDRINUSE;
                goto bindFinish;
            }
        }
        else
/*
 * Do not insert in the tree, if
 * portNumber is 0, and TCP socket is not in the tree yet.
 */
        {
            tm_bcopy(&sockTuple, &(socketEntryPtr->socTuple),
                     sizeof(sockTuple));
        }

        socketEntryPtr->socFlags |= TM_SOCF_BIND;
        if (portNumber != 0)
        {
            socketEntryPtr->socFlags2 &= ~TM_SOCF2_BIND_PORT_ZERO; 
        }
        else
        {
            socketEntryPtr->socFlags2 |= TM_SOCF2_BIND_PORT_ZERO; 
        }
    
bindFinish:
        retCode = tfSocketReturn(socketEntryPtr, socketDescriptor, errorCode,
                                 TM_ENOERROR);
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trbind.c      File Revision: 6.0.2.4 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 13211    Output checksum: 17389
 */
