/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     Cache invalidation callback registration
 * Filename:        trcshinv.c
 * Author:          Jared
 * Date Created:    06/22/2007
 * $Source: source/sockapi/trcache.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:49CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>

#ifdef TM_USE_DCACHE_MANAGEMENT

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

#define TM_DEV_CACHE_MGMT_FUNC_INVAL 1
#define TM_DEV_CACHE_MGMT_FUNC_FLUSH 2

static int tfInterfaceSaveCacheMgmtFunc (
        ttUserInterface       interfaceHandle,
        ttDevCacheMgmtFuncPtr cacheMgmtFuncPtr,
        ttUserGenericUnion    userParam,
        tt8Bit                funcType );

int tfCacheFlushThenSend(
    ttUserInterface     interfaceHandle,
    char     TM_FAR *   dataPtr,
    int                 dataLength,
    int                 flag );

int tfCacheFlushOneScatThenSend(ttUserInterface interfaceHandle,
                                ttUserPacketPtr packetUPtr);

/*
 * tfUseInterfaceCacheInvalFunc Function Description
 * Specify a function the stack will use to invalidate an area of
 * memory. This allows the stack to handle all necessary cache
 * invalidation before giving the driver a receive buffer.
 * User must call tfUseInterfaceCacheInvalFunc between the 
 * tfAddInterface and tfOpenInterface/tfConfigInterface calls.
 * Calls tfInterfaceSaveCacheMgmtFunc to check parameters and save necessary
 * data to the interface.
 */
int tfUseInterfaceCacheInvalFunc(
        ttUserInterface       interfaceHandle,
        ttDevCacheMgmtFuncPtr cacheInvalFuncPtr,
        ttUserGenericUnion    userParam)
{
      return tfInterfaceSaveCacheMgmtFunc (
                interfaceHandle,
                cacheInvalFuncPtr,
                userParam,
                TM_DEV_CACHE_MGMT_FUNC_INVAL);
}

/*
 * tfUseInterfaceCacheFlushFunc Function Description
 * Specify a function the stack will use to flush an area of
 * memory. This allows the stack to handle all necessary cache
 * flushing before giving the driver a transmit buffer.
 * User must call tfUseInterfaceCacheFlushFunc between the 
 * tfAddInterface and tfOpenInterface/tfConfigInterface calls.
 * Calls tfInterfaceSaveCacheMgmtFunc to check parameters and save necessary
 * data to the interface.
 */
int tfUseInterfaceCacheFlushFunc(
        ttUserInterface       interfaceHandle,
        ttDevCacheMgmtFuncPtr cacheFlushFuncPtr,
        ttUserGenericUnion    userParam)
{
      return tfInterfaceSaveCacheMgmtFunc (
                interfaceHandle,
                cacheFlushFuncPtr,
                userParam,
                TM_DEV_CACHE_MGMT_FUNC_FLUSH);
}

/*
 * tfInterfaceSaveCacheMgmtFunc Function Description
 * 1. Check that the interface is a valid added interface
 * 2. Add the cache invalidation function to the interface
 * 3. Save user parameter to the interface
 * Parameters
 *   interfaceHandle  Interface Handle
 *   cacheMgmtFuncPtr Pointer to callback function
 *   userParam        User parameter to pass back to callback function
 *   funcType         TM_DEV_CACHE_MGMT_FUNC_INVAL or
 *                    TM_DEV_CACHE_MGMT_FUNC_FLUSH
 * Returns
 *   TM_ENOERROR  Success
 *   TM_EINVAL    Error in parameter
 *   TM_EPERM     Device already opened or wrong link layer type
 */
static int tfInterfaceSaveCacheMgmtFunc (
        ttUserInterface       interfaceHandle,
        ttDevCacheMgmtFuncPtr cacheMgmtFuncPtr,
        ttUserGenericUnion    userParam,
        tt8Bit                funcType )
{
    ttDeviceEntryPtr      devEntryPtr;
    int                   errorCode;
    ttUserGenericUnionPtr destUnionPtr;

    errorCode = TM_ENOERROR;
    destUnionPtr = (ttUserGenericUnionPtr)0;
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check parameter */
    errorCode = tfValidInterface(devEntryPtr);
    if ( errorCode == TM_ENOERROR )
    {
        if (cacheMgmtFuncPtr == (ttDevCacheMgmtFuncPtr)0)
        {
            errorCode = TM_EINVAL;
        }
/* 
 * Cannot set after device has been opened. Also cannot use on byte-stuffing
 * interfaces (e.g., PPP and SLIP)
 */
        if (   tm_16bit_one_bit_set(devEntryPtr->devFlag, TM_DEV_OPENED)
            || tm_ll_is_byte_stuffing(devEntryPtr))
        {
            errorCode = TM_EPERM;
        }

        if (errorCode == TM_ENOERROR)
        {
/* Check which type of function needs to be stored */
            switch (funcType)
            {
            case TM_DEV_CACHE_MGMT_FUNC_INVAL:
/* Save the cache invalidation function and associated user data */
                devEntryPtr->devCacheInvalFuncPtr = cacheMgmtFuncPtr;
                destUnionPtr = &(devEntryPtr->devCacheInvalUserData);
                break;
            case TM_DEV_CACHE_MGMT_FUNC_FLUSH:
/* Save the cache flush function and associated user data */
                devEntryPtr->devCacheFlushFuncPtr = cacheMgmtFuncPtr;
                destUnionPtr = &(devEntryPtr->devCacheFlushUserData);
/* Save the original driver send routine */
                devEntryPtr->devCacheSavedSendFuncPtr =
                        devEntryPtr->devSendFuncPtr;
/* Replace driver send routine with our flush-then-send routine */
                devEntryPtr->devSendFuncPtr = tfCacheFlushThenSend;
#ifdef TM_USE_DRV_ONE_SCAT_SEND
/* 
 * If the user has set up a one-scattered-send driver function, replace it
 * with a flush-then-send function.
 * Note: tfUseInterfaceOneScatSend checks to make sure it's not called after
 *       this function has been called
 */
                if (   devEntryPtr->devOneScatSendFuncPtr
                    != (ttDevOneScatSendFuncPtr)0)
                {
                    devEntryPtr->devCacheSavedOneScatSendFuncPtr =
                            devEntryPtr->devOneScatSendFuncPtr;
                    devEntryPtr->devOneScatSendFuncPtr =
                            tfCacheFlushOneScatThenSend;
                }
#endif /* TM_USE_DRV_ONE_SCAT_SEND */
                break;
            }
/* Copy the user parameter to the appropriate variable */
            tm_bcopy(&userParam, destUnionPtr, sizeof(ttUserGenericUnion));
        }
    }
    return (errorCode);
}

/*
 * tfCacheFlushThenSend Function Description
 * This function is used to overload the driver send function if the user
 * wants all data packets flushed before being given to the driver. This
 * function replaces the driver send function in the device, flushes the
 * data, and then calls the actual driver send function.
 * 1. Flush the packet buffer from the cache
 * 2. Send the packet
 * Parameters
 *   interfaceHandle  Interface Handle
 *   dataPtr          Pointer to the data buffer
 *   dataLength       Length of the data buffer
 *   flag             Passed through to driver
 * Returns
 *   Result of driver send call
 */
int tfCacheFlushThenSend(
    ttUserInterface     interfaceHandle,
    char     TM_FAR *   dataPtr,
    int                 dataLength,
    int                 flag )
{
    ttDeviceEntryPtr      devEntryPtr;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Flush the packet out of the dcache */ 
    devEntryPtr->devCacheFlushFuncPtr(interfaceHandle,
                                      dataPtr,
                                      dataLength,
                                      devEntryPtr->devCacheFlushUserData);
/* Call the driver send routine */
    return devEntryPtr->devCacheSavedSendFuncPtr(interfaceHandle,
                                                 dataPtr,
                                                 dataLength,
                                                 flag);
}

#ifdef TM_USE_DRV_ONE_SCAT_SEND
/*
 * tfCacheFlushOneScatThenSend Function Description
 * This function is used to overload the driver one-scattered-send function if
 * the user wants all data packets flushed before being given to the driver.
 * This function replaces the driver send function in the device, flushes the
 * data, and then calls the actual driver send function.
 * 1. Flush the packet chain from the cache
 * 2. Send the packet
 * Parameters
 *   interfaceHandle  Interface Handle
 *   packetUPtr       Pointer to chain of data buffers
 * Returns
 *   Result of driver one-scattered-send call
 */
int tfCacheFlushOneScatThenSend(ttUserInterface interfaceHandle,
                                ttUserPacketPtr packetUPtr)
{
    ttDeviceEntryPtr      devEntryPtr;
    ttUserPacketPtr       tempPkt;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    tempPkt = packetUPtr;
/* Loop through each buffer in the chain */
    while (tempPkt)
    {
/* Flush the buffer out of the cache */
        devEntryPtr->devCacheFlushFuncPtr(interfaceHandle,
                                          (char *)tempPkt->pktuLinkDataPtr,
                                          tempPkt->pktuLinkDataLength,
                                          devEntryPtr->devCacheFlushUserData);
/* Get the next buffer in the chain */
        tempPkt = tempPkt->pktuLinkNextPtr;
    }
/* Call the driver send routine */
    return devEntryPtr->devCacheSavedOneScatSendFuncPtr(interfaceHandle,
                                                        packetUPtr);
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */


/*
 * tfGetCacheSafeDrvBuf Function Description
 * This function can be called by the user to allocate a packet buffer to be
 * used by a driver. The allocated buffer is invalidated in the cache before
 * being returned to the user. This function will attempt to automatically
 * align the buffer so that the IP header will be 4-byte aligned. This depends
 * on the alignment requested by the user and the length of the link layer
 * header:
 * If the alignment requested is 1 or 0 bytes, the IP header will always be
 * correctly aligned. If the alignment requested by the user is 2 bytes, the
 * IP header will be correctly aligned as long as the link layer header length
 * is a multiple of 2 bytes. In all other cases, the alignment requested by the
 * user is respected, and the IP header may not be correctly aligned.
 * This function will allocate extra space in front of the data pointer to
 * make room for the largest link-layer header in use in the system, in the
 * event that the packet is forwarded through an interface using that
 * link-layer.
 * 1. Verify the interface handle
 * 2. Calculate the alignment needs on DSP's
 * 3. Decide if the IP header can be aligned to a 4-byte boundary
 * 4. Calculate the maximum space needed before the start of data
 * 5. Allocate the buffer
 * 6. Flush the buffer from the dcache
 * 7. Align the buffer according to the user's requreiemts
 * 8. If possible, adjust the data pointer to start the IP header on a 4-byte
 *    boundary
 * Parameters
 *   interfaceHandle  Interface Handle
 *   userBufferPtr    Pointer to be filled out with the new ttPacketPtr
 *   length           Requested buffer length
 *   alignment        Requested buffer alignment
 *   flag             If set to TM_DEV_CALLED_FROM_APP, we need to lock around
 *                    the call to the user's flush function.
 * Returns
 *   Valid pointer    Pointer to the new data buffer
 *   0                Unable to allocate a new buffer
 */
char TM_FAR * tfGetCacheSafeDrvBuf (
        ttUserInterface interfaceHandle,
        ttUserBufferPtr userBufferPtr,
        int             length,
        int             alignment,
        int             flag )
{
    ttDeviceEntryPtr devEntryPtr;
    ttPacketPtr      packetPtr;
    char TM_FAR    * dataPtr;
    int              alignmentByteCount;
    int              errorCode;
    int              headerLen;
    tt8Bit           offsetExtra;

    dataPtr = (char TM_FAR *)0;
/* 1. Verify the interface handle */
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    errorCode = tfValidInterface(devEntryPtr);
    if (errorCode == TM_ENOERROR)
    {
#ifdef TM_ERROR_CHECKING
/* Parameter check */
        if (alignment < 0)
        {
            alignment = 0;
        }
#endif /* TM_ERROR_CHECKING */

        alignmentByteCount = alignment;
/* 3. Decide if the IP header can be aligned to a 4-byte boundary */
        if (   alignment < TM_ETHER_IP_ALIGN
            || (   alignment == TM_ETHER_IP_ALIGN
                && !((int)devEntryPtr->devLinkHeaderLength & 0x01)))
        {
            offsetExtra = 3;
        }
        else
        {
            offsetExtra = 0;
        }

/* 4. Calculate the maximum space needed before the start of data */
/*
 * Allocate enough room ahead of the packet to expand backwards, in case
 * we need to forward the packet through another interface.
 * Allocate datalength plus enough room to adjust for alignment ahead
 * of the buffer, plus (if possible) enough room to align the IP header to a
 * 4-byte boundary
 */
        headerLen =   TM_LL_MAX_HDR_LEN
                    - devEntryPtr->devLinkHeaderLength
                    + alignmentByteCount
                    + offsetExtra
                    ;
/* 5. Allocate the buffer */
        packetPtr = tfGetSharedBuffer(headerLen,
                                      (ttPktLen)length,
                                      TM_16BIT_ZERO);
        if (packetPtr != TM_PACKET_NULL_PTR)
        {
/* 6. Flush the buffer from the dcache */
            if (flag & TM_DEV_CALLED_FROM_APP)
            {
                tm_lock_wait(&(devEntryPtr->devDriverLockEntry));
            }
            if (devEntryPtr->devCacheInvalFuncPtr != (ttDevCacheMgmtFuncPtr)0)
            {
                devEntryPtr->devCacheInvalFuncPtr(
                        interfaceHandle,
                        (char *)packetPtr->pktLinkDataPtr -
                                tm_packed_byte_count(headerLen),
                        (ttPktLen)length + headerLen,
                        devEntryPtr->devCacheInvalUserData);
            }
            if (flag & TM_DEV_CALLED_FROM_APP)
            {
                tm_unlock(&(devEntryPtr->devDriverLockEntry));
            }

            packetPtr->pktUserFlags = TM_PKTF_USER_DEV_OWNS;
/* 7. Align the buffer according to the user's requreiemts */
            if (alignment > 0)
            {
                if (((ttUserPtrCastToInt)packetPtr->pktLinkDataPtr)
                    & (ttUserPtrCastToInt)(alignment - 1) )
                {
                    packetPtr->pktLinkDataPtr = (tt8BitPtr)
                        (    ((ttUserPtrCastToInt)packetPtr->pktLinkDataPtr)
                            & ~((ttUserPtrCastToInt)(alignment - 1)) );
                }
            }

/*
 * 8. If possible, adjust the data pointer to start the IP header on a 4-byte
 *    boundary
 */
            if (offsetExtra)
            {
                packetPtr->pktLinkDataPtr = (tt8BitPtr)
                        (  (  (  
                                 (ttUserPtrCastToInt)packetPtr->pktLinkDataPtr
                               + devEntryPtr->devLinkHeaderLength)
                            & ~((ttUserPtrCastToInt)0x03))
                         - devEntryPtr->devLinkHeaderLength);
            }

            dataPtr = (char TM_FAR *)packetPtr->pktLinkDataPtr;
            *userBufferPtr = (ttUserBuffer)packetPtr;
        }
        else
        {
            *userBufferPtr = (ttUserBuffer)0;
            dataPtr = (char TM_FAR *)0;
        }
    }
    return dataPtr;
}

#else /* TM_USE_DCACHE_MANAGEMENT */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_USE_DRV_ONE_SCAT_SEND is not defined */
int tvCacheDummy = 0;

#endif /* TM_USE_DCACHE_MANAGEMENT */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:46      For: Xilinx-Samsung
 * File Name: trcache.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  9089    Output checksum: 49908
 */
