/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (connect)
 *
 * Filename: trconn.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trconn.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:52CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * Connect for TCP and set up a machine to send to and receive from on UDP
 */
int connect(int                           socketDescriptor, 
            const struct sockaddr TM_FAR *addressPtr,
            int                           addressLength)
{
    ttSocketEntryPtr           socketPtr;
#ifdef TM_USE_TCP
    ttTcpVectPtr               tcpVectPtr;
#endif /* TM_USE_TCP */
    ttSockAddrPtrUnion         internetAddressPtrU;
    ttPacketPtr                packetPtr;
    ttIpHeader                 iphStruct; /* 20 bytes */
    ttSockTuple                sockTuple;
    tt16Bit                    treeFlag;
    int                        retCode;
    int                        errorCode;
    tt8Bit                     reinsertInTree;

    errorCode = TM_ENOERROR;
    reinsertInTree = TM_8BIT_NO;

/* PARAMETER CHECK */

/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    socketPtr = tfSocketCheckAddrLock(socketDescriptor,
                                      addressPtr,
                                      addressLength,
                                      &errorCode
                                     );
    {
        if (socketPtr == TM_SOCKET_ENTRY_NULL_PTR)
        {
/* errorCode set by tfSocketCheckAndLock */
            goto connectFinish;
        }
        if ( addressPtr == (const struct sockaddr TM_FAR *)0 )
        {
            errorCode = TM_EINVAL;
            goto connectFinish;
        }
    
/* Check the address to make sure it is valid */
        internetAddressPtrU.sockPtr = (struct sockaddr TM_FAR *)addressPtr; 
    
#ifdef TM_USE_TCP
        if ( socketPtr->socProtocolNumber == IP_PROTOTCP )
        {
/* TCP protocols */
            tcpVectPtr = (ttTcpVectPtr)((ttSocketEntryPtr)socketPtr);
/* Point to the user structure */
            tcpVectPtr->tcpsSockAddrUPtr = &internetAddressPtrU;
            errorCode = tfTcpOpen( tcpVectPtr,
                                   TM_TCPE_ACTIVE_OPEN );
/*
 * Make sure we reset the connect parameter pointer, in case tfTcpStateMachine
 * does not call tfTcpSendSyn().
 */
            tcpVectPtr->tcpsSockAddrUPtr = (ttSockAddrPtrUnionPtr)0;
        }
        else
#endif /* TM_USE_TCP */
        {
/* Non TCP protocols */
            tm_bcopy(&(socketPtr->socTuple), &sockTuple,
                     sizeof(sockTuple));
            treeFlag = (tt16Bit)(   socketPtr->socOptions
                                  & (SO_REUSEADDR | SO_REUSEPORT) );
            if (tm_16bit_bits_not_set(socketPtr->socFlags, 
                                      TM_SOCF_IN_TREE))
            {
/* Pick random port */
                treeFlag = (tt16Bit)(   treeFlag
                                      | TM_SOC_TREE_RANDOM_PORT);
            }
/*
 * Load up the socket entry unconditionally for non TCP protocols 
 */
            tm_ip_copy(
                (tt4IpAddress)(internetAddressPtrU.sockInPtr->sin_addr.s_addr),
                sockTuple.sotRemoteIpAddress);
            sockTuple.sotRemoteIpPort =
                           (ttIpPort)(internetAddressPtrU.sockInPtr->sin_port);
/* Try and set the local IP address */
            if (
/* IPv4 only */
                tm_ip_zero(socketPtr->socOurIfIpAddress)
                )
            {
                {
/* Allocate an empty ttPacket/ttSharedData */
                    packetPtr = tfGetSharedBuffer(0,
                                                  0,
                                                  TM_16BIT_ZERO);
                    if (packetPtr != TM_PACKET_NULL_PTR)
                    {

                        errorCode = tfIpDestToPacket(
                            packetPtr,
#ifdef TM_4_USE_SCOPE_ID
                            &in6addr_any,
                            &(sockTuple.sotRemoteIpAddress),
#else /* ! TM_4_USE_SCOPE_ID */
                            TM_IP_ZERO,
                            tm_4_ip_addr(sockTuple.sotRemoteIpAddress),
#endif /* ! TM_4_USE_SCOPE_ID */
                            1,
                            &socketPtr->socRteCacheStruct,
                            &iphStruct );
                        if (errorCode == TM_ENOERROR)
                        {
/* Save source IP address for getsockname() benefit */
                            tm_ip_copy(
                                   iphStruct.iphSrcAddr,
                                   tm_4_ip_addr(sockTuple.sotLocalIpAddress));
#ifdef TM_4_USE_SCOPE_ID
                            if (tm_16bit_one_bit_set(
                                    packetPtr->pktFlags,
                                    TM_OUTPUT_DEVICE_FLAG))
                            {
/* we know the output device, so use it to embed the scope ID if the IPv4
   address is link-local scope */
                                tm_4_addr_add_scope(
                                    &(sockTuple.sotLocalIpAddress),
                                    packetPtr->pktDeviceEntryPtr->devIndex);
                            }
#endif /* !TM_4_USE_SCOPE_ID */
                            reinsertInTree = TM_8BIT_YES;
                        }

                        tfFreeSharedBuffer(packetPtr, TM_SOCKET_LOCKED);
                    }
                    else
                    {
                        errorCode = TM_ENOBUFS;
                    }
                }
            }
            else
            {
                if (tm_16bit_bits_not_set(socketPtr->socFlags,
                                          TM_SOCF_IN_TREE))
                {
                    reinsertInTree = TM_8BIT_YES;
                }
            }
        
            if (errorCode == TM_ENOERROR)
            {
                if (reinsertInTree)
                {
                    retCode = tfSocketTreeReInsert(
                                        socketPtr,
                                        &sockTuple,
                                        treeFlag);

                    if (retCode != TM_SOC_RB_OKAY)
                    {
                        errorCode = TM_EADDRINUSE;
                    }
                }
                else
                {
                    tm_bcopy(&sockTuple,
                             &(socketPtr->socTuple),
                             sizeof(sockTuple));
                }
                if (errorCode == TM_ENOERROR)
                {
                    socketPtr->socFlags |= TM_SOCF_CONNECTED;
                }
            }
        }
connectFinish:
        retCode = tfSocketReturn(socketPtr, socketDescriptor,
                                 errorCode, TM_ENOERROR);
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trconn.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 34928    Output checksum: 55874
 */
