/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (tfDelArpByIpAddr, tfDelArpByPhysAddr)
 *
 * Filename: trdelarp.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trdelarp.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:54CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* Delete an ARP Entry from ARP table using IP address to locate entry*/
int tfDelArpEntryByIpAddr (
#ifdef TM_USE_STRONG_ESL
                            ttUserInterface interfaceHandle,
#endif /* TM_USE_STRONG_ESL */

                            ttUserIpAddress arpIpAddress)
{
    ttArpMapping arpMapping;
    int          errorCode;

    if (    (tm_ip_zero(arpIpAddress))
#ifdef TM_USE_STRONG_ESL
         || (tfValidInterface(interfaceHandle) != TM_ENOERROR)
#endif /* TM_USE_STRONG_ESL */
       )
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        tm_ip_copy(arpIpAddress, arpMapping.arpmIpAddress);
        arpMapping.arpmCommand = TM_4_ARP_DEL_BY_IP;
#ifdef TM_USE_STRONG_ESL
        arpMapping.arpmDevPtr = (ttDeviceEntryPtr)interfaceHandle;
#endif /* TM_USE_STRONG_ESL */
        errorCode = tfRtArpCommand(&arpMapping);
    }
    return errorCode;
}

/****************************************************************************
 * FUNCTION: tfNgDelArpEntryByIpAddr
 *
 * PURPOSE: Next Generation API replaces tfDelArpEntryByIpAddr
 *
 * PARAMETERS:
 *   arpIpAddrPtr
 *       ip address information pointer
 *
 * RETURNS:
 *   TM_ENOERROR
 *       success    
 *   TM_EINVAL
 *       arpIpAddrPtr is null pointer
 *   TM_ENOENT
 *       no ARP entry for the IP address found
 *
 ****************************************************************************/
int tfNgDelArpEntryByIpAddr(
#ifdef TM_USE_STRONG_ESL
    ttUserInterface                        interfaceHandle,
#endif /* TM_USE_STRONG_ESL */
    const struct sockaddr_storage TM_FAR * arpIpAddrPtr )
{
    int          errorCode;

    if (arpIpAddrPtr == (ttConstSockAddrStoragePtr)0)
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        switch (arpIpAddrPtr->ss_family)
        {
        case AF_INET:
            errorCode = tfDelArpEntryByIpAddr(
#ifdef TM_USE_STRONG_ESL
                                interfaceHandle,
#endif /* TM_USE_STRONG_ESL */
                                arpIpAddrPtr->addr.ipv4.sin_addr.s_addr);
            break;

        
        default:
            errorCode = TM_EAFNOSUPPORT;
            break;
        }
    }        

    return errorCode;
}

/* Delete an ARP Entry from ARP table using physical address to locate entry*/
int tfDelArpEntryByPhysAddr (
#ifdef TM_USE_STRONG_ESL
                             ttUserInterface     interfaceHandle,
#endif /* TM_USE_STRONG_ESL */
                             const char TM_FAR * physAddrPtr, int physAddrLen)
{
    ttArpMapping arpMapping;
    int          errorCode;

    if (    (physAddrPtr == (const char TM_FAR *)0)
         || (physAddrLen > TM_MAX_PHYS_ADDR)
#ifdef TM_USE_STRONG_ESL
         ||  (tfValidInterface(interfaceHandle) != TM_ENOERROR)
#endif /* TM_USE_STRONG_ESL */
       )
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        tm_bcopy( physAddrPtr, arpMapping.arpmEnetAddress, physAddrLen);
        arpMapping.arpmPhysAddrLen = (tt8Bit) physAddrLen;
        arpMapping.arpmCommand = TM_4_ARP_DEL_BY_PHYS;
#ifdef TM_USE_STRONG_ESL
        arpMapping.arpmDevPtr = (ttDeviceEntryPtr)interfaceHandle;
#endif /* TM_USE_STRONG_ESL */
        errorCode = tfRtArpCommand(&arpMapping);
    }
    return errorCode;
}

/****************************************************************************
 * FUNCTION: tfNgDelArpEntryByPhysAddr
 *
 * PURPOSE: Next Generation API replaces tfDelArpEntryByPhysAddr
 *
 *   addrFamily
 *       can be either AF_INET or AF_INET6
 *   physAddrPtr
 *       pointer to a char array that contains the physical address
 *   physAddrLen
 *       length of the physical address
 * RETURNS:
 *   TM_ENOERROR
 *      sucess
 *   TM_EAFNOSUPPORT
 *      address family not supported 
 *   TM_EINVAL
 *       physAddrPtr is null pointer or physAddrLen is too short
 *   TM_ENOENT 
 *      no such Arp entry
 *
 ****************************************************************************/
int tfNgDelArpEntryByPhysAddr (
#ifdef TM_USE_STRONG_ESL
    ttUserInterface             interfaceHandle,
#endif /* TM_USE_STRONG_ESL */
    int                         addrFamily, 
    const ttUser8Bit TM_FAR *   physAddrPtr,
    int                         physAddrLen )
{
    int          errorCode;

    if (    (physAddrPtr == (const ttUser8Bit TM_FAR *)0 )
         || (physAddrLen <= 0) )
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        switch (addrFamily)
        {
        case AF_INET:
            errorCode = tfDelArpEntryByPhysAddr(
#ifdef TM_USE_STRONG_ESL
                                interfaceHandle,
#endif /* TM_USE_STRONG_ESL */
                                (const char TM_FAR *)physAddrPtr,
                                physAddrLen);
            break;

        
        default:
            errorCode = TM_EAFNOSUPPORT;
            break;
        }
    }        

    return errorCode;
}

#ifdef TM_USE_ARP_FLUSH
/* Flush the ARP table for the given address family/families */
int tfArpFlush (int af)
{
    int errorCode;

/* We only support AF_INET, and AF_INET6 */
    af = af & (AF_INET | AF_INET6);
    af &= ~AF_INET6;
    if ((af & (AF_INET|AF_INET6)) != 0)
    {
        errorCode = TM_ENOERROR;
        tfRtArpFlush(af
#ifdef TM_LOCK_NEEDED
                   , TM_LOC_WAIT
#endif /* TM_LOCK_NEEDED */
                    );
    }
    else
    {
        errorCode = TM_EINVAL;
    }
    return errorCode;
}
#endif /* TM_USE_ARP_FLUSH */


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trdelarp.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 41642    Output checksum: 25222
 */
