/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2011
 * Copyright Elmic Systems Japan 1997 - 2011
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Device Driver options: tfInterfaceSetOptions()
 *
 * Filename: trdevopt.c
 * Author: Odile
 * Date Created: 01/18/00
 * $Source: source/sockapi/trdevopt.c $
 *
 * Modification History
 * $Revision: 6.0.2.16 $
 * $Date: 2011/12/13 14:07:30CST $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#ifdef TM_LOCK_NEEDED
#include <trglobal.h>
#else /* !TM_LOCK_NEEDED */
#endif /* !TM_LOCK_NEEDED */

/*
 * tfInterfaceSetOptions function description:
 * Called by the user to set some interface options:
 * TM_DEV_OPTIONS_RECV_COPY
 *    Make tfRecvInterface() copy received driver buffers whose length
 *    is smaller than the option threshold to a newly allocated smaller
 *    buffer. This option cannot be used
 *    on a point to point device, as a point to point link layer receive
 *    function copies the receive buffer into a new buffer, to unstuf
 *    the bytes.
 * or
 * TM_DEV_OPTIONS_XMIT_TASK
 *    Turn on/off a transmit task for flow control between the TCP/IP stack
 *    and the device driver send function. Can only be changed when device
 *    is not opened (i.e not configured).
 */
int tfInterfaceSetOptions ( ttUserInterface interfaceHandle,
                            int             optionName,
                            void   TM_FAR * optionValuePtr,
                            int             optionLength)
{
    ttDeviceEntryPtr   devEntryPtr;
    int                errorCode;
    int                optionLengthFlag;
    int                optionValue;
    ttUser32Bit        optionValue32;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    optionValue = 0; /* Compiler warning */
    optionValue32 = TM_UL(0); /* Compiler warning */
/* Check on parameters */
    errorCode = tfValidInterface(devEntryPtr);
    if (errorCode == TM_ENOERROR)
    {
        errorCode = TM_EINVAL; /* assume bad parameter */

/* Lock device */
        tm_call_lock_wait(&(devEntryPtr->devLockEntry));
/* Check on length of option */
        if (optionValuePtr != (void TM_FAR *)0)
        {
/* Extract option length flag part of the optionName */
            optionLengthFlag = optionName & (   TM_DEV_OFLAG_SHORT
                                              | TM_DEV_OFLAG_UCHAR
                                              | TM_DEV_OFLAG_INT
                                              | TM_DEV_OFLAG_INT32
                                              | TM_DEV_OFLAG_STRING);
/* Get the optionValue based on the option length flag part of the optionName */
            switch (optionLengthFlag)
            {
                case TM_DEV_OFLAG_SHORT:
                    if (optionLength == (int)sizeof(short))
                    {
                        errorCode = TM_ENOERROR;
                        optionValue = *((short TM_FAR *)optionValuePtr);
                    }
                    break;
                case TM_DEV_OFLAG_UCHAR:
                    if  (optionLength == (int)sizeof(unsigned char))
                    {
                        optionValue = *((unsigned char TM_FAR *)optionValuePtr);
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_INT:
                    if  (optionLength == (int)sizeof(int))
                    {
                        optionValue = *((int TM_FAR *)optionValuePtr);
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_INT32:
                    if  (optionLength == (int)sizeof(ttUser32Bit))
                    {
                        optionValue32 = *((ttUser32BitPtr)optionValuePtr);
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_STRING:
                    errorCode = TM_ENOERROR;
                default:
                    break;
            }
        }

/* End of check on length of option */
/* Check on value of option */
        if (    (errorCode == TM_ENOERROR)
             && (optionName & TM_DEV_OFLAG_ABSOLUTE)
             && (optionValue < 0) )
        {
            errorCode = TM_EINVAL;
        }
/* End of check on value of option. */
/* Check if option is allowed on the device */
        if (    ( errorCode == TM_ENOERROR )
             && ( optionName & TM_DEV_OFLAG_NOT_BYTE_STUFF )
             && ( tm_ll_is_byte_stuffing(devEntryPtr) ) )
/* Disallow changing some options if device requires byte stuffing */
        {
            errorCode = TM_EPERM;
        }
        if (    ( errorCode == TM_ENOERROR )
             && ( optionName & TM_DEV_OFLAG_NOT_OPEN )
             && ( tm_16bit_one_bit_set( devEntryPtr->devFlag,
                                        TM_DEV_OPENED) ) )
        {
/* Disallow changing some options after device has been configured */
            errorCode = TM_EPERM;
        }
        if (errorCode == TM_ENOERROR)
        {
/*
 * Lock device driver lock. OK to keep the device lock, since we never lock
 * the device lock after locking the device driver lock
 */
            tm_call_lock_wait(&(devEntryPtr->devDriverLockEntry));
            switch (optionName)
            {

/*
 * User gives a threshold value, so that we try and copy receive driver
 * buffers whose size are smaller than the given threshold option value
 * to a new buffer.
 */
                case TM_DEV_OPTIONS_RECV_COPY:
                    devEntryPtr->devRecvCopyThreshHold = (tt16Bit)optionValue;
                    break;

                case TM_DEV_OPTIONS_TX_ALIGN:
                    devEntryPtr->devTxAlign = (tt16Bit)optionValue;
                    break;

                case TM_DEV_OPTIONS_XMIT_TASK:
/*
 * Use a transmit task for flow control between the TCP/IP stack and
 * a device driver
 */
                    if (optionValue > 0)
                    {
                        if (tm_iface_xmit_queue_option(interfaceHandle))
/* Disallow transmit task, if a transmit buffer queue is used */
                        {
                            errorCode = TM_EPERM;
                        }
                        else
                        {
/* Allow transmit task */
                            devEntryPtr->devFlag |= TM_DEV_XMIT_TASK;
                        }
                    }
                    else
                    {
/* Reset transmit task usage */
                        devEntryPtr->devFlag &= ~TM_DEV_XMIT_TASK;
                        devEntryPtr->devXmitTaskPacketPtr = (ttPacketPtr)0;
                    }
                    break;

/* Disable gratuitous ARP's */
                case TM_DEV_OPTIONS_NO_GRAT_ARP:
                    if (optionValue > 0)
                    {
                        devEntryPtr->devFlag2 |= TM_DEVF2_NO_GRAT_ARP;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &= ~TM_DEVF2_NO_GRAT_ARP;
                    }
                    break;
/*
 * Maximum number of BOOTP/DHCP request retries.
 */
                case TM_DEV_OPTIONS_BOOT_RETRIES:
                    devEntryPtr->devBtMaxRetries = (unsigned char)optionValue;
                    break;

/*
 * Initial timeout value for BOOTP/DHCP requests.
 */
                case TM_DEV_OPTIONS_BOOT_TIMEOUT:
                    devEntryPtr->devBtInitTimeoutIntv
                        = (unsigned char)optionValue;
                    break;

#ifdef TM_USE_DRV_SCAT_RECV
/*
 * For driver scattered recv only. Set the minimum number of
 * bytes at the head of a packet that have to be contiguous.
 * If a scattered recv buffer is received with a first link length below that
 * minimum, that minimum number of bytes is copied into a new buffer.
 * (Default value is TM_DEV_DEF_RECV_CONT_HDR_LENGTH) .
 */
                case TM_DEV_OPTIONS_SCAT_RECV_LENGTH:
                    devEntryPtr->devRecvContiguousHdrLength =
                                                        (tt16Bit)optionValue;
                    break;
#endif /* TM_USE_DRV_SCAT_RECV */
/*
 * DHCP/BOOTP option: Number of ARP probes to send prior to configuring
 * the address (0 means default, -1 means no ARP probes should be sent)
 */
                case TM_DEV_OPTIONS_BOOT_ARP_RETRIES:
                    devEntryPtr->devBtArpProbes = (int)optionValue;
                    break;
/*
 * DHCP/BOOTP option: Interval in seconds between ARP probes to be sent priot
 * to configuring the address (0 means default)
 */
                case TM_DEV_OPTIONS_BOOT_ARP_INTVL:
                    devEntryPtr->devBtArpIntv = optionValue * TM_UL(1000);
                    break;
/*
 * DHCP/BOOTP option: Interval in milliseconds between ARP probes to be sent priot
 * to configuring the address (0 means default)
 */
                case TM_DEV_OPTIONS_BOOT_ARP_INTVL_MSECS:
                    if (optionValue32 > TM_8BIT_ONES * TM_UL(1000))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->devBtArpIntv = optionValue32;
                    }
                    break;
/*
 * DHCP/BOOTP option: number of seconds to wait after sending the
 * first ARP probe/Arp Request before finishing configuring
 * a DHCP/BOOTP address.
 * 0 means use default value.
 */
                case TM_DEV_OPTIONS_BOOT_ARP_TIMEOUT:
                    devEntryPtr->devBtArpTimeout = (tt8Bit)optionValue;
                    break;
#ifdef TM_USE_DHCP
/* DHCP option: Do not send a RELEASE message when closing DHCP */
                case TM_DEV_OPTIONS_NO_DHCP_RELEASE:
                    if (optionValue > 0)
                    {
                        devEntryPtr->devFlag2 |= TM_DEVF2_NO_DHCP_RELEASE;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &= ~TM_DEVF2_NO_DHCP_RELEASE;
                    }
                    break;
/*
 * DHCP option:
 * Pick which host name to send to the server in the
 * host name option and/or FQDN option.
 * 0 means pick user set values (default)
 * 1 means pick server set values
 */
                case TM_DEV_OPTIONS_BOOT_PK_HOST_NM:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |= TM_DEVF2_SRVR_HOST_NAME;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &= ~TM_DEVF2_SRVR_HOST_NAME;
                    }
                    break;
/*
 * DHCP option:
 * Pick which domain name to send to the server in the FQDN option
 * 0 means pick user set values (default)
 * 1 means pick server set values
 */
                case TM_DEV_OPTIONS_BOOT_PK_DOMAIN_NM:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |= TM_DEVF2_SRVR_DOMAIN_NAME;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &= ~TM_DEVF2_SRVR_DOMAIN_NAME;
                    }
                    break;
#ifdef TM_USE_DHCP_FQDN
                case TM_DEV_OPTIONS_DHCP_SND_BIN_FQDN:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |= TM_DEVF2_DHCP_SND_BIN_FQDN;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &= ~TM_DEVF2_DHCP_SND_BIN_FQDN;
                    }
                    break;
#endif /* TM_USE_DHCP_FQDN */
#endif /* TM_USE_DHCP */
#ifdef TM_USE_FILTERING
                case TM_DEV_OPTIONS_FILTER:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |=
                            (tt16Bit)(  TM_DEVF2_FILTER_INCOMING
                                      | TM_DEVF2_FILTER_OUTGOING);
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &=
                            ~(tt16Bit)(  TM_DEVF2_FILTER_INCOMING
                                       | TM_DEVF2_FILTER_OUTGOING);
                    }
                    break;
                case TM_DEV_OPTIONS_FILTER_INCOMING:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |=
                            (tt16Bit)TM_DEVF2_FILTER_INCOMING;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &=
                            ~(tt16Bit)TM_DEVF2_FILTER_INCOMING;
                    }
                    break;
                case TM_DEV_OPTIONS_FILTER_OUTGOING:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |=
                            (tt16Bit)TM_DEVF2_FILTER_OUTGOING;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &=
                            ~(tt16Bit)TM_DEVF2_FILTER_OUTGOING;
                    }
                    break;
#endif /* TM_USE_FILTERING */
#ifdef TM_USE_IGMP_FILTERING
                case TM_DEV_OPTIONS_FILTER_IGMP:
                    if (optionValue == 1)
                    {
                        devEntryPtr->devFlag2 |=
                                        (tt16Bit)TM_DEVF2_IGMP_FILTERING;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 &=
                                        ~(tt16Bit)TM_DEVF2_IGMP_FILTERING;
                    }
                    break;
#endif /* TM_USE_IGMP_FILTERING */
                case TM_DEV_OPTIONS_DISABLE_ICMP_RTR_ADVERT:
                    if (optionValue == 0)
                    {
                        devEntryPtr->devFlag2 &=
                            ~(tt16Bit)TM_DEVF2_DISABLE_ICMP_RTR_ADVERT;
                    }
                    else
                    {
                        devEntryPtr->devFlag2 |=
                             (tt16Bit)TM_DEVF2_DISABLE_ICMP_RTR_ADVERT;
                    }
                    break;
                case TM_DEV_OPTIONS_FORWARDING:
                    if (optionValue == 0)
                    {
                        devEntryPtr->devFlag &= ~(tt16Bit)TM_DEV_IP_FORW_ENB;
                    }
                    else
                    {
                        devEntryPtr->devFlag |= (tt16Bit)TM_DEV_IP_FORW_ENB;
                    }
                    break;
                case TM_DEV_OPTIONS_FORWARD_REFLECT:
                    if (optionValue == 0)
                    {
                        devEntryPtr->devFlag2
                                    &= ~(tt16Bit)TM_DEVF2_FORWARD_REFLECT;
                    }
                    else
                    {
                        devEntryPtr->devFlag2
                                    |= (tt16Bit)TM_DEVF2_FORWARD_REFLECT;
                    }
                    break;
                case TM_DEV_OPTIONS_IP_PROMISCUOUS:
                    if (optionValue == 0)
                    {
                        devEntryPtr->devFlag &= ~(tt16Bit)TM_DEV_IP_NO_CHECK;
                    }
                    else
                    {
                        devEntryPtr->devFlag |= (tt16Bit)TM_DEV_IP_NO_CHECK;
                    }
                    break;
#ifdef TM_USE_DHCP_COLLECT
                case TM_DEV_OPTIONS_DHCP_COLLECT_CACHE:
                    if (optionValue < 1)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->devDhcpCollectSize
                                                = (unsigned int)optionValue;
                    }
                    break;
                case TM_DEV_OPTIONS_DHCP_COLLECT_TIME:
                    devEntryPtr->devDhcpCollectTime
                                                = (unsigned int)optionValue;
                    break;
#endif /* TM_USE_DHCP_COLLECT */
#ifdef TM_6_USE_TEMPADDR
/*
 * [RFC4941] Privacy Extensions for Stateless Address Autoconfig in IPv6.
 * The following options control the autoconfiguration of temporary
 * global addresses for the specified interface.
 */
/*
 * use_tempaddr option values (similar to Linux sysctl):
 *  0: disable privacy extensions.
 *  1: enable privacy extensions but prefer public addresses over
 *     temporary addresses.
 *  2: enable privacy extensions and prefer temporary addresses over
 *     public addresses.
 */
                case TM_6_DEV_OPTIONS_USE_TEMPADDR:
                    if (optionValue > 2)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->dev6Flags2 &= (tt8Bit)
                                ~(TM_6_DEV_USE_TEMP | TM_6_DEV_PREFER_TEMP);
                        if (optionValue > 0)
                        {
                            devEntryPtr->dev6Flags2 |= TM_6_DEV_USE_TEMP;
                            if (optionValue > 1)
                            {
                                devEntryPtr->dev6Flags2
                                                    |= TM_6_DEV_PREFER_TEMP;
                            }
                        }
                    }
                    break;
/*
 * Valid lifetime (seconds) of a temporary address (default: 1 week)
 */
                case TM_6_DEV_OPTIONS_TEMP_VALID_LFT:
                    if (optionValue32 <= TM_6_TEMP_REGEN_ADVANCE)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->dev6TempValidLifetime = optionValue32;
                        (void)tf6TempAddrDevCheck(devEntryPtr);
                    }
                    break;
/*
 * Preferred lifetime (seconds) of a temporary address (default: 1 day)
 */
                case TM_6_DEV_OPTIONS_TEMP_PREFER_LFT:
                    if (optionValue32 <= TM_6_TEMP_REGEN_ADVANCE)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->dev6TempPreferLifetime = optionValue32;
                        (void)tf6TempAddrDevCheck(devEntryPtr);
                    }
                    break;
/*
 * Maximum desynchronizing factor (seconds), which is a random value
 * that helps prevent interfaces from generating temporary addresses
 * at the same time (default: 10 minutes).
 */
                case TM_6_DEV_OPTIONS_TEMP_MAX_DESYNC:
                    if (optionValue32 == TM_32BIT_ZERO)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->dev6TempMaxDesyncFactor = optionValue32;
                    }
                    break;
/*
 * Maximum attempts to generate a valid temporary address
 * (default: 3 retries).
 */
                case TM_6_DEV_OPTIONS_TEMP_MAX_RETRY:
                    if (optionValue <= 0)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        devEntryPtr->dev6TempRegenMaxRetry = optionValue;
                        (void)tf6TempAddrDevCheck(devEntryPtr);
                    }
                    break;
/*
 * Enable or disable temporary address auto-configuration for a prefix.
 */
                case TM_6_DEV_OPTIONS_TEMP_SET_FILT:
                    if (   (optionLength < sizeof(tt6TempPrefixFilter))
                        || (optionValuePtr == TM_VOID_NULL_PTR))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        errorCode = tf6TempUpdatePrefixFilter(
                                    devEntryPtr,
                                    (tt6TempPrefixFilterPtr)optionValuePtr,
                                    TM_8BIT_YES);   /* Add/update */
                    }
                    break;
/*
 * Remove all prefix filters for temporary address auto-configuration.
 */
                case TM_6_DEV_OPTIONS_TEMP_RESET_FILT:
/* Free the include/exclude prefix list */
                    tfListFree(&devEntryPtr->dev6TempPrefixFilter);
                    break;
#endif /* TM_6_USE_TEMPADDR */
#ifdef TM_6_DAD_ALLOW_NO_WAIT
                case TM_6_DEV_OPTIONS_DAD_NO_WAIT:
                    if (optionValue == 0)
                    {
                        devEntryPtr->dev6Flags2 &= ~TM_6_DEV_DAD_NO_WAIT;
                    }
                    else
                    {
                        devEntryPtr->dev6Flags2 |= TM_6_DEV_DAD_NO_WAIT;
                    }
                    break;
#endif /* TM_6_DAD_ALLOW_NO_WAIT */
                default:
                    break;
            }
/* UnLock device driver lock. */
            tm_call_unlock(&(devEntryPtr->devDriverLockEntry));
        }
/* UnLock device lock. */
        tm_call_unlock(&(devEntryPtr->devLockEntry));
    }


    return errorCode;
}

/*
 * tfInterfaceGetOptions function description:
 * Called by the user to get the value of various interface options
 */
int tfInterfaceGetOptions ( ttUserInterface interfaceHandle,
                            int             optionName,
                            void   TM_FAR * optionValuePtr,
                            int             optionLength)
{
    ttDeviceEntryPtr   devEntryPtr;
    int                errorCode;
    int                optionLengthFlag;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check on parameters */
    errorCode = tfValidInterface(devEntryPtr);
    if (errorCode == TM_ENOERROR)
    {
        errorCode = TM_EINVAL; /* assume bad parameter */
/* Lock device */
        tm_call_lock_wait(&(devEntryPtr->devLockEntry));
/* Check on length of option */
        if (optionValuePtr != (void TM_FAR *)0)
        {
/* Extract option length flag part of the optionName */
            optionLengthFlag = optionName & (   TM_DEV_OFLAG_SHORT
                                              | TM_DEV_OFLAG_UCHAR
                                              | TM_DEV_OFLAG_INT
                                              | TM_DEV_OFLAG_INT32 );
/* Get the optionValue based on the option length flag part of the optionName */
            switch (optionLengthFlag)
            {
                case TM_DEV_OFLAG_SHORT:
                    if (optionLength == (int)sizeof(short))
                    {
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_UCHAR:
                    if  (optionLength == (int)sizeof(unsigned char))
                    {
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_INT:
                    if  (optionLength == (int)sizeof(int))
                    {
                        errorCode = TM_ENOERROR;
                    }
                    break;
                case TM_DEV_OFLAG_INT32:
                    if  (optionLength == (int)sizeof(ttUser32Bit))
                    {
                        errorCode = TM_ENOERROR;
                    }
                    break;
                default:
                    break;
            }
        }

/* End of check on length of option */
        if (errorCode == TM_ENOERROR)
        {
            switch (optionName)
            {
                case TM_DEV_OPTIONS_FORWARDING:
                    if (tm_4_dev_forward(devEntryPtr))
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)1;
                    }
                    else
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)0;
                    }
                    break;
                case TM_DEV_OPTIONS_FORWARD_REFLECT:
                    if (tm_dev_forward_reflect(devEntryPtr))
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)1;
                    }
                    else
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)0;
                    }
                    break;
                case TM_DEV_OPTIONS_IP_PROMISCUOUS:
                    if (tm_dev_no_check(devEntryPtr))
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)1;
                    }
                    else
                    {
                        *(unsigned char TM_FAR *)optionValuePtr =
                                                            (unsigned char)0;
                    }
                    break;
                default:
                    break;
            }
        }
/* UnLock device lock. */
        tm_call_unlock(&(devEntryPtr->devLockEntry));
    }

    return errorCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trdevopt.c    File Revision: 6.0.2.16 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  8701    Output checksum:   488
 */
