/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Interface Transmit Queue Interface
 *
 * Filename: trdevxmt.c
 * Author: Odile
 * Date Created: 01/18/00
 * $Source: source/sockapi/trdevxmt.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:42:59CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

#if    defined(TM_LOCK_NEEDED) \
    || (    defined(TM_ERROR_CHECKING) \
         && defined(TM_USE_ANSI_LINE_FILE)  \
         &&  (   defined(TM_KERNEL_VISUAL_X86) \
              || defined(TM_KERNEL_WIN32_X86) \
              || defined(TM_MEMORY_LOGGING) ) )
#include <trglobal.h>
#endif /*    TM_LOCK_NEEDED 
          || (    TM_ERROR_CHECKING
               && TM_USE_ANSI_LINE_FILE 
               && (    TM_KERNEL_VISUAL_X86
                    || TM_KERNEL_WIN32_X86
                    || TM_MEMORY_LOGGING ) ) */
 /*
 * Local macros
 */

/* Ring of queued transmit buffers is empty */
/* #define TM_RING_EMPTY (tt8Bit)0x0 */ /* in trmacro.h */
/* Ring of queued transmit buffers is neither empty, nor full */
#define TM_RING_OTHER (tt8Bit)0x10
/* Ring of queued transmit buffers is full */
#define TM_RING_FULL  (tt8Bit)0x20

/*
 * Queue is not empty
 */
/* #define tm_iface_xmit_queue_not_empty(devXmitPtr)                        \
 *                         ( devXmitPtr->xmitFlag != TM_RING_EMPTY )
 */ /* in trmacro.h */

/*
 * Queue is full
 */
#define tm_iface_xmit_queue_full(devXmitPtr)                                \
                          ( devXmitPtr->xmitFlag  == TM_RING_FULL )
/*
 * Local functions
 */

/*
 * Try and send a buffer, queue it in the transmit queue if buffers
 * are queued ahead, and the transmit queue could not be empty, or if
 * the driver send routine returned an error.
 */
int tfDeviceSendBuffer ( ttUserInterface  interfaceHandle,
                         char     TM_FAR  *dataPtr,
                         int              dataLength,
                         int              dataFlag );

#ifdef TM_USE_DRV_ONE_SCAT_SEND
int tfDeviceOneScatSendBuffer ( ttUserInterface  interfaceHandle,
                                ttUserPacketPtr  packetPtr );
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

/*
 * Try and empty the interface transmit ring, by calling the device driver
 * send function for each buffer in the transmit ring until, the ring
 * is empty, or the send function returns an error. If the send function
 * succeeds, dequeue the buffer from the transmit ring.
 */
int tfDeviceEmptyXmitRing(ttUserInterface interfaceHandle);

/*
 * Enable/Disable the use of an interface transmit queue.
 * Can only be called when the device is not configured.
 * Cannot be called for a point to point device.
 * Parameters:
 * interfaceHandle     Interface handle as returned by tfAddInterface()
 * numberXmitDesc      Number of transmit descriptors that can be queued in
 *                     memory. If positive, the interface transmit queue
 *                     is enabled. If 0, it is disabled.
 * Returns:
 * TM_ENOERROR         no error
 * TM_ENOBUFS          Not enough memory to allocate the empty transmit ring
 * TM_EPERM            interface is already opened, or interface uses
 *                     point to point link layer that does byte stuffing
 *                     (i.e SLIP, or PPP).
 * TM_EINVAL           Invalid interface handle, or numberXmitDesc < 0
 */
int tfUseInterfaceXmitQueue ( ttUserInterface interfaceHandle,
                              short           numberXmitDesc )
{
    ttDeviceEntryPtr  devEntryPtr;
    ttDevXmitEntryPtr devXmitPtr;
    int               errorCode;

    errorCode = TM_EINVAL;
    if (numberXmitDesc >= 0)
    {
        devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check on parameters */
        if ( tfValidInterface(devEntryPtr) == TM_ENOERROR )
        {
/* Lock device */
            tm_call_lock_wait(&(devEntryPtr->devLockEntry));
/*
 * Lock device driver lock. OK to keep the device lock, since we never lock
 * the device lock after locking the device driver lock
 */
            tm_call_lock_wait(&(devEntryPtr->devDriverLockEntry));
            if (    tm_16bit_one_bit_set( devEntryPtr->devFlag,
                                          TM_DEV_OPENED)
                 || tm_ll_is_byte_stuffing(devEntryPtr)
                 || tm_16bit_one_bit_set( devEntryPtr->devFlag,
                                          TM_DEV_XMIT_TASK) )
            {
/*
 * Disallow changing transmit queue option after device has been opened
 * Disallow transmit queue option on a point to point link layer
 * device that uses byte stuffing, since 1. buffers could be small,
 * 2. the link layer send routine re-uses the buffer upon return from the
 * routine.
 * Disallow concurrent transmit buffer queue, and transmit task, otherwise
 * the transmit queue logic code would not be re-entrant, if the user
 * calls tfInterfaceSpinLock().
 */
                errorCode = TM_EPERM;
            }
            else
            {
                errorCode = TM_ENOERROR;
/*
 * Transmit queue at the driver level to let the driver postpone sending
 * the data until it is ready.
 */
                if ( devEntryPtr->devXmitEntryPtr != (ttDevXmitEntryPtr)0 )
/* If user had already set this option, free ring of buffers */
                {
                    tm_free_raw_buffer((ttRawBufferPtr)(ttVoidPtr)
                                            devEntryPtr->devXmitEntryPtr);
                    devEntryPtr->devXmitEntryPtr = (ttDevXmitEntryPtr)0;
                    devEntryPtr->devSendFuncPtr =
                                              devEntryPtr->devSavedSendFuncPtr;
                    devEntryPtr->devSavedSendFuncPtr = (ttDevSendFuncPtr)0;
#ifdef TM_USE_DRV_ONE_SCAT_SEND
                    if ( devEntryPtr->devSavedOneScatSendFuncPtr != 
                                                   (ttDevOneScatSendFuncPtr)0 )
                    {
                        devEntryPtr->devOneScatSendFuncPtr =
                                       devEntryPtr->devSavedOneScatSendFuncPtr;
                        devEntryPtr->devSavedOneScatSendFuncPtr = 
                                                   (ttDevOneScatSendFuncPtr)0;
                    }
#endif /* TM_USE_DRV_ONE_SCAT_SEND */
                    devEntryPtr->devEmptyXmitFuncPtr =
                                                    (ttDevEmptyXmitFuncPtr)0;
                }
                if (numberXmitDesc > 0)
                {
/* Allocate an empty transmit ring of buffers */
                    devXmitPtr = (ttDevXmitEntryPtr)
                     tm_get_raw_buffer((ttPktLen)
                        (   sizeof(ttDevXmitEntry)
/* numberXmitDesc - 1, because ttDevXmitEntry already contains one slot */
                         + (unsigned short)(  (numberXmitDesc - 1)
                                            * (short)sizeof(ttRingEntry) ) ));
                    if (devXmitPtr == (ttDevXmitEntryPtr)0)
                    {
/* No memory to allocate empty ring of buffers */
                        errorCode = TM_ENOBUFS;
                    }
                    else
                    {
/* Number of slots in ring of buffers */
                        devXmitPtr->xmitSlots = (tt16Bit)numberXmitDesc;
/* Read index in ring of buffers */
                        devXmitPtr->xmitReadIndex = 0;
/* Write index in ring of buffers */
                        devXmitPtr->xmitWriteIndex = 0;
/* Ring of buffers initially empty */
                        devXmitPtr->xmitFlag = TM_RING_EMPTY;
/* No buffer transmission error */
                        devXmitPtr->xmitErrorFlag = TM_8BIT_ZERO;
                        devXmitPtr->xmitBytes = 0;
                        devEntryPtr->devXmitEntryPtr = devXmitPtr;
                        devEntryPtr->devSavedSendFuncPtr =
                                                  devEntryPtr->devSendFuncPtr;
                        devEntryPtr->devSendFuncPtr = tfDeviceSendBuffer;
#ifdef TM_USE_DRV_ONE_SCAT_SEND
                        if ( devEntryPtr->devOneScatSendFuncPtr != 
                                                   (ttDevOneScatSendFuncPtr)0 )
                        {
                            devEntryPtr->devSavedOneScatSendFuncPtr =
                                           devEntryPtr->devOneScatSendFuncPtr;
                            devEntryPtr->devOneScatSendFuncPtr = 
                                                     tfDeviceOneScatSendBuffer;
                        }
#endif /* TM_USE_DRV_ONE_SCAT_SEND */
                        devEntryPtr->devEmptyXmitFuncPtr =
                                                      tfDeviceEmptyXmitRing;
                    }
                }
            }
/* UnLock device driver lock. */
            tm_call_unlock(&(devEntryPtr->devDriverLockEntry));
/* UnLock device lock. */
            tm_call_unlock(&(devEntryPtr->devLockEntry));
        }
    }
    return errorCode;
}

/*
 * Query whether an interface uses the xmit queue option
 * Parameters:
 * interfaceHandle     Interface handle as returned by tfAddInterface()
 * Returns:
 * 1                   Interface Xmit queue option is on
 * 0                   Interface Xmit queue option is not on
 * TM_DEV_ERROR        Interface Handle is not valid.
 */
int tfInterfaceXmitQueueOn ( ttUserInterface interfaceHandle )
{
    int              retCode;

    retCode = TM_DEV_ERROR; /* Assume failure */
    if ( tfValidInterface((ttDeviceEntryPtr)interfaceHandle) == TM_ENOERROR )
    {
        if ( tm_iface_xmit_queue_option(interfaceHandle) )
        {
            retCode = 1; /* Transmit queue used by the user */
        }
        else
        {
            retCode = 0; /* User is not using the Treck Transmit queue */
        }
    }
    return retCode;
}

/*
 * tfDeviceSendBuffer() function description.
 * Called from the post link layer function (i.e tfDeviceSendOnePacket()) for
 * non serial devices, when the user enabled the interface transmit queue.
 * The post link layer function was itself called from tfDeviceSend().
 * 1. First we try and empty the ring of queued transmit buffers.
 * 2. Next we check whether we should drop the current buffer. We will
 *    drop the current buffer, if we either dropped all the previous
 *    buffers from the same packet, or we dropped at least one buffer,
 *    and the current buffer is an intermediate buffer (TM_USER_BUFFER_MORE).
 * 3. Else, we are not dropping the current buffer:
 * 3.1 If the transmit ring is empty, then we try and send the current buffer.
 *     Note that if this the last buffer of a scattered packet, and we had
 *     dropped previous buffers from the same packet, we adjust the driver
 *     send function data length parameter, to make sure that the total
 *     length of the Ethernet packet is 60 bytes.
 * 3.2 If the device driver send function for the current buffer returned
 *     with no error, then update the number of transmitted/queued bytes
 *     for the current packet, but only if this is not the last buffer.
 * 3.3 If we could not send the current buffer (either because the
 *     transmit ring was not empty, or because the device driver send
 *     failed), then we try and queue the buffer to the device
 *     transmit queue.
 * 3.3.1 If the transmit ring is full, we set the current packet error flag
 *       since we will have to drop part of (or all) the packet, to remember
 *       that some piece of (or all) this packet could not get transmitted.
 *       Then if there are previously sent (or queued) buffers
 *       (xmitBytes != 0), try and walk back the transmit queue over any
 *       intermediate buffers, i.e any buffer with the TM_USER_BUFFER_MORE
 *       flag, to drop the previously queued buffers. Then we check,
 *       whether we should drop the current buffer, because no part of the
 *       current packet had been sent to the device driver (xmitBytes == 0),
 *       or because the current buffer is an intermediate buffer. Note that
 *       if we decide not to drop the packet, the transmit queue should
 *       not be full at this point.
 * 3.2.2 If we are not dropping the buffer, just store the buffer
 *       data pointer, data flag, data length in the next empty slot in the
 *       transmit ring, and update the xmitBytes count if the buffer is
 *       not the last buffer in the packet.
 * 4. If the TM_USER_BUFFER_LAST flag is set, i.e this is the last buffer
 *    in a packet:
 * 4.1 . If the transmit ring current packet error flag is set:
 * 4.1.1  . We check if we dropped that last buffer. In it is the case,
 *          this means that no previous intermediate buffer got sent or
 *          queued, so we set the errorCode to TM_ENOBUFS to let the user
 *          know that no part of the buffer was sent.
 * 4.1.2  . Otherwise if we did not drop that last buffer, we set the
 *          errorCode to TM_EIO to let the user know that part of the
 *          packet was lost.
 * 4.1.3  . We reset the transmit ring current packet error flag.
 * 4.2 We reset the xmitBytes count for the next packet to transmit.
 * 5. Return error code to the user.
 *
 * NOTE that with this scheme, if any part of a packet is sent, 
 * we are always guaranteed that the first buffer is sent, and the end of
 * the packet (the buffer with the TM_USER_BUFFER_LAST) is sent. (It is
 * possible that the end of the packet (the buffer with the
 * TM_USER_BUFFER_LAST) could not be sent, but in that case, we are
 * guaranteed that none of the other pieces of the packet were sent.)
 *
 * Parameters
 * Parameter             Meaning
 * interfaceHandle  Interface handle as returned by tfAddInterface()
 * dataPtr          Pointer to contiguous buffer
 * dataLength       Length of the data in the buffer.
 * dataFlag         Either TM_USER_BUFFER_MORE to indicate that more buffers
 *                  are coming (intermediate buffer),
 *                  Or TM_USER_BUFFER_LAST to indicate that this is the last
 *                  buffer in a packet (last buffer).
 *
 * Return:
 * Value        Meaning
 *
 * TM_ENOERROR  If flag is TM_USER_BUFFER_LAST, then entire packet was sent.
 *              If flag is TM_USER_BUFFER_MORE, buffer could have been sent
 *              or queued, or dropped.
 * TM_ENOBUFS   only returned if flag is TM_USER_BUFFER_LAST:
 *              The entire packet was dropped. The caller can free the
 *              packet, and remove it from the send queue.
 * TM_EIO       only returned if flag is TM_USER_BUFFER_LAST:
 *              part of the packet (but not all) including the last buffer
 *              was sent or queued.
 */
int tfDeviceSendBuffer ( ttUserInterface  interfaceHandle,
                         char     TM_FAR  *dataPtr,
                         int              dataLength,
                         int              dataFlag )
{
    ttDeviceEntryPtr   devEntryPtr;
    ttDevXmitEntryPtr  devXmitPtr;
    ttRingEntryPtr     xmitRingPtr;
    ttS16Bit           storeIndex;
    int                sendDataLength;
    int                retCode;
    int                errorCode;
    tt8Bit             dropTheBuffer;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    devXmitPtr = devEntryPtr->devXmitEntryPtr;
    errorCode = TM_ENOERROR;
    retCode = TM_ENOERROR;
    if ( tm_iface_xmit_queue_not_empty(devXmitPtr) )
    {
        retCode = tfDeviceEmptyXmitRing(interfaceHandle);
    }
    if (    devXmitPtr->xmitErrorFlag
         && (    (devXmitPtr->xmitBytes == 0)
              || (dataFlag == TM_USER_BUFFER_MORE) ) )
    {
/*
 * If we either dropped all the previous buffers from the same
 * packet, or we dropped at least one buffer, and the current buffer is
 * an intermediate buffer. In both of these cases, drop the current buffer.
 */
        dropTheBuffer = TM_8BIT_YES;
    }
    else
/*
 * no buffer sent so far with any error, or we had already
 * transmitted some TM_USER_BUFFER_MORE buffers, and this is the last
 * buffer.
 */
    {
        dropTheBuffer = TM_8BIT_ZERO;
        if (retCode == TM_ENOERROR)
/*
 * Queue is empty.
 * Either there were no buffers in the transmit queue, or we just
 * transmitted them all successfully. Go ahead and transmit the
 * current buffer.
 */
        {
            sendDataLength = (int)dataLength;
            if (    devXmitPtr->xmitErrorFlag
                 && (dataFlag == TM_USER_BUFFER_LAST)
                 && tm_4_ll_is_lan(devEntryPtr) )
/*
 * If we dropped at least one TM_USER_BUFFER_MORE buffer, and this is the
 * last buffer, and we are on Ethernet, make sure we send at least 60 bytes.
 */
            {
                if (   (devXmitPtr->xmitBytes + dataLength)
                     < TM_ETHER_MIN_PACKET )
                {
                    sendDataLength =   TM_ETHER_MIN_PACKET
                                     - (int)devXmitPtr->xmitBytes;
                }
            }
/*
 * Set this flag, to prevent tfSendCompleteInterface() to call
 * tfDeviceEmptyXmitRing in this thread.
 */
            devEntryPtr->devFlag |= TM_DEV_XMITING;
            retCode = (*devEntryPtr->devSavedSendFuncPtr)( interfaceHandle,
                                                           dataPtr,
                                                           sendDataLength,
                                                           dataFlag );
            devEntryPtr->devFlag &= ~TM_DEV_XMITING;
        }
        if (retCode == TM_ENOERROR)
        {
            if (dataFlag != TM_USER_BUFFER_LAST)
            {
/* Keep track of the number of bytes transmitted in the current packet */
                devXmitPtr->xmitBytes = (tt16Bit)(   devXmitPtr->xmitBytes
                                                   + (tt16Bit)dataLength );
            }
        }
        else
/*
 * We could not transmit the current buffer, either because we
 * could not empty the transmit queue, or because the current buffer send
 * failed. Queue the current buffer.
 */
        {
            if (tm_iface_xmit_queue_full(devXmitPtr))
            {
/*
 * Queue if full. We need to drop part of (or all) the current packet.
 * Set the error flag so that we remember that part of (or the entire)
 * packet did not get sent.
 */
                devXmitPtr->xmitErrorFlag = TM_8BIT_YES;
/*
 * Remove all the buffers queued in the transmit queue
 * that are part of the current packet to make room.
 */
                if (devXmitPtr->xmitBytes != 0)
                {
                    do
                    {
                        storeIndex = devXmitPtr->xmitWriteIndex;
                        if (storeIndex == 0)
                        {
                            storeIndex = (ttS16Bit)devXmitPtr->xmitSlots;
                        }
                        storeIndex = (ttS16Bit)(storeIndex - 1);
                        xmitRingPtr = &devXmitPtr->xmitRing[storeIndex];
                        if (xmitRingPtr->ringDataFlag == TM_USER_BUFFER_MORE)
                        {
#ifdef TM_ERROR_CHECKING
                            if (   devXmitPtr->xmitBytes
                                 < xmitRingPtr->ringDataLength )
                            {
                                tfKernelError("tfDeviceSendBuffer",
                                              "Corrupted transmit queue\n");
                            }
#endif /* TM_ERROR_CHECKING */
                            devXmitPtr->xmitBytes = (tt16Bit)
                                             (   devXmitPtr->xmitBytes
                                               - xmitRingPtr->ringDataLength);
                            if (storeIndex == devXmitPtr->xmitReadIndex)
                            {
                                devXmitPtr->xmitFlag = TM_RING_EMPTY;
                            }
                            else
                            {
                                devXmitPtr->xmitFlag = TM_RING_OTHER;
                            }
/* Move write index back */
                            devXmitPtr->xmitWriteIndex = storeIndex;
                        }
                    }
                    while (    (    xmitRingPtr->ringDataPtr
                                 == TM_USER_BUFFER_MORE )
                            && (tm_iface_xmit_queue_not_empty(devXmitPtr)) );

                }
/*
 * At this point, either we have removed part of (or all) the current packet
 * from the queue, or the queue is still full, because the data is not
 * scattered.
 */
                if (   (devXmitPtr->xmitBytes == 0)
/* No part of the buffer has been transmitted, or is currently queued */
                     || (dataFlag == TM_USER_BUFFER_MORE) )
/* Or this is not the last packet */
                {
                    dropTheBuffer = TM_8BIT_YES;
                }
#ifdef TM_ERROR_CHECKING
                else
                {

                    if (tm_iface_xmit_queue_full(devXmitPtr) )
                    {
                        tfKernelError("tfDeviceSendBuffer",
                                       "Corrupted transmit queue\n");
                    }
                }
#endif /* TM_ERROR_CHECKING */
            }
            if (dropTheBuffer == TM_8BIT_ZERO)
            {
/* Queue not full, insert */
                storeIndex = devXmitPtr->xmitWriteIndex;
                xmitRingPtr = &devXmitPtr->xmitRing[storeIndex];
                if (dataFlag != TM_USER_BUFFER_LAST)
                {
                    devXmitPtr->xmitBytes = (tt16Bit)(   devXmitPtr->xmitBytes
                                                       + (tt16Bit)dataLength);
                }
                xmitRingPtr->ringDataPtr = dataPtr;
                xmitRingPtr->ringDataLength = (tt16Bit)dataLength;
                xmitRingPtr->ringDataFlag = (tt16Bit)dataFlag;
/* Point to next buffer */
                storeIndex++;
                if ((tt16Bit) storeIndex == devXmitPtr->xmitSlots)
/* Wrap around of write index */
                {
                    storeIndex = 0;
                }
                devXmitPtr->xmitWriteIndex = storeIndex;
                if (storeIndex == devXmitPtr->xmitReadIndex)
                {
                    devXmitPtr->xmitFlag = TM_RING_FULL;
                }
                else
                {
                    devXmitPtr->xmitFlag = TM_RING_OTHER;
                }
            }
        }
    }
    if (dataFlag == TM_USER_BUFFER_LAST)
/* End of packet */
    {
/*
 * If any error occured during transmission of any buffer from the packet
 * corresponding to this last buffer.
 */
        if (devXmitPtr->xmitErrorFlag != TM_8BIT_ZERO)
        {
            if (dropTheBuffer != TM_8BIT_ZERO)
/* Last buffer did not get sent, i.e no buffer got sent. */
            {
                errorCode = TM_ENOBUFS;
            }
            else
            {
/*
 * Let the user know that the packet was partially stored or sent, but do nt
 * let the caller dequeue this packet from the send queue, and free it.
 * Since part of the packet was stored in the interface transmit queue,
 * or sent by the driver, including its last part, we know it will
 * eventually be dequeued, and freed via the tfSendCompleteInterface() call.
 */
                errorCode = TM_EIO;
            }
/* Reset errorFlag for the next packet to send */
            devXmitPtr->xmitErrorFlag = TM_8BIT_ZERO;
        }
/* Reset intermediate buffers transmitted bytes for next packet to send */
        devXmitPtr->xmitBytes = 0;
    }
    return errorCode;
}

#ifdef TM_USE_DRV_ONE_SCAT_SEND
/*
 * tfDeviceOneScatSendBuffer() function description.
 * Called from tfDeviceSend() for
 * non serial devices, when the user enabled the interface transmit queue,
 * and when the user enabled the single scattered send call.
 * 1. First we try and empty the ring of queued transmit packets.
 * 2. If the transmit ring is empty, then we try and send the current packet.
 * 4. If we could not send the current packet (either because the
 *    transmit ring was not empty, or because the device driver send
 *    failed), then we try and queue the packet to the device
 *    transmit queue.
 * 5. If the transmit ring is full, we drop the packet, and set the
 *    errorCode to TM_ENOBUFS to let the user know that the packet 
 *    was lost.
 * 6. Return error code to the user.
 *
 * Parameters
 * Parameter             Meaning
 * interfaceHandle  Interface handle as returned by tfAddInterface()
 * ttUserPacketPtr  pointer to packet.
 *
 * Return:
 * Value        Meaning
 *
 * TM_ENOERROR  packet was sent or queued for later transmission. 
 * TM_ENOBUFS   packet was dropped. The user can free the packet and
 *              remove it from the send queue.
 */
int tfDeviceOneScatSendBuffer ( ttUserInterface  interfaceHandle,
                                ttUserPacketPtr  packetUPtr )
{
    ttDeviceEntryPtr   devEntryPtr;
    ttDevXmitEntryPtr  devXmitPtr;
    ttRingEntryPtr     xmitRingPtr;
    ttS16Bit           storeIndex;
    int                retCode;
    int                errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    devXmitPtr = devEntryPtr->devXmitEntryPtr;
    errorCode = TM_ENOERROR;
    retCode = TM_ENOERROR;
    if ( tm_iface_xmit_queue_not_empty(devXmitPtr) )
    {
        retCode = tfDeviceEmptyXmitRing(interfaceHandle);
    }
    if (retCode == TM_ENOERROR)
/*
 * Queue is empty.
 * Either there were no packets in the transmit queue, or we just
 * transmitted them all successfully. Go ahead and transmit the
 * current packet.
 */
    {
/*
 * Set this flag, to prevent tfSendCompleteInterface() from calling
 * tfDeviceEmptyXmitRing in this thread.
 */
       devEntryPtr->devFlag |= TM_DEV_XMITING;
       retCode = (*devEntryPtr->devSavedOneScatSendFuncPtr)( 
                                                 interfaceHandle,
                                                 packetUPtr );
       devEntryPtr->devFlag &= ~TM_DEV_XMITING;
    }
    if (retCode != TM_ENOERROR)
/*
 * We could not transmit the current packet, either because we
 * could not empty the transmit queue, or because the current packet send
 * failed. Queue the current packet.
 */
    {
        if (!(tm_iface_xmit_queue_full(devXmitPtr)))
        {
/* Queue not full, insert */
            storeIndex = devXmitPtr->xmitWriteIndex;
            xmitRingPtr = &devXmitPtr->xmitRing[storeIndex];
            xmitRingPtr->ringPacketUPtr = packetUPtr;
/* Point to next packet */
            storeIndex++;
            if ((tt16Bit) storeIndex == devXmitPtr->xmitSlots)
/* Wrap around of write index */
            {
                storeIndex = 0;
            }
            devXmitPtr->xmitWriteIndex = storeIndex;
            if (storeIndex == devXmitPtr->xmitReadIndex)
            {
                devXmitPtr->xmitFlag = TM_RING_FULL;
            }
            else
            {
                devXmitPtr->xmitFlag = TM_RING_OTHER;
            }
        }
        else
        {
/*
 * Queue if full. We need to drop the current packet.
 */
            errorCode = TM_ENOBUFS;
        }
    }
    return errorCode;
}
#endif /* TM_USE_DRV_ONE_SCAT_SEND */

/*
 * Called from either tfDeviceSendBuffer() itself called from one of the
 * post link functions when the interface transmit queue is enabled,
 * or 
 * called from tfDeviceOneScatSendBuffer() itself called from tfDeviceSend
 * if the interface transmit queue is enabled and if the user has turned
 * on the single scattered device driver send mechanism
 *  or
 * called periodically by the user via the tfIoctlInterface() when the user
 * sets the TM_DEV_IOCTL_EMPTY_XMIT_FLAG to empty the interface transmit queue.
 * (which contains all the buffers that the device driver could not send
 *  right away).
 *
 * 1. Set error return code to no error.
 * 2. while the interface transmit ring is not empty
 * 2.1 . Dequeue the packet at the head of the queue, and try and call the
 *       device driver send function for that buffer, and set the error
 *       code with the return value from the device driver send.
 *       Note that if this the last buffer of a scattered packet, and we had
 *       dropped previous buffers from the same packet, we adjust the driver
 *       send function data length parameter, to make sure that the total
 *       length of the Ethernet packet is 60 bytes.
 * 2.2 . If the device driver send function returns with no error, increase
 *       a flag, to let the user know that some progress has occured.
 * 2.3 . Otherwise, requeue the packet, and break out of the loop.
 * Note that we dequeue the packet before sending the data, so that this
 * code can also work intra-machine for testing purposes, since in that
 * case another call to this routine could be made before we have a chance
 * to return from the device driver send function.
 * 3. Return error code.
 * Parameters
 * Name             Meaning
 * interfaceHandle  Interface handle as returned by tfAddInterface()
 * Return values
 * Value            Meaning
 * TM_ENOERROR    The interface transmit queue is empty
 * TM_EINPROGRESS The interface transmit queue is not empty, but some
 *                buffers got sent with no error.
 * Other          As returned from the device driver send routine. No
 *                buffer could be transmitted in this call.
 */
int tfDeviceEmptyXmitRing (ttUserInterface interfaceHandle)
{
    ttDeviceEntryPtr   devEntryPtr;
    ttDevXmitEntryPtr  devXmitPtr;
    ttRingEntryPtr     xmitRingPtr;
    ttS16Bit           readIndex;
    ttS16Bit           savReadIndex;
    int                dataLength;
    int                errorCode;
    int                progress;
    tt8Bit             savFlag;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    devXmitPtr = devEntryPtr->devXmitEntryPtr;
    errorCode = TM_ENOERROR;
    progress = 0;
/*
 * Set this flag, to prevent tfSendCompleteInterface() from calling
 * tfDeviceEmptyXmitRing in this thread.
 */
    devEntryPtr->devFlag |= TM_DEV_XMITING;
    while ( tm_iface_xmit_queue_not_empty(devXmitPtr) )
/* If there is a previously queued buffer, dequeue it, and transmit it. */
    {
        readIndex = devXmitPtr->xmitReadIndex;
        xmitRingPtr = &devXmitPtr->xmitRing[readIndex];
/*
 * Save previous readIndex, and xmitFlag, in case we have to requeue,
 * because the driver send function failed.
 */
        savReadIndex = readIndex;
        savFlag = devXmitPtr->xmitFlag;
/* Dequeue the sent buffer */
        readIndex++;
/* Point to next buffer */
        if ((tt16Bit) readIndex == devXmitPtr->xmitSlots)
/* Wrap around for read index. */
        {
            readIndex = 0;
        }
        devXmitPtr->xmitReadIndex = readIndex;
        if (readIndex == devXmitPtr->xmitWriteIndex)
        {
            devXmitPtr->xmitFlag = TM_RING_EMPTY;
        }
        else
        {
            devXmitPtr->xmitFlag = TM_RING_OTHER;
        }
#ifdef TM_USE_DRV_ONE_SCAT_SEND
        if ( devEntryPtr->devSavedOneScatSendFuncPtr != 
                                                   (ttDevOneScatSendFuncPtr)0 )
/* One scattered send call enabled */
        {
           errorCode = (*devEntryPtr->devSavedOneScatSendFuncPtr)( 
                                                 interfaceHandle,
                                                 xmitRingPtr->ringPacketUPtr);
        }
        else
#endif /* TM_USE_DRV_ONE_SCAT_SEND */
        {
            dataLength = (int)xmitRingPtr->ringDataLength;
            if (    devXmitPtr->xmitErrorFlag
                && (xmitRingPtr->ringDataFlag == TM_USER_BUFFER_LAST)
                && tm_4_ll_is_lan(devEntryPtr) )
/*
 * If we dropped some TM_USER_BUFFER_MORE packets, and this is the last
 * buffer, and we are on Ethernet, make sure we send at least 60 bytes.
 */
            {
                if (   (int)(devXmitPtr->xmitBytes + dataLength)
                    < TM_ETHER_MIN_PACKET )
                {
                    dataLength = 
                            TM_ETHER_MIN_PACKET - (int)devXmitPtr->xmitBytes;
                }
            }
            errorCode = (*devEntryPtr->devSavedSendFuncPtr)(
                                                  interfaceHandle,
                                                  xmitRingPtr->ringDataPtr,
                                                  dataLength,
                                                  xmitRingPtr->ringDataFlag );
        }
        if (errorCode == TM_ENOERROR)
        {
            progress++;
        }
        else
        {
/* Re-queue the buffer in its spot */
            devXmitPtr->xmitReadIndex = savReadIndex;
            devXmitPtr->xmitFlag = savFlag;
            break;
        }
    }
    devEntryPtr->devFlag &= ~TM_DEV_XMITING;
    if (errorCode != TM_ENOERROR)
    {
        if (progress != 0)
        {
            errorCode = TM_EINPROGRESS;
        }
    }
    return errorCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trdevxmt.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 26937    Output checksum: 57551
 */
