
/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2006
 * Copyright Elmic Systems Japan 1997 - 2006
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Interface Physical Port API functions
 *
 * Filename: trdvport.c
 * Author: Odile
 * Date Created: 04/12/06
 * $Source: source/sockapi/trdvport.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2011/02/14 16:06:31CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_STRONG_ESL

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * tfInterfaceSetPort function description
 * Allows the user to set a physical port for a given interface
 *
 * Parameters
 * Parameter        Description
 * interfaceHandle  The interface handle as returned by tfAddInterface
 * physPort         phys port set by the user.
 *
 * Returns
 * TM_ENOERROR      Success
 * TM_EINVAL        Invalid interface handle, or port.
 */
int
tfInterfaceSetPhysPort( ttUserInterface interfaceHandle,
                        int             physPort )
{
    ttDeviceEntryPtr devEntryPtr;
    int              errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check on parameters */
    if (tfValidInterface(devEntryPtr) == TM_ENOERROR)
    {
        errorCode = TM_ENOERROR;
/* Lock device driver */
        tm_call_lock_wait(&(devEntryPtr->devLockEntry));
        devEntryPtr->devPhysPort = physPort;
/* UnLock device driver */
        tm_call_unlock(&(devEntryPtr->devLockEntry));
    }
    else
    {
        errorCode = TM_EINVAL; /* bad parameter */
    }
    return errorCode;
}

/*
 * tfInterfaceGetPhysPort function description
 * Get the physical port for a given interface
 *
 * Parameters
 * Parameter        Description
 * interfaceHandle  The interface handle as returned by tfAddInterface
 *
 * Returns
 * >=               Physical port number
 * -1               Invalid interface handle.
 */
int
tfInterfaceGetPhysPort(ttUserInterface interfaceHandle)
{
    ttDeviceEntryPtr devEntryPtr;
    int              physPort;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check on parameters */
    if (tfValidInterface(devEntryPtr) == TM_ENOERROR)
    {
/* Lock device driver */
        tm_call_lock_wait(&(devEntryPtr->devLockEntry));
        physPort = devEntryPtr->devPhysPort;
/* UnLock device driver */
        tm_call_unlock(&(devEntryPtr->devLockEntry));
    }
    else
    {
        physPort = -1; /* invalid port */
    }
    return physPort;
}

/*
 * tfInterfaceGetPortMask function description
 * Get the physical port mask for a given interface
 *
 * Parameters
 * Parameter        Description
 * interfaceHandle  The interface handle as returned by tfAddInterface
 *
 * Returns
 *                   Physical port mask.
 * TM_UL(0xFFFFFFFF) is wildcard for invalid/null interfaceHandle.
 */
ttUser32Bit
tfInterfaceGetPortMask(ttUserInterface interfaceHandle)
{
    int              physPort;
    ttUser32Bit      portMask;

    physPort = tfInterfaceGetPhysPort(interfaceHandle);
/* Check on parameters */
    if (physPort == -1) /* invalid port */
    {
        portMask = TM_UL(0xFFFFFFFF);
    }
    else
    {
        portMask = 1 << physPort;
    }
    return portMask;
}

/*
 * tfInterfaceGetFromAddrPortMask Function Description
 * Called by the user/stack to get the interface handle given
 * a sock address and physical port mask.
 * physical port mask maps to a physical port if and only if one
 * bit is set, ortherwise it is a wild card physical port.
 * Parameters
 * Parameter          Description
 * sockAddrStoragePtr Pointer to a sockaddr_storage structure containing
 *                    a configured IP address on the queried interface
 *                    Null pointer means find first configured interface
 *                    on that physical port.
 * physPort           Physical port of the queried interface
 *
 * Returns
 * interfaceHandle   The interface handle as returned by tfAddInterface,
 * Null              No match found
 */
ttUserInterface
tfInterfaceGetFromAddrPortMask
                             ( ttConstSockAddrStoragePtr    sockAddrStoragePtr,
                               ttUser32Bit                  portMask )
{
    ttUserInterface interfaceHandle;
    int physPort;
    int tempPhysPort;  

    interfaceHandle = (ttUserInterface)0;
    tempPhysPort = 0;

    physPort = tfPortMaskToPhysPort(portMask);

    if (!tm_is_phys_port_wild(physPort))
    {
        interfaceHandle = 
            tfInterfaceGetFromAddrPhysPort(sockAddrStoragePtr, physPort);
    }
    else
    {
        if (tm_is_port_mask_wild(portMask))
        {
/* return the first valid interfaceHandle if one is found */
            interfaceHandle = 
                tfInterfaceGetFromAddrPhysPort(sockAddrStoragePtr, physPort);
        }
        else
        {
/* multiple physical ports selected, but not wildcard */
            while((portMask != 0) && (interfaceHandle == (ttUserInterface)0))
            {
                if (portMask & 0x01)
                {
/* try to find an interface for the selected physical port */
                    interfaceHandle = 
                        tfInterfaceGetFromAddrPhysPort(sockAddrStoragePtr, 
                                                        tempPhysPort);
                }
                tempPhysPort++;
                portMask >>= 1;
            }
        }
    }
    return interfaceHandle;
}

/*
 * tfInterfaceGetFromAddrPhysPort Function Description
 * Called by the user/stack to get the interface handle given
 * an sock address and phys port.
 * Parameters
 * Parameter          Description
 * sockAddrStoragePtr Pointer to a sockaddr_storage structure containing
 *                    a configured IP address on the queried interface
 *                    Null pointer means find first configured interface
 *                    on that physical port.
 * physPort           Physical port of the queried interface
 *
 * Returns
 * interfaceHandle   The interface handle as returned by tfAddInterface,
 * Null              No match found
 */
ttUserInterface
tfInterfaceGetFromAddrPhysPort
                             ( ttConstSockAddrStoragePtr    sockAddrStoragePtr,
                               int                          physPort )
{
    ttDeviceEntryPtr   devPtr;
#ifndef TM_SINGLE_INTERFACE_HOME
    ttDeviceEntryPtr   devNodePtr;
#endif /* !TM_SINGLE_INTERFACE_HOME */
    tt4IpAddress       ipAddress;
    ttSockAddrPtrUnion  sockAddrUnion;
#ifndef TM_SINGLE_INTERFACE_HOME
    unsigned int        indx;
    unsigned int        devCount;
    tt16Bit             ifaceMhome;
#endif /* !TM_SINGLE_INTERFACE_HOME */
    tt8Bit              addrFamily;

    devPtr = (ttDeviceEntryPtr)0;
    if (physPort >= 0 || tm_is_phys_port_wild(physPort))
    {
        tm_ip_copy(TM_IP_ZERO, ipAddress);
        addrFamily = TM_8BIT_ZERO; /* compiler warning */
        if (sockAddrStoragePtr != (ttConstSockAddrStoragePtr)0)
        {
            sockAddrUnion.sockPtr =
                                 (struct sockaddr TM_FAR *)sockAddrStoragePtr;
            addrFamily = sockAddrUnion.sockPtr->sa_family;
/* Copy always to avoid compiler warning */
            tm_ip_copy(
                    (tt4IpAddress)(sockAddrUnion.sockInPtr->sin_addr.s_addr),
                    ipAddress);
        }
/* Lock device list */
        tm_lock_wait(&tm_context(tvDeviceListLock));
#ifdef TM_SINGLE_INTERFACE_HOME
        devPtr = tm_context(tvDeviceList);
        if (devPtr != TM_DEV_NULL_PTR)
        {
/* Only allowed for IPv4 only */
            if (    tm_8bit_bits_not_set( tm_ip_dev_conf_flag(devPtr, 0),
                                          TM_DEV_IP_CONFIG )
                 || (    ( sockAddrStoragePtr != (ttConstSockAddrStoragePtr)0 )
                      && ( tm_ip_not_match( ipAddress,
                                            tm_ip_dev_addr(devPtr, 0) ) ) )
                 || ((devPtr->devPhysPort != physPort)
                        && (!tm_is_phys_port_wild(physPort)))
               )
            {
                devPtr = (ttDeviceEntryPtr)0;
            }
        }
#else /* !TM_SINGLE_INTERFACE_HOME */
        devPtr = TM_DEV_NULL_PTR;
        devCount = tm_context(tvDeviceList).listCount;
        devNodePtr = (ttDeviceEntryPtr)(ttVoidPtr)
                        tm_context(tvDeviceList).listHeadNode.nodeNextPtr;
        for ( indx = 0;
             (indx < devCount);
              indx++ )
        {
/* Match on physical port */
            if (   (devNodePtr->devPhysPort == physPort)
                || tm_is_phys_port_wild(physPort))
            {
                if (sockAddrStoragePtr != (ttConstSockAddrStoragePtr)0)
                {
                    {
                        if (tfMhomeAddrMatch(devNodePtr, ipAddress, &ifaceMhome)
                                                             == TM_ENOERROR)
                        {
                            devPtr = devNodePtr;
                            break;
                        }
                    }
                }
                else
                {
/* Wild card match on the address */
                    if (devNodePtr != tm_context(tvLoopbackDevPtr))
                    {
/*
 * Could also just check for TM_LL_CONNECTING | TM_LL_CONNECTED being set
 * on devNodePtr->devFlags
 */
                        ifaceMhome = (tt16Bit)TM_DEV_IP_CONFIG;
                        if (tfMhomeAnyConf(devNodePtr, ipAddress, &ifaceMhome)
                                                             == TM_ENOERROR)
                        {
                            devPtr = devNodePtr;
                            break;
                        }
                    }
                }
            }
            devNodePtr = (ttDeviceEntryPtr)(ttVoidPtr)
                    devNodePtr->devNodeEntry.nodeNextPtr; /* next */
        }
#endif /* !TM_SINGLE_INTERFACE_HOME */
/* UnLock device list */
        tm_unlock(&tm_context(tvDeviceListLock));
    }
    return (ttUserInterface)devPtr;
}

/*
 * tfPortMaskToPhysPort(portMask) Function description
 * Given a bit mask, return the position of the bit in the
 * mask if one and only one bit is set, otherwise return -1.
 * (Note that tfPortMaskToPhysPort() will return a valid physPort (>= 0) only
 *  if one and only one bit is set in the portMask. Returns -1 otherwise.)
 * Parameters
 * Parameter          Description
 *    portMask        32-bit bit mask.
 *
 * Returns
 *    physPort        port value corresponding to the bit set. 
 *    -1              wild card; more than one bit is set.
 */
int tfPortMaskToPhysPort (tt32Bit portMask)
{
    int     physPort;
#ifndef TM_OPTIMIZE_SIZE
    int     count;
#endif /* TM_OPTIMIZE_SIZE */

#ifndef TM_OPTIMIZE_SIZE
    union
    {
        tt8Bit  charBitMask[4];
        tt32Bit longBitMask;
    } origMaskU;
#endif /* !TM_OPTIMIZE_SIZE */

    physPort = -1; /* wild card */
    if (portMask)
    {
        if (((portMask) & (portMask - 1)) == TM_UL(0))
/* One bit set */
        {
#ifdef TM_OPTIMIZE_SIZE
            while (portMask)
            {
                physPort += 1;
                portMask >>= 1;
            }
#else /* !TM_OPTIMIZE_SIZE */
            origMaskU.longBitMask = portMask;
            count = 0;
            for (count = 0; count < 4; count++)
            {
                if (origMaskU.charBitMask[count] != 0)
                {
                    break;
                }
            }
            switch (origMaskU.charBitMask[count])
            {
                case 1:
                    physPort = 0;
                    break;
                case 2:
                    physPort = 1;
                    break;
                case 4:
                    physPort = 2;
                    break;
                case 8:
                    physPort = 3;
                    break;
                case 16:
                    physPort = 4;
                    break;
                case 32:
                    physPort = 5;
                    break;
                case 64:
                    physPort = 6;
                    break;
                case 128:
                    physPort = 7;
                    break;
                default:
                    break;
            }
            physPort = physPort + count * 8;
#endif /* !TM_OPTIMIZE_SIZE */
        }
    }
    return physPort;
}

#else /* !TM_USE_STRONG_ESL */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_USE_VCHAN is not defined */
int tvDvPortDummy = 0;
#endif /* ! TM_USE_STRONG_ESL */



/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trdvport.c    File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 62018    Output checksum: 45673
 */
