/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: inet_pton socket API.
 *
 * Filename: trpton.c
 * Author: Bryan Wang
 * Date Created:
 * $Source: source/sockapi/trpton.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2010/12/08 15:05:01CST $
 * $Author: pcarney $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* Include */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* constant declarations and enumerations */

/* type definitions */

/* local function prototypes */
int inet_pton4(ttConstCharPtr src, tt32BitPtr dst);

/* local variable definitions */

/* static const tables */

/* macro definitions */
#define TM_6_MAX_IP_SEGMENT   (TM_6_IP_ADDRESS_LENGTH / 2)

/****************************************************************************
* FUNCTION: inet_pton
*
* PURPOSE: converts an address in its standard text presentation form into
*   its numeric binary form.
*
* PARAMETERS:
*   af:
*       specifies the family of the address. Currently the AF_INET and
*       AF_INET6 address families are supported.
*   src:
*       points to the string being passed in.
*   dst:
*       points to a buffer into which the function stores the numeric address
*       The address is returned in network byte order. The calling
*       appliaction must ensure that the buffer referred to by dst is large
*       enough to hold the numeric address(4 bytes for V4 address, and
*       16 bytes for V6 address)
* RETURNS:
*   1:
*       Sucessful
*   0:
*       input is not a valid IPv4 dotted-decimal string or a valid IPv6
*       address string, or dst is NULL
*  -1:
*       af argument is unknown
* ERRORNO:
*       TM_EAFNOSUPPORT, when returns -1
* NOTES:
*       For ipv6 address, Many implementations of the existing inet_addr()
*       and inet_aton() functions accept nonstandard input: octal numbers,
*       hexadecimal numbers, and fewer than four numbers. inet_pton() does
*       not accept these formats.
*
*       DSP targets: These routines have been modified to work with
*                    correctly with 32-bit DSP's.  Since these routines
*                    always treat the output as a 16-bit array, no
*                    modification should be necessary for 16-bit DSP's.
*
*       We are returning 0 for NULL dst, though this is not specified by
*       RFC 2553
*
****************************************************************************/
int inet_pton(
    int                af,
    ttUserConstCharPtr src,
    ttUserVoidPtr      dst )
{
    struct in_addr  inAddr;
    int             errorCode;

    if ((src == (ttUserConstCharPtr)0) || (dst == (ttUserVoidPtr)0))
    {
        errorCode = 0;
        goto inetptonfinish;
    }

    switch(af)
    {
        case AF_INET:
            errorCode = inet_pton4(src, &(inAddr.s_addr));
            tm_bcopy(&inAddr,
                     dst,
                     sizeof(struct in_addr));
            break;
        default:
            errorCode = -1;
            break;
    }

inetptonfinish:
    if (errorCode == -1)
    {
/* store errno into the non socket specific globle variable */
        tm_context(tvErrorCode) = TM_EAFNOSUPPORT;
    }

    return errorCode;
}


/*
 *  return:
 *      1:   sucess
 *      0:   invalid address string
 */
int inet_pton4(ttConstCharPtr src, tt32BitPtr dst)
{
    tt32Bit addrVal;
    int     errorCode;
    int     strcmpResult;

    errorCode = 1;

/*
 * if inet_aton return 0, we can't tell wheter it is an error,
 * or the input address is 0.0.0.0, so we handle it oouselves
 */
    strcmpResult = tm_strcmp(src, "0.0.0.0");
    if (strcmpResult == 0)
    {
        *dst = TM_UL(0);
    }
    else
    {
        addrVal = inet_aton((char TM_FAR *)src);
        if (addrVal != 0)
        {
/* inet_aton returns 0 on error */
            *dst = addrVal;

        }
        else
        {
            errorCode = 0;
        }
    }
    return errorCode;
}

/*
 *  return:
 *      1:   sucess
 *      0:   invalid address string
 */

/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trpton.c      File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 11287    Output checksum: 18504
 */
