/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (tfRecvFromToInterface)
 *
 * Filename: trrcvtif.c
 * Author: Odile
 * Date Created: 12/14/00
 * $Source: source/sockapi/trrcvtif.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:00CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * tfRecvFromToInterface function description
 * Take some data from the receive queue and give it to the user.
 * Same as tfRecvFromTo, but takes 3 extra parameters to let the user get
 * the link layer source address of the packet, and the interface the packet
 * arrived on.
 * Parameters
 * Parameter          Description
 * socketDescriptor   the socket descriptor to use to recv data
 * bufferPtr          the buffer to store the received data
 * bufferLength       the length of the buffer to store the received data
 * flags              MSG_DONTWAIT: Don't wait for data
 *                    0: wait for data to come in
 * fromAddressPtr     where to store the address the data came from
 * addressLengthPtr   length of the area pointed to by fromAddressPtr, or
 *                    toAddressPtr
 * toAddressPtr       where to store the address the data was sent to
 * fromPhyAddrPtr     Where to store the link layer address the packet came
 *                    from
 * fromPhyAddrLenPtr  Pointer to length of the area pointed to by
 *                    fromPhyAddrPtr. Length updated to length of address
 *                    copied.
 * interfaceHandlePtr Pointer to interface handle (as returned by
 *                    tfAddInterface()) the packet arrived on.
 *
 * Returns
 * Value            Meaning
 * >=0              Number of bytes actually sent on the socket
 * -1               an error occured, error can be retrieved with
 *                  tfGetSocketError()
 * tfGetSocketError will return:
 * TM_EBADF         The socket descriptor is invalid
 * TM_ENOBUFS       There was insufficient user memory available to complete
 *                  the operation
 * TM_EMSGSIZE      The message was too long
 * TM_EPROTOTYPE    TCP protocol requires usage of recv, not recvfrom
 * TM_EWOULDBLOCK   The socket is marked as non-blocking and
 *                  the recv operation would block
 */
int tfRecvFromToInterface(
                    int                     socketDescriptor,
                    char            TM_FAR *bufferPtr,
                    int                     bufferLength,
                    int                     flags,
                    struct sockaddr TM_FAR *fromAddressPtr,
                    int             TM_FAR *addressLengthPtr,
                    struct sockaddr TM_FAR *toAddressPtr,
                    ttUser8BitPtr           fromPhyAddrPtr,
                    ttUserIntPtr            fromPhyAddrLenPtr,
                    ttUserInterfacePtr      interfaceHandlePtr )
{
    ttSocketEntryPtr   socketEntryPtr;
    ttPacketPtr        packetPtr;
    ttPacketPtr        recvdPacketPtr;
    ttPktLen           copyCount;
    ttPktLen           linkCopyCount;
    ttPktLen           recvdPacketLength;
    int                errorCode;
    ttSockAddrPtrUnion fromTempAddr;
    ttSockAddrPtrUnion toTempAddr;
    int                addressLength;

    errorCode = TM_ENOERROR;
    recvdPacketLength = (ttPktLen)0;
    fromTempAddr.sockPtr = fromAddressPtr; 
    toTempAddr.sockPtr = toAddressPtr; 
/* PARAMETER CHECK */
    if (addressLengthPtr != (int TM_FAR *)0)
    {
        addressLength = *addressLengthPtr;
    }
    else
    {
        addressLength = 0;
    }
/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    socketEntryPtr = tfSocketCheckAddrLenLock(  socketDescriptor,
                                                fromAddressPtr,
                                                addressLength,
                                                &errorCode
                                              );
    if ( socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR )
    {
/* errorCode set by tfSocketCheckAndLock */
        goto recvToFinish;
    }
    if (    (bufferPtr == (char TM_FAR *)0)
/* we allow bufferlength between 0 and 65534 on 16-bit architecture */
         || (bufferLength == -1)
         || (bufferLength == 0)
         || ((flags & ~TM_RECVFROM_FLAGS) != 0)
/* If user requested the to address */
         || (   (toAddressPtr != (struct sockaddr TM_FAR *) 0)
/*  and no room for the address */
             && (    (socketEntryPtr->socProtocolFamily == AF_INET)
                  && (addressLength < (int)sizeof(struct sockaddr_in)) )
            )
         || (    (fromPhyAddrPtr != (tt8BitPtr)0)
              && (    (fromPhyAddrLenPtr == (ttIntPtr)0)
/* Only support copying Ethernet MAC address */
                   || (*fromPhyAddrLenPtr < (int)sizeof(ttEnetAddress)) ) )
       )
    {
        errorCode = TM_EINVAL;
        goto recvToFinish;
    }
    
/*Check to see if it is a TCP socket */
    if (socketEntryPtr->socProtocolNumber == IP_PROTOTCP)
    {
        errorCode = TM_EPROTOTYPE;
        goto recvToFinish;
    }
/* Check to see if it is a connected socket */
    if (socketEntryPtr->socFlags & TM_SOCF_CONNECTED)
    {
        errorCode = TM_EISCONN;
        goto recvToFinish;
    }
/* If user did not set any wait/dontwait flag */
    if ((flags & TM_SOCF_MSG_WAIT_FLAGS) ==  0)
    {
/* use the socket default */
        flags |= (socketEntryPtr->socFlags & TM_SOCF_NONBLOCKING);
    }
    packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
/* Check for Data */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        errorCode = socketEntryPtr->socRecvError;
        if (errorCode != TM_ENOERROR)
        {
            socketEntryPtr->socRecvError = 0;
            goto recvToFinish;
        }
        if (flags & TM_SOCF_NONBLOCKING)
        {
            errorCode = TM_EWOULDBLOCK;
            goto recvToFinish;
        }
        else
        {
#ifdef TM_PEND_POST_NEEDED
            errorCode = tm_pend(&socketEntryPtr->socRecvPendEntry,
                                &socketEntryPtr->socLockEntry);

            if (errorCode != TM_ENOERROR)
            {
                socketEntryPtr->socRecvError = 0;
            }
            packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
/* Check for Data */
            if ((errorCode != TM_ENOERROR) || (packetPtr==TM_PACKET_NULL_PTR))

            {
                if (errorCode == TM_ENOERROR)
                {
                    errorCode = TM_EWOULDBLOCK;
                }
                goto recvToFinish;
            }
#else /* TM_PEND_POST_NEEDED */
            socketEntryPtr->socRecvError = 0;
            goto recvToFinish;
#endif /* TM_PEND_POST_NEEDED */
        }
    }
    if (fromAddressPtr != TM_SOCKADDR_NULL_PTR)
    {
/* Save the from address */
        {
            *addressLengthPtr = sizeof(struct sockaddr_in);
            fromTempAddr.sockInPtr->sin_len = sizeof(struct sockaddr_in);
            fromTempAddr.sockInPtr->sin_family=AF_INET;
            fromTempAddr.sockInPtr->sin_port=(unsigned short)
                (packetPtr->pktSharedDataPtr->
                 dataPeerPortNumber);
            tm_ip_copy(packetPtr->pktSrcIpAddress,
                       fromTempAddr.sockInPtr->sin_addr.s_addr);
        }
    }    
    if (toAddressPtr != TM_SOCKADDR_NULL_PTR)
    {
/* Save the to address */
        {
            *addressLengthPtr = sizeof(struct sockaddr_in);
            toTempAddr.sockInPtr->sin_len = sizeof(struct sockaddr_in);
            toTempAddr.sockInPtr->sin_family=AF_INET;
/* Destination port */
            toTempAddr.sockInPtr->sin_port=(unsigned short)
                (packetPtr->pktSharedDataPtr->dataLocalPortNumber);
            tm_assert(tfRecvFromToInterface,
                      (((unsigned short) (socketEntryPtr->socOurLayer4Port)
                        == TM_WILD_PORT)
                       || ((unsigned short) (socketEntryPtr->socOurLayer4Port)
                           == toTempAddr.sockInPtr->sin_port)));
/* Destination IP address extracted from the IP header */
            tm_ip_copy(packetPtr->pktDestIpAddress,
                       toTempAddr.sockInPtr->sin_addr.s_addr);
        }
    }    
    if (fromPhyAddrPtr != (ttUser8BitPtr)0)
/* Copy physical address */
    {
        if (tm_4_ll_is_lan(packetPtr->pktDeviceEntryPtr))
        {
            *fromPhyAddrLenPtr = (int)sizeof(ttEnetAddress);
            tm_bcopy((tt8BitPtr)packetPtr->pktEthernetAddress,
                     fromPhyAddrPtr, sizeof(ttEnetAddress));
        }
        else
        {
            *fromPhyAddrLenPtr = 0;
        }
    }
    if (interfaceHandlePtr != (ttUserInterfacePtr)0)
/* copy interface handle */
    {
        *interfaceHandlePtr = (ttUserInterface)
                            packetPtr->pktDeviceEntryPtr;
    }
/* Most common case. User has enough room in its buffer. */
    copyCount = packetPtr->pktChainDataLength;
    recvdPacketPtr = packetPtr;
    recvdPacketLength = copyCount;
    if (copyCount > (ttPktLen)bufferLength)
    {
/* Return truncated packet */
        errorCode = TM_EMSGSIZE;
/* Copy only up to user buffer length */
        copyCount = (ttPktLen)bufferLength;
    }
    do
    {
         if (copyCount < packetPtr->pktLinkDataLength)
         {
/* Copy only up to user buffer length */
            linkCopyCount = copyCount;
         }
         else
         {
            linkCopyCount = packetPtr->pktLinkDataLength;
         }
/* Copy this data area to the user buffer */
        {
            tm_bcopy((ttConstVoidPtr)(packetPtr->pktLinkDataPtr),
                     (ttVoidPtr)bufferPtr,
                     (unsigned)tm_packed_byte_count(
                                                linkCopyCount));
            bufferPtr += tm_packed_byte_count(linkCopyCount);
        }
/* 
 * For Datagram protocols, Keep the datagram boundary 
 */
        packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        copyCount -= linkCopyCount;
    } while (    (packetPtr != TM_PACKET_NULL_PTR) 
              && (copyCount != (ttPktLen)0) );
/* 
 * Since we copied the data area, it if the user is not peeking, unqueue
 * and free the packet
 */
    if ((flags & MSG_PEEK) == 0)
    {
/* Update receive queue */
        socketEntryPtr->socReceiveQueueNextPtr =
                                       recvdPacketPtr->pktChainNextPtr;
/* Update receive queue size */
        socketEntryPtr->socRecvQueueBytes -= (tt32Bit)recvdPacketLength;
        socketEntryPtr->socRecvQueueDgrams--;
/* Free copied packet */
        tfFreePacket(recvdPacketPtr, TM_SOCKET_LOCKED);
    }

recvToFinish:
/*
 * If errorCode not set, will return recvdPacketLength, otherwise it will set
 * the errorCode in the user error array, and return TM_SOCKET_ERROR instead
 */
    return tfSocketReturn(socketEntryPtr, socketDescriptor,
                          errorCode, (int)recvdPacketLength);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trrcvtif.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 65206    Output checksum: 24582
 */
