/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (recvfrom)
 *
 * Filename: trrecvfr.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trrecvfr.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:05CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * Take some data from the receive queue and give it to the user 
 */
int recvfrom(int                     socketDescriptor, 
             char            TM_FAR *bufferPtr, 
             int                     bufferLength, 
             int                     flags,
             struct sockaddr TM_FAR *fromAddressPtr,
             int             TM_FAR *fromAddressLengthPtr)
{
    ttSocketEntryPtr   socketEntryPtr;
    ttPacketPtr        packetPtr;
    ttPktLen           copyCount;
    ttPktLen           recvdPacketLength;
    ttPktLen           linkCopyCount;
    int                errorCode;
    ttPacketPtr        recvdPacketPtr;
    ttSockAddrPtrUnion tempSockAddr;
    int                addressLength;
    int                retCode;

    tempSockAddr.sockPtr = fromAddressPtr; 

    errorCode = TM_ENOERROR;
    recvdPacketLength = (ttPktLen)0;
/* PARAMETER CHECK */
    if (fromAddressLengthPtr != (int TM_FAR *)0)
    {
        addressLength = *fromAddressLengthPtr;
    }
    else
    {
        addressLength = 0;
    }
/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    socketEntryPtr = tfSocketCheckAddrLenLock(  socketDescriptor,
                                                fromAddressPtr,
                                                addressLength,
                                                &errorCode
                                              );
    {
        if ( socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR )
        {
/* errorCode set by tfSocketCheckAndLock */
            goto recvfromFinish;
        }
        if (    (bufferPtr == (char TM_FAR *)0)
/* we allow bufferlength between 0 and 65534 on 16-bit architecture */
             || (bufferLength == -1)
             || (bufferLength == 0)
             || ((flags & ~TM_RECVFROM_FLAGS) != 0) )
        {
            errorCode = TM_EINVAL;
            goto recvfromFinish;
        }
        
/*Check to see if it is a TCP socket */
        if (socketEntryPtr->socProtocolNumber == IP_PROTOTCP)
        {
            errorCode = TM_EPROTOTYPE;
            goto recvfromFinish;
        }
/* If user did not set any wait/dontwait flag */
        if ((flags & TM_SOCF_MSG_WAIT_FLAGS) ==  0)
        {
/* use the socket default */
            flags |= (socketEntryPtr->socFlags & TM_SOCF_NONBLOCKING);
        }
        packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
/* Check for Data */
        if (packetPtr == TM_PACKET_NULL_PTR)
        {
            errorCode = socketEntryPtr->socRecvError;
            if (errorCode != TM_ENOERROR)
            {
                socketEntryPtr->socRecvError = 0;
                goto recvfromFinish;
            }
            if (flags & TM_SOCF_NONBLOCKING)
            {
                errorCode = TM_EWOULDBLOCK;
                goto recvfromFinish;
            }
            else
            {
#ifdef TM_PEND_POST_NEEDED
                errorCode = tm_pend(&socketEntryPtr->socRecvPendEntry,
                                    &socketEntryPtr->socLockEntry);
                if (errorCode != TM_ENOERROR)
                {
                    socketEntryPtr->socRecvError = 0;
                }
                packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
/* Check for Data */
                if (    (errorCode != TM_ENOERROR) 
                     || (packetPtr==TM_PACKET_NULL_PTR) )
                {
                    if (errorCode == TM_ENOERROR)
                    {
                        errorCode = TM_EWOULDBLOCK;
                    }
                    goto recvfromFinish;
                }
#else /* TM_PEND_POST_NEEDED */
                socketEntryPtr->socRecvError = 0;
                goto recvfromFinish;
#endif /* TM_PEND_POST_NEEDED */
            }
        }
        if (fromAddressPtr != TM_SOCKADDR_NULL_PTR)
        {
/* Save the from address */
            {
                *fromAddressLengthPtr = sizeof(struct sockaddr_in);
                tempSockAddr.sockInPtr->sin_len = sizeof(struct sockaddr_in);
                tempSockAddr.sockInPtr->sin_family = AF_INET;
                tempSockAddr.sockInPtr->sin_port = (unsigned short)
                    (packetPtr->pktSharedDataPtr->dataPeerPortNumber);
                tm_ip_copy(packetPtr->pktSrcIpAddress,
                           tempSockAddr.sockInPtr->sin_addr.s_addr);
            }
            
        }    
/* Most common case. User has enough room in its buffer. */
        copyCount = packetPtr->pktChainDataLength;
        recvdPacketPtr = packetPtr;
        recvdPacketLength = copyCount;
        if (copyCount > (ttPktLen)bufferLength)
        {
/* Return truncated packet */
            errorCode = TM_EMSGSIZE;
/* Copy only up to user buffer length */
            copyCount = (ttPktLen)bufferLength;
        }
        do
        {
             if (copyCount < packetPtr->pktLinkDataLength)
             {
/* Copy only up to user buffer length */
                linkCopyCount = copyCount;
             }
             else
             {
                linkCopyCount = packetPtr->pktLinkDataLength;
             }
/* Copy this data area to the user buffer */
            {
                tm_bcopy((ttConstVoidPtr)(packetPtr->pktLinkDataPtr),
                         (ttVoidPtr)bufferPtr,
                         (unsigned)tm_packed_byte_count(
                                                linkCopyCount));
                bufferPtr += tm_packed_byte_count(linkCopyCount);
            }
/* 
 * For Datagram protocols, Keep the datagram boundary 
 */
            copyCount -= linkCopyCount;
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
        } while (    (packetPtr != TM_PACKET_NULL_PTR) 
                  && (copyCount != (ttPktLen)0) );
/* 
 * Since we copied the data area, it if the user is not peeking, unqueue
 * and free the packet
 */
        if ((flags & MSG_PEEK) == 0)
        {
/* Update receive queue */
            socketEntryPtr->socReceiveQueueNextPtr =
                                           recvdPacketPtr->pktChainNextPtr;
/* Update receive queue size */
            socketEntryPtr->socRecvQueueBytes -= (tt32Bit)recvdPacketLength;
            socketEntryPtr->socRecvQueueDgrams--;
/* Free copied packet */
            tfFreePacket(recvdPacketPtr, TM_SOCKET_LOCKED);
        }

recvfromFinish:
/*
 * If errorCode not set, will return recvdPacketLength, otherwise it will set
 * the errorCode in the user error array, and return TM_SOCKET_ERROR instead
 */
        retCode = tfSocketReturn(socketEntryPtr, socketDescriptor,
                                 errorCode, (int)recvdPacketLength);
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trrecvfr.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  1527    Output checksum: 44780
 */
