/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     Common functions used by tfIpScatteredSend(), and
 *                  tfSocketScatteredSend().
 * Filename:        trscatco.c
 * Author:          Odile
 * Date Created:    07/25/00
 * $Source: source/sockapi/trscatco.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:10CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>


/*
 * tfLinkScatDataToHeader function description
 * Function called either by tfIpScatteredSend() or tfSocScatteredSend()
 * to allocate an empty ttPacket/ttSharedData per user data block
 * in the user data block array, and point it to the corresponding
 * user buffer. Then link all the ttPacket/ttSharedData together after the
 * headPacketPtr (pointing to an allocated ttPacket/ttSharedData
 * containing (or having room) for the protocol headers).
 * Parameters
 * headPacketPtr        Pointer to first ttPacket/ttSharedData
 *                      for protocol headers storage
 * userBlockPtr         Array of data user blocks
 * userBlockCount       number of elements in the above array
 * userFreeFunction     Function to be called by the Treck stack
 *                      to let the user free a data block in the array
 * Return
 * TM_ENOERROR          success
 * TM_ENOBUFS           Could not allocate an empty ttPacket/ttSharedData
 */
int tfLinkScatDataToHeader ( ttPacketPtrPtr     headPacketPtrPtr,
                             ttUserBlockPtr     userBlockPtr,
                             int                userBlockCount,
                             ttUserFreeFuncPtr  userFreeFunction,
                             int                ipHdrIncl )
{
    ttPacketPtr     headPacketPtr;
    ttPacketPtr     packetPtr;
    ttPacketPtr     prevPacketPtr;
    ttIpHeaderPtr   iphPtr;
    ttCharPtr       pktDataPtr;
    tt16Bit         ipHdrLength;    /* IP header length */
    ttPktLen        bufferLength;
    int             copyLength;
    int             errorCode;

    headPacketPtr = *headPacketPtrPtr;
    if (ipHdrIncl)
    {
/* Point to the IP header in user data */
        iphPtr = (ttIpHeaderPtr)userBlockPtr->userDataPtr;
/* Figure out the IP header length (in bytes) */
        ipHdrLength = tm_ip_hdr_len(iphPtr->iphVersLen);
/* Check IP header length */
        if (    (ipHdrLength > TM_4_IP_MAX_HDR_LEN)
             || (ipHdrLength < TM_4_IP_MIN_HDR_LEN) )
        {
/*
 * If IP header is bigger, or smaller than allowed by RFC's, free the
 * user buffers, and return TM_EFAULT
 */
            errorCode = TM_EFAULT;
            goto hdrScatExit;
        }
/*
 * Allocate ttPacket/ttSharedData buffer to store the link layer header
 * plus the IP header. This will be the first ttPacket in the link.
 * Note that pktChainDataLength, and pktLinkDataLength will both be
 * initialized with ipHdrLength.
 */
        headPacketPtr = tfGetSharedBuffer(
            TM_LL_MAX_HDR_LEN, (ttPktLen)ipHdrLength, TM_16BIT_ZERO);
        if (headPacketPtr == (ttPacketPtr)0)
        {
/* If failed to allocate, free the user buffers, and return TM_ENOBUFS. */
            errorCode = TM_ENOBUFS;
            goto hdrScatExit;
        }
/*
 * Initialize pktIpHdrLen needed in case IP datagram need to be fragmented
 * (for tfIpScatteredSend()
 */
        headPacketPtr->pktIpHdrLen = (tt8Bit)tm_packed_byte_count(ipHdrLength);
/*
 * Copy the IP header in the newly allocated buffer, leaving room for the
 * link layer header. This will resolve three possible problems:
 *  1. non aligned IP header
 *  2. IP header spanning more than one buffer.
 *  3. If data need to be fragmented our IP fragmentation code assumes that
 *     the link layer/IP header is contiguous.
 * and will allow us to use as much written code as possible.
 * In a loop, until we reach the end of the IP header or the end of the
 * user buffer:
 * .. Check on validity of the user buffer element fields.
 * .. If IP header is strictly contained in a user buffer, update the relevant
 *    data pointers in the user block array.
 * .. else the IP header spans extra buffer(s), or IP header ends at the
 *    end of the user buffer, free the user buffer, and update
 *    userBlockPtr, and userBlockCount accordingly.
 * .. If we have finished copying the IP header, break out of the loop.
 */
/* Destination data pointer in first ttPacket/ttShared data area */
        pktDataPtr = (ttCharPtr)headPacketPtr->pktLinkDataPtr;
        do
        {
/* Check on fields in element of the array */
            if (    (userBlockPtr->userDataPtr == (ttCharPtr)0)
                 || (userBlockPtr->userBufferPtr == (ttCharPtr)0)
                 || (userBlockPtr->userDataLength <= 0) )
            {
/*
 * If invalid field(s) in an element of the array, free the user buffers,
 * and return TM_EINVAL.
 */
                errorCode = TM_EINVAL;
                goto hdrScatExit;
            }
            copyLength = (int)ipHdrLength;
            if (copyLength > userBlockPtr->userDataLength)
            {
/* Do not copy more than data contained in the current buffer */
                copyLength = userBlockPtr->userDataLength;
            }
            tm_bcopy( userBlockPtr->userDataPtr,
                      pktDataPtr,
                      (unsigned)copyLength );
            if ((int)ipHdrLength < userBlockPtr->userDataLength)
/* IP header strictly contained inside this user data block */
            {
/*
 * Update the relevant data pointers in the user block array.
 * Done with copying the IP header.
 */
                userBlockPtr->userDataPtr += tm_packed_byte_count(copyLength);
                userBlockPtr->userDataLength -= ipHdrLength;
            }
            else
/*
 * IP header spanning more than one buffer, or ends at the end of this
 * buffer.
 */
            {
                (void)(*userFreeFunction)(userBlockPtr->userBufferPtr);
                userBlockPtr++;
                userBlockCount--;
/* Update pointer to copy IP header into */
                pktDataPtr += tm_packed_byte_count(copyLength);
                if ( ((int)ipHdrLength != copyLength) && (userBlockCount == 0) )
                {
/* Missing end of IP header !! */
                    errorCode = TM_EFAULT;
                    goto hdrScatExit;
                }
            }
            ipHdrLength = (tt16Bit)(ipHdrLength - copyLength);
        } while (ipHdrLength != 0);
    }
    else
    {
/*
 * Allocate ttPacket/ttSharedData buffer to store
 * the link layer header plus the IP header, plus the UDP, or ICMP header.
 * This will be the first ttPacket in the link for a non TCP socket.
 * Note that pktChainDataLength, and pktLinkDataLength will both be zero.
 */
        headPacketPtr = tfGetSharedBuffer( TM_MAX_SEND_HEADERS_SIZE,
                                           (ttPktLen)0,
                                           TM_16BIT_ZERO );
        if (headPacketPtr == (ttPacketPtr)0)
        {
/* If failed to allocate, free the user buffers, and return TM_ENOBUFS. */
            errorCode = TM_ENOBUFS;
            goto hdrScatExit;
        }
    }
/*
 * Initialize pktLinkExtraCount field of the first ttPacket with
 * userBlockCount.
 */
    headPacketPtr->pktLinkExtraCount = userBlockCount;
    prevPacketPtr = headPacketPtr;
/*
 * Initialize an automatic variable bufferLength to zero. It will
 * be used to accumulate the user data length.
 */
    bufferLength = (ttPktLen)0;
    errorCode = TM_ENOERROR;
/*
 * . For each buffer element in the array:
 *  .. Allocate a ttPacket structure, and and empty ttSharedData, and
 *     set the TM_BUF_USER_DATA bit in the ttSharedData flag as follows:
 *     newPacketPtr = tfGetSharedBuffer(0, (ttPktLen)0, TM_BUF_USER_DATA);
 *  .. If failed to allocate, set errorCode to TM_ENOBUFS, and break out of
 *     the loop.
 *  .. Store the userFreeFunction in the last union of the ttSharedData
 *     structure
 *  .. Store the user buffer pointer in the ttSharedData structure
 *     (in the dataDeviceUserDataPtr field)
 *  .. Make the ttPacket link data pointer (pktLinkDataPtr) point the the user
 *     data pointer
 *  .. Initialize pktLinkDataLength, and pktChainDataLength with the user data
 *     length
 *  .. Add the user data length to 'bufferLength'
 *  .. Link the newly allocated ttPacket to the previous one.
 *  .. Point to next element in the array, update the array count.
 */
    while (userBlockCount != 0)
/*
 * while() instead of do while, since could be zero if called from
 * tfIpScatteredSend()
 */
    {
/* Check on fields in element of the array */
        if (    (userBlockPtr->userDataPtr == (ttCharPtr)0)
             || (userBlockPtr->userBufferPtr == (ttCharPtr)0)
             || (userBlockPtr->userDataLength <= 0) )
        {
/*
 * If invalid field(s) in an element of the array, free the user buffers,
 * and return TM_EINVAL.
 */
            errorCode = TM_EINVAL;
            goto hdrScatExit;
        }
        packetPtr = tfGetSharedBuffer(0, (ttPktLen)0, TM_BUF_USER_DATA);
        if (packetPtr == (ttPacketPtr)0)
        {
            errorCode = TM_ENOBUFS;
            goto hdrScatExit;
        }
        packetPtr->pktSharedDataPtr->dataUserFreeFuncPtr = userFreeFunction;
        packetPtr->pktSharedDataPtr->dataDeviceUserDataPtr =
                                                 userBlockPtr->userBufferPtr;
        packetPtr->pktLinkDataPtr = (tt8BitPtr)userBlockPtr->userDataPtr;
        packetPtr->pktSharedDataPtr->dataBufFirstPtr = 
                                                     packetPtr->pktLinkDataPtr;
        packetPtr->pktSharedDataPtr->dataBufLastPtr = 
            packetPtr->pktLinkDataPtr +
            tm_packed_byte_count((unsigned)(userBlockPtr->userDataLength));
        packetPtr->pktLinkDataLength = 
                                    (ttPktLen)(userBlockPtr->userDataLength);
        packetPtr->pktChainDataLength = 
                                    (ttPktLen)(userBlockPtr->userDataLength);
        bufferLength = (ttPktLen)(   bufferLength
                                  + (ttPktLen)userBlockPtr->userDataLength );
        prevPacketPtr->pktLinkNextPtr = (ttVoidPtr)packetPtr;
        userBlockPtr++;
        userBlockCount--;
        prevPacketPtr = packetPtr;
    }
/* Add bufferLength to the pktChainDataLength field of the first ttPacket */
    headPacketPtr->pktChainDataLength += bufferLength;
hdrScatExit:
    if (errorCode != TM_ENOERROR)
    {
        tfFreeScatData( headPacketPtr, userBlockPtr, userBlockCount,
                        userFreeFunction);
    }
    else
    {
        *headPacketPtrPtr = headPacketPtr;
    }
    return errorCode;
}

/*
 * tfFreeScatData function description
 * Called if an error occured in tfIpScatteredSend() or tfSocketSatteredSend()
 * . call tfFreePacket() to free the allocated ttPacket/ttSharedData, and
 *   associated user data.
 * . walk down the user blocks array, and free any user data buffers that
 *   had not been assigned to a ttPacket/ttSharedData.
 * Parameters
 * headPacketPtr        Pointer to link list of ttPacket/ttSharedData
 *                      possibly pointing to data user blocks
 * userBlockPtr         Array of data user blocks that have not yet
 *                      been pointed at by ttPacket/ttSharedData
 * userBlockCount       number of elements in the above array
 * userFreeFunction     Function to be called by the Treck stack
 *                      to let the user free a data block in the array
 * Return
 * No return value
 */
void tfFreeScatData ( ttPacketPtr        headPacketPtr,
                      ttUserBlockPtr     userBlockPtr,
                      int                userBlockCount,
                      ttUserFreeFuncPtr  userFreeFunction )
{
/*
 * tfFreePacket will free all the ttPacket/ttShared data linked to
 * headPacketPtr, and the associated user buffers.
 * tfFreePacket checks for null pointer.
 */
    tfFreePacket(headPacketPtr, TM_SOCKET_UNLOCKED);
/* Free the user buffers that have not been associated with a ttPacket */
    while ( (userBlockCount != 0) && (userBlockPtr != (ttUserBlockPtr)0) )
    {
        if (userBlockPtr->userBufferPtr != (ttCharPtr)0)
        {
            (void)(*userFreeFunction)(userBlockPtr->userBufferPtr);
        }
        userBlockCount--;
        userBlockPtr++;
    }
}
 

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trscatco.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 31337    Output checksum:   523
 */
