/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2009
 * Copyright Elmic Systems Japan 1997 - 2009
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     Device Driver scattered recv function APIs
 * Filename:        trscatdr.c
 * Author:          Odile
 * Date Created:    06/10/2002
 * $Source: source/sockapi/trscatdr.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:11CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>

#ifdef TM_USE_DRV_SCAT_RECV
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
#define TESTING
#define TESTING_TRIM
*/

static void tfFreeDrvScatRecvBuffer ( ttDeviceEntryPtr devEntryPtr,
                                      ttUserBufferPtr  bufPtr, 
                                      int              bufFlag );

/*
 * tfUseInterfaceScatRecv Function Description
 * Specify a driver driver scattered recv function. This enables the device
 * driver to use scattered recv buffers.
 * User need to call tfUseInterfaceScatRecv between the 
 * tfAddInterface, and tfOpenInterface/tfConfigInterface calls.
 * 1. Check that the interface is a valid added interface
 * 2. Check that the Recv Function is non null
 * 3. Add the recv function on the interface.
 * Parameters
 * interfaceHandle
 * devScatRecvFuncPtr
 * Returns
 * TM_ENOERROR  Success
 * TM_EINVAL    error in Parameter.
 */
int tfUseInterfaceScatRecv (
                ttUserInterface      interfaceHandle,
                ttDevScatRecvFuncPtr drvScatRecvFuncPtr )
{
    ttDeviceEntryPtr devEntryPtr;
    int              errorCode;

    errorCode = TM_ENOERROR;
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check parameter */
    errorCode = tfValidInterface(devEntryPtr);
    if ( errorCode == TM_ENOERROR )
    {
        if ( tm_16bit_one_bit_set(devEntryPtr->devFlag, TM_DEV_OPENED) )
        {
/* 
 * Cannot set after device has been opened.
 */
            errorCode = TM_EPERM;
        }
        else
        {
            if (drvScatRecvFuncPtr == (ttDevScatRecvFuncPtr)0)
            {
                errorCode = TM_EINVAL;
            }
            else
            {
                devEntryPtr->devScatRecvFuncPtr = drvScatRecvFuncPtr;
            }
        }
    }
    return (errorCode);
}

/*
 * If the user has called tfUseInterfaceScatRecv:
 * When the user has been notified that an incoming packet has arrived (via
 * either tfCheckReceiveInterface(), or tfWaitReceiveInterface()), it
 * then calls tfRecvScatInterface() so that the stack can call the
 * driver scattered receive function, and retrieve the data
 * (If the user has not called tfUseInterfaceScatRecv, then tfRecvInterface()
 *  should be used instead. See below.)
 */
int tfRecvScatInterface (ttUserInterface interfaceHandle)
{
    ttDruBlockPtr    blockPtr;
    ttPacketPtr      packetPtr;
    ttPacketPtr      newPacketPtr;
    ttPacketPtr      prevPacketPtr;
    ttDeviceEntryPtr devEntryPtr;
    tt8BitPtr        copyDataPtr;
#ifdef TM_DEV_RECV_OFFLOAD
    ttDevRecvOffloadStruct devRecvOffload;
#endif /* TM_DEV_RECV_OFFLOAD */
    ttPktLen         dataLength;
    ttPktLen         packedCopyBlockLength;
    int              blockLength;
    int              copyBlockLength;
    int              maxCopyLength;
    int              copyBlockCount;
    int              i;
    int              errorCode;
    int              retCode;
    int              blockCount;
    int              blockBufFlag;
    int              forwardSize;
#ifdef TESTING
    int              testSplitHeader;
#endif /* TESTING */
    
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    errorCode = TM_ENOERROR;
    retCode = TM_ENOERROR;
    blockCount = 0; /* In case the user does not set it */
    blockBufFlag = 0; /* In case the user does not set it */
    blockPtr = (ttDruBlockPtr)0; /* In case the user does not set it */
    packetPtr = (ttPacketPtr)0;
    prevPacketPtr = (ttPacketPtr)0;
    dataLength = (ttPktLen)0;
    i = 0;
    blockLength = 0;
#ifdef TM_DEV_RECV_OFFLOAD
/* moved here for compiler warning */
    devRecvOffload.devoRFlags = TM_16BIT_ZERO; 
#endif /* TM_DEV_RECV_OFFLOAD */
/* Call the drivers receive routine */
/* Lock the send/receive to provide single threaded access to the recv */
    tm_lock_wait(&(devEntryPtr->devDriverLockEntry));
    if ( tm_16bit_one_bit_set( devEntryPtr->devFlag, TM_DEV_OPENED) )
    {
        retCode = (*(devEntryPtr->devScatRecvFuncPtr))
                        ( interfaceHandle, 
                          &blockPtr, (int TM_FAR *)&blockCount, 
                          &blockBufFlag 
#ifdef TM_DEV_RECV_OFFLOAD
                        , &devRecvOffload
#endif /* TM_DEV_RECV_OFFLOAD */
                          );
    }
    else
    {
        retCode = TM_ENXIO;
    }
/* Unlock the send/receive */
    tm_unlock(&(devEntryPtr->devDriverLockEntry));
    if (retCode != TM_ENOERROR)
    {
        goto recvScatIntfFinish;
    }
#ifdef TM_ERROR_CHECKING
/* Check on user device values */
    if (    ( blockPtr == (ttDruBlockPtr)0 )
         || ( blockCount == 0 ) 
         || ( (blockBufFlag & (TM_DEV_SCAT_RECV_FLAGS)) == 0) )
    {
        errorCode = TM_EINVAL;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
        devEntryPtr->devInErrors++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
        goto recvScatIntfFinish;
    }
#endif /* TM_ERROR_CHECKING */
 

    maxCopyLength = (int)devEntryPtr->devRecvContiguousHdrLength;
    if ( (blockCount > 1) && (blockPtr[0].druDataLength < maxCopyLength) )
/* scattered data with first block data size less than our contiguous length */
    {
        if (tm_ll_need_32bit_align(devEntryPtr))
        {
/* 
 * We will make the Ethernet header start at a 2 bytes offset, and 
 * therefore guarantee that the IP header/ARP header are on a 4-byte boundary
 * profided that tfKernelMalloc() returns pointers on a 4-byte boundary.
 */ 
            forwardSize = TM_LL_MAX_HDR_LEN - TM_ETHER_HEADER_SIZE;
        }
        else
        {
            forwardSize = TM_LL_MAX_HDR_LEN;
        }
        packetPtr = tfGetSharedBuffer( forwardSize,
                                       (ttPktLen)maxCopyLength,
                                       TM_16BIT_ZERO );
        if ( packetPtr == TM_PACKET_NULL_PTR)
        {
            errorCode = TM_ENOBUFS;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            devEntryPtr->devInDiscards++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            goto recvScatIntfFinish;
        }
        packetPtr->pktSharedDataPtr->dataDeviceEntryPtr = devEntryPtr;
        copyDataPtr = packetPtr->pktLinkDataPtr;
        prevPacketPtr = packetPtr;
        while (    ((int)dataLength < maxCopyLength )
                && (i < blockCount)
              )
        {
            blockLength = blockPtr[i].druDataLength;
#ifdef TM_ERROR_CHECKING
/* Check on user device values */
            if (   ( blockLength <= 0 )
                || ( blockPtr[i].druStackBufferPtr == (ttUserBufferPtr)0 )
                || ( blockPtr[i].druDataPtr == (ttCharPtr)0 ) )
            {
                errorCode = TM_EINVAL;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
                devEntryPtr->devInErrors++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
                goto recvScatIntfFinish;
            }
#endif /* TM_ERROR_CHECKING */
            copyBlockLength = maxCopyLength - (int)dataLength;
            if (blockPtr[i].druDataLength < copyBlockLength)
            {
                copyBlockLength = blockLength;
            }
            packedCopyBlockLength = 
                             tm_packed_byte_count((unsigned)copyBlockLength);
            tm_bcopy( blockPtr[i].druDataPtr, 
                      copyDataPtr,
                      packedCopyBlockLength);
            dataLength = (ttPktLen)(dataLength + (ttPktLen)copyBlockLength);
            if (copyBlockLength == blockLength)
            {
                tfFreeDrvScatRecvBuffer( devEntryPtr,
                                         blockPtr[i].druStackBufferPtr, 
                                         blockBufFlag );
                copyDataPtr += packedCopyBlockLength;
                i++;
            }
            else
            {
                blockPtr[i].druDataLength -= copyBlockLength;
                blockPtr[i].druDataPtr += packedCopyBlockLength;
                break;
            }
        }
        packetPtr->pktLinkDataLength = dataLength;
        copyBlockCount = i;
    }
    else
    {
        copyBlockCount = 1;
    }
    while ( i < blockCount )
    {
        blockLength = blockPtr[i].druDataLength;
#ifdef TM_ERROR_CHECKING
/* Check on user device values */
        if (    ( blockLength <= 0 )
             || ( blockPtr[i].druStackBufferPtr == (ttUserBufferPtr)0)
             || ( blockPtr[i].druDataPtr == (ttCharPtr)0 ) )
        {
            errorCode = TM_EINVAL;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            devEntryPtr->devInErrors++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            goto recvScatIntfFinish;
        }
#endif /* TM_ERROR_CHECKING */
        if ( blockBufFlag & TM_DEV_SCAT_RECV_USER_BUFFER )
        {
/* User owned buffer; create a packet/shared data area Zero Size*/
            newPacketPtr = tfGetSharedBuffer( 0,
                                              (ttPktLen)0,
                                              TM_BUF_USER_DEVICE_DATA );
            if ( newPacketPtr == TM_PACKET_NULL_PTR)
            {
                errorCode = TM_ENOBUFS;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
                devEntryPtr->devInDiscards++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
                goto recvScatIntfFinish;
            }
/* Used by tfFreeSharedBuffer() */
            newPacketPtr->pktSharedDataPtr->dataDeviceEntryPtr = devEntryPtr;
/* Used by tfFreeSharedBuffer() */
            newPacketPtr->pktSharedDataPtr->dataDeviceUserDataPtr =
                                                    blockPtr[i].druBufferPtr;
            newPacketPtr->pktSharedDataPtr->dataBufFirstPtr = (tt8BitPtr)
                                                    blockPtr[i].druDataPtr;
            newPacketPtr->pktSharedDataPtr->dataBufLastPtr = 
                                 (tt8BitPtr)(blockPtr[i].druDataPtr)
                               + tm_packed_byte_count((unsigned)blockLength);
        }
        else
        {
            newPacketPtr = (ttPacketPtr)(blockPtr[i].druStackBufferPtr);
            newPacketPtr->pktUserFlags = 0; /* User no longer owns the packet */
        }
        if (i == 0)
        {
/* first packet */
            packetPtr = newPacketPtr;
#ifdef TM_USE_USER_PARAM
/*
 * Copy the user-defined data to the packet. This data may have been set by
 * the device driver's scattered receive function. If this packet results in
 * the creation of a new TCP socket, the user data will be copied from the
 * packet to the TCP socket. The data can then be retrieved by the
 * application using the TM_TCP_USER_PARAM socket option after a successful
 * return from accept().
 */
            tm_bcopy(   &blockPtr[i].druUserParam,
                        &packetPtr->pktUserStruct.pktuUserParam,
                        sizeof(ttUserGenericUnion) );
#endif /* TM_USE_USER_PARAM */
        }
        else
        {
/*
 * Used by HTTPD to find out the MTU. Link layer recv routine copy it only on
 * first packet
 */
            newPacketPtr->pktDeviceEntryPtr = devEntryPtr;
        }
        newPacketPtr->pktLinkDataLength = (ttPktLen)blockLength;
        newPacketPtr->pktLinkDataPtr = (tt8BitPtr)(blockPtr[i].druDataPtr);
        if (prevPacketPtr != (ttPacketPtr)0)
        {
            prevPacketPtr->pktLinkNextPtr = (ttVoidPtr)newPacketPtr;
        }
        prevPacketPtr = newPacketPtr;
        dataLength = (ttPktLen)(dataLength + ((ttPktLen)blockLength));
        i++;
    }
#if (defined(TM_SNMP_MIB) && defined(TM_USE_IPV4))
    tm_64Bit_augm(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInOctets,
                  (tt32Bit)dataLength);
#endif /* TM_SNMP_MIB && TM_USE_IPV4 */
    packetPtr->pktChainDataLength = dataLength;
    packetPtr->pktLinkExtraCount = blockCount - copyBlockCount;
#ifdef TM_ERROR_CHECKING
    newPacketPtr = packetPtr;
    i = 0;
    while (newPacketPtr && (newPacketPtr->pktLinkNextPtr != (ttVoidPtr)0))
    {
        newPacketPtr = (ttPacketPtr)newPacketPtr->pktLinkNextPtr;
        i++;
    }
    tm_assert(tfRecvScatInterface, i == packetPtr->pktLinkExtraCount);
#ifdef TESTING
/* DEBUG */
#ifdef TESTING_TRIM 
/* Testing trimming incoming scattered data */
    if (i > 0)
    {
        packetPtr->pktChainDataLength += 4;
        newPacketPtr->pktLinkDataLength += 4;
    }
#endif /* TESTING_TRIM */
/* Testing split header. Try different slit sizes */
/* Split header after 14 bytes, i.e. Ethernet header */
    testSplitHeader = packetPtr->pktLinkDataLength - 14;
/* Split header after 1 byte */
    testSplitHeader = packetPtr->pktLinkDataLength - 1;
    if (testSplitHeader > 0)
    {
        newPacketPtr = tfGetSharedBuffer( 0,
                                          (ttPktLen)testSplitHeader,
                                          TM_16BIT_ZERO );
        if (newPacketPtr != (ttPacketPtr)0)
        {
            newPacketPtr->pktLinkNextPtr = packetPtr->pktLinkNextPtr;
            packetPtr->pktLinkNextPtr = (ttVoidPtr)newPacketPtr;
            packetPtr->pktLinkDataLength -= testSplitHeader;
            tm_bcopy(    packetPtr->pktLinkDataPtr 
                       + tm_packed_byte_count(packetPtr->pktLinkDataLength),
                       newPacketPtr->pktLinkDataPtr,
                       tm_packed_byte_count((unsigned)testSplitHeader) );
            packetPtr->pktLinkExtraCount++;
        }
    }
#endif /* TESTING */
#endif /* TM_ERROR_CHECKING */
/* Call the drivers link layer recv routine */
#ifdef TM_DEV_RECV_OFFLOAD
    if (devRecvOffload.devoRFlags != TM_16BIT_ZERO)
    {
        packetPtr->pktDevOffloadPtr = (ttVoidPtr)
                            &(packetPtr->pktSharedDataPtr->dataDevRecvOffload);
        tm_bcopy((ttVoidPtr)&devRecvOffload, packetPtr->pktDevOffloadPtr,
                  sizeof(devRecvOffload));
    }
#endif /* TM_DEV_RECV_OFFLOAD */
    retCode = (*(devEntryPtr->devLinkLayerProtocolPtr->lnkRecvFuncPtr))(
                                               (void TM_FAR *)interfaceHandle,
                                               (void TM_FAR *)packetPtr );

recvScatIntfFinish:
    if (errorCode != TM_ENOERROR)
/*
 * User set invalid parameters, we could not allocate a shared buffer 
 * to point to the user data
 */
    {
/* Free data that was already queued if any */
        tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
        if ((i < blockCount) && (blockPtr != (ttDruBlockPtr)0))
        {
            while (i < blockCount)
            {
                tfFreeDrvScatRecvBuffer( devEntryPtr, 
                                         blockPtr[i].druStackBufferPtr, 
                                         blockBufFlag );
                i++;
            }
        }
        retCode = errorCode;
    }
    return retCode;
}

static void tfFreeDrvScatRecvBuffer ( ttDeviceEntryPtr devEntryPtr,
                                      ttUserBufferPtr  bufPtr, 
                                      int              blockBufFlag )
{
    if (bufPtr != (ttUserBufferPtr)0)
    {
        if ( blockBufFlag & TM_DEV_SCAT_RECV_USER_BUFFER )
        {
            if (devEntryPtr->devFreeRecvFuncPtr != (ttDevFreeRecvFuncPtr)0)
            {
/*
 * User owns the buffer. Free user buffer. Make the user device driver
 * free function re-entrant by setting the device driver lock
 */
                tm_lock_wait(&(devEntryPtr->devDriverLockEntry));
                (void)(*(devEntryPtr->devFreeRecvFuncPtr))( devEntryPtr, 
                                                    (ttCharPtr)bufPtr );
                tm_unlock(&(devEntryPtr->devDriverLockEntry));
            }
        }
        else
/* Treck owns the buffer (if any) */
        {
/* tfFreeDriverBuffer checks on zero pointer, and user ownership */
            (void)tfFreeDriverBuffer((ttPacketPtr)bufPtr);
        }
    }
}

#else /* !TM_USE_DRV_SCAT_RECV */


#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_USE_DRV_SCAT_RECV is not defined */
int tvScatDrDummy = 0;

#endif /* TM_USE_DRV_SCAT_RECV */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trscatdr.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 50978    Output checksum: 36680
 */
