/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     tfIpScatteredSend() API
 * Filename:        trscatip.c
 * Author:          Odile
 * Date Created:    07/24/00
 * $Source: source/sockapi/trscatip.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:13CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>


#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


/*
 * tfIpScatteredSend function description
 * Scattered data IP send with IP header already in the data.
 * Allow the user to send an IP datagram directly to the stack without
 * using a socket interface, and without the Treck stack changing any
 * IP header field (except if IP fragmentation is needed), and directly
 * from user owned scattered data buffers. Even the IP header itself could be
 * scattered among several data buffers. The user passes an array of user
 * block data of type ttUserBlock, and passes the number of such elements.
 * Each ttUserBlock element contains a pointer to a user buffer, a pointer
 * to the beginning of the user data in the user buffer, and the user data
 * length in the user buffer. Upon return from this routine, the user
 * can re-use the array of ttUserBlock, but the Treck stack owns the
 * user buffers that were pointed to by the ttUserBlock elements.
 * The userFreeFunction will be called by the Treck stack for each
 * user buffer, when the data has been sent out on the network, and
 * the device driver no longer need to access the data in the user buffer.
 * It is the user's responsibility to ensure that the user free function
 * is re-entrant as it could potentially be called from different threads,
 * unless the user uses a different free function per user application
 * thread, and the user does not use a Treck transmit task, and does not
 * use a Treck send (complete) task.
 * Example of tfIpScatteredSend() usage is shown in the loop back test
 * module txscatlp.c.
 * Parameters:
 * interfaceHandle  If the user knows the interface the packet is to be
 *                  sent out to, then this field is non null.
 * userBlockPtr     Pointer to the first element of the user array that
 *                  contains information about the user scattered data.
 * userBlockCount   number of elements in the above array
 * userFreeFunction Pointer to the user free function that will be called
 *                  for each user buffer when the data in the user buffer
 *                  no longer need to be accessed. This function is called
 *                  by the Treck stack with the pointer to the user buffer
 *                  as a parameter.
 * Return
 * TM_ENOERROR      success.
 * TM_EINVAL        One of the parameters is bad, i.e userBlockPtr
 *                  is null, or userBlockCount is less or equal to 0
 *                  or userFreeFunction is null, or interface handle is
 *                  invalid.
 * TM_ENOBUFS       Could not allocate Treck headers to send the data or
 *                  could not allocate an ARP entry.
 * TM_EFAULT        Bad packet. IP header size is bigger than the maximum
 *                  allowed by the RFC, i.e, 60, or it less than the
 *                  minimum, 20.
 * TM_EFAULT        Bad packet. The IP header length is bigger than
 *                  the total data length in the user scattered buffers.
 * TM_EMSGSIZE      data length of the IP datagram is bigger than the
 *                  IP MTU, and fragmentation is not allowed.
 * TM_EHOSTUNREACH  No route to host
 * TM_EDESTADDRREQ  Destination IP addess in the IP header is zero
 * TM_ENOENT        No interface could be found to send the packet out
 *                  (only if destination IP address is limited broadcast.)
 * TM_ENXIO         Outgoing interface is closed
 * TM_EIO           Interface transmit queue if full (only if interface
 *                  transmit queue is used).
 * Other error code As returned by device driver send function.
 */

int tfIpScatteredSend ( ttUserInterface     interfaceHandle,
                        ttUserBlockPtr      userBlockPtr,
                        int                 userBlockCount,
                        ttUserFreeFuncPtr   userFreeFunction )
{
    ttDeviceEntryPtr devEntryPtr;
    ttLockEntryPtr   lockEntryPtr;
    ttPacketPtr      headPacketPtr;
    ttIpHeaderPtr    iphPtr;
    int              errorCode;
    tt8Bit           needFreeBuffers;
#ifdef TM_4_USE_SCOPE_ID
    tt6IpAddress     ipv4MappedSrcAddr;
    tt6IpAddress     ipv4MappedDestAddr;
#endif /* TM_4_USE_SCOPE_ID */

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Assume bad parameters */
    errorCode = TM_EINVAL;
    needFreeBuffers = TM_8BIT_ZERO;
/* No ttSharedData/ttPacket allocated yet */
    headPacketPtr = (ttPacketPtr)0;
/* Check on user parameters */
    if (    (userBlockPtr == (ttUserBlockPtr)0)
         || (userBlockCount <= 0)
         || (userFreeFunction == (ttUserFreeFuncPtr)0) )
    {
/*
 * If Bad parameters, return TM_EINVAL. Cannot free any buffers, since
 * We either do not have an array, or the array is of unknown size, or
 * there is no free function.
 */
        goto ipScatExit;
    }
/* Check on interface handle parameter */
    if (    (interfaceHandle != (ttUserInterface)0)
         && (tfValidInterface(devEntryPtr) != TM_ENOERROR ) )
    {
/*
 * If invalid interface, free the user buffers, and return TM_EINVAL.
 */
        needFreeBuffers = TM_8BIT_YES;
        goto ipScatExit;
    }
    errorCode = tfLinkScatDataToHeader( &headPacketPtr, userBlockPtr,
                                        userBlockCount, userFreeFunction,
                                        1);
    if (errorCode != TM_ENOERROR)
    {
/* Buffers already freed in tfLinkScatDataToHeader in case of failure */
        goto ipScatExit;
    }
/*
 * Set the TM_IP_FORWARD_FLAG bit in the ttPacket flags field to prevent
 * the IP send packet function from altering the IP header.
 */
    headPacketPtr->pktFlags |= TM_IP_FORWARD_FLAG;
/* Copy the interfaceHandle to the first ttPacket pktDeviceEntryPtr field. */
    if (devEntryPtr != (ttDeviceEntryPtr)0)
    {
        headPacketPtr->pktDeviceEntryPtr = devEntryPtr;
/*
 * Set the output device flag to let the IP layer know that the outgoing
 * interface pointer is known, and valid.
 */
        headPacketPtr->pktFlags |= TM_OUTPUT_DEVICE_FLAG;
    }
/* Point to copied IP header */
    iphPtr = (ttIpHeaderPtr)(headPacketPtr->pktLinkDataPtr);
#ifdef TM_LOCK_NEEDED
    lockEntryPtr = &tm_context(tvRtIpForwCacheLock);
/* Use a lock: the tvRtIpforwCacheLock to provide single thread access. */
    tm_lock_wait(lockEntryPtr);
#else /* !TM_LOCK_NEEDED */
    lockEntryPtr = (ttLockEntryPtr)0;
#endif /* TM_LOCK_NEEDED */
#ifdef TM_USE_STRONG_ESL
    tm_context(tvRtIpForwCache).rtcDevPtr = devEntryPtr;
#endif /* TM_USE_STRONG_ESL */
/*
 * Try and find a route for the destination address, but do not update
 * the source IP address. We will use the existing tfIpDestToPacket()
 * function, which will not change the source IP address, if it already set.
 */
#ifdef TM_4_USE_SCOPE_ID
    tm_6_addr_to_ipv4_mapped(
        iphPtr->iphSrcAddr, &ipv4MappedSrcAddr);
    if (devEntryPtr != (ttDeviceEntryPtr)0)
    {
        tm_4_dev_scope_addr(&ipv4MappedSrcAddr, devEntryPtr);
    }
    tm_6_addr_to_ipv4_mapped(
        iphPtr->iphDestAddr, &ipv4MappedDestAddr);
    if (devEntryPtr != (ttDeviceEntryPtr)0)
    {
        tm_4_dev_scope_addr(&ipv4MappedDestAddr, devEntryPtr);
    }
#endif /* TM_4_USE_SCOPE_ID */
    errorCode = tfIpDestToPacket(
        headPacketPtr,
#ifdef TM_4_USE_SCOPE_ID
        &ipv4MappedSrcAddr,
        &ipv4MappedDestAddr,
#else /* ! TM_4_USE_SCOPE_ID */
        iphPtr->iphSrcAddr,
        iphPtr->iphDestAddr,
#endif /* ! TM_4_USE_SCOPE_ID */
        iphPtr->iphTos,
        &tm_context(tvRtIpForwCache),
        iphPtr);
#ifdef TM_USE_STRONG_ESL
    tm_context(tvRtIpForwCache).rtcDevPtr = (ttDeviceEntryPtr)0;
#endif /* TM_USE_STRONG_ESL */
    if (errorCode == TM_ENOERROR)
    {
/* Do not free the user buffers here as they will be freed up by the stack */
/*
 * Call the existing tfIpSendPacket() directly passing the pointer to the
 * first ttPacket, and a pointer to the tvRtIpforwCacheLock.
 */
        errorCode = tfIpSendPacket(
            headPacketPtr,
            lockEntryPtr
            );
    }
#ifdef TM_LOCK_NEEDED
    else
    {
/* UnLock */
        tm_call_unlock(lockEntryPtr);
/* Free the user buffers */
        needFreeBuffers = TM_8BIT_YES;
/* To prevent user buffers from being freed twice */
        userBlockCount = 0;
    }
#endif /* TM_LOCK_NEEDED */
ipScatExit:
    if (needFreeBuffers != TM_8BIT_ZERO)
/*
 * If an error occured before we called tfIpSendPacket, then we need
 * to free the user buffers, and allocated ttPacket/ttSharedData
 */
    {
        tfFreeScatData( headPacketPtr, userBlockPtr, userBlockCount,
                        userFreeFunction );
    }
/* Return the errorCode to the user. */
    return errorCode;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trscatip.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 57253    Output checksum: 59905
 */
