/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     tfSocketScatteredSend() API
 * Filename:        trscatso.c
 * Author:          Odile
 * Date Created:    07/25/00
 * $Source: source/sockapi/trscatso.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:14CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/*
 * tfSocketScatteredSendTo function description
 * Scattered Data Socket Send To. Allows the user to send data on a
 * non TCP socket directly from user owned scattered data buffers.
 * The user passes an array of user block data of type ttUserBlock, and
 * passes the number of such elements.
 * Each ttUserBlock element contains a pointer to a user buffer, a pointer
 * to the beginning of the user data in the user buffer, and the user data
 * length in the user buffer. Upon return from this routine, the user
 * can re-use the array of ttUserBlock, but the Treck stack owns the
 * user buffers that were pointed to by the ttUserBlock elements.
 * The only exception is when TM_SOCKET_ERROR is returned, and the errorCode
 * retrieved with tfGetSocketError() is TM_EWOULDBLOCK. In that case
 * the user still owns the buffers, and should try and re-send the same
 * buffers later on.
 * The userFreeFunction will be called by the Treck stack for each
 * user buffer, when the data has been sent out on the network, and
 * the device driver no longer need to access the data in the user buffer.
 * It is the user's responsibility to ensure that the user free function
 * is re-entrant as it could potentially be called from different threads,
 * unless the user uses a different free function per user application
 * thread, and the user does not use a Treck transmit task, and does not
 * use a Treck send (complete) task.
 * Example of tfSocketScatteredSendTo() usage is shown in the loop back test
 * module txscatlp.c.
 * Parameters:
 * socketDescriptor As returned by a socket() call
 * userBlockPtr     Pointer to the first element of the user array that
 *                  contains information about the user scattered data.
 * userBlockCount   number of elements in the above array
 * userFreeFunction Pointer to the user free function that will be called
 *                  for each user buffer when the data in the user buffer
 *                  no longer need to be accessed. This function is called
 *                  by the Treck stack with the pointer to the user buffer
 *                  as a parameter.
 * flags            0, or MSG_DONTWAIT. If MSG_DONTWAIT, the call is
 *                  non blocking for the duration of the call. If flag
 *                  is 0, then it is blocking if the socket is in blocking
 *                  mode, non blocking otherwise.
 * toAddressPtr     The address to send the data to. If the user had
 *                  called connect on the socket, then toAddressPtr should
 *                  be null.
 * toAddressLength  the length of the address
 * Return
 * > 0              Number of bytes transmitted/queued. Treck will free the
 *                  user buffers.
 * TM_SOCKET_ERROR  An error occured. The user can retrieve the error code
 *                  with the tfGetSocketError(socketDescriptor) API:
 *
 *   .. TM_EWOULDBLOCK   Not enough room to queue the data in the socket
 *                       send queue. User still owns his buffers, and is
 *                       responsible for re-sending them later on.
 *   .. TM_EINVAL        One of the parameters is bad, i.e userBlockPtr
 *                       is null, or userBlockCount is less or equal to 0
 *                       or userFreeFunction is null, or socket is a TCP
 *                       socket with toAddressPtr being non zero, or flags
 *                       is not zero, or MSG_DONTWAIT.
 *   .. TM_EFAULT        Bad packet. IP header size is bigger than the
 *                       maximum allowed by the RFC, i.e, 60, or it less
 *                       than the minimum, 20.
 *                       Bad packet. The IP header length is bigger than
 *                       the total data length in the user scattered buffers.
 *   .. TM_ENOBUFS       Not enough memory to allocate the Treck headers
 *   .. TM_EBADF         The socket descriptor is invalid
 *   .. TM_EMSGSIZE      The data length was bigger than the send queue, or
 *                       the data length caused the IP datagram to be bigger
 *                       than the IP MTU, and IP fragmentation is not
 *                       allowed.
 *   .. TM_EPROTOTYPE    Attempt to send on a connected TCP socket
 *   .. TM_EHOSTUNREACH  No route to host
 *   .. TM_ENOENT        No interface could be found to send the packet out
 *                       (limited broadcast destination IP address.)
 *   .. TM_ENXIO         Outgoing interface is closed
 *   .. TM_EIO           Interface transmit queue if full (only if
 *                       interface transmit queue is used).
 *   .. Other error code As returned by device driver send function.
 */
int tfSocketScatteredSendTo( int                      socketDescriptor,
                             ttUserBlockPtr           userBlockPtr,
                             int                      userBlockCount,
                             ttUserFreeFuncPtr        userFreeFunction,
                             int                      flags,
                       const struct sockaddr TM_FAR * toAddressPtr,
                             int                      toAddressLength )

{
    ttPacketPtr      headPacketPtr;
    ttPacketPtr      packetPtr;
    int              errorCode;
    int              retCode;
    int              inclIpHdr;
    tt8Bit           needFreePacket;

/* Assume bad parameters */
    errorCode = TM_EINVAL;
    retCode = TM_ENOERROR; /* So that errorCode is set on socket */
    needFreePacket = TM_8BIT_ZERO;
/* No ttSharedData/ttPacket allocated yet */
    headPacketPtr = (ttPacketPtr)0;
/* Check on user parameters */
    if (    (userBlockPtr == (ttUserBlockPtr)0)
         || (userBlockCount <= 0)
         || (userFreeFunction == (ttUserFreeFuncPtr)0) )
    {
/*
 * If Bad parameters, return TM_EINVAL. Cannot free any buffers, since
 * We either do not have an array, or the array is of unknown size, or
 * there is no free function.
 */
        goto socketScatExit;
    }
    if ( flags & MSG_IPHDRINCL)
    {
        flags &= ~MSG_IPHDRINCL;
        inclIpHdr = 1;
    }
    else
    {
        inclIpHdr = 0;
    }
    errorCode = tfLinkScatDataToHeader( &headPacketPtr, userBlockPtr,
                                        userBlockCount, userFreeFunction,
                                        inclIpHdr);
    if (errorCode != TM_ENOERROR)
    {
        goto socketScatExit;
    }
/*
 * Initialize the pktUserFlags field of the first ttPacket with
 * TM_PKTF_USER_OWNS so that tfZeroCopySendTo/tfZeroCopySend() will
 * accept the buffer.
 */
    headPacketPtr->pktUserFlags = TM_PKTF_USER_OWNS;
/* Tell tfZeroCopySendTo()/tfZeroCopySend() that the data is scattered */
    flags |= MSG_SCATTERED;
/* If toAddressPtr is non null, call tfZeroCopySendTo() */
    if (toAddressPtr != (const struct sockaddr TM_FAR *)0)
    {
       retCode = tfZeroCopySendTo( socketDescriptor,
                                   (ttUserMessage)headPacketPtr,
                                   (int)(headPacketPtr->pktChainDataLength),
                                   flags,
                                   toAddressPtr,
                                   toAddressLength );
    }
/* otherwise call tfZeroCopySend() */
    else
    {
       retCode = tfZeroCopySend( socketDescriptor,
                                 (ttUserMessage)headPacketPtr,
                                 (int)(headPacketPtr->pktChainDataLength),
                                 flags );
    }
/*
 * Check return value from tfZeroCopySend()
 * If return value is TM_SOCKET_ERROR, we check the error code by calling
 * tfGetSocketError(). If error code is TM_EWOULDBLOCK, then we need to
 * walk the link of allocated ttPacket/ttSharedData, and reset the
 * TM_BUF_USER_DATA flag in each one of them, so that the Treck free
 * routine tfFreePacket() will not call the user free function for
 * each data block.
 * We set needFreePacket, so that tfFreePacket() is called.
 */
    if (retCode == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(socketDescriptor);
        if (errorCode == TM_EWOULDBLOCK)
        {
/*
 * Do not free the user buffers, but do free the ttPacket/ttShared data
 * link.
 */
            needFreePacket = TM_8BIT_YES;
            packetPtr = (ttPacketPtr)headPacketPtr->pktLinkNextPtr;
            while (packetPtr)
            {
                packetPtr->pktSharedDataPtr->dataFlags &= ~TM_BUF_USER_DATA;
                packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
            }
        }
    }

socketScatExit:
    if (errorCode != TM_ENOERROR)
    {
        if (retCode != TM_SOCKET_ERROR)
        {
/* Record the error for socketDescriptor */
            tfSocketErrorRecord(socketDescriptor, errorCode);
            retCode = TM_SOCKET_ERROR;
        }
    }
    if (needFreePacket != TM_8BIT_ZERO)
    {
        tfFreePacket(headPacketPtr, TM_SOCKET_UNLOCKED);
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trscatso.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 44012    Output checksum:  9102
 */
