/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2005
 * Copyright Elmic Systems Japan 1997 - 2005
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: socket sendtofrom specifying the output interface and MAC addr
 * Filename: trsndfif.c
 * Author: Odile
 * Date Created: 05/13/05
 * $Source: source/sockapi/trsndfif.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:25CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/*
 * tfSendToFromInterface function description
 * Same as tfSendToFrom, but takes 4 extra parameters to let the user specify
 * the link layer destination address, and the interface to send the data
 * through.
 * Parameters
 * Parameter        Description
 * socketDescriptor The socket descriptor to use to send data
 * bufferPtr        The buffer to send
 * bufferLength     The length of the buffer to send
 * flags            MSG_DONTWAIT: Don't wait for data send to complete
 *                  0: wait for data send to complete
 * toAddressPtr     The address to send the data to
 * addressLength    The length of the area pointed to by toPtr and/of
 *                  fromAddressPtr.
 * fromAddressPtr   The address to send the data from
 * toPhyAddrPtr     Destination Link layer address.
 * toPhyAddrLen     Destination link layer address length.
 * interfaceHandle  Interface to send the data through as returned by
 *                  tfAddInterface
 * mhomeIndex       multihome index on the configured interface
 * Returns
 * Value            Meaning
 * >=0              Number of bytes actually sent on the socket
 * -1               An error occured, error can be retrieved with
 *                  tfGetSocketError()
 * tfGetSocketError will return:
 * TM_EBADF         The socket descriptor is invalid
 * TM_ENOBUFS       There was insufficient user memory available to complete
 *                  the operation
 * TM_EMSGSIZE      The message was too long
 * TM_EPROTOTYPE    TCP protocol requires usage of send, not sendto
 * TM_EWOULDBLOCK   The socket is marked as non-blocking and
 *                  the send operation would block
 */
int tfSendToFromInterface ( int                           socketDescriptor,
                      const char                  TM_FAR *bufferPtr,
                            int                           bufferLength,
                            int                           flags,
                      const struct sockaddr       TM_FAR *toAddressPtr,
                            int                           addressLength,
                            const struct sockaddr TM_FAR *fromAddressPtr,
                      const ttUser8BitPtr                 toPhyAddrPtr,
                            int                           toPhyAddrLen,
                            ttUserInterface               interfaceHandle,
                            unsigned int                  mhomeIndex )
{
     ttPacketPtr         packetPtr;
     ttSharedDataPtr     pktShrDataPtr;
     ttSockAddrPtrUnion  fromSockAddr;
     ttDeviceEntryPtr    devPtr;
     tt8BitPtr           ethDestAddrPtr;
     int                 errorCode;
     int                 retCode;

/*
 * We first allocate a zero copy RIP buffer of length corresponding to a
 * send request
 */
    errorCode = TM_EINVAL; /* assume failure */
    retCode = TM_SOCKET_ERROR;
    devPtr = (ttDeviceEntryPtr)interfaceHandle;
    if (    (tfValidInterface(devPtr) == TM_ENOERROR)
         && (bufferPtr != (char TM_FAR *)0)
         && (bufferLength > 0)
         && ((int)mhomeIndex < TM_MAX_IPS_PER_IF) )
    {
        fromSockAddr.sockPtr = (struct sockaddr TM_FAR *)fromAddressPtr; 
        if (fromAddressPtr != (const struct sockaddr TM_FAR *)0)
        {
            if (
                  (fromSockAddr.sockPtr->sa_family != AF_INET)
               )
            {
                goto sendtoIfFinish;
            }
/* verify that the address length is at least what we expect */
            if ((fromSockAddr.sockPtr->sa_family == AF_INET) &&
                (addressLength < (int)sizeof(struct sockaddr_in)))
            {
                goto sendtoIfFinish;
            }
        }
        packetPtr = tfGetSharedBuffer(TM_MAX_SEND_HEADERS_SIZE,
                            (ttPktLen)bufferLength, TM_16BIT_ZERO);
        if (packetPtr != TM_PACKET_NULL_PTR)
        {
            pktShrDataPtr = packetPtr->pktSharedDataPtr;
            packetPtr->pktDeviceEntryPtr = devPtr;
            packetPtr->pktMhomeIndex = (tt16Bit)mhomeIndex;
            packetPtr->pktFlags = TM_OUTPUT_DEVICE_FLAG;
/*
 * BUG ID 202: When sending out packets, we should use what user selected by
 * parameter mhomeIndex as source IP address.
 */
            packetPtr->pktFlags2 = TM_PF2_MHOMEINDEX_SET;
            packetPtr->pktUserFlags = TM_PKTF_USER_OWNS;
            tm_bcopy(bufferPtr, packetPtr->pktLinkDataPtr, bufferLength);
            if (fromAddressPtr != (const struct sockaddr TM_FAR *)0)
            {
/* Indicate that source address should be retrived from dataFromIpAddress */
                packetPtr->pktUserFlags |= TM_PKTF_USER_SEND_FROM;
                tm_ip_copy(
                      (tt4IpAddress)(fromSockAddr.sockInPtr->sin_addr.s_addr),
                      pktShrDataPtr->dataFromIpAddress);
                pktShrDataPtr->dataLocalPortNumber
                        = (ttIpPort)(fromSockAddr.sockInPtr->sin_port);
            }
            if (toPhyAddrPtr != (ttUser8BitPtr)0)
            {
                if (   (toPhyAddrLen == sizeof(ttEnetAddress)) 
                     && tm_4_ll_is_lan(devPtr))
                {
                    packetPtr->pktFlags |= TM_ETHERNET_ADDRESS_FLAG;
                    ethDestAddrPtr =
                            (tt8BitPtr)&pktShrDataPtr->dataEthernetAddress[0];
                    tm_bcopy(toPhyAddrPtr, ethDestAddrPtr, toPhyAddrLen);
                }
            }
            retCode = tfZeroCopySendTo( socketDescriptor,
                                        (ttUserMessage)packetPtr,
                                        bufferLength,
                                        flags,
                                        toAddressPtr,
                                        addressLength );
            if (retCode != TM_SOCKET_ERROR)
            {
                errorCode = TM_ENOERROR;
            }
            else
            {
                errorCode = tfGetSocketError(socketDescriptor);
                if (errorCode == TM_EWOULDBLOCK)
                {
/* If buffer has not been freed up */
                    (void)tfFreeZeroCopyBuffer((ttUserMessage)packetPtr);
                }
            }
        }
        else
        {
            errorCode = TM_ENOBUFS;
        }
    }
sendtoIfFinish:
    if (errorCode != TM_ENOERROR)
    {
/* Record the socket error for socketDescriptor */
        tfSocketErrorRecord(socketDescriptor, errorCode);
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trsndfif.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  6910    Output checksum:  8801
 */
