/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: SNMP agent cache functions (for UDP socket table, 
 *              connected TCP vectors table, routing table, Arp table.)
 * Filename: trsnmpca.c
 * Author: Odile
 * Date Created: 01/18/02
 * $Source: source/sockapi/trsnmpca.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:28CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


#include <trsocket.h>
#include <trmacro.h>
#ifdef TM_SNMP_CACHE
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* Local macros */



/* local types */
typedef ttNodePtr (TM_CODE_FAR * ttSnmpSearchFuncPtr)(
    ttSnmpCacheEntryPtr snmpCacheEntryPtr,
    ttListPtr           listPtr);

/* Local functions */

static void tfSnmpdCacheDeleteEntry (ttSnmpCacheEntryPtr snmpCacheEntryPtr);

/* Insert an entry in its corresponding cache */
static void tfSnmpdCacheInsertEntry(ttSnmpCacheEntryPtr snmpCacheEntryPtr);

/* 
 * Copy ARP entry destination address + copy device index to
 * passed ARP search structure
 */
static void tf4SnmpArpLexGraphOrder(ttRteEntryPtr       rtePtr,
                                    tt4SnmpArpSearchPtr snmp4ArpSearchPtr);
/* tfListWalk CB function ARP entry comparison */
static int tf4SnmpArpCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam);

/* Search through the IPv4 ARP table to find point of insertion */
static ttNodePtr tf4SnmpArpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                  ttListPtr           listPtr);

/* 
 * Copy Route entry dest, bit offset, hops and device index
 * to the passed Route search structure
 */
static void tf4SnmpRouteLexGraphOrder(ttRteEntryPtr         rtePtr,
                                      tt4SnmpRouteSearchPtr snmp4RouteSearchPtr);

/* tfListWalk CB function Route entry comparison */
static int tf4SnmpRouteCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam);

/* Search through the IPv4 Route table to find point of insertion */
static ttNodePtr tf4SnmpRouteSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                    ttListPtr           listPtr);

/* Copy 2-tuple IPv4 address/port , and devIndex */
static void tf4SnmpUdpLexGraphOrder(ttTupleDevPtr       tupleDevPtr,
                                    tt4SnmpUdpTuplePtr  snmp4UdpTuplePtr);

/* tfListWalk CB function UDP entry comparison */
static int tf4SnmpUdpTupleCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam);

/* Search through the IPv4 UDP table to find point of insertion */
static ttNodePtr tf4SnmpUdpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                 ttListPtr           listPtr);

/* Copy 4-tuple IPv4 address/port, and devIndex */
static void tf4SnmpTcpLexGraphOrder(ttTupleDevPtr      tupleDevPtr,
                                    tt4SnmpTcpTuplePtr snmp4TcpTuplePtr);

/* tfListWalk CB function TCP entry comparison */
static int tf4SnmpTcpTupleCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam);

/* Search through the IPv4 TCP table to find point of insertion */
static ttNodePtr tf4SnmpTcpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                 ttListPtr           listPtr);



/*
 * tfSnmpdCacheDeleteRoute function description:
 * . Common call for ARP cache entry, and IP routing entry deletion
 *   from the SNMP cache.
 * . Called with the Patricia Tree Lock on.
 * . Called from tfRtRemoveEntry() when routing/Arp cache entry
 *   is to be removed from the Patricia tree. 
 * . That routing entry may or may not be in the cache, as
 *   it could also be a broadcast, multicast entry that does
 *   not belong in the SNMP routing/Arp entries cache, but
 *   tfRtRemoveEntry() checks that the entry is in the cache
 *   prior to calling this routine.
 *   Parameters  Meaning
 *   rtePtr      Pointer to routing entry to be removed from the cache.
 *   No Return
 */
void tfSnmpdCacheDeleteRoute (ttRteEntryPtr rtePtr)
{
    tfSnmpdCacheDeleteEntry(&(rtePtr->rteSnmpCacheEntry));
    return;
}

/*
 * tfSnmpdCacheInsertRoute function description:
 * . Called with the Patricia Tree Lock on.
 * . Called from tfRtClone() when Arp cache entry
 *   has been added to the Patricia tree.  
 * . Or Called from tfRtLockAdd() when routing entry
 *   has been added to the Patricia tree. 
 * . If entry is an arp entry, use the SNMP Arp table cache
 *   and insert the ARP cache entry by devIndex/Ip address.
 * . If entry is a routing entry, use the SNMP routing table cache
 *   and insert the routing entry by Ip address.
 * . Determine which cache the entry is to be inserted to
 * . Use common routine tfSnmpCacheInsertEntry()
 *   Parameters  Meaning
 *   rtePtr      Pointer to ARP entry, or routing entry to be added to
 *               the cache.
 *   cacheIndex  indicates if rtePtr is an ARP entry (TM_SNMPC_ARP_INDEX),
 *               or a routing entry (TM_SNMPC_ROUTE_INDEX).
 *   No Return
 */
void tfSnmpdCacheInsertRoute (ttRteEntryPtr rtePtr , int cacheIndex)
{
    tt8Bit              entryType;


    if (cacheIndex == TM_SNMPC_ARP_INDEX)
    {
        entryType = TM_SNMPC_ARP_TYPE;
    }
    else
    {
        entryType = TM_SNMPC_ROUTE_TYPE;
    }
/*
 * In dual mode, convert to appropriate IPv4, or IPv6 cache.
 */

/* Store cache index for deletion */
    rtePtr->rteSnmpCacheEntry.sceTableIndex = (tt8Bit)cacheIndex;
    rtePtr->rteSnmpCacheEntry.sceTableEntryPtr = (ttVoidPtr)rtePtr;
    rtePtr->rteSnmpCacheEntry.sceParmPtr = (ttVoidPtr)rtePtr;
    rtePtr->rteSnmpCacheEntry.sceTableEntryType = entryType;
    tfSnmpdCacheInsertEntry(&(rtePtr->rteSnmpCacheEntry));
}

/*
 * tfSnmpdCacheDeleteSocket function description:
 * . Common call for UDP socket entry, and TCP socket deletion
 *   from the SNMP cache.
 * . Called with the Socket Tree Lock on from tfSocketTreeDelete(), when
 *   socket entry is to be removed from the tree.
 *   Parameters      Meaning
 *   socketPtr       Pointer to a socket entry to be deleted from the cache.
 *   No Return
 */
void tfSnmpdCacheDeleteSocket (ttSocketEntryPtr socketPtr)
{
    tfSnmpdCacheDeleteEntry(&(socketPtr->socSnmpCacheEntry));
    return;
}

/*
 * tfSnmpdCacheDeleteTmWtVect function description:
 * . Called with the Table Time Wait Table Lock on from tfTcpTmWtVectClose(),
 *   when a TCP time wait vector entry is to be removed from the Time Wait
 *   vector hash table.
 *   Parameters      Meaning
 *   tcpTmWtVectPtr  Pointer to a Time Wait TCP vector entry to be deleted
 *                   from the cache.
 *   No Return
 */
void tfSnmpdCacheDeleteTmWtVect (ttTcpTmWtVectPtr tcpTmWtVectPtr)
{
    tfSnmpdCacheDeleteEntry(&(tcpTmWtVectPtr->twsSnmpCacheEntry));
    return;
}

/*
 * tfSnmpdCacheInsertSocket function description:
 * . Called with the Socket Tree Lock on.
 * . Called from tfSocketTreeInsert() when a socket entry has been  
 *   added to the Socket Tree.  
 * . If entry is a UDP socket, use the SNMP UDP socket table cache
 *   and insert the UDP socket by local IP address/local port.
 * . If entry is a TCP socket, use the SNMP TCP vector table cache 
 *   and insert the socket entry by local IP address/local port/
 *   remote IP address/remote port.
 * . Insertion is in a null terminated doubly linked list pointed to by
 *   the SNMP cache head.
 * . Update pointers to first, and last element of the list.
 * . Update the SNMP cache serial number.
 *   Parameters      Meaning
 *   socketPtr       Pointer to a socket entry to be added to the cache.
 *   No Return
 */
void tfSnmpdCacheInsertSocket (ttSocketEntryPtr socketPtr)
{
    int                 cacheIndex;
    tt8Bit              entryType;

    entryType = (tt8Bit)TM_SNMPC_UDP_TYPE; 
/* Find which cache this entry belongs to */
    if (socketPtr->socProtocolNumber == IP_PROTOUDP)
    {
        cacheIndex = TM_SNMPC_UDP_INDEX;
/*
 * In dual mode, convert to appropriate IPv4, or IPv6 cache.
 */
    }
    else
    {
        if (socketPtr->socProtocolNumber == IP_PROTOTCP)
        {
/* Connected */
            entryType = (tt8Bit)TM_SNMPC_TCP_TYPE; 
            cacheIndex = TM_SNMPC_TCP_INDEX;
/*
 * In dual mode, convert to appropriate IPv4, or IPv6 cache.
 */
        }
        else
        {
            cacheIndex = -1; /* Non TCP, non UDP socket */
        }
    }
    if (cacheIndex != -1)
    {
        socketPtr->socSnmpCacheEntry.sceTableIndex = (tt8Bit)cacheIndex;
        socketPtr->socSnmpCacheEntry.sceParmPtr = (ttVoidPtr)
                                                      &(socketPtr->socTupleDev);
        socketPtr->socSnmpCacheEntry.sceTableEntryPtr = (ttVoidPtr)socketPtr;
        socketPtr->socSnmpCacheEntry.sceTableEntryType = entryType;
        tfSnmpdCacheInsertEntry(&(socketPtr->socSnmpCacheEntry));
    }
}

/* Insert a TCP time wait vector in the TCP cache */
void tfSnmpdCacheInsertTmWtVect (ttTcpTmWtVectPtr tcpTmWtVectPtr)
{
    int cacheIndex;

    cacheIndex = TM_SNMPC_TCP_INDEX;
    tcpTmWtVectPtr->twsSnmpCacheEntry.sceTableIndex = (tt8Bit)cacheIndex;
    tcpTmWtVectPtr->twsSnmpCacheEntry.sceTableEntryPtr = (ttVoidPtr)
                                                               tcpTmWtVectPtr;
    tcpTmWtVectPtr->twsSnmpCacheEntry.sceParmPtr = (ttVoidPtr)
                                            &(tcpTmWtVectPtr->twsTupleDev);
    tcpTmWtVectPtr->twsSnmpCacheEntry.sceTableEntryType =
                                            (tt8Bit)TM_SNMPC_TCP_TIME_WAIT_TYPE;
    tfSnmpdCacheInsertEntry(&(tcpTmWtVectPtr->twsSnmpCacheEntry));
}

/* 
 * tfSnmpdCacheDeleteEntry function description:
 * . Common call for socket, route, ARP, deletion.
 * . Map TCP socket to SNMP TCP vector cache, and UDP socket to SNMP UDP 
 *   socket cache. 
 * . Delete entry from list starting at SNMP cache head.
 *   head. 
 * . Update pointers to first, and last element of the list.
 * . Update the corresponding table SNMP cache serial number.
 */
static void tfSnmpdCacheDeleteEntry (ttSnmpCacheEntryPtr snmpCacheEntryPtr)
{
    ttSnmpdCacheHeadPtr snmpdCacheHeadPtr;
    int                 cacheIndex;

/* Find which cache this entry belongs to */
    cacheIndex = snmpCacheEntryPtr->sceTableIndex;
    tm_assert(tfSnmpdCacheDeleteEntry, cacheIndex != TM_SNMPC_NO_INDEX);
    snmpdCacheHeadPtr = &(tm_context(tvSnmpdCaches)[cacheIndex]);
#ifdef TM_LOCK_NEEDED
/* Lock the cache (to lock out SNMP cache lookup) */
    tm_call_lock_wait(&(snmpdCacheHeadPtr->snmpcLockEntry));
#endif /* TM_LOCK_NEEDED */
    if (snmpdCacheHeadPtr->snmpcNextNodePtr == &(snmpCacheEntryPtr->sceNode))
    {
/*
 * Make Insert next to scan point to next, if next to scan node is being
 * deleted
 */
        snmpdCacheHeadPtr->snmpcNextNodePtr =
                                     snmpCacheEntryPtr->sceNode.nodeNextPtr;
    }
    tfListRemove(&(snmpdCacheHeadPtr->snmpcList),
                 &(snmpCacheEntryPtr->sceNode));
    snmpCacheEntryPtr->sceTableIndex = TM_SNMPC_NO_INDEX;
#ifdef TM_SNMP_VERSION
    if (snmpdCacheHeadPtr->snmpcLastReadNodePtr ==
                                             &(snmpCacheEntryPtr->sceNode))
    {
/* Make GETNEXT point to next, if last read node is being deleted */
        snmpdCacheHeadPtr->snmpcLastReadNodePtr =
                                     snmpCacheEntryPtr->sceNode.nodeNextPtr; 
    }
#endif /* TM_SNMP_VERSION */
#ifdef TM_LOCK_NEEDED
/* Unlock the cache (to allow SNMP lookup) */
    tm_call_unlock(&(snmpdCacheHeadPtr->snmpcLockEntry));
#endif /* TM_LOCK_NEEDED */
}


/* Insert an entry in its corresponding cache */
static void tfSnmpdCacheInsertEntry(ttSnmpCacheEntryPtr snmpCacheEntryPtr)
{
    ttSnmpdCacheHeadPtr snmpdCacheHeadPtr;
    ttSnmpSearchFuncPtr snmpSearchFuncPtr;
    ttListPtr           listPtr;
    ttNodePtr           insertNodePtr;
    ttNodePtr           nodePtr;
    int                 cacheIndex;

    insertNodePtr = &(snmpCacheEntryPtr->sceNode);
#ifdef TM_ERROR_CHECKING
/* Check if already inserted in the cache */
    if (insertNodePtr->nodeNextPtr != (ttNodePtr)0)
    {
        tfKernelError("tfSnmpdCacheInsertEntry", "Entry already in cache!\n");
    }
    else
#endif /* TM_ERROR_CHECKING */
    {
        cacheIndex = snmpCacheEntryPtr->sceTableIndex;
        snmpdCacheHeadPtr = &(tm_context(tvSnmpdCaches)[cacheIndex]);
        snmpSearchFuncPtr = (ttSnmpSearchFuncPtr)0; /* compiler warning */
#ifdef TM_LOCK_NEEDED
/* Lock the cache (to lock out SNMP cache lookup) */
        tm_call_lock_wait(&(snmpdCacheHeadPtr->snmpcLockEntry));
#endif /* TM_LOCK_NEEDED */
        listPtr = &(snmpdCacheHeadPtr->snmpcList);
/* Save search criteria in local variables for loop efficiency */
        {
            switch (snmpCacheEntryPtr->sceTableEntryType)
            {
                case TM_SNMPC_ARP_TYPE:
                    snmpSearchFuncPtr = tf4SnmpArpSearch;
                    break;
                case TM_SNMPC_ROUTE_TYPE:
                    snmpSearchFuncPtr = tf4SnmpRouteSearch;
                    break;
                case TM_SNMPC_UDP_TYPE:
                    snmpSearchFuncPtr = tf4SnmpUdpSearch;
                    break;
                case TM_SNMPC_TCP_TYPE:
/* Fall through */
                case TM_SNMPC_TCP_TIME_WAIT_TYPE: 
                    snmpSearchFuncPtr = tf4SnmpTcpSearch;
                    break;
                default:
                    snmpSearchFuncPtr = (ttSnmpSearchFuncPtr)0;
                    break;
            }
        }
        if (snmpSearchFuncPtr != (ttSnmpSearchFuncPtr)0)
        {
            nodePtr = (*snmpSearchFuncPtr)(snmpCacheEntryPtr, listPtr);
            if (nodePtr == (ttNodePtr)0)
/* Insert at end of the list */
            {
                tfListAddToTail(listPtr, insertNodePtr);
            }
            else
            {
/* Insert before nodePtr */
                insertNodePtr->nodeNextPtr = nodePtr;
                insertNodePtr->nodePrevPtr = nodePtr->nodePrevPtr;
                nodePtr->nodePrevPtr->nodeNextPtr = insertNodePtr;
                nodePtr->nodePrevPtr = insertNodePtr;
                listPtr->listCount++;
            }
        }
#ifdef TM_LOCK_NEEDED
/* Unlock the cache (to allow SNMP lookup) */
        tm_call_unlock(&(snmpdCacheHeadPtr->snmpcLockEntry));
#endif /* TM_LOCK_NEEDED */
    }
}


/* 
 * Copy ARP entry destination address and device index in network
 * byte order to passed SNMP ARP search structure
 */
static void tf4SnmpArpLexGraphOrder(ttRteEntryPtr       rtePtr,
                                    tt4SnmpArpSearchPtr snmp4ArpSearchPtr)
{
    tt32Bit devIndex;

    devIndex = rtePtr->rteDevIndex;
    tm_htonl(devIndex, snmp4ArpSearchPtr->snmpa4DevIndex);
    tm_ip_copy(tm_4_ip_addr(rtePtr->rteDest), snmp4ArpSearchPtr->snmpa4Dest);
    return;
}
/*
 * tfListWalk CB function
 * Parameter is pointer to an ARP search structre
 * Copy ARP entry dest and device index to a local ARP
 * search structure
 * compare the 2 structures.
 */
static int tf4SnmpArpCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam)
{
    tt4SnmpArpSearch    snmp4ArpSearch;
    int                 retCode;
    
    tf4SnmpArpLexGraphOrder((ttRteEntryPtr)
        (((ttSnmpCacheEntryPtr)(ttVoidPtr)nodePtr)->sceTableEntryPtr),
        &snmp4ArpSearch);
    retCode = tm_memcmp(&snmp4ArpSearch,
                        genParam.genVoidParmPtr,
                        sizeof(tt4SnmpArpSearch));
    return retCode;
}

/* Search through the IPv4 ARP table to find point of insertion */
static ttNodePtr tf4SnmpArpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                  ttListPtr           listPtr)
{
    ttNodePtr           nodePtr;
    tt4SnmpArpSearch    snmp4ArpSearch;
    ttGenericUnion      genParam;

    tf4SnmpArpLexGraphOrder((ttRteEntryPtr)(snmpCacheEntryPtr->sceTableEntryPtr),
                       &snmp4ArpSearch);
    genParam.genVoidParmPtr = &snmp4ArpSearch;
    nodePtr = tfListWalk(listPtr,
                         tf4SnmpArpCompareCB, genParam);
    return nodePtr;
}

/* 
 * Copy Route entry dest, bit offset, hops and device index
 * to the passed Route search structure
 */
static void tf4SnmpRouteLexGraphOrder(ttRteEntryPtr          rtePtr,
                                      tt4SnmpRouteSearchPtr  snmp4RouteSearchPtr)
{
#ifdef TM_USE_STRONG_ESL
    tt32Bit temp32Bit;
#endif /* TM_USE_STRONG_ESL */

    tm_ip_copy(tm_4_ip_addr(rtePtr->rteDest), snmp4RouteSearchPtr->snmpr4Dest);
    snmp4RouteSearchPtr->snmpr4BitOffset = rtePtr->rteLeafNode.rtnBitOffset; 
#ifdef TM_USE_STRONG_ESL
    temp32Bit = rtePtr->rteHops;
    tm_htonl(temp32Bit, snmp4RouteSearchPtr->snmpr4Hops);
    temp32Bit = rtePtr->rteDevIndex;
    tm_htonl(temp32Bit, snmp4RouteSearchPtr->snmpr4DevIndex);
#endif /* TM_USE_STRONG_ESL */
    return;
}
/*
 * tfListWalk CB function
 * Parameter is pointer to a Route search structre
 * Copy Route entry dest, bit offset, hops and device index to a local Route
 * search structure
 * compare the 2 structures.
 */
static int tf4SnmpRouteCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam)
{
    tt4SnmpRouteSearch  snmp4RouteSearch;
    int                 retCode;
    
    tf4SnmpRouteLexGraphOrder((ttRteEntryPtr)
        (((ttSnmpCacheEntryPtr)(ttVoidPtr)nodePtr)->sceTableEntryPtr),
        &snmp4RouteSearch);
    retCode = tm_memcmp(&snmp4RouteSearch,
                        genParam.genVoidParmPtr,
                        sizeof(tt4SnmpRouteSearch));
    return retCode;
}

/* Search through the IPv4 Route table to find point of insertion */
static ttNodePtr tf4SnmpRouteSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                  ttListPtr           listPtr)
{
    ttNodePtr           nodePtr;
    tt4SnmpRouteSearch  snmp4RouteSearch;
    ttGenericUnion      genParam;

    tf4SnmpRouteLexGraphOrder((ttRteEntryPtr)(snmpCacheEntryPtr->sceTableEntryPtr),
                         &snmp4RouteSearch);
    genParam.genVoidParmPtr = &snmp4RouteSearch;
    nodePtr = tfListWalk(listPtr,
                         tf4SnmpRouteCompareCB, genParam);
    return nodePtr;
}

/* Copy 2-tuple IPv4 address/port, and devIndex */
static void tf4SnmpUdpLexGraphOrder(ttTupleDevPtr       tupleDevPtr,
                                    tt4SnmpUdpTuplePtr  snmp4UdpTuplePtr)
{
/* Copy local IP address and IP port in network byte order (lexicographical order)*/
    tm_ip_copy(tm_4_ip_addr(tupleDevPtr->tudTuple.sotLocalIpAddress),
               snmp4UdpTuplePtr->snmpu4LocalAddress); 
    snmp4UdpTuplePtr->snmpu4LocalPort = tupleDevPtr->tudTuple.sotLocalIpPort;
#ifdef TM_USE_STRONG_ESL
    if (tupleDevPtr->tudDevPtr != (ttDeviceEntryPtr)0)
    {
         snmp4UdpTuplePtr->snmpu4DevIndex = tupleDevPtr->tudDevPtr->devIndex;
    }
    else
    {
         snmp4UdpTuplePtr->snmpu4DevIndex = 0;
    }
#endif /* TM_USE_STRONG_ESL */
    return;
}

/*
 * tfListWalk CB function
 * Parameter is pointer to a 2-tuple IPv4 address, port
 * Copy node 4-tuple into a 2-tuple IPv4 address, port
 * Compare copied node parameter with passed parameter
 */
static int tf4SnmpUdpTupleCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam)
{
    tt4SnmpUdpTuple     snmp4UdpTuple;
    int                 retCode;
    
    tf4SnmpUdpLexGraphOrder(
        (ttTupleDevPtr)(((ttSnmpCacheEntryPtr)(ttVoidPtr)nodePtr)->sceParmPtr),
        &snmp4UdpTuple);
    retCode = tm_memcmp(&snmp4UdpTuple,
                        genParam.genVoidParmPtr,
                        sizeof(tt4SnmpUdpTuple));
    return retCode;
}

/* Search through the IPv4 UDP table to find point of insertion */
static ttNodePtr tf4SnmpUdpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                 ttListPtr           listPtr)
{
    ttNodePtr           nodePtr;
    tt4SnmpUdpTuple     snmp4UdpTuple;
    ttGenericUnion      genParam;

    tf4SnmpUdpLexGraphOrder(snmpCacheEntryPtr->sceParmPtr, &snmp4UdpTuple);
    genParam.genVoidParmPtr = &snmp4UdpTuple;
    nodePtr = tfListWalk(listPtr,
                         tf4SnmpUdpTupleCompareCB, genParam);
    return nodePtr;
}

/* Copy 4-tuple IPv4 address/port, and devIndex */
static void tf4SnmpTcpLexGraphOrder(ttTupleDevPtr      tupleDevPtr,
                                    tt4SnmpTcpTuplePtr snmp4TcpTuplePtr)
{
    tf4SnmpUdpLexGraphOrder(tupleDevPtr,
                      (tt4SnmpUdpTuplePtr)(ttVoidPtr)snmp4TcpTuplePtr);
    tm_ip_copy(tm_4_ip_addr(tupleDevPtr->tudTuple.sotRemoteIpAddress),
               snmp4TcpTuplePtr->snmpt4RemoteAddress);
    snmp4TcpTuplePtr->snmpt4RemotePort = tupleDevPtr->tudTuple.sotRemoteIpPort;
#ifdef TM_USE_STRONG_ESL
    if (tupleDevPtr->tudDevPtr != (ttDeviceEntryPtr)0)
    {
         snmp4TcpTuplePtr->snmpt4DevIndex = tupleDevPtr->tudDevPtr->devIndex;
    }
    else
    {
         snmp4TcpTuplePtr->snmpt4DevIndex = 0;
    }
#endif /* TM_USE_STRONG_ESL */
    return;
}

/*
 * tfListWalk CB function
 * Parameter is pointer to a 4-tuple IPv4 address, port
 * Copy node 4-tuple into a 4-tuple IPv4 address, port
 * Compare copied node parameter with passed parameter
 */
static int tf4SnmpTcpTupleCompareCB(
    ttNodePtr       nodePtr,
    ttGenericUnion  genParam)
{
    tt4SnmpTcpTuple     snmp4TcpTuple;
    int                 retCode;
    
    tf4SnmpTcpLexGraphOrder(
        (ttTupleDevPtr)(((ttSnmpCacheEntryPtr)(ttVoidPtr)nodePtr)->sceParmPtr),
        &snmp4TcpTuple);
    retCode = tm_memcmp(&snmp4TcpTuple,
                        genParam.genVoidParmPtr,
                        sizeof(tt4SnmpTcpTuple));
    return retCode;
}

/* Search through the IPv4 TCP table to find point of insertion */
static ttNodePtr tf4SnmpTcpSearch(ttSnmpCacheEntryPtr snmpCacheEntryPtr,
                                 ttListPtr           listPtr)
{
    ttNodePtr           nodePtr;
    tt4SnmpTcpTuple     snmp4TcpTuple;
    ttGenericUnion      genParam;

    tf4SnmpTcpLexGraphOrder(snmpCacheEntryPtr->sceParmPtr, &snmp4TcpTuple);
    genParam.genVoidParmPtr = &snmp4TcpTuple;
    nodePtr = tfListWalk(listPtr,
                         tf4SnmpTcpTupleCompareCB, genParam);
    return nodePtr;
}



#else /* TM_SNMP_CACHE */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_SNMP_CACHE is not defined */
int tvSnmpdCacheDummy = 0;

#endif /* TM_SNMP_CACHE */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trsnmpca.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 20827    Output checksum:  3389
 */
