/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2009
 * Copyright Elmic Systems Japan 1997 - 2009
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Create a socket by importing data from an existing socket
 *
 * Filename: trsocimp.c
 * Author: Lorne
 * Date Created: 7/30/09
 * $Source: source/sockapi/trsocimp.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:29CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#define TM_USE_TCP_IO

#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_SOCKET_IMPORT

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>



/*
 * Local macros
 */

/* Turn ON/OFF Big Window Scaling */
#define TM_TCPF_WND_SCALE      (tt16Bit)TM_TCP_WND_SCALE  /* 0x0080 */
/* Options resettable by the remote side if not set in its SYN */
#define TM_TCPF_OPTIONS        (tt16Bit)(   TM_TCPF_SEL_ACK \
                                          | TM_TCPF_WND_SCALE \
                                          | TM_TCPF_TS)
/* Turn ON/OFF Slow Start */
#define TM_TCPF_SLOW_START     (tt16Bit)TM_TCP_SLOW_START /* 0x0200 */

/* Maximum number of sqns in SACK option when time stamp option is used */
#define TM_TCP_TS_SACK_SQNS       6

/* maximum value for window scale */
#define TM_TCP_MAX_WNDSC       (tt8Bit)14

/*
 * Forward declarations
 */
static int tfTcpRcvSynOptionsSocImp(ttTcpVectPtr tcpVectPtr);

/*
 * Create a connected TCP socket based on information from the user.
 *
 * Parameter        description
 * userTcpConPtr    Pointer to TCP connection state information
 * errorCodePtr     Pointer to user variable to receive error information
 *
 * Returns a new socket descriptor if successful or TM_SOCKET_ERROR if
 * unsuccessful.
 *
 * If unsuccessful, error code returned will return one of:
 *      TM_EINVAL           Invalid parameter
 *      TM_EMFILE           Exceeded the limit on the number of sockets
 *      TM_ENOBUFS          Not enough free memory
 *      TM_EAFNOSUPPORT     Address family not supported
 *      TM_EADDRINUSE       Local address is already in use
 *      TM_EHOSTUNREACH     No route to host
 *      TM_EADDRNOTAVAIL    Remote address is not available
 *      TM_ENOENT           No matching outbound device
 */
int tfUserCreateTcpSocket(  ttUserTcpConPtr userTcpConPtr,
                            int TM_FAR *    errorCodePtr )
{
    ttSockAddrStoragePtr    localAddrPtr;
    ttSockAddrStoragePtr    remoteAddrPtr;
    tt8Bit                  family;
    ttTcpVectPtr            tcpVectPtr;
    ttSocketEntryPtr        socketPtr;
    ttSockAddrPtrUnion      internetAddressPtrU;
    int                     socketDescriptor;
    int                     errorCode;
    int                     retCode;
    tt16Bit                 tcpFlagMask;
    tt16Bit                 tcpFlagSet;

    errorCode = TM_ENOERROR;
    retCode = TM_SOCKET_ERROR;
    socketDescriptor = TM_SOCKET_ERROR;
    socketPtr = TM_SOCKET_ENTRY_NULL_PTR;

/*
 * Create the socket.
 */
    localAddrPtr = &userTcpConPtr->uconAddrLocal;
    family = localAddrPtr->ss_family;
    socketDescriptor = socket(family, SOCK_STREAM, IPPROTO_TCP);
    if (socketDescriptor == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(socketDescriptor);
        goto userCreateTcpSocketReturn;
    }
#ifdef TM_USE_STRONG_ESL

/*
 * Bind to device, if the user has specified a device name.
 */
    if (userTcpConPtr->uconBindDeviceName != (ttUserConstCharPtr)0)
    {
        errorCode = setsockopt( socketDescriptor,
                                SOL_SOCKET,
                                SO_BINDTODEVICE,
                                userTcpConPtr->uconBindDeviceName,
                                tm_strlen(userTcpConPtr->uconBindDeviceName));
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(socketDescriptor);
            tfClose(socketDescriptor);
            socketDescriptor = TM_SOCKET_ERROR;
            goto userCreateTcpSocketReturn;
        }
    }
#endif /* TM_USE_STRONG_ESL */

/*
 * Set send and receive buffer sizes.
 */
    if (userTcpConPtr->uconRcvBufSize != TM_32BIT_ZERO)
    {
        errorCode = setsockopt( socketDescriptor,
                                SOL_SOCKET,
                                SO_RCVBUF,
                                (ttCharPtr)&userTcpConPtr->uconRcvBufSize,
                                sizeof(userTcpConPtr->uconRcvBufSize) );
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(socketDescriptor);
            tfClose(socketDescriptor);
            socketDescriptor = TM_SOCKET_ERROR;
            goto userCreateTcpSocketReturn;
        }
    }
    if (userTcpConPtr->uconSndBufSize != TM_32BIT_ZERO)
    {
        errorCode = setsockopt( socketDescriptor,
                                SOL_SOCKET,
                                SO_SNDBUF,
                                (ttCharPtr)&userTcpConPtr->uconSndBufSize,
                                sizeof(userTcpConPtr->uconSndBufSize) );
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(socketDescriptor);
            tfClose(socketDescriptor);
            socketDescriptor = TM_SOCKET_ERROR;
            goto userCreateTcpSocketReturn;
        }
    }

/*
 * Bind the socket to the user specified local port and address.
 */
    errorCode = tfBindNoCheck(  socketDescriptor,
                                &localAddrPtr->addr.ipSockAddr,
                                sizeof(struct sockaddr_storage) );
    if (errorCode == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(socketDescriptor);
        tfClose(socketDescriptor);
        socketDescriptor = TM_SOCKET_ERROR;
        goto userCreateTcpSocketReturn;
    }

/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    remoteAddrPtr = &userTcpConPtr->uconAddrRemote;
    remoteAddrPtr->ss_family = family;
    internetAddressPtrU.sockPtr = &remoteAddrPtr->addr.ipSockAddr;
    socketPtr = tfSocketCheckAddrLock(socketDescriptor,
                                      internetAddressPtrU.sockPtr,
                                      sizeof(struct sockaddr_storage),
                                      &errorCode
                                     );
    if (socketPtr == TM_SOCKET_ENTRY_NULL_PTR)
    {
/* errorCode set by tfSocketCheckAddrLock */
        tfClose(socketDescriptor);
        socketDescriptor = TM_SOCKET_ERROR;
        goto userCreateTcpSocketReturn;
    }
/* TCP protocols */
    tcpVectPtr = (ttTcpVectPtr)socketPtr;
/* Make user data available to common TCP functions. */
    tcpVectPtr->tcpsUserTcpConPtr = userTcpConPtr;

/*
 * Configure socket:
 *  - disable slow start,
 *  - set selective acknowledgement option per user specification,
 *  - set window scale option per user specification,
 *  - set timestamp option per user specification,
 *  - set maximum segment size.
 */
    tcpFlagMask = (tt16Bit)~(TM_TCPF_OPTIONS | TM_TCPF_SLOW_START
                                             | TM_TCPF_PMTU_DISCOVERY);
    tcpFlagSet = 0;
    if (userTcpConPtr->uconSackOpt)
    {
        tcpFlagSet = (tt16Bit)(TM_TCPF_SEL_ACK);
    }
    if (userTcpConPtr->uconSndWndScale)
    {
        tcpFlagSet = (tt16Bit)(tcpFlagSet | TM_TCPF_WND_SCALE);
    }
    if (userTcpConPtr->uconTsOpt)
    {
        tcpFlagSet = (tt16Bit)(tcpFlagSet | TM_TCPF_TS);
    }
    tcpVectPtr->tcpsFlags = (tt16Bit)
                        ((tcpVectPtr->tcpsFlags & tcpFlagMask) | tcpFlagSet);
    tcpVectPtr->tcpsMSS = userTcpConPtr->uconMss;
    tcpVectPtr->tcpsPeerMss = userTcpConPtr->uconMss;

/*
 * Simulate a connect (send SYN) to the remote port and address.
 */
    tcpVectPtr->tcpsSockAddrUPtr = &internetAddressPtrU;
    errorCode = tfTcpOpen(tcpVectPtr, TM_TCPE_ACTIVE_OPEN);
    tcpVectPtr->tcpsSockAddrUPtr = (ttSockAddrPtrUnionPtr)0;
    if (errorCode != TM_ENOERROR)
    {
        tm_socket_checkin_unlock(socketPtr);
        socketPtr = TM_SOCKET_ENTRY_NULL_PTR;
        tfClose(socketDescriptor);
        socketDescriptor = TM_SOCKET_ERROR;
    }
    else
    {
/*
 * Simulate an accept (receive SYN-ACK) from the remote host.
 */
        tcpVectPtr->tcpsState = TM_TCPS_ESTABLISHED;
        tfTcpRcvSynOptionsSocImp(tcpVectPtr);

/*
 * Move to the established state.
 */
        tfTcpEstablishSocImp(tcpVectPtr);
    }

/* Exit socket import mode. */
    tcpVectPtr->tcpsUserTcpConPtr = (ttUserTcpConPtr)0;

/*
 * Check-in and unlock the socket and record the error code, if any.
 */
userCreateTcpSocketReturn:
    retCode = tfSocketReturn(   socketPtr, socketDescriptor,
                                errorCode, socketDescriptor  );
    *errorCodePtr = errorCode;
    return retCode;
}

/*
 * tfTcpRcvSynOptionsSocImp() function description:
 * Process received SYN options (from the peer). Initialize receive sequence
 * numbers. Also check routing entry and set MSS from peer. Initialize
 * congestion window, and slow start threshold.
 * Parameter     description
 * tcpVectPtr    Pointer to targeted TCP state vector (including socket entry)
 *
 * Return values
 * return error value if any, TM_ENOERROR otherwise.
 */
static int tfTcpRcvSynOptionsSocImp(ttTcpVectPtr tcpVectPtr)
{
    ttUserTcpConPtr userTcpConPtr;
    int             errorCode;
    tt16Bit         tcpsFlags;
    tt8Bit          option;

    userTcpConPtr = tcpVectPtr->tcpsUserTcpConPtr;
    tcpsFlags = tcpVectPtr->tcpsFlags;

/* Reset TCP options to be set by peer (assume peer has not set them) */
    tcpVectPtr->tcpsFlags = (tt16Bit)(   tcpsFlags
                                       & ((tt16Bit)~(TM_TCPF_OPTIONS)));
/* Assume minimum TCP header size */
    tcpVectPtr->tcpsHeaderSize = TM_BYTES_TCP_HDR;
#ifdef TM_TCP_SACK
    tcpVectPtr->tcpsMaxSackSqns = (tt8Bit)TM_TCP_SACK_SQNS;
#endif /* TM_TCP_SACK */
/* TM_TCPO_MSS */
    tcpVectPtr->tcpsPeerMss = userTcpConPtr->uconMss;
/* TM_TCPO_WNDSC */
    option = userTcpConPtr->uconSndWndScale;
    if (option != TM_8BIT_ZERO)
    {
        tcpVectPtr->tcpsFlags = (tt16Bit)
                                (tcpVectPtr->tcpsFlags | TM_TCPF_WND_SCALE);
        if (option > TM_TCP_MAX_WNDSC)
        {
            option = TM_TCP_MAX_WNDSC;
        }
        tcpVectPtr->tcpsSndWndScale = option;
    }
#ifdef TM_TCP_SACK
/* TM_TCPO_SACK_PERMIT */
    if (userTcpConPtr->uconSackOpt)
    {
        tcpVectPtr->tcpsFlags = (tt16Bit)
                            (tcpVectPtr->tcpsFlags | TM_TCPF_SEL_ACK);
    }
#endif /* TM_TCP_SACK */
/* TM_TCPO_TS */
    if (userTcpConPtr->uconTsOpt)
    {
        tcpVectPtr->tcpsFlags = (tt16Bit)(tcpVectPtr->tcpsFlags | TM_TCPF_TS);
/* Tcp header with time stamp option on every packet (except reset)*/
        tcpVectPtr->tcpsHeaderSize += TM_TCPO_TS_NOOP_LENGTH;
#ifdef TM_TCP_SACK
/* Maximum number of SQN SEL ACK */
        tcpVectPtr->tcpsMaxSackSqns = (tt8Bit)TM_TCP_TS_SACK_SQNS;
#endif /* TM_TCP_SACK */
/* Collect most recent time value from peer*/
        tcpVectPtr->tcpsTsRecent = userTcpConPtr->uconRcvTsRecent;
/* Time of arrival of time value from peer */
        tcpVectPtr->tcpsTsRecentAge = userTcpConPtr->uconRcvTsRecentAge;
        tcpVectPtr->tcpsTsCurrentFixup
                            = (ttS32Bit)userTcpConPtr->uconSndTsRecent
                            - (ttS32Bit)userTcpConPtr->uconSndTsRecentAge;
    }
/* Set MSS, receive window and send window to multiple of (MSS - options) */
    errorCode = tfTcpSetMssWnd(tcpVectPtr);
/* Initialize receive sequence numbers */
    tcpVectPtr->tcpsIrs = userTcpConPtr->uconIrs;
    tcpVectPtr->tcpsRcvUp = userTcpConPtr->uconRcvNxt - 1;
    tcpVectPtr->tcpsRcvNxt = userTcpConPtr->uconRcvNxt;
    tcpVectPtr->tcpsRcvAdv =
                         userTcpConPtr->uconRcvNxt + userTcpConPtr->uconRcvWnd;

    tcpVectPtr->tcpsSndWnd = userTcpConPtr->uconSndWnd;
/*
 * RFC 2001 (congestion avoidance):
 *   1.  Initialization for a given connection sets cwnd to one segment
 *      and ssthresh to 65535 bytes.
 * RFC 2581 update:
 * 1. IW (i.e. Initial Congestion Window) set to 2 MSS.
 * 2. The initial value of ssthresh MAY be arbitrarily high (for example,
 *    some implementations use the size of the advertised window), but it
 *    may be reduced in response to congestion.
 */
/* If either network is local or user turned off slow start */
    tcpVectPtr->tcpsCwnd = TM_TCP_MAX_SCALED_WND << TM_TCP_MAX_WNDSC;
/* Maximum possible send window (with Window scale added to RFC 2001) */
    tcpVectPtr->tcpsSsthresh = TM_TCP_MAX_SCALED_WND << TM_TCP_MAX_WNDSC;
    return errorCode;
}
#else /* !TM_USE_SOCKET_IMPORT */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_USE_SOCKET_IMPORT is not defined */
int tvSocimpDummy = 0;

#endif /* TM_USE_SOCKET_IMPORT */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trsocimp.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 57998    Output checksum: 51803
 */
