/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (socket)
 *
 * Filename: trsock.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trsock.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:29CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/*
 * Allocate a new Socket Index
 * It is not bound to a port number yet
 */
int socket (int family, int type, int protocol)
{
    int              errorCode;
    int              retCode;
    int              index;
    ttSocketEntryPtr socketEntryPtr;
    
    errorCode=TM_ENOERROR;
    index = TM_SOCKET_ERROR;
    socketEntryPtr = TM_SOCKET_ENTRY_NULL_PTR;
/* Check the parameters */

    if (
/* IPv4-only */
        (family != PF_INET)
        )
    {
/* per POSIX return EAFNOSUPPORT if we don't support the address family. */
        errorCode = TM_EAFNOSUPPORT;
        goto socketFinish;
    }

/* We only support Internet Protocols */
    if ( 
           ((type != SOCK_RAW) && (!(tm_treck_sock_protocols(protocol))))
        ||
/* IPv4-only */
           (protocol == IPPROTO_ICMPV6)        
        ||
/* Support only for raw, datagram and stream sockets */
           (    (type != SOCK_STREAM)
             && (type != SOCK_DGRAM)
             && (type != SOCK_RAW) )
/* If it is TCP is must be STREAM */
        || ( (protocol == IP_PROTOTCP) && (type != SOCK_STREAM) )
/* If it is UDP then it must be DATAGRAM */
        || ( (protocol == IP_PROTOUDP) && (type != SOCK_DGRAM) )
/* If it is IGMP is must be raw. */
#ifdef TM_IGMP
        || ( (protocol == IP_PROTOIGMP) && (type != SOCK_RAW) )
#endif /* TM_IGMP */
/* If it is ICMP is must be raw */
        || ( (protocol == IP_PROTOICMP) && (type != SOCK_RAW) )
        )
    {
        errorCode = TM_EPROTONOSUPPORT;
        goto socketFinish;
    }

/* If the caller did not specify a protocol to use */
    if (protocol == TM_SOC_PROTO_DEFAULT)
    {
/* For stream it is TCP */
        if (type == SOCK_STREAM)
        {
            protocol = IP_PROTOTCP;
        }
        else
        {
/* For datagram it is UDP */
            if (type == SOCK_DGRAM)
            {
                protocol = IP_PROTOUDP;
            }
            else
            {
/* For RAW it is ICMP */
                if (type == SOCK_RAW)
                {
/* IPv4-only */
                    protocol = IP_PROTOICMP;
                }
                else
                {
                    errorCode = TM_EPROTONOSUPPORT;
                    goto socketFinish;
                }
            }
        }
    }

/* Unless raw socket support is enabled, the only protocol that we support with
   IPv6 raw sockets is IPPROTO_ICMPV6 */

/* lets allocate a socket entry, and initialize protocol number 
 * owner count initialized to 1 in tfSocketAllocate() because we own
 * the socket
 */
    socketEntryPtr = tfSocketAllocate((tt8Bit)protocol, &errorCode);
 
    if (socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR)
    {
/*
 * The allocate failed either because there is no more buffers, or
 * we are exceeding the maximum number of sockets.
 */
        goto socketFinish;
    }

/* initialize the address family on the socket */
    socketEntryPtr->socProtocolFamily = (tt8Bit) family;
    if (protocol == IP_PROTOTCP)
    {
/* socket opened in connection less state. Do not allow a lookup match yet */
        socketEntryPtr->socCoFlags = (tt16Bit)
                                    (   TM_SOC_TREE_CL_FLAG
                                      | TM_SOC_TREE_NO_MATCH_FLAG);
    }
    index = tfSocketArrayInsert(socketEntryPtr);
    if ((unsigned int)index == TM_SOC_NO_INDEX)
    {
        errorCode = TM_EMFILE;
    }
    else
    {
/*
 * If an IPv4 socket is opened while IPv6 is enabled, its default address
 * should be initialized to a IPv4-mapped zero address.
 */ 
    }
socketFinish:
    {
/* If there is an error, record the socket error for TM_SOCKET_ERROR */
        retCode = tfSocketReturn( socketEntryPtr, TM_SOCKET_ERROR, errorCode, 
                                  index );
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trsock.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 52414    Output checksum:  5829
 */
