/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2011
 * Copyright Elmic Systems Japan 1997 - 2011
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (setsockopt)
 *
 * Filename: trssocko.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trssocko.c $
 *
 * Modification History
 * $Revision: 6.0.2.5 $
 * $Date: 2011/11/08 23:53:07CST $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

#ifdef TM_USE_STRONG_ESL
#include <trglobal.h>
#endif /* TM_USE_STRONG_ESL */

#ifdef TM_USE_STRONG_ESL
static int tfCheckAddressConfig (ttIpAddressPtr ipAddrPtr,
                                 ttDeviceEntryPtr devPtr);
#endif /* TM_USE_STRONG_ESL */

/*
 * Set options for a socket or protocol on a socket
 */
int setsockopt(int                socketDescriptor, 
               int                protocolLevel, 
               int                optionName,
               const char TM_FAR *optionValuePtr,
               int                optionLength)
{

    ttSocketEntryPtr socketEntryPtr;
    ttTcpVectPtr     tcpVectPtr;
#ifdef TM_USE_STRONG_ESL
    ttDeviceEntryPtr devPtr;
#endif /* TM_USE_STRONG_ESL */
    ttSockTuple      sockTuple;
    tt32Bit          rcvWndSize;
    tt32Bit          maxRcvWindow32;
    int              errorCode;
    int              retCode;
    tt16Bit          treeFlag;
#ifdef TM_USE_STRONG_ESL
#ifndef TM_SINGLE_INTERFACE_HOME
    tt16Bit          anyMhomeIndex;
#endif /* !TM_SINGLE_INTERFACE_HOME */
#endif /* TM_USE_STRONG_ESL */

/* PARAMETER CHECK */
    socketEntryPtr = tfSocketCheckValidLock(  socketDescriptor
                                           );
    {
        if (socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR)
        {
            errorCode = TM_EBADF;
            goto setsockoptFinish;
        }
        tcpVectPtr = (ttTcpVectPtr)((ttVoidPtr)socketEntryPtr);
        if (   (optionName < 0)
            || (optionValuePtr == (const char TM_FAR *)0)
            || (optionLength <= 0) )
        {
            errorCode = TM_EINVAL;
            goto setsockoptFinish;
        }
    
        errorCode = TM_ENOERROR;
    
        if (protocolLevel == SOL_SOCKET)
        {
    
/* Check the socket option for correctness and set it if it is correct */
            switch (optionName)
            {
/*
 * The following are Boolean Options so we just turn them
 * on and off
 */
/* allow local address reuse */
                case SO_REUSEADDR:      
/* No Break Because all boolean options are processed the same */
    
/* keep connections alive */
                case SO_KEEPALIVE:      
/* No Break Because all boolean options are processed the same */
    
/* leave received OOB data in line */
                case SO_OOBINLINE:      
/* No Break Because all boolean options are processed the same */
    
#ifdef TM_USE_REUSEPORT
/* allow local address & port reuse */
                case SO_REUSEPORT:      
#endif /* TM_USE_REUSEPORT */
/* No Break Because all boolean options are processed the same */

/* Allow send to broadcast address */
                case SO_BROADCAST:
/* No Break Because all boolean options are processed the same */
    
/* DSP only: user data is unpacked */
                case SO_UNPACKEDDATA:
/* No Break Because all boolean options are processed the same */
    
                    if (optionLength != (int)sizeof(int))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        if (*((int TM_FAR *)(optionValuePtr)) == 0)
                        {
/* Turn it Off */
                            socketEntryPtr->socOptions &= ~((tt16Bit)optionName);
                        }
                        else
                        {
/* Turn it On */
                            socketEntryPtr->socOptions |= ((tt16Bit)optionName);
                        }
                    }   
                    break;
    
/* linger on close if data present */
                case SO_LINGER:         
                    if (optionLength < (int)sizeof(struct linger))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        if (((struct linger TM_FAR *)(optionValuePtr))
                                                            ->l_onoff == 1)
                        {
                            socketEntryPtr->socLingerTime = (tt16Bit)
                               (((struct linger TM_FAR *)(optionValuePtr))->
                                                                    l_linger);
                            socketEntryPtr->socOptions |= SO_LINGER;
                        }
                        else
                        {
                            socketEntryPtr->socLingerTime = 0;
                            socketEntryPtr->socOptions &= ~SO_LINGER;
                        }
                    }
                    break;
    
/* send buffer size */
                case SO_SNDBUF:         
                    if (optionLength < (int)sizeof(tt32Bit))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        socketEntryPtr->socMaxSendQueueBytes = 
                            *((tt32BitPtr)(optionValuePtr));
/* Automatically adjust the low water mark */
                        socketEntryPtr->socLowSendQueueBytes =
                            tm_send_low_water(*((tt32BitPtr)(optionValuePtr)));
                    }
                    break;
    
/* receive buffer size */
                case SO_RCVBUF:         
                    if (optionLength < (int)sizeof(tt32Bit))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        rcvWndSize = *((tt32BitPtr)(optionValuePtr));
                        if (    (socketEntryPtr->socProtocolNumber ==
                                                          (tt8Bit)IP_PROTOTCP)
                             && (tcpVectPtr->tcpsState > TM_TCPS_LISTEN) )
                        {
/*
 * With TCP, after connection has started, assume we are not allowing the
 * change of the socket recv buffer. 
 * 1. We cannot allow shrinking of the receive window after the connection
 *    establishment has started. We will reset that error down below so
 *    that it is not returned to the user, so that we can silently ignore
 *    shrinking of the receive window (socketpair()).
 * 2. If the change results in a bigger recv window we will allow the
 *    change based on the TCP window scale. 
 */
                            errorCode = TM_EPERM;
                            if (rcvWndSize
                                    > socketEntryPtr->socMaxRecvQueueBytes)
/* recv window not being shrunk */
                            {
/* Allow change */
                                errorCode = TM_ENOERROR;
/*
 * Restrict the value to whatever TCP window scale is being used, but note
 * that TCP always makes that check before sending a recv window
 * adverstisement.
 */
                                if (tcpVectPtr->tcpsRcvWndScale
                                                         != TM_8BIT_ZERO)
                                {
                                    maxRcvWindow32 = TM_TCP_MAX_SCALED_WND
                                                << tcpVectPtr->tcpsRcvWndScale;
                                }
                                else
                                {
                                    maxRcvWindow32 = TM_TCP_MAX_SCALED_WND;
                                }
                                if (rcvWndSize > maxRcvWindow32)
/* 
 * Silently adjust to maximum allowed based on negotiated window scale, but
 * note that TCP will always make that check
 */
                                {
                                    rcvWndSize = maxRcvWindow32;
                                }
                            }
                        }
                        if (errorCode == TM_ENOERROR)
                        {
                            socketEntryPtr->socMaxRecvQueueBytes = rcvWndSize;
/* Automatically adjust the low water mark */
                            socketEntryPtr->socLowRecvQueueBytes =
                                                tm_recv_low_water(rcvWndSize);
                        }
                        if (errorCode == TM_EPERM)
                        {
/* Silently ignore shrinking of the receive window (socketpair()) */
                            errorCode = TM_ENOERROR;
                        }
                    }
                    break;
/* send buffer size in datagrams (non-TCP) */
                case TM_SO_SND_DGRAMS:
                    if ((optionLength < (int)sizeof(tt32Bit))
                        || (socketEntryPtr->socProtocolNumber ==
                            (tt8Bit)IP_PROTOTCP))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
/* we store this option internally as 16-bit */
                        socketEntryPtr->socMaxSendQueueDgrams =
                            (tt16Bit) *((tt32BitPtr)(optionValuePtr));
                    }
                    break;
    
/* receive buffer size in datagrams (non-TCP) */
                case TM_SO_RCV_DGRAMS:
                    if ((optionLength < (int)sizeof(tt32Bit))
                        || (socketEntryPtr->socProtocolNumber ==
                            (tt8Bit)IP_PROTOTCP))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
/* we store this option internally as 16-bit */
                        socketEntryPtr->socMaxRecvQueueDgrams =
                            (tt16Bit) *((tt32BitPtr)(optionValuePtr));
                    }
                    break;
    
/* send low-water mark */
                case SO_SNDLOWAT:       
                    if (optionLength < (int)sizeof(tt32Bit))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        socketEntryPtr->socLowSendQueueBytes = 
                            *((tt32BitPtr)(optionValuePtr));
                        
                    }
                    break;
    
/* receive low-water mark */
                case SO_RCVLOWAT:       
#ifdef TM_ERROR_CHECKING
                    if (    (optionLength < (int)sizeof(tt32Bit))
/* Do not allow low water mark bigger than maximum receive queue size */
                         || (*((tt32BitPtr)(optionValuePtr)) >
                                          socketEntryPtr->socMaxRecvQueueBytes) )
#else /* TM_ERROR_CHECKING */
                    if ( (optionLength < (int)sizeof(tt32Bit)) )
#endif /* TM_ERROR_CHECKING */
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        socketEntryPtr->socLowRecvQueueBytes = 
                            *((tt32BitPtr)(optionValuePtr));
                        
                    }
                    break;
    
/* send timeout not supported */
                case SO_SNDTIMEO:       
                    errorCode = TM_EOPNOTSUPP;
                    break;
    
/* receive timeout not supported */
                case SO_RCVTIMEO:       
                    errorCode = TM_EOPNOTSUPP;
                    break;
    
/* don't route option not supported */
                case SO_DONTROUTE:
                    errorCode = TM_EOPNOTSUPP;
                    break;

                case TM_SO_SNDAPPEND:
                    if (optionLength != (int)sizeof(unsigned int))
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        socketEntryPtr->socSendAppendThreshold =
                            (tt16Bit)
                            (*((unsigned int TM_FAR *)(optionValuePtr)));
                    }   
                    break;
                case TM_SO_RCVCOPY:
                    if (    (optionLength != (int)sizeof(unsigned int))
                         || (*((unsigned int TM_FAR*)optionValuePtr) == 0) )
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        socketEntryPtr->socRecvCopyFraction =
                            (tt16Bit)(*((unsigned int TM_FAR *)optionValuePtr));
                    }   
                    break;
#ifdef TM_USE_STRONG_ESL
                case SO_BINDTODEVICE:
                    {
/* First field in struct ifreq pointed to by optionValuePtr */
                        if (*optionValuePtr == '\0')
                        {
                            socketEntryPtr->socFlags2 &=
                                         ~TM_SOCF2_BINDTODEVICE;
                            socketEntryPtr->socRteCacheStruct.rtcDevPtr =
                                                           (ttDeviceEntryPtr)0;
#ifdef TM_SNMP_CACHE
                            socketEntryPtr->socTupleDev.tudDevPtr =
                                                           (ttDeviceEntryPtr)0;
#endif /* TM_SNMP_CACHE */
                        }
                        else
                        {
/* Convert from name to device pointer */
                            devPtr = (ttDeviceEntryPtr)
                                      tfInterfaceNameToDevice(optionValuePtr);
                            if (devPtr != (ttDeviceEntryPtr)0)
                            {
                                if (    tm_16bit_one_bit_set(
                                                socketEntryPtr->socFlags,
                                                TM_SOCF_BIND)
                                     || tm_8bit_one_bit_set(
                                                socketEntryPtr->socFlags2,
                                                TM_SOCF2_IP_SRCADDR) )
                                {
                                    errorCode = tfCheckAddressConfig(
                                            &socketEntryPtr->socOurIfIpAddress,
                                            devPtr);
                                }
                                if (errorCode == TM_ENOERROR)
                                {
                                    socketEntryPtr->socFlags2 |=
                                             TM_SOCF2_BINDTODEVICE;
                                    socketEntryPtr->socRteCacheStruct.
                                                           rtcDevPtr = devPtr;
#ifdef TM_SNMP_CACHE
                                    socketEntryPtr->socTupleDev.tudDevPtr =
                                                                       devPtr;
#endif /* TM_SNMP_CACHE */
                                }
                            }
                            else
                            {
                                errorCode = TM_ENOENT;
                            }
                        }
                    }
                    break;
#endif /* TM_USE_STRONG_ESL */
/* No Matching Option so not supported */
                default:
                    errorCode = TM_ENOPROTOOPT;
                    break;
            }  
    
        }
        else
        {
            switch (protocolLevel)
            {
/* Switch on the protocol number 6 for TCP and 17 for UDP */
                case IP_PROTOTCP:
#ifdef TM_USE_TCP
                    if (    socketEntryPtr->socProtocolNumber 
                         == (tt8Bit)IP_PROTOTCP )
                    {
                        errorCode = tfTcpSetOption( tcpVectPtr,
                                                    optionName, optionValuePtr,
                                                    optionLength);
                    }
                    else
#endif /* TM_USE_TCP */
                    {
                        errorCode = TM_EINVAL;
                    }
                    break;
/* Since IP does not have any protocol number we have a special macro */
                case IP_PROTOIP:
                    switch (optionName)
                    {
    
                        case IPO_TOS:
                            {
                                socketEntryPtr->socIpTos = 
                                  *((unsigned char TM_FAR *)(optionValuePtr));
                            }
                            break;
                        case IPO_TTL:
                            {
                                socketEntryPtr->socIpTtl = 
                                  *((unsigned char TM_FAR *)(optionValuePtr));
                            }
                            break;
    
                        case IPO_SRCADDR:
                            if (optionLength < (int)sizeof(tt4IpAddress))
                            {
                                errorCode = TM_EINVAL;
                            }
                            else
                            {
                                tm_bcopy(&(socketEntryPtr->socTuple),
                                         &sockTuple,
                                         sizeof(sockTuple));
/* This option is an IPv4 option only */
                                tm_ip_copy(
                                        *((tt4IpAddressPtr)(optionValuePtr)),
                                        tm_4_ip_addr(
                                                sockTuple.sotLocalIpAddress));
#ifdef TM_USE_STRONG_ESL
                                if (tm_8bit_one_bit_set(
                                                    socketEntryPtr->socFlags2,
                                                    TM_SOCF2_BINDTODEVICE))
                                {
                                    devPtr = socketEntryPtr->
                                                  socRteCacheStruct.rtcDevPtr;
#ifdef TM_SINGLE_INTERFACE_HOME
                                    if (  tm_8bit_bits_not_set(
                                            tm_ip_dev_conf_flag(devPtr, 0),
                                            TM_DEV_IP_CONFIG)
                                        && tm_ip_not_match(
                                              tm_4_ip_addr(
                                                 sockTuple.sotLocalIpAddress),
                                              tm_ip_dev_addr(devPtr, 0)) )
#else /* !TM_SINGLE_INTERFACE_HOME */
                                    if (tfMhomeAddrMatch(
                                             devPtr,
                                             tm_4_ip_addr(
                                                 sockTuple.sotLocalIpAddress),
                                             &anyMhomeIndex) != TM_ENOERROR )
#endif /* !TM_SINGLE_INTERFACE_HOME */
                                    {
/* could not find the specified IPv4 address on the bound interface */
                                        errorCode = TM_EINVAL;
                                    }
                                }
                                if (errorCode == TM_ENOERROR)
#endif /* TM_USE_STRONG_ESL */
                                {
                                    if (tm_16bit_one_bit_set(
                                               socketEntryPtr->socFlags,
                                               TM_SOCF_IN_TREE))
/* Socket already in the tree. Reinsert */
                                    {
                                        treeFlag = (tt16Bit) (
                                         socketEntryPtr->socCoFlags
                                       | (tt16Bit)(  socketEntryPtr->socOptions
                                                   & (  SO_REUSEADDR
                                                      | SO_REUSEPORT)) );
                                        retCode = tfSocketTreeReInsert(
                                                            socketEntryPtr,
                                                            &(sockTuple),
                                                            treeFlag );
                                        if (retCode != TM_SOC_RB_OKAY)
                                        {
                                            errorCode = TM_EADDRINUSE;
                                        }
                                    }
                                    else
/*
 * Do not insert in the tree, if socket is not in the tree yet.
 */
                                    {
                                       tm_bcopy(&sockTuple, &(socketEntryPtr->socTuple),
                                                sizeof(sockTuple));
                                    }
                                    if (errorCode == TM_ENOERROR)
                                    {
/*
 * BUG ID 203: if socket source IP address has been assigned by calling 
 * setsockopt with option IPO_SRCADDR, we should use the assigned IP
 * address.
 */
                                        socketEntryPtr->socFlags2 |= 
                                                        TM_SOCF2_IP_SRCADDR;
                                    }
                                }
                            }
                            break;
#ifdef TM_IGMP
                        case IPO_MULTICAST_TTL:
                        case IPO_MULTICAST_IF:
                        case IPO_ADD_MEMBERSHIP:
                        case IPO_DROP_MEMBERSHIP:
                        case MCAST_JOIN_GROUP:
                        case MCAST_LEAVE_GROUP:
                            errorCode = tfIgmpSetSockOpt(socketEntryPtr,
                                                         optionName,
                                                         optionValuePtr,
                                                         optionLength);
                            break;
#endif /* TM_IGMP */
                        case IPO_HDRINCL:
                            if (optionLength != (int)sizeof(int))
                            {
                                errorCode = TM_EINVAL;
                            }
                            else
                            {
                                if (    (    socketEntryPtr->socProtocolNumber
                                          != IP_PROTOUDP)
                                     && (    socketEntryPtr->socProtocolNumber
                                          != IP_PROTOTCP) )
                                {
                                    if (*((int TM_FAR *)(optionValuePtr)) == 0)
                                    {
/* Turn it Off */
                                        socketEntryPtr->socOptions
                                                            &= ~SO_IPHDRINCL;
                                    }
                                    else
                                    {
/* Turn it On */
                                        socketEntryPtr->socOptions
                                                            |= SO_IPHDRINCL;
                                    }
                                }
                                else
                                {
                                    errorCode = TM_EPERM;
                                }
                            }
                            break;
    
                        default:          
                            errorCode = TM_ENOPROTOOPT;
                            break;
                    }
                    break;
                    
                    
                default:
                    if (protocolLevel < 0)
                    {
                        errorCode = TM_EINVAL;
                    }
                    else
                    {
                        errorCode = TM_ENOPROTOOPT;
                    }
                    break;
            }
    
        }
#ifdef TM_USE_TCP
        if (socketEntryPtr->socProtocolNumber == (tt8Bit)IP_PROTOTCP)
        {
            tfTcpPurgeTcpHeaders(tcpVectPtr);
        }
#endif /* TM_USE_TCP */
setsockoptFinish:
        retCode =  tfSocketReturn(socketEntryPtr, socketDescriptor,
                                  errorCode, TM_ENOERROR);
    }
    return retCode;
}

#ifdef TM_USE_STRONG_ESL
static int tfCheckAddressConfig (ttIpAddressPtr ipAddrPtr,
                                 ttDeviceEntryPtr devPtr)

{
    tt4IpAddress ipAddress;
    int          errorCode;
#ifndef TM_SINGLE_INTERFACE_HOME
    tt16Bit      anyMhomeIndex;
#endif /* !TM_SINGLE_INTERFACE_HOME */

    errorCode = TM_ENOERROR;
/* Check that the socket source address is configured on the device */
    {
        ipAddress  = tm_4_ip_addr((*ipAddrPtr));
        if (tm_ip_not_zero(ipAddress) )
        {
#ifdef TM_SINGLE_INTERFACE_HOME
            if (  tm_8bit_bits_not_set( tm_ip_dev_conf_flag(devPtr, 0),
                                        TM_DEV_IP_CONFIG)
                && tm_ip_not_match(ipAddress, tm_ip_dev_addr(devPtr, 0)) )
#else /* !TM_SINGLE_INTERFACE_HOME */
            if (tfMhomeAddrMatch(
                    devPtr,
                    ipAddress,
                    &anyMhomeIndex) != TM_ENOERROR )
#endif /* !TM_SINGLE_INTERFACE_HOME */
            {
/* could not find the specified IPv4 address on the specified interface */
                errorCode = TM_EINVAL;
            }
        }
    }
    return errorCode;
}
#endif /* TM_USE_STRONG_ESL */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trssocko.c    File Revision: 6.0.2.5 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 22184    Output checksum:  6058
 */
