/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Virtual channel API functions
 *
 * Filename: trvchan.c
 * Author: Odile
 * Date Created: 06/14/01
 * $Source: source/sockapi/trvchan.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:52CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_VCHAN

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * tfInterfaceSetVirtualChannel function description
 * Allows the user to set an ATM virtual channel for a given multi home
 * on an interface.
 *
 * Parameters
 * Parameter        Description
 * interfaceHandle  The interface handle as returned by tfAddInterface
 * virtualChannel   ATM virtual channel set by the user.
 * mhomeIndex       Multi home index on the interface.
 *
 * Returns
 * TM_ENOERROR      Success
 * TM_EINVAL        Invalid interface handle, or multi home index.
 */
int
tfInterfaceSetVirtualChannel( ttUserInterface interfaceHandle,
                              ttUser32Bit     virtualChannel,
                              int             mhomeIndex )
{
    ttDeviceEntryPtr devEntryPtr;
    int              errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check on parameters */
    if (    ( tfValidInterface(devEntryPtr) == TM_ENOERROR )
         && ( ((unsigned int) mhomeIndex) < TM_MAX_IPS_PER_IF) )
    {
        errorCode = TM_ENOERROR;
/* Lock device driver */
        tm_call_lock_wait(&(devEntryPtr->devDriverLockEntry));
#ifdef TM_SINGLE_INTERFACE_HOME
        devEntryPtr->devVChan = virtualChannel;
#else /* !TM_SINGLE_INTERFACE_HOME */
        devEntryPtr->devVChanArray[mhomeIndex] = virtualChannel;
#endif /* TM_SINGLE_INTERFACE_HOME */
/* UnLock device driver */
        tm_call_unlock(&(devEntryPtr->devDriverLockEntry));
    }
    else
    {
        errorCode = TM_EINVAL; /* bad parameter */
    }
    return errorCode;
}

/****************************************************************************
* FUNCTION: tfNgInterfaceSetVirtualChannel
*
* PURPOSE: This function is a replacement for tfInterfaceSetVirtualChannel
*   to support IPv6.
*
* NEW PARAMETERS:
*   addrFamily:
*       Address family (either AF_INET for IPv4, or AF_INET6 for IPv6).
*
* RETURNS:
*   TM_EAFNOSUPPORT:
*       addrFamily was set to an invalid value for address family
*   TM_EINVAL:
*       Invalid interface handle, or multi home index.
*
****************************************************************************/
int tfNgInterfaceSetVirtualChannel(
    ttUserInterface interfaceHandle,
    ttUser32Bit     virtualChannel,
    int             addrFamily, 
    int             mhomeIndex )
{
    int                 errorCode;

    errorCode = TM_ENOERROR; /* assume successful */

    switch(addrFamily)
    {
    case AF_INET:
        errorCode = tfInterfaceSetVirtualChannel(
            interfaceHandle,
            virtualChannel,
            mhomeIndex );
        break;


    default:
        errorCode = TM_EAFNOSUPPORT;
        break;
    }

    return errorCode;
}

/*
 * tfInterfaceGetVirtualChannel Function Description
 * tfInterfaceGetVirtualChannel is called by the user, in the send path,
 * from the driver send routine, to get the ATM virtual channel associated
 * with the interface and multi home the data is sent from.
 * Note: this function should only be called from the device driver
 * send function, as we will check the packet currently being sent
 * in the device send queue to retrieve the associated multi-home.
 * Because of this, we also assume that the device driver is locked.
 * Parameters
 * Parameter         Description
 * interfaceHandle   The interface handle as returned by tfAddInterface,
 *                   and passed to the device driver send function.
 * virtualChannelPtr Pointer to a 32-bit variable where the virtual
 *                   channel will be stored, if the call is successful.
 * Returns
 * TM_ENOERROR       Success
 * TM_EINVAL         Invalid interface handle, or interface handle does not
 *                   match the interface handle of the packet currently
 *                   being sent.
 * TM_ENOENT         Empty send queue
 */
int
tfInterfaceGetVirtualChannel ( ttUserInterface      interfaceHandle,
                               ttUser32Bit TM_FAR * virtualChannelPtr )
{
#ifdef TM_SINGLE_INTERFACE_HOME
    ttDeviceEntryPtr devEntryPtr;
    int              errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    if (virtualChannelPtr != (tt32BitPtr)0)
    {
        errorCode = TM_ENOERROR;
        *virtualChannelPtr = devEntryPtr->devVChan;
    }
    else
    {
        errorCode = TM_EINVAL;
    }
    return errorCode;
#else /* !TM_SINGLE_INTERFACE_MHOME */
    ttDeviceEntryPtr devEntryPtr;
    ttPacketPtr      packetPtr;
    int              errorCode;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Packet currently being sent */
    packetPtr = devEntryPtr->devSendQueueCurPtr;
    if (packetPtr != TM_PACKET_NULL_PTR)
    {
        if (    (packetPtr->pktDeviceEntryPtr == devEntryPtr)
             && (virtualChannelPtr != (tt32BitPtr)0) )
        {
            switch(packetPtr->pktNetworkLayer)
            {

            default:
                *virtualChannelPtr =
                    devEntryPtr->devVChanArray[packetPtr->pktMhomeIndex];
                errorCode = TM_ENOERROR;
                break;
            }
        }
        else
        {
            errorCode = TM_EINVAL;
        }
    }
    else
    {
        errorCode = TM_ENOENT;
    }
    return errorCode;
#endif /* TM_SINGLE_INTERFACE_MHOME */
}

#else /* ! TM_USE_VCHAN */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

/* To allow link for builds when TM_USE_VCHAN is not defined */
int tvVchanDummy = 0;
#endif /* ! TM_USE_VCHAN */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trvchan.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 28955    Output checksum: 54141
 */
