/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BSD Sockets Interface (tfZeroCopyRecvFrom)
 *
 * Filename: trzrcvfr.c
 * Author: Paul & Odile
 * Date Created: 11/10/97
 * $Source: source/sockapi/trzrcvfr.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:44:57CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
    
/*
 * Take some data from the receive queue and give it to the user 
 */
/*
 * Question  What exactly is zero copy?  How is zero copy different from the 
 *  "regular" way? Does the use of zero copy functions affect how my device 
 *  driver is written? Is the zero copy facility proprietary to the Elmic stack?
 * 
 * Answer  Zero copy means that your application uses one of our buffers, so 
 *  you avoid the Elmic stack doing an extra copy from your application buffer 
 *  to our buffer in the transmit-path, or vice versa in the receive-path. In 
 *  the receive-path, nothing needs to change in your device driver. Basically, 
 *  in the receive-path you still pass a received packet to the stack in your 
 *  driverReceive function the same way as you currently do, only once the stack 
 *  queues the data on a socket to be received, calling tfZeroCopyRecv/
 *  tfZeroCopyRecvFrom on that socket avoids the extra copy from the socket 
 *  receive queue buffer to your application buffer, since instead we return 
 *  to you a pointer to our buffer (bufferHandlePtr) and a pointer to the packet 
 *  data in that buffer (dataPtrPtr) as well as the length of the received data. 
 *  Next, you need to give that buffer back to the Treck stack by calling 
 *  tfFreeZeroCopyBuffer. In the transmit-path, you need to get one of our 
 *  buffers to put your application data into before you call tfZeroCopySend/
 *  tfZeroCopySendTo, and you do that by calling tfGetZeroCopyBuffer. Once you 
 *  have made the call to tfZeroCopySend/tfZeroCopySendTo, that zero copy buffer 
 *  is no longer yours, since you have given it back to the Elmic stack. 
 * 
 *  No changes are required to your device driver to use zero copy functions. 
 * 
 * The zero copy functions are Elmic proprietary extensions to the BSD sockets 
 * API. 
 */
int tfZeroCopyRecvFrom(int                     socketDescriptor, 
                       ttUserMessage   TM_FAR *bufferHandlePtr, 
                       char   TM_FAR * TM_FAR *dataPtrPtr,
                       int                     maxDataLength,
                       int                     flags, 
                       struct sockaddr TM_FAR *fromAddressPtr,
                       int             TM_FAR *fromAddressLengthPtr)
{
    ttSocketEntryPtr    socketEntryPtr;
    ttPacketPtr         packetPtr;
    ttPacketPtr         retPacketPtr;
    ttPktLen            copyCount;
    ttPktLen            datagramLength;
    int                 errorCode;
    ttSockAddrPtrUnion  tempSockAddr;
    int                 addressLength;

    tempSockAddr.sockPtr = fromAddressPtr;
    errorCode = TM_ENOERROR;
    copyCount = (ttPktLen)0;
    retPacketPtr = TM_PACKET_NULL_PTR;
/* PARAMETER CHECK */

    if (fromAddressLengthPtr != (int TM_FAR *)0)
    {
        addressLength = *fromAddressLengthPtr;
    }
    else
    {
        addressLength = 0;
    }

/*
 * Map from socket descriptor to a locked socket entry pointer and verify
 * that the address length is at least what we expect for this address type.
 */    
    socketEntryPtr = tfSocketCheckAddrLenLock(  socketDescriptor,
                                                fromAddressPtr,
                                                addressLength,
                                                &errorCode
                                              );
    if ( socketEntryPtr == TM_SOCKET_ENTRY_NULL_PTR )
    {
/* errorCode set by tfSocketCheckAndLock */
        goto zeroCopyRecvFromFinish;
    }
    if (    (bufferHandlePtr == (ttUserMessage TM_FAR *)0)
         || (dataPtrPtr == (char TM_FAR * TM_FAR *)0)
/* we allow bufferlength between 0 and 65534 on 16-bit architecture */
         || (maxDataLength == -1)
         || (maxDataLength == 0)
         || ((flags & ~TM_ZEROCOPYRECVFROM_FLAGS) != 0) )
    {
        errorCode = TM_EINVAL;
        goto zeroCopyRecvFromFinish;
    }
    
/*Check to see if it is a TCP socket */
    if (socketEntryPtr->socProtocolNumber == IP_PROTOTCP)
    {
        errorCode = TM_EPROTOTYPE;
        goto zeroCopyRecvFromFinish;
    }
/* If user did not set any wait/dontwait flag */
    if ((flags & TM_SOCF_MSG_WAIT_FLAGS) ==  0)
    {
/* use the socket default */
        flags |= (socketEntryPtr->socFlags & TM_SOCF_NONBLOCKING);
    }
    packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
/* Check for Data */
    if (packetPtr == TM_PACKET_NULL_PTR)
    {
        errorCode = socketEntryPtr->socRecvError;
        if (errorCode != TM_ENOERROR)
        {
            socketEntryPtr->socRecvError = 0;
            goto zeroCopyRecvFromFinish;
        }
        if (flags & TM_SOCF_NONBLOCKING)
        {
            errorCode = TM_EWOULDBLOCK;
            goto zeroCopyRecvFromFinish;
        }
        else
        {
#ifdef TM_PEND_POST_NEEDED
            errorCode = tm_pend(&socketEntryPtr->socRecvPendEntry,
                                &socketEntryPtr->socLockEntry);
            if (errorCode != TM_ENOERROR)
            {
                socketEntryPtr->socRecvError = 0;
            }
            packetPtr = socketEntryPtr->socReceiveQueueNextPtr;
            /* Check for Data */
            if ((errorCode != TM_ENOERROR) || (packetPtr==TM_PACKET_NULL_PTR))
            {
                if (errorCode == TM_ENOERROR)
                {
                    errorCode = TM_EWOULDBLOCK;
                }
                goto zeroCopyRecvFromFinish;
            }
#else /* TM_PEND_POST_NEEDED */
            socketEntryPtr->socRecvError = 0;
            goto zeroCopyRecvFromFinish;
#endif /* TM_PEND_POST_NEEDED */
        }
    }
    datagramLength = packetPtr->pktChainDataLength;
    if (datagramLength > (ttPktLen)maxDataLength)
    {
        errorCode = TM_EMSGSIZE;
        if ((flags & MSG_PEEK) == 0)
/* User not peeking, trim excess data */
        {
            tfPacketTailTrim(packetPtr, (ttPktLen)maxDataLength,
                             TM_SOCKET_LOCKED);
        }
    }
    if (fromAddressPtr != TM_SOCKADDR_NULL_PTR)
    {
/* Save the from address */
        {
            *fromAddressLengthPtr = sizeof(struct sockaddr_in);
            (tempSockAddr.sockInPtr)->sin_len = 
                sizeof(struct sockaddr_in);
            (tempSockAddr.sockInPtr)->sin_family=AF_INET;
            (tempSockAddr.sockInPtr)->sin_port=
                (unsigned short)(packetPtr->pktSharedDataPtr->
                                 dataPeerPortNumber);
            tm_ip_copy(packetPtr->pktSrcIpAddress,
                       (tempSockAddr.sockInPtr)->sin_addr.s_addr);
        }
    }    
    copyCount = packetPtr->pktChainDataLength;
    if (copyCount > (ttPktLen)maxDataLength)
/* If user is peeking, we have not trimmed the excess data */
    {
        copyCount = (ttPktLen)maxDataLength;
    }
/* Check for IP Fragmented data */
    if (    (    packetPtr->pktLinkDataLength 
              != copyCount )
         && (    (flags & MSG_SCATTERED) == 0 )
       )
    {
/*
 * Because the incoming data is fragmented and we need to deliver the whole
 * datagram back to the user (to preserve datagram boundaries), because
 * the user does not want scattered data, and we are not going to copy the
 * data anyways at the end of this routine for DSP, we are forced to copy to 
 * make the data contiguous. 
 */
        retPacketPtr = tfGetSharedBuffer(TM_MAX_SEND_HEADERS_SIZE,
                                         copyCount,
                                         TM_16BIT_ZERO);
        if (retPacketPtr == TM_PACKET_NULL_PTR)
        {
            errorCode = TM_ENOBUFS;
            goto zeroCopyRecvFromFinish;
        }
        else
        {
            if ((flags & MSG_PEEK) == 0)
/* If the user is not peeking */
            {
/* Remove the packet from the receive queue, user owns the packet */
                socketEntryPtr->socReceiveQueueNextPtr=
                                               packetPtr->pktChainNextPtr;
/* Update receive queue size */
                socketEntryPtr->socRecvQueueBytes -= (tt32Bit)datagramLength;
                socketEntryPtr->socRecvQueueDgrams--;
/* User owns the packet and is allowed to free it */
                retPacketPtr->pktUserFlags = TM_PKTF_USER_OWNS;
                tfCopyPacket(packetPtr, retPacketPtr);
/* Since we copied the whole data area, and we are not peeking, free it */
                tfFreePacket(packetPtr, TM_SOCKET_LOCKED);
            }
            else
            {
                tfCopyPacket(packetPtr, retPacketPtr);
            }
        }
    }
    else
    {
/*
 * If the packet is NOT IP fragmented, or the user wants scattered data,
 * or we know we are going to copy the packet anyways at the end of 
 * this routine to unpack DSP data, then we just hand the scattered packet 
 */
        retPacketPtr = packetPtr;
        if ((flags & MSG_PEEK) == 0)
/* If the user is not peeking */
        {
/* Remove the packet from the receive queue, user owns the packet */
            socketEntryPtr->socReceiveQueueNextPtr =
                                               packetPtr->pktChainNextPtr;
/* Update receive queue size */
            socketEntryPtr->socRecvQueueBytes -=
                                             (tt32Bit)datagramLength;
            socketEntryPtr->socRecvQueueDgrams--;
/* User owns the packet and is allowed to free it */
            retPacketPtr->pktUserFlags = TM_PKTF_USER_OWNS;
        }
    }
zeroCopyRecvFromFinish:
    if ( (errorCode == TM_ENOERROR) || (errorCode == TM_EMSGSIZE) )
/* no error, or truncated */
    {

        {
#ifdef TM_LINT
LINT_NULL_PTR_BEGIN(bufferHandlePtr)
LINT_NULL_PTR_BEGIN(retPacketPtr)
LINT_NULL_PTR_BEGIN(dataPtrPtr)
#endif /* TM_LINT */
/* Return buffer handle to the user */
        *bufferHandlePtr = (ttUserMessage)retPacketPtr;
/* Return data pointer to the user */
        *dataPtrPtr = (char TM_FAR *)retPacketPtr->pktLinkDataPtr;
#ifdef TM_LINT
LINT_NULL_PTR_END(bufferHandlePtr)
LINT_NULL_PTR_END(retPacketPtr)
LINT_NULL_PTR_END(dataPtrPtr)
#endif /* TM_LINT */
        }
    }
/*
 * Return read bytes to the user:
 * If errorCode not set, will return copyCount, otherwise it will set
 * the errorCode in the user error array, and return TM_SOCKET_ERROR instead
 */
    return tfSocketReturn(socketEntryPtr, socketDescriptor,
                          errorCode, (int)copyCount);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trzrcvfr.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 23602    Output checksum:  2461
 */
