/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:     tfZeroCopyUserBufferSend() API
 * Filename:        trzsendu.c
 * Author:          Odile
 * Date Created:    11/13/02
 * $Source: source/sockapi/trzsendu.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:45:01CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h> 
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/*
 * tfZeroCopyUserBufferSend function description
 * Data Socket Send. Allows the user to send data on a
 * socket directly from user owned data buffers without having to copy
 * the data.
 * Upon return from this routine, the user no longer owns its buffer.
 * The only exception is when TM_SOCKET_ERROR is returned, and the errorCode
 * retrieved with tfGetSocketError() is TM_EWOULDBLOCK. In that case
 * the user still owns its buffer, and should try and re-send the same
 * buffer later on.
 * The userFreeFunction will be called by the Turbo Treck stack for each
 * user buffer if an error other than TM_EWOULDBLOCK occurs, or when the data
 * has been sent out on the network, and the device driver no longer need to 
 * access the data in the user buffer.
 * It is the user's responsibility to ensure that the user free function
 * is re-entrant as it could potentially be called from different threads,
 * unless the user uses a different free function per user application
 * thread, and the user does not use a Turbo Treck transmit task, and does not
 * use a Turbo Treck send (complete) task.
 * Example of tfZeroCopyUserBufferSend() usage is shown in the loop back test
 * module txscatlp.c.
 * Parameters
 * socketDescriptor     As returned by a socket() call
 * userBufferPtr        pointer to user buffer to be used 
 *                      when calling the user free function
 * userDataPtr          pointer to user data
 * userDataLength       size of user data
 * userFreeFunction     User free function called by the stack, when
 *                      the buffer is no longer accessed by the stack.
 * flags                MSG_DONTWAIT, MSG_DONTROUTE, MSG_WAITALL, MSG_OOB, 
 *                      or 0
 *
 * Return
 * > 0              Number of bytes transmitted/queued. Treck will free the
 *                  user buffers.
 * TM_SOCKET_ERROR  An error occured. The user can retrieve the error code
 *                  with the tfGetSocketError(socketDescriptor) API:
 *
 *   .. TM_EWOULDBLOCK   Not enough room to queue the data in the socket
 *                       send queue. User still owns his buffer, and is
 *                       responsible for re-sending later on.
 *   .. TM_EINVAL        One of the parameters is bad, i.e userBufferPtr,
 *                       or userDataPtr, or userDataLength, or user free
 *                       function is 0, userDataLength is -1, or flags
 *                       has a bit that is not allowed.
 *   .. TM_ENOBUFS       Not enough memory to allocate the Turbo Treck headers
 *   .. TM_EBADF         The socket descriptor is invalid
 *   .. TM_EMSGSIZE      The data length was bigger than the send queue.
 *   .. TM_EHOSTUNREACH  No route to host
 */
int tfZeroCopyUserBufferSend ( int                  socketDescriptor,
                               char TM_FAR *        userBufferPtr, 
                               char TM_FAR *        userDataPtr, 
                               int                  userDataLength, 
                               ttUserFreeFuncPtr    userFreeFunction,
                               int                  flags)
{
    ttPacketPtr      packetPtr;
    int              errorCode;
    int              retCode;
    tt8Bit           needFreePacket;

    retCode = TM_ENOERROR; /* So that errorCode is set on socket */
    packetPtr = (ttPacketPtr)0;
    needFreePacket = TM_8BIT_YES;
/* 
 * Check on user parameters 
 * (userDataLength && flags checked in tfZeroCopySend)
 */
    if (    (userDataPtr == (ttCharPtr)0) 
         || (userBufferPtr == (ttCharPtr)0)
         || (userFreeFunction == (ttUserFreeFuncPtr)0)
       )
    {
/*
 * If Bad parameters, return TM_EINVAL, free user buffer is free function and
 * userBufferPtr is non null.
 */
        errorCode = TM_EINVAL;
        goto zeroCopyUserExit;
    }
/*
 *  .. Allocate a ttPacket structure, and and empty ttSharedData, and
 *     set the TM_BUF_USER_DATA bit in the ttSharedData flag as follows:
 *     newPacketPtr = tfGetSharedBuffer(0, (ttPktLen)0, TM_BUF_USER_DATA);
 *  .. If failed to allocate, set errorCode to TM_ENOBUFS, and break out of
 *     the loop.
 *  .. Store the userFreeFunction in the ttSharedData structure
 *  .. Store the user buffer pointer in the ttSharedData structure
 *     (in the dataDeviceUserDataPtr field)
 *  .. Make the ttPacket link data pointer (pktLinkDataPtr) point to the user
 *     data pointer
 *  .. Initialize pktLinkDataLength, and pktChainDataLength with the user data
 *     length
 */
    packetPtr = tfGetSharedBuffer(0, (ttPktLen)0, TM_BUF_USER_DATA);
    if (packetPtr == (ttPacketPtr)0)
    {
        errorCode = TM_ENOBUFS;
        goto zeroCopyUserExit;
    }
    packetPtr->pktSharedDataPtr->dataUserFreeFuncPtr = userFreeFunction;
    packetPtr->pktSharedDataPtr->dataDeviceUserDataPtr = userBufferPtr;
    packetPtr->pktLinkDataPtr = (tt8BitPtr)userDataPtr;
    packetPtr->pktSharedDataPtr->dataBufFirstPtr = packetPtr->pktLinkDataPtr;
    packetPtr->pktSharedDataPtr->dataBufLastPtr = 
                            packetPtr->pktLinkDataPtr +
                            tm_packed_byte_count((unsigned)(userDataLength));
    packetPtr->pktLinkDataLength = (ttPktLen)(userDataLength);
    packetPtr->pktChainDataLength = (ttPktLen)(userDataLength);
/*
 * Initialize the pktUserFlags field of the first ttPacket with
 * TM_PKTF_USER_OWNS so that tfZeroCopySend() will
 * accept the buffer.
 */
    packetPtr->pktUserFlags = TM_PKTF_USER_OWNS;
/*
 * Tell tfZeroCopySend() that a buffer might need to be allocated for 
 * network headers 
 */
    flags |= MSG_SCATTERED;
    needFreePacket = TM_8BIT_ZERO;
    retCode = tfZeroCopySend( socketDescriptor,
                             (ttUserMessage)packetPtr,
                             userDataLength,
                             flags );
/*
 * Check return value from tfZeroCopySend()
 * If return value is TM_SOCKET_ERROR, we check the error code by calling
 * tfGetSocketError(). If error code is TM_EWOULDBLOCK, then we need to
 * reset the TM_BUF_USER_DATA flag in the shared data flags, so that the 
 * Treck free routine tfFreePacket() will not call the user free function for
 * the data block.
 */
    if (retCode == TM_SOCKET_ERROR)
    {
        errorCode = tfGetSocketError(socketDescriptor);
        if (errorCode == TM_EWOULDBLOCK)
        {
/*
 * Do not free the user buffers, but do free the ttPacket/ttShared data
 * link.
 */
            packetPtr->pktSharedDataPtr->dataFlags &= ~TM_BUF_USER_DATA;
            needFreePacket = TM_8BIT_YES;
        }
    }
    else
    {
        errorCode = TM_ENOERROR;
    }

zeroCopyUserExit:
    if (errorCode != TM_ENOERROR)
    {
        if (retCode != TM_SOCKET_ERROR)
        {
/* Record the error for socketDescriptor */
            tfSocketErrorRecord(socketDescriptor, errorCode);
            retCode = TM_SOCKET_ERROR;
        }
        if (needFreePacket != TM_8BIT_ZERO)
        {
            if (packetPtr != TM_PACKET_NULL_PTR)
            {
/*
 * Free packetPtr, and associated user buffer if needed.
 */
                tfFreeSharedBuffer(packetPtr, TM_SOCKET_UNLOCKED);
            }
            else
            {
/* Free the user buffer if it has not been associated with a ttPacket */
                if (    (userFreeFunction != (ttUserFreeFuncPtr)0)
                     && (userBufferPtr != (ttCharPtr)0) )
                {
                    (void)(*userFreeFunction)(userBufferPtr);
                }
            }
        }
    }
    return retCode;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trzsendu.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  7705    Output checksum: 39338
 */
