/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: BASE64 encoding and decoding routines.
 *
 * Filename: trbase64.c
 * Author: Bryan Wang
 * Date Created: 9/30/03
 * $Source: source/trbase64.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:47:05CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <trsocket.h>
#include <trmacro.h> /* For TM_CONST_QLF */
#include <trtype.h>
#include <trproto.h>

#define TM_BASE64_DECODE_ARRAY_SIZE 80
#define TM_BASE64_DECODE_START_CHAR 43
#define TM_BASE64_DECODE_INVAL_BYTE 255

static const char TM_CONST_QLF tlPadChar   = '=';
static const char TM_CONST_QLF tlCharSet[] = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const unsigned char TM_CONST_QLF tlDecodeArray[] =
{
    62,  255, 255, 255, 63,  52,  53,  54,
    55,  56,  57,  58,  59,  60,  61,  255,
    255, 255, 255, 255, 255, 255, 0,   1,
    2,   3,   4,   5,   6,   7,   8,   9,
    10,  11,  12,  13,  14,  15,  16,  17,
    18,  19,  20,  21,  22,  23,  24,  25,
    255, 255, 255, 255, 255, 255, 26,  27,
    28,  29,  30,  31,  32,  33,  34,  35,
    36,  37,  38,  39,  40,  41,  42,  43,
    44,  45,  46,  47,  48,  49,  50,  51
};


/* 
 * Function tfBase64OutputLen()
 * DESCRIPTION:
 *  Given a buffer and length of the buffer, compute the length of
 *  the Base64 encoded stream.
 * PARAMETERS
 *  inPtr:  The input buffer
 *  inLen:  The input buiffer len

 * RETURNS:
 *  >=0 length needed for the encoded buffer 
 *  -1  inPtr is invalid
 */
int tfBase64OutputLen(ttCharPtr inPtr, int inLen)
{
    int len;

    len = inLen;
    TM_UNREF_IN_ARG(inPtr);

/* align the length to 3 and times 4, you get the encoded length */
    len = ((len + 2) / 3) * 4;

/* for the end of each encoded line , we need to and CRLF*/
    len += (len / TM_BASE64_LINE_LEN + 2);

    return len;
}


/* 
 * Function tfBase64InputLen()
 * DESCRIPTION:
 *  Given the length of a buffer to hold a base64 encoded stream, 
 *  compute the length of input buffer that can be accomodated.
 *  Note that this cannot be done acurately, since the actuall 
 *  encode length of a buffer depends on the real content of the 
 *  input buffer. This would be a estimation at the lower end. 
 * PARAMETERS
 *  outLen:  Length of the output buffer
 * RETURNS:
 *  length of input buffer that can be encoded.
 */
int tfBase64InputLen(int outLen)
{
    int len;

    len = outLen;

/* Reserve space for holding the CRLF at the end of each encoded line */
    len -= (len / TM_BASE64_LINE_LEN + 1) * 2;

/* every 3 char is encoded as a 4 char */
    return len / 4 * 3;
}



/* 
 * Function tfBase64Encode()
 * DESCRIPTION:
 *  BASE64 encoding routing.
 * PARAMETERS
 *  inPtr:  The input buffer
 *  inLen:  The input buiffer len
 *  outPtr: The output buffer
 *  outLen: The output buffer len

 * RETURNS:
 *   >0:   successful, returns actual length of the encoded buffer
 *  <=0:   One of the pointer is NULL or the output len is not enough.
 */
int tfBase64Encode(ttCharPtr inPtr, int inLen, ttCharPtr outPtr, int outLen)
{
    ttCharPtr outBasePtr;
    ttCharPtr linePtr;
    int       i;
    int       inChar;
    int       codedLen;
   
    if (    inLen <= 0
         || outLen < tfBase64OutputLen(inPtr, inLen)
         || inPtr == (ttCharPtr)0
         || outPtr == (ttCharPtr)0 )
    {
        codedLen = -1;
    }
    else
    {
        outBasePtr = outPtr;
        linePtr = outPtr;

        for (i = 0; i < inLen; ++ i)
        {
           inChar = (inPtr[i] >> 2) & 0x3f;
           *(outPtr ++) = tlCharSet[inChar];

           inChar = (inPtr[i] << 4) & 0x3f;
           if (++i < inLen)
           {
               inChar |= (inPtr[i] >> 4) & 0x0f;
           }
           *(outPtr ++) = tlCharSet[inChar];

           if (i < inLen)
           {
               inChar = (inPtr[i] << 2) & 0x3f;
               if (++i < inLen)
                   inChar |= (inPtr[i] >> 6) & 0x03;

               *(outPtr ++) = tlCharSet[inChar];
           }
           else
           {
               ++i;
               *(outPtr ++) = tlPadChar;
           }

           if (i < inLen)
           {
               inChar = inPtr[i] & 0x3f;
               *(outPtr ++) = tlCharSet[inChar];
           }
           else
           {
               *(outPtr ++) = tlPadChar;
           }

           if (outPtr - linePtr >= TM_BASE64_LINE_LEN - 2)
           {
               *(outPtr ++) = '\r';
               *(outPtr ++) = '\n';
               linePtr = outPtr;
           }
        }
/* NULL terminate the output buffer */
        *outPtr = 0;
        codedLen = (int) (outPtr - outBasePtr);
    }

    return codedLen;
}


/* 
 * Function tfBase64Decode()
 * DESCRIPTION:
 *  Decode the string in inPtr, saving the decoded text in outPtr.
 * PARAMETERS
 *  inPtr:  Pointer to the input (encoded) string.
 *  inLen:  Number of bytes, starting at inPtr[0], to be processed.
 *  outPtr: Pointer to string where output will be written.
 *  outLen: Length of output string. Note that this length must include the
 *          null byte for string termination.
 * RETURNS:
 *  Length of decoded string (not including null string terminator).
 */
int tfBase64Decode(ttCharPtr inPtr, int inLen, ttCharPtr outPtr, int outLen)
{
    int           outBytesUsed;
    int           inBytesUsed;
    unsigned char decodedBits;
    unsigned char bitsLeft;
    char          curInByte;

    bitsLeft = 8;
    outBytesUsed = 0;
    inBytesUsed = 0;

/* Ignore any terminating '='. They're padding */
    while (inLen && inPtr[inLen-1] == tlPadChar)
    {
        inLen--;
    }

/* Make sure the target string is empty */
    tm_memset(outPtr, 0, outLen);

/* Setting bitsLeft to 8 (above) consumed the first byte of the
 * output string */
    if (outLen)
    {
        outLen--;
    }

/* As long as there's more to process and room to do it...*/
    while (inLen && outLen)
    {
/* Get the current byte to process */
        curInByte = inPtr[inBytesUsed++];
/* Make sure it's within the valid range of characters */
        if (curInByte < TM_BASE64_DECODE_START_CHAR ||
            curInByte >= TM_BASE64_DECODE_ARRAY_SIZE +
                         TM_BASE64_DECODE_START_CHAR)
        {
            break;
        }
/* Look up character in decode array */
        decodedBits = tlDecodeArray[curInByte-TM_BASE64_DECODE_START_CHAR];
/* Make sure character was valid input */
        if (decodedBits == TM_BASE64_DECODE_INVAL_BYTE)
        {
            break;
        }
/* If the byte we're outputting to doesn't have enough space... */
        if (bitsLeft < 6)
        {
/* Consume another byte in the output string */
            outLen--;
/* Write what bits we can to the current output byte */
            outPtr[outBytesUsed++] |= decodedBits >> (6 - bitsLeft);
/* Write the rest to the next output byte */
            outPtr[outBytesUsed]   |= decodedBits << (2 + bitsLeft);
/* Update bitsLeft to be accurate for the new output byte */
            bitsLeft += 2;
        }
/* Else this output byte has enough space */
        else
        {
/* Subtract the bits used */
            bitsLeft -= 6;
/* Save the bits to the output byte */
            outPtr[outBytesUsed] |= decodedBits << bitsLeft;
/* If we've filled the output byte, get the next one */
            if (!bitsLeft)
            {
                outBytesUsed++;
                outLen--;
                bitsLeft = 8;
            }
        }
/* Mark that we've consumed an input byte */
        inLen--;
    }

/* If we decoded the whole thing, return successfully */
    if (inLen)
    {
        outBytesUsed = -1;
    }

    return outBytesUsed;
}


/* 
 * Function tfBase64DecodedLen()
 * DESCRIPTION:
 *  Calculate the length of the string created by decoding the given
 *  encoded string.
 * PARAMETERS
 *  encodedPtr:  Pointer to the input (encoded) string.
 *  encodedLen:  Number of bytes (not counting null terminator) in the
 *               encoded string.
 * RETURNS:
 *  Length of decoded string (not including null string terminator).
 */
int tfBase64DecodedLen(ttCharPtr encodedPtr, int encodedLen)
{
    int decodedLen;

    decodedLen = (encodedLen * 3)/4;
/* Account for pad chars at the end of the string */
    while (decodedLen && encodedPtr[encodedLen-1] == tlPadChar)
    {
        decodedLen--;
        encodedLen--;
    }
    return (decodedLen);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trbase64.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 41015    Output checksum: 36331
 */
