/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: PPP/SLIP Dialer
 *
 * Filename: trdialer.c
 * Author: Jason
 * Date Created: 11/10/97
 * $Source: source/trdialer.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:47:16CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 *  This is a module designed for dialing up to an ISP to establish a PPP/SLIP
 *  connection.  It includes basic 'scripting' support which allows the user
 *  to enter any number of send-expect pairs along with timeout values and
 *  retry counts. 
 *
 *  Upon initiliazation (tfUseDialer), the dialer allocates memory for its
 *  state structure, and attaches this to the device entry.  The existence of
 *  this entry signals the device interface that the user wants to call the 
 *  dialer before opening the link layer.  Thus when the user calls 
 *  tfConfigInteface, instead of calling the link layer open function 
 *  (e.g. tfPppClientOpen) the dialer open (tfDialerOpen) is called, which 
 *  kicks off the dialing sequence.  When the dialer is opening, it 
 *  temporarily sets the link layer recieve function to itself (tfDialerRecv),
 *  so any incoming data is diverted to the dialer.  When the dialer has 
 *  completed, this function is returned to its previous value 
 *  (e.g. tfPppAsyncRecv).
 *
 *  When the dialer is opened, it determines what its first action should be:
 *  either to wait for an expect string, or send a string.  When this action 
 *  is completed, we check the other string in this state, and take the 
 *  necessary action.  When this is done, we move to the next state (if any) 
 *  and repeat this process.  If there are no further states, then the dialer 
 *  has completed succesfully, so we notify the user and call the link layer 
 *  open. In the case that the dialer has failed (has recieved an error 
 *  string, for instance) we notify the user and close the device.
 */ 

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#if    defined(TM_LOCK_NEEDED) \
    || (    defined(TM_ERROR_CHECKING) \
         && defined(TM_USE_ANSI_LINE_FILE)  \
         &&  (   defined(TM_KERNEL_VISUAL_X86) \
              || defined(TM_KERNEL_WIN32_X86) \
              || defined(TM_MEMORY_LOGGING) ) )
#include <trglobal.h>
#endif /*    TM_LOCK_NEEDED 
          || (    TM_ERROR_CHECKING
               && TM_USE_ANSI_LINE_FILE 
               && (    TM_KERNEL_VISUAL_X86
                    || TM_KERNEL_WIN32_X86
                    || TM_MEMORY_LOGGING ) ) */


#define TM_DIALER_CLOSING 1

static void tfDialerFreeBuffersTimers(ttDialerInfoPtr dialInfoPtr);

/* Recieve data from the device (the dialer acts as pseudo-link layer) */
TM_NEARCALL int tfDialerRecv(void TM_FAR *interfaceId, void TM_FAR *bufferHandle);

/* Send data to the device */
void tfDialerSend(ttDeviceEntryPtr devEntryPtr);

/* Called when a send timeout occurs */
void tfDialerTimeout (ttVoidPtr      timerBlockPtr,
                      ttGenericUnion userParm1,
                      ttGenericUnion userParm2);

/* Opens the dialer and starts the dialing sequence */
int tfDialerOpen(void TM_FAR * interfaceHandle,
                 tt16Bit       multiHomeIndex,
                 int           protocolFamily);
            
/* Process the current dialer state, and preforms the necessary action */                 
int tfDialerProcess(ttDeviceEntryPtr devEntryPtr);

/*
 * Close the dialer, notify the user, and either start the link layer (for
 * sucessful dialing) or close the device (for failure).
 */
int tfDialerClose(void TM_FAR *interfaceHandle);


/*
 * Calls tfDialerAddSendExpect with different flags 
 */
int tfDialerAddExpectSend(ttUserInterface interfaceId,
                          char TM_FAR *   sendString,
                          char TM_FAR *   expectString,
                          char TM_FAR *   errorString,
                          int             numRetries,
                          int             timeout,
                          unsigned char   flags)
{
    int errorCode;

    errorCode = tfDialerAddSendExpect(interfaceId,
                                      sendString,
                                      expectString,
                                      errorString,
                                      numRetries,
                                      timeout,
                                      (unsigned char) 
                                              (flags | TM_DIALER_EXPECT_FIRST));

    return errorCode;
}                          

/*
 * tfDialerAddSendExpect Function Description:
 *   Called by the user to add a 'send-expect' pair.  This must be called 
 *   *before* tfConfigInterface (and tfDialerOpen) is called.  This routine
 *   adds a new ttDialerInfo structure to the linked list of send-expect 
 *   pairs.
 *
 * Internals
 *  1. Check parameters.
 *  2. Allocate memory for info structure and associated strings.
 *  3. Set values in info structure
 *  4. Attach new dialer info structure to interface
 *
 * Parameter         Description
 *  interfaceHandle   interface to dial on
 *  sendString        string to send to device
 *  expectString      string from device indicating success
 *  errorString       string from device indicating failure
 *  numRetries        number of times to retry sending 'sendString' befor 
 *                     giving up.
 *  timeout           number of seconds to wait between retries.
 *  flags             TM_DIALER_FAIL_ON_ERROR: if this flag is set, if the 
 *                      dialer recieves errorString the dialing sequence is
 *                      terminated.  If this flag is not set, if the dialer
 *                      recieves errorString, it simply goes to the next
 *                      send/expect pair. 
 *    
 *
 * Returns
 * Value       Meaning
 *  TM_ENOMEM   Not enough memory to allocate new send-expect entry
 *  TM_EINVAL   sendString and expectString not specified
 *  TM_EINVAL   invalid timeout/retry values (numRetries < 0 OR timeout <= 0)
 *  TM_EPERM    Dialer not yet initialized.
 */
int tfDialerAddSendExpect(ttUserInterface interfaceId,
                          char TM_FAR *   sendString,
                          char TM_FAR *   expectString,
                          char TM_FAR *   errorString,
                          int             numRetries,
                          int             timeout,
                          unsigned char   flags)
{
    ttDialerInfoPtr dialInfoPtr;
    ttDialerInfoPtr newDialInfoPtr;
    tt16Bit         sendLength, expectLength, errorLength;
    tt16Bit         length;
    int             errorCode;

#define devEntryPtr ((ttDeviceEntryPtr) interfaceId)
    
    errorCode = TM_ENOERROR;

/*  1. Check parameters. */
    if ( ((sendString == (char *)0) && (expectString == (char *)0)) ||
         (timeout == 0) || 
         (devEntryPtr == (ttDeviceEntryPtr)0) )
    {
        errorCode = TM_EINVAL;
        goto errorReturn;
    }

/* Dialer must be initialized first */
    if (devEntryPtr->devDialerEntryPtr == (ttDialerEntryPtr) 0)
    {
        errorCode = TM_EPERM;
        goto errorReturn;
    }

    tm_call_lock_wait(&(devEntryPtr->devLockEntry));
    
/*  2. Allocate memory for info structure and associated strings. */
    if (expectString != (char *)0)
    {
        expectLength = (tt16Bit) (tm_strlen(expectString) + 1);
    }
    else
    {
        expectLength = 0;
    }

    if (sendString != (char *)0)
    {
        sendLength = (tt16Bit) (tm_strlen(sendString) + 1);        
    }
    else
    {
        sendLength = 0;
    }    

    if (errorString != (char *)0) 
    {
        errorLength = (tt16Bit)(tm_strlen(errorString) + 1);
    }
    else
    {
        errorLength=0;
    }
    
    length = (tt16Bit) (sizeof(ttDialerInfo) + sendLength + expectLength + 
        errorLength);
    newDialInfoPtr = (ttDialerInfoPtr)tm_get_raw_buffer(length);
    if (newDialInfoPtr == TM_DIALER_INFO_NULL_PTR)
    {
        errorCode = TM_ENOMEM;
    }
    else
    {
/*  3. Set values in info structure  */

/* Clear the structure */
        tm_bzero( newDialInfoPtr, length );

/* Set string pointer values, and copy strings into new buffer */
/* Copy send string */
        if (sendLength > 0)
        {
            newDialInfoPtr->dialiSendStringPtr =             
                (tt8BitPtr) newDialInfoPtr + sizeof(ttDialerInfo);
            tm_strcpy(newDialInfoPtr->dialiSendStringPtr,sendString);    
        }
        else
        {
            newDialInfoPtr->dialiSendStringPtr = TM_8BIT_NULL_PTR;            
        }

/* Copy expect string */
        if (expectLength > 0)
        {
            newDialInfoPtr->dialiExpectStringPtr = 
                (tt8BitPtr) newDialInfoPtr + sizeof(ttDialerInfo) + 
                sendLength;            
            tm_strcpy(newDialInfoPtr->dialiExpectStringPtr,expectString);
        }
        else
        {
            newDialInfoPtr->dialiExpectStringPtr = TM_8BIT_NULL_PTR;
        }

/* Copy error string */
        if (errorLength > 0) 
        {
            newDialInfoPtr->dialiErrorStringPtr = 
                (tt8BitPtr) newDialInfoPtr + sizeof(ttDialerInfo) + 
                sendLength + expectLength;
            tm_strcpy(newDialInfoPtr->dialiErrorStringPtr,errorString);    
        }
        else
        {
            newDialInfoPtr->dialiErrorStringPtr = TM_8BIT_NULL_PTR;
        }        

/* Set timeout, retry, flag values */            
        newDialInfoPtr->dialiNumRetries = (tt8Bit) numRetries;
        newDialInfoPtr->dialiTimeout = (tt8Bit) timeout;
        newDialInfoPtr->dialiFlags = flags;
        newDialInfoPtr->dialiExpectMatch = TM_8BIT_ZERO;
        newDialInfoPtr->dialiErrorMatch = TM_8BIT_ZERO;
        newDialInfoPtr->dialiNextEntryPtr = TM_DIALER_INFO_NULL_PTR;
        
/*  4. Attach new dialer info structure to device entry  */
        if (devEntryPtr->devDialerEntryPtr->dialInfoPtr != 
                                            TM_DIALER_INFO_NULL_PTR)
        {
            dialInfoPtr = devEntryPtr->devDialerEntryPtr->dialInfoPtr;
            /* Find end of chain */
            while (dialInfoPtr->dialiNextEntryPtr != 
                TM_DIALER_INFO_NULL_PTR) 
            {
                dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
            }
            dialInfoPtr->dialiNextEntryPtr = newDialInfoPtr;
        }
        else
        {
            devEntryPtr->devDialerEntryPtr->dialInfoPtr = newDialInfoPtr;
        }
        
    }

    tm_call_unlock(&(devEntryPtr->devLockEntry));
    
#undef devEntryPtr
errorReturn:
    return errorCode;
}

/*
 * tfDialerInit Function Description:
 *   Initializes and enables the dialer.
 *
 * Internals
 *  1. Allocate memory for the dialer state structure and attach to device 
 *     entry 
 *  2. Set notification function in dialer entry.
 *  3. Initialize info (send-req pairs) pointer to null 
 *  4. Set success flag to failed until we successfully dial
 *  5. Set protocol family to zero - this indicates that tfDialerOpen has not
 *     been called and is used to prevent the user from attempting to open
 *     the dialer again before it has completed (needed for IPv4/IPv6 mode)
 *
 * Parameter         Description
 *   interfaceId     interface to dial on
 *   notifyFuncPtr   user supplied function to callback when dial operation
 *                   has completed.  May be NULL.
 *
 * Returns
 * Value        Meaning
 * TM_ENOERROR  Dialer sucessfully initialized
 * TM_EINVAL    Bad interface handle
 * TM_EALREADY  Dialer already initialized
 */
int tfUseDialer(ttUserInterface        interfaceId, 
                ttUserLnkNotifyFuncPtr notifyFuncPtr)
{
    int              errorCode;
    ttDeviceEntryPtr devEntryPtr;
    ttDialerEntryPtr dialerEntryPtr;
    ttDialerInfoPtr  dialInfoPtr;

    devEntryPtr = (ttDeviceEntryPtr)interfaceId;

    if (interfaceId != (ttUserInterface) 0)
    {
        tm_call_lock_wait(&(devEntryPtr->devLockEntry));
        
        dialerEntryPtr = devEntryPtr->devDialerEntryPtr;
        if (dialerEntryPtr != TM_DIALER_ENTRY_NULL_PTR)
        {
/*
 * The dialer has already been initialized, so we should reset the state.
 * This involves free all of the send expect pairs, and freeing the
 * dialer state vector.
 */ 
            dialInfoPtr = dialerEntryPtr->dialInfoPtr;
            tfDialerFreeBuffersTimers(dialInfoPtr);

/* Free dialer entry */
            tm_free_raw_buffer((ttRawBufferPtr)dialerEntryPtr);
            devEntryPtr->devDialerEntryPtr = TM_DIALER_ENTRY_NULL_PTR;
        }

        {

/*  
 * 1. Allocate memory for the dialer state structure and attach to device 
 *    entry 
 */                     
            devEntryPtr->devDialerEntryPtr = (ttDialerEntryPtr)
                                     tm_get_raw_buffer(sizeof(ttDialerEntry));

            if (devEntryPtr->devDialerEntryPtr == TM_DIALER_ENTRY_NULL_PTR)
            {
                errorCode = TM_ENOMEM;
            }
            else
            {
                dialerEntryPtr = devEntryPtr->devDialerEntryPtr;

/*  2. Set notification function in dialer vector. */
                dialerEntryPtr->dialNotifyFuncPtr = notifyFuncPtr;
    
/*  3. Initialize info (send-req pairs) pointer to null */
                dialerEntryPtr->dialInfoPtr = TM_DIALER_INFO_NULL_PTR;
        
/* 4. Set success flag to failed until we successfully dial */
                dialerEntryPtr->dialSuccess = TM_8BIT_ZERO;

                dialerEntryPtr->dialOpenFuncPtr = tfDialerOpen;
                
/*                
 *  5. Set protocol family to zero - this indicates that tfDialerOpen has not
 *     been called and is used to prevent the user from attempting to open
 *     the dialer again before it has completed (needed for IPv4/IPv6 mode)
 */
                dialerEntryPtr->dialProtocolFamily = 0;

                
                errorCode = TM_ENOERROR;
            }
        }

        tm_call_unlock(&(devEntryPtr->devLockEntry));
    }
    else
    {
/* Bad interface handle */
        errorCode = TM_EINVAL;
    }

    return errorCode;
}                 


/*
 * tfDialerProcess
 *
 * Internals
 *  1. Find first active dialer entry
 *  2. If we have a string to send, send it now 
 *
 * Parameter         Description
 *   devEntryPtr     interface to dial on
 *
 * Returns
 * Value        Meaning
 * TM_ENOERROR  Dialer sucessfully initialized
 * TM_EINVAL    Bad interface handle
 * TM_EALREADY  Dialer already initialized 
 */
int tfDialerProcess(ttDeviceEntryPtr devEntryPtr)
{
    ttDialerInfoPtr dialInfoPtr;
    tt8Bit          timeoutSecs;
    int             errorCode;
    ttGenericUnion  timerParm1;

    errorCode = TM_ENOERROR;

/*  1. Find first active dialer entry */
    dialInfoPtr = 
        (ttDialerInfoPtr) devEntryPtr->devDialerEntryPtr->dialInfoPtr;
    while ((dialInfoPtr != TM_DIALER_INFO_NULL_PTR) &&
           (dialInfoPtr->dialiFlags & TM_DIALER_COMPLETE)) 
    {
        dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
    }

    if (dialInfoPtr == TM_DIALER_INFO_NULL_PTR)
    {
/* ERROR!  Traversed entire list with no incomplete entries */    
        tfKernelWarning(
            "tfDialerSend",
            "Corrupted send-expect string list #1");
        (void)tfDialerClose(devEntryPtr);
        errorCode = TM_EALREADY;
    }
    else
    {
/*  2. If we have a string to send, send it now */
        if (!(dialInfoPtr->dialiFlags & TM_DIALER_EXPECT_FIRST) ||
            (dialInfoPtr->dialiExpectStringPtr == TM_8BIT_NULL_PTR))
        {
            tfDialerSend(devEntryPtr);

/* Only timeout if we're waiting for something */
            if (dialInfoPtr->dialiExpectStringPtr != TM_8BIT_NULL_PTR)
            {        
                timeoutSecs = dialInfoPtr->dialiTimeout;
                timerParm1.genVoidParmPtr = (ttVoidPtr)devEntryPtr;
                dialInfoPtr->dialiTimerPtr = 
                    tfTimerAdd(tfDialerTimeout,
                               timerParm1,
                               timerParm1, /* unused */
                               timeoutSecs*TM_MSECS_PER_SECOND,
                               0);
            }
        }
    }

    return errorCode;
}

/*
 * tfDialerRecv Function Description:
 *
 * Internals
 *   1. If string from device matches expect string, move to next send-expect
 *      pair.
 *     a. If this was the last send-expect pair, indicate to user/stack that
 *        dialing has been successful, and the link-layer can now start up.
 *     b. If this is a 'expect-send' pair, send the specified string.
 *   2. If string from device matches error string:
 *     a. If TM_DIALER_FAIL_ON_ERROR is set, tear down the dialer and return
 *         to the user.
 *     b. If TM_DIALER_FAIL_ON_ERROR is not set, continue to the next
 *        send-expect pair.
 *
 * Parameter         Description
 *  interfaceId     
 *  bufferHandle    
 *
 * Returns
 * Value       Meaning
 *
 */
int tfDialerRecv(void TM_FAR *interfaceId, void TM_FAR *bufferHandle)
{ 
    ttDeviceEntryPtr devEntryPtr;
    tt8Bit           sourceLength;
    tt8BitPtr        sourceStringPtr;
    ttDialerInfoPtr  dialInfoPtr;
    tt8Bit           srcOffset;
    tt8Bit           matchStrLength;

    devEntryPtr = (ttDeviceEntryPtr) interfaceId;


/* Lock the device entry */
    tm_call_lock_wait(&(devEntryPtr->devLockEntry));

    if (devEntryPtr->devDialerEntryPtr == (ttDialerEntryPtr) 0)
    {
        goto DIALER_RECV_END;
    }

/* Get current entry */

    dialInfoPtr = 
        (ttDialerInfoPtr) devEntryPtr->devDialerEntryPtr->dialInfoPtr;
    while ((dialInfoPtr != TM_DIALER_INFO_NULL_PTR) &&
           (dialInfoPtr->dialiFlags & TM_DIALER_COMPLETE)) 
    {
        dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
    }

    if (dialInfoPtr == TM_DIALER_INFO_NULL_PTR)
    {
/* ERROR!  Traversed entire list with no incomplete entries */    
        tfKernelWarning(
            "tfDialerSend",
            "Corrupted send-expect string list #2");
        (void)tfDialerClose(devEntryPtr);
    }
    else
    {

        sourceLength = (tt8Bit) ((ttPacketPtr) bufferHandle)->pktLinkDataLength;
        sourceStringPtr = ((ttPacketPtr) bufferHandle)->pktLinkDataPtr;

        srcOffset=0;
        while (srcOffset < sourceLength)
        {

/*
 * Check to see if current character matches the current character in the 
 * expect string.  If it doesn't, reset the expect string counter.
 */

            if (dialInfoPtr->dialiExpectStringPtr != TM_8BIT_NULL_PTR)
            {
                if (
                    sourceStringPtr[srcOffset]
                    == dialInfoPtr->dialiExpectStringPtr
                    [dialInfoPtr->dialiExpectMatch])
                {
                    dialInfoPtr->dialiExpectMatch++;
                }
                else 
                {
                    dialInfoPtr->dialiExpectMatch=0;
                }
            }    

/*
 * Check to see if current character matches the current character in the 
 * error string.  If it doesn't, reset the expect string counter.
 */       
            if (dialInfoPtr->dialiErrorStringPtr != TM_8BIT_NULL_PTR)
            {
                if (
                    sourceStringPtr[srcOffset]
                    == dialInfoPtr->dialiErrorStringPtr
                    [dialInfoPtr->dialiErrorMatch])
                {
                    dialInfoPtr->dialiErrorMatch++;
                }                
                else
                {
                    dialInfoPtr->dialiErrorMatch=0;        
                }
            }    

/* If the expect string counter equals the length of the expect string, it
 * means we've matched the string, so do post expect processing 
 */    
            if (dialInfoPtr->dialiExpectStringPtr != TM_8BIT_NULL_PTR)
            {
                matchStrLength =
                    (tt8Bit) tm_strlen(dialInfoPtr->dialiExpectStringPtr);
            }
            else
            {
                matchStrLength = 0;
            }
            if (dialInfoPtr->dialiExpectMatch == matchStrLength &&
                matchStrLength > 0)
            {
        
                if ((dialInfoPtr->dialiFlags & TM_DIALER_EXPECT_FIRST) &&
                        (dialInfoPtr->dialiSendStringPtr != TM_8BIT_NULL_PTR))
                {
                    tfDialerSend(devEntryPtr);

/* 
 * There's a chance that we just sent our last string, and the dialing 
 * sequence is complete.  Check, and if it is, exit this routine.
 */
                    if (devEntryPtr->devDialerEntryPtr == 
                        TM_DIALER_ENTRY_NULL_PTR)
                    {
                        goto DIALER_RECV_END;
                    }



                }
        
/*   1. If string from device matches expect string, move to next send-expect
 *      pair.
 */
                if (dialInfoPtr->dialiNextEntryPtr != TM_DIALER_INFO_NULL_PTR)
                {


/* Mark this entry as completed */
                    dialInfoPtr->dialiFlags |= TM_DIALER_COMPLETE;

                    if (dialInfoPtr->dialiTimerPtr != TM_TMR_NULL_PTR)
                    {
                        tm_timer_remove(dialInfoPtr->dialiTimerPtr);
                        dialInfoPtr->dialiTimerPtr = TM_TMR_NULL_PTR;
                    }

/* Transmit next sequence */
                    (void)tfDialerProcess(devEntryPtr);

/* Move to next dialer entry */
                    dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
                }   
                else
                {
/*        
 *     a. If this was the last send-expect pair, indicate to user/stack that        
 *        dialing has been successful, and the link-layer can now start up.
 */
                    devEntryPtr->devDialerEntryPtr->dialSuccess = TM_8BIT_YES;

                    (void)tfDialerClose(devEntryPtr);

                    break;
                }
    
/* Reset the expect string counter */
                dialInfoPtr->dialiExpectMatch=0;
            }

/* If the error string counter equals the length of the error string, it
 * means we've matched the string, so do post error processing 
 */
            if (dialInfoPtr->dialiErrorStringPtr != TM_8BIT_NULL_PTR)
            {
                matchStrLength =
                    (tt8Bit) tm_strlen(dialInfoPtr->dialiErrorStringPtr);
            }
            else
            {
                matchStrLength = 0;
            }
            if ((dialInfoPtr->dialiErrorMatch == matchStrLength) &&
                (matchStrLength > 0))
            {

/*   2. If string from device matches error string:  */

                if (dialInfoPtr->dialiFlags & TM_DIALER_FAIL_ON_ERROR)
                {
/*     a. If TM_DIALER_FAIL_ON_ERROR is set, tear down the dialer and return
 *         to the user.
 */
                    devEntryPtr->devDialerEntryPtr->dialSuccess=TM_8BIT_ZERO;
                    (void)tfDialerClose(devEntryPtr);
                    break;                                 
                }
                else
                {
/*     b. If TM_DIALER_FAIL_ON_ERROR is not set, continue to the next
 *        send-expect pair.
 */            
    
                    if (dialInfoPtr->dialiNextEntryPtr != 
                                                      TM_DIALER_INFO_NULL_PTR)
                    {
/* Mark this entry as completed */
                        dialInfoPtr->dialiFlags |= TM_DIALER_COMPLETE;

/* Remove retransmission timer */
                        if (dialInfoPtr->dialiTimerPtr != TM_TMR_NULL_PTR)
                        {
                            tm_timer_remove(dialInfoPtr->dialiTimerPtr);
                            dialInfoPtr->dialiTimerPtr = TM_TMR_NULL_PTR;
                        }

/* Process next dialer entry */
                        (void)tfDialerProcess(devEntryPtr);

/* Move to next dialer entry */
                        dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
                    }
                    else
                    {
/* 
 * Even though FAIL_ON_ERROR isn't set, this is the last entry, so mark that
 * the dial process has failed and tear down the dialer.
 */
                        devEntryPtr->devDialerEntryPtr->dialSuccess=TM_8BIT_ZERO;
                        (void)tfDialerClose(devEntryPtr);
                        break;                 
                    }    
                }
       
/* Reset the error string counter */
                dialInfoPtr->dialiErrorMatch=0;        
            }        
    
            srcOffset++;
        }

    }

DIALER_RECV_END:
/* Unlock the device entry */
    tm_call_unlock(&(devEntryPtr->devLockEntry));

    tfFreePacket((ttPacketPtr)bufferHandle, TM_SOCKET_UNLOCKED);

    return TM_ENOERROR;
}


/*
 * tfDialerSend Function Description:
 *   Transmits a send sequence to the device.
 *
 * Internals
 *   1. Find first entry in the send-expect list that is not completed (ie, 
 *      does not have its TM_DIALER_COMPLETE bit set)
 *   2. Send associated data to the driver.
 *   3. If sequence is expect->send OR expect string is empty, move on to the
 *      the next pair.
 *
 * Parameter         Description
 *  interfaceId     
 *
 * Returns
 * Value       Meaning
 *
 */
void tfDialerSend(ttDeviceEntryPtr devEntryPtr)
{

    ttDialerInfoPtr dialInfoPtr;
    tt16Bit         length;
    tt16Bit         sendCount;

/*   1. Find first entry in the send-expect list that is not completed (ie, 
 *      does not have its TM_DIALER_COMPLETE bit set)
 */
    dialInfoPtr = devEntryPtr->devDialerEntryPtr->dialInfoPtr;
    while ((dialInfoPtr != TM_DIALER_INFO_NULL_PTR) &&
           (dialInfoPtr->dialiFlags & TM_DIALER_COMPLETE)) 
    {
        dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
    }

    if (dialInfoPtr == TM_DIALER_INFO_NULL_PTR)
    {
/* ERROR!  Traversed entire list with no incomplete entries */    
        tfKernelWarning(
            "tfDialerSend",
            "Corrupted send-expect string list #3");
        (void)tfDialerClose(devEntryPtr);
    }    
    else
    {

/*   2. Send associated data to the driver.  */
        length = (tt16Bit) tm_strlen(dialInfoPtr->dialiSendStringPtr);
        sendCount=0;

        if (length > 0)
        {
/* Lock the device driver */
            tm_call_lock_wait(&(devEntryPtr->devDriverLockEntry));
            while ((sendCount+1) < length)
            {
                (void)(*(devEntryPtr->devSendFuncPtr))
                    (devEntryPtr,
                     (char TM_FAR *) dialInfoPtr->dialiSendStringPtr+sendCount,
                     1,
                     TM_USER_BUFFER_MORE);
                sendCount++;
            }

            (void)(*(devEntryPtr->devSendFuncPtr))
                    (devEntryPtr,
                     (char TM_FAR *) dialInfoPtr->dialiSendStringPtr+sendCount,
                     1,
                    TM_USER_BUFFER_LAST);

            tm_call_unlock(&(devEntryPtr->devDriverLockEntry));
        }



    
/*   3. If sequence is expect->send OR expect string is empty, move on to the
 *      the next pair.
 */
        if ((dialInfoPtr->dialiFlags & TM_DIALER_EXPECT_FIRST) ||
            (dialInfoPtr->dialiExpectStringPtr == TM_8BIT_NULL_PTR))
        {
            dialInfoPtr->dialiFlags |= TM_DIALER_COMPLETE;    

            if (dialInfoPtr->dialiNextEntryPtr != TM_DIALER_INFO_NULL_PTR)
            {
                (void)tfDialerProcess(devEntryPtr);
            }
            else
            {
                devEntryPtr->devDialerEntryPtr->dialSuccess = TM_8BIT_YES;
                (void)tfDialerClose(devEntryPtr);
            }

        }        
    }

} 


/*
 * tfDialerTimeout Function Description:
 *   Called when a transmit timeout occurs.
 *
 * Internals
 *   1. Decrement current send-expect retry count.
 *   2. Check retry count:
 *     a. If zero, we've run out of retries.  Somehow ?????? notify 
 *        user/stack that dialing has failed.
 *     b. If not zero, call tfDialerSend to resend the current send sequence
 *        and set new timer.
 *
 * Parameter         Description
 *  timerArgPtr
 *
 * Returns
 * Value       Meaning
 *
 */
void tfDialerTimeout (ttVoidPtr      timerBlockPtr,
                      ttGenericUnion userParm1,
                      ttGenericUnion userParm2)
{
    ttDeviceEntryPtr devEntryPtr;
    ttDialerInfoPtr  dialInfoPtr;
    tt8Bit           timeoutSecs;
    ttGenericUnion   timerParm1;

/* Avoid compiler warnings about unused parameters */
    TM_UNREF_IN_ARG(userParm2);
    TM_UNREF_IN_ARG(timerBlockPtr);
    
    devEntryPtr = (ttDeviceEntryPtr) userParm1.genVoidParmPtr;

    if (devEntryPtr->devDialerEntryPtr == (ttDialerEntryPtr) 0)
    {
        goto dialerTimeoutExit;
    }

/* Lock the device entry for single threaded access to the dialer */
    tm_call_lock_wait(&(devEntryPtr->devLockEntry));

/* Get current entry */
    dialInfoPtr = devEntryPtr->devDialerEntryPtr->dialInfoPtr;
    while ((dialInfoPtr != TM_DIALER_INFO_NULL_PTR) &&
           (dialInfoPtr->dialiFlags & TM_DIALER_COMPLETE)) 
    {
        dialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
    }

    if (dialInfoPtr == TM_DIALER_INFO_NULL_PTR)
    {
/* ERROR!  Traversed entire list with no incomplete entries */    
        tfKernelWarning(
            "tfDialerSend",
            "Corrupted send-expect string list #4");
        (void)tfDialerClose(devEntryPtr);
    }
    else
    {
    
/*   1. Decrement current send-expect retry count.  */
        dialInfoPtr->dialiNumRetries--;

/*   2. Check retry count:  */
        if (dialInfoPtr->dialiNumRetries == 0)
        {
/*     a. If zero, we've run out of retries.  Notify     
 *        user/stack that dialing has failed.
 */
            devEntryPtr->devDialerEntryPtr->dialSuccess = TM_8BIT_ZERO;
            (void)tfDialerClose(devEntryPtr);     
        }
        else
        {
/*     b. If not zero, call tfDialerSend to resend the current send sequence    
 *        and set new timer.
 */
            tfDialerSend(devEntryPtr);
            timeoutSecs = dialInfoPtr->dialiTimeout;
            timerParm1.genVoidParmPtr = (ttVoidPtr)devEntryPtr;
            dialInfoPtr->dialiTimerPtr=
                tfTimerAdd(tfDialerTimeout,
                           timerParm1,
                           timerParm1, /* unused */
                           timeoutSecs*TM_MSECS_PER_SECOND,
                           0);        
        }
    }

/* Unlock the device entry */
    tm_call_unlock(&(devEntryPtr->devLockEntry));

dialerTimeoutExit:
    return;
}


/*
 * tfDialerOpen Function Description:
 *   Called by tfConfigInterface, so the device driver's open routine should
 *   have been called (we can communicate with the device) but the link
 *   layer open would not have been called (e.g., PPP won't interfere with 
 *   us).
 *
 * Internals
 *   1. If the protocol family has already been set in the dialer entry,
 *      this indicates that the dialer has already been opened and is in
 *      the process of dialing.  Return TM_EPERM.
 *   2. Set current send-expect entry to first entry in list
 *   3. Transmit first send/expect sequence.
 *   4. Start transmit timeout timer.
 *
 * Parameter         Description
 *  interfaceHandle     
 *
 * Returns
 * Value       Meaning
 *  TM_EINVAL   bad interface handle pointer
 *
 */
int tfDialerOpen(void TM_FAR * interfaceId,
                 tt16Bit       mHomeIndex,
                 int           protocolFamily)
{
    tt8Bit              timeoutSecs;
    ttDialerInfoPtr     dialInfoPtr;
    ttLinkLayerEntryPtr linkLayerPtr;
    ttDeviceEntryPtr    devEntryPtr;
    ttDialerEntryPtr    dialEntryPtr;
    int                 errorCode;
    ttGenericUnion      timerParm1;


    mHomeIndex=mHomeIndex;
    errorCode = TM_EINPROGRESS;
    devEntryPtr = (ttDeviceEntryPtr) interfaceId;

/* No need to lock here, already locked in tfConfigInterface */
    dialEntryPtr = devEntryPtr->devDialerEntryPtr;
    
    if (dialEntryPtr->dialProtocolFamily == 0)
    {
/* Save the protocol family (IPv4, IPv6) in the dialer entry */
        dialEntryPtr->dialProtocolFamily = protocolFamily;
    
/* 
 * Save the link layer recieve routine, and set the LL recieve routine to 
 * point at the dialer 
 */
        linkLayerPtr = devEntryPtr->devLinkLayerProtocolPtr;
        dialEntryPtr->dialLinkRecvFuncPtr = linkLayerPtr->lnkRecvFuncPtr;
        linkLayerPtr->lnkRecvFuncPtr = tfDialerRecv;

        dialEntryPtr->dialLinkCloseFuncPtr = linkLayerPtr->lnkCloseFuncPtr;
        linkLayerPtr->lnkCloseFuncPtr = tfDialerClose;

        dialEntryPtr->dialCloseFlag = TM_8BIT_ZERO;

        dialInfoPtr = dialEntryPtr->dialInfoPtr;    
    
        if (!(dialInfoPtr->dialiFlags & TM_DIALER_EXPECT_FIRST) ||
            (dialInfoPtr->dialiExpectStringPtr == TM_8BIT_NULL_PTR))
        {
/*   1. Transmit first send/expect sequence.  */
            tfDialerSend(devEntryPtr);
    
/*   2. Start transmit timeout timer.  */
            if (dialInfoPtr->dialiExpectStringPtr != TM_8BIT_NULL_PTR)
            {
                timeoutSecs = dialInfoPtr->dialiTimeout;
                timerParm1.genVoidParmPtr = (ttVoidPtr) devEntryPtr;
                dialInfoPtr->dialiTimerPtr = 
                    tfTimerAdd(tfDialerTimeout,
                               timerParm1,
                               timerParm1, /* unused */
                               timeoutSecs*TM_MSECS_PER_SECOND,
                               0);
            }        
        }
    }
    else
    {
        errorCode = TM_EPERM;
    }
        
    return errorCode;
}

/*
 * Frees a chain of dialer info structs (send/expect pairs).
 * This routine does not free the basic dialer state vector, so
 * this should be done elsewhere.
 */ 
static void tfDialerFreeBuffersTimers(ttDialerInfoPtr dialInfoPtr)
{
    ttDialerInfoPtr nextDialInfoPtr;
    
    while (dialInfoPtr != TM_DIALER_INFO_NULL_PTR) 
    {

/* Remove timer */    
        if (dialInfoPtr->dialiTimerPtr != TM_TMR_NULL_PTR)
        {
            tm_timer_remove(dialInfoPtr->dialiTimerPtr);
            dialInfoPtr->dialiTimerPtr = TM_TMR_NULL_PTR;
        }

/* Remove strings and info block */
        nextDialInfoPtr = dialInfoPtr->dialiNextEntryPtr;
        tm_free_raw_buffer((ttRawBufferPtr) dialInfoPtr);

            
        dialInfoPtr = nextDialInfoPtr;
    }
}

/*
 * tfDialerClose Function Description:
 *
 * Internals
 *   1. Free all associated memory and timers.
 *   2. Check if dialer is still enabled.
 *     a. If sucessful, call link layer open routine and notify user.
 *     b. If not sucessful, close device and notify user.
 *
 * Parameter         Description
 *  interfaceHandle     
 *
 * Returns
 * Value       Meaning
 *  TM_EINVAL   bad interface handle pointer
 *
 */
int tfDialerClose(void TM_FAR * interfaceId)
{
    ttDialerInfoPtr     dialInfoPtr;
    ttLinkLayerEntryPtr linkLayerPtr;
    ttDialerEntryPtr    dialerEntryPtr;
    ttDeviceEntryPtr    devEntryPtr;

    devEntryPtr = (ttDeviceEntryPtr) interfaceId;


    dialerEntryPtr = devEntryPtr->devDialerEntryPtr;
    linkLayerPtr = devEntryPtr->devLinkLayerProtocolPtr;

    if (dialerEntryPtr == (ttDialerEntryPtr) 0)
    {
        goto DIALER_CLOSE_END;
    }

    if (dialerEntryPtr->dialCloseFlag != TM_DIALER_CLOSING)
    {

        dialerEntryPtr->dialCloseFlag = TM_DIALER_CLOSING;

/*   1. Free all associated memory and timers.  */
        dialInfoPtr = dialerEntryPtr->dialInfoPtr;
        tfDialerFreeBuffersTimers(dialInfoPtr);

/*
 * Since the dialer cannot use the normal 'tfDeviceSend' mechanism, but rather
 * sends data directly to the driver, it is necessary to clear the send 
 * complete counter once the dialer has completed.  This is to avoid any future
 * calls to tfSendCompleteInterface from attempting to free packets that were
 * never placed in the send queue.
 */
        devEntryPtr->devSendCompltCount = 0;

    
/*   2. Check if dialer is still enabled.  */
        if (dialerEntryPtr->dialSuccess == TM_8BIT_YES)
        {
/*     a. If enabled, call link layer open routine and notify user.  */

/* notify user */
            if (dialerEntryPtr->dialNotifyFuncPtr != (ttLnkNotifyFuncPtr) 0) 
            {
                (*(dialerEntryPtr->dialNotifyFuncPtr))
                    (interfaceId,TM_DIALER_OPEN_COMPLETE);
            }

/* Restore link layer recieve pointer */
            linkLayerPtr->lnkRecvFuncPtr = dialerEntryPtr->dialLinkRecvFuncPtr;
            linkLayerPtr->lnkCloseFuncPtr = 
                dialerEntryPtr->dialLinkCloseFuncPtr;
 /* 
  * open regular link layer (typically PPP,SLIP) 
  */
            (void)(*(devEntryPtr->devLinkLayerProtocolPtr->lnkOpenFuncPtr))
                (interfaceId, 0, dialerEntryPtr->dialProtocolFamily);
            
        }
        else
        {
/*     b. If not enabled, close device and notify user.    */ 
            if (dialerEntryPtr->dialNotifyFuncPtr != (ttLnkNotifyFuncPtr) 0)
            {
                (*(dialerEntryPtr->dialNotifyFuncPtr))
                    (interfaceId,TM_DIALER_OPEN_FAILED);
            }

            (void)tfDeviceClose(devEntryPtr);

/* Restore link layer recieve pointer */
            linkLayerPtr->lnkRecvFuncPtr = dialerEntryPtr->dialLinkRecvFuncPtr;
            linkLayerPtr->lnkCloseFuncPtr = 
                dialerEntryPtr->dialLinkCloseFuncPtr;
        }

/* Free dialer entry */
        tm_free_raw_buffer((ttRawBufferPtr) dialerEntryPtr);
        devEntryPtr->devDialerEntryPtr = TM_DIALER_ENTRY_NULL_PTR;

        dialerEntryPtr->dialCloseFlag = TM_8BIT_ZERO;
    }

DIALER_CLOSE_END:
    
    return TM_ENOERROR;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trdialer.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 19574    Output checksum: 57201
 */
