/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Ethernet Link Layer
 *
 * Filename: trether.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: source/trether.c $
 *
 * Modification History
 * $Revision: 6.0.2.5 $
 * $Date: 2011/07/25 15:12:19CDT $
 * $Author: pcarney $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


int tfEtherSend(ttPacketPtr packetPtr);
int tfEtherIoctl(int operation);
int tfEtherError(int type);


/*
 * Used to initialize the link layer in the link layer list
 */

ttUserLinkLayer tfUseEthernet(void)
{
    ttLinkLayerEntryPtr linkLayerEntryPtr;
    int                 strcmpResult;

/* Search for Ethernet Protocol in Link Layer List */

/* Lock the Link Layer list */
    tm_call_lock_wait(&tm_context(tvLinkLayerListLock));
    linkLayerEntryPtr = tm_context(tvLinkLayerListPtr);
/* Check for duplicate */
    while (linkLayerEntryPtr != TM_LINK_NULL_PTR)
    {
        strcmpResult =
            tm_strcmp((char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),
                      "ETHDIX");
        if (strcmpResult == TM_STRCMP_EQUAL)
        {
            goto useEthernetUnlock; /* entry already in */
        }
        linkLayerEntryPtr = linkLayerEntryPtr->lnkNextPtr;
    }
/* if none exists then malloc an entry */
    linkLayerEntryPtr = (ttLinkLayerEntryPtr)tm_kernel_malloc(
                                          (unsigned)sizeof(ttLinkLayerEntry));
    if (linkLayerEntryPtr==TM_LINK_NULL_PTR)
    {
/* if the malloc fails then return a null pointer */
        goto useEthernetUnlock;
    }

/*
 * The default MTU for an Ethernet interface should continue to be 1500 for
 * both IPv4 and IPv6.  [RFC2464].R2:10
 */
    linkLayerEntryPtr->lnkMtu = 1500;

    linkLayerEntryPtr->lnkOpenFuncPtr = tfLinkOpen;
    linkLayerEntryPtr->lnkCloseFuncPtr = tfEtherCommonClose;
    linkLayerEntryPtr->lnkSendFuncPtr = tfEtherSend;
    linkLayerEntryPtr->lnkRecvFuncPtr = tfEtherRecv;
    linkLayerEntryPtr->lnkIoctlFuncPtr = tfEtherIoctl;
    linkLayerEntryPtr->lnkErrorFuncPtr = tfEtherError;
    linkLayerEntryPtr->lnkMcastFuncPtr = tm_ether_mcast_func;
    linkLayerEntryPtr->lnkNotifyFuncPtr = TM_LL_NOTIFY_FUNC_NULL_PTR;
    linkLayerEntryPtr->lnkPostFuncPtr = tfDeviceSendOnePacket;
    linkLayerEntryPtr->lnkHeaderLength = TM_ETHER_HEADER_SIZE;
    linkLayerEntryPtr->lnkLinkLayerProtocol = TM_LINK_LAYER_ETHERNET;
    linkLayerEntryPtr->lnkPhysAddrLen = TM_ETHER_ADDRESS_LENGTH;
    tm_ether_copy( tvEtherBroadcast.enuAddress,
                   linkLayerEntryPtr->lnkPhysBroadAddr );
    tm_strcpy((char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),"ETHDIX");

/* ARP for IPv4... */
    linkLayerEntryPtr->lnkResolveFuncPtr = tfArpResolve;
    linkLayerEntryPtr->lnkSendRequestFuncPtr =
        (ttLnkSendRequestFuncPtr) tfArpSendRequest;
    linkLayerEntryPtr->lnkResolveCleanFuncPtr = tfArpResolveClean;


/* Stuff it into the table */
    linkLayerEntryPtr->lnkNextPtr = tm_context(tvLinkLayerListPtr);
    tm_context(tvLinkLayerListPtr) = linkLayerEntryPtr;
useEthernetUnlock:
    tm_call_unlock(&tm_context(tvLinkLayerListLock));
    return((ttUserLinkLayer)linkLayerEntryPtr);
/* return the link layer pointer */
}

/*
 * Create the header for an packet destined for ethernet
 */
int tfEtherSend(ttPacketPtr packetPtr)
{
    ttEtherHeaderPtr etherHeaderPtr;
    int              padding;
    int              retCode;


    retCode = TM_LL_OKAY;

/* Put the ethernet header in front of the packet */
    packetPtr->pktLinkDataLength += TM_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength += TM_ETHER_HEADER_SIZE;
    packetPtr->pktLinkDataPtr -= TM_ETHER_HEADER_SIZE;

/* Min size met? */
    if (packetPtr->pktChainDataLength < TM_ETHER_MIN_PACKET)
    {
/*
 * We need to pad with zeroes
 */
        padding =  TM_ETHER_MIN_PACKET - (int)packetPtr->pktChainDataLength;
        if (    (packetPtr->pktChainDataLength == packetPtr->pktLinkDataLength)
             && ( (  packetPtr->pktSharedDataPtr->dataBufLastPtr
                       - packetPtr->pktLinkDataPtr) >= TM_ETHER_MIN_PACKET )
           )
/*
 * Contiguous buffer (i.e. non TCP packet, or TCP packet with data copied at
 * the end of the header), and allocated packet has room to add the padding.
 */
        {
            tm_bzero(   packetPtr->pktLinkDataPtr
                      + packetPtr->pktLinkDataLength,
                      padding);
            packetPtr->pktLinkDataLength = TM_ETHER_MIN_PACKET;
            packetPtr->pktChainDataLength = TM_ETHER_MIN_PACKET;
        }
        else
        {
/*
 * Scattered buffer, or allocated packet too short for padding:
 *  add an extra link for the padding initialized with zeroes.
 * Note that if the device cannot handle the extra link,
 * tfDeviceSend() will copy to a new buffer. No need to worry about copying
 * here.
 */
            retCode = tfPacketTailAdd(packetPtr, padding);
        }
    }
    if (!(packetPtr->pktSharedDataPtr->dataFlags & TM_BUF_HDR_RECYCLED))
    {
        etherHeaderPtr = (ttEtherHeaderPtr)(packetPtr->pktLinkDataPtr);

/* Copy the device address into the packet */
        tm_ether_copy( (packetPtr->pktDeviceEntryPtr->devPhysAddrArr),
                       (etherHeaderPtr->ethSrcAddr));
/* Copy the destination address into the packet */
        tm_ether_copy( (packetPtr->pktEthernetAddress),
                       (etherHeaderPtr->ethDstAddr));
/* Set the packet type */
        if (packetPtr->pktNetworkLayer == TM_NETWORK_IPV4_LAYER)
        {
            etherHeaderPtr->ethType = TM_ETHER_IP_TYPE;
        }
        else
        {
            {
                if (packetPtr->pktNetworkLayer == TM_NETWORK_ARP_LAYER)
                {
                    etherHeaderPtr->ethType = TM_ETHER_ARP_TYPE;
                }
                else
                {
/* Should never happen */
                    retCode = TM_EPROTONOSUPPORT;
                }
            }
        }
    }
    return (retCode);
}

/*
 * Process an incoming Ethernet
 * We convert from the exposed USER API functions to the
 * Internal form since this is called from the device code
 */
int tfEtherRecv(void TM_FAR *interfaceId, void TM_FAR *bufferHandle)
{
    ttEtherHeaderPtr        etherHeaderPtr;
    ttPacketPtr             packetPtr;
    ttDeviceEntryPtr        devEntryPtr;
    tt8BitPtr               pktEtherAddressPtr;
    tt8BitPtr               etherHeaderAddressPtr;
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
    tt16BitPtr              typePtr;
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
#ifdef TM_DEV_RECV_OFFLOAD
    ttDevRecvOffloadPtr     pktDevRecvOffloadPtr;
#endif /* TM_DEV_RECV_OFFLOAD */
    int                     flags;
    int                     retCode;
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
	int                     tagCount;
    tt16Bit                 type;
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */

    retCode = TM_ENOERROR;
    packetPtr = (ttPacketPtr)bufferHandle;
    devEntryPtr = (ttDeviceEntryPtr)interfaceId;
/* Store the device the packet came from */
    packetPtr->pktDeviceEntryPtr = devEntryPtr;
#ifdef TM_USE_DRV_SCAT_RECV
/* If Ethernet header is not contiguous, make it contiguous */
    if (tm_pkt_hdr_is_not_cont(packetPtr, TM_ETHER_HEADER_SIZE))
    {
        packetPtr = tfContiguousSharedBuffer(packetPtr, TM_ETHER_HEADER_SIZE);
        if (packetPtr == (ttPacketPtr)0)
        {
            retCode = TM_ENOBUFS;
            goto etherRecvExit;
        }
    }
#endif /* TM_USE_DRV_SCAT_RECV */

/* write the packet into the pcap file */
#ifdef TM_USE_SNIFF
#ifdef TM_KERNEL_WIN32_X86
    if(packetPtr != TM_PACKET_NULL_PTR)
    {
        tfPcapWritePacket(packetPtr);
    }
#endif /* TM_KERNEL_WIN32_X86 */
#endif /* TM_USE_SNIFF */


/* Check the type field */
    etherHeaderPtr = (ttEtherHeaderPtr)(packetPtr->pktLinkDataPtr);
#ifdef TM_DEV_RECV_OFFLOAD
    pktDevRecvOffloadPtr = (ttDevRecvOffloadPtr)packetPtr->pktDevOffloadPtr;
    if (    (pktDevRecvOffloadPtr != (ttDevRecvOffloadPtr)0)
         && (   pktDevRecvOffloadPtr->devoRFlags
              & (TM_DEVOR_PACKET_RECV_CHKSUM_COMPUTED ) ) )
    {
        tfDevoRecvPeelChecksum(packetPtr, (tt16BitPtr)etherHeaderPtr,
                               (tt16Bit)TM_ETHER_HEADER_SIZE);
    }
#endif /* TM_DEV_RECV_OFFLOAD */
    packetPtr->pktLinkDataPtr += TM_ETHER_HEADER_SIZE;
    packetPtr->pktLinkDataLength -= TM_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength -= TM_ETHER_HEADER_SIZE;
/* Copy the source ethernet address into the packet */
    pktEtherAddressPtr = (tt8BitPtr)packetPtr->pktEthernetAddress;
    etherHeaderAddressPtr = (tt8BitPtr)etherHeaderPtr->ethSrcAddr;
    tm_ether_copy(etherHeaderAddressPtr, pktEtherAddressPtr);
/*
 * Look at the ethernet destination address to check for destination
 * Ethernet broadcast/multicast.
 */
    etherHeaderAddressPtr = (tt8BitPtr)etherHeaderPtr->ethDstAddr;
    if (tm_ether_broadcast(etherHeaderAddressPtr))
    {
        flags = TM_LL_BROADCAST_FLAG;
    }
    else
    {
        if (tm_ether_mcast(etherHeaderAddressPtr))
        {
            flags = TM_LL_MULTICAST_FLAG;
        }
        else
        {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            devEntryPtr->devInUcastPkts++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            flags = TM_LL_UNICAST_FLAG;
        }
    }

#ifdef TM_CHECK_TAGGED_ETHER_HEADER
    tagCount = -1;
	typePtr = &etherHeaderPtr->ethType;
    do
    {
        type = *typePtr;
        switch (type)
#else /* TM_CHECK_TAGGED_ETHER_HEADER */
        switch (etherHeaderPtr->ethType)
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
        {
            case TM_ETHER_ARP_TYPE:
                packetPtr->pktFlags = 0;
                retCode = tfArpIncomingPacket(packetPtr);
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
                type = TM_16BIT_ZERO; /* exit do while loop */
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
                break;

            case TM_ETHER_IP_TYPE:
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
                if (flags != TM_LL_UNICAST_FLAG)
                {
                    if (flags == TM_LL_BROADCAST_FLAG)
                    {
                        tm_64Bit_incr(
                            devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInBcastPkts);
                        tm_64Bit_incr(
                            tm_context(tvDevIpv4Data).ipIfStatsHCInBcastPkts);
                    }
                    else
                    {
                        tm_64Bit_incr(
                            devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInMcastPkts);
                        tm_64Bit_incr(
                             tm_context(tvDevIpv4Data).ipIfStatsHCInMcastPkts);
                    }
                }
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
                packetPtr->pktFlags = (tt16Bit)flags;
                tfIpIncomingPacket(packetPtr
                                  );
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
                type = TM_16BIT_ZERO; /* exit do while loop */
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
                break;

            default:
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
                if ((type != TM_16BIT_ZERO) && (tagCount != 0))
/* First time through the loop, check for Tagged Ethernet header */
                {
#ifdef TM_USE_DRV_SCAT_RECV
/* If Ethernet header is not contiguous, make it contiguous */
                    if (tm_pkt_hdr_is_not_cont(packetPtr,
                                               2*sizeof(tt16Bit)))
                    {
                        packetPtr = tfContiguousSharedBuffer(packetPtr,
                                                         2*sizeof(tt16Bit));
                        if (packetPtr == (ttPacketPtr)0)
                        {
                            retCode = TM_ENOBUFS;
                            goto etherRecvExit;
                        }
                    }
#endif /* TM_USE_DRV_SCAT_RECV */
#ifdef TM_DEV_RECV_OFFLOAD
                    pktDevRecvOffloadPtr = (ttDevRecvOffloadPtr)
                                                 packetPtr->pktDevOffloadPtr;
                    if (    (pktDevRecvOffloadPtr != (ttDevRecvOffloadPtr)0)
                         && (   pktDevRecvOffloadPtr->devoRFlags
                              & (TM_DEVOR_PACKET_RECV_CHKSUM_COMPUTED ) ) )
                    {
                        tfDevoRecvPeelChecksum(packetPtr,
                                         (tt16BitPtr)packetPtr->pktLinkDataPtr,
                                         (tt16Bit)(2*sizeof(tt16Bit)));
                    }
#endif /* TM_DEV_RECV_OFFLOAD */
/* Skip over tagged type and control info */
                    typePtr = typePtr + 2;
                    tagCount++;
/* Skip over tagged control info, and eth type */
                    packetPtr->pktLinkDataPtr += 2*sizeof(tt16Bit);
                    packetPtr->pktLinkDataLength -= 2*sizeof(tt16Bit);
                    packetPtr->pktChainDataLength -= 2*sizeof(tt16Bit);
                }
                else
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
                {
/* Packet Not For TRECK */
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
                    type = TM_16BIT_ZERO; /* exit do while loop */
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
#ifdef TM_SNMP_MIB
                    devEntryPtr->dev4Ipv4Mib.ipIfStatsInUnknownProtos++;
                    tm_context(tvDevIpv4Data).ipIfStatsInUnknownProtos++;
#endif /* TM_SNMP_MIB */
                    tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
                    retCode = TM_EPROTONOSUPPORT;
                }
                break;
        }
#ifdef TM_CHECK_TAGGED_ETHER_HEADER
    } while (type);
#endif /* TM_CHECK_TAGGED_ETHER_HEADER */
#ifdef TM_USE_DRV_SCAT_RECV
etherRecvExit:
#endif /* TM_USE_DRV_SCAT_RECV */

    return(retCode);
}

int tfEtherIoctl(int operation)
{
    TM_UNREF_IN_ARG(operation);
    return TM_LL_OKAY;
}

int tfEtherError(int type)
{
    TM_UNREF_IN_ARG(type);
    return TM_LL_OKAY;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:47      For: Xilinx-Samsung
 * File Name: trether.c     File Revision: 6.0.2.5 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 10115    Output checksum: 48072
 */
