/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: VLAN Tagged Ethernet Link Layer
 *
 * Filename: trethtag.c
 * Author: Tim
 * Date Created: 09/08/2004
 * $Source: source/trethtag.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:47:23CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_VCHAN
/* Virtual channels must be enabled to use VLAN tagging options. */
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


int tfEtherSendTagged(ttPacketPtr packetPtr);
int tfEtherRecvTagged(void TM_FAR *interfaceId, void TM_FAR *bufferHandle);
int tfEtherIoctlTagged(int operation);
int tfEtherErrorTagged(int type);

/* 
 * Used to initialize the link layer in the link layer list
 */
 
ttUserLinkLayer tfUseEthernetTagged(void)
{
    ttLinkLayerEntryPtr linkLayerEntryPtr;

/* Search for Ethernet Protocol in Link Layer List */

/* Lock the Link Layer list */
    tm_call_lock_wait(&tm_context(tvLinkLayerListLock));
    linkLayerEntryPtr = tm_context(tvLinkLayerListPtr);
/* Check for duplicate */
    while (linkLayerEntryPtr != TM_LINK_NULL_PTR)
    {
        if (tm_strcmp( (char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),
                       "ETHTAG") == TM_STRCMP_EQUAL)
        {
            goto useEthernetUnlock; /* entry already in */
        }
        linkLayerEntryPtr = linkLayerEntryPtr->lnkNextPtr;
    }
/* if none exists then malloc an entry */
    linkLayerEntryPtr = (ttLinkLayerEntryPtr)tm_kernel_malloc(
                                         (unsigned)sizeof(ttLinkLayerEntry));
    if (linkLayerEntryPtr==TM_LINK_NULL_PTR)
    {
/* if the malloc fails then return a null pointer */
        goto useEthernetUnlock;
    }

/*
 * The default MTU for an Ethernet interface should continue to be 1500 for
 * both IPv4 and IPv6.  [RFC2464].R2:10
 */ 
    linkLayerEntryPtr->lnkMtu = 1500;
    
    linkLayerEntryPtr->lnkOpenFuncPtr = tfLinkOpen;
    linkLayerEntryPtr->lnkCloseFuncPtr = tfEtherCommonClose;
    linkLayerEntryPtr->lnkSendFuncPtr = tfEtherSendTagged; 
    linkLayerEntryPtr->lnkRecvFuncPtr = tfEtherRecvTagged;
    linkLayerEntryPtr->lnkIoctlFuncPtr = tfEtherIoctlTagged;
    linkLayerEntryPtr->lnkErrorFuncPtr = tfEtherErrorTagged;
    linkLayerEntryPtr->lnkMcastFuncPtr = tm_ether_mcast_func;
    linkLayerEntryPtr->lnkNotifyFuncPtr = TM_LL_NOTIFY_FUNC_NULL_PTR;
    linkLayerEntryPtr->lnkPostFuncPtr = tfDeviceSendOnePacket;
    linkLayerEntryPtr->lnkHeaderLength = TM_TAGGED_ETHER_HEADER_SIZE;
    linkLayerEntryPtr->lnkLinkLayerProtocol = TM_LINK_LAYER_ETHER_TAG;
    linkLayerEntryPtr->lnkPhysAddrLen = TM_ETHER_ADDRESS_LENGTH;
    tm_ether_copy(tvEtherBroadcast.enuAddress,
                  linkLayerEntryPtr->lnkPhysBroadAddr);
    tm_strcpy((char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),"ETHTAG");

/* ARP for IPv4... */    
    linkLayerEntryPtr->lnkResolveFuncPtr = tfArpResolve;
    linkLayerEntryPtr->lnkResolveCleanFuncPtr = tfArpResolveClean;
    linkLayerEntryPtr->lnkSendRequestFuncPtr =
        (ttLnkSendRequestFuncPtr) tfArpSendRequest;
    
    
/* Stuff it into the table */
    linkLayerEntryPtr->lnkNextPtr = tm_context(tvLinkLayerListPtr);
    tm_context(tvLinkLayerListPtr) = linkLayerEntryPtr;
useEthernetUnlock:
    tm_call_unlock(&tm_context(tvLinkLayerListLock));
    return((ttUserLinkLayer)linkLayerEntryPtr);    
/* return the link layer pointer */
}

/*
 * Create the tagged header for an packet destined for ethernet
 */
int tfEtherSendTagged(ttPacketPtr packetPtr)
{
    ttTaggedEtherHeaderPtr  etherTaggedHeaderPtr;
    int                     padding;
    int                     retCode;
    tt16Bit                 vlanInfo;
    tt16Bit                 sTemp;
    tt32Bit                 myTag;
    ttDeviceEntry           *devPtr;
    
    retCode = TM_LL_OKAY;
    vlanInfo = 0;

    devPtr = packetPtr->pktDeviceEntryPtr;

/* Put the ethernet header in front of the packet */
    packetPtr->pktLinkDataLength += TM_TAGGED_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength += TM_TAGGED_ETHER_HEADER_SIZE;

    packetPtr->pktLinkDataPtr -= TM_TAGGED_ETHER_HEADER_SIZE;

/* Min size met? */
    if (packetPtr->pktChainDataLength < TM_ETHER_MIN_PACKET)
    {
/* 
 * We need to pad with zeroes
 */
        padding =  TM_ETHER_MIN_PACKET - (int)packetPtr->pktChainDataLength;
        if (    (packetPtr->pktChainDataLength == packetPtr->pktLinkDataLength)
             && ( (  packetPtr->pktSharedDataPtr->dataBufLastPtr
                       - packetPtr->pktLinkDataPtr) >= TM_ETHER_MIN_PACKET )
           )
/* 
 * Contiguous buffer (i.e. non TCP packet, or TCP packet with data copied at
 * the end of the header), and allocated packet has room to add the padding.
 */
        {
            tm_bzero(   packetPtr->pktLinkDataPtr
                      + packetPtr->pktLinkDataLength,
                      padding);
            packetPtr->pktLinkDataLength = TM_ETHER_MIN_PACKET;
            packetPtr->pktChainDataLength = TM_ETHER_MIN_PACKET;
        }
        else
        {
/*
 * Scattered buffer, or allocated packet too short for padding:
 *  add an extra link for the padding initialized with zeroes.
 * Note that if the device cannot handle the extra link,
 * tfDeviceSend() will copy to a new buffer. No need to worry about copying
 * here.
 */
            retCode = tfPacketTailAdd(packetPtr, padding);
        }
    }
    if (!(packetPtr->pktSharedDataPtr->dataFlags & TM_BUF_HDR_RECYCLED))
    {
        etherTaggedHeaderPtr = 
                        (ttTaggedEtherHeaderPtr)(packetPtr->pktLinkDataPtr);
        
/* Copy the device address into the packet */
        tm_ether_copy( (devPtr->devPhysAddrArr),
                    (etherTaggedHeaderPtr->ethSrcAddr));
/* Copy the destination address into the packet */
        tm_ether_copy( (packetPtr->pktEthernetAddress),
                    (etherTaggedHeaderPtr->ethDstAddr));
/* Device Virtual Channel  */
        myTag = devPtr->devVChanArray[TM_VLAN_MHOME_INDEX];
/* Copy the VLAN Tag Control Info field into the packet */
/* This should be the 16 LSBs of the devVChan in the device entry pointer */
        sTemp = (0xFFFF & (tt16Bit)(myTag));
        tm_htons(sTemp, vlanInfo);
        etherTaggedHeaderPtr->ethQTagControlInfo = vlanInfo;

/* Copy the VLAN length/type field into the packet */
/* This should be the 16 MSBs of the devVChan in the device entry pointer */
        myTag = myTag >> 16;
        sTemp = (0xFFFF & (tt16Bit)myTag);
        tm_htons(sTemp, vlanInfo);
        etherTaggedHeaderPtr->ethQTagLengthType = vlanInfo;

/* Set the packet type */
        if (packetPtr->pktNetworkLayer == TM_NETWORK_IPV4_LAYER)
        {
            etherTaggedHeaderPtr->ethType = TM_ETHER_IP_TYPE;
        }
        else
        {
            {    
                if (packetPtr->pktNetworkLayer == TM_NETWORK_ARP_LAYER)
                {
                    etherTaggedHeaderPtr->ethType = TM_ETHER_ARP_TYPE;
                }
                else
                {
/* Should never happen */
                    retCode = TM_EPROTONOSUPPORT;
                }
            }
        }

    }
   return (retCode);
}
/* 
 * Process an incoming tagged Ethernet packet
 * We convert from the exposed USER API functions to the
 * Internal form since this is called from the device code
 */

int tfEtherRecvTagged(void TM_FAR *interfaceId, void TM_FAR *bufferHandle)
{
    ttTaggedEtherHeaderPtr  etherTaggedHeaderPtr;
    ttPacketPtr             packetPtr;
    ttDeviceEntryPtr        devEntryPtr;
    tt8BitPtr               pktEtherAddressPtr;
    tt8BitPtr               etherHeaderAddressPtr;
    tt16Bit                 candidateQTag;
    tt16Bit                 candidateVID;
    tt32Bit                 currentVlanTag;
    tt32Bit                 candidateVlanTag;
    int                     flags;
    int                     retCode;

    currentVlanTag = TM_UL(0);
    candidateVlanTag = TM_UL(0);

#ifdef TM_DEV_RECV_OFFLOAD
    ttDevRecvOffloadPtr     pktDevRecvOffloadPtr;
#endif /* TM_DEV_RECV_OFFLOAD */

    retCode = TM_ENOERROR;
    packetPtr = (ttPacketPtr)bufferHandle;
    devEntryPtr = (ttDeviceEntryPtr)interfaceId;
/* Store the device the packet came from */
    packetPtr->pktDeviceEntryPtr = devEntryPtr;
#ifdef TM_USE_DRV_SCAT_RECV
/* If Ethernet header is not contiguous, make it contiguous */
    if (tm_pkt_hdr_is_not_cont(packetPtr, TM_TAGGED_ETHER_HEADER_SIZE))
    {
        packetPtr = tfContiguousSharedBuffer(packetPtr, 
                                                TM_TAGGED_ETHER_HEADER_SIZE);
        if (packetPtr == (ttPacketPtr)0)
        {
            retCode = TM_ENOBUFS;
            goto etherTaggedRecvExit;
        }
    }
#endif /* TM_USE_DRV_SCAT_RECV */

/* write the packet into the pcap file */
#ifdef TM_USE_SNIFF
#ifdef TM_KERNEL_WIN32_X86
    if(packetPtr != TM_PACKET_NULL_PTR)
    {
        tfPcapWritePacket(packetPtr);
    }
#endif /* TM_KERNEL_WIN32_X86 */
#endif /* TM_USE_SNIFF */

/* Check the type field */
    etherTaggedHeaderPtr = 
                    (ttTaggedEtherHeaderPtr)(packetPtr->pktLinkDataPtr);

/* Get the current valid VLAN tag from the interface */
/* The current mHome index will be TM_VLAN_MHOME_INDEX defined in trsocket.h */
    currentVlanTag = devEntryPtr->devVChanArray[TM_VLAN_MHOME_INDEX];

/* Check the current VLANtag against the incoming packet's VLANtag */
    tm_ntohs(etherTaggedHeaderPtr->ethQTagLengthType, candidateQTag);
    candidateVlanTag = (tt32Bit)candidateQTag;
    candidateVlanTag = (TM_UL(0xFFFF0000) & (candidateVlanTag << 16));

    tm_ntohs(etherTaggedHeaderPtr->ethQTagControlInfo, candidateVID);

    candidateVlanTag = (tt32Bit)(candidateVlanTag | (tt32Bit)candidateVID);
    
    if(currentVlanTag != candidateVlanTag)
    {
        /* VLAN tag doesn't match */
        retCode = TM_EPROTONOSUPPORT;
        tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
        goto etherTaggedRecvExit;
    }
    else
    {
        retCode = TM_ENOERROR;
    }

#ifdef TM_DEV_RECV_OFFLOAD
    pktDevRecvOffloadPtr = (ttDevRecvOffloadPtr)packetPtr->pktDevOffloadPtr;
    if (    (pktDevRecvOffloadPtr != (ttDevRecvOffloadPtr)0)
         && (   pktDevRecvOffloadPtr->devoRFlags 
              & (TM_DEVOR_PACKET_RECV_CHKSUM_COMPUTED ) ) )
    {
        tfDevoRecvPeelChecksum(packetPtr, (tt16BitPtr)etherHeaderPtr, 
                               (tt16Bit)TM_TAGGED_ETHER_HEADER_SIZE);
    }
#endif /* TM_DEV_RECV_OFFLOAD */
    packetPtr->pktLinkDataPtr += TM_TAGGED_ETHER_HEADER_SIZE;
    packetPtr->pktLinkDataLength -= TM_TAGGED_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength -= TM_TAGGED_ETHER_HEADER_SIZE;
/* Copy the source ethernet address into the packet */
    pktEtherAddressPtr = (tt8BitPtr)packetPtr->pktEthernetAddress;
    etherHeaderAddressPtr = (tt8BitPtr)etherTaggedHeaderPtr->ethSrcAddr;
    tm_ether_copy(etherHeaderAddressPtr, pktEtherAddressPtr);
/*
 * Look at the ethernet destination address to check for destination
 * Ethernet broadcast/multicast.
 */
    etherHeaderAddressPtr = (tt8BitPtr)etherTaggedHeaderPtr->ethDstAddr;
    if (tm_ether_broadcast(etherHeaderAddressPtr))
    {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
        tm_64Bit_incr(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInBcastPkts);
        tm_64Bit_incr(tm_context(tvDevIpv4Data).ipIfStatsHCInBcastPkts);
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
        flags = TM_LL_BROADCAST_FLAG;

    }
    else
    {
        if (tm_ether_mcast(etherHeaderAddressPtr))
        {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            tm_64Bit_incr(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInMcastPkts);
            tm_64Bit_incr(tm_context(tvDevIpv4Data).ipIfStatsHCInMcastPkts);
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            flags = TM_LL_MULTICAST_FLAG;
        }
        else
        {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            devEntryPtr->devInUcastPkts++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            flags = TM_LL_UNICAST_FLAG;
        }
    }
    switch (etherTaggedHeaderPtr->ethType)
    {
        case TM_ETHER_ARP_TYPE:
            packetPtr->pktFlags = 0;
            retCode = tfArpIncomingPacket(packetPtr);
            break;

        case TM_ETHER_IP_TYPE:
            packetPtr->pktFlags = (tt16Bit) flags;
            tfIpIncomingPacket(packetPtr
                               );
            break;

            
        default:
/* Packet Not For TRECK */
#ifdef TM_SNMP_MIB
            devEntryPtr->dev4Ipv4Mib.ipIfStatsInUnknownProtos++;
            tm_context(tvDevIpv4Data).ipIfStatsInUnknownProtos++;
#endif /* TM_SNMP_MIB */
            tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
            retCode = TM_EPROTONOSUPPORT;
            break;
    }
etherTaggedRecvExit:

    return(retCode);
}

int tfEtherIoctlTagged(int operation)
{
    TM_UNREF_IN_ARG(operation);
    return TM_LL_OKAY;
}

int tfEtherErrorTagged(int type)
{
    TM_UNREF_IN_ARG(type);
    return TM_LL_OKAY;
}

#else /* !TM_USE_VCHAN */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */


/* To allow link for builds when TM_USE_VCHAN is not defined */
int tvEthTagDummy = 0;

#endif /* TM_USE_VCHAN */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:48      For: Xilinx-Samsung
 * File Name: trethtag.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 27081    Output checksum: 57021
 */
