/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2010
 * Copyright Elmic Systems Japan 1997 - 2010
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Interface Loopback Driver. Can be used with any link layer.
 *              Supports multiple context.
 *
 * Filename: trindrmc.c
 * Author: Bryan
 * Date Created: 05/28/2002
 * $Source: source/trindrmc.c $
 *
 * Modification History
 * $Revision: 6.0.2.4 $
 * $Date: 2011/09/07 00:33:20CDT $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>


typedef struct tsIntfDrvInfo
{
/* Device parameter array index */
    int             infoDeviceIndex;
} ttIntfDrvInfo;

typedef ttIntfDrvInfo TM_FAR * ttIntfDrvInfoPtr;

#define TM_INTFDRV_INFO_PTR_NULL (ttIntfDrvInfoPtr)0

/* Bytes to add to malloc to insure 8-byte alignment */
#define TM_INTFDRV_ALIGNMENT_MODIFIER   8     

/* Mask for rounding to 8-byte multiple */
#define TM_INTFDRV_ALIGNMENT_MOD_MASK   0xfffffff8UL

#define TM_INTF_MTU 5000

/*
 * Define the following symbol to restrict Ethernet sends to broadcast,
 * multicast and exact match unicast.
 * This avoids the catastrophic network failure that can occur when there
 * are 3 or more hosts on the same network id (tfNetworkSetId) and 2 or
 * more of the hosts support IP forwarding (which can lead to exponential
 * packet duplication on the network segment).
 * ** Note ** When using this option, make sure that all interface names start
 * with an ASCII character that has the least significant bit clear because
 * names are used as Ethernet addresses and the first bit of the destination
 * address transmitted is the Ethernet multicast/unicast bit.
 */
/* #define TM_INTF_ETHERNET_EXACT */

/*
 * Support for network cable disconnect.
 * Define the following symbol to be the network identifier that you have
 * reserved as a "bit bucket".
 * Any interface having the following net id will not be allowed to send
 * to to any other interface, as though the cable had been unplugged.
 * The net id for an interface can be changed by calling tfNetworkSetId().
 */
/* #define TM_INTF_DISCONNECT_NETID 0 */


/* put a packet to a interface in the right context */
static int tfIntfMCtxDrvTransmit(
    ttDeviceEntryPtr    sendDevPtr,
    ttDeviceEntryPtr    recvDevPtr,
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext       sendContextHandle,
    ttUserContext       recvContextHandle,
#endif
    ttCharPtr           dataPtr,
    int                 dataLength,
    int                 flag );

int tfIntfMCtxDrvOpen(  ttUserInterface  interfaceHandle);
int tfIntfMCtxDrvClose( ttUserInterface  interfaceHandle);

int tfIntfMCtxDrvSend(
    ttUserInterface     interfaceHandle,
    char TM_FAR *       dataPtr,
    int                 dataLength,
    int                 flag );
int tfIntfMCtxDrvRecv(
    ttUserInterface     interfaceHandle,
    ttCharPtr TM_FAR *  dataPtrPtr,
    ttIntPtr            dataSizePtr,
    ttUserBufferPtr     packetPtrPtr
#ifdef TM_DEV_RECV_OFFLOAD
  , ttDevRecvOffloadPtr recvOffloadPtr
#endif /* TM_DEV_RECV_OFFLOAD */
                     );

int tfIntfMCtxDrvGetPhyAddr( 
    ttUserInterface     interfaceHandle,
    char TM_FAR *       physicalAddress );

int tfIntfMCtxDrvIoctl( 
    ttUserInterface     interfaceHandle, 
    int                 flag,
    void TM_FAR *       optionPtr, 
    int optionLen );

ttUserInterface tfUseIntfMCtxDriver(
        char TM_FAR *   namePtr,
        ttUserLinkLayer linkLayerHandle,
        int TM_FAR *    errorCodePtr )
{
    ttUserInterface interfaceHandle;

    interfaceHandle = tfAddInterface(
                            namePtr,
                            linkLayerHandle,
                            tfIntfMCtxDrvOpen,
                            tfIntfMCtxDrvClose,
                            tfIntfMCtxDrvSend,
                            tfIntfMCtxDrvRecv,
                            (ttDevFreeRecvFuncPtr)0,
                            tfIntfMCtxDrvIoctl,
                            tfIntfMCtxDrvGetPhyAddr,
                            errorCodePtr);
    return interfaceHandle;
}

int tfIntfMCtxDrvGetPhyAddr (
    ttUserInterface interfaceHandle,
    char TM_FAR *   physicalAddress )
{
    ttDeviceEntryPtr devEntryPtr;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    tm_bcopy(&devEntryPtr->devNameArray[0], physicalAddress, 6);
    return 0;
}



int tfIntfMCtxDrvOpen( ttUserInterface  interfaceHandle)
{
#ifdef TM_MULTIPLE_CONTEXT

    ttIntfDrvInfoPtr        devInfoPtr;
    int                     devIndex;
    int                     errorCode;

    errorCode   = TM_DEV_OKAY;
    devInfoPtr  = TM_INTFDRV_INFO_PTR_NULL;

/* Find an unused ISR parameter structure */
    for(devIndex = 0; devIndex < TM_MAX_NUMBER_DEVICES; devIndex++ )
    {
        if(    tvDeviceIsrPara[devIndex].deviInterfaceHandle 
            == (ttUserInterface)0 )
            break;
    }

    if(devIndex >= TM_MAX_NUMBER_DEVICES)
    {
#ifdef TM_ERROR_CHECKING
        tfKernelError("tfIntfMCtxDrvOpen",
           "Device driver parameter structure used up : corrupted memory.");
#endif /* TM_ERROR_CHECKING */
        errorCode = TM_DEV_ERROR;
        goto intfMctxDrvOpenEnd;
    }
    else
    {
        tvDeviceIsrPara[devIndex].deviInterfaceHandle = interfaceHandle;
        tvDeviceIsrPara[devIndex].deviContextHandle = tfGetCurrentContext();
    }

/* 
 * Get memory for the tsIntfDrvInfo structure and make sure it is
 * 8-byte aligned, this is to store device specific information on 
 * the devEntry
 */
    devInfoPtr =   (ttIntfDrvInfoPtr)tm_get_raw_buffer( sizeof(ttIntfDrvInfo)
                 + TM_INTFDRV_ALIGNMENT_MODIFIER );

    if (devInfoPtr == TM_INTFDRV_INFO_PTR_NULL) 
    {
        errorCode = TM_DEV_ERROR;
        goto intfMctxDrvOpenEnd;
    }

/*    devInfoPtr = (ttIntfDrvInfoPtr)*/
/*                     ((((tt32Bit)(devInfoPtr)) +*/
/*                     (TM_INTFDRV_ALIGNMENT_MODIFIER-1)) &*/
/*                     TM_INTFDRV_ALIGNMENT_MOD_MASK);*/

/* this is what we want to store and retrieve later */
    devInfoPtr->infoDeviceIndex = devIndex;

/* Store driver info pointer */
    if(    tfDeviceStorePointer(interfaceHandle, (ttVoidPtr)devInfoPtr )
        != TM_ENOERROR )
    {
#ifdef TM_ERROR_CHECKING
        tfKernelError("tfIntfMCtxDrvOpen",
                      "Device driver already opened: corrupted memory.");
#endif /* TM_ERROR_CHECKING */
        errorCode = TM_DEV_ERROR;
        goto intfMctxDrvOpenEnd;
    }

intfMctxDrvOpenEnd:
    if (errorCode != TM_DEV_OKAY && devInfoPtr != TM_INTFDRV_INFO_PTR_NULL)
    {
        tm_free_raw_buffer((ttVoidPtr)devInfoPtr);
    }

    return(errorCode);            

#else /* !TM_MULTIPLE_CONTEXT */
    TM_UNREF_IN_ARG(interfaceHandle);
    return TM_DEV_OKAY;
#endif /* TM_MULTIPLE_CONTEXT */
}


/*
 * tfIntfMCtxDrvClose
 *
 * Closes the serial interface:  restore old interrupt handler and disable
 * IRQ4/IRQ3 interupts in the PIC. As with the interrupt initialization code,
 * this is very hardware specific.
 */
int tfIntfMCtxDrvClose(ttUserInterface interfaceHandle)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttIntfDrvInfoPtr        devInfoPtr;
    int                     errorCode;

    errorCode = TM_DEV_OKAY;

/* Get the device info pointer */
    devInfoPtr = tm_device_get_pointer(interfaceHandle);
    if( devInfoPtr == TM_INTFDRV_INFO_PTR_NULL )
    {
#ifdef TM_ERROR_CHECKING
        tfKernelError("tfIntfMCtxDrvClose",
                      "Device driver already closed: corrupted memory.");
#endif /* TM_ERROR_CHECKING */
        errorCode = TM_DEV_ERROR;
        goto intfMctxDrvCloseEnd;
    }

/* Free ISR handler parameters structure */
    tvDeviceIsrPara[devInfoPtr->infoDeviceIndex].deviInterfaceHandle = 
                                                        (ttUserInterface)0;
/* Free device driver variable */
    if( tfDeviceClearPointer(interfaceHandle) != devInfoPtr )
    {
#ifdef TM_ERROR_CHECKING
        tfKernelError("tfIntfMCtxDrvClose",
            "Fail to clear device driver pointer: corrupted memory!");
#endif /* TM_ERROR_CHECKING */
        errorCode = TM_DEV_ERROR;
    }
    else
    {
/* Free malloc'd memory */
        tm_free_raw_buffer((ttVoidPtr)devInfoPtr);
    }

intfMctxDrvCloseEnd:
    return(errorCode);

#else /* !TM_MULTIPLE_CONTEXT */
    TM_UNREF_IN_ARG(interfaceHandle);
    return TM_DEV_OKAY;
#endif /* TM_MULTIPLE_CONTEXT */

}

/* put a packet to a interface in the right context */
static int tfIntfMCtxDrvTransmit(
    ttDeviceEntryPtr    sendDevPtr,
    ttDeviceEntryPtr    recvDevPtr,
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext       sendContextHandle,
    ttUserContext       recvContextHandle,
#endif
    ttCharPtr           dataPtr,
    int                 dataLength,
    int                 flag )
{
    ttPacketPtr         packetPtr;
    int                 errorCode;
    int                 hdrSize;
    ttPktLen            dataSize;

    errorCode = TM_ENOERROR;

/*
 * Check if there is an existing packet that we are writing into.
 * In scattered send case, we concatenate onto the end of an existing packet
 * buffer when we do the copy.
 */
    tm_kernel_set_critical;
#ifdef TM_INDRV_INLINE_SEND_RECV
/* Single packet queue */
    packetPtr = recvDevPtr->devReceiveQueuePtr;
#else /* TM_INDRV_INLINE_SEND_RECV */
/* Multiple packet queue. Check packet at the end of the queue. */
    packetPtr = recvDevPtr->devRcvLastQueuePtr;
#endif /* TM_INDRV_INLINE_SEND_RECV */
    tm_kernel_release_critical;

/* copy into a single buffer */
    if (    (packetPtr == TM_PACKET_NULL_PTR)
#ifndef TM_INDRV_INLINE_SEND_RECV
         || tm_16bit_bits_not_set( packetPtr->pktUserFlags,
                                   TM_PKTF_USER_DEV_WRITE )
#endif /* TM_INDRV_INLINE_SEND_RECV */
       )
    {
/*
 * Either there is no packet currently in the queue, or packet at the
 * end of the queue can't receive any more data, we need to allocate space
 * for a packet in the receiver's queue before we put the data onto it.
 */
        hdrSize = TM_LL_MAX_HDR_LEN - sendDevPtr->devLinkHeaderLength;
        dataSize = TM_INTF_MTU + recvDevPtr->devLinkHeaderLength;
        if (tm_ll_need_32bit_align(recvDevPtr))
        {
            hdrSize += TM_ETHER_HW_ALIGN;
            dataSize += TM_ETHER_IP_ALIGN;
        }
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(recvContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
        packetPtr = tfGetSharedBuffer(hdrSize, dataSize, TM_16BIT_ZERO);
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(sendContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
        if ( packetPtr == TM_PACKET_NULL_PTR)
        {
#ifdef TM_SNMP_MIB
            sendDevPtr->dev4Ipv4Mib.ipIfStatsOutDiscards++;
            tm_context(tvDevIpv4Data).ipIfStatsOutDiscards++;
#endif /* TM_SNMP_MIB */
            errorCode = TM_ENOBUFS;
            goto intfMctxDrvTransmitEnd;
        }
        if (tm_ll_need_32bit_align(recvDevPtr))
        {
            if (   ((ttUserPtrCastToInt) packetPtr->pktLinkDataPtr)
                 & (ttUserPtrCastToInt)(TM_ETHER_HW_ALIGN - 1) )
            {
/* Align the data area (according to TM_ETHER_HW_ALIGN) */
                packetPtr->pktLinkDataPtr = (tt8BitPtr)
                    (((ttUserPtrCastToInt) packetPtr->pktLinkDataPtr)
                     & ~((ttUserPtrCastToInt) (TM_ETHER_HW_ALIGN - 1)) );
            }
/*
 * Offset of ethernet frame to allow IP area to be long word aligned.
 */
            packetPtr->pktLinkDataPtr += TM_ETHER_IP_ALIGN;
        }
/* Queue packet */
#ifdef TM_INDRV_INLINE_SEND_RECV
/* Single packet queue */
        recvDevPtr->devReceiveQueuePtr = packetPtr;
#else /* TM_INDRV_INLINE_SEND_RECV */
/* Packet is being written into */
        packetPtr->pktUserFlags |= TM_PKTF_USER_DEV_WRITE;
        packetPtr->pktRecvNextPtr = (ttPacketPtr)0;
        tm_kernel_set_critical;
        if (recvDevPtr->devReceiveQueuePtr == TM_PACKET_NULL_PTR)
        {
/* Empty queue. */
            recvDevPtr->devReceiveQueuePtr = packetPtr;
            recvDevPtr->devRcvLastQueuePtr = packetPtr;
        }
        else
        {
/* Non Empty queue. Queue at the end of the queue */
            recvDevPtr->devRcvLastQueuePtr->pktRecvNextPtr = packetPtr;
            recvDevPtr->devRcvLastQueuePtr = packetPtr;
        }
        tm_kernel_release_critical;
#endif /* TM_INDRV_INLINE_SEND_RECV */
        recvDevPtr->devReceiveDataPtr = packetPtr->pktLinkDataPtr;
        recvDevPtr->devReceiveDataLength = 0;
    }

/* copy the sending data to the receiver's queue */
    {
        tm_bcopy(dataPtr,
                 recvDevPtr->devReceiveDataPtr,
                 tm_packed_byte_count(dataLength));
        recvDevPtr->devReceiveDataPtr += tm_packed_byte_len(dataLength);
    }
    recvDevPtr->devReceiveDataLength =
        (tt16Bit)(recvDevPtr->devReceiveDataLength + (tt16Bit)dataLength );

    if (flag != TM_USER_BUFFER_MORE)
    {
/* The final segment, so send it up */
/*
 * We can inline the tfRecvInterface() function keeping only what we need
 * (if TM_INDRV_INLINE_SEND_RECV is defined), otherwise the user need
 * to call tfRecvInterface().
 */

#ifdef TM_INDRV_INLINE_SEND_RECV
        dataLength = (int)recvDevPtr->devReceiveDataLength;
        packetPtr->pktLinkDataLength = (ttPktLen)dataLength;
        packetPtr->pktChainDataLength = (ttPktLen)dataLength;
#ifdef TM_SNMP_MIB
/* We bypass tfRecvInterface */
        recvDevPtr->devInOctets += (tt32Bit)dataLength;
#endif /* TM_SNMP_MIB */
        recvDevPtr->devReceiveQueuePtr = TM_PACKET_NULL_PTR;
/* Unlock the device send/receive */
        tm_unlock(&(sendDevPtr->devDriverLockEntry));
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(recvContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
        errorCode = (*(recvDevPtr->devLinkLayerProtocolPtr->lnkRecvFuncPtr))(
                                               (void TM_FAR *)recvDevPtr,
                                               (void TM_FAR *)packetPtr);
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(sendContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
        tm_lock_wait(&(sendDevPtr->devDriverLockEntry));
#else /* !TM_INDRV_INLINE_SEND_RECV */
        packetPtr->pktLinkDataLength = recvDevPtr->devReceiveDataLength;
        packetPtr->pktUserFlags = TM_PKTF_USER_DEV_OWNS;
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(recvContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
/* Notify the Treck stack that we've received a packet */
        tfNotifyInterfaceTask((ttUserInterface)recvDevPtr , 1, 0, 0UL, 0UL);
#ifdef TM_MULTIPLE_CONTEXT
        tfSetCurrentContext(sendContextHandle);
#endif /* TM_MULTIPLE_CONTEXT */
#endif /* TM_INDRV_INLINE_SEND_RECV */
    }

intfMctxDrvTransmitEnd:
    return errorCode;

}



int tfIntfMCtxDrvSend (
    ttUserInterface  interfaceHandle,
    char     TM_FAR  *dataPtr,
    int              dataLength,
    int              flag )
{
    ttDeviceEntryPtr    sendDevPtr;
    ttDeviceEntryPtr    recvDevPtr;
    int                 errorCode;
#ifdef TM_USE_VCHAN
    ttUser32Bit         vchan;
#endif /* TM_USE_VCHAN */
#ifdef TM_MULTIPLE_CONTEXT
    ttDeviceEntryPtr    devPtr;
    int                 devIndex;
    ttDeviceIsrParaPtr  sendDevParaPtr;
    ttDeviceIsrParaPtr  recvDevParaPtr;
    ttUserContext       sendContext;
    ttUserContext       recvContext;
    ttUser32Bit         networkId;
    ttIntfDrvInfoPtr    devInfoPtr;
    tt32Bit             copies;
#endif

    errorCode = TM_DEV_OKAY;
    sendDevPtr = (ttDeviceEntryPtr) interfaceHandle;
    recvDevPtr = (ttDeviceEntryPtr)0;

#ifdef TM_MULTIPLE_CONTEXT
/* Get the device info pointer */
    devInfoPtr = tm_device_get_pointer(interfaceHandle);
    if( devInfoPtr == TM_INTFDRV_INFO_PTR_NULL )
    {
#ifdef TM_ERROR_CHECKING
        tfKernelError("tfIntfMCtxDrvSend",
                      "Device driver already closed: corrupted memory.");
#endif /* TM_ERROR_CHECKING */
        errorCode = TM_DEV_ERROR;
        goto intfMctxDrvSendEnd;
    }

    sendDevParaPtr = &tvDeviceIsrPara[devInfoPtr->infoDeviceIndex];
    networkId = tm_device_get_network_id(interfaceHandle);
    copies = 0;
#endif /* TM_MULTIPLE_CONTEXT */


#ifdef TM_USE_VCHAN
    errorCode = tfInterfaceGetVirtualChannel(interfaceHandle, &vchan);
    tm_assert(tfIntfMCtxDrvSend, (errorCode == TM_ENOERROR));
#endif /* TM_USE_VCHAN */

/*
 * dump the packet to the file system in libpcap format,
 * to be opened by tcpdump, windump, ethereal, or other libpcap file readers.
 */
#ifdef TM_USE_INF_DRV_SNIFF
/*
 * PPP and SLIP sniffing is done in tfPppStuffPacket and tfSlipStuffPacket,
 * before framing data and escape data is applied.
 */
    if ((   ((ttDeviceEntryPtr)interfaceHandle)->devLinkLayerProtocol
          & TM_LINK_PT2PT_MASK) == 0)
    {
        tfPcapWriteFrame(
            ((ttDeviceEntryPtr)interfaceHandle)->devLinkLayerProtocol,
            dataPtr,
            dataLength);
    }
#endif /* TM_USE_INF_DRV_SNIFF */


#ifdef TM_MULTIPLE_CONTEXT
#ifdef TM_INTF_DISCONNECT_NETID
/*
 * Don't send if sender's network cable is "disconnected".
 * We never need to check if the cable is disconnected at the receiver's
 * end because a connected sender net id will never match a disconnected
 * receiver net id.
 */
    if (networkId == TM_INTF_DISCONNECT_NETID)
    {
        goto intfMctxDrvSendEnd;
    }
#endif /* TM_INTF_DISCONNECT_NETID */
    sendContext = tfGetCurrentContext();
    recvContext = (ttUserContext)0;

/* For each interface configured under the multiple context interface system,
 * we examine the following condidtion to decide whether a copy of the packet
 * should be deliverd:
 *  1. not the sending interface itself
 *  2. same network ID
 *  3. same link layer protocol
 *  4. Different contexts.
 */
    for (devIndex=0; devIndex< TM_MAX_NUMBER_DEVICES; devIndex++)
    {
        recvDevParaPtr = &(tvDeviceIsrPara[devIndex]);
        devPtr = (ttDeviceEntryPtr)recvDevParaPtr->deviInterfaceHandle;
        if (    (recvDevParaPtr->deviInterfaceHandle != (ttUserInterface)0)
             && (recvDevParaPtr->deviInterfaceHandle != interfaceHandle )
             && (   tm_device_get_network_id(recvDevParaPtr->deviInterfaceHandle)
                 == networkId )
             && (   devPtr->devLinkLayerProtocol
                 == sendDevPtr->devLinkLayerProtocol)
             && (sendContext != recvDevParaPtr->deviContextHandle)
#ifdef TM_INTF_ETHERNET_EXACT
/* Restrict Ethernet to broadcast, multicast and exact match unicast */
             && (   (devPtr->devLinkLayerProtocol & TM_LINK_ETHER_MASK) == 0
                 || (dataPtr[0] & 1) != 0
                 || tm_memcmp( dataPtr,
                               devPtr->devPhysAddrArr,
                               TM_ETHERNET_PHY_ADDR_LEN ) == 0 )
#endif /* TM_INTF_ETHERNET_EXACT */
           )
        {
/* we should transmit a copy of the packet to this interface interface */
            recvDevPtr = (ttDeviceEntryPtr)recvDevParaPtr->deviInterfaceHandle;
            recvContext = recvDevParaPtr->deviContextHandle;
            tfIntfMCtxDrvTransmit(
                    sendDevPtr,
                    recvDevPtr,
                    sendContext,
                    recvContext,
                    dataPtr,
                    dataLength,
                    flag);
            copies ++;
        }
    }

    if (copies == 0)
    {
        errorCode = TM_EHOSTDOWN;
    }

intfMctxDrvSendEnd:
#else /* TM_MULTIPLE_CONTEXT */
    recvDevPtr = (ttDeviceEntryPtr)interfaceHandle;

    tfIntfMCtxDrvTransmit(
        sendDevPtr,
        recvDevPtr,
        dataPtr,
        dataLength,
        flag);

#endif /* TM_MULTIPLE_CONTEXT */

#ifndef TM_TASK_SEND
    if (flag != TM_USER_BUFFER_MORE)
    {
        tfSendCompleteInterface((ttUserInterface)sendDevPtr,
                                 TM_DEV_SEND_COMPLETE_DRIVER );
    }
#endif /* TM_TASK_SEND */

    return errorCode;
}


int tfIntfMCtxDrvRecv (
    ttUserInterface     interfaceHandle,
    ttCharPtr TM_FAR *  dataPtrPtr,
    ttIntPtr            dataSizePtr,
    ttUserBufferPtr     packetPtrPtr
#ifdef TM_DEV_RECV_OFFLOAD
  , ttDevRecvOffloadPtr recvOffloadPtr
#endif /* TM_DEV_RECV_OFFLOAD */
                      )
{
#ifdef TM_INDRV_INLINE_SEND_RECV
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtrPtr);
    TM_UNREF_IN_ARG(dataSizePtr);
    TM_UNREF_IN_ARG(packetPtrPtr);
#ifdef TM_DEV_RECV_OFFLOAD
    TM_UNREF_IN_ARG(recvOffloadPtr); /* compiler warning for unused parameter */
#endif /* TM_DEV_RECV_OFFLOAD */
    return TM_ENOERROR;
#else /* !TM_INDRV_INLINE_SEND_RECV */
    ttDeviceEntryPtr devEntryPtr;
    ttPacketPtr      packetPtr;
    int              errorCode;

#ifdef TM_DEV_RECV_OFFLOAD
    TM_UNREF_IN_ARG(recvOffloadPtr); /* compiler warning for unused parameter */
#endif /* TM_DEV_RECV_OFFLOAD */
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    tm_kernel_set_critical;
/* Check at head of the queue */
    packetPtr = devEntryPtr->devReceiveQueuePtr;
    if (    (packetPtr != TM_PACKET_NULL_PTR)
         && tm_16bit_bits_not_set(packetPtr->pktFlags,
                                  TM_PKTF_USER_DEV_WRITE) )
/* Packet in the queue, and ready to be received */
    {
/* unlink this packet from the head of the queue of received packets */
        devEntryPtr->devReceiveQueuePtr =
                devEntryPtr->devReceiveQueuePtr->pktRecvNextPtr;
        if (devEntryPtr->devReceiveQueuePtr == (ttPacketPtr)0)
        {
/* Empty queue */
            devEntryPtr->devRcvLastQueuePtr = (ttPacketPtr)0;
        }
        packetPtr->pktRecvNextPtr = (ttPacketPtr)0;
/* Back to the caller */
        *dataSizePtr = (int)packetPtr->pktLinkDataLength;
        *dataPtrPtr = (ttCharPtr)packetPtr->pktLinkDataPtr;
        errorCode = TM_ENOERROR;
    }
    else
    {
        packetPtr = (ttPacketPtr)0;
        errorCode = TM_ENOBUFS;
    }
    tm_kernel_release_critical;
    *packetPtrPtr = packetPtr;
    return errorCode;
#endif /* TM_INDRV_INLINE_SEND_RECV */
}

/*
 * Can be used to configure the device or refill the receive pool, or
 * enable reception of a list of multicast addresses.
 */
int tfIntfMCtxDrvIoctl ( ttUserInterface interfaceHandle, int flag,
                     void TM_FAR * optionPtr, int optionLen )
{

    TM_UNREF_IN_ARG(interfaceHandle);
    optionPtr       = optionPtr;
    optionLen       = optionLen;

#ifdef TM_IGMP
    flag            = flag;

/*
 * if (flag == TM_DEV_SET_MCAST_LIST)
 * Enable reception for all multicast addresses on the list
 */
/*
 * Else if (flag == TM_DEV_SET_ALL_MCAST)
 * Enable reception of all multicast addresses
 */

#else /* !TM_IGMP */
    flag            = flag;

#endif /* TM_IGMP */

    return TM_ENOERROR;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:48      For: Xilinx-Samsung
 * File Name: trindrmc.c    File Revision: 6.0.2.4 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 45632    Output checksum: 46284
 */
