/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Interface Loopback Driver. Can be used with any link layer.
 *
 * Filename: trindrv.c
 * Author: Odile
 * Date Created: 10/28/99
 * $Source: source/trindrv.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:47:42CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#if defined(TM_LOCK_NEEDED) || (defined(TM_SNMP_MIB) && defined(TM_USE_IPV4))
#include <trglobal.h>
#endif /* TM_LOCK_NEEDED || (TM_SNMP_MIB && TM_USE_IPV4) */

#define TM_INTF_MTU 5000


ttUserInterface tfUseIntfDriver ( ttCharPtr         namePtr,
                                  ttUserLinkLayer   linkLayerHandle,
                                  int TM_FAR      * errorCodePtr )
{
    ttUserInterface interfaceHandle;

    interfaceHandle = tfAddInterface( namePtr,
                                      linkLayerHandle,
                                      (ttDevOpenCloseFuncPtr)0,
                                      (ttDevOpenCloseFuncPtr)0,
                                      tfIntfDrvSend,
                                      tfIntfDrvRecv,
                                      (ttDevFreeRecvFuncPtr)0,
                                      tfIntfDrvIoctl,
                                      tfIntfDrvGetPhyAddr,
                                      errorCodePtr);
    return interfaceHandle;
}

int tfIntfDrvGetPhyAddr ( ttUserInterface interfaceHandle,
                          char TM_FAR *physicalAddress )
{
    ttDeviceEntryPtr devEntryPtr;

    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
    tm_bcopy(&devEntryPtr->devNameArray[0], physicalAddress, 6);
    return 0;
}


int tfIntfDrvSend ( ttUserInterface  interfaceHandle,
                    char     TM_FAR  *dataPtr,
                    int              dataLength,
                    int              flag )
{
    ttDeviceEntryPtr devEntryPtr;
    ttPacketPtr      packetPtr;
    int              errorCode;

    errorCode = TM_ENOERROR;
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;

/* 
 * dump the packet to the file system in libpcap format,
 * to be opened by tcpdump, windump, ethereal, or other libpcap file readers.
 */
#ifdef TM_USE_INF_DRV_SNIFF
/* 
 * PPP and SLIP sniffing is done in tfPppStuffPacket and tfSlipStuffPacket,
 * before framing data and escape data is applied. 
 */
    if ((   ((ttDeviceEntryPtr)interfaceHandle)->devLinkLayerProtocol
          & TM_LINK_PT2PT_MASK) == 0)
    {
        tfPcapWriteFrame(
            ((ttDeviceEntryPtr)interfaceHandle)->devLinkLayerProtocol,
            dataPtr, 
            dataLength);
    }
#endif /* TM_USE_INF_DRV_SNIFF */

/* 
 * Check if there is an existing packet that we are writing into.
 * In scattered send case, we concatenate onto the end of an existing packet
 * buffer when we do the copy. 
 */
#ifdef TM_INDRV_INLINE_SEND_RECV
/* Single packet queue */
    packetPtr = devEntryPtr->devReceiveQueuePtr;
#else /* TM_INDRV_INLINE_SEND_RECV */
/* Multiple packet queue. Check packet at the end of the queue. */
    packetPtr = devEntryPtr->devRcvLastQueuePtr;
#endif /* TM_INDRV_INLINE_SEND_RECV */
/* copy into a single buffer */
    if (    (packetPtr == TM_PACKET_NULL_PTR)
#ifndef TM_INDRV_INLINE_SEND_RECV
         || tm_16bit_bits_not_set( packetPtr->pktUserFlags, 
                                   TM_PKTF_USER_DEV_WRITE )
#endif /* TM_INDRV_INLINE_SEND_RECV */
       )
    {
/* 
 * Either there is no packet currently in the queue, or packet at the
 * end of the queue is ready to be received. 
 */
        packetPtr = tfGetSharedBuffer(
                                 TM_LL_MAX_HDR_LEN + TM_ETHER_HW_ALIGN,
                                 (ttPktLen)(TM_INTF_MTU + TM_ETHER_IP_ALIGN),
                                 TM_16BIT_ZERO );
        if ( packetPtr == TM_PACKET_NULL_PTR)
        {
#ifdef TM_SNMP_MIB
            devEntryPtr->dev4Ipv4Mib.ipIfStatsOutDiscards++;
            tm_context(tvDevIpv4Data).ipIfStatsOutDiscards++;
#endif /* TM_SNMP_MIB */
            errorCode = TM_ENOBUFS;
            goto testSendFinish;
        }
        if (tm_ll_need_32bit_align(devEntryPtr))
        {
            if (   ((ttUserPtrCastToInt) packetPtr->pktLinkDataPtr)
                 & (ttUserPtrCastToInt)(TM_ETHER_HW_ALIGN - 1) )
            {
/* Align the data area (according to TM_ETHER_HW_ALIGN) */
                packetPtr->pktLinkDataPtr = (tt8BitPtr)
                    (((ttUserPtrCastToInt) packetPtr->pktLinkDataPtr)
                     & ~((ttUserPtrCastToInt) (TM_ETHER_HW_ALIGN - 1)) );
            }
/*
 * Offset of ethernet frame to allow IP area to be long word aligned.
 */
            packetPtr->pktLinkDataPtr += TM_ETHER_IP_ALIGN;
        }
/* Queue packet */
#ifndef TM_INDRV_INLINE_SEND_RECV
/* Packet is being written into */
        packetPtr->pktUserFlags |= TM_PKTF_USER_DEV_WRITE;
        packetPtr->pktRecvNextPtr = (ttPacketPtr)0;
        if (devEntryPtr->devReceiveQueuePtr == TM_PACKET_NULL_PTR)
        {
/* Empty queue. */
            devEntryPtr->devReceiveQueuePtr = packetPtr;
            devEntryPtr->devRcvLastQueuePtr = packetPtr;
        }
        else
        {
/* Non Empty queue. Queue at the end of the queue */
            devEntryPtr->devRcvLastQueuePtr->pktRecvNextPtr = packetPtr;
            devEntryPtr->devRcvLastQueuePtr = packetPtr;
        }
#else /* TM_INDRV_INLINE_SEND_RECV */
/* Single packet queue */
        devEntryPtr->devReceiveQueuePtr = packetPtr;
#endif /* TM_INDRV_INLINE_SEND_RECV */
        devEntryPtr->devReceiveDataPtr = packetPtr->pktLinkDataPtr;
        devEntryPtr->devReceiveDataLength = 0;
    }
    {
        tm_bcopy(dataPtr, 
                 devEntryPtr->devReceiveDataPtr, 
                 tm_packed_byte_count(dataLength));
        devEntryPtr->devReceiveDataPtr += tm_packed_byte_len(dataLength);
    }
    devEntryPtr->devReceiveDataLength = (tt16Bit)
                                       (   devEntryPtr->devReceiveDataLength
                                         + (tt16Bit)dataLength );
    if (flag != TM_USER_BUFFER_MORE)
    {
        tfSendCompleteInterface( interfaceHandle,
                                 TM_DEV_SEND_COMPLETE_DRIVER );
/* The final segment so send it up */
/*
 * We can inline the tfRecvInterface() function keeping only what we need
 * (if TM_INDRV_INLINE_SEND_RECV is defined), otherwise the user need
 * to call tfRecvInterface().
 */
#ifdef TM_INDRV_INLINE_SEND_RECV
        dataLength = (int)devEntryPtr->devReceiveDataLength;
        packetPtr->pktLinkDataLength = (ttPktLen)dataLength;
        packetPtr->pktChainDataLength = (ttPktLen)dataLength;
#ifdef TM_SNMP_MIB
/* We bypass tfRecvInterface */
        devEntryPtr->devInOctets += (tt32Bit)dataLength;
#endif /* TM_SNMP_MIB */
        devEntryPtr->devReceiveQueuePtr = TM_PACKET_NULL_PTR;
/* Unlock the device send/receive */
        tm_unlock(&(devEntryPtr->devDriverLockEntry));
        errorCode = (*(devEntryPtr->devLinkLayerProtocolPtr->lnkRecvFuncPtr))(
                                               (void TM_FAR *)interfaceHandle,
                                               (void TM_FAR *)packetPtr);
        tm_lock_wait(&(devEntryPtr->devDriverLockEntry));
/* End of in line tfRecvInterface */
#else /* !TM_INDRV_INLINE_SEND_RECV */
        packetPtr->pktLinkDataLength = devEntryPtr->devReceiveDataLength;
        packetPtr->pktUserFlags = TM_PKTF_USER_DEV_OWNS;
/* Notify the Treck stack that we've received a packet */
        tfNotifyInterfaceTask(interfaceHandle, 1, 0, 0UL, 0UL);
#endif /* TM_INDRV_INLINE_SEND_RECV */
    }
testSendFinish:
    return errorCode;    

}

int tfIntfDrvRecv ( ttUserInterface    interfaceHandle,
                    ttCharPtr TM_FAR  *dataPtrPtr,
                    ttIntPtr           dataSizePtr,
                    ttUserBufferPtr    packetPtrPtr
#ifdef TM_DEV_RECV_OFFLOAD
                  , ttDevRecvOffloadPtr recvOffloadPtr
#endif /* TM_DEV_RECV_OFFLOAD */
                  )
{
#ifdef TM_INDRV_INLINE_SEND_RECV
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(dataPtrPtr);
    TM_UNREF_IN_ARG(dataSizePtr);
    TM_UNREF_IN_ARG(packetPtrPtr);
#ifdef TM_DEV_RECV_OFFLOAD
    TM_UNREF_IN_ARG(recvOffloadPtr); /* compiler warning for unused parameter */
#endif /* TM_DEV_RECV_OFFLOAD */
    return TM_ENOERROR;
#else /* !TM_INDRV_INLINE_SEND_RECV */
    ttDeviceEntryPtr devEntryPtr;
    ttPacketPtr      packetPtr;
    int              errorCode;

#ifdef TM_DEV_RECV_OFFLOAD
    TM_UNREF_IN_ARG(recvOffloadPtr); /* compiler warning for unused parameter */
#endif /* TM_DEV_RECV_OFFLOAD */
    devEntryPtr = (ttDeviceEntryPtr)interfaceHandle;
/* Check at head of the queue */
    packetPtr = devEntryPtr->devReceiveQueuePtr;
    if (    (packetPtr != TM_PACKET_NULL_PTR) 
         && tm_16bit_bits_not_set(packetPtr->pktFlags,
                                  TM_PKTF_USER_DEV_WRITE) )
/* Packet in the queue, and ready to be received */
    {
/* unlink this packet from the head of the queue of received packets */
        devEntryPtr->devReceiveQueuePtr = 
                devEntryPtr->devReceiveQueuePtr->pktRecvNextPtr;
        if (devEntryPtr->devReceiveQueuePtr == (ttPacketPtr)0)
        {
/* Empty queue */
            devEntryPtr->devRcvLastQueuePtr = (ttPacketPtr)0;
        }
        packetPtr->pktRecvNextPtr = (ttPacketPtr)0;
/* Back to the caller */
        *dataSizePtr = (int)packetPtr->pktLinkDataLength;
        *dataPtrPtr = (ttCharPtr)packetPtr->pktLinkDataPtr;
        errorCode = TM_ENOERROR;
    }
    else
    {
        packetPtr = (ttPacketPtr)0;
        errorCode = TM_ENOBUFS;
    }
    *packetPtrPtr = packetPtr;
    return errorCode;
#endif /* TM_INDRV_INLINE_SEND_RECV */
}

/*
 * Can be used to configure the device or refill the receive pool, or
 * enable reception of a list of multicast addresses.
 */
int tfIntfDrvIoctl ( ttUserInterface interfaceHandle, int flag,
                     void TM_FAR * optionPtr, int optionLen )
{

    TM_UNREF_IN_ARG(interfaceHandle);
    optionPtr       = optionPtr;
    optionLen       = optionLen;

#ifdef TM_IGMP
    flag            = flag;

/*
 * if (flag == TM_DEV_SET_MCAST_LIST)
 * Enable reception for all multicast addresses on the list
 */
/*
 * Else if (flag == TM_DEV_SET_ALL_MCAST)
 * Enable reception of all multicast addresses
 */

#else /* !TM_IGMP */
    flag            = flag;

#endif /* TM_IGMP */

    return TM_ENOERROR;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:48      For: Xilinx-Samsung
 * File Name: trindrv.c     File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 19210    Output checksum:  2399
 */
