/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: SLIP Link Layer
 *
 * Filename: trslip.c
 * Author: Jason
 * Date Created: 7/1/98
 * $Source: source/trslip.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:48:29CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>


#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * Local macros
 */

#define TM_SLIP_VECT_NULL_PTR (ttSlipVectPtr)0

#define TM_SLIP_ESCAPE_CHAR   (tt8Bit)0xDB
#define TM_SLIP_ESCAPE_END    (tt8Bit)0xDC
#define TM_SLIP_ESCAPE_ESC    (tt8Bit)0xDD
#define TM_SLIP_END_CHAR      (tt8Bit)0xC0

/*
 * local functions
 */
int tfSlipOpen(void TM_FAR * interfaceId,
               tt16Bit       multiHomeIndex,
               int           protocolFamily);
int tfSlipClose(void TM_FAR * interfaceId);
int tfSlipSend(ttPacketPtr packetPtr);
int tfSlipStuffPacket(ttPacketPtr packetPtr);
int tfSlipRecv(void TM_FAR *interfaceId, void TM_FAR *bufferHandle);
int tfSlipIoctl(int operation);
int tfSlipError(int type);

ttUserLinkLayer tfUseSlip (void)
{
    ttLinkLayerEntryPtr linkLayerEntryPtr;

/* Search for Loopback Protocol in Link Layer List */
/* Lock the Link Layer list */
    tm_call_lock_wait(&tm_context(tvLinkLayerListLock));
    linkLayerEntryPtr = tm_context(tvLinkLayerListPtr);
/* Check for duplicate */
    while (linkLayerEntryPtr != TM_LINK_NULL_PTR)
    {
        if (    tm_strcmp( (char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),
                           "SLIP" )
             == TM_STRCMP_EQUAL )
        {
            goto useSlipUnLock; /* entry already in */
        }
        linkLayerEntryPtr = linkLayerEntryPtr->lnkNextPtr;
    }
/* if none exists then malloc an entry */
    linkLayerEntryPtr = (ttLinkLayerEntryPtr)
                         tm_kernel_malloc((unsigned)sizeof(ttLinkLayerEntry));
    if (linkLayerEntryPtr == TM_LINK_NULL_PTR)
    {
/* if the malloc fails then return a null pointer */
        goto useSlipUnLock;
    }
    tm_bzero(linkLayerEntryPtr, sizeof(ttLinkLayerEntry));
    linkLayerEntryPtr->lnkMtu = tm_context(tvSlipMtu);
    linkLayerEntryPtr->lnkOpenFuncPtr = tfSlipOpen;
    linkLayerEntryPtr->lnkCloseFuncPtr = tfSlipClose;
    linkLayerEntryPtr->lnkSendFuncPtr = tfSlipSend;
    linkLayerEntryPtr->lnkRecvFuncPtr = tfSlipRecv;
    linkLayerEntryPtr->lnkIoctlFuncPtr = tfSlipIoctl;
    linkLayerEntryPtr->lnkErrorFuncPtr = tfSlipError;
    linkLayerEntryPtr->lnkPostFuncPtr = tfSlipStuffPacket;
    linkLayerEntryPtr->lnkLinkLayerProtocol = TM_LINK_LAYER_SLIP;
    tm_strcpy((char TM_FAR *)(linkLayerEntryPtr->lnkNameArray), "SLIP");
/* Stuff it into the table */
    linkLayerEntryPtr->lnkNextPtr = tm_context(tvLinkLayerListPtr);
    tm_context(tvLinkLayerListPtr) = linkLayerEntryPtr;
useSlipUnLock:
    tm_call_unlock(&tm_context(tvLinkLayerListLock));
    return ((ttUserLinkLayer)linkLayerEntryPtr);
}


/*
 * Open an Slip Link Layer
 */
int tfSlipOpen ( void TM_FAR * interfaceId,
                 tt16Bit       multiHomeIndex,
                 int           protocolFamily)
{
    ttSlipVectPtr slipVectPtr;
    int           errorCode;
#define devEntryPtr ((ttDeviceEntryPtr)interfaceId)

    TM_UNREF_IN_ARG(protocolFamily);
    
    slipVectPtr = (ttSlipVectPtr)(devEntryPtr->devStateVectorPtr);
    if (slipVectPtr == TM_SLIP_VECT_NULL_PTR)
    {
        errorCode = tfSlipVectInit(devEntryPtr);
    }
    else
    {
/* Clean up previous session, if necessary */
        (void)tfSlipClose(interfaceId);
        errorCode = TM_ENOERROR;
    }
/* Add it to our local routing table */
    if (errorCode == TM_ENOERROR)
    {
        errorCode = tfDeviceStart(devEntryPtr, multiHomeIndex, PF_INET);
    }
    return errorCode;
#undef devEntryPtr
}

/*
 * Close a SLIP Link Layer
 */
int tfSlipClose ( void TM_FAR *interfaceId )
{
    ttSlipVectPtr slipVectPtr;
    ttPacketPtr   packetPtr;
#define devEntryPtr ((ttDeviceEntryPtr)interfaceId)

    slipVectPtr = (ttSlipVectPtr)(devEntryPtr->devStateVectorPtr);
    packetPtr = slipVectPtr->slipsPacketPtr;
    if (packetPtr != TM_PACKET_NULL_PTR)
    {
        tfFreeSharedBuffer(packetPtr, TM_SOCKET_UNLOCKED);
        slipVectPtr->slipsPacketPtr = TM_PACKET_NULL_PTR;
    }
    slipVectPtr->slipsWorkingPtr = TM_8BIT_NULL_PTR;
    slipVectPtr->slipsLastByteStuffState = 0;
    return TM_LL_OKAY;
#undef devEntryPtr
}

int tfSlipSend ( ttPacketPtr packetPtr )
{
    TM_UNREF_IN_ARG(packetPtr);
    return TM_ENOERROR;
}

/*
 * Encapsulate the packet and send it on to the device
 */
int tfSlipStuffPacket ( ttPacketPtr packetPtr )
{
    ttDevSendFuncPtr devSendFuncPtr;
    ttDeviceEntryPtr devEntryPtr;
    ttSlipVectPtr    slipVectPtr;
    tt8BitPtr        dataPtr;
    tt8BitPtr        sendDataPtr;
    tt32Bit          length;
#ifdef TM_SNMP_MIB
    tt32Bit          outOctets;
#endif /* TM_SNMP_MIB */
    int              sendBufSize;
    int              sendSize;
    int              errorCode;
    tt8Bit           mapChar;

    devEntryPtr = packetPtr->pktDeviceEntryPtr;
    slipVectPtr = (ttSlipVectPtr)devEntryPtr->devStateVectorPtr;
    devSendFuncPtr = devEntryPtr->devSendFuncPtr;
    length = packetPtr->pktLinkDataLength;
    dataPtr = packetPtr->pktLinkDataPtr;
#ifdef TM_SNMP_MIB
    outOctets = length;
#endif /* TM_SNMP_MIB */
/* Chain in the packet in the device send queue */
    packetPtr->pktChainNextPtr = TM_PACKET_NULL_PTR;
    sendBufSize = (int)slipVectPtr->slipsSendBufSize;
    sendDataPtr = slipVectPtr->slipsSendDataPtr;
    sendSize = 0;
/* Loop thru all the characters, checking to see if they need escaping */
    while (length != TM_UL(0))
    {
        switch (*dataPtr)
        {
/* End character, Escape character */
            case TM_SLIP_END_CHAR:
                sendDataPtr[sendSize] = TM_SLIP_ESCAPE_CHAR;
                mapChar = TM_SLIP_ESCAPE_END;
#ifdef TM_SNMP_MIB
/*
 * Increase by 1 (instead of 2), because the escaped character is
 * already included in outOctets.
 */
                outOctets++;
#endif /* TM_SNMP_MIB */
                break;
/* Escape character */
            case TM_SLIP_ESCAPE_CHAR:
                sendDataPtr[sendSize] = TM_SLIP_ESCAPE_CHAR;
                mapChar = TM_SLIP_ESCAPE_ESC;
#ifdef TM_SNMP_MIB
/*
 * Increase by 1 (instead of 2), because the escaped character is
 * already included in outOctets.
 */
                outOctets++;
#endif /* TM_SNMP_MIB */
                break;
            default:
                sendDataPtr[sendSize] = *dataPtr;
                mapChar = 0;
                break;
        }
        sendSize++;
    
        if (sendSize >= sendBufSize)
        {
            (void)(*(devSendFuncPtr))( (ttUserInterface)devEntryPtr,
                                       (char TM_FAR *)sendDataPtr,
                                       sendSize,
                                       TM_USER_BUFFER_MORE );
            sendSize = 0;
        }
        dataPtr++;
        if (mapChar != 0)
        {
            sendDataPtr[sendSize] = mapChar;
            sendSize++;
            if (sendSize >= sendBufSize)
            {
                (void)(*(devSendFuncPtr))( (ttUserInterface)devEntryPtr,
                                           (char TM_FAR *)sendDataPtr,
                                           sendSize,
                                           TM_USER_BUFFER_MORE );
                sendSize = 0;
            }
        }
/* Check for scattered data */
        if (--length == TM_UL(0))
        {
            packetPtr = (ttPacketPtr)packetPtr->pktLinkNextPtr;
            if (packetPtr != TM_PACKET_NULL_PTR)
            {
                length = packetPtr->pktLinkDataLength;
                dataPtr = (tt8BitPtr)packetPtr->pktLinkDataPtr;
#ifdef TM_SNMP_MIB
                outOctets += length;
#endif /* TM_SNMP_MIB */
            }
        }
    }
    sendDataPtr[sendSize] = TM_SLIP_END_CHAR;
    sendSize++;
#ifdef TM_SNMP_MIB
/* Add the tlEndChar byte (not yet added to outOctets) */
    outOctets++;
#endif /* TM_SNMP_MIB */
/*
 * If the device driver function returns an error on the last buffer,
 * this function returns an error to its caller (either tfDeviceSend(),
 * or tfXmitInterface() (if a transmit task is used), in which case
 * the caller will call tfDeviceSendFailed() which will remove the
 * last queued packet from the send queue.
 */
    errorCode = (*(devSendFuncPtr))( (ttUserInterface)devEntryPtr,
                                     (char TM_FAR *)sendDataPtr,
                                     sendSize,
                                     TM_USER_BUFFER_LAST );
#ifdef TM_SNMP_MIB
    if (errorCode == TM_ENOERROR)
    {
        tm_64Bit_augm(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCOutOctets,
                      outOctets);
        tm_64Bit_augm(tm_context(tvDevIpv4Data).ipIfStatsHCOutOctets,
                      outOctets);
    }
/*
 * In case of an error interface SNMP MIB device variables updated in
 * tfDeviceSendFailed(), itself called from either tfXmitInterface() or
 * tfDeviceSend() when this function returns an error.
 */
#endif /* TM_SNMP_MIB */
    return errorCode;
}

/* 
 * Unstuff and process an incoming packet
 */
int tfSlipRecv ( void TM_FAR *interfaceId, void TM_FAR *bufferHandle )
{
    ttPacketPtr   packetPtr;
    tt8BitPtr     dataPtr;
    tt8BitPtr     outDataPtr;
    ttSlipVectPtr slipVectPtr;
    tt32Bit       length;
    int           retCode;
    tt16Bit       maxMtu;
    tt8Bit        prevData;
#define devEntryPtr ((ttDeviceEntryPtr)interfaceId)

    retCode = TM_ENOERROR;
    length = ((ttPacketPtr)bufferHandle)->pktLinkDataLength;
    dataPtr = ((ttPacketPtr)bufferHandle)->pktLinkDataPtr;

    tm_lock_wait(&(devEntryPtr->devLockEntry));
/* Restore our state information */
    slipVectPtr = (ttSlipVectPtr)devEntryPtr->devStateVectorPtr;

/*
 * Check if the SLIP state vector has not been allocated.  This could occur
 * if data is received after the device is opened but before SLIP is opened.  
 * Free the incoming data, unlock the device and return.
 */
    if (slipVectPtr == TM_SLIP_VECT_NULL_PTR)
    {
        goto slipRecvReturn; 
    }

    do
    {
        packetPtr = slipVectPtr->slipsPacketPtr;
        if (packetPtr == TM_PACKET_NULL_PTR)
        {
/* REVNOTE: use TM_MAX_HEADERS_SIZE instead of TM_LL_MAX_HDR_LEN for CSLIP */
            maxMtu = (tt16Bit)(tm_max(tm_context(tvSlipMtu),
                                      devEntryPtr->devMtu));
            packetPtr = tfGetSharedBuffer( TM_LL_MAX_HDR_LEN,
                                           (ttPktLen)(maxMtu+2),
                                           TM_16BIT_ZERO );
            if (packetPtr == TM_PACKET_NULL_PTR)
            {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
                devEntryPtr->dev4Ipv4Mib.ipIfStatsInDiscards++;
                tm_context(tvDevIpv4Data).ipIfStatsInDiscards++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
                break; /* out of the do/while loop */
            }
            slipVectPtr->slipsPacketPtr = packetPtr;
            slipVectPtr->slipsWorkingPtr = packetPtr->pktLinkDataPtr;
            slipVectPtr->slipsLastByteStuffState = 0;
            packetPtr->pktLinkDataLength = 0;
        }

        prevData = slipVectPtr->slipsLastByteStuffState;
        outDataPtr = slipVectPtr->slipsWorkingPtr;

        while (length--)
        {
            if (prevData == TM_SLIP_ESCAPE_CHAR)
            {
/*
 * If this would exceed the boundary of the packet, we need to free the
 * packet and break out of this while() loop, so we can process the next frame.
 */
                if (outDataPtr >= packetPtr->pktSharedDataPtr->dataBufLastPtr)
                {
                    length++;
                    tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
                    packetPtr = TM_PACKET_NULL_PTR;
                    slipVectPtr->slipsPacketPtr = TM_PACKET_NULL_PTR;
                    break;
                }

                if (*dataPtr == TM_SLIP_ESCAPE_END)
                {
                    *outDataPtr++ = TM_SLIP_END_CHAR;
                    packetPtr->pktLinkDataLength++;
                }
                else
                {
                    *outDataPtr++ = TM_SLIP_ESCAPE_CHAR;
                    packetPtr->pktLinkDataLength++;
                }
            }
            else
            {
/* For escape character, skip to next one */
                if (*dataPtr != TM_SLIP_ESCAPE_CHAR)
                {
                    if (*dataPtr == TM_SLIP_END_CHAR)
                    {
                        if (packetPtr->pktLinkDataLength != 0)
                        {
/* End of packet, so send it up */
                            packetPtr->pktNetworkLayer =
                                TM_NETWORK_IPV4_LAYER;
                            packetPtr->pktDeviceEntryPtr = devEntryPtr;
                            packetPtr->pktChainDataLength =
                                                packetPtr->pktLinkDataLength;
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
                            devEntryPtr->devInUcastPkts++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
/* Make sure current packet cannot be used for next frame */
                            slipVectPtr->slipsPacketPtr =
                                                           TM_PACKET_NULL_PTR;
                            tm_unlock(&(devEntryPtr->devLockEntry));
                            tfIpIncomingPacket(packetPtr
                                               );
                            tm_lock_wait(&(devEntryPtr->devLockEntry));
                            if ( length != 0 )
                            {
/*
 * If there's another SLIP frame, then allocate a new buffer. (Zero
 * packetPtr to continue the outer loop to allocate the new buffer.)
 */
                                packetPtr = TM_PACKET_NULL_PTR;
                                dataPtr++;
/* out of inner while (length--) loop */
                                break;
                            }
/*
 * Else (length == 0) we will break out of the 2 loops, because length
 * is 0, and packetPtr is non null
 */
                        }
/*
 * Else SLIP END CHAR is at beginning of a packet. Do not copy it, and
 * keep processing the characters handed to us by the device driver.
 */
                    }
                    else
                    {
/* A regular character, just add it to our packet */
/*
 * If this would exceed the boundary of the packet, we need to free the
 * packet and break out of this while() loop, so we can process the next frame.
 */
                        if (   (outDataPtr)
                            >= (packetPtr->pktSharedDataPtr->dataBufLastPtr))
                        {
                            length++;
                            tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
                            packetPtr = TM_PACKET_NULL_PTR;
                            slipVectPtr->slipsPacketPtr = TM_PACKET_NULL_PTR;
                            break;
                        }
                        *outDataPtr++ = *dataPtr;
                        packetPtr->pktLinkDataLength++;
                    }
                }
            }

            prevData = *dataPtr++;
        }
/* Save our state information */
        slipVectPtr->slipsLastByteStuffState = prevData;
        slipVectPtr->slipsWorkingPtr = outDataPtr;
    }
/*
 * Continue the do while loop only if we need to re-allocate a packet
 * because there is still some data in the buffer handed to us from
 * the device driver, and the data belongs to a new frame.
 */
    while (packetPtr == TM_PACKET_NULL_PTR);

slipRecvReturn:
    tm_unlock(&(devEntryPtr->devLockEntry));

/* Free the incoming packet since we are done with it */
    tfFreePacket((ttPacketPtr)bufferHandle, TM_SOCKET_UNLOCKED);

    return (retCode);
#undef devEntryPtr
}

/*
 * Stub functions
 */

int tfSlipIoctl ( int operation )
{
    TM_UNREF_IN_ARG(operation);
    return TM_LL_OKAY;
}

int tfSlipError ( int type )
{
    TM_UNREF_IN_ARG(type);
    return TM_LL_OKAY;
}

int tfSlipVectInit (ttDeviceEntryPtr devEntryPtr)
{
    ttSlipVectPtr slipVectPtr;
    int           errorCode;

    slipVectPtr = (ttSlipVectPtr)tm_kernel_malloc(
                                                (unsigned)sizeof(ttSlipVect));
    if (slipVectPtr == TM_SLIP_VECT_NULL_PTR)
    {
        errorCode = TM_ENOBUFS;
    }
    else
    {
        devEntryPtr->devStateVectorPtr = (ttVoidPtr)slipVectPtr;
        slipVectPtr->slipsPacketPtr = TM_PACKET_NULL_PTR;
        slipVectPtr->slipsWorkingPtr = TM_8BIT_NULL_PTR;
        slipVectPtr->slipsLastByteStuffState = 0;
        slipVectPtr->slipsSendDataPtr = &(slipVectPtr->slipsSendBuf[0]);
#if (TM_SLIP_DEF_SEND_SIZE > TM_SLIP_DEF_BUF_SIZE)
/* Do not let the user make the send size bigger than the buffer size */
        slipVectPtr->slipsSendBufSize = TM_SLIP_DEF_BUF_SIZE;
#else /* !(TM_SLIP_DEF_SEND_SIZE > TM_SLIP_DEF_BUF_SIZE) */
        slipVectPtr->slipsSendBufSize = TM_SLIP_DEF_SEND_SIZE;
#endif /* !(TM_SLIP_DEF_SEND_SIZE > TM_SLIP_DEF_BUF_SIZE) */
        errorCode = TM_ENOERROR;
    }
    return errorCode;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:52      For: Xilinx-Samsung
 * File Name: trslip.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 25017    Output checksum: 42740
 */
