/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2010
 * Copyright Elmic Systems Japan 1997 - 2010
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description:
 *
 *     When one of the interface drivers(trinfdrv.c, trindrmc.c or tfinscdr.c)
 *     is used for testing, we can't get the packet captured from the wire.
 *     By adding trsniff.c and #define TM_USE_SNIFF in your project,
 *     "*.pcap" will be created in you working directory, if you run it in
 *     Visual C, it is the directory when your project file is.
 *
 *     The file is in the standard libpcap format and therefore can be viewed
 *     by most packet capture softwares.
 *
 *     When using ethereal, turn off the option for "enable network name
 *     translation" option to avoid the delay for trying to find the host names
 *     for the IP address.
 *
 *     When use tcpdump,  using "tcpdump -n -r <pcap>" "-n" disables name
 *     resolution. There is a Windows version of tcpdump, available here,
 *     http://windump.polito.it/install/default.htm. It works like the
 *     UNIX/Linux version except that it displays all IPv6 address as <null>.
 *
 * Filename:    trsniff.c
 * Author:      Bryan Wang
 * Date Created:11/19/02
 * $Source: source/trsniff.c $
 *
 * Modification History
 * $Revision: 6.0.2.4 $
 * $Date: 2011/09/07 00:35:25CDT $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */


#include <trsocket.h>

#ifdef TM_USE_SNIFF

#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

#ifdef TM_USE_CLIB
/* for timezone offset purpose only */
#include <time.h>
#define TM_TIMEZONE     timezone
#else /* TM_USE_CLIB */
#define TM_TIMEZONE     (8L * 60 * 60) /* pacific timezone offset */
#endif /* TM_USE_CLIB */

#define TM_LINKTYPE_NULL            0   /* 4 bytes network protocol ID */
#define TM_LINKTYPE_ETHERNET        1   /* Ethernet */
#define TM_LINKTYPE_SLIP            8   /* Serial Line IP */
#define TM_LINKTYPE_PPP             9   /* Point-to-point Protocol */
#define TM_LINKTYPE_FDDI            10  /* FDDI */

#define TM_TCPDUMP_MAGIC            0xa1b2c3d4
#define TM_PCAP_VERSION_MAJOR       2
#define TM_PCAP_VERSION_MINOR       4

#define TM_PCAP_FILE_NAME_NULL          "null.cap"
#define TM_PCAP_FILE_NAME_ETHER         "ether.cap"
#define TM_PCAP_FILE_NAME_SLIP          "slip.cap"
#define TM_PCAP_FILE_NAME_PPP           "ppp.cap"
#define TM_PCAP_FILE_NAME_TUNNEL        "tunnel.cap"
#define TM_PCAP_FILE_NAME_MN_VIRTUAL    "mnvirtual.cap"


typedef struct tsPcapFileHeader {
  tt32Bit   magic;
  tt16Bit   version_major;
  tt16Bit   version_minor;
  ttS32Bit  thiszone;   /* gmt to local correction */
  tt32Bit   sigfigs;    /* accuracy of timestamps */
  tt32Bit   snaplen;    /* max length saved portion of each pkt */
  tt32Bit   linktype;   /* data link type (LINKTYPE_*) */
} ttPcapFileHeader;


typedef struct tsPcapRecordHeader
{
  struct timeval    timeStamp;     /* time stamp */
  tt32Bit           capLen; /* length of portion captured */
  tt32Bit           pktLen; /* length this packet (off wire) */
} ttPcapRecordHeader;


/* open a file, writes the libpcap header into it, and close it */
static int tfPcapOpenFile(
    const char *    fileName,
    tt8Bit          linkType);

#ifdef TM_USE_INF_DRV_SNIFF
/* writes an packet as a record into the opened libpcap file*/
static int tfPcapWriteRecord(
    ttVoidPtr               filePtr,
    tt8Bit                  linkType,
    ttPcapRecordHeader *    recHdrPtr,
    const char*             pktDataPtr);
#endif /* TM_USE_INF_DRV_SNIFF */

/* to avoid reopen file when write a packet */
void TM_FAR * tlEtherFilePtr = 0;
void TM_FAR * tlPppFilePtr  = 0;
void TM_FAR * tlSlipFilePtr = 0;
void TM_FAR * tlNullFilePtr = 0;
void TM_FAR * tlTunnelFilePtr = 0;
void TM_FAR * tlCnVirtualFilePtr = 0;

/* write a single packet into the pcap file, supporting scattered packets */
int tfPcapWritePacket(ttPacketPtr packetPtr)
{
    int                 errorCode;
    char*               fileName;
    tt8Bit              linkType;
    ttPcapRecordHeader  recHdr;
    ttVoidPtrPtr        filePtrPtr;
    ttDeviceEntryPtr    devPtr;
/* 4 byte network protocol ID for null link layer record fromat */
    tt32Bit             networkProtocol;
    tt32Bit             seconds;

    errorCode = TM_ENOERROR;
    devPtr = packetPtr->pktDeviceEntryPtr;

    tm_call_lock_wait(&tm_global(tvSniffLock));

    recHdr.pktLen = packetPtr->pktChainDataLength;
    switch(packetPtr->pktDeviceEntryPtr->devLinkLayerProtocol)
    {
    case TM_LINK_LAYER_NULL:
        linkType = TM_LINKTYPE_NULL;
        recHdr.pktLen += sizeof(networkProtocol);
        fileName = TM_PCAP_FILE_NAME_NULL;
        filePtrPtr = &tlNullFilePtr;
        break;
    case TM_LINK_LAYER_E8023:
    case TM_LINK_LAYER_ETHERNET:
        linkType = TM_LINKTYPE_ETHERNET;
        fileName = TM_PCAP_FILE_NAME_ETHER;
        filePtrPtr = &tlEtherFilePtr;
        break;
    case TM_LINK_LAYER_SLIP:
/*
 * SLIP is actually raw IP with header, trailer and escape chars,
 * we encapsulate it as NULL link-layer frames
 * with a file name indicates SLIP
 */
        linkType = TM_LINKTYPE_NULL;
        recHdr.pktLen += sizeof(networkProtocol);
        fileName = TM_PCAP_FILE_NAME_SLIP;
        filePtrPtr = &tlSlipFilePtr;
        break;
    case TM_LINK_LAYER_PPP:
        linkType = TM_LINKTYPE_PPP;
        fileName = TM_PCAP_FILE_NAME_PPP;
        filePtrPtr = &tlPppFilePtr;
/* Allow room for the device name to be appended */
        recHdr.pktLen += (tt32Bit)tm_strlen(devPtr->devNameArray);
        break;
    case TM_LINK_LAYER_V6_TUNNEL:
        linkType = TM_LINKTYPE_NULL;
        fileName = TM_PCAP_FILE_NAME_TUNNEL;
        filePtrPtr = &tlTunnelFilePtr;
        break;
    default:
        errorCode = TM_EPROTONOSUPPORT;
        goto writeFrameEnd;
    }

    recHdr.capLen = recHdr.pktLen;

    tm_kernel_set_critical;
/* use tvTime for the timestap */
    seconds = tvTime / 1000;
    recHdr.timeStamp.tv_sec = seconds + TM_TIMEZONE;
    recHdr.timeStamp.tv_usec = 1000*(tvTime - (seconds*1000));
    tm_kernel_release_critical;

    if(*filePtrPtr == (void*)0)
    {
/*
 * file not opened yet, let's open the file in write mode and
 * write the header into it, if the file doesn't exist, it'll be
 * created, if exists, its contents are lost, therefore, we start
 * always start with a empty file here.
 * file is closed after writing the header
 */
        tfPcapOpenFile(fileName, linkType);

/*
 * open the file in append mode, so that following call to tfPcapWriteFrame
 * can append into it without reopen it
 */
        *filePtrPtr = tfFSOpenFile ((ttVoidPtr)0,
                                    fileName,
                                    TM_FS_APPEND,
                                    TM_TYPE_BINARY,
                                    TM_STRU_RECORD);
        if(*filePtrPtr == (void*)0)
        {
            errorCode = TM_EIO;
            goto writeFrameEnd;
        }
    }

/*
 * Here we have the proper (ttVoidPtr)0, filePtrPtr, linkType, recHdr,
 * read to write
 */

/* write the record header */
    errorCode = tfFSWriteFile((ttVoidPtr)0,
                              *filePtrPtr,
                              (char*)&recHdr,
                              sizeof(ttPcapRecordHeader));
    if (errorCode == TM_SOCKET_ERROR)
    {
        errorCode = TM_EIO;
        goto writeFrameEnd;
    }

    if (linkType == TM_LINKTYPE_NULL)
    {
/* write the network protocol ID into the record first*/
        switch (packetPtr->pktNetworkLayer)
        {
        case TM_NETWORK_IPV4_LAYER:
            networkProtocol = htons(TM_ETHER_IP_TYPE);
            break;
        case TM_NETWORK_ARP_LAYER:
            networkProtocol = htons(TM_ETHER_ARP_TYPE);
                break;
        default:
            errorCode = TM_EPROTONOSUPPORT;
            goto writeFrameEnd;
        }

        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                  *filePtrPtr,
                                  (char*)&networkProtocol,
                                  sizeof(networkProtocol));
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeFrameEnd;
        }
    }

/* write the whole packet into the file, link by link */
    while(packetPtr != TM_PACKET_NULL_PTR)
    {
        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                  *filePtrPtr,
                                  (char*)(packetPtr->pktLinkDataPtr),
                                  packetPtr->pktLinkDataLength);
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeFrameEnd;
        }
        packetPtr = packetPtr->pktLinkNextPtr;
    }

/* if this is PPP, append the device name to identify the direction */
    if (linkType == TM_LINKTYPE_PPP)
    {
        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                    *filePtrPtr,
                                    (char*)devPtr->devNameArray,
                                    (int)tm_strlen(devPtr->devNameArray));
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeFrameEnd;
        }
    }

/* flush the file, this is 50% faster than closing and reopen the file */
    tfFSFlushFile((ttVoidPtr)0, *filePtrPtr);

writeFrameEnd:
    tm_call_unlock(&tm_global(tvSniffLock));
    return errorCode;
}


/* open a file, writes the libpcap header into it, and close it */
int tfPcapOpenFile(
    const char *    fileName,
    tt8Bit          linkType)
{
    int                 errorCode;
    void TM_FAR *       filePtr;
    ttPcapFileHeader    fileHdr;

    errorCode = TM_ENOERROR;

/* if fileName exists, it'll be overwritten */
    filePtr = tfFSOpenFile ((ttVoidPtr)0,
                            fileName,
                            TM_FS_WRITE,
                            TM_TYPE_BINARY,
                            TM_STRU_RECORD);
    if(filePtr == (void*)0)
    {
        errorCode = TM_EIO;
        goto crateFileEnd;
    }

    fileHdr.magic         = TM_TCPDUMP_MAGIC;
    fileHdr.version_major = TM_PCAP_VERSION_MAJOR;
    fileHdr.version_minor = TM_PCAP_VERSION_MINOR;

    fileHdr.thiszone  = 0;      /* int time zone offset */
    fileHdr.snaplen   = 65535;  /* max num of packets in the file */
    fileHdr.sigfigs   = 0;
    fileHdr.linktype  = linkType;

/* writes the libpcap file header into the opened file*/
    errorCode = tfFSWriteFile((ttVoidPtr)0,
                              filePtr,
                              (char *)&fileHdr,
                              sizeof(fileHdr));

    if (errorCode == TM_SOCKET_ERROR)
    {
        errorCode = TM_EIO;
    }

    tfFSCloseFile((ttVoidPtr)0, filePtr);

crateFileEnd:
    return errorCode;
}



/*
 * writes an packet as a record into the opened libpcap file, for sniffing
 * in the interface dirvers
 */
#ifdef TM_USE_INF_DRV_SNIFF
static int tfPcapWriteRecord(
    ttVoidPtr               filePtr,
    tt8Bit                  linkType,
    ttPcapRecordHeader *    recHdrPtr,
    const char*             pktDataPtr)
{
    int     errorCode;
    tt32Bit networkProtocol;

    errorCode = TM_ENOERROR;


    if (linkType == TM_LINKTYPE_NULL)
    {
/* adjust the packet length for the extra 4 byte network proto ID */
        recHdrPtr->capLen += 4;
        recHdrPtr->pktLen += 4;
    }

    errorCode = tfFSWriteFile((ttVoidPtr)0,
                              filePtr,
                              (char*)recHdrPtr,
                              sizeof(ttPcapRecordHeader));
    if (errorCode == TM_SOCKET_ERROR)
    {
        errorCode = TM_EIO;
        goto writeRecordEnd;
    }

    if (linkType == TM_LINKTYPE_NULL)
    {
/*
 * for null link layer, we need to extract the IP version (4 or 6) out of the
 * packet data to obtain the network protocol ID and write it in after the
 * record header, in host order
 */
        if (((*pktDataPtr & 0xF0) >> 4) == 4 )
        {
/* networkProtocol = TM_ETHER_IP_TYPE;*/
            networkProtocol = PF_INET;
        }
        else if (((*pktDataPtr & 0xF0) >> 4) == 6 )
        {
/* networkProtocol = TM_6_ETHER_IP_TYPE;*/
            networkProtocol = PF_INET6;
        }
        else
        {
            errorCode = TM_EPROTONOSUPPORT;
            goto writeRecordEnd;
        }

        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                  filePtr,
                                  (char*)&networkProtocol,
                                  4);
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeRecordEnd;
        }

/* write the network layer packet */
        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                  filePtr,
                                  (char*)pktDataPtr,
                                  recHdrPtr->capLen - 4);
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeRecordEnd;
        }
    }
    else
    {
        errorCode = tfFSWriteFile((ttVoidPtr)0,
                                  filePtr,
                                  (char*)pktDataPtr,
                                  recHdrPtr->capLen);
        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = TM_EIO;
            goto writeRecordEnd;
        }
    }
writeRecordEnd:
    return errorCode;
}
#endif /* TM_USE_INF_DRV_SNIFF */


/*
 * write a single packet into the pcap file, for sniffing
 * in the interface dirvers
 */
#ifdef TM_USE_INF_DRV_SNIFF
int tfPcapWriteFrame(
    tt16Bit linkLayerProtocol,
    char*   dataPtr,
    int     dataLen)
{
    int                 errorCode;
    char*               fileName;
    tt8Bit              linkType;
    ttPcapRecordHeader  recHdr;
    ttVoidPtrPtr        filePtrPtr;

    recHdr.capLen = dataLen;
    recHdr.pktLen = dataLen;

    errorCode = TM_ENOERROR;

    switch(linkLayerProtocol)
    {
    case TM_LINK_LAYER_NULL:
        linkType = TM_LINKTYPE_NULL;
        fileName = TM_PCAP_FILE_NAME_NULL;
        filePtrPtr = &tlNullFilePtr;
        break;
    case TM_LINK_LAYER_E8023:
    case TM_LINK_LAYER_ETHERNET:
        linkType = TM_LINKTYPE_ETHERNET;
        fileName = TM_PCAP_FILE_NAME_ETHER;
        filePtrPtr = &tlEtherFilePtr;
        break;
    case TM_LINK_LAYER_SLIP:
/*
 * SLIP is actually raw IP, we encapsulate it as NULL link-layer frames
 * though the file name indicates slip
 */
        linkType = TM_LINKTYPE_NULL;
        fileName = TM_PCAP_FILE_NAME_SLIP;
        filePtrPtr = &tlSlipFilePtr;
        break;
    case TM_LINK_LAYER_PPP:
        linkType = TM_LINKTYPE_PPP;
        fileName = TM_PCAP_FILE_NAME_PPP;
        filePtrPtr = &tlPppFilePtr;
        break;
    case TM_LINK_LAYER_V6_TUNNEL:
        linkType = TM_LINKTYPE_NULL;
        fileName = TM_PCAP_FILE_NAME_TUNNEL;
        filePtrPtr = &tlTunnelFilePtr;
        break;
    default:
        errorCode = TM_EPROTONOSUPPORT;
        goto writeFrameEnd;
    }

    tm_kernel_set_critical;
/* use tvTime for the timestap */
#ifdef TM_USE_CLIB
    recHdr.timeStamp.tv_sec = tvTime / 1000 + timezone;
#else /* TM_USE_CLIB */
    recHdr.timeStamp.tv_sec = tvTime / 1000 + 8 * 60 * 60; /* pacific timezone offset */
#endif /* TM_USE_CLIB */
    recHdr.timeStamp.tv_usec = tvTime - recHdr.timeStamp.tv_sec;
    tm_kernel_release_critical;

    if(*filePtrPtr == (void*)0)
    {
/*
 * file not opened yet, let's open the file in write mode and
 * write the header into it, if the file doesn't exist, it'll be
 * created, if exists, its contents are lost, therefore, we start
 * always start with a empty file here.
 * file is closed after writing the header
 */
        tfPcapOpenFile(fileName, linkType);

/*
 * open the file in append mode, so that following call to tfPcapWriteFrame
 * can append into it without reopen it
 */
        *filePtrPtr = tfFSOpenFile ((ttVoidPtr)0,
                                    fileName,
                                    TM_FS_APPEND,
                                    TM_TYPE_BINARY,
                                    TM_STRU_RECORD);
        if(*filePtrPtr == (void*)0)
        {
            errorCode = TM_EIO;
            goto writeFrameEnd;
        }
    }

/* write this frame into the file */
    tfPcapWriteRecord(*filePtrPtr, linkType, &recHdr, dataPtr);

/* flush the file, this is 50% faster than closing and reopen the file */
    tfFSFlushFile((ttVoidPtr)0, *filePtrPtr);

writeFrameEnd:
    return errorCode;
}
#endif /* TM_USE_INF_DRV_SNIFF */



#else /* TM_USE_SNIFF */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */

int tvSniffDummy = 0;

#endif /* TM_USE_SNIFF */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:52      For: Xilinx-Samsung
 * File Name: trsniff.c     File Revision: 6.0.2.4 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 27196    Output checksum: 41067
 */
