/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Telnet server Protocol
 * Filename: trtelntd.c
 * Author: Odile
 * Date Created: 06/07/99
 * $Source: source/trtelntd.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2010/07/08 05:31:31CDT $
 * $Author: pcarney $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_USE_TELNETD
#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/*
 * This module contains the Telnet server implementation, and follows the
 * Telnet protocol from RFC 854, and RFC 1123.
 *
 * Features of the Telnet server implementation:
 *
 * . Task requirement:
 * The Telnet server can run either in the context of a task (for
 * multi-tasking systems), or from an executive loop (for single-tasking
 * systems). In both cases it allows concurrent incoming connections.
 * For multi-tasking systems, only one task (the Telnet server task)
 * is required.
 *
 * . Managing concurrent imcoming connections. Telnet server run q:
 * To manage concurrent incoming connections, The Telnet server uses a run q
 * of active connection blocks and sends and recvs data to and from the
 * network in non-blocking mode.
 * It uses the call back feature of the Treck TCP/IP stack to enter active
 * blocks in the run q.
 *
 * Limitations: The only supported TCP options are SGA, ECHO, and binary
 *
 */

/* Compile options definitions. */
#ifdef TM_LINT
LINT_UNREF_MACRO(tm_teld_q_empty)
LINT_UNREF_MACRO(TM_NVT_SPACE)
LINT_UNREF_MACRO(TM_TC_GA)
LINT_UNREF_MACRO(TM_TC_EL)
LINT_UNREF_MACRO(TM_TC_EC)
LINT_UNREF_MACRO(TM_TC_AO)
LINT_UNREF_MACRO(TM_TC_BREAK)
LINT_UNREF_MACRO(TM_TC_NOP)
LINT_UNREF_MACRO(TM_TC_ABORT)
LINT_UNREF_MACRO(TM_TC_SUSP)
LINT_UNREF_MACRO(TM_TC_EOF)
LINT_UNREF_MACRO(TM_TEL_NO_SOCKET)
LINT_UNREF_MACRO(tconNextQEntryPtr)
LINT_UNREF_MACRO(tconPrevQEntryPtr)
LINT_UNREF_MACRO(tconSga)
LINT_UNREF_MACRO(tconOutputEcho)
LINT_UNREF_MACRO(tconInputSga)
LINT_UNREF_MACRO(tconOutputSga)
#endif /* TM_LINT */

/*
 * Local macros
 */

/* Mininum TELNET idle timeout value */
#define TM_TELD_MIN_IDLE_TIMEOUT    TM_UL(300) /* 300 secs., or 5 minutes */

/* 
 * Time (in ms) before we retry the user's receive function if TM_EWOULDBLOCK
 * was returned.
 */
#define TM_TELD_INCOMING_TIMEOUT 10

/* tfTeldSend flag values */
/* #define TM_TELD_SEND_EOL     0x01 */ /* set in trsocket.h */
/* #define TM_TELD_SEND_COMMAND 0x02 */ /* set in trsocket.h */
/* #define TM_TELD_SEND_DONT_FLUSH 0x04 */ /* set in trsocket.h */

/*
 * tconExecFlags values (only set or checked in the context of the
 * Telnet server execute thread.)
 */
#define TM_TELEF_COMMAND_BUFFERED    0x1000
#define TM_TELEF_TIMEOUT             0x2000

/*
 * tconSendFlags values (protected with the tconSendLockEntry flag)
 */
#define TM_TELSF_SENDING             0x1000
#define TM_TELSF_BUFFER_ALLOCATED    0x2000
/*
 * Flag indicates the context of the tfTeldSend caller
 */
/* tfTeldSend called from the execute thread */
#define TM_TELD_EXECUTE_TASK        0
/* tfTeldSend called from user send thread */
#define TM_TELD_SEND_TASK           1

/* Telnet server listening port */
#define TM_TELD_PORT            tm_const_htons(23)

/* Run queue and list of Telnet connections queue index*/
#define TM_TELD_RUN_Q           0
#define TM_TELD_CONN_LIST       1

/* Size of an option reply */
#define TM_TELD_REPLY_SIZE      3 /* (sizeof(tconTelIacBuf[]) -1) */

/* Size of a command without option */
#define TM_TELD_COMMAND_SIZE    2 /* (sizeof(tconTelIacBuf[]) - 2) */

/* Queue not empty */
#define tm_teld_q_not_empty(i)                                   \
  (tm_context(tvTeldConQHead).servQHeadCount[i] != 0)

/* Queue empty */
#define tm_teld_q_empty(i)                                       \
  (tm_context(tvTeldConQHead).servQHeadCount[i] == 0)

/* Telnet connection output in binary mode */
#define tm_teld_out_binary(teldConnEntryPtr)  \
      ( (teldConnEntryPtr)->tconOutputBinary == TM_TC_WILL )


#define tm_teld_get_input_char(telChar, telInputPtr, telIndex, packedIndex)  \
        (telChar = telInputPtr[telIndex])

#define tm_teld_set_iac(replyPtr, iacChar)                                   \
        (replyPtr[0] = iacChar)
#define tm_teld_set_command(replyPtr, replyCommand)                          \
        (replyPtr[1] = replyCommand)
#define tm_teld_set_option(replyPtr, option)                                 \
        (replyPtr[2] = option)
#define tm_teld_get_command(replyPtr, replyCommand)                          \
        (replyCommand = replyPtr[1])




/* Timeout call back */
#define TM_CB_TIMEOUT          0x4000

 
/*
 * telnet input command line size, and telnet input raw buffer size
 */
#define TM_CMD_LINE_SIZE        512

/*
 * Buffer size for send buffer.
 */
#define TM_BUF_SEND_SIZE        2052

/* Extra bytes for option reply */
#define TM_BUF_EXTRA_BYTES      4

/*
 * Telnet NVT special characters
 */
#define TM_NVT_NULL   0  /* no operation */
#define TM_NVT_LF   '\n'/* next print line, same horizontal position (10) */
#define TM_NVT_CR   '\r'/* left margin of current line (13) */
#define TM_NVT_SPACE ' ' /* space */

/*
 * Telnet command codes.
 */
#define TM_TC_IAC    (tt8Bit)255 /* interpret as command escape */
#define TM_TC_DONT   (tt8Bit)254 /* request peer not to do option */
#define TM_TC_DO     (tt8Bit)253 /* request peer to do option */
#define TM_TC_WONT   (tt8Bit)252 /* indicates refusal to do option */
#define TM_TC_WILL   (tt8Bit)251 /* indicates request to do option */
#define TM_TC_SB     (tt8Bit)250 /* start of subnegotiation of the option */

#define TM_TC_GA     (tt8Bit)249 /* "Go Ahead" signal */
#define TM_TC_EL     (tt8Bit)248 /* erase the current line function */
#define TM_TC_EC     (tt8Bit)247 /* erase the current character function */
#define TM_TC_AYT    (tt8Bit)246 /* "Are You There" function */
#define TM_TC_AO     (tt8Bit)245 /* abort output function */
#define TM_TC_IP     (tt8Bit)244 /* Interrupt process */
#define TM_TC_BREAK  (tt8Bit)243 /* NVT character BRK */
#define TM_TC_DM     (tt8Bit)242 /* data mark-- Synch signal (with TCP URG) */
#define TM_TC_NOP    (tt8Bit)241 /* no operation */
#define TM_TC_SE     (tt8Bit)240 /* End of subnegotiation parameters */
#define TM_TC_EOR    (tt8Bit)239 /* end of record (transparent mode) */
#define TM_TC_ABORT  (tt8Bit)238 /* Abort process */
#define TM_TC_SUSP   (tt8Bit)237 /* Suspend process */
#define TM_TC_EOF    (tt8Bit)236 /* End of file */

/*
 * Telnet options we support
 */
#define TM_TO_BINARY (tt8Bit)0
#define TM_TO_ECHO   (tt8Bit)1
#define TM_TO_SGA    (tt8Bit)3

/* Maximum option value of options we support */
#define TM_TO_MAXOPTION TM_TO_SGA

/*
 * Telnet states
 */
 /* Processing regular characters */
#define TM_TEL_DATA_STATE       (tt8Bit)0x00
/* An IAC has been received */
#define TM_TEL_IAC_STATE        (tt8Bit)0x01
/* A WILL option has been received */
#define TM_TEL_WILL_STATE       (tt8Bit)0x02
/* A DO option has been received */
#define TM_TEL_DO_STATE         (tt8Bit)0x04
/* A WONT option has been received */
#define TM_TEL_WOPT_STATE       (tt8Bit)0x08
/* A DONT option has been received */
#define TM_TEL_DOPT_STATE       (tt8Bit)0x10
/* TCP Urgent mode, or sequence IAC, IP (interrupt process) received */
#define TM_TEL_URGENT_STATE     (tt8Bit)0x20
 /* IAC in urgent mode */
#define TM_TEL_URGENT_IAC_STATE (tt8Bit)(TM_TEL_IAC_STATE|TM_TEL_URGENT_STATE)
/* Carriage return as been received */
#define TM_TEL_CR_STATE         (tt8Bit)0x40
/* SB has been received */
#define TM_TEL_SB_STATE         (tt8Bit)0x80
/* IAC has been received in SB_STATE */
#define TM_TEL_SB_IAC_STATE     (tt8Bit)(TM_TEL_IAC_STATE|TM_TEL_SB_STATE)

/* Telnet socket index */
/* TELNET invalid socket index */
#define TM_TEL_NO_SOCKET  -1


/*
 * Local types
 */
/* Connection entry data type */
#ifdef TM_LINT
LINT_UNREF_MEMBER_BEGIN
#endif /* TM_LINT */
typedef struct tsTeldConnEntry
{
#define tconNextQEntryPtr(qIndex)     \
  tconQueueEntry.servQEntry[qIndex].qNextEntryPtr
#define tconPrevQEntryPtr(qIndex)     \
  tconQueueEntry.servQEntry[qIndex].qPrevEntryPtr
/* Run q, and open connection list */
    struct tsServerQueue           tconQueueEntry;
/* One timer to check on client time out */
    ttTimerPtr                     tconTimerPtr;
/* One timer to call tfTeldIncoming if TM_EWOULDBLOCK is ever returned */
    ttTimer                        tconTeldIncTimer;
    ttTimerPtr                     tconTeldIncTimerPtr;
/*
 * Telnet send buffer pointer, points to either tconTelSendBuf[], or
 * memory allocated buffer, if tconTelSendBuf[] is too small.
 */
    ttCharPtr                      tconSendBufPtr;
/* IP address of the telnet client */
    struct sockaddr_storage        tconClientAddress;
#ifdef TM_LOCK_NEEDED
/*
 * Lock use when sending data since send could be called either from the
 * user task, or the telnet server execute task.
 * tconSendLockEntry protects the following fields:
 * . tconSendBufPtr
 * . tconSendFlags
 * . tconSendBytes
 * . tconSendWordIndex;
 * . tconSendBitPosition;
 */
    ttLockEntry                    tconSendLockEntry;
#endif /* TM_LOCK_NEEDED */
/* Socket descriptor of the connection */
    int                            tconDesc;
/* Current index in telnet command line on socket */
    int                            tconTelCmdLineIndex;
/* Current read index in telnet input buffer on socket */
    int                            tconTelBufReadIndex;
/* Current processing index in telnet raw input buffer on socket */
    int                            tconTelBufProcessIndex;
/*
 * Number of bytes waiting to be flushed in the telnet network output
 * buffer.
 */
    int                            tconSendBytes;
/* CB flags for connection */
    tt16Bit                        tconCBFlags;
/* Connection flags in the context of the execute thread */
    tt16Bit                        tconExecFlags;
/* Connection send flags */
    tt16Bit                        tconSendFlags;
/* Telnet state on input data */
    tt8Bit                         tconTelState;
    tt8Bit                         tconOwnerCount;
/* Network telnet input command line buffer after NVT transformation */
    tt8Bit                         tconTelCmdLine[TM_CMD_LINE_SIZE];
/* Network telnet input buffer before NVT transformation */
    tt8Bit                         tconTelInputBuf[TM_CMD_LINE_SIZE];
/* Telnet send buffer pointer to send NVT data on the network */
    tt8Bit                         tconTelSendBuf[  TM_BUF_SEND_SIZE
                                                  + TM_BUF_EXTRA_BYTES];
/*
 * Buffer to send IAC commands (padded with an extra byte to make
 * it a multiple of 4 bytes
 */
    tt8Bit                         tconTelIacBuf[4];
/* Size of first dimension of our tconOption array */
#define TM_TOPTIONS  ((int)TM_TO_MAXOPTION + 1)
#define TM_TELD_OPTION_INPUT   0 /* second index in tconOption: input */
#define TM_TELD_OPTION_OUTPUT  1 /* second index in tconOption: output */
/* Current telnet state on the network input buffer (necessary for NVT) */
    tt8Bit                         tconOption[TM_TOPTIONS]
                                             [TM_TELD_OPTION_OUTPUT + 1];
/* First index in tconOption */
#define tconEcho            tconOption[TM_TO_ECHO]
#define tconBinary          tconOption[TM_TO_BINARY]
#define tconSga             tconOption[TM_TO_SGA]
#define tconInputBinary     tconBinary[TM_TELD_OPTION_INPUT]
#define tconOutputBinary    tconBinary[TM_TELD_OPTION_OUTPUT]
#define tconInputEcho       tconEcho[TM_TELD_OPTION_INPUT]
#define tconOutputEcho      tconEcho[TM_TELD_OPTION_OUTPUT]
#define tconInputSga        tconSga[TM_TELD_OPTION_INPUT]
#define tconOutputSga       tconSga[TM_TELD_OPTION_OUTPUT]
} ttTeldConnEntry;
#ifdef TM_LINT
LINT_UNREF_MEMBER_END
#endif /* TM_LINT */

typedef  ttTeldConnEntry TM_FAR * ttTeldConnEntryPtr;


/*
 * Call back functions.
 */

/* Signal the Telnet server process with call back events */
TM_NEARCALL void tfTeldSignal( int               socketDesc,
                   ttVoidPtr         teldConnEntryPtr,
                   int               CBFlags );

/* Signal the TELNET server process with client TELNET connection request */
void tfTeldSignalAccept(int listenDesc, int flags);



/*
 * Local functions
 */

/* Execute the main FTP server loop */
static void tfTeldExecute (void);

/* Close a socket, resetting CB flags */
static void tfTeldCloseSocket (int socketDesc);

/* Close the listening socket */
static void tfTeldCloseListen (void);

/* Accept a new connection request */
static void tfTeldAccept(void);

/* Receive data on the connection */
static int tfTeldRecv(ttTeldConnEntryPtr teldConnEntryPtr);

/*
 * Send data on connection. If TM_TELD_SEND_EOL flag is set, indicates whether
 * end of line has been reached, and means that we need to send CR, LF
 * in ASCII mode, or IAC, EOR in binary mode. If TM_TELD_SEND_COMMAND flag
 * is set, indicates that IAC should not be mapped.
 */
static int tfTeldSend( ttTeldConnEntryPtr  teldConnEntryPtr,
                       ttCharPtr           bufferPtr,
                       int                 bytes,
                       int                 flag,
                       int                 thread );

/* Flush send queue */
static int tfTeldFlush( ttTeldConnEntryPtr teldConnEntryPtr,
                        int                thread,
                        tt8Bit             needSendLock );

/*
 * Queue telnet control block to the run queue if there is any outstanding
 * input data that have not been given to the user yet.
 */
static void tfTeldQueueUnScannedCtrlInput (
                                        ttTeldConnEntryPtr teldConnEntryPtr );

/* Close a telnet connection, freeing the telnet connection block. */
static void tfTeldClose( ttTeldConnEntryPtr teldConnEntryPtr,
                         int                errorCode);

/*
 * Verify that the telnet handle given by the user is valid (in the open
 * connection list), and that the socket is still open. If so, then
 * increase telnet connection entry ownership.
 */
static int tfTeldCheckOutHandle(ttTeldConnEntryPtr teldConnEntryPtr);
/*
 * Decrease the telnet connection entry ownership. Free the telnet connection
 * entry, if ownership has reached 0.
 */
static void tfTeldCheckInHandle (ttTeldConnEntryPtr teldConnEntryPtr);

static void tfTeldTimerRemove(ttVoidPtr timerArgPtr);

void tfTeldIdleTimeout (ttVoidPtr      timerBlockPtr,
                        ttGenericUnion userParm1,
                        ttGenericUnion userParm2);

void tfTeldIncomingTimeout (ttVoidPtr      timerBlockPtr,
                        ttGenericUnion userParm1,
                        ttGenericUnion userParm2);

static void tfTeldTimerCleanup(ttVoidPtr vptr, tt8Bit flags);

/*
 * tfTeldUserStart function description:
 * Start the Telnet server. Should only be called once, or after
 * tfTeldUserStop.
 * 0.  Check on validity of parameters
 * 1.  Lock the telnet server execute (for re-entrancy)
 * 2.  Fail if server has already been started
 * 3.  Save passed parameters in local variables, and initialize
 *     local variables.
 * 4.  Open listening socket
 * 5.  Set listening socket to non blocking state
 * 6.  Bind to Telnet server port
 * 7.  Register for ACCEPT call back on listening socket.
 * 8.  Listen up to maximum number of connections allowed by the user
 * 9.  If all succeed and in blocking mode, create a Pend semaphore, and
 *     call the Telnet server main loop tfTeldExecute().
 * 10. If any error occurs, retrieve socket error if needed, and close
 *     the listening socket
 * 11. Unlock the Telnet server execute
 * 12. Return errorCode
 *
 * If blockingState is TM_BLOCKING_ON, then tfTeldUserStart will block
 * inside tfTeldExecute() main loop and return only when tfTeldUserStop
 * is called. If blockingState is TM_BLOCKING_OFF, the tfTeldUserStart
 * will return, and the user will need to call tfTeldUserExecute()
 * periodically to keep executing the Telnet server code.
 * Parameters
 * Parameter      Description
 * maxConnections Maximum number of incoming connections at any one time
 * maxBackLog     Number of outstanding incoming connections waiting to be
 *                accepted.
 * blockingState  Blocking or non blocking
 * Return value
 * Value                Meaning
 * 0                    success
 * TM_EALREADY          Telnetd already running
 * other socket error   failure
 */
int tfTeldUserStart ( int telnetOptionsAllowed,
                      int maxConnections,
                      int maxBackLog,
                      int blockingState )
{
    struct sockaddr_storage sockAddress;
    ttS32Bit            availableConnections;
    int                 addressLength;
    int                 errorCode;
    int                 tempInt;
    int                 listenDesc;
    ttSockAddrPtrUnion  tempSockAddr;

    tempSockAddr.sockNgPtr = &sockAddress;
    errorCode = TM_ENOERROR;
/* Number of available sockets in the system - 1 for the listening socket */
    availableConnections = (ttS32Bit)(tm_context(tvMaxNumberSockets)
                            - tm_context(tvNumberOpenSockets) - 1);
    if (    ( maxBackLog <= 0 )
         || ( maxBackLog > (int)availableConnections )
         || (    ( maxConnections != 0 )
              && (    ( maxConnections < maxBackLog )
                   || ( maxConnections > (int)availableConnections ) ) )
         || (    (blockingState != TM_BLOCKING_ON)
              && (blockingState != TM_BLOCKING_OFF) )
         || (     (blockingState == TM_BLOCKING_ON)
              &&  (tvPendPostEnabled == TM_8BIT_ZERO) ) )
    {
        errorCode = TM_EINVAL;
    }
    else
    {
        tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
/* Check if not already started */
        if (    (tm_context(tvTeldExecuting) != 0)
             || (tm_context(tvTeldListenDesc) != TM_SOCKET_ERROR) )
        {
            errorCode = TM_EALREADY;
        }
        else
        {
            tm_context(tvTeldIdleTimeout) = 
                TM_TELD_MIN_IDLE_TIMEOUT * TM_UL(1000);
            if (tm_context(tvTeldIdleTimeout) < tm_context(tvTcpMaxRtoTime))
            {
                tm_context(tvTeldIdleTimeout) = tm_context(tvTcpMaxRtoTime);
            }
            tm_context(tvTeldShutdown) = 0;
            tm_context(tvTeldOptionsAllowed) = telnetOptionsAllowed;
            tm_context(tvTeldMaxConnections) = maxConnections;
/* Blocking state */
            tm_context(tvTeldBlockingState) = blockingState;
/* Initialize pending sempahore */
#ifdef TM_PEND_POST_NEEDED
            tm_bzero(&tm_context(tvTeldPendSem), sizeof(ttPendEntry));
#endif /* TM_PEND_POST_NEEDED */
/* Number of pending connections. */
            tm_context(tvTeldPendingConnections) = 0;
/* Open a listening socket - If IPv6 is enabled, open an AF_INET6 socket,
 * which can accept both IPv4 and IPv4 connections.  Otherwise, just open a
 * regular AF_INET socket.
 */ 
            listenDesc = socket(AF_INET, SOCK_STREAM, IP_PROTOTCP);
            if (listenDesc == TM_SOCKET_ERROR)
            {
                errorCode = TM_SOCKET_ERROR;
            }
            else
            {
                tm_context(tvTeldListenDesc) = listenDesc;
/* Non blocking state */
                (void)tfBlockingState(listenDesc, TM_BLOCKING_OFF);
                addressLength = sizeof(struct sockaddr_storage);
                errorCode = getsockname(
                                     listenDesc,
                                     tempSockAddr.sockPtr,
                                     &addressLength);
                if (errorCode == TM_ENOERROR)
                {
/*
 * Need to set SO_REUSEADDR in case there is still an accept socket from a
 * previous server session that is still in time wait state.
 */
                    tempInt = 1;
                    errorCode = setsockopt( listenDesc,
                                            SOL_SOCKET,
                                            SO_REUSEADDR,
                                            (const char TM_FAR *)&tempInt,
                                            sizeof(int) );
                    if (errorCode == TM_ENOERROR)
                    {
/* Telnet server port */
                        sockAddress.ss_port = TM_TELD_PORT;
/* Bind to server port */
                        errorCode = bind(
                             listenDesc,
                             (const struct sockaddr TM_FAR *)
                                 tempSockAddr.sockPtr,
                             sizeof(struct sockaddr_storage));
                        if (errorCode == TM_ENOERROR)
                        {
/* Note: order of calls does not matter, since we own the execute lock */
/* Register CB_ACCEPT */
                            errorCode = tfRegisterSocketCB(
                                                       listenDesc,
                                                       tfTeldSignalAccept,
                                                       TM_CB_ACCEPT);
                            if (errorCode == TM_ENOERROR)
                            {
/* Listen for incoming connections */
                                errorCode = listen(listenDesc, maxBackLog);
                                if (errorCode == TM_ENOERROR)
                                {
#ifdef TM_PEND_POST_NEEDED
                                    if (blockingState != TM_BLOCKING_OFF)
                                    {
                                        tfTeldExecute();
                                    }
#endif /* TM_PEND_POST_NEEDED */
                                }
                            }
                        }
                    }
                }
            }

            if (errorCode != TM_ENOERROR)
            {
                errorCode = tfGetSocketError(listenDesc);
/* We unlock, and re-lock in this routine */
                if (listenDesc != TM_SOCKET_ERROR)
                {
                    tfTeldCloseListen();
                }
            }
        }
        tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
    }
    return errorCode;
}


/*
 * tfTeldUserIdleTimeout function description:
 * This function allows the user to set a Telnet connection's idle
 * timeout to a value larger than the default (TM_TELD_MIN_IDLE_TIMEOUT).
 *
 * 1. Default errorCode return to TM_EINVAL (assume failure).
 * 1. Check that user timeout is bigger than minimum (5 minutes)
 * 2. If bigger, set tvTeldIdleTimeout to new value, and set errorCode
 *    to no error.
 * 3. Return error code.
 *
 * Parameter:
 *     timeout      The desired idle timeout, in seconds.
 *
 * Return value:
 *     Value        Meaning
 *     0            Success
 *     TM_EINVAL    The desired idle timeout is less than the default
 *                  or smaller than the TCP maximum retransmission time.
 *
 * Notes:
 *     The new timeout value will need to be converted from seconds to
 *     milliseconds and stored in a 32-bit variable. This means that
 *     the maxium allowed value, in seconds, is 2^32 / 1000, or
 *     4,294,967 or about 49.7 days.
 */
int tfTeldUserIdleTimeout(ttUser32Bit timeout)
{
    int errorCode;

    errorCode = TM_EINVAL;
    if (    (timeout >= TM_TELD_MIN_IDLE_TIMEOUT)
         && (timeout >= (tm_context(tvTcpMaxRtoTime) / 1000))
         && (timeout <= (TM_32BIT_MAX / 1000)) )
    {
        errorCode = TM_ENOERROR;
        tm_context(tvTeldIdleTimeout) = timeout * TM_UL(1000);
    }

    return errorCode;
}


/*
 * tfTeldUserStop function description:
 * Stop execution of the Telnet server.
 * 1. Lock the Telnet server execute.
 * 2. If server not running, set error code to TM_EALREADY.
 * 3. Else If server running:
 * 3.1  set the tvTeldShutdown flag
 * 3.2  . if in non blocking mode, call tfTeldExecute() to close all active
 *        connections if not already in the execute loop
 *      . else (blocking mode), post on the server pending
 *        semaphore to let the execute loop close all open connections.
 * 4. Unlock.
 * 5. Return error code.
 * Parameter:
 * none
 * Return value:
 * Value        Meaning
 * 0            success
 * TM_EALREADY  Telnet server was already stopped.
 */
int tfTeldUserStop (void)
{
    int errorCode;

    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
    if (tm_context(tvTeldListenDesc) == TM_SOCKET_ERROR)
    {
        errorCode = TM_EALREADY;
    }
    else
    {
        tm_context(tvTeldShutdown) = 1;
        if (tm_context(tvTeldBlockingState) == TM_BLOCKING_OFF)
        {
/* Non blocking mode */
            if (tm_context(tvTeldExecuting) == 0)
/* Call execute loop if not already in there */
            {
/* clean up */
                tfTeldExecute();
            }
        }
#ifdef TM_PEND_POST_NEEDED
        else
        {
/*
 * Post (for the shutdown event).
 */
            tm_post( &tm_context(tvTeldPendSem),
                           &tm_context(tvTeldExecuteLockEntry),
                           TM_ENOERROR );
        }
#endif /* TM_PEND_POST_NEEDED */
        errorCode = TM_ENOERROR;
    }
    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
    return errorCode;
}



/*
 * tfTeldUserExecute function description:
 * Execute the Telnet server main loop. Call valid only if tfTeldUserStart
 * had been called in non blocking mode, and nobody is currently executing
 * the Telnet server code main loop.
 * Parameters
 * none
 * Return value
 * Value                Meaning
 * 0                    success
 * TM_EPERM             Telnet server executing or not started or stopped
 * error                failure
 *
 */
int tfTeldUserExecute (void)
{
    int errorCode;

    tm_lock_wait(&tm_context(tvTeldExecuteLockEntry));
    if (    (tm_context(tvTeldListenDesc) != TM_SOCKET_ERROR)
         && (tm_context(tvTeldExecuting) == 0) )
    {
/*
 * tfTeldUserStart has been called, and nobody is executing the server main
 * loop
 */
        tfTeldExecute();
        errorCode = TM_ENOERROR;
    }
    else
    {
/* tfTeldUserStart has not been called, or tfTeldExecute is running */
        errorCode = TM_EPERM;
    }
    tm_unlock(&tm_context(tvTeldExecuteLockEntry));
    return errorCode;
}

/*
 * tfTeldUserClose function description
 * User wants to close the specified telnet connection
 * Parameters
 * Parameter      Description
 * teldHandle     Telnet handle (pointer to a telnet connection entry)
 * Return value
 * Value          Meaning
 * 0              Success
 * TM_EINVAL      Invalid telnet handle
 */
int tfTeldUserClose( ttUserTeldHandle teldHandle )
{
    int errorCode;

/* Check that the handle is valid */
    errorCode = tfTeldCheckOutHandle((ttTeldConnEntryPtr)teldHandle);
    if (errorCode == TM_ENOERROR)
    {
/* Tell the execute thread to close the socket */
        tfTeldSignal( ((ttTeldConnEntryPtr)teldHandle)->tconDesc,
                      (ttTeldConnEntryPtr)teldHandle,
                      TM_CB_RESET );
        tfTeldCheckInHandle((ttTeldConnEntryPtr)teldHandle);
    }
    return errorCode;
}

/*
 * tfTeldUserSend function description
 * Send some data on the telnet connection
 * Parameters
 * Parameter      Description
 * teldHandle     Telnet handle (pointer to a telnet connection entry)
 * telSendBufPtr  Pointer to user data to send
 * teldBytes      Number of bytes to send
 * flag           If TM_TELD_SEND_EOL is set, indicates that the server needs
 *                to send CR,LF in ascii mode, or IAC, EOR in binary mode
 *                if TM_TELD_SEND_COMMAND, indicates that IAC should not be
 *                mapped.
 * Return value
 * Value          Meaning
 * 0              Success
 * TM_EWOULDBLOCK Data not sent. Not enough room for the data.
 * TM_EINVAL      Invalid telnet handle
 */
int tfTeldUserSend( ttUserTeldHandle teldHandle,
                    char TM_FAR *    teldSendBufPtr,
                    int              teldBytes,
                    int              flag )
{
    int errorCode;
    int allowedFlag;

    errorCode = TM_EINVAL;
    allowedFlag = flag & (  TM_TELD_SEND_EOL
                          | TM_TELD_SEND_COMMAND
                          | TM_TELD_SEND_DONT_FLUSH);
    if (    ( (teldSendBufPtr != (char TM_FAR *)0) && (teldBytes > 0) )
        || ( (teldBytes == 0) && (allowedFlag & TM_TELD_SEND_EOL) ) )
    {
/* call tfTeldCheckOutHandle */
        errorCode = tfTeldCheckOutHandle((ttTeldConnEntryPtr)teldHandle);
        if (errorCode == TM_ENOERROR)
        {
            errorCode = tfTeldSend( (ttTeldConnEntryPtr)teldHandle,
                                   teldSendBufPtr,
                                   teldBytes,
                                   allowedFlag,
                                   TM_TELD_SEND_TASK );
            tfTeldCheckInHandle((ttTeldConnEntryPtr)teldHandle);
        }
    }
    return errorCode;
}

/*
 * tfTeldCheckOutHandle function description
 * Check that the user provided telnet handle, teldConnEntryPtr is a valid
 * telnet connection entry pointer, and that the corresponding socket
 * is still opened. If so, Increase telnet entry owner count.
 * Parameters
 * Parameter          Description
 * teldConnEntryPtr   Telnet handle (pointer to a telnet connection entry)
 * Return value
 * TM_EINVAL          no such telnet handle
 * TM_ESHUTDOWN       Telnet connection closed
 */
static int tfTeldCheckOutHandle (ttTeldConnEntryPtr teldConnEntryPtr)
{
    ttServerQueuePtr  servQueuePtr;
    int               errorCode;

    errorCode = TM_EINVAL;
    tm_lock_wait(&tm_context(tvTeldExecuteLockEntry));
/* Loop through TM_TELD_CONN_LIST, and check for a match */
    servQueuePtr = (ttServerQueuePtr)
                   tm_context(tvTeldConQHead).
                                    servHeadNextQEntryPtr(TM_TELD_CONN_LIST);
    while (servQueuePtr != (ttServerQueuePtr)(ttVoidPtr)
                                            (&(tm_context(tvTeldConQHead))))
    {
        if (teldConnEntryPtr == (ttTeldConnEntryPtr)servQueuePtr)
        {
            errorCode = TM_ENOERROR;
            break;
        }
        servQueuePtr = (ttServerQueuePtr)
                       servQueuePtr->servNextQEntryPtr(TM_TELD_CONN_LIST);
    }
    if (errorCode == TM_ENOERROR)
    {
        if (teldConnEntryPtr->tconDesc == TM_SOCKET_ERROR)
        {
            errorCode = TM_ESHUTDOWN;
        }
        else
        {
            teldConnEntryPtr->tconOwnerCount++;
        }
    }
    tm_unlock(&tm_context(tvTeldExecuteLockEntry));
    return errorCode;
}

/*
 * tfTeldCheckInHandle function description
 * Called when returning from a user call where the user passed a
 * pointer to an active telnet connection entry.
 * . Decrease telnet connection ownership.
 * . Check telnet connection ownership.
 * . If zero, free the telnet connection entry.
 * Parameters
 * Parameter          Description
 * teldConnEntryPtr   Telnet handle (pointer to a telnet connection entry)
 * Return value
 * No return value
 */
static void tfTeldCheckInHandle (ttTeldConnEntryPtr teldConnEntryPtr)
{
    tt8Bit ownerCount;

    tm_lock_wait(&tm_context(tvTeldExecuteLockEntry));
    teldConnEntryPtr->tconOwnerCount--;
    ownerCount = teldConnEntryPtr->tconOwnerCount;
    tm_unlock(&tm_context(tvTeldExecuteLockEntry));
    if (ownerCount == 0)
    {
        tm_free_raw_buffer((ttRawBufferPtr)teldConnEntryPtr);
    }
}

/*
 * tfTeldSignalAccept function description:
 * Call back function when an incoming Telnet control connection arrives.
 * 1 Obtain exclusive access to the Telnet server.
 * 2 If the Telnet server is not in the process of shutting down, increase
 *   the number of pending connections, and post to the Telnet server
 * 3 Release access to the Telnet server.
 * Parameters:
 * Parameter    Description
 * listenDesc   Telnet server listening socket
 * flags        TM_CB_ACCEPT (only)
 * Return value:
 * no return
 */
void tfTeldSignalAccept (int listenDesc, int flags)
{
    TM_UNREF_IN_ARG(flags);
    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
    if (    (tm_context(tvTeldShutdown) == 0)
         && (tm_context(tvTeldListenDesc) == listenDesc) )
    {
        tm_context(tvTeldPendingConnections)++;
#ifdef TM_PEND_POST_NEEDED
        if (tm_context(tvTeldBlockingState) != TM_BLOCKING_OFF)
        {
/* Post */
            tm_post( &tm_context(tvTeldPendSem),
                           &tm_context(tvTeldExecuteLockEntry),
                           TM_ENOERROR );
        }
#endif /* TM_PEND_POST_NEEDED */
    }
    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
}


/*
 * tfTeldSignal function description:
 * Call back function when one or more registered events occur on the
 * control connection.
 * 1. Obtain exclusive access to the Telnet server.
 * 2. If Telnet server is not in the process of shutting down, and if
 *    the socket descriptor match the one stored in the telnet connection
 *    entry and if CB flags non zero, add the CB flags to the current CB
 *    flags of the connection. If Telnet connection not already in the run Q,
 *    append to the run q, and post on the Telnet server.
 * 3. Release exclusive access to the Telnet server.
 * Parameters:
 * Parameter    Description
 * socketDesc   socket CB occurs on
 * teldPtr      pointer to the corresponding TELNETD connection entry.
 * CBflags      call back flags on.
 * Return:
 * no return value.
 */
void tfTeldSignal ( int               socketDesc,
                    ttVoidPtr         teldPtr,
                    int               CBFlags )
{
#define teldConnEntryPtr ((ttTeldConnEntryPtr)teldPtr)

    if (teldConnEntryPtr != (ttTeldConnEntryPtr)0)
    {
        tm_lock_wait(&tm_context(tvTeldExecuteLockEntry));
        if (tm_context(tvTeldShutdown) == 0)
/* Not in the process of shutting down */
        {
            if (teldConnEntryPtr->tconDesc == socketDesc)
            {
                if (CBFlags == TM_CB_TIMEOUT)
                {
/*
 * Note that the following is coded here as opposed to in the time out
 * routine, since we have the execute lock on here, and not in the timeout
 * routine.
 */
/* If the execute loop had just removed the timer, ignore the timeout */
                    if (teldConnEntryPtr->tconTimerPtr == TM_TMR_NULL_PTR)
                    {
                        CBFlags = 0;
                    }
                    else
                    {
/*
 * The timer is non auto, so it is automatically removed by the timer thread,
 * upon return from this routine. Just indicate that it is gone.
 */
                        teldConnEntryPtr->tconTimerPtr = TM_TMR_NULL_PTR;
/*
 * If there is any activity on the connection
 * reset the CB_TIMEOUT CBFLags to prevent timeout, no need to signal
 * or to add the timeout timer, since if any of these flags are non
 * zero, we are guaranteed to be in the run Queue, and the timer
 * is automatically added in the execute loop.
 */
                        if (teldConnEntryPtr->tconCBFlags != 0)
                        {
                            CBFlags = 0;
                        }
                    }
                }
                if (CBFlags != 0)
                {
                    teldConnEntryPtr->tconCBFlags |= (tt16Bit)CBFlags;
/* tfServerAppendTo Queue checks whether the entry is already in the queue */
                    tfServerAppendToQueue( &tm_context(tvTeldConQHead),
                                           (ttServerQueuePtr)teldConnEntryPtr,
                                           TM_TELD_RUN_Q );
#ifdef TM_PEND_POST_NEEDED
                    if (tm_context(tvTeldBlockingState) != TM_BLOCKING_OFF)
                    {
/* Post */
                        tm_post( &tm_context(tvTeldPendSem),
                                       &tm_context(tvTeldExecuteLockEntry),
                                       TM_ENOERROR );
                    }
#endif /* TM_PEND_POST_NEEDED */
                }
            }
        }
        tm_unlock(&tm_context(tvTeldExecuteLockEntry));
    }
    return;
#undef teldConnEntryPtr
}

/*
 * tfTeldExecute function description:
 * This is the main Telnet server loop. Called with the execute lock on from
 * either tfTeldUserStart(), tfTeldUserExecute() (in non blocking mode), or
 * from tfTeldUserStop().
 * A. First, set the tvTeldExecuting variable and start executing a do while
 *    loop until tvTeldExecuting is reset
 * B. While there is an entry in the run Q, or the user wants to shutdown the
 *    server, or there is any pending connection:
 * B.1. We first check whether the user wants to stop the Telnet server. In
 *    that case, we close the listenining socket, reset the pending
 *    connection count, and close each open connection (we unlock and
 *    relock the execute lock to close each open connection).
 * B.2. While there is any pending connection, try and accept each one of
 *    them (we unlock and relock the execute lock to accept).
 * B.3. If there is an entry in the RUN Q:
 *   B.3.a Save the socket call back flags, reset them in the connection
 *         entry, and remove the entry from the run Q.
 *   B.3.b Unlock the execute lock.
 *   B.3.c Examine the CB flags on the socket
 *       . If CB send complete flag is on, Try and send data on the
 *         connection.
 *       . If a recv or oob recv occurred, call the telnet
 *         receive function. If any error occurred on the connection
 *         in the receive function, close the connection
 *         with that error and go back to the beginning of the loop (B.3.e).
 *       . If a remote close or reset occurred, close the connection
 *         with a reset error, and go back to the beginning of the loop
 *         (B.3.e).
 *   B.3.d If any error occurred on the connection while processing
 *        the above call back flags,
 *        close the connection with that error and go back to
 *        the beginning of the loop.
 *   B.3.e Relock the execute lock.
 * C. No more work to do. If the the user has stopped the Telnet server, reset
 *    the execute variable, tvTeldExecuting, to exit this function.
 * D. Otherwise If the user called tfTeldUserStart in non blocking mode,
 *    also reset the execute variable, tvTeldExecuting to exit this function.
 * E. Otherwise (tvTeldExecuting still on) pend to wait for the call back
 *    functions to post to our TELNETD pend semaphore.
 * F. Exit the do while loop (A), if tvTeldExecuting is reset.
 *
 * Parameters
 * none
 * Return value
 * 0                    success
 * TM_EPERM             Telnet server running in blocking mode or not started
 * error                failure
 *
 */
static void tfTeldExecute (void)
{
    ttTeldConnEntryPtr teldConnEntryPtr;
    tt32Bit            timeout;
    int                CBFlags;
    int                errorCode;
    ttGenericUnion     timerParm1;

    tm_context(tvTeldExecuting) = 1;
    do
    {
/* Check if there is any TELNETD work queued in */
        while (    ( tm_teld_q_not_empty(TM_TELD_RUN_Q) )
                || ( tm_context(tvTeldPendingConnections) != 0 )
                || ( tm_context(tvTeldShutdown) != TM_8BIT_ZERO  ) )
        {
            if (tm_context(tvTeldShutdown) != TM_8BIT_ZERO)
            {
/* Close the listening socket */
                tfTeldCloseListen();
                tm_context(tvTeldPendingConnections) = 0;
                teldConnEntryPtr = (ttTeldConnEntryPtr)
                                (tm_context(tvTeldConQHead).
                                  servHeadNextQEntryPtr(TM_TELD_CONN_LIST));
                while (    teldConnEntryPtr
                        != (ttTeldConnEntryPtr)&tm_context(tvTeldConQHead) )
                {
                    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
                    tfTeldClose(teldConnEntryPtr, TM_ECONNABORTED);
                    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
                    teldConnEntryPtr = (ttTeldConnEntryPtr)
                                (tm_context(tvTeldConQHead).
                                    servHeadNextQEntryPtr(TM_TELD_CONN_LIST));
                }
                tm_context(tvTeldShutdown) = 0;
            }
            while (tm_context(tvTeldPendingConnections) != 0)
            {
                tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
                tfTeldAccept();
                tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
                tm_context(tvTeldPendingConnections)--;
            }
            errorCode = TM_ENOERROR;
            if (tm_teld_q_not_empty(TM_TELD_RUN_Q))
            {
                teldConnEntryPtr = (ttTeldConnEntryPtr)
                                      (tm_context(tvTeldConQHead).
                                          servHeadNextQEntryPtr(TM_TELD_RUN_Q));
                CBFlags = teldConnEntryPtr->tconCBFlags;
                teldConnEntryPtr->tconCBFlags = 0;
                tfServerRemoveFromQueue( &tm_context(tvTeldConQHead),
                                         (ttServerQueuePtr)teldConnEntryPtr,
                                         TM_TELD_RUN_Q );
/*
 * remove the idle timeout timer (with execute lock on to avoid race
 * conditions with the timeout function)
 */
                tfTeldTimerRemove(teldConnEntryPtr);
                if ( CBFlags == TM_CB_TIMEOUT )
                {
                    if (teldConnEntryPtr->tconExecFlags & TM_TELEF_TIMEOUT)
                    {
/*
 * Close the connection since the client did not ACK or respond to our
 * AYT message
 */
                        errorCode = TM_ETIMEDOUT;
                        CBFlags = 0;
                    }
                    else
                    {
/* Connection has been idle, send an AYT message */
                        teldConnEntryPtr->tconExecFlags |= TM_TELEF_TIMEOUT;
                    }
                }
                else
                {
/* Reset timeout condition */
                    teldConnEntryPtr->tconExecFlags &= ~TM_TELEF_TIMEOUT;
                }
                tm_unlock(&tm_context(tvTeldExecuteLockEntry));
                if ( CBFlags == TM_CB_TIMEOUT )
                {
/* Send an AYT message */
                   tm_teld_set_command(teldConnEntryPtr->tconTelIacBuf,
                                       TM_TC_AYT);
                   (void)tfTeldSend(
                               teldConnEntryPtr,
                               (ttCharPtr)&teldConnEntryPtr->tconTelIacBuf[0],
                               TM_TELD_COMMAND_SIZE,
                               TM_TELD_SEND_COMMAND,
                               TM_TELD_EXECUTE_TASK );
                }
                if ( CBFlags & (TM_CB_SEND_COMPLT) )
                {
                    errorCode = tfTeldFlush( teldConnEntryPtr,
                                             TM_TELD_EXECUTE_TASK,
                                             TM_8BIT_YES );
                    if (errorCode != TM_ENOERROR)
                    {
/* Process next control block */
                        goto nextControlBlock;
                    }
                }
                if ( CBFlags & (TM_CB_RECV_OOB|TM_CB_RECV) )
                {
                    errorCode = tfTeldRecv(teldConnEntryPtr);
                    if (errorCode != TM_ENOERROR)
                    {
/* Process next control block */
                        goto nextControlBlock;
                    }
                }
                if (   CBFlags
                     & (   TM_CB_REMOTE_CLOSE
                         | TM_CB_RESET
                         | TM_CB_SOCKET_ERROR ) )
                {
/* Connection reset or closed by peer */
                    errorCode = TM_ECONNRESET;
/* Process next control block */
                    goto nextControlBlock;
                }
nextControlBlock:
                if (errorCode != TM_ENOERROR)
                {
                    tfTeldClose(teldConnEntryPtr, errorCode);
                }
                else
                {
/*
 * Add idle timeout timer
 */
                    if (CBFlags == TM_CB_TIMEOUT)
                    {
/* If we timed out, set timeout to AYT response */
                        timeout = tm_context(tvTcpMaxRtoTime);
                    }
                    else
                    {
/* Set default timeout interval */
                        timeout =   tm_context(tvTeldIdleTimeout)
                                  - tm_context(tvTcpMaxRtoTime);
                    }
                    timerParm1.genVoidParmPtr = (ttVoidPtr) teldConnEntryPtr;
                    teldConnEntryPtr->tconTimerPtr =
                                      tfTimerAdd( tfTeldIdleTimeout,
                                                  timerParm1,
                                                  timerParm1, /* unused */
                                                  timeout,
                                                  0 );
                }
                tm_lock_wait(&tm_context(tvTeldExecuteLockEntry));
            }
        }
/* No more work to do. Check whether the user has stopped the Telnet server */
        if (tm_context(tvTeldListenDesc) == TM_SOCKET_ERROR)
        {
/* User stopped the Telnet server. */
            tm_context(tvTeldExecuting) = 0;
        }
        else
/* No more work to do, and Telnet server still up, pend if allowed */
        {
#ifdef TM_PEND_POST_NEEDED
            if (tm_context(tvTeldBlockingState) != TM_BLOCKING_OFF)
            {
                (void)tm_pend( &tm_context(tvTeldPendSem),
                               &tm_context(tvTeldExecuteLockEntry) );
            }
            else
#endif /* TM_PEND_POST_NEEDED */
            {
/* otherwise, exit the loop */
                tm_context(tvTeldExecuting) = 0;
            }
        }
    }
    while ( tm_context(tvTeldExecuting) != 0 );
}

/*
 * tfTeldClose function description:
 * Close a Telnet control connection. Only called from the execute loop.
 * removing the Telnet connection from any queue, 
 * and then free the TELNETD connection entry.
 *
 * Save socket descriptor from the telnet connection entry.
 * If socket is open:
 *  1. Lock the TELNETD execute lock.
 *  2. Lock the TELNETD connection send lock (so that a user send will
 *     not get an invalid socket descriptor, since we are going to reset
 *     tconDesc here in this routine).
 *  3. Reset execute flags.
 *  4. Reset the socket field in the connection entry.
 *  5. Remove the TELNETD connection entry from the run queue, and the open
 *     connections list.
 *  6. Decrease entry owner count (entry no longer in the open connections
 *     list).
 *  7. Decrease the number of open connections.
 *  8. Unlock the TELNETD connection send lock.
 *  9. Unlock the TELNETD execute lock.
 * 10. Close the telnet entry socket.
 * 11. Tell the user about the closed socket.
 * 12. If owner count had reached 0, free the TELNETD connection entry.
 *
 * Parameters:
 * Parameter            Description
 * teldConnEntryPtr     Pointer to TELNETD connection entry
 * errorCode            errorCode that happens on the control connection.
 * Return code:
 * no return.
 */
static void tfTeldClose (ttTeldConnEntryPtr teldConnEntryPtr,
                         int                errorCode)
{
    int               socketDesc;
    tt8Bit            ownerCount;

/*
 * We know this field is always valid, because tfTeldClose() is
 * called only from the execute loop.
 */
    socketDesc = teldConnEntryPtr->tconDesc;
    TM_UNREF_IN_ARG(errorCode);
    if (socketDesc != TM_SOCKET_ERROR)
    {
        tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
        tm_call_lock_wait(&teldConnEntryPtr->tconSendLockEntry);
        teldConnEntryPtr->tconDesc = TM_SOCKET_ERROR;
        teldConnEntryPtr->tconExecFlags = TM_16BIT_ZERO;
/*
 * Remove the timer with timer lock on to avoid race conditions with the
 * timeout function.
 */
        tfTeldTimerRemove(teldConnEntryPtr);
/* Remove entry from runq, and open connections list */
        tfServerRemoveFromQueue( &tm_context(tvTeldConQHead),
                                 (ttServerQueuePtr)teldConnEntryPtr,
                                 TM_TELD_RUN_Q );
        tfServerRemoveFromQueue( &tm_context(tvTeldConQHead),
                                 (ttServerQueuePtr)teldConnEntryPtr,
                                  TM_TELD_CONN_LIST );
/* Connection list ownersip */
        teldConnEntryPtr->tconOwnerCount--;
/* Save owner count (while lock is on) */
        ownerCount = teldConnEntryPtr->tconOwnerCount;
        tm_call_unlock(&teldConnEntryPtr->tconSendLockEntry);
        tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
/*
 * Close the socket to get all pending send tasks to wake up and have a
 * chance to exit.
 */
        tfTeldCloseSocket(socketDesc);
/* Let the user know about the closed socket */
        tfTeldClosed((ttUserTeldHandle)teldConnEntryPtr);
/* Free the entry, if last one to own the telnet connection entry */
        if (ownerCount == 0)
        {
            tm_free_raw_buffer((ttRawBufferPtr)teldConnEntryPtr);
        }
    }
}

/*
 * tfTeldCloseListen function description:
 * Function called with execute lock on.
 * 1. Unlock the TELNETD execute lock to prevent deadlock with call back
 *    function.
 * 2. CLose the listening socket. This will cause all pending connections
 *    to be reset.
 * 3. Relock the TELNETD execute lock.
 * Parameters:
 * none
 * Return:
 * none
 */
static void tfTeldCloseListen (void)
{
     int    socketDesc;

     socketDesc = tm_context(tvTeldListenDesc);
     tm_context(tvTeldListenDesc) = TM_SOCKET_ERROR;
     tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
     tfTeldCloseSocket(socketDesc);
     tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
}

/*
 * tfTeldCloseSocket function description:
 *
 * 1. Reset all call back flags.
 * 2. Close the socket.
 *
 * Parameters:
 * Parameter    Description
 * socketDesc   socket descriptor to close.
 *
 * Return:
 * none
 */
static void tfTeldCloseSocket (int socketDesc)
{
    (void)tfRegisterSocketCB(socketDesc, (ttUserSocketCBFuncPtr)0, 0);
    (void)tfClose(socketDesc);
}

/*
 * tfTeldAccept function description:
 * Accept a control connection
 * 1. Call accept (in non blocking mode) on the listening socket, to accept
 *    an incoming Telnet control connection.
 * 2. If we got a valid socket descriptor:
 * 2.1 Make the socket non blocking.
 * 2.2 Check that we did not exceed the maximum number of connections allowed.
 *     If we did not:
 * 2.2.1 Allocate a TELNETD connection entry and initialize it.
 * 2.2.2 Increase number of open connections.
 * 2.2.3 Append to open connection entry list
 * 2.2.4 Store socket descriptor in control socket field of TELNETD connection
 *       entry.
 * 2.2.6 Disable Nagle algorithm (so that the TCP send queue gets flushed
 *       to the network if there is no data in the TCP send queue waiting
 *       to be acked.
 * 2.2.7 Handle urgent data in line.
 * 2.2.8 Initialize Telnet transfer parameters with default values.
 * 2.2.9 Register for socket call back for RECV, RECV OOB, REMOTE CLOSE,
 *       SEND COMPLETE and RESET.
 * 2.2.10 If the register succeeds, copy the telnet client IP address, and
 *        notify the user of the accepted telnet connection.
 * 2.2.11 If we could not allocate a TELNETD connection entry set
 *        errorCode so that we close the socket at the bottom of the routine.
 * 2.3 If we exceeded the number of available connections, set errorCode so
 *        that we close the socket at the bottom of the routine.
 * 2.4 If an error occurred earlier close the control connection: if we
 *     had allocated a TELNETD connection entry successfully, use
 *     tfTeldClose with error code. If we had not allocated a TELNETD
 *     connection, use tfTeldCloseSocket().
 *
 * Parameters:
 * none
 *
 * Return:
 * none
 */
static void tfTeldAccept (void)
{
    struct sockaddr_storage TM_FAR * sockAddressPtr;    
    ttTeldConnEntryPtr           teldConnEntryPtr;
    struct sockaddr_storage      sockAddress;
    int                          acceptDesc;
    int                          tempInt;
    int                          errorCode;
    ttSockAddrPtrUnion           tempSockAddr;

    tempSockAddr.sockNgPtr = &sockAddress;
    errorCode = TM_SOCKET_ERROR;
    teldConnEntryPtr = (ttTeldConnEntryPtr)0;
    tempInt = sizeof(struct sockaddr_storage);
    acceptDesc = accept( tm_context(tvTeldListenDesc),
                         tempSockAddr.sockPtr,
                         &tempInt);
    if (acceptDesc != TM_SOCKET_ERROR)
    {
/* Connection socket is in non blocking state */
        (void)tfBlockingState(acceptDesc, TM_BLOCKING_OFF);
        if (    tm_context(tvTeldMaxConnections) == 0
            ||  (   tm_context(tvTeldConQHead).
                                servQHeadCount[TM_TELD_CONN_LIST]
                  < tm_context(tvTeldMaxConnections) ) )
        {
            teldConnEntryPtr = (ttTeldConnEntryPtr)
                           tm_get_raw_buffer((ttPktLen)sizeof(ttTeldConnEntry));
            if ( teldConnEntryPtr != (ttTeldConnEntryPtr)0 )
            {
                tm_bzero(teldConnEntryPtr, sizeof(ttTeldConnEntry));
/* Initialize tconTelIacBuf */
                tm_teld_set_iac(teldConnEntryPtr->tconTelIacBuf, TM_TC_IAC);
                tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
/* Insert in the open connection entries queue */
                tfServerAppendToQueue( &tm_context(tvTeldConQHead),
                                       (ttServerQueuePtr)teldConnEntryPtr,
                                       TM_TELD_CONN_LIST );
                teldConnEntryPtr->tconDesc = acceptDesc;
/* Owner count is 1 for TM_TELD_CONN_LIST ownership */
                teldConnEntryPtr->tconOwnerCount = 1;
                tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
/* Disable Nagle algorithm (TCP_NODELAY on) */
                tempInt = 1; /* ON */
                errorCode = setsockopt( acceptDesc,
                                        IPPROTO_TCP,
                                        TCP_NODELAY,
                                        (const char TM_FAR *)&tempInt,
                                        sizeof(int) );
/* Handle Urgent Data inline */
                if (errorCode == TM_ENOERROR)
                {
/* SO_OOBINLINE on */
                    errorCode = setsockopt( acceptDesc,
                                            SOL_SOCKET,
                                            SO_OOBINLINE,
                                            (const char TM_FAR *)&tempInt,
                                            sizeof(int) );

                    if (errorCode == TM_ENOERROR)
                    {
                        errorCode = tfRegisterSocketCBParam(
                                                       acceptDesc,
                                                       tfTeldSignal,
                                                       teldConnEntryPtr,
                                                        TM_CB_RECV
                                                      | TM_CB_RECV_OOB
                                                      | TM_CB_SEND_COMPLT
                                                      | TM_CB_REMOTE_CLOSE
                                                      | TM_CB_SOCKET_ERROR
                                                      | TM_CB_RESET );
                        if (errorCode == TM_ENOERROR)
                        {
/* Copy client IP address for our own bookeeping */
                            sockAddressPtr =
                                        &teldConnEntryPtr->tconClientAddress;

                            if (sockAddress.ss_family == AF_INET)
                            {
                                tm_ip_copy(
                                    sockAddress.addr.ipv4.sin_addr.s_addr,
                                    sockAddressPtr->addr.ipv4.sin_addr.s_addr);
                            }
                            else
                            {
                            }

                            sockAddressPtr->ss_port = sockAddress.ss_port;
                            sockAddressPtr->ss_family = sockAddress.ss_family;

/* Let the user know about new connection  - if IPv6 is not enabled, we
 * use the old tfTeldOpened API, to maintain backward compatibility.
 * If IPv6 is enabled, we use the new API so we can pass back a
 * sockaddr_storage address structure.
 */
                            tempSockAddr.sockNgPtr = sockAddressPtr;
                            tfTeldOpened(
                                       (ttUserTeldHandle)teldConnEntryPtr,
                                       tempSockAddr.sockInPtr );
                        }
                    }
                }
            }
            else
            {
/*
 * We could not allocate an ttTeldConnEntry structure, close
 * the connection with tfTeldCloseSocket().
 */
                errorCode = TM_ENOBUFS;
            }
        }
        else
        {
/*
 * We reached the maximum number of connections. Close the connection
 * with tfTeldCloseSocket().
 */
            errorCode = TM_EWOULDBLOCK;
        }
        if (errorCode != TM_ENOERROR)
        {
            if (teldConnEntryPtr != (ttTeldConnEntryPtr)0)
            {
                tfTeldClose(teldConnEntryPtr, errorCode);
            }
            else
            {
/*
 * We either could not allocate an ttTeldConnEntry structure, or we reached
 * the maximum number of connections.
 */
                tfTeldCloseSocket(acceptDesc);
            }
        }
    }
}

/*
 * tfTeldRecv function description.
 * Recv characters on the connection.
 * This routine handles telnet default options (and replies DONT/WONT to any
 * telnet options it does not support), telnet input NVT ascii EOL <CR,LF>,
 * TCP and telnet urgent data (with or without telnet IP), and multiple
 * telnet SYNCH's (to end urgent data), handles not receiving a full telnet
 * command in one recv call, handles receiving more than one telnet command
 * in a recv call.
 * When a full telnet command has been received (the telnet EOL (i.e <CR,LF>
 * in telnet ascii mode, or <IAC,EOR> in telnet binary mode have been
 * received in our buffer)), or if there is no more data to receive and
 * we have some data buffered, the function tfTeldIncoming() function
 * is called.
 * We use 2 input buffers, one raw input telnet buffer where we recv the
 * raw telnet network data input, and one telnet command buffer where we
 * store the data after applying the NVT telnet transformation to the
 * raw telnet network data.
 * Details of the algorithm:
 * 0. Copy connection variables in local variables
 * 1. In a do while loop until either we find a full telnet command, or there
 *    is nothing to read:
 * 1.a If we are not in telnet urgent mode, we check if TCP is in urgent
 *     mode, in which case we transition to telnet urgent mode, and reset
 *     the current telnet raw input (or portion thereof), and current telnet
 *     command input buffer.
 * 1.b If there is a full buffered telnet command in the telnet command buffer
 *     (from the bottom of the loop, or from a previous unsuccessful
 *     tfTeldIncoming), or if there is no more data to recv from the TCP
 *     internal input buffers, then we go ahead and call tfTeldIncoming,
 *     and break out of the loop to give other connections a chance to run.
 * 1.c Otherwise we recv the input data into a raw telnet input buffer,
 *     keeping track of beginning and end of the raw telnet input buffer.
 * 1.d In an inner loop (while raw telnet input buffer is non empty):
 * 1.d.1 we process telnet characters, implementing a subset of the telnet
 *       protocol. We just reply DONT to WILL and WONT client telnet options,
 *       and reply WONT to DO and DONT to options we do not support. If the
 *       user allowed binary and echo options, we allow binary in both
 *       direction, and echo from us. We also allow SGA options in both
 *       directions. In ascii mode, we do our internal NVT conversion
 *       transforming "\r\n", or "\r\0" to our eol flag as specified in
 *       the telnet protocol. In binary mode, we convert end of block
 *       marker (IAC, EOR) to our eol flag. We check
 *       for the (IAC, IP) telnet interrupt process sequence, in which case
 *       we transition to urgent mode (we are in urgent mode if either the
 *       underlying TCP is in urgent mode of we see a telnet interrupt process
 *       (IAC, IP)). In urgent mode, we skip all characters until we reach a
 *       Telnet SYNCH (IAC, DM) which marks the end of telnet urgent data.
 * 1.d.2 Copy telnet data character (if any) in a telnet command buffer.
 * 1.d.3 When telnet command buffer contains a full telnet command ("\r\n",
 *       or "\r\0" has been seen in the raw telnet input buffer in ascii
 *       mode, IAC, EOR in binary mode), or telnet raw input buffer is full,
 *       we make sure we save unprocessed telnet characters to the beginning
 *       of the raw input telnet buffer, in case we had received beyond
 *       one telnet command. We break out of the inner loop.
 * 1.e   Out of the inner loop:
 *       If the raw input telnet buffer is empty, we reset the telnet
 *       indexes to the beginning of the raw telnet buffer. We save all
 *       local indexes and variable back into the TELNETD connection block.
 *       Go back to the outer loop (step 1.a above) if the previous recv
 *       did not fail.
 *
 * Parameters:
 * Value                Meaning
 * teldConnEntryPtr     Pointer to TELNETD connection entry
 *
 * Return:
 * TM_ENOERROR          no error is returned
 */
static int tfTeldRecv (ttTeldConnEntryPtr teldConnEntryPtr)
{
    ttGenericUnion u1;
    tt8BitPtr   telInputBufferPtr;
    tt8BitPtr   teldCmdBufferPtr;
    tt8BitPtr   replyBufferPtr;
    int         bytesToRecv;
    int         socketDesc;
    int         recvBytes;
    int         telCmdLineIndex;
    int         telBufReadIndex;
    int         telBufProcessIndex;
    int         unprocessedBytes;
    int         i;
    int         retCode;
    int         errorCode;
    int         flag;
    int         optionIndex;
    tt16Bit     execFlags;
    tt8Bit      telState;
    tt8Bit      telChar;
    tt8Bit      processCurrentChar;
    tt8Bit      replyCommand;
    tt8Bit      replySent;
    tt8Bit      userWouldBlock;

    errorCode = TM_ENOERROR;
    replySent = TM_8BIT_ZERO;
    userWouldBlock = TM_8BIT_NO;
    recvBytes = 0;
    socketDesc = teldConnEntryPtr->tconDesc;
    execFlags = teldConnEntryPtr->tconExecFlags;
/* Telnet read index in telnet input buffer */
    telBufReadIndex = teldConnEntryPtr->tconTelBufReadIndex;
/* telnet command line */
    teldCmdBufferPtr = &teldConnEntryPtr->tconTelCmdLine[0];
/* Telnet input buffer */
    telInputBufferPtr = &teldConnEntryPtr->tconTelInputBuf[0];
/* telnet command line index */
    telCmdLineIndex = teldConnEntryPtr->tconTelCmdLineIndex;
/* Telnet processing index in telnet input buffer */
    telBufProcessIndex = teldConnEntryPtr->tconTelBufProcessIndex;
/* Remember previous Telnet state */
    telState = teldConnEntryPtr->tconTelState;
    do
    {
/*
 * If we are not in telnet urgent state, check if we have urgent data in
 * the stream, in which case we need to flush our telnet command buffer,
 * telnet input buffer, and enter the telnet urgent state.
 */
        if ( tm_8bit_bits_not_set(telState, TM_TEL_URGENT_STATE) )
        {
/* check if we have received urgent data */
            retCode = tfGetOobDataOffset(socketDesc);
            if (retCode != TM_SOCKET_ERROR)
            {
                telState = (tt8Bit)( (tt8Bit)(telState & TM_TEL_IAC_STATE)
                                    | TM_TEL_URGENT_STATE );
/* Reset current telnet command if any */
                telCmdLineIndex = 0;
                execFlags = (tt16Bit)
                            (execFlags & (tt16Bit)~TM_TELEF_COMMAND_BUFFERED);
/* Reset Telnet input buffer also */
                telBufProcessIndex = 0;
                telBufReadIndex = 0;
            }
        }
        if (    (tm_16bit_one_bit_set(execFlags, TM_TELEF_COMMAND_BUFFERED)
             || (telCmdLineIndex != 0))
             && (userWouldBlock == TM_8BIT_NO))
        {
/* Flush telnet command to the user */
            if ( tm_16bit_one_bit_set(execFlags, TM_TELEF_COMMAND_BUFFERED) )
            {
                flag = TM_TELD_SEND_EOL;
            }
            else
            {
                flag = 0;
            }
            if ( teldConnEntryPtr->tconOutputEcho == TM_TC_WILL )
            {
/* tfTeldSend will flush the reply as well */
                replySent = TM_8BIT_ZERO;
                (void)tfTeldSend( teldConnEntryPtr,
                                  (ttCharPtr)teldCmdBufferPtr,
                                  telCmdLineIndex,
                                  flag,
                                  TM_TELD_EXECUTE_TASK );
            }
            else
            {
                if (replySent != TM_8BIT_ZERO)
                {
                    (void)tfTeldFlush( teldConnEntryPtr,
                                 TM_TELD_EXECUTE_TASK,
                                 TM_8BIT_YES);
                    replySent = TM_8BIT_ZERO;
                }
            }
            errorCode = tfTeldIncoming( (ttUserTeldHandle)teldConnEntryPtr,
                                        (char TM_FAR *)teldCmdBufferPtr,
                                        telCmdLineIndex,
                                        flag );
            if (errorCode == TM_EWOULDBLOCK)
            {
/* start the tconTeldIncTimer */
                if (teldConnEntryPtr->tconTeldIncTimerPtr == TM_TMR_NULL_PTR)
                {
                    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
                    u1.genVoidParmPtr = (ttVoidPtr)teldConnEntryPtr;
                    teldConnEntryPtr->tconTeldIncTimerPtr = 
                            tfTimerAddExt( &teldConnEntryPtr->tconTeldIncTimer,
                                        tfTeldIncomingTimeout,
                                        tfTeldTimerCleanup,
                                        u1,
                                        u1, /* unused */
                                        TM_TELD_INCOMING_TIMEOUT,
                                        TM_TIM_AUTO);
                    teldConnEntryPtr->tconOwnerCount++;
                    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
                }
                else
                {
/* reset the timer */
                    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
                    tm_timer_new_time(teldConnEntryPtr->tconTeldIncTimerPtr, 
                                        TM_TELD_INCOMING_TIMEOUT);
                    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
                }
                userWouldBlock = TM_8BIT_YES;
            }
            else if (errorCode == TM_ENOERROR)
            {
/* if the timer is running, remove it */
                if (teldConnEntryPtr->tconTeldIncTimerPtr != TM_TMR_NULL_PTR)
                {
                    tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
                    tm_timer_remove(teldConnEntryPtr->tconTeldIncTimerPtr);
                    teldConnEntryPtr->tconTeldIncTimerPtr = TM_TMR_NULL_PTR;
                    tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
                }

                teldConnEntryPtr->tconExecFlags = (tt16Bit)
                                    (   execFlags
                                      & (tt16Bit)~TM_TELEF_COMMAND_BUFFERED );
                teldConnEntryPtr->tconTelCmdLineIndex = 0;
                break;
            }
            else
            {
/* Just execute one command to give everybody else a chance at the CPU */
                break;
            }
        }
/*
 * Don't need to check for 0 return (EOF) since we have registered for
 * remote close call back.
 */
        if (telCmdLineIndex > telBufReadIndex)
        {
            bytesToRecv = TM_CMD_LINE_SIZE - telCmdLineIndex;
        }
        else
        {
            bytesToRecv = TM_CMD_LINE_SIZE - telBufReadIndex;
        }

        recvBytes = recv( socketDesc,
                          (ttCharPtr)&telInputBufferPtr[telBufReadIndex],
                          bytesToRecv,
                          MSG_DONTWAIT );
        if (recvBytes > 0)
        {
/* Update read telnet index */
            telBufReadIndex += recvBytes;
        }
        while (telBufProcessIndex < telBufReadIndex)
        {
/* get a new character */
            tm_teld_get_input_char( telChar, telInputBufferPtr,
                                    telBufProcessIndex, packedBufIndex );
            telBufProcessIndex++;
/*
 * Assume current character is special telnet character. If it is not,
 * processCurrentChar will be set below.
 */
            processCurrentChar = TM_8BIT_ZERO;
            replyBufferPtr = (tt8BitPtr)0;
            switch (telState)
            {
                case TM_TEL_IAC_STATE:
/*
 * We have seen an IAC character previously:
 * Process the Telnet command. Just implement a subset of the telnet
 * commands.
 */
                    switch (telChar)
                    {
                        case TM_TC_WILL:
/* IAC, WILL makes us transit to WILL state */
                            telState = TM_TEL_WILL_STATE;
                            break;
                        case TM_TC_WONT:
/* IAC, WONT makes us transit to WOPT state */
                            telState = TM_TEL_WOPT_STATE;
                            break;
                        case TM_TC_DO:
/* IAC, DO makes us transit to DO state */
                            telState = TM_TEL_DO_STATE;
                            break;
                        case TM_TC_DONT:
/* IAC, DONT makes us transit to DOPT state */
                            telState = TM_TEL_DOPT_STATE;
                            break;
                        case TM_TC_IP:
/* Interrupt process */
                            telState = TM_TEL_URGENT_STATE;
/* Flush telnet command buffer */
                            telCmdLineIndex = 0;
                            break;
                        case TM_TC_EOR:
                            telState = TM_TEL_DATA_STATE;
                            if (    teldConnEntryPtr->tconInputBinary
                                 == TM_TC_WILL )
                            {
/* In binary iput mode IAC,EOR is end of record */
                                execFlags = (tt16Bit)
                                      (execFlags | TM_TELEF_COMMAND_BUFFERED);
                            }
                            break;
                        case TM_TC_SB:
                            telState = TM_TEL_SB_STATE;
/* Look for IAC SE */
                            break;
                        case TM_TC_IAC:
/* IAC, IAC means escaped IAC character, process the second IAC */
                            processCurrentChar = TM_8BIT_YES;
/* fall through to default */
                        default:
/* Ignore (skip) all other commands. Go back to data state */
                            telState = TM_TEL_DATA_STATE;
                            break;
                    }
                    break;
                case TM_TEL_WILL_STATE:
                    replyBufferPtr = &teldConnEntryPtr->tconTelIacBuf[0];
/* Default reply is refusal */
                    tm_teld_set_command(replyBufferPtr, TM_TC_DONT);
                    if ( telChar == TM_TO_ECHO)
                    {
/*
 * Supersedes default reply only for supported options. Also refusal.
 * We do not allow the telnet client to echo our characters back to us.
 */
                        replyCommand = TM_TC_DONT;
                    }
                    else
                    {
/* Supersedes default reply only for supported options */
                        replyCommand = TM_TC_DO;
                    }
                    optionIndex = TM_TELD_OPTION_INPUT;
                    goto replyOption;
                case TM_TEL_DO_STATE:
                    replyBufferPtr = &teldConnEntryPtr->tconTelIacBuf[0];
/* Default reply is refusal */
                    tm_teld_set_command(replyBufferPtr, TM_TC_WONT);
/* Supersedes default reply only for supported options */
                    replyCommand = TM_TC_WILL;
                    optionIndex = TM_TELD_OPTION_OUTPUT;
                    goto replyOption;
                case TM_TEL_WOPT_STATE:
                    replyBufferPtr = &teldConnEntryPtr->tconTelIacBuf[0];
/* Default reply is refusal */
                    tm_teld_set_command(replyBufferPtr, TM_TC_DONT);
/* Supersedes default reply only for supported options. Also refusal here. */
                    replyCommand = TM_TC_DONT;
                    optionIndex = TM_TELD_OPTION_INPUT;
                    goto replyOption;
                case TM_TEL_DOPT_STATE:
                    replyBufferPtr = &teldConnEntryPtr->tconTelIacBuf[0];
/* Default reply is refusal */
                    tm_teld_set_command(replyBufferPtr, TM_TC_WONT);
/* Supersedes default reply only for supported options. Also refusal here. */
                    replyCommand = TM_TC_WONT;
                    optionIndex = TM_TELD_OPTION_OUTPUT;
replyOption:
/* Transition to Data state */
                    telState = TM_TEL_DATA_STATE;
                    if (    (telChar == TM_TO_SGA)
                         || (telChar == TM_TO_ECHO)
                         || (telChar == TM_TO_BINARY) )
/* We only support SGA, ECHO, BINARY options */
                    {
                        if (    (    (telChar == TM_TO_ECHO)
                                  && !(   tm_context(tvTeldOptionsAllowed)
                                        & TM_TELD_ECHO_ON) )
                            ||  (    (telChar == TM_TO_BINARY)
                                  && !(   tm_context(tvTeldOptionsAllowed)
                                        & TM_TELD_BINARY_ON) ) )
                        {
/* If we do not allow the option to be negotiated, then refuse it */
                            tm_teld_get_command( replyBufferPtr,
                                                 replyCommand );
                        }
                        if (    teldConnEntryPtr->tconOption[telChar]
                                                            [optionIndex]
                             != replyCommand )
/* If we have not already replied in the same manner */
                        {
                            tm_teld_set_command(replyBufferPtr, replyCommand);
                            teldConnEntryPtr->tconOption[telChar]
                                                        [optionIndex]
                                                          = replyCommand;
                        }
                        else
                        {
/* We have already replied */
                            replyBufferPtr = (tt8BitPtr)0;
                        }
                    }
/* else: Just send the default replyBufferPtr */
                    break;
                case TM_TEL_URGENT_STATE:
/* In urgent state, ignore any character but IAC */
                    if (telChar == TM_TC_IAC)
                    {
                        telState = TM_TEL_URGENT_IAC_STATE;
                    }
                    break;
                case TM_TEL_URGENT_IAC_STATE:
/* In urgent state, after an IAC, Look for a DM to end urgent data */
                    if (telChar == TM_TC_DM)
                    {
/*
 * Telnet SYNCH, signal end of urgent data, we will check if there is more
 * urgent data ahead at the top of the outer loop.
 */
                        telState = TM_TEL_DATA_STATE;
                    }
                    else
                    {
                        telState = TM_TEL_URGENT_STATE;
                    }
                    break;
                case TM_TEL_CR_STATE:
                    telState = TM_TEL_DATA_STATE;
/*
 * "\r\n" marks end of line.
 * "\r\0" marks end of line.
 * "\r" followed by any other character is illegal in telnet ascii mode.
 */
                    if ((telChar != TM_NVT_LF) && (telChar != '\0') )
                    {
/*
 * '\r' followed by any other character besides "\n" or "\0" is illegal, but
 * we tolerate it for the broken Microsoft 2000/XP Telnet client. In this case,
 * process the current character later.
 */
                        telBufProcessIndex--;
                    }
                    break;
                case TM_TEL_DATA_STATE:
                    if (telChar == TM_NVT_CR)
                    {
                        if (    teldConnEntryPtr->tconInputBinary
                             != TM_TC_WILL )
                        {
/*
 * First '\r' in non ascii mode, just remember it (transition to CR_STATE
 * state)
 */
                            telState = TM_TEL_CR_STATE;
/*
 * ecr 2/4/02: added hack for broken Microsoft 2000/XP Telnet client, which
 * sends "\r" for end of line.
 */
                            execFlags =  (tt16Bit) (
                                execFlags | TM_TELEF_COMMAND_BUFFERED);
                        }
                        else
                        {
/* Binary mode. Just process the TM_NVT_CR as data */
                            processCurrentChar = TM_8BIT_YES;
                        }
                    }
                    else
                    {
                        if (telChar == TM_TC_IAC)
                        {
/* Current (IAC) character makes us transition to IAC state */
                            telState = TM_TEL_IAC_STATE;
                        }
                        else
                        {
/* Character is neither '\r' or IAC and we are in data state, process it */
                            processCurrentChar = TM_8BIT_YES;
                        }
                    }
                    break;
                case TM_TEL_SB_STATE:
                    if (telChar == TM_TC_IAC)
                    {
                        telState = TM_TEL_SB_IAC_STATE;
                    }
                    break;
                case TM_TEL_SB_IAC_STATE:
                    if (telChar == TM_TC_SE)
/* End of subnegotiation */
                    {
                        telState = TM_TEL_DATA_STATE;
                    }
                    else
                    {
                        telState = TM_TEL_SB_STATE;
                    }
                    break;

                default:
                    break;
            }
            if (replyBufferPtr != (tt8BitPtr)0)
            {
/* Process DO, DONT, WILL, WONT commands */
/* Set the Telnet option following the IAC, command */
               tm_teld_set_option(replyBufferPtr, telChar);
               (void)tfTeldSend(
                               teldConnEntryPtr,
                               (ttCharPtr)replyBufferPtr,
                               TM_TELD_REPLY_SIZE,
                               TM_TELD_SEND_COMMAND | TM_TELD_SEND_DONT_FLUSH,
                               TM_TELD_EXECUTE_TASK );
                replySent = TM_8BIT_YES;
            }
            if (processCurrentChar != TM_8BIT_ZERO)
            {
/* Process current character. */
                teldCmdBufferPtr[telCmdLineIndex] = telChar;
                telCmdLineIndex++;
            }
/* If telnet command buffer is full, or if we have received a full command */
            if (    tm_16bit_one_bit_set(execFlags, TM_TELEF_COMMAND_BUFFERED)
                 || (telCmdLineIndex == TM_CMD_LINE_SIZE) )
            {
                unprocessedBytes = telBufReadIndex - telBufProcessIndex;
                if (unprocessedBytes != 0)
                {
/* Non empty telnet buffer, copy unprocessed bytes to beginning */
                    for (i = 0; i < unprocessedBytes; i++)
                    {
/* Avoid overlapping bcopy by copying up one byte at a time */
                        telInputBufferPtr[i] =  
                            telInputBufferPtr[telBufProcessIndex + i];
                    }
                    telBufReadIndex = unprocessedBytes;
                    telBufProcessIndex = 0;
                }
/*
 * Break to Process current telnet command or flush the command buffer.
 */
                break;
            }
        }
        if (telBufProcessIndex == telBufReadIndex)
/* If telnet input buffer is empty, reset to beginning */
        {
            telBufProcessIndex = 0;
            telBufReadIndex = 0;
        }
/* save current state */
        teldConnEntryPtr->tconTelState = telState;
/* Save current position in command line */
        teldConnEntryPtr->tconTelCmdLineIndex = telCmdLineIndex;
/* Save telnet command flags */
        teldConnEntryPtr->tconExecFlags = execFlags;
/* Save current Telnet indexes before executing the command */
        teldConnEntryPtr->tconTelBufReadIndex = telBufReadIndex;
        teldConnEntryPtr->tconTelBufProcessIndex = telBufProcessIndex;
/*
 * Try and read more data from the network.
 */
    }
    while ( recvBytes > 0 );
    if (replySent != TM_8BIT_ZERO)
    {
        (void)tfTeldFlush(teldConnEntryPtr, TM_TELD_EXECUTE_TASK, TM_8BIT_YES);
    }
/* Queue unprocessed bytes for later processing */
    if(teldConnEntryPtr->tconTelCmdLineIndex < TM_CMD_LINE_SIZE)
    {
        tfTeldQueueUnScannedCtrlInput(teldConnEntryPtr);
    }
    return TM_ENOERROR;
}


/*
 * tfTeldQueueUnScannedCtrlInput function description:
 * This function is called after each telnet command execution
 * to check whether there is any outstanding
 * Telnet input that needs to be queued at the end of the TELNETD run
 * Q, to give other connections a fair share of the CPU.
 * Detailed algorithm:
 * If our connection is not in the run queue for recv, and if there is a
 * buffered telnet command, or are unprocessed telnet input bytes, or more to
 * recv from the socket, we need to insert into the run queue for later
 * input processing.
 * Parameters:
 * Parameter          Description
 * teldConnEntryPtr   Pointer to TELNETD connection entry
 *
 * Return:
 * None
 */
static void tfTeldQueueUnScannedCtrlInput (
                                        ttTeldConnEntryPtr teldConnEntryPtr )
{
    int tempInt;

    if (   (   teldConnEntryPtr->tconCBFlags
             & (TM_CB_RECV|TM_CB_RECV_OOB) ) == 0 )
/* Not in the run queue for recv */
    {
        if (     ( tm_16bit_one_bit_set( teldConnEntryPtr->tconExecFlags,
                                         TM_TELEF_COMMAND_BUFFERED ) )
             ||  (    teldConnEntryPtr->tconTelBufReadIndex
                   != teldConnEntryPtr->tconTelBufProcessIndex ) )
/* Buffered FTP command, or unprocessed telnet input buffer bytes */
        {
            tempInt = 1;
        }
        else
        {
/* Bytes in the socket recv queue? */
            tempInt = tfGetWaitingBytes(teldConnEntryPtr->tconDesc);
        }
        if ( tempInt > 0 )
        {
/* Queue this entry in the run queue */
           tfTeldSignal( teldConnEntryPtr->tconDesc,
                         (ttVoidPtr)teldConnEntryPtr,
                         TM_CB_RECV );
        }
    }
}

/*
 * tfTeldSend function description:
 * Send data on the telnet connection.
 * 1. Compute the maximum amount of space needed to copy the data, and to
 *    do the telnet NVT transformation.
 * 2. Lock
 * 3. If previous data not yet flushed to the network, do so
 * 4. Check if there is room for all the data to be copied including NVT
 *    transformation. Try and copy (if there is no outstanding telnet output
 *    buffer pointer not yet flushed, and if there is room for all the user
 *    data) to the telnet connection static send buffer, or append to the
 *    telnet connection static send buffer if the outstanding telnet send
 *    buffer pointer points to the static telnet connection entry send
 *    buffer, and if there is room for all of our data, otherwise if if
 *    there is no outstanding telnet send buffer not yet flushed, allocate
 *    a new buffer.
 * 5. If there is room, copy data. In Ascii mode map IAC to IAC,IAC, CR to
 *    CR,NUL, and eol to CR, LF in ascii mode. In binary mode, map IAC to
 *    IAC,IAC, and eol to IAC, EOR.
 * 6. Flush data to the network.
 * 7. Unlock
 * 8. Return 
 *
 * Parameters:
 * Parameter          Description
 * teldConnEntryPtr     Pointer to telnet connection entry
 * bufferPtr            pointer to telnet data to send
 * bytes                number of bytes to send
 * flag                 Indicate whether we need to send CR/LF
 *                      Indicate whether we want to map IAC
 * thread               Context in which this function was called
 *                      (either from the send thread, or execute thread)
 *
 * Return:
 * Value              Meaning
 * TM_ENOERROR        success
 * TM_EWOULDBLOCK     no room for the data
 * TM_ENOBUFS         could not allocate a buffer
 */
static int tfTeldSend ( ttTeldConnEntryPtr  teldConnEntryPtr,
                        ttCharPtr           bufferPtr,
                        int                 bytes,
                        int                 flag,
                        int                 thread )
{
    tt8BitPtr sendBufPtr;
    int       copyBytes;
    int       sendBytes;
    int       errorCode;
    int       spaceNeeded;
    int       extraBufSpace;
    int       loopCount;
    tt8Bit    copyFlag;
    tt8Bit    endChar1;
    tt8Bit    endChar2;

    if (flag & TM_TELD_SEND_COMMAND)
    {
        extraBufSpace = TM_BUF_EXTRA_BYTES;
        spaceNeeded = bytes;
    }
    else
    {
/* Worst case scenario */
        spaceNeeded = 2 * bytes;
        extraBufSpace = 0;
    }
    if (flag & TM_TELD_SEND_EOL)
    {
        spaceNeeded += 2;
    }

    loopCount = 0;
    tm_lock_wait(&teldConnEntryPtr->tconSendLockEntry);
    do
    {
/* assume no error */
        errorCode = TM_ENOERROR;
        loopCount++;
        copyFlag = TM_8BIT_ZERO;
        if (teldConnEntryPtr->tconSendBufPtr != (ttCharPtr)0)
        {
/* Output not yet flushed */
            if (    tm_16bit_bits_not_set( teldConnEntryPtr->tconExecFlags,
                                           TM_TELSF_BUFFER_ALLOCATED ) )
            {
/* Using tconTelSendBuf[]. Append data if there is room */
                if (   teldConnEntryPtr->tconSendBytes + spaceNeeded
                     < TM_BUF_SEND_SIZE + extraBufSpace)
                {
                    copyFlag = TM_8BIT_YES;
                }
            }
        }
        else
        {
            copyFlag = TM_8BIT_YES;
            if (spaceNeeded <= TM_BUF_SEND_SIZE)
            {
                teldConnEntryPtr->tconSendBufPtr = (ttCharPtr)
                                     &teldConnEntryPtr->tconTelSendBuf[0];
            }
            else
            {
                teldConnEntryPtr->tconSendBufPtr = (ttCharPtr)
                              tm_get_raw_buffer((ttPktLen)spaceNeeded);
                if (teldConnEntryPtr->tconSendBufPtr == (ttCharPtr)0)
                {
                    copyFlag = TM_8BIT_ZERO;
                    errorCode = TM_ENOBUFS;
                }
                else
                {
                    teldConnEntryPtr->tconExecFlags = (tt16Bit)
                                        (   teldConnEntryPtr->tconExecFlags
                                          | TM_TELSF_BUFFER_ALLOCATED );
                }
            }
        }
        if (copyFlag != TM_8BIT_ZERO)
        {

            sendBufPtr = (tt8BitPtr)
                         (   teldConnEntryPtr->tconSendBufPtr
                           + teldConnEntryPtr->tconSendBytes );

            for (  copyBytes = 0, sendBytes = 0;
                   copyBytes < bytes;
                   copyBytes++, sendBytes++)
            {
                sendBufPtr[sendBytes] = (tt8Bit)bufferPtr[copyBytes];
                if ( !(flag & TM_TELD_SEND_COMMAND) )
                {
                    if (sendBufPtr[sendBytes] == TM_TC_IAC)
                    {
/* Map IAC, to IAC,IAC */
                        sendBytes++;
                        sendBufPtr[sendBytes] = TM_TC_IAC;
                    }
                    if ( !tm_teld_out_binary(teldConnEntryPtr) )
                    {
/* Map CR to CR NUL */
                        if (sendBufPtr[sendBytes] == TM_NVT_CR)
                        {
                            sendBytes++;
                            sendBufPtr[sendBytes] = TM_NVT_NULL;
                        }
                    }
                }
            }
            if (flag & TM_TELD_SEND_EOL)
            {
                if ( tm_teld_out_binary(teldConnEntryPtr) )
                {
/* Map eol to IAC,EOR */
                    endChar1 = TM_TC_IAC;
                    endChar2 = TM_TC_EOR;
                }
                else
                {
/* Map eol to CR,LF */
                    endChar1 = TM_NVT_CR;
                    endChar2 = TM_NVT_LF;
                }
                sendBufPtr[sendBytes++] = endChar1;
                sendBufPtr[sendBytes++] = endChar2;
            }

            teldConnEntryPtr->tconSendBytes += sendBytes;
/* Output waiting to be flushed */
            if ( !(flag & TM_TELD_SEND_DONT_FLUSH) )
            {
                (void)tfTeldFlush(teldConnEntryPtr, thread, TM_8BIT_ZERO);
            }
        }
        else
        {
/* We have no room. flush the data */
            if (errorCode == TM_ENOERROR)
            {
                errorCode = TM_EWOULDBLOCK;
            }
            (void)tfTeldFlush(teldConnEntryPtr, thread, TM_8BIT_ZERO);
        }
    } while ( (copyFlag == TM_8BIT_ZERO) && (loopCount <= 1) );
    tm_unlock(&teldConnEntryPtr->tconSendLockEntry);
    return errorCode;
}

/*
 * tfTeldFlush function description
 * 1. Obtain telnet connection send lock (if not done so already)
 * 2. If we need to send data, and no other thread is in the process
 *    of sending data
 * 2.1 Save send buffer pointer and send bytes from the telnet connection
 * 2.2 In blocking mode and in the send thread use a send blocking flag
 * 2.3 Release the telnet connection lock
 * 2.4 send the data
 * 2.5 Obtain telnet connection send lock
 * 2.6 Update send buffer pointer, and send bytes in the telnet
 *     connection. If all data have been flushed, and send buffer
 *     was allocated, free it.
 * 3. Release telnet connection send lock (if obtained in step 1)
 * Parameters:
 * Parameter          Description
 * teldConnEntryPtr     Pointer to telnet connection entry
 * thread               Context in which this function was called
 *                      (either from the send thread, or execute thread)
 *
 * Return:
 * Value              Meaning
 * TM_ENOERROR        success, or if send call had returned TM_EWOULDBLOCK
 * other              same as blocking send error codes on failure
 */
static int tfTeldFlush ( ttTeldConnEntryPtr teldConnEntryPtr,
                         int                thread,
                         tt8Bit             needSendLock )
{
    ttCharPtr sendBufPtr;
    int       sendBytes;
    int       errorCode;
    int       bytes;
    int       flag;
    int       socketDesc;
    tt8Bit    freeBuffer;

    errorCode = TM_ENOERROR;
    if (needSendLock != TM_8BIT_ZERO)
    {
        tm_lock_wait(&teldConnEntryPtr->tconSendLockEntry);
    }
    if (    (teldConnEntryPtr->tconSendBufPtr != (ttCharPtr)0)
         && tm_16bit_bits_not_set( teldConnEntryPtr->tconSendFlags,
                                   TM_TELSF_SENDING) )
    {
        teldConnEntryPtr->tconSendFlags = (tt16Bit)
                                           (   teldConnEntryPtr->tconSendFlags
                                             | TM_TELSF_SENDING );
        sendBufPtr = teldConnEntryPtr->tconSendBufPtr;
        sendBytes = teldConnEntryPtr->tconSendBytes;
        if (    (thread == TM_TELD_EXECUTE_TASK)
             || (tm_context(tvTeldBlockingState) == TM_BLOCKING_OFF) )
        {
            flag = MSG_DONTWAIT;
        }
        else /* BLOCKING ON */
        {
            flag = MSG_WAITALL;
        }
/*
 * Save socket descriptor, since it could be reset if connection were to be
 * closed, while we are sending without the tconSendLockEntry on
 */
        socketDesc = teldConnEntryPtr->tconDesc;
        tm_unlock(&teldConnEntryPtr->tconSendLockEntry);
        bytes = send( socketDesc,
                      sendBufPtr,
                      sendBytes,
                      flag );
        tm_lock_wait(&teldConnEntryPtr->tconSendLockEntry);
        freeBuffer = TM_8BIT_ZERO;
        if (bytes >= 0)
        {
            teldConnEntryPtr->tconSendBytes -= bytes;
            if (teldConnEntryPtr->tconSendBytes == 0)
            {
/* Done with the buffer (content appended to TCP send queue) */
                freeBuffer  = TM_8BIT_YES;
            }
            else
            {
                teldConnEntryPtr->tconSendBufPtr += bytes;
            }
        }
        else /* Bytes < 0 */
/* An error occured */
        {
            errorCode = tfGetSocketError(socketDesc);
            if (errorCode == TM_EWOULDBLOCK)
/* Send queue is full, or below low water mark. Try and send later */
            {
                errorCode = TM_ENOERROR;
            }
            else
            {
/* An error occured. We should free the buffer */
                freeBuffer  = TM_8BIT_YES;
                teldConnEntryPtr->tconSendBytes = 0;
            }
        }
        if (freeBuffer)
        {
            if ( tm_16bit_one_bit_set( teldConnEntryPtr->tconSendFlags,
                                       TM_TELSF_BUFFER_ALLOCATED ) )
            {
                teldConnEntryPtr->tconSendFlags = (tt16Bit)
                                    (   teldConnEntryPtr->tconSendFlags
                                      & (tt16Bit)~TM_TELSF_BUFFER_ALLOCATED );
                tm_free_raw_buffer(
                                 (tt8BitPtr)teldConnEntryPtr->tconSendBufPtr);
            }
            teldConnEntryPtr->tconSendBufPtr = (ttCharPtr)0;
        }
        teldConnEntryPtr->tconSendFlags = (tt16Bit)
                                          (   teldConnEntryPtr->tconSendFlags
                                            & (tt16Bit)~TM_TELSF_SENDING );
    }
    if (needSendLock != TM_8BIT_ZERO)
    {
        tm_unlock(&teldConnEntryPtr->tconSendLockEntry);
    }
    return errorCode;
}

/*
 * tfTeldIdleTimeout function description:
 * Idle timeout function. Called when a TELNET connection has been
 * inactive for tvTeldIdleTimeout milliseconds.
 * We just call tfTeldSignal with a Call back timeout flag, to use
 * common code.
 * The tconTimerPtr is reset in tfTeldSignal with
 * the execute lock on to avoid race conditions with the TELNETD server
 * execute loop.
 *
 * Parameters:
 * Value          Meaning
 * timerArgPtr  Pointer to an Telnet connection entry
 *
 * Return:
 * no return value
 */
void tfTeldIdleTimeout (ttVoidPtr      timerBlockPtr,
                        ttGenericUnion userParm1,
                        ttGenericUnion userParm2)
{
/* Avoid compiler warnings about unused parameters */
    TM_UNREF_IN_ARG(userParm2);
    TM_UNREF_IN_ARG(timerBlockPtr);
    
    tfTeldSignal( ((ttTeldConnEntryPtr)(userParm1.genVoidParmPtr))->tconDesc,
                  userParm1.genVoidParmPtr,
                  TM_CB_TIMEOUT );
}
void tfTeldIncomingTimeout (ttVoidPtr      timerBlockPtr,
                        ttGenericUnion userParm1,
                        ttGenericUnion userParm2)
{
/* Avoid compiler warnings about unused parameters */
    TM_UNREF_IN_ARG(userParm2);
    TM_UNREF_IN_ARG(timerBlockPtr);

/* 
 * User's tfTeldIncoming function returned TM_EWOULDBLOCK and tfTeldIncoming
 * has not returned TM_ENOERROR due to more incoming data within the last
 * TM_TELD_INCOMING_TIMEOUT seconds.
 */
    tfTeldSignal( ((ttTeldConnEntryPtr)(userParm1.genVoidParmPtr))->tconDesc,
                userParm1.genVoidParmPtr,
                TM_CB_RECV );
}
static void tfTeldTimerCleanup(ttVoidPtr vptr, tt8Bit flags)
{
    ttTimerPtr            timerPtr;
    ttTeldConnEntryPtr teldConnEntryPtr;

    timerPtr = (ttTimerPtr)vptr;
    if (!(flags & TM_TIMER_CLEANUP_LOCKED))
    {
        tm_call_lock_wait(&tm_context(tvTeldExecuteLockEntry));
    }
    teldConnEntryPtr = 
            (ttTeldConnEntryPtr)timerPtr->tmrUserParm1.genVoidParmPtr;
    if (teldConnEntryPtr->tconOwnerCount > 1)
    {
        teldConnEntryPtr->tconOwnerCount--;
    }
    else
    {
        tm_free_raw_buffer((ttRawBufferPtr)teldConnEntryPtr);
    }
    if (!(flags & TM_TIMER_CLEANUP_LOCKED))
    {
        tm_call_unlock(&tm_context(tvTeldExecuteLockEntry));
    }
}
/*
 * tfTeldTimerRemove function description:
 * Remove and reset the idle connection timer, called with the execute
 * lock on to protect tconTimerPtr. Called from either tfTeldClose()
 * when a connection control is being closed, or from the execute loop
 * while processing an TELNETD connection entry.
 * 1. If the TELNETD connection entry timer pointer is non null,
 *    remove the timer.
 * 2. Reset the timer pointer in the Telnet connection entry.
 * Parameters:
 * Value                Meaning
 * timerArgPtr          Pointer to TELD connection entry
 * Return:
 * No return
 */
static void tfTeldTimerRemove (ttVoidPtr timerArgPtr)
{
#define teldConnEntryPtr    ((ttTeldConnEntryPtr)timerArgPtr)
    ttTimerPtr  timerPtr;

    timerPtr = teldConnEntryPtr->tconTimerPtr;
    if (timerPtr != TM_TMR_NULL_PTR)
    {
        tm_timer_remove(timerPtr);
        teldConnEntryPtr->tconTimerPtr = TM_TMR_NULL_PTR;
    }
#undef teldConnEntryPtr
}

/* 
 * Returns the number of bytes currently in the TCP socket send queue for this 
 * telnet session. 
 */
int tfTeldSendQueueBytes(ttUserTeldHandle teldHandle)
{
    int              sockDesc;
    int              retCode;
    ttSocketEntryPtr socketEntryPtr;
    tt32Bit          totalInFlightBytes;

    sockDesc = ((ttTeldConnEntryPtr) teldHandle)->tconDesc;

    socketEntryPtr = tfSocketCheckValidLock(  sockDesc
                                           );
    if (socketEntryPtr != (ttSocketEntryPtr) 0)
    {
        totalInFlightBytes = socketEntryPtr->socSendQueueBytes;
        retCode = (int)totalInFlightBytes;
    }
    else
    {
        retCode = TM_SOCKET_ERROR;
    }
    return tfSocketReturn(socketEntryPtr, sockDesc, TM_ENOERROR, retCode);
}

/*
 * Returns the total size of the TCP socket send queue for this telnet session.
 */
int tfTeldSendQueueSize(ttUserTeldHandle teldHandle)
{
    int              sockDesc;
    int              retCode;
    ttSocketEntryPtr socketEntryPtr;

    sockDesc = ((ttTeldConnEntryPtr) teldHandle)->tconDesc;

    socketEntryPtr = tfSocketCheckValidLock(  sockDesc
                                           );
    if (socketEntryPtr != (ttSocketEntryPtr) 0)
    {
        retCode = (int) socketEntryPtr->socMaxSendQueueBytes;
    }
    else
    {
        retCode = TM_SOCKET_ERROR;
    }
    return tfSocketReturn(socketEntryPtr, sockDesc, TM_ENOERROR, retCode);
}

#ifdef TM_USE_STOP_TRECK
void tfTeldDeInit (
#ifdef TM_MULTIPLE_CONTEXT
                ttUserContext contextHandle
#else /* !TM_MULTIPLE_CONTEXT */
                void
#endif /* !TM_MULTIPLE_CONTEXT */
                )
{
    ttTeldConnEntryPtr teldConnEntryPtr;

    tm_call_lock_wait(&tm_context_var(tvTeldExecuteLockEntry));
    if (tm_context_var(tvTeldListenDesc) != TM_SOCKET_ERROR)
    {
/* Close the listening socket */
        tfTeldCloseListen();
    }
    tm_context_var(tvTeldPendingConnections) = 0;
    teldConnEntryPtr = (ttTeldConnEntryPtr)
                                (tm_context_var(tvTeldConQHead).
                                  servHeadNextQEntryPtr(TM_TELD_CONN_LIST));
    while (teldConnEntryPtr
                        != (ttTeldConnEntryPtr)&tm_context_var(tvTeldConQHead) )
    {
        tm_call_unlock(&tm_context_var(tvTeldExecuteLockEntry));
        tfTeldClose(teldConnEntryPtr, TM_ECONNABORTED);
        tm_call_lock_wait(&tm_context_var(tvTeldExecuteLockEntry));
        teldConnEntryPtr = (ttTeldConnEntryPtr)
                                (tm_context_var(tvTeldConQHead).
                                    servHeadNextQEntryPtr(TM_TELD_CONN_LIST));
    }
    tm_call_unlock(&tm_context_var(tvTeldExecuteLockEntry));
}
#endif /* TM_USE_STOP_TRECK */

#else /* !TM_USE_TELNETD */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */


/* To allow link for builds when TM_USE_TELNETD is not defined */
int tvTelnetdDummy = 0;

#endif /* !TM_USE_TELNETD */

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:53      For: Xilinx-Samsung
 * File Name: trtelntd.c    File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 49553    Output checksum: 61884
 */
