/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Treck test application
 *
 * Filename: trtest.c
 * Author: Jason
 * Date Created: 3/21/2000
 * $Source: source/trtest.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2011/02/14 15:53:07CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */

#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trglobal.h>
#include <trproto.h>
#include <trtest.h>

/* Default listen backlog entries for server tests */
#define TM_TEST_LISTEN_BACKLOG  10

/* Various states of the test: initial, connecting, sending, receiving */
#define TM_TEST_PHASE_INITIAL    0
#define TM_TEST_PHASE_SEND       1
#define TM_TEST_PHASE_RECV       2
#define TM_TEST_PHASE_CONNECTING 3
#define TM_TEST_PHASE_CLOSE      4

/* Used for the fill and validate routines to write/read a long word */
#define TM_TEST_FILL_INIT TM_UL(0x00010203)
#define TM_TEST_FILL_INC  TM_UL(0x04040404)

/* Default echo and discards port (works for UNIX and NT) */
#define TM_STD_ECHO_PORT    htons(7)
#define TM_STD_DISCARD_PORT htons(9)


/* Returns the group of the test (UDP, TCP, etc) */
#define tm_test_group(type) ((tt32Bit)(type) & TM_UL(0xffffff00))

/* 
 * Checks for test type (send, recv, etc).  These overlap each other; That is,
 * a echo client is a send type as well as a echo type test.
 */

#define tm_test_udp_echo_type(type) \
( ((type) == TM_TEST_UDP_ECHO_CLIENT) || \
  ((type) == TM_TEST_UDP_ECHO_SERVER) )

#define tm_test_udp_echo_server_type(type) \
  ((type) == TM_TEST_UDP_ECHO_SERVER)

#define tm_test_udp_send_type(type) \
( ((type) == TM_TEST_UDP_ECHO_CLIENT) || \
  ((type) == TM_TEST_UDP_SEND) )

#define tm_test_udp_recv_type(type) \
( ((type) == TM_TEST_UDP_ECHO_SERVER) || \
  ((type) == TM_TEST_UDP_RECV) ) 

#define tm_test_tcp_echo_type(type) \
( ((type) == TM_TEST_TCP_ECHO_CLIENT) || \
  ((type) == TM_TEST_TCP_ECHO_SERVER) )

#define tm_test_tcp_echo_server_type(type) \
  ((type) == TM_TEST_TCP_ECHO_SERVER)

#define tm_test_tcp_echo_client_type(type) \
  ((type) == TM_TEST_TCP_ECHO_CLIENT)

#define tm_test_tcp_send_type(type) \
( ((type) == TM_TEST_TCP_ECHO_CLIENT) || \
  ((type) == TM_TEST_TCP_SEND) )

#define tm_test_tcp_recv_type(type) \
( ((type) == TM_TEST_TCP_ECHO_SERVER) || \
  ((type) == TM_TEST_TCP_RECV) )

#define tm_test_recv_type(type) \
( tm_test_tcp_recv_type(type) || \
  tm_test_udp_recv_type(type) )

#define tm_test_send_type(type) \
( tm_test_tcp_send_type(type) || \
  tm_test_udp_send_type(type) )

#define tm_test_echo_type(type) \
( tm_test_tcp_echo_type(type) || \
  tm_test_udp_echo_type(type) )

#define tm_test_echo_client_type(type) \
( ((type) == TM_TEST_TCP_ECHO_CLIENT) || \
  ((type) == TM_TEST_UDP_ECHO_CLIENT) ) 

#define tm_test_connect_type(type) \
( ((type) == TM_TEST_TCP_CONNECT) )


/* State transition table */
static const unsigned char  TM_CONST_QLF  tlTestStateTable[5] = 
    { 0,
      TM_TEST_PHASE_RECV,
      TM_TEST_PHASE_SEND,
      0,
      TM_TEST_PHASE_CONNECTING,
    };


static int tfTestUdp( int           sock,
                      int           testType,
                      struct sockaddr_storage TM_FAR * ipAddrPtr,
                      int           portNumber,
                      int           dataSize,
                      ttTestInfoPtr testInfoPtr,
                      ttUser32Bit   flags );

static int tfTestTcp( int                    sock,
                      int                    testType,
                      struct sockaddr_storage TM_FAR * ipAddrPtr,
                      int                    portNumber,
                      int                    dataSize,
                      ttTestInfoPtr          testInfoPtr,
                      ttUser32Bit            flags );

static int tfTestMemValidate(char TM_FAR * dataPtr, int dataLen);

static void tfTestMemFill(char TM_FAR * dataPtr, int dataLen);

TM_NEARCALL static void tfTestSocketSignal( int           sock,
                                void TM_FAR * sockParam, 
                                int           eventFlags );

static int tfTestTcpConnect( struct sockaddr_storage TM_FAR * ipAddrPtr,
                             int           portNumber,
                             ttTestInfoPtr testInfoPtr, 
                             ttUser32Bit   flags,
                             tt8Bit addrFamily);

static void tfTestRandom( int TM_FAR *         testType,
                          int TM_FAR *         dataSize,
                          ttUser32Bit TM_FAR * flags );


/*
 * tfTestMemFill()
 *
 *   Fills an outgoing data buffer with the byte sequence 0x00, 0x01, 0x02, etc.
 *   We start by filling in the long words, and then catch the extra 16-bit
 *   and 8-bit value at the end.
 */
static void tfTestMemFill(char TM_FAR * dataPtr, int dataLen)
{
    tt32Bit       data;
    int           dataCount;
    unsigned char numCounter;

    numCounter = 0;

    data = TM_TEST_FILL_INIT; /* For compiler warnings */
    for (dataCount=0; (unsigned)dataCount < ((unsigned)dataLen >> 2);
         dataCount++)
    {
/* 
 * Every 64 long words the counter value overflows and needs to be reset.  This
 * code also handles the initialization of the counter value.
 */
        if ( (dataCount % 64) == 0 )
        {
            data = TM_TEST_FILL_INIT;
        }       

        *((tt32BitPtr) dataPtr) = data;
        dataPtr += sizeof(tt32Bit);

        data += TM_TEST_FILL_INC;

        numCounter += 4;
    }       

/* 16-bit aligned value at end of buffer */
    if (dataLen & 2)
    {
        *dataPtr++ = (char)numCounter++;
        *dataPtr++ = (char)numCounter++;
    }

/* 8-bit aligned value at end of buffer */
    if (dataLen & 1)
    {
        *dataPtr++ = (char)numCounter;
    }
    
    return; 

}

/*
 * tfTestMemValidate()
 *
 *   This routine performs the inverse of the tfTestMemFill routine above; it
 *   compares the incoming data buffer to what tfTestMemFill would have filled
 *   it with.  This routine also uses the same method to compare the data as
 *   tfTestMemFill uses to fill the buffer.  If any data doesn't match, return
 *   TM_EIO to the user.
 */
static int tfTestMemValidate(char TM_FAR * dataPtr, int dataLen)
{
    int           errorCode;
    int           dataCount;
    tt32Bit       data;
    unsigned char numCounter;

    errorCode = TM_ENOERROR;
    data = TM_TEST_FILL_INIT; /* For compiler warnings */

    numCounter = 0;

    for (dataCount=0; (unsigned)dataCount < ((unsigned)dataLen >> 2);
         dataCount++)
    {
        if ( (dataCount % 64) == 0 )
        {
            data = TM_TEST_FILL_INIT;
        }

        if ( *((tt32BitPtr) dataPtr) != data )
        {
            errorCode = TM_EIO;
            break;
        }

        dataPtr += sizeof(tt32Bit);

        data += TM_TEST_FILL_INC;

        numCounter += 4;
    }       

    if (dataLen & 2)
    {
        if ( (((unsigned char)*dataPtr++) != numCounter++) || 
             ( ((unsigned char)*dataPtr++) != numCounter++) )
        {
            errorCode = TM_EIO;
        }
    }

    if (dataLen & 1)
    {
        if ( ((unsigned char)*dataPtr++) != numCounter )
        {
            errorCode = TM_EIO;
        }
    }
    
    return errorCode;
}


/*
 * tfTestUdp
 *
 *   Called by the main test routine,tfTestTreck, and tfTestUserExecute for most
 *   UDP tests (the exception being connected UDP, which uses the TCP test
 *   routines).  This routine performs four primary tests: send, receive, 
 *   echo client, and echo server.  It does this through a rudimentary FSM,
 *   consisting of three states: initial, sending and receiving.  After one 
 *   phase has completed, the test type is checked, and the FSM moves to the
 *   next phase, assuming no error has occurred.  For instance, when a sending
 *   phase is complete, we examine the test type:  if the test is an echo test,
 *   the next phase will be a receiving state.  Otherwise, we stay in the 
 *   current sending state.  This allows us to handle all four tests with only
 *   2 sets of code (send and receive).
 */
static int tfTestUdp( int                    sock,
                      int                    testType,
                      struct sockaddr_storage TM_FAR * ipAddrPtr,
                      int                    portNumber,
                      int                    dataSize,
                      ttTestInfoPtr          testInfoPtr,
                      ttUser32Bit            flags )
{
    int                testCountDec;
    int                errorCode;
    int                retCode;
    int                addrLen;
    struct sockaddr_storage addr;
    unsigned char      testForever;
    char TM_FAR *      dataPtr;
    ttUserMessage      bufHandle;
    ttUser32Bit        testCount;
    unsigned char      phase;
    ttSockAddrPtrUnion tempSockAddr;

    errorCode = TM_ENOERROR;
    retCode = TM_SOCKET_ERROR; /* to avoid compiler warning */

/* Restore info from the state vector */
    dataPtr   = testInfoPtr->testDataPtr;
    bufHandle = testInfoPtr->testBufHandle;
    testCount = testInfoPtr->testCount;
    phase     = testInfoPtr->testPhase;

    tm_bcopy(
        ipAddrPtr,
        &addr,
        sizeof(struct sockaddr_storage));
    addr.ss_port        = (ttIpPort) portNumber;
    tempSockAddr.sockNgPtr = &addr;
/*
 * INITIAL phase
 */
    if (phase == TM_TEST_PHASE_INITIAL)
    {

/* 
 * According to the test type, set the next phase.  If this is a server test,
 * bind to a port, and if this a echo test, double the times we run this test.
 */
        if ( tm_test_udp_send_type(testType) )
        {
            phase = TM_TEST_PHASE_SEND;
            if (flags & TM_TEST_FLAG_UDP_CONNECT)
            {
                errorCode = connect(
                    sock, tempSockAddr.sockPtr, sizeof(struct sockaddr_storage) );

                if (errorCode != TM_ENOERROR)
                {
                    errorCode = tfGetSocketError(sock);
                    goto testUdpEnd;
                }
            }
        }
        if (tm_test_udp_recv_type(testType))
        {
            phase = TM_TEST_PHASE_RECV;
            errorCode = bind( sock, 
                              tempSockAddr.sockPtr, 
                              sizeof(struct sockaddr_storage) );

            if (errorCode == TM_SOCKET_ERROR)
            {
                errorCode = tfGetSocketError(sock);
                goto testUdpEnd;
            }
        }
        if ( tm_test_udp_echo_type(testType) )
        {
            testCount <<= TM_UL(1);
        }

/* 
 * If we're not using the Treck zero copy extensions, preallocate a normal
 * buffer for use with send/recv.
 */
        if ( !(flags & TM_TEST_FLAG_ZEROCOPY) )
        {
            dataPtr = (char TM_FAR *) tm_get_raw_buffer((ttPktLen)dataSize);

            if (dataPtr == (char TM_FAR *) 0)
            {
                errorCode = TM_ENOBUFS;
            }

        }


    }

    if (testCount > TM_UL(0))
    {
        testCountDec = 1;
        testForever  = 0;
    }
    else
    {
        testCountDec = 0;
        testForever  = 1;
    }

    while (    ((testCount > TM_UL(0)) || (testForever)) 
            && (errorCode == TM_ENOERROR) )
    {

/*
 * SEND phase
 */
        if (phase == TM_TEST_PHASE_SEND)
        {

/* 
 * If we don't currently have a zero copy buffer, allocate one (this could be
 * executed depending on the test being executed - for instance, if this is 
 * a send test, we'll need to allocate a buffer each time, because the previous
 * send 'freed' the buffer.  However, with an echo test, we'll never need to
 * allocate a buffer, because one will be available to us from the receive.
 * The same works with freeing the zero copy buffer below.
 */
            if ( (dataPtr == (char TM_FAR *) 0) && 
                                               (flags & TM_TEST_FLAG_ZEROCOPY) )
            {
                bufHandle = tfGetZeroCopyBuffer(dataSize, &dataPtr);

                if (dataPtr == (char TM_FAR *) 0)
                {
                    errorCode = TM_ENOBUFS;
                    goto testUdpEnd;
                }
            }

/* Fill the send buffer with data */
            if ( flags & TM_TEST_FLAG_FILL_DATA )
            {
                tfTestMemFill(dataPtr, dataSize);
            }

/* Send the data buffer */
            if ( flags & TM_TEST_FLAG_ZEROCOPY )
            {
                if (flags & TM_TEST_FLAG_UDP_CONNECT)
                {
                    retCode = tfZeroCopySend(sock, bufHandle, dataSize, 0);
                }
                else
                {
                    retCode = tfZeroCopySendTo( sock, bufHandle, dataSize, 0,
                                                tempSockAddr.sockPtr, 
                                                sizeof(struct sockaddr_storage) );
                }
                if (retCode > 0)
                {
/* Zero copy buffer has been freed by tfZeroCopySendTo */
                    bufHandle = (ttUserMessage) 0;
                    dataPtr   = (char TM_FAR *) 0;
                }
                else
                {
                    if (retCode < 0)
                    {
                        errorCode = tfGetSocketError(sock);
                        if (errorCode != TM_EWOULDBLOCK)
                        {
/* ZC buffer was freed by tfZeroCopySend */
                            bufHandle = (ttUserMessage) 0;
                            dataPtr   = (char TM_FAR *) 0;
                        }
                    }
                }
            }
            else
            {
                if (flags & TM_TEST_FLAG_UDP_CONNECT)
                {
                    retCode = send(sock, dataPtr, dataSize, 0);
                }
                else
                {
                    retCode = sendto( sock, dataPtr, dataSize, 0,
                                      tempSockAddr.sockPtr, 
                                      sizeof(struct sockaddr_storage) );
                }
            }

        }

/*
 * RECEIVE phase 
 */
        if (phase == TM_TEST_PHASE_RECV)
        {

/* 
 * If a zero copy buffer is still allocated, free it here, since 
 * tfZeroCopyRecvFrom will give us a new one.
 */
            if ( (flags & TM_TEST_FLAG_ZEROCOPY) && 
                 (dataPtr != (char TM_FAR *) 0) )
            {
                (void)tfFreeZeroCopyBuffer(bufHandle);
                bufHandle = (ttUserMessage) 0;
                dataPtr   = (char TM_FAR *) 0;
            }

/* Receive the data */
            addrLen = sizeof(struct sockaddr_storage);
            if ( flags & TM_TEST_FLAG_ZEROCOPY )
            {
                if (flags & TM_TEST_FLAG_UDP_CONNECT)
                {
                    retCode = tfZeroCopyRecv(sock, &bufHandle, &dataPtr, 
                                             testInfoPtr->testDataSize, 0);
                }
                else
                {
                    retCode = tfZeroCopyRecvFrom( sock, &bufHandle, &dataPtr, 
                                                  testInfoPtr->testDataSize, 0, 
                                                  tempSockAddr.sockPtr, 
                                                  &addrLen );
                }
            }
            else
            {
                if (flags & TM_TEST_FLAG_UDP_CONNECT)
                {
                    retCode = recv(sock, dataPtr,
                                   testInfoPtr->testDataSize, 0);
                }
                else
                {
                    retCode = recvfrom( sock, dataPtr,
                                        testInfoPtr->testDataSize, 0,
                                        tempSockAddr.sockPtr, 
                                        &addrLen );
                }
            }        

/* If enabled, validate that the incoming data is what we expect */
            if ( (flags & TM_TEST_FLAG_VALIDATE) && (retCode > 0) )
            {
                errorCode = tfTestMemValidate(dataPtr, dataSize);
            }
            if (retCode == 0)
            {
                if (flags & TM_TEST_FLAG_ZEROCOPY)
                {
                    if (dataPtr != (char TM_FAR *) 0)
                    {
/* Received a zero byte datagram. free the zero copy buffer if any. */
                        (void)tfFreeZeroCopyBuffer(bufHandle);
                        bufHandle = (ttUserMessage) 0;
                        dataPtr   = (char TM_FAR *) 0;
                    }
                }
/* Received a zero byte datagram. Ignore */
                break;
            }
        }

/* Handle errors from both the send and receive phases: */
        if (retCode < 0)
        {
/* An error occured - record the error and exit the loop */
            errorCode = tfGetSocketError(sock);
            break;
        }
        else
        {
            testCount -= (ttUser32Bit)testCountDec;
/* Send or receive sucessful: decrement count and (maybe) move to next phase */
            if ( tm_test_udp_echo_type(testType) )
            {
                phase = tlTestStateTable[phase];
                if (    (testType == TM_TEST_UDP_ECHO_SERVER)
                     && (phase == TM_TEST_PHASE_SEND) )
                {
                    dataSize = retCode;
                    continue;
                }
            }
        }

/* If in non-blocking mode, exit this loop and return to the user */
        if ( (flags & TM_TEST_FLAG_NONBLOCKING) && 
             (errorCode == TM_ENOERROR) && 
             ((testCount > TM_UL(0)) || testForever) )
        {
            errorCode = TM_EWOULDBLOCK;
        }

    }

/* 
 * If we're actually done with this test (ie, not just exiting with a 
 * EWOULDBLOCK) free up the send/receive buffers.
 */
    if ( (dataPtr != (char TM_FAR *) 0) && (errorCode != TM_EWOULDBLOCK) )
    {
        if (flags & TM_TEST_FLAG_ZEROCOPY)
        {
            (void)tfFreeZeroCopyBuffer(bufHandle);
            bufHandle = (ttUserMessage) 0;
            dataPtr   = (char TM_FAR *) 0;
        }
        else
        {
            tm_free_raw_buffer( (ttRawBufferPtr) dataPtr);
            dataPtr = (char TM_FAR *) 0;
        }
    }

/* Save info to the state vector */
    tm_bcopy(
        &addr,
        &(testInfoPtr->testIpAddress),
        sizeof(struct sockaddr_storage));
    testInfoPtr->testPortNum   = addr.ss_port;
    testInfoPtr->testDataPtr   = dataPtr;
    testInfoPtr->testBufHandle = bufHandle;
    testInfoPtr->testPhase     = phase;
    testInfoPtr->testCount     = testCount;

testUdpEnd:
    return (errorCode);
}

/*
 * tfTestTcpConnect()
 *
 *   Performs the TCP connect test.  Repeatedly create a socket, connect to a
 *   remote host and then close the socket.  On many desktop OS's, connecting
 *   this many times upsets the OS, so the remote peer may start rejecting 
 *   connections with a TCP RST.
 */
static int tfTestTcpConnect( struct sockaddr_storage TM_FAR * ipAddrPtr,
                             int           portNumber,
                             ttTestInfoPtr testInfoPtr, 
                             ttUser32Bit   flags,
                             tt8Bit addrFamily)
{

    int                 errorCode;
    int                 sock;
    ttUser32Bit         testCount;
    unsigned char       testCountDec;
    unsigned char       testForever;
    ttSockAddrPtrUnion  tempSockAddr;

    errorCode = TM_ENOERROR;

/* Restore data from state vector */
    testCount = testInfoPtr->testCount;
    sock      = testInfoPtr->testSock;

    ipAddrPtr->ss_port        = (ttIpPort) portNumber;

    tempSockAddr.sockNgPtr = ipAddrPtr;

    if (testCount > TM_UL(0))
    {
        testCountDec = 1;
        testForever  = 0;
    }
    else
    {
        testCountDec = 0;
        testForever  = 1;
    }

    while (    ((testCount > TM_UL(0)) || testForever) 
            && (errorCode == TM_ENOERROR) )
    {
    
        errorCode = connect(sock, tempSockAddr.sockPtr, 
                            sizeof(struct sockaddr_storage));

        if (errorCode == TM_SOCKET_ERROR)
        {
            errorCode = tfGetSocketError(sock);

/* 
 * In non-blocking mode, if connect hasn't completed, it first returns
 * EINPROGRESS, and EALREADY each time after that (returning another error
 * code (hopefully ENOERROR) when the operation is complete).  For our purposes,
 * both of these return values are equivalent to EWOULDBLOCK.
 */
            if ((errorCode == TM_EINPROGRESS) || (errorCode == TM_EALREADY))
            {
                errorCode = TM_EWOULDBLOCK;
            }
        }
        else
        {
            testCount -= testCountDec;

            (void)tfClose(sock);

            sock = socket(addrFamily, SOCK_STREAM, 0);
          
            if (sock == TM_SOCKET_ERROR)
            {
                errorCode = TM_SOCKET_ERROR;
                break;            
            } 
            
            if (flags & TM_TEST_FLAG_NONBLOCKING)
            {
                (void)tfBlockingState(sock, TM_BLOCKING_OFF);
            }                       
        }


    }


    testInfoPtr->testCount = testCount;
    testInfoPtr->testSock  = sock;

    return errorCode;
}


/*
 * tfTestSocketSignal()
 *
 *   Socket callback function.  Called in non-blocking mode when a connect
 *   has completed or when an accept is ready.  Mark the flag in the session
 *   structure.
 */
static void tfTestSocketSignal( int           sock, 
                                void TM_FAR * sockParam, 
                                int           eventFlags )
{
    ttTestInfoPtr testInfoPtr;

    TM_UNREF_IN_ARG(sock);

    testInfoPtr = (ttTestInfoPtr) sockParam;

    testInfoPtr->testEvents =
        eventFlags & (TM_CB_CONNECT_COMPLT | TM_CB_ACCEPT);
    
}


/*
 * tfTestTcp()
 *
 *   Called by the main test routine, tfTestTreck, to perform the TCP send,
 *   receive, echo client and echo server tests.  This routine is very similar
 *   to the tfTestUdp() routine above, and uses the same method of a FSM to
 *   perform various tests.  tfTestTcp adds an additional state, 'connecting'
 *   to deal with connection establishment.  In this state, we'll wait for
 *   either a successful 'connect' or 'accept'.  In blocking mode, this is
 *   done through the blocking connect/accept calls.  In non-blocking mode,
 *   we wait for callback notification.
 */
static int tfTestTcp( int                    sock,
                      int                    testType,
                      struct sockaddr_storage TM_FAR * ipAddrPtr,
                      int                    portNumber,
                      int                    dataSize,
                      ttTestInfoPtr          testInfoPtr,
                      ttUser32Bit            flags )
{
    int                testCountDec;
    int                errorCode;
    int                retCode;
    struct sockaddr_storage addr;
    int                addrLen;
    unsigned char      testForever;
    char TM_FAR *      dataPtr;
    char TM_FAR *      bufPtr;
    ttUserMessage      bufHandle;
    int                dataSock;
    int                recvdBytes;
    int                sentBytes;
    int                echoBytes;
    ttUser32Bit        testCount;
    unsigned char      phase;
    ttSockAddrPtrUnion tempSockAddr;
    int                receivedSize;
    int                tempInt;


    receivedSize = dataSize;
    errorCode = TM_ENOERROR;
    retCode = TM_SOCKET_ERROR; /* to avoid compiler warning */

/* Restore info from the state vector */
    dataPtr    = testInfoPtr->testDataPtr;
    bufHandle  = testInfoPtr->testBufHandle;
    dataSock   = testInfoPtr->testDataSock;
    recvdBytes = testInfoPtr->testRecvdBytes;
    sentBytes  = testInfoPtr->testSentBytes;
    echoBytes  = testInfoPtr->testEchoBytes;
    testCount  = testInfoPtr->testCount;
    phase      = testInfoPtr->testPhase;


    tm_bcopy(
        ipAddrPtr,
        &addr,
        sizeof(struct sockaddr_storage));
    addr.ss_port        = (ttIpPort) portNumber;
    tempSockAddr.sockNgPtr = &addr;

/*
 * INITIAL phase
 */
    if (phase == TM_TEST_PHASE_INITIAL)
    {

        if ( tm_test_tcp_send_type(testType) )
        {
            phase = TM_TEST_PHASE_SEND;
        }
        if ( tm_test_tcp_recv_type(testType) )
        {
            phase = TM_TEST_PHASE_RECV;
        }

/* 
 * If we're not using the Treck zero copy extensions, preallocate a normal
 * buffer for use with send/recv.
 */
        if ( !(flags & TM_TEST_FLAG_ZEROCOPY) )
        {
            dataPtr = (char TM_FAR *) tm_get_raw_buffer((ttPktLen)dataSize);

            if (dataPtr == (char TM_FAR *) 0)
            {
                errorCode = TM_ENOBUFS;
                goto testTcpEnd;
            }
        }

/*
 * If this test is a send test (ie, client) attempt to connect to the remote
 * host. 
 */
        if ( tm_test_tcp_send_type(testType) )
        {

            if (flags & TM_TEST_FLAG_NONBLOCKING)
            {
                testInfoPtr->testEvents = 0;

                errorCode = tfRegisterSocketCBParam( sock, 
                                                     tfTestSocketSignal,
                                                     (void TM_FAR *) 
                                                     testInfoPtr,
                                                     TM_CB_CONNECT_COMPLT );
            }

            errorCode = connect(
                sock, tempSockAddr.sockPtr, sizeof(struct sockaddr_storage) );
 
/* 
 * If 'connect' returns EINPROGRESS (meaning that we're in non-blocking mode
 * move to CONNECTING phase where we'll wait for a connect complete callback.
 * Otherwise, if the connect was successful, set the data socket to the
 * connected socket.
 */ 
            if (errorCode == TM_ENOERROR)
            {
                dataSock = sock;
            }
            else
            {
                errorCode = tfGetSocketError(sock);

                if (errorCode == TM_EINPROGRESS)
                {
                    phase = TM_TEST_PHASE_CONNECTING;
                    errorCode = TM_EWOULDBLOCK;
                }
            }

        }

/* Accept connection from remote host */
        else
        {
/*
 * Need to set SO_REUSEADDR in case there is still an accept socket from a
 * previous server session that is still in time wait state.
 */
            if(tm_test_tcp_echo_type(testType))
            {
                tempInt = 1;
                errorCode = setsockopt( sock,
                                        SOL_SOCKET,
                                        SO_REUSEADDR,
                                        (const char TM_FAR *)&tempInt,
                                        sizeof(int) );
            }
            if (errorCode == TM_ENOERROR)
            {
                errorCode = bind( sock, 
                                  tempSockAddr.sockPtr, 
                                   sizeof(struct sockaddr_storage) );

                if (errorCode == TM_ENOERROR)
                {     
                    if (flags & TM_TEST_FLAG_NONBLOCKING)
                    {
                        testInfoPtr->testEvents = 0;

                        errorCode = tfRegisterSocketCBParam( sock, 
                                                             tfTestSocketSignal,
                                                             (void TM_FAR *) 
                                                             testInfoPtr,
                                                             TM_CB_ACCEPT );
                    }

/*
 * In both blocking and non-blocking mode, call listen, and if there was no
 * error, move to the CONNECTING phase, where accept will be called.
 */
                    errorCode = listen(sock, TM_TEST_LISTEN_BACKLOG);

                    if (errorCode != TM_ENOERROR)
                    {
                        errorCode = tfGetSocketError(sock);
                    }

                    if ( (errorCode == TM_ENOERROR)
                         || (errorCode == TM_EWOULDBLOCK) )
                    {
                        phase = TM_TEST_PHASE_CONNECTING;
                    }

                }
                else
                {
                    errorCode = tfGetSocketError(sock);
                }

            }
        }
    }


    if (testCount > TM_UL(0))
    {
        testCountDec = 1;
        testForever = 0;
    }
    else
    {
        testCountDec = 0;
        testForever = 1;
    }


    while (    ((testCount > TM_UL(0)) || (testForever)) 
            && (errorCode == TM_ENOERROR) 
            && (phase != TM_TEST_PHASE_INITIAL) )
    {
/*
 * CONNECTING phase
 */
        if (phase == TM_TEST_PHASE_CONNECTING)
        {

/* 
 * In blocking case, just go ahead and call accept - we'll pend until there is
 * a connection available.  In non-blocking mode, only call accept if we
 * know there is a connection available (ie, callback has occurred) 
 */
            if ((testInfoPtr->testEvents & TM_CB_ACCEPT) ||
                                              !(flags & TM_TEST_FLAG_NONBLOCKING) )
            {
                addrLen = sizeof(struct sockaddr_storage);
                dataSock = accept( sock, 
                                   tempSockAddr.sockPtr, 
                                   &addrLen );

                if (dataSock < 0)
                {
                    errorCode = tfGetSocketError(sock);
                }
                else 
                {
/* Set blocking state */
                    if (flags & TM_TEST_FLAG_NONBLOCKING)
                    {
                        (void)tfBlockingState( dataSock, TM_BLOCKING_OFF );
                    }
                }

                if(errorCode == TM_ENOERROR)
                {
                    phase = TM_TEST_PHASE_RECV;
                }
            }
            else
            {
/* Call to 'connect' complete: set data socket and move to the next phase */
                if (testInfoPtr->testEvents & TM_CB_CONNECT_COMPLT)
                {
                    dataSock = sock;

                    phase = TM_TEST_PHASE_SEND;
                }
                else
                {
                    errorCode = TM_EWOULDBLOCK;
                }
            }

        }

/*
 * SEND phase 
 */
        if (phase == TM_TEST_PHASE_SEND)
        {

/* 
 * If we don't currently have a zero copy buffer, allocate one (this could be
 * executed depending on the test being executed - for instance, if this is 
 * a send test, we'll need to allocate a buffer each time, because the previous
 * send 'freed' the buffer.  However, with an echo test, we'll never need to
 * allocate a buffer, because one will be available to us from the receive.
 * The same works with freeing the zero copy buffer below.
 */
            if (    (dataPtr == (char TM_FAR *) 0)
                 && (flags & TM_TEST_FLAG_ZEROCOPY) )
            {
                bufHandle = tfGetZeroCopyBuffer(dataSize, &dataPtr);

                if (dataPtr == (char TM_FAR *) 0)
                {
                    errorCode = TM_ENOBUFS;
                }
            }

            bufPtr = dataPtr;
            if (testType == TM_TEST_TCP_ECHO_SERVER)
            {
                if (echoBytes != 0)
                {
                    receivedSize = echoBytes;
                    if (!(flags & TM_TEST_FLAG_ZEROCOPY))
                    {
                        bufPtr = dataPtr + recvdBytes - echoBytes;
                    }
                }
            }
            else
            {
/* If enabled, fill the send buffer with data */
                if ( flags & TM_TEST_FLAG_FILL_DATA )
                {
                    tfTestMemFill(dataPtr, dataSize);
                }
            }
            while ( (sentBytes < receivedSize) && (errorCode == TM_ENOERROR) )
            {
                if (flags & TM_TEST_FLAG_ZEROCOPY)
                {
                    retCode =
                        tfZeroCopySend(dataSock, bufHandle, receivedSize, 0);
        
                    if (retCode > 0)
                    {
/* ZC buffer was freed by tfZeroCopySend */
                        bufHandle = (ttUserMessage) 0;
                        dataPtr   = (char TM_FAR *) 0;
                    }
                    else
                    {
                        if (retCode < 0)
                        {
                            errorCode = tfGetSocketError(dataSock);
                            if (errorCode != TM_EWOULDBLOCK)
                            {
/* ZC buffer was freed by tfZeroCopySend */
                                bufHandle = (ttUserMessage) 0;
                                dataPtr   = (char TM_FAR *) 0;
                            }
                        }
                    }
        
                }
                else
                {
#ifdef TM_LINT
LINT_NULL_PTR_BEGIN(dataPtr)
#endif /* TM_LINT */
                    retCode =
                        send(dataSock, bufPtr+sentBytes, receivedSize-sentBytes, 0);
                }

                if (retCode < 0)
                {
                    errorCode = tfGetSocketError(dataSock);
                    break;
                }
                else
                {
                    sentBytes += retCode;
                }

            }

            if (errorCode == TM_ENOERROR)
            {
                receivedSize = dataSize;
                sentBytes = 0;    
            }        
        }

/*
 * RECEIVE phase 
 */
        if (phase == TM_TEST_PHASE_RECV)
        {

            if (dataSize == recvdBytes)
            {
                recvdBytes = 0;
            }
            if (testType == TM_TEST_TCP_ECHO_SERVER)
            {
                echoBytes = 0;
            }
/* Loop until we receive all of the data we're expecting. */
            while ( (recvdBytes < dataSize) && (errorCode == TM_ENOERROR) )
            {

                if ( flags & TM_TEST_FLAG_ZEROCOPY )
                {
/* 
 * If we still have a zero copy buffer at this point, free it, since it will
 * be overwritten by the upcoming call to tfZeroCopyRecv.
 */
                    if (dataPtr != (char TM_FAR *) 0)
                    {
                        (void)tfFreeZeroCopyBuffer(bufHandle);
                        bufHandle = (ttUserMessage) 0;
                        dataPtr   = (char TM_FAR *) 0;
                    }

                    retCode = tfZeroCopyRecv( dataSock, &bufHandle, &dataPtr, 
                                              dataSize-recvdBytes, 0 );

                }
                else
                {
                    retCode = recv(
                        dataSock, dataPtr+recvdBytes, dataSize-recvdBytes, 0);
#ifdef TM_LINT
LINT_NULL_PTR_END(dataPtr)
#endif /* TM_LINT */
                }

                if (retCode < 0)
                {
                    errorCode = tfGetSocketError(dataSock);
                    break;
                }
                else
                {
                    recvdBytes += retCode;
                    if (tm_test_tcp_echo_type(testType))
                    {
                        if (retCode == 0)
                        {
/* TCP echo test: end of file */
                            tfClose(dataSock);
                            phase = TM_TEST_PHASE_CLOSE;
                            break;
                        }
                        else if (testType == TM_TEST_TCP_ECHO_SERVER)
                        {
                            echoBytes += retCode;
                            if (recvdBytes < dataSize)
                            {
                                if  (     (flags & TM_TEST_FLAG_ZEROCOPY)
                                       || (tfGetWaitingBytes(dataSock) == 0) )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
/* If enabled, verify that the incoming data is as we expect it to be */
            if (    (flags & TM_TEST_FLAG_VALIDATE)
                 && (recvdBytes == dataSize))
            {
                errorCode = tfTestMemValidate(dataPtr, dataSize);
            }
        }

/* Send or receive sucessful: decrement count and (maybe) move to next phase */
        if (errorCode == TM_ENOERROR)        
        {
            if ( tm_test_tcp_echo_type(testType) )
            {
                phase = tlTestStateTable[phase];
                if (recvdBytes == dataSize)
                {
                    if (    tm_test_tcp_echo_server_type(testType) 
                         && (phase == TM_TEST_PHASE_RECV))
                    {
                        testCount -= (ttUser32Bit)testCountDec;
                    }
                    if (    tm_test_tcp_echo_client_type(testType) 
                         && (phase == TM_TEST_PHASE_SEND))
                    {
                        testCount -= (ttUser32Bit)testCountDec;
                    }
                }
            }
            else
            {
                testCount -= (ttUser32Bit)testCountDec;
            }
        }

/* If in non-blocking mode, exit this loop and return to the user */
        if ( (flags & TM_TEST_FLAG_NONBLOCKING) && 
             (errorCode == TM_ENOERROR) && 
             ((testCount > TM_UL(0)) || testForever) )
        {
            errorCode = TM_EWOULDBLOCK;
        }

    }

/* 
 * If we're actually done with this test (ie, not just returning with an
 * EWOULDBLOCK free up any leftover buffers.
 */
    if ( (dataPtr != (char TM_FAR *) 0) && (errorCode != TM_EWOULDBLOCK) )
    {
        if (flags & TM_TEST_FLAG_ZEROCOPY)
        {
            (void)tfFreeZeroCopyBuffer(bufHandle);
            bufHandle = (ttUserMessage) 0;
            dataPtr   = (char TM_FAR *) 0;
        }
        else
        {
            tm_free_raw_buffer( (ttRawBufferPtr) dataPtr);
            dataPtr = (char TM_FAR *) 0;
        }
    }

/* Save info to the state vector */
    testInfoPtr->testDataPtr    = dataPtr;
    testInfoPtr->testBufHandle  = bufHandle;
    testInfoPtr->testDataSock   = dataSock;
    testInfoPtr->testRecvdBytes = recvdBytes;
    testInfoPtr->testSentBytes  = sentBytes;
    testInfoPtr->testEchoBytes  = echoBytes;
    testInfoPtr->testPhase      = phase;
    testInfoPtr->testCount      = testCount;

testTcpEnd:
    return (errorCode);
}

/*
 * tfTestUserExecute()
 *
 *   Called by the user in non-blocking mode to execute one more 'iteration'
 *   of the test.  This routine basically calls the general UDP or TCP routine
 *   and if this returns something other than TM_EWOULDBLOCK the socket is
 *   closed and the state vector is freed, since the test is now complete.
 */
int tfTestUserExecute(ttUserTestHandle testSessionPtr)
{
    ttTestInfoPtr infoPtr;
    unsigned int  i;
    int           errorCode;

    infoPtr = (ttTestInfoPtr) testSessionPtr;

    errorCode = TM_ENOERROR;

/* UDP tests */
    if ( tm_test_group(infoPtr->testType) == TM_TEST_UDP )
    {
        errorCode = tfTestUdp( infoPtr->testSock,
                               infoPtr->testType,
                               &(infoPtr->testIpAddress),
                               infoPtr->testPortNum,
                               infoPtr->testDataSize, 
                               infoPtr,
                               infoPtr->testFlags );

/* 
 * Test has completed (either sucessfully or unsuccessfully): free the state 
 * structure and close the socket.
 */
        if (errorCode != TM_EWOULDBLOCK)
        {
            (void)tfClose(infoPtr->testSock);

            tm_free_raw_buffer( (ttRawBufferPtr) infoPtr );
        }
    }

/* TCP tests */
    if ( tm_test_group(infoPtr->testType) == TM_TEST_TCP )
    {

/* TCP send, receive and echo tests */
        if (infoPtr->testType != TM_TEST_TCP_CONNECT)
        {
            errorCode = tfTestTcp( infoPtr->testSock,
                                   infoPtr->testType,
                                   &(infoPtr->testIpAddress),
                                   infoPtr->testPortNum,
                                   infoPtr->testDataSize,
                                   infoPtr,
                                   infoPtr->testFlags );
        }
        else
        {
/* TCP connection test */
            errorCode = tfTestTcpConnect( &(infoPtr->testIpAddress),
                                          infoPtr->testPortNum,
                                          infoPtr,
                                          infoPtr->testFlags,
                                          AF_INET);
        }

/* 
 * Test has completed (either sucessfully or unsuccessfully): free the state 
 * structure and close the socket.
 */
        if (errorCode != TM_EWOULDBLOCK)
        {
            (void)tfClose(infoPtr->testSock);

            if (infoPtr->testDataSock != TM_SOCKET_ERROR)
            {
                (void)tfClose(infoPtr->testDataSock);
            }

            tm_free_raw_buffer( (ttRawBufferPtr) infoPtr);

/* Give the FTP socket enough time to linger and close */
            for (i=0; i<tm_ms_to_ticks(60000); i++)
            {
                tfTimerUpdate();
                if ((i%10) == 0)
                {
                    tfTimerExecute();
                }
            }
        }
    }


    return errorCode;
}

/*
 * tfTestRandom()
 *
 */
static void tfTestRandom( int TM_FAR *         testType,
                          int TM_FAR *         dataSize,
                          ttUser32Bit TM_FAR * flags )
{
    tt32Bit     newTestType;
    ttUser32Bit newFlags;

/* Retrieve a random test type: must be a client (send or connect) test */
    do {

        newTestType = tfGetRandom() % (TM_TEST_TCP_CONNECT + TM_UL(1));

    } while ( !(tm_test_send_type(newTestType) ||
                                           tm_test_connect_type(newTestType)) );

    *testType = (int) newTestType;

/* Get random test flag (but don't change non-blocking flag */
    newFlags = *flags & TM_TEST_FLAG_NONBLOCKING;
    newFlags |= (tfGetRandom() % (TM_TEST_VALID_FLAGS)) & 
                                                      ~TM_TEST_FLAG_NONBLOCKING;

    *flags = newFlags;

/* get random data size.  dataSize > 0 AND dataSize <= user's maximum size */
    *dataSize = (int) (tfGetRandom() % (tt32Bit) *dataSize) + 1;

    return;
}

/*
 * tfTestTreck()
 *
 *   The main routine of the test module.  This routine first allocates and
 *   initializes the state structure which will be used to pass (and possibly
 *   store) information for this test session.
 *   If this test is a UDP or a TCP test, an appropriate socket is created,
 *   blocking mode set, and various options are set for this socket.  Then
 *   the general TCP or UDP test routine is called.  If this routine returns
 *   with a TM_EWOULDBLOCK message, info is saved in the state vector.  The
 *   routine then returns to the user.
 */
int tfTestTreck( int                 testType,
                 char TM_FAR *       ipAddrStr,
                 int                 portNumber,
                 int                 dataSize,
                 ttUser32Bit         testCount,
                 ttUser32Bit         flags,
                 ttUserTestHandlePtr testHandlePtr,
                 void TM_FAR *       testParam )
{

    int                         errorCode;
    int                         sock;
    int                         optParam;
    unsigned int                i;
    ttTestInfoPtr               infoPtr;
    tt8Bit                      addrFamily;
    struct sockaddr_storage     ipNgAddr;

/* Avoid compiler warnings */
    TM_UNREF_IN_ARG(testParam);

    errorCode = TM_ENOERROR;

    tm_memset(&ipNgAddr, 0, sizeof(struct sockaddr_storage));
    ipNgAddr.ss_len = sizeof(struct sockaddr_storage);

    addrFamily = AF_INET; /* assume IPv4 format */
            errorCode = inet_pton(
                AF_INET,
                ipAddrStr,
                (void*) &ipNgAddr.addr.ipv4.sin_addr);

    if (errorCode != 1)
    {
/* inet_pton failed, address string must be invalid */
        errorCode = TM_EINVAL;
        goto testTreckEnd;
    }
    ipNgAddr.ss_family = addrFamily;


/* If enabled, pick random values for the test type, data size and flags */
    if ( flags & TM_TEST_FLAG_RANDOM )
    {
        tfTestRandom( &testType, &dataSize, &flags );

        if (tm_test_echo_type(testType))
        {
            portNumber = TM_STD_ECHO_PORT;
        }
        else
        {
            portNumber = TM_STD_DISCARD_PORT;
        }

    }
/*
 * Some flag/test combinations are invalid.  We can either return an error to
 * the user, or simply fix them.  Currently, we simply correct the problem.
 */
#ifndef TM_AUTO_TEST
/* 
 * Can only validate data on a TCP or UDP echo client,
 * except in automated test code where we recv our own data.
 */
    if (flags & (TM_TEST_FLAG_VALIDATE))
    {
        if (!tm_test_echo_client_type(testType))
        {
/* Non automated test code: Do not validate data if not on an Echo client. */
            flags &= ~TM_TEST_FLAG_VALIDATE;
        }
    }
#endif /* TM_AUTO_TEST */

/* 
 * Using a zero copy receive for the TCP echo or TCP recv is not possible if
 * we want to validate the data, since the data may be received in multiple
 * blocks.
 */
    if (flags & (TM_TEST_FLAG_ZEROCOPY | TM_TEST_FLAG_VALIDATE))
    {
        if (    (testType == TM_TEST_TCP_ECHO_CLIENT)
             || (testType == TM_TEST_TCP_RECV)
             || (testType == TM_TEST_TCP_ECHO_SERVER) )
        {
            flags &= ~TM_TEST_FLAG_ZEROCOPY;
        }
    }

/* UDP connect may not be used with UDP server tests */
    if (    (tm_test_udp_recv_type(testType)) 
         && (flags & TM_TEST_FLAG_UDP_CONNECT) )
    {
        flags &= ~TM_TEST_FLAG_UDP_CONNECT;
    }

/* 
 * If validation is enabled for an echo client test, we also have to fill
 * the buffer.
 */
    if (    (tm_test_echo_client_type(testType))
         && (flags & TM_TEST_FLAG_VALIDATE) )
    {
        flags |= TM_TEST_FLAG_FILL_DATA;
    }

/* 
 * If non-blocking mode, testHandlePtr must contain a pointer to the state 
 * vector */
    if ( (flags & TM_TEST_FLAG_NONBLOCKING)
         && (testHandlePtr == (void TM_FAR *) 0) )
    {
        errorCode = TM_EINVAL;
        goto testTreckEnd;
    }

/* Allocate and initialize the state structure for this test session */
    infoPtr = (ttTestInfoPtr) tm_get_raw_buffer((ttPktLen)sizeof(ttTestInfo));

    if (infoPtr != (ttTestInfoPtr)0)
    {
        tm_bzero(infoPtr, sizeof(*infoPtr));
        infoPtr->testDataSize   = dataSize;
        infoPtr->testCount      = testCount;
        infoPtr->testDataSock   = TM_SOCKET_ERROR;
/*
 * UDP test: send, receive, echo client, echo server
 */
        if ( tm_test_group(testType) == TM_TEST_UDP )
        {

            sock = socket( addrFamily, SOCK_DGRAM, 0);

/* Set blocking state */
            if (flags & TM_TEST_FLAG_NONBLOCKING)
            {
                (void)tfBlockingState( sock, TM_BLOCKING_OFF );
            }

/* Turn UDP checksum on or off, according to flags from user */
            (void)tfSetTreckOptions( TM_OPTION_UDP_CHECKSUM, 
                               (ttUser32Bit) !(flags & TM_TEST_FLAG_UDP_CS_OFF) );

            errorCode = tfTestUdp( sock,
                                   testType,
                                   &ipNgAddr,
                                   portNumber,
                                   dataSize, 
                                   infoPtr, 
                                   flags );
            if (errorCode == TM_EWOULDBLOCK)
            {
/* 
 * Test returned with a EWOULDBLOCK, which means we're in blocking mode, and
 * should save our info, and return to user.
 */
                infoPtr->testFlags     = flags;
                infoPtr->testSock      = sock;
                infoPtr->testType      = testType;                    
                infoPtr->testDataSize  = dataSize;

                if (testHandlePtr != (void TM_FAR *) 0)
                {
                    *((struct tsTestInfo TM_FAR * TM_FAR *)testHandlePtr)
                        = infoPtr;
                }
            }
            else
            {
/* 
 * Test has completed (either sucessfully or unsuccessfully): free the state 
 * structure and close the socket.
 */
                tm_free_raw_buffer((ttRawBufferPtr)infoPtr);

                (void)tfClose(sock);
            }

        }

/*
 * TCP test: send, receive, echo client, echo server, connect
 */
        if ( tm_test_group(testType) == TM_TEST_TCP )
        {

            sock = socket( addrFamily, SOCK_STREAM, 0);

            if (flags & TM_TEST_FLAG_NONBLOCKING)
            {
                (void)tfBlockingState( sock, TM_BLOCKING_OFF );
            }

            if (flags & TM_TEST_FLAG_TCP_NODELAY)
            {
                optParam = 1;
                (void)setsockopt( sock, IP_PROTOTCP, TCP_NODELAY,
                            (char TM_FAR *) &optParam, sizeof(optParam) );
            }
            

/* TCP send, receive, echo client, echo server tests */
            if (testType != TM_TEST_TCP_CONNECT)
            {

                errorCode = tfTestTcp( sock,
                                       testType,
                                       &ipNgAddr,
                                       portNumber,
                                       dataSize, 
                                       infoPtr, 
                                       flags );
            }
            else
            {
/* TCP connect test */
                infoPtr->testSock      = sock;
                tm_bcopy(
                    &ipNgAddr,
                    &(infoPtr->testIpAddress),
                    sizeof(struct sockaddr_storage));
                infoPtr->testPortNum   = portNumber;

                errorCode = tfTestTcpConnect( &ipNgAddr,
                                              portNumber,
                                              infoPtr,
                                              flags,
                                              addrFamily);
            }


            if (errorCode == TM_EWOULDBLOCK)
            {
/* 
 * Test returned with a EWOULDBLOCK, which means we're in blocking more, and
 * should save info, and return to user.
 */
                infoPtr->testFlags     = flags;
                infoPtr->testSock      = sock;
                infoPtr->testType      = testType;                    
                infoPtr->testDataSize  = dataSize;

                if (testHandlePtr != (void TM_FAR *) 0)
                {
                    *((struct tsTestInfo TM_FAR * TM_FAR *)testHandlePtr)
                        = infoPtr;
                }
            }
            else
            {
/* 
 * Test has completed (either sucessfully or unsuccessfully): free the state
 * structure and free the sockets (both the listening and accepted socket if
 * necessary).
 */
                (void)tfClose(sock);

                if (infoPtr->testDataSock != TM_SOCKET_ERROR)
                {
                    (void)tfClose(infoPtr->testDataSock);
                }

                tm_free_raw_buffer((ttRawBufferPtr)infoPtr);

/* Give the FTP socket enough time to linger and close */
                for (i=0; i<tm_ms_to_ticks(60000); i++)
                {
                    tfTimerUpdate();
                    if ((i%10) == 0)
                    {
                        tfTimerExecute();
                    }
                }
            }
        }

        if ( tm_test_group(testType) == TM_TEST_MISC )
        {

            switch (testType)
            {
                case TM_TEST_LOCK:
                    errorCode = TM_EINVAL;
#ifdef TM_ERROR_CHECKING
#ifdef TM_LOCK_NEEDED
                    while (testCount--)
                    {
                        tfLockCheck();
                        errorCode = TM_ENOERROR;
                    }
#endif /* TM_LOCK_NEEDED */
#endif /* TM_ERROR_CHECKING */
                    break;

                default:
                    errorCode = TM_EINVAL;
            }

        }
    }
    else
    {
        errorCode = TM_ENOBUFS;
    }

testTreckEnd:
    return (errorCode);
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:53      For: Xilinx-Samsung
 * File Name: trtest.c      File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 34531    Output checksum:  3650
 */
