/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Transparent Ethernet PCI bus Link Layer
 *
 * Filename: trtransp.c
 * Author: Odile
 * Date Created: 9/08/99
 * $Source: source/trtransp.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:48:58CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */


/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>


#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>

/* Transparent link layer not yet supported by TI C3 or C5 DSP */
/*
 * local functions
 */
int tfTranspClose(void TM_FAR * interfaceId);
int tfTranspSend(ttPacketPtr packetPtr);
int tfTranspRecv(void TM_FAR *interfaceId, void TM_FAR *bufferHandle);
int tfTranspIoctl(int operation);
int tfTranspError(int type);

ttUserLinkLayer tfUseTransp (void)
{
    ttLinkLayerEntryPtr linkLayerEntryPtr;

/* Search for Loopback Protocol in Link Layer List */
/* Lock the Link Layer list */
    tm_call_lock_wait(&tm_context(tvLinkLayerListLock));
    linkLayerEntryPtr = tm_context(tvLinkLayerListPtr);
/* Check for duplicate */
    while (linkLayerEntryPtr != TM_LINK_NULL_PTR)
    {
        if (    tm_strcmp( (char TM_FAR *)(linkLayerEntryPtr->lnkNameArray),
                           "TRANSP" )
             == TM_STRCMP_EQUAL )
        {
            goto useTranspUnLock; /* entry already in */
        }
        linkLayerEntryPtr = linkLayerEntryPtr->lnkNextPtr;
    }
/* if none exists then malloc an entry */
    linkLayerEntryPtr = (ttLinkLayerEntryPtr)
                         tm_kernel_malloc((unsigned)sizeof(ttLinkLayerEntry));
    if (linkLayerEntryPtr == TM_LINK_NULL_PTR)
    {
/* if the malloc fails then return a null pointer */
        goto useTranspUnLock;
    }
    tm_bzero(linkLayerEntryPtr, sizeof(ttLinkLayerEntry));
    linkLayerEntryPtr->lnkMtu = 1500;
    linkLayerEntryPtr->lnkOpenFuncPtr = tfLinkOpen;
    linkLayerEntryPtr->lnkCloseFuncPtr = tfTranspClose;
    linkLayerEntryPtr->lnkSendFuncPtr = tfTranspSend;
    linkLayerEntryPtr->lnkRecvFuncPtr = tfTranspRecv;
    linkLayerEntryPtr->lnkIoctlFuncPtr = tfTranspIoctl;
    linkLayerEntryPtr->lnkErrorFuncPtr = tfTranspError;
    linkLayerEntryPtr->lnkPostFuncPtr = tfDeviceSendOnePacket;
    linkLayerEntryPtr->lnkLinkLayerProtocol = TM_LINK_LAYER_TRANSP;
    tm_strcpy((char TM_FAR *)(linkLayerEntryPtr->lnkNameArray), "TRANSP");
/* Stuff it into the table */
    linkLayerEntryPtr->lnkNextPtr = tm_context(tvLinkLayerListPtr);
    tm_context(tvLinkLayerListPtr) = linkLayerEntryPtr;

useTranspUnLock:
    tm_call_unlock(&tm_context(tvLinkLayerListLock));
    return ((ttUserLinkLayer)linkLayerEntryPtr);
}

/*
 * Close a Transp Link Layer
 */
int tfTranspClose ( void TM_FAR *interfaceId )
{
    TM_UNREF_IN_ARG(interfaceId);
    return TM_LL_OKAY;
}

int tfTranspSend ( ttPacketPtr packetPtr )
{
    ttEtherHeaderPtr etherHeaderPtr;
    int              padding;
    int              retCode;
    
    retCode = TM_LL_OKAY;
    
/* Put the ethernet header in front of the packet */
    packetPtr->pktLinkDataLength += TM_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength += TM_ETHER_HEADER_SIZE;
    packetPtr->pktLinkDataPtr -= TM_ETHER_HEADER_SIZE;
/* Min size met? */
    if (packetPtr->pktChainDataLength < TM_ETHER_MIN_PACKET)
    {
/* 
 * We need to pad with zeroes
 */
        padding =  TM_ETHER_MIN_PACKET - (int)packetPtr->pktChainDataLength;
        if (    (packetPtr->pktChainDataLength == packetPtr->pktLinkDataLength)
             && ( (  packetPtr->pktSharedDataPtr->dataBufLastPtr
                       - packetPtr->pktLinkDataPtr) >= TM_ETHER_MIN_PACKET )
           )
/* 
 * Contiguous buffer (i.e. non TCP packet, or TCP packet with data copied at
 * the end of the header), and allocated packet has room to add the padding.
 */
        {
            tm_bzero(   packetPtr->pktLinkDataPtr
                      + packetPtr->pktLinkDataLength,
                      padding);
            packetPtr->pktLinkDataLength = TM_ETHER_MIN_PACKET;
            packetPtr->pktChainDataLength = TM_ETHER_MIN_PACKET;
        }
        else
        {
/*
 * Scattered buffer, or allocated packet too short for padding:
 *  add an extra link for the padding initialized with zeroes.
 * Note that if the device cannot handle the extra link,
 * tfDeviceSend() will copy to a new buffer. No need to worry about copying
 * here.
 */
            retCode = tfPacketTailAdd(packetPtr, padding);
        }
    }
    if (!(packetPtr->pktSharedDataPtr->dataFlags & TM_BUF_HDR_RECYCLED))
    {
        etherHeaderPtr = (ttEtherHeaderPtr)(packetPtr->pktLinkDataPtr);
        
/* Copy the device address into the packet */
        tm_ether_copy( (packetPtr->pktDeviceEntryPtr->devPhysAddrArr),
                       (etherHeaderPtr->ethSrcAddr));
/* Copy the destination address into the packet */
        tm_ether_copy( (packetPtr->pktEthernetAddress),
                       (etherHeaderPtr->ethDstAddr));
/* Set the packet type */
        if (packetPtr->pktNetworkLayer == TM_NETWORK_IPV4_LAYER)
        {
            etherHeaderPtr->ethType = TM_ETHER_IP_TYPE;
        }
        else
        {
            if (packetPtr->pktNetworkLayer == TM_NETWORK_ARP_LAYER)
            {
                etherHeaderPtr->ethType = TM_ETHER_ARP_TYPE;
            }
            else
            {
/* Should never happen */
                retCode = TM_EPROTONOSUPPORT;
            }
        }
    }
    return (retCode);
}

/* 
 * Unstuff and process an incoming packet
 */
int tfTranspRecv ( void TM_FAR *interfaceId, void TM_FAR *bufferHandle )
{
    ttEtherHeaderPtr etherHeaderPtr;
    tt8BitPtr        pktEtherAddressPtr;
    tt8BitPtr        etherHeaderAddressPtr;
    ttArpHeaderPtr   arpHeaderPtr;
    tt8BitPtr        sendAddressPtr;
    tt8BitPtr        targetAddressPtr;
    ttPacketPtr      packetPtr;
    ttDeviceEntryPtr devEntryPtr;
    int              flags;
    int              retCode;
    int              i;
    tt8Bit           tempByte;

    retCode = TM_ENOERROR;
    packetPtr = (ttPacketPtr)bufferHandle;
    devEntryPtr = (ttDeviceEntryPtr)interfaceId;
#ifdef TM_USE_DRV_SCAT_RECV
/* If Ethernet header is not contiguous, make it contiguous */
    if (tm_pkt_hdr_is_not_cont(packetPtr, TM_ETHER_HEADER_SIZE))
    {
        packetPtr = tfContiguousSharedBuffer(packetPtr, TM_ETHER_HEADER_SIZE);
        if (packetPtr == (ttPacketPtr)0)
        {
            retCode = TM_ENOBUFS;
            goto transpRecvExit;
        }
    }
#endif /* TM_USE_DRV_SCAT_RECV */
/* Check the type field */
    etherHeaderPtr = (ttEtherHeaderPtr)(packetPtr->pktLinkDataPtr);
    packetPtr->pktLinkDataPtr += TM_ETHER_HEADER_SIZE;
    packetPtr->pktLinkDataLength -= TM_ETHER_HEADER_SIZE;
    packetPtr->pktChainDataLength -= TM_ETHER_HEADER_SIZE;
/* Store the device the packet came from */
    packetPtr->pktDeviceEntryPtr = devEntryPtr;
/* Copy the source ethernet address into the packet */
    pktEtherAddressPtr = (tt8BitPtr)packetPtr->pktEthernetAddress;
    etherHeaderAddressPtr = (tt8BitPtr)etherHeaderPtr->ethSrcAddr;
    tm_ether_copy(etherHeaderAddressPtr, pktEtherAddressPtr);
/*
 * Look at the ethernet destination address to check for destination
 * Ethernet broadcast/multicast.
 */
    etherHeaderAddressPtr = (tt8BitPtr)etherHeaderPtr->ethDstAddr;
    if (tm_ether_broadcast(etherHeaderAddressPtr))
    {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
        if (etherHeaderPtr->ethType == TM_ETHER_IP_TYPE)
        {
            tm_64Bit_incr(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInBcastPkts);
            tm_64Bit_incr(tm_context(tvDevIpv4Data).ipIfStatsHCInBcastPkts);
        }
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
        flags = TM_LL_BROADCAST_FLAG;

    }
    else
    {
        if (tm_ether_mcast(etherHeaderAddressPtr))
        {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            if (etherHeaderPtr->ethType == TM_ETHER_IP_TYPE)
            {
                tm_64Bit_incr(devEntryPtr->dev4Ipv4Mib.ipIfStatsHCInMcastPkts);
                tm_64Bit_incr(tm_context(tvDevIpv4Data).ipIfStatsHCInMcastPkts);
            }
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            flags = TM_LL_MULTICAST_FLAG;
        }
        else
        {
#if (defined(TM_SNMP_MIB) || defined(TM_PPP_LQM))
            devEntryPtr->devInUcastPkts++;
#endif /* TM_SNMP_MIB or TM_PPP_LQM */
            flags = TM_LL_UNICAST_FLAG;
        }
    }
    switch (etherHeaderPtr->ethType)
    {
        case TM_ETHER_ARP_TYPE:
            packetPtr->pktFlags = 0;
            packetPtr->pktNetworkLayer =
                TM_NETWORK_ARP_LAYER;
/* Reply with our fake address */
    
            arpHeaderPtr = (ttArpHeaderPtr)(packetPtr->pktLinkDataPtr);
            if ( arpHeaderPtr->arpOperation == TM_ARP_REQUEST )
            {

                arpHeaderPtr->arpOperation = TM_ARP_REPLY;
                sendAddressPtr =
                               (tt8BitPtr)(arpHeaderPtr->arpSendPhyAddress);
                targetAddressPtr =
                             (tt8BitPtr)(arpHeaderPtr->arpTargetPhyAddress);
/* Swap the sender and receiver  (FASTER THAN BCOPYS) */
                for ( i = 0;
                      i < (TM_ETHER_ADDRESS_LENGTH + TM_IP_ADDRESS_LENGTH);
                      i++)
                {
                    tempByte = *sendAddressPtr;
                    *sendAddressPtr = *targetAddressPtr;
                    *targetAddressPtr = tempByte;
                    sendAddressPtr++;
                    targetAddressPtr++;
                }
            /* Now plug in our physical address */
                tm_ether_copy(devEntryPtr->devPhysAddrArr,
                              arpHeaderPtr->arpSendPhyAddress);
#ifdef TM_SNMP_MIB
                packetPtr->pktFlags &= ~(   TM_LL_BROADCAST_FLAG
                                          | TM_LL_MULTICAST_FLAG);
#endif /* TM_SNMP_MIB */
                retCode = tfDeviceSend(packetPtr);
            }
            else
            {
/* Ignore ARP replies. There should not be any */
                tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
                retCode = TM_EPROTONOSUPPORT;
            }
            break;
        case TM_ETHER_IP_TYPE:
            packetPtr->pktFlags = (tt16Bit) flags;
            packetPtr->pktNetworkLayer =
                TM_NETWORK_IPV4_LAYER;
            tfIpIncomingPacket(packetPtr
                               );
            break;
        default:
/* Packet Not For TRECK */
#ifdef TM_SNMP_MIB
            devEntryPtr->dev4Ipv4Mib.ipIfStatsInUnknownProtos++;
#endif /* TM_SNMP_MIB */
            tfFreePacket(packetPtr, TM_SOCKET_UNLOCKED);
            retCode = TM_EPROTONOSUPPORT;
            break;
    }
#ifdef TM_USE_DRV_SCAT_RECV
transpRecvExit:
#endif /* TM_USE_DRV_SCAT_RECV */
    return(retCode);
}

/*
 * Stub functions
 */

int tfTranspIoctl ( int operation )
{
    TM_UNREF_IN_ARG(operation);
    return TM_LL_OKAY;
}

int tfTranspError ( int type )
{
    TM_UNREF_IN_ARG(type);
    return TM_LL_OKAY;
}


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:53      For: Xilinx-Samsung
 * File Name: trtransp.c    File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum:  2423    Output checksum: 33705
 */
