/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2011
 * Copyright Elmic Systems Japan 1997 - 2011
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Automated test code for IPv6 DNS resolver
 *
 * Filename: tudnsr.c
 * Author: Jason Schmidlapp
 * Date Created: 1/17/2003
 * $Source: test/unittest/tudnsres.c $
 *
 * Modification History
 * $Revision: 6.0.2.4 $
 * $Date: 2011/11/09 23:55:37CST $
 * $Author: lfox $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */



/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>

#ifdef TM_MULTIPLE_CONTEXT

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
/* for the automated testing */
#include "tutest.h"



/* constant declarations and enumerations */

/* type definitions */

/* DNS Request/Response Header */
typedef struct tsDnsHeader
{

/* Identifier used to match queries and replies */
    tt16Bit dnsId;

/*
 * Various flags: query/response, opcode, authoritative answer, truncation,
 * recursion desired/available, response code.
 */
    tt16Bit dnsFlags;

/* Number of query records */
    tt16Bit dnsQuestionCount;

/* Number of answer records */
    tt16Bit dnsAnswerCount;

/* Number of name server records */
    tt16Bit dnsNameServerCount;

/* Number of 'additional' type records */
    tt16Bit dnsAdditionalCount;

} ttDnsHeader;

typedef ttDnsHeader TM_FAR * ttDnsHeaderPtr;

#define TM_DNS_HEADER_SIZE 12

/*
 * local function prototypes
 * */
/* Test functions */
static void tfxDnsTestNumericFlag(void);
static void tfxDnsTestReverseLookup(int family);
static void tfxDnsCleanUp(void);
static void tfxDnsTestMultiServers(void);
static void tfxDnsTestQueryType(int family);
static void tfxDnsTestStrError(void);
static void tfxDnsTestQueryResponse(int family);
static void tfxDnsTestQueryDual(void);

static void tfxDnsOldTestQueryType(void);
static void tfxDnsOldTestQueryResponse(void);
static void tfxDnsOldTestReverseLookup(void);

#if (defined(TM_USE_RFC3484) && defined(TM_USE_STOP_TRECK))
static void tfxDnsTestRfc3484(void);
#define TM_DNS_TEST_ADDR1_TEMPORARY     0x1
#define TM_DNS_TEST_ADDR2_TEMPORARY     0x2
#define TM_DNS_TEST_ADDR1_DEPRECATED    0x4
#define TM_DNS_TEST_ADDR2_DEPRECATED    0x8
#define TM_DNS_TEST_ADDR1_MIP_HOME      0x10
#define TM_DNS_TEST_ADDR2_MIP_CAREOF    0x20

static void tfxDnsTestRfc3484RunTest(ttUserInterface     interfaceHandle,
                                     ttCharPtrPtr        sourceIpStringPtrPtr,
                                     tt8Bit              sourceIpStringCount,
                                     ttCharPtr           destIpStringPtr,
                                     ttCharPtr           expectedIpStringPtr,
                                     int                 flags);
#endif /* TM_USE_RFC3484 && TM_USE_STOP_TRECK */

/* Utility functions */
static int tfxAddResponseRec(tt8BitPtr     dataPtr,
                             tt16Bit       recType,
                             tt32Bit       ttl,
                             tt8BitPtr addrStrPtr,
                             tt8BitPtr rrNamePtr);
static void tfxDnsRunTimer(void);
static ttUserMessage tfxDnsCreateReply( tt8BitPtr    requestBuf,
                                        tt16Bit      queryType,
                                        tt8BitPtr    ipStr1,
                                        tt8BitPtr    ipStr2,
                                        tt8BitPtr    canonStr,
                                        tt16Bit      dnsFlags,
                                        int TM_FAR * dataLengthPtr);

/* local variable definitions */
static int tlServerSd;

/* static const tables */
const char ipv4TestAddr[] = "10.1.1.20";
const char ipv4TestAddr2[] = "192.168.1.100";

/* Any letters need to be in caps, since this is the format that inet_ntop
   uses. */
const char ipv6TestAddr[] = "3FFE:10:1:2:A:B:C:D";
const char ipv6TestAddr2[] = "FE80::10:20:30:40";

const char testHostnameOne[] = "www.elmic.com";
const char testHostnameOneRec[] = "\3www\5elmic\3com";

/* Make sure that this is at least as long as testHostnameOne */
const char testCanonName[] = "server.elmic.com";
const char testCanonRecName[] = "\6server\5elmic\3com";

/* MX Response packet */
/*
 * Response is formatted like so:
 * Transaction ID: 0x0000 (Placeholder to be populated by test)
 * Flags: 0x8500 (Response, Authoritative, Recursive)
 * Questions: 1
 * Answer RRs: 4
 * Authority RRs: 3
 * Additional RRs: 4
 * Queries: example.com: type MX, class IN
 * Answers: example.com: type MX, class IN, pref 10, mx mxserver10.example.com
 *          example.com: type MX, class IN, pref 10, mx mx6server10.example.com
 *          example.com: type MX, class IN, pref 10, mx mxserver15.example.com
 *          example.com: type MX, class IN, pref 10, mx mx6server15.example.com
 * Authoritative nameservers:
 *          example.com: type NS, class IN, ns server.example.com
 *          example.com: type NS, class IN, ns ns1.ip-provider.net
 *          example.com: type NS, class IN, ns ns2.ip-provider.net
 * Additional records:
 *          mxserver10.example.com: type A, class IN, addr 172.30.79.10
 *          mx6server10.example.com: type AAAA, class IN, addr 3ffe::10
 *          mxserver15.example.com: type A, class IN, addr 172.30.79.15
 *          mx6server15.example.com: type AAAA, class IN, addr 3ffe::15
 */
tt8Bit mxResponse[] =
{
    0x00, 0x00, 0x85, 0x00, 0x00, 0x01, 0x00, 0x04, 0x00, 0x03, 0x00, 0x04,
    0x07, 0x65, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x03, 0x63, 0x6f, 0x6d,
    0x00, 0x00, 0x0f, 0x00, 0x01, 0xc0, 0x0c, 0x00, 0x0f, 0x00, 0x01, 0x00,
    0x01, 0x51, 0x80, 0x00, 0x0f, 0x00, 0x0a, 0x0a, 0x6d, 0x78, 0x73, 0x65,
    0x72, 0x76, 0x65, 0x72, 0x31, 0x30, 0xc0, 0x0c, 0xc0, 0x0c, 0x00, 0x0f,
    0x00, 0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x10, 0x00, 0x0a, 0x0b, 0x6d,
    0x78, 0x36, 0x73, 0x65, 0x72, 0x76, 0x65, 0x72, 0x31, 0x30, 0xc0, 0x0c,
    0xc0, 0x0c, 0x00, 0x0f, 0x00, 0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x0f,
    0x00, 0x0f, 0x0a, 0x6d, 0x78, 0x73, 0x65, 0x72, 0x76, 0x65, 0x72, 0x31,
    0x35, 0xc0, 0x0c, 0xc0, 0x0c, 0x00, 0x0f, 0x00, 0x01, 0x00, 0x01, 0x51,
    0x80, 0x00, 0x10, 0x00, 0x0f, 0x0b, 0x6d, 0x78, 0x36, 0x73, 0x65, 0x72,
    0x76, 0x65, 0x72, 0x31, 0x35, 0xc0, 0x0c, 0xc0, 0x0c, 0x00, 0x02, 0x00,
    0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x09, 0x06, 0x73, 0x65, 0x72, 0x76,
    0x65, 0x72, 0xc0, 0x0c, 0xc0, 0x0c, 0x00, 0x02, 0x00, 0x01, 0x00, 0x01,
    0x51, 0x80, 0x00, 0x15, 0x03, 0x6e, 0x73, 0x31, 0x0b, 0x69, 0x70, 0x2d,
    0x70, 0x72, 0x6f, 0x76, 0x69, 0x64, 0x65, 0x72, 0x03, 0x6e, 0x65, 0x74,
    0x00, 0xc0, 0x0c, 0x00, 0x02, 0x00, 0x01, 0x00, 0x01, 0x51, 0x80, 0x00,
    0x06, 0x03, 0x6e, 0x73, 0x32, 0xc0, 0xb0, 0xc0, 0x2b, 0x00, 0x01, 0x00,
    0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x04, 0xac, 0x1e, 0x4f, 0x0a, 0xc0,
    0x46, 0x00, 0x1c, 0x00, 0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x10, 0x3f,
    0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x10, 0xc0, 0x62, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x51,
    0x80, 0x00, 0x04, 0xac, 0x1e, 0x4f, 0x0f, 0xc0, 0x7d, 0x00, 0x1c, 0x00,
    0x01, 0x00, 0x01, 0x51, 0x80, 0x00, 0x10, 0x3f, 0xfe, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15
};

/* macro definitions */
#define TM_HOSTNAME_BUF_LEN 128
#define TM_INPUT_BUF_LEN 128
#define TM_DNS_TEST_TIMEOUT_SECS 1

#define TM_DNS_QTYPE_A     1
#define TM_DNS_QTYPE_CNAME 5
#define TM_DNS_QTYPE_REV   12
#define TM_DNS_QTYPE_MX    15
#define TM_DNS_QTYPE_AAAA  28

#define TM_DNS_HDR_RESP            0x8000
#define TM_DNS_HDR_TRUNCATION      0x0200

#define TM_DNS_RR_HDR_LEN 10

#define TM_TEST_TTL_LONG  600
#define TM_TEST_TTL_SHORT 60

static void tfxDnsRunTimer(void)
{
    int timeoutCount;

    timeoutCount =
        (TM_DNS_TEST_TIMEOUT_SECS * TM_UL(1000)) / tvTimerTickLength;

    while (timeoutCount-- >= 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

}

void tfxTestDnsResolver(void)
{
    int errorCode;
    int addrFamily;
    struct sockaddr_storage bindAddr;

/* Switch context to the client to initiliaze the DNS resolver */
    tfSetCurrentContext(tvContextHandleClient);
    errorCode = tfDnsInit(TM_BLOCKING_OFF);

    if (errorCode == TM_EPROTONOSUPPORT)
    {
        printf("DNS resolver is not available.\n");
        goto testDnsExit;
    }

    tm_assert(tfxTestDnsResolver,
              (errorCode == TM_ENOERROR) || (errorCode == TM_EALREADY) );

/* minimize the retries since we know we're either going to get and
   immediate response, or we're not going to get an answer at all. */
    errorCode = tfDnsSetOption(TM_DNS_OPTION_RETRIES, 0);
    tm_assert(tfxTestDnsResolver, errorCode == TM_ENOERROR);

    errorCode = tfDnsSetOption(TM_DNS_OPTION_TIMEOUT,
                               TM_DNS_TEST_TIMEOUT_SECS);
    tm_assert(tfxTestDnsResolver, errorCode == TM_ENOERROR);

/*
 * Setup the DNS 'server'.  This is really just a UDP socket listening on
 * port 53, which will receive packets, verify the contents and possibly
 * send a simple response.
 */
    tfSetCurrentContext(tvContextHandleServer);

    addrFamily = AF_INET;

    tlServerSd = socket(addrFamily, SOCK_DGRAM, 0);
    tm_assert(tfxTestDnsResolver, tlServerSd >= 0);

    errorCode = tfBlockingState(tlServerSd, TM_BLOCKING_OFF);
    tm_assert(tfxTestDnsResolver, errorCode == TM_ENOERROR);

    tm_bzero(&bindAddr, sizeof(bindAddr));
    bindAddr.ss_len = sizeof(struct sockaddr_storage);
    bindAddr.ss_family = (tt8Bit) addrFamily;
#ifdef TM_TEST_WILD_PORT
    bindAddr.ss_port = htons(TM_WILD_PORT);
#else /* ! TM_TEST_WILD_PORT */
    bindAddr.ss_port = htons(53);
#endif /* ! TM_TEST_WILD_PORT */

    errorCode =
        bind(tlServerSd, (struct sockaddr TM_FAR *) &bindAddr,
             bindAddr.ss_len);
    tm_assert(tfxTestDnsResolver, errorCode == TM_ENOERROR);


/* The server is set up, so run the tests. */

/* Test the AI_NUMERICFLAG (getaddrinfo) and NI_NUMERICFLAG (getnameinfo) */
    tm_test(tfxDnsTestNumericFlag);


/* Test that the proper DNS queries are sent, for IPv4 */
    tm_test_with_1_param(tfxDnsTestQueryType, AF_INET);
/* Test that DNS queries are processed correctly, for IPv4 */
#ifdef TM_4_USE_SCOPE_ID
    tm_test_with_1_param(tfxDnsTestQueryResponse, AF_INET6);
#else /* TM_4_USE_SCOPE_ID */
    tm_test_with_1_param(tfxDnsTestQueryResponse, AF_INET);
#endif /* TM_4_USE_SCOPE_ID */
/* Test reverse (PTR) queries for IPv4 */
#ifdef TM_4_USE_SCOPE_ID
    tm_test_with_1_param(tfxDnsTestReverseLookup,AF_INET6);
#else /* TM_4_USE_SCOPE_ID */
    tm_test_with_1_param(tfxDnsTestReverseLookup,AF_INET);
#endif /* TM_4_USE_SCOPE_ID */

    tm_test(tfxDnsTestMultiServers);


/* Test 'old' DNS APIs for IPv4 */
#if defined(TM_USE_IPV4) && !defined(TM_4_USE_SCOPE_ID)
    tm_test(tfxDnsOldTestQueryType);
    tm_test(tfxDnsOldTestQueryResponse);
    tm_test(tfxDnsOldTestReverseLookup);
#endif /* TM_USE_IPV4 and not TM_4_USE_SCOPE_ID */

#ifdef TM_TEST_DNS_VERBOSE
    tm_test(tfxDnsTestStrError);
#endif /* TM_TEST_DNS_VERBOSE*/

    tfSetCurrentContext(tvContextHandleServer);
    tfClose(tlServerSd);

    tfSetCurrentContext(tvContextHandleClient);
    tfDnsCacheInvalidate();

    tfxDnsCleanUp();

#if (defined(TM_USE_RFC3484) && defined(TM_USE_STOP_TRECK))
    tfxDnsTestRfc3484();
#endif /* TM_USE_RFC3484 && TM_USE_STOP_TRECK */

testDnsExit:
    return;
}

static void tfxDnsCleanUp(void)
{
    ttDnsCacheEntryPtr  curCacheEntryPtr;
    ttDnsCacheEntryPtr  nextCacheEntryPtr;
#ifdef TM_DNS_USE_SRV
    ttAddrInfoExtPtr    curAddrInfo;
#endif /* TM_DNS_USE_SRV */

/* Lock the DNS global lock */
    tm_call_lock_wait(&tm_context(tvDnsLockEntry));
    curCacheEntryPtr = tm_context(tvDnsCachePtr);
    if (curCacheEntryPtr != (ttDnsCacheEntryPtr)0)
    {
        do
        {
            nextCacheEntryPtr = curCacheEntryPtr->dnscNextEntryPtr;
/* 5.  Deallocate memory associated with this entry: */
/* 5.1.    Free timer. */
            if (curCacheEntryPtr->dnscTimerPtr != (ttTimerPtr) 0)
            {
                tm_timer_remove(curCacheEntryPtr->dnscTimerPtr);
                curCacheEntryPtr->dnscTimerPtr = (ttTimerPtr) 0;
            }
/* Free chain of address information structures, if present. */
            if (curCacheEntryPtr->dnscAddrInfoPtr != (struct addrinfo TM_FAR *)0)
            {
#ifdef TM_DNS_USE_SRV
#define TM_DNS_QTYPE_SRV 33
/* SRV entries use a different structure, and must be free'd differently */
                if (curCacheEntryPtr->dnscQueryType == TM_DNS_QTYPE_SRV)
                {
                    curAddrInfo =
                    (ttAddrInfoExtPtr)(curCacheEntryPtr->dnscAddrInfoPtr);

/* 1. Loop through each ttAddrInfoExt in the chain */
                    while (curAddrInfo)
                    {
/* 2. Free any associated addrinfo chains */
                        if (curAddrInfo->aie_ip4addrs)
                        {
                            freeaddrinfo(curAddrInfo->aie_ip4addrs);
                        }
                        curAddrInfo = (ttAddrInfoExtPtr)curAddrInfo->aie_next;
                    }
/* 3. Free the ttAddrInfoExt chain */
                    freeaddrinfo((struct addrinfo TM_FAR *)
                        (curCacheEntryPtr->dnscAddrInfoPtr));
                }
                else
#endif /* TM_DNS_USE_SRV */
                {
                    freeaddrinfo(curCacheEntryPtr->dnscAddrInfoPtr);
                }
            }

/*  5.2.    Free hostname string. */
            if (curCacheEntryPtr->dnscHostnameStr != (ttCharPtr) 0)
            {
                tm_free_raw_buffer(
                    (ttRawBufferPtr)curCacheEntryPtr->dnscHostnameStr);
            }

/*  5.3.    Free request string. */
            if (curCacheEntryPtr->dnscRequestStr != (ttCharPtr) 0)
            {
                tm_free_raw_buffer(
                    (ttRawBufferPtr)curCacheEntryPtr->dnscRequestStr);
            }

/*  5.4     Free reverse lookup hostname string. */
            if (curCacheEntryPtr->dnscRevHostnameStr != (ttCharPtr) 0)
            {
                tm_free_raw_buffer(
                    (ttRawBufferPtr)curCacheEntryPtr->dnscRevHostnameStr);
            }

/*  5.6.    Free DNS entry structure. */
            tm_free_raw_buffer((ttRawBufferPtr)curCacheEntryPtr);
            curCacheEntryPtr = nextCacheEntryPtr;
        } while (curCacheEntryPtr != tm_context(tvDnsCachePtr));
        tm_context(tvDnsCachePtr) = (ttDnsCacheEntryPtr)0;
    }
    tm_context(tvDnsCacheEntries) = 0;

    tm_call_unlock(&tm_context(tvDnsLockEntry));
    return;
}

/*
 * Multiple DNS servers:
 *
 * Test series #1: verify operation with multiple DNS servers.
 * Test #1.1: Two DNS servers.
 * Test #1.2: remove second name server, and verify that it isn't used.
 */
static void tfxDnsTestMultiServers(void)
{
    struct sockaddr_storage serverAddr;
    struct sockaddr_storage destAddr;
    ttSockAddrPtrUnion       sockAddr;
    int addrLength;
    struct addrinfo TM_FAR * addrInfoPtr;
    struct addrinfo addrHint;
    int errorCode;
    char inputBuf[TM_INPUT_BUF_LEN];
    tt8Bit match;
    int addrFamily;
#if defined(TM_USE_IPV4) && !defined(TM_SINGLE_INTERFACE_HOME)
    tt4IpAddress ip4Addr;
#endif /* defined(TM_USE_IPV4) && !defined(TM_SINGLE_INTERFACE_HOME) */

    tm_bzero(&addrHint, sizeof(struct addrinfo));
    tfSetCurrentContext(tvContextHandleClient);

/*
 * Test series #1: verify operation with multiple DNS servers.
 */

/* Test #1.1: Two DNS servers. */
/* setup DNS resolver with two servers. */
    tm_bcopy(&tv4ServerSockAddr[0], &serverAddr, sizeof(tv4ServerSockAddr[0]));
    addrFamily = AF_INET;
    errorCode = tfNgDnsSetServer(&serverAddr, 0);
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_ENOERROR);

#ifndef TM_SINGLE_INTERFACE_HOME
#ifdef TM_4_USE_SCOPE_ID
    tm_bcopy(
        &tv4MappedServerSockAddr[1], &serverAddr, sizeof(tv4ServerSockAddr[1]));
#else /* ! TM_4_USE_SCOPE_ID */
    tm_bcopy(&tv4ServerSockAddr[1], &serverAddr, sizeof(tv4ServerSockAddr[1]));
#endif /* ! TM_4_USE_SCOPE_ID */
    errorCode = tfNgDnsSetServer(&serverAddr, 1);
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_ENOERROR);
#endif /* !TM_SINGLE_INTERFACE_HOME */

    addrHint.ai_flags = 0;
    addrHint.ai_family = addrFamily;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the first query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(struct sockaddr_storage);
    sockAddr.sockNgPtr = &destAddr;
    errorCode = tfRecvFromTo( tlServerSd,
                              inputBuf,
                              sizeof(inputBuf),
                              0,
                              (struct sockaddr TM_FAR *) 0,
                              &addrLength,
                              sockAddr.sockPtr );
    tm_assert(tfxDnsTestMultiServers, errorCode > 0);

/* Verify that request was sent to first name server. */
    match = TM_8BIT_ZERO;
    if ( tm_ip_match( sockAddr.sockInPtr->sin_addr.s_addr,
                      tv4ServerSockAddr[0].addr.ipv4.sin_addr.s_addr ))
    {
        match = TM_8BIT_YES;
    }
    tm_assert(tfxDnsTestMultiServers, match == TM_8BIT_YES);
    tm_assert(tfxDnsTestMultiServers,
              (sockAddr.sockInPtr->sin_port == tm_const_htons(53)));

/* Allow timeout to expire, and then check for second request. */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

#ifndef TM_SINGLE_INTERFACE_HOME
/* Get request from the second server. */
    tfSetCurrentContext(tvContextHandleServer);
    addrLength = sizeof(struct sockaddr_storage);
    sockAddr.sockNgPtr = &destAddr;
    errorCode = tfRecvFromTo( tlServerSd,
                              inputBuf,
                              sizeof(inputBuf),
                              0,
                              (struct sockaddr TM_FAR *) 0,
                              &addrLength,
                              sockAddr.sockPtr );
    tm_assert(tfxDnsTestMultiServers, errorCode > 0);

    match = TM_8BIT_ZERO;
/* IPv4 address return in IPv4-mapped IPv6 address format: */
    ip4Addr = sockAddr.sockInPtr->sin_addr.s_addr;

/* Verify that request was sent to the second name server. */
    if ( tm_ip_match( ip4Addr,
                      tv4ServerSockAddr[1].addr.ipv4.sin_addr.s_addr ))
    {
        match = TM_8BIT_YES;
    }
    tm_assert(tfxDnsTestMultiServers, match == TM_8BIT_YES);
    tm_assert(tfxDnsTestMultiServers,
              (sockAddr.sockInPtr->sin_port == tm_const_htons(53)));

    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_ETIMEDOUT);

/* Test #1.2: remove second name server, and verify that it isn't used. */
    tfSetCurrentContext(tvContextHandleClient);

    errorCode = tfNgDnsSetServer((struct sockaddr_storage TM_FAR *)0, 1);
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_ENOERROR);

    addrHint.ai_flags = 0;
    addrHint.ai_family = addrFamily;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Read request. */
    tfSetCurrentContext(tvContextHandleServer);
    addrLength = sizeof(struct sockaddr_storage);
    sockAddr.sockNgPtr = &destAddr;
    errorCode = tfRecvFromTo( tlServerSd,
                              inputBuf,
                              sizeof(inputBuf),
                              0,
                              (struct sockaddr TM_FAR *) 0,
                              &addrLength,
                              sockAddr.sockPtr );
    tm_assert(tfxDnsTestMultiServers, errorCode > 0);

/* Verify that the resolver times out after first request. */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();
#endif /* !TM_SINGLE_INTERFACE_HOME */

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ETIMEDOUT);

/* Verify that no additional requests were sent. */
    tfSetCurrentContext(tvContextHandleServer);
    addrLength = sizeof(struct sockaddr_storage);
    sockAddr.sockNgPtr = &destAddr;
    errorCode = tfRecvFromTo( tlServerSd,
                              inputBuf,
                              sizeof(inputBuf),
                              0,
                              (struct sockaddr TM_FAR *) 0,
                              &addrLength,
                              sockAddr.sockPtr );
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_SOCKET_ERROR);
    errorCode = tfGetSocketError(tlServerSd);
    tm_assert(tfxDnsTestMultiServers, errorCode == TM_EWOULDBLOCK);

}


/*
 * DNS Query Type tests:
 *
 * Test series #1: Normal queries.
 *     Test #1.1: Normal query (A, AAAA)
 *     Test #1.2: Verify that if we immediately query again for the same address,
 *                the cache entry is used, and no request is sent.
 *     Test #1.3: Verify that the cache entry times out correctly.
 * Test series #2: Verify that IPv6 query is sent when AF_UNSPEC is
 *                 provided as address family.
 * Test #2.1 unspecified address family
 */
static void tfxDnsTestQueryType(int family)
{
    struct sockaddr_storage serverAddr;
    struct sockaddr_storage TM_FAR * ipv4AddrPtr;
    struct addrinfo TM_FAR * addrInfoPtr;
    struct addrinfo addrHint;
    int errorCode;
    char inputBuf[TM_INPUT_BUF_LEN];
    tt8BitPtr dataPtr;
    tt16Bit queryType;
    tt16Bit expectedQueryType;

    tm_bzero(&addrHint, sizeof(struct addrinfo));

    if (family == AF_INET)
    {
        expectedQueryType = TM_DNS_QTYPE_A;
    }
    else
    {
        expectedQueryType = TM_DNS_QTYPE_AAAA;
    }

    tfSetCurrentContext(tvContextHandleClient);

/* setup DNS resolver with a single server. */
    ipv4AddrPtr = &tv4ServerSockAddr[0];

#ifdef TM_4_USE_SCOPE_ID
    tm_bcopy(ipv6AddrPtr, &serverAddr, sizeof(struct sockaddr_storage));
#else /* ! TM_4_USE_SCOPE_ID */
    {
        tm_bcopy(ipv4AddrPtr, &serverAddr, sizeof(struct sockaddr_storage));
    }
#endif /* ! TM_4_USE_SCOPE_ID */

    errorCode = tfNgDnsSetServer(&serverAddr, 0);
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ENOERROR);

/*
 * Test series #1: Normal queries.
 */

/* Test #1.1: Normal query (A, AAAA) */
    addrHint.ai_flags = 0;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsTestQueryType, errorCode > 0);

/* We know the format of the query, and we know that we'll only be sending
   a single question with this query, so rather than parse through the whole
   packet just to determine what the question type is, we look from the end
   of the packet.  We have to do a 'bcopy' since there's no guarantee on
   alignment... */
    dataPtr = (tt8BitPtr) inputBuf + errorCode - 4;
    tm_bcopy(dataPtr, &queryType, sizeof(queryType));

    queryType = ntohs(queryType);

/* Make sure that the query was an AAAA (IPv6) type */
    tm_assert(tfxDnsTestQueryType, queryType == expectedQueryType);

/* Test #1.2: Verify that if we immediately query again for the same address,
 *            the cache entry is used, and no request is sent.
 */
    tfSetCurrentContext(tvContextHandleClient);
    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_SOCKET_ERROR);

    errorCode = tfGetSocketError(tlServerSd);
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Test #1.3: Verify that the cache entry times out correctly. */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

    addrHint.ai_flags = 0;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ETIMEDOUT);

/*
 * Test series #2: Verify that IPv6 query is sent when AF_UNSPEC is
 *                 provided as address family.
 */
    tfSetCurrentContext(tvContextHandleClient);
    addrHint.ai_flags = 0;
    addrHint.ai_family = AF_UNSPEC;

/* Test #2.1 unspecified address family */
    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsTestQueryType, errorCode > 0);
    dataPtr = (tt8BitPtr) inputBuf + errorCode - 4;
    tm_bcopy(dataPtr, &queryType, sizeof(queryType));

    queryType = ntohs(queryType);
/* IPv6 is not enabled, so default should be IPv4 */
    tm_assert(tfxDnsTestQueryType, queryType == TM_DNS_QTYPE_A);

/* Make sure that this request times out... */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ETIMEDOUT);

}

/*
 * Tests the AI_NUMERICFLAG with getaddrinfo and NI_NUMERICFLAG with
 * getnameinfo.
 *
 * Test series #1: getaddrinfo with the AI_NUMERICHOST flag
 *                 This of course assumes that inet_pton works correctly...
 *     Test #1.1: getaddrinfo, AI_NUMERICHOST, IPv4
 *     Test #1.2: getaddrinfo, AI_NUMERICFLAG, IPv6
 * Test series #2: getnameinfo with NI_NUMERICHOST flag
 *     Test #2.1: getnameinfo, NI_NUMERICHOST, IPv4
 *     Test #2.2: getnameinfo, NI_NUMERICHOST, IPv6
 */
static void tfxDnsTestNumericFlag(void)
{
    struct addrinfo          addrHint;
    struct addrinfo TM_FAR * addrInfoPtr;
    struct addrinfo TM_FAR * aiWalkPtr;
    struct sockaddr_storage  origAddr;
    ttSockAddrPtrUnion       sockAddr;
    int                      sockTypeMask;
    int                      errorCode;
    char                     hostname[TM_HOSTNAME_BUF_LEN];
    tt8Bit                   match;

    tm_bzero(&addrHint, sizeof(struct addrinfo));
    tfSetCurrentContext(tvContextHandleClient);

/*
 * Test series #1: getaddrinfo with the AI_NUMERICHOST flag
 * This of course assumes that inet_pton works correctly...
 */

/* Test #1.1: getaddrinfo, AI_NUMERICHOST, IPv4 */
    errorCode = inet_pton(AF_INET,
                          ipv4TestAddr,
                          (ttVoidPtr) &(origAddr.addr.ipv4.sin_addr.s_addr));
    tm_assert(tfxDnsTestNumericFlag, errorCode == 1);

    addrHint.ai_flags = AI_NUMERICHOST;
    addrHint.ai_family = AF_INET;

    errorCode = getaddrinfo( ipv4TestAddr, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestNumericFlag, errorCode == TM_ENOERROR);
    tm_assert(tfxDnsTestNumericFlag, addrInfoPtr != 0);
/* Check that all supported socket types are represented exactly once */
    sockTypeMask = (1 << SOCK_DGRAM)
#ifndef TM_DISABLE_TCP
                 + (1 << SOCK_STREAM)
#endif /* !TM_DISABLE_TCP */
#if defined(TM_USE_RAW_SOCKET) || defined(TM_6_USE_RAW_SOCKET)
                 + (1 << SOCK_RAW)
#endif /* TM_USE_RAW_SOCKET || TM_6_USE_RAW_SOCKET */
                ;
    for (aiWalkPtr = addrInfoPtr; aiWalkPtr; aiWalkPtr = aiWalkPtr->ai_next)
    {
        tm_assert(tfxDnsTestNumericFlag, aiWalkPtr->ai_family == AF_INET);
        tm_assert(tfxDnsTestNumericFlag,
                  aiWalkPtr->ai_addrlen == sizeof(struct sockaddr_in));
        tm_assert(tfxDnsTestNumericFlag, aiWalkPtr->ai_addr != 0);
        if (  (   aiWalkPtr->ai_socktype == SOCK_DGRAM
               && aiWalkPtr->ai_protocol == IPPROTO_UDP )
#ifndef TM_DISABLE_TCP
           || (   aiWalkPtr->ai_socktype == SOCK_STREAM
               && aiWalkPtr->ai_protocol == IPPROTO_TCP )
#endif /* !TM_DISABLE_TCP */
#if defined(TM_USE_RAW_SOCKET) || defined(TM_6_USE_RAW_SOCKET)
           || (   aiWalkPtr->ai_socktype == SOCK_RAW
               && aiWalkPtr->ai_protocol == 0 )
#endif /* TM_USE_RAW_SOCKET || TM_6_USE_RAW_SOCKET */
           )
        {
            match = 1;
        }
        else
        {
            match = 0;
        }
        tm_assert(tfxDnsTestNumericFlag, match == 1);
        sockTypeMask -= 1 << aiWalkPtr->ai_socktype;

        sockAddr.sockPtr = aiWalkPtr->ai_addr;
        if (tm_ip_match(sockAddr.sockInPtr->sin_addr.s_addr,
                        origAddr.addr.ipv4.sin_addr.s_addr) )
        {
            match = 1;
        }
        else
        {
            match = 0;
        }

        tm_assert(tfxDnsTestNumericFlag, match == 1);
    }
    tm_assert(tfxDnsTestNumericFlag, sockTypeMask == 0);

    freeaddrinfo(addrInfoPtr);

/* Test #1.2: getaddrinfo, AI_NUMERICFLAG, IPv6 */

/*
 * Test series #2: getnameinfo with NI_NUMERICHOST flag
 */

/* Test #2.1: getnameinfo, NI_NUMERICHOST, IPv4 */
    errorCode = inet_pton(AF_INET,
                          ipv4TestAddr,
                          (ttVoidPtr) &(origAddr.addr.ipv4.sin_addr.s_addr));
    tm_assert(tfxDnsTestNumericFlag, errorCode == 1);

    origAddr.addr.ipv4.sin_family = AF_INET;
    origAddr.addr.ipv4.sin_len = sizeof(struct sockaddr_in);

    sockAddr.sockNgPtr = &origAddr;
    errorCode = getnameinfo( sockAddr.sockPtr,
                             sizeof(struct sockaddr_in),
                             hostname,
                             TM_HOSTNAME_BUF_LEN,
                             0,
                             0,
                             NI_NUMERICHOST );
    tm_assert(tfxDnsTestNumericFlag, errorCode == TM_ENOERROR);

    match = (tt8Bit) tm_strcmp(hostname, ipv4TestAddr);
    tm_assert(tfxDnsTestNumericFlag, match == 0);

/* Test #2.2: getnameinfo, NI_NUMERICHOST, IPv6 */

}

/*
 * IPv6 reverse lookup tests.
 *
 * Test series #1: verify that both IP6 and ARPA domains are attempted.
 *     Test #1.1: Attempt to do a reverse lookup, verify that both domains
 *                are used.
 */
static void tfxDnsTestReverseLookup(int family)
{
    int                      errorCode;
    int                      addrLength;
    int                      dataLength;
    ttSockAddrPtrUnion       sockAddr;
    ttSockAddrPtrUnion       dnsServerSockAddr;
    struct sockaddr_storage  dnsServerAddr;
    ttUserMessage            bufferHandle;
    struct sockaddr_storage  serverAddr;
    struct sockaddr_storage  origAddr;
    struct sockaddr_storage  clientAddr;
    struct sockaddr_storage TM_FAR * ipv4AddrPtr;
    char                     hostname[TM_HOSTNAME_BUF_LEN];
    char                     inputBuf[TM_INPUT_BUF_LEN];
    tt8BitPtr                dataPtr;
    tt8Bit                   match;
    ttDnsHeaderPtr           dnsHdrPtr;

    tfSetCurrentContext(tvContextHandleClient);

    ipv4AddrPtr = &tv4ServerSockAddr[0];

    {
        tm_bcopy(ipv4AddrPtr, &serverAddr, sizeof(struct sockaddr_storage));
    }

    errorCode = tfNgDnsSetServer(&serverAddr, 0);
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ENOERROR);


/*
 * Test series #1: verify that both INT and ARPA domains are attempted.
 *     Test #1.1: Attempt to do a reverse lookup, verify that both domains
 *                are used.
 */

    tfSetCurrentContext(tvContextHandleClient);

    addrLength = 0; /* avoid compiler warnings */
    {
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr,
                              (ttVoidPtr) &(origAddr.addr.ipv4.sin_addr.s_addr));
        addrLength = sizeof(struct sockaddr_in);
    }

    tm_assert(tfxDnsTestReverseLookup, errorCode == 1);

    origAddr.ss_family = (tt8Bit) family;
    origAddr.ss_len = (tt8Bit) addrLength;

    sockAddr.sockNgPtr = &origAddr;
    errorCode = getnameinfo( sockAddr.sockPtr,
                             sockAddr.sockPtr->sa_len,
                             hostname,
                             TM_HOSTNAME_BUF_LEN,
                             0,
                             0,
                             0 );
    tm_assert(tfxDnsTestReverseLookup, errorCode == TM_EWOULDBLOCK);

/* Get query from client */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    dnsServerSockAddr.sockNgPtr = &dnsServerAddr;
    errorCode = tfRecvFromTo( tlServerSd,
                              inputBuf,
                              sizeof(inputBuf),
                              0,
                              sockAddr.sockPtr,
                              &addrLength,
                              dnsServerSockAddr.sockPtr);
    tm_assert(tfxDnsTestReverseLookup, errorCode > 0);
    tm_assert(tfxDnsTestMultiServers,
              (dnsServerSockAddr.sockInPtr->sin_port == tm_const_htons(53)));

/* Check that the domain is ".arpa". */
    dataPtr = (tt8BitPtr) (inputBuf + errorCode - 9);
    match = (tt8Bit) tm_strcmp(dataPtr, "arpa");
    tm_assert(tfxDnsTestReverseLookup, match == 0);

    if (family == AF_INET6)
    {
/* For IPv6, Send a "no answer" response back, which should cause the ".int"
   domain to be used. */
        dnsHdrPtr = (ttDnsHeaderPtr) inputBuf;

/* Set flags to response. */
        dnsHdrPtr->dnsFlags = htons(TM_DNS_HDR_RESP);

/* Clear "question count" field */
        dnsHdrPtr->dnsAnswerCount = 0;

/* Send response back to resolver, with no answer */
        errorCode = tfSendToFrom( tlServerSd,
                                  inputBuf,
                                  12,
                                  0,
                                  sockAddr.sockPtr,
                                  addrLength,
                                  dnsServerSockAddr.sockPtr);

        tfSetCurrentContext(tvContextHandleClient);
        tfxDnsRunTimer();

/* Get query from client */
        tfSetCurrentContext(tvContextHandleServer);

        addrLength = sizeof(struct sockaddr_storage);
        sockAddr.sockNgPtr = &clientAddr;
        errorCode = tfRecvFromTo( tlServerSd,
                                  inputBuf,
                                  sizeof(inputBuf),
                                  0,
                                  sockAddr.sockPtr,
                                  &addrLength,
                                  dnsServerSockAddr.sockPtr);
        tm_assert(tfxDnsTestReverseLookup, errorCode > 0);
        tm_assert(tfxDnsTestMultiServers,
                  (dnsServerSockAddr.sockInPtr->sin_port
                   == tm_const_htons(53)));

/* Check that the domain is ".int". */
        dataPtr = (tt8BitPtr) (inputBuf + errorCode - 8);
        match = (tt8Bit) tm_strcmp(dataPtr, "int");
        tm_assert(tfxDnsTestReverseLookup, match == 0);
    }

    tfSetCurrentContext(tvContextHandleServer);

/* Send response to client. */
    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      TM_DNS_QTYPE_REV,
                                      0,
                                      0,
                                      (tt8BitPtr) testHostnameOneRec,
                                      0,
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);

    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

/* Get results from reverse lookup */
    tfSetCurrentContext(tvContextHandleClient);

    {
/* AF_INET */
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr,
                              (ttVoidPtr) &(origAddr.addr.ipv4.sin_addr.s_addr));
    }
    tm_assert(tfxDnsTestReverseLookup, errorCode == 1);

    sockAddr.sockNgPtr = &origAddr;
    errorCode = getnameinfo( sockAddr.sockPtr,
                             sockAddr.sockPtr->sa_len,
                             hostname,
                             TM_HOSTNAME_BUF_LEN,
                             0,
                             0,
                             0 );
    tm_assert(tfxDnsTestReverseLookup, errorCode == TM_ENOERROR);

    match = (tt8Bit) tm_strcmp(hostname, testHostnameOne);
    tm_assert(tfxDnsTestReverseLookup, match == 0);

}

/*
 * Verify that gai_strerror returns proper strings for all error codes,
 * including an unrecognized one.
 */
#ifdef TM_TEST_DNS_VERBOSE
static void tfxDnsTestStrError(void)
{
    char TM_FAR * stringPtr;

    printf("Verify that gai_strerror returns proper strings for all error codes");

    stringPtr = gai_strerror(EAI_AGAIN);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_AGAIN: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_BADFLAGS);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_BADFLAGS: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_FAIL);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_FLAGS: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_FAMILY);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_FAMILY: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_MEMORY);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_MEMORY: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_NONAME);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_NONAME_: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_SERVICE);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_SERVICE: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_SOCKTYPE );
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_SOCKTYPE: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_SYSTEM );
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_SYSTEM: %s\n", stringPtr);

    stringPtr = gai_strerror(EAI_OVERFLOW );
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("EAI_OVERFLOW: %s\n", stringPtr);

    stringPtr = gai_strerror(7777);
    tm_assert(tfxDnsTestStrError, stringPtr != 0);
    printf("INVALID: %s\n", stringPtr);

    return;
}
#endif /* TM_TEST_DNS_VERBOSE*/

static void tfxDnsTestQueryResponse(int family)
{
    struct sockaddr_storage TM_FAR *    ipv4AddrPtr;
    struct addrinfo TM_FAR *            addrInfoPtr;
    struct addrinfo TM_FAR *            nextAddrInfoPtr;
    ttSockAddrPtrUnion                  sockAddr;
    ttUserMessage                       bufferHandle;
    tt8BitPtr                           addrStrPtr;
    tt8BitPtr                           addrStr2Ptr;
    struct sockaddr_storage             serverAddr;
    struct sockaddr_storage             clientAddr;
    struct sockaddr_storage             mxAddr;
    struct addrinfo                     addrHint;
    int                                 errorCode;
    int                                 addrLength;
    int                                 dataLength;
    int                                 timeoutCount;
    tt16Bit                             queryType;
    char                                inputBuf[TM_INPUT_BUF_LEN];
    tt8Bit                              match;

    tm_bzero(&addrHint, sizeof(struct addrinfo));

    if (family == AF_INET)
    {
        queryType = TM_DNS_QTYPE_A;
        addrLength = 4;
    }
    else
    {
        queryType = TM_DNS_QTYPE_AAAA;
        addrLength = 16;
    }

    tfSetCurrentContext(tvContextHandleClient);

/* setup DNS resolver with a single server. */
    ipv4AddrPtr = &tv4ServerSockAddr[0];

    {
        tm_bcopy(ipv4AddrPtr, &serverAddr, sizeof(struct sockaddr_storage));
    }
    errorCode = tfNgDnsSetServer(&serverAddr, 0);
    tm_assert(tfxDnsTestQueryType, errorCode == TM_ENOERROR);

/*
 * Test series #1: Single protocol family.  Send a normal A or AAAA query.
 *                 The "server" will respond with an answer that includes
 *                 multiple answer records as well as a CNAME record.
 *                 Each individual test below then verifies that various
 *                 parts of this response were received correctly.
 */

    addrHint.ai_flags  = AI_CANONNAME;;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryType, errorCode > 0);

/* Add first A/AAAA record */
    if (family == AF_INET)
    {
        addrStrPtr = (tt8BitPtr) ipv4TestAddr;
        addrStr2Ptr = (tt8BitPtr) ipv4TestAddr2;
    }
    else
    {
/* family == AF_INET6 */
        addrStrPtr = (tt8BitPtr) ipv6TestAddr;
        addrStr2Ptr = (tt8BitPtr) ipv6TestAddr2;
    }

    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      queryType,
                                      addrStrPtr,
                                      addrStr2Ptr,
                                      (tt8BitPtr) testCanonRecName,
                                      0,
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);

    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

    tfSetCurrentContext(tvContextHandleClient);

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);

/* Test #1.1: Verify that addresses are returned correctly in the
 *            addrinfo structures.
 */

/* Verify that two (and only two) addrinfo structures are returned, and that
 * they both have valid address pointers included.. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_next != 0);
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addr != 0);

/* We expect 2 addresses each with up to 3 sock types */
    while (addrInfoPtr->ai_socktype !=
#ifndef TM_DISABLE_TCP
           SOCK_STREAM
#else /* TM_DISABLE_TCP */
           SOCK_DGRAM
#endif /* TM_DISABLE_TCP */
           )
    {
        addrInfoPtr = addrInfoPtr->ai_next;
        tm_assert(tfxDnsTestQueryResponse, addrInfoPtr != 0);
    }

    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addr != 0);

    nextAddrInfoPtr = addrInfoPtr->ai_next;
    while (nextAddrInfoPtr->ai_socktype !=
#ifndef TM_DISABLE_TCP
           SOCK_STREAM
#else /* TM_DISABLE_TCP */
           SOCK_DGRAM
#endif /* TM_DISABLE_TCP */
           )
    {
        nextAddrInfoPtr = nextAddrInfoPtr->ai_next;
        tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr != 0);
    }

/* Verify that the address families match. */
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_family == family);
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_family == family);

    tm_assert(tfxDnsTestQueryResponse,
              nextAddrInfoPtr->ai_addr->sa_family == family);
    tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr->ai_family == family);

/* Verify that the address structure lengths are set correctly. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addrlen > 0);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_len == addrInfoPtr->ai_addrlen);

    tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr->ai_addrlen > 0);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_len == nextAddrInfoPtr->ai_addrlen);

/* Verify that the addresses match the expected values.  Note that the
 * address are returned in reverse order (#2 first, and then #1).
 */
    if (family == AF_INET)
    {
/* First address returned. */
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr2,
                              &serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, errorCode == 1);

        sockAddr.sockPtr = addrInfoPtr->ai_addr;
        match = (tt8Bit) tm_ip_match(sockAddr.sockInPtr->sin_addr.s_addr,
                                     serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, match != 0);

/* Second address returned */
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr,
                              &serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, errorCode == 1);

        sockAddr.sockPtr = nextAddrInfoPtr->ai_addr;
        match = (tt8Bit) tm_ip_match(sockAddr.sockInPtr->sin_addr.s_addr,
                                     serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, match != 0);

    }
    else
    {
    }

/* Test #1.2: Verify that canonical name is returned. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_canonname != 0);

    match = (tt8Bit) tm_strcmp(addrInfoPtr->ai_canonname, testCanonName);
    tm_assert(tfxDnsTestQueryResponse, match == 0);

/* Test #1.3: For IPv6 only - verify that link-local address is scoped
 * correctly.
 */

    freeaddrinfo(addrInfoPtr);

/* Test #1.4: Verify that the TTL received in the response is used
 *            correctly - the shortest response should be used.  Verify
 *            that cache entries are not timed out too early or too late.
 */

/* Wait half of the TTL time, try an indentical query, and verify that no
 * new request packet is sent.
 */
    timeoutCount = (TM_TEST_TTL_SHORT*1000) / tvTimerTickLength;
    timeoutCount = timeoutCount >> 1;

    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);

    freeaddrinfo(addrInfoPtr);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_SOCKET_ERROR);

    errorCode = tfGetSocketError(tlServerSd);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_EWOULDBLOCK);

/* Allow the TTL to expire, try an identical query, and verify that a new
 * request packet is sent (the cache entry times out).
 */
    tfSetCurrentContext(tvContextHandleClient);

    timeoutCount = (TM_TEST_TTL_SHORT*1000) / tvTimerTickLength;
    timeoutCount = timeoutCount >> 1;

    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryResponse, errorCode > 0);

/* Don't respond, and allow the request to time out. */
    tfSetCurrentContext(tvContextHandleClient);

    tfxDnsRunTimer();

/* Get error code for this hostname, so entry can be deleted. */
    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ETIMEDOUT);

/*
 * Test series #2: The form of a test is the same as #1.
 *                 However, in this time, the "server" will respond with an
 *                 answer that is truncated, i.e. the TC bit is set because the
 *                 512-byte limit has bee reached and some information will not
 *                 fit in the response.
 *                 Each individual test below then verifies that resolver can
 *                 handle receiving a truncated response from the DNS server.
 */

    tfSetCurrentContext(tvContextHandleClient);

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryType, errorCode > 0);

/* Add first A/AAAA record */
    if (family == AF_INET)
    {
        addrStrPtr = (tt8BitPtr) ipv4TestAddr;
        addrStr2Ptr = (tt8BitPtr) ipv4TestAddr2;
    }
    else
    {
/* family == AF_INET6 */
        addrStrPtr = (tt8BitPtr) ipv6TestAddr;
        addrStr2Ptr = (tt8BitPtr) ipv6TestAddr2;
    }

/* Create reply whose TC bit is set */
    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      queryType,
                                      addrStrPtr,
                                      addrStr2Ptr,
                                      (tt8BitPtr) testCanonRecName,
                                      TM_DNS_HDR_TRUNCATION,/* Set TC bit */
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);


    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

    tfSetCurrentContext(tvContextHandleClient);

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);

/*
 * Test #2.1: Verify that addresses are returned correctly in the
 *            addrinfo structures.
 */

/* Verify that two (and only two) addrinfo structures are returned, and that
 * they both have valid address pointers included.. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_next != 0);
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addr != 0);

/* We expect 2 addresses each with up to 3 sock types */
    while (addrInfoPtr->ai_socktype !=
#ifndef TM_DISABLE_TCP
           SOCK_STREAM
#else /* TM_DISABLE_TCP */
           SOCK_DGRAM
#endif /* TM_DISABLE_TCP */
           )
    {
        addrInfoPtr = addrInfoPtr->ai_next;
        tm_assert(tfxDnsTestQueryResponse, addrInfoPtr != 0);
    }

    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addr != 0);

    nextAddrInfoPtr = addrInfoPtr->ai_next;
    while (nextAddrInfoPtr->ai_socktype !=
#ifndef TM_DISABLE_TCP
           SOCK_STREAM
#else /* TM_DISABLE_TCP */
           SOCK_DGRAM
#endif /* TM_DISABLE_TCP */
           )
    {
        nextAddrInfoPtr = nextAddrInfoPtr->ai_next;
        tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr != 0);
    }

/* Verify that the address families match. */
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_family == family);
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_family == family);

    tm_assert(tfxDnsTestQueryResponse,
              nextAddrInfoPtr->ai_addr->sa_family == family);
    tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr->ai_family == family);

/* Verify that the address structure lengths are set correctly. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_addrlen > 0);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_len == addrInfoPtr->ai_addrlen);

    tm_assert(tfxDnsTestQueryResponse, nextAddrInfoPtr->ai_addrlen > 0);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr->ai_addr->sa_len == nextAddrInfoPtr->ai_addrlen);

/* Verify that the addresses match the expected values.  Note that the
 * address are returned in reverse order (#2 first, and then #1).
 */
    if (family == AF_INET)
    {
/* First address returned. */
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr2,
                              &serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, errorCode == 1);

        sockAddr.sockPtr = addrInfoPtr->ai_addr;
        match = (tt8Bit) tm_ip_match(sockAddr.sockInPtr->sin_addr.s_addr,
                                     serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, match != 0);

/* Second address returned */
        errorCode = inet_pton(AF_INET,
                              ipv4TestAddr,
                              &serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, errorCode == 1);

        sockAddr.sockPtr = nextAddrInfoPtr->ai_addr;
        match = (tt8Bit) tm_ip_match(sockAddr.sockInPtr->sin_addr.s_addr,
                                     serverAddr.addr.ipv4.sin_addr.s_addr);
        tm_assert(tfxDnsTestQueryResponse, match != 0);

    }
    else
    {
    }

/* Test #2.2: Verify that canonical name is returned. */
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_canonname != 0);

    match = (tt8Bit) tm_strcmp(addrInfoPtr->ai_canonname, testCanonName);
    tm_assert(tfxDnsTestQueryResponse, match == 0);

/* Test #2.3: For IPv6 only - verify that link-local address is scoped
 * correctly.
 */

    freeaddrinfo(addrInfoPtr);

/*
 * Test series #3: The form of a test is the same as #2.
 *                 However, in this time, TC bit is set and there is no
 *                 resource record in the response.
 *                 This test verifies that resolver can return TM_DNS_EANSWER.
 */

/* Allow the TTL of previous query to expire, */
    timeoutCount = (TM_TEST_TTL_LONG*1000) / tvTimerTickLength;
    timeoutCount = timeoutCount >> 1;

    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

    tfSetCurrentContext(tvContextHandleClient);

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

    tfxDnsRunTimer();

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryType, errorCode > 0);

/* Create reply which does not have resource record */
    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      queryType,
                                      0,/* no address information */
                                      0,/* no address information */
                                      0,/* no canonical name */
                                      TM_DNS_HDR_TRUNCATION,/* Set TC bit */
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);

    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

    tfSetCurrentContext(tvContextHandleClient);

    addrHint.ai_flags  = AI_CANONNAME;
    addrHint.ai_family = family;

/* Verify that resolver returns TM_DNS_EANSWER. */
    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_DNS_EANSWER);


/*
 * 4.0: Proper MX record parsing and sorting
 *      The expected response order is v6:pref=10;
 *                                     v4:pref=10;
 *                                     v6:pref=15;
 *                                     v4:pref=15;
 *      This order depends on if v4, v6 or both are defined.
 */
    tfDnsCacheInvalidate();

/* 4.1: hintsPtr is NULL */
    tm_bzero(&addrHint, sizeof(struct addrinfo));
    errorCode = tfNgDnsGetMailHost("example.com",
                                   (struct addrinfo *)0,
                                   &addrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);
    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom(tlServerSd,
                         inputBuf,
                         sizeof(inputBuf),
                         0,
                         sockAddr.sockPtr,
                         &addrLength);
    tm_assert(tfxDnsTestQueryResponse, errorCode > 0);

/* Copy the Transaction ID from the query */
    tm_bcopy(inputBuf, mxResponse, 2);

    errorCode = sendto(tlServerSd,
                       (ttCharPtr)mxResponse,
                       sizeof(mxResponse),
                       0,
                       sockAddr.sockPtr,
                       addrLength);

    tfSetCurrentContext(tvContextHandleClient);
    errorCode = tfNgDnsGetMailHost("example.com",
                                   (struct addrinfo *)0,
                                   &addrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr != (struct addrinfo TM_FAR *)0);
    sockAddr.sockPtr = addrInfoPtr->ai_addr;
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_mxpref == 10);
    tm_bzero(&mxAddr, sizeof(struct sockaddr_storage));
    inet_pton(AF_INET,
              "172.30.79.10",
              &(mxAddr.addr.ipv4.sin_addr));
    mxAddr.ss_len = sizeof(struct sockaddr_storage);
    mxAddr.ss_family = AF_INET;
    errorCode = tm_ip_match_sockaddr(*(sockAddr.sockNgPtr), mxAddr);
    tm_assert(tfxDnsTestQueryResponse, errorCode != 0);

    errorCode = tfNgDnsGetNextMailHost("example.com",
                                       addrInfoPtr,
                                       (unsigned short)addrInfoPtr->ai_mxpref,
                                       (struct addrinfo *)0,
                                       &nextAddrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);
    freeaddrinfo(addrInfoPtr);
    addrInfoPtr = nextAddrInfoPtr;
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr != (struct addrinfo TM_FAR *)0);
    sockAddr.sockPtr = addrInfoPtr->ai_addr;
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_mxpref == 15);
    tm_bzero(&mxAddr, sizeof(struct sockaddr_storage));
    inet_pton(AF_INET,
              "172.30.79.15",
              &(mxAddr.addr.ipv4.sin_addr));
    mxAddr.ss_len = sizeof(struct sockaddr_storage);
    mxAddr.ss_family = AF_INET;
    errorCode = tm_ip_match_sockaddr(*(sockAddr.sockNgPtr), mxAddr);
    tm_assert(tfxDnsTestQueryResponse, errorCode != 0);


    errorCode = tfNgDnsGetNextMailHost("example.com",
                                       addrInfoPtr,
                                       (unsigned short)addrInfoPtr->ai_mxpref,
                                       (struct addrinfo *)0,
                                       &addrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_DNS_EANSWER);

    freeaddrinfo(addrInfoPtr);

/* 4.2: hintsPtr favors IPv4 */
    tm_bzero(&addrHint, sizeof(struct addrinfo));
    addrHint.ai_family = AF_INET;

    errorCode = tfNgDnsGetMailHost("example.com",
                                   &addrHint,
                                   &addrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr != (struct addrinfo TM_FAR *)0);
    sockAddr.sockPtr = addrInfoPtr->ai_addr;
    tm_assert(tfxDnsTestQueryResponse, addrInfoPtr->ai_mxpref == 10);

    tm_bzero(&mxAddr, sizeof(struct sockaddr_storage));
    inet_pton(AF_INET,
              "172.30.79.10",
              &(mxAddr.addr.ipv4.sin_addr));
    mxAddr.ss_len = sizeof(struct sockaddr_storage);
    mxAddr.ss_family = AF_INET;
    errorCode = tm_ip_match_sockaddr(*(sockAddr.sockNgPtr), mxAddr);
    tm_assert(tfxDnsTestQueryResponse, errorCode != 0);

    errorCode = tfNgDnsGetNextMailHost("example.com",
                                       addrInfoPtr,
                                       (unsigned short)addrInfoPtr->ai_mxpref,
                                       &addrHint,
                                       &nextAddrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_ENOERROR);
    freeaddrinfo(addrInfoPtr);
    addrInfoPtr = nextAddrInfoPtr;
    tm_assert(tfxDnsTestQueryResponse,
              addrInfoPtr != (struct addrinfo TM_FAR *)0);
    sockAddr.sockPtr = addrInfoPtr->ai_addr;

    tm_bzero(&mxAddr, sizeof(struct sockaddr_storage));
    inet_pton(AF_INET,
              "172.30.79.15",
              &(mxAddr.addr.ipv4.sin_addr));
    mxAddr.ss_len = sizeof(struct sockaddr_storage);
    mxAddr.ss_family = AF_INET;
    errorCode = tm_ip_match_sockaddr(*(sockAddr.sockNgPtr), mxAddr);
    tm_assert(tfxDnsTestQueryResponse, errorCode != 0);

    errorCode = tfNgDnsGetNextMailHost("example.com",
                                       addrInfoPtr,
                                       (unsigned short)addrInfoPtr->ai_mxpref,
                                       &addrHint,
                                       &addrInfoPtr);
    tm_assert(tfxDnsTestQueryResponse, errorCode == TM_DNS_EANSWER);

    freeaddrinfo(addrInfoPtr);


    tfDnsCacheInvalidate();

    return;
}

/*
 * Creates a DNS response record at specified location (assumes that enough
 * space is available).  Returns the total length of the record.
 */
static int tfxAddResponseRec(tt8BitPtr     dataPtr,
                             tt16Bit       queryType,
                             tt32Bit       ttl,
                             tt8BitPtr     addrStrPtr,
                             tt8BitPtr rrNamePtr)
{
    int          family;
    int          addrLen;
    int          nameLen;
    int          recordLen;
    int          errorCode;
    tt8Bit       addrBuf[16];
    tt8BitPtr    origDataPtr; /* used to compute length of RR. */
    tt16Bit      recClass;
    tt16Bit      recLength;
    tt16Bit      recType;

    origDataPtr = dataPtr;

    addrLen = 0; /* Avoid compiler warning */
    if (queryType == TM_DNS_QTYPE_A)
    {
        family = AF_INET;
        addrLen = 4;
    }
    else if (queryType == TM_DNS_QTYPE_AAAA)
    {
        family = AF_INET6;
        addrLen = 16;
    }
    else
    {
/* must be a CNAME record, so no address is used. */
        family = 0;
    }

/* For A/AAAA records, convert address from ASCII to binary. */
    if (family > 0)
    {
        errorCode = inet_pton(family,
                              (const char TM_FAR *) addrStrPtr,
                              (void TM_FAR *) addrBuf);
        tm_assert(tfxAddResponseRec, errorCode == 1);
    }

/* First, copy the resource record name. */
    tm_strcpy(dataPtr, rrNamePtr);
    dataPtr += (int)tm_strlen(rrNamePtr) + 1;

/* Resource record type (A, MX, etc) */
    recType = htons(queryType);
    tm_bcopy(&recType, dataPtr, 2);
    dataPtr += 2;

/* Resource record class (should always be 'IN') */
    recClass = htons(1);
    tm_bcopy(&recClass, dataPtr, 2);
    dataPtr += 2;

/* Time that this record is allowed to be cached */
    ttl = htonl(ttl);
    tm_bcopy(&ttl, dataPtr, 4);
    dataPtr += 4;

    if (    (queryType == TM_DNS_QTYPE_A)
         || (queryType == TM_DNS_QTYPE_AAAA) )
    {
/* Length of record data that follows this header */
        recLength = htons(addrLen);
        tm_bcopy(&recLength, dataPtr, 2);
        dataPtr += 2;

/* Copy the record data */
        tm_bcopy(addrBuf, dataPtr, addrLen);

        dataPtr += addrLen;
    }
    else if (   (queryType == TM_DNS_QTYPE_CNAME)
             || (queryType == TM_DNS_QTYPE_REV))
    {
        nameLen = (int)tm_strlen(addrStrPtr) + 1;

        /* Length of record data that follows this header */
        recLength = htons(nameLen);
        tm_bcopy(&recLength, dataPtr, 2);
        dataPtr += 2;

        tm_bcopy(addrStrPtr, dataPtr, nameLen);

        dataPtr += nameLen;
    }

    recordLen = (int)(ttUserPtrCastToInt)(dataPtr - origDataPtr);

    return recordLen;
}


static ttUserMessage tfxDnsCreateReply( tt8BitPtr    requestBuf,
                                        tt16Bit      queryType,
                                        tt8BitPtr    ipStr1,
                                        tt8BitPtr    ipStr2,
                                        tt8BitPtr    canonRevStr,
                                        tt16Bit      dnsFlags,
                                        int TM_FAR * dataLengthPtr)
{
    ttDnsHeaderPtr  dnsHdrPtr;
    ttUserMessage   bufferHandle;
    tt8BitPtr       questionDomainPtr;
    tt8BitPtr       dataPtr;
    int             addrLength;
    int             allocSize;
    int             dataLength;
    int             recLength;
    tt16Bit         id;
    tt16Bit         answerCount;

    questionDomainPtr = (tt8BitPtr) (requestBuf + TM_DNS_HEADER_SIZE);

    tm_bcopy(requestBuf, &id, 2);

    if (queryType == TM_DNS_QTYPE_A)
    {
        addrLength = 5;
    }
    else
    {
        addrLength = 16;
    }

/* Allocate a buffer to form the response in. */
    allocSize = TM_DNS_HEADER_SIZE;
    answerCount = 0;
    if (canonRevStr != 0)
    {
        allocSize += (int)tm_strlen(canonRevStr) + TM_DNS_RR_HDR_LEN +
            (int)tm_strlen(questionDomainPtr) + 2;
        answerCount++;
    }
    if (ipStr1 != 0)
    {
        allocSize +=
            TM_DNS_RR_HDR_LEN + addrLength + (int)tm_strlen(questionDomainPtr) + 1;
        answerCount++;
    }
    if (ipStr2 != 0)
    {
        allocSize +=
            TM_DNS_RR_HDR_LEN + addrLength + (int)tm_strlen(questionDomainPtr) + 1;
        answerCount++;
    }

    bufferHandle = tfGetZeroCopyBuffer(allocSize,
                                       (char TM_FAR * TM_FAR *) &dataPtr);
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);

/* Clear out the DNS header. */
    tm_bzero(dataPtr, allocSize);

    dnsHdrPtr = (ttDnsHeaderPtr) dataPtr;

    dnsHdrPtr->dnsId = id;

/* Set flags to response. */
    dnsHdrPtr->dnsFlags = htons(TM_DNS_HDR_RESP);
    dnsHdrPtr->dnsFlags |= htons(dnsFlags);

/* Clear "question count" field */
    dnsHdrPtr->dnsAnswerCount = htons(answerCount);

    dataPtr += TM_DNS_HEADER_SIZE;
    dataLength = TM_DNS_HEADER_SIZE;

    if (ipStr1 != 0)
    {
        recLength = tfxAddResponseRec(dataPtr,
                                      queryType,
                                      TM_TEST_TTL_LONG,
                                      ipStr1,
                                      questionDomainPtr);
        tm_assert(tfxDnsTestQueryResponse, recLength > 0);

        dataPtr += recLength;
        dataLength += recLength;
    }

    if (canonRevStr != 0)
    {
        if (queryType == TM_DNS_QTYPE_REV)
        {
/* Add the CNAME or PTR record. */
            recLength = tfxAddResponseRec(dataPtr,
                                          queryType,
                                          TM_TEST_TTL_SHORT,
                                          canonRevStr,
                                          questionDomainPtr);
        }
        else
        {
            recLength = tfxAddResponseRec(dataPtr,
                                          TM_DNS_QTYPE_CNAME,
                                          TM_TEST_TTL_LONG,
                                          canonRevStr,
                                          questionDomainPtr);
        }
        tm_assert(tfxDnsTestQueryResponse, recLength > 0);

        dataPtr += recLength;
        dataLength += recLength;
    }

    if (ipStr2 != 0)
    {
        recLength = tfxAddResponseRec(dataPtr,
                                      queryType,
                                      TM_TEST_TTL_SHORT,
                                      ipStr2,
                                      questionDomainPtr);
        tm_assert(tfxDnsTestQueryResponse, recLength > 0);

        dataPtr += recLength;
        dataLength += recLength;
    }

    tm_assert(tfxDnsCreateReply, dataLength <= allocSize);
    *dataLengthPtr = dataLength;

    return bufferHandle;
}

/*
 * Tests for 'Old' DNS API (for IPv4 only)
 */

/*
 * DNS Query Type tests:
 *
 * Test series #1: Normal queries.
 *     Test #1.1: Normal query (A, AAAA)
 *     Test #1.2: Verify that if we immediately query again for the same address,
 *                the cache entry is used, and no request is sent.
 *     Test #1.3: Verify that the cache entry times out correctly.
 * Test series #2: Verify that IPv6 query is sent when AF_UNSPEC is
 *                 provided as address family.
 * Test #2.1 unspecified address family
 * Test series #3: MX records
 *          3.1: Proper return code tests for tfGetMailHost()
 *          3.2: Proper return code tests for tfGetNextMailHost()
 */
static void tfxDnsOldTestQueryType(void)
{
    struct addrinfo TM_FAR *    addrInfoPtr;
    tt8BitPtr                   dataPtr;
    struct addrinfo             addrHint;
    char                        inputBuf[TM_INPUT_BUF_LEN];
    tt4IpAddress                ip4Addr;
    int                         errorCode;
    tt16Bit                     queryType;
    unsigned short              mxPref;

    tm_bzero(&addrHint, sizeof(struct addrinfo));

    tfSetCurrentContext(tvContextHandleClient);

/* setup DNS resolver with a single server. */
    errorCode =
        tfDnsSetServer(tv4ServerSockAddr[0].addr.ipv4.sin_addr.s_addr, 0);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_ENOERROR);

/*
 * Test series #1: Normal queries.
 */

/* Test #1.1: Normal query (A) */

    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsOldTestQueryType, errorCode > 0);

/* We know the format of the query, and we know that we'll only be sending
   a single question with this query, so rather than parse through the whole
   packet just to determine what the question type is, we look from the end
   of the packet.  We have to do a 'bcopy' since there's no guarantee on
   alignment... */
    dataPtr = (tt8BitPtr) inputBuf + errorCode - 4;
    tm_bcopy(dataPtr, &queryType, sizeof(queryType));

    queryType = ntohs(queryType);

/* Make sure that the query was an AAAA (IPv6) type */
    tm_assert(tfxDnsOldTestQueryType, queryType == TM_DNS_QTYPE_A);

/* Test #1.2: Verify that if we immediately query again for the same address,
 *            the cache entry is used, and no request is sent.
 */
    tfSetCurrentContext(tvContextHandleClient);
    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_SOCKET_ERROR);

    errorCode = tfGetSocketError(tlServerSd);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EWOULDBLOCK);

/* Test #1.3: Verify that the cache entry times out correctly. */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_ETIMEDOUT);

/*
 * Test series #2: Verify that IPv6 query is sent when AF_UNSPEC is
 *                 provided as address family.
 */
    tfSetCurrentContext(tvContextHandleClient);
    addrHint.ai_flags = 0;
    addrHint.ai_family = AF_UNSPEC;

/* Test #2.1 unspecified address family */
    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);

    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          (struct sockaddr TM_FAR *) 0,
                          (int TM_FAR *) 0 );
    tm_assert(tfxDnsOldTestQueryType, errorCode > 0);
    dataPtr = (tt8BitPtr) inputBuf + errorCode - 4;
    tm_bcopy(dataPtr, &queryType, sizeof(queryType));

    queryType = ntohs(queryType);
/* IPv6 is not enabled, so default should be IPv4 */
    tm_assert(tfxDnsOldTestQueryType, queryType == TM_DNS_QTYPE_A);

/* Make sure that this request times out... */
    tfSetCurrentContext(tvContextHandleClient);
    tfxDnsRunTimer();

    errorCode = getaddrinfo( testHostnameOne, 0, &addrHint, &addrInfoPtr );
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_ETIMEDOUT);

/* Test series #3: MX records */
    tfDnsCacheInvalidate();

/* 3.1: Proper return code tests for tfGetMailHost() */
    ip4Addr = 0;
    mxPref = 0;
    errorCode = tfDnsGetMailHost((const char *)0, &ip4Addr, &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);
    errorCode = tfDnsGetMailHost(testCanonName, (tt4IpAddressPtr)0, &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);
    errorCode = tfDnsGetMailHost(testCanonName, &ip4Addr, (unsigned short *)0);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);

/* 3.2: Proper return code tests for tfGetNextMailHost() */
    errorCode = tfDnsGetNextMailHost((const char *)0,
                                     ip4Addr,
                                     mxPref,
                                     &ip4Addr,
                                     &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);
    errorCode = tfDnsGetNextMailHost(testCanonName,
                                     (tt4IpAddress)0,
                                     mxPref,
                                     &ip4Addr,
                                     &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);
    errorCode = tfDnsGetNextMailHost(testCanonName,
                                     ip4Addr,
                                     mxPref,
                                     (tt4IpAddressPtr)0,
                                     &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);
    errorCode = tfDnsGetNextMailHost(testCanonName,
                                     ip4Addr,
                                     mxPref,
                                     &ip4Addr,
                                     (unsigned short *)0);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EINVAL);

    return;
}

static void tfxDnsOldTestQueryResponse(void)
{
    ttSockAddrPtrUnion      sockAddr;
    ttUserMessage           bufferHandle;
    struct sockaddr_storage clientAddr;
    tt4IpAddress            ip4Addr;
    tt4IpAddress            testIp4Addr;
    int                     errorCode;
    int                     addrLength;
    int                     dataLength;
    int                     timeoutCount;
    unsigned short          mxPref;
    char                    inputBuf[TM_INPUT_BUF_LEN];
    tt8Bit                  match;

    tfSetCurrentContext(tvContextHandleClient);

/* setup DNS resolver with a single server. */
    errorCode =
        tfDnsSetServer(tv4ServerSockAddr[0].addr.ipv4.sin_addr.s_addr, 0);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_ENOERROR);

/*
 * Test series #1: Single protocol family.  Send a normal A or AAAA query.
 *                 The "server" will respond with an answer that includes
 *                 multiple answer records as well as a CNAME record.
 *                 Each individual test below then verifies that various
 *                 parts of this response were received correctly.
 */
    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsOldTestQueryResponse, errorCode > 0);

    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      TM_DNS_QTYPE_A,
                                      (tt8BitPtr) ipv4TestAddr,
                                      (tt8BitPtr) ipv4TestAddr2,
                                      (tt8BitPtr) testCanonRecName,
                                      0,
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsOldTestQueryResponse, bufferHandle != 0);

    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

    tfSetCurrentContext(tvContextHandleClient);

    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_ENOERROR);

/* Test #1.1: Verify that addresses are returned correctly */

/* Verify that the addresses match the expected values.  */
    errorCode = inet_pton(AF_INET,
                          ipv4TestAddr2,
                          &testIp4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == 1);

/* The old API's now return the IP address in network byte order. */
    match = (tt8Bit) tm_ip_match(ip4Addr, testIp4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, match != 0);

/* Test #1.4: Verify that the TTL received in the response is used
 *            correctly.  Verify that cache entries are not timed out
 *            too early or too late.
 */

/* Wait half of the TTL time, try an indentical query, and verify that no
 * new request packet is sent.
 */
    timeoutCount = (TM_TEST_TTL_SHORT*1000) / tvTimerTickLength;
    timeoutCount = timeoutCount >> 1;

    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_ENOERROR);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_SOCKET_ERROR);

    errorCode = tfGetSocketError(tlServerSd);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_EWOULDBLOCK);

/* Allow the TTL to expire, try an identical query, and verify that a new
 * request packet is sent (the cache entry times out).
 */
    tfSetCurrentContext(tvContextHandleClient);

    timeoutCount = (TM_TEST_TTL_SHORT*1000) / tvTimerTickLength;
    timeoutCount = timeoutCount >> 1;

    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }

    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_EWOULDBLOCK);

/* Allow the server to process the query. */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestQueryResponse, errorCode > 0);

/* Don't respond, and allow the request to time out. */
    tfSetCurrentContext(tvContextHandleClient);

    tfxDnsRunTimer();

/* Get error code for this hostname, so entry can be deleted. */
    errorCode = tfDnsGetHostByName( testHostnameOne, &ip4Addr);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_ETIMEDOUT);


/* 3.3: Proper MX record parsing and sorting */
    tfDnsCacheInvalidate();
    errorCode = tfDnsGetMailHost("example.com", &ip4Addr, &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_EWOULDBLOCK);

    tfSetCurrentContext(tvContextHandleServer);
    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom(tlServerSd,
                         inputBuf,
                         sizeof(inputBuf),
                         0,
                         sockAddr.sockPtr,
                         &addrLength);
    tm_assert(tfxDnsTestQueryResponse, errorCode > 0);

/* Copy the Transaction ID from the query */
    tm_bcopy(inputBuf, mxResponse, 2);

    errorCode = sendto(tlServerSd,
                       (ttCharPtr)mxResponse,
                       sizeof(mxResponse),
                       0,
                       sockAddr.sockPtr,
                       addrLength);

    tfSetCurrentContext(tvContextHandleClient);
    errorCode = tfDnsGetMailHost("example.com", &ip4Addr, &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_ENOERROR);
    tm_assert(tfxDnsOldTestQueryType, ip4Addr == inet_addr("172.30.79.10"));
    tm_assert(tfxDnsOldTestQueryType, mxPref == 10);
    errorCode = tfDnsGetNextMailHost("example.com",
                                     ip4Addr,
                                     mxPref,
                                     &ip4Addr,
                                     &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_ENOERROR);
    tm_assert(tfxDnsOldTestQueryType, ip4Addr == inet_addr("172.30.79.15"));
    tm_assert(tfxDnsOldTestQueryType, mxPref == 15);
    errorCode = tfDnsGetNextMailHost("example.com",
                                     ip4Addr,
                                     mxPref,
                                     &ip4Addr,
                                     &mxPref);
    tm_assert(tfxDnsOldTestQueryType, errorCode == TM_DNS_EANSWER);

    tfDnsCacheInvalidate();

    return;
}

static void tfxDnsOldTestReverseLookup(void)
{
    int                      errorCode;
    int                      addrLength;
    int                      dataLength;
    ttSockAddrPtrUnion       sockAddr;
    ttUserMessage            bufferHandle;
    struct sockaddr_storage  clientAddr;
    char                     hostname[TM_HOSTNAME_BUF_LEN];
    char                     inputBuf[TM_INPUT_BUF_LEN];
    tt8BitPtr                dataPtr;
    tt8Bit                   match;
    tt4IpAddress             ip4Addr;
    tt4IpAddress             testIp4Addr;
    tt32Bit                  timeoutCount;

    tfSetCurrentContext(tvContextHandleClient);

/* Make sure that any previous queries have timed out. */
    timeoutCount = (TM_TEST_TTL_SHORT*1000) / tvTimerTickLength;
    while (timeoutCount-- > 0)
    {
        tfTimerUpdate();
        tfTimerExecute();
    }


/* setup DNS resolver with a single server. */
    errorCode =
        tfDnsSetServer(tv4ServerSockAddr[0].addr.ipv4.sin_addr.s_addr, 0);
    tm_assert(tfxDnsOldTestQueryResponse, errorCode == TM_ENOERROR);

/*
 * Test series #1: verify that both INT and ARPA domains are attempted.
 *     Test #1.1: Attempt to do a reverse lookup, verify that both domains
 *                are used.
 */
    tfSetCurrentContext(tvContextHandleClient);

    errorCode = inet_pton(AF_INET,
                          ipv4TestAddr,
                          &ip4Addr);
    tm_assert(tfxDnsOldTestReverseLookup, errorCode == 1);

    errorCode = tfDnsGetHostByAddr(ip4Addr,
                                   hostname,
                                   TM_HOSTNAME_BUF_LEN);
    tm_assert(tfxDnsOldTestReverseLookup, errorCode == TM_EWOULDBLOCK);

/* Get query from client */
    tfSetCurrentContext(tvContextHandleServer);

    addrLength = sizeof(clientAddr);
    sockAddr.sockNgPtr = &clientAddr;
    errorCode = recvfrom( tlServerSd,
                          inputBuf,
                          sizeof(inputBuf),
                          0,
                          sockAddr.sockPtr,
                          &addrLength );
    tm_assert(tfxDnsTestReverseLookup, errorCode > 0);

/* Check that the domain is ".arpa". */
    dataPtr = (tt8BitPtr) (inputBuf + errorCode - 9);
    match = (tt8Bit) tm_strcmp(dataPtr, "arpa");
    tm_assert(tfxDnsTestReverseLookup, match == 0);

    tfSetCurrentContext(tvContextHandleServer);

/* Send response to client. */
    bufferHandle = tfxDnsCreateReply( (tt8BitPtr) inputBuf,
                                      TM_DNS_QTYPE_REV,
                                      0,
                                      0,
                                      (tt8BitPtr) testHostnameOneRec,
                                      0,
                                      (int TM_FAR *) &dataLength );
    tm_assert(tfxDnsTestQueryResponse, bufferHandle != 0);

    errorCode = tfZeroCopySendTo(tlServerSd,
                                 bufferHandle,
                                 dataLength,
                                 0,
                                 sockAddr.sockPtr,
                                 addrLength );

/* Get results from reverse lookup */
    tfSetCurrentContext(tvContextHandleClient);

    errorCode = inet_pton(AF_INET,
                          ipv4TestAddr,
                          &testIp4Addr);
    tm_assert(tfxDnsOldTestReverseLookup, errorCode == 1);

    errorCode =  tfDnsGetHostByAddr(ip4Addr,
                                    hostname,
                                    TM_HOSTNAME_BUF_LEN);
    tm_assert(tfxDnsOldTestReverseLookup, errorCode == TM_ENOERROR);

    match = (tt8Bit) tm_strcmp(hostname, testHostnameOne);
    tm_assert(tfxDnsOldTestReverseLookup, match == 0);

}


#if (defined(TM_USE_RFC3484) && defined(TM_USE_STOP_TRECK))
/* Source Address Selection tests from RFC3484 Section 10.1 */
static void tfxDnsTestRfc3484(void)
{
    ttUserContext               contextHandle;
    ttUserContext               origContextHandle;
    ttUserLinkLayer             linkLayerHandle;
    ttUserInterface             interfaceHandle;
    ttCharPtr                   sourceAddrArray[2];
    char                        sourceAddr1[INET6_ADDRSTRLEN];
    char                        sourceAddr2[INET6_ADDRSTRLEN];
    int                         errorCode;
    int                         flags;
/* Specify the 64-bit interface ID to force the link-local address to fe80::1 */
    tt8Bit                      macAddr[] = {0x02, 0x00, 0x00, 0x00,
                                             0x00, 0x00, 0x00, 0x01 };

    origContextHandle = tfGetCurrentContext();

    sourceAddrArray[0] = sourceAddr1;
    sourceAddrArray[1] = sourceAddr2;

    contextHandle = tfCreateTreckContext();
    tm_assert(tfxDnsTestRfc3484, contextHandle != (ttUserContext)0);

    tfSetCurrentContext(contextHandle);
    errorCode = tfStartTreck();
    tm_assert(tfxDnsTestRfc3484, errorCode == TM_ENOERROR);

#ifdef TM_USE_LOGGING_LEVELS
    tfLogSetLevel(TM_LOG_MODULE_ALL, TM_LOG_LEVEL_MAX);
    tfLogSetVerbosity(TM_LOG_VERBOSE);
#endif /* TM_USE_LOGGING_LEVELS */

#ifdef TM_TEST_NULL_LINK_LAYER
    linkLayerHandle = tfUseNullLinkLayer();
#else /* ! TM_TEST_NULL_LINK_LAYER */
    linkLayerHandle = tfUseEthernet();
#endif /* ! TM_TEST_NULL_LINK_LAYER */
    tm_assert(tfxDnsTestRfc3484, linkLayerHandle != (ttUserLinkLayer)0);

    interfaceHandle = tfUseIntfMCtxDriver("RFC3484",
                                          linkLayerHandle, 
                                          &errorCode);
    tm_assert(tfxDnsTestRfc3484, interfaceHandle != (ttUserInterface)0);
    tm_assert(tfxDnsTestRfc3484, errorCode == TM_ENOERROR);

    errorCode = tf6Eui64SetInterfaceId(interfaceHandle,
                                       macAddr,
                                       TM_6_DEV_CALLED_FROM_APP);
    tm_assert(tfxDnsTestRfc3484, errorCode == TM_ENOERROR);

/* 10.1.1: Prefer appropriate scope */
    flags = 0;
    tm_strcpy(sourceAddr1, "3ffe::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             1,
                             "2001::1",
                             "3FFE::1",
                             flags);

/* 10.1.2: Prefer appropriate scope */
    flags = 0;
    tm_strcpy(sourceAddr1, "fec0::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             1,
                             "2001::1",
                             "fec0::1",
                             flags);

/* 10.1.3: Prefer appropriate scope */
    flags = 0;
    tm_strcpy(sourceAddr1, "2001::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             1,
                             "fec0::1",
                             "2001::1",
                             flags);

/* 10.1.4: Prefer appropriate scope */
    flags = 0;
    tm_strcpy(sourceAddr1, "fec0::1");
    tm_strcpy(sourceAddr2, "2001::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "ff05::1",
                             "fec0::1",
                             flags);

/* 10.1.5: Prefer same address */
    flags = TM_DNS_TEST_ADDR1_DEPRECATED;
    tm_strcpy(sourceAddr1, "2001::1"); /* deprecated */
    tm_strcpy(sourceAddr2, "2002::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "2001::1",
                             "2001::1",
                             flags);

/* 10.1.6: Prefer appropriate scope */
    flags = TM_DNS_TEST_ADDR1_DEPRECATED;
    tm_strcpy(sourceAddr1, "fec0::2"); /* deprecated */
    tm_strcpy(sourceAddr2, "2001::1");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "fec0::1",
                             "fec0::2",
                             flags);

/* 10.1.7: Longest matching prefix */
    flags = 0;
    tm_strcpy(sourceAddr1, "2001::2");
    tm_strcpy(sourceAddr2, "3ffe::2");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "2001::1",
                             "2001::2",
                             flags);

#if 0
/* TODO: Make this test functional */
/* 10.1.8: Prefer home address */
    flags = TM_DNS_TEST_ADDR1_MIP_HOME | TM_DNS_TEST_ADDR2_MIP_CAREOF;
    tm_strcpy(sourceAddr1, "3ffe::2"); /* Home */
    tm_strcpy(sourceAddr2, "2001::2"); /* Care-of */
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "2001::1",
                             "3ffe::2",
                             flags);
#endif /* 0 */

/* 10.1.9: Prefer matching label */
    flags = TM_DNS_TEST_ADDR1_TEMPORARY;
    tm_strcpy(sourceAddr1, "2002:836b:2179::d5e3:7953:13eb:22e8"); /* temporary */
    tm_strcpy(sourceAddr2, "2001::2");
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "2002:836b:2179::1",
                             "2002:836b:2179::d5e3:7953:13eb:22e8",
                             flags);

/* 10.1.10: Prefer public address */
    flags = TM_DNS_TEST_ADDR2_TEMPORARY;
    tm_strcpy(sourceAddr1, "2001::2");
    tm_strcpy(sourceAddr2, "2001::d5e3:7953:13eb:22e8"); /* temporary */
    tfxDnsTestRfc3484RunTest(interfaceHandle,
                             sourceAddrArray,
                             2,
                             "2001::d5e3:0:0:1",
#ifdef TM_6_PREFER_TEMPADDR
                             "2001::d5e3:7953:13eb:22e8",
#else /* !TM_6_PREFER_TEMPADDR */
                             "2001::2",
#endif /* !TM_6_PREFER_TEMPADDR */
                             flags);

    errorCode = tfStopTreck(contextHandle);
    tm_assert(tfxDnsTestRfc3484, errorCode == TM_ENOERROR);

    tfSetCurrentContext(origContextHandle);
    return;
}

static void tfxDnsTestRfc3484RunTest(ttUserInterface     interfaceHandle,
                                     ttCharPtrPtr        sourceIpStringPtrPtr,
                                     tt8Bit              sourceIpStringCount,
                                     ttCharPtr           destIpStringPtr,
                                     ttCharPtr           expectedIpStringPtr,
                                     int                 flags)
{
    struct sockaddr_storage     sourceIpAddr;
    tt6IpAddress                destIpAddr;
    tt6IpAddress                expectedIpAddr;
    tt6IpAddress                deprecatePrefix;
    int                         errorCode;
    int                         ipv6Flags;
    tt8Bit                      destIpAddrType;
    tt8Bit                      i;

    destIpAddrType = 0;

    errorCode = inet_pton(AF_INET6, destIpStringPtr, &destIpAddr);
    tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 1);

    for (i=0; i<sourceIpStringCount; i++)
    {
        ipv6Flags = 0;
        tm_bzero(&sourceIpAddr, sizeof(struct sockaddr_storage));
        sourceIpAddr.ss_family = AF_INET6;
        sourceIpAddr.ss_len = sizeof(struct sockaddr_storage);
        errorCode = inet_pton(AF_INET6,
                              sourceIpStringPtrPtr[i],
                              &(sourceIpAddr.addr.ipv6.sin6_addr));
        tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 1);

        if (   ((i==0) && (flags & TM_DNS_TEST_ADDR1_TEMPORARY))
            || ((i==1) && (flags & TM_DNS_TEST_ADDR2_TEMPORARY)) )
        {
            ipv6Flags |= TM_6_DEV_IP_TEMPORARY;
        }
        errorCode = tfNgConfigInterface(interfaceHandle,
                                        &sourceIpAddr,
                                        64,
                                        TM_DEV_MCAST_ENB,
                                        ipv6Flags,
                                        1,
                                        0,
                                        i);
        tm_assert(tfxDnsTestRfc3484RunTest,
                  (   (errorCode == TM_ENOERROR)
                   || (errorCode == TM_EINPROGRESS)) );

        while (!(((ttDeviceEntryPtr)interfaceHandle)->
                        dev6IpAddrFlagsArray[i] & TM_6_DEV_IP_CONFIG))
        {
            tfTimerUpdate();
            tfTimerExecute();
        }

        if (   ((i==0) && (flags & TM_DNS_TEST_ADDR1_DEPRECATED))
            || ((i==1) && (flags & TM_DNS_TEST_ADDR2_DEPRECATED)) )
        {
/* We assume 64-bit prefix len when deprecating the address */
            tm_bzero(&deprecatePrefix, sizeof(tt6IpAddress));
            deprecatePrefix.s6LAddr[0] =
                    sourceIpAddr.addr.ipv6.sin6_addr.s6LAddr[0];
            deprecatePrefix.s6LAddr[1] =
                    sourceIpAddr.addr.ipv6.sin6_addr.s6LAddr[1];
            tfx6DeprecatePrefix((ttDeviceEntryPtr)interfaceHandle,
                                &deprecatePrefix);
        }
    }

    errorCode = inet_pton(AF_INET6, expectedIpStringPtr, &expectedIpAddr);
    tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 1);

    tm_bzero(&(sourceIpAddr.addr.ipv6.sin6_addr), sizeof(tt6IpAddress));
    errorCode = tf6SelectSourceAddress(
                            interfaceHandle,
                            &(sourceIpAddr.addr.ipv6.sin6_addr),
                            (ttDeviceEntryPtrPtr)0,
                            TM_16BIT_NULL_PTR,
                            &destIpAddr,
                            &destIpAddrType,
                            0);
    tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 0);

    errorCode = tm_memcmp(&(sourceIpAddr.addr.ipv6.sin6_addr),
                          &expectedIpAddr,
                          sizeof(tt6IpAddress));
    tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 0);

    errorCode = tfCloseInterface(interfaceHandle);
    tm_assert(tfxDnsTestRfc3484RunTest, errorCode == 0);

    return;
}
#endif /* TM_USE_RFC3484 && TM_USE_STOP_TRECK */

#else /* TM_MULTIPLE_CONTEXT */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */
static int tltxDnsDummy = 0;

#endif /* TM_MULTIPLE_CONTEXT */



/***************** End Of File *****************/

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:53      For: Xilinx-Samsung
 * File Name: tudnsres.c    File Revision: 6.0.2.4 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 22741    Output checksum:   958
 */
