/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Kernel Hooks
 *              NOTE: This File is intended to be modified by the user
 *                    In order to add a new operating system
 * Filename: tukrnl.c
 * Author: Paul
 * Date Created: 11/10/97
 * $Source: test/unittest/tukrnl.c $
 *
 * Modification History
 * $Revision: 6.0.2.3 $
 * $Date: 2010/04/01 12:05:42CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/* to avoid an warning in MS VC++ header file */
struct _RPC_ASYNC_STATE;
#include <windows.h>
#include <time.h>
#define TM_BSD_CLASH

#include <stdio.h>
#include <stdlib.h>
#include <trsocket.h>
#include <trmacro.h>
#include <trtype.h>
#include <trproto.h>

/*
 * Dynamically allocate the Treck simple heap. Only called when
 * TM_DYNAMIC_CREATE_SHEAP has been #define'd.
 * User can change the size of the created heap by updating *sheapSizePtr
 * prior to calling malloc.
 * The start of the simple heap (i.e. the pointer returned) being
 * aligned on a 32-bit boundary.
 * Return:
 * Pointer to beginning of dynamically created heap.
 */
#ifdef TM_DYNAMIC_CREATE_SHEAP
ttUser32Bit TM_FAR * tfKernelSheapCreate (ttUser32Bit TM_FAR * sheapSizePtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library malloc() is not re-entrant. If malloc() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    return (ttUser32Bit TM_FAR *)(malloc((*sheapSizePtr)));
}

void tfKernelSheapFree (ttUser32Bit TM_FAR * sheapPtr)
{
/* *sheapSizePtr can be changed by the user */

/*
 * NOTE 'C' library free() is not re-entrant. If free() were to be used
 * for TM_TRECK_PREEMPTIVE_KERNEL, it would have to be protected (with a
 * mutex for example).
 */
    free(sheapPtr);
    return;
}
#endif /* TM_DYNAMIC_CREATE_SHEAP */

void  tfKernelError(const char TM_FAR *functionName,
                    const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)tfKernelPrint(TM_PRINT_TYPE_FAIL,
                        "tfKernelError! %s: %s\n",
                        functionName,
                        errorMessage);
    tm_set_current_context(contextHandle);
    tm_thread_stop;
}

void  tfKernelWarning(const char TM_FAR *functionName,
                      const char TM_FAR *errorMessage)
{
#ifdef TM_MULTIPLE_CONTEXT
    ttUserContext contextHandle;
#endif /* TM_MULTIPLE_CONTEXT */
    tm_get_current_context(contextHandle);
    (void)tfKernelPrint(
            TM_PRINT_TYPE_WARN,
            "tfKernelWarning! %s: %s\n", functionName, errorMessage);
    tm_set_current_context(contextHandle);
}

#define TM_COLOR_PRINT_MAX_LINE 512
/* print a string in the right location with good color */
int tfKernelPrint(int msgType, const char TM_FAR * format, ...)
{
    int                         nResult;
    HANDLE                      outputHandle;
    CONSOLE_SCREEN_BUFFER_INFO  bufferInfo;
    COORD *                     pPos;
    WORD                        foregroundColor;
#ifndef TM_DISABLE_ALL_FILE_LOGGING
    FILE *                      logFilePtr;
#endif /* !TM_DISABLE_ALL_FILE_LOGGING */
    char                        buffer[TM_COLOR_PRINT_MAX_LINE];
    va_list                     args;

    va_start(args, format);
    nResult = tfVSPrintF(buffer, format, args);
    va_end(args);

    outputHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    GetConsoleScreenBufferInfo(outputHandle, &bufferInfo);
    pPos = &bufferInfo.dwCursorPosition;


    switch(msgType)
    {
    case TM_PRINT_TYPE_TEST:
        foregroundColor = FOREGROUND_GREEN;
        pPos->X         = 0;
        break;
    case TM_PRINT_TYPE_INFO:
        foregroundColor = FOREGROUND_GREEN;
        break;
    case TM_PRINT_TYPE_PASS:
        foregroundColor = FOREGROUND_GREEN;
        pPos->X         = 74;
        pPos->Y         -= 1;
        break;
    case TM_PRINT_TYPE_FAIL:
        foregroundColor = FOREGROUND_RED;
        pPos->X         = 74;
        pPos->Y         -= 1;
        break;
    case TM_PRINT_TYPE_WARN:
        foregroundColor = FOREGROUND_BLUE;
        pPos->X         = 74;
        break;
    case TM_PRINT_TYPE_STAT:
        foregroundColor = FOREGROUND_GREEN | FOREGROUND_RED;
        pPos->X         = 0;
        break;
    default:
        foregroundColor = FOREGROUND_GREEN | FOREGROUND_RED |FOREGROUND_BLUE;
        pPos->X         = 0;
        pPos->Y         = 0;
        break;
    }

    if (msgType != TM_PRINT_TYPE_FILE)
    {
        SetConsoleCursorPosition(outputHandle, *pPos);

        foregroundColor |= FOREGROUND_INTENSITY;
        SetConsoleTextAttribute(
                outputHandle,
                foregroundColor);

#ifdef TM_USE_SHEAP
        printf("%s", buffer);
#else /* !TM_USE_SHEAP */
        printf("%s\n", buffer);
#endif /* !TM_USE_SHEAP */

        SetConsoleTextAttribute(
                GetStdHandle(STD_OUTPUT_HANDLE),
                FOREGROUND_GREEN | FOREGROUND_RED |FOREGROUND_BLUE);

#ifndef TM_DISABLE_ALL_FILE_LOGGING
        logFilePtr = fopen(TM_DEBUG_AUTO_LOG_FILE, "a+");
        if (logFilePtr )
        {
            fprintf(  logFilePtr , "%s\n", buffer);
            fclose(logFilePtr );
        }
#endif /* !TM_DISABLE_ALL_FILE_LOGGING */
    }
#ifndef TM_DISABLE_ALL_FILE_LOGGING
    else
    {
        logFilePtr = fopen(TM_DEBUG_AUTO_LOG_FILE, "a+");
        if (logFilePtr )
        {
            fprintf(  logFilePtr , "%s\n", buffer);
            fclose(logFilePtr );
        }
    }
#endif /* !TM_DISABLE_ALL_FILE_LOGGING */

    return nResult;
}


/*
 * Treck Stack calls this fucntion to ask for the system time. It is
 * used in the appliction protocols such as the web server and SMTP client.
 * User could get the time from the system clock, from network,
 * or from application user interface.
 *
 * Parameters:
 *  secondsPtr:
 *      Pointer to store number of seconds elapsed since midnight (00:00:00),
 *      January 1, 1970, Coordinated Universal Time (UTC).
 *      UTC is the international time standard. It is the current term for
 *      what was commonly referred to as Greenwich Meridian Time (GMT).
 *      This matches the return value of the ANSI C library
 *      function time_t time(time_t *)
 *  daysPtr:
 *      The maximum number of seconds a 32 bit unsigned integer can hold is
 *      2**32 = 4G. this is approximately 126 year.
 *      In case this value ever wrap arounds, we need another space to hold the
 *      a bigger units of time. This is the purpose of daysPtr.
 *  NOTE: The total laps of time is the days stored in daysPtr plus the seconds
 *        stored in the secondsPtr
 * Returns:
 *  TM_KERN_OKAY:   Sucessful
 *  TM_KERN_ERROR:  Failed to obtain the system time.
 *
 */
int tfKernelGetSystemTime(ttUser32BitPtr daysPtr, ttUser32BitPtr secondsPtr)
{
    int errorCode;

    errorCode = TM_KERN_OKAY;
    *daysPtr = 0;
    *secondsPtr = (ttUser32Bit) time(NULL);

    return errorCode;
}

/*
 * The user can replace Treck's internal random number generator by defining
 * TM_USE_KERNEL_RNG in trsystem.h and replacing the contents of this function
 * to return a random 32-bit value.
 */
#ifdef TM_USE_KERNEL_RNG
ttUser32Bit tfKernelGetRandom (void)
{
    ttUser32Bit retVal;

    retVal = 0;
    return retVal;
}
#endif /* TM_USE_KERNEL_RNG */

int tfKernelCreateCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    /*
    Example:
    int kernelSemaphore;
    int errorCode;

    tm_get_current_context(contextHandle);
    kernelSemaphore=kernelCreateSemaphore();
    tm_set_current_context(contextHandle);
    if(kernelSemaphore == -1)
    {
        errorCode = TM_KERN_ERROR;
    }
    else
    {
        countingSemaphore->genIntParm=kernelSemaphore;
        errorCode = TM_KERN_OKAY;
    }
    return errorCode;
    */
    TM_UNREF_IN_ARG(countingSemaphore);
    return TM_KERN_OKAY;
}

/*
 * Delete a counting semaphore
 */
int tfKernelDeleteCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    /*
    Example:
    tm_get_current_context(contextHandle);
    if (kernelDeleteSemaphore(countingSemaphore->genIntParm)==0)
    {
        errorCode = TM_KERN_OKAY;
    }
    else
    {
        errorCode = TM_KERN_ERROR;
    }
    tm_set_current_context(contextHandle);
    return errorCode;
    */
    TM_UNREF_IN_ARG(countingSemaphore);
    return TM_KERN_OKAY;
}

void  tfKernelInitialize(void)
{
    return;
}

void  tfKernelUnInitialize(void)
{
    return;
}

int tfKernelPendCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    TM_UNREF_IN_ARG(countingSemaphore);
    return TM_KERN_OKAY;
}

int tfKernelPostCountSem(ttUserGenericUnionPtr countingSemaphore)
{
    TM_UNREF_IN_ARG(countingSemaphore);
    return TM_KERN_OKAY;
}

void tfKernelTaskYield(void)
{
    return;
}

/*
 * Allocate a memory block of size from the kernel
 */
void TM_FAR * tfKernelMalloc(unsigned size)
{
    return (void TM_FAR *)(malloc((unsigned)size));
}

/*
 * Free a memory block back to the kernel
 */
void tfKernelFree(void TM_FAR *memoryBlockPtr)
{
   free(memoryBlockPtr);
   return;
}

/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:53      For: Xilinx-Samsung
 * File Name: tukrnl.c      File Revision: 6.0.2.3 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 28971    Output checksum: 22005
 */
