/*
 * Copyright Notice:
 * Copyright Treck Incorporated  1997 - 2004
 * Copyright Elmic Systems Japan 1997 - 2004
 * No portions or the complete contents of this file may be copied or
 * distributed in any form (including but not limited to printed or electronic
 * forms) without the expressed written consent of Treck Incorporated OR
 * Elmic Systems.  Copyright laws and International Treaties protect the
 * contents of this file.  Unauthorized use is prohibited.
 * All rights reserved.
 *
 * Description: Automated test code for TFTP client and server.
 *
 * Filename: tutftp.c
 * Author: Jason Schmidlapp
 * Date Created: 1/31/2003
 * $Source: test/unittest/tutftp.c $
 *
 * Modification History
 * $Revision: 6.0.2.2 $
 * $Date: 2010/01/18 06:49:55CST $
 * $Author: odile $
 * $ProjectName: /home/mks/proj/tcpip.pj $
 * $ProjectRevision: 6.0.1.23 $
 */

/*
 * Include
 */
#include <trsocket.h>
#include <trmacro.h>

#if defined(TM_MULTIPLE_CONTEXT) && defined(TM_USE_TFTP) && defined(TM_USE_TFTPD)

#include <trtype.h>
#include <trproto.h>
#include <trglobal.h>
/* for the automated testing */
#include "tutest.h"
#include "trramfs.h"

/* constant declarations and enumerations */

/* type definitions */

/* local function prototypes */

/* Test functions */
static void tfxTestTftpGet(int family, int size);
static void tfxTestTftpPut( int family, int size );
static void tfxTestTftpGetTimeout( int family );
static void tfxTestTftpPutTimeout( int family );

/* Utility functions */
static void tfxInitTftpFilesystem(const char TM_FAR * filename,
                                  tt8Bit              startData,
                                  int                 size);
static int tfxTftpRunClientServer(void);
static int tfxTftpRunClient(void);
static int tfxTftpReadFile(const char TM_FAR * filename,
                           ttCharPtr           dataBufPtr,
                           int                 size);
static int tfxTftpVerifyFile( const char TM_FAR * filename,
                              ttCharPtr fileDataBuf,
                              int size);

/* local variable definitions */

/* static const tables */

/* macro definitions */
#define TM_RAMFS_BLOCKS 256 /* 32K (one block = 128 bytes) */

#define TM_START_FILE_SIZE 256
#define TM_END_FILE_SIZE 8*1024
#define TM_FILE_SIZE_INC 1024

#define TM_TFTP_TEST_FILENAME "testfile1.dat"

/*
 * expects that some
 * sort of filesystem is available, and has already been initialized.
 */
void tfxTestTftp(void)
{

    int    size;
    int    errorCode;
    
/*
 * Initialize TFTP server.
 */
    tfSetCurrentContext(tvContextHandleServer);    

    if (tm_context(tvTftpSrvPtr) == (ttVoidPtr)0)
    {
/* Can only call tfTftpInit once, otherwise memory is leaked (because we
   re-init the lock structure). */
        tfTftpdInit();
    }

    errorCode = tfTftpdUserStart(1,  /* maximum connections */
                                 1,  /* send timeout, 1 second */
                                 60, /* total timeout, 60 seconds */
                                 TM_BLOCKING_OFF);
    tm_assert(tfxTestTftp, errorCode == TM_ENOERROR);

    tfSetCurrentContext(tvContextHandleClient);

/*
 * Initialize the TFTP client.  Can only call tfTftpInit once, otherwise
 * memory is leaked (because we re-init the lock structure).
 */
    if (tm_context(tvTftpClientPtr) == (ttVoidPtr)0)
    {
        tfTftpInit();
    }

/* Run the transfer tests. */
    size = TM_START_FILE_SIZE;
    while (size <= TM_END_FILE_SIZE)
    {
        tm_test_with_2_param_int(tfxTestTftpGet, AF_INET, size);
        tm_test_with_2_param_int(tfxTestTftpPut, AF_INET, size);


        size += TM_FILE_SIZE_INC;
    }
    
/*
 * Stop the TFTP server.
 */
    tfSetCurrentContext(tvContextHandleServer);        

    errorCode = tfTftpdUserStop();
    tm_assert(tfxTestTftp, errorCode == TM_ENOERROR);

/* Run the timeout tests. */

    tm_test_with_1_param(tfxTestTftpGetTimeout, AF_INET);
    tm_test_with_1_param(tfxTestTftpPutTimeout, AF_INET);

    
}

/* Initialize the filesystem on the TFTP server, by creating a file with
   sequential data starting at a random point. */
static void tfxInitTftpFilesystem(const char TM_FAR * filename,
                                  tt8Bit              startData,
                                  int                 size)
{
    int           errorCode;
    int           i;
    ttVoidPtr     filePtr;
    tt8BitPtr     dataPtr;
    tt8BitPtr     currDataPtr;    
    tt8Bit        data;
    
    
    tfSetCurrentContext(tvContextHandleServer);

/* Generate random file */
    dataPtr = (tt8BitPtr) tm_get_raw_buffer(size);
    tm_assert(tfxInitTftpFilesystem, dataPtr != TM_8BIT_NULL_PTR);

    data = startData;
    currDataPtr = dataPtr;
    for (i=0;i<size;i++)
    {
        *currDataPtr++ = data;
        data = (tt8Bit) ((data + 1) % 255);
    }
   
    
/* Open file */
    filePtr = tfFSOpenFile( (ttVoidPtr) 0,
                            filename,
                            TM_FS_WRITE,
                            TM_TYPE_ASCII,
                            TM_STRU_STREAM );
    tm_assert(tfxInitTftpFilesystem, filePtr != TM_VOID_NULL_PTR);

/* Write test file */
    errorCode = tfFSWriteFile( (ttVoidPtr) 0,
                               filePtr,
                               (const char TM_FAR *) dataPtr,
                               size );
    tm_assert(tfxInitTftpFilesystem, errorCode != TM_SOCKET_ERROR);

/* Close test file */
    errorCode = tfFSCloseFile( TM_VOID_NULL_PTR,
                               filePtr );
    tm_assert(tfxInitTftpFilesystem, errorCode == TM_ENOERROR);
    
    tm_free_raw_buffer((ttRawBufferPtr)dataPtr);
}

/* Read a file from the filessystem, verify the length. */
static int tfxTftpReadFile(const char TM_FAR * filename,
                           ttCharPtr           dataBufPtr,
                           int                 size)
{
    ttVoidPtr filePtr;
    int copyDataLength;
    int retCode;
    int dataLength;
    int errorCode;
    ttCharPtr copyDataPtr;
    
    /* Open file */
    filePtr = tfFSOpenFile( (ttVoidPtr) 0,
                            filename,
                            TM_FS_READ,
                            TM_TYPE_ASCII,
                            TM_STRU_STREAM );
    tm_assert(tfxTftpReadFile, filePtr != TM_VOID_NULL_PTR);

    copyDataLength = size;
    copyDataPtr = dataBufPtr;
    dataLength = 0;
    do
    {
        retCode = tfFSReadFile( TM_VOID_NULL_PTR,
                                filePtr,
                                copyDataPtr,
                                copyDataLength );

        if (retCode > 0)
        {
            copyDataPtr += retCode;
            copyDataLength -= retCode;
            dataLength += retCode;
        }
        else
        {
            break;
        }
        
    } while (copyDataLength != 0);

/* Try to read from the file again - since we're at the end of the file,
   it should return zero... */
    errorCode = tfFSReadFile( TM_VOID_NULL_PTR,
                              filePtr,
                              copyDataPtr,
                              copyDataLength );
    tm_assert(tfxTftpReadFile, errorCode == TM_ENOERROR);
    
/* Close test file */
    errorCode = tfFSCloseFile( TM_VOID_NULL_PTR,
                               filePtr );
    tm_assert(tfxTftpReadFile, errorCode == TM_ENOERROR);
    
    return dataLength;
}

/* Verify that the given file matches the data in the buffer. */
static int tfxTftpVerifyFile( const char TM_FAR * filename,
                              ttCharPtr fileDataBuf,
                              int size)
{
    int errorCode;
    int i;
    ttCharPtr realFileDataBuf;
    
    realFileDataBuf = (char TM_FAR *) tm_get_raw_buffer(size);
    tm_assert(tfxTestTftpGet, realFileDataBuf != 0);
    
    errorCode = tfxTftpReadFile( filename,
                                 realFileDataBuf,
                                 size );
    tm_assert(tfxTftpVerifyFile, errorCode == size);

    errorCode = TM_ENOERROR;
    for (i=0;i<size;i++)
    {
        if (realFileDataBuf[i] != fileDataBuf[i])
        {
            errorCode = TM_EINVAL;
            break;
        }
    }
    
    tm_free_raw_buffer( (ttRawBufferPtr)realFileDataBuf );
    
    return errorCode;
}

/*
 * Tests that a TFTP GET command works correctly, both on the client and
 * server side.  Verify that the file transferred is identical to the
 * original file.
 */ 
static void tfxTestTftpGet( int family, int size )
{
    struct sockaddr_storage remoteAddr;
    ttCharPtr               fileDataBuf;
    int                     errorCode;
    int                     i;
    tt8Bit                  randomData;

    tfSetCurrentContext(tvContextHandleServer);
    
    randomData = (tt8Bit) (tfGetRandom() % 255);
    tfxInitTftpFilesystem(TM_TFTP_TEST_FILENAME,
                          randomData,
                          size);
    
    tfSetCurrentContext(tvContextHandleClient);

    if (family == AF_INET)
    {
#ifdef TM_4_USE_SCOPE_ID
        tm_bcopy(&tv4MappedServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4MappedServerSockAddr[0]));
#else /* ! TM_4_USE_SCOPE_ID */
        tm_bcopy(&tv4ServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4ServerSockAddr[0]));
#endif /* ! TM_4_USE_SCOPE_ID */
    }
    else
    {
    }
        

    remoteAddr.ss_port = htons(69);

/* Get working buffer and clear it out. */
    fileDataBuf = (char TM_FAR *) tm_get_raw_buffer(size);
    tm_assert(tfxTestTftpGet, fileDataBuf != 0);
    
    for (i=0;i<size;i++)
    {
        fileDataBuf[i] = 0;
    }
    
    errorCode = tfTftpGet( TM_TFTP_TEST_FILENAME,
                           (struct sockaddr TM_FAR *) &remoteAddr,
                           fileDataBuf,
                           size,
                           TM_TYPE_BINARY,
                           TM_BLOCKING_OFF );
    tm_assert(tfxTestTftpGet, errorCode == TM_ENOERROR);

/* Run the client & server, allow the command to complete (successfully) */    
    errorCode = tfxTftpRunClientServer();
    tm_assert(tfxTestTftpGet, errorCode == size);

    tfSetCurrentContext(tvContextHandleServer);

/* Verify that the transferred file matches the original. */
    errorCode = tfxTftpVerifyFile( TM_TFTP_TEST_FILENAME, fileDataBuf, size );
    tm_assert(tfxTestTftpGet, errorCode == TM_ENOERROR);    

/* Remove the original file. */
    errorCode = tfFSDeleteFile( TM_VOID_NULL_PTR,
                                TM_TFTP_TEST_FILENAME );
    tm_assert(tfxTestTftpGet, errorCode == TM_ENOERROR);
    
    tfSetCurrentContext(tvContextHandleClient);
/* Free file data buffer in client context */
    tm_free_raw_buffer( (ttRawBufferPtr)fileDataBuf );
}

/*
 * Tests that a TFTP PUT command works correctly, both on the client and
 * server side.  Verify that the file transferred is identical to the
 * original file.
 */ 
static void tfxTestTftpPut( int family, int size )
{
    struct sockaddr_storage remoteAddr;
    ttCharPtr               fileDataBuf;
    int                     errorCode;
    int                     i;
    tt8Bit                  dataChar;
    ttUser32Bit             retCode;
    
    tfSetCurrentContext(tvContextHandleClient);

    if (family == AF_INET)
    {
#ifdef TM_4_USE_SCOPE_ID
        tm_bcopy(&tv4MappedServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4MappedServerSockAddr[0]));
#else /* ! TM_4_USE_SCOPE_ID */
        tm_bcopy(&tv4ServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4ServerSockAddr[0]));
#endif /* ! TM_4_USE_SCOPE_ID */
    }
    else
    {
    }

    remoteAddr.ss_port = htons(69);

/* Get working buffer and fill it with sequential data, starting at an
   random value. */
    fileDataBuf = (char TM_FAR *) tm_get_raw_buffer(size);
    tm_assert(tfxTestTftpPut, fileDataBuf != 0);

    dataChar = (tt8Bit) (tfGetRandom() % 255);
    for (i=0;i<size;i++)
    {
        fileDataBuf[i] = dataChar;
        dataChar = (tt8Bit) ((dataChar + 1) % 255);
    }

    retCode = tfTftpPut( TM_TFTP_TEST_FILENAME,
                         (struct sockaddr TM_FAR *) &remoteAddr,
                         fileDataBuf,
                         size,
                         TM_TYPE_BINARY,
                         TM_BLOCKING_OFF );
    tm_assert(tfxTestTftpPut, retCode == (ttUser32Bit)0);

    tfSetCurrentContext(tvContextHandleServer);

/* Run the client & server, allow the command to complete (successfully) */
    errorCode = tfxTftpRunClientServer();
    tm_assert(tfxTestTftpPut, errorCode == TM_ENOERROR);

/* Verify that the received file matches the original file. */
    errorCode = tfxTftpVerifyFile( TM_TFTP_TEST_FILENAME, fileDataBuf, size );
    tm_assert(tfxTestTftpPut, errorCode == TM_ENOERROR);    

/* Remove the original file */
    errorCode = tfFSDeleteFile( TM_VOID_NULL_PTR,
                                TM_TFTP_TEST_FILENAME );
    tm_assert(tfxTestTftpPut, errorCode == TM_ENOERROR);
    
    tfSetCurrentContext(tvContextHandleClient);
/* Free the working buffer in client context. */
    tm_free_raw_buffer( (ttRawBufferPtr)fileDataBuf );
    
}

/*
 * Tests that a GET command by the client correctly times out when no
 * response is received from the TFTP server.
 */ 
static void tfxTestTftpGetTimeout( int family )
{
    struct sockaddr_storage remoteAddr;
    int                     errorCode;
/* Size doesn't matter, as we're not going to receive any data back... */
    char                    fileDataBuf[16];

    tfSetCurrentContext(tvContextHandleClient);

    if (family == AF_INET)
    {
#ifdef TM_4_USE_SCOPE_ID
        tm_bcopy(&tv4MappedServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4MappedServerSockAddr[0]));
#else /* ! TM_4_USE_SCOPE_ID */
        tm_bcopy(&tv4ServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4ServerSockAddr[0]));
#endif /* ! TM_4_USE_SCOPE_ID */
    }
    else
    {
    }

    remoteAddr.ss_port = htons(69);

    errorCode = tfTftpGet( TM_TFTP_TEST_FILENAME,
                           (struct sockaddr TM_FAR *) &remoteAddr,
                           fileDataBuf,
                           16,
                           TM_TYPE_BINARY,
                           TM_BLOCKING_OFF );
    tm_assert(tfxTestTftpGetTimeout, errorCode == TM_ENOERROR);

/* Run the client and allow the GET command to timeout, check return value. */
    errorCode = tfxTftpRunClient();
    tm_assert(tfxTestTftpGetTimeout, errorCode == TM_TFTP_TIMEOUT);
}

/*
 * Tests that a PUT command by the client correctly times out when no
 * response is received from the TFTP server.
 */ 
static void tfxTestTftpPutTimeout( int family )
{
    struct sockaddr_storage remoteAddr;
    int                     errorCode;
/* Size doesn't matter, as we're not going to receive any data back... */
    char                    fileDataBuf[16];

    tfSetCurrentContext(tvContextHandleClient);

    if (family == AF_INET)
    {
#ifdef TM_4_USE_SCOPE_ID
        tm_bcopy(&tv4MappedServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4MappedServerSockAddr[0]));
#else /* ! TM_4_USE_SCOPE_ID */
        tm_bcopy(&tv4ServerSockAddr[0],
                 &remoteAddr,
                 sizeof(tv4ServerSockAddr[0]));
#endif /* ! TM_4_USE_SCOPE_ID */
    }
    else
    {
    }

    remoteAddr.ss_port = htons(69);
    
    errorCode = tfTftpPut( TM_TFTP_TEST_FILENAME,
                           (struct sockaddr TM_FAR *) &remoteAddr,
                           fileDataBuf,
                           16,
                           TM_TYPE_BINARY,
                           TM_BLOCKING_OFF );
    tm_assert(tfxTestTftpPutTimeout, errorCode == TM_ENOERROR);

/* Run the client and allow the PUT command to timeout, check return value. */
    errorCode = tfxTftpRunClient();
    tm_assert(tfxTestTftpPutTimeout, errorCode == TM_TFTP_TIMEOUT);
    
}

/*
 * Runs the TFTP client in blocking mode.
 * This allows the client until the current command is complete.  Also runs
 * the timer, which allows the command the time out.
 */ 
static int tfxTftpRunClient(void)
{
    int errorCode;
    
    do
    {
        tfSetCurrentContext(tvContextHandleClient);    
        errorCode = tfTftpUserExecute();

        tfTimerUpdate();
        tfTimerExecute();
        
    } while (errorCode == TM_TFTP_EXECUT);

    return errorCode;

}

/*
 * Runs the TFTP client and TFTP server in blocking mode.
 * This allows the client and server to execute until the current command
 * is complete.
 */ 
static int tfxTftpRunClientServer(void)
{
    int errorCode;
    
    do
    {
        tfSetCurrentContext(tvContextHandleServer);    
        errorCode = tfTftpdUserExecute();

        tfSetCurrentContext(tvContextHandleClient);    
        errorCode = tfTftpUserExecute();
        
    } while (errorCode == TM_TFTP_EXECUT);

    return errorCode;
}

#else /* ! (TM_MULTIPLE_CONTEXT && TM_USE_TFTP && TM_USE_TFTPD) */

#ifdef TM_LINT
LINT_UNUSED_HEADER
#endif /* TM_LINT */
static int tltfxFtpDummy = 0;

#endif /* ! (TM_MULTIPLE_CONTEXT && TM_USE_TFTP && TM_USE_TFTPD) */


/* Processed by Treck Shipping Tool v2.60 
 * 2012/02/06 10:46:54      For: Xilinx-Samsung
 * File Name: tutftp.c      File Revision: 6.0.2.2 
 * Project Name: tcpip.pj   Project Release: 6.0.1.23
 * Input checksum: 48552    Output checksum:  6866
 */
