#include <trsocket.h>
#include "cache.h"
#ifndef __PPC__
#include "mb_interface.h"
#endif


/******************************************************************************
 * Perform any necessary setups to the data/instruction caches
 *****************************************************************************/
void SetupCache(void)
{
#ifdef __PPC__
    static Xuint32 Iccr = 0;
    static Xuint32 Dccr = 0;

    /* Enable instruction and data cache for application space */
    Iccr = cacheGetSegment(XPAR_DDR2_SDRAM_MEM_BASEADDR);
    Dccr = Iccr;
    XCache_EnableICache(Iccr);
    XCache_EnableDCache(Dccr);
#else
    microblaze_invalidate_icache();
    microblaze_enable_icache();
    microblaze_invalidate_dcache();
    microblaze_enable_dcache();
#endif
}


/*******************************************************************************
 * Intg_CacheInvalidate
 *
 * Description:
 *   Invalidate the data cache around the given address.
 *
 * Parameters:
 *   Buffer - Starting address where cache should be invalidated.
 *   Bytes - Number of bytes to flush
 *
 * Note:
 *   This implementation is for the PPC.
 ******************************************************************************/
void cacheInvalidate(void* Buffer, unsigned Bytes)
{
#ifdef __PPC__
    unsigned Addr;
    unsigned LastAddr;
#endif

    /* Nothing to do */
    if (Bytes == 0)
    {
        return;
    }

#ifdef __PPC__
    /* Back the starting address up to the start of a cache line
     * perform cache operations until Buffer+Bytes
     */
    Addr = (unsigned)Buffer & ~(CACHE_LINE - 1);
    LastAddr = (unsigned)Buffer + Bytes;

    while(Addr < LastAddr)
    {
        /* For some reason, the standalone BSP does not have a data cache
         * invalidate function. Implement one here with inline assembly.
         * This will work only with the PPC.
         */
        __asm__ __volatile__("dcbi  0,%0; sync; isync\n" : : "r" (Addr));

        Addr += CACHE_LINE;
    }
#else

    microblaze_invalidate_dcache_range((int)Buffer, (int)Bytes);
#endif
}

void cacheFlush(void *Buffer, unsigned Bytes)
{
#ifdef __PPC__
    unsigned Addr;
    unsigned LastAddr;

    /* Nothing to do */
    if (Bytes == 0)
    {
        return;
    }

    /* Back the starting address up to the start of a cache line
     * perform cache operations until Buffer+Bytes
     */
    Addr = (unsigned)Buffer & ~(CACHE_LINE - 1);
    LastAddr = (unsigned)Buffer + Bytes;

    while(Addr < LastAddr)
    {
        __asm__ __volatile__("dcbst  0,%0; sync; isync\n" : : "r" (Addr));

        Addr += CACHE_LINE;
    }
#else
    /* Usually MicroBlaze Cache is write through. In this case do nothing. */
#ifdef XPAR_MICROBLAZE_0_DCACHE_USE_WRITEBACK
    microblaze_flush_dcache_range((int)Buffer, Bytes);
#endif /* XPAR_MICROBLAZE_0_DCACHE_USE_WRITEBACK */
#endif
}

/****************************************************************************
* Utility to map a memory segment to a 128 MB region used to enable/disable
* cache areas in the PPC405.
*
* Parameters:
*   MemSeg - Base address of memory segment to map the cache control bit for.
*
* Returns:
*   Bitmask compatible with PPC ICCR and DCCR registers. The bit set correlates
*   to the 128 MB memory segment for which the MemSeg parameter appears.
*
*****************************************************************************/
Xuint32 cacheGetSegment(Xuint32 MemSeg)
{
    Xuint32 Seg128;

    Seg128 = MemSeg >> 27;
    return(1 << (31 - Seg128));
}
#if 0
void tfPpcCacheInval(ttUserInterface         interfaceHandle,
                            char TM_FAR *            dataPtr,
                            int                        dataLength,
                            ttUserGenericUnion    unused)
{
    cacheInvalidate(dataPtr, dataLength);
}

void tfPpcCacheFlush(ttUserInterface         interfaceHandle,
                            char TM_FAR *            dataPtr,
                            int                        dataLength,
                            ttUserGenericUnion    unused)
{
    cacheFlush(dataPtr, dataLength);
}
#endif /* 0 */

#ifdef TM_USE_DCACHE_MANAGEMENT
void tfDevCacheInvalFuncCb(ttUserInterface      interfaceHandle,
                           char *               dataPtr,
                           int                  dataLength,
                           ttUserGenericUnion   userParam)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(userParam);

    cacheInvalidate(dataPtr, dataLength);

    return;
}

void tfDevCacheFlushFuncCb(ttUserInterface      interfaceHandle,
                           char *               dataPtr,
                           int                  dataLength,
                           ttUserGenericUnion   userParam)
{
    TM_UNREF_IN_ARG(interfaceHandle);
    TM_UNREF_IN_ARG(userParam);

    cacheFlush(dataPtr, dataLength);

    return;
}
#endif /* TM_USE_DCACHE_MANAGEMENT */
