/**************************************************************************
*
*     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
*     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
*     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
*     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
*     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*     FOR A PARTICULAR PURPOSE.
*
*     (c) Copyright 2008 Xilinx, Inc.
*     All rights reserved.
*
**************************************************************************/

/**************************************************************************
* Filename:     marvell_88e1111.c
*
* Description:
* Routines to control and display the status of a Marvell 88E1111 Ethernet
* PHY.
*
*
* MODIFICATION HISTORY:
*
* Ver    Who   Date     Changes
* ------ ------- -------- -----------------------------------------------------
* 1.00.a bhill   3/6/08   Initial Release
* 2.00.a pcarney 12/17/10 Added support for AXI bus
*
*
* $Id: Xilinx/Demos/app_source/ml50x/marvell_88e1111.c 1.3 2009/08/20 15:26:10EDT tcarney Exp  $
******************************************************************************/

/***************************** Include Files *********************************/
#include "trsystem.h"
#ifdef TM_TRECK_PREEMPTIVE_KERNEL
#include "xmk.h"
#include "sys/timer.h"
#endif /* TM_TRECK_PREEMPTIVE_KERNEL */

#include "xutil.h"
#include "marvell_88e1111.h"
#ifdef TM_XILINX_USE_AXI4
#include "xtr_axiethernet.h"
#else /* TM_XILINX_USE_AXI4 */
#include "xtr_lltemac.h"
#endif /* TM_XILINX_USE_AXI4 */
#include "util_time.h"


#define PHY_IDO_REG  2
#define PHY_ID1_REG  3


#ifdef TM_XILINX_USE_AXI4
static void AxiEthernetUtilPhyDelay(unsigned int Seconds);
#endif /* TM_XILINX_USE_AXI4 */

/*
 * Pointers to functions which will read/write the PHY registers.
 * These must be set with marvell_phy_setvectors() before any other functions
 * in this file are called.
 */
static PhyWrite_t *PhyWrite_fp;
static PhyRead_t  *PhyRead_fp;

extern int PhyAddr;
extern TemacDevData XpsTemac;

/***************************************************************************/
/**
 * marvell_phy_setvectors:
 * initialize read/write vectors to the driver which will provide access
 * to the PHY registers.
 *
 ***************************************************************************/
void marvell_phy_setvectors (PhyRead_t *read_fp, PhyWrite_t *write_fp)
{
    PhyWrite_fp = write_fp;
    PhyRead_fp = read_fp;
}

/***************************************************************************/
/**
 * marvell_phy_detected:
 * Returns XTRUE if Marvell PHY detected at PhyAddress, otherwise returns
 * XFALSE.
 *
 ***************************************************************************/
int marvell_phy_detected (void *phy_inst, u32 PhyAddress)
{
    u16 val;

    if (PhyRead_fp == NULL)
     {
        xil_printf("%s ERROR: PHY accessor vectors not set.\n\r",
                   __FUNCTION__);
        return XFALSE;
    }

    (*PhyRead_fp)(phy_inst, PhyAddress, MPHY_ID0_REG, &val);
    if (val == MPHY_ID0_MARVELL_OUI)
     {
        return XTRUE;
    }
    return XFALSE;
}

/***************************************************************************/
/**
 * marvell_phy_link_status:
 * Returns XTRUE if LINK is UP
 * Returns XFALSE if LINK is DOWN
 *
 ***************************************************************************/
int marvell_phy_link_status (void *phy_inst, u32 PhyAddress)
{
    u16 val;

    if (PhyRead_fp == NULL)
     {
        xil_printf("%s ERROR: PHY accessor vectors not set.\n\r",
                   __FUNCTION__);
        return XFALSE;
    }

    (*PhyRead_fp)(phy_inst, PhyAddress, MPHY_SPCFC_STATUS_REG, &val);
    if (val & MPHY_SPCFC_STAT_LINK_RT)
     {
        return XTRUE;
    }
    return XFALSE;
}

/***************************************************************************/
/**
 * marvell_phy_display_status: Prints status of PHY
 *
 ***************************************************************************/
void marvell_phy_display_status (void *phy_inst, u32 PhyAddress)
{
    u16 val, speed_status;

    if (PhyRead_fp == NULL)
     {
        xil_printf("%s ERROR: PHY accessor vectors not set.\n\r",
                   __FUNCTION__);
        return;
    }

    (*PhyRead_fp)(phy_inst, PhyAddress, MPHY_SPCFC_STATUS_REG, &val);
    if (val & MPHY_SPCFC_STAT_LINK_RT)
     {
        xil_printf("PHY Status: LINK_OK ");
    }
     else
     {
        xil_printf("PHY Status: LINK_DOWN\n\r");
        return;
    }

    speed_status = val >> MPHY_SPCFC_STAT_SPD_SHIFT;
    switch (speed_status)
     {
         case MPHY_SPCFC_STAT_SPD_RESVD:
              xil_printf("SPEED-RESERVED ");
              break;
         case MPHY_SPCFC_STAT_SPD_1000:
              xil_printf("SPEED-1000MB ");
              break;
         case MPHY_SPCFC_STAT_SPD_100:
              xil_printf("SPEED-100MB ");
              break;
         case MPHY_SPCFC_STAT_SPD_10:
              xil_printf("SPEED-10MB ");
              break;
    }

    if (val & MPHY_SPCFC_STAT_DUPLEX)
     {
        xil_printf("FULL-DUPLEX ");
    }
     else
     {
        xil_printf("HALF-DUPLEX ");
    }

    if (val & MPHY_SPCFC_STAT_SPD_DUP_RSLVD)
     {
        xil_printf("SPD_DPLX_RSLVD ");
    }
     else
     {
        xil_printf("SPD_DPLX_NOT_RSLVD ");
    }

    if (val & MPHY_SPCFC_STAT_MDIX)
     {
        xil_printf("MDIX ");
    }
     else
     {
        xil_printf("MDI ");
    }
    xil_printf("\n\r");
}


/***************************************************************************/
/**
 * marvell_phy_print_status: Easier way to print PHY status from other
 * source files
 *
 ***************************************************************************/
void marvell_phy_print_status (void){
    TemacDevData * Mac;
     Mac = &XpsTemac;

     marvell_phy_display_status (&Mac->TemacInst, PhyAddr);

}


/***************************************************************************/
/**
 * marvell_phy_autonegotitate: This function initializes the
 * Marvell 88e1111 PHY and autodetects the link speed
 *
 ***************************************************************************/
void marvell_phy_autonegotiate(void)
{
    TemacDevData * Mac;
    u16   val;

    val = 0;
    Mac = &XpsTemac;

    marvell_phy_setvectors(
#ifdef TM_XILINX_USE_AXI4
                           (PhyRead_t*)XAxiEthernet_PhyRead,
                           (PhyWrite_t*)XAxiEthernet_PhyWrite
#else /* TM_XILINX_USE_AXI4 */
                           (PhyRead_t*)XLlTemac_PhyRead,
                           (PhyWrite_t*)XLlTemac_PhyWrite
#endif /* TM_XILINX_USE_AXI4 */
                           );

/*
 * Verify that a Marvell PHY is present
 */
   val = marvell_phy_detected(&Mac->TemacInst, PhyAddr);
   if (val == XFALSE) {
       xil_printf("Marvell PHY not detected.\n\r");
       return;
   }
/* Autonegotiate speed.  The Duplex argument will be ignored*/
   marvell_phy_set_link(MPHY_USER_INPUT_SPD_AUTO, DUPLEX_FULL);

}

/***************************************************************************/
/**
 * marvell_phy_set_link: This function sets the link speed by
 * interacting with the registers on the Marvell PHY
 *
 ***************************************************************************/
void marvell_phy_set_link(int Speed, int Duplex){
     const char *Fname = "marvell_phy_set_link: ";
     TemacDevData * Mac;
     u16   val;
#ifdef TM_TRECK_NO_KERNEL
     XTime startNegotiation;
#else /* TM_TRECK_NO_KERNEL */
    time_t startNegotiation;
    time_t currentTime;
#endif /* TM_TRECK_NO_KERNEL */


     val = 0;
     Mac = &XpsTemac;

     switch(Speed){
     case MPHY_USER_INPUT_SPD_10:
         (*PhyRead_fp)(&Mac->TemacInst, PhyAddr, MPHY_CONTROL_REG, &val);
/* for 10Mbps MSB = 0, LSB = 0 */
         val &= ~MPHY_CONTROL_SPD_SEL_MSB;
         val &= ~MPHY_CONTROL_SPD_SEL_LSB;
        break;
     case MPHY_USER_INPUT_SPD_100:
         (*PhyRead_fp)(&Mac->TemacInst, PhyAddr, MPHY_CONTROL_REG, &val);
/* for 100Mbps MSB = 0, LSB = 1 */
         val &= ~MPHY_CONTROL_SPD_SEL_MSB;
         val |= MPHY_CONTROL_SPD_SEL_LSB;
        break;
     case MPHY_USER_INPUT_SPD_1000:
         (*PhyRead_fp)(&Mac->TemacInst, PhyAddr, MPHY_CONTROL_REG, &val);
/* for 1000Mbps MSB = 1, LSB = 0 */
         val |= MPHY_CONTROL_SPD_SEL_MSB;
         val &= ~MPHY_CONTROL_SPD_SEL_LSB;
        break;
     case MPHY_USER_INPUT_SPD_AUTO:
         (*PhyRead_fp)(&Mac->TemacInst, PhyAddr, MPHY_CONTROL_REG, &val);
         val |= MPHY_CONTROL_AUTONEG;
         val |= MPHY_CONTROL_RSTRT_AUTONEG;
        break;
     default:
         xil_printf("%s Unsupported speed of %d Mbps requested\n", Fname, Speed);
         break;
     }

/*
 * If autonegotation isn't being used it must be turned off and duplex must
 * be set.
 */
     if(Speed != MPHY_USER_INPUT_SPD_AUTO)
     {
        val &= ~MPHY_CONTROL_AUTONEG;
        if(Duplex == DUPLEX_HALF)
             val &= ~MPHY_CONTROL_DUPLEX;
         else
             val |= MPHY_CONTROL_DUPLEX;
     }
/* PHY must be reset after speed changes */
     val |= MPHY_CONTROL_RST;

/* write val to MPHY control register and wait for PHY*/
     (*PhyWrite_fp)(&Mac->TemacInst, PhyAddr, MPHY_CONTROL_REG, val);
/* Use timestamps to timeout phy negotiation if it takes more than 30 sec*/
#ifdef TM_TRECK_NO_KERNEL
     startNegotiation = TimeStampGet();
#else /* TM_TRECK_NO_KERNEL */
     time(&startNegotiation);
#endif /* TM_TRECK_NO_KERNEL */

     do {
          val = marvell_phy_link_status(&Mac->TemacInst, PhyAddr);
#ifdef TM_TRECK_NO_KERNEL
          if(TimeStampDiffMs(startNegotiation, TimeStampGet()) > 10000)
          {
              xil_printf("PHY negotiation unable to complete, check if board is connected via Ethernet.\n\r");
              break;
          }
#else /* TM_TRECK_NO_KERNEL */
        time(&currentTime);
/* With a 400 MHz clock (PPC) the time will rollover every 10.7 sec, so this
 * can't be more than a 10 second wait unless we start to count rollovers.
 */
        if (currentTime >= startNegotiation + 10)
        {
            xil_printf("PHY negotiation unable to complete, check if board is connected via Ethernet.\n\r");
            break;
        }
#endif /* TM_TRECK_NO_KERNEL */

     } while (val == XFALSE);

/*
 * When the device is used in a multispeed environment, the link speed must be
 * explicitly set using XLlTemac_SetOperatingSpeed() and must match the speed
 * the PHY has negotiated. If the speeds are mismatched, then the MAC will not
 * pass traffic.
 */
    if(Speed == MPHY_USER_INPUT_SPD_AUTO)
     {
          (*PhyRead_fp)(&Mac->TemacInst, PhyAddr, MPHY_SPCFC_STATUS_REG, &val);
        Speed = val >> MPHY_SPCFC_STAT_SPD_SHIFT;
        switch (Speed)
          {
            case MPHY_SPCFC_STAT_SPD_1000:
                Speed = 1000;
                break;
            case MPHY_SPCFC_STAT_SPD_100:
                Speed = 100;
                break;
            case MPHY_SPCFC_STAT_SPD_10:
                Speed = 10;
                break;
          }
    }

#ifdef TM_XILINX_USE_AXI4
    XAxiEthernet_SetOperatingSpeed(&Mac->TemacInst, Speed);
    /*
     * pdmc: From xaxiethernet_example_intr_sgdma.c
     * Setting the operating speed of the MAC needs a delay.  There
     * doesn't seem to be register to poll, so please consider this
     * during your application design.
     */
    AxiEthernetUtilPhyDelay(2);
#else /* TM_XILINX_USE_AXI4 */
    XLlTemac_SetOperatingSpeed(&Mac->TemacInst, Speed);
#endif /* TM_XILINX_USE_AXI4 */

/* Print PHY status */
    marvell_phy_display_status(&Mac->TemacInst, PhyAddr);
}

#ifdef TM_XILINX_USE_AXI4
/******************************************************************************/
/**
*
* For Microblaze we use an assembly loop that is roughly the same regardless of
* optimization level, although caches and memory access time can make the delay
* vary.  Just keep in mind that after resetting or updating the PHY modes,
* the PHY typically needs time to recover.
*
* @return   None
*
* @note     None
*
******************************************************************************/
static void AxiEthernetUtilPhyDelay(unsigned int Seconds)
{
    static int WarningFlag = 0;

    /* If MB caches are disabled or do not exist, this delay loop could
     * take minutes instead of seconds (e.g., 30x longer).  Print a warning
     * message for the user (once).  If only MB had a built-in timer!
     */
    if (((mfmsr() & 0x20) == 0) && (!WarningFlag)) {
        WarningFlag = 1;
    }

#define ITERS_PER_SEC   (XPAR_CPU_CORE_CLOCK_FREQ_HZ / 6)
    asm volatile ("\n"
            "1:               \n\t"
            "addik r7, r0, %0 \n\t"
            "2:               \n\t"
            "addik r7, r7, -1 \n\t"
            "bneid  r7, 2b    \n\t"
            "or  r0, r0, r0   \n\t"
            "bneid %1, 1b     \n\t"
            "addik %1, %1, -1 \n\t"
            :: "i"(ITERS_PER_SEC), "d" (Seconds));

}
#endif /* TM_XILINX_USE_AXI4 */
