/* $Id: */
/********************************************************************
 * Copyright(C) 2005 by Xilinx, Inc. All rights reserved.
 * This text/file contains proprietary, confidential
 * information of Xilinx, Inc., is distributed under license
 * from Xilinx, Inc., and may be used, copied and/or
 * disclosed only pursuant to the terms of a valid license
 * agreement with Xilinx, Inc.  Xilinx hereby grants you a
 * license to use this text/file solely for design, simulation,
 * implementation and creation of design files limited
 * to Xilinx devices or technologies. Use with non-Xilinx
 * devices or technologies is expressly prohibited and
 * immediately terminates your license unless covered by
 * a separate agreement.
 *
 * Xilinx is providing this design, code, or information
 * as-is solely for use in developing programs and
 * solutions for Xilinx devices, with no obligation on the
 * part of Xilinx to provide support.  By providing this design,
 * code, or information as one possible implementation of
 * this feature, application or standard, Xilinx is making no
 * representation that this implementation is free from any
 * claims of infringement.  You are responsible for
 * obtaining any rights you may require for your implementation.
 * Xilinx expressly disclaims any warranty whatsoever with
 * respect to the adequacy of the implementation, including
 * but not limited to any warranties or representations that this
 * implementation is free from claims of infringement, implied
 * warranties of merchantability or fitness for a particular
 * purpose.
 *
 * Xilinx products are not intended for use in life support
 * appliances, devices, or systems. Use in such applications is
 * expressly prohibited.
 *
 * Any modifications that are made to the Source Code are
 * done at the user's sole risk and will be unsupported.
 * The Xilinx Support Hotline does not have access to source
 * code and therefore cannot answer specific questions related
 * to source HDL.
 *
 * This copyright and support notice must be retained as part
 * of this text at all times.
 * (c) Copyright 2008 Xilinx, Inc.
 * All rights reserved.
 ********************************************************************/

/******************************************************************************
*
* util_userio.c - Various console input/output routines
*
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1     rmm  03/08/05 First release
*
*****************************************************************************/

/***************************** Include Files ********************************/
#include <stdio.h>
#include <string.h>
#include "xparameters.h"
#include "trsocket.h"
#include "app.h"


/**************************** Type Definitions ******************************/

/***************** Macros (Inline Functions) Definitions ********************/

/* Goto a lower level than inbyte to peek to see if data is present on the
 * console. This may not work if both types of UARTs are present in the
 * system.
 */
#ifdef XPAR_XUARTNS550_NUM_INSTANCES
#  include "xuartns550_l.h"
#  define INPUT_PRESENT() (XUartNs550_IsReceiveData(XPAR_RS232_UART_1_BASEADDR))
#elif defined XPAR_XUARTLITE_NUM_INSTANCES
#  include "xuartlite_l.h"
#  define INPUT_PRESENT() (!XUartLite_IsReceiveEmpty(XPAR_RS232_UART_1_BASEADDR))
#else
#  error No supported console UART present
#endif

/************************** Variable Definitions ****************************/

/************************** Function Prototypes *****************************/
extern char inbyte(void);

/*******************************************************************************
 * IsKeyPressed
 *
 * Description:
 *   Detects the presence of input on stdin. If input is present, then the
 *   first byte is read and flushed.
 *
 * Parameters:
 *   Flush - Flush the character from the console if this parameter is >=1
 *
 * Returns:
 *   Number of bytes queued
 ******************************************************************************/
int IsKeyPressed(int Flush)
{
    if (INPUT_PRESENT())
    {
        if (Flush) inbyte();
        return(1);
    }
    else
    {
        return(0);
    }
}


/*******************************************************************************
 * WaitForKeyPress
 *
 * Description:
 *   Wait for the user to enter any character on the console
 *
 * Parameters:
 *   None
 *
 * Returns:
 *   Nothing
 ******************************************************************************/
void WaitForKeyPress(void)
{
    while(!inbyte());
}

/*******************************************************************************
 * GetUserInput
 *
 * Description:
 *   Retrieve a line of input from the user. A line is defined as all characters
 *   up to a newline.
 *
 * Parameters:
 *   Prompt - Printed before string is accepted to que the user to enter something.
 *   Response - User entered string with newline stripped
 *   MaxChars - Maximum number of characters to read
 *
 * Returns:
 *   Number of characters read (excluding newline)
 *
 ******************************************************************************/
int GetUserInput(char* Prompt, char* Response, int MaxChars)
{
    int i;
    int Finished;

    /* display prompt */
    if (Prompt) xil_printf(Prompt);

    /* This basically implements a fgets function. The standalone EDK stdin
     * is apparantly buggy because it behaves differently when newline is entered
     * by itself vs. when it is entered after a number of regular chars have been
     * entered.
     *
     * Characters entered are echoed back.
     */
    Finished = 0;
    i = 0;

    while(!Finished)
    {
        /* flush out any output pending in stdout */
        fflush(stdout);

        if( i >= MaxChars){
            Response[i] = 0;
            Finished = 1;
            continue;
        }

        /* wait for a character to arrive */
#ifdef TM_TRECK_NO_KERNEL
        while (IsKeyPressed(0) == 0)
        {
            PollIO();
        }
#endif /* TM_TRECK_NO_KERNEL */
        Response[i] = inbyte();

        /* normal chars, add them to the string and keep going */
        if ((Response[i] >= 0x20) && (Response[i] <=0x7E))
        {
            xil_printf("%c", Response[i++]);
            continue;
        }

        /* control chars */
        switch(Response[i])
        {
            /* carriage return */
            case 0x0D:
                Response[i] = 0;
                Finished = 1;
                xil_printf("\n\r");
                break;

            /* backspace */
            case 0x08:
                if (i != 0)
                {
                    /* erase previous character and move cursor back one space */
                    xil_printf("\b \b");
                    Response[--i] = 0;
                }
                break;

            /* ignore all other control chars */
            default:
                continue;
        }

    }

    return i;
}

/******************************************************************************
 * Prompt the user to answer a yes/no question
 *
 * Parameters:
 *   Prompt - Question to ask
 *   DefaultYes - 1 then the default answer is yes, 0 is no
 *
 * Returns:
 *   1 if the user answers yes, 0 is no
 *****************************************************************************/
int AskYesNo(char *Prompt, int DefaultYes)
{
    int rc = DefaultYes;
    char UserResponse[5];

    while(1)
    {
        xil_printf("%s ", Prompt);

        if (DefaultYes)
        {
            xil_printf("[Y/n]: ");
        }
        else
        {
            xil_printf("[y/N]: ");
        }

        /* If the user just presses return without entering anything, return
         * default answer */
        if (GetUserInput(NULL, UserResponse, sizeof(UserResponse) - 1) == 0)
        {
            return(rc);
        }

        if (strlen(UserResponse) == 1)
        {
            switch(UserResponse[0])
            {
                case 'y':
                case 'Y':
                    return(1);

                case 'n':
                case 'N':
                    return(0);

                default:
                    break;
            }
        }
    }
}


