/* $Id: */
/********************************************************************
 * Copyright(C) 2008 by Xilinx, Inc. All rights reserved.
 * This text/file contains proprietary, confidential
 * information of Xilinx, Inc., is distributed under license
 * from Xilinx, Inc., and may be used, copied and/or
 * disclosed only pursuant to the terms of a valid license
 * agreement with Xilinx, Inc.  Xilinx hereby grants you a
 * license to use this text/file solely for design, simulation,
 * implementation and creation of design files limited
 * to Xilinx devices or technologies. Use with non-Xilinx
 * devices or technologies is expressly prohibited and
 * immediately terminates your license unless covered by
 * a separate agreement.
 *
 * Xilinx is providing this design, code, or information
 * as-is solely for use in developing programs and
 * solutions for Xilinx devices, with no obligation on the
 * part of Xilinx to provide support.  By providing this design,
 * code, or information as one possible implementation of
 * this feature, application or standard, Xilinx is making no
 * representation that this implementation is free from any
 * claims of infringement.  You are responsible for
 * obtaining any rights you may require for your implementation.
 * Xilinx expressly disclaims any warranty whatsoever with
 * respect to the adequacy of the implementation, including
 * but not limited to any warranties or representations that this
 * implementation is free from claims of infringement, implied
 * warranties of merchantability or fitness for a particular
 * purpose.
 *
 * Xilinx products are not intended for use in life support
 * appliances, devices, or systems. Use in such applications is
 * expressly prohibited.
 *
 * Any modifications that are made to the Source Code are
 * done at the user's sole risk and will be unsupported.
 * The Xilinx Support Hotline does not have access to source
 * code and therefore cannot answer specific questions related
 * to source HDL.
 *
 * This copyright and support notice must be retained as part
 * of this text at all times.
 * (c) Copyright 2008 Xilinx, Inc.
 * All rights reserved.
 ********************************************************************/

/******************************************************************************
*
* xbuffermem.c - Entry point for performance test suite
*
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1     drg  05/09/08 First release
*
*****************************************************************************/

/***************************** Include Files ********************************/
#include "xbasic_types.h"
#include "xbuffermem.h"
#include "cache.h"
#include "debug.h"
#include "defaults.h"

/************************** Constant Definitions ****************************/

        /* The amount of space needed must be greater than the size of the
         * socket used in the application.
         */
#define BUFFER_SPACE_SIZE    (2*1024*1024)

#define MEM_FREE_FLAG        0x00001
#define MEM_ALLOCED_FLAG    0x00002

/**************************** Type Definitions ******************************/
typedef struct BufHeader *BufHeadPtr;

typedef struct BufHeader{
    char * DataPtr;
    u32 DataLen;
    BufHeadPtr NextHeader;
    u32 Flags;
    BufHeadPtr  NextFree;
}HeapNode;

/***************** Macros (Inline Functions) Definitions ********************/

/************************** Function Prototypes *****************************/

/************************** Variable Definitions ****************************/

static unsigned char DriverBufferSpace[BUFFER_SPACE_SIZE];
static BufHeadPtr FreeListHead;
static BufHeadPtr LastHeaderPtr;

void xbuf_init(){
    /* Initialize the free list */
    FreeListHead = NULL;

    /* Point the LastHead to the start */
    LastHeaderPtr = (BufHeadPtr)DriverBufferSpace;

    /* Setup the start of the list, put one empty header, buffer at the beginning */
    LastHeaderPtr->DataPtr = (char *)(LastHeaderPtr + 1);
    LastHeaderPtr->DataLen = 0;
    LastHeaderPtr->NextHeader = NULL;
    LastHeaderPtr->NextFree = NULL;
    LastHeaderPtr->Flags = MEM_FREE_FLAG;
    DEBUG_LOG(DEBUG_MEM, "xbuf_init():Header Ptr %08x, Driver Space %08x, End %08x",LastHeaderPtr, DriverBufferSpace, LastHeaderPtr->DataPtr);
}

char * xbuf_alloc(u16 BufLen){
    u32 TotalLength;
    u32 EndofBuf;
    BufHeadPtr CurNodePtr, PrvNodePtr;

    if(0 == BufLen){
        DEBUG_LOG(DEBUG_ERROR, "xbuf_alloc():0 Buffer Length");
        return (NULL);
    }

    /* First check the free list for something big enough */
    CurNodePtr = FreeListHead;

    while(CurNodePtr != NULL){
        /* Is the current one big enough? */
        if(CurNodePtr->DataLen >= BufLen){
            /* It fits, don't try for the best fit!
             * Remove it from the free list
             */
            if(CurNodePtr == FreeListHead){
                /* First item in the list */
                FreeListHead = CurNodePtr->NextFree;
            }
            else{
                PrvNodePtr->NextFree = CurNodePtr->NextFree;
            }

            /* Update the flags and the memory is allocated */
            CurNodePtr->Flags = MEM_ALLOCED_FLAG;
            CurNodePtr->NextFree = NULL;

            DEBUG_LOG(DEBUG_MEM, "xbuf_alloc():Free list Last Ptr %08x, Last Data %08x, Data Len %d",CurNodePtr, CurNodePtr->DataPtr, CurNodePtr->DataLen);
            return(CurNodePtr->DataPtr);
        }

        /* Keep the previous pointer, so we can remove the current from the list
         * */
        PrvNodePtr = CurNodePtr;

        CurNodePtr = PrvNodePtr->NextFree;
    }

    /* Use the last head pointer so we don't have to traverse the list */
    if(NULL == LastHeaderPtr){
        /* Oops, this should never be null */
        DEBUG_LOG(DEBUG_MEM, "xbuf_alloc():LastHeaderPtr is null");
        return (NULL);
    }

    /*
     * To avoid cache flush and invalidate anomalies, the buffer MUST
     * start and end on a cache line boundary. Otherwise, cache
     * operations on the contents of the buffer could affect data
     * beyond the limits of the buffer.
     * The buffer header must immediately precede the buffer in memory
     * with no padding, so that the header can be easily located when
     * the buffer is freed.
     * This may add padding prior to the buffer header to align the
     * following data on a cache line.
     */
    EndofBuf = (u32)LastHeaderPtr->DataPtr + LastHeaderPtr->DataLen + sizeof(HeapNode);
    EndofBuf = ROUND_UP(EndofBuf, CACHE_LINE);
    BufLen = ROUND_UP(BufLen, CACHE_LINE);

    /* Check the length for past the end of the heap space */
    TotalLength = EndofBuf + BufLen - (u32)DriverBufferSpace;
    if(TotalLength > BUFFER_SPACE_SIZE){
        DEBUG_LOG(DEBUG_ERROR, "xbuf_alloc():Out of buffer space %08x, End %08x",TotalLength, EndofBuf);
        /* Nothing big enough for us, try to reclaim memory
        * Call this an error now, maybe we don't have to do it. */
        return (NULL);
    }

    /* Set the next Head pointer to the space preceding the data buffer */
    LastHeaderPtr->NextHeader = (BufHeadPtr)EndofBuf - 1;

    /* Move the LastHeaderPtr to the new area. */
    LastHeaderPtr = LastHeaderPtr->NextHeader;

    /* Set the BufHeader */
    LastHeaderPtr->DataPtr = (char *)EndofBuf;
    LastHeaderPtr->DataLen = BufLen;
    LastHeaderPtr->NextHeader = NULL;
    LastHeaderPtr->NextFree = NULL;
    LastHeaderPtr->Flags = MEM_ALLOCED_FLAG;

    /* Return the pointer to the data space */
    return(LastHeaderPtr->DataPtr);

}

void xbuf_free(char * BufPtr){
    BufHeadPtr CurHeader;

    /* Make sure the BufPtr matches the header */
    CurHeader = (BufHeadPtr)BufPtr - 1;

    if((MEM_ALLOCED_FLAG != CurHeader->Flags) ||
            (CurHeader->DataPtr != BufPtr)){
        /* Invalid Free, output an error */
        DEBUG_LOG(DEBUG_ERROR, "xbuf_free():Invalid free DataPtr 0x%08x, Flags %08x", CurHeader->DataPtr,CurHeader->Flags);
        return;
    }

    /* Change the flags */
    CurHeader->Flags = MEM_FREE_FLAG;

    /* Push the freed pointer onto the Free List */
    CurHeader->NextFree = FreeListHead;

    FreeListHead = CurHeader;

    return;
}
