/* $Id: */
/********************************************************************
 * Copyright(C) 2008 by Xilinx, Inc. All rights reserved.
 * This text/file contains proprietary, confidential
 * information of Xilinx, Inc., is distributed under license
 * from Xilinx, Inc., and may be used, copied and/or
 * disclosed only pursuant to the terms of a valid license
 * agreement with Xilinx, Inc.  Xilinx hereby grants you a
 * license to use this text/file solely for design, simulation,
 * implementation and creation of design files limited
 * to Xilinx devices or technologies. Use with non-Xilinx
 * devices or technologies is expressly prohibited and
 * immediately terminates your license unless covered by
 * a separate agreement.
 *
 * Xilinx is providing this design, code, or information
 * as-is solely for use in developing programs and
 * solutions for Xilinx devices, with no obligation on the
 * part of Xilinx to provide support.  By providing this design,
 * code, or information as one possible implementation of
 * this feature, application or standard, Xilinx is making no
 * representation that this implementation is free from any
 * claims of infringement.  You are responsible for
 * obtaining any rights you may require for your implementation.
 * Xilinx expressly disclaims any warranty whatsoever with
 * respect to the adequacy of the implementation, including
 * but not limited to any warranties or representations that this
 * implementation is free from claims of infringement, implied
 * warranties of merchantability or fitness for a particular
 * purpose.
 *
 * Xilinx products are not intended for use in life support
 * appliances, devices, or systems. Use in such applications is
 * expressly prohibited.
 *
 * Any modifications that are made to the Source Code are
 * done at the user's sole risk and will be unsupported.
 * The Xilinx Support Hotline does not have access to source
 * code and therefore cannot answer specific questions related
 * to source HDL.
 *
 * This copyright and support notice must be retained as part
 * of this text at all times.
 * (c) Copyright 2008 Xilinx, Inc.
 * All rights reserved.
 ********************************************************************/

/******************************************************************************
*
* xtr_scatdma.c - The treck stack to DMA interface.
* Use scattered send an receive.
*
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1     drg  03/31/08 First release
* 2     pdmc 03/10/11 Modified for AXI bus and added full checksum
*                     offloading support
*
*****************************************************************************/

/***************************** Include Files ********************************/
#include <trsocket.h>
#include "xaxidma.h"
#include "xbuffermem.h"
#include "xtr_axiethernet.h"
#include "util_time.h"
#include "xtr_axidma.h"
#include "cache.h"
#include "debug.h"
#include "defaults.h"

/************************** Constant Definitions ****************************/
#define BD_ALIGNMENT     XAXIDMA_BD_MINIMUM_ALIGNMENT     /* Alignment of buffer descriptors */

/* Reallocate BD when there are this many free. */
#define RX_BD_FREE_THRESHOLD    16
    /* Threshold to disable the interrupts, so buffers can get freed */
#define RX_BD_IRPTDIS_THRESHOLD    32


#define CSUM_ENABLE_PARTIAL     0x00000001
#define CSUM_ENABLE_FULL        0x00000002

#define BLOCK_COUNT    (RX_BD_COUNT * 2)


#undef USE_TX_ISR
/**************************** Type Definitions ******************************/

/***************** Macros (Inline Functions) Definitions ********************/

/************************** Function Prototypes *****************************/
static void ScatTxISR(void *Callback);

/************************** Variable Definitions ****************************/

extern int IfaceMtu;
extern int DmaInitialized;

/************************** Function Definitions ****************************/
/******************************************************************************
 * Initialize and setup the device with DMA capability
 *
 * Parameters:
 *   Handle - User interface object
 *
 * Returns:
 *   TM_DEV_ERROR if open fails, TM_DEV_OKAY otherwise
 *****************************************************************************/
int AxiDma_ScatOpen(ttUserInterface Handle){
    XAxiDma_BdRing * RxRingPtr;
    XAxiDma_BdRing * TxRingPtr;
    TemacDevData *   Mac;
    int Status;

    Mac = (TemacDevData*)tfDeviceGetPointer(Handle);

    /* Re-initialize DMA -- this also re-initializes Ethernet */
    Status = InitializeDmaAndEthernet(Mac);
    if (Status != XST_SUCCESS)
    {
        return Status;
    }

    /*
     * Setup the memory area for RX Buffers. Allow the Treck stack to manage the
     * memory. The driver just allocates and uses buffers.
     * For the RX space, get all available BD, and tie them to buffers, then
     * let the hardware have the BD ring to use for RX.
     */
    if (AxiDma_RefillRxRing(Handle) < 0) {
        DEBUG_LOG(DEBUG_ERROR, "ScatOpen: AxiDma_RefillRxRing failed");
        return(TM_DEV_ERROR);
    }

    if (IfaceMtu > XAE_MTU)
    {
        /* Enable Jumbo frames */

        XAxiEthernet_SetOptions(&Mac->TemacInst, XAE_JUMBO_OPTION);
    }
    /* Setup the interrupt handlers and enable the interrupts here. The
     * interrupt handler needs the Treck UI Handle.
     */
    Status = Setup_Interrupt(Mac->TemacInst.Config.AxiDmaTxIntr,
                             (XInterruptHandler) ScatTxISR,
                             Handle);

    Status |= Setup_Interrupt(Mac->TemacInst.Config.AxiDmaRxIntr,
                              (XInterruptHandler) SgRxISR,
                              Handle);

    /* Install master interrupt handler */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "AxiDma_ScatOpen: Failed to install interrupt handler");
        return(Status);
    }

    /* For TX, do nothing, the buffer will be handed to the driver in the
     * AxiDma_Send()
     */

    /*
     * Enable transmit and receive related interrupts
     */
    TxRingPtr = XAxiDma_GetTxRing(&(Mac->DmaInst));
    RxRingPtr = XAxiDma_GetRxRing(&(Mac->DmaInst));

#ifdef USE_TX_ISR
    XAxiDma_BdRingIntEnable(TxRingPtr, XAXIDMA_IRQ_ALL_MASK);
#endif
    XAxiDma_BdRingIntEnable(RxRingPtr, XAXIDMA_IRQ_ALL_MASK);

    /* Start the DMA engines */
    Status = XAxiDma_BdRingStart(TxRingPtr);
    Status |= XAxiDma_BdRingStart(RxRingPtr);
    if (Status != XST_SUCCESS) {
        DEBUG_LOG(DEBUG_ERROR, "ScatOpen: Error starting BdRing, code %d\n", Status);
        return(TM_DEV_ERROR);
    }

    return(TM_DEV_OKAY);
}

#ifndef USE_TX_ISR
/******************************************************************************
 * Return transmit buffer(s) from the driver back to the stack. This is done by
 * querying HW as to whether it has any processed descriptors for us to go
 * through. If there are some, then free up the associated buffer.
 *
 * Parameters:
 *   Handle - User interface object
 *
 * Returns:
 *   TM_DEV_ERROR if error accessing the DMA engine
 *   TM_DEV_OKAY if no errors
 *****************************************************************************/
static int AxiDma_FreeTxBuffer(ttUserInterface Handle)
{
    TemacDevData *Mac = (TemacDevData*)tfDeviceGetPointer(Handle);
    XAxiDma_BdRing * TxRingPtr;
    XAxiDma_Bd *BdPtr, *BdCurPtr;
    XStatus Status;
    register int NumBds, i;


    TxRingPtr = XAxiDma_GetTxRing(&(Mac->DmaInst));

    if ((NumBds = XAxiDma_BdRingFromHw(TxRingPtr, XAXIDMA_ALL_BDS, &BdPtr)) > 0)
    {
        for (i=0, BdCurPtr=BdPtr; i<NumBds; i++)
        {
            /*Find the last Bd for a packet, then tell the Treck stack to free
             * the buffers */
            if (XAxiDma_BdGetCtrl(BdCurPtr) & XAXIDMA_BD_CTRL_TXEOF_MASK)
            {
                DEBUG_LOG(DEBUG_TX, "AxiDma_FreeTxBuffer(): SendComplete interface signalled");
                tfSendCompleteInterface(Handle,
                                  TM_DEV_SEND_COMPLETE_DRIVER);
                Mac->EthStats.TxFrames++;
            }
            if (XAxiDma_BdGetSts(BdCurPtr) & XAXIDMA_BD_STS_ALL_ERR_MASK)
            {
                DEBUG_LOG(DEBUG_ERROR,"DMA ERROR\r\n");
            }
            BdCurPtr = XAxiDma_BdRingNext(TxRingPtr, BdCurPtr);
        }

        Status = XAxiDma_BdRingFree(TxRingPtr, NumBds, BdPtr);
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR,"AxiDma_FreeTxBuffer(): Error freeing TxBDs, code %d",
                      Status);
            return(TM_DEV_ERROR);
        }
    }

    DEBUG_LOG(DEBUG_TX, "AxiDma_FreeTxBuffer(): Freed %d buffers", NumBds);

    return(TM_DEV_OKAY);
}
#endif

/******************************************************************************
 * Send one packet which may be scattered in multiple buffers.
 *
 * Parameters:
 *   Handle - User interface object
 *   DataPtr - Address of packet
 *   DataLen - Length of packet
 *   Flag - Flags from the stack
 *
 * Returns:
 *   TM_DEV_ERROR if an error occurred setting up the transmission of the frame
 *   TM_DEV_OKAY if frame transmission was kicked off without problems
 *
 *****************************************************************************/
int AxiDma_ScatSend(ttUserInterface Handle, ttUserPacketPtr PktPtr)
{
    XAxiDma_Bd *                BdPtr;
    XAxiDma_Bd *                BdCurPtr;
    XAxiDma_BdRing *            TxRingPtr;
    TemacDevData *              Mac;
    ttDevSendOffloadStruct *    CsumOff;
    XStatus                     Status;
    int                         BdCnt;
    int                         iter;
    char                        csumEnable;

    Mac = (TemacDevData*)tfDeviceGetPointer(Handle);

    if (PktPtr == NULL)
    {
        DEBUG_LOG(DEBUG_ERROR,"SgScatSend received NULL packet pointer");
        return(TM_DEV_ERROR);
    }

#ifndef USE_TX_ISR
    /* Free any Tx Buffers still held by the hardware. Do this so no TX
     * Interrupt is used.
     */
    AxiDma_FreeTxBuffer(Handle);
#endif


    /* For scatter send how many buffers are there */
    BdCnt = PktPtr->pktuLinkExtraCount + 1;
    DEBUG_LOG(DEBUG_TX, "\r\nLlDma_ScatSend():Sending packet with %d Frags, total len %d",
                    BdCnt, PktPtr->pktuChainDataLength);

    /* Get the checksum info from the first PktPtr */
    CsumOff = PktPtr->pktuDevOffloadPtr;

    TxRingPtr = XAxiDma_GetTxRing(&(Mac->DmaInst));

    /* Allocate BDs, prepare it */
    Status = XAxiDma_BdRingAlloc(TxRingPtr, BdCnt, &BdPtr);
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "Scat_Send(): error allocating TxBD, code %d", Status);
        return(TM_DEV_ERROR);
    }

    /* Map each fragment to a TxBD */
    BdCurPtr = BdPtr;

    for (iter = 0; iter < BdCnt; iter++)
    {
        DEBUG_LOG(DEBUG_TX, "ScatSend(): Current Packet Pointer: %08X", PktPtr);
        /* Set the BD data and length pointers, be sure to flush the buffer from the
        * cache. The BD will be flushed in the ToHw call, so foggetaboutit.
        */
        cacheFlush(PktPtr->pktuLinkDataPtr, PktPtr->pktuLinkDataLength);
        XAxiDma_BdSetBufAddr(BdCurPtr, (Xuint32)PktPtr->pktuLinkDataPtr);
        XAxiDma_BdSetLength(BdCurPtr, PktPtr->pktuLinkDataLength);

        /* Store the packet idenifier */
        XAxiDma_BdSetId(BdCurPtr, PktPtr);

        DEBUG_LOG(DEBUG_TX, "ScatSend(): Frag at %08X for %d bytes",
                XAxiDma_BdGetBufAddr(BdCurPtr), XAxiDma_BdGetLength(BdCurPtr));

        /* Clear the control register, and disable CSUM */
        XAxiDma_BdSetCtrl(BdCurPtr, 0);

        DEBUG_LOG(DEBUG_TX, "ScatSend(): Next Current Packet Pointer: %08X", PktPtr->pktuLinkNextPtr);
        PktPtr = PktPtr->pktuLinkNextPtr;
        if((PktPtr != NULL) && (iter != (BdCnt-1))){
                /*
                 * PktPtr->next is not null when doing SCATTER RECV. Why?
                 */
                BdCurPtr = XAxiDma_BdRingNext(TxRingPtr, BdCurPtr);
        }
    }

    /* Set the start of packet flag */
    XAxiDma_BdSetCtrl(BdPtr, XAXIDMA_BD_CTRL_TXSOF_MASK);

    /* Map the last fragment */
    XAxiDma_BdSetCtrl(BdCurPtr,
                    (XAxiDma_BdGetCtrl(BdCurPtr) | XAXIDMA_BD_CTRL_TXEOF_MASK));

/* If checksum offloading, then setup the 1st descriptor */
    csumEnable = 0;
    if (   (CsumOff)
        && (CsumOff->devoFlags & (  TM_DEVO_IP_CHKSUM_OFFLOAD
                                  | TM_DEVO_TCP_CHKSUM_OFFLOAD
                                  | TM_DEVO_UDP_CHKSUM_OFFLOAD )) )
    {
/* Enable hardware checksum computation for the buffer descriptor */
        if (XAxiEthernet_IsTxFullCsum(&Mac->TemacInst))
        {
            csumEnable = CSUM_ENABLE_FULL;
        }
        else if (XAxiEthernet_IsTxPartialCsum(&Mac->TemacInst))
        {
            csumEnable = CSUM_ENABLE_PARTIAL;
        }
    }

    if (csumEnable != 0)
    {
        XAxiDma_BdWrite((BdPtr),
                        XAXIDMA_BD_USR0_OFFSET,
                        (  XAxiDma_BdRead((BdPtr), XAXIDMA_BD_USR0_OFFSET)
                         | csumEnable) );

/* Write Start Offset and Insert Offset into BD */
        XAxiDma_BdWrite(BdPtr,
                        XAXIDMA_BD_USR1_OFFSET,
                        (CsumOff->devoTucss << 16) | CsumOff->devoTucso);

/* Write 0, as the seed value, to the BD */
        XAxiDma_BdWrite(BdPtr, XAXIDMA_BD_USR2_OFFSET, 0);
    }
    else
    {
        XAxiDma_BdWrite((BdPtr),
                        XAXIDMA_BD_USR0_OFFSET,
                        (  XAxiDma_BdRead((BdPtr), XAXIDMA_BD_USR0_OFFSET)
                         & ~(CSUM_ENABLE_FULL | CSUM_ENABLE_PARTIAL)));
        DEBUG_LOG(DEBUG_TX, "Checksum offload is disabled on this Tx packet\r\n");
    }


    /* Enqueue the descriptor */
    DEBUG_LOG(DEBUG_TX, "AxiDma_ScatSend():Enque packet with %d Bd, Ctrl = 0x%08x, Sts = 0x%08x",
                    iter, XAxiDma_BdGetCtrl(BdPtr), XAxiDma_BdGetSts(BdPtr));

    Status = XAxiDma_BdRingToHw(TxRingPtr, BdCnt, BdPtr);
    if (Status != XST_SUCCESS){
        DEBUG_LOG(DEBUG_ERROR, "Scat_Send(): error committing TxBD, code %d", Status);
        Status = XAxiDma_BdRingUnAlloc(TxRingPtr, BdCnt, BdPtr);
        return(TM_DEV_ERROR);
    }

    return(TM_DEV_OKAY);
}


/******************************************************************************
 * Handle SGDMA transmit interrupts
 * Get the used Bd from the hardware, count the number of packets sent and
 * inform the Treck stack.
 *
 * Parameters:
 *   Callback - Device instance
 *****************************************************************************/
static void ScatTxISR(void *Callback) {
    ttUserInterface *TreckUI = (ttUserInterface*)Callback;
    TemacDevData *Mac;
    XAxiDma_BdRing * TxRingPtr;
    XAxiDma_Bd *BdPtr, *BdCurPtr;
    XStatus Status;
    register int NumBds, i;
    int NumFrames, TotalBytes;
    u32 IrqStatus;

    Mac = (TemacDevData*)tfDeviceGetPointer(TreckUI);

    TxRingPtr = XAxiDma_GetTxRing(&(Mac->DmaInst));

    /* Read pending interrupts */
    IrqStatus = XAxiDma_BdRingGetIrq(TxRingPtr);
    /* Acknowledge pending interrupts */
    XAxiDma_BdRingAckIrq(TxRingPtr, IrqStatus);

    /* If error interrupt is asserted, raise error flag, reset the
     * hardware to recover from the error, and return with no further
     * processing.
     */
    if ((IrqStatus & XAXIDMA_IRQ_ALL_MASK)) {
        DEBUG_LOG(DEBUG_ERROR, "ScatTxISR: Error: IrqStatus: 0x%08x\n\r", IrqStatus & XAXIDMA_IRQ_ALL_MASK);
        xil_printf("ScatTxISR error: Resetting DMA, Status: 0x%08x\n\r",
                        XAxiEthernet_ReadReg(Mac->DmaInst.RegBase, XAXIDMA_SR_OFFSET + XAXIDMA_TX_OFFSET));
        XAxiDma_Reset(&Mac->DmaInst);
        return;
    }

    /*
     * If Reception done interrupt is asserted, call TX call back function
     * to handle the processed BDs and then raise the according flag.
     */
    if (!(IrqStatus & (XAXIDMA_DELAY_MASK | XAXIDMA_COALESCE_MASK))) {
        return;
    }

    DEBUG_LOG(DEBUG_ERROR,"ScatTxISR: was invoked");
    NumFrames = 0;
    TotalBytes = 0;

    if ((NumBds = XAxiDma_BdRingFromHw(TxRingPtr, XAXIDMA_ALL_BDS, &BdPtr)) > 0)
    {
        for (i=0, BdCurPtr=BdPtr; i<NumBds; i++)
        {
            /*Find the last Bd for a packet, then tell the Treck stack to free
             * the buffers */
            TotalBytes = XAxiDma_BdGetLength(BdCurPtr);

            if (XAxiDma_BdGetCtrl(BdCurPtr) & XAXIDMA_BD_CTRL_TXEOF_MASK)
            {
                NumFrames++;
            }
            if (XAxiDma_BdGetSts(BdCurPtr) & XAXIDMA_BD_STS_ALL_ERR_MASK)
            {
                DEBUG_LOG(DEBUG_ERROR,"DMA ERROR\r\n");
            }
            BdCurPtr = XAxiDma_BdRingNext(TxRingPtr, BdCurPtr);
        }

        Status = XAxiDma_BdRingFree(TxRingPtr, NumBds, BdPtr);
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR,"ScatTxISR(): Error freeing TxBDs, code %d",
                      Status);
            return;
        }
    }

    DEBUG_LOG(DEBUG_TX, "ScatTxISR(): Freed %d buffers", NumBds);

    /* Notify the Treck stack we have sent the buffers */
    tfNotifyInterfaceIsr(TreckUI, 0, NumFrames, TotalBytes, 0);

    return;
}

